# - Find ROOT instalation
# This module tries to find the ROOT installation on your system.
# It tries to find the root-config script which gives you all the needed information.
# If the system variable ROOTSYS is set this is straight forward.
# If not the module uses the pathes given in ROOT_CONFIG_SEARCHPATH.
# If you need an other path you should add this path to this varaible.  
# The root-config script is then used to detect basically everything else.
# This module defines a number of key variables and macros.
include(CmakePackage/root_options_extraction.cmake)

MESSAGE(STATUS "Looking for Root...")

root_get_feature(CXX_STANDARD_STRING "^cxx")
message(STATUS "CXX_STANDARD_STRING: [${CXX_STANDARD_STRING}]")
set(CMAKE_CXX_STANDARD ${CXX_STANDARD_STRING} CACHE STRING "")
set(CMAKE_CXX_STANDARD_REQUIRED ON)

message(STATUS "ROOTSYS: $ENV{ROOTSYS}")

SET(ROOT_CONFIG_SEARCHPATH
  ${SIMPATH}/tools/root/bin
  $ENV{ROOTSYS}/bin
  )

SET(ROOT_DEFINITIONS "")
SET(ROOT_PYTHON_VERSION "")
SET(ROOT_MATHMORE "")
SET(ROOT_DATAFRAME "")
SET(ROOT_INSTALLED_VERSION_TOO_OLD FALSE)

SET(ROOT_CONFIG_EXECUTABLE ROOT_CONFIG_EXECUTABLE-NOTFOUND)

FIND_PROGRAM(ROOT_CONFIG_EXECUTABLE NAMES root-config PATHS
  ${ROOT_CONFIG_SEARCHPATH}
  NO_DEFAULT_PATH)

IF (${ROOT_CONFIG_EXECUTABLE} MATCHES "ROOT_CONFIG_EXECUTABLE-NOTFOUND")
  MESSAGE( FATAL_ERROR "ROOT not installed in the searchpath and ROOTSYS is not set. Please
 set ROOTSYS or add the path to your ROOT installation in the Macro FindROOT.cmake in the
 subdirectory cmake/modules.")
ELSE (${ROOT_CONFIG_EXECUTABLE} MATCHES "ROOT_CONFIG_EXECUTABLE-NOTFOUND")
  STRING(REGEX REPLACE "(^.*)/bin/root-config" "\\1" test ${ROOT_CONFIG_EXECUTABLE}) 
  SET( ENV{ROOTSYS} ${test})
  set( ROOTSYS ${test})
ENDIF (${ROOT_CONFIG_EXECUTABLE} MATCHES "ROOT_CONFIG_EXECUTABLE-NOTFOUND")  


IF (NOT ROOT_MIN_VERSION)
  SET(ROOT_MIN_VERSION "5.99")
ENDIF (NOT ROOT_MIN_VERSION)

IF (NOT ROOT_MAX_VERSION)
  SET(ROOT_MAX_VERSION "6.44")
ENDIF (NOT ROOT_MAX_VERSION)

IF (ROOT_CONFIG_EXECUTABLE)
  
  SET(ROOT_FOUND FALSE)
  IF(UNIX)
    EXEC_PROGRAM(${ROOT_CONFIG_EXECUTABLE} ARGS "--version" OUTPUT_VARIABLE ROOTVERSION)
    
    MESSAGE(STATUS "Looking for Root... - found $ENV{ROOTSYS}/bin/root")
    MESSAGE(STATUS "Looking for Root... - version ${ROOTVERSION} ")   

    if(ROOTVERSION VERSION_LESS ROOT_MIN_VERSION)
      message(FATAL_ERROR "Invalid ROOT version: ${ROOTVERSION} found: at least ${ROOT_MIN_VERSION} required.")
    endif()

    if(ROOTVERSION VERSION_GREATER ROOT_MAX_VERSION)
      message(FATAL_ERROR "Invalid ROOT version: ${ROOTVERSION} found: at most ${ROOT_MAX_VERSION} required.")
    endif()

    set(ROOT_FOUND True)
  ENDIF(UNIX)
  IF(WIN32)
    #Trouver un moyen de tester la version de root
    SET(ROOT_FOUND TRUE)
  ENDIF(WIN32)

  root_get_configuration(ROOT_PYTHON_VERSION "--python-version")

ENDIF (ROOT_CONFIG_EXECUTABLE)


IF (ROOT_FOUND)
  IF(UNIX)
    # ask root-config for the library dir
    # Set ROOT_LIBRARY_DIR
    EXEC_PROGRAM( ${ROOT_CONFIG_EXECUTABLE}
      ARGS "--libdir"
      OUTPUT_VARIABLE ROOT_LIBRARY_DIR_TMP )
    
    IF(EXISTS "${ROOT_LIBRARY_DIR_TMP}")
      SET(ROOT_LIBRARY_DIR ${ROOT_LIBRARY_DIR_TMP} )
    ELSE(EXISTS "${ROOT_LIBRARY_DIR_TMP}")
      MESSAGE("Warning: ROOT_CONFIG_EXECUTABLE reported ${ROOT_LIBRARY_DIR_TMP} as library path,")
      MESSAGE("Warning: but ${ROOT_LIBRARY_DIR_TMP} does NOT exist, ROOT must NOT be installed correctly.")
    ENDIF(EXISTS "${ROOT_LIBRARY_DIR_TMP}")
    
    # ask root-config for the binary dir
    EXEC_PROGRAM(${ROOT_CONFIG_EXECUTABLE}
      ARGS "--bindir"
      OUTPUT_VARIABLE root_bins )
    SET(ROOT_BINARY_DIR ${root_bins})
    
    # ask root-config for the include dir
    EXEC_PROGRAM( ${ROOT_CONFIG_EXECUTABLE}
      ARGS "--incdir" 
      OUTPUT_VARIABLE root_headers )
    SET(ROOT_INCLUDE_DIR ${root_headers})
    # CACHE INTERNAL "")

    # ask root-config for the mathmore package
    EXEC_PROGRAM( ${ROOT_CONFIG_EXECUTABLE}
      ARGS "--has-mathmore" 
      OUTPUT_VARIABLE root_mathmore )
    SET(ROOT_MATHMORE ${root_mathmore})
    MESSAGE(STATUS "Looking for Mathmore in Root... - results: [${ROOT_MATHMORE}]")
    # Check that the library is there as well
    if("${ROOT_MATHMORE}" STREQUAL "yes")
      SET(lib lib-NOTFOUND)
      FIND_LIBRARY(lib libMathMore.so ${ROOT_LIBRARY_DIR_TMP})
      IF(lib MATCHES "lib-NOTFOUND")
        MESSAGE("Looking for Mathmore library in [${ROOT_LIBRARY_DIR_TMP}]... - libMathMore not found, MathMore discarded !")
        SET(ROOT_MATHMORE "no")
      ELSE(lib MATCHES "lib-NOTFOUND")
        MESSAGE(STATUS "Looking for Mathmore library in Root... - libMathMore detected...") 
      ENDIF(lib MATCHES "lib-NOTFOUND")
    endif("${ROOT_MATHMORE}" STREQUAL "yes")
    
    # ask root-config for the library variables
    EXEC_PROGRAM( ${ROOT_CONFIG_EXECUTABLE}
      #    ARGS "--noldflags --noauxlibs --libs" 
      ARGS "--glibs" 
      OUTPUT_VARIABLE root_flags )
    
    SET(ROOT_LIBRARIES ${root_flags})
    
    # check for ROOT Minuit2
    EXEC_PROGRAM(root-config
      ARGS --has-minuit2
      OUTPUT_VARIABLE HAS_MINUIT2
      RETURN_VALUE has_minuit_ret )
    IF(NOT has_minuit_ret)
      # determine if minuit2 was included with ROOT
      IF(HAS_MINUIT2 STREQUAL "yes")
	MESSAGE(STATUS "Looking for Minuit2 in Root... - found Minuit2 library")
      ELSE(HAS_MINUIT2 STREQUAL "yes")
	MESSAGE(STATUS "Looking for Minuit2 in Root... - library Minuit2 was not found !!!!")
      ENDIF(HAS_MINUIT2 STREQUAL "yes")
    ENDIF(NOT has_minuit_ret)

    # check for XML
    EXEC_PROGRAM(root-config
        ARGS --has-xml
        OUTPUT_VARIABLE HAS_XML
        RETURN_VALUE has_xml_ret)

    IF(NOT has_xml_ret)
        IF(HAS_XML STREQUAL "yes")
            MESSAGE(STATUS "Looking for XML interface in Root... - found")
        ELSE(HAS_XML STREQUAL "yes")
            MESSAGE(STATUS "Looking for XML interface in Root... - not found !!!!")
        ENDIF(HAS_XML STREQUAL "yes")
    ENDIF(NOT has_xml_ret)
    
    # check for LibFFTW
    EXEC_PROGRAM(root-config
      ARGS --has-fftw3
      OUTPUT_VARIABLE HAS_FFTW
      RETURN_VALUE has_fftw3_ret )

    IF(NOT has_fftw3_ret)
      IF(HAS_FFTW STREQUAL "yes")
	SET(FFTW_FOUND TRUE)
	MESSAGE(STATUS "Looking for FFTW in Root... - found FFTW library")
      ELSE(HAS_FFTW STREQUAL "yes") 
	SET(FFTW_FOUND FALSE)
 	MESSAGE(STATUS "Looking for FFTW in Root... - library FFTW was not found !!!!")
      ENDIF(HAS_FFTW STREQUAL "yes") 
    ENDIF(NOT has_fftw3_ret)
    
  ENDIF(UNIX)
  
  IF(WIN32)
    SET(ROOT_LIBRARY_DIR $ENV{ROOTSYS}/lib)
    SET(ROOT_BINARY_DIR $ENV{ROOTSYS}/bin)
    SET(ROOT_INCLUDE_DIR $ENV{ROOTSYS}/include)
    MESSAGE(STATUS "ROOT:${ROOT_LIBRARY_DIR}")
    SET(ROOT_LIBRARIES libPostscript libMatrix libPhysics libThread libRint libGpad libGraf3d libGraf libHist libCore  libMathCore  libTreePlayer libTree libXMLIO libRIO libMLP)
    FOREACH(i ${ROOT_LIBRARIES})
      SET(lib lib-NOTFOUND)
      FIND_LIBRARY(lib ${i} ${ROOT_LIBRARY_DIR})
      IF(lib MATCHES "lib-NOTFOUND")
        MESSAGE("Library ${i} not found")
      ELSE(lib MATCHES "lib-NOTFOUND")
        MESSAGE(STATUS "Library ${i} detected...")
        SET(ROOT_LIBRARIES_FOUND ${ROOT_LIBRARIES_FOUND} ${i})
      ENDIF(lib MATCHES "lib-NOTFOUND")
    ENDFOREACH(i)

    # check for ROOT Minuit2 for WIN32
    FIND_LIBRARY(HAS_MINUIT2 libMinuit2 ${ROOT_LIBRARY_DIR})
    IF(HAS_MINUIT2)
      MESSAGE(STATUS "Looking for Root... - found minuit2 library")
    ELSE(HAS_MINUIT2)
      MESSAGE(STATUS "Looking for Root... - library minuit2 was not found !!!!")
    ENDIF(HAS_MINUIT2)
    
    # check for ROOT XML interface for WIN32
    FIND_LIBRARY(HAS_XML libXMLParser ${ROOT_LIBRARY_DIR})
    IF(HAS_XML)
        MESSAGE(STATUS "Looking for XML interface in Root... - found")
    ELSE(HAS_XML)
        MESSAGE(STATUS "Looking for XML interface in Root... - not found !!!!")
    ENDIF(HAS_XML)

    # check for ROOT LibFFTW for WIN32
    FIND_LIBRARY(HAS_FFTW libFFTW ${ROOT_LIBRARY_DIR})
    IF(HAS_FFTW)
      SET(FFTW_FOUND TRUE)
      MESSAGE(STATUS "Looking for Root... - found FFTW library")
    ELSE(HAS_FFTW)
      SET(FFTW_FOUND FALSE)
      MESSAGE(STATUS "Looking for Root... - library FFTW was not found !!!!")
    ENDIF(HAS_FFTW)

  ENDIF(WIN32)

# Check that the library is there

if(WIN32)
    list(APPEND CMAKE_FIND_LIBRARY_PREFIXES "lib")
endif()

SET(lib lib-NOTFOUND)
FIND_LIBRARY(lib NAMES ROOTDataFrame PATHS ${ROOT_LIBRARY_DIR})
execute_process(COMMAND ${CMAKE_COMMAND} -E echo_append
  "-- Looking for Dataframe library in Root... ")
IF(lib MATCHES "lib-NOTFOUND")
    execute_process(COMMAND ${CMAKE_COMMAND} -E echo
        "- not found in [${ROOT_LIBRARY_DIR}], discarded!")
    SET(ROOT_DATAFRAME "no")
ELSE(lib MATCHES "lib-NOTFOUND")
    execute_process(COMMAND ${CMAKE_COMMAND} -E echo
        "- found in [${ROOT_LIBRARY_DIR}]")
    SET(ROOT_DATAFRAME "yes")
ENDIF(lib MATCHES "lib-NOTFOUND")

if(WIN32)
    list(REMOVE_ITEM CMAKE_FIND_LIBRARY_PREFIXES "lib")
endif()

  # Make variables changeble to the advanced user
  MARK_AS_ADVANCED(ROOT_LIBRARY_DIR ROOT_INCLUDE_DIR ROOT_DEFINITIONS ROOT_LIBRARIES_FOUND lib)
  
  # Set ROOT_INCLUDES
  SET( ROOT_INCLUDES ${ROOT_INCLUDE_DIR})
  include_directories(${ROOT_INCLUDE_DIR})
  SET(LD_LIBRARY_PATH ${LD_LIBRARY_PATH} ${ROOT_LIBRARY_DIR})
  
  #######################################
  #
  #       Check the executables of ROOT 
  #          ( rootcint ) 
  #
  #######################################
  
  FIND_PROGRAM(ROOT_CINT_EXECUTABLE
    NAMES rootcint
    PATHS ${ROOT_BINARY_DIR}
    NO_DEFAULT_PATH
    )
  
ENDIF (ROOT_FOUND)



###########################################
#
#       Macros for building ROOT dictionary
#
###########################################

MACRO (ROOT_GENERATE_DICTIONARY_OLD )
  
  set(INFILES "")    
  
  foreach (_current_FILE ${ARGN})
    
    IF (${_current_FILE} MATCHES "^.*\\.h$")
      IF (${_current_FILE} MATCHES "^.*Link.*$")
        set(LINKDEF_FILE ${_current_FILE})
      ELSE (${_current_FILE} MATCHES "^.*Link.*$")
        set(INFILES ${INFILES} ${_current_FILE})
      ENDIF (${_current_FILE} MATCHES "^.*Link.*$")
    ELSE (${_current_FILE} MATCHES "^.*\\.h$")
      IF (${_current_FILE} MATCHES "^.*\\.cxx$")
        set(OUTFILE ${_current_FILE})
      ELSE (${_current_FILE} MATCHES "^.*\\.cxx$")
        set(INCLUDE_DIRS ${INCLUDE_DIRS} -I${_current_FILE})   
      ENDIF (${_current_FILE} MATCHES "^.*\\.cxx$")
    ENDIF (${_current_FILE} MATCHES "^.*\\.h$")
    
  endforeach (_current_FILE ${ARGN})
  
  STRING(REGEX REPLACE "(^.*).cxx" "\\1.h" bla "${OUTFILE}")
  SET (OUTFILES ${OUTFILE} ${bla})
  
  ADD_CUSTOM_COMMAND(OUTPUT ${OUTFILES}
    COMMAND ${ROOT_CINT_EXECUTABLE}
    ARGS -f ${OUTFILE} -c -DHAVE_CONFIG_H ${INCLUDE_DIRS} ${INFILES} ${LINKDEF_FILE} DEPENDS ${INFILES})
  
  #   MESSAGE("ROOT_CINT_EXECUTABLE has created the dictionary ${OUTFILE}")
  
ENDMACRO (ROOT_GENERATE_DICTIONARY_OLD)

###########################################
#
#       Macros for building ROOT dictionary
#
###########################################

MACRO (ROOT_GENERATE_DICTIONARY INFILES LINKDEF_FILE OUTFILE INCLUDE_DIRS_IN)
  
  set(INCLUDE_DIRS)
  
  foreach (_current_FILE ${INCLUDE_DIRS_IN})
    ##MESSAGE(STATUS "  ***** ROOT_GENERATE_DICTIONARY _current_FILE[${_current_FILE}]")
    set(INCLUDE_DIRS ${INCLUDE_DIRS} -I${_current_FILE})   
  endforeach (_current_FILE ${INCLUDE_DIRS_IN})
  
  STRING(REGEX REPLACE "^(.*)\\.(.*)$" "\\1.h" bla "${OUTFILE}")
  SET (OUTFILES ${OUTFILE} ${bla})
  
  
  ADD_CUSTOM_COMMAND(OUTPUT ${OUTFILES}
    COMMAND ${ROOT_CINT_EXECUTABLE}
    ARGS -f ${OUTFILE} -c -DHAVE_CONFIG_H ${INCLUDE_DIRS} ${INFILES} ${LINKDEF_FILE} DEPENDS ${INFILES})
  
ENDMACRO (ROOT_GENERATE_DICTIONARY)

##################################################
#
#       Macros for building ROOT dictionary
#        with a DEFINE definitions for rootcint
#
##################################################

MACRO (ROOT_GENERATE_DICTIONARY_DEFINES INFILES LINKDEF_FILE OUTFILE INCLUDE_DIRS_IN WITH_DEFINES)
  
  set(INCLUDE_DIRS)
  
  foreach (_current_FILE ${INCLUDE_DIRS_IN})
    ##MESSAGE(STATUS "  ***** ROOT_GENERATE_DICTIONARY _current_FILE[${_current_FILE}]")    
    set(INCLUDE_DIRS ${INCLUDE_DIRS} -I${_current_FILE})   
  endforeach (_current_FILE ${INCLUDE_DIRS_IN})
  
  STRING(REGEX REPLACE "^(.*)\\.(.*)$" "\\1.h" bla "${OUTFILE}")
  SET (OUTFILES ${OUTFILE} ${bla})
  
  
  ADD_CUSTOM_COMMAND(OUTPUT ${OUTFILES}
    COMMAND ${ROOT_CINT_EXECUTABLE}
    ARGS -f ${OUTFILE} -c ${WITH_DEFINES} -DHAVE_CONFIG_H ${INCLUDE_DIRS} ${INFILES} ${LINKDEF_FILE} DEPENDS ${INFILES})
  
ENDMACRO (ROOT_GENERATE_DICTIONARY_DEFINES)



###############################################################
#
# Exchange file extention of LIST from
# FILE_EXT1 to FILE_EXT2 and assign the
# newly created list to OUTVAR. The input
# list LIST is not changed at all
# Ex: CHANGE_FILE_EXTENSION(*.cxx *.h TRD_HEADERS "${TRD_SRCS}")
#
################################################################

MACRO (CHANGE_FILE_EXTENSION FILE_EXT1 FILE_EXT2 OUTVAR LIST)
  SET (BLA "")
  IF (${FILE_EXT1} MATCHES "^[*][.]+.*$")
    STRING(REGEX REPLACE "^[*]+([.].*)$" "\\1" FILE_EXT1_NEW ${FILE_EXT1})
  ENDIF  (${FILE_EXT1} MATCHES "^[*][.]+.*$")
  
  IF (${FILE_EXT2} MATCHES "^[*][.]+.*$")
    STRING(REGEX REPLACE "^[*]+([.].*)" "\\1" FILE_EXT2_NEW ${FILE_EXT2})
  ENDIF  (${FILE_EXT2} MATCHES "^[*][.]+.*$")
  
  foreach (_current_FILE ${LIST})
    
    STRING(REGEX REPLACE "^(.*)${FILE_EXT1_NEW}$" "\\1${FILE_EXT2_NEW}" test ${_current_FILE})
    SET (BLA ${BLA} ${test})
    
  endforeach (_current_FILE ${ARGN})
  
  SET (${OUTVAR} ${BLA})
  
ENDMACRO (CHANGE_FILE_EXTENSION)
