//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TAttribute.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   $Date$
 
 \brief  Implementation of the class URANIE::DataServer::TAttribute
 
 Created: Fri Oct  1 13:37:23 2004
 
 \par Modifications
\li JB Blanchard (2016-02-15)
 - Add two cases in EType in order to store vectors of double (kVector)  and strings (kString)
 \li M. Salvoldi (2011-09-28)
 - comparison of string by using strcmp method in setFileKey()
 -  replace the Form(%s/%s...) by gSystem->ConcatFileName(...)  in setFileKey)
 \li M. Salvoldi (OCC/Areva) (2011/07/27 08:11:51): Add a parameter "Attribute Son" in the composition of the object attribut
 - add initialisation of these varirables _attSon,_nlevel,_haveSon
 - add methods associated to these new variables:   getSonAttribute(), setSonAttribute(TAttribute * tatt), getLevel() , setLevel(Int_t nlevel), haveSon();
 \li F. Gaudier (2010-08-06 15:16:33)
 - add the enum kNewRow to replace the enum kNewLine (deprecated from v2.1.10 to v2.3.0)
 \li F. Gaudier (2010-08-06 10:15:43)
 - Change the log message for the setKey deprecated method; add the name of the attribute
 \li Cedric Forgues (C.S.) 2010/06/24
 Exceptions are added for setStd() and setMean(), setLowerBound(), setUpperBound().
 \li Cedric Forgues (C.S.) 2010/06/11
 An exception is added to warn user he used bad bounds.
 \li Cedric Forgues (C.S.) 2010/04/23
 - The setKey() method is deprecated till URANIE 2.3 (2010-12-31). It will now launch the method setFileKey(), which fullfill the same functions as setKey(). 
 \li Cedric Forgues (C.S.) 2009/12/04
 methode setKey()
 \li Cedric Forgues (C.S.) 2009/12/03
 methode setKey()
 \li Cedric Forgues (C.S.) 2009/12/03
 methode getFileNameOfKey(), mise en deprecated;
 *********************************************************************
 */

#include <iostream>
using std::cout;
using std::endl;
#include "TAttribute.h"
#include "UExceptions.h"
#include "UEntry.h"

ClassImp (URANIE::DataServer::TAttribute)

namespace URANIE
{
namespace DataServer
{

  // -----------------------------------------------------------------------
  //  Constructors_stepValue 
  // -----------------------------------------------------------------------
  //! Default constructor
  TAttribute::TAttribute() :
    TNamed("_uranie_attribut_att_", "attribute title"), _sunity(0), _snote(0), _blog(kFALSE), _nshare(1), _attSon(NULL),
    _nlevel(0), _haveSon(kFALSE), _sFormatSubstitute("%e"), _KeyList(), _bHaveUpperBound(kFALSE), _bHaveLowerBound(kFALSE),
    _bHaveDefaultValue(kFALSE), _bHaveDefaultVector(kFALSE), _bHaveDefaultString(kFALSE), _bHaveStepValue(kFALSE), _vbuffer(NULL),
    _norigin(kAttribute), _nAttribute(kInput), _nType(kInconnu)
  {
    //     cout << "Constructeur URANIE::DataServer::TAttribute vide[" << this << "]" << endl;
    lowerBound = -1.0 * DLE_MAX;
    upperBound = DLE_MAX;
    //_nfield = NULL;
    clearVectors();
  }

  //! Constructor by name and title 
  TAttribute::TAttribute(const char* name, const char* title) :
    TNamed(name, title),
    _sunity(0), _snote(0), _blog(kFALSE), _nshare(1), _attSon(NULL),
    _nlevel(0), _haveSon(kFALSE), _sFormatSubstitute("%e"), _KeyList(), _bHaveUpperBound(kFALSE), _bHaveLowerBound(kFALSE),
    _bHaveDefaultValue(kFALSE), _bHaveDefaultVector(kFALSE), _bHaveDefaultString(kFALSE), _bHaveStepValue(kFALSE), _vbuffer(NULL),
    _norigin(kAttribute), _nAttribute(kInput), _nType(kDefault)
  {
    //     cout << "Constructeur URANIE::DataServer::TAttribute[" << this << "]str[" << str<<"]"<<endl;
    lowerBound = -1.0 * DLE_MAX;
    upperBound = DLE_MAX;
    clearVectors();
  }

  //! Constructor by name only
  TAttribute::TAttribute(TString str) :
    TNamed(str.Data(), str.Data()),
    _sunity(0), _snote(0), _blog(kFALSE), _nshare(1), _attSon(NULL),
    _nlevel(0), _haveSon(kFALSE), _sFormatSubstitute("%e"), _KeyList(), _bHaveUpperBound(kFALSE), _bHaveLowerBound(kFALSE),
    _bHaveDefaultValue(kFALSE), _bHaveDefaultVector(kFALSE), _bHaveDefaultString(kFALSE), _bHaveStepValue(kFALSE), _vbuffer(NULL),
    _norigin(kAttribute), _nAttribute(kInput), _nType(kDefault)
  {
    //      cout << "Constructeur URANIE::DataServer::TAttribute[" << this << "]str[" << str<<"]"<<endl;
    lowerBound = -1.0 * DLE_MAX;
    upperBound = DLE_MAX;
    clearVectors();
  }

  //! Constructor by name and type only
  TAttribute::TAttribute(TString str, EType type) :
    TNamed(str.Data(), str.Data()),
    _sunity(0), _snote(0), _blog(kFALSE), _nshare(1), _attSon(NULL),
    _nlevel(0), _haveSon(kFALSE), _sFormatSubstitute("%e"), _KeyList(), _bHaveUpperBound(kFALSE), _bHaveLowerBound(kFALSE),
    _bHaveDefaultValue(kFALSE), _bHaveDefaultVector(kFALSE), _bHaveDefaultString(kFALSE), _bHaveStepValue(kFALSE), _vbuffer(NULL),
    _norigin(kAttribute), _nAttribute(kInput), _nType(type)
  {
    //      cout << "Constructeur URANIE::DataServer::TAttribute[" << this << "]str[" << str<<"]"<<endl;
    lowerBound = -1.0 * DLE_MAX;
    upperBound = DLE_MAX;
    clearVectors();
  }
//! Constructor by name , slurm value and type
  TAttribute::TAttribute(TString str, TString value, EType type) :
    TNamed(str.Data(), str.Data()),
    _sunity(0), _snote(0), _blog(kFALSE), _nshare(1), _attSon(NULL),
    _nlevel(0), _haveSon(kFALSE), _sFormatSubstitute("%s"), _KeyList(), _bHaveUpperBound(kFALSE), _bHaveLowerBound(kFALSE),
    _bHaveDefaultValue(kFALSE), _bHaveDefaultVector(kFALSE), _bHaveDefaultString(kFALSE), _bHaveStepValue(kFALSE), _vbuffer(NULL),
    _norigin(kAttribute), _nAttribute(kInput), _nType(type), _slurmValue(value)
  {
    //      cout << "Constructeur URANIE::DataServer::TAttribute[" << this << "]str[" << str<<"]"<<endl;
    lowerBound = -1.0 * DLE_MAX;
    upperBound = DLE_MAX;

    clearVectors();
  }
  
  //! Constructor by name and range
  TAttribute::TAttribute(TString str, Double_t lower, Double_t upper) :
    TNamed(str.Data(), str.Data()),
    _sunity(0), _snote(0), _blog(kFALSE), _nshare(1), _attSon(NULL),
    _nlevel(0), _haveSon(kFALSE), _sFormatSubstitute("%e"), _KeyList(), upperBound(upper), _bHaveUpperBound(kTRUE),
    lowerBound(lower), _bHaveLowerBound(kTRUE), _bHaveDefaultValue(kFALSE), _bHaveDefaultVector(kFALSE), _bHaveDefaultString(kFALSE),
    _bHaveStepValue(kFALSE), _vbuffer(NULL), _norigin(kAttribute), _nAttribute(kInput), _nType(kDefault)
  {
    //     cout << "Constructeur URANIE::DataServer::TAttribute[" << this << "]str[" << str<<"] lower[" << lower <<"] upper[" << upper << "]" << endl;
    if (upper < lower)
      try
        {
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     "LowerBound is superior to UpperBound");
        } catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
	  ue.printMessage();
	  throw ue;
        }
    clearVectors();
  }

  //! Copy constructor
  TAttribute::TAttribute(const TAttribute &attr) :
    TNamed(attr.GetName(), attr.GetTitle()),
    //     _slabel(attr._slabel),
    _sunity(attr._sunity), _snote(attr._snote), _blog(attr._blog), _nshare(attr._nshare), _attSon(attr._attSon),
    _nlevel(attr._nlevel), _haveSon(attr._haveSon), _sFormatSubstitute(attr._sFormatSubstitute), _KeyList(attr._KeyList), 
    upperBound(attr.upperBound), _bHaveUpperBound(attr._bHaveUpperBound), lowerBound(attr.lowerBound), _bHaveLowerBound(attr._bHaveLowerBound),
    _defaultValue(attr._defaultValue),  _bHaveDefaultValue(attr._bHaveDefaultValue), _bHaveDefaultVector(attr._bHaveDefaultVector),
    _bHaveDefaultString(attr._bHaveDefaultString),  _stepValue(attr._stepValue), _bHaveStepValue(attr._bHaveStepValue), _vbuffer(NULL),
    _norigin(attr._norigin), _nAttribute(attr._nAttribute),
    _nType(attr._nType)
  {
    // // cout << "Constructeur de copie URANIE::DataServer::TAttribute[" << this << "]" << endl;
    // for(int iel=0; iel<attr.getMinimumSize(); iel++) _vminimum.push_back( attr.getMinimum(iel) );
    // for(int iel=0; iel<attr.getMaximumSize(); iel++) _vmaximum.push_back( attr.getMaximum(iel) );
    // for(int iel=0; iel<attr.getMeanSize(); iel++) _vmean.push_back( attr.getMean(iel) );
    // for(int iel=0; iel<attr.getStdSize(); iel++) _vstd.push_back( attr.getStd(iel) );
  }

  TAttribute *TAttribute::Clone(const char *newname) const
  {
    if (_blog)
      cout << " ************************" << endl
	   << " ** Clone TAttribute newname[" << newname << "]" << endl;
    TAttribute *newatt = (TAttribute*) TNamed::Clone(newname);
    if (_blog)
      cout << " ************************" << endl
	   << " ** End Of Clone TAttribute newname[" << newname << "]"
	   << endl;
    return newatt;
  }

  TAttribute::~TAttribute()
  {
    if (_blog)
      cout << "Destructeur::TAttribute[" << this << "] sname[" << GetName()
	   << "] Share[" << getShare() << "]" << endl;

    // Clean
    for (list<TAttributeFileKey*>::iterator it = _KeyList.begin();
	 it != _KeyList.end(); it++)
      {
          if ((*it)->_sEventList != NULL && !gROOT->GetList()->IsEmpty())
	  {
            delete (*it)->_sEventList;
            (*it)->_sEventList = NULL;
	  }
        delete (*it);
        (*it) = NULL;
      }
    _KeyList.clear();
    clearVectors();
    
  }
  // -----------------------------------------------------------------------
  //  Getter and setter methods
  // -----------------------------------------------------------------------
  void TAttribute::setDataType(TString str)
  {
    
    str.ToLower();
    
    if (str == "real" || str == "double" || str=="d")
      _nType = kReal;
    else if (str == "vector" || str=="v")
      _nType = kVector;
    else if (str == "string" || str=="s")
      _nType = kString;
    else if (TString(str).BeginsWith("{"))
      _nType = kCategory;
    else
      _nType = kInconnu;
    
  }
  
  TString TAttribute::getDataTypeStr()
  {
    TString thestr="D";
    switch(_nType)
      {
      case kVector: thestr="V"; break;
      case kString: thestr="S"; break;
      case kCategory: thestr="C"; break;
      case kInconnu: thestr="U"; break;
      default: thestr="D"; break;
	  
      }
    return thestr;
  }
  
  TString TAttribute::getLegend()
  {

    TString slegend = GetTitle();

    if (_sunity.Length() > 0)
      slegend += TString(Form(" [%s]", _sunity.Data()));

    return slegend;
  }


  void TAttribute::setFileKey(TString sfile, TString skey,
			      TString sformatToSubstitute, TAttributeFileKey::EFileType sFileType)
  {
    //     try{
    //       if(!_KeyList.empty())
    //         for(list<TAttributeFileKey *>::iterator it =_KeyList.begin(); it!=_KeyList.end();it++)
    //           if ((*it)->_sKey == skey && (*it)->_sFile == Form("%s/%s",gSystem->pwd(),sfile.Data()) )
    //         throw URANIE::Exceptions::UErrorExceptions(__FILE__,__LINE__,Form("TAttribute::setKey The key [%s] was already defined for the file [%s]",((*it)->_sKey).Data(),((*it)->_sFile).Data()));
    //     }
    //     catch (URANIE::Exceptions::UErrorExceptions& ue) {
    //       ue.printMessage();
    //       throw(ue);
    //     }

    //     try{
    //       if(!_KeyList.empty())
    //         for(list<TAttributeFileKey *>::iterator it =_KeyList.begin(); it!=_KeyList.end();it++)
    //           if ((*it)->_sFile == Form("%s/%s",gSystem->pwd(),sfile.Data()) && sFileType == URANIE::DataServer::TAttributeFileKey::kKey)
    //         throw URANIE::Exceptions::UErrorExceptions(__FILE__,__LINE__,Form("TAttribute::setKey A key (%s) was already defined for the file [%s] with type kKey",skey.Data(),((*it)->_sFile).Data()));
    //     }
    //     catch (URANIE::Exceptions::UErrorExceptions& ue) {
    //       ue.printMessage();
    //       throw(ue);
    //     }

    //The 2 previous "try" are gathered in this one.
    if (sformatToSubstitute.Length() == 0)
      sformatToSubstitute = "%e";
    try
      {
        if (!_KeyList.empty())
	  for (list<TAttributeFileKey *>::iterator it = _KeyList.begin();
	       it != _KeyList.end(); it++)
	    if (!strcmp(gSystem->BaseName((*it)->_sFile),
                        gSystem->BaseName(sfile.Data())))
	      throw URANIE::Exceptions::UErrorExceptions(__FILE__,
							 __LINE__,
							 Form(
							      "TAttribute::setFileKey A key (%s) was already defined for the file [%s] with type kKey",
							      skey.Data(), ((*it)->_sFile).Data()));
      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
        throw(ue);
      }

    TAttributeFileKey * Args = new TAttributeFileKey;
    char *result=NULL;
    if ((gSystem->DirName(sfile.Data()) == TString(".")) || // Relative path on Linux
	(gSystem->DirName(sfile.Data()) == TString(""))) // Relative path on Windows
      {
	result = gSystem->ConcatFileName(gSystem->pwd(), sfile.Data());
	sfile=TString(result);
	delete[] result; result=NULL;
      }
    Args->_sFile = sfile.Data();
    if (skey == "" ? Args->_sKey = GetName() : Args->_sKey = skey)
      Args->_sFormat = sformatToSubstitute;
    Args->_sFileType = sFileType;
    Args->_sEventList = NULL;
    _KeyList.push_back(Args);
  }


  void TAttribute::setFieldOfKey(Int_t ind)
  {
      UNUSED(ind);
  }

  void TAttribute::setFieldOfKey(Int_t Index, Int_t ind)
  {
    list<TAttributeFileKey *>::iterator it = _KeyList.begin();
    advance(it, Index);
    TString filename = (*it)->_sFile;
    if ((*it)->_sEventList == NULL)
      {
        (*it)->_sEventList = new TEventList(Form("tel_%s_%s_%s",
                                                 GetName(),
                                                 ((*it)->_sKey).Data(),
                                                 filename.Data()),
                                            Form("List of indexes from name[%s] with key[%s] in file[%s] and flag is with format",
                                                 GetName(),
                                                 ((*it)->_sKey).Data(),
                                                 filename.Data()));

      }
    _nfields.push_back((*it)->_sEventList);
    (*it)->_sEventList->Enter(ind);
  }

  Bool_t TAttribute::setDefault(TString  value)
  {
    Bool_t ok;
    
    if (_blog)
      cout << "TAttribute::setDefault [" << value << "]" << endl;

    switch(_nType)
    {
    case kString:
        ok=setDefaultString( value );
        break;
        
    case kVector:
    {
        vector<double> vecvalue;
        URANIE::DataServer::UEntry::fillVectorFromStr(value, ",", vecvalue);
        ok=setDefaultVector( vecvalue );
        break;
    }
    
    default:
    {
        double douvalue = atof(value);
        ok=setDefaultValue( douvalue );
        break;
    }
    }

    return ok;
  }

TString TAttribute::getDefault(TString format)
{

    TString output="";
    string delim=",";
    
    switch(_nType)
    {
    case kString:
        if( _bHaveDefaultString )
            output=_defaultString;
        break;

    case kVector:
    {
        if( _bHaveDefaultVector )
            output=URANIE::DataServer::UEntry::dumpVectorStr(&_defaultVector, -1, delim, format);
        break;
    }
    
    default:
    {
        if( _bHaveDefaultValue )
            output=Form(format,_defaultValue);
        break;
    }
    }

    return output;
}


  Bool_t TAttribute::setDefaultValue(Double_t value)
  {
    double indic;
    Bool_t ok;
    
    if (_blog)
      cout << "TAttribute::setDefaultValue [" << value << "]" << endl;

    // types
    try
    {
        if( _nType == kVector)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TAttribute::setDefaultValue: attribute is a vector, so use setDefaultVector or setDefault instead");
        if( _nType == kString)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TAttribute::setDefaultValue: attribute is a string, so use setDefaultString or setDefault instead");
        
    } catch (URANIE::Exceptions::UErrorExceptions& ue){
        ue.printMessage();
        throw ue;
    }
    
    // bounds
    try {
        if (_bHaveUpperBound && value > upperBound) {
            value = upperBound;
	        throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
		            Form("TAttribute::setDefaultValue: value [%20.20e] is bigger than upper bound [%20.20e], set to it!",
						    value, upperBound));
		}
		else if (_bHaveLowerBound && value < lowerBound) {
		    value = lowerBound;
	        throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
		            Form("TAttribute::setDefaultValue: value [%20.20e] is smaller than lower bound [%20.20e], set to it!",
						    value, lowerBound));
		}
    } catch (URANIE::Exceptions::UWarningExceptions& ue) {
        ue.printMessage();
    }
    
    _defaultValue = value;
    _bHaveDefaultValue = kTRUE;
    return _bHaveDefaultValue;
  }

Bool_t TAttribute::setDefaultVector(vector<double> &value) 
  {
    double indic;
    Bool_t ok;
    
    if (_blog)
    {
        cout << "TAttribute::setDefaultVector [" << URANIE::DataServer::UEntry::dumpVectorStr(&value) << "]" << endl;
    }
    
    try
    {
        if( _nType != kVector)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TAttribute::setDefaultVector: attribute is a not vector, so use setDefault instead");
    } catch (URANIE::Exceptions::UErrorExceptions& ue){
        ue.printMessage();
        throw ue;
    }

    _defaultVector.clear();    
    for(unsigned int ival=0; ival<value.size(); ival++)
    {
        _defaultVector.push_back( value.at(ival) );
    }
    _bHaveDefaultVector = kTRUE;
    
    return _bHaveDefaultVector;
  }

  Bool_t TAttribute::hasDefaultString()
  {
      try{
          if( _nType != kString)
              throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                         "TAttribute::hasDefaultString should not be called for non-string attributes. Use hasDefault() instead");
      } catch (URANIE::Exceptions::UErrorExceptions& ue){
          ue.printMessage();
          throw ue;
      }

      return _bHaveDefaultString;
  }

  Bool_t TAttribute::hasDefaultVector()
  {
      try{
          if( _nType != kVector)
              throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                         "TAttribute::hasDefaultVector should not be called for non-vector attributes. Use hasDefault() instead");
      } catch (URANIE::Exceptions::UErrorExceptions& ue){
          ue.printMessage();
          throw ue;
      }

      return _bHaveDefaultVector;
  }

  Bool_t TAttribute::hasDefaultValue()
  {
      try{
          if( _nType == kString || _nType == kVector )
              throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                         "TAttribute::hasDefaultValue should not be called for non-real attributes. Use hasDefault() instead");
      } catch (URANIE::Exceptions::UErrorExceptions& ue){
          ue.printMessage();
          throw ue;
      }

      return _bHaveDefaultValue;
  }

  Bool_t TAttribute::setDefaultString(TString value)
  {
    Bool_t ok;
    
    if (_blog)
      cout << "TAttribute::setDefaultString [" << value << "]" << endl;

    try
    {
        if( _nType != kString)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TAttribute::setDefaultString: attribute is not a string one, so use setDefault instead");
        
    } catch (URANIE::Exceptions::UErrorExceptions& ue){
        ue.printMessage();
        throw ue;
    }
    
    _defaultString = value;
    _bHaveDefaultString = kTRUE;

    return _bHaveDefaultString;
  }

  Bool_t TAttribute::getDefaultValue(Double_t &val)
  {      
      
    Bool_t res = kFALSE;
    try
    {
        if( _nType == kVector)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TAttribute::getDefaultValue: default value requested while attribute is a vector. Use getDefaultVector instead");
        if( _nType == kString)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TAttribute::getDefaultValue: default value requested while attribute is a vector. Use getDefaultString instead");
        
    } catch (URANIE::Exceptions::UErrorExceptions& ue){
        ue.printMessage();
        throw ue;
    }
    
    if (_bHaveDefaultValue)
      {
        val = _defaultValue;
        res = kTRUE;
      }
    return res;
  }

Bool_t TAttribute::getDefaultVector(vector<double> *val)
  {      
      
    Bool_t res = kFALSE;
    try
    {
        if( _nType != kVector)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TAttribute::getDefaultVector: default vector requested while attribute is not a vector.");
    } catch (URANIE::Exceptions::UErrorExceptions& ue){
        ue.printMessage();
        throw ue;
    }

    if (_bHaveDefaultVector)
    {
        val->clear();
        for(unsigned int ival=0; ival<_defaultVector.size(); ival++)
            val->push_back( _defaultVector.at(ival) );
        res = kTRUE;
    }

    return res;
  }

  Bool_t TAttribute::getDefaultString(string &val)
  {      
      
    Bool_t res = kFALSE;
    try
    {
        if( _nType != kString)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TAttribute::getDefaultString: default string requested while attribute is not a string.");
        
    } catch (URANIE::Exceptions::UErrorExceptions& ue){
        ue.printMessage();
        throw ue;
    }
    
    if (_bHaveDefaultString)
      {
          val = _defaultString;
          res = kTRUE;
      }
    return res;
  }

  void TAttribute::setStepValue(Double_t value)
  {
    _stepValue = value;
    _bHaveStepValue = kTRUE;
  }

  Bool_t TAttribute::getStepValue(Double_t &val)
  {
    Bool_t res = kFALSE;
    if (_bHaveStepValue)
      {
        val = _stepValue;
        res = kTRUE;
      }
    return res;
  }

  void TAttribute::setStd(Double_t val, int iel)
  {

    try
      {
        if (val < 0)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     "Standard deviation can't be negative!");
      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
        throw ue;
        exit(-1);
      }
    checkSizeBeforePush("Std", iel, val);
  }

  void TAttribute::setMean(Double_t val, int iel)
  {

    try
      {
        if (val < getLowerBound() || val > getUpperBound())
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form("For Attribute [%s], given mean [%g] isn't between boundaries [%g,%g]!",GetName(),val,getLowerBound(),getUpperBound()));
      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
        throw ue;
        exit(-1);
      }
    checkSizeBeforePush("Mean", iel, val);
  }


  ///////////////////////////////////////////////////////////
  //  Treatment the boundaries of an attribute
  ///////////////////////////////////////////////////////////
  void TAttribute::setBounds(Double_t lower, Double_t upper)
  {
    try
      {
        if (_blog)
	  cout << " ** TAttribute::setBounds name[" << GetName() << "] lower["
	       << lower << "] Upper [" << upper << "]" << endl;
        if (lower > upper)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form(
							  "TAttribute::setBounds method\n For attribute name[%s], the lower value[%e] is greater than the upper value[%e].",
							  GetName(), lower, upper));
        lowerBound = lower;
        _bHaveLowerBound = kTRUE;

        upperBound = upper;
        _bHaveUpperBound = kTRUE;

      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
        throw(ue);
      }
  }

  void TAttribute::setLowerBound(Double_t val, bool internalcall)
  {
    try
      {
        if (val > upperBound)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     "Given lower bound is superior to upperbound!");
      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
        throw ue;
        exit(-1);
      }
    lowerBound = val;
    if(!internalcall)
        _bHaveLowerBound = kTRUE;
  }

  void TAttribute::setUpperBound(Double_t val, bool internalcall)
  {
    try
      {
        if (val < lowerBound)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     "Given upper bound is inferior to lowerbound!");
      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
        throw ue;
        exit(-1);
      }
    upperBound = val;
    if(!internalcall)
        _bHaveUpperBound = kTRUE;
  }
  //////////////////////////
  // private  methodes
  ////////////////
  void TAttribute::setSonAttribute(TAttribute * tatt)
  {
    if ((tatt->haveSon() == kFALSE) && (this->getLevel() == 0))
      {
        _attSon = tatt;
	tatt->addShare(); //added by JBB for mem leak
        _attSon->setLevel(getLevel() + 1);
        _haveSon = kTRUE;
      }
    else
      throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						 "TAttribute::setSonAttribute - ERROR: Multi level doesn't accepted!!!");
  }
  void TAttribute::removeSonAttribute()
  {
    if ((this->haveSon() == kTRUE) && (_attSon->getLevel() == 1))
      {
	_attSon->setLevel(_attSon->getLevel()-1);
	_attSon->delShare();
	if(_attSon->getShare()==0)
	  delete _attSon;
        _attSon = NULL;
        _haveSon = kFALSE;
      }

  }
  Int_t TAttribute::getLevel()
  {
    return _nlevel;
  }
  void TAttribute::setLevel(Int_t nlevel)
  {
    _nlevel = nlevel;
  }
  TAttribute *TAttribute::getSonAttribute()
  {
    return _attSon;
  }
  Bool_t TAttribute::haveSon()
  {
    return _haveSon;
  }


  void TAttribute::setQuantile(double prob, double quantile, int iel)
  {
    try
      {
	if(iel < 0)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form("TAttribute::setQuantile - ERROR: cannot add element with index %d",
							  iel));
	if(UInt_t(iel) > _vquantile.size())
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form("TAttribute::setQuantile - ERROR: cannot add element with index %d, the vector size is only %lu",
							  iel, _vquantile.size()));
	else if(UInt_t(iel)==_vquantile.size())
	  _vquantile.push_back( new map<double,double>() );
      }catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
	ue.printMessage();
	throw ue;
      }
    
    if( _vquantile.at(iel)->count(prob)==0 )
      _vquantile.at(iel)->insert( std::pair<double,double>(prob,quantile) );
    else
      (*_vquantile.at(iel))[prob]=quantile;
    
  }

  void TAttribute::getQuantilesSize(int &nb, int iel)
  {
    try
      {
  	if(iel < 0)
  	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
  						     Form("TAttribute::getQuantileSize - ERROR: cannot access element with index %d",
  							  iel));
  	if(UInt_t(iel) >= _vquantile.size())
  	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
  						     Form("TAttribute::getQuantileSize - ERROR: cannot access element with index %d, the vector size is only %lu",
  							  iel, _vquantile.size()));
      }catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
  	ue.printMessage();
  	throw ue;
      }

    nb=_vquantile.at(iel)->size();
  }

  void TAttribute::getQuantiles(double *prob, double *quantile, int iel)
  {
    try
      {
	if(iel < 0)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form("TAttribute::getQuantiles - ERROR: cannot access element with index %d",
							  iel));
	if(UInt_t(iel) >= _vquantile.size())
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form("TAttribute::getQuantiles - ERROR: cannot access element with index %d, the vector size is only %lu",
							  iel, _vquantile.size()));
      }catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
	ue.printMessage();
	throw ue;
      }

    unsigned int i=0;
    for(map<double,double>::iterator itmap=_vquantile.at(iel)->begin(); itmap!=_vquantile.at(iel)->end();  itmap++)
      {
	prob[i] = itmap->first;
	quantile[i] = itmap->second;
	i++;
      }
    
  }

  void TAttribute::getQuantile(double &prob, double &quantile, int iel)
  {
    try
      {
	if(iel < 0)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form("TAttribute::getQuantiles - ERROR: cannot access element with index %d",
							  iel));
	if(UInt_t(iel) >= _vquantile.size())
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form("TAttribute::getQuantiles - ERROR: cannot access element with index %d, the vector size is only %lu",
							  iel, _vquantile.size()));
      }catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
	ue.printMessage();
	throw ue;
      }

    try
      {
	if( _vquantile.at(iel)->count(prob)==0  )
	  throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
						       Form("TAttribute::getQuantile. Quantile %e is not computed for element %d",prob,iel));    
      }catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
	ue.printMessage();
	throw ue;
      }

    quantile=(*_vquantile.at(iel))[prob];
    
  }

  void TAttribute::getQuantileVector(double &prob, vector<double> &quantile)
  {
    quantile.clear();
    for(unsigned int iel=0; iel<_vquantile.size(); iel++)
      {	
	try
	  {
	    if( _vquantile.at(iel)->count(prob)==0 )
	      throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							   Form("TAttribute::getQuantileVector. Quantile %e is not computed for element %d",prob,iel));    
	  }catch (URANIE::Exceptions::UWarningExceptions& ue)
	  {
	    ue.printMessage();
	    quantile.clear();
	    throw ue;
	  }

	quantile.push_back( (*_vquantile.at(iel))[prob] );
      }    
  }


  void TAttribute::getStatisticalVector(const char *type, vector<double> &vec)
  {
    vec.clear();
    vector<Double_t> *buffer=NULL; 
    if(strcmp(type,"Minimum")==0 || strcmp(type,"min")==0) buffer=&_vminimum;
    else if(strcmp(type,"Maximum")==0 || strcmp(type,"max")==0) buffer=&_vmaximum;
    else if(strcmp(type,"Mean")==0 || strcmp(type,"mean")==0) buffer=&_vmean;
    else if(strcmp(type,"Std")==0 || strcmp(type,"std")==0) buffer=&_vstd;
    
    for(unsigned int iel=0; iel<buffer->size()-1; iel++)
      vec.push_back(buffer->at(iel));
  }
  
  //Function to check whether the vector is correct (size=elementnb) to push the element
  //or only a re-assignement is needed. Throw exception else.
  void TAttribute::checkSizeBeforePush(const char* type, int iel, Double_t val)
  {
    _vbuffer=NULL;
    if(strcmp(type,"Minimum")==0) _vbuffer=&_vminimum;
    else if(strcmp(type,"Maximum")==0) _vbuffer=&_vmaximum;
    else if(strcmp(type,"Mean")==0) _vbuffer=&_vmean;
    else if(strcmp(type,"Std")==0) _vbuffer=&_vstd;

    try
      {
	if(iel < 0)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form("TAttribute::set%s - ERROR: cannot add element with index %d",
							  type, iel));
	if(UInt_t(iel) > _vbuffer->size())
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form("TAttribute::set%s - ERROR: cannot add element with index %d, the vector size is only %lu",
							  type, iel, _vbuffer->size()));
	else if(UInt_t(iel)==_vbuffer->size())
	  _vbuffer->push_back(val);
	else
	  _vbuffer->at(iel) = val;
      }catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
	ue.printMessage();
	throw ue;
      }
    
    _vbuffer=NULL;
  }

  //Function to check whether the vector is correct (size=elementnb) to push the element
  //or only a re-assignement is needed. Throw exception else.
  Double_t TAttribute::checkSizeBeforeGet(const char* type, int iel)
  {
    _vbuffer=NULL; double Default=0;
    if(strcmp(type,"Minimum")==0) { _vbuffer=&_vminimum; Default=DLE_MAX;}
    else if(strcmp(type,"Maximum")==0) { _vbuffer=&_vmaximum; Default= -1.0 / DLE_MAX;}
    else if(strcmp(type,"Mean")==0) { _vbuffer=&_vmean; Default=0;}
    else if(strcmp(type,"Std")==0) { _vbuffer=&_vstd; Default=0;}

    try
      {
	if(iel < 0)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form("TAttribute::set%s - ERROR: cannot add element with index %d in attribute %s",
							  type, iel, GetName()));
	if(UInt_t(iel) > _vbuffer->size())
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form("TAttribute::set%s - ERROR: cannot add element with index %d, the vector size of attribute %s is only %lu",
							  type, iel, GetName(), _vbuffer->size()));
      }catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
	ue.printMessage();
	throw ue;
      }

	try
      {
	if( _vbuffer->size()==0 )
	  throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
						       Form("TAttribute::get%s WARNING. No statistic computed for attribute %s. Default value returned",type, GetName()));
      }catch(URANIE::Exceptions::UWarningExceptions &ue)
      {
	ue.printMessage();
	return Default;
      }
    
    Default = _vbuffer->at(iel);
    return Default;
  }
  
  void TAttribute::clearVectors()
  {
    _vminimum.clear(); _vmaximum.clear(); _vmean.clear();  _vstd.clear();
    for(vector<map<double,double>* >::iterator it = _vquantile.begin() ; it != _vquantile.end(); ++it)
      {
	delete (*it);
	(*it) = NULL ;
      }
      _vquantile.clear();
  }
  
  ///////////////////////////////////////////////////////////
  //  Log
  ///////////////////////////////////////////////////////////
  void TAttribute::printLog()
  {
    cout << "*******************************" << endl;
    cout << "** TAttribute::printLog  ******" << endl;
    cout << "**    Name[" << GetName() << "]" << endl;
    cout << "**   Title[" << GetTitle() << "]" << endl;
    cout << "**   unity[" << _sunity << "]" << endl;
    cout << "**    type[" << _nType << "]" << endl;
    cout << "**   share[" << getShare() << "]" << endl;
    switch (_norigin)
      {
      case kInternal:
	cout << "**   Origin[kInternal]" << endl;
	break;
      case kIterator:
	cout << "**   Origin[kIterator]" << endl;
	break;
      case kDeleted:
	cout << "**   Origin[kDeleted]" << endl;
	break;
      case kConstant:
	cout << "**   Origin[kConstant]" << endl;
	break;
      case kAttribute:
	cout << "**   Origin[kAttribute]" << endl;
	break;
      }
    switch (_nAttribute)
      {
      case kInput:
	cout << "**   Attribute[kInput]" << endl;
	break;
      case kOutput:
	cout << "**   Attribute[kOutput]" << endl;
	break;
      }

    cout << "**   _snote[" << _snote << "]" << endl;
    for (list<TAttributeFileKey*>::iterator it = _KeyList.begin();
	 it != _KeyList.end(); it++)
      {
	cout << "**   Key [" << (*it)->_sKey << "] format to substitute["
	     << (*it)->_sFormat << "], File [" << (*it)->_sFile << "]";
	//        cout<<"**   The substitute format is "<<(*it)->_sFormat<<endl;
	switch ((*it)->_sFileType)
	  {
	  case TAttributeFileKey::kUnknown:
	    cout << " Type[kUnknown]" << endl;
	    break;
	  case TAttributeFileKey::kKey:
	    cout << " Type[kKey]" << endl;
	    break;
	  case TAttributeFileKey::kFlag:
	    cout << " Type[kFlag]" << endl;
	    break;
	  case TAttributeFileKey::kNewKey:
	    cout << " Type[kNewKey]" << endl;
	    break;
	  case TAttributeFileKey::kNewLine:
	    cout << " Type[kNewLine]" << endl;
	    break;
	  case TAttributeFileKey::kNewRow:
	    cout << " Type[kNewRow]" << endl;
	    break;
	  case TAttributeFileKey::kNewColumn:
	    cout << " Type[kNewColumn]" << endl;
	    break;
	  case TAttributeFileKey::kXMLAttribute:
	      cout << " Type[kXMLAttribute]" << endl;
	      break;
	  case TAttributeFileKey::kXMLField:
	      cout << " Type[kXMLField]" << endl;
	      break;
        case TAttributeFileKey::kFMU:
            cout << " Type[kFMU]" << endl;
            break;
          case TAttributeFileKey::kNewTDS:
	      cout << " Type[kNewTDS]" << endl;
	      break;
	  }
      }
    cout
      << "--------------------------------------------------------------------------------------------"
      << endl;
    if (!_nfields.empty())
      for (list<TAttributeFileKey*>::iterator it = _KeyList.begin();
	   it != _KeyList.end(); it++)
	{
	  if ((*it)->_sEventList != NULL)
	    ((*it)->_sEventList)->Print("all");
	  else
	    cout << "    ** TEventList NULL" << endl;
	}
    cout
      << "--------------------------------------------------------------------------------------------"
      << endl;

    cout << "** - min[";
    unsigned iel;
    for(iel=0; iel<_vminimum.size(); iel++)
      cout << _vminimum.at(iel) << ((iel!=_vminimum.size()-1) ? "," : "");
    cout << "] max[" ;
    for(iel=0; iel<_vmaximum.size(); iel++)
      cout << _vmaximum.at(iel) << ((iel!=_vmaximum.size()-1) ? "," : "");
    cout << "]" << endl;
    
    cout << "** - mean[";
    for(iel=0; iel<_vmean.size(); iel++)
      cout << _vmean.at(iel) << ((iel!=_vmean.size()-1) ? "," : "");
    cout << "] std[";
    for(iel=0; iel<_vstd.size(); iel++)
      cout << _vstd.at(iel) << ((iel!=_vstd.size()-1) ? "," : "");
    cout << "]" << endl;
    
    cout << "**   lowerBound[" << lowerBound << "] upperBound[" << upperBound
	 << "]" << endl;
    if (_bHaveDefaultValue)
      cout << "** _defaultValue [" << _defaultValue << "]" << endl;
    else
      cout << "** NOT _defaultValue[]" << endl;
    if (_bHaveStepValue)
      cout << "** _stepValue [" << _stepValue << "]" << endl;
    else
      cout << "** NOT _stepValue[]" << endl;

    if (_attSon)
      {
	cout << " ** Attribute to substitute level[" << _nlevel << "]" << endl;
	_attSon->printLog();
      }
    else
      cout << " ** No Attribute to substitute level[" << _nlevel << "]"
	   << endl;

    cout << "*******************************" << endl;

  }

} //  //namespace DataServer
} //namespace URANIE
