//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TBetaDistribution.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   2009-11-20 10:09:30

 \brief Interface of the class URANIE::DataServer::TBetaDistribution

 */
/**

 * @class URANIE::DataServer::TBetaDistribution
 *
 * @brief Description of the class TBetaDistribution.
 *
 This class defines a stochastic attribute following a Beta distribution.

 The distribution is defined over the interval [x_{min}, x_{max}]. The other user 
 parameters are \f$ \alpha \f$ and \f$ \beta \f$. Both are strictly positive real 
 values controlling the "shape" of the distribution:
 \li if \f$ \alpha = \beta \f$, the density function is symmetric;
 \li if \f$ \alpha > \beta \f$, the density is greater on the right of the interval;
 \li if \f$ \alpha < \beta \f$, the density is greater on the left of the interval;
 \li if \f$ \alpha = \beta = 1.0 \f$, the density uniform;
 \li if \f$ \alpha \f$ and \f$ \beta < 1.0 \f$, the density U-shaped;
 \li if \f$ \alpha \f$ and \f$ \beta > 1.0 \f$, the density unimodal;
 \li if \f$ \alpha >= 1.0 \f$ and \f$ \beta <= 1.0 \f$, or the reverse, the density is monotonic (increasing or decreasing respectively)

 \par Probability Density Function (PDF):
 Given \f$ Y = (x-x_{min})/(x_{max}-x_{min}) \f$ and \f$ B(\alpha,\beta) \f$ the Beta function: 
 \li \f$ f(x) = Y^{\alpha - 1} * ( 1 - Y )^{\beta - 1} / B(\alpha,\beta)\f$ for \f$ x \in [x_{min}, x_{max}] \f$
 \li \f$ f(x) = 0 \f$ otherwise

 \par Cumulative Distribution Function (CDF):
 Given \f$ Y = (x-x_{min})/(x_{max}-x_{min}) \f$, \f$ B(\alpha,\beta) \f$ the Beta function, and \f$ B_{Y}(\alpha,\beta) \f$ the Incomplete Beta function  
 \li \f$ F(x) = 0.0 \f$ for \f$ x < x_{min} \f$
 \li \f$ F(x) = B_{Y}(\alpha, \beta)/B(\alpha,\beta)\f$ for \f$ x \in [x_{min},x_{max}] \f$
 \li \f$ F(x) = 1.0 \f$ for \f$ x > x_{max} \f$

 \par Inverse CDF:
 There is no analytical representation of the Inverse CDF of the Beta distribution. 
 The value \f$ F^{-1}(p) \f$ is estimated by dichotomy from the CDF.

 */
#ifndef TBETADISTRIBUTION_H
#define TBETADISTRIBUTION_H

//Uranie
#include "TStochAttribut.h"

namespace URANIE
{
namespace DataServer
{
class TBetaDistribution: public TStochasticAttribute
{
private:

    Double_t _dparameterAlpha;  ///< The shape parameter \f$ \alpha\f$ 
    Double_t _dparameterBeta;   ///< The shape parameter \f$ \beta \f$

public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    /*!
     Default range is [0,1], \f$ \alpha = 0.5 \f$ and \f$ \beta = 0.5 \f$ 
     */
    TBetaDistribution();

    //! Constructor setting the name
    /*! 
     Default range is [0,1], \f$ \alpha = 0.5 \f$ and \f$ \beta = 0.5 \f$
     
     \param name (TString) Name of the attribute
     */
    TBetaDistribution(TString name);

    //! Constructor setting name, shape parameters and range
    /*!                 
     \param name (TString) Name of the attribute
     \param alpha (Double_t) Shape parameter \f$ \alpha \f$
     \param beta (Double_t) Shape parameter \f$ \beta \f$
     \param min (Double_t) Lower bound (Default = 0.0)
     \param max (Double_t) Upper bound (Default = 1.0)               
     */
    TBetaDistribution(TString name, Double_t alpha, Double_t beta,
            Double_t min = 0.0, Double_t max = 1.0);

    //! Construction of a TBetaDistribution from an existing TAttribute object
    /*!
     \warning this copy constructor is incomplete. Please do not use.
     (Nicolas Gilardi, 2010.05.28)
     
     \param att (TAttribute *) Pointer to an existing TAttribute object
     \param alpha (Double_t) Shape parameter \f$ \alpha \f$
     \param beta (Double_t) Shape parameter \f$ \beta \f$                
     */
    TBetaDistribution(URANIE::DataServer::TAttribute *att, Double_t alpha,
            Double_t beta);

    //! Default destructor
    virtual ~TBetaDistribution();
    //@}

    /** @name Law parameters
     This law has four parameters: the lower bound, the upper bound, the 
     shape parameter \f$ \alpha \f$ and the shape parameter \f$ \beta \f$. 
     The bounds parameters are both defined in the class 
     URANIE::DataServer::TAttribute. The shape parameters are defined here.
     **/
    //@{
    //! Set the shape parameter \f$ \alpha\f$
    /*!
     \param alpha (Double_t) Shape parameter \f$ \alpha \f$
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE)
     
     \exception UErrorExceptions if alpha < 1e-4.
     */
    virtual void setParameterAlpha(Double_t alpha, Bool_t recompute = kTRUE);

    //! Return the \f$ \alpha\f$ parameter
    Double_t getParameterAlpha()
    {
        return _dparameterAlpha;
    }

    //! Set the shape parameter \f$ \beta \f$
    /*!
     \param beta (Double_t) Shape parameter \f$ \beta \f$
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE)
     
     \exception UErrorExceptions if beta < 1e-4.
     */
    virtual void setParameterBeta(Double_t beta, Bool_t recompute = kTRUE);

    //! Return the \f$ \beta\f$ parameter
    Double_t getParameterBeta()
    {
        return _dparameterBeta;
    }

    //! Set all the parameters
    /*!
     \param alpha (Double_t) Shape parameter \f$ \alpha \f$
     \param beta (Double_t) Shape parameter \f$ \beta \f$
     \param lower (Double_t) Lower bound (Default = 0.0)
     \param upper (Double_t) Upper bound (Default = 1.0)               
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE)
     */
    virtual void setParameters(Double_t alpha, Double_t beta, Double_t lower =
            0.0, Double_t upper = 1.0, Bool_t recompute = kTRUE);
    //@}

    /** @name Distribution specific methods
     PDF, CDF, etc.
     **/
    //@{
    //! Computes the theoretical values of various statistical measures, like mean, standard deviation, mode, etc.
    virtual void computeTheoreticalInformation();

    //! Returns the probability density for the value x
    /*!
     We are using ROOT's implementation of the Beta density function.
     
     \param x (Double_t) An acceptable value for the attribute                 
     */
    virtual Double_t getPDF(Double_t x);

    //! Returns the cumulative distribution for the value x
    /*!
     We are using ROOT's implementation of the Beta distribution function.
     
     \param x (Double_t) An acceptable value for the attribute                
     */
    virtual Double_t getCDF(Double_t x);

    //! Returns the value corresponding to the cumulative density p
    /*!                 
     The output value x is estimated by dichotomy from the CDF.
     
     \param p (Double_t) A valid probability
     
     \exception UErrorExceptions if p is outside of the interval [0,1]                
     */
    virtual Double_t getInvCDF(Double_t p);
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::DataServer::TBetaDistribution, ID_DATASERVER)
    //Classe de
};

} // Fin du namespace DataServer
} // Fin du namespace URANIE
#endif
