//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TComposedDistribution.h
 \author Jean-Baptiste Blanchard <jean-baptiste.blanchard@cea.fr>
 \date   Wed Aug 24 12:00:00 CET 2022


 \brief Interface of the class URANIE::DataServer::TComposedDistribution

 */
/**

 * @class URANIE::DataServer::TComposedDistribution
 *
 * @brief Description of the class TComposedDistribution.
 *
 * This class allows to define a composition of stochastic laws which can be weighted in order to 
 * create a new law.
 *
 */
#ifndef TCOMPOSEDDISTRIBUTION_H
#define TCOMPOSEDDISTRIBUTION_H

//Uranie
#include "TStochAttribut.h"

namespace URANIE
{
namespace DataServer
{
class TComposedDistribution: public TInfiniteDistribution
{

private:
protected:
    int                                   _iNbLaws;                  ///< Number of laws in the composition
    double                                _dSumWeights;              ///< Sum of all weights
    vector<TStochasticAttribute *>        _inputLaws;                ///< Pointers of laws in the composition
    vector<double>                        _inputWeights;             ///< Weights of laws in the composition
    
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    TComposedDistribution();

    //! Constructor by name only
    /*!      
     \param name (TString) Name of the attribute
     */
    TComposedDistribution(TString name);


    //! Default destructor
    virtual ~TComposedDistribution();
    //@}
    //---------------------------------------------
    /** @name Getters and Setters methods
     Getting and setting attribute values.
     **/
    //@{
    //! add a distribution
    /*!
      \param statt: a pointer to the TStochasticAttribute to be added
      \param weight: this is the weight used to normalise the composition
      \exception if the weight is ==0 or negative and if the distribution is already found in the input list
    */
    
    int addDistribution(URANIE::DataServer::TStochasticAttribute *statt, double weight=1.);

    int getNbComponent() { return _iNbLaws;}
    
    //@}

    /** @name Statistical methods
     The statistical methods ( pdf, cdf, quantile, ...).
     **/
    //@{                
    //! Mean and Variance are meaningless for non ordered discrete values.
    /*!
     \exception always throws a UWarningExceptions.
     */
    virtual void computeTheoreticalInformation() {computeTheoreticalInformation(0);}
    void computeTheoreticalInformation(int info=0);

    //! CDF is meaningless for non ordered discrete values.
    /*!
     This function returns 0.0.
     \exception always throws a UWarningExceptions.
     */
    virtual Double_t getCDF(Double_t x);

    //! Returns the probability of a value.
    /*!
     If x belongs to the distribution, its probability will be equal 
     to the corresponding value in the pFrequencies array divided by the 
     sum of the pFrequencies values.
     NOTE: only values belonging to the distribution have a non-zero
     probability. 
     \param x (Double_t): the value for which the probability is computed 
     */
    virtual Double_t getPDF(Double_t x);

    //! Returns a value from the distribution.
    /*! 
     The choice of the value is based on a pseudo-cumulative 
     function where the order of the discrete values is the order 
     of appearance in the pValues array.
     \param p (Double_t): a probability value between 0 and 1
     */
    virtual Double_t getInvCDF(Double_t p);
    
    virtual double getTheoreticalMean() {computeTheoreticalInformation(0); return _dtheoreticalMean; }
    virtual double getTheoreticalStdDev() {computeTheoreticalInformation(1); return _dtheoreticalStdDev; }
    virtual double getTheoreticalMode() {computeTheoreticalInformation(2); return _dtheoreticalMode; }
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::DataServer::TComposedDistribution, ID_DATASERVER)
};
} // Fin du namespace DataServer
} // Fin du namespace URANIE
#endif
