//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//
//
//
//
//
//////////////////////////////////////////////////////////////
/*!
  \par Modifications
  \li JB Blanchard (CEA) 2016/02/15
  Changing the all logic, making this class inheriting from the TTree directly (to allow vectors of double and strings) 
  \li Cedric Forgues (C.S.) 2010/06/22
  Adding of une exception for the getBranchData() method.
  // *****************************************************************************

  */

#include "TDSNtupleD.h"
#include "TLeaf.h"
#include "TStyle.h"
#include "TObjArray.h"
#include "TTreeFormula.h"
#include "TObjString.h"
#include <algorithm>
#include <iostream>
#include <iomanip>
using std::ostringstream;
#include <fstream>

// Uranie
#include "UExceptions.h"

ClassImp (URANIE::DataServer::TDSNtupleD)

namespace URANIE
{
namespace DataServer
{
    
//
// ========== Constructors & destructor ================
//
TDSNtupleD::TDSNtupleD():
    TTree(),
    _varname(""),
    _vartype("")
{
}

//Destructor
TDSNtupleD::~TDSNtupleD()
{
    ClearVectorsAndMaps();
}


void TDSNtupleD::ClearVectorsAndMaps()
{
    if( _vfArgs.size() != 0)
    {
	for(std::vector<UEntry*>::iterator it=_vfArgs.begin(); it!=_vfArgs.end(); it++)
	    if( dynamic_cast<URANIE::DataServer::UEntry*>( (*it) ) != NULL )
		delete (*it);
    }
    _vfArgs.clear();
    _vfDArgs.clear();
    for(std::vector<std::string*>::iterator it=_vfSArgs.begin(); it!=_vfSArgs.end(); it++)
	delete (*it);
    _vfSArgs.clear();
    for(std::vector<std::vector<Double_t>* >::iterator it=_vfVArgs.begin(); it!=_vfVArgs.end(); it++)
	delete (*it);
    _vfVArgs.clear();
    _mTypeNb.clear(); _mVarTypeNb.clear();
}
  
//Copy constructor from TTree
TDSNtupleD::TDSNtupleD(TTree* ntd, const char* varlist, const char* cut, bool NewIsMainOne) :
    TTree()
{

    if(NewIsMainOne)
    {
	TString name = ntd->GetName(), title=ntd->GetTitle();
	ntd->SetTitle(title+TString("_copied")); ntd->SetName(name+TString("_copied"));
	_initTTree( name, title);
    }
    else
	_initTTree( (std::string(ntd->GetName())+"_copy").c_str(), (std::string(ntd->GetTitle())+"_copy").c_str() );

    // Check the branches
    TObjArray* branchList = ntd->GetListOfBranches();
    TBranch* bObject = (TBranch*) branchList->First();

    std::vector<std::string> buflist; 
    UEntry::StringParser(std::string(varlist),":",buflist);
    //Get the types and names
    TObjArray* leafList = new TObjArray();//ntd->GetListOfLeaves();
    std::vector<std::string> notpresentleaf; notpresentleaf.clear(); //for vector if the leaf is not present in the tree, the size of the vector itself is used
    while (bObject)
    {
	if( (std::string(varlist)!="*" && find(buflist.begin(), buflist.end(), std::string(bObject->GetName()) ) != buflist.end())
	    || std::string(varlist)=="*" )
	{    
	    _varname+=std::string(bObject->GetName())+":";
	    std::string thetype = std::string(bObject->GetClassName());
	    if(thetype=="string") _vartype+="S:"; 
	    else if(thetype=="vector<double>")
	    {    
		_vartype+="V:";
		if ( ((std::string(varlist)!="*" && find(buflist.begin(), buflist.end(), SIZENAME+"__"+std::string(bObject->GetName()) ) != buflist.end())
		      && ntd->FindLeaf( (SIZENAME+"__"+std::string(bObject->GetName())).c_str() )==0 ) || 
		     (std::string(varlist)!="*" && find(buflist.begin(), buflist.end(), SIZENAME+"__"+std::string(bObject->GetName()) ) == buflist.end()) ) 
		    notpresentleaf.push_back( std::string(bObject->GetName()) );
	    }
	    else _vartype+="D:";
	    leafList->Add((TLeaf*)bObject->GetLeaf(bObject->GetName()));
	}
	bObject = (TBranch*) branchList->After(bObject);
    }
    //Remove the final useless separation (":")
    _varname=_varname.substr(0, _varname.size()-1);
    _vartype=_vartype.substr(0, _vartype.size()-1);

    //Construct the tree from the information
    _branchConstruction(32000);

    //Loop over event to fill the new tree by copy
    TLeaf* lObject;
    std::vector<double> buffer;  int shiftForV=0;
    TTreeFormula *cutForm = NULL;
    if ( strcmp(cut,"")!=0 )
    {
	cutForm = new TTreeFormula("Cut", cut, ntd);
	try
	{
	    if (cutForm->Compile() != 0)
		throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							   Form(
							       "TDSNtupleD::ntupleDataRead  ERROR: the formula [%s] not correct",
							       cutForm->GetTitle()));
	} catch (URANIE::Exceptions::UErrorExceptions& ue)
	{
	    ue.printMessage();
	    ClearVectorsAndMaps();
	    throw ue;
	}
    }

    for (UInt_t entry = 0; entry < ntd->GetEntries(); entry++)
    {
	ntd->GetEntry(entry); shiftForV=0;
	if (cutForm != 0 && cutForm->EvalInstance(0) != 0.0)
	{
	    //	    cout<<"cutForm  "<<cutForm->GetTitle()<<"  is  "<<cutForm->EvalInstance(0)<<endl;
	    continue;
	}
	
	for (_i=0;_i<_fNvar;_i++)
	{
	    lObject = (TLeaf*) leafList->At(_i-shiftForV);
	    _vfArgs[_i]->assignEntry( (void*)(lObject->GetValuePointer()), strcmp(lObject->GetTypeName(),"Float_t") );
	    if(_vfArgs[_i]->tag==UEntry::Vector) 
	    {    
		if ( find( notpresentleaf.begin(), notpresentleaf.end(), std::string(lObject->GetName()) ) != notpresentleaf.end())
		{
		    _i++;
		    *(_vfArgs[_i]->val.dou) = _vfArgs[_i-1]->val.vec->size();
		    shiftForV++;
		}
	    }
	}	
	TTree::Fill();
    }

    if(cutForm!=NULL)
	delete cutForm;
    cutForm=NULL;

    if(leafList!=NULL)
	delete leafList;
    leafList=NULL;
}

  
TDSNtupleD::TDSNtupleD(const char* name, const char* title, const char* varlist,
		       const char* vartype, Int_t bufsize) :
    TTree(),
    _varname(varlist),
    _vartype(vartype)
{
    _initTTree(name, title);
    _branchConstruction(bufsize);
}

  
TDSNtupleD::TDSNtupleD(const char* name, const char* title, const char* varlist,
		       Int_t bufsize) :
    TTree(),
    _varname(varlist),
    _vartype("")
{
    _initTTree(name, title);
    _branchConstruction(bufsize);
}

  
TDSNtupleD::TDSNtupleD(const char* name, const char* title, const char* varlist,
		       TMatrixD data, Int_t bufsize) :
    TTree(),
    _varname(varlist),
    _vartype("")
{
    _initTTree(name, title);
    _branchConstruction(bufsize);

    // Check the data size
    try
    {
	if (UInt_t(data.GetNcols()) != _fNvar)
	{
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form(
							   "TDSNtupleD: the number of column of the matrix (%d) is different from the number of branches of the TDSNtupleD (%d)",
							   data.GetNcols(), _fNvar));
	}
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
	ue.printMessage();
	ClearVectorsAndMaps();
	throw ue;
    }

    // Fill the tree
    for (Int_t line = 0; line < data.GetNrows(); line++)
	Fill(data[line].GetPtr());
}

  
//
// ========== Initialisation ================
//
void TDSNtupleD::_initTTree(const char* name, const char* title)
{
    // TNamed params
    SetName(name);
    SetTitle(title);

    // gDirectory->Append(this);
    fDirectory = 0;//gDirectory;

    // Defaults color and form
    SetMarkerColor(2);
    SetMarkerSize(4);
    SetMarkerStyle(7);


    // Fill default
    SetFillColor(gStyle->GetHistFillColor());
    SetFillStyle(gStyle->GetHistFillStyle());
}

void TDSNtupleD::_branchConstruction(Int_t bufsize)
{
    //Reset everything
    _fNvar = 0;
    ClearVectorsAndMaps();
    _mTypeNb.insert( std::pair<std::string,unsigned int>("D",0) ); _mTypeNb.insert( std::pair<std::string,unsigned int>("V",0) ); _mTypeNb.insert( std::pair<std::string,unsigned int>("S",0) );
		     
    //Check the list of names and types
    std::vector<std::string> v_varname, v_vartype;
    UEntry::StringParser(std::string(_varname), std::string(":"), v_varname);
    UEntry::StringParser(std::string(_vartype), std::string(":"), v_vartype);

    //if list of types and names have not the same number
    try{
	if( v_vartype.size() != v_varname.size())
	{
	    if( v_vartype.size() == 0 )
		for(_i=0; _i<v_varname.size(); _i++)
		    v_vartype.push_back("D");
	    else
		throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							   Form(
							       "TDSNtupleD: the number of elements in varname and vartype does not match"));
	}
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
	ue.printMessage();
	ClearVectorsAndMaps();
	throw ue;
    }

    //Detect if the size branch is in the list of variable... if not put the uefull index in map for later modification
    map<unsigned int, std::string> inname;
    for(_i=0; _i<v_vartype.size(); _i++)
    {
	if(v_vartype.at(_i)=="V")
	    if( !TString(_varname).Contains( (SIZENAME+"__"+v_varname.at(_i)).c_str() ) ) 
		inname[ _i +1 + inname.size()]=SIZENAME+"__"+v_varname.at(_i);
    }

    //Modifying the vector of attribute
    for(map<unsigned int, std::string>::iterator itmap=inname.begin(); itmap!=inname.end(); itmap++)
    {
	v_varname.insert( v_varname.begin()+itmap->first, itmap->second);
	v_vartype.insert( v_vartype.begin()+itmap->first, "D" );
    }

    //Check that the size is always right after the vector
    for(_i=0; _i<v_vartype.size(); _i++)
    {
	if( v_vartype.at(_i)=="V" )
	    if( find( v_varname.begin(), v_varname.end(), SIZENAME+"__"+v_varname.at(_i) ) != v_varname.begin()+_i+1 )
	    {
		unsigned int toch = std::distance(v_varname.begin(), find( v_varname.begin(), v_varname.end(), SIZENAME+"__"+v_varname.at(_i) ));
		iter_swap( v_varname.begin()+toch, v_varname.begin()+_i+1);
		iter_swap( v_vartype.begin()+toch, v_vartype.begin()+_i+1);
		_i=0;
	    }
    }

    //Transfer back the possible modification of the list of atribute to the string that contains type/name
    _varname=""; _vartype="";
    for(_i=0; _i<v_varname.size(); _i++)
    {
	_varname+=v_varname.at(_i)+":";
	_vartype+=v_vartype.at(_i)+":";
    }
    _varname=_varname.substr(0, _varname.size()-1);
    _vartype=_vartype.substr(0, _vartype.size()-1);
    
    //Setting types and orders
    _fNvar =  v_varname.size();  
    for(_i=0; _i<_fNvar; _i++)
    {
	_mVarTypeNb.insert( std::pair<unsigned int,unsigned int>(_i, _mTypeNb[v_vartype.at(_i)]) );
	_mTypeNb[v_vartype.at(_i)]=_mTypeNb[v_vartype.at(_i)]+1;

	if( v_vartype.at(_i) == "D" )
	    _vfArgs.push_back( new UEntry(UEntry::Double, v_varname.at(_i).c_str(), NULL) );
	else if( v_vartype.at(_i) == "S" )
	{
	    _vfSArgs.push_back( new std::string("") );
	    _vfArgs.push_back( new UEntry(UEntry::String, v_varname.at(_i).c_str(), _vfSArgs.back() ) );
	}
	else if( v_vartype.at(_i) == "V" )
	{
	    _vfVArgs.push_back( new std::vector<Double_t>() );
	    _vfArgs.push_back( new UEntry(UEntry::Vector, v_varname.at(_i).c_str(), _vfVArgs.back() ) );
	}
    }

    
    //Save memory for objects    
    if ( _mTypeNb["D"] != 0)
    {
	_vfDArgs.resize(_mTypeNb["D"],0.);
	for(_i=0; _i<_vfArgs.size(); _i++)
	    if( _vfArgs.at(_i)->tag == URANIE::DataServer::UEntry::Double )
	    {
		//XYZ<<"  Memory mapping of entry "<<_i<<"  with _i "<<_i<<" index "<< _mVarTypeNb[_i] << endl;
		_vfArgs.at(_i)->setAddress( (void*)( &_vfDArgs[ _mVarTypeNb[_i] ] ) );
	    }
    }

    //  Create one branch for each variable
    for (_i=0;_i<_fNvar;_i++) 
    {
	if(_vfArgs[_i]->tag==UEntry::Double)
	    TTree::Branch(v_varname[_i].c_str(), _vfArgs[_i]->val.dou, (v_varname[_i]+"/"+v_vartype.at(_i)).c_str(), bufsize);
	else if (_vfArgs[_i]->tag==UEntry::String)
	    TTree::Branch(v_varname[_i].c_str(), "string", &_vfArgs[_i]->val.str);
	else if (_vfArgs[_i]->tag==UEntry::Vector)
	    TTree::Branch(v_varname[_i].c_str(), "vector<double>", &_vfArgs[_i]->val.vec, bufsize);
    }

    SetDirectory( gROOT );
}

  
//
// ========== Fillers ================
//

Int_t TDSNtupleD::Fill()
{
    if( _varname!=""  && _vfArgs.size() == 0 )
    {
	int iv=0, is=0;
	for (_i = 0; _i < _fNvar; _i++)
	{
	    TBranch *branch = (TBranch*)fBranches.UncheckedAt(_i);
	    if (branch) 
	    {
		std::string thetype = std::string( branch->GetClassName() );
		if ( thetype=="vector<double>" ) _vfArgs.push_back( new UEntry(UEntry::Vector, branch->GetName(), _vfVArgs[iv++] ) );
		else if ( thetype=="string" ) _vfArgs.push_back( new UEntry(UEntry::String, branch->GetName(), _vfSArgs[is++] ) );
		else
		{
		    if( _mVarTypeNb.size() == 0 )
			_vfArgs.push_back( new UEntry(UEntry::Double, branch->GetName(), &_vfDArgs[_i]) );
		    else
			_vfArgs.push_back( new UEntry(UEntry::Double, branch->GetName(), &_vfDArgs[ _mVarTypeNb[_i] ]) );
		}
	    }
	}
    }
    return TTree::Fill();
}

Int_t TDSNtupleD::Fill(Double_t x0, Double_t x1, Double_t x2, Double_t x3, Double_t x4, Double_t x5,
		       Double_t x6, Double_t x7, Double_t x8, Double_t x9, Double_t x10, Double_t x11,
		       Double_t x12, Double_t x13, Double_t x14)
{
    try
    {
	if ( _fNvar != _mTypeNb["D"] )
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__,
						       __LINE__,
						       Form(
							   "TDSNtupleD::Fill(Double_t, Doublet_t ...): Method not available when dealing with a TDSNtupleD that contains %d vector<double> and %d strings",
							   _mTypeNb["V"],_mTypeNb["S"]));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
	ue.printMessage();
	throw ue;
    }
      
    if (_fNvar >  0) *_vfArgs[0]->val.dou  = x0;
    if (_fNvar >  1) *_vfArgs[1]->val.dou  = x1;
    if (_fNvar >  2) *_vfArgs[2]->val.dou  = x2;
    if (_fNvar >  3) *_vfArgs[3]->val.dou  = x3;
    if (_fNvar >  4) *_vfArgs[4]->val.dou  = x4;
    if (_fNvar >  5) *_vfArgs[5]->val.dou  = x5;
    if (_fNvar >  6) *_vfArgs[6]->val.dou  = x6;
    if (_fNvar >  7) *_vfArgs[7]->val.dou  = x7;
    if (_fNvar >  8) *_vfArgs[8]->val.dou  = x8;
    if (_fNvar >  9) *_vfArgs[9]->val.dou  = x9;
    if (_fNvar > 10) *_vfArgs[10]->val.dou = x10;
    if (_fNvar > 11) *_vfArgs[11]->val.dou = x11;
    if (_fNvar > 12) *_vfArgs[12]->val.dou = x12;
    if (_fNvar > 13) *_vfArgs[13]->val.dou = x13;
    if (_fNvar > 14) *_vfArgs[14]->val.dou = x14;

    return TTree::Fill();
}


Int_t TDSNtupleD::Fill(Double_t* AllValues)
{

    try
    {
	if ( _fNvar != _mTypeNb["D"] )
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__,
						       __LINE__,
						       Form(
							   "TDSNtupleD::Fill(Double_t*): Method not available when dealing with a TDSNtupleD that contains %d vector<double> and %d strings",
							   _mTypeNb["V"],_mTypeNb["S"]));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
	ue.printMessage();
	throw ue;
    }
    
    // *-*- Store array x into buffer
    for (_i=0;_i<_fNvar;_i++)  
	*_vfArgs[_i]->val.dou = AllValues[_i];

    return TTree::Fill();
}

//______________________________________________________________________________
Int_t TDSNtupleD::Fill(std::string line, int IterIndex)
{

    // Parse the input line
    std::vector<std::string> theval, buffer;
    UEntry::StringParser(line, " ", theval);

    try
    {
	if ( _fNvar != theval.size()+_mTypeNb["V"] )
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__,
						       __LINE__,
						       Form(
							   "TDSNtupleD::Fill(string): Number of field (%lu) different from number of branches (%d:%s) : %s",
							   theval.size(),_fNvar,_varname.c_str(), line.c_str()));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
	ue.printMessage();
	throw ue;
    }

    TString beg,end;
    // icol was defined in case the iterator is in the input file, but not as first column...
    // shifticolV was introduced to take into account the shift induced by the double attribute corresponding to vector size.
    UInt_t icol=0, shiftV=0; 
    // *-*- Store array x into buffer
    for (_i=0;_i<_fNvar;_i++)
    {

	if(IterIndex==-1) icol=_i;
	else
	{
	    if(_i==0) icol=IterIndex;
	    else if ((_i-shiftV)>UInt_t(IterIndex)) icol=_i;
	    else if ((_i-shiftV)<=UInt_t(IterIndex)) icol=_i-1;
	}
	icol -= shiftV;

	switch(_vfArgs[_i]->tag)
	{
	case URANIE::DataServer::UEntry::String: beg=end="\""; break;
	case URANIE::DataServer::UEntry::Vector: beg="["; end="]"; break;
	default: beg=end=""; break;
	}
	
	_vfArgs[_i]->fillEntry( TString(theval[icol]), ",", beg, end );
	if(_vfArgs[_i]->tag==UEntry::Vector) 
	{
	    _i++;
	    *_vfArgs[_i]->val.dou = (double)_vfArgs[_i-1]->val.vec->size();
	    shiftV++;
	}
    }
    return TTree::Fill();
}


Int_t TDSNtupleD::Fill(std::vector<URANIE::DataServer::UEntry*> *event)
{
    try
    {
	if ( _fNvar != event->size()+_mTypeNb["V"] )
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__,
						       __LINE__,
						       Form(
							   "TDSNtupleD::Fill(string): Number of field (%lu) different from number of branches (%d:%s) ",
							   event->size(),_fNvar,_varname.c_str()));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
	ue.printMessage();
	throw ue;
    }
    
    // icol was defined in case the iterator is in the input file, but not as first column...
    // shifticolV was introduced to take into account the shift induced by the double attribute corresponding to vector size.
    UInt_t shiftV=0; 
    // *-*- Store array x into buffer
    for (_i=0;_i<_fNvar;_i++)
    {

	_vfArgs[_i]->copyEntry( event->at(_i-shiftV) );
	if(_vfArgs[_i]->tag==UEntry::Vector) 
	{
	    _i++;
	    *_vfArgs[_i]->val.dou = (double)_vfArgs[_i-1]->val.vec->size();
	    shiftV++;
	}
    }
    return TTree::Fill();

}


Int_t TDSNtupleD::Fill(std::vector<URANIE::DataServer::UEntry*> *vinput, std::vector<URANIE::DataServer::UEntry*> *voutput)
{
    try
    {
	if ( _fNvar != (vinput->size() + voutput->size())+_mTypeNb["V"] )
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__,
						       __LINE__,
						       Form(
							   "TDSNtupleD::Fill(string): Number of field in both vectors (%lu, %lu) different from number of branches (%d:%s) ",
							   vinput->size(), voutput->size(),_fNvar,_varname.c_str()));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
	ue.printMessage();
	throw ue;
    }
    
    // icol was defined in case the iterator is in the input file, but not as first column...
    // shifticolV was introduced to take into account the shift induced by the double attribute corresponding to vector size.
    UInt_t shiftV=0; 
    // *-*- Store array x into buffer
    for (_i=0;_i<_fNvar;_i++)
    {

        if( _i-shiftV < vinput->size() ) 
            _vfArgs[_i]->copyEntry( vinput->at(_i-shiftV) );
        else
            _vfArgs[_i]->copyEntry( voutput->at(_i-shiftV-vinput->size()) );
	if(_vfArgs[_i]->tag==UEntry::Vector) 
	{
	    _i++;
	    *_vfArgs[_i]->val.dou = (double)_vfArgs[_i-1]->val.vec->size();
	    shiftV++;
	}
    }
    return TTree::Fill();

}


//
// ========== Operation on branches ==============
//


//______________________________________________________________________________
void TDSNtupleD::ResetBranchAddress(TBranch *branch)
{
    // Reset the branch addresses to the internal fArgs array. Use this
    // method when the addresses were changed via calls to SetBranchAddress().
//	XYZZY;
    if (branch)
    {
	UInt_t index = fBranches.IndexOf(branch);
	if (index>0)
	{
	    //XYZ<<" Looking at branch "<<branch->GetName()<<" with index "<<index<<" for vect size "<<_vfArgs.size()<<endl;
	    if(_vfArgs[index]->tag==UEntry::Double) branch->SetAddress(_vfArgs[index]->val.dou);
	    else if(_vfArgs[index]->tag==UEntry::String) branch->SetAddress(&_vfArgs[index]->val.str);
	    else if(_vfArgs[index]->tag==UEntry::Vector) branch->SetAddress(&_vfArgs[index]->val.vec);
	}
    }
}


//______________________________________________________________________________
void TDSNtupleD::ResetBranchAddresses()
{
    // Reset the branch addresses to the internal fArgs array. Use this
    // method when the addresses were changed via calls to SetBranchAddress().
//	XYZZY;
    for (_i = 0; _i < _fNvar; _i++)
    {
	TBranch *branch = (TBranch*)fBranches.UncheckedAt(_i);
	if (branch) 
	{
	    if(_vfArgs[_i]->tag==UEntry::Double) branch->SetAddress(_vfArgs[_i]->val.dou);
	    else if(_vfArgs[_i]->tag==UEntry::String) branch->SetAddress(&_vfArgs[_i]->val.str);
	    else if(_vfArgs[_i]->tag==UEntry::Vector) branch->SetAddress(&_vfArgs[_i]->val.vec);
	}
    }
}


TBranch* TDSNtupleD::Branch(const char* name, void* address, const char* leaflist, Int_t bufsize)
{
//	XYZZY;
    
    TString sleaflist = Form("%s", leaflist);
    Int_t nLeaf = sleaflist.CountChar(':') + 1;

    const char* standardLea_fName = Form("%s/D", name);   
    if(sleaflist=="vector<double>")
    {
	_mTypeNb["V"] = _mTypeNb["V"]+1;
	_vfVArgs.push_back( new std::vector<Double_t>() );
	_vfArgs.push_back( new UEntry(UEntry::Vector, name, _vfVArgs.back()) );
    }
    else if(sleaflist=="string")
    {
	_mTypeNb["S"] = _mTypeNb["S"]+1;
	_vfSArgs.push_back( new std::string() );
	_vfArgs.push_back( new UEntry(UEntry::String, name, _vfSArgs.back()) );	  
    }
    else
    {	  	 
	for(int i=0; i<nLeaf; i++)
	    _mVarTypeNb.insert( std::pair<unsigned int, unsigned int>(_fNvar+i,_vfDArgs.size()+i) );
	_vfDArgs.clear();
	_vfDArgs.resize(_mTypeNb["D"]+nLeaf, 0.);
	_vfArgs.push_back( new UEntry(UEntry::Double, name, &_vfDArgs.at(_mTypeNb["D"]) ) );
	_mTypeNb["D"] = _mTypeNb["D"]+nLeaf;

	for(_i=0; _i< _vfArgs.size(); _i++)
	    if( _vfArgs.at(_i)->tag == URANIE::DataServer::UEntry::Double )
	    {
		//XYZ<<"  Memory mapping of entry "<<_i<<" index "<< _mVarTypeNb[_i] << endl;
		_vfArgs.at(_i)->setAddress( (void*)( &_vfDArgs[ _mVarTypeNb[_i] ] ) );
	    }
    }
      
    _fNvar++;
    for (_i = 0; _i < UInt_t(fBranches.GetEntries()); _i++)
    {
	TBranch* bObject = (TBranch*) fBranches.At(_i);
	std::string theclass = std::string(bObject->GetClassName());
	if( sleaflist=="vector<double>" && theclass=="vector<double>")
	    bObject->SetAddress(&_vfArgs[_i]->val.vec);
	else if( sleaflist=="string" && theclass=="string")
	    bObject->SetAddress(&_vfArgs[_i]->val.str);
	else if( sleaflist!="vector<double>" && sleaflist!="string" && theclass=="" )
	    bObject->SetAddress(_vfArgs[_i]->val.dou);
    }
    
    if (sleaflist == "")
	return TTree::Branch(name, ( (address==0) ? _vfArgs[_fNvar-1]->val.dou : address), standardLea_fName, bufsize);
    else if( sleaflist=="vector<double>" )
	return TTree::Branch(name, leaflist, ( (address==0) ? &_vfArgs[_fNvar-1]->val.vec : address), bufsize);
    else if( sleaflist=="string" )
	return TTree::Branch(name,  leaflist, ( (address==0) ? &_vfArgs[_fNvar-1]->val.str : address));
    else
	return TTree::Branch(name, ( (address==0) ? _vfArgs[_fNvar-1]->val.dou : address), leaflist, bufsize);

}

//
// ========== Get data / extract data / readfiles ==============
//

Double_t* TDSNtupleD::GetArgs() 
{
    return _vfArgs[0]->val.dou;
}

  
Double_t* TDSNtupleD::getBranchData(const char* branchName, const char* selection, Int_t* nbDataPtr)
{
    
    try
    {
	if (!GetBranch(branchName))
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form(
							   "TDSNtupleD::getBranchData: Branch name [%s] doesn't exist in the TDSNtupleD!",
							   branchName));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
	ue.printMessage();
	throw ue;
    }

    //Test whether this branch is one that contains regular double
    std::string thetype = std::string( GetBranch(branchName)->GetClassName() );
    try
    {
	if ( thetype=="vector<double>" || thetype == "string"  )
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form(
							   "TDSNtupleD::getBranchData: Branch name [%s] exist but contains %s",
							   branchName,thetype.c_str()));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
	ue.printMessage();
	throw ue;
    }

    Long64_t n =  Draw(branchName, selection, "goff");
    if( n > 1000000)
    {
        SetEstimate(n);
        n = Draw(branchName, selection, "goff");

    }   
    if (nbDataPtr)
	*nbDataPtr = n;

    return GetV1();
}

void TDSNtupleD::copyBranchData(Double_t* copyArray, Int_t arraySize, const char* branchName, const char* selection, Int_t* nbDataPtr)
{

    try
    {
	if (!GetBranch(branchName))
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form(
							   "TDSNtupleD::copyBranchData: Branch name [%s] doesn't exist in the TDSNtupleD!",
							   branchName));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
	ue.printMessage();
	throw ue;
    }

    //Test whether this branch is one that contains regular double
    std::string thetype = std::string( GetBranch(branchName)->GetClassName() );
    try
    {
	if ( thetype=="vector<double>" || thetype == "string"  )
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form(
							   "TDSNtupleD::copyBranchData: Branch name [%s] exist but contains %s",
							   branchName,thetype.c_str()));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
	ue.printMessage();
	throw ue;
    }
    
    Long64_t totalData =  Draw(branchName, selection, "goff");
    if (nbDataPtr)
	*nbDataPtr = totalData;

    try
    {
	if (totalData > arraySize)
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form(
							   "TDSNtupleD::copyBranchData: the provided array is too small to contain the selected data: array size [%d], selected data [%lld]",
							   arraySize, totalData));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
	ue.printMessage();
	throw ue;
    }

    if(totalData>1000000)
    {
        SetEstimate(totalData);
        totalData = Draw(branchName, selection, "goff");

    }   
    Double_t *data = GetV1();
    for (int id = 0; id < totalData; id++)
	copyArray[id] = data[id];
}

void TDSNtupleD::extractData(Double_t* dataOut, Int_t arraySize,
			     const char* varexpinput, const char* selection, const Option_t* option,
			     Int_t* nbDataPtr)
{
    TString sOption = TString(option);
    sOption.ToLower();

    // Construction of the event list
    if( fDirectory != gROOT ) fDirectory=gROOT;
    Int_t nbEvents = (Int_t) Draw(">>evList", ((strcmp(selection,"")==0) ? "1>0" : selection), "goff");
    TEventList* evList = (TEventList*) gROOT->Get("evList");

    // construction of the list of leaves
    TObjArray* leafList;
    Bool_t leafList_is_new = kFALSE;
    TString varInputs(varexpinput);
    if (varInputs.CompareTo("") == 0 || varInputs.CompareTo("*") == 0)
    {
	//Exit if one of the branch is not a regular double one
	try
	{
	    if ( _fNvar != _mTypeNb["D"] )
		throw URANIE::Exceptions::UErrorExceptions(__FILE__,
							   __LINE__,
							   Form(
							       "TDSNtupleD::extractData(Double_t* ...): Method not available when dealing with a TDSNtupleD that contains %d vector<double> and %d strings",
							       _mTypeNb["V"],_mTypeNb["S"]));
	} catch (URANIE::Exceptions::UErrorExceptions& ue)
	{
	    ue.printMessage();
	    if( gROOT->FindObject("evList") ) gROOT->Delete("evList");
	    throw ue;
	}	
	leafList = GetListOfLeaves();
    }
    else
    {
	TObjArray *varInputsList = varInputs.Tokenize(":");
	leafList = new TObjArray();
	leafList_is_new = kTRUE;
	for (_i = 0; _i < UInt_t(varInputsList->GetEntries()); _i++)
	{
	    const char* varName = (varInputsList->At(_i))->GetName();
	    TLeaf * l = GetLeaf(varName);
	    try
	    {
		if (!l)
		    throw URANIE::Exceptions::UErrorExceptions(__FILE__,
							       __LINE__,
							       Form(
								   "TDSNtupleD::extractData: Branch name [%s] doesn't exist in the TDSNtupleD!",
								   varName));
	    } catch (URANIE::Exceptions::UErrorExceptions& ue)
	    {
		ue.printMessage();
		if( gROOT->FindObject("evList") ) gROOT->Delete("evList");
		if(leafList_is_new)
		    delete leafList;
		leafList=NULL;
		delete varInputsList;
		throw ue;
	    }
	    std::string theclass = GetBranch((varInputsList->At(_i))->GetName())->GetClassName();
	    try
	    {
		if ( theclass=="vector<double>" || theclass=="string" )
		    throw URANIE::Exceptions::UErrorExceptions(__FILE__, 
							       __LINE__, 
							       Form( 
								   "TDSNtupleD::extractData(Double_t* ...): Method not available when asking to extract %s which is a %s branch",
								   varName,theclass.c_str()));
	    } catch (URANIE::Exceptions::UErrorExceptions& ue)
	    {
		ue.printMessage();
		if( gROOT->FindObject("evList") ) gROOT->Delete("evList");
		delete varInputsList;
		if (leafList_is_new)
		    delete leafList;
		leafList = NULL;
		throw ue;
	    }
	    leafList->Add(l);
	}
	delete varInputsList;
    }
    Int_t nbLeaves = leafList->GetEntries(); // total number of useful leaves

    // Check if the size of the provided array is enough to store the data.
    Int_t totalData = nbEvents * nbLeaves;
    try
    {
	if (totalData > arraySize)
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form(
							   "TDSNtupleD::extractData: the provided array is too small to contain the selected data: array size [%d], selected data [%d]",
							   arraySize, totalData));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
	ue.printMessage();
	delete evList;//if( gROOT->FindObject("evList") ) gROOT->Delete("evList");
	if (leafList_is_new)
	    delete leafList;
	leafList = NULL;
	throw ue;
    }
    if (nbDataPtr)
	*nbDataPtr = totalData;

    // Construction of the array that will store the data.            
    Double_t *values = new Double_t[totalData];

    for (Int_t ev = 0; ev < nbEvents; ev++)
    {
	GetEntry(evList->GetEntry(ev));
	for (Int_t l = 0; l < nbLeaves; l++)
	    values[ev * nbLeaves + l] = ((TLeaf*) leafList->At(l))->GetValue();
    }
    // If the list of leaves was created in this function, we delete it.
    if (leafList_is_new)
	delete leafList;
    leafList = NULL;

    // Store the collected data into dataOut, in the correct order (hopefully...)
    if (sOption == "column")
    {
	for (Int_t sEv = 0; sEv < nbEvents; sEv++)
	{
	    for (Int_t l = 0; l < nbLeaves; l++)
		dataOut[sEv + l * nbEvents] = values[sEv * nbLeaves + l];
	}
    }
    else
    {
	for (Int_t sEv = 0; sEv < nbEvents; sEv++)
	{
	    for (Int_t l = 0; l < nbLeaves; l++)
		dataOut[sEv * nbLeaves + l] = values[sEv * nbLeaves + l];
	}
    }
    delete[] values;
    values = NULL;
    if( gROOT->FindObject("evList") ) gROOT->Delete("evList");
}

void TDSNtupleD::extractData(TMatrixD* outputMatrix, const char* varexpinput,
			     const char* selection)
{
  
    // Construction of the event list
    if( fDirectory != gROOT ) fDirectory=gROOT;
    Int_t nbEvents = (Int_t) Draw(">>evList", ((strcmp(selection,"")==0) ? "1>0" : selection), "goff");
    TEventList* evList = (TEventList*) gROOT->Get("evList");

    // construction of the list of leaves
    TObjArray* leafList;
    Bool_t leafList_is_new = kFALSE;
    TString varInputs(varexpinput);
    if (varInputs.CompareTo("") == 0 || varInputs.CompareTo("*") == 0)
    {
	//if one the branch is not filled with regular double
	try
	{
	    if ( _fNvar != _mTypeNb["D"] )
		throw URANIE::Exceptions::UErrorExceptions(__FILE__,
							   __LINE__,
							   Form(
							       "TDSNtupleD::extractData(TMatrixD* ...): Method not available when dealing with a TDSNtupleD that contains %d vector<double> and %d strings",
							       _mTypeNb["V"],_mTypeNb["S"]));
	} catch (URANIE::Exceptions::UErrorExceptions& ue)
	{
	    ue.printMessage();
	    if( gROOT->FindObject("evList") ) gROOT->Delete("evList");
	    throw ue;
	}
	leafList = GetListOfLeaves();
    }
    else
    {
	TObjArray *varInputsList = varInputs.Tokenize(":");
	leafList = new TObjArray();
	leafList_is_new = kTRUE;
	for (_i = 0; _i < UInt_t(varInputsList->GetEntries()); _i++)
	{
	    const char* varName = (varInputsList->At(_i))->GetName();
	    TLeaf * l = GetLeaf(varName);
	    try
	    {
		if (!l)
		    throw URANIE::Exceptions::UErrorExceptions(__FILE__,
							       __LINE__,
							       Form(
								   "TDSNtupleD::extractData: Branch name [%s] doesn't exist in the TDSNtupleD!",
								   varName));
	    } catch (URANIE::Exceptions::UErrorExceptions& ue)
	    {
		ue.printMessage();
		delete varInputsList;
		if (leafList_is_new)
		    delete leafList;
		leafList = NULL;
		if( gROOT->FindObject("evList") ) gROOT->Delete("evList");
		throw ue;
	    }
	    std::string theclass = GetBranch((varInputsList->At(_i))->GetName())->GetClassName();
	    try
	    {
		if ( theclass=="vector<double>" || theclass=="string" )
		    throw URANIE::Exceptions::UErrorExceptions(__FILE__, 
							       __LINE__, 
							       Form( 
								   "TDSNtupleD::extractData(TMatrix* ...): Method not available when asking to extract %s which is a %s branch",
								   varName,theclass.c_str()));
	    } catch (URANIE::Exceptions::UErrorExceptions& ue)
	    {
		ue.printMessage();
		delete varInputsList;
		if (leafList_is_new)
		    delete leafList;
		leafList = NULL;
		if( gROOT->FindObject("evList") ) gROOT->Delete("evList");
		throw ue;
	    }
	    leafList->Add(l);
	}
	delete varInputsList;
    }
    Int_t nbLeaves = leafList->GetEntries(); // total number of useful leaves

    // Initialise the matrix that will store the data.
    outputMatrix->ResizeTo(nbEvents, nbLeaves);
    outputMatrix->Zero();

    // Fill the matrix
    for (Int_t ev = 0; ev < nbEvents; ev++)
    {
	GetEntry(evList->GetEntry(ev));
	for (Int_t l = 0; l < nbLeaves; l++)
	    (*outputMatrix)(ev, l) = ((TLeaf*) leafList->At(l))->GetValue();
    }

    // If the list of leaves was created in this function, we delete it.
    if (leafList_is_new)
	delete leafList;
    leafList = NULL;

    if( gROOT->FindObject("evList") ) gROOT->Delete("evList");

}

Long64_t TDSNtupleD::ReadFile(const char* filename)
{
    Long64_t nlines = 0;

    std::ifstream file(filename, ios::in);
    std::string line;
    while ( !UEntry::safeGetline(file, line).eof() )
    {
	if (!(line.length() > 0) || UEntry::isWhiteLine(line))
	    break;

	std::stringstream sstr; sstr << nlines << " ";
	if(nlines==0)
	{
	    if( std::string(GetName())=="" && std::string(GetTitle())=="" )
		_initTTree("__no_name_TDSNtupleD__", "__no_title_TDSNtupleD__");
	    if(_varname=="")
	    {
		std::vector<std::string> nbEntree;
		UEntry::StringParser(line, " ", nbEntree);
		_varname="__no_name_tdsn_iter__";
		for(_i=0; _i<nbEntree.size(); _i++)
		{
		    std::stringstream sstr2; sstr2 << "col" <<  _i;
		    _varname+=":"+sstr2.str();
		}
		_branchConstruction(32000);
	    }
	}
	  
	Fill( sstr.str()+line );
	nlines += 1.0;
    }

    file.close();
    return nlines;
}


  
//ToBeCorrected - JBB
void TDSNtupleD::Merger(TDSNtupleD *file, const std::string newbrname, const std::string oldbrname, bool tdsmerge, const char* iter1, const char* iter2)
{

    UInt_t ibr;
    UInt_t Old_fNvar=_fNvar;

    try
    {
	if(GetEntries() != file->GetEntries())
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form("TDSNtupleD::Merger. The TDSNtuples don't have the same number of entries (%lld and %lld)",
							    GetEntries(), file->GetEntries() ));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
	ue.printMessage();
	throw ue;
    }

    //Get the name of the input to be read and their type
    std::vector<std::string> v_oldbrname, v_newbrname, v_type;
    UEntry::StringParser(newbrname, ":", v_newbrname);
    UEntry::StringParser(oldbrname, ":", v_oldbrname);
    for(ibr=0; ibr<v_newbrname.size();ibr++)
    {
	std::string theclass = std::string(file->GetBranch( v_oldbrname.at(ibr).c_str() )->GetClassName());
	Branch( v_newbrname.at(ibr).c_str(), 0, theclass.c_str());	
    }

    // check whether the iterator are actually leaves of the ntuples
    bool iterfounds=false;
    if( GetLeaf(iter1) && file->GetLeaf(iter2) )
	iterfounds=true;
    //Create a map 
    map<unsigned int,unsigned int> EntryNBIterNB1; // IterValue of this wrt entry value
    map<unsigned int,unsigned int> IterNBEntryNB2; // Entry value of file wrt its iter value
    if( iterfounds )
    {
	for (_i = 0; _i < GetEntries(); _i++)
	{
	    file->GetEntry(_i); GetEntry(_i);
	    EntryNBIterNB1.insert( std::pair<unsigned int, unsigned int>(_i, GetLeaf(iter1)->GetValue() ));
	    IterNBEntryNB2.insert( std::pair<unsigned int, unsigned int>(file->GetLeaf(iter2)->GetValue(),_i) );
	}
	try
	{
	    for (_i = 0; _i < GetEntries(); _i++)
		if( IterNBEntryNB2.count( EntryNBIterNB1[_i] ) == 0 )
		{
		    iterfounds=false;
		    EntryNBIterNB1.clear();
		    IterNBEntryNB2.clear();
		    throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
								 Form("TDSNtupleD::Merger WARNING : iterator names are provided but value %u from %s is not found in %s \n ==> Will concatenate the ntuples as they come!! (might be normal in some cases: TOAT sampling for instance)",EntryNBIterNB1[_i],iter1,iter2));
		}
	}catch(URANIE::Exceptions::UWarningExceptions &ue)
	{
	    ue.printMessage();
	}
    }

    //Loop over event to fill the new tree by copy
    TLeaf* lObject;
    std::vector<double> buffer;
    for (_i = 0; _i < GetEntries(); _i++)
    {
	GetEntry( _i );
	file->GetEntry( ((iterfounds) ? IterNBEntryNB2[EntryNBIterNB1[_i]] : _i) ); 
	for (ibr=Old_fNvar; ibr<_fNvar; ibr++)
	{
	    std::string theclass = std::string(file->GetBranch( v_oldbrname.at(ibr-Old_fNvar).c_str() )->GetClassName());
	    lObject = file->GetBranch( v_oldbrname.at(ibr-Old_fNvar).c_str() )->GetLeaf( ((tdsmerge) ? v_newbrname.at(ibr-Old_fNvar).c_str() : v_oldbrname.at(ibr-Old_fNvar).c_str()) );
	    try
	    {
		if(!lObject)
		    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							       Form(
								   "TDSNtupleD::Merger. The Leaf [%s] is not found in the ntuple %s",
								   v_newbrname.at(ibr-Old_fNvar).c_str(), file->GetName()));
	    } catch (URANIE::Exceptions::UErrorExceptions& ue)
	    {
		ue.printMessage();
		throw ue;
	    }

	    _vfArgs[ibr]->assignEntry( (void*)(lObject->GetValuePointer()) );
	    // if(_vfArgs[ibr]->tag==UEntry::Double) _vfArgs[ibr]->val.dou = lObject->GetValue();
	    // if(_vfArgs[ibr]->tag==UEntry::String) *_vfArgs[ibr]->val.str = *(string*)(lObject->GetValuePointer());
	    // if(_vfArgs[ibr]->tag==UEntry::Vector) 
	    // {
	    // 	  _vfArgs[ibr]->val.vec->clear();
	    // 	  buffer = *(vector<double>*)(lObject->GetValuePointer());
	    // 	  for(unsigned int ival=0; ival<buffer.size(); ival++)
	    // 	      _vfArgs[ibr]->val.vec->push_back( buffer.at(ival) );
	    // }
	      
	    GetBranch( v_newbrname.at(ibr-Old_fNvar).c_str() )->Fill();
	}	
    }
}

void TDSNtupleD::Browse(TBrowser *b)
{
    // Browse content.

    fLeaves.Browse( b );
}

  
}
}

