//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TExponentialDistribution.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Thu Jan 18 17:08:58 CET 2007

 \brief Implementation of the class URANIE::DataServer::TExponentialDistribution

 */
#include <iostream>
using std::cout;
using std::endl;

#include "TMath.h"
#include "TExponentialDistribution.h"

ClassImp (URANIE::DataServer::TExponentialDistribution)

namespace URANIE
{
namespace DataServer
{
// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------

TExponentialDistribution::TExponentialDistribution() :
        TInfiniteDistribution("", TStochasticAttribute::kExponential), _dparameterLambda(
                0.1), _dparameterMode(0.0)
{
    if (_blog)
        cout << "Constructing TExponentialDistribution" << endl;

    if (lowerBound < _dparameterMode)
        TAttribute::setLowerBound(_dparameterMode);
    computeTheoreticalInformation();
}

TExponentialDistribution::TExponentialDistribution(TString str) :
        TInfiniteDistribution(str, TStochasticAttribute::kExponential), _dparameterLambda(
                0.1), _dparameterMode(0.0)
{
    if (_blog)
        cout
                << "Constructing URANIE::DataServer::TExponentialDistribution with: Name["
                << str << "]" << endl;

    if (lowerBound < _dparameterMode)
        TAttribute::setLowerBound(_dparameterMode);
    computeTheoreticalInformation();
}

TExponentialDistribution::TExponentialDistribution(TString str, Double_t lambda,
        Double_t mode) :
        TInfiniteDistribution(str, TStochasticAttribute::kExponential)
{
    if (_blog)
        cout
                << "Constructing URANIE::DataServer::TExponentialDistribution with: Name["
                << str << "] lambda[" << lambda << "] mode[" << mode << "]"
                << endl;
    setParameters(lambda, mode);
}

TExponentialDistribution::TExponentialDistribution(
        URANIE::DataServer::TAttribute *att, Double_t lambda, Double_t mode) :
        TInfiniteDistribution(att, TStochasticAttribute::kExponential)
{
    if (_blog)
        cout
                << "Constructing URANIE::DataServer::TExponentialDistribution  from URANIE::DataServer::TAttribute object "
                << att->GetName() << endl;
    setParameters(lambda, mode);
}

// -----------------------------------------------------------------------
//  Destructors
// -----------------------------------------------------------------------       

TExponentialDistribution::~TExponentialDistribution()
{
    if (_blog)
        cout
                << "Destroying URANIE::DataServer::TExponentialDistribution object "
                << GetName() << endl;

}

// -----------------------------------------------------------------------
//  Parameters
// -----------------------------------------------------------------------        

void TExponentialDistribution::setParameterLambda(Double_t lambda,
        Bool_t recompute)
{
    try
    {
        if (lambda < 1e-4)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TExponentialDistribution::setParameterLambda: parameter lambda [%f] is too close to zero: lower limit is 1e-4",
                            lambda));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    _dparameterLambda = lambda;

    if (recompute)
        computeTheoreticalInformation();
}

void TExponentialDistribution::setParameterMode(Double_t mode, Bool_t recompute)
{
    _dparameterMode = mode;

    if (lowerBound < mode)
        TAttribute::setLowerBound(mode);

    if (recompute)
        computeTheoreticalInformation();
}

void TExponentialDistribution::setParameters(Double_t lambda, Double_t mode,
        Bool_t recompute)
{
    setParameterLambda(lambda, kFALSE);
    setParameterMode(mode, kFALSE);
    if (recompute)
        computeTheoreticalInformation();
}

// -----------------------------------------------------------------------
//  Theoretical Information
// -----------------------------------------------------------------------        

void TExponentialDistribution::computeTheoreticalInformation()
{
    _dtheoreticalMean = _dparameterMode + 1.0 / _dparameterLambda;

    _dtheoreticalStdDev = 1.0 / _dparameterLambda;

    _dtheoreticalMode = _dparameterMode;
}

// -----------------------------------------------------------------------
//  Distribution
// -----------------------------------------------------------------------        

Double_t TExponentialDistribution::getPDF(Double_t x)
{
    Double_t pdf = 0.0;
    if ((x >= lowerBound) && (x <= upperBound))
    {
        pdf = _dparameterLambda
                * TMath::Exp(-1.0 * _dparameterLambda * (x - _dparameterMode));
        // Take into account the case where at least one of the bounds is finite
        pdf /= _pmax - _pmin;
    }
    return pdf;
}

Double_t TExponentialDistribution::getCDF(Double_t x)
{
    Double_t cdf = 0.0;
    if ((x >= lowerBound) && (x <= upperBound))
    {
        cdf = 1 - TMath::Exp(-1.0 * _dparameterLambda * (x - _dparameterMode));
        // Take into account the case where at least one of the bounds is finite
        //cdf /= _pmax - _pmin;
        cdf = (cdf - _pmin) / (_pmax - _pmin);
    }
    else if (x > upperBound)
        cdf = 1.0;

    return cdf;
}

Double_t TExponentialDistribution::getInvCDF(Double_t p)
{
    // Verify that p is a valid probability
    try
    {
        if ((p < 0.0) || (p > 1.0))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TExponentialDistribution::getInvCDF: parameter p [%f] is not a valid probability: p must be in the interval [0,1]",
                            p));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }
    // Take into account the case where at least one of the bounds is finite            
    Double_t realP = p * (_pmax - _pmin) + _pmin;

    Double_t x = _dparameterMode - TMath::Log(1.0 - realP) / _dparameterLambda;

    return x;
}

// -----------------------------------------------------------------------
//  Log
// -----------------------------------------------------------------------

void TExponentialDistribution::printLog(Option_t *option)
{
    TString soption = option;
    soption.ToLower();
    TInfiniteDistribution::printLog();
    cout << "*******************************" << endl;
    cout << "** TExponentialDistribution::printLog  ******" << endl;
    cout << "**   Law[" << _lawType << "] Exponential" << endl;
    cout << "**   Bounds[" << lowerBound << ", " << upperBound << "]" << endl;
    cout << "**   Rate parameter lambda[" << _dparameterLambda << "]" << endl;
    cout << "**   Mode parameter xmin[" << _dparameterMode << "]" << endl;
    cout << "**   Theoretical Mean[" << _dtheoreticalMean << "]" << endl;
    cout << "**   Theoretical StdDev[" << _dtheoreticalStdDev << "]" << endl;
    cout << "**   Theoretical Mode[" << _dtheoreticalMode << "]" << endl;
    cout << "*******************************" << endl;
}

} // Fin du namespace DataServer
} // Fin du namespace URANIE
