//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TGenParetoDistribution.cxx
 \author Roman NIKOLAEV <roman.nikolaev@opencascade.com>
 \date   Tue Feb 19 12:00:00 CET 2013

 \brief Implementation of the class URANIE::DataServer::TGenParetoDistribution

 */
#include <iostream>
using std::cout;
using std::endl;

#include "TMath.h"
#include "TGenParetoDistribution.h"
#include "TLeaf.h"
#include "TMap.h"

ClassImp (URANIE::DataServer::TGenParetoDistribution)

namespace URANIE
{
namespace DataServer
{

// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------
TGenParetoDistribution::TGenParetoDistribution() :
        TInfiniteDistribution("", TStochasticAttribute::kGenPareto), _dparameterXi(0.25), _dparameterMu(0.0), _dparameterSigma(1.0)
{
    if (_blog)
        cout << "Constructing TGenParetoDistribution" << endl;
    setLowerBound(_dparameterMu);
    computeTheoreticalInformation();
}

TGenParetoDistribution::TGenParetoDistribution(TString name) :
        TInfiniteDistribution(name, TStochasticAttribute::kGenPareto), _dparameterXi(0.25), _dparameterMu(0.0), _dparameterSigma(1.0)
{
    if (_blog)
        cout
                << "Constructing URANIE::DataServer::TGenParetoDistribution with: Name["
                << name << "]" << endl;
    setLowerBound(_dparameterMu);
    computeTheoreticalInformation();
}

TGenParetoDistribution::TGenParetoDistribution(TString name, Double_t mu,
        Double_t sigma, Double_t xi) :
        TInfiniteDistribution(name, TStochasticAttribute::kGenPareto)
{
    if (_blog)
        cout
                << "Constructing URANIE::DataServer::TGenParetoDistribution with: Name["
                << name << "] Mu [ " << mu << "] " << "Sigma [ " << sigma
                << "] " << "Xi [" << xi << "] " << endl;
    setParameters(mu, sigma, xi);
    setLowerBound(_dparameterMu);
}

TGenParetoDistribution::TGenParetoDistribution(
        URANIE::DataServer::TAttribute *att, Double_t mu, Double_t sigma,
        Double_t xi) :
        TInfiniteDistribution(att, TStochasticAttribute::kGenPareto)
{
    if (_blog)
        cout
                << "Constructing URANIE::DataServer::TGenParetoDistribution with: Name["
                << att->GetName() << "] Mu [ " << mu << "] " << "Sigma [ "
                << sigma << "] " << "Xi [" << xi << "] " << endl;
    setParameters(mu, sigma, xi);
    setLowerBound(_dparameterMu);
}

// -----------------------------------------------------------------------
//  Destructor
// -----------------------------------------------------------------------

TGenParetoDistribution::~TGenParetoDistribution()
{
    if (_blog)
        cout << "Destroying URANIE::DataServer::TGenParetoDistribution "
                << GetName() << endl;

}
// -----------------------------------------------------------------------
//  Parameters
// -----------------------------------------------------------------------
void TGenParetoDistribution::setParameterMu(Double_t mu, Bool_t recompute)
{
    _dparameterMu = mu;
    setLowerBound(_dparameterMu);
    if (recompute)
        computeTheoreticalInformation();
}

void TGenParetoDistribution::setParameterSigma(Double_t sigma, Bool_t recompute)
{
    try
    {
        if (sigma <= 0.0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TGenParetoDistribution::setParameterSigma: parameter sigma is less equla to zero: sigma should be > 0"));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    _dparameterSigma = sigma;
    if (recompute)
        computeTheoreticalInformation();
}

void TGenParetoDistribution::setParameterXi(Double_t xi, Bool_t recompute)
{
    _dparameterXi = xi;
    if (recompute)
        computeTheoreticalInformation();
}

void TGenParetoDistribution::setParameters(Double_t mu, Double_t sigma,
        Double_t xi, Bool_t recompute)
{

    setParameterSigma(sigma, kFALSE);
    setParameterXi(xi, kFALSE);
    setParameterMu(mu, kFALSE);
    if (recompute)
        computeTheoreticalInformation();
}

Double_t TGenParetoDistribution::getParameterMu()
{
    return _dparameterMu;
}

Double_t TGenParetoDistribution::getParameterSigma()
{
    return _dparameterSigma;
}

Double_t TGenParetoDistribution::getParameterXi()
{
    return _dparameterXi;

}

// -----------------------------------------------------------------------
//  Distribution
// -----------------------------------------------------------------------        

Double_t TGenParetoDistribution::getPDF(Double_t x)
{
    Double_t pdf = 0.;
    if ((x > lowerBound) && (x < upperBound))
    {
        pdf = TMath::Power(1 + _dparameterXi * getZ(x),
                -(1. / _dparameterXi + 1.)) / _dparameterSigma;
	pdf /= ( ((_pmax-_pmin)!=0) ? (_pmax-_pmin) : 1);
    }
    return pdf;
}

Double_t TGenParetoDistribution::getCDF(Double_t x)
{
    Double_t cdf;
    if (x < lowerBound)
        cdf = 0.0;
    else if (x > upperBound)
        cdf = 1.0;
    else
    {
        cdf = 1 - TMath::Power(1 + _dparameterXi * getZ(x),
			       -1. / _dparameterXi);

	cdf = (cdf - _pmin) / (_pmax - _pmin) ;
    }
    return cdf;
}
/*-------------------------------------------------------------------------*/

Double_t TGenParetoDistribution::getInvCDF(Double_t p)
{
    // Verify that p is a valid probability
    try
    {
        if ((p < 0.0) || (p > 1.0))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TGenParetoDistribution::getInvCDF: parameter p [%f] is not a valid probability: p must be in the interval [0,1]",
                            p));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    Double_t realP = p * (_pmax - _pmin) + _pmin;
    
    Double_t x = ((TMath::Power(1 - realP, -_dparameterXi) - 1) * _dparameterSigma)
            / _dparameterXi + _dparameterMu;

    return x;
}
/*-------------------------------------------------------------------------*/

void TGenParetoDistribution::computeTheoreticalInformation()
{

    _dtheoreticalMean = ((_dparameterXi>=1) ? THEO_DEF : _dparameterMu + _dparameterSigma / (1 - _dparameterXi)) ;

    _dtheoreticalStdDev = ((_dparameterXi>= 0.5) ? THEO_DEF :
                           _dparameterSigma * _dparameterSigma / ( (1 - _dparameterXi)*(1 - _dparameterXi)*(1 - 2*_dparameterXi)) ) ;

    _dtheoreticalMode = _dparameterMu;
    
}

/*-------------------------------------------------------------------------*/

/*-------------------------------------------------------------------------*/
Double_t TGenParetoDistribution::getXiEquation(Double_t x, Double_t skewness)
{
    return (2 * (1. + x) * TMath::Sqrt(1 - 2. * x) / (1 - 3. * x) - skewness);
}

/*-------------------------------------------------------------------------*/
Double_t TGenParetoDistribution::getZ(Double_t x)
{
    return (x - _dparameterMu) / _dparameterSigma;
}

/*-------------------------------------------------------------------------*/

void TGenParetoDistribution::printLog(Option_t *option)
{
    TString soption = option;
    soption.ToLower();
    TInfiniteDistribution::printLog();
    cout << "*******************************" << endl;
    cout << "** TGenParetoDistribution::printLog  ******" << endl;
    cout << "**   Law[" << _lawType << "] TGenParetoDistribution" << endl;
    cout << "**   Bounds[" << lowerBound << ", " << upperBound << "]" << endl;
    cout << "**   Location mu[" << _dparameterMu << "]" << endl;
    cout << "**   Scale sigma[" << _dparameterSigma << "]" << endl;
    cout << "**   Shape  Xi[" << _dparameterXi << "]" << endl;
    cout << "*******************************" << endl;
}

} // Fin du namespace DataServer
} // Fin du namespace URANIE
