//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TGumbelMaxDistribution.cxx
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 \date   Mon Jan 19 12:00:00 CET 2009

 \brief Implementation of the class URANIE::DataServer::TGumbelMaxDistribution

 totot
 */
#include <iostream>
using std::cout;
using std::endl;

#include "TMath.h"
#include "TGumbelMaxDistribution.h"

ClassImp (URANIE::DataServer::TGumbelMaxDistribution)

namespace URANIE
{
namespace DataServer
{

// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------

TGumbelMaxDistribution::TGumbelMaxDistribution() :
        TInfiniteDistribution("", TStochasticAttribute::kGumbel), _dparameterMode(
                0.0), _dparameterScale(1.0)
{
    //  cout << "build TGumbelMaxDistribution" << endl;
    computeTheoreticalInformation();
}

TGumbelMaxDistribution::TGumbelMaxDistribution(TString str) :
        TInfiniteDistribution(str, TStochasticAttribute::kGumbel), _dparameterMode(
                0.0), _dparameterScale(1.0)
{
    //  cout << "Constructor of URANIE::DataServer::TGumbelMaxDistribution[" << this << "]str[" << str<<"]"<<endl;
    computeTheoreticalInformation();
}

TGumbelMaxDistribution::TGumbelMaxDistribution(TString str, Double_t mu,
        Double_t beta) :
        TInfiniteDistribution(str, TStochasticAttribute::kGumbel)
{
    //  cout << "Constructeur URANIE::DataServer::TGumbelMaxDistribution[" << this << "]str[" << str<<"] Location [" << mu <<"] Scale [" << beta << "]" << endl;
    setParameters(mu, beta);
}

TGumbelMaxDistribution::TGumbelMaxDistribution(
        URANIE::DataServer::TAttribute *att, Double_t mu, Double_t beta) :
        TInfiniteDistribution(att, TStochasticAttribute::kGumbel)
{
    //  cout << "Constructor of URANIE::DataServer::TGumbelMaxDistribution att[" << this << "]att->getName()[" << att->getName()<<"] Location [" << mu <<"] Scale [" << beta << "]" << endl;
    setParameters(mu, beta);
}

// -----------------------------------------------------------------------
//  Destructors
// -----------------------------------------------------------------------

TGumbelMaxDistribution::~TGumbelMaxDistribution()
{
    //  cout << "detruit TGumbelMaxDistribution" << endl; 

}

// -----------------------------------------------------------------------
//  Parameters
// -----------------------------------------------------------------------

void TGumbelMaxDistribution::setParameterMode(Double_t mu, Bool_t recompute)
{
    _dparameterMode = mu;
    if (recompute)
        computeTheoreticalInformation();
}

void TGumbelMaxDistribution::setParameterScale(Double_t beta, Bool_t recompute)
{
    try
    {
        if (beta < 1e-6 * _dparameterMode)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TGumbelMaxDistribution::setParameterScale: parameter beta [%f] is too close to zero: lower limit is 1e-6 * mu [%f]",
                            beta, 1e-6 * _dparameterMode));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    _dparameterScale = beta;
    if (recompute)
        computeTheoreticalInformation();
}

void TGumbelMaxDistribution::setParameters(Double_t mu, Double_t beta,
        Bool_t recompute)
{
    setParameterMode(mu, kFALSE);
    setParameterScale(beta, kFALSE);

    if (recompute)
        computeTheoreticalInformation();
}

// -----------------------------------------------------------------------
//  Theoretical Information
// -----------------------------------------------------------------------

void TGumbelMaxDistribution::computeTheoreticalInformation()
{
    _dtheoreticalMean = _dparameterMode
            + _dparameterScale * TMath::EulerGamma();

    _dtheoreticalStdDev = _dparameterScale * TMath::Pi() / TMath::Sqrt(6.0);

    _dtheoreticalMode = _dparameterMode;
}

// -----------------------------------------------------------------------
//  Distribution
// -----------------------------------------------------------------------

Double_t TGumbelMaxDistribution::getPDF(Double_t x)
{
    Double_t pdf = 0.0;

    if ((x >= lowerBound) && (x <= upperBound))
    {
        Double_t z = TMath::Exp(-(x - _dparameterMode) / _dparameterScale);

        pdf = z * TMath::Exp(-z) / _dparameterScale;
        // Take into account the case where the law is truncated
        pdf /= _pmax - _pmin;
    }
    return pdf;
}

Double_t TGumbelMaxDistribution::getCDF(Double_t x)
{
    Double_t cdf = 0.0;

    if ((x >= lowerBound) && (x <= upperBound))
    {
        Double_t z = TMath::Exp(-(x - _dparameterMode) / _dparameterScale);

        cdf = TMath::Exp(-z);
        // Take into account the case where the law is truncated
        //cdf /= _pmax - _pmin; 
        cdf = (cdf - _pmin) / (_pmax - _pmin);
    }
    else if (x > upperBound)
        cdf = 1.0;

    return cdf;
}

Double_t TGumbelMaxDistribution::getInvCDF(Double_t p)
{
    // Verify that p is a valid probability
    try
    {
        if ((p < 0.0) || (p > 1.0))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TGumbelMaxDistribution::getInvCDF: parameter p [%f] is not a valid probability: p must be in the interval [0,1]",
                            p));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    // Take into account the case where the law is truncated
    Double_t realP = p * (_pmax - _pmin) + _pmin;

    Double_t x = _dparameterMode
            - _dparameterScale * TMath::Log(-TMath::Log(realP));

    return x;
}

// -----------------------------------------------------------------------
//  Log
// -----------------------------------------------------------------------

void TGumbelMaxDistribution::printLog(Option_t *option)
{
    //  TInfiniteDistribution::printLog();
    TString soption = option;
    soption.ToLower();
    cout << "*******************************" << endl;
    cout << "** TGumbelMaxDistribution::printLog  ******" << endl;
    cout << "**   Law[" << _lawType << "] GumbelMax" << endl;
    cout << "**   Bounds[" << lowerBound << ", " << upperBound << "]" << endl;
    cout << "**   Mode parameter mu[" << _dparameterMode << "]" << endl;
    cout << "**   Scale parameter beta[" << _dparameterScale << "]" << endl;
    cout << "**   Theoretical Mean[" << _dtheoreticalMean << "]" << endl;
    cout << "**   Theoretical StdDev[" << _dtheoreticalStdDev << "]" << endl;
    cout << "**   Theoretical Mode[" << _dtheoreticalMode << "]" << endl;
    cout << "*******************************" << endl;
}

} // Fin du namespace DataServer
} // Fin du namespace URANIE

