//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TInvGammaDistribution.cxx
 \author Guillaume Damblin <guillaume.damblin@cea.fr>
 \date   2009-11-20 10:09:30 \brief Interface of the class URANIE::DataServer::TInvGammaDistribution

 \brief Implementation of the class URANIE::DataServer::TInvGammaDistribution
 */

#include <iostream>
using std::cout;
using std::endl;

#include "TMath.h"
#include "TInvGammaDistribution.h"

ClassImp (URANIE::DataServer::TInvGammaDistribution)

namespace URANIE
{
namespace DataServer
{

// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------

TInvGammaDistribution::TInvGammaDistribution() :
  TInfiniteDistribution("", TStochasticAttribute::kInvGamma), 
	_dparameterAlpha(1.0), _dparameterScale(1.0), _dparameterLocation(0.0)
{

    if (_blog)
        cout << "Constructing TInvGammaDistribution" << endl;

    setLowerBound(_dparameterLocation);
    computeTheoreticalInformation();
}

TInvGammaDistribution::TInvGammaDistribution(TString str) :
        TInfiniteDistribution(str, TStochasticAttribute::kInvGamma),
	_dparameterAlpha(1.0), _dparameterScale(1.0), _dparameterLocation(0.0)
{

    if (_blog)
        cout
            << "Constructing URANIE::DataServer::TInvGammaDistribution with: Name["
            << str << "]" << endl;

    setLowerBound(_dparameterLocation);
    computeTheoreticalInformation();
}

  TInvGammaDistribution::TInvGammaDistribution(TString str,  Double_t alpha, Double_t scale, Double_t location) :
    TInfiniteDistribution(str, TStochasticAttribute::kInvGamma),
    _dparameterAlpha(alpha), _dparameterScale(scale), _dparameterLocation(location)
 
{
    if (_blog)
        cout
            << "Constructing URANIE::DataServer::TInvGammaDistribution with: Name["
            << str << "] alpha[" << alpha << "] scale[" << scale << "]"
            << endl;
  
    setParameterAlpha(alpha, kFALSE);
    setParameterScale(scale, kFALSE);
    setParameterLocation(location, kFALSE);

    setLowerBound(_dparameterLocation);
    computeTheoreticalInformation();
}

TInvGammaDistribution::TInvGammaDistribution(URANIE::DataServer::TAttribute *att, Double_t alpha, Double_t scale, Double_t location) :
  TInfiniteDistribution(att, TStochasticAttribute::kInvGamma),
  _dparameterAlpha(alpha), _dparameterScale(scale), _dparameterLocation(location)
{

    if (_blog)
        cout<< "Constructing URANIE::DataServer::TInvGammaDistribution  from URANIE::DataServer::TAttribute object "
            << att->GetName() << endl;
  
    setParameterAlpha(alpha, kFALSE);
    setParameterScale(scale, kFALSE);
    setParameterLocation(location, kFALSE);

    setLowerBound(_dparameterLocation);
    computeTheoreticalInformation();
}


// -----------------------------------------------------------------------
//  Destructors
// -----------------------------------------------------------------------       

TInvGammaDistribution::~TInvGammaDistribution()
{
  if (_blog)
    cout
      << "Destroying URANIE::DataServer::TInvGammaDistribution object "
      << GetName() << endl;
}

// -----------------------------------------------------------------------
//  Parameters
// -----------------------------------------------------------------------        

void TInvGammaDistribution::setParameterAlpha(Double_t alpha, Bool_t recompute)
{
  
    _dparameterAlpha = alpha;
    try
    {
        if(alpha <= 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form(
                                                           "TInvGammaDistribution::setParameterAlpha: alpha parameter value [%f] should be greater than 0",
                                                           alpha));
    }catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

  
    if (recompute)
        computeTheoreticalInformation();
}


void TInvGammaDistribution::setParameterScale(Double_t scale, Bool_t recompute)
{
  
    _dparameterScale = scale;
    try
    {
        if(scale <= 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form(
                                                           "InvTGammaDistribution::setParameterScale: scale parameter value [%f] should be greater than 0",
                                                           scale));
    }catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }
      
    if (recompute)
        computeTheoreticalInformation();
}


void TInvGammaDistribution::setParameterLocation(Double_t location, Bool_t recompute)
{
  
    _dparameterLocation = location;
        
    if (recompute)
        computeTheoreticalInformation();
}

void TInvGammaDistribution::setParameters( Double_t alpha, Double_t scale, Double_t location, Bool_t recompute)
{
  setParameterAlpha(alpha, kFALSE);
  setParameterScale(scale, kFALSE);
  setParameterLocation(location, kFALSE);
  
  if (recompute)
    computeTheoreticalInformation();
}

// -----------------------------------------------------------------------
//  Theoretical Information
// -----------------------------------------------------------------------        

void TInvGammaDistribution::computeTheoreticalInformation()
{
    if(_blog)
        cout << " ***  TInvGammaDistribution::computeTheoreticalInformation " << endl;

    _dtheoreticalMean = ((_dparameterAlpha>1) ? _dparameterScale / (_dparameterAlpha - 1) + _dparameterLocation : THEO_DEF);
    
    _dtheoreticalStdDev =  ((_dparameterAlpha>2) ? _dparameterScale*TMath::Sqrt(1./((_dparameterAlpha-1)*(_dparameterAlpha-1)*(_dparameterAlpha-2))) : THEO_DEF);
    
    _dtheoreticalMode = _dparameterScale / (_dparameterAlpha + 1);
}


// -----------------------------------------------------------------------
//  Distribution
// -----------------------------------------------------------------------        

Double_t TInvGammaDistribution::getPDF(Double_t x)
{	

    Double_t y=(x-_dparameterLocation)/_dparameterScale;
    if(y <=0)
        return 0;
    
    Double_t pdf = TMath::Power(1./y,_dparameterAlpha+1)*TMath::Exp(-1./y)/(TMath::Gamma(_dparameterAlpha)*_dparameterScale);
    if ((x >= lowerBound) && (x <= upperBound))
        pdf /= _pmax - _pmin;
    else
        pdf=0;
    
    return pdf;
}

Double_t TInvGammaDistribution::getCDF(Double_t x)
{    

    Double_t cdf = 0.0;
    if ((x >= lowerBound) && (x <= upperBound))
    {
        cdf = 1-TMath::Gamma(_dparameterAlpha,_dparameterScale/(x-_dparameterLocation));
        cdf = (cdf - _pmin) / (_pmax - _pmin);
        if (cdf != cdf ) cdf = 0.0;
    }
    else if (x > upperBound)
        cdf = 1.0;

    return cdf;
}

Double_t TInvGammaDistribution::getInvCDF(Double_t p)
{

    // Verify that p is a valid probability
    try
    {
        if ((p < 0.0) || (p > 1.0))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TInvGammaDistribution::getInvCDF: parameter p [%f] is not a valid probability: p must be in the interval [0,1]",
                            p));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    if (_blog)
        cout << " *** Looking for the value x corresponding to probability p["
                << p << "]" << endl;

    Double_t x0 = lowerBound, x1 = upperBound, px;
    Double_t xcrt = (upperBound + lowerBound) / 2.0;
   //Double_t xtol = x0 * 0.000000000000001, ptol = 0.000000000000001;
    //Double_t xtol = ((x0!=0)?x0:0.1) * 1e-15, ptol = 1e-15;
    Double_t xtol = ((x0!=0 && x0!= -DLE_MAX)?TMath::Abs(x0):0.1) * 1e-14, ptol = 1e-16; 
    Int_t ncpt = 0;

    bool hasbroken=false;
    while (TMath::Abs(x1 - x0) > xtol)
    {
        px = getCDF(xcrt);
        if (_blog)
            cout << " ** number of computations [" << ncpt
                    << "], current solution[" << xcrt
                    << "], corresponding probability[" << px
                    << "], lower limit[" << x0 << "], upper limit[" << x1
                 << "]  diff versus tol ["<<TMath::Abs(x1 - x0)<<"/"<<xtol<<"]";
        if (TMath::Abs(px - p) < ptol)
        {
            if (_blog)
                cout << " ** solution found[" << xcrt
                        << "], corresponding probability[" << px
                        << "], requested probability[" << p << "]" << endl;
            return xcrt;
        }
        else if (px < p)
            x0 = xcrt;
        else if (px > p)
            x1 = xcrt;

        if (_blog)
            cout << " --- new lower limit[" << x0 << "], new upper limit[" << x1
                    << "]";

        xcrt = 0.5 * (x0 + x1);
        if (_blog)
            cout << " ************* new proposed solution[" << xcrt << "]"
                    << endl;
        ncpt++;

        
        if(ncpt>500)
        {
            hasbroken=true;
            break;
        }
                    
    }

    if(hasbroken)
    {
        URANIE::DataServer::UMessageLogger _localmess("TInvGamma");
        _localmess <<  " TInvGammaDistribution::getInvCDF the requested invCDF value might be not accurate, the numerical precisions limit is reached"<< Endl;
    }
    
    return xcrt;
}

// -----------------------------------------------------------------------
//  Log
// -----------------------------------------------------------------------

void TInvGammaDistribution::printLog(Option_t *option)
{
    UNUSED(option);
    
    //TInfiniteDistribution::printLog();
    cout << "*******************************" << endl;
    cout << "** TInvGammaDistribution::printLog  ******" << endl;
    cout << "**   Law[" << _lawType << "] InvGamma" << endl;
    cout << "**   Bounds[" << lowerBound << ", " << upperBound << "]" << endl;
    cout << "**   Parameter Alpha[" << _dparameterAlpha << "]" << endl;
    cout << "**   Parameter Scale[" << _dparameterScale << "]" << endl;
    cout << "**   Parameter Location[" << _dparameterLocation << "]" << endl;
    cout << "**   Theoretical Mean[" << _dtheoreticalMean << "]" << endl;
    cout << "**   Theoretical StdDev[" << _dtheoreticalStdDev << "]" << endl;
    cout << "**   Theoretical Mode[" << _dtheoreticalMode << "]" << endl;
    cout << "*******************************" << endl;

}

} // Fin du namespace DataServer
} // Fin du namespace URANIE

// End Of File
