//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TLogNormalDistribution.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Thu Jan 18 17:08:58 CET 2007

 \brief Interface of the class URANIE::DataServer::TLogNormalDistribution

 */
/**

 * @class URANIE::DataServer::TLogNormalDistribution
 *
 * @brief Description of the class TLogNormalDistribution.
 *

 This class defines a stochastic attribute following an LogNormal distribution.

 A positive random variable \f$ x \f$ is said to follow a \em LogNormal law
 when \f$ \ln{x}\f$ follows a Normal law.

 In URANIE, the LogNormal law is defined by three parameters: the <b> minimum value</b> 
 of the distribution \f$ x_{min} \f$ , the <b>mean of the LogNormal law</b> 
 \f$ M = \mathrm{I\!E}\left[~x~\right] \f$ and a coefficient, called  <b>Error Factor</b>
 which represents the ratio between the value of the \f$95^\textrm{th}\f$ percentile 
 and the median value of the LogNormal distribution. The mean is strictly positive, and 
 the error factor is greater than 1.0.

 The error factor can be computed from the mean and variance of the LogNormal law:

 \li \f$ ErrorFactor = \exp(1.645 * \sqrt{\ln(A+1)}) \f$ where \f$ A = V/(M-x_{min})^2   \f$ and \f$ V \f$ is the variance of the LogNormal law.

 In addition, each LogNormal law is related to a Normal law. The relation between 
 the Normal law parameters and the LogNormal law parameters is given by:

 \li \f$ \sigma = \ln{\left(ErrorFactor\right)} / 1.645 \f$

 \li \f$ \mu = \ln{\left(M-x_{min}\right)} - 0.5 * \sigma^2 \f$


 In the equations below, the parameters \f$ \mu \f$ and \f$ \sigma \f$ are the 
 parameters of the related Normal law.

 \par Probability Density Function (PDF):
 \li \f$ f(x) = 1/((x-x_{min})\sigma\sqrt{2\pi}) * \exp\left( -\left(\ln(x-x_{min})-\mu\right)^2 / 2\sigma^2 \right) \f$ for \f$ x > x_{min} \f$

 \par Cumulative Distribution Function (CDF):
 \li \f$ F(x) = 0.5 * \left( 1 + erf\left( \left(\ln(x-x_{min}) - \mu\right) / \sqrt{2\sigma^2} \right) \right)\f$ for \f$ x > 0 \f$

 \par Inverse CDF:
 For a given \f$ p \in [0,1] \f$, if \f$ X \f$ is generated from the inverse of the related 
 Normal law: 
 \li \f$ F^{-1}(p) = \exp(X) + x_{min}\f$ 



 *
 */
#ifndef TLOGNORMALDISTRIBUTION_H
#define TLOGNORMALDISTRIBUTION_H

//Uranie
#include "TStochAttribut.h"
#include "TNormalDistribution.h"

namespace URANIE
{
namespace DataServer
{
class TLogNormalDistribution: public TInfiniteDistribution
{
private:
    TNormalDistribution *_unormal;   ///< The underlying Normal law
    Double_t _dparameterXMin; ///< The minimum value parameter of the LogNormal law
    Double_t _dparameterMean;       ///< The mean parameter of the LogNormal law
    Double_t _dparameterErrorFactor; ///< The error factor parameter of the LogNormal law

    //! Creates the underlying normal distribution
    /*!
     \param M (Double_t) the mean of the LogNormal law
     \param errorfactor (Double_t) the error factor of the LogNormal law
     */
    void createUnderlyingNormal(Double_t M, Double_t errorfactor);

public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    /*!
     Default mean is 0.01, default error factor is 10.0, default minimum value is 0.0
     */
    TLogNormalDistribution();

    //! Constructor setting the name
    /*!
     Default mean is 0.01, default error factor is 10.0, default minimum value is 0.0
     
     \param name (TString) Name of the attribute
     */
    TLogNormalDistribution(TString name);

    //! Constructor setting name, mean and error factor
    /*!
     \param name (TString) Name of the attribute
     \param M (Double_t) Mean parameter \f$ M \f$
     \param errorfactor (Double_t) Error factor parameter
     \param xmin (Double_t) Minimum value parameter (Default = 0.0)
     */
    TLogNormalDistribution(TString name, Double_t M, Double_t errorfactor,
            Double_t xmin = 0.0);

    //! Construction of a TLogNormalDistribution from an existing TAttribute object
    /*!
     \warning this copy constructor is incomplete. Please do not use.
     (Nicolas Gilardi, 2010.06.11)
     
     \param att (TAttribute *) Pointer to an existing TAttribute object
     \param M (Double_t) Mean parameter \f$ M \f$
     \param errorfactor (Double_t) Error factor parameter
     \param xmin (Double_t) Minimum value parameter (Default = 0.0)
     */
    TLogNormalDistribution(URANIE::DataServer::TAttribute *att, Double_t M,
            Double_t errorfactor, Double_t xmin = 0.0);

    //! Default destructor
    virtual ~TLogNormalDistribution();

    //@}

    /** @name Law parameters
     This law has three parameters: the mean \f$ M \f$, the error factor 
     and the minimum value \f$ x_{min} \f$.
     All are defined here. Another function allows to set directly
     the underlying normal law parameters.
     Finally, as this law is defined only over the interval \f$ ]x_{min}, +\inf)\f$, 
     the functions defining the lower and upper bounds must be redefined.
     **/
    //@{
    //! Set the mean parameter
    /*!
     \param M (Double_t) Mean parameter \f$ \mu \f$
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE)

     \exception UErrorExceptions if \f$ M - x_{min} < x_{min} * 1e-6 \f$                
     */
    virtual void setParameterMean(Double_t M, Bool_t recompute = kTRUE);

    //! Return the mean parameter
    Double_t getParameterMean()
    {
        return _dparameterMean;
    }

    //! Set the error factor parameter
    /*!
     \param errorfactor (Double_t) Error factor parameter
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE)
     
     \exception UErrorExceptions if (errorfactor - 1) <  1e-6                
     */
    virtual void setParameterErrorFactor(Double_t errorfactor,
            Bool_t recompute = kTRUE);

    //! Return the error factor parameter
    Double_t getParameterErrorFactor()
    {
        return _dparameterErrorFactor;
    }

    //! Set the minimum value parameter
    /*!                
     \param xmin (Double_t) Minimum value parameter \f$ x_{min} \f$
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE)
     
     \exception UErrorExceptions if \f$ M - x_{min} < x_{min} * 1e-6 \f$                
     */
    virtual void setParameterXMin(Double_t xmin, Bool_t recompute = kTRUE);

    //! Return the minimum value parameter
    Double_t getParameterXMin()
    {
        return _dparameterXMin;
    }

    //! Set all the parameters
    /*!
     \param M (Double_t) Mean parameter \f$ M \f$
     \param errorfactor (Double_t) Error factor parameter
     \param xmin (Double_t) Minimum value parameter (Default = 0.0)                
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE)
     
     \exception UErrorExceptions if \f$ M - x_{min} < x_{min} * 1e-6 \f$                
     */
    virtual void setParameters(Double_t M, Double_t errorfactor, Double_t xmin =
            0.0, Bool_t recompute = kTRUE);

    //! Set the parameters of the underlying normal law
    /*!
     This function allows to set the parameters of the underlying Normal law. The LogNormal
     law parameters (mean and error factor) are automatically updated.
     
     \param mu (Double_t) Mean parameter of the underlying normal law
     \param sigma (Double_t) Standard deviation parameter of the underlying normal law
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE)
     
     \warning the minimum value \f$ x_{min} \f$ of the LogNormal law 
     is unknown to the underlying normal law. If 
     \f$ x_{min} \neq 0.0\f$, it must be set using 
     TLogNormalDistribution::setParameterXMin()
     */
    virtual void setUnderlyingNormalParameters(Double_t mu, Double_t sigma,
            Bool_t recompute = kTRUE);

    //! Return the mean parameter of the underlying normal law
    Double_t getUnderlyingNormalParameterMu()
    {
        return _unormal->getParameterMu();
    }

    //! Return the standard deviation parameter of the underlying normal law
    Double_t getUnderlyingNormalParameterSigma()
    {
        return _unormal->getParameterSigma();
    }

    //! Sets the lower bound of the attribute
    /*!
     As we are dealing with a law defined over the interval \f$ ]0, +\inf)\f$, 
     setting a lower bound greater than 0 means the law is truncated. 
     The _pmin value must be recalculated.
     
     \param val (Double_t) The value of the new lower bound
     
     \exception UErrorExceptions if val is superior to the upper bound
     \exception UErrorExceptions if val is smaller than \f$ x_{min} \f$                
     */
    virtual void setLowerBound(Double_t val);

    //! Sets the upper bound of the attribute
    /*!
     As we are dealing with a law defined over the interval \f$ ]0, +\inf)\f$, 
     setting a finite upper bound means that the law is truncated. 
     The _pmax value must be recalculated. 
     
     \param val (Double_t) The value of the new upper bound
     
     \exception UErrorExceptions if val is inferior to the lower bound
     */
    virtual void setUpperBound(Double_t val);

    //! Sets the lower and upper bounds of the attribute
    /*!
     As we are dealing with a law defined over the interval \f$ ]0, +\inf)\f$, 
     setting a finite upper bound or a lower bound greater than 0.0 means 
     that the law is truncated. The _pmin and _pmax values must be recalculated. 

     \param lower (Double_t) The value of the new lower bound
     \param upper (Double_t) The value of the new upper bound
     
     \exception UErrorExceptions if the upper bound is inferior to the lower bound
     */
    virtual void setBounds(Double_t lower, Double_t upper);

    //@}

    /** @name Distribution specific methods
     PDF, CDF, etc.
     **/
    //@{
    //! Computes the theoretical values of various statistical measures.
    virtual void computeTheoreticalInformation();

    //! Returns the probability density for the value x
    /*!
     \param x (Double_t) An acceptable value for the attribute                                  
     */
    virtual Double_t getPDF(Double_t x);

    //! Returns the cumulative distribution for the value x
    /*!
     \param x (Double_t) An acceptable value for the attribute
     */
    virtual Double_t getCDF(Double_t x);

    //! Returns the value corresponding to the cumulative density p
    /*!
     \param p (Double_t) A valid probability
     
     \exception UErrorExceptions if p is outside of the interval [0,1]
     */
    virtual Double_t getInvCDF(Double_t p);
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}


    ClassDef(URANIE::DataServer::TLogNormalDistribution, ID_DATASERVER)
    //Classe de
};

} // Fin du namespace DataServer
} // Fin du namespace URANIE
#endif
