//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TLogTriangularDistribution.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Thu Jan 18 17:08:58 CET 2007

 \brief Implementation of the class URANIE::DataServer::TLogTriangularDistribution

 */
#include <iostream>
using std::cout;
using std::endl;

#include "TMath.h"
#include "TLogTriangularDistribution.h"

ClassImp (URANIE::DataServer::TLogTriangularDistribution)

namespace URANIE
{
namespace DataServer
{

// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------

TLogTriangularDistribution::TLogTriangularDistribution() :
        TStochasticAttribute("", TStochasticAttribute::kLogTriangular, 0.01,
                10.0)
{
    if (_blog)
        cout << "Constructing URANIE::DataServer::TLogTriangularDistribution"
                << endl;
    setParameterMode(5.0);
}

TLogTriangularDistribution::TLogTriangularDistribution(TString str) :
        TStochasticAttribute(str, TStochasticAttribute::kLogTriangular, 0.01,
                10.0)
{
    if (_blog)
        cout
                << "Constructing URANIE::DataServer::TLogTriangularDistribution with: Name["
                << str << "]" << endl;
    setParameterMode(5.0);
}

TLogTriangularDistribution::TLogTriangularDistribution(TString str,
        Double_t min, Double_t max) :
        TStochasticAttribute(str, TStochasticAttribute::kLogTriangular, 0.01,
                10.0)
{
    if (_blog)
        cout
                << "Constructing URANIE::DataServer::TLogTriangularDistribution with: Name["
                << str << "] lowerBound[" << min << "] upperBound[" << max
                << "]" << endl;
    setBounds(min, max, kFALSE);
    setParameterMode((max + min) / 2.);
}

TLogTriangularDistribution::TLogTriangularDistribution(TString str,
        Double_t min, Double_t max, Double_t mode) :
        TStochasticAttribute(str, TStochasticAttribute::kLogTriangular, 0.01,
                10.0)
{
    if (_blog)
        cout
                << "Constructing URANIE::DataServer::TLogTriangularDistribution with: Name["
                << str << "] lowerBound[" << min << "] upperBound[" << max
                << "] Mode[" << mode << "]" << endl;
    setBounds(min, max, kFALSE);
    setParameterMode(mode);
}

TLogTriangularDistribution::TLogTriangularDistribution(
        URANIE::DataServer::TAttribute *att, Double_t mode) :
        TStochasticAttribute(att, TStochasticAttribute::kLogTriangular)
{
    if (_blog)
        cout
                << "Constructing URANIE::DataServer::TLogTriangularDistribution from existing TAttribute ["
                << att->GetName() << "] and Mode[" << mode << "]" << endl;
    setParameterMode(mode);
}

// -----------------------------------------------------------------------
//  Destructors
// -----------------------------------------------------------------------

TLogTriangularDistribution::~TLogTriangularDistribution()
{
    if (_blog)
        cout << "Destroying URANIE::DataServer::TLogTriangularDistribution "
                << GetName() << endl;

}

// -----------------------------------------------------------------------
//  Parameters
// -----------------------------------------------------------------------

void TLogTriangularDistribution::setLowerBound(Double_t lower, Bool_t recompute)
{
    try
    {
        if (lower <= 0.0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TLogTriangularDistribution::setLowerBound: parameter lower [%f] is negative or null",
                            lower));
        else if (lower >= upperBound)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TLogTriangularDistribution::setLowerBound: parameter lower [%f] is greater than or equal to upperBound [%f]",
                            lower, upperBound));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }
    TAttribute::setLowerBound(lower);

    if (recompute)
        computeTheoreticalInformation();
}

void TLogTriangularDistribution::setUpperBound(Double_t upper, Bool_t recompute)
{
    try
    {
        if (upper <= 0.0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TLogTriangularDistribution::setUpperBound: parameter upper [%f] is negative or null",
                            upper));
        else if (upper <= lowerBound)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TLogTriangularDistribution::setUpperBound: parameter upper [%f] is less than or equal to lowerBound [%f]",
                            upper, lowerBound));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }
    TAttribute::setUpperBound(upper);

    if (recompute)
        computeTheoreticalInformation();
}

void TLogTriangularDistribution::setBounds(Double_t lower, Double_t upper,
        Bool_t recompute)
{
    try
    {
        if (lower <= 0.0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TLogTriangularDistribution::setBounds: parameter lower [%f] is negative or null",
                            lower));
        else if (upper <= 0.0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TLogTriangularDistribution::setBounds: parameter upper [%f] is negative or null",
                            upper));
        else if (upper <= lower)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TLogTriangularDistribution::setBounds: parameter upper [%f] is lesser or equal to parameter lower [%f]",
                            upper, lower));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }
    TAttribute::setLowerBound(lower);
    TAttribute::setUpperBound(upper);

    if (recompute)
        computeTheoreticalInformation();
}

void TLogTriangularDistribution::setParameterMode(Double_t mode,
        Bool_t recompute)
{
    // Verify that mode is between lowerBound and upperBound
    try
    {
        if ((mode < lowerBound) || (mode > upperBound))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TLogTriangularDistribution::setParameterMode: parameter mode [%f] is out of bounds [%f,%f]",
                            mode, lowerBound, upperBound));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    _dparameterMode = mode;

    if (recompute)
        computeTheoreticalInformation();
}

void TLogTriangularDistribution::setParameters(Double_t lower, Double_t upper,
        Double_t mode, Bool_t recompute)
{
    // Verify that mode is between lower and upper
    try
    {
        if ((mode < lower) || (mode > upper))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TLogTriangularDistribution::setParameters: the new parameter mode [%f] is out of the new bounds [%f,%f]",
                            mode, lower, upper));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    setBounds(lower, upper, kFALSE);
    _dparameterMode = mode;

    if (recompute)
        computeTheoreticalInformation();
}

// -----------------------------------------------------------------------
//  Theoretical Information
// -----------------------------------------------------------------------

void TLogTriangularDistribution::computeTheoreticalInformation()
{
    Double_t lnModeMin = TMath::Log(_dparameterMode / lowerBound);
    Double_t lnMaxMin = TMath::Log(upperBound / lowerBound);
    Double_t lnMaxMode = TMath::Log(upperBound / _dparameterMode);

    _dtheoreticalMean = 2
            * (upperBound * lnModeMin - _dparameterMode * lnMaxMin
                    + lowerBound * lnMaxMode)
            / (lnModeMin * lnMaxMin * lnMaxMode);

    // WARNING: I didn't had enough time to derive the formula for the 
    //          variance of a log-triangular distribution. The proposed 
    //          value for _dtheoreticalStdDev is thus a computed approximation.
    //          (Nicolas Gilardi, 2010.06.10)
    Double_t tmpVar = 0.0;
    for (Double_t x = lowerBound; x <= upperBound;
            x += (upperBound - lowerBound) / 1000.0)
    {
        Double_t density = getPDF(x);
        tmpVar += (x * density - _dtheoreticalMean)
                * (x * density - _dtheoreticalMean);
    }
    _dtheoreticalStdDev = TMath::Sqrt(tmpVar / 1000.0);

    _dtheoreticalMode = _dparameterMode;

    _dleftArea = getCDF(_dparameterMode);
}

// -----------------------------------------------------------------------
//  Distribution
// -----------------------------------------------------------------------        

Double_t TLogTriangularDistribution::getPDF(Double_t x)
{
    Double_t pdf = 0.;
    if ((x >= lowerBound) && (x <= upperBound))
    {
        if (x < _dparameterMode)
            pdf = 2.0 * TMath::Log(x / lowerBound)
                    / (TMath::Log(upperBound / lowerBound)
                            * TMath::Log(_dparameterMode / lowerBound));
        else
            pdf = 2.0 * TMath::Log(upperBound / x)
                    / (TMath::Log(upperBound / lowerBound)
                            * TMath::Log(upperBound / _dparameterMode));
    }
    return pdf;
}

Double_t TLogTriangularDistribution::getCDF(Double_t x)
{
    Double_t cdf;
    if (x < lowerBound)
        cdf = 0.0;
    else if (x > upperBound)
        cdf = 1.0;
    else
    {
        if (x < _dparameterMode)
            cdf = (TMath::Log(x / lowerBound) * TMath::Log(x / lowerBound))
                    / (TMath::Log(upperBound / lowerBound)
                            * TMath::Log(_dparameterMode / lowerBound));
        else
            cdf = 1.0
                    - (TMath::Log(upperBound / x) * TMath::Log(upperBound / x))
                            / (TMath::Log(upperBound / lowerBound)
                                    * TMath::Log(upperBound / _dparameterMode));
    }
    return cdf;
}

Double_t TLogTriangularDistribution::getInvCDF(Double_t p)
{
    // Verify that p is a valid probability
    try
    {
        if ((p < 0.0) || (p > 1.0))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TLogTriangularDistribution::getInvCDF: parameter p [%f] is not a valid probability: p must be in the interval [0,1]",
                            p));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    Double_t x;
    if (p < _dleftArea)
        x = TMath::Exp(
                TMath::Log(lowerBound)
                        + TMath::Sqrt(
                                p * TMath::Log(upperBound / lowerBound)
                                        * TMath::Log(
                                                _dparameterMode / lowerBound)));
    else
        x = TMath::Exp(
                TMath::Log(upperBound)
                        - TMath::Sqrt(
                                (1 - p) * TMath::Log(upperBound / lowerBound)
                                        * TMath::Log(
                                                upperBound / _dparameterMode)));

    return x;
}

// -----------------------------------------------------------------------
//  Log
// -----------------------------------------------------------------------      

void TLogTriangularDistribution::printLog(Option_t *option)
{
    TString soption = option;
    soption.ToLower();
    cout << "*******************************" << endl;
    cout << "** TLogTriangularDistribution::printLog  ******" << endl;
    cout << "**   Law[" << _lawType << "] TTriangularDistribution" << endl;
    cout << "**   Bounds[" << lowerBound << ", " << upperBound << "]" << endl;
    cout << "**   Mode parameter[" << _dparameterMode << "]" << endl;
    cout << "**   Theoretical Mean[" << _dtheoreticalMean << "]" << endl;
    cout << "**   Theoretical StdDev[" << _dtheoreticalStdDev << "]" << endl;
    cout << "**   Theoretical Mode[" << _dtheoreticalMode << "]" << endl;
    cout << "*******************************" << endl;
}

} // Fin du namespace DataServer
} // Fin du namespace URANIE

