//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TNormalDistribution.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Thu Jan 18 17:08:58 CET 2007

 \brief Implementation of the class URANIE::DataServer::TNormalDistribution

 */
#include <iostream>
using std::cout;
using std::endl;

#include "TMath.h"
#include "TNormalDistribution.h"
 #include "Math/QuantFuncMathCore.h"

ClassImp (URANIE::DataServer::TNormalDistribution)

namespace URANIE
{
namespace DataServer
{
// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------

TNormalDistribution::TNormalDistribution() :
        TInfiniteDistribution("", TStochasticAttribute::kNormal), _dparameterMu(
                0.0), _dparameterSigma(1.0)
{
    //  cout << "construit TNormalDistribution" << endl;
    computeTheoreticalInformation();
}

TNormalDistribution::TNormalDistribution(TString str) :
        TInfiniteDistribution(str, TStochasticAttribute::kNormal), _dparameterMu(
                0.0), _dparameterSigma(1.0)
{
    //  cout << "Constructeur URANIE::DataServer::TNormalDistribution[" << this << "]str[" << str<<"]"<<endl;
    computeTheoreticalInformation();
}

TNormalDistribution::TNormalDistribution(TString str, Double_t mu,
        Double_t sigma) :
        TInfiniteDistribution(str, TStochasticAttribute::kNormal)
{
    //  cout << "Constructeur URANIE::DataServer::TNormalDistribution[" << this << "]str[" << str<<"] Mean[" << mean <<"] Sigma[" << sigma << "]" << endl;
    setParameters(mu, sigma);
}

TNormalDistribution::TNormalDistribution(URANIE::DataServer::TAttribute *att,
        Double_t mu, Double_t sigma) :
        TInfiniteDistribution(att, TStochasticAttribute::kNormal)
{
    //  cout << "Constructeur URANIE::DataServer::TNormalDistribution att[" << this << "]att->getName()[" << att->getName()<<"] Mean[" << mean <<"] Sigma[" << sigma << "]" << endl;
    setParameters(mu, sigma);
}

// -----------------------------------------------------------------------
//  Destructors
// -----------------------------------------------------------------------

TNormalDistribution::~TNormalDistribution()
{
    //  cout << "detruit TNormalDistribution" << endl; 

}

// -----------------------------------------------------------------------
//  Parameters
// -----------------------------------------------------------------------        

void TNormalDistribution::setParameterMu(Double_t mu, Bool_t recompute)
{
    _dparameterMu = mu;

    if (recompute)
        computeTheoreticalInformation();
}

void TNormalDistribution::setParameterSigma(Double_t sigma, Bool_t recompute)
{
    try
    {
        if (sigma < 1e-6 * _dparameterMu)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TNormalDistribution::setParameterSigma: parameter sigma [%f] is too close to zero: lower limit is 1e-6 * mu [%f]",
                            sigma, 1e-6 * _dparameterMu));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    _dparameterSigma = sigma;

    if (recompute)
        computeTheoreticalInformation();
}

void TNormalDistribution::setParameters(Double_t mu, Double_t sigma,
        Bool_t recompute)
{
    setParameterMu(mu, kFALSE);
    setParameterSigma(sigma, kFALSE);

    if (recompute)
        computeTheoreticalInformation();
}

// -----------------------------------------------------------------------
//  Theoretical Information
// -----------------------------------------------------------------------        

void TNormalDistribution::computeTheoreticalInformation()
{
    _dtheoreticalMean = _dparameterMu;

    _dtheoreticalStdDev = _dparameterSigma;

    _dtheoreticalMode = _dparameterMu;
}

// -----------------------------------------------------------------------
//  Distribution
// -----------------------------------------------------------------------        

Double_t TNormalDistribution::getPDF(Double_t x)
{
    Double_t pdf = 0;
    if ((x >= lowerBound) && (x <= upperBound))
    {
        pdf = (1 / (TMath::Sqrt(2 * TMath::Pi()) * _dparameterSigma))
                * TMath::Exp(
                        (-0.5) * (x - _dparameterMu) * (x - _dparameterMu)
                                / (_dparameterSigma * _dparameterSigma));
        // Take into account the case where at least one of the bounds is finite
        pdf /= _pmax - _pmin;
    }
    return pdf;
}

Double_t TNormalDistribution::getCDF(Double_t x)
{
    Double_t cdf = 0.0;
    if ((x >= lowerBound) && (x <= upperBound))
    {
        cdf = 0.5
                * (1
                        + TMath::Erf(
                                (x - _dparameterMu)
                                        / (_dparameterSigma * TMath::Sqrt(2))));
        // Take into account the case where at least one of the bounds is finite
        //cdf /= _pmax - _pmin;
        cdf = (cdf - _pmin) / (_pmax - _pmin);
    }
    else if (x > upperBound)
        cdf = 1.0;

    return cdf;
}

Double_t TNormalDistribution::getInvCDF(Double_t p)
{
    Bool_t use_ROOT_algorithm = kTRUE;
    Double_t x;

    // Verify that p is a valid probability
    try
    {
        if ((p < 0.0) || (p > 1.0))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TNormalDistribution::getInvCDF: parameter p [%f] is not a valid probability: p must be in the interval [0,1]",
                            p));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }
    // Take into account the case where at least one of the bounds is finite            
    Double_t realP = p * (_pmax - _pmin) + _pmin;

    static Double_t borne = nextafter(0., HUGE_VAL);//1.e-13;

    
    if (realP < borne)
        realP = borne;
    else if (realP >= 1. - borne)
        realP = 1. - borne;
    
    if (_blog)
        cout << "p [" << p << "], realP [" << realP << "]" << endl;

    //          ROOT implementation of AS241 second algorithm
    // (Michael J. Wichura, Applied Statistics, vol. 37, 1988, pp. 477-484)                    
    if (use_ROOT_algorithm)
    {
        if(this->isLowerBounded() && realP<=borne)
            return getLowerBound();
        else if(this->isUpperBounded() && realP>=(1-borne))
            return getUpperBound();
        else
            x = ROOT::Math::normal_quantile(realP,1.);

    }
    else
    {
        //           Peter John Acklam algorithm ? 
        // (http://home.online.no/~pjacklam/notes/invnorm/)


        const Double_t a[6] =
        { -3.969683028665376e+01, 2.209460984245205e+02, -2.759285104469687e+02,
                1.383577518672690e+02, -3.066479806614716e+01,
                2.506628277459239e+00 };
        const Double_t b[5] =
        { -5.447609879822406e+01, 1.615858368580409e+02, -1.556989798598866e+02,
                6.680131188771972e+01, -1.328068155288572e+01 };
        const Double_t c[6] =
        { -7.784894002430293e-03, -3.223964580411365e-01,
                -2.400758277161838e+00, -2.549732539343734e+00,
                4.374664141464968e+00, 2.938163982698783e+00 };
        const Double_t d[4] =
        { 7.784695709041462e-03, 3.224671290700398e-01, 2.445134137142996e+00,
                3.754408661907416e+00 };

        if (realP < borne)
            realP = borne;
        else if (realP > 1. - borne)
            realP = 1. - borne;

        if (_blog)
            cout << "p [" << p << "], realP [" << realP << "]" << endl;

        Double_t q = realP < 1 - realP ? realP : 1.0 - realP;
        Double_t t, u;

        if (q > 0.02425)
        {
            /* Rational approximation for central region. */
            u = q - 0.5;
            t = u * u;
            u = u
                    * (((((a[0] * t + a[1]) * t + a[2]) * t + a[3]) * t + a[4])
                            * t + a[5])
                    / (((((b[0] * t + b[1]) * t + b[2]) * t + b[3]) * t + b[4])
                            * t + 1);
        }
        else
        {
            /* Rational approximation for tail region. */
            t = sqrt(-2 * log(q));
            u = (((((c[0] * t + c[1]) * t + c[2]) * t + c[3]) * t + c[4]) * t
                    + c[5])
                    / ((((d[0] * t + d[1]) * t + d[2]) * t + d[3]) * t + 1);
        }

        t = TMath::Erfc(-u / sqrt(2.)) / 2. - q;

        t = t * TMath::Sqrt(TMath::TwoPi()) * exp(u * u / 2); /* f(u)/df(u) */
        u = u - t / (1 + u * t / 2); /* Halley's method */

        x = realP > 0.5 ? -u : u;
    }

    return x * _dparameterSigma + _dparameterMu;
}

// -----------------------------------------------------------------------
//  Log
// -----------------------------------------------------------------------

void TNormalDistribution::printLog(Option_t *option)
{
    //  TInfiniteDistribution::printLog();
    TString soption = option;
    soption.ToLower();
    cout << "*******************************" << endl;
    cout << "** TNormalDistribution::printLog  ******" << endl;
    cout << "**   Law[" << _lawType << "] Normal" << endl;
    cout << "**   Bounds[" << lowerBound << ", " << upperBound << "]" << endl;
    cout << "**   Mean parameter mu[" << _dparameterMu << "]" << endl;
    cout << "**   StdDev parameter sigma [" << _dparameterSigma << "]" << endl;
    cout << "**   Theoretical Mean[" << _dtheoreticalMean << "]" << endl;
    cout << "**   Theoretical StdDev[" << _dtheoreticalStdDev << "]" << endl;
    cout << "**   Theoretical Mode[" << _dtheoreticalMode << "]" << endl;
    cout << "*******************************" << endl;
}

} // Fin du namespace DataServer
} // Fin du namespace URANIE

