//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TPermutation.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   2020/09/29

 \brief Header of the URANIE::DataServer::TPermutation class
 */
#ifndef TPERMUTATION_H
#define TPERMUTATION_H

/// ROOT
#include "TEventList.h"
#include "TPatternsEventList.h"
#include "TRandom.h"
#include "TTree.h"

/// Uranie
#include "DataServer.h"

namespace URANIE
{
    namespace DataServer
    {
    class TPermutation: public TEventList
    {

    public:
        //---------------------------------------------
        /** @name Constructor and Destructor
         **/
        //@{
        //! Default constructor
        TPermutation();
        //! Default constructor with a name, a title and the init size
        TPermutation(const char *name, const char *title, Int_t initsize, Bool_t bExact=kFALSE);

        //! Default destructor
        virtual ~TPermutation();
        //@}

        //---------------------------------------------
        /** @name Manipulate the entries
         **/
        //@{
        //!  Enter element entry into the list
        virtual void Enter(Long64_t entry);
        //!  Enter element entry into the list
        /*!
         \see Enter
         */
        virtual void enter(Long64_t entry)
        {
            Enter(entry);
        }

        virtual void generatePermutation(Int_t *vecindexes, Option_t *option="");


        //@}

        //---------------------------------------------
        /** @name The Seed
         **/
        //@{
        //! Init the seed
        /*!
         If the value is zero, the seed is initialized with the hour in milliseconds

         \param ind (Int_t) the seed value
         */
        void setSeed(Int_t ind = 0);
        //! Returns the seed value
        Int_t getSeed()
        {
            return _nSeed;
        }

        //! Return the exact attribute
        Bool_t getExact() {
            return _bExact;
        }
        TString getArchiveName() {
            return _sArchiveName;
        }

        TTree *getExactTree() {
            return _ntdOfExact;
        }

        TTree *getSampledTree() {
            return _ntdOfSampled;
        }

        //@}

        //---------------------------------------------
        /** @name Printing Log
         **/
        //@{
        void setLog()
        {
            _blog = kTRUE;
        }
        void unsetLog()
        {
            _blog = kFALSE;
        }
        void changeLog()
        {
            _blog = _blog ? kFALSE : kTRUE;
        }
        Bool_t getLog()
        {
            return _blog;
        }
        virtual void Print(Option_t *option = "") const;
        void printLog(Option_t *option = "")
        {
            Print(option);
        }
        //@}


    private:
        // methods
        void fillAllPermutations();

        void generateBySwapping(Option_t *option="");

        Bool_t isPermutationAlreadyExist();

        // Attributes
    protected:
        Int_t _nSeed;                          ///< The seed parameter
        Long64_t _nMaxSize;                    ///< The total size of Exact method
        Bool_t _bExact;                        ///< Boolean to set exact or not exact list of permutation
        TString _sArchiveName;                 ///< The name of the ROOT File
        TTree *_ntdOfExact;                    ///< TTree of all the permutations
        TPatternsEventList * _pelExactMixed;   ///< The indexes permuted of the exact permutation list
        Long64_t _nExactIndex;                 ///< The current position
        TTree *_ntdOfSampled;                  ///< TTree of all the permutations
        Int_t *_vecCurrent;                    ///< The current permutation
    public:
        Bool_t _blog;          /// log printing

        ClassDef(TPermutation, ID_DATASERVER)
    };

    } // End Of namespace DataServer
} // End Of namespace URANIE

#endif
// End Of File
