//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TTemporalTree.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   2014/11/27

 \brief Implementation of the class URANIE::DataServer::TTemporalTree

*/
#include <iostream>
using std::cout;
using std::endl;

// ROOT
#include "TCanvas.h"
#include "TH2F.h"
#include "TFile.h"
#include "TLeaf.h"
#include "TMath.h"

// URANIE
#include "TTemporalTree.h"
#include "UExceptions.h"
#include "UTimer.h"

ClassImp (URANIE::DataServer::TTemporalTree)

namespace URANIE
{
namespace DataServer
{
// -----------------------------------------------------------------------
//  Constructor
// -----------------------------------------------------------------------
//! Default empty constructor
TTemporalTree::TTemporalTree():
    _sIter(""), _sTime(""), _sY(""), _sXs(""),
    _lstOfYTAttributes(0), _lstOfXAttributes(0),
    _sAdditional(""), _bHaveAdditionnal(kFALSE), _bHaveXs(kFALSE), _bdrawProgressBar(kTRUE),
    _tTree(NULL), _tStatTree(NULL),
    _tds(0), _nNMaxTime(-1),_nNMinTime(-1), _blog(kFALSE)
{
    SetNameTitle ("__TTT__", "Title of TTemporalTree");
}

//! Default constructor
TTemporalTree::TTemporalTree(TDataServer *tds, TString stime, TString sy, TString sxs):
    _sIter(""), _sTime(stime), _sY(sy), _sXs(sxs),
    _lstOfYTAttributes(0), _lstOfXAttributes(0),
    _sAdditional(""), _bHaveAdditionnal(kFALSE), _bHaveXs(kFALSE), _bdrawProgressBar(kTRUE),
    _tTree(0), _tStatTree(0),
    _tds(tds), _nNMaxTime(-1),_nNMinTime(-1), _blog(kFALSE)
{
    SetName( Form("TTT_%s_%s_", _sTime.Data(), _sY.Data ()));
    SetTitle ( Form("title for TTT_%s_%s_", _sTime.Data(), _sY.Data ()));

    try {
        if( _tds == NULL )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                         Form("TTemporalTree::TTemporalTree constructor:\n The TDS is NULL with sY[%s] sTime[%s] and sX[%s]", _sY.Data (), _sTime.Data (), _sXs.Data ()));
        if ( _sY.CountChar(':') != 0 )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                         Form("TTemporalTree::TTemporalTree constructor:\n Only one output attribute sY[%s]", _sY.Data ()));
        if ( ! _tds->isAttribute (_sY) )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                         Form("TTemporalTree::TTemporalTree constructor:\n The output attribute [%s] does not exist in the TDS", _sY.Data ()));
        if ( _sTime.CountChar(':') != 0 )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                         Form("TTemporalTree::TTemporalTree constructor:\n Only one time attribute sTime[%s]", _sTime.Data ()));
        if ( ! _tds->isAttribute (_sTime) )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                         Form("TTemporalTree::TTemporalTree constructor:\n The time attribute [%s] does not exist in the TDS", _sTime.Data ()));

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    _lstOfYTAttributes = _tds->getListOfAttributes(Form("%s:%s", _sTime.Data (), _sY.Data ()));
    if ( _blog )
        _lstOfYTAttributes->Print ("all");

    if ( _sXs.Length () > 0 ) {
        _lstOfXAttributes = _tds->getListOfAttributes(_sXs.Data ());
        if ( _blog )
            _lstOfXAttributes->Print ("all");
        _bHaveXs = kTRUE;
    }

}

TTemporalTree::~TTemporalTree()
{
  if ( _blog ) {
      cout << endl << " ** Begin of Destructor::TTemporalTree[" << GetName() << "]" << endl;
      gDirectory->pwd();
      gDirectory->ls();
      cout << "~TTemporalTree" << endl << endl;
    }

  if ( _blog )
    cout << " ** _lstOfXAttributes ... ";
  if ( _lstOfXAttributes ) {
      if ( _blog )
        cout << " to delete **"<<endl;
      delete _lstOfXAttributes; _lstOfXAttributes = NULL;
    } else {
      if ( _blog )
        cout << " NOT to delete **"<<endl;
    }

  if ( _blog )
    cout << " ** _lstOfYTAttributes ... ";
  if ( _lstOfYTAttributes ) {
      if ( _blog )
        cout << " to delete **"<<endl;
      delete _lstOfYTAttributes; _lstOfYTAttributes = NULL;
    } else {
      if ( _blog )
        cout << " NOT to delete **"<<endl;
    }

  if ( _blog )
    cout << " ** _tTree ... ";
  if ( _tTree ) {
      if ( _blog )
        cout << " to delete **"<<endl;
      delete _tTree; _tTree = NULL;
    } else {
      if ( _blog )
        cout << " NOT to delete **"<<endl;
    }

  if ( _blog )
    cout << " ** _tStatTree ... ";
  if ( _tStatTree ) {
      if ( _blog )
        cout << " to delete **"<<endl;
      delete _tStatTree; _tStatTree = NULL;
    } else {
      if ( _blog )
        cout << " NOT to delete **"<<endl;
    }
  if ( _blog )
    cout << "End Of Destructor::TTemporalTree[" << GetName() << "]" << endl;
}

void TTemporalTree::setIterator(TString siter)
{
    try {
        if ( _sY.CountChar(':') != 0 )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                         Form("TTemporalTree::setIterator method:\n Only one iterator attribute sIter[%s]", siter.Data ()));
        if ( ! _tds->isAttribute (siter) )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                         Form("TTemporalTree::setIterator method:\n The iterator attribute [%s] does not exist in the TDS", siter.Data ()));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    _sIter = siter;
}

void TTemporalTree::setAdditinalAttribute(TString satt)
{
    try {
        if ( satt.CountChar(':') != 0 )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                         Form("TTemporalTree::setAdditinalAttribute method:\n Only one iterator attribute [%s]", satt.Data ()));
        if ( ! _tds->isAttribute (satt) )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                         Form("TTemporalTree::setAdditinalAttribute method:\n The iterator attribute [%s] does not exist in the TDS", satt.Data ()));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    _sAdditional = satt;
    _bHaveAdditionnal = kTRUE;
}



Bool_t TTemporalTree::buildTree (Option_t *option)
{

  if ( _blog )
    cout << " ** TTemporalTree::buildTree with Option_t[" << option << "]" << endl;

//  TFile * rootfile = TFile::Open("temporaltree.root", "recreate");
//  rootfile->cd();
//  cout << " ***************************** PWD *********************" << endl;
//  gDirectory->pwd();
//  gDirectory->ls();
//  cout << " ***************************** PWD *********************" << endl;

  if ( _sIter.Length () == 0 )
    _sIter = _tds->getIteratorName();
  if ( _blog )
    cout << " ** sIter[" << _sIter << "]" << endl;

  _tds->computeStatistic (_sIter);
  Int_t nFirstComputation = Int_t(_tds->getTuple()->GetMinimum(_sIter.Data()));
  Int_t nLastComputation = Int_t(_tds->getTuple()->GetMaximum(_sIter.Data()));
  if ( _blog )
    cout << " ** Number of Computations : [" << nLastComputation-nFirstComputation+1  << "]" << endl;

  Int_t nsize;
  Int_t niter;
  Double_t dadditional = 0.0;

  if ( _blog ) {
      // Clean
      cout << " ***************************** Clean *********************" << endl;
      gROOT->pwd();
      gROOT->ls();
      gROOT->Delete(Form("temporal_%s_%s_%s;*", _tds->GetName (), _sY.Data(), _sTime.Data()));
      cout << endl << " -----------------------------------" << endl;
      gROOT->ls();
      gROOT->pwd();
      cout << " ***************************** End Of Clean *********************" << endl;
    }
  _tTree =  new TTree(Form("temporal_%s_%s_%s", _tds->GetName (), _sY.Data(), _sTime.Data()),
                      Form("Temporal attribute %s versus time %s for tuple %s", _sY.Data(), _sTime.Data(), _tds->GetName()));
  _tTree->SetMarkerStyle(7);
  _tTree->SetMarkerSize(1.5);
  _tTree->SetLineWidth(4);
  _tTree->SetMarkerColor(kBlack);
  _tTree->SetLineColor(kBlack);
  _tTree->Branch("nT", &nsize,"nT/I");
  _tTree->Branch("nS", &niter,"nS/I");

  // Inputs
  Double_t *dvecXs = NULL;
  TList *lstOfXLeaves = NULL;

  if ( _lstOfXAttributes != NULL ) {
      if ( _blog )
        _lstOfXAttributes->Print ("all");
      Int_t nx = _lstOfXAttributes->GetEntries ();
      dvecXs = new Double_t[nx];
      for(Int_t ii=0; ii<nx ; ii++)
        dvecXs[ii] = 0.0;
      TString sxlabel = _sXs;
      sxlabel.ReplaceAll (":", "/D:");
      sxlabel.Append ("/D");
      if ( _blog )
        cout << endl << " **** sX[" << _sXs << "] sxlabel[" << sxlabel << "]" << endl;

      lstOfXLeaves = new TList();
      TIter next(_lstOfXAttributes);
      while (TAttribute *att = (TAttribute *) next())
        {
          lstOfXLeaves->Add(_tds->GetTuple ()->GetLeaf(att->GetName()));
        }
      if ( _blog )
        lstOfXLeaves->Print ("all");

      _tTree->Branch("sX", &dvecXs[0], sxlabel);
    }

  // Additional Attribute if request
  TLeaf *tlAdditional = NULL;
  if ( _bHaveAdditionnal ) {
      tlAdditional = _tds->GetTuple ()->GetLeaf(_sAdditional);
      _tTree->Branch(_sAdditional, &dadditional, Form("%s/D", _sAdditional.Data ()));
    }

  TString sselect = "--vide--";
  if (  _blog )
    cout << " ** sselect[" << sselect << "]" << endl;

  TEventList *tlstevents;
  URANIE::DataServer::UTimer timer(nLastComputation-nFirstComputation, "Uranie Temporal Tree");

  // Find the maximum of number of times
  cout << " ** Looking for the (Min, Max) number of Times ..."<<endl;
  if (  _blog )
    cout << " *** Begin : _nNMinTime[" << _nNMinTime << "]_nNMaxTime[" << _nNMaxTime << "]" << endl;
  for(niter=nFirstComputation; niter<=nLastComputation; niter++) {
      sselect = Form("(%s==%d)", _sIter.Data(), niter);
      if (  _blog )
        cout << " ** sselect[" << sselect << "]";
      // Clean the event list
      gROOT->Delete("evlistUranie;*");
      _tds->draw(">>evlistUranie", sselect,"goff");
      tlstevents = (TEventList*) gDirectory->Get("evlistUranie");
      nsize = tlstevents->GetN ();

      if ( nsize > _nNMaxTime )
        _nNMaxTime = nsize;
      if ( ( _nNMinTime < 0 ) || ( _nNMinTime > nsize ) )
        _nNMinTime = nsize;

      if (  _blog )
        cout << " **** During : ndraw = [" << nsize << "] _nNMinTime[" << _nNMinTime << "] _nNMaxTime[" << _nNMaxTime << "]" << endl;
#ifndef WIN32
      if (_bdrawProgressBar)
        {
          timer.DrawProgressBar(niter);
        }
#endif
    }
  cout << endl << " ** time elapsed " << timer.GetElapsedTime() << endl;
  cout << " ** Number of Times : _nNMinTime[" << _nNMinTime << "]_nNMaxTime[" << _nNMaxTime << "]" << endl;


  // Output and Time attributes
  Double_t *dvecY = new Double_t[_nNMaxTime];
  Double_t *dvecT = new Double_t[_nNMaxTime];
  TList *listOfYTLeaves = new TList();
  listOfYTLeaves->Add (_tds->GetTuple ()->GetLeaf(_lstOfYTAttributes->At(0)->GetName()) );
  listOfYTLeaves->Add (_tds->GetTuple ()->GetLeaf(_lstOfYTAttributes->At(1)->GetName()) );

  if (  _blog )
    listOfYTLeaves->Print ("all");

  _tTree->Branch(_sTime, &dvecT[0], Form("%s[nT]/D", _sTime.Data ()));
  _tTree->Branch(_sY, &dvecY[0], Form("%s[nT]/D", _sY.Data ()));

  // Loop on the computations
  cout << " ** Create the TTree with times ..." << endl;
  for(niter=nFirstComputation; niter<=nLastComputation; niter++) {
      sselect = Form("(%s==%d)", _sIter.Data(), niter);
      if (  _blog )
        cout << " ** sselect[" << sselect << "]";

      // Clean the EventList
      gROOT->Delete("evlistUranie;*");
      _tds->draw(">>evlistUranie", sselect,"goff");
      tlstevents = (TEventList*) gDirectory->Get("evlistUranie");
      nsize = tlstevents->GetN ();
      if (  _blog ) {
          cout << " ** nsize[" << nsize << "]" << endl;
          tlstevents->Print ("all");
        }
      // Output and Time attributes
      for(Int_t ii=0; ii<nsize; ii++) {
          _tds->GetTuple()->GetEntry(tlstevents->GetEntry(ii));
          dvecT[ii] = ((TLeaf *)listOfYTLeaves->At(0))->GetValue();
          dvecY[ii] = ((TLeaf *)listOfYTLeaves->At(1))->GetValue();
        }

      // X attributes
      if ( _bHaveXs ) {
          // Locate at the first Entry of the TEventList to have the X values
          _tds->GetTuple()->GetEntry(tlstevents->GetEntry(0));

          Int_t icrtx = 0;
          // Loop on the X attributes
          TIter nextLeaf(lstOfXLeaves);
          while (TLeaf *leaf = (TLeaf *) nextLeaf())
            {
              dvecXs[icrtx++] = leaf->GetValue();
            }
        }

      // Additional attribute
      if ( _bHaveAdditionnal ) {
          // Locate at the first Entry of the TEventList to have the X values
          _tds->GetTuple()->GetEntry(tlstevents->GetEntry(0));
          dadditional = tlAdditional->GetValue ();
        }

      // Fill all the branchs of the TTree
      _tTree->Fill();

      // Clean
      gROOT->Delete("evlistUranie;*");
      gROOT->Delete("elist1;*");
      gROOT->Delete("htemp;*");
      gDirectory->Delete("evlistUranie;*");
      gDirectory->Delete("elist1;*");
      gDirectory->Delete("htemp;*");

#ifndef WIN32
      if (_bdrawProgressBar)
        {
          timer.DrawProgressBar(niter);
        }
#endif
    }
  cout << endl << " ** time elapsed " << timer.GetElapsedTime() << endl;

  // Clean
  if ( _blog ) {
      cout << " ***************************** The TTree created *********************";
      cout << endl << " ------------- gROOT ---------------------------" << endl;
      gROOT->pwd();
      gROOT->ls();
  }
  gROOT->Delete("evlistUranie;*");
  gROOT->Delete("elist1;*");
  gROOT->Delete(Form("Stat_temp_%s_%s_%s;*", _tds->GetName (), _sY.Data(), _sTime.Data()));
  if ( _blog ) {
      cout << endl << " ------------- gROOT ---------------------------" << endl;
      gROOT->ls();
      cout << endl << " ------------- gDirectory ----------------------" << endl;
      gDirectory->pwd();
      gDirectory->ls();
      cout << " ***************************** End TTree created *********************" << endl;
  }


  cout << " ** Compute Statistics (Mean, Median, Standard-Deviation) ..." << endl;
  _tStatTree =  new TTree(Form("Stat_temp_%s_%s_%s", _tds->GetName (), _sY.Data(), _sTime.Data()),
                          Form("Stat. Time %s versus time %s for %s", _sY.Data(), _sTime.Data(), _tds->GetName()));
  _tStatTree->SetMarkerStyle(8);
  _tStatTree->SetMarkerSize(1.5);
  _tStatTree->SetLineWidth(4);
  _tStatTree->SetMarkerColor(kTeal);
  _tStatTree->SetLineColor(kTeal);


  Int_t ntime;
  Double_t dmeantime, dsigmatime;
  Double_t dmeanY, dsigmaY, dmedianY;

  _tStatTree->Branch("nT", &ntime,"nT/I");

  _tStatTree->Branch(_sTime, &dmeantime, Form("%s/D", _sTime.Data()));
  _tStatTree->Branch(Form("sigma_%s", _sTime.Data ()), &dsigmatime, Form("sigma_%s/D", _sTime.Data()));

  _tStatTree->Branch(_sY, &dmeanY, Form("%s/D", _sY.Data()));
  _tStatTree->Branch(Form("sigma_%s", _sY.Data ()), &dsigmaY, Form("sigma_%s/D", _sY.Data()));
  _tStatTree->Branch(Form("median_%s", _sY.Data ()), &dmedianY, Form("median_%s/D", _sY.Data()));

  sselect = "--vide--";
  if (  _blog )
    cout << " ** sselect(-" << sselect << "-)" << endl << endl;

  for(Int_t itime=0; itime< _nNMaxTime; itime++) {
      sselect = Form("%s[%d]:%s[%d]", _sY.Data(), itime, _sTime.Data(), itime);
      if ( _blog )
        cout << endl << " ** itime(" << itime << ") sselect[" << sselect << "]";

      ntime = Int_t(_tTree->Draw(sselect, "", "goff"));
      if ( _blog )
        cout << " ** ntime[" << ntime << "]";

      if ( ntime > 1 ) {
          dmeantime = TMath::Mean(ntime, _tTree->GetV2());
          dsigmatime = TMath::RMS(ntime, _tTree->GetV2());

          dmeanY = TMath::Mean(ntime, _tTree->GetV1());
          dsigmaY = TMath::RMS(ntime, _tTree->GetV1());
          dmedianY = TMath::Median(ntime, _tTree->GetV1());
        }

      if ( _blog )
        cout << " ** Time[" << dmeantime << "] SigmaTime[" << dsigmatime << "] Y[" << dmeanY << "] SigmaY[" << dsigmaY << "]" << endl;

      if ( _tStatTree != NULL ) {
          _tStatTree->Fill();
        } else {
          cout << " EROOR : The _tStatTree is a pointer NULL" << endl;
          exit(-1);
        }
      // Clean
      gROOT->Delete("htemp;*");
    }

  // Clean
  if ( _blog ) {
      cout << " ***************************** The Stat TTree created *********************";
      cout << endl << " ------------- gROOT ---------------------------" << endl;
      gROOT->pwd();
      gROOT->ls();
  }
  gROOT->Delete("evlistUranie;*");
  gROOT->Delete("elist1;*");
  gROOT->Delete("htemp;*");
  if ( _blog ) {
      cout << endl << " ------------- gROOT ---------------------------" << endl;
      gROOT->ls();
      cout << endl << " ------------- gDirectory ----------------------" << endl;
      gDirectory->pwd();
      gDirectory->ls();
      cout << " ***************************** End Stat TTree created *********************" << endl;
  }


  // Write the two TTrees in a ROOT File
  TFile * rootfile = TFile::Open("temporaltree.root", "recreate");
  _tTree->SetDirectory (0);
  _tTree->Write ();

  _tStatTree->SetDirectory(0);
  _tStatTree->Write ();

  // Close the TFile to store the two TTrees
  rootfile->Close ();
  _tTree->SetDirectory(gDirectory);
  _tStatTree->SetDirectory(gDirectory);

  // Clean
  if ( dvecT != NULL ) { delete [] dvecT; dvecT = NULL;}
  if ( dvecY != NULL ) { delete [] dvecY; dvecY = NULL;}
  if ( dvecXs != NULL ) { delete [] dvecXs; dvecXs = NULL;}
  if ( lstOfXLeaves != NULL ) {
      delete lstOfXLeaves;   lstOfXLeaves = NULL;
    }
  if ( listOfYTLeaves != NULL ) {
      delete listOfYTLeaves;   listOfYTLeaves = NULL;
    }

  if ( _blog )
    cout << " ** End Of TTemporalTree::buildTree with Option_t[" << option << "]" << endl;
  
  return kTRUE;
}


void TTemporalTree::draw(Option_t *option)
{

    if ( _blog ) {
        cout << " ** Begin Of draw" << endl;
        cout << " ** Name[" << GetName() << "]" << endl;
        gDirectory->pwd();
        gDirectory->ls();
        cout << " ** Begin Of draw" << endl;
    }

    TCanvas *c = new TCanvas(Form("Canvas_%s", GetName()), "c1",880,47,726,839);
    c->Divide(1,2);
    c->cd(1);

    _tTree->SetMarkerColor(kBlack);
    _tTree->Draw(Form("%s:%s", _sY.Data (), _sTime.Data ()), "","p");
    TH2F *htemp = (TH2F*)gPad->GetPrimitive("htemp");
    htemp->GetXaxis()->SetTitle(_tds->getAttribute(_sTime)->getLegend());
    htemp->GetYaxis()->SetTitle(_tds->getAttribute(_sY)->getLegend());

    for(Int_t ns=0; ns< _nNMaxTime; ns++) {
        _tTree->SetMarkerColor(ns%10);
        _tTree->Draw(Form("%s[%d]:%s[%d]", _sY.Data (), ns, _sTime.Data (), ns), "","psame");
    }
    _tTree->SetMarkerColor(kOrange);
    _tTree->SetMarkerStyle(8); _tTree->SetMarkerSize(1.5);
    _tTree->Draw(Form("%s:%s", _sY.Data (), _sTime.Data ()), Form("nT!=%d", _nNMaxTime),"psame");

    if ( _tStatTree != NULL ) {
        _tStatTree->SetMarkerColor(kTeal); _tStatTree->SetLineColor(_tStatTree->GetMarkerColor());
        _tStatTree->Draw(Form("%s:%s", _sY.Data (), _sTime.Data ()), "","lsame");
        _tStatTree->Draw(Form("%s+2.0*sigma_%s:%s", _sY.Data (), _sY.Data (), _sTime.Data ()), "","lsame");
        _tStatTree->Draw(Form("%s-2.0*sigma_%s:%s", _sY.Data (), _sY.Data (), _sTime.Data ()), "","lsame");
        _tStatTree->SetMarkerColor(kYellow); _tStatTree->SetLineColor(_tStatTree->GetMarkerColor());
        _tStatTree->Draw(Form("median_%s:%s", _sY.Data (), _sTime.Data ()), "","lsame");
      }

    c->cd(2);
    _tTree->SetMarkerColor(kBlack);
    _tTree->SetMarkerStyle(7); _tTree->SetMarkerSize(1.5);
    _tTree->Draw(Form("%s:%s", _sY.Data (), _sTime.Data ()), "","p");
    for(Int_t niter=0; niter<_tTree->GetEntries(); niter++) {
        _tTree->SetMarkerColor(niter%10); _tTree->SetLineColor(niter%10);
        _tTree->Draw(Form("%s:%s", _sY.Data (), _sTime.Data ()), Form("nS==%d", niter),"lpsame");
        if ( niter%100 == 0 ) {
            gPad->Modified(); gPad->Update();
        }
    }
    htemp = (TH2F*)gPad->GetPrimitive("htemp");
    htemp->GetXaxis()->SetTitle(_tds->getAttribute(_sTime)->getLegend());
    htemp->GetYaxis()->SetTitle(_tds->getAttribute(_sY)->getLegend());
}


TTree * TTemporalTree::getTTree ()
{
//    TFile *trootfile = new TFile(_tds->getArchiveFileName ());
//    if ( _blog ) {
//        trootfile->Print ();
//        trootfile->ls ();
//    }

//    _tTree = (TTree *) trootfile->Get (Form("temporal_%s_%s_%s", _tds->GetName (), _sY.Data(), _sTime.Data()));
    return(_tTree);
}



TTree * TTemporalTree::getStatTree ()
{
    return(_tStatTree);
}



void TTemporalTree::printLog()
{
    cout << "Event TTemporalTree Name[" << GetName() << "] sYime(" << _sTime << ") sY(" << _sY << ")" << endl;
    if ( _bHaveXs )
        cout << " ** sX(" << _sXs << ")" << endl;
    if ( _bHaveAdditionnal )
        cout << " ** _sAdditional[" << _sAdditional << "]" << endl;
    cout << " ** TDS[" << _tds->GetName () << "]" << endl;
}


} //  //namespace DataServer
} //namespace URANIE
