//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TTriangularDistribution.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Thu Jan 18 17:08:58 CET 2007

 \brief Interface of the class URANIE::DataServer::TTriangularDistribution

 */
/**

 * @class URANIE::DataServer::TTriangularDistribution
 *
 * @brief Description of the class TTriangularDistribution.

 This class defines a stochastic attribute following a Triangular distribution.

 The form of the density function is a triangle of base \f$ [x_{min}, x_{max}] \f$, 
 with a maximum density at \f$ x_{mode} \f$. By definition, the area of the 
 triangle is equal to 1.

 \par Probability Density Function (PDF):
 \li \f$ f(x) = 2 * (x-x_{min}) / ( (x_{max}-x_{min}) * (x_{mode}-x_{min}) ) \f$ for \f$ x \in [x_{min},x_{mode}] \f$
 \li \f$ f(x) = 2 * (x_{max}-x) / ( (x_{max}-x_{min}) * (x_{max}-x_{mode}) ) \f$ for \f$ x \in [x_{mode},x_{max}] \f$
 \li \f$ f(x) = 0 \f$ otherwise

 \par Cumulative Distribution Function (CDF):
 \li \f$ F(x) = 0 \f$ for \f$ x < x_{min} \f$ 
 \li \f$ F(x) = (x-x_{min})^2 / ( (x_{max}-x_{min}) * (x_{mode}-x_{min}) ) \f$ for \f$ x \in [x_{min},x_{mode}] \f$
 \li \f$ F(x) = 1 - (x_{max}-x)^2 / ( (x_{max}-x_{min}) * (x_{max}-x_{mode}) ) \f$ for \f$ x \in [x_{mode},x_{max}] \f$
 \li \f$ F(x) = 1 \f$ for \f$ x > x_{max} \f$ 

 \par Inverse CDF:
 \li \f$ F^{-1}(p) = x_{min} + \sqrt{p * (x_{max}-x_{min}) * (x_{mode}-x_{min})} \f$ for \f$ p \in [0,F(x_{mode})] \f$
 \li \f$ F^{-1}(p) = x_{max} - \sqrt{(1-p) * (x_{max}-x_{min}) * (x_{max}-x_{mode})} \f$ for \f$ p \in [F(x_{mode}),1] \f$

 */
#ifndef TTRIANGULARDISTRIBUTION_H
#define TTRIANGULARDISTRIBUTION_H

//Uranie
#include "TStochAttribut.h"

namespace URANIE
{
namespace DataServer
{
class TTriangularDistribution: public TStochasticAttribute
{
private:
    Double_t _dleftArea; ///< The probability that \f$ x \leq x_{mod} \f$, equivalent to \f$ F(x_{mod}) \f$ or getCDF(\f$ x_{mod} \f$) 

protected:
    Double_t _dparameterMode;  ///< The mode parameter

public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    /*!
     The default range is [0,1] with mode = 0.5
     */
    TTriangularDistribution();

    //! Constructor setting the name
    /*! 
     The default range is [0,1] with mode = 0.5
     
     \param name (TString) Name of the attribute
     */
    TTriangularDistribution(TString name);

    //! Constructor setting name and range
    /*!                 
     The default value for the mode is (max-min)/2
     
     \param name (TString) Name of the attribute
     \param min (Double_t) Lower bound
     \param max (Double_t) Upper bound
     */
    TTriangularDistribution(TString name, Double_t min, Double_t max);

    //! Constructor setting name, range and mode
    /*!                 
     \param name (TString)  Name of the attribut
     \param min  (Double_t) Lower bound
     \param max  (Double_t) Upper bound
     \param mode (Double_t) Mode value
     */
    TTriangularDistribution(TString name, Double_t min, Double_t max,
            Double_t mode);

    //! Construction of a TUniformDistribution from an existing TAttribute object
    /*!
     \warning this copy constructor is incomplete. Please do not use.
     (Nicolas Gilardi, 2010.05.28)
     
     \param att  (TAttribute *) Pointer to an existing TAttribute object
     \param mode (Double_t)     Mode value
     */
    TTriangularDistribution(URANIE::DataServer::TAttribute *att, Double_t mode);

    //! Default destructor
    virtual ~TTriangularDistribution();
    //@}

    /** @name Law parameters
     This law has three parameters: the lower bound, the upper bound and 
     the mode. The bounds parameters are defined in the class URANIE::DataServer::TAttribute.
     The mode parameter is defined here.
     **/
    //@{
    //! Set the value of the mode parameter
    /*!
     \param mode (Double_t) Value of the mode parameter
     
     \exception UErrorExceptions if mode is outside of the interval [lowerBound,upperBound]
     */
    virtual void setParameterMode(Double_t mode);

    //! Returns the value of the mode parameter
    virtual Double_t getParameterMode()
    {
        return _dparameterMode;
    }

    //! Set all parameters at once
    /*!
     \param lower  (Double_t) value of the lower bound
     \param upper  (Double_t) value of the upper bound
     \param mode   (Double_t) value of the mode
     
     \exception UErrorExceptions if mode is outside of the interval [lower,upper]
     */
    virtual void setParameters(Double_t lower, Double_t upper, Double_t mode);

    //@}

    /** @name Distribution specific methods
     PDF, CDF, etc.
     **/
    //@{                
    //! Computes the theoretical values of various statistical measures, like mean, standard deviation, mode, etc.
    virtual void computeTheoreticalInformation();

    //! Returns the probability density for the value x
    /*!
     \param x (Double_t) An acceptable value for the attribute                 
     */
    virtual Double_t getPDF(Double_t x);

    //! Returns the cumulative distribution for the value x
    /*!
     \param x (Double_t) An acceptable value for the attribute                
     */
    virtual Double_t getCDF(Double_t x);

    //! Returns the value corresponding to the cumulative density p
    /*!
     \param p (Double_t) A valid probability
     
     \exception UErrorExceptions if p is outside of the interval [0,1]                
     */
    virtual Double_t getInvCDF(Double_t p);
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    //! Display information specific to this class
    /*!
     \param option (Option_t *) character chain containing a list of options (Default = "")
     
     \note definir une liste d'options et un exemple d'utilisation (Nicolas Gilardi 2010.05.31)
     */
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::DataServer::TTriangularDistribution, ID_DATASERVER)
    //Classe de
};

} // Fin du namespace DataServer
} // Fin du namespace URANIE
#endif
