//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   UMessageLogger.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   jeu avr 26 14:14:11 CEST 2007
 
 $Revision$
 $Date$
 
 \brief Interface of the class URANIE::DataServer::UMessageLogger
 
 \par Modifications
 F. Gaudier (2010-08-06 10:08:43)
 - Add the \b Deprecated item between kINFO and kWARNIG in green color
 */
#include <stdlib.h>
// STL include(s):
#include <iomanip>
#include <iostream>

// ROOT include(s):
#include "TObject.h"

// Local include(s):
#include "UMessageLogger.h"
#include <iostream>
#include <fstream>

// uncomment this line to inhibit colored output
#ifndef WIN32
#define USE_COLORED_CONSOLE
#endif

ClassImp (URANIE::DataServer::UMessageLogger);

namespace URANIE
{
namespace DataServer
{

// this is the hard-coded maximum length of the source names
static const std::string::size_type MAXIMUM_SOURCE_NAME_LENGTH = 0;
// this is the hardcoded prefix
//     static const string PREFIX = "--- ";
static const std::string PREFIX = " ";
// this is the hardcoded suffix
//     static const string SUFFIX = " : ";
static const std::string SUFFIX = " ";
std::map<UMessageLogger::EMsgType, std::string>* UMessageLogger::fTypeMap  = 0;
std::map<UMessageLogger::EMsgType, std::string>* UMessageLogger::fColorMap = 0;
Int_t                            UMessageLogger::fgInstanceCounter = 0;

UMessageLogger::UMessageLogger(const TObject* source, EMsgType minType) :
        fObjSource(source), fStrSource(""), fPrefix(PREFIX), fSuffix(SUFFIX), fActiveType(
                minType), fMaxSourceSize(MAXIMUM_SOURCE_NAME_LENGTH), fMinType(
                minType)
{
    // constructor
   fgInstanceCounter++;
    InitMaps();

    TString sname = fObjSource->GetName();
}

UMessageLogger::UMessageLogger(const std::string& source, EMsgType minType) :
        fObjSource(0), fStrSource(source), fPrefix(PREFIX), fSuffix(SUFFIX), fActiveType(
                minType), fMaxSourceSize(MAXIMUM_SOURCE_NAME_LENGTH), fMinType(
                minType)
{
    // constructor
   fgInstanceCounter++;
    InitMaps();
}

UMessageLogger::UMessageLogger(EMsgType minType) :
        fObjSource(0), fStrSource("Unknown"), fPrefix(PREFIX), fSuffix(SUFFIX), fActiveType(
                minType), fMaxSourceSize(MAXIMUM_SOURCE_NAME_LENGTH), fMinType(
                minType)
{
    // constructor
   fgInstanceCounter++;
    InitMaps();
}

UMessageLogger::UMessageLogger(const UMessageLogger& parent) :
        //   basic_ios< UMessageLogger::char_type, UMessageLogger::traits_type >( new UMessageLogger::__stringbuf_type() ),
        basic_ios<UMessageLogger::char_type, UMessageLogger::traits_type>(), std::ostringstream(), TObject(), fPrefix(
                PREFIX), fSuffix(SUFFIX), fMaxSourceSize(
                MAXIMUM_SOURCE_NAME_LENGTH)
{
    // copy constructor
   fgInstanceCounter++;
    InitMaps();
    *this = parent;
}

UMessageLogger::~UMessageLogger()
{
    // destructor
   fgInstanceCounter--;
   if (fgInstanceCounter == 0) {
      // last MsgLogger instance has been deleted, can also delete the maps
      delete fTypeMap;  fTypeMap  = 0;
      delete fColorMap; fColorMap = 0;
   } else if(fgInstanceCounter < 0) {
        //fgInstanceCounter = 0;
   }
}

UMessageLogger& UMessageLogger::operator=(const UMessageLogger& parent)
{
    // assingment operator
    if (&parent != this)
    {
        fObjSource = parent.fObjSource;
        fStrSource = parent.fStrSource;
        fActiveType = parent.fActiveType;
        fMinType = parent.fMinType;
    }

    return *this;
}

std::string UMessageLogger::GetFormattedSource() const
{
    // make sure the source name is no longer than fMaxSourceSize:
    std::string source_name;
    if (fObjSource)
    {
        source_name = fObjSource->GetName();

        TString sname = fObjSource->GetName();
        Int_t nfirst = sname.Index("::");
        if (nfirst > 0)
        {
            Int_t nlast = sname.Index("::", nfirst + 2);
            if (nlast > 0)
                source_name = sname.Remove(nfirst + 2, nlast - nfirst + 1);
        }
    }
    else
        source_name = fStrSource;

    if (source_name.size() > fMaxSourceSize)
    {
        source_name = source_name.substr(0, fMaxSourceSize - 3);
        source_name += "...";
    }

//       return source_name;
    return "";
}

std::string UMessageLogger::GetPrintedSource() const
{
    // the full logger prefix
    std::string source_name = GetFormattedSource();
    if (source_name.size() < fMaxSourceSize)
        for (std::string::size_type i = source_name.size(); i < fMaxSourceSize; i++)
            source_name.push_back(' ');

    return fPrefix + source_name + fSuffix;
}

void UMessageLogger::Send()
{
    // activates the logger writer

    // make sure the source name is no longer than fMaxSourceSize:
    std::string source_name = GetFormattedSource();

    std::string message = this->str();
    std::string::size_type previous_pos = 0, current_pos = 0;

    // slice the message into lines:
    for (;;)
    {
        current_pos = message.find('\n', previous_pos);
        std::string line = message.substr(previous_pos, current_pos - previous_pos);

        std::ostringstream message_to_send;
        // must call the modifiers like this, otherwise g++ get's confused with the operators...
        message_to_send.setf(std::ios::adjustfield, std::ios::left);
        message_to_send.width(fMaxSourceSize);
        message_to_send << source_name << fSuffix << line;
        this->WriteMsg(fActiveType, message_to_send.str());

        if (current_pos == message.npos)
            break;
        previous_pos = current_pos + 1;
    }

    // reset the stream buffer:
    this->str("");
    return;
}

void UMessageLogger::WriteMsg(EMsgType type, const std::string& line) const
{
    // putting the output string, the message type, and the color
    // switcher together into a single string

    if (type < fMinType)
        return;
    std::map<EMsgType, std::string>::const_iterator stype;
    if ((stype = fTypeMap->find(type)) == fTypeMap->end())
        return;
#ifdef USE_COLORED_CONSOLE
    // no text for kVERBOSE
    if (type == kVERBOSE)
        // no color for info
        std::cout << fPrefix << line << std::endl;
    else
        std::cout << fColorMap->find(type)->second << fPrefix << "<URANIE::"
                << stype->second << ">" << line << "\033[0m" << std::endl;
#else
    if (type == kVERBOSE)
    std::cout << fPrefix << line << std::endl;
    else
    std::cout << fPrefix << "<URANIE::" << stype->second << ">" << line << std::endl;
#endif // USE_COLORED_CONSOLE

    // take decision to stop if fatal error
    if (type == kFATAL)
    {
        std::cout << "***> abort program execution" << std::endl;
        exit(1);
    }
}

UMessageLogger& UMessageLogger::Endmsg(UMessageLogger& logger)
{
    // end line
    logger.Send();
    return logger;
}

void UMessageLogger::InitMaps()
{
   // Create the message type and color maps
   if (fTypeMap != 0 && fColorMap != 0) return;

   fTypeMap  = new std::map<EMsgType, std::string>();
   fColorMap = new std::map<EMsgType, std::string>();
    // fill maps that assign a string and a color to echo message level
    (*fTypeMap)[kVERBOSE] = std::string("VERBOSE");
    (*fTypeMap)[kDEBUG] = std::string("DEBUG");
    (*fTypeMap)[kINFO] = std::string("INFO");
    (*fTypeMap)[kDEPRECATED] = std::string("DEPRECATED");
    (*fTypeMap)[kWARNING] = std::string("WARNING");
    (*fTypeMap)[kERROR] = std::string("ERROR");
    (*fTypeMap)[kFATAL] = std::string("FATAL");
    (*fTypeMap)[kALWAYS] = std::string("ALWAYS");

    (*fColorMap)[kVERBOSE] = std::string("\033[1;34m");
    (*fColorMap)[kDEBUG] = std::string("\033[34m");
    (*fColorMap)[kINFO] = std::string("\033[1;34m");
    (*fColorMap)[kDEPRECATED] = std::string("\033[32m");
    (*fColorMap)[kWARNING] = std::string("\033[1;35m");
    (*fColorMap)[kERROR] = std::string("\033[31m");
    (*fColorMap)[kFATAL] = std::string("\033[1;31;40m");
    (*fColorMap)[kALWAYS] = std::string("\033[30m");
}

} //  //namespace DataServer
} //namespace URANIE
