//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   UTimer.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   jeu avr 26 14:14:11 CEST 2007
 
 $Revision$
 $Date$
 
 \brief Interface of the class URANIE::DataServer::UTimer

 Timing information for training and evaluation of Uranie methods

 */
/*!
 @class URANIE::DataServer::UTimer

 @brief Description of the class UTimer.

 Timing information for training and evaluation of Uranie methods

 
 Usage:

 TMVA::Timer timer( Nloops, "MyClassName" ); 
 for (Int_t i=0; i<Nloops; i++) {
 ... // some code

 // now, print progress bar:
 timer.DrawProgressBar( i );

 // **OR** text output of left time (never both !)
 fLogger << " time left: " << timer.GetLeftTime( i ) << Endl;

 } 
 fLogger << "MyClassName" << ": elapsed time: " << timer.GetElapsedTime() 
 << Endl;    

 Remark: in batch mode, the progress bar is quite ugly; you may 
 want to use the text output then
 
 */

#ifndef UTIMER
#define UTIMER

#include "time.h"
#include "TString.h"
#include "TStopwatch.h"

#include "UMessageLogger.h"

// ensure that clock_t is always defined
#if defined(__SUNPRO_CC) && defined(_XOPEN_SOURCE) && (_XOPEN_SOURCE - 0 == 500 )
#ifndef _CLOCK_T
#define _CLOCK_T
typedef long clock_t; // relative time in a specified resolution
#endif /* ifndef _CLOCK_T */

#endif // SUN and XOPENSOURCE=500

namespace URANIE
{
namespace DataServer
{

class UTimer: public TStopwatch
{

public:

    UTimer(Bool_t colourfulOutput = kTRUE);
    UTimer(Int_t ncounts, TString prefix = "", Bool_t colourfulOutput = kTRUE);
    virtual ~UTimer(void);

    void Init(Int_t ncounts);
    void Reset(void);

    // when the "Scientific" flag set, time is returned with subdecimals
    // for algorithm timing measurement
    TString GetElapsedTime(Bool_t Scientific = kTRUE);
    TString GetLeftTime(Int_t icounts);
    void DrawProgressBar(Int_t icounts, Int_t nFailure = 0);

private:

    Double_t ElapsedSeconds(void);
    TString SecToText(Double_t, Bool_t);

    Int_t fNcounts;               // reference number of "counts" 
    TString fPrefix;                // prefix for outputs
    Bool_t fColourfulOutput;       // flag for use of colors

    static const TString fgClassName; // used for output
    static const Int_t fgNbins;     // number of bins in progress bar

    UMessageLogger* fLogger;     // the output logger

    ClassDef(UTimer, ID_DATASERVER) // Timing information for training and evaluation of Uranie methods
    ;
};

} // Fin du namespace DataServer
} // Fin du namespace URANIE

#endif
