//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestQuantiles.cxx
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 \date   $Date$
 
 \brief  Test file of the TDataServer quantile computation functions
 
 Created: 2010-02-17 09:14:10
 
 */

// CPPUNIT
#define DBL_PRECISION_COMPARE 1.e-6

// Uranie
#include "TestQuantiles.h"

QuantilesTest::QuantilesTest()
{
}

QuantilesTest::~QuantilesTest()
{
}

void QuantilesTest::setUp()
{
    nx = 8;
    nProba = 20;

    x = new Double_t[nx];
    mixed_x = new Double_t[nx];
    proba = new Double_t[nProba];
    quantiles = new Double_t[nProba];
    quantileValueRef = new Double_t[nProba];

    x[0] = 1.;
    x[1] = 2.;
    x[2] = 3.;
    x[3] = 5.;
    x[4] = 6.;
    x[5] = 8.;
    x[6] = 9.;
    x[7] = 10.;

    mixed_x[0] = 5.;
    mixed_x[1] = 2.;
    mixed_x[2] = 9.;
    mixed_x[3] = 1.;
    mixed_x[4] = 6.;
    mixed_x[5] = 10.;
    mixed_x[6] = 3.;
    mixed_x[7] = 8.;

    quantileValueRef[0] = 1.00;
    quantileValueRef[1] = 1.35;
    quantileValueRef[2] = 1.70;
    quantileValueRef[3] = 2.05;
    quantileValueRef[4] = 2.40;
    quantileValueRef[5] = 2.75;
    quantileValueRef[6] = 3.20;
    quantileValueRef[7] = 3.90;
    quantileValueRef[8] = 4.60;
    quantileValueRef[9] = 5.15;
    quantileValueRef[10] = 5.50;
    quantileValueRef[11] = 5.85;
    quantileValueRef[12] = 6.40;
    quantileValueRef[13] = 7.10;
    quantileValueRef[14] = 7.80;
    quantileValueRef[15] = 8.25;
    quantileValueRef[16] = 8.60;
    quantileValueRef[17] = 8.95;
    quantileValueRef[18] = 9.30;
    quantileValueRef[19] = 9.65;

    for (Int_t i = 0; i < nProba; i++)
        proba[i] = 0.05 * i;

    tds = new URANIE::DataServer::TDataServer("tds", "Testing Quantiles");
    tds->addAttribute("x");
    tds->addAttribute("mixed_x");
    tds->createTuple();
    for (Int_t i = 0; i < nx; i++)
        tds->getTuple()->Fill(Double_t(i), x[i], mixed_x[i]);
}

void QuantilesTest::tearDown()
{
    delete[] x;
    x = NULL;
    delete[] mixed_x;
    mixed_x = NULL;
    delete[] proba;
    proba = NULL;
    delete[] quantiles;
    quantiles = NULL;
    delete[] quantileValueRef;
    quantileValueRef = NULL;
    delete tds;
    tds = NULL;
}

////////////////////////////////
///                          ///                            
///         Tests            ///
///                          ///
////////////////////////////////

void QuantilesTest::testMethods()
{
    cout << endl;
    // The reference values are taken from R
    Double_t methodValueRef[9][20] =
    {
    { 1, 1, 1, 2, 2, 2, 3, 3, 5, 5, 5, 6, 6, 8, 8, 8, 9, 9, 10, 10 },
    { 1.0, 1.0, 1.0, 2.0, 2.0, 2.5, 3.0, 3.0, 5.0, 5.0, 5.5, 6.0, 6.0, 8.0, 8.0,
            8.5, 9.0, 9.0, 10.0, 10.0 },
    { 1, 1, 1, 1, 2, 2, 2, 3, 3, 5, 5, 5, 6, 6, 8, 8, 8, 9, 9, 10 },
    { 1.0, 1.0, 1.0, 1.2, 1.6, 2.0, 2.4, 2.8, 3.4, 4.2, 5.0, 5.4, 5.8, 6.4, 7.2,
            8.0, 8.4, 8.8, 9.2, 9.6 },
    { 1.0, 1.0, 1.3, 1.7, 2.1, 2.5, 2.9, 3.6, 4.4, 5.1, 5.5, 5.9, 6.6, 7.4, 8.1,
            8.5, 8.9, 9.3, 9.7, 10.0 },
    { 1.00, 1.00, 1.00, 1.35, 1.80, 2.25, 2.70, 3.30, 4.20, 5.05, 5.50, 5.95,
            6.80, 7.70, 8.30, 8.75, 9.20, 9.65, 10.00, 10.00 },
    { 1.00, 1.35, 1.70, 2.05, 2.40, 2.75, 3.20, 3.90, 4.60, 5.15, 5.50, 5.85,
            6.40, 7.10, 7.80, 8.25, 8.60, 8.95, 9.30, 9.65 },
    { 1.000000, 1.000000, 1.166667, 1.583333, 2.000000, 2.416667, 2.833333,
            3.500000, 4.333333, 5.083333, 5.500000, 5.916667, 6.666667,
            7.500000, 8.166667, 8.583333, 9.000000, 9.416667, 9.833333,
            10.000000 },
    { 1.0000, 1.0000, 1.2000, 1.6125, 2.0250, 2.4375, 2.8500, 3.5250, 4.3500,
            5.0875, 5.5000, 5.9125, 6.6500, 7.4750, 8.1500, 8.5625, 8.9750,
            9.3875, 9.8000, 10.0000 } };

    for (Int_t method = 1; method < 10; method++)
    {
        tds->computeQuantile("x", nProba, proba, quantiles, method);
        for (Int_t i = 0; i < nProba; i++)
            CPPUNIT_ASSERT_DOUBLES_EQUAL(methodValueRef[method - 1][i],
                    quantiles[i], DBL_PRECISION_COMPARE);
    }
}

void QuantilesTest::testOrder()
{
    tds->computeQuantile("mixed_x", nProba, proba, quantiles);
    for (Int_t i = 0; i < nProba; i++)
        CPPUNIT_ASSERT_DOUBLES_EQUAL(quantileValueRef[i], quantiles[i],
                DBL_PRECISION_COMPARE);
}
