//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTDSNtupleD.cxx
 \author C�dric Forgues (C.S.) <uranietm@cea.fr>
 \date   $Date$

 \brief  Fichier de test de la classe URANIE::DataServer::TDSNtupleD

 Cr�e le : mer jun 16 10:00:00 2010

 \par Modifications
 \li Cedric Forgues (C.S.) 2010/06/16
 Ajout des m�thodes pour la DI 49
 // *****************************************************************************

 */

#define DBL_PRECISION_COMPARE 1.e-12

#include "TestTDSNtupleD.h"
#include "TString.h"
#include "TFile.h"

#include <iostream>

TDSNtupleDTest::TDSNtupleDTest()
{

}

TDSNtupleDTest::~TDSNtupleDTest()
{

}

void TDSNtupleDTest::setUp()
{
  _tdsntuple=NULL;
}

void TDSNtupleDTest::tearDown()
{
  delete _tdsntuple;
  _tdsntuple=NULL;
}

////////////////////////////////
///                          ///
///         Tests            ///
///                          ///
////////////////////////////////

void TDSNtupleDTest::testTDSNtupleDConstructorFromTuple()
{
  TNtupleD * tuple = new TNtupleD("Nom", "Titre", "x1:x2");
  Double_t * line1 = new Double_t[2];
  line1[0] = 1, line1[1] = 2;
  Double_t * line2 = new Double_t[2];
  line2[0] = 3, line2[1] = 4;
  tuple->Fill(line1);
  tuple->Fill(line2);

  _tdsntuple = new URANIE::DataServer::TDSNtupleD(tuple);

  Double_t * dataX1;
  Double_t * dataX2;
  Int_t nbDataX1, nbDataX2;

  dataX1 = _tdsntuple->getBranchData("x1", "", &nbDataX1);
  CPPUNIT_ASSERT(nbDataX1 == _tdsntuple->GetEntries());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(dataX1[0], 1., DBL_PRECISION_COMPARE);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(dataX1[1], 3., DBL_PRECISION_COMPARE);

  dataX2 = _tdsntuple->getBranchData("x2", "", &nbDataX2);
  CPPUNIT_ASSERT(nbDataX2 == _tdsntuple->GetEntries());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(dataX2[0], 2., DBL_PRECISION_COMPARE);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(dataX2[1], 4., DBL_PRECISION_COMPARE);

  tearDown();

  _tdsntuple = new URANIE::DataServer::TDSNtupleD(tuple,"x1","x1>2");
  dataX1 = _tdsntuple->getBranchData("x1", "", &nbDataX1);
  CPPUNIT_ASSERT(nbDataX1 == _tdsntuple->GetEntries());
  CPPUNIT_ASSERT(1 == _tdsntuple->GetEntries());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(dataX1[0], 1, DBL_PRECISION_COMPARE);
  
  delete[] line1; line1 = NULL;
  delete[] line2; line2 = NULL;
  delete tuple; tuple = NULL;
}

void TDSNtupleDTest::testTDSNtupleDConstructorFromHomemadeTuple()
{
  URANIE::DataServer::TDataServer *tds = new URANIE::DataServer::TDataServer("myTDS","mon tds a moi");
  tds->fileDataRead("myDataStrVector2.dat");
  tds->saveTuple("RECREATE");
  delete tds; tds=NULL;

  TFile *fin = TFile::Open("myDataStrVector2.root");
  _tdsntuple = new URANIE::DataServer::TDSNtupleD( (TTree*)fin->Get("myTDS__data__tree__") );

  _tdsntuple->GetEntry(0);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(1, ((vector<double>*)((TLeaf*)_tdsntuple->GetLeaf("xfactor"))->GetValuePointer())->at(0), DBL_PRECISION_COMPARE);
  CPPUNIT_ASSERT( ("odd"==*((string*)((TLeaf*)_tdsntuple->GetLeaf("parity"))->GetValuePointer())) );
  
  _tdsntuple->GetEntry( 9 );
  CPPUNIT_ASSERT_DOUBLES_EQUAL(16, ((vector<double>*)((TLeaf*)_tdsntuple->GetLeaf("yfactor"))->GetValuePointer())->at(4), DBL_PRECISION_COMPARE);
  CPPUNIT_ASSERT( ("even"==*((string*)((TLeaf*)_tdsntuple->GetLeaf("parity"))->GetValuePointer())) );
  
}


void TDSNtupleDTest::testTDSNtupleDConstructorWithVarlist()
{
  Int_t nbBranches = 10000;
  TString largeList = "";
  for (Int_t i = 1; i <= nbBranches; i++)
    largeList += Form("x%d:", i);
  largeList = largeList.Remove(largeList.Length() - 1);

  _tdsntuple = new URANIE::DataServer::TDSNtupleD("TDSNtupleD", "Titre du TDSNtupleD", largeList.Data());
  CPPUNIT_ASSERT_EQUAL(_tdsntuple->GetNbranches(), nbBranches);

}

void TDSNtupleDTest::testTDSNtupleDConstructorWithMatrix()
{
  Int_t nbBranches = 3;
  Int_t nbEvents = 3;
  Double_t matrixArray[9] =
    { 1.0, 0.0, 0.0, 2.0, 4.0, 16.0, 3.0, 6.0, 36.0 };
  TMatrixD matrix(nbEvents, nbBranches);
  matrix.SetMatrixArray(matrixArray);

  _tdsntuple = new URANIE::DataServer::TDSNtupleD ("TDSNtupleD", "Titre du TDSNtupleD", "iterator:X:sqrX", matrix);
  CPPUNIT_ASSERT_EQUAL(_tdsntuple->GetNbranches(), nbBranches);
  CPPUNIT_ASSERT_EQUAL(Int_t(_tdsntuple->GetEntries()), nbEvents);
  double *data = _tdsntuple->getBranchData("iterator");
  CPPUNIT_ASSERT_DOUBLES_EQUAL(data[0], 1.0, DBL_PRECISION_COMPARE);
  data = _tdsntuple->getBranchData("X");
  CPPUNIT_ASSERT_DOUBLES_EQUAL(data[2], 6.0, DBL_PRECISION_COMPARE);
  data = _tdsntuple->getBranchData("sqrX");
  CPPUNIT_ASSERT_DOUBLES_EQUAL(data[1], 16.0, DBL_PRECISION_COMPARE);
}

void TDSNtupleDTest::testTDSNtupleDFillDoubleMethod()
{
  _tdsntuple = new URANIE::DataServer::TDSNtupleD ("TDSNtupleD", "Titre du TDSNtupleD", "x:y:z", "D:D:D");
  double pouet[3]={1.0,2.0,3.0};
  _tdsntuple->Fill(pouet);
  _tdsntuple->Fill(2.0,4.0,6.0);
  int foo=3; _tdsntuple->Fill(foo,6.0,9.0);

  for(unsigned int i=0; i<3; i++)
    {
      _tdsntuple->GetEntry(i);
      CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsntuple->getBranchData( "x" )[i], 1.0*(i+1), DBL_PRECISION_COMPARE);
      CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsntuple->getBranchData( "y" )[i], 2.0*(i+1), DBL_PRECISION_COMPARE);
      CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsntuple->getBranchData( "z" )[i], 3.0*(i+1), DBL_PRECISION_COMPARE);
    }

}

void TDSNtupleDTest::testTDSNtupleDFillWithString()
{
  _tdsntuple = new URANIE::DataServer::TDSNtupleD ("TDSNtupleD", "Titre du TDSNtupleD", "V:S:D", "V:S:D");
  _tdsntuple->Fill("1,2,3 nous 2.5");
  _tdsntuple->Fill("4,5,6 vous 5.0");

  _tdsntuple->GetEntry(0);
  CPPUNIT_ASSERT_DOUBLES_EQUAL( ((vector<double>*)_tdsntuple->GetBranch( "V" )->GetLeaf( "V" )->GetValuePointer())->at(0),1, DBL_PRECISION_COMPARE);
  CPPUNIT_ASSERT_DOUBLES_EQUAL( ((vector<double>*)_tdsntuple->GetBranch( "V" )->GetLeaf( "V" )->GetValuePointer())->at(1),2, DBL_PRECISION_COMPARE);
  CPPUNIT_ASSERT_DOUBLES_EQUAL( ((vector<double>*)_tdsntuple->GetBranch( "V" )->GetLeaf( "V" )->GetValuePointer())->at(2),3, DBL_PRECISION_COMPARE);
  CPPUNIT_ASSERT( *(string*)(_tdsntuple->GetBranch( "S" )->GetLeaf( "S" )->GetValuePointer()) == "nous" );
  CPPUNIT_ASSERT_DOUBLES_EQUAL( _tdsntuple->GetBranch( "D" )->GetLeaf( "D" )->GetValue(),2.5, DBL_PRECISION_COMPARE);
  _tdsntuple->GetEntry(1);
  CPPUNIT_ASSERT_DOUBLES_EQUAL( ((vector<double>*)_tdsntuple->GetBranch( "V" )->GetLeaf( "V" )->GetValuePointer())->at(0),4, DBL_PRECISION_COMPARE);
  CPPUNIT_ASSERT_DOUBLES_EQUAL( ((vector<double>*)_tdsntuple->GetBranch( "V" )->GetLeaf( "V" )->GetValuePointer())->at(1),5, DBL_PRECISION_COMPARE);
  CPPUNIT_ASSERT_DOUBLES_EQUAL( ((vector<double>*)_tdsntuple->GetBranch( "V" )->GetLeaf( "V" )->GetValuePointer())->at(2),6, DBL_PRECISION_COMPARE);
  CPPUNIT_ASSERT( *(string*)(_tdsntuple->GetBranch( "S" )->GetLeaf( "S" )->GetValuePointer()) == "vous" );
  CPPUNIT_ASSERT_DOUBLES_EQUAL( _tdsntuple->GetBranch( "D" )->GetLeaf( "D" )->GetValue(),5, DBL_PRECISION_COMPARE);


}

void TDSNtupleDTest::testGetBranchData()
{
  URANIE::DataServer::TDataServer * tds = new URANIE::DataServer::TDataServer("tds", "title of tds");    
  tds->fileDataRead("geyser.dat");        
  URANIE::DataServer::TDSNtupleD *tdsntuple = tds->getTuple();        
  CPPUNIT_ASSERT_EQUAL(tdsntuple->GetNbranches(), 3);
        
  Double_t * dataIter;
  Double_t * dataX1;
  Double_t * dataX2;
  dataIter = tdsntuple->getBranchData("tds__n__iter__");        
  CPPUNIT_ASSERT_DOUBLES_EQUAL(dataIter[0], 1., DBL_PRECISION_COMPARE);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(dataIter[271], 272., DBL_PRECISION_COMPARE);        
  dataX1 = tdsntuple->getBranchData("x1");        
  CPPUNIT_ASSERT_DOUBLES_EQUAL(dataX1[0], 3.600, DBL_PRECISION_COMPARE);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(dataX1[271], 4.467, DBL_PRECISION_COMPARE);       
  dataX2 = tdsntuple->getBranchData("x2");        
  CPPUNIT_ASSERT_DOUBLES_EQUAL(dataX2[0], 79.000, DBL_PRECISION_COMPARE);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(dataX2[271], 74.000, DBL_PRECISION_COMPARE);    

  delete tds;
  tds = NULL;
}

void TDSNtupleDTest::testCopyBranchData()
{
    // Construction of the tuple
    _tdsntuple = new URANIE::DataServer::TDSNtupleD("tuple", "copy test", "x1:x2:x3");

    // Filling the tuple
    Double_t inputData[3] =
    { 1.0, 2.0, 3.0 };
    _tdsntuple->Fill(inputData);
    inputData[0] = -1.0;
    inputData[1] = -4.0;
    inputData[2] = -9.0;
    _tdsntuple->Fill(inputData);
    inputData[0] = 1.0;
    inputData[1] = 4.0;
    inputData[2] = 9.0;
    _tdsntuple->Fill(inputData);
    inputData[0] = 0.5;
    inputData[1] = 1.0;
    inputData[2] = 1.5;
    _tdsntuple->Fill(inputData);
    inputData[0] = -2.0;
    inputData[1] = -1.0;
    inputData[2] = 0.0;
    _tdsntuple->Fill(inputData);

    // Construction of the output array
    Int_t maxData = (Int_t) _tdsntuple->GetEntries();
    Double_t *outDataX1 = new Double_t[maxData];
    Double_t *outDataX2Pos = new Double_t[maxData];
    Double_t *outDataX3Neg = new Double_t[maxData];
    Int_t nbX1, nbX2Pos, nbX3Neg;

    _tdsntuple->copyBranchData(outDataX1, maxData, "x1", "", &nbX1);
    _tdsntuple->copyBranchData(outDataX2Pos, maxData, "x2", "x2>0.0", &nbX2Pos);
    _tdsntuple->copyBranchData(outDataX3Neg, maxData, "x3", "x1<0.0", &nbX3Neg);

    CPPUNIT_ASSERT_EQUAL(nbX1, 5);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(outDataX1[0], 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(outDataX1[1], -1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(outDataX1[2], 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(outDataX1[3], 0.5, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(outDataX1[4], -2.0, DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_EQUAL(nbX2Pos, 3);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(outDataX2Pos[0], 2.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(outDataX2Pos[1], 4.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(outDataX2Pos[2], 1.0, DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_EQUAL(nbX3Neg, 2);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(outDataX3Neg[0], -9.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(outDataX3Neg[1], 0.0, DBL_PRECISION_COMPARE);

    delete[] outDataX1;
    outDataX1 = NULL;
    delete[] outDataX2Pos;
    outDataX2Pos = NULL;
    delete[] outDataX3Neg;
    outDataX3Neg = NULL;
}

void TDSNtupleDTest::testExtractData()
{
    // Construction of the tuple
  _tdsntuple = new URANIE::DataServer::TDSNtupleD("tuple", "copy test", "x1:x2:x3");

    // Filling the tuple
    Double_t inputData[3] =
    { 1.0, 2.0, 3.0 };
    _tdsntuple->Fill(inputData);
    inputData[0] = -1.0;
    inputData[1] = -4.0;
    inputData[2] = -9.0;
    _tdsntuple->Fill(inputData);
    inputData[0] = 1.0;
    inputData[1] = 4.0;
    inputData[2] = 9.0;
    _tdsntuple->Fill(inputData);
    inputData[0] = 0.5;
    inputData[1] = 1.0;
    inputData[2] = 1.5;
    _tdsntuple->Fill(inputData);
    inputData[0] = -2.0;
    inputData[1] = -1.0;
    inputData[2] = 0.0;
    _tdsntuple->Fill(inputData);

    // Construction of the output array
    TMatrixD outMatrix(5, 3);
    Double_t outData[15];
    Double_t outDataRef[15] =
    { 1.0, 2.0, 3.0, -1.0, -4.0, -9.0, 1.0, 4.0, 9.0, 0.5, 1.0, 1.5, -2.0, -1.0,
            0.0 };
    Double_t outDataX1X3lin[10];
    Double_t outDataX1X3linRef[10] =
    { 1.0, 3.0, -1.0, -9.0, 1.0, 9.0, 0.5, 1.5, -2.0, 0.0 };
    Double_t outDataX2X1col[10];
    Double_t outDataX2X1colRef[10] =
    { 2.0, -4.0, 4.0, 1.0, -1.0, 1.0, -1.0, 1.0, 0.5, -2.0 };

    // The matrix is created too small on purpose: it should be resized 
    // automatically by the function extractData.                                       
    TMatrixD outMatrixSelect(2, 2);
    Double_t outDataSelect[6];
    Double_t outDataSelectRef[6] =
    { -1.0, -4.0, -9.0, -2.0, -1.0, 0.0 };

    // data extraction to array
    Int_t nbData;
    _tdsntuple->extractData(outData, 15, "x1:x2:x3");
    _tdsntuple->extractData(outDataX1X3lin, 10, "x1:x3", "", "line");
    _tdsntuple->extractData(outDataX2X1col, 10, "x2:x1", "", "column");
    _tdsntuple->extractData(outDataSelect, 6, "*", "x1<0.0", "line", &nbData);

    // data extraction to matrix
    _tdsntuple->extractData(&outMatrix);
    _tdsntuple->extractData(&outMatrixSelect, "", "x1<0.0");

    // tests
    for (Int_t i = 0; i < 15; i++)
    {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(outMatrix.GetMatrixArray()[i],
                outDataRef[i], DBL_PRECISION_COMPARE);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(outData[i], outDataRef[i],
                DBL_PRECISION_COMPARE);
    }
    for (Int_t i = 0; i < 10; i++)
    {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(outDataX1X3lin[i], outDataX1X3linRef[i],
                DBL_PRECISION_COMPARE);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(outDataX2X1col[i], outDataX2X1colRef[i],
                DBL_PRECISION_COMPARE);
    }

    CPPUNIT_ASSERT_EQUAL(nbData, 6);
    for (Int_t i = 0; i < 6; i++)
    {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(outMatrixSelect.GetMatrixArray()[i],
                outDataSelectRef[i], DBL_PRECISION_COMPARE);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(outDataSelect[i], outDataSelectRef[i],
                DBL_PRECISION_COMPARE);
    }

}

void TDSNtupleDTest::testReadFile()
{
    _tdsntuple = new URANIE::DataServer::TDSNtupleD();
    _tdsntuple->ReadFile("basic.dat");
    CPPUNIT_ASSERT_EQUAL(1000, (Int_t)(_tdsntuple->GetEntries()));
}

////////////////////////////////
///                          ///
///       Exceptions         ///
///                          ///
////////////////////////////////

void TDSNtupleDTest::throwExceptionGetBranchData()
{
    URANIE::DataServer::TDataServer * tds = new URANIE::DataServer::TDataServer("tds", "title of tds");
    tds->fileDataRead("geyser.dat");
    URANIE::DataServer::TDSNtupleD *tdsntuple = tds->getTuple();
    CPPUNIT_ASSERT_THROW(tdsntuple->getBranchData("x3"), URANIE::Exceptions::UErrorExceptions);

    delete tds;
    
    _tdsntuple = new URANIE::DataServer::TDSNtupleD ("TDSNtupleD", "Titre du TDSNtupleD", "x:y:z", "D:D:S");  
    CPPUNIT_ASSERT_THROW(_tdsntuple->getBranchData("z"), URANIE::Exceptions::UErrorExceptions);
    tearDown();

    _tdsntuple = new URANIE::DataServer::TDSNtupleD ("TDSNtupleD", "Titre du TDSNtupleD", "x:y:z", "D:D:V");  
    CPPUNIT_ASSERT_THROW(_tdsntuple->getBranchData("z"), URANIE::Exceptions::UErrorExceptions);
}

void TDSNtupleDTest::throwExceptionCopyBranchData()
{
    Int_t nbData;
    Double_t * dataArray;
    URANIE::DataServer::TDataServer * tds = new URANIE::DataServer::TDataServer(
            "tds", "title of tds");

    tds->fileDataRead("geyser.dat");
    URANIE::DataServer::TDSNtupleD *tdsntuple = tds->getTuple();

    nbData = (Int_t) tdsntuple->GetEntries();
    dataArray = new Double_t[nbData];
    CPPUNIT_ASSERT_THROW(tdsntuple->copyBranchData(dataArray, nbData, "x3"),
            URANIE::Exceptions::UErrorExceptions);
    delete[] dataArray;

    nbData = nbData / 2;
    dataArray = new Double_t[nbData];
    CPPUNIT_ASSERT_THROW(tdsntuple->copyBranchData(dataArray, nbData, "x1"),
            URANIE::Exceptions::UErrorExceptions);
    delete[] dataArray;

    delete tds;

    _tdsntuple = new URANIE::DataServer::TDSNtupleD ("TDSNtupleD", "Titre du TDSNtupleD", "x:y:z", "D:D:S");  
    CPPUNIT_ASSERT_THROW(_tdsntuple->copyBranchData(dataArray, nbData, "z"), URANIE::Exceptions::UErrorExceptions);
    tearDown();

    _tdsntuple = new URANIE::DataServer::TDSNtupleD ("TDSNtupleD", "Titre du TDSNtupleD", "x:y:z", "D:D:V");  
    CPPUNIT_ASSERT_THROW(_tdsntuple->copyBranchData(dataArray, nbData, "z"), URANIE::Exceptions::UErrorExceptions);
}

void TDSNtupleDTest::throwExceptionConstructWithMatrix()
{

  Int_t nbBranches = 4;
  Int_t nbEvents = 3;
  Double_t matrixArray[12] =
    { 1.0, 0.0, 0.0, 2.0, 4.0, 16.0, 3.0, 6.0, 36.0,1.0,1.0,1.0 };
  TMatrixD matrix(nbEvents, nbBranches);
  matrix.SetMatrixArray(matrixArray);

  CPPUNIT_ASSERT_THROW(new URANIE::DataServer::TDSNtupleD ("TDSNtupleD", "Titre du TDSNtupleD", "iterator:X:sqrX", matrix),
		       URANIE::Exceptions::UErrorExceptions);
}

void TDSNtupleDTest::throwExceptionConstructNameAndTypeNotMatching()
{
  CPPUNIT_ASSERT_THROW(new URANIE::DataServer::TDSNtupleD ("TDSNtupleD", "Titre du TDSNtupleD", "x:y:z", "D:D"),
		       URANIE::Exceptions::UErrorExceptions);
}

void TDSNtupleDTest::throwExceptionMergerNotSameNumberOfEntries()
{

  _tdsntuple = new URANIE::DataServer::TDSNtupleD ("t1", "Titre de t1", "x:y:z", "D:D:D");
  URANIE::DataServer::TDSNtupleD *t2 = new URANIE::DataServer::TDSNtupleD ("t2", "Titre de t2", "a:b:c", "D:D:D");
  _tdsntuple->Fill(string("0.1 0.1 0.1"));
  t2->Fill(string("0.2 0.2 0.2"));
  t2->Fill(string("0.3 0.3 0.3"));
  
  CPPUNIT_ASSERT_THROW(_tdsntuple->Merger(t2,"a:b:c","a:b:c"),URANIE::Exceptions::UErrorExceptions);
  delete t2;
}


void TDSNtupleDTest::throwExceptionFillDoublesWhenStringOrVectorInTree()
{
  _tdsntuple = new URANIE::DataServer::TDSNtupleD ("TDSNtupleD", "Titre du TDSNtupleD", "x:y:z", "D:D:S");  
  CPPUNIT_ASSERT_THROW(_tdsntuple->Fill(1.0,2.0,3.0), URANIE::Exceptions::UErrorExceptions);
  tearDown();

  _tdsntuple = new URANIE::DataServer::TDSNtupleD ("TDSNtupleD", "Titre du TDSNtupleD", "x:y:z", "D:D:V");  
  CPPUNIT_ASSERT_THROW(_tdsntuple->Fill(1.0,2.0,3.0), URANIE::Exceptions::UErrorExceptions);
}

void TDSNtupleDTest::throwExceptionFillDoubleArrayWhenStringOrVectorInTree()
{
  double pouet[3]={1.0,2.0,3.0};
  
  _tdsntuple = new URANIE::DataServer::TDSNtupleD ("TDSNtupleD", "Titre du TDSNtupleD", "x:y:z", "D:D:S");  
  CPPUNIT_ASSERT_THROW(_tdsntuple->Fill(pouet), URANIE::Exceptions::UErrorExceptions);
  tearDown();

  _tdsntuple = new URANIE::DataServer::TDSNtupleD ("TDSNtupleD", "Titre du TDSNtupleD", "x:y:z", "D:D:V");  
  CPPUNIT_ASSERT_THROW(_tdsntuple->Fill(pouet), URANIE::Exceptions::UErrorExceptions);
}


void TDSNtupleDTest::throwExceptionExtractData()
{
  Int_t nbData;
  Double_t * dataArray;
  TMatrixD mat(2, 2);
  URANIE::DataServer::TDataServer * tds = new URANIE::DataServer::TDataServer(
									      "tds", "title of tds");

  tds->fileDataRead("geyser.dat");
  URANIE::DataServer::TDSNtupleD *tdsntuple = tds->getTuple();

  CPPUNIT_ASSERT_THROW(tdsntuple->extractData(&mat, "x1:x3"),
		       URANIE::Exceptions::UErrorExceptions);

  nbData = 2 * (Int_t) tdsntuple->GetEntries();
  dataArray = new Double_t[nbData];
  CPPUNIT_ASSERT_THROW(tdsntuple->extractData(dataArray, nbData, "x1:x3"),
		       URANIE::Exceptions::UErrorExceptions);
  delete[] dataArray;

  nbData = (Int_t) tdsntuple->GetEntries();
  dataArray = new Double_t[nbData];
  CPPUNIT_ASSERT_THROW(tdsntuple->extractData(dataArray, nbData, "x1:x2"),
		       URANIE::Exceptions::UErrorExceptions);

  delete tds;

  _tdsntuple = new URANIE::DataServer::TDSNtupleD ("TDSNtupleD", "Titre du TDSNtupleD", "x:y:z", "D:D:S");  
  CPPUNIT_ASSERT_THROW(_tdsntuple->extractData(dataArray, nbData, "x:z"), URANIE::Exceptions::UErrorExceptions);
  CPPUNIT_ASSERT_THROW(_tdsntuple->extractData(dataArray, nbData, "*"), URANIE::Exceptions::UErrorExceptions);
  CPPUNIT_ASSERT_THROW(_tdsntuple->extractData(&mat, "x:z"), URANIE::Exceptions::UErrorExceptions);
  CPPUNIT_ASSERT_THROW(_tdsntuple->extractData(&mat, "*"),   URANIE::Exceptions::UErrorExceptions);
  tearDown();
    
  _tdsntuple = new URANIE::DataServer::TDSNtupleD ("TDSNtupleD", "Titre du TDSNtupleD", "x:y:z", "D:D:V");
  CPPUNIT_ASSERT_THROW(_tdsntuple->extractData(dataArray, nbData, "x:e"), URANIE::Exceptions::UErrorExceptions);
  CPPUNIT_ASSERT_THROW(_tdsntuple->extractData(dataArray, nbData, "*"), URANIE::Exceptions::UErrorExceptions);
  CPPUNIT_ASSERT_THROW(_tdsntuple->extractData(&mat, "x:z"), URANIE::Exceptions::UErrorExceptions);
  CPPUNIT_ASSERT_THROW(_tdsntuple->extractData(&mat, "*"),   URANIE::Exceptions::UErrorExceptions);

  delete[] dataArray;   
}

void TDSNtupleDTest::throwExceptionFillWithStringArgumentProblem()
{
  _tdsntuple = new URANIE::DataServer::TDSNtupleD ("TDSNtupleD", "Titre du TDSNtupleD", "V:S:D", "V:S:D");
  CPPUNIT_ASSERT_THROW(_tdsntuple->Fill("1,2,3 nous 2.5 535"),URANIE::Exceptions::UErrorExceptions);
  CPPUNIT_ASSERT_THROW(_tdsntuple->Fill("1,2,3 nous"),URANIE::Exceptions::UErrorExceptions);
}
