//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TCluster.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mon august 4th 14:21:10 CEST 2008
 
 \brief Interface of the class URANIE::Launcher::TCluster
 
 */
/**
 
 * @class URANIE::Launcher::TCluster
 *
 * @brief Description of the class TCluster.
 *
 * To be written by the developper.
 *
 */
#ifndef TCLUSTER_H
#define TCLUSTER_H

#include <stdlib.h>
#include <iostream>

//ROOT
#include "Rtypes.h"
#include "TNamed.h"

// Uranie
#include "Launcher.h"

#ifdef WITH_LIBSSH
#include "TCode.h"
#include "TInputFileFlag.h"
#include "TAttribute.h"
#include "TInputFileRecreate.h"

//SSH
#define LIBSSH_STATIC 1
#include <libssh/libssh.h>
#include <libssh/sftp.h>
#endif

namespace URANIE
{
namespace Launcher
{
class TCluster: public TNamed
{
public:
    Bool_t _blog;                          ///< Boolean to edit the log
#ifdef WITH_LIBSSH    
    enum EDistrib { kUnknown, kLSB, kSLURM, kMPI, kPBS, kSGE, kEc };
    EDistrib  _edistrib;
    TString _eType;
#endif
    
private:
    Int_t _nnode;                          ///< The number of nodes by proc
    TString _slistOfNodes;                   ///< The list of nodes
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{

    TCluster(const char* name, const char *nodeList, Int_t nnode);
#ifdef WITH_LIBSSH    
    TCluster(URANIE::DataServer::TDataServer *tds, TCode *code);
    TCluster(TCode *code);
#endif

    //! Default destructor
    virtual ~TCluster();

    //! Constructs the list of nodes
    virtual void init() = 0;
    //@}

#ifdef WITH_LIBSSH    
    //! setWorkingDirectory
    /*!
     * localWorkingDirectory Tstring
     */
    void setWorkingDirectory(TString localWorkingDirectory);
    void setRemoteWorkingDirectory();

    TCode * _code;
    URANIE::DataServer::TDataServer* _tds;
#endif
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    void setNNodes(Int_t nnode){
        _nnode = nnode;
    }
    TString getListOfNodes(){
        return _slistOfNodes;
    }
    // Adds the node name
    void addNodeName(const TString& snodename, Int_t n = -1);
    virtual void getCommandPrefix(char *, char*) = 0;
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    void setLog(){
        _blog = kTRUE;
    }
    void unsetLog(){
        _blog = kFALSE;
    }
    void changeLog(){
        _blog = _blog ? kFALSE : kTRUE;
    }
    Bool_t getLog(){
        return _blog;
    }

    virtual void printLog(Option_t *option = "");
    //@}

#ifdef WITH_LIBSSH
    //////////////////////////
    /// Return different cluster
    /////////////////////////

    /////////////////////////
    /// Get and sets
    /////////////////////////
    void setClusterUserAuth(TString user ,TString  authentification_method="public_key"){
        _sClusterUser=user;
        _sClusterUserAuthMethod=authentification_method;
    }
    //! Sets the Cluster name
    /*!
     * \param cluster_name (TString) name of the cluster
     */
    virtual void setCluster(TString cluster_name){
        _sClusterName = cluster_name;

    }
    virtual void setCluster(TCluster *cluster);

    /*!
     * \brief Get the Cluster Name
     * \return cluster Name
     */
    TString getClusterName() const{
        return _sClusterName;
    }
    /*!
     * \brief Get the Cluster Ip
     * \return cluster IP
     */
    TString getClusterIP() const{
        return _sClusterIP;
    }
    /*!
     * \brief Get the Cluster Ip
     * \return cluster IP
     */
    TString setClusterIP() const{
        return _sClusterIP;
    }
    //! Sets the Skeleton path
    /*!
     * \param skeleton header (TString) name of the cluster
     */
    void setSkeleton(TString skeleton){
        _sSkeletonHeader = skeleton;
    }
    /*!
     * \brief Get the Header Name
     * \return skeleton Header Name
     */
    TString getSkeleton() const{
        return _sSkeletonHeader;
    }
    //! Sets the generate header output path
    /*!
     * \param generate header (TString) output of the cluster
     */

    void setOutputHeaderName(TString oHeaderName){
#ifdef WIN32
        const char *tail = strrchr(oHeaderName.Data(), '/');
        _sOutputHeader = tail ? tail + 1 : oHeaderName.Data();   
#else
        _sOutputHeader=basename(oHeaderName);
#endif
    }
    //        //! Sets the path to generate header output path
    //    /*!
    //     * \param generate header (TString) output of the cluster
    //     */

    //     * \brief Get the path to remoteWorkingDir
    //     * \return Output Path Header Name
    //     */

    //! setJobID
    /*!
     * \brief set the split_jobid
     * \return Output Path Header Name
     */
    void setJobID(TString jobid){
        _sJobID.push_back(jobid);

    }
    //! setNRetries
    /*!
     * \brief Get the path to remoteWorkingDir
     * \return Output Path Header Name
     */
    void setNRetries(Int_t retries){
        _sRetries = retries;

    }
    //! getOutputHeaderName
    /*!
     * \brief Get the Header Name
     * \return Output Header Name
     */
    TString getOutputHeaderName() const{
        return _sOutputHeader;
    }
    //! getOutputHeaderPath
    /*!
     * \brief Get the path to Header
     * \return Output Path Header Name
     */
    TString getOutputHeaderPath() const{
        return _sOutputHeaderPath;
    }
    //! getClusterUser
    /*!
     * \brief Get the path to Header
     * \return Output Path Header Name
     */
    TString getClusterUser() const{
        return _sClusterUser;
    }
    //! getClusterUserAuthMethod
    /*!
     * \brief Get the path to Header
     * \return Output Path Header Name
     */
    TString getClusterUserAuthMethod() const{
        return _sClusterUserAuthMethod;
    }
    //! getRemotePath
    /*!
     * \brief Get the path to remoteWorkingDir
     * \return Output Path Header Name
     */
    TString getRemotePath() const{
        return _sRemoteWorkingDirectory;
    }
    //! getOutputSsh
    /*!
     * \brief Get the path to remoteWorkingDir
     * \return Output Path Header Name
     */
    TString getOutputSsh() const{
        return _sSshOutput;
    }
    //! getOutputSsh
    /*!
     * \brief Get the script name on remote platform
     * \return headerfinalname
     */
    TString getSplitName() const{
        return _sSplitName;
    }
    //! getJobID
    /*!
     * \brief Get JobID
     * \return jobID
     */
    TString getJobID(Int_t index) const{
        return _sJobID.at(index);
    }
    //! Sets the number of cores requested
    /*!
     * \param ncores (Int_t) number of cores requested
     */
    void setNumberOfCores(Int_t ncores){
        _nAllocatedCores = ncores;
    }
    /*!
     * \brief Get the Number of cores allocated
     * \return Number of cores allocated
     */
    Int_t getNumberOfCores() const{
        return _nAllocatedCores;
    }
    //////////////////////////
    //CLUSTER CODE
    //////////////////////////
    void preTreatment(Option_t *option="");

    //! addClusterDirective
    /*!
     * \param directive  (TString) name
     * \param directive  (TString) value
     */
    void addClusterDirective(TString directive, TString value);

    //! getSplitName
    /*!
     * returns the script for the current split
     */
    void setJobName(TString path){
        _sJobName=path;
    }
    //! getJobName
    /*!
     * returns the script for the current split
     */
    TString getJobName(){
        return _sJobName;
    }

    /////////////////////////
    // SSH
    /////////////////////////

    //! connectSSH
    /*!
     * \param session  (ssh_session)
     * \param cluster_name  (TString)
     * \param user  (TString)
     * \param authentification_method  (TString)
     */
    Int_t connectSSH(TString cluster_name ,  TString  user,  TString authentification_method);
    //! verify_knownhostSSH
    /*!
     * \param session  (ssh_session)
     */
    Int_t verify_knownhostSSH();
    //! openSFTPChannel
    /*!
     * \param session  (ssh_session)
     * \param sftp  (sftp_session)
     */
    Int_t openSFTPChannel();
    //! sendHeader
    Int_t sendHeader();
    //! sendCompiledBinary
    /*!
     */
    Int_t sendCompiledBinary(std::list<TString> filenames);
    //! sendInputFiles
    /*!
     * \param start  (Int_t)
     * \param end  (Int_t)
     */
    Int_t sendInputFiles(Int_t start, Int_t end);

    //! retrieveInputFiles
    /*!
     * \param TString localFolder  (sftp_session)
     *
     */
    Int_t retrieveInputFiles(const char * localFolder);

    //! closeSFTPChannel
    /*!
     * \param sftp  (sftp_session)
     *
     */
    Int_t closeSFTPChannel();

    //! setRemotePath
    /*!
     * \brief set the remote working directory
     */
    void setRemotePath(TString remotePath){
        if (!remotePath.EndsWith("/"))
        {
            remotePath.Append("/");
        }
        _sRemoteWorkingDirectory = remotePath;
    }

    //! setDataServer
    /*!
     * \brief set the remote working directory
     */
    void setDataServer(URANIE::DataServer::TDataServer *tds){
        _tds =tds;
    }
    //! getSplitName
    /*!
     * returns the script for the current split
     */
    TString getGateway()
    {
        return _sGateWay;
    }

    Bool_t isConnected(){
        return _sConnected;
    }

    virtual void selectHeader(TString SkeletonHeaderName="./",TString localHeaderFolderPath=gSystem->pwd() )   {UNUSED(SkeletonHeaderName,localHeaderFolderPath); cout << "base" <<endl;};
    virtual void generateHeader(Option_t *option = "")   {UNUSED(option); cout << "base" <<endl;};
    virtual void generateHeaderSplit(Option_t *option = "")  {UNUSED(option); cout << "base" <<endl;};
    TString getJobStatusCommand(std::list<std::pair<TString, std::pair<int, int> > > JobList) {UNUSED(JobList); return "base";};
    TString getJobStatusCommand(TString JobIDs) {UNUSED(JobIDs); return "base";};
    TString checkJobStatus(TString JobId, TString Command) {UNUSED(JobId,Command); return "base";};
    void  substituteDirectives(TString infilePath, TString outFilePath, std::map<TString,TString> ListOfDirectives);
        //@}

    void execRemoteCommand(TString command);
    virtual TString submitJob(TString command);

    void replaceHeaderKey(std::string& source, const std::string& from, const std::string& to);
#endif
    
protected:

    /////////////////////////
    // SSH
    /////////////////////////
#ifdef WITH_LIBSSH    
    ssh_session _sSshSession;
    /*! A private variable
      ! sFtpSession
      * sftp_session holder
      */
    sftp_session _sFtpSession;
    ssh_session *_ptrsshSession;
    /*! A private variable
      ! sFtpSession
      * sftp_session holder
      */
    sftp_session * _ptrsftp;

    /*! A private variable
     * Store the results of last command executed
     */
    TString _sSshOutput;
    Bool_t _sConnected;

    /////////////////////////
    // Header
    /////////////////////////
    TString _sSkeletonHeader;
    /*! A private variable
     * cluster name of the script to use
     */
    TString _sOutputHeader;
    /*! A private variable
     * Contains the script of the split number of a job
     */
    TString _sSplitName;
    /*! A private variable
     * Contains the workingDirectory
     */
    TString _sWorkingDirectory;
    /*! A private variable
     * Path to cluster Skeleton header to use
     */
    TString _sOutputHeaderPath;
    /*! set job folder Name
     * Path to job folderto use
     */
    TString _sJobName;
    /*! A private variable
     * Path to localHeaderFolderPath
     */
    TString _slocalHeaderFolderPath;
    /*! A private variable
     * Contain the Number of cores allocated
     */
    Int_t _nAllocatedCores;
    /*! A private variable
      ! _sRetries
      * Number Of retries if an inputfile is not found
      */
    Int_t _sRetries;

    /*! A private variable
     * Path to remoteWorkingDirectory
     */
    TString _sRemoteWorkingDirectory;

    /////////////////////////
    // Cluster
    /////////////////////////
    /*! A private variable
      ! _sClusterUser
      * Set cluster user
      */
    TString _sClusterUser;
    /*! A private variable
      ! _sClusterUserAuthMethod
      * Set the authentification method (default public_key)
      */
    TString _sClusterUserAuthMethod;
    /*! A private variable
     * Contain the Name of the Cluster
     */

    TString _sClusterName;
    /*! A private variable
     * Contain the ip of the Cluster
     */
    TString _sClusterIP;

    /*! A private variable
     * Contains the gateway of the Cluster
     */
    TString _sGateWay;
     /*! A private variable
     * Store the jobid submitted
     */
    vector<TString> _sJobID;
    /*! A private boolean
     * Describes if a binary has been send
     */
    Bool_t _sUserBinary=kFALSE;
    /*! A private Tstring
     * Describes the protocol of the connection
     */
    TString _sConnectionProtocol="ssh";

    /*! A private variable
     * List of Cluster Directives
     */
    std::map<TString,TString> _sListClusterDirectives;
#endif

    Int_t getProcsPerJob(Char_t separator, const Char_t* procs) const;

    ClassDef(URANIE::Launcher::TCluster, ID_LAUNCHER)

};


} // Fin du namespace Launcher
} // Fin du namespace URANIE
#endif
