//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
// TCode lun oct 9 14:49:25 CEST 2006
/*!
 \file   TCode.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mon oct 9th 14:49:25 CEST 2006
 
 \brief Implementation of the class URANIE::Launcher::TCode

 \par Modifications
\li Jean-Baptiste Blanchard 2016/04/08
 - Very large modification so that one can read vectors/strings.
 - Part of TLauncher code is transmitted here (more consistent with every one role)
 \li M. Salvoldi (Areva/OCC) 2011/09/28
 - suppress IFDEF WIN32 in init(),
 - replace Form("%s/%s",....) by gSystem->COncatFileName(...) in init() and addInputFile
 - add IFDEF WIN32 to replace "ls -l" by "dir" on Window OS
 \li F. Gaudier (2010-12-14 11:38:09)
 - resolve the DI111 and DI113
 \li F. Cedric Forgues (C.S.) 2010-10-05: Comments on the warning exception noticing that a working directory doesn't exist. Now a working directory can be specified with on ly a string without any sub directory: you can do setWorkingDirectory("toto") not setWorkingDirectory("toto/toto2")
 \li F. Gaudier 2010-05-20
 Put the recursive option to kTRUE when create the working directory in the init method
 throw and catch two exceptions
 \li Fabrice Gaudier ()
 Store the index of simulation in the vector of output in the postTraitment method.
 \li Cedric Forgues (C.S.) 2010/04/08
 The default working directory (when not specified by user) is now _sReferenceDirectory/URANIE so as to prevent the reference directoty to be deleted if setClean() method is used.
 \li Cedric Forgues (C.S.) 2009/12/08
 Inclusion of the _code->initFileType() method which is the first part of the init() method.
 *********************************************************************  
 */
// Test writting permisions
#include <iostream>
#include <sstream>
#include <cstdlib>
#include <string>
#include <algorithm>    // std::find

using std::cout;
using std::endl;
// ROOT
#include "TSystem.h"
#include "TLeaf.h"

// URANIE
#include "TCode.h"
#include "UExceptions.h"
#include "TInputFileRecreate.h"
#ifdef WITH_LIBXSLT
#include "TInputFileXML.h"
#endif
#ifdef WITH_FMI
#include "TInputFileFMU.h"
#endif

#include "TInputFileKey.h"
#include "TInputFileFlag.h"

#ifndef OLDCLUB
#include "TamponTexte.h"
#else
#include "club/TamponTexte.h"
#endif

ClassImp (URANIE::Launcher::TCode)

namespace URANIE {
namespace Launcher {

TCode::TCode() :
    _listOfInputFiles(NULL), _sWorkingDirectory(""), _sReferenceDirectory(""), _sCommand(""), 
    _nCurrentSimulation(1), _nOfInputToSubstitute(0), _nIndexOfInputToSubstitute(NULL), _blog(kFALSE)
{
    if (_blog)
	cout << "construit TCode" << endl;
    SetNameTitle("Uranie_Unknown_Code", "Unknown Code with uranie");
    _listOfInputFiles = new TList();

}
TCode::TCode(const TString &name, const TString &title) :
    _listOfInputFiles(NULL), _sWorkingDirectory(""), _sReferenceDirectory(""), _sCommand(""), _nCurrentSimulation(1),
      _nOfInputToSubstitute(0), _nIndexOfInputToSubstitute(NULL), _blog(kFALSE), _tds(NULL)
{
    if (_blog)
	cout << "construit TCode with name[" << name << "]" << endl;
    SetNameTitle(name, title);
    _listOfInputFiles = new TList();
}

// -----------------------------------------------------------------------
//  Constructor with a TDataServer
// -----------------------------------------------------------------------
TCode::TCode(URANIE::DataServer::TDataServer *tds, TString scmd) :
    _sWorkingDirectory(""), _sReferenceDirectory(gSystem->pwd()), _sCommand(scmd),
    _nCurrentSimulation(1), _nOfInputToSubstitute(0), _nIndexOfInputToSubstitute(NULL), _blog(kFALSE), _tds(tds)
{

    try {
        if ( !_tds )
            throw URANIE::Exceptions::UErrorExceptions(
                    __FILE__,
                    __LINE__,
                    Form("Method TCode::%s :\n The type TDS is empty.",
                         __func__));

        if (_blog)
            cout << endl << " ** Constructor::TCode with TDataServer["
                 << tds->GetName() << "] Command[" << _sCommand << "]" << endl;
        SetNameTitle(Form("code%s", tds->GetName()), tds->GetTitle());
        _listOfInputFiles = new TList();

        if (_blog)
        {
            cout << endl << " ** List the Input File " << endl;
            _listOfInputFiles->Print();
            cout << endl << " ** End Of Link the attributes with the Input File "
                 << endl;
        }

        if (_blog)
            cout << " ** End Of Constructor::TCode with TDataServer["
                 << tds->GetName() << "] Command[" << _sCommand << "]" << endl;
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

TCode::TCode(URANIE::Launcher::TCode &mycode):
    TNamed(mycode),
    _outputFiles(mycode._outputFiles), _listOfInputFiles(NULL), 
    _sWorkingDirectory(mycode._sWorkingDirectory), _sReferenceDirectory(mycode._sReferenceDirectory), 
    _sCommand(mycode._sCommand), _nCurrentSimulation(mycode._nCurrentSimulation),    
    _nOfInput(mycode._nOfInput), _nOfInputToSubstitute(mycode._nOfInputToSubstitute),
    _nIndexOfInputToSubstitute(NULL), _blog(mycode._blog), _tds(mycode._tds)
{

    if (_blog)
	cout << endl << " ** Copy Constructor::TCode with TCode[" << mycode.GetName() << "]" << endl;
    SetNameTitle(Form("copy_%s", mycode.GetName()), Form("copy_%s",mycode.GetTitle()) );
	
    //Creating the list of index to substitute
    if( mycode._nIndexOfInputToSubstitute != NULL )
    {
	_nIndexOfInputToSubstitute = new Int_t[_nOfInputToSubstitute];
	for(int iindex=0; iindex<_nOfInputToSubstitute; iindex++)
	    _nIndexOfInputToSubstitute[iindex] = mycode._nIndexOfInputToSubstitute[iindex];
    }
      	
    //Creating the list of input files
    _listOfInputFiles = new TList();
    if(mycode._listOfInputFiles!=NULL)
    {
	TInputFile *inputFileToAdd= NULL;
	TIter next(mycode._listOfInputFiles);
	while( (inputFileToAdd = (TInputFile *)next()) )
	{
	    if( dynamic_cast<TInputFileRecreate *>(inputFileToAdd) )
		_listOfInputFiles->Add( new TInputFileRecreate( inputFileToAdd->GetName() ) );
	    else if( dynamic_cast<TInputFileKey *>(inputFileToAdd) )
		_listOfInputFiles->Add( new TInputFileKey( inputFileToAdd->GetName() ) );
	    else if( dynamic_cast<TInputFileFlag *>(inputFileToAdd) )
		_listOfInputFiles->Add( new TInputFileFlag( inputFileToAdd->GetName() ) );
#ifdef WITH_LIBXSLT
	    else if( dynamic_cast<TInputFileXML *>(inputFileToAdd) )
		_listOfInputFiles->Add( new TInputFileXML( inputFileToAdd->GetName() ) );
#endif
	    for( int iatt=0; iatt<inputFileToAdd->getAttributesNumber(); iatt++)
		((TInputFile *)_listOfInputFiles->Last())->addAttribute( inputFileToAdd->getAttribute( iatt ) );
	}
    }
	
    if (_blog)
	cout << " ** End Of Constructor::TCode with TCode[" << mycode.GetName() << "] "<< endl;
}

  
// -----------------------------------------------------------------------
//  Destructor
// -----------------------------------------------------------------------
TCode::~TCode()
{
    if (_blog)
        cout << " *** ==============" << endl <<
                " *** Method [TCode::" << __func__  << "] Name[" << GetName() << "]" << endl;

    if (_nIndexOfInputToSubstitute != NULL)
    {
	delete[] _nIndexOfInputToSubstitute;
	_nIndexOfInputToSubstitute = NULL;
    }

    if(_listOfInputFiles!=NULL)
    {
	_listOfInputFiles->Delete();
	delete _listOfInputFiles;
    }
    _listOfInputFiles=NULL;

    if (_blog)
        cout << " *** End of Method [TCode::" << __func__  <<
                "] Name[" << GetName() << "]" << endl <<
                " *** ==============" << endl;
}

// ----------------------------------------------------------------------
// ----------------------------------------------------------------------
// Encapsulate a code
// ----------------------------------------------------------------------
// ----------------------------------------------------------------------
///////////////////////////////////////////////////////////
//  runTEST Method
///////////////////////////////////////////////////////////
void TCode::runTest(Option_t *option)
{
    TString soption = option ;
    soption.ToLower();
    if (_blog)
	cout << "TCode::runTest" << endl;
    //      init(option);

}

///////////////////////////////////////////////////////////
//  Inits the type of InputFiles
///////////////////////////////////////////////////////////
void TCode::initFileType(Option_t *option)
{
    FUNCT_TRACE();
    TString soption = option ;
    soption.ToLower();
    if (_blog)
        cout << endl << " ** TCode::initFileType _nCurrentSimulation[" << _nCurrentSimulation << "]"
	     << endl;
    
    // Find the good information between attributes and Input File
    if (_blog)
	cout << endl << " ** Link the attributes with the Input File " << endl;
    TList *listOfAttributes = _tds->getListOfAttributes();
    
    Int_t nx = listOfAttributes->GetEntries();
    if (_blog)
    {
        cout << endl << " ***  Attributes listing [" << nx << "]" << endl;
        listOfAttributes->Print();
        cout << " ***  End Of Attributes listing [" << nx << "]" << endl << endl;
    }
    
    try
    {
        URANIE::DataServer::TAttribute *att;
        for (Int_t i = 0; i < nx; i++)
	{
            att = (URANIE::DataServer::TAttribute *) (listOfAttributes->At(i));

    
            if (att->isOutput()) {
                if (_blog)
                    cout << " *** Case of Ouput Attribute[" << att->GetName() << "]" << endl;
                continue;
            }
            if (_blog)
                cout << " *** Case of input Attribute[" << att->GetName() << "]" << endl;
    
            TString sname = TString(att->GetName());
            list<URANIE::DataServer::TAttributeFileKey *> KeyList = att->getKeyList();

            if ( KeyList.size() == 0 ) {
                if (_blog)
                    cout << "   *** Case of input Attribute[" << att->GetName() << "] without KeyList" << endl;
                continue;
            }
            if (_blog)
                cout << "   *** Case of input Attribute[" << att->GetName() << "]  KeyList size[" << KeyList.size() << "]" << endl;

            for (list<URANIE::DataServer::TAttributeFileKey *>::iterator it =  KeyList.begin(); it != KeyList.end(); ++it) {
                TString sfile = (*it)->_sFile;
                TString sKey = (*it)->_sKey;
    
                if (sfile.Length() > 0)
		{
                    if (_blog) {
                        cout << "   *** Att " << i << ": Name[" << sname
			     << "], Key[" << sKey << "], File[" << sfile
			     << "]" << endl;
                        _listOfInputFiles->Print("all");
                    }
                    TInputFile *inputFileToAdd =
			dynamic_cast<TInputFile *>(_listOfInputFiles->FindObject(
						       sfile));
                    if (inputFileToAdd == NULL)
		    {
                        if (_blog) {
                            cout << "   *** InputFile[" << sfile << "] not found in the KeyList" << endl <<
                                    "     *** Goto Add it ..." << endl;
			}
			    
                        switch ((*it)->_sFileType)
			{
			case (URANIE::DataServer::TAttributeFileKey::kKey):
                            inputFileToAdd = new TInputFileKey(sfile);
                            break;
			case (URANIE::DataServer::TAttributeFileKey::kFlag):
                            inputFileToAdd = new TInputFileFlag(sfile);
                            break;
			case (URANIE::DataServer::TAttributeFileKey::kNewLine):
                            try
			    {
                                throw URANIE::Exceptions::UDeprecatedExceptions(
				    __FILE__, __LINE__, "2.1.10", "2.3.0",
				    "TCode", "initFileType",
				    "The URANIE::DataServer::TAttributeFileKey::kNewLine enum must be replaced by the URANIE::DataServer::TAttributeFileKey::kNewRow enum");
			    } catch (URANIE::Exceptions::UDeprecatedExceptions& ue)
			    {
                                ue.printMessage();
                                if (ue.getMessageType()
				    > URANIE::Exceptions::UExceptions::kWARNING)
				    exit(-1);
                                inputFileToAdd = new TInputFileRecreate(sfile);
			    }
                            break;
			case (URANIE::DataServer::TAttributeFileKey::kNewKey):
			case (URANIE::DataServer::TAttributeFileKey::kNewRow):
			case (URANIE::DataServer::TAttributeFileKey::kNewTDS):
			case (URANIE::DataServer::TAttributeFileKey::kNewColumn):
                            inputFileToAdd = new TInputFileRecreate(sfile);
			break;
			case (URANIE::DataServer::TAttributeFileKey::kXMLAttribute):
			case (URANIE::DataServer::TAttributeFileKey::kXMLField):
#ifdef WITH_LIBXSLT
                            inputFileToAdd = new TInputFileXML(sfile);
#else
                            cout << " *** ERROR :: LibXSLT not found, then class TInputFileXML not available.\n" << endl;
                            exit(-1);
#endif                  
			break;
                        // Case of FMI
                        case (URANIE::DataServer::TAttributeFileKey::kFMU):
                            if (_blog)
                                cout << "     *** case FMU[" << (*it)->_sFileType << "] with sfile[" << sfile << "]" << endl;
#ifdef WITH_FMI
                            inputFileToAdd = new TInputFileFMU(sfile);
#else
                            throw URANIE::Exceptions::UErrorExceptions(
                                        __FILE__, __LINE__,
                                        Form("TCode::%s Method\n the library \"FMILibrary\" not found, then the class TInputFileFMU is not available.",
                                             __func__));
#endif
                            break;

			default:
                            try
			    {
                                throw URANIE::Exceptions::UErrorExceptions(
				    __FILE__, __LINE__,
				    Form(
					"TCode::initFileType Method\n The input file enum [%d] is not yet recognized",
					(*it)->_sFileType));

			    } catch (URANIE::Exceptions::UErrorExceptions& ue)
			    {
                                ue.printMessage();
				delete listOfAttributes;
				listOfAttributes=NULL;
                                exit(-1);
			    }
			}
			    
                        _listOfInputFiles->Add(inputFileToAdd);
			    
		    }
                    else
		    {
                        if (_blog)
                            cout << "   *** InputFile[" << sfile << "] found in the list" << endl;
		    }
                    // Add the attribute in the inputfile
                    if (_blog)
			cout << " ************************** Add attribute ["
			     << att->GetName() << "] in the InputFile["
			     << inputFileToAdd->GetName() << "]" << endl;

#ifndef OLDCLUB		    
		    try
		    {
			inputFileToAdd->addAttribute(att);
		    } catch (TamponTexteError& tte)
		    {
			delete listOfAttributes;
			listOfAttributes=NULL;
			throw tte;
		    }
#else 
            inputFileToAdd->addAttribute(att);
#endif
		}
	    }
            if (_blog)
		att->printLog();
	}
    } catch (URANIE::Exceptions::UExceptions& ue)
    {
	    
        if (ue.getMessageType() == URANIE::Exceptions::UExceptions::kDEBUG)
	    ue.printMessage();
        if (ue.getMessageType() > URANIE::Exceptions::UExceptions::kWARNING)
	{
            ue.printMessage();
            exit(-1);
	}
    }

    if (_blog) {
        cout << endl << " *** End of list of attributes[" << nx << "]" << endl;
        _listOfInputFiles->Print();
        for (Int_t ifile = 0; ifile < _listOfInputFiles->GetSize(); ifile++) {
            cout << " *** printLog of the ifile(" << ifile << "/" << _listOfInputFiles->GetSize() << ")" << endl;
            ((TInputFile *) (_listOfInputFiles->At(ifile)))->printLog();
        }
    }
    
    // Create the tuple if necessary
    if (!_tds->hasTuple())
        _tds->createTuple();
    
    if (_blog)
        _tds->getTuple()->Print();
    
    // Fill the vector of _nIndexOfInputToSubstitute vector
    TObject *crtobj = NULL;
    TList *listOfInputToSubstitue = new TList();
    TList *listOfInput = new TList();
    for (Int_t ifile = 0; ifile < getNInputFiles(); ifile++)
    {
	TString NAMES="", TYPES="";
        for (Int_t ivar = 0; ivar < getNInputAttributes(ifile); ivar++)
	{
	        
            TString sname = getInputAttribute(ifile, ivar)->GetName();
	    list<URANIE::DataServer::TAttributeFileKey *> KeyList = getInputAttribute(ifile, ivar)->getKeyList();
            list<URANIE::DataServer::TAttributeFileKey *>::iterator it =  KeyList.begin(); 
	    if(ivar==0 && (*it)->_sFileType == URANIE::DataServer::TAttributeFileKey::kNewTDS )
	    {
		NAMES="#COLUMN_NAMES: "+sname;
		TYPES="#COLUMN_TYPES: "+getInputAttribute(ifile, ivar)->getDataTypeStr();
	    }
	    else if ( (*it)->_sFileType == URANIE::DataServer::TAttributeFileKey::kNewTDS )
	    {
		NAMES+="|"+sname;
		TYPES+="|"+getInputAttribute(ifile, ivar)->getDataTypeStr();
	    }
	    
            if (_blog)
		cout << " ** input [" << ifile << "," << ivar << "] :: sname["
		     << sname << "] -------";
            crtobj = dynamic_cast<TObject *>(listOfInput->FindObject(sname));
            // Inputs to substitute
            listOfInputToSubstitue->Add((TObject*) (_tds->getTuple()->GetLeaf(sname.Data())));
            _nOfInputToSubstitute++;
	        
            // Inputs to save
            if (crtobj == NULL)
	    {
                if (_blog)
		    cout << " ** [ not  found ]";
                listOfInput->Add( (TObject*) (_tds->getTuple()->GetLeaf(sname.Data())));
	    }
            else
	    {
                if (_blog)
		    cout << " -- found in input";
	    }
            if (_blog)
		cout << " :: nOfInputToSubstitue[" << _nOfInputToSubstitute
		     << "]" << endl << endl;
	}

	if( NAMES != "" )
	{
	    TInputFileRecreate *tinobj = dynamic_cast<TInputFileRecreate *>( getInputFile(ifile) );
	    if( tinobj != NULL)
		tinobj->setHeader( NAMES+"\n"+TYPES+"\n" );
	}
    }
        
    if (_blog)
    {
        cout << " ** Input Attributes number[" << listOfInput->GetEntries()
	     << "] " << listOfInput << endl;
        listOfInput->Print();
        cout << " ** InputToSubstitue Attributes number["
	     << listOfInputToSubstitue->GetEntries()
	     << "] nOfInputToSubstitue[" << _nOfInputToSubstitute << "]"
	     << endl;
        listOfInputToSubstitue->Print();
        cout << endl;
    }
    

    // Auxiliary attributes
    if (_blog)
	cout << " ** Auxiliary Attributes :: Search from the ["
	     << _tds->getNAttributes() << "] attributes" << endl;
    TList *listOfAuxiliaryAttributes = new TList();
    TIter next(listOfAttributes);
    Int_t nauxcrt = 0;
        
    while (URANIE::DataServer::TAttribute *att =
	   (URANIE::DataServer::TAttribute *) next())
    {
        if (_blog)
	    cout << " ** aux [" << nauxcrt << "] :: " << att->GetName()
		 << " -------";
        TObject *auxatt = dynamic_cast<TObject *>(listOfInput->FindObject(
						      att->GetName()));
	    
        if (auxatt == NULL)
	{
            if (_blog)
		cout << " ** [not found]" << endl;
            listOfAuxiliaryAttributes->Add(
		(TObject*) (_tds->getTuple()->GetLeaf(att->GetName())));

	}
        else
	{
            if (_blog)
		cout << " -- found in input" << endl;
	}
        nauxcrt++;
    }
    
    _nOfInput = listOfInput->GetEntries();
    
    //Clean////
    if(_nIndexOfInputToSubstitute != NULL)
    {
        delete[] _nIndexOfInputToSubstitute;
        _nIndexOfInputToSubstitute = NULL;
    }
    //////////

    
    // Fill the index of attributes to substitute
    // only in case of the two list of attributes is different
    _nIndexOfInputToSubstitute = new Int_t[_nOfInputToSubstitute];
    
    TIter nextInputToSubstitue(listOfInputToSubstitue);
    Int_t niter = 0;
    
    while (TNamed *obj = (TNamed *) nextInputToSubstitue())
    {
	if (_blog)
	    obj->Print();
	Int_t nIndexOfAttribute = listOfInput->IndexOf(obj);
	if (_blog)
	    cout << " ** Attribute ToSubstitue [" << obj->GetName()
		 << "] nIndexOfAttribute[" << nIndexOfAttribute << "]"
		 << endl;
	_nIndexOfInputToSubstitute[niter++] = nIndexOfAttribute + 1;
    }
    
    if (_blog)
    {
	cout << endl << " ** nIndexOfInputToSubstitute vector **" << endl;
	for (Int_t i = 0; i < niter; i++)
	    cout << _nIndexOfInputToSubstitute[i] << " ";
	cout << endl;
	cout<<"List of input to substitute "<<endl;
	listOfInputToSubstitue->Print();
	cout<<"List of input "<<endl;
	listOfInput->Print();
    }

    // Clean
    delete listOfAttributes;
    listOfAttributes=NULL;
    delete listOfInputToSubstitue;
    listOfInputToSubstitue = NULL;
    delete listOfInput;
    listOfInput = NULL;
    delete listOfAuxiliaryAttributes;
    listOfAuxiliaryAttributes = NULL;

    if (_blog)
        cout << endl <<
                " *** End of Method [TCode::" << __func__  << "]" <<
                " _nCurrentSimulation[" << _nCurrentSimulation << "]" << endl <<
                " ********************************" << endl;


}

///////////////////////////////////////////////////////////
//  Init Method
///////////////////////////////////////////////////////////
void TCode::init(Option_t *option)
{
    TString soption = option ;
    soption.ToLower();
    FUNCT_TRACE();
    if (_blog)
        cout << endl << " ** TCode::init[" << _nCurrentSimulation << "]" << endl;

    // Find the good information between attributes and Input File
    if (_blog)
	cout << endl << " ** Link the attributes with the Input File " << endl;
    TList *listOfAttributes = _tds->getListOfAttributes();

    Int_t nx = listOfAttributes->GetEntries();
    if (_blog)
    {
        cout << endl << " **  Attributes listing [" << nx << "]" << endl;
        listOfAttributes->Print();
        cout << endl << " **  End Of Attributes listing [" << nx << "]" << endl;
    }

    // Places in the good directory
    char *foo=NULL; //buffer used to prevent memleak, as ConcatFileNAme result must be deleted
    if (_sWorkingDirectory.Length() == 0)
    {
	foo = gSystem->ConcatFileName(_sReferenceDirectory.Data(), "URANIE");
	_sWorkingDirectory=foo;
	delete[] foo; foo=NULL;
    }
    else
	// TODO: CFO 2010/10/05
	// Ce if permet de rajouter un working directory en relatif mais ne peut construire un repertoire avec
	// sous-repertoire (type toto1/toto2 fourni en argument)
	if (!gSystem->IsAbsoluteFileName(_sWorkingDirectory.Data()))
	{
	    foo = gSystem->ConcatFileName(_sReferenceDirectory.Data(), _sWorkingDirectory.Data());
	    _sWorkingDirectory = foo;
	    delete[] foo; foo=NULL;
	}

    // Creates the working directory if it doesn't exist
    try
    {
        TString sdir = _sWorkingDirectory;
        void * openFileResult = gSystem->OpenDirectory(sdir.Data());
        try
	{
            if (openFileResult == 0)
	    {
                // TODO: CFO 2010/10/05: le warning a deja provoque un crash sous WIN32
                // A investiguer pour peut-etre trouver une nouvelle facon d'ecrire cette exception.
                // cout<<Form("** TCode::init Method\n The directory \"%s\" does not exist\n URANIE creates it for you",_sWorkingDirectory.Data())<<endl;
                throw URANIE::Exceptions::UInfoExceptions(__FILE__, __LINE__,
							  Form(
							      "TCode::init Method\n The working directory \"%s\" does not exist\n URANIE creates it for you",
							      sdir.Data()));
	    }
	} catch (URANIE::Exceptions::UExceptions& ue)
	{
            ue.printMessage();
            if (gSystem->mkdir(sdir, kTRUE) == -1)
		throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							   Form(
							       "TCode::init ERROR : the directory \"%s\" can not be created",
							       sdir.Data()));
	}
        gSystem->FreeDirectory(openFileResult);
        gSystem->cd(sdir);

        // Creates the "UranieResult" directory if it doesn't exist
#ifdef WIN32
        sdir = Form("%s\\UranieResults", _sWorkingDirectory.Data());
#else
        sdir = Form("%s/UranieResults", _sWorkingDirectory.Data());
#endif
        openFileResult = gSystem->OpenDirectory(sdir);
        try
	{
            if (openFileResult == 0)
	    {
                throw URANIE::Exceptions::UInfoExceptions(__FILE__, __LINE__,
							  Form(
							      "TCode::init Method\n The result directory \"%s\" does not exist\n URANIE creates it for you",
							      sdir.Data()));
	    }
	} catch (URANIE::Exceptions::UExceptions& ue)
	{
            ue.printMessage();
            if (gSystem->mkdir(sdir, kTRUE) == -1)
		throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							   Form(
							       "TCode::init ERROR : the result directory \"%s\" can not be created",
							       sdir.Data()));
	}
        gSystem->FreeDirectory(openFileResult);

        // Creates the launching "UranieLauncher_i" directory if it doesn't exist
#ifdef WIN32
        sdir = Form("%s\\UranieLauncher_%d", _sWorkingDirectory.Data(), _nCurrentSimulation);
#else
        sdir = Form("%s/UranieLauncher_%d", _sWorkingDirectory.Data(),
		    _nCurrentSimulation);
#endif
        openFileResult = gSystem->OpenDirectory(sdir);
        try
	{
            if (openFileResult == 0)
	    {
                throw URANIE::Exceptions::UInfoExceptions(__FILE__, __LINE__,
							  Form(
							      "TCode::init Method\n The launching directory \"%s\" does not exist\n URANIE creates it for you",
							      sdir.Data()));
	    }
	} catch (URANIE::Exceptions::UExceptions& ue)
	{
            if (_nCurrentSimulation == 1)
		ue.printMessage();
            if (gSystem->mkdir(sdir, kTRUE) == -1)
		throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							   Form(
							       "TCode::init ERROR : the result directory \"%s\" can not be created",
							       sdir.Data()));
	}
        gSystem->FreeDirectory(openFileResult);
        gSystem->cd(sdir);

    } catch (URANIE::Exceptions::UExceptions& ue)
    {
        ue.printMessage();
        if (ue.getMessageType() > URANIE::Exceptions::UExceptions::kWARNING)
	    exit(-1);
    }

    if (_blog)
    {
        cout << endl << "*****************************" << endl
	     << "*** Copy Input File" << endl;
        cout << "*** Current Directory [" << gSystem->pwd() << "]" << endl;
    }
    Bool_t bcopy;

    Int_t ninputfiles = (Int_t)(_listOfInputFiles->GetEntries());
    if (_blog)
    {
        cout << endl << " **  List Of Input Files [" << ninputfiles << "]"
	     << endl;
        _listOfInputFiles->Print();
        for (Int_t ifile = 0; ifile < ninputfiles; ifile++)
	    ((TInputFile *) (_listOfInputFiles->At(ifile)))->printLog();
        cout << endl << " **  End Of List Of Input Files [" << ninputfiles
	     << "]" << endl;
    }

    try
    {

        if (_blog)
	    cout << " *** Number of Input Files[" << ninputfiles << "]" << endl;
        for (Int_t indx = 0; indx < ninputfiles; indx++)
	{
            bcopy = kTRUE;
            TString sfile = getInputFile(indx)->getFileName();
            Int_t natt = getInputFile(indx)->getAttributesNumber();

            //C.F. 2009/10/08 NEW 2.0.1
            //If the file was given with a relative path, reference directory is set as dirname. 
            if (gSystem->DirName(sfile.Data()) == TString("."))
	    {
		foo = gSystem->ConcatFileName(_sReferenceDirectory.Data(), sfile.Data());
		sfile = foo;
		delete[] foo; foo=NULL;
	    }

            TString sfilename = gSystem->BaseName(sfile.Data());
            if (_blog)
		cout << " *** input File(" << indx + 1 << "/" << ninputfiles
		     << ") sfilename[" << sfilename << "] natt[" << natt
		     << "] sfile[" << sfile << "]" << endl;
            if (natt > 0)
	    {
                ////////////////////////////////////////
                // FGA::MODIFICATION 2010-12-13 10:10:02
                // Test non pertinent
                //  remplacer par le dynamic cast
                ////////////////////////////////////////
		//           list < URANIE::DataServer::TAttributeFileKey *> KeyList = getInputFile(indx)->getAttribute(0)->getKeyList();
		//           list< URANIE::DataServer::TAttributeFileKey *>::iterator first = KeyList.begin();      
		//           if ( ((*first)->_sFileType == URANIE::DataServer::TAttributeFileKey::kNewKey) ||
		//            ((*first)->_sFileType == URANIE::DataServer::TAttributeFileKey::kNewRow) ||
		//            ((*first)->_sFileType == URANIE::DataServer::TAttributeFileKey::kNewColumn) ) {
		//         cout << " *********** kNewKey || kNewRow || kNewColumn ************" << endl;
		//         bcopy = kFALSE;
		//           }
                TInputFileRecreate *typeOfInputFile =
		    dynamic_cast<TInputFileRecreate *>(getInputFile(indx));

                if (typeOfInputFile)
		    bcopy = kFALSE;
                ////////////////////////////////////////
                // End Of FGA::MODIFICATION 2010-12-13 10:10:02
                ////////////////////////////////////////
	    }

            // Case of Copy
            if (bcopy)
	    {
                if (_blog)
		    cout << "  *** Copy input sfile[" << sfile << "] to[" << sfilename << "]" << endl;

                Int_t nret = gSystem->CopyFile(sfile, sfilename, kTRUE);
                if (nret == -1)
		    throw URANIE::Exceptions::UExceptions(__FILE__, __LINE__,
							  Form("TCode::init ERROR : for copy file \"%s\"",
							       sfile.Data()));
	    }
            else
	    {
                if (_blog)
		    cout << "  *** NO Copy input File[" << sfilename << "]"
			 << endl;

	    }

            //       if (  _blog )
            //         gSystem->Exec(Form("more %s", sfilename.Data()));
	}

    } catch (URANIE::Exceptions::UExceptions& ue)
    {
        ue.printMessage();
        if (ue.getMessageType() > URANIE::Exceptions::UExceptions::kWARNING)
	{
	    delete listOfAttributes;
	    listOfAttributes=NULL;
	    exit(-1);
	}
    }

    if (_blog)
    {
        cout << "*** End Of Copy Input File" << endl << "***" << endl;
        cout << "*** Listing of Working Directory [" << gSystem->pwd() << "]"
	     << endl;
#ifdef WIN32
        gSystem->Exec("dir");
#else
        gSystem->Exec("ls -al");
#endif
        cout << "*** End of Listing of Working Directory" << endl
	     << "*****************************" << endl << endl;
    }

    delete listOfAttributes;
    listOfAttributes=NULL;
}


// ----------------------------------------------------------------------
//  PRE-TRAITMENT
// ----------------------------------------------------------------------
void TCode::preTraitment(Int_t n, Double_t *val, Int_t *index, Option_t *option)
{

    UNUSED(n, index);//never used in function but called by methods
    TString soption = option;
    soption.ToLower();
    if (_blog)
        cout << endl << "TCode::preTraitment[" << _nCurrentSimulation << "]"
	     << endl;
    _nCurrentSimulation++;
    
    // Gets the current value of the sample
    // The first value is the index \todo to modify.
    // fga::note 2007-07-05 13:43
    // With Launcher this index must be equal beacause there a problems with a design.
    // While the optimisation is performed, the information has to be known 
    // fin de fga::note
    Int_t ncrt = 1;
    // Runs the input list to replace the values
    Int_t ninput = (Int_t)(_listOfInputFiles->GetEntries());
    for (Int_t indx = 0; indx < ninput; indx++)
    {
        TString sfile = getInputFile(indx)->getFileName();
        if (_blog)
            cout << "TTCode::preTraitment sfile[" << sfile << "] #["
		 << getInputFile(indx)->getAttributesNumber() << "]" << endl;
        TString sfilename = gSystem->BaseName(sfile.Data());
        if (_blog)
            cout << "TTCode::preTraitment sfilename[" << sfilename << "]"
		 << endl;

        // Recherche des indices des champs dans le fichier d'entree,
        // OU (depuis 2.0.1) ecrase des fichiers on the fly
        if (getInputFile(indx)->getAttributesNumber() > 0)
        {
            if (_blog)
		getInputFile(indx)->printLog();

            getInputFile(indx)->substitue(ncrt, val, _nIndexOfInputToSubstitute);
            // Increment the index ncrt by the number of uncertain attribut in the current input file
        }

    }

    if (_blog)
        cout << endl << "Enf of TCode::preTreatment[" << _nCurrentSimulation
	     << "]" << endl;

}


// void TCode::UpdateInputList()
// {
//   // _listOfInput->Clear();
//   for (Int_t ifile = 0; ifile < getNInputFiles(); ifile++)
//     for (Int_t ivar = 0; ivar < getNInputAttributes(ifile); ivar++)
// 	{
// 	  TString sname = getInputAttribute(ifile, ivar)->GetName();
// 	  _listOfInput->Add( (TObject*) (_tds->getTuple()->GetLeaf(sname.Data())));	  
// 	}
// }
  
// ----------------------------------------------------------------------
//  PRE-TRAITMENT introduced to cope with vectors/strings
// ----------------------------------------------------------------------
void TCode::preTraitment(Int_t n, vector<URANIE::DataServer::UEntry*> *in, Option_t *option)
{


    TString soption = option ;
    soption.ToLower();
    if (_blog)
	cout << endl << "TCode::preTraitment[" << _nCurrentSimulation << "]"
	     << endl;
    _nCurrentSimulation++; //Needed to create the correct output folder

    // Create the input file with correct values
    for(vector<URANIE::DataServer::UEntry*>::iterator it=in->begin(); it !=in->end(); it++)
	(*it)->clearEntry();
    
    Int_t icrt = 0;
    in->at(icrt++)->pushEntry(1+n);
    
    vector<TString> names;
    for (Int_t ifile = 0; ifile < getNInputFiles(); ifile++)
	for (Int_t ivar = 0; ivar < getNInputAttributes(ifile); ivar++)
	{
	    TString sname = getInputAttribute(ifile, ivar)->GetName();
	    if( find(names.begin(), names.end(), sname ) != names.end() )
		continue;
	    in->at(icrt++)->assignEntry( (void*)(((TLeaf*) (_tds->getTuple()->GetLeaf(sname.Data())))->GetValuePointer() ) );
	    names.push_back(sname);
	}
    icrt = 1;

    // Gets the current value of the sample
    // The first value is the index \todo to modify.
    // fga::note 2007-07-05 13:43
    // With Launcher this index must be equal beacause there a problems with a design.
    // While the optimisation is performed, the information has to be known 
    // fin de fga::note

    // Runs the input list to replace the values
    Int_t ninput = (Int_t)(_listOfInputFiles->GetEntries());
    int BegAtt=0;
    for (Int_t indx = 0; indx < ninput; indx++)
    {
	icrt = 1;
        TString sfile = getInputFile(indx)->getFileName();
	if (_blog)
	    cout << "TTCode::preTraitment sfile[" << sfile << "] #["   << getInputFile(indx)->getAttributesNumber() << "]" << endl;
        TString sfilename = gSystem->BaseName(sfile.Data());
	if (_blog)
	    cout << "TTCode::preTraitment sfilename[" << sfilename << "]"   << endl;
        // Recherche des indices des champs dans le fichier d'entree,
        // OU (depuis 2.0.1) ecrase des fichiers on the fly
        if (getInputFile(indx)->getAttributesNumber() > 0)
	{
	    if (_blog)
	    {
		getInputFile(indx)->printLog();
		XYZ<<" for file "<<sfilename<<" attributenumber is "<<getInputFile(indx)->getAttributesNumber()<<" icrt "<<icrt<<endl;
	    }
	    
            getInputFile(indx)->substitue(icrt, in, ((_nIndexOfInputToSubstitute!=NULL) ? (_nIndexOfInputToSubstitute+BegAtt) : _nIndexOfInputToSubstitute ));
            BegAtt+=getInputFile(indx)->getAttributesNumber();

	}

    }

    if (_blog)
        cout << endl << "End of TCode::preTreatment[" << _nCurrentSimulation << "]" << endl;
}

// ----------------------------------------------------------------------
//  RUN
// ----------------------------------------------------------------------
void TCode::run(Option_t *option)
{
    TString soption = option ;
    soption.ToLower();
    if (_blog)
	cout << "TCode::run [" << _nCurrentSimulation << "]" << endl;

    TString scommand = _sCommand;

    if (scommand.Contains("%D") == kTRUE)
    {
        TString scurrentdirectory = gSystem->pwd();
        scommand.ReplaceAll("%D", scurrentdirectory.Data());
    }

    if (_blog)
	cout << "TCode::run cmd [" << scommand << "]" << endl;

    ////////////////////////////////////////
    // VBE::PROPOSITION 2011-09-28 09:13:56
    // use execvp to launch the code instead of gSystem::Exec method
    //  DI 141
    ////////////////////////////////////////
    // std::vector<std::string> elems;
    // std::stringstream ss(scommand.Data());
    // std::string item;
    // while(std::getline(ss, item,' ')) {
    //   elems.push_back(item);
    // }
    // char**argtable=new char*[elems.size()+1];
    // for (Int_t i=0; i<(Int_t)elems.size(); i++)
    //   argtable[i]=const_cast<char*>(elems[i].c_str());
    //
    // argtable[elems.size()]=NULL;
    // execvp (argtable[0],argtable);
    // delete[] argtable;
    ////////////////////////////////////////
    // End Of VBE::PROPOSITION 2011-09-28 09:13:56
    ////////////////////////////////////////

    gSystem->Exec(scommand.Data());

}


// ----------------------------------------------------------------------
//  POST-TRAITMENT introduced to cope with vectors/strings
// ----------------------------------------------------------------------
Bool_t TCode::postTraitment(vector<URANIE::DataServer::UEntry*> *out, Option_t *option)
{
    TString soption = option ;
    soption.ToLower();
    if (_blog)
	cout << "TCode::postTraitment" << endl;

    // Create the input file with correct values
    for(vector<URANIE::DataServer::UEntry*>::iterator it=out->begin(); it !=out->end(); it++)
	(*it)->clearEntry();
    
    // Run the input list to replace the values
    Int_t noutput = (Int_t)(_outputFiles.size());
    Bool_t bFailure = kTRUE;
    // Fill the current number of simulation
    // loop on the output files
    int shift=2;
    for (Int_t indy = 0; indy < noutput; indy++)
    {

        if (_blog)
	{
            cout << " ** TCode::postTraitment[" << indy << "/" << noutput << "]"
		 << endl;
            _outputFiles[indy]->setLog();
	}
        TString sfile = getOutputFile(indy)->getFileName();
        if (_blog)
	    cout << " ** TCode::postTraitment[" << indy << "/" << noutput
		 << "] sfile[" << sfile << "]" << endl;

	vector<URANIE::DataServer::UEntry*> pointer(out->begin()+shift, out->end()); 
        if (_outputFiles[indy]->searchOutput(&pointer) == kFALSE)
	    bFailure = kFALSE;
	
	shift+=_outputFiles[indy]->getAttributesNumber();
    }

    out->at(1)->pushEntry(_nCurrentSimulation);
    
    // stringstream sstr; sstr <<  _nCurrentSimulation << " ";
    // if( sout->count("ncalc")==0 )
    // 	sout->insert( std::pair<string,string>("ncalc",sstr.str()) );
    // else
    // 	(*sout)["ncalc"]=sstr.str();
    
    return bFailure;

}

// ----------------------------------------------------------------------
//  POST-TRAITMENT
// ----------------------------------------------------------------------
Bool_t TCode::postTraitment(Int_t &nout, Double_t *dvalout, Option_t *option,  TList *lstOut)
{
    TString soption = option ;
    soption.ToLower();
    if (_blog)
	cout << "TCode::postTraitment" << endl;
    // Run the input list to replace the values
    Int_t noutput = (Int_t)(_outputFiles.size());
    Bool_t bFailure = kTRUE;
    // Fill the current number of simulation
    if (nout == 0)
	dvalout[nout++] = 1.0 * _nCurrentSimulation;
    // loop on the output files
    for (Int_t indy = 0; indy < noutput; indy++)
    {

        if (_blog)
	{
            cout << " ** TCode::postTraitment[" << indy << "/" << noutput << "]"
		 << endl;
            _outputFiles[indy]->setLog();
	}
        TString sfile = getOutputFile(indy)->getFileName();
        if (_blog)
	    cout << " ** TCode::postTraitment[" << indy << "/" << noutput
		 << "] sfile[" << sfile << "]" << endl;

        if (_outputFiles[indy]->searchOutput(nout, dvalout, lstOut) == kFALSE)
	    bFailure = kFALSE;
    }

    return bFailure;

}

void TCode::terminate(Option_t *option)
{
    TString soption = option ;
    soption.ToLower();
    if (_blog)
	cout << "TCode::terminate" << endl;
    gSystem->cd("..");
    //gSystem->Exec("ls -alR");
}

// ----------------------------------------------------------------------
// Adding files
// ----------------------------------------------------------------------
void TCode::addInputFile(TInputFile *ifile)
{
    TInputFile *inputfile = dynamic_cast<TInputFile *>(_listOfInputFiles->FindObject(ifile->GetName()));

    // if it does not yet exist in the list, we add it
    if (!inputfile)
        _listOfInputFiles->Add(ifile);
}

void TCode::addInputFile(TString str)
{
    if (_sReferenceDirectory.Length() == 0)
        _sReferenceDirectory = gSystem->pwd();

    // \todo Enlever TInputFile en abstraite pure pour le mettre en virtuelle.
    // feature : juste sert pour copier
    TInputFile *ifile;
    if (gSystem->IsAbsoluteFileName(str.Data())) {
        ifile = new TInputFileRecreate(str.Data());
    } else {
        char *sFile = gSystem->ConcatFileName(_sReferenceDirectory.Data(), str.Data());
        ifile = new TInputFileRecreate(sFile);
        delete[] sFile;
    }

    // The file must be exist
    try{
        if ( !ifile->isExist() )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TCode::addInputFile ERROR\n The file [%s] set at [%s] not found", str.Data(), ifile->getFileName().Data()));
    }   catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        delete ifile; ifile = nullptr;
        throw ue;
    }

    _listOfInputFiles->Add(ifile);
}

void TCode::addOutputFile(TOutputFile *outputfile)
{
    if (_blog)
	cout << " ** TCode::addOutputFile [" << outputfile->getFileName()
	     << "] natt[" << outputfile->getAttributesNumber() << "]"
	     << endl;

    // Set the "Output" flag for each attributes
    for (Int_t iatt = 0; iatt < outputfile->getAttributesNumber(); iatt++)
    {
	
	try
        {
	    if (!(_tds->isAttribute(outputfile->getAttribute(iatt)->GetName())))
           {
	       outputfile->getAttribute(iatt)->setOutput();
           }
	    else
           {
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form("TCode::addOutputFile[%s] Exist already an input attribute with the same name '%s'",
                            outputfile->GetName(),outputfile->getAttribute(iatt)->GetName()));
           }
        } catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
        ue.printMessage();
        throw ue;
        exit(-1);
        }        
    }

    _outputFiles.push_back(outputfile);
    if (_blog)
	cout << " ** End of TCode::addOutputFile [" << outputfile->getFileName()
	     << "]" << endl;
}

Int_t TCode::getNOutputAttributes()
{
    Int_t nvar = 0;
    //output
    Int_t noutput = (Int_t)(_outputFiles.size());
    for (Int_t indx = 0; indx < noutput; indx++)
    {
        nvar += _outputFiles[indx]->getAttributesNumber();
    }
    return nvar;

}
Int_t TCode::getNInputAttributes()
{
    Int_t nvar = 0;
    Int_t ninput = (Int_t)(_listOfInputFiles->GetEntries());
    for (Int_t indx = 0; indx < ninput; indx++)
    {
        nvar += getInputFile(indx)->getAttributesNumber();
    }
    return nvar;

}
Int_t TCode::getNInputAttributes(Int_t nfile)
{

    Int_t nvar = 0;
    if (nfile < (Int_t)(_listOfInputFiles->GetEntries()))
	nvar += getInputFile(nfile)->getAttributesNumber();
    return nvar;

}

Int_t TCode::getNOutputAttributes(Int_t nfile)
{

    Int_t nvar = 0;
    if (nfile < (Int_t)(_outputFiles.size()))
	nvar += _outputFiles[nfile]->getAttributesNumber();
    return nvar;

}

URANIE::DataServer::TAttribute * TCode::getAttribute(Int_t nfile, Int_t natt)
{
    URANIE::DataServer::TAttribute *att = 0;
    if (nfile < (Int_t)(_outputFiles.size()))
	att = _outputFiles[nfile]->getAttribute(natt);
    return att;

}
// ----------------------------------------------------------------------
// Get input or output files
// ----------------------------------------------------------------------
TInputFile * TCode::getInputFile(TString sfile)
{
    TInputFile * inputfile = NULL;
    try
    {
        inputfile = dynamic_cast<TInputFile *>(_listOfInputFiles->FindObject(
						   sfile));
        if (_blog)
	    _listOfInputFiles->Print("all");
        if (inputfile)
	    return inputfile;
        else
	{
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 Form(
							     "TCode::getInputFile ERROR: InputFile %s doesn't seem to exist",
							     sfile.Data()));
	}
    } catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
        return inputfile;
    }
}

TInputFile * TCode::getInputFile(Int_t nfile)
{

    try
    {
        Int_t nsize = (Int_t)(_listOfInputFiles->GetSize());
        if (nfile > nsize)
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form(
							   "TCode::getInputFile ERROR\n The number [%d] is too large: The max index is[%d]",
							   nfile, nsize));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    return dynamic_cast<TInputFile *>(_listOfInputFiles->At(nfile));

}

TInputFile * TCode::getLastInputFile()
{
    Int_t nsize = -1;
    try {
        if ( _listOfInputFiles == nullptr )
            _listOfInputFiles = new TList();
        nsize = _listOfInputFiles->GetEntries();
        cout << " *** nsize[" << nsize << "]" << endl;
        if ( nsize == 0 )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__, "TCode::getLastInputFile ERROR\n The list of Input Files is empty");
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    return dynamic_cast<TInputFile *>(_listOfInputFiles->At(nsize-1));
}


// ----------------------------------------------------------------------
// Get output files from name
// ----------------------------------------------------------------------
TOutputFile * TCode::getOutputFile(TString sfile)
{
    FUNCT_TRACE();
    if (_blog)
        cout << " *** Method TCode::" << __func__ << " with sfile[" << sfile << "]" << endl;

    TOutputFile * outputfile = NULL;
    try
    {
        size_t nout = _outputFiles.size();
        if (_blog)
            cout << " *** nout(" << nout << ")" << endl;

        if (nout == 0)
            return outputfile;

        for(size_t i=0; i<nout; i++) {
            outputfile = _outputFiles[i];
            if ( outputfile->getFileName () == sfile) {
                return outputfile;
            }
        }
        return outputfile;
    } catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
        return outputfile;
    }
}


TOutputFile * TCode::getOutputFile(Int_t nfile)
{
    try
    {
        Int_t nsize = (Int_t)(_outputFiles.size());

        if (nfile > nsize)
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form(
							   "TCode::getOutputFile ERROR\n The number [%d] is too large: The max index is[%d]",
							   nfile, nsize));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    return dynamic_cast<TOutputFile *>(_outputFiles[nfile]);
}

URANIE::DataServer::TAttribute * TCode::getInputAttribute(Int_t nfile,
							  Int_t natt)
{
    URANIE::DataServer::TAttribute *att = 0;
    if (nfile < (Int_t)(_listOfInputFiles->GetEntries()))
	att = getInputFile(nfile)->getAttribute(natt);
    return att;

}

void TCode::setWorkingDirectory(TString sPathName)
{
    if (_blog)
        cout << " *** Method TCode::" << __func__ << " with sPathName[" << sPathName << "]" << endl;

    // buffer used to prevent memleak, as gSystem::ConcatFileName result must be deleted
    char *foo=NULL;

    if ( sPathName.Length() == 0 ) {
        foo = gSystem->ConcatFileName(_sReferenceDirectory.Data(), "URANIE");
        _sWorkingDirectory=foo;
        delete[] foo; foo=NULL;
    } else { if (gSystem->IsAbsoluteFileName(sPathName.Data())) {
            _sWorkingDirectory = sPathName;
        } else {
            foo = gSystem->ConcatFileName(_sReferenceDirectory.Data(), sPathName.Data());
            _sWorkingDirectory = foo;
            delete[] foo; foo=NULL;
        }
    }

    if ( _blog )
        cout << " *** Final _sWorkingDirectory Directory[" << _sWorkingDirectory << endl;

    // Clean
    if ( !foo )
        delete[] foo;
    foo=NULL;

    if (_blog)
        cout << " *** End of Method TCode::" << __func__ << " with sPathName[" << sPathName << "]" << endl;

}

// ----------------------------------------------------------------------
//  createDirectory Method
// ----------------------------------------------------------------------
void TCode::createDirectory(TString sPathName, TString sDirectoryName)
{
    if (_blog)
        cout << " *** Method TCode::" << __func__ << " with sPathName[" << sPathName << "] sDirectoryName[" << sDirectoryName << "]" << endl;

    TString sdir = sPathName;
    if ( sDirectoryName.Length() > 0 ) {
        //buffer used to prevent memleak, as ConcatFileNAme result must be deleted
       char *foo=NULL;
       foo = gSystem->ConcatFileName(sPathName.Data(), sDirectoryName);
       sdir = foo;
       delete[] foo; foo=NULL;
       if ( _blog)
           cout << " *** sdir[" << sdir << "]" << endl;
    }

    try {
        // Creates the directory if it doesn't exist
        void * openFileResult = gSystem->OpenDirectory(sdir.Data());
        try {
            if (openFileResult == 0) {
                throw URANIE::Exceptions::UInfoExceptions(
                            __FILE__,
                            __LINE__,
                            Form("TCode::createDirectory Method\n The directory \"%s\" does not exist\n URANIE creates it for you", sdir.Data()));
            }
        } catch (URANIE::Exceptions::UExceptions& ue)
        {
            if (_nCurrentSimulation == 1 )
                ue.printMessage();
            if (gSystem->mkdir(sdir, kTRUE) == -1)
                throw URANIE::Exceptions::UErrorExceptions(
                        __FILE__,
                        __LINE__,
                        Form("TCode::createDirectory ERROR : The directory \"%s\" can not be created", sdir.Data()));
        }
        gSystem->FreeDirectory(openFileResult);
        gSystem->cd(sdir);

        if ( _blog )
            cout << " *** Current Directory[" << gSystem->pwd() << endl;
    } catch (URANIE::Exceptions::UExceptions& ue)
    {
        ue.printMessage();
        if (ue.getMessageType() > URANIE::Exceptions::UExceptions::kWARNING)
            exit(-1);
    }
    if (_blog)
        cout << " *** End of Method TCode::" << __func__ << " with sPathName[" << sPathName << "] sDirectoryName[" << sDirectoryName << "]" << endl;

}

// ----------------------------------------------------------------------
//  copyFiles Method
// ----------------------------------------------------------------------
void TCode::copyFiles(Option_t *option)
{
    TString soption = option ;
    soption.ToLower();
    FUNCT_TRACE();

    if (_blog)
        cout << endl << "*****************************" << endl <<
                " *** Method TCode::" << __func__ << " with soption[" << soption << "]" << endl <<
                " *** Current Directory [" << gSystem->pwd() << "]" << endl;

    Bool_t bcopy;

    Int_t ninputfiles = (Int_t)(_listOfInputFiles->GetEntries());
    if (_blog) {
        cout << endl << " ***  List Of Input Files [" << ninputfiles << "]" << endl;
        _listOfInputFiles->Print();
        for (Int_t ifile = 0; ifile < ninputfiles; ifile++)
            ((TInputFile *) (_listOfInputFiles->At(ifile)))->printLog();
        cout << endl << " ***  End Of List Of Input Files [" << ninputfiles << "]" << endl;
    }

    try {
        if (_blog)
            cout << " *** Number of Input Files[" << ninputfiles << "]" << endl;
        for (Int_t indx = 0; indx < ninputfiles; indx++) {
            if (_blog)
                cout << " *** input File(" << indx + 1 << "/" << ninputfiles << ")"<<endl;

            // Copy by default (with or without attributes)
            bcopy = kTRUE;

            TString sfile = getInputFile(indx)->getFileName();
            Int_t natt = getInputFile(indx)->getAttributesNumber();

            // With attributes to substitute, get the copy flag
            if (natt > 0)
                bcopy = getInputFile(indx)->isFileToCopy();

            if (_blog)
                cout << " *** input File(" << indx + 1 << "/" << ninputfiles << ")" <<
                        " sfile[" << sfile << "] natt[" << natt << "]" <<
                        " bcopy[" << bcopy << "]" << endl;

            //C.F. 2009/10/08 NEW 2.0.1
            // \todo 2021/10/28 To verify
            //If the file was given with a relative path, reference directory is set as dirname.
            if (gSystem->DirName(sfile.Data()) == TString(".")) {
                char *foo=NULL; //buffer used to prevent memleak, as ConcatFileNAme result must be deleted
                foo = gSystem->ConcatFileName(_sReferenceDirectory.Data(), sfile.Data());
                sfile = foo;
                delete[] foo; foo=NULL;
            }

            TString sfilename = gSystem->BaseName(sfile.Data());

            if (_blog)
                cout << " *** input File(" << indx + 1 << "/" << ninputfiles <<
                        ") sfilename[" << sfilename << "] natt[" << natt <<
                        "] sfile[" << sfile << "]" <<
                        "] bcopy[" << bcopy << "]" << endl;

            // Case of Copy
            if (bcopy) {
                if (_blog)
                    cout << "  *** Copy input sfile[" << sfile << "] to[" << sfilename << "]" << endl;

                Int_t nret = gSystem->CopyFile(sfile, sfilename, kTRUE);
                if (nret == -1)
                    throw URANIE::Exceptions::UExceptions(__FILE__, __LINE__,
                                                          Form("TCode::%s ERROR : for copy file \"%s\"",
                                                               __func__, sfile.Data()));
            } else {
                if (_blog)
                    cout << "  *** NO Copy input File[" << sfilename << "]"
                         << endl;
            } // End of if (bcopy)
        } // End of for (Int_t indx = 0
    } catch (URANIE::Exceptions::UExceptions& ue)
    {
        ue.printMessage();
        if (ue.getMessageType() > URANIE::Exceptions::UExceptions::kWARNING) {
            throw ue;
        }
    }

    if (_blog)
    {
        cout << "*** End Of Copy Input File" << endl << "***" << endl;
        cout << "*** Listing of Working Directory [" << gSystem->pwd() << "]"
             << endl;
#ifdef WIN32
        gSystem->Exec("dir");
#else
        gSystem->Exec("ls -al");
#endif
        cout << "*** End of Listing of Working Directory" << endl
             << "*****************************" << endl << endl;
    }

    if ( _blog )
        cout << " *** End of Method TCode::" << __func__ << " with soption[" << soption << "]" << endl;
}



void TCode::resetCurrentSimulation()
{
    _nCurrentSimulation = 9999;
    init();
}

void TCode::cleanOutputFiles(Option_t *option)
{
    TString soption = option ;
    soption.ToLower();
    for (Int_t i = 0; i < Int_t(_outputFiles.size()); i++)
	if (!gSystem->AccessPathName(_outputFiles[i]->GetName(), kFileExists))
	    gSystem->Unlink(_outputFiles[i]->GetName());
}

TString TCode::getInputVarNames()
{
    TString sinput;
    for (Int_t ifile = 0; ifile < this->getNInputFiles(); ifile++)
    {
        for (Int_t ivar = 0; ivar < this->getNInputAttributes(ifile); ivar++)
	{
            TString sname = this->getInputAttribute(ifile, ivar)->GetName();
            sinput += sname + ":";
	}
    }
    if (sinput.Length() > 0)
	sinput.Remove(sinput.Last(':'));
    else
    {
        for (Int_t iattribute = 0; iattribute < _tds->getNAttributes();
	     iattribute++)
	{
            if (URANIE::DataServer::TAttribute::kInput
		== (_tds->getAttribute(iattribute))->getInputOutput())
	    {
                sinput += TString((_tds->getAttribute(iattribute))->GetName())
		    + ":";
	    }
	}
        if (sinput.Length() > 0)
	    sinput.Remove(sinput.Last(':'));
    }
    return sinput;
}

TString TCode::getOutputVarNames()
{
    TString soutput;
    for (Int_t ioutfile = 0; ioutfile < this->getNOutputFiles(); ioutfile++)
    {
        for (Int_t iattfile = 0;
	     iattfile < this->getNOutputAttributes(ioutfile); iattfile++)
	{
            TString sattname =
		this->getAttribute(ioutfile, iattfile)->GetName();
            soutput += sattname + ":";
	}
    }
    if (soutput.Length() > 0)
	soutput.Remove(soutput.Last(':'));
    else
    {
        for (Int_t iattribute = 0; iattribute < _tds->getNAttributes();
	     iattribute++)
	{
            if (URANIE::DataServer::TAttribute::kOutput
		== (_tds->getAttribute(iattribute))->getInputOutput())
	    {
                soutput += TString((_tds->getAttribute(iattribute))->GetName())
		    + ":";
	    }
	}
        if (soutput.Length() > 0)
	    soutput.Remove(soutput.Last(':'));
    }
    return soutput;
}

// ----------------------------------------------------------------------
// Printing Log files
// ----------------------------------------------------------------------
void TCode::printLog(Option_t *option)
{
    cout << "TCode::printLog[" << option << "]" << endl;
    cout << "Name[" << GetName() << "] Title[" << GetTitle() << "]" << endl;
    cout << "_sWorkingDirectory[" << _sWorkingDirectory << "]" << endl;
    cout << "_sReferenceDirectory[" << _sReferenceDirectory << "]" << endl;

    //input
    if ( _listOfInputFiles ) {
        Int_t ninput = (Int_t)(_listOfInputFiles->GetEntries());
        for (Int_t indx = 0; indx < ninput; indx++) {
            cout << "inputFiles index[" << indx + 1 << "/" << ninput << "]" << endl;
            getInputFile(indx)->printLog();
        }
    } else {
        cout << "  ** empty inputFiles" << endl;
    }

    //output
    Int_t noutput = (Int_t)(_outputFiles.size());
    for (Int_t indx = 0; indx < noutput; indx++) {
        cout << "outputFiles index[" << indx + 1 << "/" << noutput << "]" << endl;
        _outputFiles[indx]->printLog();
    }

    // Index to Substitute
    cout << "  ** nIndexOfInputToSubstitute vector _nOfInputToSubstitute[" << _nOfInputToSubstitute << "]" << endl;
    if( _nIndexOfInputToSubstitute != NULL) {

	for (Int_t indx = 0; indx < _nOfInputToSubstitute; indx++)
	    cout << _nIndexOfInputToSubstitute[indx] << " ";
	cout << endl;
    }
}

} // End of namespace Launcher
} // End of namespace URANIE
// End of File

