//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TCode.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mon oct 9th 14:49:25 CEST 2006
 
 \brief Interface of the class URANIE::Launcher::TCode
 
 */
/**
 
 * @class URANIE::Launcher::TCode
 *
 * @brief Description of the class TCode.
 *
 * To be written by the developper.
 *
 */
#ifndef TCODE_H
#define TCODE_H

// URANIE
#include "TAttribute.h"
#include "TDataServer.h"

#include "TInputFile.h"
#include "TOutputFile.h"

namespace URANIE {
namespace Launcher {
class TCode: public TNamed {
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    /*!
     * \brief Default constructor
     */
    TCode();
    /*!
     * \brief constructor with a  name and a title
     * \param name
     * \param title
     */
    TCode(const TString &name, const TString &title);
    /*!
     * \brief Constructor from TDataServer and Command as string
     * \param tds
     * \param scmd
     */
    TCode(URANIE::DataServer::TDataServer *tds, TString scmd);

    /*!
     * \brief Copy constructor
     * \param mycode
     */
    TCode(URANIE::Launcher::TCode &mycode);

    /*!
     * \brief Default destructor
     */
    virtual ~TCode();
    //@}

    //---------------------------------------------
    /** @name Get Input  Files or Output Files
     **/
    //@{    
    /*!
     * \brief Get the input file file from his name
     * \param sfile
     * \return
     */
    TInputFile * getInputFile(TString sfile);
    /*!
     * \brief Get the input file from his index
     * \param nfile
     * \return
     */
    TInputFile * getInputFile(Int_t nfile);
    /*!
     * \brief Get the \b last input file
     * \return
     */
    TInputFile * getLastInputFile();
    /*!
     * \brief Get the output file from his index
     * \param nfile
     * \return
     */
    TOutputFile * getOutputFile(Int_t nfile);

    /*!
     * \brief getOutputFile
     * \param sfile
     * \return
     */
    TOutputFile * getOutputFile(TString sfile);

    //@}

    //---------------------------------------------
    /** @name Setting and Getting attributs
     **/
    //@{
    /*!
     * \brief Set the working directory
     * \param str
     */
    void setWorkingDirectory(TString sPathName = "");

    /*!
     * \brief Get the working directory
     * \return
     */
    TString getWorkingDirectory() {
        return _sWorkingDirectory;
    };
    /*!
     * \brief Set the reference directory where inputs no change are located
     * \param str
     */
    void setReferenceDirectory(TString str) {
        _sReferenceDirectory = str;
    };
    /*!
     * \brief Get the reference directory where inputs no change are located
     * \return
     */
    TString getReferenceDirectory() {
        return _sReferenceDirectory;
    };
    /*!
     * \brief Set the code command
     * \param str
     */
    void setCommand(TString str) {
        _sCommand = str;
    };
    /*!
     * \brief Get the code command
     * \return
     */
    TString getCommand() {
        return _sCommand;
    };
    /*!
     * \brief Get the number of output files
     * \return
     */
    Int_t getNOutputFiles() {
        return (Int_t)(_outputFiles.size());
    };
    /*!
     * \brief Get the number of output attributes
     * \return
     */
    Int_t getNOutputAttributes();
    /*!
     * \brief Get the number of output attributes in the output file given by his index
     * \param nfile
     * \return
     */
    Int_t getNOutputAttributes(Int_t nfile);
    /*!
     * \brief Get the number of input files
     * \return
     */
    Int_t getNInputFiles() {
        return (Int_t)(_listOfInputFiles->GetEntries());
    };
    /*!
     * \brief Get the number of input attributes
     * \return
     */
    Int_t getNInputAttributes();
    /*!
     * \brief Get the number of input attributes in the input file given by his index
     * \param nfile
     * \return
     */
    Int_t getNInputAttributes(Int_t nfile);
    /*!
     * \brief Get an attribute of an input file
     * \param nfile
     * \param natt
     * \return
     */
    URANIE::DataServer::TAttribute * getInputAttribute(Int_t nfile, Int_t natt);
    /*!
     * \brief Get an attribute of an output file
     * \param nfile
     * \param natt
     * \return
     */
    URANIE::DataServer::TAttribute * getAttribute(Int_t nfile, Int_t natt);
    /*!
     * \brief Get the names of the input variables separated by ":"
     * \return
     */
    TString getInputVarNames();
    /*!
     * \brief Get the names of the output variables separated by ":"
     * \return
     */
    TString getOutputVarNames();
    /*!
     * \brief Get the pointer to the TDS
     * \return
     */
    URANIE::DataServer::TDataServer * getTDataServer() {
        return _tds;
    };
    /*!
     * \brief Set the pointer to the TDS
     * \param atds
     */
    void setTDataServer(URANIE::DataServer::TDataServer *atds) {
        _tds = atds;
        //      UpdateInputList();
    };
    //@}

    //---------------------------------------------
    /** @name Encapsulates a code
     *
     * The states of the TCode are follow:
     * \dot
     * digraph samplerTSamplerMachineEtat {
     *    rankdir=LR;
     *    node [shape=record, fontname=Helvetica, fontsize=10];
     *    stepInit [ label="init" ];
     *    stepPreTraitment [ label="preTraitment"];
     *    stepRun [ label="run"];
     *    stepPostTraitment [ label="postTraitment"];
     *    stepTerminate [ label="terminate"];
     *    stepRunTest [ label="runTest"];
     *    stepInit -> stepPreTraitment -> stepRun -> stepPostTraitment -> stepTerminate [ arrowhead="open", style="dashed" ];
     *    stepPostTraitment -> stepInit [ arrowhead="open", style="doted" ];
     *    stepRunTest;
     *    }
     * \enddot
     * with
     * \li \b init : the initialisation phase
     * \li \b preTraitment : the preTraitment phase (change in the inputs files the current values of the input attributes)
     * \li \b run : Evaluate the code with the current inputs values
     * \li \b postTraitment : the postTraitment phase (cath in the output files the values of teh ouput attributes)
     * \li \b terminate : the terminate phase
     * \li \b runTest : run a test \b \em "Not Yest implemented" (just to verify all are good with the first computation)
     **/
    //@{
    /*!
     * \brief Run a test for the code
     * \param option
     */
    virtual void runTest(Option_t *option = "");
    /*!
     * \brief Init file typing
     * \param option
     */
    virtual void initFileType(Option_t *option = "");
    /*!
     * \brief Init phase of the code
     * \param option
     */
    virtual void init(Option_t *option = "");
    /*!
     * \brief Pre-traitment phase of the code
     * \param n
     * \param val
     * \param index
     * \param option
     */
    virtual void preTraitment(Int_t n, Double_t *val, Int_t *index = NULL, Option_t *option = "");
    /*!
     * \brief Pre-traitment phase of the code that deals with vectors/strings
     * \param n
     * \param input
     * \param option
     */
    virtual void preTraitment(Int_t n, vector<URANIE::DataServer::UEntry*> *input, Option_t *option = "");
    /*!
     * \brief Run the code
     * \param option
     */
    virtual void run(Option_t *option = "");
    /*!
     * \brief Post-traitment phase of the code
     * \param nout
     * \param dvalout
     * \param option
     * \param lstOut
     * \return
     */
    virtual Bool_t postTraitment(Int_t &nout, Double_t *dvalout, Option_t *option = "", TList *lstOut = NULL);
    /*!
     * \brief Post-traitment phase of the code that deals with vectors/strings
     * \param out
     * \param option
     * \return
     */
    virtual Bool_t postTraitment(vector<URANIE::DataServer::UEntry*> *out, Option_t *option);
    /*!
     * \brief Terminate phase of the code
     * \param option
     */
    virtual void terminate(Option_t *option = "");
    //@}

    //---------------------------------------------
    /** @name Manipulates the code
     **/
    //@{

    /*!
     * \brief create the directory sPathName+"/"+sDirectoryName
     *
     * \param sPathName
     * \param sDirectoryName
     */
    void createDirectory(TString sPathName, TString sDirectoryName = "");

    /*!
     * \brief copyFiles
     * \param option
     */
    virtual void copyFiles(Option_t *option = "");

    /*!
     * \brief resetCurrentSimulation
     *
     * Creates the uranieLancher_0 subdirectory to store all the computation
     */
    void resetCurrentSimulation();
    /*!
     * \brief Cleans the output files
     * \param option
     */
    virtual void cleanOutputFiles(Option_t *option = "");
    //@}

    //---------------------------------------------
    /** @name Adding Input and Output files
     **/
    //@{
    /*!
     * \brief Add a TInputFile file
     * \param ifile
     */
    void addInputFile(TInputFile *ifile);
    /*!
     * \brief Add an input file as a TString
     * This input file does not contain input attributes.
     * It is a \b mandatory input file for the code evaluation.
     * \param str
     */
    void addInputFile(TString str);
    /*!
     * \brief Add a TOutputFile file
     * The TOutputFile file contains the output attributes of teh study.
     * \param ifile
     */
    void addOutputFile(TOutputFile *ifile);
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    /*!
     * \brief Force the log
     */
    void setLog() {
        _blog = kTRUE;
    }
    /*!
     * \brief Force to desable the log
     */
    void unsetLog() {
        _blog = kFALSE;
    }
    /*!
     * \brief Swith the status of the log
     */
    void changeLog() {
        _blog = _blog ? kFALSE : kTRUE;
    }
    /*!
     * \brief Get the status of the log
     * \return
     */
    Bool_t getLog() {
        return _blog;
    }
    /*!
     * \brief Print the log of the object/class
     * \param option
     */
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Launcher::TCode, ID_LAUNCHER)

    //---------------------------------------------
    /** @name protected attributes
     *
     **/
    //@{
    protected:
        vector<TOutputFile *> _outputFiles;         ///< vector of output files
        TList *_listOfInputFiles;                   ///< The list of input
        TString _sWorkingDirectory;                 ///< The Working directory
        TString _sReferenceDirectory;               ///< The Working directory
        TString _sCommand;                          ///< The Command of the code
        Int_t _nCurrentSimulation;            ///< The Current number of  simulation
        Int_t _nOfInput;              ///< The size of the vector listOfInput scanning input files.
        Int_t _nOfInputToSubstitute; ///< The size of the vector _nIndexOfInputToSubstitute
        Int_t *_nIndexOfInputToSubstitute;    ///< The vector of Index to Substitute
        Bool_t _blog;                               ///< Boolean for edit the log
        URANIE::DataServer::TDataServer *_tds;      //! Pointeur vers un TDS
    //@}
};

} // End of namespace Launcher
} // End of namespace URANIE

#endif
// End of File
