//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TInputFile.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mon oct 9th 14:50:04 CEST 2006
 
 \brief Interface of the class URANIE::Launcher::TInputFile
 
 */
/*!
 @class URANIE::Launcher::TInputFile
 
 @brief Abstract class for an input file of a code.
 
 \warning This class contains two pure abstracted methods
 \li \ref searchIndexeAttribute : the method is to find the indexes of the attribute in the input file
 \li \ref substitue : the method is to substitute the value in the input file must to implement in a daughter class.
 
 */
#ifndef TINPUTFILE_H
#define TINPUTFILE_H

// URANIE
#include "Launcher.h"
#include "TCodeFile.h"
#include "UExceptions.h"


#define DEFAULT_FMU_WORKING_DIRECTORY "URANIE_FMU"
#define CTEST_RETURN_SUCCESS 0
#define CTEST_RETURN_FAIL 1

namespace URANIE
{
namespace Launcher
{
class TInputFile: public TCodeFile
{
public:

    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    TInputFile(TString str);
    //! Default destructor
    virtual ~TInputFile();
    //@}

    //---------------------------------------------
    /** @name Setting and Getting attributes
     **/
    //@{
    virtual void addAttribute(URANIE::DataServer::TAttribute*);

    Bool_t isExist();
    //@}

    //---------------------------------------------
    /** @name Methods to perform initialisation and terminate states
     **/
    //@{
    //! The init step
    /*!
      \warning not yet used
    */
    virtual void init();

    //! The terminate step
    /*!
      \warning not yet used
    */
    virtual void terminate();
    //@}


    //---------------------------------------------
    /** @name File properties
     **/
    //@{
    /*!
     * \brief Get if the file must be copied (default is kTRUE)
     * \return
     */
    virtual Bool_t isFileToCopy() {
            return kTRUE;
        }
    //@}

    //---------------------------------------------
    /** @name Abstract pure methods
	these two methods must be implement in a daughter class.
    **/
    //@{
    //! search the indexes of the attribute in the input file
    /*!
      \param att (URANIE::DataServer::TAttribute*) the attribute to search the indexes
    */
    virtual void searchIndexeAttribute(URANIE::DataServer::TAttribute* att) = 0;

    //! Substitute the values of the attributes linked in the input file and write it
    /*!
      The first value is defined par the parameter \em n and the values are stored in the parameter \em val.
      After the method, the parameter is increased by the number of attributes contained in the input file.
      \param n (Int_t &) the index of the first value in the vector of data \e val
      \param val (Double_t *) the pointer of the data
      \param index (Int_t *)the pointer of the index of data
    */
    virtual void substitue(Int_t &n, Double_t *val, Int_t *index = NULL) = 0;
    virtual void substitue(Int_t &n, vector<URANIE::DataServer::UEntry*> *ent, Int_t *index = NULL) =0;
    //virtual void substitue();
    //@}

    //---------------------------------------------
    /** @name The Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Launcher::TInputFile, ID_LAUNCHER)
    //Classe de
};

} // Fin du namespace Launcher
} // Fin du namespace URANIE
#endif
