//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TInputFileFlag.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mon oct 9th 14:52:50 CEST 2006
 
 \brief Implementation of the class URANIE::Launcher::TInputFileFlag
 
 \par Modifications
 \li F. Gaudier (2010-12-14 11:38:09)
 - resolve the DI111 and DI113

 */
#include <iostream>
using std::cout;
using std::endl;

// ROOT
#include "TMath.h"

// URANIE
#include "TInputFileFlag.h"

// Club
#ifdef OLDCLUB
#include "club/IterateurChamp.h"
#include "club/TamponTexte.h"
#include "club/ClubErreurs.h"
#else
#include "TamponTexte.h"
#endif

ClassImp (URANIE::Launcher::TInputFileFlag)

namespace URANIE
{
namespace Launcher
{

TInputFileFlag::TInputFileFlag(TString str) :
        TInputFile(str)
{
    if (_blog)
        cout << "constructing TInputFileFlag file[" << str << "]" << endl;
    setSeparatorCharacter(" \t\r\n");
    _sstrbeg = _sstrend = "";
    _svecbeg = ""; _svecend = ""; _svecdel = ",";
}

TInputFileFlag::~TInputFileFlag()
{
    if (_blog)
        cout << "destroying TInputFileFlag" << endl;
}

void TInputFileFlag::searchIndexeAttribute(URANIE::DataServer::TAttribute* att)
{
    list<URANIE::DataServer::TAttributeFileKey *> KeyList = att->getKeyList();
    for (list<URANIE::DataServer::TAttributeFileKey *>::iterator it =
            KeyList.begin(); it != KeyList.end(); ++it)
    {
        Int_t nfield = -1;
        TString skey = (*it)->_sKey;
        if (_blog)
            cout << "skey[" << skey << "]      -------- " << endl;

        TString sfilecrt = (*it)->_sFile;
        if (sfilecrt.CompareTo(_sFileName, TString::kExact) == 0)
        {
            try
            {
                TamponTexte fnomFichier(((*it)->_sFile).Data(), 0);
                // Define the comment and separator character
                fnomFichier.initCarCommentaires(_cComments);
                fnomFichier.initSeparateurs(_cSeparators);
#ifndef OLDCLUB
		fnomFichier.initFieldSeparateur(_cFieldSeparator);
#endif
		
                // Elimination des lignes de commentaires
                //fnomFichier.elimineCommentaires();

                // Elimination des lignes vides
                //fnomFichier.elimineBlancs();

                // Test si le fichier ainsi obtenu est vide
                if (fnomFichier.estVide())
                {
                    throw "URANIE::Launcher::TCode::searchIndexeAttribute the file is empty";
                }
                if (_blog)
                {
                    cout << " ** Opening the file " << fnomFichier.nom()
                            << " ended." << endl;
                    cout << " ** file " << fnomFichier.nom() << " : "
                            << fnomFichier.nombreChamps() << " champs" << endl;
                }

                // acces direct
                Int_t nchamps = fnomFichier.nombreChamps();
                for (Int_t i = 1; i <= nchamps; i++)
                {
                    const char* champ = fnomFichier.champ(i);
                    TString strchamp(champ);
                    if (_blog)
                        cout << "champ[" << i << "] valeur[" << strchamp
                                << "] skey[" << skey << "]      -------- "
                                << endl;

                    if (strchamp.CompareTo(skey.Data(), TString::kExact) == 0)
                    {
                        if (_blog)
                            cout << "********** champ[" << i << "] valeur["
                                    << strchamp << "] skey[" << skey << "]";
                        list<URANIE::DataServer::TAttributeFileKey *>::iterator first =
                                KeyList.begin();
                        att->setFieldOfKey(distance(first, it), i);
                        nfield = i;
                        if (_blog)
                            cout << "---------- champ[" << nfield << "] valeur["
                                    << strchamp << "] [" << champ << "]"
                                    << endl;
                    }
#ifndef OLDCLUB
                    delete[] champ;
#endif
                }

                if (nfield == -1)
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__,
                            __LINE__,
                            Form(
                                    "TInputFileFlag::searchIndexeAttribute ERROR\n The key [%s] in the TInputFileFlag[%s] not found",
                                    skey.Data(), ((*it)->_sFile).Data()));
            }
#ifdef OLDCLUB
            catch (ClubErreurs ce)
            {
                ce.affiche();
                throw ce;
            }
#else
            catch (TamponTexteError& tte)
            {
                tte.printMessage();
                throw tte;
            }
            
            catch (URANIE::Exceptions::UErrorExceptions& ue)
            {
                ue.printMessage();
                throw ue;
            }
#endif
        }
    }
}

void TInputFileFlag::substitue(Int_t &n, Double_t *val, Int_t *index)
{

    for (Int_t i = 0; i < getAttributesNumber(); i++)
    {
        if (_blog)
            cout << endl << " ** TInputFileFlag::substitue file[" << _sFileName
                    << "] i[" << i << "/" << getAttributesNumber() << "]"
                    << endl;
        list<URANIE::DataServer::TAttributeFileKey *> KeyList =
                _attributes[i]->getKeyList();
        for (list<URANIE::DataServer::TAttributeFileKey *>::iterator it =
                KeyList.begin(); it != KeyList.end(); ++it)
        {
            TString sfilename = gSystem->BaseName((*it)->_sFile.Data());

            if (_blog)
                cout << "\nTInputFileFlag::substitue sfilename[" << sfilename
                        << "]" << endl;

            TString sfilecrt = (*it)->_sFile;
            if (sfilecrt.CompareTo(_sFileName, TString::kExact) == 0)
            {
                try
                {
                    TamponTexte fnomFichier(sfilename, 0);

                    // Define the comment and separator character
                    fnomFichier.initCarCommentaires(_cComments);
                    fnomFichier.initSeparateurs(_cSeparators);
#ifndef OLDCLUB
		    fnomFichier.initFieldSeparateur(_cFieldSeparator);
#endif
                    // Elimination des lignes de commentaires
                    //fnomFichier.elimineCommentaires();

                    // Elimination des lignes vides
                    //fnomFichier.elimineBlancs();

                    // Test si le fichier ainsi obtenu est vide
                    if (fnomFichier.estVide())
                    {
                        throw "URANIE::Launcher::TCode::searchIndexeAttribute the file is empty";
                    }
                    if (_blog)
                    {
                        cout << " ** Opening the file " << fnomFichier.nom()
                                << " ended." << endl;
                        cout << " ** file " << fnomFichier.nom() << " : "
                                << fnomFichier.nombreChamps() << " champs"
                                << endl;
                        cout << " ** _cComments [" << _cComments << "]" << endl;
                    }

                    TString sformattosubstitute = TString((*it)->_sFormat);
                    Double_t dValue = 0.0;
                    // in case of the vector of index is not empty, use it to catch the value
                    if (index != NULL)
                    {
                        dValue = val[index[n - 1]];
                        if (_blog)
                            cout << " **** sfilename[" << sfilename << "] n["
                                    << n << "] index[n]=" << index[n - 1]
                                    << ", dValue[" << dValue
                                    << "] FormatSubstitute["
                                    << sformattosubstitute << "]" << endl;
                    }
                    else
                    {
                        dValue = val[n];
                        if (_blog)
                            cout << " **** sfilename[" << sfilename << "] n["
                                    << n << "] index[n][NULL] dValue[" << dValue
                                    << "] FormatSubstitute["
                                    << sformattosubstitute << "]" << endl;
                    }
                    for (Int_t j = 0; j < (*it)->_sEventList->GetN(); j++)
                    {
                        if (_blog)
                            cout << " ********************** key["
                                    << Int_t((*it)->_sEventList->GetEntry(j))
                                    << "] value[" << dValue << "]" << endl;
                        if (sformattosubstitute.Contains("%d"))
                        {
                            Int_t nval = (Int_t)(TMath::Ceil(dValue));
                            fnomFichier.remplaceChamp(
                                    Int_t((*it)->_sEventList->GetEntry(j)),
                                    Form(sformattosubstitute.Data(), nval));
                        }
                        else
                            fnomFichier.remplaceChamp(
                                    Int_t((*it)->_sEventList->GetEntry(j)),
                                    Form(sformattosubstitute.Data(), dValue));
                    }

                    fnomFichier.ecrit();
                }
#ifdef OLDCLUB
                catch (ClubErreurs ce)
                {
                    throw ("TAttribute::searchIndexe Error");
                }
#else
                catch (TamponTexteError& ex)
                {
                    ex.printMessage();
                    throw("TAttribute::searchIndexe Error");
                } catch (URANIE::Exceptions::UErrorExceptions&)
                {
                    throw("TAttribute::searchIndexe Error");
                }
#endif
            }
        }
        n++;
    }
}

void TInputFileFlag::substitue(Int_t &n, vector<URANIE::DataServer::UEntry*> *ent, Int_t *index)
{
    TString beg, end;
    for (Int_t i = 0; i < getAttributesNumber(); i++)
    {
        if (_blog)
            cout << endl << " ** TInputFileFlag::substitue file[" << _sFileName
		 << "] i[" << i << "/" << getAttributesNumber() << "]"
		 << endl;
        list<URANIE::DataServer::TAttributeFileKey *> KeyList =
	    _attributes[i]->getKeyList();
        for (list<URANIE::DataServer::TAttributeFileKey *>::iterator it =
		 KeyList.begin(); it != KeyList.end(); ++it)
        {
            TString sfilename = gSystem->BaseName((*it)->_sFile.Data());

            if (_blog)
                cout << "\nTInputFileFlag::substitue sfilename[" << sfilename
		     << "]" << endl;

            TString sfilecrt = (*it)->_sFile;
            if (sfilecrt.CompareTo(_sFileName, TString::kExact) == 0)
            {
                try
                {
                    TamponTexte fnomFichier(sfilename, 0);

                    // Define the comment and separator character
                    fnomFichier.initCarCommentaires(_cComments);
                    fnomFichier.initSeparateurs(_cSeparators);
#ifndef OLDCLUB
		    fnomFichier.initFieldSeparateur(_cFieldSeparator);
#endif
	
                    // Elimination des lignes de commentaires
                    //fnomFichier.elimineCommentaires();

                    // Elimination des lignes vides
                    //fnomFichier.elimineBlancs();

                    // Test si le fichier ainsi obtenu est vide
                    if (fnomFichier.estVide())
                    {
                        throw "URANIE::Launcher::TCode::searchIndexeAttribute the file is empty";
                    }
                    if (_blog)
                    {
                        cout << " ** Opening the file " << fnomFichier.nom()
			     << " ended." << endl;
                        cout << " ** file " << fnomFichier.nom() << " : "
			     << fnomFichier.nombreChamps() << " champs"
			     << endl;
                        cout << " ** _cComments [" << _cComments << "]" << endl;
                    }

                    TString sformattosubstitute = TString((*it)->_sFormat);
		    getBegEndProperties( _attributes[i], beg, end);
		    for (Int_t j = 0; j < (*it)->_sEventList->GetN(); j++)
                    {
                        if (_blog)
                            cout << " ********************** key["
				 << Int_t((*it)->_sEventList->GetEntry(j))
				 << "] value[" << ent->at( ((index==NULL) ? n : index[n - 1]) )->dumpStr() << "]" << endl;
			
			fnomFichier.remplaceChamp(
			    Int_t((*it)->_sEventList->GetEntry(j)),
			    ent->at( ((index==NULL) ? n : index[n - 1]) )->dumpStr( _svecdel, beg, end, (*it)->_sFormat).c_str() );
		    }

                    fnomFichier.ecrit();
                }
#ifdef OLDCLUB
                catch (ClubErreurs ce)
                {
                    throw ("TAttribute::searchIndexe Error");
                }
#else
                catch (TamponTexteError& ex)
                {
                    ex.printMessage();
                    throw("TAttribute::searchIndexe Error");
                } catch (URANIE::Exceptions::UErrorExceptions&)
                {
                    throw("TAttribute::searchIndexe Error");
                }
#endif
            }
        }
        n++;
    }
}
void TInputFileFlag::substitue()
{
    for (Int_t i = 0; i < getAttributesNumber(); i++)
    {
        if (_blog)
            cout << endl << " ** TInputFileFlag::substitue file[" << _sFileName
                    << "] i[" << i << "/" << getAttributesNumber() << "]"
                    << endl;

        list<URANIE::DataServer::TAttributeFileKey *> KeyList = _attributes[i]->getKeyList();
        for (list<URANIE::DataServer::TAttributeFileKey *>::iterator it = KeyList.begin(); it != KeyList.end(); ++it)
        {
            TString sfilename = gSystem->BaseName((*it)->_sFile.Data());

            if (_blog)
                cout << "\nTInputFileFlag::substitue sfilename[" << sfilename << "]" << endl;

            TString sfilecrt = (*it)->_sFile;
            if (sfilecrt.CompareTo(_sFileName, TString::kExact) == 0)
            {
                try
                {
                    TamponTexte fnomFichier(sfilename, 0);

                    // Define the comment and separator character
                    //fnomFichier.initCarCommentaires(_cComments);
                    fnomFichier.initSeparateurs(_cSeparators);
#ifndef OLDCLUB
		    fnomFichier.initFieldSeparateur(_cFieldSeparator);
#endif
                    // Elimination des lignes de commentaires
                    //fnomFichier.elimineCommentaires();

                    // Elimination des lignes vides
                    //fnomFichier.elimineBlancs();

                    // Test si le fichier ainsi obtenu est vide
                    if (fnomFichier.estVide())
                    {
                        throw "URANIE::Launcher::TCode::searchIndexeAttribute the file is empty";
                    }
                    if (_blog)
                    {
                        cout << " ** Opening the file " << fnomFichier.nom()
                                << " ended." << endl;
                        cout << " ** file " << fnomFichier.nom() << " : "
                                << fnomFichier.nombreChamps() << " champs"
                                << endl;
                        cout << " ** _cComments [" << _cComments << "]" << endl;
                    }

                    TString sformattosubstitute = TString((*it)->_sFormat);


                    TString value = _attributes[i]->getSlurmOption();

                    for (Int_t j = 0; j < (*it)->_sEventList->GetN(); j++) {

                        if (_blog)
                            cout << " **** sfilename[" << sfilename << "] n["
                                 << 0 << "] index[n][NULL] value[" << value
                                 << "] FormatSubstitute["
                                 << sformattosubstitute << "]" << endl;

                        fnomFichier.remplaceChamp(Int_t((*it)->_sEventList->GetEntry(j)), value);
                        fnomFichier.ecrit();
                    }
                }
#ifdef OLDCLUB
                catch (ClubErreurs ce)
                {
                    throw ("TAttribute::searchIndexe Error");
                }
#else
                catch (TamponTexteError& ex)
                {
                    ex.printMessage();
                    throw("TAttribute::searchIndexe Error");
                } catch (URANIE::Exceptions::UErrorExceptions&)
                {
                    throw("TAttribute::searchIndexe Error");
                }
#endif
            }
        }

    }
}

} // Fin du namespace Launcher
} // Fin du namespace URANIE

