//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
// TLSFCluster lun aoû 4 14:19:19 CEST 2008
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TLSFCluster.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   2011-05-20 10:15:42
 
 \brief Implementation of the class URANIE::Launcher::TLSFCluster
 
 This class implement the LSF wrapper to submit parralel launch in the Titane cluster at CCRT
 
 <example>
 Sender: LSF System <lsfadmin@titane1514>
 Subject: Job 413994: <samplFLOWRATE> Done

 Job <samplFLOWRATE> was submitted from host <titane1000> by user <gaudier> in cluster <titane>.
 Job was executed on host(s) <8*titane1514>, in queue <testext>, as user <gaudier> in cluster <titane>.
 <2*titane1516>
 </example>

 The nodes allocated to the job are noticed in the environment variable \bf LSB_HOSTS
 This variable contains all the nodes allocated:
 <example>
 LSB_HOSTS=titane1514 titane1514 titane1514 titane1514 titane1514 titane1514 titane1514 titane1514 titane1516 titane1516
 </example>

 An another environmant variable \bf LSB_MCPU_HOSTS contains this information, but we have to treat it to construct the nodes list:
 <example>
 LSB_MCPU_HOSTS=titane1514 8 titane1516 2
 </example>

 All the envrionment variables are listed :
 <code>
 LSB_EXEC_CLUSTER=titane
 LSB_BATCH_JID=413994
 LSB_TRAPSIGS=trap # 15 10 12 2 1
 LSB_ERRORFILE=samplFLOWRATE.error.log
 LSB_JOBRES_CALLBACK=51944@titane1514
 LSB_JOB_EXECUSER=gaudier
 LSB_JOBID=413994
 LSB_JOBRES_PID=5286
 LSB_JOBNAME=samplFLOWRATE
 LSB_EEXEC_REAL_UID=
 LSB_DJOB_RU_INTERVAL=15
 LSB_HOSTS=titane1514 titane1514 titane1514 titane1514 titane1514 titane1514 titane1514 titane1514 titane1516 titane1516
 LSB_UNIXGROUP_INT=den
 LSB_DJOB_HB_INTERVAL=15
 LSB_JOBFILENAME=/usr/share/lsf/spool/1305720816.413994
 LSB_JOBINDEX=0
 LSB_EXIT_PRE_ABORT=99
 LSB_JOBEXIT_STAT=0
 LSB_CHKFILENAME=/usr/share/lsf/spool/1305720816.413994
 LSB_DJOB_HOSTFILE=/usr/share/lsf/spool/1305720816.413994.hostfile
 LSB_JOB_STARTER=/usr/share/lsf/7.0/linux2.6-glibc2.3-x86_64/etc/jobstarter
 LSB_ACCT_FILE=/tmp/.1305720816.413994.acct
 LSB_SUB_HOST=titane1000
 LSB_QUEUE=testext
 LSB_MCPU_HOSTS=titane1514 8 titane1516 2
 LSB_OUTPUTFILE=samplFLOWRATE.output.log
 LSB_ECHKPNT_RSH_CMD=ssh
 LSB_EEXEC_REAL_GID=  
 </code>
 
 */

#include "TLSFCluster.h"

#include "TSystem.h"
#include <sstream>

ClassImp (URANIE::Launcher::TLSFCluster)

namespace URANIE
{
namespace Launcher
{
// class method
bool TLSFCluster::eligible()
{
    const char *tmp = gSystem->Getenv("LSB_MCPU_HOSTS");
    return (tmp != NULL && tmp[0]);
}

const char *TLSFCluster::nodeList()
{
    // LSB_HOSTS is not set beyond 4096 bytes
    //it is necessary to analyse LSB_MCPU_HOSTS
    
    return gSystem->Getenv("LSB_MCPU_HOSTS");
    
//     const char* tmp = gSystem->Getenv("LSB_MCPU_HOSTS");
// 
//     istringstream is(tmp);
//     std::ostringstream os;
//     while (!is.eof())
//     {
//         string node;
//         int nodenumber;
//         is >> node >> nodenumber;
//         for (int i = 0; i < nodenumber; i++)
//         {
//             os << node;
//             os << " ";
//         }
//     }
//     os.flush();
//     string* nodelist = new string(os.str(), 0, os.str().size() - 1); //removing the last character
//     
//     return nodelist->c_str();
}

// Constructor
TLSFCluster::TLSFCluster() :
        TCluster("LSF", TLSFCluster::nodeList(), 1)
{
    if (_blog)
        cout << " ** Constructor TLSFCluster::TLSFCluster [" << GetName()
                << "] Title[" << GetTitle() << "]" << endl;
    init();
}
TLSFCluster::~TLSFCluster()
{
    if (_blog)
        cout << " ** Destroy TLSFCluster [" << GetName() << "] Title["
                << GetTitle() << "]" << endl;
}

///////////////////////////////////////////////////////////
//  Construct and return the list of node
///////////////////////////////////////////////////////////
void TLSFCluster::init()
{
    TString snodes = GetTitle();
    if (_blog)
        cout << " ** TCluster::getListOfNodes Node[" << snodes << "]" << endl;
    if (snodes.CountChar('[') == 0)
    {
//         addNodeName(snodes);
        istringstream is(snodes.Data());
        while (!is.eof())
        {
            string node;
            int nodenumber;
            is >> node >> nodenumber;
            addNodeName(node,nodenumber);
        }
    }
    else
    {
        Int_t nbegin = snodes.Index('[');
        Int_t nend = snodes.Index(']');
        TString ssnodes = snodes(nbegin + 1, nend - 1 - nbegin);
        if (_blog)
            cout << " ** snode[" << snodes << "] ssnode[" << ssnodes << "]"
                    << endl;
        ssnodes.ReplaceAll(",", " ");
        istringstream inStream((char *) ssnodes.Data());
        string inName;
        size_t found;
        while (inStream >> inName)
        {
            if (_blog)
                cout << " ** inName[" << inName << "] -------- ";
            found = inName.find("-");
            if (found != string::npos)
            {
                if (_blog)
                    cout << "found at: " << int(found);
                inName.replace(int(found), 1, " ");
                if (_blog)
                    cout << " **** inName[" << inName << "]" << endl;
                istringstream inStream2(inName);
                string sfirst, slast;
                inStream2 >> sfirst;
                Int_t nfirst = atoi(sfirst.c_str());
                inStream2 >> slast;
                Int_t nlast = atoi(slast.c_str());
                if (_blog)
                    cout << " ** ****** First\"" << sfirst << "\" nfirst["
                            << nfirst << "] -- last\"" << slast << "\" nlast["
                            << nlast << "]" << endl;
                for (Int_t icrt = nfirst; icrt <= nlast; icrt++)
                    addNodeName(Form("%s%d", GetName(), icrt));

            }
            else
            {
                if (_blog)
                    cout << "not found" << endl;
                addNodeName(Form("%s%s", GetName(), inName.c_str()));
            }
        }

    }

    if (_blog)
        cout << " **  End Of TCluster::getListOfNodes _slistOfNodes["
                << TCluster::getListOfNodes() << "]" << endl;
    TCluster::getListOfNodes();
}

void TLSFCluster::getCommandPrefix(char *mpiexe, char* procs)
{
    cout << "MPIRUN FOR LSF" << endl;
    cout << "procs = " << procs << endl;

    Char_t* procptr = procs;
    Int_t procsperjob = 1;
    //counts the number of comma in the string to define number of procs per job
    //WARNING : it works only if string is structured as 'host1,host2,host3'
    while (*procptr != '\0')
    {
        if (*procptr == ',')
            procsperjob++;
        procptr++;
    }
    Char_t npstring[20];
    sprintf(npstring, " -np %d ", procsperjob);

    strcpy(mpiexe, "mpirun ");
    strcat(mpiexe, npstring);
    strcat(mpiexe, " -host ");
    strcat(mpiexe, procs);
#ifdef WITH_MPIOLD
    strcat(mpiexe, " --bind-to-none");
#else
    strcat(mpiexe, " --bind-to none");
#endif
    cout << "TLSFCluster mpiexe: " << mpiexe << endl;
}

///////////////////////////////////////////////////////////
//  Log
///////////////////////////////////////////////////////////
void TLSFCluster::printLog(Option_t *option)
{
    cout << "TLSFCluster::printLog[" << option << "]" << endl;
    TCluster::printLog(option);
}

} // Fin du namespace Launcher
} // Fin du namespace URANIE

