//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////

/*!
 \file   TLauncherByStepRemote.cxx
 \author Daniel Beltran Mora  <daniel.beltran@bsc.es>
 \date 2019-02-14

 \brief

 */

#include "TLauncherByStepRemote.h"

#include <iostream>
#include <fstream>
#include <stdlib.h>


using std::cout;
using std::endl;

// ROOT
#include "TFile.h"
#include "TLeaf.h"
#include "TPad.h"
#include "TObjString.h"
#include "TRegexp.h"

// URANIE
#include "UExceptions.h"
#include "UTimer.h"
#include "TAttribute.h"
#include "TSlurmCluster.h"

ClassImp (URANIE::Launcher::TLauncherByStepRemote);

namespace URANIE
{
namespace Launcher
{

////////////////////////////////////////////////////////////
//  Default constructor
////////////////////////////////////////////////////////////
TLauncherByStepRemote::TLauncherByStepRemote(URANIE::DataServer::TDataServer *tds, TCode *code,EDistrib eDistrib, Option_t *option ) :
    TLauncher(tds, code, option), _nAllocatedSeconds(1800), _nAllocatedCores(0), _sProjectName("")
{
    setSave();
    setClean();
    static const char alphanum[] =
        "0123456789"
        "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz";
    srand(time(NULL));
    _sJobName = alphanum[ rand() % sizeof(alphanum)];
    _edistrib =eDistrib;
    if (_edistrib == kUnknown)
    {
        _cluster = new TSlurmCluster(tds,code);
        _cluster->_eType = "kSLURM";
    }
    else if (_edistrib == kEc)
    {
        _cluster = new TEcCluster(tds,code);
        _cluster->_eType = "kEc";
    }
    else if (_edistrib == kSLURM)
    {
        _cluster = new TSlurmCluster(tds,code);
        _cluster->_eType = "kSLURM";
    }
    else
    {
        _cluster = new TSlurmCluster(tds,code);
        _cluster->_eType = "kSLURM";
    }
    eType=_cluster->_eType;

}

TLauncherByStepRemote::TLauncherByStepRemote(URANIE::DataServer::TDataServer *tds, TCode *code,TCluster* cluster, Option_t *option ):
    TLauncher(tds, code, option), _nAllocatedSeconds(1800), _nAllocatedCores(0), _sProjectName("")
{

    _cluster=cluster;
    eType=cluster->_eType;
    _cluster->_code=cluster->_code;
}
////////////////////////////////////////////////////////////
//  Default destructor
////////////////////////////////////////////////////////////
TLauncherByStepRemote::~TLauncherByStepRemote()
{

}


////////////////////////////////////////////////////////////
//  preTreatment
////////////////////////////////////////////////////////////

TCluster* TLauncherByStepRemote::getCluster()
{
    return _cluster;
}
void TLauncherByStepRemote::preTreatment(Option_t *option)
{
    if (_sSingleJob)
    {
        _sJobName=".";
        _cluster->setJobName(_sJobName);
    }


    Bool_t runAll = kFALSE;
    if (_sworkingDirectory.Length() == 0 || !_sJobList.empty() )
    {
        _sworkingDirectory=_sdirectory;
    }
    TString aux=gSystem->ConcatFileName(_sworkingDirectory,"URANIE");
    _sworkingDirectory=TString(gSystem->ConcatFileName(aux,_sJobName));
    if ( _sworkingDirectory.Contains(_sJobName+"/URANIE"))
    {
        int i=0;
        while (i < 6)
        {
            aux.Chop();
            i++;
        }
    }
    _cluster->setWorkingDirectory(_sworkingDirectory);
    _code->setWorkingDirectory(_sworkingDirectory);
    if (_blog)
        cout << " ** TLauncherByStepRemote::preTreatment" << endl;
    TString soption =option;
    soption.ToLower();
    Int_t npat = (Int_t)(_tds->getNPatterns());
    Int_t start=0,end=npat,maxIterations=1,initialSplit=0;
    _sChunkSize = (npat + _sNumberOfChunks - 1) / _sNumberOfChunks;
    // ----------------------------------------------------------
    //  LOOP
    // ----------------------------------------------------------
    if ( soption.Contains("nsplit=") && !soption.Contains("nsplit=all")  && _sNumberOfChunks > 1 )
    {
        std::vector<TRegexp> regular;
        regular.push_back(TRegexp("nsplit=[0-9]+"));
        regular.push_back(TRegexp("nsplit=\\[[0-9]+-[0-9]+\\]"));
        TString out="";
        unsigned int i=0;
        do
        {
            out = soption(regular.at(i));
            i++;
        }while(out.Length() == 0 && i < regular.size() );
        if (i > regular.size())
        {
            if(_DoLogger)
                (*_fLogger) << Endl << "Invalid value for nsplit= it must be a number {accepted formats are  [N-N] or N}";
            //exit(-1);
        }

        TString nsplit = out.Remove(0, 7);
        soption.ReplaceAll(Form("nsplit=%s", out.Data()), "");
        //        if (nsplit.Length() == 1 && nsplit.IsDigit()  )
        if ( nsplit.IsDigit() )
        {

            initialSplit= nsplit.Atoi();
            maxIterations = 1;
        }
        else if (nsplit.BeginsWith("[") && nsplit.EndsWith("]") )
        {

            initialSplit = TString(nsplit[1]).Atoi();
            maxIterations = TString(nsplit[3]).Atoi() -initialSplit;
        } else
        {
            if(_DoLogger)
                (*_fLogger) << Endl << "Invalid format";
            //exit(-1);
        }

        if( _blog && _DoLogger)
            (*_fLogger)  << endl << " ** Nsplit read in option[" << nsplit << "]" << endl;

        if ( initialSplit+1 > _sNumberOfChunks || (initialSplit+maxIterations) > _sNumberOfChunks)
        {
            if(_DoLogger)
                (*_fLogger) << Endl << " Split number is outbound" ;
            //exit(-1);
        }
        start = initialSplit  * _sChunkSize;
        end = TMath::Min( start + _sChunkSize, npat);

    }
    else if ( !soption.Contains("nsplit=") || soption.Contains("nsplit=all") )
    {
        runAll=kTRUE;
        initialSplit = start = 0;
        end = _sChunkSize;
        maxIterations=_sNumberOfChunks;
    }
    TString soptionCluster;
    soptionCluster.Form("start=%d,end=%d", start, end+(_sChunkSize*(maxIterations-1)));
    soptionCluster.Form("start=%d,end=%d",start,end+(_sChunkSize*(maxIterations-1)));
    _sLastSplit = std::make_pair(start, end+(_sChunkSize*(maxIterations-1)));
    if(_DoLogger)
        (*_fLogger) << Endl<< " ** TLauncherByStepRemote::preTreatment" << Endl;

    if (_bsave && _DoLogger)
        (*_fLogger) << " ** Save : Yes" << Endl;
    else if (_DoLogger)
        (*_fLogger) << " ** Save : No" << Endl;
    if (_sJobList.empty() )
    {
        // Intialisation and Define the attributs _sworkingDirectory by default
        initiateTheEntryVector();
        init(option);
        // Clean if necessary
        clean(option);
        _code->init(option);

    }

    _cluster->preTreatment(soptionCluster);

    //if (_sListCompilerDirectives.size() > 0 && !_cluster->isConnected())
    if (_sListCompilerDirectives.size() > 0)
    {
        addCompilerDirective("workingDir",_sworkingDirectory);
        compileBinary();

    }
    _cluster->generateHeader(soptionCluster);
    if (soption.Contains("simultaneous"))
    {
        clusterRun(option);
    }

#ifndef WIN32
    URANIE::DataServer::UTimer timer(npat, "Uranie Launcheur");
#endif
    if (_cluster->openSFTPChannel() != SSH_OK) {cout << "(OPEN)Error with ftp session" << endl;}
    cout << "SSH_OK: " << SSH_OK << endl;
    for (Int_t Chunk = 0; Chunk < maxIterations; Chunk++)
    {
        for (Int_t ipat = start + Chunk * _sChunkSize; ipat < end + Chunk * _sChunkSize; ipat++) {
            _code->init(option);
            _tds->getTuple()->GetEntry(ipat);
            _code->preTraitment(ipat, &_InputVector, option);
            if (_blog)
                cout << endl << " ** TLauncherByStepRemote::preTreatment iter[" << 1.0 + ipat << "]" << endl;
#ifndef WIN32
            if (_bdrawProgressBar)
                timer.DrawProgressBar(ipat, 0);
#endif
        }  // End of LOOP
        if (soption.Contains("simultaneous")) {
            if (_cluster->sendInputFiles(start, end) != SSH_OK) cout << "Error" << endl;
        }
    }

    if (!soption.Contains("simultaneous"))
    {
        if (_cluster->sendInputFiles(0 , npat)  != SSH_OK) cout << "Error" <<endl ;
    }

    if ( _cluster->closeSFTPChannel() != SSH_OK) cout <<  "Error";
#ifndef WIN32
    if(_DoLogger)
        (*_fLogger) << Endl<< "time elapsed " << timer.GetElapsedTime() << Endl;
#endif        

    // Terminate
    gSystem->cd(_sdirectory);

    if (_blog)
    {
        cout << " **************************************" << endl;
        cout << " ** End Of TLauncherByStepRemote::preTreatment" << endl;
        _tds->getTuple()->Print();
        cout << " **************************************" << endl;
    }

}
////////////////////////////////////////////////////////////
//  run
////////////////////////////////////////////////////////////
void TLauncherByStepRemote::run(Option_t *option)
{

    TString soption = option ;
    soption.ToLower();
    preTreatment(soption);
    if (!soption.Contains("simultaneous"))
    {
        clusterRun(soption);
    }
    gSystem->cd(_sdirectory);

    //postTreatment();
}
void TLauncherByStepRemote::compileBinary()
{
    std::map<TString, TString>::iterator it = _sListCompilerDirectives.begin();
    std::list<TString> filenames;
    TString scriptName;
    bool found=false;
    while(it != _sListCompilerDirectives.end() && !found)
    {
        if ((*it).first.Contains("exe"))
        {
            scriptName=(*it).second;
            found=true;
        }
        it++;
    }
    if ( !found)
    {
        cout << "ERROR CAN'T COMPILE SCRIPT , NAME DIRECTIVE NOT FOUND" << endl;
        exit(-1);
    }
    gSystem->cd(_sworkingDirectory);

    gSystem->Exec("mkdir bin; mkdir CODE ; mkdir lib");
    void *dirp = gSystem->OpenDirectory(gSystem->ConcatFileName(_sdirectory,"CODE"));
    Char_t *afile;
    int i=0;
    while ( (afile = const_cast<Char_t *>(gSystem->GetDirEntry(dirp)) ) ) // now it doesn't return . and .. i > 1 is not needed.
    {
        gSystem->CopyFile(gSystem->ConcatFileName(_sdirectory,"CODE/"+TString(afile)), gSystem->ConcatFileName(_sworkingDirectory,"CODE/"+TString(afile)), kTRUE);
        i++;
    }
    gSystem->FreeDirectory(dirp);

    _cluster->substituteDirectives(gSystem->ConcatFileName(_sdirectory,_sCompilerSkeleton),gSystem->ConcatFileName(_sworkingDirectory,"CMakeLists.txt"),  _sListCompilerDirectives);
    substituteCompilerCode();

    TString command="mkdir build ;  cd build ; cmake -DCMAKE_INSTALL_PREFIX="+TString(gSystem->ConcatFileName(_sworkingDirectory,+"bin"))+" .. ; make";

    gSystem->Exec(command);
    filenames.push_back(scriptName);
    _cluster->openSFTPChannel();
    _cluster->sendCompiledBinary(filenames);
    _cluster->closeSFTPChannel();


    gSystem->cd(_sdirectory);
}
void TLauncherByStepRemote::addCompilerDirective(TString directive, TString value)
{
    directive = "@"+directive+"@";
    _sListCompilerDirectives.insert(std::make_pair(directive,value));
}

void TLauncherByStepRemote::addCodeLibrary(TString libName, TString sourcesList)
{
    _sCodeLibs.insert(std::make_pair(libName,sourcesList));
}
void TLauncherByStepRemote::addCodeExternalLibrary(TString libName, TString libPath , TString includePath)
{
    _sExternalLib.emplace_back(make_pair(libName,std::make_pair(libPath,includePath)));

}
void TLauncherByStepRemote::addCodeExternalLibrary(TString libName)
{
    _sExternalLib.emplace_back(make_pair(libName,std::make_pair("","")));}
void TLauncherByStepRemote::addCodeDependency(TString destinationLib, TString sourceList)
{
    _sCodeDependencies.insert(std::make_pair(destinationLib,sourceList));
}

void TLauncherByStepRemote::substituteCompilerCode() {
    TString path = gSystem->ConcatFileName(_sdirectory, "CODE");
    std::ifstream inputfile(gSystem->ConcatFileName(_sworkingDirectory,"CMakeLists.txt")); // this should be cmakelists.txt
    std::stringstream infile;
    infile << inputfile.rdbuf();
    inputfile.close();
    std::ofstream os(gSystem->ConcatFileName(_sworkingDirectory,"CMakeLists.txt"), ios::trunc);
    std::string line;
    TString auxLine;

    while (std::getline(infile, line))
    {
        TString context;
        auxLine = TString(line);
        if (auxLine.Contains("@addLibrary@") )
        {
            TString pathToSource="${PROJECT_SOURCE_DIR}/CODE/";
            auto libPtr = _sCodeLibs.begin();
            while (libPtr != _sCodeLibs.end())
            {
                TString text = "\nset("+libPtr->first+"_SRCS ";
                TObjArray * libSources;
                Bool_t singleSource=kFALSE;
                if ( libPtr->second.Contains(",") )
                    libSources= libPtr->second.Tokenize(",");
                else if( libPtr->second.Contains(" ") )
                    libSources= libPtr->second.Tokenize(" ");
                else if( libPtr->second.Contains("|") )
                    libSources= libPtr->second.Tokenize("|");
                else
                    singleSource =kTRUE;
                if (!singleSource)
                {
                    for ( auto it = 0; it < libSources->GetSize(); it++)
                    {
                        text.Append(pathToSource+((TObjString *)(libSources->At(it)))->String()+" ");
                    }
                    text.Chop();
                }
                else
                {
                    text.Append(pathToSource+libPtr->second);
                }
                text.Append(")\n");
                text.Append("add_library("+libPtr->first+" STATIC ${"+libPtr->first+"_SRCS}) #Always STATIC (FOR NOW) \n");
                text.Append("target_include_directories("+libPtr->first+" PUBLIC ${LIBRARY_OUTPUT_PATH})");
                os << text << endl;
                libPtr++;
            }
        }
        else if (auxLine.Contains("@addDependency@"))
        {
            auto dependenciesPtr = _sCodeDependencies.begin();
            while (dependenciesPtr != _sCodeDependencies.end())
            {
                TString text = "\ntarget_link_libraries("+dependenciesPtr->first+" ";
                TObjArray * libs;
                Bool_t singleLib=kFALSE;
                if ( dependenciesPtr->second.Contains(",") )
                    libs= dependenciesPtr->second.Tokenize(",");
                else if( dependenciesPtr->second.Contains(" ") )
                    libs= dependenciesPtr->second.Tokenize(" ");
                else if( dependenciesPtr->second.Contains("|") )
                    libs= dependenciesPtr->second.Tokenize("|");
                else
                    singleLib =kTRUE;
                if (!singleLib)
                {
                    for ( auto it = 0; it < libs->GetSize(); it++)
                    {
                        text+=((TObjString *)(libs->At(it)))->String()+" ";
                    }
                    text.Chop();
                }
                else
                {
                    text+=dependenciesPtr->second;
                }
                text.Append(")");
                os << text << endl;
                dependenciesPtr++;
            }
        }
        else if (auxLine.Contains("@addExternalLibrary@"))
        {
            auto extLibPtr = _sExternalLib.begin();
            while (extLibPtr != _sExternalLib.end())
            {
                TString path = gSystem->ConcatFileName(_sworkingDirectory, "CODE/" + extLibPtr->first + ".cmake");
                gSystem->Exec("touch " + path);
                os << "\nfind_package("+extLibPtr->first+ " REQUIRED)" << endl;
                if ( extLibPtr->second.first.Length() > 0 &&  extLibPtr->second.second.Length() > 0  )
                {
                    TObjArray * libs;

                    Bool_t singleLib,singlePath=kFALSE;
                    TString libsStr,frontLib;
                    if ( extLibPtr->first.Contains(",") )
                        libs = extLibPtr->first.Tokenize(",");
                    else if( extLibPtr->first.Contains(" ") )
                        libs= extLibPtr->first.Tokenize(" ");
                    else if( extLibPtr->first.Contains("|") )
                        libs= extLibPtr->first.Tokenize("|");
                    else
                    {
                        singleLib =kTRUE;
                        frontLib=libsStr=extLibPtr->first;
                    }
                    if (!singleLib)
                    {
                        for ( auto it = 0; it < libs->GetSize(); it++)
                        {
                            libsStr+=((TObjString *)(libs->At(it)))->String()+" ";
                        }
                        libsStr.Chop();
                        frontLib=((TObjString *)(libs->At(0)))->String();
                    }
                    TString libLocation=frontLib+"_LIB_LOCATIONS";
                    TString includeLocation=frontLib+"_INCLUDE_LOCATIONS";
                    TString text;
                    std::ofstream osCmake(gSystem->ConcatFileName(_sworkingDirectory,"CODE/"+extLibPtr->first+".cmake"), ios::trunc);

                    TString pathsLibStr="";
                    TString pathsHeaderStr="";
                    if ( extLibPtr->second.first.Length() > 0 ) // LIBS
                    {
                        TObjArray * pathsLib;

                        Bool_t singlePath= kFALSE;
                        if ( extLibPtr->second.first.Contains(",") )
                            pathsLib= extLibPtr->second.first.Tokenize(",");
                        else if( extLibPtr->second.first.Contains(" ") )
                            pathsLib= extLibPtr->second.first.Tokenize(" ");
                        else if( extLibPtr->second.first.Contains("|") )
                            pathsLib= extLibPtr->second.first.Tokenize("|");
                        else {
                            singlePath = kTRUE;
                            pathsLibStr=extLibPtr->second.first;
                        }
                        if (!singlePath)
                            for ( auto it = 0; it < pathsLib->GetSize(); it++)
                            {
                                pathsLibStr+=((TObjString *)(pathsLib->At(it)))->String()+" ";
                            }


                        text+="set(_"+libLocation+" "+pathsLibStr+")";
                        text += "\nfind_library(" + frontLib + "_LIBRARY NAMES" + libsStr + " HINTS " +"$("+libLocation+")";
                    } else
                    {
                        text += "\nfind_library(" + frontLib + "_LIBRARY NAMES" + libsStr + " NO_DEFAULT_PATH)";
                        text += "\nfind_library(" + frontLib + "_LIBRARY NAMES" + libsStr + ")";
                    }
                    if ( extLibPtr->second.second.Length() > 0 ) // HEADER
                    {
                        TObjArray * pathsHeader;
                        Bool_t singlePathHeader= kFALSE;
                        if ( extLibPtr->second.second.Contains(",") )
                            pathsHeader= extLibPtr->second.second.Tokenize(",");
                        else if( extLibPtr->second.second.Contains(" ") )
                            pathsHeader= extLibPtr->second.second.Tokenize(" ");
                        else if( extLibPtr->second.second.Contains("|") )
                            pathsHeader= extLibPtr->second.second.Tokenize("|");
                        else {
                            singlePathHeader = kTRUE;
                            pathsHeaderStr=extLibPtr->second.second;
                        }
                        if (!singlePathHeader)
                            for ( auto it = 0; it < pathsHeader->GetSize(); it++)
                            {
                                pathsHeaderStr+=((TObjString *)(pathsHeader->At(it)))->String()+" ";
                            }
                        text+="set(_"+includeLocation+" "+pathsHeaderStr+")";
                        text += "\nfind_path(" + frontLib + "_INCLUDE_DIR NAMES" + libsStr + ".h HINTS " + "$("+includeLocation+")";
                    }
                    else{
                        text ="\nfind_path("+frontLib+"_INCLUDE_DIR "+frontLib+".h NO_DEFAULT_PATH)";
                        text ="\nfind_path("+frontLib+"_INCLUDE_DIR "+frontLib+".h)";

                    }
                    text+="\ninclude(FindPackageHandleStandardArgs)";
                    text+="\nfind_package_handle_standard_args("+frontLib+" DEFAULT_MSG "+frontLib+"_LIBRARY "+frontLib+"_INCLUDE_DIR)";
                    text+="\nmark_as_advanced("+includeLocation+" "+libLocation+")";
                    osCmake << text << endl;
                }
                extLibPtr++;
            }
        }
        else
        {
            os << line << endl;
        }
    }
    os.close();
}


void TLauncherByStepRemote::clusterRun(Option_t *option)
{
    UNUSED(option);

    if (_blog)
    {
        cout << " **************************************" << endl;
        cout << " ** TLauncherByStepRemote::run" << endl;
    }
    if(_DoLogger)
        (*_fLogger) << Endl<< " ** TLauncherByStepRemote::run" << Endl;

    Int_t npat = (Int_t)(_tds->getNPatterns());
#ifndef WIN32
    URANIE::DataServer::UTimer timer(npat, "Uranie Launcheur");
#endif
    //if ( gSystem->AccessPathName(gSystem->ConcatFileName(_sworkingDirectory,"joblist_"+_cluster->getSplitName())) == kFALSE )
    //{
    TString command =TString("sbatch ") +gSystem->ConcatFileName(_cluster->getRemotePath(), _cluster->getSplitName());
    if (eType == "kSLURM" ) {
        if (strcmp(_cluster->getClusterName().Data(), "marenostrum4") == 0) {
            TString command =
                TString("sbatch ") +
                gSystem->ConcatFileName(_cluster->getRemotePath(), _cluster->getSplitName());
            TString jobid = _cluster->submitJob(command);
            //BACKUP
            std::ofstream os(gSystem->ConcatFileName(_sworkingDirectory, "joblist_" + _cluster->getJobName()),
                             ios::app);
            os << jobid << "|" << _sLastSplit.first << "|" << _sLastSplit.second << endl;
            os.close();
            _sJobList.push_back(std::make_pair(jobid, _sLastSplit)); // MAKE A CONTROL IF NOT WORKING
        } else {
            TString command =
                TString("sbatch ") +
                gSystem->ConcatFileName(_cluster->getRemotePath(), _cluster->getSplitName());
            TString jobid = _cluster->submitJob(command);

            _sJobList.push_back(std::make_pair(jobid, _sLastSplit));
        }// SLURM DEFAULT
    }
    else if (eType == "kEc" )
    {
        TString command =
            TString("ecaccess-job-submit -distant -queueName ")  + _cluster->getGateway() +
            + " " + _cluster->getGateway()+":"+gSystem->ConcatFileName(_cluster->getRemotePath(), _cluster->getSplitName());
        TString jobid = _cluster->submitJob(command);
        std::ofstream os(gSystem->ConcatFileName(_sworkingDirectory, "joblist_" + _cluster->getJobName()),
                         ios::app);
        os << jobid << "|" << _sLastSplit.first << "|" << _sLastSplit.second << endl;
        os.close();
        _sJobList.push_back(std::make_pair(jobid, _sLastSplit));
    }
    //}


#ifndef WIN32
    if(_DoLogger)
        (*_fLogger) << Endl<< "time elapsed " << timer.GetElapsedTime() << Endl;
#endif
    // Terminate
    //gSystem->cd(_sdirectory);
    if (_blog)
    {
        cout << endl;
        cout << " ** End Of TLauncherByStepRemote::run" << endl;
        cout << " **************************************" << endl;
    }

}

////////////////////////////////////////////////////////////
//  postTreatment
////////////////////////////////////////////////////////////

void TLauncherByStepRemote::UpdateJobListStatus()
{

    auto auxActiveJobs=_sActiveJobs;
    auto job = auxActiveJobs.begin();
    TString command;
    if (_edistrib == kEc)
        command = _cluster->getJobStatusCommand(auxActiveJobs);
    else
    {
        command = _cluster->getJobStatusCommand(_sJobIDs);
    }
    while ( job !=  auxActiveJobs.end() )
    {
        auto jobFound = std::find(_sJobStatus.begin(), _sJobStatus.end(), (*job).first);
        TString jobStatus = _cluster->checkJobStatus((*job).first,command);
        if ( jobStatus.Contains("COMPLETED") )
        {
            if (jobFound != _sJobStatus.end())
            {
                (*jobFound)="[COMPLETED|"+ (*job).first + "]";
            }
            _sCompletedJobs.push_back((*job));
            _sActiveJobs.remove((*job));
        } else if (jobStatus.Contains("FAILED") )
        {
            if (jobFound != _sJobStatus.end())
            {
                (*jobFound)= "[FAILED|"+ (*job).first + "]";
            }
            _sCompletedJobs.push_back((*job));
            _sActiveJobs.remove((*job));
        } else
        {
            if (jobFound != _sJobStatus.end())
            {
                (*jobFound)="[RUNNING|"+ (*job).first + "]";
            }

        }
        //SafeWaiting
        job++;
        sleep(2);
    }
    if(_DoLogger)
        (*_fLogger) << Endl << "JOB STATUS ";
    TString output="{";
    for (auto it = _sJobStatus.begin(); it != _sJobStatus.end() ; it++ )
    {
        output+=(*it)+",";
    }
    output.Chop();
    output+="}";
    if(_DoLogger)
        (*_fLogger) << output << Endl;
}

void TLauncherByStepRemote::RetrieveJobOutput(std::pair<Int_t,Int_t> split)
{
    Int_t npat = (Int_t)(_tds->getNPatterns());
    if (_edistrib == kSLURM)
        if (_cluster->openSFTPChannel() != SSH_OK) {cout << "(OPEN)Error with ftp session" << endl; exit(-1);}

    for (Int_t ipat = split.first; ipat < TMath::Min( split.second, npat); ipat++)
    {
        TString sdir = Form("UranieLauncher_%d", ipat + 1);
        _cluster->retrieveInputFiles(sdir);
    }
    if (_edistrib == kSLURM)
        if (_cluster->closeSFTPChannel() != SSH_OK) {cout << "(OPEN)Error with ftp session" << endl; exit(-1);}
}

void TLauncherByStepRemote::postTreatment(Option_t *option) {
    if (_blog)
        cout << " ** TLauncherByStepRemote::postTreatment" << endl;
    if(_DoLogger)
        (*_fLogger) << endl << " ** TLauncherByStepRemote::postTreatment" << endl;
    TString srootfile = _sworkingDirectory;
    if (_blog)
        cout << " ** srootfile    getWorkingDirectory[" << srootfile << "]" << endl;
    srootfile += TString("/UranieResults/uranielauncher.root");
    if (_blog)
        cout << " ** srootfile[" << srootfile << "]" << endl;
    //Deal with the intermediate file
    TString soption(option);
    soption.ToLower();
    if (soption.Contains("nointermed"))
        setNoIntermediateSaved();
    Int_t npat = (_tds->getNPatterns());
    //Create the ntuple and go through the final results output file
    std::stringstream sstr;
    sstr.precision(SSTR_PREC);
    TString sntlname;
    TString slabelvarname = "Launcher__n__iter__:n__calculs__";
    TString slabelvartype = "D:D";
    URANIE::DataServer::TDSNtupleD *ntl = createNtupleDFile(sntlname, slabelvarname, slabelvartype);
    Int_t nFailure = 0;
    TFile *rootfile;
    // -------------------------------------------------------------
    //  LOOP
    // -------------------------------------------------------------
    URANIE::DataServer::UTimer timer(npat, "Uranie Launcheur");
    auto aJob = _sJobList.begin();
    while (aJob != _sJobList.end())
    {
        _sJobStatus.push_back((*aJob).first);
        _sJobIDs+=(*aJob).first+",";

        _sActiveJobs.push_back(*aJob);
        aJob++;
    }
    _sJobIDs.Chop();
    _sJobList.clear();
    do{
        UpdateJobListStatus();
        auto completedJobs=_sCompletedJobs;
        auto job = completedJobs.begin();
        while (job != completedJobs.end())
        {
            RetrieveJobOutput((*job).second);
            _sCompletedJobs.remove((*job));
            job++;
        };
    } while (!_sActiveJobs.empty() || !_sCompletedJobs.empty()   );


    for (Int_t ipat = 0; ipat < npat; ipat++)
    {
        TString sdir = Form("%s/UranieLauncher_%d", _sworkingDirectory.Data(), ipat + 1);
        gSystem->cd(sdir);
        if (_blog)
            cout << endl << " ** TLauncher::run iter[" << 1.0 + ipat << "]" << endl;
        if (_code->postTraitment(&_OutputVector, _sopt) == kFALSE)
            nFailure++;
        // Store the Output Files if request by addSaveFile(TString) method
        if (_listOfSaveFile)
        {
            cout << endl << " ** Save the files" << endl;
            _listOfSaveFile->Print();
            TNamed *obj = NULL;
            TIter next(_listOfSaveFile);
            while ((obj = (TNamed *) next()))
            {
                TString sfilename = obj->GetName();
                TString soutname = Form("%s.%d", obj->GetTitle(), ipat + 1);
                Int_t nret = gSystem->CopyFile(sfilename.Data(), Form("%s/%s", _sdirectory.Data(), soutname.Data()),kTRUE);
                if (nret == -1)
                    throw URANIE::Exceptions::UExceptions(__FILE__, __LINE__,
                                                          Form("TLauncher::run ERROR : for copy output file \"%s\" to \"%s\"",sfilename.Data(), soutname.Data()));
            }
        }
        _OutputVector.at(0)->copyEntry(_InputVector.at(0));
        ntl->Fill(&_OutputVector);
        if (_blog) {
            cout << " ** End Of Analysis of several outputs" << endl;
            cout << " *************************************" << endl << endl;
        }
        // ---- NEWS ----
        // Save the TDSNtupleD in a rootfile
        if (!_bnoIntermediateSaved) {
            rootfile = TFile::Open(srootfile.Data(), "RECREATE");
            if (_blog) {
                cout << " ** TLauncherByStepRemote Current directory[" << gSystem->pwd()
                     << "]" << endl;
                cout << " **** srootfile[" << srootfile << "]" << endl;
            }
            ntl->SetDirectory(0);
            ntl->Write();
            rootfile->Close();
            ntl->SetDirectory(gROOT);
            if (ipat % 5 == 0)
                gSystem->CopyFile(srootfile.Data(), Form("%s.save", srootfile.Data()), kTRUE);
        }
#ifndef WIN32
        if (_bdrawProgressBar)
            timer.DrawProgressBar(ipat, nFailure);
#endif
    }  // End of LOOP
    gSystem->cd(srootfile);
    if(_bnoIntermediateSaved )
    {
        rootfile = TFile::Open(srootfile.Data(), "RECREATE");
        if (_blog)
        {
            cout << " ** TLauncherByStepRemote Current directory[" << gSystem->pwd()
                 << "]" << endl;
            cout << " **** srootfile[" << srootfile << "]" << endl;
        }
        ntl->SetDirectory(0);
        ntl->Write();
        rootfile->Close();
    }

    // Merge the input and output ntupleD
    vector<string> buffer;
    URANIE::DataServer::UEntry::StringParser( string(slabelvarname), ":", buffer);
    string thevars="";
    for (unsigned int ivar=1; ivar<buffer.size(); ivar++)
        thevars+=buffer.at(ivar)+((ivar!=buffer.size()-1)?":":"");

    _tds->getTuple()->Merger(ntl, thevars, thevars,false,_tds->getIteratorName(),"Launcher__n__iter__");
    if(_DoLogger)
        (*_fLogger) << Endl<< "time elapsed " << timer.GetElapsedTime() << Endl;
    if (nFailure > 0 && _DoLogger)
        (*_fLogger) << "     Failure : \033[0;31m" << nFailure
                    << "\033[0m/\033[0;31m" << npat << "\033[0m"
                    << " --- \033[0;31m"
                    << setw(1) << 100. * Double_t(nFailure)/Double_t(npat) << "%\033[0m" << Endl;
    // Terminate
    terminate(option);
    gSystem->cd(_sdirectory);
    // Affect the current TTree to the TDataServer
    if (_blog)
        cout << " ** End of Lancher ntall" << endl;
    delete ntl;
    if (_blog)
    {
        cout << " ** End of Lancher tds ****************" << endl;
        _tds->getTuple()->Print();
    }
    //if ( gSystem->AccessPathName(gSystem->ConcatFileName(_sworkingDirectory,"joblist_"+_cluster->getSplitName()))  == kFALSE )
    //    gSystem->Unlink(gSystem->ConcatFileName(_sworkingDirectory,"joblist_"+_cluster->getSplitName()));

}
////////////////////////////////////////////////////////////
//  print the Log
////////////////////////////////////////////////////////////
void TLauncherByStepRemote::printLog(Option_t *option)
{
    cout << "TLauncherByStepRemote::printLog[" << option << "]" << endl;
}

} // End Of namespace Launcher
} // End Of namespace URANIE

//  End Of File TLauncherByStepRemote.cxx
