//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TLauncherFunction.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   thu jun 14 11:49:19 CEST 2007

 \brief Implementation of the class URANIE::Launcher::TLauncherFunction

 \par Modifications
 \li  Cedric Forgues (C.S.) 2010/03/02
 New input given to run method
 \li Cedric Forgues (C.S.) 2010/01/28
 run(). Now for only one output attribute, the name of the leaf is the name of the attribute.
 */
#include <sstream>
#include <string>

// ROOT
//#include "Api.h"
#include "TLeaf.h"
#include "TList.h"

// Uranie
#include "TLauncherFunction.h"
#include "UExceptions.h"
#include "UTimer.h"

using std::cout;
using std::endl;

ClassImp (URANIE::Launcher::TLauncherFunction)

namespace URANIE
{
namespace Launcher
{

//////////////////////
//  Constructeur
//////////////////////
  TLauncherFunction::TLauncherFunction(URANIE::DataServer::TDataServer *tds, const char *functionanme, TString sinput, TString soutput) :
      _fMethodCall(NULL), _fMethod(NULL), _soutput(soutput), _listOfInputAttributes(NULL), _blog(kFALSE), 
      _bdrawProgressBar(kTRUE), _internal(NULL)
  {
    if (_blog) {
        cout << "************* TLauncherFunction Constructor *******************" << endl;
        cout << " ** functionanme[" << functionanme << "]" << endl;
        cout << " ** sinput[" << sinput << "]" << endl;
        cout << " ** soutput[" << soutput << "]" << endl;
      }
    
    SetNameTitle(functionanme, Form("Launcher the Analystical Function [%s]", functionanme));

    try   {
        if (tds == NULL)
	        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__, 
	            "TLauncherFunction::TLauncherFunction ERROR : the TDS  is null");

        _tds = tds;

        if (_blog)
            cout << "***************** TMethodCall with [" << functionanme << "]**************" << endl;
        _fMethodCall = new TMethodCall();
        _fMethodCall->InitWithPrototype(functionanme, "Double_t*,Double_t*");

	if ( _fMethodCall->IsValid() == kFALSE ) {
	    _fMethodCall->Print();
	        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
	            Form("TLauncherFunction::TLauncherFunction ERROR : the Function[%s] was not found", functionanme));
      
	  }
        if (_blog) _fMethodCall->Print();
      
        // Get the current list of attributes
        _listOfInputAttributes = _tds->getListOfAttributes(sinput.Data());
    
        Int_t nvars = _listOfInputAttributes->GetEntries();
        if (_blog) {
            cout << endl << " **  Attributes listing  [" << nvars << "]" << endl;
            _listOfInputAttributes->Print();
        }
    
        // The output of the function
        if (_soutput.Length() == 0) _soutput = functionanme;
    
    _nOutput = soutput.CountChar(':') + 1;
  } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
      ue.printMessage();
      if(_fMethodCall!=NULL)
	delete _fMethodCall;
      throw ue;
    }

}

#ifdef ROOT5
  TLauncherFunction::TLauncherFunction(URANIE::DataServer::TDataServer *tds, void *fcn, TString sinput, TString soutput) :
     _fMethodCall(NULL), _fMethod(NULL), _soutput(soutput), _listOfInputAttributes(NULL), 
     _blog(kFALSE), _bdrawProgressBar(kTRUE), _internal(NULL)
  {
    if (_blog) {
        cout << "************* TLauncherFunction Constructor *******************" << endl;
        cout << " ** sinput[" << sinput << "]" << endl;
        cout << " ** soutput[" << soutput << "]" << endl;
    }
    try   {
      if (tds == NULL)
	throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__, "TLauncherFunction::TLauncherFunction ERROR : the TDS  is null");

      _tds = tds;

      char *funcname = G__p2f2funcname(fcn);
      
      if (_blog)
        cout << " ** funcname[" << funcname << "]" << endl;
      
      SetNameTitle(TString(funcname), Form("Launcher the Analystical Function [%s]", funcname));
      
      _fMethodCall = new TMethodCall();
      _fMethodCall->InitWithPrototype(funcname, "Double_t*,Double_t*");

      if ( _fMethodCall->IsValid() == kFALSE )
	throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__, Form("TLauncherFunction::TLauncherFunction ERROR : the Function[%s] was not found", funcname));
      
      if (_blog) _fMethodCall->Print();
      
      //_fMethod = UranieLauncherFunctionInteractiveMethod;
      //gPlexnum = this;
      
      
      // Get the current list of attributes
      _listOfInputAttributes = _tds->getListOfAttributes(sinput.Data());
      
      Int_t nvars = _listOfInputAttributes->GetEntries();
      if (_blog) {
	cout << endl << " **  Attributes listing  [" << nvars << "]" << endl;
	_listOfInputAttributes->Print();
      }
      
      if (_soutput.Length() == 0)
        _soutput = TString(funcname);
      
      _nOutput = _soutput.CountChar(':') + 1;
      
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
	ue.printMessage();
	if(_fMethodCall!=NULL)
	  delete _fMethodCall;
	throw ue;
      }
}
#else
TLauncherFunction::TLauncherFunction(URANIE::DataServer::TDataServer *tds,
        void (*fcn)(Double_t*,Double_t*),
        TString sinput, TString soutput) :
    _fMethodCall(NULL), _fMethod(NULL), _soutput(soutput), _listOfInputAttributes(NULL),
    _blog(kFALSE), _bdrawProgressBar(kTRUE), _internal(NULL)
{

    if (_blog) {
        cout << "************* TLauncherFunction Constructor *******************" << endl;
        cout << " ** sinput[" << sinput << "]" << endl;
        cout << " ** soutput[" << soutput << "]" << endl;
    }
    
    try   {
        if (tds == NULL)
	        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__, 
	            "TLauncherFunction::TLauncherFunction ERROR : the TDS  is null");
        if (fcn == NULL)                
	        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__, 
	            "TLauncherFunction::TLauncherFunction ERROR : null function pointer");
        if (_soutput.Length() == 0)
  	        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__, 
	            "TLauncherFunction::TLauncherFunction ERROR : empty string for outputs");
        
        _fMethod = fcn;
        _tds = tds;
        SetNameTitle(soutput.Data(), Form("Launcher the Analystical Function [%s]", soutput.Data()));
        // Get the current list of attributes
        _listOfInputAttributes = _tds->getListOfAttributes(sinput.Data());
        Int_t nvars = _listOfInputAttributes->GetEntries();
        _nOutput = _soutput.CountChar(':') + 1;
      
        if (_blog) {
            cout << endl << " **  Attributes listing  [" << nvars << "]" << endl;
            _listOfInputAttributes->Print();
        }
      
    }
    catch (URANIE::Exceptions::UErrorExceptions& ue) {
	    ue.printMessage();
	    throw ue;
    }
    
}
#endif

TLauncherFunction::~TLauncherFunction()
{
    if (_blog)
        cout << "detruit TLauncherFunction" << endl;

    if (_listOfInputAttributes)
    {
        delete _listOfInputAttributes;
    }

    if (_fMethodCall)
        delete _fMethodCall;
}

//////////////////////
//  init
//////////////////////

//////////////////////
//  Run
//////////////////////
void TLauncherFunction::run(TString sinput, TString soutput, Option_t *option)
{
  TString soption = option;
  soption.ToLower();

  if (_blog)
    cout << "TLauncherFunction::run Function[" << GetName() << "]" << endl;

  TList * listOfRunningInputAttributes = NULL;
  bool listOfRunningInputAttributes_is_new=false;
  Int_t nRunningInput;

  try
    {
      if (_listOfInputAttributes == NULL)
	throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						   Form(
							"TLauncherFunction::run ERROR : the list of attributes for TDataServer \"%s\"  is NULL",
							_tds->GetName()));
      if (_blog)
	_listOfInputAttributes->Print();

      if (_tds->getTuple() == NULL)
	throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
						     Form(
							  "TLauncherFunction::run ERROR : the TTree of TDataServer \"%s\"  is NULL",
							  _tds->GetName()));

      // Get the list of attributes for current running
      if (sinput.IsNull())
        {
	  listOfRunningInputAttributes = _listOfInputAttributes;
	  nRunningInput = _listOfInputAttributes->GetEntries();
        }
      else
        {
	  try
            {
	      listOfRunningInputAttributes = _tds->getListOfAttributes(sinput.Data());
	      listOfRunningInputAttributes_is_new=true;
            } catch (URANIE::Exceptions::UErrorExceptions& ue)
            {
	      throw ue;
            }

	  nRunningInput = listOfRunningInputAttributes->GetEntries();
        }

    } catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
      ue.printMessage();
      throw ue;
    }

  // Sets output attributes for current running
  TString sRunningOutput;
  Int_t nRunningOutput;
  if (soutput.IsNull())
    {
      sRunningOutput = _soutput;
      nRunningOutput = _nOutput;
    }
  else
    {
      sRunningOutput = TString(soutput);
      nRunningOutput = sRunningOutput.CountChar(':') + 1;
    }

  Int_t nCalc = (Int_t)(_tds->getTuple()->GetEntries());
  if (_blog)
    cout << " ** nCalc[" << nCalc << "]" << endl;

  Double_t *valInp = new Double_t[nRunningInput];
  Double_t *valOut = new Double_t[nRunningOutput];

  TString satt(sRunningOutput.Data());
  satt.ReplaceAll(":", " ");
  satt += " ";
  istringstream inStream((char *) satt.Data());

  // At least, first output attribute
  string inName;
  inStream >> inName;
  try
    {
      _tds->addAttributeByNameTitle( TString(inName), Form("Output of function[%s]", inName.c_str()));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
      delete[] valInp;
      valInp = NULL;
      delete[] valOut;
      valOut = NULL;
      throw ue;
    }
  _tds->getAttribute(inName.c_str())->setOutput();
  if (_blog)
    cout << "run - " << inName << endl;

  string invar=inName, intype="D";
  // After, other attributes if exist
  while (inStream >> inName)
    {
      if (_blog)
	cout << "run - " << inName << endl;

      _tds->addAttributeByNameTitle(TString(inName), Form("Output of function[%s]", inName.c_str()));
      _tds->getAttribute(inName.c_str())->setOutput();
      
      invar+=":"+inName;
      intype+=":D";
    }

 _internal = new URANIE::DataServer::TDSNtupleD( (string(GetName())+"internaltree").c_str(),"internalone",invar.c_str(), intype.c_str());
  
  // Get The list of Leaf (from TTree) of the list of attributes ( _listOfInputAttributes)
  // Get The list of Leaf (from TTree) of the list of attributes ( listOfRunningInputAttributes)
  TList *listOfInputLeaves = new TList();
  //       TIter next(_listOfInputAttributes);
  TIter next(listOfRunningInputAttributes);

  while (URANIE::DataServer::TAttribute *att =
	 (URANIE::DataServer::TAttribute *) next())
    {
      if (_blog)
	cout << " ** Attribut - name[" << att->GetName() << "]" << endl;
      listOfInputLeaves->Add(_tds->getTuple()->GetLeaf(att->GetName()));
    }

  Int_t ivar;
  URANIE::DataServer::UTimer timer(nCalc, "Uranie Launcher");

  for (Int_t iCalc = 0; iCalc < nCalc; iCalc++)
    {
      // SPO (InckA) - add for DI153; solve problems if using multiple TLauncherFunction
      // gPlexnum = this;
      // get the current pattern
      _tds->getTuple()->GetEntry(iCalc);
      // Fill the input vector
      ivar = 0;
      TIter nextLeaf(listOfInputLeaves);
      while (TLeaf *leaf = (TLeaf *) nextLeaf())
        {
	  valInp[ivar++] = leaf->GetValue(0);
        }
        // Evaluate  y = f(x)
        if (_fMethod) (*_fMethod)(valInp, valOut);
        else if (_fMethodCall) {
            uintptr_t args[2];
            Double_t result = -1.0;

            args[0] = reinterpret_cast<uintptr_t>(valInp);
            args[1] = reinterpret_cast<uintptr_t>(valOut);
            _fMethodCall->SetParamPtrs(args, 2);
            _fMethodCall->Execute(&result);
        }
        else {
            std::cout << "NoEvaluator\n";
            break;
        }
      //Fill the output
      _internal->Fill( valOut );
     
#ifndef WIN32
      if (_bdrawProgressBar)
        {
	  timer.DrawProgressBar(iCalc);
        }
#endif

    }
  if (_bdrawProgressBar)
    cout << endl << " ** time elapsed " << timer.GetElapsedTime() << endl;

  _tds->getTuple()->Merger(_internal,invar.c_str(), invar.c_str());


  if(listOfRunningInputAttributes_is_new && listOfRunningInputAttributes!=NULL)
    delete listOfRunningInputAttributes;
  listOfRunningInputAttributes=NULL;
  delete listOfInputLeaves;
  delete[] valInp;
  valInp = NULL;
  delete[] valOut;
  valOut = NULL;
  delete _internal;
  _internal=NULL;
}

void TLauncherFunction::delOutputAttributes()
{
    if (_blog)
        cout << "TLauncherFunction::delOutputAttributes[" << _soutput << "]"
                << endl;
    TString satt(_soutput.Data());
    satt.ReplaceAll(":", " ");
    satt += " ";
    istringstream inStream((char *) satt.Data());
    string inName;

    // Boucle sur la liste de variable d'entr�e pr�sent dans str
    URANIE::DataServer::TAttribute *att = 0;

    while (inStream >> inName)
    {
        att = _tds->getAttribute(inName.c_str());
        if (att != NULL)
        {
            _tds->delAttribute(inName.c_str());
        }
    }
}

void TLauncherFunction::setNOutput(Int_t n)
{
  if (_blog) {
      cout << " ** TLauncherFunction::setNOutput(" << n << ")" << endl;
      cout << " ** Before _soutput[" << _soutput << "]" << endl;
    }
  try   {
    if (n < 1)
      throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                            Form(
                                              "TLauncherFunction::setNOutput ERROR : the number of output attributes %d must be greater or equal to 1.",
                                              n));

    // Nothing to do if _nOutput == n ( unused method)
    if (_nOutput == n) {
        return;
      } else {
        if (_nOutput > 1)   {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form(
                                                         "TLauncherFunction::setNOutput ERROR : the number of output attributes %d do not match the list of attributes \"%s\".",
                                                         n, _soutput.Data()));
          }
        TString srac = _soutput;
        for (Int_t i = 1; i < n; i++)
          _soutput += Form("_%d:%s", i, srac.Data());
        _soutput += Form("_%d", n);
        if (_blog)
          cout << "** After _soutput[" << _soutput << "]" << endl;

        _nOutput = _soutput.CountChar(':') + 1;
      }
  } catch (URANIE::Exceptions::UErrorExceptions& ue)
  {
    ue.printMessage();
    throw ue;
  }
}

//////////////////////
//  Printing Log
//////////////////////
void TLauncherFunction::printLog(Option_t *option)
{
    cout << "TLauncherFunction::printLog[" << option << "]" << endl;
    cout << "Name[" << GetName() << "] Title[" << GetTitle() << "]" << endl;
    if (_tds)
        _tds->printLog();
}

} // Fin du namespace Launcher
} // Fin du namespace URANIE

