//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TLauncherFunction.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   thu jun 14 11:49:19 CEST 2007
 
 \brief Interface de la classe URANIE::Launcher::TLauncherFunction
 
 */
/**
 
 * @class URANIE::Launcher::TLauncherFunction
 *
 * @brief Execute an analytical function on a dataserver.
 *
 * This class extends the old class TPleXNum which is devoted to exectute an analytical function on a numerical experiment plant.
 *
 */
#ifndef ULAUNCHERFUNCTION_H
#define ULAUNCHERFUNCTION_H

#include <iostream>
using std::cout;
using std::endl;

//ROOT
#include "Rtypes.h"
#include "TNamed.h"
#include "TMethodCall.h"

// Uranie
#include "Launcher.h"
#include "TDataServer.h"

namespace URANIE
{
namespace Launcher
{
class TLauncherFunction: public TNamed
{
private:
    TMethodCall *_fMethodCall; ///< Pointer to MethodCall in case of interpreted function
    void (*_fMethod)(Double_t *gin, Double_t *f);
    TString _soutput;                         ///< Name of the output attributes
    Int_t _nOutput;                      ///< The dimension of the output vector
    URANIE::DataServer::TDataServer *_tds;        ///< Pointer to the DataServer
    TList *_listOfInputAttributes;               ///< List of the input branches
    Bool_t _blog;                                   ///< Boolean to edit the log
    Bool_t _bdrawProgressBar; ///< Boolean to know if the progress bar has to be drawn
    URANIE::DataServer::TDSNtupleD *_internal;
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Constructor by a TDataServer, the name of the function and a list of inputs (optional). User MUST specify that functionanme IS a TString by forecasting it so as to use the fiest constructor. 
    TLauncherFunction(URANIE::DataServer::TDataServer *tds, const char* functionanme, 
            TString sinput = "", TString soutput = "");
#ifdef ROOT5
    //! Constructor by a TDataServer, a function and  list of inputs and outputs (optional)
    TLauncherFunction(URANIE::DataServer::TDataServer *tds, void *fcn, 
            TString sinput = "", TString soutput = "");
#else
    //! Constructor by a TDataServer, a function and  list of inputs and outputs (optional)
    TLauncherFunction(URANIE::DataServer::TDataServer *tds, void (*fcn)(Double_t*,Double_t*),
            TString sinput, TString soutput);
#endif
    //! Default destructor
    virtual ~TLauncherFunction();
    //@}

    //---------------------------------------------
    /** @name Setting and Getting attributs
     **/
    //@{
    //! Sets the number of output
    void setNOutput(Int_t n);

    //! Set the "draw progress bar" flag
    /*!
     The progress bar will be drawn if and only if the "draw progress bar"
     flag is true.
     */
    void setDrawProgressBar(Bool_t bbool = kTRUE)
    {
        _bdrawProgressBar = bbool;
    }

    //! Get the "draw progress bar" flag
    /*!
     \return (Bool_t) the "draw progress bar" flag (boolean)
     */
    Bool_t getDrawProgressBar()
    {
        return _bdrawProgressBar;
    }

    //@}

    //---------------------------------------------
    /** @name Evaluate the function
     **/
    //@{
    //! Runs phase of the code
    virtual void run(TString sinput = "", TString soutput = "",
            Option_t *option = "");
    //@}

    void delOutputAttributes();

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    void setLog()
    {
        _blog = kTRUE;
    }
    void unsetLog()
    {
        _blog = kFALSE;
    }
    void changeLog()
    {
        _blog = _blog ? kFALSE : kTRUE;
    }
    Bool_t getLog()
    {
        return _blog;
    }
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Launcher::TLauncherFunction, ID_LAUNCHER)
    //Classe de
};

} // Fin du namespace Launcher
} // Fin du namespace URANIE

#endif
