//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TOutputFile.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mon oct 9th 14:52:50 CEST 2006
 
 \brief Implementation of the class URANIE::Launcher::TOutputFile
 
 \par Modifications
 \li F. Gaudier (2010-08-06 11:28:25)
 - change the deprecated setKey method by setFileKey method
 \li F. Gaudier 2010-04-14 10:21:52
 For the addAttribute method, test the cases
 1. if the keyList is not empty,
 1.1 test if it contains one object
 1.1.1 if the filename is the same than the current filename is good
 1.1.2 else throw an error exception
 1.2 if it contains more than one element, throw an error exception
 2. if the keyList is empty, applique the setFileKey method
 \li F. Gaudier 2010-04-12 14:00:55
 Increase the index \b indcrt in the searchOutput method
 \li F. Gaudier 2010/02/23
 throw an exceptions when the key of one attribute is not found in the output file
 set more readable log
 
 */
#include <stdlib.h>
#include <iostream>
using std::cout;
using std::endl;

// ROOT
#include "TRegexp.h"
#include "TMatrixT.h"

// Uranie
#include "TOutputFile.h"
#include "UExceptions.h"

ClassImp (URANIE::Launcher::TOutputFile)

namespace URANIE
{
namespace Launcher
{

TOutputFile::TOutputFile(TString str) :
        TCodeFile(str)
{
    if (_blog)
        cout << " *** Method TOutputFile::" << __func__ << " with str[" << str << "]" << endl;

    if (_blog)
        cout << " *** End of Method TOutputFile::" << __func__ << " with str[" << str << "]" << endl;
}

TOutputFile::~TOutputFile()
{
    if (_blog)
        cout << " *** Destroy TOutputFile _sFileName[" << _sFileName << "]" << endl;
}

/////////////////////
//// Add attribut to the Output File
/////////////////////
void TOutputFile::addAttribute(TString str)
{
  URANIE::DataServer::TAttribute *att = new URANIE::DataServer::TAttribute(str);
  att->delShare();
  addAttribute(att);
}

void TOutputFile::addAttribute(URANIE::DataServer::TAttribute* att)
{
    if (_blog)
    {
        cout << "********************************" << endl;
        cout << " ***TOutputFile::addAttribute attribute[" << att->GetName()
                << "] TOutputFile FileName[" << _sFileName << "]" << endl;
    }
    try
    {
        list<URANIE::DataServer::TAttributeFileKey *> KeyList =
                att->getKeyList();

        if (!KeyList.empty())
        {
            Int_t nOutputfile = KeyList.size();
            if (nOutputfile > 1)
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                        Form(
                                "TOutputFile::addAttribute\n The Output attribute [%s] have %d Output Files: only one is autorised",
                                att->GetName(), nOutputfile));

            if (_blog)
                cout << " ** KeyList :: Size[" << nOutputfile << "]" << endl;
            list<URANIE::DataServer::TAttributeFileKey *>::iterator it =
                    KeyList.begin();
            TString skey = (*it)->_sKey;
            TString sfilename = gSystem->BaseName((*it)->_sFile.Data());
            if (_blog)
                cout << " ** KeyList :: skey[" << skey << "] sfilename["
                        << sfilename << "]" << endl;
            if (_sFileName != sfilename)
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                        Form(
                                "TOutputFile::addAttribute\n The Output attribute [%s] have already an Output File [%s] which is diffrent than the current [%s]",
                                att->GetName(), sfilename.Data(),
                                _sFileName.Data()));

        }
        else
            att->setFileKey(_sFileName);

        TCodeFile::addAttribute(att);
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }
    if (_blog)
    {
        cout << " *** End Of TOutputFile::addAttribute attribute["
                << att->GetName() << "] TOutputFile FileName[" << _sFileName
                << "]" << endl;
        cout << "********************************";
    }
}

void TOutputFile::printLog(Option_t *option)
{
    cout << "TOutputFile::printLog[" << option << "]" << endl;
    cout << "_sFileName[" << _sFileName << "]" << endl;
    Int_t natt = (Int_t)(_attributes.size());
    for (Int_t indx = 0; indx < natt; indx++)
    {
        cout << "index[" << indx + 1 << "/" << natt << "]" << endl;
        _attributes[indx]->printLog();
    }
}

} // Fin du namespace Launcher
} // Fin du namespace URANIE

