//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TOutputFileKey.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mon oct 9th 14:52:50 CEST 2006
 
 \brief Interface of the class URANIE::Launcher::TOutputFileKey
 
 */
/**
 
 @class URANIE::Launcher::TOutputFileKey

 @brief Description of the class TOutputFileKey.

 This classe enables to raise information from an ASCII file with the "Clef = valeur ;" formalism

 \verbatim
 coutMoy = 3.741194e-01 ;
 coutSig = 1.385948e-02 ;
 coutInc = 1.000000e+02 ;
 coutMax = 5.293728e-01 ;
 \endverbatim

 Thus, to raise the 2 informations \em coutSig and \em coutMax from file \em _coutwithkey_.dat exposed above,
 the code of \uranie is as follows :

 \verbatim
 TOutputFileKey *fout = new TOutputFileKey("_coutwithkey_.dat");
 
 fout->addAttribute(new TAttribute("coutSig"));
 fout->addAttribute(new TAttribute("coutMax"));
 \endverbatim

 \see TOutputFileKeyLine


 */
#ifndef TOUTPUTFILEKEY_H
#define TOUTPUTFILEKEY_H

// ROOT
#include "Rtypes.h"
#include "TString.h"
#include "TList.h"
// URANIE
#include "Launcher.h"
#include "TOutputFile.h"

#include <vector>

namespace URANIE
{
namespace Launcher
{
class TOutputFileKey: public TOutputFile
{
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    TOutputFileKey(TString str);
    //! Default destructor
    virtual ~TOutputFileKey();
    //@}

    //---------------------------------------------
    /** @name Add attribut to the File
     **/
    //@{
    //!   Add the attribute in the file with the key if defined
    /*!
     \param att (URANIE::DataServer::TAttribute*) the pointer of attribute
     \param skey (TString) [""] the key associated for the attribute in the file. If empty, the name was use.
     */
    void addAttribute(URANIE::DataServer::TAttribute* att, TString skey = "");
    //! Add the attribute by this name in the file
    /*!
     \param sname (TString)  the name of the attribute to created.
     \param skey (TString) [""] the key associated for the attribute in the file. If empty, the name was use.
     */
    void addAttribute(TString sname, TString skey = "");
    //@}

    //---------------------------------------------
    /** @name Computes Output
     **/
    //@{
    //! Looking for output usindouble only and transmitting matrix (not used by launcher anymore)
    Bool_t searchOutput(Int_t &n, Double_t *val, TList *lstOut = NULL);
    //! Looking for output using a map to be able to cope with vectors/strings
    Bool_t searchOutput(vector<URANIE::DataServer::UEntry*> *out);
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Launcher::TOutputFileKey, ID_LAUNCHER)
    //Classe de
};

} // Fin du namespace Launcher
} // Fin du namespace URANIE
#endif
