//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TOutputFileRow.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mon sept 3rd 14:31:26 CEST 2007
 
 \brief Interface of the class URANIE::Launcher::TOutputFileRow


 
 */
/*!
 
 @class URANIE::Launcher::TOutputFileRow

 @brief Description of the class TOutputFileRow.

 \par The File format
 The output data stored in the \b Row \b format likes :
 \verbatim
 0.20E+05   0.1125274E-25   0.5626372E-26
 0.30E+05   0.6558204E-22   0.3279102E-22
 0.40E+05   0.5512052E-19   0.2756026E-19
 0.50E+05   0.1204756E-16   0.6023779E-17
 0.60E+05   0.9529597E-15   0.4764799E-15
 0.70E+05   0.3464354E-13   0.1732177E-13
 0.80E+05   0.6793988E-12   0.3396994E-12
 0.90E+05   0.8090790E-11   0.4045395E-11
 0.10E+06   0.6406688E-10   0.3203344E-10
 0.20E+06   0.3121793E-06   0.1560896E-06
 0.30E+06   0.1549709E-06   0.7748547E-07
 0.40E+06   0.3679952E-08   0.1839976E-08
 0.50E+06   0.2374279E-10   0.1187140E-10
 0.60E+06   0.7977330E-13   0.3988665E-13
 0.70E+06   0.1085446E-40   0.5427229E-41
 0.80E+06   0.8481527E-39   0.4240764E-39
 0.90E+06   0.3601835E-37   0.1800918E-37
 0.10E+07   0.9537987E-36   0.4768994E-36
 0.20E+07   0.1399458E-27   0.6997289E-28
 0.30E+07   0.4081473E-24   0.2040736E-24
 0.40E+07   0.1816580E-22   0.9082898E-23
 0.50E+07   0.1019476E-21   0.5097381E-22
 0.60E+07   0.2019173E-21   0.1009586E-21
 0.70E+07   0.2611224E-21   0.1305612E-21
 0.80E+07   0.3016021E-21   0.1508010E-21
 0.90E+07   0.3424757E-21   0.1712378E-21
 \endverbatim
 where the first row is the \b time, and the the two last rows are called \b y1 and \b y2.
 This file contains three attributes.
 When you have only one row, the file looks like:
 \verbatim
 0.20E+05   0.1125274E-25   0.5626372E-26
 \endverbatim

 
 \par Example code source
 Let the file \b "data.out" shown in the last section. Suppose that we want to store the \b "time" (first row) and \b "y2" (the last row).
 For the \b y2 attribute, we set the default value to \b -1.0

 \code
 TOutputFileRow * fout = new TOutputFileRow("data.out");

 fout->addAttribute("time",  1);
 fout->addAttribute(new TAttribute("y1"));

 TAttribute *attdose =  new TAttribute("y2");
 attdose->setDefaultValue(-1.0);
 fout->addAttribute(attdose,  3);
 \endcode

 \li For the \b "time" attribute, we use the \ref addAttribute method with the name only. then, the row number is set automatically with 1.
 \li For the \b "y1" attribute, we use the \ref addAttribute method with create a TAttribute object. then, the row number is set automatically with 2 because it exists already an attribute for the first row.
 \li For the \b "y2" attribute, we use the \ref addAttribute method with create a TAttribute object ans define the row number
 
 \see TOutputFileColumn, TOutputFileKey and TOutputFileDataServer

 */
#ifndef TOUTPUTFILEROW_H
#define TOUTPUTFILEROW_H

#include <iostream>

//ROOT
#include "Rtypes.h"

// Uranie
#include "Launcher.h"
#include "TOutputFile.h"

namespace URANIE
{
namespace Launcher
{
class TOutputFileRow: public TOutputFile
{
private:
    Int_t _nLines;                ///< The number of lines in the file.
    Int_t _nColumns;              ///< The number of columnsin the file
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    TOutputFileRow(TString str);
    //! Default destructor
    virtual ~TOutputFileRow();
    //@}

    //---------------------------------------------
    /** @name Add attribut to the File
     **/
    //@{
    //!   Add the attribute and the row number in the file
    /*!
     The row number begins to 1.
     */
    void addAttribute(URANIE::DataServer::TAttribute*, Int_t nrow = 0);
    //! Add the attribute by this name and the row number in the file
    /*!
     The row number begins to 1.
     */
    void addAttribute(TString name, Int_t nrow = 0);
    //@}

    /** @name Search the Output values
     **/
    //@{
    //! Looking for output usindouble only and transmitting matrix (not used by launcher anymore)
    Bool_t searchOutput(Int_t &indcrt, Double_t *val, TList *lstOut = NULL);
    //! Looking for output using a map to be able to cope with vectors/strings
    Bool_t searchOutput(vector<URANIE::DataServer::UEntry*> *out);
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Launcher::TOutputFileRow, ID_LAUNCHER)
    //Classe de
};

} // Fin du namespace Launcher
} // Fin du namespace URANIE
#endif
