//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TamponTexte.cxx
 \author Vincent Bergeaud <vincent.bergeaud@cea.fr>

 \brief Implementation of the class URANIE::Launcher::TamponTexte

 This class is intended for the replacement of the TamponTexte class
 from the CLUB library.It is used for parsing the ascii files, with special treatments
 for whitespaces, separators and comments.
 Items can be accessed by token number (ie champ) or by (line, token) pair (ie ligneChamp).  
 
 */

#include "TamponTexte.h"
#include <sstream>
#include <cstring>
#include <iostream>
#include <fstream>

/*!
 Creates a TamponTexte object by reading a file and storing it in the working buffer
 \param filename name of the ascii file read for the buffer
 \param dummy unused param in order to reproduce the club syntax
 */
TamponTexte::TamponTexte(const char* filename, int dummy) :
        _name(filename)
{
    _cFieldSeparator=" ";
    
    (void) dummy;
    std::ifstream is((filename && *filename) ? filename : _name.c_str());
    if (is.good())
    {
        std::string str((std::istreambuf_iterator<char>(is)),
                std::istreambuf_iterator<char>());
        _buffer = str;
        is.close();
    }
    else
      {
	try{
	  std::string str("TamponTexte - impossible to open file ");
	  str += filename;
	  throw TamponTexteError(str);
	}catch(TamponTexteError &tte)
	  {
	    tte.printMessage();
	    throw tte;
	  }
      }
    if (_buffer[_buffer.size() - 1] != '\n')
        _buffer.append("\n");
}

const char* TamponTexte::nom() const
{
    return _name.c_str();
}

/*!
 \param iline line number
 \param ifield field number
 \return field
 retrieves the field designated by a (field, line) double (with 1-indexing)

 For instance, on file :
 aa bb
 cc dd
 ee ff

 champLigne (1,3) will return "ee"
 */
const char* TamponTexte::champLigne(int ifield, int iline)
{
    std::string token, token2, token3;
  try{
    std::string stline;
    std::istringstream str(_buffer);
    for (int i = 1; i <= iline; i++)
      std::getline(str, stline);
    if (stline.empty())
      throw TamponTexteError("the line is empty in champLigne");
    std::string line(stline);
    for (unsigned int i = 0; i < _separators.size(); i++)
    {
//	if( _separators[i] == "\t" || _separators[i] == "=")
	    replaceAll(line, _separators[i], " ");
    }
    line.append("\n");

    std::istringstream strline(line);

    if( _cFieldSeparator == " " )
    {
	for (int i = 1; i <= ifield; i++)
	    strline >> token;
    }
    else
    {
	for (int i = 1; i <= ifield; i++){
            std::getline(strline, token2);
            std::stringstream ss(token2);
            while (getline(ss, token, _cFieldSeparator.c_str()[0]) && i<ifield)
                i++;
        }
    }

    if (strline.eof() && _cFieldSeparator == " ")
      throw TamponTexteError("wrong number of fields in champLigne");
  }catch(TamponTexteError &tte)
    {
      tte.printMessage();
      throw tte;
    }

  for (unsigned int i = 0; i < _separators.size(); i++)
    replaceAll(token, _separators[i], "");

  if( _cFieldSeparator != " " && token==std::string(_cFieldSeparator))
      token="";
      
  char *token_char = new char[token.length() + 1];
  strcpy(token_char, token.c_str());
  return token_char;
}

std::string TamponTexte::ligne(int iline)
{
    std::string line;
    try
    {
	std::string stline;
	std::istringstream str(_buffer);
	for (int i = 1; i <= iline; i++)
	    std::getline(str, stline);

	if (stline.empty())
	    throw TamponTexteError("["+stline+"]\n");

	line = stline;
    }catch(TamponTexteError &tte)
    {
	tte.printMessage();
	throw tte;
    }
  
    return line;
}

/*!
 Adds a separator character to the separator list used for parsing tokens
 */
void TamponTexte::initSeparateurs(const char* separator)
{
    int len = strlen(separator);
    for (int i = 0; i < len; i++)
    {
        std::string str(separator + i, 1);
        _separators.push_back(str);
    }
}

/*!
 Adds a separator character to the separator list used for parsing tokens
 */
void TamponTexte::initFieldSeparateur(const char* separator)
{
    int len = strlen(separator);
    _cFieldSeparator = separator;
}

/*!
 Defines comment symbols
 \param comment list of chars for symbols starting comments
 */
void TamponTexte::initCarCommentaires(const char* comment)
{
    int len = strlen(comment);
    for (int i = 0; i < len; i++)
    {
        std::string str(comment + i, 1);
        _comments.push_back(str);
    }
}

/*!
 Eliminates comments. Lines starting with comments will be entirely removed.
 For lines with a comment symbol in the middle, the line will be deleted after the comment symbol.
 */
void TamponTexte::elimineCommentaires()
{
    for (unsigned int i = 0; i < _comments.size(); i++)
    {
        size_t pos = _buffer.find(_comments[i]);
        while (pos < _buffer.size())
        {
            size_t pos2 = _buffer.find('\n', pos);
            if (pos == 0 || _buffer[pos - 1] == '\n')
                _buffer.erase(pos, pos2 - pos + 1); // erase the entire line
            else
                _buffer.erase(pos, pos2 - pos); //erase only the line after the comment symbol
            pos = pos + 1;
            pos = _buffer.find(_comments[i]);
        }
    }
}

/*!
 Deletes lines with white caracters only.
 */
void TamponTexte::elimineBlancs()
{
    size_t pos = 0;

    while (pos < _buffer.size())
    {
        size_t eol = _buffer.find('\n', pos);
        size_t pos2 = eol < _buffer.size() ? eol : _buffer.size();
        bool iswhite = true;
        for (size_t i = pos; i < pos2; i++)
        {
            if (_buffer[i] != ' ')
            {
                iswhite = false;
                break;
            }
        }

        if (iswhite)
        {
            _buffer.erase(pos, pos2 - pos + 1); // line is deleted and pos is left unchanged
        }
        else
        {
            pos = pos2 + 1; // moving pos to the next line
        }
    }
}

bool TamponTexte::estVide() const
{
    return _buffer.empty();
}

/*! \return number of lines in the buffer
 */
int TamponTexte::nombreLignes()
{
    int nblines = 0;
    std::string line;
    std::istringstream str(_buffer);
    while (std::getline(str, line))
        nblines++;
    return nblines;
}

/*! \return number of tokens in the buffer 
 */
int TamponTexte::nombreChamps()
{
    int nbchamps = 0;
    std::istringstream str(_buffer);
    std::string token, token2, token3;
    if( _cFieldSeparator == " " )
    {
	while (str >> token)
	    nbchamps++;
    }
    else
    {
	while(std::getline(str, token2)){
            std::stringstream ss1(token2);
            while(std::getline(ss1, token3,'\n')){
                std::stringstream ss(token3);
                while(getline(ss, token,_cFieldSeparator.c_str()[0]))
                    nbchamps++;
            }
        }
    }

    return nbchamps;
    
}

/*! \param ifield token number (1-indexing)
 \return token 
 Parses the buffer with separators (" ", "\n" and "\t" being default) as delimiters. 
 */
const char* TamponTexte::champ(int ifield)
{

    std::string stline;
    std::string work(_buffer);
    for (unsigned int i = 0; i < _separators.size(); i++)
        replaceAll(work, _separators[i], " ");

    std::istringstream str(work);
    std::string token, token2, token3;

    if( _cFieldSeparator == " " )
    {
	for (int i = 1; i <= ifield; i++)
	    str >> token;
    }
    else
    {
        int i=0;
        while(std::getline(str, token2))
        {
            std::stringstream ss(token2);
            while(getline(ss, token, _cFieldSeparator.c_str()[0]))// && i<ifield)
            {
                i++;
                if(i>=ifield)
                    break;
            }
            if(i>=ifield)
                break;
        }
    }
    
    try{
      if (token == "")          
        throw TamponTexteError("wrong number of fields in champ()");
    }catch(TamponTexteError &tte)
      {
	tte.printMessage();
	throw tte;
      }
    char *token_char = new char[token.length() + 1];
    strcpy(token_char, token.c_str());
    return token_char;
}

/*!
 \param \iline line number (0-indexing)
 \return number of tokens on line \a iline

 */
int TamponTexte::nombreChampsLigne(int iline)
{
    std::string stline;
    std::istringstream str(_buffer);
    for (int i = 1; i <= iline; i++)
        std::getline(str, stline);    
    std::string line(stline);
    for (unsigned int i = 0; i < _separators.size(); i++)
        replaceAll(line, _separators[i], " ");
    std::string token, token2;
    int nbfields = 0;
    std::istringstream strline(line);

    if( _cFieldSeparator == " ")	
	while (strline >> token)
	    nbfields++;
    else
	while(std::getline(strline, token2)){
            std::stringstream ss(token2);
            while(getline(ss, token, _cFieldSeparator.c_str()[0]))
                nbfields++;
        }
    	
    return nbfields;
}

/*!
 \param itoken token number (0-indexing)
 \param replace string for replacement
 Replaces token number \a itoken with string \a replace
 */
void TamponTexte::remplaceChamp(int itoken, const char* replace)
{
    std::string work = _buffer;
    for (unsigned int i = 0; i < _separators.size(); i++)
        replaceAll(work, _separators[i], " ");
       
    std::istringstream str(work);
    std::string token, token2, token3;
    size_t pos1 = 0;
    std::stringstream ss;
    if( _cFieldSeparator == " ")
    {
        for (int i = 1; i < itoken; i++)
	    str >> token;
    }
    else
    {
        int i=1;
        while(std::getline(str, token2))
        {
            std::stringstream ss(token2);
            while(getline(ss, token, _cFieldSeparator.c_str()[0]))// && i<ifield)
            {
                i++;
                if(i>itoken)
                    break;
            }
            if(i>itoken)
                break;
        }
    }

    if( _cFieldSeparator == " ")
	str >> token;
    
    size_t pos2 = (size_t) str.tellg(); //first char after token
    std::string rep=(replace);
    if ( !str.eof() )
    {
	pos2 = pos2 -1; //last char of token    
	pos1 = pos2 - token.size() + 1; //first char of token
    }
    if( _cFieldSeparator != " " )
    {
	if( str.str().substr(pos1,1).compare(0,1," ")==0 )//.compare(" ") == 0)
	    pos1=pos1+1;
        rep+='\n';
    }
    _buffer.replace(pos1, token.size(), rep.c_str());
}

/*!
 \param name file name for writing the buffer (default : input file)
 */
void TamponTexte::ecrit(const char* name)
{
    std::string filename;
    if (name == 0)
        filename = _name;
    else
        filename = name;

    std::ofstream ostr(filename.c_str());
    if (ostr)
        ostr.write(_buffer.c_str(), _buffer.size());
    else
      {
	try{
	  throw TamponTexteError("TamponTExte::ecrit()-unable to open file");
	}catch(TamponTexteError &tte)
	  {
	    tte.printMessage();
	    throw tte;
	  }
      }
    ostr.close();
}

/*! Replace all occurrences of str1 by str2 in buffer
 \param buffer string to perform the replacement
 \param str1 string to be replaced by str2
 \param str2 string that replaces str1
 */
void TamponTexte::replaceAll(std::string& buffer, const std::string& str1,
        const std::string& str2)
{
    if (str1.compare(str2) == 0)
        return; //equal strings

    size_t pos = buffer.find(str1); //find first occurence

    while (pos != std::string::npos)
    {
        buffer.replace(pos, str1.size(), str2);
        pos = buffer.find(str1, pos + str2.size());
    }
}
