//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTLauncherFunction.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   $Date$
 
 \brief  Fichier de test de la classe URANIE::Launcher::TestTLauncherFunction
 
 */
//CPPUNIT
#define DBL_PRECISION_COMPARE 1.e-8

// Uranie
#include "TestTLauncherFunction.h"
#include "TAttribute.h"
#include "TUniformDistribution.h"
#include "uranie_commands.h"

#define ASCIIDATAFILE  "flowrateUniformDesign.dat"
#define ISHIGAMIDATAFILE  "ishigamiUniformDesign.dat"

// ROOT
#include "TString.h"
#include "TMath.h"
#include "TFile.h"
#include "TestUserFunctions.h"
#ifdef ROOT5
#include "Api.h"
#endif

#include <iostream>

TLauncherFunctionTest::TLauncherFunctionTest() :
    tds1(NULL), tds2(NULL), tds3(NULL), tlf1(NULL), _bLog(kFALSE)
{
    gROOT->LoadMacro("UserFunctions.C");

    _workingTestPath = gSystem->pwd();

#ifdef ROOT5
    // Get the void * from the "rosenbrock" analytical function
    G__ClassInfo globalscope;
    G__MethodInfo method;
    long dummy;
    
    // resolve function overloading :: flowrateModel
    method = globalscope.GetMethod("flowrateModel", "double*,double*", &dummy);
    _p2fFlowrate = method.PointerToFunc();

    // resolve function overloading :: HoXuSurrogateModel
    method = globalscope.GetMethod("HoXuSurrogateModel", "double*,double*", &dummy);
    _p2fHoXuSurrogateModel = method.PointerToFunc();

    // resolve function overloading :: flowrateModelSquared
    method = globalscope.GetMethod("flowrateModelSquared", "double*,double*", &dummy);
    _p2fFlowrateSquared = method.PointerToFunc();
    
    // resolve function overloading :: fSum
    method = globalscope.GetMethod("fSum", "double*,double*", &dummy);
    _p2fSum = method.PointerToFunc();
    
    // resolve function overloading :: fProd
    method = globalscope.GetMethod("fProd", "double*,double*", &dummy);
    _p2fProd = method.PointerToFunc();
    
    // resolve function overloading :: fctNotDefined (Function Not Defined)
    method = globalscope.GetMethod("fctNotDefined", "double*,double*", &dummy);
    _p2fUndefined = method.PointerToFunc();
    
    // resolve function overloading :: FctIshigami
    method = globalscope.GetMethod("FctIshigami", "double*,double*", &dummy);
    _p2FctIshigami = method.PointerToFunc();
    
    // resolve function overloading :: FctIshigamiPlus10
    method = globalscope.GetMethod("FctIshigamiPlus10", "double*,double*", &dummy);
    _p2FctIshigamiPlus10 = method.PointerToFunc();
    
    // resolve function overloading :: FctIshigamiPlus20
    method = globalscope.GetMethod("FctIshigamiPlus20", "double*,double*", &dummy);
    _p2FctIshigamiPlus20 = method.PointerToFunc();
    
    // resolve function overloading :: FctIshigamiPlus30
    method = globalscope.GetMethod("FctIshigamiPlus30", "double*,double*", &dummy);
    _p2FctIshigamiPlus30 = method.PointerToFunc();
#endif
}

TLauncherFunctionTest::~TLauncherFunctionTest()
{
}


void TLauncherFunctionTest::setUp()
{
    tds1 = new URANIE::DataServer::TDataServer("tds1", "my title 1");
    tds1->fileDataRead(ASCIIDATAFILE);
}

void TLauncherFunctionTest::tearDown()
{
    delete tlf1;
    tlf1 = NULL;
    delete tds1;
    tds1 = NULL;

    if(tds2){ delete tds2; tds2=NULL;}
    if(tds3){ delete tds3; tds3=NULL;}
    
    UranieTests::removeFile(_workingTestPath.Data(), "test_composition.dat");
    UranieTests::removeFile(_workingTestPath.Data(),"10_input_setNOuptut_leaf_test.root");
    UranieTests::removeFile(_workingTestPath.Data(),"flowrateUniformDesign.root");
    UranieTests::removeFile(_workingTestPath.Data(),"ishigamiUniformDesign.root");
    UranieTests::removeFile(_workingTestPath.Data(),"TdsIshigamiSample.root");
    UranieTests::removeFile(_workingTestPath.Data(),"_dataserver_.root");

}

////////////////////////////////
///                          ///
///         Tests            ///
///                          ///
////////////////////////////////
void TLauncherFunctionTest::testNames()
{
    TString sExpected("fSum");

    tlf1 = new URANIE::Launcher::TLauncherFunction(tds1, sExpected);
    if (_bLog)
        cout << " **TLauncherFunctionTest::testNames Name[" << tlf1->GetName()
                << "]" << endl;
    CPPUNIT_ASSERT(!strcmp(tlf1->GetName(), sExpected));
}

void TLauncherFunctionTest::testTitles()
{
    TString sFunctionName("flowrateModel");
#ifdef ROOT5
    tlf1 = new URANIE::Launcher::TLauncherFunction(tds1, _p2fFlowrate);
#else
    tlf1 = new URANIE::Launcher::TLauncherFunction(tds1, &flowrateModel, "", sFunctionName);
#endif
    //if (_bLog)
        cout << " **TLauncherFunctionTest::testNames GetTitle["
                << tlf1->GetTitle() << "]" << endl;
    TString sExpected(
            Form("Launcher the Analystical Function [%s]",
                    sFunctionName.Data()));
    CPPUNIT_ASSERT(!strcmp(tlf1->GetTitle(), sExpected));
}

////////////////////////////////
///                          ///
///     Tests  Function      ///
///                          ///
////////////////////////////////

void TLauncherFunctionTest::testLauncherFunctionWithoutInputNames()
{
    Int_t ninput = tds1->getNAttributes();
    Int_t noutput = 0;
    Int_t natt = ninput;
#ifdef ROOT5
    tlf1 = new URANIE::Launcher::TLauncherFunction(tds1, _p2fFlowrate);
#else
    tlf1 = new URANIE::Launcher::TLauncherFunction(tds1, flowrateModel, "", "flowrateModel");
#endif
    tlf1->unsetLog();
    tlf1->setDrawProgressBar(kFALSE);
    tlf1->run();
    natt++;
    noutput++;
    CPPUNIT_ASSERT_EQUAL(natt, tds1->getNAttributes());
    CPPUNIT_ASSERT_EQUAL(ninput, tds1->getNInputAttributes());
    CPPUNIT_ASSERT_EQUAL(noutput, tds1->getNOutputAttributes());
    TTree* tuple = (TNtupleD*) tds1->getTuple();
    TObjArray * Leaves = tuple->GetListOfLeaves();
    CPPUNIT_ASSERT(tds1->isAttribute("flowrateModel"));
    CPPUNIT_ASSERT(Leaves->FindObject("flowrateModel") != 0);
    if (tds1->isAttribute("flowrateModel"))
    {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getValue("flowrateModel", 0),
                tds1->getValue("ystar", 0), 1e-2);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getValue("flowrateModel", 31),
                tds1->getValue("ystar", 31), 1e-2);
    }
    delete tlf1;
    tlf1 = NULL;

#ifdef ROOT5
    tlf1 = new URANIE::Launcher::TLauncherFunction(tds1, _p2fHoXuSurrogateModel,
            "", "HoXu");
#else
    tlf1 = new URANIE::Launcher::TLauncherFunction(tds1, HoXuSurrogateModel, 
            "", "HoXu");
#endif
    tlf1->unsetLog();
    tlf1->setDrawProgressBar(kFALSE);
    tlf1->run();
    natt++;
    noutput++;
    CPPUNIT_ASSERT_EQUAL(natt, tds1->getNAttributes());
    CPPUNIT_ASSERT_EQUAL(ninput, tds1->getNInputAttributes());
    CPPUNIT_ASSERT_EQUAL(noutput, tds1->getNOutputAttributes());
    CPPUNIT_ASSERT(tds1->isAttribute("HoXu"));
    CPPUNIT_ASSERT(Leaves->FindObject("HoXu") != 0);
    if (tds1->isAttribute("HoXu"))
    {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getValue("HoXu", 0),
                26.2731696405384, DBL_PRECISION_COMPARE);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getValue("HoXu", 31),
                157.56595244191, DBL_PRECISION_COMPARE);
    }
    delete tlf1;
    tlf1 = NULL;

#ifdef ROOT5
    tlf1 = new URANIE::Launcher::TLauncherFunction(tds1, _p2fFlowrateSquared,
            "", "y2:y3");
#else
    tlf1 = new URANIE::Launcher::TLauncherFunction(tds1, flowrateModelSquared,
            "", "y2:y3");
#endif
    tlf1->unsetLog();
    tlf1->setDrawProgressBar(kFALSE);
    tlf1->run();
    natt += 2;
    noutput += 2;
    CPPUNIT_ASSERT_EQUAL(natt, tds1->getNAttributes());
    CPPUNIT_ASSERT_EQUAL(ninput, tds1->getNInputAttributes());
    CPPUNIT_ASSERT_EQUAL(noutput, tds1->getNOutputAttributes());
    CPPUNIT_ASSERT(tds1->isAttribute("y2"));
    CPPUNIT_ASSERT(tds1->isAttribute("y3"));
    CPPUNIT_ASSERT(Leaves->FindObject("y2") != 0);
    CPPUNIT_ASSERT(Leaves->FindObject("y3") != 0);
    if (tds1->isAttribute("y2") && tds1->isAttribute("y3"))
    {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getValue("y2", 0),
                tds1->getValue("ystar", 0), 1e-2);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getValue("y2", 31),
                tds1->getValue("ystar", 31), 1e-2);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getValue("y3", 0),
                tds1->getValue("rw", 0) * tds1->getValue("rw", 0),
                DBL_PRECISION_COMPARE);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getValue("y3", 31),
                tds1->getValue("rw", 31) * tds1->getValue("rw", 31),
                DBL_PRECISION_COMPARE);
    }
}

void TLauncherFunctionTest::testLauncherFunctionWithInputNames()
{
    Int_t ninput = tds1->getNAttributes();
    Int_t noutput = 0;
    Int_t natt = ninput;

    TTree* tuple = (TNtupleD*) tds1->getTuple();
    TObjArray * Leaves = tuple->GetListOfLeaves();
    tlf1 = new URANIE::Launcher::TLauncherFunction(tds1, flowrateModel,
            "rw:r:tu:tl:hu:hl:l:kw", "flowrateModel");
    tlf1->setDrawProgressBar(kFALSE);
    tlf1->run();
    natt++;
    noutput++;
    CPPUNIT_ASSERT_EQUAL(natt, tds1->getNAttributes());
    CPPUNIT_ASSERT_EQUAL(ninput, tds1->getNInputAttributes());
    CPPUNIT_ASSERT_EQUAL(noutput, tds1->getNOutputAttributes());
    CPPUNIT_ASSERT(tds1->isAttribute("flowrateModel"));
    CPPUNIT_ASSERT(Leaves->FindObject("flowrateModel") != 0);
    if (tds1->isAttribute("flowrateModel"))
    {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getValue("flowrateModel", 0),
                tds1->getValue("ystar", 0), 1e-2);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getValue("flowrateModel", 31),
                tds1->getValue("ystar", 31), 1e-2);
    }
    delete tlf1;
    tlf1 = NULL;

    tlf1 = new URANIE::Launcher::TLauncherFunction(tds1, flowrateModel,
            "rw:r:tu:tl:hu:hl:l:kw", "y1");
    tlf1->setDrawProgressBar(kFALSE);
    tlf1->run();
    natt++;
    noutput++;
    CPPUNIT_ASSERT_EQUAL(natt, tds1->getNAttributes());
    CPPUNIT_ASSERT_EQUAL(ninput, tds1->getNInputAttributes());
    CPPUNIT_ASSERT_EQUAL(noutput, tds1->getNOutputAttributes());
    CPPUNIT_ASSERT(tds1->isAttribute("y1"));
    CPPUNIT_ASSERT(Leaves->FindObject("y1") != 0);
    if (tds1->isAttribute("y1"))
    {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getValue("y1", 0),
                tds1->getValue("ystar", 0), 1e-2);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getValue("y1", 31),
                tds1->getValue("ystar", 31), 1e-2);
    }
    delete tlf1;
    tlf1 = NULL;

    tlf1 = new URANIE::Launcher::TLauncherFunction(tds1, flowrateModel,
            "rw:r:tu:tl:hl:hu:l:kw", "y2");
    tlf1->setDrawProgressBar(kFALSE);
    tlf1->run();
    natt++;
    noutput++;
    CPPUNIT_ASSERT_EQUAL(natt, tds1->getNAttributes());
    CPPUNIT_ASSERT_EQUAL(ninput, tds1->getNInputAttributes());
    CPPUNIT_ASSERT_EQUAL(noutput, tds1->getNOutputAttributes());
    CPPUNIT_ASSERT(tds1->isAttribute("y2"));
    CPPUNIT_ASSERT(Leaves->FindObject("y2") != 0);
    if (tds1->isAttribute("y2"))
    {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getValue("y2", 0),
                -(tds1->getValue("ystar", 0)), 1e-2);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getValue("y2", 31),
                -(tds1->getValue("ystar", 31)), 1e-2);
    }
    delete tlf1;
    tlf1 = NULL;

    tlf1 = new URANIE::Launcher::TLauncherFunction(tds1, flowrateModelSquared,
            "rw:r:tu:tl:hu:hl:l:kw", "y3:y4");
    tlf1->setDrawProgressBar(kFALSE);
    tlf1->run();
    natt += 2;
    noutput += 2;
    CPPUNIT_ASSERT_EQUAL(natt, tds1->getNAttributes());
    CPPUNIT_ASSERT_EQUAL(ninput, tds1->getNInputAttributes());
    CPPUNIT_ASSERT_EQUAL(noutput, tds1->getNOutputAttributes());
    CPPUNIT_ASSERT(tds1->isAttribute("y3"));
    CPPUNIT_ASSERT(tds1->isAttribute("y4"));
    CPPUNIT_ASSERT(Leaves->FindObject("y3") != 0);
    CPPUNIT_ASSERT(Leaves->FindObject("y4") != 0);
    if (tds1->isAttribute("y3") && tds1->isAttribute("y4"))
    {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getValue("y3", 0),
                tds1->getValue("ystar", 0), 1e-2);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getValue("y3", 31),
                tds1->getValue("ystar", 31), 1e-2);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getValue("y4", 0),
                tds1->getValue("rw", 0) * tds1->getValue("rw", 0),
                DBL_PRECISION_COMPARE);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getValue("y4", 31),
                tds1->getValue("rw", 31) * tds1->getValue("rw", 31),
                DBL_PRECISION_COMPARE);
    }
}

void TLauncherFunctionTest::testLauncherFunctionRun()
{
    Int_t ninput = tds1->getNAttributes();
    Int_t noutput = 0;
    Int_t natt = ninput;
    
#ifdef ROOT5
    tlf1 = new URANIE::Launcher::TLauncherFunction(tds1, _p2fFlowrate);
#else
    tlf1 = new URANIE::Launcher::TLauncherFunction(tds1, "flowrateModel");
#endif
    tlf1->setDrawProgressBar(kFALSE);
    tlf1->run("rw:r:tu:tl:hu:hl:l:kw", "y1");
    natt++;
    noutput++;
    CPPUNIT_ASSERT_EQUAL(natt, tds1->getNAttributes());
    CPPUNIT_ASSERT_EQUAL(ninput, tds1->getNInputAttributes());
    CPPUNIT_ASSERT_EQUAL(noutput, tds1->getNOutputAttributes());
    TTree* tuple = (TNtupleD*) tds1->getTuple();
    TObjArray * Leaves = tuple->GetListOfLeaves();
    CPPUNIT_ASSERT(tds1->isAttribute("y1"));
    CPPUNIT_ASSERT(Leaves->FindObject("y1") != 0);
    if (tds1->isAttribute("y1"))
    {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getValue("y1", 0),
                tds1->getValue("ystar", 0), 1e-2);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getValue("y1", 31),
                tds1->getValue("ystar", 31), 1e-2);
    }

    tlf1->run("rw:r:tu:tl:hl:hu:l:kw", "y2");
    natt++;
    noutput++;
    CPPUNIT_ASSERT_EQUAL(natt, tds1->getNAttributes());
    CPPUNIT_ASSERT_EQUAL(ninput, tds1->getNInputAttributes());
    CPPUNIT_ASSERT_EQUAL(noutput, tds1->getNOutputAttributes());
    CPPUNIT_ASSERT(tds1->isAttribute("y2"));
    CPPUNIT_ASSERT(Leaves->FindObject("y2") != 0);
    if (tds1->isAttribute("y2"))
    {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getValue("y2", 0),
                -(tds1->getValue("ystar", 0)), 1e-2);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getValue("y2", 31),
                -(tds1->getValue("ystar", 31)), 1e-2);
    }

    tlf1->run("", "");
    natt++;
    noutput++;
    CPPUNIT_ASSERT_EQUAL(natt, tds1->getNAttributes());
    CPPUNIT_ASSERT_EQUAL(ninput, tds1->getNInputAttributes());
    CPPUNIT_ASSERT_EQUAL(noutput, tds1->getNOutputAttributes());
    CPPUNIT_ASSERT(tds1->isAttribute("flowrateModel"));
    CPPUNIT_ASSERT(Leaves->FindObject("flowrateModel") != 0);
    if (tds1->isAttribute("flowrateModel"))
    {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getValue("flowrateModel", 0),
                tds1->getValue("y1", 0), DBL_PRECISION_COMPARE);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getValue("flowrateModel", 31),
                tds1->getValue("y1", 31), DBL_PRECISION_COMPARE);
    }
}

void TLauncherFunctionTest::testLauncherFunctionSetNOutput()
{
    Int_t ninput = tds1->getNAttributes();
    Int_t noutput = 0;
    Int_t natt = ninput;
#ifdef ROOT5
    tlf1 = new URANIE::Launcher::TLauncherFunction(tds1, _p2fFlowrateSquared);
#else
    tlf1 = new URANIE::Launcher::TLauncherFunction(tds1, flowrateModelSquared, 
            "*", "flowrateModelSquared");
#endif
    tlf1->setNOutput(2);
    tlf1->setDrawProgressBar(kFALSE);
    tlf1->run();
    natt += 2;
    noutput += 2;
    CPPUNIT_ASSERT_EQUAL(natt, tds1->getNAttributes());
    CPPUNIT_ASSERT_EQUAL(ninput, tds1->getNInputAttributes());
    CPPUNIT_ASSERT_EQUAL(noutput, tds1->getNOutputAttributes());
    CPPUNIT_ASSERT(tds1->isAttribute("flowrateModelSquared_1"));
    CPPUNIT_ASSERT(tds1->isAttribute("flowrateModelSquared_2"));
    TTree* tuple = (TNtupleD*) tds1->getTuple();
    TObjArray * Leaves = tuple->GetListOfLeaves();
    CPPUNIT_ASSERT(Leaves->FindObject("flowrateModelSquared_1") != 0);
    CPPUNIT_ASSERT(Leaves->FindObject("flowrateModelSquared_2") != 0);
    if (tds1->isAttribute("flowrateModelSquared_1")
            && tds1->isAttribute("flowrateModelSquared_2"))
    {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(
                tds1->getValue("flowrateModelSquared_1", 0),
                tds1->getValue("ystar", 0), 1e-2);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(
                tds1->getValue("flowrateModelSquared_1", 31),
                tds1->getValue("ystar", 31), 1e-2);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(
                tds1->getValue("flowrateModelSquared_2", 0),
                tds1->getValue("rw", 0) * tds1->getValue("rw", 0),
                DBL_PRECISION_COMPARE);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(
                tds1->getValue("flowrateModelSquared_2", 31),
                tds1->getValue("rw", 31) * tds1->getValue("rw", 31),
                DBL_PRECISION_COMPARE);
    }

}

void TLauncherFunctionTest::testLauncherFunctionSetNOutputAndTupleGestion()
{
    delete tds1;
    tds1 = new URANIE::DataServer::TDataServer("tutu","toto");
    tds1->fileDataRead("10_input_setNOuptut_leaf_test.dat");
    
#ifdef ROOT5
    tlf1 = new URANIE::Launcher::TLauncherFunction(tds1, _p2FctIshigami, "x:y:z");
#else
    tlf1 = new URANIE::Launcher::TLauncherFunction(tds1, FctIshigami, "x:y:z", "Ishigami");
#endif
    tlf1->setNOutput(4);
    tlf1->setDrawProgressBar(kFALSE);
    tlf1->run();

    //First treatment
    tds1->saveTuple("RECREATE"); // save this to the tuple
    tds1->addAttribute("sum","x+y+z");

    //Read if from the ntupleDataRead method
    tds2 = new URANIE::DataServer::TDataServer("tata","titi");
    tds2->ntupleDataRead("10_input_setNOuptut_leaf_test.root","tutu__data__tree__");
    tds2->addAttribute("ratio","x/y"); // Add a new varriable

    //Read it from loadtree
    tds3 = new URANIE::DataServer::TDataServer("tete","tyty");
    TFile *inputfile = TFile::Open("_dataserver_.root");
    URANIE::DataServer::TDSNtupleD *inputtree = (URANIE::DataServer::TDSNtupleD*)inputfile->Get("tutu__data__tree__;2");
    tds3->loadTree(inputtree);      
    tds3->computeRank("sum");

    //Getting values
    tds2->Draw("ratio","","goff");
    double *rat=tds2->getTuple()->GetV1();
    tds3->Draw("sum:Rk_sum","","goff");
    double *sum=tds3->getTuple()->GetV1(), *rks=tds3->getTuple()->GetV2();

    double sum_db[10] = {0.521788006905,1.43045982269,0.725950447563,1.48558581353,2.35546807281,1.55544265218,1.35185429743,2.11452357178,1.36201718403,1.85137710229};
    double rat_db[10] = {1.62876730881,2.2566281687,1.11764767573,3.69178860417,1.13694709245,0.103142551369,1.96776616864,1.05110365089,0.401566418991,0.980880973692};
    double rks_db[10] = {1,5,2,6,10,7,3,9,4,8};

    for(unsigned int ival=0; ival<10; ival++)
    {
	CPPUNIT_ASSERT_DOUBLES_EQUAL(sum_db[ival], sum[ival], DBL_PRECISION_COMPARE);
	CPPUNIT_ASSERT_DOUBLES_EQUAL(rks_db[ival], rks[ival], DBL_PRECISION_COMPARE);
	CPPUNIT_ASSERT_DOUBLES_EQUAL(rat_db[ival], rat[ival], DBL_PRECISION_COMPARE);       	
    }
    
}

void TLauncherFunctionTest::throwExceptionLauncherFunctionConstructorTdsNull()
{
#ifdef ROOT5
    CPPUNIT_ASSERT_THROW( new URANIE::Launcher::TLauncherFunction(NULL, "Flowrate"), 
            URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW( new URANIE::Launcher::TLauncherFunction(NULL, _p2fFlowrate), URANIE::Exceptions::UErrorExceptions);
#else
    CPPUNIT_ASSERT_THROW( 
        new URANIE::Launcher::TLauncherFunction(NULL, flowrateModel,
                "rw:r:tu:tl:hu:hl:l:kw", "y1")
        , URANIE::Exceptions::UErrorExceptions
    );
#endif
}

void TLauncherFunctionTest::throwExceptionLauncherFunctionConstructorFctUndefined()
{
  CPPUNIT_ASSERT_THROW( new URANIE::Launcher::TLauncherFunction(tds1, "_totoFunction"), 
        URANIE::Exceptions::UErrorExceptions);
#ifdef ROOT5
  CPPUNIT_ASSERT_THROW( new URANIE::Launcher::TLauncherFunction(tds1, (void *)NULL), URANIE::Exceptions::UErrorExceptions);
  CPPUNIT_ASSERT_THROW( new URANIE::Launcher::TLauncherFunction(tds1, _p2fUndefined), URANIE::Exceptions::UErrorExceptions);
#else
  CPPUNIT_ASSERT_THROW( new URANIE::Launcher::TLauncherFunction(tds1, 
        (void (*)(Double_t*,Double_t*))NULL, "", "rien"), URANIE::Exceptions::UErrorExceptions);
#endif
}

void TLauncherFunctionTest::throwExceptionOutputExists()
{
    cout << endl;
#ifdef ROOT5
    tlf1 = new URANIE::Launcher::TLauncherFunction(tds1, _p2fFlowrate);
#else
    tlf1 = new URANIE::Launcher::TLauncherFunction(tds1, flowrateModel, "", "flowrate");
#endif
    tlf1->setDrawProgressBar(kFALSE);
    tlf1->run();
    CPPUNIT_ASSERT_THROW(tlf1->run("", ""),
            URANIE::Exceptions::UErrorExceptions);
}

void TLauncherFunctionTest::throwExceptionInputNotExists()
{
    cout << endl;
#ifdef ROOT5
    tlf1 = new URANIE::Launcher::TLauncherFunction(tds1, _p2fFlowrate);
#else
    tlf1 = new URANIE::Launcher::TLauncherFunction(tds1, flowrateModel, "", "flowrate");
#endif
    tlf1->setDrawProgressBar(kFALSE);
    CPPUNIT_ASSERT_THROW(tlf1->run("rw:r:ru", "y2"),
            URANIE::Exceptions::UErrorExceptions);
}

void TLauncherFunctionTest::testMultipleLauncherFunction()
{
    URANIE::Launcher::TLauncherFunction * tlfSum =
            new URANIE::Launcher::TLauncherFunction(tds1, "fSum", "rw:r",
                    "ySum");
    URANIE::Launcher::TLauncherFunction * tlfProd =
            new URANIE::Launcher::TLauncherFunction(tds1, "fProd", "rw:r",
                    "yProd");
    tlfSum->setDrawProgressBar(kFALSE);
    tlfProd->setDrawProgressBar(kFALSE);
    tlfSum->run();
    tlfProd->run();

    //int index[] =  { 0, 1, 20, 31 };
    for (int i = 0; i < 4; i++)
    {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getValue("ySum", i),
                tds1->getValue("rw", i) + tds1->getValue("r", i), 1e-12);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getValue("yProd", i),
                tds1->getValue("rw", i) * tds1->getValue("r", i), 1e-12);
    }

    delete tlfSum;
    tlfSum = NULL;
    delete tlfProd;
    tlfProd = NULL;
}


void TLauncherFunctionTest::testMultipleLauncherFunctionWithMultidimensionalOutput()
{
    // construction de tds avec 3 variables aléatoires
    URANIE::DataServer::TDataServer *tds = new URANIE::DataServer::TDataServer("tdsishigami","Ex. Ishigami");
    tds->fileDataRead("TdsIshigamiSample.dat");
    
    URANIE::Launcher::TLauncherFunction * tlfFctIshigami =
#ifdef ROOT5
            new URANIE::Launcher::TLauncherFunction(tds, _p2FctIshigami,
#else
            new URANIE::Launcher::TLauncherFunction(tds, &FctIshigami,
#endif
                    "", "y1:y2:y3:y4");
    URANIE::Launcher::TLauncherFunction * tlfFctIshigamiPlus10 =
#ifdef ROOT5
            new URANIE::Launcher::TLauncherFunction(tds, _p2FctIshigamiPlus10,
#else
            new URANIE::Launcher::TLauncherFunction(tds, &FctIshigamiPlus10,
#endif
                    "", "z1:z2:z3:z4");
    URANIE::Launcher::TLauncherFunction * tlfFctIshigamiPlus20 =
#ifdef ROOT5
            new URANIE::Launcher::TLauncherFunction(tds, _p2FctIshigamiPlus20, 
#else
            new URANIE::Launcher::TLauncherFunction(tds, &FctIshigamiPlus20,
#endif
                    "", "w1:w2:w3:w4");
    URANIE::Launcher::TLauncherFunction * tlfFctIshigamiPlus30 =
#ifdef ROOT5
            new URANIE::Launcher::TLauncherFunction(tds, _p2FctIshigamiPlus30, 
#else
            new URANIE::Launcher::TLauncherFunction(tds, &FctIshigamiPlus30, 
#endif
                    "", "v1:v2:v3:v4");
            
            
            
    tlfFctIshigami->setDrawProgressBar(kFALSE);
    tlfFctIshigamiPlus10->setDrawProgressBar(kFALSE);
    tlfFctIshigamiPlus20->setDrawProgressBar(kFALSE);
    tlfFctIshigamiPlus30->setDrawProgressBar(kFALSE);
    
    tlfFctIshigami->run();
    tlfFctIshigamiPlus10->run();
    tlfFctIshigamiPlus20->run();
    tlfFctIshigamiPlus30->run();

    TString varName;
    TString varComp;
    for (int i = 0; i < 8; i++)
    {
        for(Int_t jt=1; jt<=4 ;jt++)
        {
            varComp = Form("y%d",jt);
            
            varName = Form("z%d",jt);
            CPPUNIT_ASSERT_DOUBLES_EQUAL(tds->getValue(varName.Data(), i),
                    tds->getValue(varComp.Data(), i) + 10.0, 1e-12);
            
            varName = Form("w%d",jt);
            CPPUNIT_ASSERT_DOUBLES_EQUAL(tds->getValue(varName.Data(), i),
                    tds->getValue(varComp.Data(), i) + 20.0, 1e-12);
            
            varName = Form("v%d",jt);
            CPPUNIT_ASSERT_DOUBLES_EQUAL(tds->getValue(varName.Data(), i),
                    tds->getValue(varComp.Data(), i) + 30.0, 1e-12);
            
        }
    }

    delete tlfFctIshigami;
    tlfFctIshigami = NULL;
    delete tlfFctIshigamiPlus10;
    tlfFctIshigamiPlus10 = NULL;
    delete tlfFctIshigamiPlus20;
    tlfFctIshigamiPlus20 = NULL;
    delete tlfFctIshigamiPlus30;
    tlfFctIshigamiPlus30 = NULL;
    
    delete tds;
}


void TLauncherFunctionTest::testFunctionComposition()
{
    URANIE::DataServer::TDataServer *tdsIshigami; 
#ifdef ROOT5
    //Fct Ishigami ///////////////////////////////////
    void * p2fIshigami; ///< Fct Ishigami
   
    // Get the void * from the "Ishigami" analytical function
    G__ClassInfo globalscope;
    G__MethodInfo method;
    long dummy;
    
    // resolve function overloading
    method = globalscope.GetMethod("FctIshigami", "double*,double*", &dummy);
    p2fIshigami = method.PointerToFunc();

    //end Fct Ishigami //////////////////////////////////
#endif

    tdsIshigami = new URANIE::DataServer::TDataServer("TdsIshigami",
            "TDS for Ishigami Use Case");

    tdsIshigami->fileDataRead(ISHIGAMIDATAFILE);
//     tdsIshigami->fileDataRead("ishigamiUniformDesign.dat");

    string listofparam = "x1:x2:x3";
    string listofres = "y1:u1:u2:u3";

    URANIE::Launcher::TLauncherFunction * ishig;
#ifdef ROOT5
    ishig = new URANIE::Launcher::TLauncherFunction(tdsIshigami, p2fIshigami, listofparam, listofres);
#else
    ishig = new URANIE::Launcher::TLauncherFunction(tdsIshigami, FctIshigami,
            listofparam, listofres);
#endif
    ishig->setDrawProgressBar(kFALSE);
    ishig->run();
   
    listofparam = "y1:u2:u3";
    listofres = "y2:v1:v2:v3";

    URANIE::Launcher::TLauncherFunction * ishig2;
#ifdef ROOT5
    ishig2 = new URANIE::Launcher::TLauncherFunction(tdsIshigami, p2fIshigami, listofparam, listofres);
#else
    ishig2 = new URANIE::Launcher::TLauncherFunction(tdsIshigami, FctIshigami,
            listofparam, listofres);
#endif
    ishig2->setDrawProgressBar(kFALSE);
    ishig2->run();
    
    Double_t* par = new Double_t[3];
    Double_t* res = new Double_t[4];
    
    tdsIshigami->exportData("test_composition.dat");
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsIshigami->getValue("x1",0), 
                                 tdsIshigami->getValue("u1",0),
                                 DBL_PRECISION_COMPARE);
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsIshigami->getValue("x2",0), 
                                 tdsIshigami->getValue("u2",0),
                                 DBL_PRECISION_COMPARE);
     
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsIshigami->getValue("x3",0), 
                                 tdsIshigami->getValue("u3",0),
                                 DBL_PRECISION_COMPARE);
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsIshigami->getValue("x1",1), 
                                 tdsIshigami->getValue("u1",1),
                                 DBL_PRECISION_COMPARE);
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsIshigami->getValue("x2",1), 
                                 tdsIshigami->getValue("u2",1),
                                 DBL_PRECISION_COMPARE);
     
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsIshigami->getValue("x3",1), 
                                 tdsIshigami->getValue("u3",1),
                                 DBL_PRECISION_COMPARE);
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsIshigami->getValue("x1",0), 
                                 tdsIshigami->getValue("v1",0),
                                 DBL_PRECISION_COMPARE);
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsIshigami->getValue("x2",0), 
                                 tdsIshigami->getValue("v2",0),
                                 DBL_PRECISION_COMPARE);
     
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsIshigami->getValue("x3",0), 
                                 tdsIshigami->getValue("v3",0),
                                 DBL_PRECISION_COMPARE);
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsIshigami->getValue("y1",1), 
                                 tdsIshigami->getValue("v1",1),
                                 DBL_PRECISION_COMPARE);
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsIshigami->getValue("x2",1), 
                                 tdsIshigami->getValue("v2",1),
                                 DBL_PRECISION_COMPARE);
     
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsIshigami->getValue("x3",1), 
                                 tdsIshigami->getValue("v3",1),
                                 DBL_PRECISION_COMPARE);
    
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsIshigami->getValue("y1",2), 
                                 tdsIshigami->getValue("v1",2),
                                 DBL_PRECISION_COMPARE);
    
    par[0] = 1; par[1] = 1; par[2] = 1;
    FctIshigami(par,res);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsIshigami->getValue("y1",1), 
                                 res[0],
                                 DBL_PRECISION_COMPARE);
    
    par[0] = res[0]; par[1] = res[1]; par[2] = res[2];
    FctIshigami(par,res);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsIshigami->getValue("y2",1), 
                                 res[0],
                                 DBL_PRECISION_COMPARE);
    
    
    par[0] = -1; par[1] = -1; par[2] = -1;
    FctIshigami(par,res);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsIshigami->getValue("y1",2), 
                                 res[0],
                                 DBL_PRECISION_COMPARE);
    
    par[0] = res[0]; par[1] = res[1]; par[2] = res[2];
    FctIshigami(par,res);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsIshigami->getValue("y2",2), 
                                 res[0],
                                 DBL_PRECISION_COMPARE);
    
    
    delete ishig2; 
    delete ishig; 

    if (tdsIshigami)
        delete tdsIshigami;
    
    if(par)
        delete[] par;
    if(res)
        delete[] res;
}



// End Of File
