//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////

#include "TestTamponTexte.h"
#include "uranie_commands.h"
#include "UExceptions.h"
#include <iostream>

using namespace URANIE::Launcher;

TestTamponTexte::TestTamponTexte()
{
    _workingTestPath = gSystem->pwd();
}

TestTamponTexte::~TestTamponTexte()
{
}

void TestTamponTexte::setUp()
{
    std::string filename("testTamponTexte.txt");
    std::ofstream fstr(filename.c_str());
    fstr << "# youpi" << '\n';
    fstr << "OP 5" << '\n';
    fstr << "OP 6" << '\n';
    fstr << "\n";
    fstr << " \n";
    fstr << "# very youpi\n";
    fstr << " KP 4 # youpi\n";
    fstr << " \n";
    fstr << " PL 5.345435\n";
    fstr << " KL|4|TR";
}


void TestTamponTexte::tearDown()
{
    UranieTests::removeFile(_workingTestPath.Data(), "testTamponTexte*.txt");
    UranieTests::removeFile(_workingTestPath.Data(), "toto.txt");
}


void TestTamponTexte::testWrongFile()
{
#ifndef OLDCLUB
    CPPUNIT_ASSERT_THROW(TamponTexte tt("wrongFile.txt"),TamponTexteError);
#endif    
}

void TestTamponTexte::testNombreLignes()
{
TamponTexte tt("testTamponTexte.txt");
CPPUNIT_ASSERT_EQUAL(10, tt.nombreLignes());
}

void TestTamponTexte::testElimineBlancs()
{
TamponTexte tt("testTamponTexte.txt");
tt.elimineBlancs();
CPPUNIT_ASSERT_EQUAL(7, tt.nombreLignes());
}

void TestTamponTexte::testElimineCommentaires()
{
TamponTexte tt("testTamponTexte.txt");
tt.elimineCommentaires();
CPPUNIT_ASSERT_EQUAL(10, tt.nombreLignes());
TamponTexte tt2("testTamponTexte.txt");
tt2.initCarCommentaires("#");
tt2.elimineCommentaires();
CPPUNIT_ASSERT_EQUAL(8, tt2.nombreLignes());
}

void TestTamponTexte::testChampLigne()
{
TamponTexte tt("testTamponTexte.txt");
tt.initCarCommentaires("#");
tt.elimineCommentaires();
tt.elimineBlancs();
const char *val1 = tt.champLigne(2, 2);
CPPUNIT_ASSERT(strcmp("6", val1) == 0);
#ifndef OLDCLUB
delete[] val1;
#endif
const char *val2 = tt.champLigne(1, 3);
CPPUNIT_ASSERT(strcmp("KP", val2) == 0);
#ifndef OLDCLUB
delete[] val2;
#endif
#ifndef OLDCLUB
CPPUNIT_ASSERT_THROW(tt.champLigne(1, 11), TamponTexteError);
CPPUNIT_ASSERT_THROW(tt.champLigne(3, 1), TamponTexteError);
#endif
}

void TestTamponTexte::testChamp()
{
TamponTexte tt("testTamponTexte.txt");
tt.initCarCommentaires("#");
tt.elimineCommentaires();
tt.elimineBlancs();
const char* val1 = tt.champ(4);
CPPUNIT_ASSERT(strcmp("6", val1) == 0);
const char* val2 = tt.champ(1);
CPPUNIT_ASSERT(strcmp("OP", val2) == 0);
#ifndef OLDCLUB
tt.initSeparateurs("|");
const char* val3 = tt.champ(11);
CPPUNIT_ASSERT(strcmp("TR", val3) == 0);
delete[] val1;
delete[] val2;
delete[] val3;
#endif
}
void TestTamponTexte::testNombreChampsLigne()
{
TamponTexte tt("testTamponTexte.txt");
tt.initCarCommentaires("#");
tt.elimineCommentaires();
tt.elimineBlancs();
CPPUNIT_ASSERT_EQUAL(2, tt.nombreChampsLigne(1));
}

void TestTamponTexte::testRemplaceChamp()
{
TamponTexte tt("testTamponTexte.txt");
tt.initCarCommentaires("#");
tt.elimineCommentaires();
tt.elimineBlancs();
CPPUNIT_ASSERT_EQUAL(5, tt.nombreLignes());
tt.remplaceChamp(4, "tt");
tt.ecrit("toto.txt");
const char* res = tt.champLigne(2, 2);
CPPUNIT_ASSERT(strcmp(res, "tt") == 0);
#ifndef OLDCLUB
delete[] res;
#endif
}

void TestTamponTexte::testEcrit()
{
TamponTexte tt("testTamponTexte.txt");
tt.initCarCommentaires("#");
tt.elimineCommentaires();
tt.elimineBlancs();
tt.remplaceChamp(4, "tt");
tt.ecrit("testTamponTexte2.txt");
TamponTexte tt2("testTamponTexte2.txt");
const char* res = tt.champ(4);
CPPUNIT_ASSERT(strcmp(res, "tt") == 0);
#ifndef OLDCLUB
delete[] res;
#endif
}
