//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   flowrate.cxx
 \author Forgues Cedric <uranietm@cea.fr>
 \date   lun nov 23 11:08:44 2009
 
 \brief  "Pseudo" Code \b flowrate allowing to show functionalities of Uranie Launcher. Synthesis of flowrate and flowrateStop executables.

 \section origin_flow_rate_water Origine
 Example given in <em>Design and Modeling for Computers Experiments</em> de <em>Kai-Tai Fang</em>, <em>Runze Li</em> et <em>Agus Sudjianto</em>

 Analytic function \f$ \re^{8}\f$ in \f$\re\f$ which formula is:
 \f[
 y = f(x) = \frac{2 \pi T_u\left( H_u - H_l\right)}{\ln(\frac{r}{r_{\omega}})\left[ 1 + \frac{2 L T_u}{\ln(\frac{r}{r_{\omega}}) r_{\omega}^2 K_{\omega}} + \frac{T_u}{T_l}\right]}
 \f]

 The parameters of these function are :
 \li \f$ r_{\omega} \in [0.05, 0.15]\f$ ( \f$ m\f$) : radius of borehole
 \li \f$ r \in [100, 50~000]\f$  ( \f$ m\f$) : radius of influence
 \li \f$ T_u \in [63~070, 115~600]\f$ ( \f$ m^2/yr\f$) : transmissivity of upper aquifer 
 \li \f$ T_l \in [63.1, 116]\f$ ( \f$ m^2/yr\f$ ) : transmissivity of lower aquifer
 \li \f$ H_u \in [990, 1~110]\f$ ( \f$ m\f$) : potentiometric head of upper aquifer
 \li \f$ H_l \in [700, 820]\f$ ( \f$ m\f$) : potentiometric head of lower aquifer
 \li \f$ L   \in [1~120, 1~680]\f$ ( \f$ m\f$) : length of borehole
 \li \f$ K_{\omega} \in [9~855, 12~045]\f$ ( \f$ m/yr\f$) : hydraulic conductivity of borehole

 $Revision$
 $State$
 $Date$

 \par Revision
 \li SPO (InckA) 2012-11-08
 Addition of XML output functionalities to flowrate. The code now writes XML output if libXML2 was found.
 \li SPO (InckA) 2012-08-01
 Addition of XML input functionalities to flowrate. The code now accepts parameters -x to read input data
 from an XML file, and parameter -xf to read from XML and simulate error.
 \li CFO (CS) 2010/12/15
 Default values are set for constraint variables chu, chl and cr + silent mode instead of verbose mode.
 2010-06-15 16:59:39 (FGA)
 Add the loop parameter to simultate long external code. Normal use, it will be nLoop = 1;
 \verbatim time flowrate -k
 nLoop =  10000000 : 1.160u 0.000s 0:01.16 100.0%
 nLoop = 100000000 : 11.600u 0.010s 0:11.61 100.0%
 \endverbatim
 */
#define MAX_LENGTH 256

//!  Output file with values in row and a header
#define OUTPUT_FILE_WITHROW "_output_flowrate_withRow_.dat"
//! second output file. Contains cumulated sums for each file of all variables.
#define OUTPUT_FILE2 "_output2_.dat"
//#define  bdebug 0
//!  Output file with the  "Key = value" format.
#define OUTPUT_FILE_WITHKEY "_output_flowrate_withKey_.dat"
//!  Output file with XML format.
#define OUTPUT_FILE_WITHXML "_output_flowrate_withXML_.dat"

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <iostream>     // for I/O
#include <fstream>      // for file I/O
#include <sstream>   
#include <cstring>

// Club
#ifdef OLDCLUB
#include "club/TamponTexte.h"
#include "club/IterateurChamp.h"
#include "club/BaseErreurs.h"
#else
#include "TamponTexte.h"
#endif

// LibXML2
#ifdef WITH_LIBXML2
#include <libxml/parser.h>
#include <libxml/xpathInternals.h>
#endif

int bdebug = 0;
int nloop = 1;

//---------------------------------------------
/** @name Methods to construct flowrate executable.
 **/
//@{
/*!
 Returns the list of options available with flowrate.
 -d:    debug mode.
 -s:    silent mode.
 -k:    input file flowrate_input_with_keys.in with keys.
 -f:    input file with flowrate_input_with_flags.in with flags.
 -kf:   input file flowrate_input_with_keys.in with failure mode.  
 -ff:   input file flowrate_input_with_flags.in with failure mode.
 -h,-?: prints help message.  
 */
void usage(const char *scmd)
{
    printf("Usage: %s [-d] [-s] [-k|-f|-r|-x|-kf|-ff|-xf|-mkk|-mkf|-mff|-Folkk|-Folkf|-Folff] [-h|-?] [-mt] [file]\n",
            scmd);
    printf("     -d:   debug mode\n");
    printf("     -s:   silent mode\n");
    printf("     -k:   input file with keys  [flowrate_input_with_keys.in]\n");
    printf("     -f:   input file with flags [flowrate_input_with_flags.in]\n");
    printf("     -r:   input file with values in rows [flowrate_input_with_values_rows.in]\n");
#ifdef WITH_LIBXML2
    printf("     -x:   input file with XML [flowrate_input_with_xml.in]\n");
#endif
    printf("     -kf:  input file with keys with failure mode [flowrate_input_with_keys.in]\n");
    printf("     -rf:  input file with vlaues in row with failure mode [flowrate_input_with_rows.in]\n");
    printf("     -ff:  input file with flags with failure mode [flowrate_input_with_flags.in]\n");
    printf("     -mkf:  multi input files with shared variables using keys and flags [flowrate_input_with_keys.in] [flowrate_input_with_flags.in]\n");
    printf("     -mkk:  multi input files with shared variables using keys [flowrate_input_with_keys.in] and [flowrate_input_with_keys2.in]\n");
    printf("     -mff:  multi input files with shared variables using flags [flowrate_input_with_flags.in] and [flowrate_input_with_flags2.in]\n");
    printf("     -Folkf:  multi input files without shared variable using keys and flags [flowrate_input_with_keys.in] [flowrate_input_with_flags.in]\n");
    printf("     -Folkk:  multi input files without shared variable using keys [flowrate_input_with_keys.in] and [flowrate_input_with_keys2.in]\n");
    printf("     -Folff:  multi input files without shared variable using flags [flowrate_input_with_flags.in] and [flowrate_input_with_flags2.in]\n");
#ifdef WITH_LIBXML2
    printf("     -xf:  input file with XML with failure mode [flowrate_input_with_xml.in]\n");
#endif
    printf("   -h,-?:   this help message\n\n");

}

/*!
 Reads file with values in rows.
 */
int readInputFileWithValuesRows(const char *sfile, double *input,
        double * dcenter)
{
    if (bdebug)
        std::cout << "***** readInputFileWithValuesRows file[" << sfile << "] ... "
                << std::endl;

    int nfound = -1;

    int nparamToRead = 8;
    int nparamExcludeToread = 3;

    std::ifstream fileInput(sfile, std::ios::in);
    if (fileInput)
    {
        if (bdebug)
            std::cout << " ** file open ifstream " << std::endl;
    }
    else
    {
        if (bdebug)
            std::cout << "********************* " << std::endl;
        return nfound;
    }
    nfound++;
    for (int i = 0; i < nparamToRead; i++)
    {
        fileInput >> input[i];
        if (!fileInput.good())
            break;
        if (bdebug)
            std::cout << " ** i[" << i << "/" << nparamToRead << "] value["
                    << input[i] << "]" << std::endl;
        nfound++;
    }
    for (int i = 0; i < nparamExcludeToread; i++)
    {
        fileInput >> dcenter[i];
        if (!fileInput.good())
            break;
        if (bdebug)
            std::cout << " ** i[" << i << "/" << nparamExcludeToread << "] value["
                    << dcenter[i] << "]" << std::endl;
    }

    fileInput.close();
    return nfound;
}

/*!
 Reads file with flags.
 */
int readInputFileWithFlags(const char *sfile, double *input, double * dcenter)
{

    if (bdebug)
        std::cout << "***** readInputFileWithFlags file[" << sfile << "] ... "
                << std::endl;

    int nfound = 0;

    FILE *fin = fopen(sfile, "r");
    if (fin)
    {
        if (bdebug)
            std::cout << " ** file open" << std::endl;
    }
    else
    {
        if (bdebug)
            std::cout << "********************* " << std::endl;
        return -1;
    }

    TamponTexte fnomFichier(sfile, 0);
    //   fnomFichier.initSeparateurs("\t\n");
    //   fnomFichier.initCarCommentaires("#");
    fnomFichier.initSeparateurs(" \n");
    fnomFichier.initCarCommentaires("#");

    //   // elimination of white lines
    //   fnomFichier.elimineBlancs();
    //   // elimination des lignes de commentaires
    //fnomFichier.elimineCommentaires();
    // Test si le fichier ainsi obtenu est vide
    if (fnomFichier.estVide())
    {
        std::cerr << "\n le fichier: " << fnomFichier.nom() << " est vide.\n";
        return -1;
    }
    if (bdebug)
    {
        std::cout << " ** Lines : " << fnomFichier.nombreLignes() << std::endl;
    }
    for (int iline = 1; iline <= fnomFichier.nombreLignes(); iline++)
    {
        if (bdebug)
            std::cout << " ** Line " << iline << "/" << fnomFichier.nombreLignes()
                    << " : Field [ " << fnomFichier.nombreChampsLigne(iline)
                    << "]";
        if (fnomFichier.nombreChampsLigne(iline) > 1)
        {
            const char *strclef = fnomFichier.champLigne(1, iline);
            if (bdebug)
                std::cout << " - Key[" << strclef << "]";

            // Case of parameter Rw and R
            if (strcmp("frottement_paroi", strclef) == 0)
            {
                nfound++;
                input[0] = atof(fnomFichier.champLigne(3, iline));
                nfound++;
                input[1] = atof(fnomFichier.champLigne(4, iline));
            }
            // Case of parameter tu
            if (strcmp("information_Tu", strclef) == 0)
            {
                nfound++;
                input[2] = atof(fnomFichier.champLigne(4, iline));
            }
            // Case of parameter tl
            if (strcmp("information_Tl", strclef) == 0)
            {
                nfound++;
                input[3] = atof(fnomFichier.champLigne(4, iline));
            }
            // Case of parameter hu
            if (strcmp("dt_min", strclef) == 0)
            {
                nfound++;
                input[4] = atof(fnomFichier.champLigne(2, iline));
            }

            // Case of parameter chu
            if (strcmp("ch_abcsissa_hu", strclef) == 0)
            {
                dcenter[0] = atof(fnomFichier.champLigne(2, iline));
            }
            // Case of parameter chl
            if (strcmp("ch_ordinate_hl", strclef) == 0)
            {
                dcenter[1] = atof(fnomFichier.champLigne(2, iline));
            }
            // Case of parameter cr
            if (strcmp("c_radius", strclef) == 0)
            {
                dcenter[2] = atof(fnomFichier.champLigne(2, iline));
            }
            // Case of parameter nLoop
            if (strcmp("nLoop", strclef) == 0)
            {
                nloop = (int) (atoi(fnomFichier.champLigne(2, iline)));
            }
            // Case of parameter hl
            if (strcmp("dt_max", strclef) == 0)
            {
                nfound++;
                input[5] = atof(fnomFichier.champLigne(2, iline));
            }
            // Case of parameter L
            if (strcmp("information_L", strclef) == 0)
            {
                if (strcmp("precision", fnomFichier.champLigne(1, iline + 1))
                        == 0)
                {
                    nfound++;
                    input[6] = atof(fnomFichier.champLigne(2, iline + 1));
                }
            }
            // Case of parameter Kw
            if (strcmp("kW", strclef) == 0)
            {
                nfound++;
                input[7] = atof(fnomFichier.champLigne(2, iline));
            }

        }
        if (bdebug)
            std::cout << std::endl;
    }

    return nfound;
}

/*!
 Reads file with keys.
 */
int readInputFileWithKeys(const char *sfile, double *input, double *dcenter)
{
    if (bdebug)
        std::cout << "***** readInputFileWithKeys file[" << sfile << "] ... "
                << std::endl;

    int nfound = 0;

    FILE *fin = fopen(sfile, "r");
    if (fin)
    {
        if (bdebug)
            std::cout << " ** file open" << std::endl;
    }
    else
    {
        if (bdebug)
            std::cout << "********************* " << std::endl;
        return -1;
    }

    TamponTexte fnomFichier(sfile, 0);
    fnomFichier.initSeparateurs(" =\n");
    fnomFichier.initCarCommentaires("#");

    // elimination des lignes vides
    fnomFichier.elimineBlancs();
    // elimination des lignes de commentaires
    fnomFichier.elimineCommentaires();
    // Test si le fichier ainsi obtenu est vide
    if (fnomFichier.estVide())
    {
        std::cerr << "\n le fichier: " << fnomFichier.nom() << " est vide.\n";
        return -1;
    }
    if (bdebug)
    {
        std::cout << " ** Lines : " << fnomFichier.nombreLignes() << std::endl;
    }

    for (int iline = 1; iline <= fnomFichier.nombreLignes(); iline++)
    {

        if (bdebug)
        {
            std::cout << " ** Line " << iline << "/" << fnomFichier.nombreLignes()
                    << " : Field [ " << fnomFichier.nombreChampsLigne(iline)
                    << "]";
        }
        if (fnomFichier.nombreChampsLigne(iline) > 1)
        {
            const char *strclef = fnomFichier.champLigne(1, iline);
            if (bdebug)
                std::cout << " - Key[" << strclef << "]";

            // Case of parameter Rw
            if (strcmp("Rw", strclef) == 0)
            {
                nfound++;
                input[0] = atof(fnomFichier.champLigne(2, iline));
            }
            // Case of parameter R
            if (strcmp("R", strclef) == 0)
            {
                nfound++;
                input[1] = atof(fnomFichier.champLigne(2, iline));
            }
            // Case of parameter tu
            if (strcmp("Tu", strclef) == 0)
            {
                nfound++;
                input[2] = atof(fnomFichier.champLigne(2, iline));
            }
            // Case of parameter tl
            if (strcmp("Tl", strclef) == 0)
            {
                nfound++;
                input[3] = atof(fnomFichier.champLigne(2, iline));
            }
            // Case of parameter hu
            if (strcmp("Hu", strclef) == 0)
            {
                nfound++;
                input[4] = atof(fnomFichier.champLigne(2, iline));
            }
            // Case of parameter chu
            if (strcmp("chu", strclef) == 0)
            {
                dcenter[0] = atof(fnomFichier.champLigne(2, iline));
            }
            // Case of parameter chl
            if (strcmp("chl", strclef) == 0)
            {
                dcenter[1] = atof(fnomFichier.champLigne(2, iline));
            }
            // Case of parameter cr
            if (strcmp("cr", strclef) == 0)
            {
                dcenter[2] = atof(fnomFichier.champLigne(2, iline));
            }
            // Case of parameter nLoop
            if (strcmp("nLoop", strclef) == 0)
            {
                nloop = (int) (atof(fnomFichier.champLigne(2, iline)));
            }
            // Case of parameter hl
            if (strcmp("Hl", strclef) == 0)
            {
                nfound++;
                input[5] = atof(fnomFichier.champLigne(2, iline));
            }
            // Case of parameter L
            if (strcmp("L", strclef) == 0)
            {
                nfound++;
                input[6] = atof(fnomFichier.champLigne(2, iline));
            }
            // Case of parameter Kw
            if (strcmp("Kw", strclef) == 0)
            {
                nfound++;
                input[7] = atof(fnomFichier.champLigne(2, iline));
            }

        }
        if (bdebug)
            std::cout << std::endl;
    }

    // Close the input file
    fclose(fin);

    if (bdebug)
        std::cout << "**** End of readInputFileWithKeys file[" << sfile
                << "] - Found [" << nfound << "]" << std::endl;

    return nfound;

}

/*!
 Reads file with XML format.
 */
#ifdef WITH_LIBXML2
int readInputFileWithXML(const char *sfile, double *input, double * dcenter)
{
    if ( bdebug )
    std::cout << "***** readInputFileWithXML file[" << sfile << "] ... " << std::endl;

    xmlInitParser();

    // Parsing of the file
    xmlDocPtr doc = xmlParseFile(sfile);
    if (doc == NULL)
    {
        std::cerr << "Error: unable to parse file \"" << sfile << "\"" << std::endl;
        return(-1);
    }

    // Create xpath evaluation context 
    xmlXPathContextPtr xpathCtx = xmlXPathNewContext(doc);
    if(xpathCtx == NULL)
    {
        std::cerr << "Error: unable to create new XPath context" << std::endl;
        xmlFreeDoc(doc);
        return(-1);
    }

    // Get the values
    int nfound = 0;
    const char* xmlPaths[12] =
    {   "//wall_friction/@rw",
        "//wall_friction/@r",
        "//parameter[@name='Tu']/value/double",
        "//parameter[@name='Tl']/value/double",
        "//parameter[tonode='mesher' and toport='dt_hu']/value/double",
        "//parameter[tonode='mesher' and toport='dt_hl']/value/double",
        "//parameter[@name='L']/@precision",
        "//kW/@value",
        "//stop_criterium/@ch_abscissa_hu",
        "//stop_criterium/@ch_ordinate_hl",
        "//stop_criterium/@c_radius",
        "//stop_criterium/@nLoop"};

    xmlXPathObjectPtr xpathObj;
    xmlNodePtr cur;
    double value;
    for(int i=0; i<12; i++)
    {
        xpathObj = xmlXPathEvalExpression((const xmlChar *)xmlPaths[i], xpathCtx);
        if(xpathObj == NULL || xpathObj->nodesetval == NULL)
        {
            std::cerr << "Error: unable to evaluate xpath expression \"" << (const xmlChar *)xmlPaths[i] << "\"" << std::endl;
            exit(1);
        }
        cur = xpathObj->nodesetval->nodeTab[0];
        xmlXPathFreeObject(xpathObj);
        value = atof((char *)xmlNodeGetContent(cur));

        // mandatory values
        if(i<8)
        {
            input[i] = value;
            nfound ++;
        }
        // optional value: center
        else if(i<11)
        {
            dcenter[i-8] = value;
        }
        // nLoop
        else
        {
            nloop = (int)value;
        }
    }

    // Cleanup 
    xmlXPathFreeContext(xpathCtx);
    xmlFreeDoc(doc);
    xmlCleanupParser();

    if ( bdebug )
    std::cout << "**** End of readInputFileWithXML file[" << sfile << "] - Found [" << nfound << "]"<<std::endl;
    return nfound;
}
#endif

double waterborehole(double *x)
{
    double drw = x[0], dr = x[1];
    double dtu = x[2], dtl = x[3];
    double dhu = x[4], dhl = x[5];
    double dl = x[6], dkw = x[7];

    double DPI = 3.14159;

    double dnum = 2.0 * DPI * dtu * (dhu - dhl);
    double dlnronrw = log(dr / drw);
    double dden =
            dlnronrw
                    * (1.0 + (2.0 * dl * dtu) / (dlnronrw * drw * drw * dkw)
                            + dtu / dtl);
    return dnum / dden;
}
//@}

//! 
/*! 
 
 \param argc le nombre d'argument
 \param argv la liste des arguments

 \return indice de l'erreur de retour
 */
int main(int argc, char *argv[])
{
    FILE * fout;
    char sfile[MAX_LENGTH];
    char sfile2[MAX_LENGTH];
    char * option = NULL;
    // Read the arguments
    bool bFileDouble = false;
    bool bFilesFollow = false;
    bool bFileByKey = false;
    bool bFileByFlag = false;
    bool bFileByRecreate = false;
    bool bFileByXML = false;
    bool bSilent = true;
    for (int i = 0; i < argc; i++)
    {

        if (!strcmp(argv[i], "-d"))
        {
            bdebug = 1;
        }
        if (!strcmp(argv[i], "-s"))
        {
            bSilent = false;
        }

        if (!strcmp(argv[i], "-r") || !strcmp(argv[i], "-rf"))
        {
            option = argv[i];
            bFileByRecreate = true;
            if (i != argc - 1)
                sprintf(sfile, "%s", argv[i + 1]);
            else
                sprintf(sfile, "%s", "flowrate_input_with_values_rows.in");
        }

        if (!strcmp(argv[i], "-k") || !strcmp(argv[i], "-kf") || !strcmp(argv[i], "-ks"))
        {
            option = argv[i];
            bFileByKey = true;
            if (i != argc - 1)
                sprintf(sfile, "%s", argv[i + 1]);
            else
                sprintf(sfile, "%s", "flowrate_input_with_keys.in");
        }

        if (!strcmp(argv[i], "-f") || !strcmp(argv[i], "-ff") || !strcmp(argv[i], "-ff"))
        {
            option = argv[i];
            bFileByFlag = true;
            if (i != argc - 1)
                sprintf(sfile, "%s", argv[i + 1]);
            else
                sprintf(sfile, "%s", "flowrate_input_with_flags.in");
        }
        
        if (!strcmp(argv[i], "-mkk") )
        {
            option = argv[i];
            bFileByKey = true;
            bFileDouble=true;
            sprintf(sfile, "%s", "flowrate_input_with_keys.in");
            sprintf(sfile2, "%s", "flowrate_input_with_keys2.in");
        }

        if (!strcmp(argv[i], "-mff") )
        {
            option = argv[i];
            bFileByFlag = true;
            bFileDouble=true;
            sprintf(sfile, "%s", "flowrate_input_with_flags.in");
            sprintf(sfile2, "%s", "flowrate_input_with_flags2.in");
        }
        
        if (!strcmp(argv[i], "-mkf") )
        {
            option = argv[i];
            bFileByKey = true;
            bFileByFlag = true;
            bFileDouble=true;
            sprintf(sfile, "%s", "flowrate_input_with_keys.in");
            sprintf(sfile2, "%s", "flowrate_input_with_flags.in");
        }

        if (!strcmp(argv[i], "-Folkk") )
        {
            option = argv[i];
            bFileByKey = true;
            bFileDouble=true;
            bFilesFollow=true;
            sprintf(sfile, "%s", "flowrate_input_with_keys.in");
            sprintf(sfile2, "%s", "flowrate_input_with_keys2.in");
        }

        if (!strcmp(argv[i], "-Folff") )
        {
            option = argv[i];
            bFileByFlag = true;
            bFileDouble=true;
            bFilesFollow=true;
            sprintf(sfile, "%s", "flowrate_input_with_flags.in");
            sprintf(sfile2, "%s", "flowrate_input_with_flags2.in");
        }
        
        if (!strcmp(argv[i], "-Folkf") )
        {
            option = argv[i];
            bFileByKey = true;
            bFileByFlag = true;
            bFileDouble=true;
            bFilesFollow=true;
            sprintf(sfile, "%s", "flowrate_input_with_keys.in");
            sprintf(sfile2, "%s", "flowrate_input_with_flags.in");
        }
        
#ifdef WITH_LIBXML2
        if (!strcmp(argv[i], "-x") || !strcmp(argv[i], "-xf") || !strcmp(argv[i], "-xf"))
        {
            option=argv[i];
            bFileByXML = true;
            if ( i != argc-1 )
		sprintf(sfile, "%s", argv[i+1]);
            else
		sprintf(sfile, "%s", "flowrate_input_with_xml.in");
        }
#endif

        if (!strcmp(argv[i], "-h") || !strcmp(argv[i], "-?"))
        {
            usage(argv[0]);
            exit(0);
        }
    }

    if ((bFileByRecreate == false) && (bFileByKey == false)
	&& (bFileByFlag == false) && (bFileByXML == false))
    {
        std::cerr
	    << " ** ERROR : You must select the flags (-r) OR (-f, -ff) OR keys (-k, -kf) OR XML (-x, -xf) option"
	    << std::endl;
        usage(argv[0]);
        exit(-1);
    }
    if (bdebug)
        std::cout << "with debug " << std::endl;

    // Allocate parameters vectors
    int nparam = 8;
    int nfound = 0;
    double *dparam = new double[nparam];
    double *dexclude = new double[3];
    //Default parameters for conditions 
    dexclude[0] = 1050;
    dexclude[1] = 770;
    dexclude[2] = 1100;

    //Reads the input parameters with two files
    if(bFileDouble)
    {
        double *dparam1 = new double[nparam];
        double *dparam2 = new double[nparam];
        int nfound2 = 0;
        if (bFileByKey && bFileByFlag)
        {
            nfound  = readInputFileWithKeys(sfile, dparam1, dexclude);
            nfound2 = readInputFileWithFlags(sfile2, dparam2, dexclude);
        }
        else if(bFileByKey && !bFileByFlag)
        {
            nfound  = readInputFileWithKeys(sfile, dparam1, dexclude);
            nfound2 = readInputFileWithKeys(sfile2, dparam2, dexclude);
        }
        else if(!bFileByKey && bFileByFlag)
        {
            nfound  = readInputFileWithFlags(sfile, dparam1, dexclude);
            nfound2 = readInputFileWithFlags(sfile2, dparam2, dexclude);
        }

        if(bFilesFollow)
        {
            dparam[0] = dparam1[0];
            dparam[1] = dparam1[1];
            dparam[2] = dparam1[2];
            dparam[3] = dparam1[3];
            dparam[4] = dparam2[4];
            dparam[5] = dparam2[5];
            dparam[6] = dparam2[6];
            dparam[7] = dparam2[7];
        }
        else
        {
            dparam[0] = dparam1[0];
            dparam[1] = dparam2[1];
            dparam[2] = dparam2[2];
            dparam[3] = dparam2[3];
            dparam[4] = dparam1[4];
            dparam[5] = dparam1[5];
            dparam[6] = dparam2[6];
            dparam[7] = dparam1[7];
        }
        delete[] dparam1;
        delete[] dparam2;
    }
    else
    {
        //Reads the input parameters with one file
        if (bFileByRecreate)
            nfound = readInputFileWithValuesRows(sfile, dparam, dexclude);

        if (bFileByKey)
            nfound = readInputFileWithKeys(sfile, dparam, dexclude);

        if (bFileByFlag)
            nfound = readInputFileWithFlags(sfile, dparam, dexclude);

#ifdef WITH_LIBXML2
        if ( bFileByXML)
            nfound = readInputFileWithXML(sfile, dparam, dexclude);
#endif
    }
    
    if (nfound != nparam)
    {
        std::cerr << " ** ERROR : Pb in read file [" << sfile << "] found[" << nfound
	     << "] param[" << nparam << "]" << std::endl;
        exit(-1);
    }
    if (bSilent)
    {
        std::cout << "******************" << std::endl;
        std::cout << "*** Input File[" << sfile << "] NbParam[" << nparam << "]" << std::endl;
        std::cout << "******************" << std::endl;
        for (int i = 0; i < nparam; i++)
            std::cout << "** i[" << i << "] value[" << dparam[i] << "]" << std::endl;
        std::cout << "******************" << std::endl;
        std::cout << "** Conditions : Center hu(" << dexclude[0] << "), hl("
	     << dexclude[1] << ") radius(" << dexclude[2] << ")" << std::endl;
        std::cout << "******************" << std::endl << std::endl;
    }

    // Computes the outputs
    double y = waterborehole(dparam);
    // Used when we want simulate failure of a code (here flowrate) for files with keys ("-kf") or with flags "-ff"
    double ddistance = (dparam[4] - dexclude[0]) * (dparam[4] - dexclude[0]);
    ddistance += (dparam[5] - dexclude[1]) * (dparam[5] - dexclude[1]);

    if (nloop < 0)
        nloop = 1;
    if (nloop > 1)
        if (bSilent)
        {
            std::cout << " ** nLoop " << nloop << " ...." << std::endl;
        }
    double yloop = 0.0;
    double ddloop = 0.0;
    for (int iloop = 0; iloop < nloop; iloop++)
    {
        // Computes the outputs
        yloop = waterborehole(dparam);
        // Used when we want simulate failure of a code (here flowrate) for files with keys ("-kf") or with flags "-ff"
        ddloop = (dparam[4] - dexclude[0]) * (dparam[4] - dexclude[0]);
        ddloop += (dparam[5] - dexclude[1]) * (dparam[5] - dexclude[1]);
    }
    if (nloop > 1)
        std::cout << " ** nLoop " << nloop << " .... Done" << std::endl;

    if (!strcmp(option, "-kf") || !strcmp(option, "-ff")
	|| !strcmp(option, "-xf") || !strcmp(option, "-rf"))
    {
        if (ddistance < dexclude[2])
        {
            //cerr << "** " << dparam[4] << " (" << dexclude[0] << ") " << dparam[5] << " (" << dexclude[1] << ") -- ddistance[" << ddistance << "]" << std::endl;
            exit(-1);
        }
    }

    if (!strcmp(option, "-ks") || !strcmp(option, "-fs")
	|| !strcmp(option, "-xs"))
    {
        if (ddistance < dexclude[2])
        {
            //cerr << "** " << dparam[4] << " (" << dexclude[0] << ") " << dparam[5] << " (" << dexclude[1] << ") -- ddistance[" << ddistance << "]" << std::endl;
            int *p=NULL;
            *p=8;
        }
    }

    ////////////////////////////////
    // Writes the ouput in files
    ////////////////////////////////
    // File with Row values and a header
    fout = fopen(OUTPUT_FILE_WITHROW, "w");
    fprintf(fout, "#COLUMN_NAMES: yhat | d\n\n%e  %e", y, ddistance);
    fclose(fout);
    // File with "key = value" format; 
    fout = fopen(OUTPUT_FILE_WITHKEY, "w");
    fprintf(fout, "yhat\t=\t%e;\n", y);
    fprintf(fout, "d = %e;\n", ddistance);
    fclose(fout);
#ifdef WITH_LIBXML2
    // File with XML format;
    xmlDocPtr doc = xmlNewDoc((const xmlChar*)"1.0");
    xmlNodePtr root_node = xmlNewNode(0, (const xmlChar*)"steady_state");
    xmlDocSetRootElement(doc, root_node);
    xmlNewProp(root_node, (const xmlChar*)"name", (const xmlChar*)"flowrate");
    xmlNodePtr param_node = xmlNewChild(root_node, 0, (const xmlChar*)"parameter", 0);
    xmlNewChild(param_node, 0, (const xmlChar*)"tonode", (const xmlChar*)"mesher");
    xmlNewChild(param_node, 0, (const xmlChar*)"toport", (const xmlChar*)"dt_hl");
    xmlNodePtr value_node = xmlNewChild(param_node, 0, (const xmlChar*)"value", 0);
    char yhat[15]; sprintf(yhat, "%.6e", y);
    xmlNewChild(value_node, 0, (const xmlChar*)"double", (const xmlChar*)yhat);
    xmlNodePtr dist_node = xmlNewChild(root_node, 0, (const xmlChar*)"distance", 0);
    char dist[15]; sprintf(dist, "%.6e", ddistance);
    xmlNewProp(dist_node, (const xmlChar*)"value", (const xmlChar*)dist);
    xmlSaveFormatFile(OUTPUT_FILE_WITHXML, doc, 1);
    xmlFreeDoc(doc);
    xmlCleanupParser();
#endif
    ////////////////////////////////

    if (bSilent)
    {
        std::cout << "******************" << std::endl;
        std::cout << "***  Ouput Files [" << OUTPUT_FILE_WITHROW << "] and ["
	     << OUTPUT_FILE_WITHKEY << "]";

#ifdef WITH_LIBXML2
        std::cout << "\n                ["<< OUTPUT_FILE_WITHXML <<"]";
#endif
        std::cout << std::endl;
        std::cout << "******************" << std::endl;
        std::cout << "** y[0] value[" << y << "]" << std::endl;
        std::cout << "** y[1] value[" << ddistance << "]" << std::endl;
        std::cout << "******************" << std::endl;
    }

    delete[] dparam;
    dparam = NULL;
    
    delete[] dexclude;
    dexclude = NULL;
    
    return 0;

}
