//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   multitype.cxx
 \author Jean-Baptiste Blanchard <jean-baptiste.blanchard@©ea.fr>
 \date   lun nov 23 11:08:44 2009
 
 \brief  "Pseudo" Code \b created to play with multitype with both launcher and relauncher modules

 $Revision$
 $State$
 $Date$

 \par Revision
 \li JBB 2016-11-15
 Creation
 \endverbatim
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string>

#include <iostream>     // for I/O
#include <fstream>      // for file I/O
#include <sstream>   
#include <cstring>

#include "TRandom3.h"
#include "TString.h"
#include "TRegexp.h"
#include <time.h>

// LibXML2
#ifdef WITH_LIBXML2
#include <libxml/xpathInternals.h>
#endif


//---------------------------------------------
/** @name Methods to construct multitype executable.
 **/
//@{
/*!
 Returns the list of options available with multitype.
 -h,-?: prints help message.  
 */
void usage(const char *scmd)
{
    printf("\n");
    printf("Usage: %s [-mt|-mtDS|-mtCol|_mtFlat|-mtXML|-mtRow] [-Readmt|-ReadmtDS|-ReadmtCol|-ReadmtFlat|-ReadmtXML|-ReadmtRow] -empty -delim \n", scmd);
    printf("\n");
    printf("  -mt:   produce multitype output. This used to test String and Vector reading as output.\n");
    printf("\t It needs an input file [multitype_input.dat] (recreated on the flight) only containing one number (a seed) \n");
    printf("\t It dumps an output file whose format depends on the -mt flag: \n");
    printf("\t   -mt|-mtDS    : DataServer file type [_output_multitype_mt_DataServer_.dat]\n");
    printf("\t   -mtKey       : Key file type        [_output_multitype_mt_Key_.dat && _output_multitype_mt_Key_condensate_.dat]\n");
    printf("\t   -mtCol       : Column file type     [_output_multitype_mt_Column_.dat]\n");
    printf("\t   -mtFlat      : Flat file type       [_output_multitype_mt_Flat_.dat]\n");
    printf("\t   -mtXML       : XML file type        [_output_multitype_mt_XML_.dat]\n");
    printf("\t   -mtRow       : Row file type        [_output_multitype_mt_Row_.dat]\n\n");

    printf("  -Readmt:   Read the multitype output. This is used to test String and Vector reading as input.\n");
    printf("\t It uses previous output as input file (see list above). \n");
    printf("\t It dumps an output file whose format depends on the -mt flag: \n");
    printf("\t   -ReadmtDS    : DataServer file type [_output_multitype_mt_DataServer_.dat]     -> [_output_multitype_readmt_DataServer_.dat] \n");
    printf("\t   -ReadmtKey   : Key file type        [_output_multitype_mt_Key_condensate_.dat] -> [_output_multitype_readmt_Key_.dat] \n");
    printf("\t   -ReadmtCol   : Column file type     [_output_multitype_mt_Column_.dat]         -> [_output_multitype_readmt_Column_.dat] \n");
    printf("\t   -ReadmtFlat  : Flat file type       [_output_multitype_mt_Flat_.dat]           -> [_output_multitype_readmt_Flat_.dat] \n");
    printf("\t   -ReadmtXML   : XML file type        [_output_multitype_mt_.xml]                -> [_output_multitype_readmt_.xml] \n");
    printf("\t   -ReadmtRow   : Row file type        [_output_multitype_mt_Row_.dat]            -> [_output_multitype_readmt_Row_.dat] \n\n");

    printf("  -empty: Allow to generate empty vectors\n");
    printf("  -delim: When not embedded in the output format, use \"[\", \",\", \"]\" as delimiters for vectors\n");
    printf("  -h|-?: Dump this help\n\n");
}

void multitype(std::string type, bool allow_empty, bool vec_delim)
{
  
  double seed=123456.;
  std::string words[10]={"zero","one","two","three","four","five","six","seven","eight","nine"};
  std::string keys[5]={"w1", "v1","v2","f1","w2"};

  //Getting the seed
  std::fstream  afile;
  afile.open("multitype_input.dat", std::ios::in );
  afile >> seed;
  afile.close();

  if( allow_empty == true )
      vec_delim=true;
  
  TRandom3 *rand = new TRandom3();
  rand->SetSeed(int(seed));
  // // unsigned int time_ui = (unsigned int)time(NULL) ;
  // // cout<<time_ui<<endl;
  // // rand->SetSeed(time_ui);
  // rand->SetSeed(1629438581);
      
  //Fill variables
  std::string w1=words[rand->Integer(10)], w2=words[rand->Integer(10)];
  std::vector<double> v1, v2;
  int sizemax=15;
  int size1 = rand->Integer(sizemax);
  if( !allow_empty)
      while (size1==0){ size1 = rand->Integer(sizemax); }
  int size2 = rand->Integer(sizemax);
  if( !allow_empty)
      while (size2==0){ size2 = rand->Integer(sizemax); }
  for(int i=0; i<size1; i++) v1.push_back( rand->Gaus(0,1) );
  for(int i=0; i<size2; i++) v2.push_back( rand->Gaus(10,3) );
  double f1 = rand->Gaus(50,10.2);

  //Write file
  std::ofstream out; std::stringstream sstr; sstr.precision(13); sstr.str("");
  if(type=="DS")
    {
      //DataServer Output
      out.open("_output_multitype_mt_DataServer_.dat");
      sstr << w1 << " ";
      if( vec_delim ) sstr << "[";
      for(unsigned int i=0; i<v1.size(); i++) sstr << v1.at(i) << ((i!=v1.size()-1)?",":"");
      if( vec_delim ) sstr << "] [";
      else sstr << " ";
      for(unsigned int i=0; i<v2.size(); i++) sstr << v2.at(i) << ((i!=v2.size()-1)?",":"");
      if (vec_delim) sstr << "]";
      sstr << " ";
      sstr << f1 << " " << w2 << "\n";
      out << "#COLUMN_NAMES: w1|v1|v2|f1|w2\n#COLUMN_TYPES: S|V|V|D|S\n\n" << sstr.str(); sstr.str("");
      out.close();
    }
  else if(type=="Col")
    {
      //Column Output
      out.open("_output_multitype_mt_Column_.dat");
      sstr << w1 << "\n";
      if(vec_delim) sstr << "[";
      for(unsigned int i=0; i<v1.size(); i++) sstr << v1.at(i) << ((i!=v1.size()-1)?" ":"");
      if(vec_delim) sstr << "]";
      sstr << "\n";
      if(vec_delim) sstr << "[";
      for(unsigned int i=0; i<v2.size(); i++) sstr << v2.at(i) << ((i!=v2.size()-1)?" ":"");
      if(vec_delim) sstr << "]";
      sstr << "\n";
      sstr << f1 << "\n"; 
      sstr << w2 << "\n";
      out << sstr.str(); sstr.str("");
      out.close();
    }
  else if(type=="Flat")
    {
      //Column Output
      out.open("_output_multitype_mt_Flat_.dat");
      sstr << w1 << " ";
      if(vec_delim) sstr << "[";
      for(unsigned int i=0; i<v1.size(); i++) sstr << v1.at(i) << ((i!=v1.size()-1)?",":"");
      if(vec_delim)sstr << "] [";
      else sstr << " ";
      for(unsigned int i=0; i<v2.size(); i++) sstr << v2.at(i) << ((i!=v2.size()-1)?",":"");
      if(vec_delim)sstr << "]";      
      sstr << " ";
      sstr << f1 << " "; 
      sstr << w2 << "\n";
      out << sstr.str(); sstr.str("");
      out.close();
    }
  else if(type=="XML")
    {
      //Column Output
      out.open("_output_multitype_mt_.xml");
      sstr << "<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>\n";
      sstr << "<multitypeMT>\n";
      sstr << " <w1 value=\""<<w1<<"\"/>\n";
      sstr << " <v1 n=\"" << v1.size() << "\">\n  <values>\n   ";
      for(unsigned int i=0; i<v1.size(); i++) sstr << v1.at(i) << ((i!=v1.size()-1)?" ":"");
      sstr << "\n  </values>\n </v1>\n"; 
      sstr << " <v2 n=\"" << v2.size() << "\">\n  <values>\n   ";
      for(unsigned int i=0; i<v2.size(); i++) sstr << v2.at(i) << ((i!=v2.size()-1)?" ":"");
      sstr << "\n  </values>\n </v2>\n"; 
      sstr << " <f1 value=\""<<f1<<"\"/>\n";
      sstr << " <w2 value=\""<<w2<<"\"/>\n";
      sstr << "</multitypeMT>\n";
      out << sstr.str(); sstr.str("");
      out.close();
    }
  else if (type == "Key")
    {
      //Key Output
      out.open("_output_multitype_mt_Key_.dat");
      sstr << keys[0] << " = " << w1 << "\n"; 
      for(unsigned int i=0; i<v1.size(); i++) sstr << keys[1] << " = " << v1.at(i) << "\n"; 
      for(unsigned int i=0; i<v2.size(); i++) sstr << keys[2] << " = " << v2.at(i) << "\n"; 
      sstr << keys[3] << " = " << f1 << "\n"; 
      sstr << keys[4] << " = " << w2 << "\n";
      out << sstr.str(); sstr.str("");
      out.close();

      //Key Output 2
      out.open("_output_multitype_mt_Key_condensate_.dat");
      sstr << keys[0] << " = " << w1 << "\n"; 
      sstr << keys[1] << " = [ "; for(unsigned int i=0; i<v1.size(); i++) sstr << v1.at(i) << ((i!= (v1.size()-1))? "," : "");
      sstr << "]\n";
      sstr << keys[2] << " = [ "; for(unsigned int i=0; i<v2.size(); i++) sstr << v2.at(i) << ((i!= (v2.size()-1))? "," : "");
      sstr << "]\n";
      sstr << keys[3] << " = " << f1 << "\n"; 
      sstr << keys[4] << " = " << w2 << "\n";
      out << sstr.str(); sstr.str("");
      out.close();
    }
  else if (type == "Row")
    {
      //Row Output
      unsigned int max=v1.size();
      if( v1.size() < v2.size() ) max=v2.size();
      out.open("_output_multitype_mt_Row_.dat");
      std::string sep="; ";
      for(unsigned int i=0; i<max; i++)
      {
          if(i==0)
          {
              if( v1.size()!=0 && v2.size()!=0)
                  sstr << w1 << sep << v1[0] << sep << v2[0] << sep << f1 << sep << w2 << "\n";
              if(v1.size()==0 && v2.size()!=0)
                  sstr << w1 << sep << " " << sep << v2[0] << sep << f1 << sep << w2 << "\n";
              if(v1.size()!=0 && v2.size()==0)
                  sstr << w1 << sep << v1[0] << sep << " " << sep << f1 << sep << w2 << "\n";
              if(v1.size()==0 && v2.size()==0)
                  sstr << w1 << sep << " " << sep << " " << sep << f1 << sep << w2 << "\n";
          }
	  else
	    {
	      sstr << sep;
	      if( i < v1.size()) sstr << v1[i] << sep;
	      else sstr << sep;
	      if( i < v2.size()) sstr << v2[i] << sep;
	      else sstr << sep;	  
	      sstr <<  sep  << sep << "\n";
	    }      
	}
      out << sstr.str(); sstr.str("");
      out.close();
    }
  delete rand;
}

void readMultiType(std::string type, bool allow_empty)
{
  
    std::string words[10]={"zero","one","two","three","four","five","six","seven","eight","nine"};
    std::string inkeys[5]={"w1", "v1","v2","f1","w2"};
    std::string keys[4]={"thev1", "thev2"};
    int _index[5]={0,0,1,0,1};
    int i=0;

    //input variables
    std::string Word[2];
    std::vector<double> Vector[2];
    double Double[1];

    //Output variables
    double thev1=987654321, thev2=987654321;

    // //Read input file
    std::string inname;
    if( type=="DS")
	inname = "_output_multitype_mt_DataServer_.dat";
    else if ( type=="Col")
	inname = "_output_multitype_mt_Column_.dat";
    else if ( type=="Flat")
	inname = "_output_multitype_mt_Flat_.dat";
    else if (type=="Key")
	inname = "_output_multitype_mt_Key_condensate_.dat";
    else if (type=="Row")
	inname = "_output_multitype_mt_Row_.dat";
    else if (type=="XML")
	inname = "_output_multitype_mt_.xml";
    
    std::ifstream fin(inname.c_str(), std::ios::in); std::stringstream sstr; sstr.precision(13); sstr.str(""); TString line;
    if (!fin)
    {
	std::cout << "ouverture impossible " << inname << std::endl;
	return ;
    }
    if(type=="DS")
    {
	//DataServer Output
	i=0; 
	while (line.ReadLine(fin))
	    if (line[0] != '#' )
		break;	  

	std::istringstream pouet(line.Data()); std::string test;
	while( pouet >> test )
	{
	    if( i==0 || i==4 ) Word[_index[i] ] = test; 
	    else if( i==1 || i==2 )
	    {
		Vector[_index[i] ].clear();
		std::istringstream pouet2(test.c_str()); std::string value; 
		while(std::getline(pouet2, value, ','))
		    Vector[ _index[i] ].push_back( atof(value.c_str()) );
	    }
	    else if( i==3 ) Double[ _index[i] ] = atof( test.c_str() );
	    i++;
	}
    }
    else if(type=="Row")
    {
	//Row Output
	i=0;
	Vector[0].clear(); Vector[1].clear();
	TString delim(";");
	while (line.ReadLine(fin))
	{
	    std::istringstream pouet(line.Data()); std::string test;
	    i=0;
	    while(std::getline(pouet, test, delim.Data()[0]) )
	    {
		if (test != " " && test != "" && test != "\t")
		{
		    std::string val  = std::string(TString(test).ReplaceAll(" ","") );
		    if( i==1 || i==2 ) Vector[_index[i] ].push_back( atof(val.c_str()) );
		    else if( i==0 || i==4 ) Word[_index[i] ] = val; 
		    else if( i==3 ) Double[ _index[i] ] = atof( val.c_str() );
		}
		i++;
	    }
	}
    }
    else if(type=="XML")
    {
	//XML Output
	Vector[_index[0] ].clear();
	Vector[_index[1] ].clear();
	std::string test; TString readit;
	while (line.ReadLine(fin))
	{
	    std::istringstream pouet(line.Data());
	    while( pouet >> test )
	    {
		readit = TString(test);
		if( readit.Contains("<w1") ){ pouet >> test; readit = TString(test);  Word[ 0 ] = std::string(readit.ReplaceAll("value=","").ReplaceAll("/","").ReplaceAll(">","").ReplaceAll("\"","")); }
		if( readit.Contains("<w2") ){ pouet >> test; readit = TString(test);  Word[ 1 ] = std::string(readit.ReplaceAll("value=","").ReplaceAll("/","").ReplaceAll(">","").ReplaceAll("\"","")); }
		if( readit.Contains("<f1") ){ pouet >> test; readit = TString(test);  Double[ 0 ] = atof( readit.ReplaceAll("value=","").ReplaceAll("/","").ReplaceAll(">","").ReplaceAll("\"","") ); }
		if( readit.Contains("<v1"))
		{
		    line.ReadLine(fin); line.ReadLine(fin);
		    std::istringstream value(line.Data());
		    while( value >> test )
			Vector[ 0 ].push_back( atof( test.c_str() ) );
		}
		if( readit.Contains("<v2") )
		{
		    line.ReadLine(fin); line.ReadLine(fin);
		    std::istringstream value(line.Data());
		    while( value >> test )
			Vector[ 1 ].push_back( atof( test.c_str() ) );
		}
	    }
	}
    }
    else if(type=="Col")
    {
	//Column Output
	i=0; 
	while (line.ReadLine(fin))
	{
	    std::istringstream pouet(line.Data()); std::string test;
	    if( i==1 || i==2 ) Vector[_index[i] ].clear();
	    while( pouet >> test )
	    {
		if( i==0 || i==4 ) Word[_index[i] ] = test; 
		else if( i==1 || i==2 ) Vector[ _index[i] ].push_back( atof( test.c_str()) );
		else if( i==3 ) Double[ _index[i] ] = atof( test.c_str() );
	    }
	    i++;
	}
    }
    else if(type=="Flat")
    {
	//Column Output
	i=0;
        line.ReadLine(fin);
        std::istringstream pouet(line.Data()); std::string test;
        if( i==1 || i==2 ) Vector[_index[i] ].clear();
        while( pouet >> test )
        {
            if( i==0 || i==4 ) Word[_index[i] ] = test; 
            else if( i==1 || i==2 )
            {
                TString vec(test);
                vec.ReplaceAll("[","");
                vec.ReplaceAll("]","");
                vec.ReplaceAll(","," ");

                std::istringstream pouet2(vec.Data()); std::string test2;
                while( pouet2 >> test2 )
                    Vector[ _index[i] ].push_back( atof( test2.c_str()) );
            }
            else if( i==3 ) Double[ _index[i] ] = atof( test.c_str() );
	    i++;
	}
    }
    else if (type == "Key")
    {
	//Key Output

	i=0; TString beg="[", del=",", end="]";
	bool found=false;
	while (line.ReadLine(fin))
	{
	    beg="[", del=",", end="]";
	    TString name(inkeys[i]);

	    TRegexp reg[3] = {
		Form("[ \t]*%s[ \t]*=[ \t]*%s[ \t]*[-+0-9.e%s]+[ \t]*%s[ \t]*;?[ \t]*", name.Data(), beg.Data(), del.Data(), end.Data()),
		Form("[ \t]*%s[ \t]*=[ \t]*[a-zA-Z_]+[ \t]*", name.Data()),
		Form("[ \t]*%s[ \t]*=[ \t]*[-+0-9.e]+[ \t]*;?[ \t]*", name.Data())
	    };
  
	    TRegexp re_beg[3]={
		Form("=[ \t]*%s[ \t]*[-+0-9.]", beg.Data()),
		"=[ \t]*[a-zA-Z_]",
		"=[ \t]*[-0-9.e]"
	    };
  
	    TRegexp re_end[3]={
		Form("[-+0-9.e][ \t]*%s", end.Data()),
		"[A-Za-z_]+",
		"[-+0-9.e]+"
	    };

	    Ssiz_t _istart=0, _istop=0, _length=0, _end=0, tmp=0;
	    TString value="";//, newline="", repl="toto";
	    for(unsigned int ireg=0; ireg<3; ireg++)
	    {
		_istart=0; _istop=0; tmp=0;
		if(ireg==0) {beg="["; end="]";}
		else {beg=""; end="";}

		if( line.Contains(reg[ireg]) )
		{
		    _istart = line.Index(re_beg[ireg], &tmp, 0); _istart+=tmp-1;
		    _istop = line.Index(re_end[ireg], &tmp, _istart); _istop+=((ireg==0) ? 1 : tmp);

		    value=line;
		    value.Replace(_istop,line.Sizeof(),"").Replace(0,_istart,"");
		    //newline.Replace(_istart, _istop-_istart ,"toto");
		    found=true;
		    break;
		}
	    }

	    if( !found ) break;
	  
	    if(i==0) Word[0]=value;
	    if(i==1)
	    {
		std::istringstream pouet(value.Data()); std::string test;
		while(std::getline(pouet, test, ','))
		    Vector[0].push_back( atof( test.c_str() ) );
	    }
	    if(i==2) 
	    {
		std::istringstream pouet(value.Data()); std::string test;
		while(std::getline(pouet, test, ','))
		    Vector[1].push_back( atof( test.c_str() ) );
	    }
	    if(i==3) Double[0]= atof( value.Data() );
	    if(i==4) Word[1]=value;
	    i++;
	}
    }

    fin.close();  
    for(unsigned int iday=0; iday<10; iday++)
    {

	if( Word[0] == words[iday] )
	    thev1 = ( (iday<Vector[0].size()) ? Vector[0].at(iday) : -123456789 );
      
	if( Word[1] == words[iday] )
	    thev2 = ( (iday<Vector[1].size()) ? Vector[1].at(iday) : -123456789 );
    }

    // //Write Output File
    std::ofstream out; sstr.str("");
    if(type=="DS")
    {
	//DataServer Output
	out.open("_output_multitype_readmt_DataServer_.dat");
	sstr << thev1 << " " << thev2 << "\n";
	out << "#COLUMN_NAMES: thev1|thev2\n#COLUMN_TYPES: D|D\n\n" << sstr.str(); sstr.str("");
	out.close();
    }
    else if(type=="XML")
    {
      //Column Output
      out.open("_output_multitype_readmt_.xml");
      sstr << "<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>\n";
      sstr << "<multitypeREADMT>\n";
      sstr << " <thev1 value=\""<<thev1<<"\"/>\n";
      sstr << " <thev2 value=\""<<thev2<<"\"/>\n";
      sstr << "</multitypeREADMT>\n";
      out << sstr.str(); sstr.str("");
      out.close();
    }
    else if(type=="Row")
    {
	//Column Output
	out.open("_output_multitype_readmt_Row_.dat");
	sstr << thev1 << "; " << thev2 << "\n"; 
	out << sstr.str(); sstr.str("");
	out.close();
    }
    else if(type=="Col")
    {
	//Column Output
	out.open("_output_multitype_readmt_Column_.dat");
	sstr << thev1 << "\n";
	sstr << thev2 << "\n"; 
	out << sstr.str(); sstr.str("");
	out.close();
    }
    else if(type=="Flat")
    {
	//Column Output
	out.open("_output_multitype_readmt_Flat_.dat");
	sstr << thev1 << " ";
	sstr << thev2 << "\n"; 
	out << sstr.str(); sstr.str("");
	out.close();
    }
    else if (type == "Key")
    {
	//Key Output
	out.open("_output_multitype_readmt_Key_.dat");
	sstr << keys[0] << " = " << thev1 << "\n"; 
	sstr << keys[1] << " = " << thev2 << "\n";
	out << sstr.str(); sstr.str("");
	out.close();
    }
}

int main(int argc, char *argv[])
{
    bool allow_empty=false;
    bool vec_delim=false;
    for (int i = 1; i < argc; i++)
    {
        if (!strcmp(argv[i], "-empty") )
            allow_empty=true;
        if (!strcmp(argv[i], "-delim") )
            vec_delim=true;
    }
    
    for (int i = 1; i < argc; i++)
    {

	if (!strcmp(argv[i], "-h") || !strcmp(argv[i], "-?"))
	{
	    usage("multitype");
	    return 0;
	}
	
	//multitype new function
	if (!strcmp(argv[i], "-mt") || !strcmp(argv[i], "-mtDS"))
	{
	    multitype("DS", allow_empty, vec_delim);
	    return 0;
	}
	if (!strcmp(argv[i], "-mtCol"))
	{
	    multitype("Col", allow_empty, vec_delim);
	    return 0;
	}
        if (!strcmp(argv[i], "-mtFlat"))
	{
	    multitype("Flat", allow_empty, vec_delim);
	    return 0;
	}
	if (!strcmp(argv[i], "-mtXML"))
	{
	    multitype("XML", allow_empty, vec_delim);
	    return 0;
	}
	if (!strcmp(argv[i], "-mtKey"))
	{
	    multitype("Key", allow_empty, vec_delim);
	    return 0;
	}
	if (!strcmp(argv[i], "-mtRow"))
	{
	    multitype("Row", allow_empty, vec_delim);
	    return 0;
	}

	//readmultitype new test
	if (!strcmp(argv[i], "-ReadmtDS"))
	{
	    readMultiType("DS", allow_empty);
	    return 0;
	}
	if (!strcmp(argv[i], "-ReadmtCol"))
	{
	    readMultiType("Col", allow_empty);
	    return 0;
	}
	if (!strcmp(argv[i], "-ReadmtFlat"))
	{
	    readMultiType("Flat", allow_empty);
	    return 0;
	}
	if (!strcmp(argv[i], "-ReadmtXML"))
	{
	    readMultiType("XML", allow_empty);
	    return 0;
	}
	if (!strcmp(argv[i], "-ReadmtKey"))
	{
	    readMultiType("Key", allow_empty);
	    return 0;
	}
	if (!strcmp(argv[i], "-ReadmtRow"))
	{
	    readMultiType("Row", allow_empty);
	    return 0;
	}

    }
    return 0;
}
