"""
Example of flowrate code launching with XML input files
"""
from rootlogon import ROOT, DataServer, Launcher, Sampler

nS = 100
# Define the DataServer
tds = DataServer.TDataServer("tdsFlowrate", "DoE for Flowrate")

# Add the study attributes ( min, max and nominal values)
tds.addAttribute(DataServer.TUniformDistribution("rw", 0.05, 0.15))
tds.addAttribute(DataServer.TUniformDistribution("r", 100.0, 50000.0))
tds.addAttribute(DataServer.TUniformDistribution("tu", 63070.0, 115600.0))
tds.addAttribute(DataServer.TUniformDistribution("tl", 63.1, 116.0))
tds.addAttribute(DataServer.TUniformDistribution("hu", 990.0, 1110.0))
tds.addAttribute(DataServer.TUniformDistribution("hl", 700.0, 820.0))
tds.addAttribute(DataServer.TUniformDistribution("l", 1120.0, 1680.0))
tds.addAttribute(DataServer.TUniformDistribution("kw", 9855.0, 12045.0))

# The reference input file
sIn = "flowrate_input_with_xml.in"
AttIn = DataServer.TAttributeFileKey.kXMLAttribute
FieIn = DataServer.TAttributeFileKey.kXMLField
XmlKey = "parameter[tonode='mesher' and toport='dt_%s']/value/double"
tds.getAttribute("rw").setFileKey(sIn, "wall_friction/@rw", "%e", AttIn)
tds.getAttribute("r").setFileKey(sIn, "wall_friction/@r", "%e", AttIn)
tds.getAttribute("tu").setFileKey(sIn, "parameter[@name='Tu']/value/double",
                                  "%e", FieIn)
tds.getAttribute("tl").setFileKey(sIn, "parameter[@name='Tl']/value/double",
                                  "%e", FieIn)
tds.getAttribute("hu").setFileKey(sIn, XmlKey % ("hu"), "%e", FieIn)
tds.getAttribute("hl").setFileKey(sIn, XmlKey % ("hl"), "%e", FieIn)
tds.getAttribute("l").setFileKey(sIn, "parameter[@name='L']/@precision",
                                 "%e", AttIn)
tds.getAttribute("kw").setFileKey(sIn, "kW/@value", "%e", AttIn)

# Generate the Design of Experiments
sampling = Sampler.TSampling(tds, "lhs", nS)
sampling.generateSample()

# The output file of the code
fout = Launcher.TOutputFileXML("_output_flowrate_withXML_.dat")
# The attribute in the output file
fout.addAttribute(DataServer.TAttribute("yhat"),
                  "/steady_state[@name='flowrate']/parameter/value/double",
                  FieIn)

# Instanciation de mon code
mycode = Launcher.TCode(tds, "flowrate -s -x")
# on ajoute le fichier de sortie du code
mycode.addOutputFile(fout)

# Lancement du code
lanceur = Launcher.TLauncher(tds, mycode)
lanceur.setSave()
lanceur.setClean()
lanceur.setVarDraw("yhat:rw", "", "")

lanceur.run()

tds.exportData("_flowrate_sampler_launcher_.dat")

# Visualisation
Canvas = ROOT.TCanvas("c1", "Graph for the Macro launchCodeFlowrateOATMinMax",
                      5, 64, 1270, 667)
pad = ROOT.TPad("pad", "pad", 0, 0.03, 1, 1)
pad.Draw()
ROOT.gStyle.SetPalette(1)
pad.Divide(2, 2)
pad.cd(1)
tds.draw("yhat:rw")

pad.cd(3)
tds.draw("yhat")
pad.cd(2)
tds.draw("yhat:rw", "", "colz")

pad.cd(4)
# The parallel plot
tds.draw("rw:r:tu:tl:hu:hl:l:kw:yhat", "", "para")
para = ROOT.gPad.GetListOfPrimitives().FindObject("ParaCoord")

# The output attribute
axis = para.GetVarList().FindObject("yhat")
axis.AddRange(ROOT.TParallelCoordRange(axis, 15.0, 80.0))
para.AddSelection("blue")
para.GetCurrentSelection().SetLineColor(ROOT.kBlue)
