//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
// TABC
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.2 $
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TABC.h
 \author Rudy Chocat
 \date   2020/07/17
 
 \brief Interface of class URANIE::ABC::TABC 

 
 */

#ifndef TABC_H
#define TABC_H
 
////// ROOT
#include "TTree.h"
#include "TMatrixD.h"
#include "TRandom3.h"
#include "TMath.h"
 
// Uranie
#include "TCalibration.h"
#include "TDataServer.h"
 
namespace URANIE
{
namespace Calibration
{
 
class TABC : public URANIE::Calibration::TCalibration
{
public :
    vector<double>                                     _values;
    URANIE::DataServer::TDataServer *_tdsPosterior;     ///< TDS containing posterior members for calibration
    
protected:    

    double _threshold; 
    Int_t _kPosterior;
    Int_t _kEval;
    
    TRandom3 *_rand;

    Bool_t _boolEpsilon;
    Double_t _Epsilon;


    //@}
public :

    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    
    /*! 
     * \brief Default ABC constructor with the runner argument: it contains the assessor to be used
     * \param tds : the dataserver that contains no data but only one attribute per parameter to be calibrated
     * \param run : the runner that contains the assessor to be used on the reference data
     * \param ns : number of sample to be generated (depending on the method)
     */
    TABC(URANIE::DataServer::TDataServer *tds, URANIE::Relauncher::TRun *run, int ns = 100, Option_t *option = "");
    
    /*! 
     * \brief Default ABC constructor with the function argument: it contains the assessor to be used   
     * \param tds : the dataserver that contains no data but only one attribute per parameter to be calibrated
     * \param fcn : the pointer to a function
     * \param ns : number of sample to be generated (depending on the method)
     * \param varexpinput : the input variable for the function in the correct order (both input and parameters)
     * \param varexpoutput : the output of the function in the correct order 
     */
    TABC(URANIE::DataServer::TDataServer *tds, void (*fcn)(Double_t*,Double_t*),  const char *varexpinput, const char *varexpoutput, int ns = 100, Option_t *option = "");
    
    /*! 
     * \brief Default ABC constructor with the function argument: it contains the assessor to be used
     * \param tds : the dataserver that contains no data but only one attribute per parameter to be calibrated
     * \param fcn : the name of the function
     * \param ns : number of sample to be generated (depending on the method)
     * \param varexpinput : the input variable for the function in the correct order (both input and parameters)
     * \param varexpoutput : the output of the function in the correct order 
     */
    TABC(URANIE::DataServer::TDataServer *tds, const char *fcn, const char *varexpinput, const char *varexpoutput, int ns = 100, Option_t *option = "");
    
    /*! 
     * \brief Default ABC constructor with the code argument: it contains the assessor to be used
     * \param tds : the dataserver that contains no data but only one attribute per parameter to be calibrated
     * \param code : the code object that will be runned
     * \param ns : number of sample to be generated (depending on the method)
     */
    TABC(URANIE::DataServer::TDataServer *tds, URANIE::Launcher::TCode *fcode, int ns = 1, Option_t *option = "");
    
    //! Default destructor
    virtual ~TABC();
    //@}

    //---------------------------------------------
    /** @name Initialisation methods
     **/
    //@{
    void initABC();
    
    //@}
    
    //---------------------------------------------
    /** @name Internal methods
     **/
    
    //@{  
    //---------------------------------------------
    /** @name ABC functions
     **/ 
    void setUpABC();
//     void evalABC();
    void computeABC(int nEval);
    void computeABCscore(Double_t newEpsilon);
    
    void posteriorToPar();

    /*!
     * Check the content of the input dataserver
     */
    
    void checktdsParContent();
    
    //@}
    //---------------------------------------------
    /** @name Setting and Getting attributs
     **/
    //@{
    /*! Add gaussian noise for deterministic model
     * \param stdname : name of the attribute in the observed dataserver that will be used to introduce gaussian noise
     */
    void setGaussianNoise(const char *stdname);
       
    //!  Set the percentile 
    void setPercentile(Double_t eps)
    {
        try
        {
            if ((eps>1.0) || (eps<0.0))
            {
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form( "TABC::setPercentile: the percentile must be set between 0 and 1: however it is fix at %4.2f ! ", eps));
            }
        }
        catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            throw ue;
        }
        
        _boolEpsilon = true;
        _Epsilon = eps;
    }

    //!  Get the posterior tds 
    URANIE::DataServer::TDataServer *getTdsPosterior()
    {
        return _tdsPosterior;
    }
    //!  Get the posterior tds    
    double getThreshold()
    {
        try
        {
            if (_threshold==12345678.9)
            {
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                         "TABC::getThreshold: the threshold is not yet computed");
            }
        }
        catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            throw ue;
        } 
        return _threshold;
    }
    int getNEval()
    {
        return _kEval;
    }
    
    
    
    
    //@}

    ClassDef(URANIE::Calibration::TABC, ID_CALIBRATION)
 
};
 
} // End of namespace ABC
} // End of namespace URANIE
#endif
