//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TDistanceFunction.h
 \author Jean-Baptiste Blanchard (jean-baptiste.blanchard@cea.fr)
 \date   

 \brief Interface de la classe URANIE::Calibration::TDistanceFunction

 */
/**

 * @class URANIE::Calibration::TDistanceFunction
 *
 * @brief Description of the class TDistanceFunction.
 *
 * This class defines the structure of all distance function (DF) and rely on the relauncher
 * structure to do so. Every DF will be based on this class, inherinting from TSimpleEval.
 * The aim of the distance function is to 
 * \li handle the data (input / reference values and possible weights)
 * \li get parameters value from the TCalibration inheriting object and recompute estimated outputs
 * for all events in reference dataserver
 * \li store the apriori evaluation for residuals plots (and possibly all evaluations if requested)
 * \li compute a distance that will be used as a way to quantify how close the new parameters set is to the reference.
 *
 */
#ifndef TDISTANCEFUNCTION_H
#define TDISTANCEFUNCTION_H    

// Root
#include "TMatrixD.h"
#include "TRandom3.h"

//// URANIE
#include "Calibration.h"
#include "TStandardEval.h"
#include "TRun.h"
#include "TCode.h"
#include "TLauncher.h"

namespace URANIE
{
namespace Calibration
{

class TDistanceFunction: public URANIE::Relauncher::TDoubleEval
{


public:

    URANIE::DataServer::TDataServer *_tdsEval;        ///< pointer toward the evaluation dataserver
    URANIE::DataServer::TDataServer *_tdsObs;         ///< pointer toward the reference dataserver
    URANIE::DataServer::TDSNtupleD  *_allData;         ///< pointer toward the reference dataserver

protected:
    //Name of the variables
    vector<string>                   _vrefName;       ///< Vector of reference name
    vector<string>                   _voutName;       ///< Vector of output name
    vector<string>                   _vparName;       ///< Vector of parameters name

    //Storing data
    vector<vector<double> >          _yObs;           ///< Vector of observations outputs (size: _nVar < _nObs >)
    vector<vector<double> >          _yExp;           ///< Vector of tested values  (size: _nVar <_nObs>)
    TMatrixD                         _mObsCovMat;     ///< Observation Covariance matrix
    TMatrixD                         _myObs;          ///< Vector of observations outputs (size: _nVar < _nObs >)
    TMatrixD                         _myExp;          ///< Vector of tested values  (size: _nVar <_nObs>)
    unsigned int                     _nObs;           ///< Number of observations
    unsigned int                     _nVar;           ///< Number of variables;
    unsigned int                     _nPar;           ///< Number of Parameters;
    vector<double>                   _vParValues;
    
    //Weights for events and variables
    vector<double>                   _varWeights;     ///< Vector of variable weights (size: _nVar)
    vector<vector<double> >          _obsWeights;     ///< Vector of observation weights outputs (size: _nObs)

    URANIE::Relauncher::TRun        *_run;
    TString                          _tlcType;
    TString                          _codeLauncherOpt;///< Option for code launcher
    URANIE::Launcher::TCode         *_code;
    TString                         _funcName;
    void (*_pFunction)(double *, double *);           ///< Function pointer
    TString                         _varexpinput;
    TString                         _varexpoutput;        
        
    // Options
    bool                             _baprioriSet;    ///< Whether apriori is set to evaluator
    bool                             _bsaveAll;       ///< Whether all evaluations should be saved, not only a priori and a posteriori
    bool                             _buseMatrix;     ///< Use matrix instead of vectors
    bool                             _blog;           ///< Boolean to decide if the log information is shown or not.
    bool                             _dumpAllTds;     ///< dump all tested dataserver;
    bool                             _keepParValue;   ///< Keep the parameters values in runner case
    bool                             _bdumpDetails;   ///< Dump the details in Distance Computation
    
    // Iterators
    int                              _icalc;          ///< Calcul iterator;
    unsigned int                     _ivar, _iobs;

    // Attribute
    TList                           *_listOfParameters;///< list of parameters

    //Random noise
    TRandom3                        *_rand;           ///< The  random generator 
    int                              _nSeed;          ///< The seed of the random generator
    TMatrixD                         _mSigma;         ///< Vector of variance of errors
    bool                             _boolNoise;      ///< Whether noise is add to the output
    
public:

    /**
     * @class hide_DistanceConstructor
     * \li the evaluation dataserver 
     * \li the reference dataserver
     * \li the list of variable that are the output reference (_nVar in total)
     * \li the outputs of the code (_nVar as well)
     * \li the observation weight (uncertainty for instance)
    */
    
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{            
    //! Standard constructor
    /*!
     * A TDistanceFunction object is created by providing
     * @copydoc hide_DistanceConstructor
     */
    TDistanceFunction(URANIE::DataServer::TDataServer *tdsEval, URANIE::DataServer::TDataServer *tdsObs, const char *reference, const char *output, const char *weight="");

    //! Default destructor
    virtual ~TDistanceFunction();
    //@}


    //---------------------------------------------
    /** @name Initialisation and evaluation
     **/
    //@{

    //! Internal initialisation method
    void fillInOutAtt(URANIE::DataServer::TDataServer *tdsPar);
    void initParameters(URANIE::DataServer::TDataServer *tdsPar, URANIE::Relauncher::TRun *run);
    void initParameters(URANIE::DataServer::TDataServer *tdsPar, const char* funcName, const char *varexpinput, const char *varexpoutput);
    void initParameters(URANIE::DataServer::TDataServer *tdsPar, void (*fcn)(Double_t*,Double_t*), const char *varexpinput, const char *varexpoutput);
    void initParameters(URANIE::DataServer::TDataServer *tdsPar, URANIE::Launcher::TCode *Code);

    /*! 
      \brief Official method inherited from TSimpleEval, it is the main method of a DF
      It calls three methods in a row : runConfiguration, getData and localeval
      \param theta: value of the parameters to be tested (dimension _nPDar)
      \param res: output value of the DF (dimension 1)
    */
    Int_t eval(Double_t* theta, Double_t* res, int=0);

    /*! 
      \brief Runs the code provided on the reference input to get estimations
      \param theta: value of the parameters to be tested (dimension _nPar)
    */
    void runConfiguration(double *theta);

    //@}
    

    //---------------------------------------------
    /** @name Get and Set methods
     **/
    //@{

    //! Return the observation vector of vector
    vector<vector<double> > getObservationVector() {
        try{
            if( _buseMatrix == true )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           "TDistanceFunction::getObservationVector. This distancefunction is using matrix. Please call getObservationMatrix() instead");
            
        } catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            throw ue;
        }
        return _yObs;
    }

    //! Return the observation matrix 
    TMatrixD getObservationMatrix() {
        try{
            if( _buseMatrix == false )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           "TDistanceFunction::getObservationMatrix. This distancefunction is using vectors. Please call getObservationVector() instead");
            
        } catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            throw ue;
        }
        return _myObs;
    }

    //! Choose whether every evaluations should be kept (might slown down the process)
    void setSaveAllEvaluations(bool value)
    {
        _bsaveAll = value;
    }

    //! Change the code launcher
    void changeLauncher(TString tlcName);

    //! Change the option of the code launcher
    void changeCodeLauncherOpt(TString opt){
        _codeLauncherOpt=opt;
    }

    void dumpAllDataservers(){
        _dumpAllTds=true;
    }

    void keepParametersValue(){
        _keepParValue=true;
    }

    void dumpDetails(){
        _bdumpDetails=true;
    }

    //! add options to the already  of the code launcher
    void addCodeLauncherOpt(TString opt){
        _codeLauncherOpt+=","+opt;
    }

    //! Choose whether vector of vector or matrix should be used in the distance function localeval method
    void setUseMatrix(bool value)
    {
        _buseMatrix = value;
        getData(true);
    }

    //! Set a weight to ponderate variable when _nVar >1
    void setVarWeights( int nwei, double *wei);
    //! Set a weight to ponderate variable when _nVar >1
    void setVarWeights(vector<double> wei);

    //! set the observatiton covariance matrix
    void setObservationCovarianceMatrix(TMatrixD &mat);

    //! Check the value of the saveAll option
    bool getSaveAllEvaluation() {
       return _bsaveAll;
    }

    //! Check the value of the useMatrix option
    bool getUseMatrix() {
       return _buseMatrix;
    }
    
    //! Set Normal random noise for an output variable
    void setGaussianRandomNoise(const char *stdname);
    
        //! Set the seed of the random generator if one is used
    void setSeed(UInt_t nval)
    {
        _nSeed = nval;
    }


    //@}

    
    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    void setLog()
    {
        _blog = kTRUE;
    }
    void unsetLog()
    {
        _blog = kFALSE;
    }
    void changeLog()
    {
        _blog = !_blog;
    }
    Bool_t getLog()
    {
        return _blog;
    }
    virtual void printLog(Option_t *option = "");
    //@}
    
    
protected:
    //---------------------------------------------
    /** @name Internal methods
     **/
    //@{            

    /*! 
      \brief Get the data either the reference one, or the estimated one for a given parameter configuration
      \param reference: boolean by default set to false.
    */
    void getData(bool reference=false);  

    /*! 
      \brief Use the reference and the newly obtained estimations to compute a distance.
      \param res: output value of the DF (dimension 1)
    */    
    virtual Int_t localeval(Double_t* res)=0;

    //@}


    ClassDef(URANIE::Calibration::TDistanceFunction, ID_CALIBRATION)
};

} // Fin du namespace Calibration
} // Fin du namespace URANIE
#endif
