//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
// TLinearBayesian
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TLinearBayesian.h
 \author Jean-Baptiste Blanchard
 \date   2020/07/17

 \brief Interface of class URANIE::Calibration::TLinearBayesian

 */
/**

 * @class URANIE::Calibration::TLinearBayesian
 *
 * @brief Description of the class TLinearBayesian.
 *

 *
 */
#ifndef TLINEARBAYESIAN_H
#define TLINEARBAYESIAN_H

// Uranie
#include "TCalibration.h"

namespace URANIE
{
namespace Calibration
{

class TLinearBayesian: public URANIE::Calibration::TCalibration {

protected:
    
    void (*_fTransfoParam)(double *, double*);        ///< Parameter transformation function
    vector<double> _aPrioriMode;                      ///< A priori modes of the laws

    TMatrixD _mParCovariance;                         ///< Parameters covariance matrix
    TMatrixD _mParValues;                             ///< Parametres matrix
    TMatrixD _mTransfoParValues;                      ///< Parametres matrix

    TMatrixD _mH;                                     ///< Regressor matrix
    vector<string> _vRegName;                         ///< Regressor names
    string _regname;                                  ///< regressor name
public:
    
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{

    //! Default constructor with TRun arg
    /*! 
     * \brief Default constructor with the runner argument: it contains the assessor to be used
     * \param tds : the dataserver that contains no data but only one attribute per parameter to be calibrated
     * \param run : the runner that contains the assessor to be used on the reference data
     * \param ns : number of sample to be generated (depending on the method)
     * \param option : see parseOption below
     */
    TLinearBayesian(URANIE::DataServer::TDataServer *tds, URANIE::Relauncher::TRun *run, Int_t ns=1, Option_t * option = "");

    /*! 
     * \brief Default Calibration constructor with the function argument: it contains the assessor to be used
     * \param tds : the dataserver that contains no data but only one attribute per parameter to be calibrated
     * \param fcn : the pointer to a function
     * \param ns : number of sample to be generated (depending on the method)
     * \param varexpinput : the input variable for the function in the correct order (both input and parameters)
     * \param varexpoutput : the output of the function in the correct order 
     */
    TLinearBayesian(URANIE::DataServer::TDataServer *tds, void (*fcn)(Double_t*,Double_t*), const char *varexpinput, const char *varexpoutput, int ns = 1, Option_t * option = "");

    /*! 
     * \brief Default Calibration constructor with the function argument: it contains the assessor to be used
     * \param tds : the dataserver that contains no data but only one attribute per parameter to be calibrated
     * \param fcn : the name of the function
     * \param ns : number of sample to be generated (depending on the method)
     * \param varexpinput : the input variable for the function in the correct order (both input and parameters)
     * \param varexpoutput : the output of the function in the correct order 
     */
    TLinearBayesian(URANIE::DataServer::TDataServer *tds, const char *fcn, const char *varexpinput, const char *varexpoutput, int ns = 1, Option_t * option = "");

    /*! 
     * \brief Default Calibration constructor with the code argument: it contains the assessor to be used
     * \param tds : the dataserver that contains no data but only one attribute per parameter to be calibrated
     * \param code : the code object that will be runned
     * \param ns : number of sample to be generated (depending on the method)
     */
    TLinearBayesian(URANIE::DataServer::TDataServer *tds, URANIE::Launcher::TCode *fcode, int ns = 1, const char *option = "");
    
    //! Default destructor
    virtual ~TLinearBayesian();
    //@}

    
    //---------------------------------------------
    /** @name Predictions methods
     **/
    //@{

    /*! 
     * \brief Set parameter transformation function
     * Sometimes the calibration is performed on transformed variables (for linear problem for instance)
     * This method allows to provide a custom function to trasform the estimated parameters to the original one.
     * the first argument of the fonction must an array of double of size _nPar and the second argument is also an
     * array of double of size _nPar
     */
    void computePredictionVariance(URANIE::DataServer::TDataServer *tds_new, string outname);
    //@}

    
    //---------------------------------------------
    /** @name Setting and Getting attributs
     **/
    //@{

    /*! 
     * \brief Set parameter transformation function
     * Sometimes the calibration is performed on transformed variables (for linear problem for instance)
     * This method allows to provide a custom function to trasform the estimated parameters to the original one.
     * the first argument of the fonction must an array of double of size _nPar and the second argument is also an
     * array of double of size _nPar
     */
    void setParameterTransformationFunction( void (*fTransfoParam)(double *, double*)){ _fTransfoParam = fTransfoParam; }

    /*!
     * \brief Set the regressor matrix by providing the variables to be extracted from tdsObs
     */
    void setRegressorName(const char *regName);
    
    /*!
     * \brief Set the distance function and some needed informations
     * \param funcName: name of the distance function chosen among the already implemented ones:
     \li "LS" for least square
     \li "weighterLS" for weighted least square
     \li "relativeLS" for relative least square
     \li "L1" for L1 norm 
     \li "Mahalanobis" for Mahalanobis distance function
     * \param tdsRef : the dataserver that contains all needed information detailled below 
     \li input: the input attributes used of the assessors in run
     \li reference: the reference attributes that will be used to compared the newly done estimations for every iterations
     \li weight: a weight to reweight every event, one-by-one (an experimental uncertainty for instance)
    */
    void setDistanceAndReference(const char *funcName, URANIE::DataServer::TDataServer *tdsRef, const char *input, const char *output, const char *weight="");

    /*!
     * \brief Set the distance function and some needed informations
     * \param distFunc: a pointer to a TDistanceFunction object (not usual, recommended only when dealing with a home-made distance unction compiled on the spot)
     * \param tdsRef : the dataserver that contains all needed information detailled below 
     \li input: the input attributes used of the assessors in run
     \li reference: the reference attributes that will be used to compared the newly done estimations for every iterations
     \li weight: a weight to reweight every event, one-by-one (an experimental uncertainty for instance)
    */
    void setDistanceAndReference(URANIE::Calibration::TDistanceFunction *distFunc, URANIE::DataServer::TDataServer *tdsRef, const char *input, const char *output, const char *weight="");  

    /*!
     * \brief get the matrix of parameter values
     */
    TMatrixD getParameterValueMatrix(){ return _mParValues;}

    /*!
     * \brief get the matrix of parameter values
     */
    TMatrixD getTransfParameterValueMatrix(){ return _mTransfoParValues;}
    
    /*!
     * \brief get the matrix of parameter covariances
     */
    TMatrixD getParameterCovarianceMatrix(){ return _mParCovariance;}
    
    //@}

    //---------------------------------------------
    /** @name Visualisation methods
     **/
    //@{

    /*!     
     * brief Draws the parameters as functions (ROOT's TF1s)
     * The estimateParameters method hass computed the gaussian parameters, so this one plots the parameters as continous function TF1
     * \param sTitre: title of the plots if specified
     * \param variable: list of variables that should be plotted
     * \param select: useless for this method
     * \param option: possible option for this plot
     \li "nonewcanvas": don't create a new canvas, but use the one active 
     \li "vertical" : if there are several parameters to be shown put them on top of another instead of side by side
     \li "transformed" : use the transformed central alues and not the nominal ones
     \li "same" : plot the function on the existing pad, using the existing axis.
    */
    void drawParameters(TString sTitre, const char *variable = "*", const char *select = "1>0", Option_t * option = "");
    
    /*!
     * \brief dump content
     */
    void printLog(Option_t *option = "");
    
    //@}

protected:
    void computeParameters(Option_t *option="");
    /*!
     * Check the content of the input dataserver
     */ 
    void checktdsParContent();
    
    ClassDef(URANIE::Calibration::TLinearBayesian, ID_CALIBRATION)

    
};

} // Fin du namespace Calibration
} // Fin du namespace URANIE

#endif
