//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
// TMinimisation
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
  \file   TMinimisation.cxx
  \author Jean-Baptiste Blanchard
  \date   2020/07/17

  \brief Implementation of the class URANIE::Calibration::TMinimisation


*/

// Uranie
#include "TMinimisation.h"
#include "TSequentialRun.h"
#include "TThreadedRun.h"
#include "TVizir2.h"
#include "TNloptDirect.h"
#include "TNloptCobyla.h"

ClassImp (URANIE::Calibration::TMinimisation)

namespace URANIE
{
namespace Calibration
{

// -----------------------------------------------------------------------
//  Constructors & Initialisation
// -----------------------------------------------------------------------

TMinimisation::TMinimisation(URANIE::DataServer::TDataServer *tds, URANIE::Relauncher::TRun *runner, Int_t ns, Option_t * option):
    TCalibration(tds, runner, ns, option),
#ifdef WITH_NLOPT
    _solv(NULL),
#endif
#ifdef WITH_VIZIR    
    _solviz(NULL),
#endif
    _optimRun(NULL), _optimMaster(NULL)
{
    setMethodName("Minimisation");

    parseOption(option);
}

TMinimisation::TMinimisation(URANIE::DataServer::TDataServer *tds, void (*fcn)(Double_t*,Double_t*), const char *varexpinput, const char *varexpoutput, int ns, Option_t *option):
    TCalibration(tds, fcn, varexpinput, varexpoutput, ns, option)
{
    try{
        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                   "This constructor is not implemented as the minimisation relies on the Reoptimizer module");
    }catch(URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }        
}

TMinimisation::TMinimisation(URANIE::DataServer::TDataServer *tds, const char *fcn, const char *varexpinput, const char *varexpoutput, int ns, Option_t *option):
    TCalibration(tds, fcn, varexpinput, varexpoutput, ns, option)   
{
    try{
        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                   "This constructor is not implemented as the minimisation relies on the Reoptimizer module");
    }catch(URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }        
}

TMinimisation::TMinimisation(URANIE::DataServer::TDataServer *tds, URANIE::Launcher::TCode *fcode, int ns, Option_t *option):
    TCalibration(tds, fcode, ns, option)
{
    try{
        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                   "This constructor is not implemented as the minimisation relies on the Reoptimizer module");
    }catch(URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }        
}

// -----------------------------------------------------------------------
//  Destructors
// -----------------------------------------------------------------------

TMinimisation::~TMinimisation()
{
    if (_blog)
        cout << " *** destroy TMinimisation(" << GetName() << ") Title[" << GetTitle() << "] *** " << endl;

    if( _optimRun != NULL )
    {
        delete _optimRun;
        _optimRun = NULL;
    }

    if( _optimMaster != NULL )
    {
        delete _optimMaster;
        _optimMaster = NULL;
    }

}


// -----------------------------------------------------------------------
//  Define properties
// -----------------------------------------------------------------------

#ifdef WITH_NLOPT
void TMinimisation::setOptimProperties(URANIE::Reoptimizer::TNloptSolver *solv, Option_t *option)
{
    // Check the option to search for sequential flag for instance    
    parseOption(option);

    // Assess the solver
    _solv = solv;

    try
    {
        if( _solv == NULL )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TMinimisation::setOptimProperties \n The nlopt-based solver provided is empty (null)");
    }catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
   
    // Set the runner depending on the number of initialisati$on points
    _optimRun = new URANIE::Relauncher::TSequentialRun(_dFunc);
    if(_optimRun->onMaster())
    {
        // Create the single-objective constrained optimizer
        _optimMaster = new URANIE::Reoptimizer::TNlopt(_tdsPar, _optimRun, _solv);
    }
}
#endif

#ifdef WITH_VIZIR
void TMinimisation::setOptimProperties(URANIE::Reoptimizer::TVizirSolverShare *solv, Option_t *option)
{
    // Check the option to search for sequential flag for instance    
    parseOption(option);

    // Assess the solver
    _solviz = solv;

    try
    {
        if( _solviz == NULL )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TMinimisation::setOptimProperties \n The vizir-based solver provided is empty (null)");
    }catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
   
    // Set the runner depending on the number of initialisation points
    _optimRun = new URANIE::Relauncher::TSequentialRun(_dFunc);
    if(_optimRun->onMaster())
    {
        // Create the single-objective constrained optimizer
        _optimMaster = new URANIE::Reoptimizer::TVizir2(_tdsPar, _optimRun, _solviz);
    }
}
#endif

void TMinimisation::computeParameters(Option_t *option)
{

    parseOption(option);

    try
    {
        if( _optimRun == NULL || _optimMaster==NULL )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TMinimisation::computeParameters \n \
The method estimateParameters should not be called while setOptimProperties has not been called");
    }catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

#ifdef WITH_NLOPT
    URANIE::Reoptimizer::TNlopt  *nloptMaster=dynamic_cast<URANIE::Reoptimizer::TNlopt  *>(_optimMaster);
    // Check that the optimMaster has been configured with startingpoint
    // If not, check whether parameters have default values

    if(_blog)
        XYZZY;
    
    if(nloptMaster != NULL)
    {
        if(nloptMaster->getNStartingPoint()==0)
        {
            vector<double> point(_nPar);

            if(_blog)
                XYZZY;
            
            try
            {
                for(int ipar=0; ipar<_nPar; ipar++)
                    if( !_tdsPar->getAttribute(ipar)->getDefaultValue(point[ipar]) )
                        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                                   Form("TMinimisation::computeParameters \n No initialisation is done, either through the optimisation master directly \
or as default value for attribute [%s]",_tdsPar->getAttribute(ipar)->GetName()));                
            } catch (URANIE::Exceptions::UErrorExceptions& ue)
            {
                ue.printMessage();
                throw ue;
            }

            nloptMaster->setStartingPoint(_nPar,&point[0]);
            if(_blog)
            {
                for(int ipar=0; ipar<_nPar; ipar++)
                    XYZ<<"Par("<<ipar<<")= "<<point[ipar]<<"  ";
                cout<<endl;
            }
        }
    }
#endif
#ifdef WITH_VIZIR
    URANIE::Reoptimizer::TVizir2 *vizirMaster=dynamic_cast<URANIE::Reoptimizer::TVizir2 *>(_optimMaster);
#endif
    // Choose a cost function to be minimized 
    _optimRun->startSlave(); 

    if(_optimRun->onMaster())
    {
        // add the objective
        _optimMaster->addObjective(_tdsPar->getAttribute("agreement"));
        
        if(_blog) XYZZY;
        _optimMaster->solverLoop();
        if(_blog) XYZZY;
                
        _optimRun->stopSlave();
    }
    
    _tdsPar->getTuple()->SetScanField(-1);
    _tdsPar->scan("*");

    // Save the aposteriori results    
    for(unsigned int ii=0; ii<_vrefName.size(); ii++)
        _tdsEval->getTuple()->Merger(_tdsObs->getTuple(), ( _vrefName.at(ii)+"_aposteriori").c_str(), _voutName.at(ii).c_str() );
}

void TMinimisation::parseOption(Option_t *option)
{

    TString soption=option;
    soption.ToLower();

    URANIE::Calibration::TCalibration::parseOption(option);
}

void TMinimisation::printLog(Option_t *option)
{
    cout << " ** TMinimisation::printLog[" << option << "]" << endl;
    cout << " ** Name[" << GetName() << "] Title[" << GetTitle() << "]" << endl;
    cout << " **   ID[" << getID() << "]" << endl;
    cout << " ** _nSam[" << _nSam << "]" << endl;
    cout << " ** _nSeed[" << _nSeed << "]" << endl;
    cout << " **  Log of DataServer (unactive)" << endl;
    if ( 0 ) {
        _tdsPar->printLog();
        cout << " **   End of Log of DataServer" << endl;
    }
    cout << " ** End of TMinimisation::printLog[" << option << "]" << endl;
}

} // End of namespace Calibration
} // End of namespace URANIE

