//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.2 $
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TPMCABC.cxx
 \author Rudy Chocat
 \date   2020/07/17

 \brief Function to calibrate with population monte carlo abc 
 */

// Root
#include "TDecompSVD.h"
#include "TRegexp.h"
#include "TLeaf.h"

// Uranie
#include "TPMCABC.h"
#include "TNormalDistribution.h"
#include "TSampling.h"
#include "TCustomDistribution.h"

ClassImp (URANIE::Calibration::TPMCABC)
namespace URANIE
{
namespace Calibration
{
    
// -----------------------------------------------------------------------
//  Constructors & Initialisation
// ---------------------------------------------------------------------
    
TPMCABC::TPMCABC(URANIE::DataServer::TDataServer *tds, URANIE::Relauncher::TRun *run,  int ns, Option_t *option):
    TABC(tds, run, ns, option)
{
    TPMCABC::initPMCABC();
}
    
TPMCABC::TPMCABC(URANIE::DataServer::TDataServer *tds, void (*fcn)(Double_t*,Double_t*), const char *varexpinput, const char *varexpoutput, int ns, Option_t *option):
    TABC(tds, fcn, varexpinput, varexpoutput, ns, option) 
{
    TPMCABC::initPMCABC();
}

TPMCABC::TPMCABC(URANIE::DataServer::TDataServer *tds, const char *fcn, const char *varexpinput, const char *varexpoutput, int ns, Option_t *option):
    TABC(tds, fcn, varexpinput, varexpoutput, ns, option)
{
    TPMCABC::initPMCABC();
}

TPMCABC::TPMCABC(URANIE::DataServer::TDataServer *tds, URANIE::Launcher::TCode *fcode, int ns, Option_t *option):
    TABC(tds, fcode, ns, option)
{
    TPMCABC::initPMCABC();
}

// -----------------------------------------------------------------------
//  Destructor
// -----------------------------------------------------------------------
TPMCABC::~TPMCABC()
{
    if (_blog)
        cout << "************ Destructor URANIE::Calibration::TPMCABC" << endl;

    if (_blog)
        cout << "************ End of Destructor URANIE::Calibration::TPMCABC"<< endl;
    
    if( _tdsParStep!=NULL)
        delete _tdsParStep;
    _tdsParStep=NULL;
}

// -----------------------------------------------------------------------
//  Initialisation
// -----------------------------------------------------------------------
    
void TPMCABC::initPMCABC()
{
    _nSteps = 3;
    XYZZY;
    _Epsilon = 0.1;
    _vEpsilon.clear();
    _kStep = 0;
    _tdsParStep = new URANIE::DataServer::TDataServer( Form("step%s", _tdsPar->GetName()), Form("The intermediate tds of %s", _tdsPar->GetTitle()));
    for (int i=0;i<_nPar;i++)
    {
        _tdsParStep->addAttribute(_tdsPar->getAttribute(i));
    }
    _tdsParStep->addAttribute( new URANIE::DataServer::TAttribute("weight"));
    _tdsParStep->getAttribute("weight")->delShare();
    
    _Sigmat.ResizeTo(_nPar,_nPar);
}

// -----------------------------------------------------------------------
//  Compute the parameters
// -----------------------------------------------------------------------
void TPMCABC::computeParameters(Option_t *option)
{
    TABC::setUpABC();
    
    _nIterMax = (int)(_nSam / _Epsilon);
    try
    {
        if (!_boolEpsilon)
        {
            throw URANIE::Exceptions::UInfoExceptions(__FILE__, __LINE__,
                    Form( "TPMCABC::computeParameters:: the percentile is not specified : we fix it at %4.2f \% ! ", 100.0*_Epsilon));
        }
    } catch (URANIE::Exceptions::UInfoExceptions& ue)
    {
        ue.printMessage();
    }
    try
    {
        throw URANIE::Exceptions::UInfoExceptions(__FILE__, __LINE__,
                Form( "TPMCABC::computeParameters:: %d evaluations will be performed ! ", _nSteps*_nIterMax));       
    } catch (URANIE::Exceptions::UInfoExceptions& ue)
    {
        ue.printMessage();
    }
    
    
    if(_vEpsilon.empty())
    {
        for (int k=0; k<_nSteps; k++)
        {
            _vEpsilon.push_back(_Epsilon);
        }
    }
    _Epsilon= _vEpsilon[_kStep];
    TString soption = TString(option);
    soption.ToLower();
    soption.ReplaceAll(" ", "");

    _nIterMax = (int)(_nSam / _Epsilon);
    _tdsPar->deleteTuple();
    
    ////// generate and evaluate the sample
    TABC::computeABC(_nIterMax);
    
    ////// Evaluate the threshold by quantile
    _tdsPar->computeQuantile("agreement",_Epsilon,_threshold);
    if(_blog)
        cout << "** in TPMCABC::computeParameters, at step " << _kStep <<",the "<<_Epsilon<<"-Quantile is " << _threshold << endl;
    
    ////// Evaluate and generate the sample
    TABC::computeABCscore(_threshold);
    for (_kStep=1; _kStep<_nSteps; _kStep++)
    {
        TPMCABC::generateNewSample();
        TABC::computeABCscore(_threshold);
    }
    TABC::posteriorToPar();
    
}

// -----------------------------------------------------------------------
//  Generate a new sample
// -----------------------------------------------------------------------
void TPMCABC::generateNewSample()
{
    if(_blog)
        cout << "** in TPMCABC::generateNewSample, step --- " << _kStep << endl;
    
    stringstream sstr_corr;
    sstr_corr.str("");
    for (int i=0;i<_nPar;i++)
    {
        sstr_corr << _tdsParStep->getAttribute(i)->GetName();
        if(i!= (_nPar-1))
            sstr_corr << ":";
    }
    
    int kIndex = 0;
    TMatrixD MU(_nPar,1);
//      _Sigmat(_nPar,_nPar);
    TMatrixD Sigmat2(_nPar,_nPar);
    TMatrixD U(_nPar,_nPar);
    TMatrixD L(_nPar,_nPar);
    TMatrixD Z(_nPar,1);
    TMatrixD X(_nPar,1);
    
    TMatrixD Thetat;
     vector<double> weightt;
    URANIE::DataServer::TCustomDistribution *tcd;
    vector<double> dPar(_nPar + 2);
    double res=0;
    double denom=0.0;
    double num=0.0;
    int iatt=0;
    int nPost = _tdsPosterior->getTuple()->GetEntries();
    
    TList *listOfParametersLeaves = new TList;
    for (int i=0;i<_nPar;i++)
    {
        listOfParametersLeaves->Add(_tdsPosterior->getTuple()->GetLeaf( _listOfParameters->At(i)->GetName() ) );
    }

    if(_kStep>1)
    {
        Thetat.ResizeTo(_tdsParStep->getTuple()->GetEntries(),_nPar);
        Thetat = _tdsParStep->getMatrix(sstr_corr.str().c_str());
        weightt = _frequencies;
    }
    
    _tdsParStep->createTuple();
    _frequencies.clear();
    _indices.clear();
    for(int k = 0; k < _tdsPosterior->getTuple()->GetEntries(); k++)
    {
        _tdsPosterior->getTuple()->GetEntry(k);
        TIter nextLeaf(listOfParametersLeaves);
        dPar[0]=double(k);
        iatt=0;
        while (TLeaf *leaf = (TLeaf *) nextLeaf())
        {
            dPar[1+iatt] = leaf->GetValue();
            MU(iatt,0) = leaf->GetValue();
            iatt++;
        }
        if(_kStep>1)
        {
            denom = 0.0;
            for(int i = 0; i<(int)weightt.size(); i++)
            {
                for(int j = 0; j<_nPar; j++)
                {
                    X(j,0) = Thetat(i,j);
                }
                denom += weightt[i]*TPMCABC::pdfGaussianMultivariate(X,MU,_Sigmat);
            }
            num = 1.0;
            for(int i = 0; i<_nPar; i++)
            {
                num*= _vatt.at(i)->getPDF(MU(i,0));
            }
            dPar[1+iatt] = num / denom;
        }
        else
        {
            dPar[1+iatt] = 1.0/(double)nPost;
        }
        _frequencies.push_back(dPar[1+iatt]);
        if(_blog)
            cout << "** in TPMCABC::generateNewSample, weight --- " << _frequencies[k] << endl;
        _indices.push_back((double)k);
        _tdsParStep->getTuple()->Fill(&dPar[0]);
    }
    
    _tdsParStep->computeStatistic();
    TMatrixD mCorSigma = _tdsParStep->computeCorrelationMatrix(sstr_corr.str().c_str());
    for(int i=0; i< _nPar;i++)
    {
        for(int j=0; j< _nPar;j++)
        {
            _Sigmat = sqrt(mCorSigma(i,j)*_tdsParStep->getAttribute(i)->getStd()*_tdsParStep->getAttribute(j)->getStd());
            Sigmat2(i,j) = 2.0*mCorSigma(i,j)*_tdsParStep->getAttribute(i)->getStd()*_tdsParStep->getAttribute(j)->getStd();
        }
    }

    TDecompSVD svdcorr(Sigmat2);
    try 
    {
        if (svdcorr.Decompose())
        {
            for(int i=0; i< _nPar;i++)
            {
                for(int j=0; j< _nPar;j++)
                {
                    if(i==j)
                        L(i,j) = sqrt( svdcorr.GetSig()(i) );
                    else
                    {
                        L(i,j) = 0.0;
                    }
                }
            }
            U = svdcorr.GetU() * L;
        }
        else
        {
            if (_blog)
            {
                cerr << "The correlation matrix is not positive definite" << endl;
            }
            // throw "The correlation matrix is not positive definite\n";
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
            "TPMCABC::generateSampleByDakota ERROR : The correlation matrix is not positive definite");
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        throw ue;
    }
    
    
    tcd = new URANIE::DataServer::TCustomDistribution("index", _indices.size(), &_indices[0],&_frequencies[0]);
    
    listOfParametersLeaves = new TList;
    for (int i=0;i<_nPar;i++)
    {
        listOfParametersLeaves->Add(_tdsParStep->getTuple()->GetLeaf( _listOfParameters->At(i)->GetName() ) );
    }
    _tdsPar->createTuple();
    
    _Epsilon= _vEpsilon[_kStep];
    _nIterMax = (int)(_nSam / _Epsilon);
    if(_blog)
        cout <<"** in TPMCABC::generateNewSample, number of eval at step " << _kStep << " --- "<< _nIterMax << endl;
    for(int k = 0; k <_nIterMax; k++)
    {
        res=0.0;
        dPar[0]=double(k);
        kIndex = (int)tcd->getInvCDF( _rand->Rndm() );
        _tdsParStep->getTuple()->GetEntry(kIndex);
        TIter nextLeaf(listOfParametersLeaves);
        iatt=0;
        if(_blog)
            cout << "** in TPMCABC::computeParameters, index value --- " << kIndex<< endl;
        while (TLeaf *leaf = (TLeaf *) nextLeaf())
        {
            MU(iatt,0) = leaf->GetValue();
            Z(iatt,0) = _rand->Gaus();
            iatt++;
        }
        for (int i=0;i<_nPar; i++)
        {
            X(i,0) = MU(i,0) ;
            for (int j=0;j<_nPar; j++)
            {
                X(i,0) += U(i,j)*Z(j,0);
            }
        }
        _values.clear();
        for (int ix=0;ix<_nPar; ix++)
        {
            _values.push_back(X[ix][0]);
            if(_blog)
                cout<<"** in TPMCABC::computeParameters, Theta " << ix << " --- " << _values[ix] <<endl;
            dPar[1+ix] = _values[ix];      
        }
        _dFunc->eval(&_values[0],&res);_kEval++;
        if(_blog)
            cout<<"** in TPMCABC::computeParameters, res --- " << res <<endl;
        dPar[1+_nPar] = res;
    _tdsPar->getTuple()->Fill(&dPar[0]);
    }
    
    _tdsPar->computeQuantile("agreement",_Epsilon,_threshold);
    if(_blog)
        cout << "** in TPMCABC::generateNewSample, at step " << _kStep <<",the "<<_Epsilon<<"-Quantile is " << _threshold << endl;
    
    ////// Evaluate and generate the sample
    delete tcd;
    delete listOfParametersLeaves;

}

// -----------------------------------------------------------------------
//  pdf of the mulvarate gaussian distribution
// -----------------------------------------------------------------------
Double_t TPMCABC::pdfGaussianMultivariate(TMatrixD X,TMatrixD MU,TMatrixD Sigma)
{
    int dimX = X.GetNrows();
    Double_t detSigma;
    TMatrixD Sinv(Sigma); 
    Sinv.Invert(&detSigma);
    Double_t den = TMath::Sqrt(detSigma*TMath::Power(2.0 * TMath::Pi(), (double)dimX));
    TMatrixD X_MU = X-MU;
    Double_t value_exp = (TMatrixD(X_MU,TMatrixD::kTransposeMult,Sinv)*(X-MU))[0][0];
    Double_t term1 = TMath::Exp(-0.5*value_exp);
    return term1/den;
}
    
} // End of namespace ABC
} // End of namespace URANIE
