//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.2 $
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TPMCABC.h
 \author Rudy Chocat
 \date   2020/07/17

 \brief Function to calibrate with population monte carlo abc 
 */

#ifndef __PMC_ABC__
#define __PMC_ABC__

// Root
#include "TTree.h"
#include "TMatrixD.h"

// Uranie
#include "TABC.h"
#include "TDataServer.h"

namespace URANIE
{
namespace Calibration
{
    
class TPMCABC : public URANIE::Calibration::TABC
{

public :
    URANIE::DataServer::TDataServer *_tdsParStep;   ///< Pointer of the intermediate TDS
    vector<double> _frequencies;                    ///< Vector of the frequencies for sampling
    vector<double> _indices;                        ///< Vector of possible indices for sampling
    vector<double> _vEpsilon;                       ///< Vector of the thresholds
    TMatrixD _Sigmat;                               ///< Sqrt covariance matrix
    int _kStep;                                     ///< Iterator of steps
    int _nSteps;                                    ///< Number of steps
    
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{

    /*!
      Constructor of PMC ABC where nABC is set to be the number of the produced a posteriori sample.
      \param tds: the pointer of the TDataServer. It must contains attribuutes that inherit from TStochasticAttribute
      \param run: the runner that would be used to produce evaluations 
      \param nABC:  size of a posteriori sample [100 by default], by default also size of the sample evaluated at each bloc, 
    */
    TPMCABC(URANIE::DataServer::TDataServer *tds, URANIE::Relauncher::TRun *run, int ns = 1, Option_t *option = "");
    
    /*!
      Constructor of PMC ABC where nABC is set to be the number of the produced a posteriori sample.
      \param tds: the pointer of the TDataServer. It must contains attribuutes that inherit from TStochasticAttribute
      \param fcn: the function that would be used to produce evaluations 
      \param nABC:  size of a posteriori sample [100 by default], by default also size of the sample evaluated at each bloc, 
    */
    TPMCABC(URANIE::DataServer::TDataServer *tds, void (*fcn)(Double_t*,Double_t*),  const char *varexpinput, const char *varexpoutput, int ns = 100, Option_t *option = "");
    
    
    /*!
      Constructor of PMC ABC where nABC is set to be the number of the produced a posteriori sample.
      \param tds: the pointer of the TDataServer. It must contains attribuutes that inherit from TStochasticAttribute
      \param fcn: the function that would be used to produce evaluations 
      \param nABC:  size of a posteriori sample [100 by default], by default also size of the sample evaluated at each bloc, 
    */
    TPMCABC(URANIE::DataServer::TDataServer *tds, const char *fcn, const char *varexpinput, const char *varexpoutput, int ns = 100, Option_t *option = "");
    
    /*!
      Constructor of PMC ABC where nABC is set to be the number of the produced a posteriori sample.
      \param tds: the pointer of the TDataServer. It must contains attribuutes that inherit from TStochasticAttribute
      \param code: the code that would be used to produce evaluations 
      \param nABC:  size of a posteriori sample [100 by default], by default also size of the sample evaluated at each bloc, 
    */    
    TPMCABC(URANIE::DataServer::TDataServer *tds, URANIE::Launcher::TCode *fcode, int ns = 1, Option_t *option = "");
    

    //! Default destructor
    virtual ~TPMCABC();
    //@}
    
    //---------------------------------------------
    /** @name PMC ABC function
     **/
       
    //@{
    /*! 
     * \brief Generate the sample 
     * \param option: possble options are
     * \li "clean": remove internal branche from the tdsPar object at the end
     * \li "gd": another implementation of this algorithm
     */
    void computeParameters(Option_t *option = "");
    //@}
    
    void initPMCABC();
    void generateNewSample();
    double pdfGaussianMultivariate(TMatrixD X,TMatrixD MU,TMatrixD Sigma);
    
    int getNSteps()
    {
        return _nSteps;
    }
    
    
    URANIE::DataServer::TDataServer *getTdsStep()
    {
        return _tdsParStep;
    }
    
    void setThresholds(vector<double> vectTh)
    {
        _vEpsilon = vectTh;
        try
        {
            for(int i=0; i < (int)_vEpsilon.size(); i++)
            {
                if ((_vEpsilon[i]>1.0) or (_vEpsilon[i]<0.0))
                {
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                            Form( "TABC::setThresholds: the percentile must be set between 0 and 1: however it is fix at %4.2f at step %d ! ",_vEpsilon[i],i));
                }
            }
        }
        catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            throw ue;
        }
        _nSteps = (int)_vEpsilon.size();
    }
    void setNSteps(int steps)
    {
        try
        {
            if ((steps<1))
            {
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                        Form( "TABC::setNSteps: the number of steps must be strictly upper than 0: however it is fix at %d ! ", steps));
            }
        }
        catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            throw ue;
        }
        _nSteps = steps;
    }
    
    
    ClassDef(URANIE::Calibration::TPMCABC, ID_CALIBRATION)
    
};
    
} // Fin du namespace ABC
} // Fin du namespace URANIE
#endif
