
//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TStandardDistanceFunction.cxx
 \author Jean-Baptiste Blanchard (jean-baptiste.blanchard@cea.fr)
 \date   

 \brief Implementation of the class  URANIE::Calibration::TStandardDistanceFunction

 */

//// URANIE
#include "TStandardDistanceFunction.h"

ClassImp (URANIE::Calibration::TL1DistanceFunction)
ClassImp (URANIE::Calibration::TLSDistanceFunction)
ClassImp (URANIE::Calibration::TWeightedLSDistanceFunction)
ClassImp (URANIE::Calibration::TRelativeLSDistanceFunction)
ClassImp (URANIE::Calibration::TMahalanobisDistanceFunction)

namespace URANIE
{
namespace Calibration
{


// -----------------------------------------------------------------------
//   LEAST SQUARE
// -----------------------------------------------------------------------
TLSDistanceFunction::TLSDistanceFunction(URANIE::DataServer::TDataServer *tdsEval, URANIE::DataServer::TDataServer *tdsObs, const char *reference, const char *output, const char *weight) :
    TDistanceFunction(tdsEval, tdsObs, reference,  output, weight)
{
}

TLSDistanceFunction::~TLSDistanceFunction()
{
}

int TLSDistanceFunction::localeval(Double_t* res)
{
   
    res[0] = 0.0;
    double inter=0.0;
    if(_bdumpDetails)
        cout<<"Computing LS distance "<<endl;
    for(_ivar=0; _ivar<_nVar; _ivar++)
    {
        if(_bdumpDetails)
            cout<<"For variable "<<_ivar<<": sum set to  "<<inter<<endl;
        
        inter=0;
        for(_iobs=0; _iobs<_nObs; _iobs++)
        {
            if(_bdumpDetails)
                cout<<"  evt("<<_iobs<<"), sum += ("<<_yObs[_ivar].at(_iobs)<<" - "<<_yExp[_ivar].at(_iobs)<<" )^2 = "<<inter<<endl;
            inter +=  (_yObs[_ivar].at(_iobs) - _yExp[_ivar].at(_iobs) )*(_yObs[_ivar].at(_iobs) - _yExp[_ivar].at(_iobs) );        
        }
        res[0] += sqrt(_varWeights[_ivar] * inter);

        if(_bdumpDetails)
            cout<<" sum is sqrt() and weighted by varaible weights ("<<_varWeights[_ivar]<<") => Res is so far "<<res[0]<<endl;        
    }
    return 1.0;
}


// -----------------------------------------------------------------------
//  WEIGHTED LEAST SQUARE
// -----------------------------------------------------------------------
TWeightedLSDistanceFunction::TWeightedLSDistanceFunction(URANIE::DataServer::TDataServer *tdsEval, URANIE::DataServer::TDataServer *tdsObs, const char *reference, const char *output, const char *weight) :
    TDistanceFunction(tdsEval, tdsObs, reference,  output, weight)
{

    try{
        string sweight=weight;
        vector<string> weiName;
        URANIE::DataServer::UEntry::StringParser(sweight,":", weiName);
        
        if( sweight =="" )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       "TWeightedLSDistanceFunction::constructor\n\
The distance function WeightedLS should be used with a weight. The string provided is empty here.");

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    
}

TWeightedLSDistanceFunction::~TWeightedLSDistanceFunction()
{
}

int TWeightedLSDistanceFunction::localeval(Double_t* res)
{

    res[0] = 0.0;
    double inter=0.0;

   if(_bdumpDetails)
        cout<<"Computing WeightedLS distance "<<endl;
    for(_ivar=0; _ivar<_nVar; _ivar++)
    {
        if(_bdumpDetails)
            cout<<"For variable "<<_ivar<<": sum set to  "<<inter<<endl;
        
        inter=0;
        for(_iobs=0; _iobs<_nObs; _iobs++)
        {
            if(_bdumpDetails)
                cout<<"  evt("<<_iobs<<"), sum += ("<<_yObs[_ivar].at(_iobs)<<" - "<<_yExp[_ivar].at(_iobs)<<" )^2 * "<<_obsWeights.at(_ivar).at(_iobs)<<" = "<<inter<<endl;
            inter +=  (_yObs[_ivar].at(_iobs) - _yExp[_ivar].at(_iobs) )*(_yObs[_ivar].at(_iobs) - _yExp[_ivar].at(_iobs) ) * _obsWeights.at(_ivar).at(_iobs);            
        }
        res[0] += sqrt(_varWeights[_ivar] * inter);

        if(_bdumpDetails)
            cout<<" sum is sqrt() and weighted by varaible weights ("<<_varWeights[_ivar]<<") => Res is so far "<<res[0]<<endl;        
    }

    return 1.0;
}


// -----------------------------------------------------------------------
//  RELATIVE LEAST SQUARE
// -----------------------------------------------------------------------
TRelativeLSDistanceFunction::TRelativeLSDistanceFunction(URANIE::DataServer::TDataServer *tdsEval, URANIE::DataServer::TDataServer *tdsObs, const char *reference, const char *output, const char *weight) :
    TDistanceFunction(tdsEval, tdsObs, reference,  output, weight)
{
}

TRelativeLSDistanceFunction::~TRelativeLSDistanceFunction()
{
}

int TRelativeLSDistanceFunction::localeval(Double_t* res)
{
    res[0] = 0.0;
    double inter=0.0;

    if(_bdumpDetails)
        cout<<"Computing RelativeLS distance "<<endl;
    for(_ivar=0; _ivar<_nVar; _ivar++)
    {
        if(_bdumpDetails)
            cout<<"For variable "<<_ivar<<": sum set to  "<<inter<<endl;
        
        inter=0;
        for(_iobs=0; _iobs<_nObs; _iobs++)
        {
            if(_bdumpDetails)
                cout<<"  evt("<<_iobs<<"), sum += (("<<_yObs[_ivar].at(_iobs)<<" - "<<_yExp[_ivar].at(_iobs)<<") / "<<_yExp[_ivar].at(_iobs)<<")^2 = "<<inter<<endl;
            inter +=  (_yObs[_ivar].at(_iobs) - _yExp[_ivar].at(_iobs) )*(_yObs[_ivar].at(_iobs) - _yExp[_ivar].at(_iobs) ) / (_yObs[_ivar].at(_iobs)*_yObs[_ivar].at(_iobs));
        }
        res[0] += sqrt(_varWeights[_ivar] * inter);

        if(_bdumpDetails)
            cout<<" sum is sqrt() and weighted by varaible weights ("<<_varWeights[_ivar]<<") => Res is so far "<<res[0]<<endl;        
    }

    return 1.0;
}


// -----------------------------------------------------------------------
//  L1 NORM
// -----------------------------------------------------------------------
TL1DistanceFunction::TL1DistanceFunction(URANIE::DataServer::TDataServer *tdsEval, URANIE::DataServer::TDataServer *tdsObs, const char *reference, const char *output, const char *weight) :
    TDistanceFunction(tdsEval, tdsObs, reference,  output, weight)
{
}

TL1DistanceFunction::~TL1DistanceFunction()
{
}

int TL1DistanceFunction::localeval(Double_t* res)
{
    res[0] = 0.0;
    double inter=0;

    if(_bdumpDetails)
        cout<<"Computing L1 distance "<<endl;
    for(_ivar=0; _ivar<_nVar; _ivar++)
    {
        if(_bdumpDetails)
            cout<<"For variable "<<_ivar<<": sum set to  "<<inter<<endl;
        
        inter=0;
        for(_iobs=0; _iobs<_nObs; _iobs++)
        {
            if(_bdumpDetails)
                cout<<"  evt("<<_iobs<<"), sum += |"<<_yObs[_ivar].at(_iobs)<<" - "<<_yExp[_ivar].at(_iobs)<<"|  = "<<inter<<endl;
            inter +=  fabs(_yObs[_ivar].at(_iobs) - _yExp[_ivar].at(_iobs) );
        }
        res[0] += _varWeights[_ivar] * inter;

        if(_bdumpDetails)
            cout<<" sum is sqrt() and weighted by varaible weights ("<<_varWeights[_ivar]<<") => Res is so far "<<res[0]<<endl;        
    }
    
    return 1.0;
}


// -----------------------------------------------------------------------
//  MAHALANOBIS
// -----------------------------------------------------------------------
TMahalanobisDistanceFunction::TMahalanobisDistanceFunction(URANIE::DataServer::TDataServer *tdsEval, URANIE::DataServer::TDataServer *tdsObs, const char *reference, const char *output, const char *weight) :
    TDistanceFunction(tdsEval, tdsObs, reference,  output, weight)
{
    setUseMatrix(true);   
}

TMahalanobisDistanceFunction::~TMahalanobisDistanceFunction()
{
}

int TMahalanobisDistanceFunction::localeval(Double_t* res)
{

    res[0] = 0.0;
    _myExp -= _myObs;    
    for(_ivar=0; _ivar<_nVar; _ivar++)
    {
        TMatrixD Res(_myExp.GetSub(_ivar,_ivar,0,_nObs-1) *_mObsCovMat , TMatrixD::kMultTranspose, _myExp.GetSub(_ivar,_ivar,0,_nObs-1));        
        res[0] += sqrt(_varWeights[_ivar] * Res(0,0));
    }
    
    return 1.0;
}

} // Fin du namespace Calibration
} // Fin du namespace URANIE


