//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTLSDistanceFunction.cxx
 \author Jean-Baptiste Blanchard
 \date   2020/08/27

 \brief  test file for Linear bayesian classes

 */

// URANIE
#include "UExceptions.h"
#include "TestTLSDistanceFunction.h"

// CPPUNIT
#include <cppunit/extensions/HelperMacros.h>
#include <cppunit/Test.h>

#define DBL_PRECISION_FINE 1.e-5
#define DBL_PRECISION_LARGE 1.e-2
#define Flow1DFile "Ex2DoE_n100_sd1.75.dat"

#define DEL(x) if(x!=NULL){delete x; x=NULL;} 

// Constructor Destructor
TLSDistanceFunctionTest::TLSDistanceFunctionTest():
    _tdsEval(NULL), _tdsRef(NULL), _tdf(NULL)
{

}

TLSDistanceFunctionTest::~TLSDistanceFunctionTest()
{
    tearDown();
}

// pre and post treatment
void TLSDistanceFunctionTest::setUp()
{
    _tdsEval = new TDataServer("tdsEval","pouet");
    
    _tdsRef = new TDataServer("tdsRef","pouet");
}

void TLSDistanceFunctionTest::tearDown()
{  
    DEL(_tdsRef);
    DEL(_tdsEval);
    DEL(_tdf);
}

void TLSDistanceFunctionTest::throwExceptionConstructor()
{
    // wrong tdseval
    CPPUNIT_ASSERT_THROW(new TLSDistanceFunction(NULL, NULL, "", "", ""), URANIE::Exceptions::UErrorExceptions);

    // empty tdseval
    CPPUNIT_ASSERT_THROW(new TLSDistanceFunction(_tdsEval, NULL, "", "", ""), URANIE::Exceptions::UErrorExceptions);

    _tdsEval->fileDataRead(Flow1DFile);
    // wrong tdsobs
    CPPUNIT_ASSERT_THROW(new TLSDistanceFunction(_tdsEval, NULL, "", "", ""), URANIE::Exceptions::UErrorExceptions);

    // empty tdsobs
    CPPUNIT_ASSERT_THROW(new TLSDistanceFunction(_tdsEval, _tdsRef, "", "", ""), URANIE::Exceptions::UErrorExceptions);

    _tdsRef->fileDataRead(Flow1DFile);
    // empty ref
    CPPUNIT_ASSERT_THROW(new TLSDistanceFunction(_tdsEval, _tdsRef, "", "", ""), URANIE::Exceptions::UErrorExceptions);

    // wrong ref
    CPPUNIT_ASSERT_THROW(new TLSDistanceFunction(_tdsEval, _tdsRef, "QexP", "out", ""), URANIE::Exceptions::UErrorExceptions);

    // empty out
    CPPUNIT_ASSERT_THROW(new TLSDistanceFunction(_tdsEval, _tdsRef, "Qexp", "", ""), URANIE::Exceptions::UErrorExceptions);

    // wrong numbers
    CPPUNIT_ASSERT_THROW(new TLSDistanceFunction(_tdsEval, _tdsRef, "Qexp", "out:foo", ""), URANIE::Exceptions::UErrorExceptions);
    
    // wrong numbers
    CPPUNIT_ASSERT_THROW(new TLSDistanceFunction(_tdsEval, _tdsRef, "Qexp:rw", "out", ""), URANIE::Exceptions::UErrorExceptions);
}

void TLSDistanceFunctionTest::throwExceptionSetObservationCovMatrix()
{

    _tdsEval->fileDataRead(Flow1DFile);
    _tdsRef->fileDataRead(Flow1DFile);
      
    // a distancefunction
    _tdf = new TLSDistanceFunction(_tdsEval, _tdsRef, "Qexp", "out", "");
    
    TMatrixD foo(100,99);
    CPPUNIT_ASSERT_THROW( _tdf->setObservationCovarianceMatrix(foo), URANIE::Exceptions::UErrorExceptions);

    foo.T();
    CPPUNIT_ASSERT_THROW( _tdf->setObservationCovarianceMatrix(foo), URANIE::Exceptions::UErrorExceptions);
}

void TLSDistanceFunctionTest::throwExceptionSetVarWeights()
{

    _tdsEval->fileDataRead(Flow1DFile);
    _tdsRef->fileDataRead(Flow1DFile);
      
    double tutu[2]={1.0, 2.0};
    _tdf = new TLSDistanceFunction(_tdsEval, _tdsRef, "Qexp", "out", "");
    CPPUNIT_ASSERT_THROW( _tdf->setVarWeights(2,tutu), URANIE::Exceptions::UErrorExceptions);

    vector<double> foo={1.0, 2.0};
    CPPUNIT_ASSERT_THROW( _tdf->setVarWeights(foo), URANIE::Exceptions::UErrorExceptions);
}

void TLSDistanceFunctionTest::throwExceptionSetGaussianRandomNoise()
{

    _tdsEval->fileDataRead(Flow1DFile);
    _tdsRef->fileDataRead(Flow1DFile);
    _tdsRef->addAttribute("foo","1");
    _tdsRef->addAttribute("par","1");
      
    _tdf = new TLSDistanceFunction(_tdsEval, _tdsRef, "Qexp", "out", "rw");
    // Wrong name
    CPPUNIT_ASSERT_THROW( _tdf->setGaussianRandomNoise("tutu"), URANIE::Exceptions::UErrorExceptions);
    // Wrong number
    CPPUNIT_ASSERT_THROW( _tdf->setGaussianRandomNoise("foo:par"), URANIE::Exceptions::UErrorExceptions);

}

void TLSDistanceFunctionTest::throwExceptionGetObservation()
{

    _tdsEval->fileDataRead(Flow1DFile);
    _tdsRef->fileDataRead(Flow1DFile);
      
    _tdf = new TLSDistanceFunction(_tdsEval, _tdsRef, "Qexp", "out", "");

    vector<vector<double> > tutu;
    TMatrixD foo;
    CPPUNIT_ASSERT_THROW( _tdf->getObservationMatrix(), URANIE::Exceptions::UErrorExceptions);

    _tdf->setUseMatrix(true);
    CPPUNIT_ASSERT_THROW( _tdf->getObservationVector(), URANIE::Exceptions::UErrorExceptions);
}

