//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
  \file   TestTMetropHasting.cxx
  \author Jean-Baptiste Blanchard
  \date   2020/08/27

  \brief  test file for Linear bayesian classes

*/

// ROOT
#include "TROOT.h"
#include "TSystem.h"

// URANIE
#include "UExceptions.h"
#include "TestTMetropHasting.h"
#include "TUniformDistribution.h"
#include "TestUserFunctions.h"
#include "code_paths.h"
#include "uranie_commands.h"

// CPPUNIT
#include <cppunit/extensions/HelperMacros.h>
#include <cppunit/Test.h>

#define DBL_PRECISION_FINE 1.e-5
#define DBL_PRECISION_LARGE 1.e-2
#define Flow1DFile "Ex2DoE_n100_sd1.75.dat"
#define DEL(x) if(x!=NULL){delete x; x=NULL;} 

// Constructor Destructor
TMetropHastingTest::TMetropHastingTest():
    _tds(NULL),_tdsRef(NULL), _tmh(NULL), _seqRun(NULL), _thrRun(NULL), _evalJit(NULL), _evalCode(NULL), _code(NULL),
    _outRow(NULL), _inp(NULL), _outp(NULL), _norm(NULL), _out(NULL), _threaded(false)
{
    _workingTestPath = gSystem->pwd();
}

TMetropHastingTest::~TMetropHastingTest()
{
    tearDown();
}


void TMetropHastingTest::cleanWorkingDirectories()
{
    UranieTests::removeFile(_workingTestPath.Data(),"_dataserver_.root");
    UranieTests::removeFile(_workingTestPath.Data(),"Ex2DoE_n100_sd1.root");
#ifndef WIN32
    UranieTests::removeFolder(_workingTestPath.Data(), "URA_*");
    UranieTests::removeFolder(_workingTestPath.Data(), "URANIE*");
#endif

}

// pre and post treatment
void TMetropHastingTest::setUp()
{
    _tds = new TDataServer("tds","pouet");
    _norm = new TNormalDistribution("norm", 0., 1.0);
    
    _out=new TAttribute("out");

    _tdsRef = new TDataServer("tdsRef","pouet");
}

void TMetropHastingTest::setUpException()
{
    _evalJit = new TCJitEval(EmptyFunction);
    _evalJit->addInput(_norm);
    _evalJit->addOutput(_out);

    _seqRun = new TSequentialRun(_evalJit);
    
}

void TMetropHastingTest::setUpFlow1D(string type)
{

    _tdsRef->fileDataRead(Flow1DFile);
    _tdsRef->addAttribute("wei_exp","1./(sd_eps*sd_eps)");
        
    // define the parameters
    _tds->addAttribute( new URANIE::DataServer::TUniformDistribution("hl", 700, 760) );
    _tds->getAttribute( "hl" )->delShare();

    if(type=="TCJitEval")
    {
        // Create interface to assessors
        _evalJit= new TCJitEval(flowrateModel);
        // Add inputs in the correct order
        _evalJit->addInput(_tds->getAttribute("hl"));
        _evalJit->addInput(_tdsRef->getAttribute("rw"));
        _evalJit->addInput(_tdsRef->getAttribute("l"));
    
        //Add the output: the cost value 
        _evalJit->addOutput(_out);

        _seqRun = new TSequentialRun(_evalJit);    
    }
    else if(type=="TCodeEval")
    {
        // Create interface to assessors
        _evalCode= new TCodeEval((UranieTests::code_path("flowrate") + " -s -k").data());

        // Add inputs in the correct orde
        _inp = new TKeyScript("flowrate_input_with_keys.in");
        _inp->addInput(_tds->getAttribute("hl"),"Hl");    
        _inp->addInput(_tdsRef->getAttribute("rw"),"Rw");
        _inp->addInput(_tdsRef->getAttribute("l"),"L");

        // Add inputs in the correct orde
        _outp = new TKeyResult("_output_flowrate_withKey_.dat");
        _outp->addOutput(_out, "yhat" );    
    
        //Add the output: the cost value 
        _evalCode->addInputFile(_inp);
        _evalCode->addOutputFile(_outp);

        if( ! _threaded )
            _seqRun = new TSequentialRun(_evalCode);
        else
            _thrRun = new TThreadedRun(_evalCode,4);    
    }
    else if(type=="TCode")
    { 
        TString sIn = TString("flowrate_input_with_keys.in");

        // Set the reference input file and the key for each input attributes
        _tdsRef->getAttribute("rw")->setFileKey(sIn, "Rw");
        _tdsRef->getAttribute("l")->setFileKey(sIn, "L");
        _tds->getAttribute("hl")->setFileKey(sIn, "Hl");
 
        // The output file of the code
        _outRow = new TOutputFileRow("_output_flowrate_withRow_.dat");
        // The attribute in the output file
        _outRow->addAttribute(_out);
    
        // Instanciation de mon code
        _code= new TCode(_tdsRef, UranieTests::code_path("flowrate") + " -s -k");
        // on ajoute le fichier de sortie du code
        _code->addOutputFile( _outRow );

    }
}

void TMetropHastingTest::tearDown()
{
    DEL(_evalJit);
    DEL(_evalCode);
    DEL(_code);
    DEL(_outRow);
    DEL(_inp);
    DEL(_outp);
    DEL(_seqRun);
    DEL(_thrRun);
    DEL(_tmh);
    DEL(_tdsRef);
    DEL(_tds);
    DEL(_norm);
    DEL(_out);
    _threaded=false;

    cleanWorkingDirectories();
}

void TMetropHastingTest::runFlow1DCJitNoPlots()
{
    setUpFlow1D("TCJitEval");

    _tmh = new URANIE::Calibration::TMetropHasting(_tds, _seqRun, 2000, "");
    _tmh->setDistanceAndReference("weightedLS",_tdsRef,"rw:l","Qexp","wei_exp");
    _tmh->setSeed(132456);
    _tmh->setAcceptationRatioRange(0.4, 0.45);
    _tmh->estimateParameters();

    int burn=0;
    vector<int> lag={1,5,10,20,50,100,200};
    vector<double> autoCorr;
    _tmh->getAutoCorrelation(lag, &autoCorr, burn);
    
    cout<<"Autocorrelation are "<<autoCorr.size()<<":"<<endl;
    for(unsigned il=0; il<lag.size(); il++)
    {
        cout<<"*** for lag="<<lag.at(il)<<": ";
        for(int ip=0; ip<_tmh->getNPar(); ip++)            
            cout<<autoCorr.at(ip*lag.size()+il)<<";  ";
        cout<<endl;
    }


    _tds->computeStatistic();
    
    int nb=_tds->getTuple()->Draw("hl","","goff");
    cout<<std::setprecision(12)<<_tds->getAttribute("hl")->getMean()<<"  "<<_tds->getAttribute("hl")->getStd()<<endl;
    CPPUNIT_ASSERT_EQUAL(2000, nb);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(749.633353726, _tds->getAttribute("hl")->getMean(), DBL_PRECISION_FINE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.01401175762, _tds->getAttribute("hl")->getStd(), DBL_PRECISION_FINE);
}

void TMetropHastingTest::runFlow1DCJitNoPlotsVerySmall()
{
    setUpFlow1D("TCJitEval");

    _tmh = new URANIE::Calibration::TMetropHasting(_tds, _seqRun, 15, "");
    _tmh->setDistanceAndReference("weightedLS",_tdsRef,"rw:l","Qexp","wei_exp");
    _tmh->setNbDump(1);
    _tmh->setSeed(132456);
    _tmh->setAcceptationRatioRange(0.4, 0.45);
    _tmh->estimateParameters();

    int burn=0;
    vector<int> lag={1,5,10,20};
    vector<double> autoCorr;
    _tmh->getAutoCorrelation(lag, &autoCorr, burn);
    
    cout<<"Autocorrelation are "<<autoCorr.size()<<":"<<endl;
    for(unsigned il=0; il<lag.size(); il++)
    {
        cout<<"*** for lag="<<lag.at(il)<<": ";
        for(int ip=0; ip<_tmh->getNPar(); ip++)            
            cout<<autoCorr.at(ip*lag.size()+il)<<";  ";
        cout<<endl;
    }

    _tds->computeStatistic("hl", Form("%s > 5", _tds->getIteratorName()));
    
    int nb=_tds->getTuple()->Draw("hl","","goff");
    cout<<std::setprecision(12)<<_tds->getAttribute("hl")->getMean()<<"  "<<_tds->getAttribute("hl")->getStd()<<endl;
    CPPUNIT_ASSERT_EQUAL(15, nb);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(750.246075930416, _tds->getAttribute("hl")->getMean(), DBL_PRECISION_FINE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.91803084546647, _tds->getAttribute("hl")->getStd(), DBL_PRECISION_FINE);
}

void TMetropHastingTest::runFlow1DNoPlotsVerySmallCodeEval()
{
    setUpFlow1D("TCodeEval");

    _tmh = new URANIE::Calibration::TMetropHasting(_tds, _seqRun, 15, "");
    _tmh->setDistanceAndReference("weightedLS",_tdsRef,"rw:l","Qexp","wei_exp");
    _tmh->setSeed(132456);
    _tmh->setNbDump(1);
    _tmh->setAcceptationRatioRange(0.4, 0.45);
    _tmh->estimateParameters();

    int burn=0;
    vector<int> lag={1,5,10,20};
    vector<double> autoCorr;
    _tmh->getAutoCorrelation(lag, &autoCorr, burn);
    
    cout<<"Autocorrelation are "<<autoCorr.size()<<":"<<endl;
    for(unsigned il=0; il<lag.size(); il++)
    {
        cout<<"*** for lag="<<lag.at(il)<<": ";
        for(int ip=0; ip<_tmh->getNPar(); ip++)            
            cout<<autoCorr.at(ip*lag.size()+il)<<";  ";
        cout<<endl;
    }


    _tds->computeStatistic("hl", Form("%s > 5", _tds->getIteratorName()));
    
    int nb=_tds->getTuple()->Draw("hl","","goff");
    cout<<std::setprecision(12)<<_tds->getAttribute("hl")->getMean()<<"  "<<_tds->getAttribute("hl")->getStd()<<endl;
    CPPUNIT_ASSERT_EQUAL(15, nb);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(750.246075930416, _tds->getAttribute("hl")->getMean(), DBL_PRECISION_FINE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.91803084546647, _tds->getAttribute("hl")->getStd(), DBL_PRECISION_FINE);
}

void TMetropHastingTest::runFlow1DNoPlotsVerySmallCodeEvalThreaded()
{
    _threaded=true;
    setUpFlow1D("TCodeEval");

    _tmh = new URANIE::Calibration::TMetropHasting(_tds, _thrRun, 15, "");
    _tmh->setDistanceAndReference("weightedLS",_tdsRef,"rw:l","Qexp","wei_exp");
    _tmh->setNbDump(1);
    _tmh->setSeed(132456);
    _tmh->setAcceptationRatioRange(0.4, 0.45);
    _tmh->estimateParameters();

    int burn=0;
    vector<int> lag={1,5,10,20};
    vector<double> autoCorr;
    _tmh->getAutoCorrelation(lag, &autoCorr, burn);
    
    cout<<"Autocorrelation are "<<autoCorr.size()<<":"<<endl;
    for(unsigned il=0; il<lag.size(); il++)
    {
        cout<<"*** for lag="<<lag.at(il)<<": ";
        for(int ip=0; ip<_tmh->getNPar(); ip++)            
            cout<<autoCorr.at(ip*lag.size()+il)<<";  ";
        cout<<endl;
    }


    _tds->computeStatistic("hl", Form("%s > 5", _tds->getIteratorName()));
    
    int nb=_tds->getTuple()->Draw("hl","","goff");
    cout<<std::setprecision(12)<<_tds->getAttribute("hl")->getMean()<<"  "<<_tds->getAttribute("hl")->getStd()<<endl;
    CPPUNIT_ASSERT_EQUAL(15, nb);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(750.246075930416, _tds->getAttribute("hl")->getMean(), DBL_PRECISION_FINE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.91803084546647, _tds->getAttribute("hl")->getStd(), DBL_PRECISION_FINE);
}

void TMetropHastingTest::runFlow1DNoPlotsVerySmallCode()
{
    setUpFlow1D("TCode");

    _tmh = new URANIE::Calibration::TMetropHasting(_tds, _code, 15, "");
    _tmh->setDistanceAndReference("weightedLS",_tdsRef,"rw:l","Qexp","wei_exp");
    _tmh->setNbDump(1);
    _tmh->setSeed(132456);
    _tmh->setAcceptationRatioRange(0.4, 0.45);
    _tmh->estimateParameters();

    int burn=0;
    vector<int> lag={1,5,10,20};
    vector<double> autoCorr;
    _tmh->getAutoCorrelation(lag, &autoCorr, burn);
    
    cout<<"Autocorrelation are "<<autoCorr.size()<<":"<<endl;
    for(unsigned il=0; il<lag.size(); il++)
    {
        cout<<"*** for lag="<<lag.at(il)<<": ";
        for(int ip=0; ip<_tmh->getNPar(); ip++)            
            cout<<autoCorr.at(ip*lag.size()+il)<<";  ";
        cout<<endl;
    }


    _tds->computeStatistic("hl", Form("%s > 5", _tds->getIteratorName()));
    
    int nb=_tds->getTuple()->Draw("hl","","goff");
    cout<<std::setprecision(12)<<_tds->getAttribute("hl")->getMean()<<"  "<<_tds->getAttribute("hl")->getStd()<<endl;
    CPPUNIT_ASSERT_EQUAL(15, nb);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(750.246075930416, _tds->getAttribute("hl")->getMean(), DBL_PRECISION_FINE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.91803084546647, _tds->getAttribute("hl")->getStd(), DBL_PRECISION_FINE);
}

void TMetropHastingTest::runFlow1DPointerFunctionNoPlots()
{
    setUpFlow1D("Function");

    TString inFile="myflow.C";
    gROOT->LoadMacro( gSystem->PrependPathName(gSystem->pwd(), inFile) );
    
    _tmh = new URANIE::Calibration::TMetropHasting(_tds, "flowrateModel" , "hl:rw:l", "out",2000, "");
    _tmh->setDistanceAndReference("weightedLS",_tdsRef,"rw:l","Qexp","wei_exp");
    _tmh->setSeed(132456);
    _tmh->setAcceptationRatioRange(0.4, 0.45);
    _tmh->estimateParameters();

    int burn=0;
    vector<int> lag={1,5,10,20,50,50,200};
    vector<double> autoCorr;
    _tmh->getAutoCorrelation(lag, &autoCorr, burn);
    
    cout<<"Autocorrelation are "<<autoCorr.size()<<":"<<endl;
    for(unsigned il=0; il<lag.size(); il++)
    {
        cout<<"*** for lag="<<lag.at(il)<<": ";
        for(int ip=0; ip<_tmh->getNPar(); ip++)            
            cout<<autoCorr.at(ip*lag.size()+il)<<";  ";
        cout<<endl;
    }


    _tds->computeStatistic();
    
    int nb=_tds->getTuple()->Draw("hl","","goff");
    cout<<std::setprecision(12)<<_tds->getAttribute("hl")->getMean()<<"  "<<_tds->getAttribute("hl")->getStd()<<endl;
    CPPUNIT_ASSERT_EQUAL(2000, nb);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(749.633353726, _tds->getAttribute("hl")->getMean(), DBL_PRECISION_FINE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.01401175762, _tds->getAttribute("hl")->getStd(), DBL_PRECISION_FINE);
}

void TMetropHastingTest::runFlow1DFunctionNoPlots()
{
    setUpFlow1D("Function");

    _tmh = new URANIE::Calibration::TMetropHasting(_tds, flowrateModel , "hl:rw:l", "out",2000, "");
    _tmh->setDistanceAndReference("weightedLS",_tdsRef,"rw:l","Qexp","wei_exp");
    _tmh->setSeed(132456);
    _tmh->setAcceptationRatioRange(0.4, 0.45);
    _tmh->estimateParameters();

    int burn=0;
    vector<int> lag={1,5,10,20,50,100,200};
    vector<double> autoCorr;
    _tmh->getAutoCorrelation(lag, &autoCorr, burn);
    
    cout<<"Autocorrelation are "<<autoCorr.size()<<":"<<endl;
    for(unsigned il=0; il<lag.size(); il++)
    {
        cout<<"*** for lag="<<lag.at(il)<<": ";
        for(int ip=0; ip<_tmh->getNPar(); ip++)            
            cout<<autoCorr.at(ip*lag.size()+il)<<";  ";
        cout<<endl;
    }


    _tds->computeStatistic();
    
    int nb=_tds->getTuple()->Draw("hl","","goff");

    cout<<std::setprecision(12)<<_tds->getAttribute("hl")->getMean()<<"  "<<_tds->getAttribute("hl")->getStd()<<endl;
    CPPUNIT_ASSERT_EQUAL(2000, nb);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(749.633353726, _tds->getAttribute("hl")->getMean(), DBL_PRECISION_FINE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.01401175762, _tds->getAttribute("hl")->getStd(), DBL_PRECISION_FINE);
}

void TMetropHastingTest::throwExceptionConstruction()
{
    setUpException();
    
    // no tds
    CPPUNIT_ASSERT_THROW(new TMetropHasting(NULL, (URANIE::Relauncher::TRun*)NULL, 10), URANIE::Exceptions::UErrorExceptions);

    // tds without attributes
    CPPUNIT_ASSERT_THROW(new TMetropHasting(_tds, (URANIE::Relauncher::TRun*)NULL, 10), URANIE::Exceptions::UErrorExceptions);

    // empty runner
    CPPUNIT_ASSERT_THROW(new TMetropHasting(_tds, (URANIE::Relauncher::TRun*)NULL, 10), URANIE::Exceptions::UErrorExceptions);
    
    // wrong kind of attribute
    _tds->addAttribute("tutu");
    CPPUNIT_ASSERT_THROW(new TMetropHasting(_tds, _seqRun, 10), URANIE::Exceptions::UErrorExceptions);

    // wrong kind of laws
    CPPUNIT_ASSERT_THROW(new TMetropHasting(_tds, (URANIE::Relauncher::TRun*)NULL, 10), URANIE::Exceptions::UErrorExceptions);

}


void TMetropHastingTest::throwExceptionInitialisation()
{
    setUpException();

    _tds->addAttribute(_norm);
    // wrong kind of attribute
    _tmh=new TMetropHasting(_tds, _seqRun, 10);

    double v[1]={10}, s[1]={-1};
    CPPUNIT_ASSERT_THROW(_tmh->setInitialisation(1,v,s), URANIE::Exceptions::UErrorExceptions);

    double v2[2]={10,20}, s2[2]={1,2};
    CPPUNIT_ASSERT_THROW(_tmh->setInitialisation(2,v2,s2), URANIE::Exceptions::UErrorExceptions);
    
    vector<double> mean={10}, std{1.,2.};
    CPPUNIT_ASSERT_THROW(_tmh->setInitialisation(mean,std), URANIE::Exceptions::UErrorExceptions);
}

void TMetropHastingTest::throwExceptionAccRange()
{
    setUpFlow1D("TCJitEval");

    _tmh = new URANIE::Calibration::TMetropHasting(_tds, _seqRun, 2000, "");
    _tmh->setDistanceAndReference("weightedLS",_tdsRef,"rw:l","Qexp","wei_exp");  
    
    // low > hig
    CPPUNIT_ASSERT_THROW(_tmh->setAcceptationRatioRange(0.5,0.4), URANIE::Exceptions::UErrorExceptions);
    // low > 1.0
    CPPUNIT_ASSERT_THROW(_tmh->setAcceptationRatioRange(1.5,2.4), URANIE::Exceptions::UErrorExceptions);
    // low < 0.0
    CPPUNIT_ASSERT_THROW(_tmh->setAcceptationRatioRange(-0.5,0.4), URANIE::Exceptions::UErrorExceptions);
    // hig > 1.0
    CPPUNIT_ASSERT_THROW(_tmh->setAcceptationRatioRange(0.5,1.4), URANIE::Exceptions::UErrorExceptions);
    // hig > 1.0
    CPPUNIT_ASSERT_THROW(_tmh->setAcceptationRatioRange(0.5,-0.4), URANIE::Exceptions::UErrorExceptions);
}

void TMetropHastingTest::throwExceptionReference()
{
    setUpException();

    _tds->addAttribute(_norm);
    // wrong kind of attribute
    _tmh=new TMetropHasting(_tds, _seqRun, 10);

    // no tdsref
    CPPUNIT_ASSERT_THROW(_tmh->setDistanceAndReference("toutou",NULL,"piou","piou","piou"), URANIE::Exceptions::UErrorExceptions);

    // no tdsref empty
    CPPUNIT_ASSERT_THROW(_tmh->setDistanceAndReference("toutou",_tdsRef,"piou","piou","piou"), URANIE::Exceptions::UErrorExceptions);

    _tdsRef->fileDataRead(Flow1DFile);
    // wrong input
    CPPUNIT_ASSERT_THROW(_tmh->setDistanceAndReference("LS",_tdsRef,"rw:lout","Qexp","piou"), URANIE::Exceptions::UErrorExceptions);
    // wrong reference
    CPPUNIT_ASSERT_THROW(_tmh->setDistanceAndReference("LS",_tdsRef,"rw:l","Qexpmero","piou"), URANIE::Exceptions::UErrorExceptions);
    // wrong weight
    CPPUNIT_ASSERT_THROW(_tmh->setDistanceAndReference("LS",_tdsRef,"rw:l","Qexp","piou"), URANIE::Exceptions::UErrorExceptions);
    // wrong reference number
    CPPUNIT_ASSERT_THROW(_tmh->setDistanceAndReference("LS",_tdsRef,"rw:l","Qexp:sd_eps","sd_eps"), URANIE::Exceptions::UErrorExceptions);
    
    // wrong function name
    CPPUNIT_ASSERT_THROW(_tmh->setDistanceAndReference("toutou", _tdsRef, "rw:l", "Qexp", "bisou"), URANIE::Exceptions::UErrorExceptions);

}
