// --- C++ ---
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2019 CEA/DEN
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   TEGO.cxx
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2019-10-22
 
 \brief  efficient global optimization algorithm
*/


#include "TEGO.h"

#include <functional>
// ROOT
#include "TLeaf.h"

// Uranie
#include "TAttribute.h"
#include "TSampling.h"
#include "TCJitEval.h"
#include "TComposeEval.h"
#include "TSequentialRun.h"
#include "TGPBuilder.h"
#include "TVizirGenetic.h"

#include "TEgoKBModeler.h"
#include "TEgoHjDynSolver.h"
//#include "TEgoStdSolver.h"

#define _USE_MATH_DEFINES
#include <math.h>

//#define ALPHA 0.2
#define Nmore 3
#define DOEDONE
//#define NOLAMBDA

ClassImp(URANIE::MetaModelOptim::TEGO);
ClassImp(URANIE::MetaModelOptim::TEgoModeler);
ClassImp(URANIE::MetaModelOptim::TEgoSolver);

#ifdef NOLAMBDA
/*** fitness functions ***/ 
/* expected improvement */
int URANIE::MetaModelOptim::egocrit(double* in, double *out, void *info)
{
    double fmin, sigma, a, ei;

    fmin = *((double *) info);
    //std::cout << "fmin " << fmin << std::endl;
    if (in[1] < 1e-40) ei = (fmin > in[0])? fmin - in[0] : 0. ;
    else {
        sigma = sqrt(in[1]) ;
        a = (fmin - in[0]) / sigma ;
        ei = sigma * ( a * (1. + erf(a / sqrt(2.))) / 2.  +  // sigma * (a * PDF(a) +
                1/sqrt(2*M_PI) * exp(-0.5*a*a) ); //  CDF(a))
    }

    out[0] = ei ;
    return 1;
}

/* metamodel */
int URANIE::MetaModelOptim::krigeage(double *in, double *out, void *info)
{
    URANIE::Modeler::TKriging *mmf = *((URANIE::Modeler::TKriging **) info);
    URANIE::Modeler::TKriging *mm = (URANIE::Modeler::TKriging *) mmf->rebuild(0);
    int ret = mm->eval(in, out, 0);
    if (mm->unbuild()) delete mm;
    return ret;
}
#endif

/*** CLASS ***/
/** constructeur **/
URANIE::MetaModelOptim::TEGO::TEGO(URANIE::DataServer::TDataServer *tds, URANIE::Relauncher::TRun *run) :
    TOptimShare(tds, run), _crit(NULL), _solv(NULL), _model(NULL)
{
    // default
    _NK = 50;
    _NC = 500;
}

URANIE::MetaModelOptim::TEGO::~TEGO()
{
    if (_crit) delete _crit;
    //if (_solv) delete _solv;
}

/* setup */
void URANIE::MetaModelOptim::TEGO::setSize(int first, int all)
{
    _NK = first;
    _NC = all;
    _NR = _run->_ncpus;
}

void URANIE::MetaModelOptim::TEGO::addObjective(URANIE::DataServer::TAttribute *obj, 
        URANIE::Reoptimizer::TOptimFit *fit)
{
    if (!_crit) {
        _y = obj;
        _crit = new URANIE::Reoptimizer::TFitGetter(obj, _fun->getOutputList(), fit);
    }
    else std::cout << "EGO is single objective\n";
}

void URANIE::MetaModelOptim::TEGO::setModeler(TEgoModeler *model)
{
    _model = model;
}

void URANIE::MetaModelOptim::TEGO::setSolver(TEgoSolver *algo)
{
    _solv = algo;
}

/** methodes utilitaires **/
/* copy d'un tds dans un tableau C
 ***/
int URANIE::MetaModelOptim::TEGO::copyTDSin(vector<URANIE::DataServer::TAttribute*>& in,
    URANIE::DataServer::TDataServer *tds, 
    vector<vector<URANIE::DataServer::UEntry*>> &sampl,
    vector<char> &mark)
{
    TList leaves;
    TLeaf *leaf;
    int i, j;

    for( i=0; in[i]; i++ ) {
        leaves.Add(tds->GetTuple()->GetLeaf(in[i]->GetName()));
    }

    TIter nextl(&leaves);
    for (i=0; i<_NR+_NK; i++) {
        *sampl[i][0]->val.dou = i; //index
        tds->GetTuple()->GetEntry(i);
        j = 1;
        while ((leaf = (TLeaf *) nextl())) {
            *sampl[i][j++]->val.dou = leaf->GetValue(0);
        }
        mark[i] = FILL;
        nextl.Reset();
    }

    return 1;
}

int URANIE::MetaModelOptim::TEGO::copyTDSall(URANIE::DataServer::TDataServer *tds, 
    vector<vector<URANIE::DataServer::UEntry*>> &sampl,
    vector<char> &mark)
{
    URANIE::DataServer::TAttribute *att;
    TList leaves, *tmplist;
    TLeaf *leaf;
    int i, j;

    tmplist = tds->getListOfAttributes("*");
    TIter next(tmplist);
    while ( (att = (URANIE::DataServer::TAttribute *) next()) ) {
        leaves.Add(tds->GetTuple()->GetLeaf(att->GetName()));
    }

    TIter nextl(&leaves);
    for (i=0; i<_NR+_NK; i++) {
        *sampl[i][0]->val.dou = i; //index
        tds->GetTuple()->GetEntry(i);
        j = 1;
        while ((leaf = (TLeaf *) nextl())) {
            sampl[i][j++]->assignEntry((void*)(leaf->GetValuePointer()));
        }
        mark[i] = EVALUATED;
        nextl.Reset();
    }

    delete tmplist;
    return 1;
}

/* sampling initial pour les NK premiers points nécéssaires pour le krigging
 ***/
int URANIE::MetaModelOptim::TEGO::init_sampling(vector<URANIE::DataServer::TAttribute*>& in,
    vector<vector<URANIE::DataServer::UEntry*>> &sampl, 
    vector<char> &mark)
{
    int i;

    URANIE::DataServer::TDataServer tds("tmptds", "tds initial sample");
    tds.keepFinalTuple(kFALSE);
    for (i=0; in[i] != NULL; i++) tds.addAttribute(in[i]);

    // sampling
    URANIE::Sampler::TSampling sampling(&tds, "lhs", _NK+_NR);
    sampling.generateSample();
    tds.exportData("lhs.dat");

    // copie of sampling part
    mark.resize(_NC);
    for (i=0; i<_NC; i++) mark[i] = EMPTY;
    sampl.resize(_NC);
    allocDataEntries(_NC, sampl);
    copyTDSin(in, &tds, sampl, mark);

    // sortie
    return 1;
}

int URANIE::MetaModelOptim::TEGO::reuse_sampling(vector<URANIE::DataServer::TAttribute*> &in,
    URANIE::DataServer::TDataServer *tds, 
    vector<vector<URANIE::DataServer::UEntry*>> &sampl, 
    vector<char> &mark, bool done)
{
    int i;

    _NK = tds->GetTuple()->GetEntries() - _NR;

    // copie of sampling part
    mark.resize(_NC);
    for (i=0; i<_NC; i++) mark[i] = EMPTY;
    sampl.resize(_NC);
    allocDataEntries(_NC, sampl);
    if (done) copyTDSall(tds, sampl, mark);
    else copyTDSin(in, tds, sampl, mark);

    // sortie
    return 1;
}

void URANIE::MetaModelOptim::TEGO::prepareInput(vector<URANIE::DataServer::UEntry*> *sampl, 
    vector<URANIE::DataServer::UEntry*> *in)
{
    vector<URANIE::DataServer::UEntry*> *tmpa =
        new vector<URANIE::DataServer::UEntry*>(sampl->begin()+1, sampl->end());
    vector<URANIE::DataServer::UEntry*> *tmpi =
        new vector<URANIE::DataServer::UEntry*>(in->begin(), in->end());

    inputForCode(tmpa, tmpi);
    delete tmpa;
}

void URANIE::MetaModelOptim::TEGO::fillSample(long ev, vector<URANIE::DataServer::UEntry*> *in,
    vector<URANIE::DataServer::UEntry*> *out,
    vector<URANIE::DataServer::UEntry*> *sampl)
{
    vector<URANIE::DataServer::UEntry*> *tmpi =
        new vector<URANIE::DataServer::UEntry*>(in->begin(), in->end());
    vector<URANIE::DataServer::UEntry*> *tmpo =
        new vector<URANIE::DataServer::UEntry*>(out->begin(), out->end());
    vector<URANIE::DataServer::UEntry*> *tmpa =
        new vector<URANIE::DataServer::UEntry*>(sampl->begin()+1, sampl->end());

    sampl->at(0)->pushEntry((double) ev);
    allFromCode(tmpi, tmpo, tmpa);
    _tds->getTuple()->Fill(sampl);

}

/*** protocole ***/
// solve
void URANIE::MetaModelOptim::TEGO::solverLoop()
{
    int i, ret;
    long j, jj, ev;
    //double ei;
    //double tmp[_NinDat+5]; // tmpout[2];
    // TRandom3 rnd;

    // liste des entrées
#ifdef OLD
    TList *lin = _tds->getListOfAttributes("*");
    TIter nexti(lin);
    URANIE::DataServer::TAttribute *in[lin->GetEntries()];
    i = 0;
    while ((in[i] = (URANIE::DataServer::TAttribute*) nexti())) i++;
#else
    if (_Naux != 0) exit(1);
    TList *lin = _tds->getListOfAttributes("*");
    TIter nexti(lin);
    TIter nextp(&_param);
    vector<URANIE::DataServer::TAttribute*> in(_NinDat+1);
    in[_NinDat] = nullptr;
    URANIE::Relauncher::TMasterVar *var;
    while ((var = (URANIE::Relauncher::TMasterVar*) nextp())) {
        if (var->isExported()) {
            i=0;
            while ( var->getRef() != (URANIE::DataServer::TAttribute *) nexti()) i++;
            in[i] = var->getRef();
            nexti.Reset();
        }
    }
    delete lin;
#endif
    // init tds
    fillDS(); 
    /** initial sampling **/
    vector<vector<URANIE::DataServer::UEntry*>> sampl;
    vector<char> mark;
    if (_tds->hasTuple()) {
        URANIE::Relauncher::TOutputVar *var;
        TIter next(&_sortie);
        bool filled = true;
        while ((var = (URANIE::Relauncher::TOutputVar*) next())) {
            if (!var->isPre()) filled = false;
        }
        reuse_sampling(in, _tds, sampl, mark, filled);
    }
    else {
        init_sampling(in, sampl, mark);
    }
    _tds->createTuple();

    /* ego local variables */
    auto yvar = new URANIE::DataServer::TAttribute("y_var");
    auto ego = new URANIE::DataServer::TAttribute("ego");
    
    // expected improvement
    double fmin = 0.0;
#ifdef NOLAMBDA
    URANIE::Relauncher::TCJitEval egofun(&egocrit, &fmin);
#else
    std::function<int(double*,double*)> egocrit = [&](double *in, double *out) {
        double sigma, a, ei;

        //std::cout << "fmin " << fmin << std::endl;
        if (in[1] < 1e-40) ei = (fmin > in[0])? fmin - in[0] : 0. ;
        else {
            sigma = sqrt(in[1]) ;
            a = (fmin - in[0]) / sigma ;
            ei = sigma * ( a * (1. + erf(a / sqrt(2.))) / 2.  +  // sigma * (a * PDF(a) +
                    1/sqrt(2*M_PI) * exp(-0.5*a*a) ); //  CDF(a))
        }

        out[0] = ei ;
        return 1;
    };
    URANIE::Relauncher::TCJitEval egofun(&egocrit);
#endif
    egofun.addInput(_y);
    egofun.addInput(yvar);
    egofun.addOutput(ego);

    /** métamodèle construction objects **/
    TEgoModeler *model = (_model) ? _model : new TEgoKBModeler();
    model->build(_tds, _crit, in , _y, yvar, &fmin, _NC);
    URANIE::Modeler::TKriging *krig2;
    
    /** ei search objects **/
    // metamodele indirection
#ifdef NOLAMBDA
    URANIE::Relauncher::TCJitEval dynkrig(&krigeage, &krig2);
#else
    std::function<int(double*, double*)> krigeage = [&krig2](double *in, double *out) {
        auto mm = (URANIE::Modeler::TKriging *) krig2->rebuild(0);
        int ret = mm->eval(in, out, 0);
        if (mm->unbuild()) delete mm;
        return ret;
    };
    URANIE::Relauncher::TCJitEval dynkrig( &krigeage);
#endif
    for (i=0; in[i] != NULL; i++) dynkrig.addInput(in[i]);
    dynkrig.addOutput(_y);
    dynkrig.addOutput(yvar);
    // eval
    URANIE::Relauncher::TComposeEval ceval;
    ceval.addEval(&dynkrig);
    ceval.addEval(&egofun);
    URANIE::Relauncher::TSequentialRun srun(&ceval);
    srun.startSlave();
    // optim
    TEgoSolver *solv = (_solv)? _solv : new TEgoHjDynSolver();
    //TEgoSolver *solv = (_solv)? _solv : new TEgoStdSolver();
    solv->build(&srun, in, ego);

    /*** running ***/ 
    // init
    long *index = new long[_NR];
    for (ev=0; ev<_NR; ev++) index[ev] = -1;
    vector<vector<URANIE::DataServer::UEntry *>> codin, codout;
    codin.resize(_NR);
    allocInCodeEntries(_NR, codin);
    codout.resize(_NR);
    allocOutCodeEntries(_NR, codout);
    
    int unused = _NR;
    for (ev=0; mark[ev] == EVALUATED; ev++) {
        _tds->getTuple()->Fill(&sampl[ev]);
    }
    // fill resources
    for (; ev<_NR; ev++) {
        index[ev] = ev;
        prepareInput(&sampl[ev], &codin[ev]);
        mark[ev] = PENDING;
        _run->put(ev, &codin[ev], &codout[ev]);
        unused--;
    }

    // sampling phase
    for (; ev<_NR+_NK; ev++) {
        ret = _run->get(j);
        jj = index[j];
        fillSample(jj, &codin[j], &codout[j], &sampl[jj]);
        mark[jj] = EVALUATED;
        for (auto it : codout[j]) it->clearEntry();

        index[j] = ev;
        prepareInput(&sampl[ev], &codin[j]);
        mark[ev] = PENDING;
        _run->put(j, &codin[j], &codout[j]);
    }
    
    //_tds->exportData("first.dat");
    
    /** searching phase **/
    //int error = 0;
    int toolow = 0;
    int dodoe = 0;
    int nb;
    // EGO searching phase loop
    for (; ev<_NC; ) {
        // metamodele
        krig2 = model->next(mark, sampl, dodoe);
        if (!krig2) {
            //error = 1;
            break;
        }

        // ego
        nb = 0;
        if (srun.onMaster()) {
            solv->next();
            std::cout << "more";
            for (int l=0; 
                    !unused && !_run->readyToGet() && l<Nmore; 
                    l++)
            {
                solv->more();
                std::cout << l ;
            }
            std::cout << std::endl;
            nb = solv->getNext(ev, _NC-ev-1, sampl, mark);
            if (!nb) toolow++;
        }
        
        // clean
        delete krig2;
        // prepare next
        //dodoe = 1;
        dodoe = (3*ev > 2*_NC && ev % 2) ? 2 : 1;
        if (toolow>4) break;
 
        // wait for next result
        while (_run->readyToGet() || unused < 1) {
            ret = _run->get(j);
            jj = index[j];
            fillSample(jj, &codin[j], &codout[j], &sampl[jj]);
            index[j] = -1;
            mark[jj] = EVALUATED;

            std::cout << "EGO get " << jj;
            for (i=0; i<_NoutDat; i++)  std::cout << " " << *sampl[jj][_NinDat+1+i]->val.dou;
            std::cout << std::endl;
            unused++;
        }
        std::cout << "unused " << unused << std::endl;

        nb = (nb > unused) ? unused : nb ;
        j = 0;
        for (int l=0; l<nb ; l++, ev++, unused--) {
            for (; index[j] != -1; j++);
            index[j] = ev;
            prepareInput(&sampl[ev], &codin[j]);
            mark[ev] = PENDING;
            _run->put(j, &codin[j], &codout[j]);
        }
    }

    // last running resource end loop
    for (i=unused; i<_NR; i++) {
        ret = _run->get(j);
        jj = index[j];
        fillSample(jj, &codin[j], &codout[j], &sampl[jj]);
        index[j] = -1;
        mark[jj] = EVALUATED;
        std::cout << "EGO get " << i << " " << jj << " " << *sampl[jj][_NinDat+1]->val.dou << std::endl;
    }
    // result
    for (i=0; i<ev; i++) {
        for (j=0; j<_NinDat+_NoutDat+1; j++) std::cout << *sampl[i][j]->val.dou << " ";
        std::cout << std::endl;
    }
    // cleanup
    if (!_model) delete model;
    if (!_solv) delete solv;
    for (i=0; i<_NR ; i++) {
        for (auto it : codin[i]) delete it;
        codin[i].clear();
        for (auto it : codout[i]) delete it;
        codout[i].clear();
    }
    delete[] index;
    for (i=0; i<_NC ; i++) {
        for (auto it : sampl[i]) delete it;
        sampl[i].clear();
    }
    if (yvar->getShare() == 1) delete yvar; else yvar->delShare();
    if (ego->getShare() == 1) delete ego; else ego->delShare();

}

