// --- C++ ---
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2020 CEA/DEN
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   TestTEgo.cxx
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2020-11-03
 
 \brief  test de l algorithme EGO
*/


#include "TestTEgo.h"

// CPPUNIT
#include <cppunit/extensions/HelperMacros.h>
#include <cppunit/Test.h>

#define DBL_PREC_STD  0.1
#define DBL_PREC_LARGE 0.25

// ROOT

// URANIE
#include <iostream>
#include "TDataServer.h"
#include "TVizirGenetic.h"
#include "TEGO.h"
#include "TEgoKBModeler.h"
#ifdef WITH_VIZIR
#include "TEgoDynSolver.h"
#include "TEgoStdSolver.h"
#endif
#include "TEgoHjDynSolver.h"
#include "TEgoHjStdSolver.h"
#include "TThreadedRun.h"
#include "TSequentialRun.h"

//#define LENT 1000000000
#define LENT 40000000


#define NK 40 // nombre minimal for a kriging model
#define NR 8 // nombre de ressources
//#define NC 300 // nombre de calcul max
#define NC 120 // nombre de calcul max
#define NP 3
#define NG 10000
#define NT 10 // tested item

// pbl
#define ROSEN 0
#define HIMMEL 1

// data
#define NODATA 0
#define INDATA 1
#define ALLDATA 2

#define MOD


/* rosenbrock */
int userfunl(double* in, double *out)
{
    double ros, ret, tmp, mem;
    int i;

    // rosenbrock
    ros = 0.;
    tmp = in[0] - 1.;
    ros += tmp*tmp;
    for (i=1; i<NP; i++) {
        tmp = in[i] - in[i-1]*in[i-1];
        ros += tmp*tmp;
    }

    // himmelblau
    mem = 0.;
    for (long j=0; j<LENT; j++)
    {
        ret = 0;
        tmp = in[0]*in[0] - in[1] - 6;
        ret += tmp*tmp;
        for (i=1; i<NP; i++) {
            tmp = (i%2) ? in[i]*in[i] - in[i-1] - 6 : in[i] - in[i-1];
            ret += tmp*tmp;
        }
        if (ret > mem) mem=ret;
    }

    out[0] = ros;
    out[1] = mem;
    return 1;
}

/* rosenbrock */
int userfun(double* in, double *out)
{
    double ros, ret, tmp, mem;
    int i;

    // rosenbrock
    ros = 0.;
    tmp = in[0] - 1.;
    ros += tmp*tmp;
    for (i=1; i<NP; i++) {
        tmp = in[i] - in[i-1]*in[i-1];
        ros += tmp*tmp;
    }

    // himmelblau
    ret = 0;
    tmp = in[0]*in[0] - in[1] - 6;
    ret += tmp*tmp;
    for (i=1; i<NP; i++) {
        tmp = (i%2) ? in[i]*in[i] - in[i-1] - 6 : in[i] - in[i-1];
        ret += tmp*tmp;
    }

    out[0] = ros;
    out[1] = ret;
    return 1;
}

// Constructor destructor
TEgoTest::TEgoTest() {}

TEgoTest::~TEgoTest() {}

// pre and post traitement
void TEgoTest::setUp()
{
    int i;

    // input
    _x1 = new URANIE::DataServer::TUniformDistribution("x1", -8., 8.);
    _x2 = new URANIE::DataServer::TUniformDistribution("x2", -8., 8.);
    _x3 = new URANIE::DataServer::TUniformDistribution("x3", -8., 8.);

    // output
    _ros = new URANIE::DataServer::TAttribute("ros");
    _y = new URANIE::DataServer::TAttribute("y");

    // user fun
    URANIE::DataServer::TAttribute *cin[4] = {_x1, _x2, _x3, NULL};
    //std::cout << "toto2\n";
    _fun = new URANIE::Relauncher::TCJitEval(&userfun);
    for (i=0; cin[i] != NULL; i++) _fun->addInput(cin[i]);
    _fun->addOutput(_ros);
    _fun->addOutput(_y);

    _funl = new URANIE::Relauncher::TCJitEval(&userfunl);
    for (i=0; cin[i] != NULL; i++) _funl->addInput(cin[i]);
    _funl->addOutput(_ros);
    _funl->addOutput(_y);

    //std::cout << "toto3\n";
}

void TEgoTest::tearDown()
{
    delete _fun;
    delete _funl;
    delete _y;
    delete _ros;
    delete _x3;
    delete _x2;
    delete _x1;
}

// utilitaires
/* modeler param */
URANIE::MetaModelOptim::TEgoModeler *TEgoTest::regulationMod()
{
    auto egomod = new URANIE::MetaModelOptim::TEgoKBModeler();
    egomod->setModel("matern7/2", "const", 1e-8);
    //egomod.setModel("matern7/2", "const", 0.0);
    //egomod.setSolver("ML", "Praxis", 300, 800);
    egomod->setSolver("ML", "Bobyqa", 300, 800);
    return egomod;
}

void TEgoTest::commonRoutine(URANIE::Relauncher::TEval *fun, int pbl, int data,
    URANIE::MetaModelOptim::TEgoModeler *model,
    URANIE::MetaModelOptim::TEgoSolver *eisolv)
{
    int i;
    URANIE::DataServer::TAttribute *in[4] = {_x3, _x1, _x2, NULL};

    URANIE::Relauncher::TThreadedRun run(fun, NR+1);
    run.startSlave();
    if (run.onMaster()) {
        // tds
        std::cout << "\nmaster\n";
        URANIE::DataServer::TDataServer tds("tds", "ego tds");
        tds.keepFinalTuple(kFALSE);
        for (i=0; in[i] != NULL; i++) tds.addAttribute(in[i]);
        switch (data) {
        case INDATA :
            tds.fileDataRead("lhsIn.dat", kFALSE, kTRUE);
            break;
        case ALLDATA :
            tds.addAttribute(_ros);
            tds.addAttribute(_y);
            tds.fileDataRead("lhsAll.dat", kFALSE, kTRUE);
            break;
        default :
            break;
        }
        
        // meta modele to use
        //URANIE::MetaModelOptim::TEgoModeler *model=NULL;

        // ei optimiser
        //URANIE::MetaModelOptim::TEgoSolver *eisolv=NULL;

        // master
        std::cout << "ego\n";
        URANIE::MetaModelOptim::TEGO egosolv(&tds, &run);
        egosolv.setSize(NK, NC);
        if (model) egosolv.setModeler(model);
        if (eisolv) egosolv.setSolver(eisolv);
        std::cout << "run\n";
        
        if (pbl) egosolv.addObjective(_y);
        else egosolv.addObjective(_ros);

        egosolv.solverLoop();
        run.stopSlave();
        if (pbl) validateHimmelblau(tds);
        else validateRosen(tds);
    }

}

void TEgoTest::modalRoutine(bool hjma)
{
    int i;
    double decal;
    URANIE::DataServer::TAttribute *in[4] = {_x1, _x2, _x3, NULL};

    URANIE::Relauncher::TSequentialRun run(_fun);
    run.startSlave();
    if (run.onMaster()) {
        // tds
        URANIE::DataServer::TDataServer tds("tds", "ego tds");
        tds.keepFinalTuple(kFALSE);
        for (i=0; in[i] != NULL; i++) tds.addAttribute(in[i]);

        // optimiser
        if (hjma) {
            URANIE::MetaModelOptim::THJMA vizir(&tds, &run);
            vizir.setSize(128, 32);
            vizir.addObjective(_y);

            vizir.solverLoop();
            //tds.exportData("modHJ1.dat");
            decal -= 1.;
            vizir.solverLoop();
            //tds.exportData("modHJ2.dat");
            decal -= 1.;
            vizir.solverLoop();
        }
        else {
#ifdef WITH_VIZIR
            URANIE::Reoptimizer::TVizirGenetic viz;
            viz.setSize(800, 200000, 400);
            viz.setModalDiversity(32);
            viz.setStoppingCriteria(1);
            viz.setQuiet();
            URANIE::Reoptimizer::TVizir2 vizir(&tds, &run, &viz);
            vizir.addObjective(_y);

            vizir.solverLoop();
#endif
        }

        //tds.exportData("modC.dat");
        run.stopSlave();
        validateHimmelblau(tds);
    }
}

void TEgoTest::validateHimmelblau(URANIE::DataServer::TDataServer &tds)
{
    int i, ii, top;
    double mem, tmp;
    
    Int_t pat = tds.getNPatterns();
    //CPPUNIT_ASSERT_EQUAL(NC, pat);
    // attributes
    //CPPUNIT_ASSERT_EQUAL(5, tds.getNAttributes());
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("x1"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("x3"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("y"));
    // values
    ii = 1;
    top = (pat > NT) ? NT : pat ;
    mem = tds.getValue("y", pat-1);
    for (i=2; i<top; i++) {
        tmp = tds.getValue("y", pat - i);
        if (tmp < mem ) {
            mem = tmp;
            ii = i;
        }
    }
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, mem, DBL_PREC_STD);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds.getValue("x2", pat-ii), 
        tds.getValue("x3", pat-ii), DBL_PREC_LARGE);
}

void TEgoTest::validateRosen(URANIE::DataServer::TDataServer &tds)
{
    int i, ii, top;
    double mem, tmp;
    
    Int_t pat = tds.getNPatterns();
    //CPPUNIT_ASSERT_EQUAL(NC, pat);
    // attributes
    CPPUNIT_ASSERT_EQUAL(5, tds.getNAttributes());
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("x2"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("x3"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("ros"));
    // values
    ii = 1;
    mem = tds.getValue("ros", pat-1);
    top = (pat > NT) ? NT : pat ;
    for (i=2; i<top; i++) {
        tmp = tds.getValue("ros", pat - i);
        if (tmp < mem ) {
            mem = tmp;
            ii = i;
        }
    }
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, mem, DBL_PREC_STD);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tds.getValue("x1", pat-ii), DBL_PREC_LARGE);
}

// TEST_SUITE
void TEgoTest::modalHjma()
{
    modalRoutine(true);
}

void TEgoTest::rosEmptyDft()
{
    auto mod = regulationMod();
    commonRoutine(_funl, ROSEN, NODATA, mod, NULL);
    delete mod;
}

void TEgoTest::himEmptyDft()
{
    URANIE::MetaModelOptim::TEgoHjDynSolver solv;
    solv.setManyNewItem(2);
    commonRoutine(_funl, HIMMEL, NODATA, NULL, &solv);
}

/* solver param */
#ifdef WITH_VIZIR
void TEgoTest::modalVizir()
{
    modalRoutine(false);
}

void TEgoTest::rosInVizirStd()
{
    auto mod = regulationMod();
    // solver
    URANIE::Reoptimizer::TVizirGenetic viz;
    viz.setSize(320, 100000, 160);
    viz.setQuiet();
    URANIE::MetaModelOptim::TEgoStdSolver ssolv;
    ssolv.setSolver(&viz);
    
    commonRoutine(_fun, ROSEN, INDATA, mod, &ssolv);
    delete mod;
}

void TEgoTest::himAllVizirDyn()
{
    auto mod = regulationMod();
    // solver
    URANIE::Reoptimizer::TVizirGenetic viz;
    viz.setSize(1200, 120000, 400);
    viz.setQuiet();
    URANIE::MetaModelOptim::TEgoDynSolver dsolv;
    dsolv.setSolver(&viz);
    dsolv.setStepSize(NG);

    commonRoutine(_fun, HIMMEL, ALLDATA, mod, &dsolv);
    delete mod;
}
#endif

void TEgoTest::himAllHjmaStd()
{
    auto mod = regulationMod();
    // solver
    URANIE::MetaModelOptim::TEgoHjStdSolver hjssolv;
    hjssolv.setSize(64, 32);

    commonRoutine(_fun, HIMMEL, ALLDATA, mod, &hjssolv);
    delete mod;
}

void TEgoTest::rosInHjmaDyn()
{
    auto mod = regulationMod();
    // solver
    URANIE::MetaModelOptim::TEgoHjDynSolver solv;
#ifdef HUIT
    solv.setSize(128, 32);
#endif

    commonRoutine(_fun, ROSEN, INDATA, mod, &solv);
    delete mod;
}

