/*//////////////////////////////////////////////////////////////
 // Copyright (C) 2013-2024 CEA/DES
 //
 // This program is free software: you can redistribute it and/or modify
 // it under the terms of the GNU Lesser General Public License as published
 // by the Free Software Foundation, either version 3 of the License, or any
 // later version.
 //
 // This program is distributed in the hope that it will be useful,
 // but WITHOUT ANY WARRANTY; without even the implied warranty of
 // MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 // GNU Lesser General Public License for more details.
 //
 // You should have received a copy of the GNU Lesser General Public License
 // along with this program. If not, see <http://www.gnu.org/licenses/>.
 //////////////////////////////////////////////////////////////
 //////////////////////////////////////////////////////////////
 // $Id$
 // $Author$
 // $Date$
 // $Revision$
 // $State$
 //////////////////////////////////////////////////////////////*/
/*!
 \file   gpCor.c
 \author Jean-Marc Martinez <jean-marc.martinez@cea.fr>
 \date   2012/07/19 15:38:00

 \brief
 */
#include <math.h>
#ifdef WIN32
#include "libpgauss_Export.h"
#else
#define libpgauss_EXPORT
#endif

/* Covariance de Gauss                                         */
/* Methode gpCorGauss                                        */
/* Input X1 : matrice d x n1                                   */
/* Input X2 : matrice d x n2                                   */
/* Input n1 : nombre de points de X1                            */
/* Input n2 : nombre de points de X2                            */
/* Input d  : dimension des points                              */
/* Input lc : d longueurs de correlation                     */
/* Output C : matrice n1 x n2 correlations entre X1 et X2     */
libpgauss_EXPORT int gpCorGauss(double *X1, double *X2, int n1, int n2, int d, double *lc,
        double *C)
{
    int i, j, k;
    double *px1, *px2, *pl, h, s;

    for (i = 0; i < n1; i++)
    {
        px2 = X2;
        for (j = 0; j < n2; j++)
        {
            px1 = X1 + i * d;
            pl = lc;
            for (s = 0., k = 0; k < d; k++)
            {
                h = (*px1++ - *px2++) / *pl++;
                s += h * h;
            }
            C[i + j * n1] = exp(-s); /* matrice n1 x n2 */
        }
    }
    return (0);
}

/* Covariance de Gauss (fast square matrix)                     */
/* Methode gpCorGauss2                                        */
/* Input X : matrice d x n                                  */
/* Input n : nombre de points de X                                */
/* Input d  : dimension des points                              */
/* Input lc : d longueurs de correlation                     */
/* Output C : matrice n x n correlations             */
libpgauss_EXPORT int gpCorGauss2(double *X, int n, int d, double *lc, double *C)
{
    int i, j, k;
    double *px1, *px2, *pl, h, s;

    for (i = 0; i < n; i++)
    {
        px2 = X;
        C[i + i * n] = 1;
        for (j = 0; j < i; j++)
        {
            px1 = X + i * d;
            pl = lc;
            for (s = 0., k = 0; k < d; k++)
            {
                h = (*px1++ - *px2++) / *pl++;
                s += h * h;
            }
            C[i + j * n] = exp(-s); /* matrice n x n */
            C[j + i * n] = C[i + j * n];
        }
    }
    return (0);
}

/* Covariance de Gauss - Isotropique                            */
/* Methode gpCorGaussIso                                       */
/* Input X1 : matrice d x n1                                   */
/* Input X2 : matrice d x n2                                   */
/* Input n1 : nombre de points de X1                            */
/* Input n2 : nombre de points de X2                            */
/* Input d  : dimension                                         */
/* Input lc : une longueur de correlation                     */
/* Output C : matrice n1 x n2 correlations entre X1 et X2     */
libpgauss_EXPORT int gpCorGaussIso(double *X1, double *X2, int n1, int n2, int d, double lc,
        double *C)
{
    int i, j, k;
    double *px1, *px2, lc2, h, s;

    lc2 = lc * lc;
    for (i = 0; i < n1; i++)
    {
        px2 = X2;
        for (j = 0; j < n2; j++)
        {
            px1 = X1 + i * d;
            for (s = 0., k = 0; k < d; k++)
            {
                h = (*px1++ - *px2++);
                s += h * h;
            }
            C[i + j * n1] = exp(-s / lc2); /* matrice n1 x n2 */
        }
    }
    return (0);
}

/* Covariance de Gauss - Isotropique (fast square matrix)       */
/* Methode gpCorGaussIso2                                       */
/* Input X : matrice d x n                                   */
/* Input n : nombre de points de X1                              */
/* Input d  : dimension                                         */
/* Input lc : une longueur de correlation                     */
/* Output C : matrice n x n de correlation            */
libpgauss_EXPORT int gpCorGaussIso2(double *X, int n, int d, double lc, double *C)
{
    int i, j, k;
    double *px1, *px2, lc2, h, s;

    lc2 = lc * lc;
    for (i = 0; i < n; i++)
    {
        px2 = X;
        C[i + i * n] = 1;
        for (j = 0; j < i; j++)
        {
            px1 = X + i * d;
            for (s = 0., k = 0; k < d; k++)
            {
                h = (*px1++ - *px2++);
                s += h * h;
            }
            C[i + j * n] = exp(-s / lc2); /* matrice n x n */
            C[j + i * n] = C[i + j * n];
        }
    }
    return (0);
}

/* Covariance exponentielle puissance                           */
/* Methode gpCorExponential                                    */
/* Input X1 : matrice d x n1                                   */
/* Input X2 : matrice d x n2                                   */
/* Input n1 : nombre de points de X1                            */
/* Input n2 : nombre de points de X2                            */
/* Input d  : dimension                                         */
/* Input lc : d longueurs de correlation                     */
/* Input po : d puissances                    */
/* Output C : matrice n1 x n2 correlations entre X1 et X2     */
libpgauss_EXPORT int gpCorExponential(double *X1, double *X2, int n1, int n2, int d, double *lc,
        double *po, double *C)
{
    int i, j, k;
    double *px1, *px2, *pl, *pp, h, s;

    for (i = 0; i < n1; i++)
    {
        px2 = X2;
        for (j = 0; j < n2; j++)
        {
            px1 = X1 + i * d;
            pl = lc;
            pp = po;
            for (s = 0., k = 0; k < d; k++)
            {
                h = (*px1++ - *px2++) / *pl++;
                s += pow(fabs(h), *pp++);
            }
            C[i + j * n1] = exp(-s); /* matrice n1 x n2 */
        }
    }
    return (0);
}

/* Covariance exponentielle puissance (fast square matrix)      */
/* Methode gpCorExponential2                                    */
/* Input X : matrice d x n                                   */
/* Input n : nombre de points de X                             */
/* Input d  : dimension                                         */
/* Input lc : d longueurs de correlation                     */
/* Input po : d puissances                    */
/* Output C : matrice n x n correlations              */
libpgauss_EXPORT int gpCorExponential2(double *X, int n, int d, double *lc, double *po,
        double *C)
{
    int i, j, k;
    double *px1, *px2, *pl, *pp, h, s;

    for (i = 0; i < n; i++)
    {
        px2 = X;
        C[i + i * n] = 1;
        for (j = 0; j < i; j++)
        {
            px1 = X + i * d;
            pl = lc;
            pp = po;
            for (s = 0., k = 0; k < d; k++)
            {
                h = (*px1++ - *px2++) / *pl++;
                s += pow(fabs(h), *pp++);
            }
            C[i + j * n] = exp(-s); /* matrice n x n */
            C[j + i * n] = C[i + j * n];
        }
    }
    return (0);
}

/* Covariance Matern 1/2                                     */
/* Methode gpCMatern12                                        */
/* Input X1 : matrice d x n1                                   */
/* Input X2 : matrice d x n2                                   */
/* Input n1 : nombre de points de X1                            */
/* Input n2 : nombre de points de X2                            */
/* Input d  : dimension                                         */
/* Input lc : d longueurs de correlation                     */
/* Output C : matrice n1 x n2 correlations entre X1 et X2     */
libpgauss_EXPORT int gpCorMatern12(double *X1, double *X2, int n1, int n2, int d, double *lc,
        double *C)
{
    int i, j, k;
    double *px1, *px2, *pl, h, s, a, as;

    a = sqrt(2);
    for (i = 0; i < n1; i++)
    {
        px2 = X2;
        for (j = 0; j < n2; j++)
        {
            px1 = X1 + i * d;
            pl = lc;
            for (s = 0., k = 0; k < d; k++)
            {
                h = (*px1++ - *px2++) / *pl++;
                s += h * h;
            }
            s = sqrt(s);
            as = a * s;
            C[i + j * n1] = exp(-as) ; /* matrice n1 x n2 */
        }
    }
    return (0);
}

/* Covariance Matern 1/2 (fast square matrix)                   */
/* Methode gpCMatern122                                        */
/* Input X : matrice d x n                                  */
/* Input n : nombre de points de X                                */
/* Input d  : dimension                                         */
/* Input lc : d longueurs de correlation                     */
/* Output C : matrice n x n correlations             */
libpgauss_EXPORT int gpCorMatern122(double *X, int n, int d, double *lc, double *C)
{
    int i, j, k;
    double *px1, *px2, *pl, h, s, a, as;

    a = sqrt(2);
    for (i = 0; i < n; i++)
    {
        px2 = X;
        C[i + i * n] = 1;
        for (j = 0; j < i; j++)
        {
            px1 = X + i * d;
            pl = lc;
            for (s = 0., k = 0; k < d; k++)
            {
                h = (*px1++ - *px2++) / *pl++;
                s += h * h;
            }
            s = sqrt(s);
            as = a * s;
            C[i + j * n] = exp(-as) ; /* matrice n x n */
            C[j + i * n] = C[i + j * n];
        }
    }
    return (0);
}

/* Covariance Matern 3/2                                     */
/* Methode gpCMatern32                                        */
/* Input X1 : matrice d x n1                                   */
/* Input X2 : matrice d x n2                                   */
/* Input n1 : nombre de points de X1                            */
/* Input n2 : nombre de points de X2                            */
/* Input d  : dimension                                         */
/* Input lc : d longueurs de correlation                     */
/* Output C : matrice n1 x n2 correlations entre X1 et X2     */
libpgauss_EXPORT int gpCorMatern32(double *X1, double *X2, int n1, int n2, int d, double *lc,
        double *C)
{
    int i, j, k;
    double *px1, *px2, *pl, h, s, a, as;

    a = sqrt(6);
    for (i = 0; i < n1; i++)
    {
        px2 = X2;
        for (j = 0; j < n2; j++)
        {
            px1 = X1 + i * d;
            pl = lc;
            for (s = 0., k = 0; k < d; k++)
            {
                h = (*px1++ - *px2++) / *pl++;
                s += h * h;
            }
            s = sqrt(s);
            as = a * s;
            C[i + j * n1] = exp(-as) * (1. + as); /* matrice n1 x n2 */
        }
    }
    return (0);
}

/* Covariance Matern 3/2 (fast square matrix)                   */
/* Methode gpCMatern322                                        */
/* Input X : matrice d x n                                  */
/* Input n : nombre de points de X                                */
/* Input d  : dimension                                         */
/* Input lc : d longueurs de correlation                     */
/* Output C : matrice n x n correlations             */
libpgauss_EXPORT int gpCorMatern322(double *X, int n, int d, double *lc, double *C)
{
    int i, j, k;
    double *px1, *px2, *pl, h, s, a, as;

    a = sqrt(6);
    for (i = 0; i < n; i++)
    {
        px2 = X;
        C[i + i * n] = 1;
        for (j = 0; j < i; j++)
        {
            px1 = X + i * d;
            pl = lc;
            for (s = 0., k = 0; k < d; k++)
            {
                h = (*px1++ - *px2++) / *pl++;
                s += h * h;
            }
            s = sqrt(s);
            as = a * s;
            C[i + j * n] = exp(-as) * (1. + as); /* matrice n x n */
            C[j + i * n] = C[i + j * n];
        }
    }
    return (0);
}

/* Covariance Matern 5/2                                     */
/* Methode gpCMatern52                                       */
/* Input X1 : matrice d x n1                                   */
/* Input X2 : matrice d x n2                                   */
/* Input n1 : nombre de points de X1                            */
/* Input n2 : nombre de points de X2                            */
/* Input d  : dimension                                         */
/* Input lc : d longueurs de correlation                     */
/* Output C : matrice n1 x n2 correlations entre X1 et X2     */
libpgauss_EXPORT int gpCorMatern52(double *X1, double *X2, int n1, int n2, int d, double *lc,
        double *C)
{
    int i, j, k;
    double *px1, *px2, *pl, h, s, a, as, b;

    a = sqrt(10);
    b = 10. / 3.;
    for (i = 0; i < n1; i++)
    {
        px2 = X2;
        for (j = 0; j < n2; j++)
        {
            px1 = X1 + i * d;
            pl = lc;
            for (s = 0., k = 0; k < d; k++)
            {
                h = (*px1++ - *px2++) / *pl++;
                s += h * h;
            }
            s = sqrt(s);
            as = a * s;
            C[i + j * n1] = exp(-as) * (b * s * s + as + 1); /* matrice n1 x n2 */
        }
    }
    return (0);
}

/* Covariance Matern 5/2 (fast square matrix)                   */
/* Methode gpCMatern522                                       */
/* Input X : matrice d x n                                   */
/* Input n : nombre de points de X                                */
/* Input d  : dimension                                         */
/* Input lc : d longueurs de correlation                     */
/* Output C : matrice n x n correlations             */
libpgauss_EXPORT int gpCorMatern522(double *X, int n, int d, double *lc, double *C)
{
    int i, j, k;
    double *px1, *px2, *pl, h, s, a, as, b;

    a = sqrt(10);
    b = 10. / 3.;
    for (i = 0; i < n; i++)
    {
        px2 = X;
        C[i + i * n] = 1;
        for (j = 0; j < i; j++)
        {
            px1 = X + i * d;
            pl = lc;
            for (s = 0., k = 0; k < d; k++)
            {
                h = (*px1++ - *px2++) / *pl++;
                s += h * h;
            }
            s = sqrt(s);
            as = a * s;
            C[i + j * n] = exp(-as) * (b * s * s + as + 1); /* matrice n x n */
            C[j + i * n] = C[i + j * n];
        }
    }
    return (0);
}

/* Covariance Matern 7/2                                     */
/* Methode gpCMatern72                                       */
/* Input X1 : matrice d x n1                                   */
/* Input X2 : matrice d x n2                                   */
/* Input n1 : nombre de points de X1                            */
/* Input n2 : nombre de points de X2                            */
/* Input d  : dimension                                         */
/* Input lc : d longueurs de correlation                     */
/* Output C : matrice n1 x n2 correlations entre X1 et X2     */
libpgauss_EXPORT int gpCorMatern72(double *X1, double *X2, int n1, int n2, int d, double *lc,
        double *C)
{
    int i, j, k;
    double *px1, *px2, *pl, h, s, a, b1, b2;

    a = sqrt(14.);
    b1 = 14. * a / 15.;
    b2 = 28. / 5;
    for (i = 0; i < n1; i++)
    {
        px2 = X2;
        for (j = 0; j < n2; j++)
        {
            px1 = X1 + i * d;
            pl = lc;
            for (s = 0., k = 0; k < d; k++)
            {
                h = (*px1++ - *px2++) / *pl++;
                s += h * h;
            }
            s = sqrt(s);
            C[i + j * n1] = exp(-a * s) * (s * (s * (b1 * s + b2) + a) + 1); /* matrice n1 x n2 */
        }
    }
    return (0);
}

/* Covariance Matern 7/2 (fast square matrix)                   */
/* Methode gpCMatern722                                       */
/* Input X : matrice d x n                                   */
/* Input n : nombre de points de X                              */
/* Input d  : dimension                                         */
/* Input lc : d longueurs de correlation                     */
/* Output C : matrice n x n correlations              */
libpgauss_EXPORT int gpCorMatern722(double *X, int n, int d, double *lc, double *C)
{
    int i, j, k;
    double *px1, *px2, *pl, h, s, a, b1, b2;

    a = sqrt(14.);
    b1 = 14. * a / 15.;
    b2 = 28. / 5;
    for (i = 0; i < n; i++)
    {
        px2 = X;
        C[i + i * n] = 1;
        for (j = 0; j < i; j++)
        {
            px1 = X + i * d;
            pl = lc;
            for (s = 0., k = 0; k < d; k++)
            {
                h = (*px1++ - *px2++) / *pl++;
                s += h * h;
            }
            s = sqrt(s);
            C[i + j * n] = exp(-a * s) * (s * (s * (b1 * s + b2) + a) + 1); /* matrice n x n */
            C[j + i * n] = C[i + j * n];
        }
    }
    return (0);
}

/* Covariance de MaternI                                     */
/* Methode gpCorMaternI                                        */
/* Input X1 : matrice d x n1                                   */
/* Input X2 : matrice d x n2                                   */
/* Input n1 : nombre de points de X1                            */
/* Input n2 : nombre de points de X2                            */
/* Input d  : dimension                                         */
/* Input lc : d longueurs de correlation                     */
/* Input nu : d valeurs de nu                             */
/* Output C : matrice n1 x n2 correlations entre X1 et X2     */
libpgauss_EXPORT int gpCorMaternI(double *X1, double *X2, int n1, int n2, int d, double *lc,
        double *nu, double *C)
{
    double gpMaternFunction(double nu, double x);
    int i, j, k;
    double *px1, *px2, *pl, *pnu, h, c;

    for (i = 0; i < n1; i++)
    {
        px2 = X2;
        for (j = 0; j < n2; j++)
        {
            px1 = X1 + i * d;
            pl = lc;
            pnu = nu;
            for (c = 1., k = 0; k < d; k++)
            {
                h = fabs((*px1++ - *px2++)) / *pl++;
                c = c * gpMaternFunction(*pnu++, h);
                if (c < 0.0)
                    return (1);
            }
            C[i + j * n1] = c; /* matrice n1 x n2 */
        }
    }
    return (0);
}

/* Covariance de MaternI (fast square matrix)                   */
/* Methode gpCorMaternI2                                        */
/* Input X : matrice d x n                                   */
/* Input n : nombre de points de X                                */
/* Input d  : dimension                                         */
/* Input lc : d longueurs de correlation                     */
/* Input nu : d valeurs de nu                             */
/* Output C : matrice n x n correlations              */
libpgauss_EXPORT int gpCorMaternI2(double *X, int n, int d, double *lc, double *nu, double *C)
{
    double gpMaternFunction(double nu, double x);
    int i, j, k;
    double *px1, *px2, *pl, *pnu, h, c;

    for (i = 0; i < n; i++)
    {
        px2 = X;
        C[i + i * n] = 1;
        for (j = 0; j < i; j++)
        {
            px1 = X + i * d;
            pl = lc;
            pnu = nu;
            for (c = 1., k = 0; k < d; k++)
            {
                h = fabs((*px1++ - *px2++)) / *pl++;
                c = c * gpMaternFunction(*pnu++, h);
                if (c < 0.0)
                    return (1);
            }
            C[i + j * n] = c; /* matrice n x n */
            C[j + i * n] = C[i + j * n];
        }
    }
    return (0);
}

/* Covariance de MaternII                                     */
/* Methode gpCorMaternII                                        */
/* Input X1 : matrice d x n1                                   */
/* Input X2 : matrice d x n2                                   */
/* Input n1 : nombre de points de X1                            */
/* Input n2 : nombre de points de X2                            */
/* Input d  : dimension                                         */
/* Input lc : d longueurs de correlation                     */
/* Input nu : valeur de nu                             */
/* Output C : matrice n1 x n2 correlations entre X1 et X2     */
libpgauss_EXPORT int gpCorMaternII(double *X1, double *X2, int n1, int n2, int d, double *lc,
        double nu, double *C)
{
    double gpMaternFunction(double nu, double x);
    int i, j, k;
    double *px1, *px2, *pl, h, c;

    for (i = 0; i < n1; i++)
    {
        px2 = X2;
        for (j = 0; j < n2; j++)
        {
            px1 = X1 + i * d;
            pl = lc;
            for (c = 1., k = 0; k < d; k++)
            {
                h = fabs((*px1++ - *px2++)) / *pl++;
                c = c * gpMaternFunction(nu, h);
                if (c < 0.0)
                    return (1);
            }
            C[i + j * n1] = c; /* matrice n1 x n2 */
        }
    }
    return (0);
}

/* Covariance de MaternII (fast square matrix)                  */
/* Methode gpCorMaternII2                                        */
/* Input X : matrice d x n                                   */
/* Input n : nombre de points de X                                */
/* Input d  : dimension                                         */
/* Input lc : d longueurs de correlation                     */
/* Input nu : valeur de nu                             */
/* Output C : matrice n x n correlations    */
libpgauss_EXPORT int gpCorMaternII2(double *X, int n, int d, double *lc, double nu, double *C)
{
    double gpMaternFunction(double nu, double x);
    int i, j, k;
    double *px1, *px2, *pl, h, c;

    for (i = 0; i < n; i++)
    {
        px2 = X;
        C[i + i * n] = 1;
        for (j = 0; j < i; j++)
        {
            px1 = X + i * d;
            pl = lc;
            for (c = 1., k = 0; k < d; k++)
            {
                h = fabs((*px1++ - *px2++)) / *pl++;
                c = c * gpMaternFunction(nu, h);
                if (c < 0.0)
                    return (1);
            }
            C[i + j * n] = c; /* matrice n x n */
            C[j + i * n] = C[i + j * n];
        }
    }
    return (0);
}

/* Covariance de MaternIII                                     */
/* Methode gpCorMaternIII                                        */
/* Input X1 : matrice d x n1                                   */
/* Input X2 : matrice d x n2                                 */
/* Input n1 : nombre de points de X1                            */
/* Input n2 : nombre de points de X2                            */
/* Input d  : dimension                                         */
/* Input lc : d longueurs de correlation                     */
/* Input nu : valeur de nu                             */
/* Output C : matrice n1 x n2 correlations entre X1 et X2     */
libpgauss_EXPORT int gpCorMaternIII(double *X1, double *X2, int n1, int n2, int d, double *lc,
        double nu, double *C)
{
    double gpMaternFunction(double nu, double x);
    int i, j, k;
    double *px1, *px2, *pl, h, s;

    for (i = 0; i < n1; i++)
    {
        px2 = X2;
        for (j = 0; j < n2; j++)
        {
            px1 = X1 + i * d;
            pl = lc;
            for (s = 0., k = 0; k < d; k++)
            {
                h = (*px1++ - *px2++) / *pl++;
                s += h * h;
            }
            C[i + j * n1] = gpMaternFunction(nu, sqrt(s)); /* matrice n1 x n2 */
            if (C[i + j * n1] < 0.0)
                return (1);
        }
    }
    return (0);
}

/* Methode gpCorMaternIII  (fast square matrix)           */
/* Input X : matrice d x n                              */
/* Input n : nombre de points de X                      */
/* Input d  : dimension                                 */
/* Input lc : d longueurs de correlation                */
/* Input nu : valeur de nu                         */
/* Output C : matrice n x n correlations          */
libpgauss_EXPORT int gpCorMaternIII2(double *X, int n, int d, double *lc, double nu, double *C)
{
    double gpMaternFunction(double nu, double x);
    int i, j, k;
    double *px1, *px2, *pl, h, s;

    for (i = 0; i < n; i++)
    {
        px2 = X;
        C[i + i * n] = 1;
        for (j = 0; j < i; j++)
        {
            px1 = X + i * d;
            pl = lc;
            for (s = 0., k = 0; k < d; k++)
            {
                h = (*px1++ - *px2++) / *pl++;
                s += h * h;
            }
            C[i + j * n] = gpMaternFunction(nu, sqrt(s)); /* matrice n x n */
            if (C[i + j * n] < 0.0)
                return (1);
            C[j + i * n] = C[i + j * n];
        }
    }
    return (0);
}
