/*//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////*/

#include <stdlib.h>
#include <math.h>
#include "linpack_d.h"
#ifdef WIN32
#include "libpgauss_Export.h"
#else
#define libpgauss_EXPORT
#endif

/* Calcul des erreurs par LOO                             */
/* Methode gpLooh                                         */
/* Input Hin : matrice des regresseurs ou U issue via SVD */
/* Input Kin : matrice de variance                        */
/* Input yin : vecteur des observations                   */
/* Input n : dimension Kin[n x n], yin[n]                 */
/* Input p : dimension Hin[n x p]                         */
/* Output errorLoo : n erreurs LOO                        */
/* Output varLoo : n variances                            */
/* Output eqmLoo   : erreur quadratique moyenne           */
libpgauss_EXPORT int gpLoohError(double *Hin, double *Kin, double *yin, int n, int p, double *errorLoo, double *varLoo, double *eqmLoo) {
  void gpDposl ( double a[], int lda, int n, double b[] );
  int i, j, jn, info, bidon_jpvt;
  double xk, xa, dummy;
  double *qraux, work, *qty, *v, *ay, *z, *At, dT, eqm;
  double determinant[2];

  /* Recopie des donnees qui seront modifiees */
  double *H, *pH, *K, *pK, *y, *py;
  H = calloc(sizeof(double),n*p);
  K = calloc(sizeof(double),n*n);
  y = calloc(sizeof(double),n);
  for(pH = H,i=0;i<n*p;i++) *pH++ = *Hin++;
  for(pK = K,i=0;i<n*n;i++) *pK++ = *Kin++;
  for(py = y,i=0;i<n;i++)   *py++ = *yin++;
  /*******************************/

  qraux = calloc(sizeof(double),p);
  /* work  = calloc(sizeof(double),p); */
  qty   = calloc(sizeof(double),n);
  v     = calloc(sizeof(double),n);
  ay    = calloc(sizeof(double),p);
  z     = calloc(sizeof(double),n);
  At    = calloc(sizeof(double),n * p);
  info = dpofa (K, n, n); /* Decomposition Choleski K = Lt L*/
  if (info == 0) {
    for(i=0;i<n;i++) v[i]=y[i];
    dposl (K, n, n, v);                            /* Resolution y = K v */
    for (j=0;j<p;j++) gpDposl(K, n, n, &(H[j*n])); /* Resolution H = L M */
    dqrdc(H,n,n,p,qraux,&bidon_jpvt,&work,0);       /* Decomposition QR de M */
    /* calcul de A = Qt inv(Lt) */
    dtrdi(K,n,n,determinant,11); /* inverse d'une triangulaire superieure inv(L) */
    for(i=0;i<n;i++) {
      for(j=0;j<i;j++)
	  z[j]=0.;

      for(j=i;j<n;j++)
	  z[j]=K[i + j*n];
      dqrsl(H,n,n,p,qraux,z,&dummy,qty,&dummy,&dummy,&dummy,1000);
      for(j=0;j<p;j++) At[i+j*n] = qty[j];
    }
    /* Calcul de A * y */
    for(j=0;j<p;j++) {
      ay[j]=0.;
      jn=j*n;
      for(i=0;i<n;i++) ay[j] += At[i+jn] * y[i];
    }
    /* Elements diagonaux de inv(K)*/
    for(i=0;i<n;i++) {
      dT = 0.;
      for(j=i;j<n;j++) { xk = K[i+j*n];  dT += xk * xk; }
      for(j=0;j<p;j++) { xa = At[i+j*n]; dT -= xa * xa; }
      errorLoo[i] = v[i];
      for(j=0;j<p;j++) errorLoo[i] -= At[i+j*n] * ay[j];
      errorLoo[i] /= dT;
      varLoo[i]    = 1. /dT;
    }
    for(eqm=0., i=0;i<n;i++) eqm += errorLoo[i] * errorLoo[i];
    eqmLoo[0] = eqm / n;
  }
  else {
    /* printf("gpLooh : The leading minor %d is not positive\n", info); */
    for(i=0;i<n;i++) errorLoo[i]=0.;
    for(i=0;i<n;i++) varLoo[i]=0.;
    eqmLoo[0] = -1;
  }
    free(qraux);    qraux = NULL;
  /* free(work); */
    free(qty);      qty = NULL;
    free(v);        v = NULL;
    free(ay);       ay = NULL;
    free(z);        z = NULL;
    free(At);       At = NULL;
    free(H);        H = NULL;
    free(K);        K = NULL;
    free(y);        y = NULL;
  return info;
}

