/*//////////////////////////////////////////////////////////////
 // Copyright (C) 2013-2024 CEA/DES
 //
 // This program is free software: you can redistribute it and/or modify
 // it under the terms of the GNU Lesser General Public License as published
 // by the Free Software Foundation, either version 3 of the License, or any
 // later version.
 //
 // This program is distributed in the hope that it will be useful,
 // but WITHOUT ANY WARRANTY; without even the implied warranty of
 // MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 // GNU Lesser General Public License for more details.
 //
 // You should have received a copy of the GNU Lesser General Public License
 // along with this program. If not, see <http://www.gnu.org/licenses/>.
 //////////////////////////////////////////////////////////////
 //////////////////////////////////////////////////////////////
 // $Id$
 // $Author$
 // $Date$
 // $Revision$
 // $State$
 //////////////////////////////////////////////////////////////*/
/*!
 \file  gpMLh.c
 \author Jean-Marc Martinez <jean-marc.martinez@cea.fr>
 \date   2012/07/19 15:38:00

 \brief
 */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "linpack_d.h"
#ifdef WIN32
#include "libpgauss_Export.h"
#else
#define libpgauss_EXPORT
#endif

/* Calcul du moins log de la vraisemblance                */
/* Methode gpMLh                                          */
/* Input Hin : matrice des regresseurs ou U issue via SVD */
/* Input Kin : matrice de variance (array double Fortran) */
/* Input yin : vecteur des observations (array double)    */
/* Input n : dimension K[n x n], y[n]                   */
/* Input p : dimension Hin[n x p]                         */
/* cas = 0 general sinon variance en facteur        */
/* Output ml[2] : critere + variance si cas != 0    */
/* Si ml[1] = -1 => echec lors decomposition Cholesky    */
libpgauss_EXPORT int gpMLh(double *Hin, double *Kin, double *yin, int n, int p, double *ml,
        int cas)
{
    void gpDposl(double a[], int lda, int n, double b[]);
    int i, j, info, bidon_jpvt;
    double bidon, *qraux, *qty, v1, v2, v3;

    /* Recopie des donnees qui seront modifiees */
    double *H, *pH, *K, *pK, *y, *py;
	
    H = calloc(sizeof(double), n * p);
    K = calloc(sizeof(double), n * n);
    y = calloc(sizeof(double), n);
    for (pH = H, i = 0; i < n * p; i++)
        *pH++ = *Hin++;
    for (pK = K, i = 0; i < n * n; i++)
        *pK++ = *Kin++;
    for (py = y, i = 0; i < n; i++)
        *py++ = *yin++;
    /*******************************/

    qraux = calloc(sizeof(double), p);
    qty = calloc(sizeof(double), n);

    info = dpofa(K, n, n); /* Decomposition Choleski K = Lt L*/
    if (info == 0)
    {
        gpDposl(K, n, n, y); /* Resolution y = Lt x */
        for (j = 0; j < p; j++)
            gpDposl(K, n, n, &(H[j * n])); /* Resolution H = Lt M */
        dqrdc(H, n, n, p, qraux, &bidon_jpvt, 0, 0); /* Decomposition QR */
        dqrsl(H, n, n, p, qraux, y, &bidon, qty, &bidon, &bidon, &bidon, 1000); /* Resolution Qt y */
        for (v1 = 0., i = 0; i < n; i++)
            v1 += y[i] * y[i];
        for (v2 = 0., j = 0; j < p; j++)
            v2 += qty[j] * qty[j];
        for (v3 = 0., i = 0; i < n; i++)
            v3 += log(K[i + i * n]);
        if (cas == 0)
        {
            ml[0] = v1 - v2 + 2. * v3;
            ml[1] = -1.; /* pas utilise dans ce cas */
        }
        else
        {
            ml[0] = n * log(v1 - v2) + 2. * v3;
            ml[1] = (v1 - v2) / n;
        }
    }
    else
    {
        /* printf("gpMLh : The leading minor %d is not positive\n", info); */
        ml[0] = 0.;
        ml[1] = -1.; /* Cholesky fails */
    }
	
    free(qraux);    qraux = NULL;
    free(qty);      qty = NULL;
    free(H);        H = NULL;
    free(K);        K = NULL;
    free(y);        y = NULL;
    return info;
}

