/*//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////*/

#include <stdlib.h>
#include <math.h>
#include "linpack_d.h"
#ifdef WIN32
#include "libpgauss_Export.h"
#else
#define libpgauss_EXPORT
#endif

/*
              IMPLEMENTED ON 2019/09/25
*/
/* Calcul de des predictions - avec tendance deterministe       */
/* Methode gpPredh                                      	*/
/* Input Hp : matrice new x p des regresseurs			*/
/* Input Kp : matrice new x n de covariances    		*/
/* Input Rp : matrice de variance new x new			*/
/* Input variance : variance du processus			*/
/* Input hbeta : coefficients de la tendance			*/
/* Input iL : matrice n x n calcule par gpModh    	    	*/
/* Input gamma : vecteur de taille n calcule par gpModh         */
/* Input iR : matrice p x p calculee par gpModh    	    	*/
/* Input M  : matrice n x p calculee par gpModh         	*/
/* Input new : nombre de predictions                  		*/
/* Input n : nombre de points utilises pour la selection	*/
/* Input p : nombre de regresseurs				*/
/* Output yp : vecteur des new predictions			*/
/* Output vp : vecteur des new variances			*/
libpgauss_EXPORT void gpPredhCov(double *Hp, double *Kp, double *Rp, double variance, double *hbeta, double *gamma, double *iL, double *iR, double *M, int new, int n, int p, double *yp, double *vp) {
    int i, j, l, inew, jnew, k;
    double vt1, vt2, sum, *pgamma;

    /* calcul des predictions */
    for(i=0;i<new;i++) {
        pgamma = gamma;
        for(vt1=0.,j=0;j<n;j++) vt1 += Kp[i+j*new] * *pgamma++;
        for(vt2=0.,j=0;j<p;j++) vt2 += hbeta[j] * Hp[i+j*new];
        yp[i] = vt2 + vt1;
    }
    
    /* compute the matrix Kp * iL*/
    double *KpiL = calloc(sizeof(double),new*n);
    for(i=0;i<new;i++) {
        for(j=0;j<n;j++) {
            KpiL[i*n+j]=0;
            for(k=0;k<=j;k++) KpiL[i*n+j] += Kp[i+k*new] * iL[k+j*n];
        }
    }
    
    /* compute the matrix (Hnew - Rnew*K^(-1)*H) here known as (Hp - KpiL*M) */
    double *Bloc = calloc(sizeof(double),new*p);
    for(i=0;i<new;i++) {
        for(j=0;j<p;j++) {
            Bloc[i*p+j]=Hp[i+j*new];
            for(k=0;k<n;k++) Bloc[i*p+j] -= KpiL[i*n+k] * M[k+j*n];
        }
    }
    
    /* calcul des variances */
    for(inew=0;inew<new;inew++) {
        for(jnew=0;jnew<new;jnew++) {

            /* premiere partie : algo identitique a celle sans tendance deterministe */
            vp[inew*new+jnew] = Rp[inew*new+jnew] ;
            for(vt2=0.,k=0;k<n;k++) vt2 += KpiL[inew*n+k] * KpiL[jnew*n+k]; 
            vp[inew*new+jnew] -= vt2;
            
            /* seconde partie : Massacrage JBB */
            for(sum=0.,k=0;k<p;k++) {
                for(vt1=0.,l=0;l<p;l++) vt1 += Bloc[inew*p+l] * iR[l+k*p] ;
                for(vt2=0.,l=0;l<p;l++) vt2 += Bloc[jnew*p+l] * iR[l+k*p] ;
                sum += vt1*vt2;
            }
            vp[inew*new+jnew] += sum;
            if ( inew==jnew && vp[inew*new+jnew] < 0.) vp[inew*new+jnew] = 0.;
        }
    }
    
    free(KpiL); KpiL=NULL;
    free(Bloc); Bloc=NULL;
}
