/*
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunSoft, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice 
 * is preserved.
 * ====================================================
 */
 

#include "utilWIN32.h"

double tgamma(double x)
//****************************************************************************80
//
//  Purpose:
//
//    R8_GAMMA evaluates Gamma(X) for a real argument.
//
//  Discussion:
//
//    This routine calculates the gamma function for a real argument X.
//
//    Computation is based on an algorithm outlined in reference 1.
//    The program uses rational functions that approximate the gamma
//    function to at least 20 significant decimal digits.  Coefficients
//    for the approximation over the interval (1,2) are unpublished.
//    Those for the approximation for 12 <= X are from reference 2.
//
//  Licensing:
//
//    This code is distributed under the GNU LGPL license. 
//
//  Modified:
//
//    18 January 2008
//
//  Author:
//
//    Original FORTRAN77 version by William Cody, Laura Stoltz.
//    C++ version by John Burkardt.
//
//  Reference:
//
//    William Cody,
//    An Overview of Software Development for Special Functions,
//    in Numerical Analysis Dundee, 1975,
//    edited by GA Watson,
//    Lecture Notes in Mathematics 506,
//    Springer, 1976.
//
//    John Hart, Ward Cheney, Charles Lawson, Hans Maehly,
//    Charles Mesztenyi, John Rice, Henry Thatcher,
//    Christoph Witzgall,
//    Computer Approximations,
//    Wiley, 1968,
//    LC: QA297.C64.
//
//  Parameters:
//
//    Input, double X, the argument of the function.
//
//    Output, double R8_GAMMA, the value of the function.
//
{
//
//  Coefficients for minimax approximation over (12, INF).
//
   typedef int bool;
   #define false 0
   #define true 1
    double eps;
    double fact;
    double half;
    int i;
    int n;
    double one;
    bool parity;
    double pi;
    double res;
    double sqrtpi;
    double sum;
    double twelve;
    double two;
    double value;
    double xbig;
    double xden;
    double xinf;
    double xminin;
    double xnum;
    double y;
    double y1;
    double ysq;
    double z;
    double zero;

	double c[7] = {-1.910444077728E-03,
	               8.4171387781295E-04,
				   -5.952379913043012E-04,
				   7.93650793500350248E-04,
				   -2.777777777777681622553E-03,
				   8.333333333333333331554247E-02,
				   5.7083835261E-03};

    double p[8] = { -1.71618513886549492533811E+00, 
	             2.47656508055759199108314E+01,
                 -3.79804256470945635097577E+02, 
				 6.29331155312818442661052E+02,
                 8.66966202790413211295064E+02, 
				 -3.14512729688483675254357E+04,
                 -3.61444134186911729807069E+04, 
				 6.64561438202405440627855E+04 };
				 
    double q[8] = { -3.08402300119738975254353E+01,
	              3.15350626979604161529144E+02,
                  -1.01515636749021914166146E+03, 
				  -3.10777167157231109440444E+03,
                  2.25381184209801510330112E+04, 
				  4.75584627752788110767815E+03,
                  -1.34659959864969306392456E+05, 
				  -1.15132259675553483497211E+05 };

	sqrtpi = 0.9189385332046727417803297;		
    eps	= 2.22E-16;
    half	= 0.5;
    one	= 1.0;
    pi	= 3.1415926535897932384626434;   
    twelve	= 12.0;
    two	= 2.0;
    xbig	= 171.624;
    xinf	= 1.79E+308;
    xminin	= 2.23E-308;
    zero	= 0.0;
	
    parity = false;
    fact = one;
    n = 0;
    y = x;
//
//  Argument is negative.
//
    if (y <= zero)
    {
        y = -x;
        y1 = (double) (int) (y);
        res = y - y1;

        if (res != zero)
        {
            if (y1 != (double) (int) (y1 * half) * two)
            {
                parity = true;
            }

            fact = -pi / sin(pi * res);
            y = y + one;
        }
        else
        {
            res = xinf;
            value = res;
            return value;
        }
    }
//
//  Argument is positive.
//
    if (y < eps)
    {
//
//  Argument < EPS.
//
        if (xminin <= y)
        {
            res = one / y;
        }
        else
        {
            res = xinf;
            value = res;
            return value;
        }
    }
    else if (y < twelve)
    {
        y1 = y;
//
//  0.0 < argument < 1.0.
//
        if (y < one)
        {
            z = y;
            y = y + one;
        }
//
//  1.0 < argument < 12.0.
//  Reduce argument if necessary.
//
        else
        {
            n = (int) (y) - 1;
            y = y - (double) (n);
            z = y - one;
        }
//
//  Evaluate approximation for 1.0 < argument < 2.0.
//
        xnum = zero;
        xden = one;
        for (i = 0; i < 8; i++)
        {
            xnum = (xnum + p[i]) * z;
            xden = xden * z + q[i];
        }
        res = xnum / xden + one;
//
//  Adjust result for case  0.0 < argument < 1.0.
//
        if (y1 < y)
        {
            res = res / y1;
        }
//
//  Adjust result for case 2.0 < argument < 12.0.
//
        else if (y < y1)
        {
            for (i = 1; i <= n; i++)
            {
                res = res * y;
                y = y + one;
            }
        }
    }
    else
    {
//
//  Evaluate for 12.0 <= argument.
//
        if (y <= xbig)
        {
            ysq = y * y;
            sum = c[6];
            for (i = 0; i < 6; i++)
            {
                sum = sum / ysq + c[i];
            }
            sum = sum / y - y + sqrtpi;
            sum = sum + (y - half) * log(y);
            res = exp(sum);
        }
        else
        {
            res = xinf;
            value = res;
            return value;
        }
    }
//
//  Final adjustments and return.
//
    if (parity)
    {
        res = -res;
    }

    if (fact != one)
    {
        res = fact / res;
    }

    value = res;

    return value;
}
//****************************************************************************80


double lgamma(double xvalue)
/******************************************************************************/
/*
  Purpose:

    ALNGAM computes the logarithm of the gamma function.

  Licensing:

    This code is distributed under the GNU LGPL license. 

  Modified:

    05 November 2010

  Author:

    Original FORTRAN77 version by Allan Macleod.
    C version by John Burkardt.

  Reference:

    Allan Macleod,
    Algorithm AS 245,
    A Robust and Reliable Algorithm for the Logarithm of the Gamma Function,
    Applied Statistics,
    Volume 38, Number 2, 1989, pages 397-402.

  Parameters:

    Input, double XVALUE, the argument of the Gamma function.

    Output, int IFAULT, error flag.
    0, no error occurred.
    1, XVALUE is less than or equal to 0.
    2, XVALUE is too big.

    Output, double ALNGAM, the logarithm of the gamma function of X.
*/
{
  int * ifault;
  double alr2pi;
  double value;
  double x;
  double x1;
  double x2;
  double xlge;
  double xlgst;
  double y;

  
   double r1[9] = {
    -2.66685511495, 
    -24.4387534237, 
    -21.9698958928, 
     11.1667541262, 
     3.13060547623, 
     0.607771387771, 
     11.9400905721, 
     31.4690115749, 
     15.2346874070 };
   double r2[9] = {
    -78.3359299449, 
    -142.046296688, 
     137.519416416, 
     78.6994924154, 
     4.16438922228, 
     47.0668766060, 
     313.399215894, 
     263.505074721, 
     43.3400022514 };
   double r3[9] = {
    -2.12159572323E+05, 
     2.30661510616E+05, 
     2.74647644705E+04, 
    -4.02621119975E+04, 
    -2.29660729780E+03, 
    -1.16328495004E+05, 
    -1.46025937511E+05, 
    -2.42357409629E+04, 
    -5.70691009324E+02 };
  double r4[5] = {
     0.279195317918525, 
     0.4917317610505968, 
     0.0692910599291889, 
     3.350343815022304, 
     6.012459259764103 };
	 
	 
	 
   alr2pi = 0.918938533204673;
  xlge = 510000.0;
  xlgst = 1.0E+30;
  
  x = xvalue;
  value = 0.0;
/*
  Check the input.
*/
  if ( xlgst <= x )
  {
    *ifault = 2;
    return value;
  }

  if ( x <= 0.0 )
  {
    *ifault = 1;
    return value;
  }

  *ifault = 0;
/*
  Calculation for 0 < X < 0.5 and 0.5 <= X < 1.5 combined.
*/
  if ( x < 1.5 )
  {
    if ( x < 0.5 )
    {
      value = - log ( x );
      y = x + 1.0;
/*
  Test whether X < machine epsilon.
*/
      if ( y == 1.0 )
      {
        return value;
      }
    }
    else
    {
      value = 0.0;
      y = x;
      x = ( x - 0.5 ) - 0.5;
    }

    value = value + x * (((( 
        r1[4]   * y 
      + r1[3] ) * y 
      + r1[2] ) * y 
      + r1[1] ) * y 
      + r1[0] ) / (((( 
                  y 
      + r1[8] ) * y 
      + r1[7] ) * y 
      + r1[6] ) * y 
      + r1[5] );

    return value;
  }
/*
  Calculation for 1.5 <= X < 4.0.
*/
  if ( x < 4.0 )
  {
    y = ( x - 1.0 ) - 1.0;

    value = y * (((( 
        r2[4]   * x 
      + r2[3] ) * x 
      + r2[2] ) * x 
      + r2[1] ) * x 
      + r2[0] ) / (((( 
                  x 
      + r2[8] ) * x 
      + r2[7] ) * x 
      + r2[6] ) * x 
      + r2[5] );
  }
/*
  Calculation for 4.0 <= X < 12.0.
*/
  else if ( x < 12.0 ) 
  {
    value = (((( 
        r3[4]   * x 
      + r3[3] ) * x 
      + r3[2] ) * x 
      + r3[1] ) * x 
      + r3[0] ) / (((( 
                  x 
      + r3[8] ) * x 
      + r3[7] ) * x 
      + r3[6] ) * x 
      + r3[5] );
  }
/*
  Calculation for 12.0 <= X.
*/
  else
  {
    y = log ( x );
    value = x * ( y - 1.0 ) - 0.5 * y + alr2pi;

    if ( x <= xlge )
    {
      x1 = 1.0 / x;
      x2 = x1 * x1;

      value = value + x1 * ( ( 
             r4[2]   * 
        x2 + r4[1] ) * 
        x2 + r4[0] ) / ( ( 
        x2 + r4[4] ) * 
        x2 + r4[3] );
    }
  }

  return value;
}

