//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TANNModeler.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   2008-02-07 17:14:24
 
 \brief Interface of the class URANIE::Modeler::TANNModeler
 
 */
#ifndef TANNMODELER_H
#define TANNMODELER_H

// Newmat
#include "newmat.h"
using namespace NEWMAT;

#include <iostream>

//ROOT
#include "Rtypes.h"
#include "TNamed.h"

//Uranie
#include "Modeler.h"
#include "TModeler.h"
#include "TDataServer.h"
#include "TPatternsEventList.h"
#include "UMLP.h"

//   GPU training of ANN
#if defined(WITH_GPUANN) && !defined(__CINT__)
#include "nn_headers.h"
namespace NNLM = NeuralNetworkLevenbergMarquardt;
#endif


namespace URANIE
{
namespace Modeler
{

class TempSolution;

class TANNModeler: public TModeler
{
public:
    enum EProblem {
        kRegression, kClassification
    };

    enum ENorm
    {
        kCR, kMinusOneOne, kZeroOne
    };
    // Attributes
protected:
    Int_t _nSeed;                             ///! Seed for the random weight initialization

private:
    Int_t _nInput;                                    ///! The number of inputs
    Int_t _nHidden;                                   ///! The number of hidden neurons
    Int_t _nHiddenLayer;                              ///! the number of hidden layer
    Int_t _nOutput;                                   ///! The number of outputs
    Int_t _nNeurons;                                  ///! The number of neurons

    Int_t _nApp;                            ///! Number of samples in the training set dedicated to back-propagation
    Int_t _nTest;                         ///! Number of samples in the training set dedicated to validation
    Double_t _dLearn;                                  ///! The learning quality
    Double_t _dTest;                                    ///! The test quality
    Double_t _dWeightDecay;                       ///! The weight decay parameter - when used, favorize a smoother ANN
    URANIE::DataServer::TPatternsEventList * _elAll; ///! The mix list of patterns
    Bool_t _bMix;               ///! True if data set must be mixed

    TString _sArchi;                                  ///! The architecture, e.g. "x:y:z,3,yhat" specifying inputs, number of hidden neurons, and output
    TString _sInput;
    TString _sHidden;
    TString _sOutput;

    Int_t _nWeight;                                   ///! The number of weights

    /**! The vector of weights, stored like this:
     *   [a][b(1) b(2) b(3).... b(_nInput+2)][c(1) c(2) c(3).... c(_nInput+2)][ ... ]
     *   a: scalar value giving the weight of the output biais (link betw. single output and biais neuron in hidden layer)
     *   b, c, d, ...: _nHidden blocks of (_nInput+2) scalars representing:
     *     b(1): weight link between second hidden neuron and output (c(1) for the 3rd hidden neuron, d(1) for the 4th, etc ...)
     *     b(2): weight between biais input neuron and second hidden neuron
     *     b(2+i): weight between i-th (real) input neuron and second hidden neuron
     *  See impl. of UMPL::evaluateInput().
    */

    ColumnVector _vecWeights;
    TempSolution * _tmpSolution;


    ColumnVector _vecMinValues;                       ///! Min values, used for normalization
    ColumnVector _vecMaxValues;                       ///! Max values, used for normalization

	
    TList * _listOfAttributes;
    //---------------------------------------------
    /** @name Parameters of algorithm
     **/
    //@{
    Double_t _dFcnTol;                  ///! The function tolerance for the trust region algorithm
    ENorm _nNormType;                   ///! The normalized for input and output
    EProblem _nProlemType;              ///! The type of problem (Regression, Classification)

    //! Temporary array used during training
    Double_t * _dval;  // size: _nInput+1 (a full training vector = input + (a single) output)

    //@}
    static const int OPTIM_MAX_ITER = 10000;   //<! Maximal number of iterations in the optimization algorithms
    static const int OPTIM_MAX_FEVAL = 50000;  //<! Maximal number of evaluations in the optimization algorithms
public:
    Bool_t _blog;                             ///< Boolean for edit the log
    URANIE::DataServer::TDataServer *_tds;     //! Pointeur vers un TDS
    UMLP *_UMLP;
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Constructor with a dataserver
    TANNModeler(URANIE::DataServer::TDataServer *tds, TString architecture,
            Double_t dratio = 0.80, Option_t *option = "");
#ifdef	WITH_LIBXML2
    //! Constructor with a dataserver and a PMML File
    TANNModeler (URANIE::DataServer::TDataServer *tds, const char* pmmlfile, const char* ANNname, Double_t dratio = 0.80, Option_t *option ="");
#endif
    //! Default destructor
    virtual ~TANNModeler();
    //@}
    //---------------------------------------------
    /** @name Setter and Getter function
     **/
    //@{
    //! Gets the number of input
    Int_t getNInput()
    {
        return _nInput;
    }
    //! Gets the number of hidden neurons
    Int_t getNHidden()
    {
        return _nHidden;
    }
    //! Gets the number of output
    Int_t getNOutput()
    {
        return _nOutput;
    }
    //! Gets the number of pure training patterns ("APPrentissage")
    Int_t getNAPP()
    {
        return _nApp;
    }
    //! Gets the number of pure training patterns ("APPrentissage")
    Int_t getNTest()
    {
        return _nTest;
    }

    //! Sets the weigts
    void setWeights(ColumnVector weights)
    {
        _vecWeights = weights;
    }
    ColumnVector getWeights()
    {
        return _vecWeights;
    }

    //! Returns the weight givent by the index
    /*!
     the index begins to 0 and ends to nW-1
     */
    Double_t getWeight(Int_t ind)
    {
        return _vecWeights(ind + 1);
    }

    Double_t getMinValue(Int_t ind)
    {
        return _vecMinValues(ind);
    }
    Double_t getMaxValue(Int_t ind)
    {
        return _vecMaxValues(ind);
    }
    Double_t * getdval()
    {
        return _dval;
    }

    void setValue(Int_t ind, Int_t &ncrt, Double_t *xy, Bool_t normalize=true);

    void setDataSet(URANIE::DataServer::TPatternsEventList* tel)
    {
        _elAll = tel;
    }

    void setMix(Bool_t bmix = kTRUE)
    {
        _bMix = bmix;
    }
    void setFcnTol(Double_t dprec)
    {
        _dFcnTol = dprec;
    }
    Double_t getFcnTol()
    {
        return _dFcnTol;
    }

    //! Sets the normalisation
    void setNormalization(ENorm n);
    //! Get the normalisation
    ENorm getNormalization()
    {
        return _nNormType;
    }

    //! Get the problem type
    EProblem getProblem() { return _nProlemType;}
    //! Init the seed
    /*!
     If the value is zero, the seed is initialized with the hour in milliseconds
     
     \param ind (Int_t) the seed value
     */
    void setSeed(Int_t ind = 0);
    //! Returns the seed value
    Int_t getSeed()
    {
        return _nSeed;
    }
    //! Return the name of input attributes indexed by i
    const char * getInputName(Int_t i)
    {
        URANIE::DataServer::TAttribute *att;
        att = (URANIE::DataServer::TAttribute *) _listOfAttributes->At(i);
        return att->GetName();
    }
    //! Get the attribute name of the output
    const char * getOutputName()
    {
        return _sOutput.Data();
    }
    //@}

    //---------------------------------------------
    /** @name Training
     **/
    //@{
    // Init the ANN
    void init();
    //! randomize the weights
    void randomizeWeights();
    /** Launch the training of the ANN. This potentially uses the GPU if available.
    * \param niter number of times (a random permutation of) the test base will be presented for training
    * \param nInit number of times the ANN is trained (from random start weights) with a given permutation
    *              of the test database
    */
#ifdef WITH_GPUANN
    void train(Int_t niter = 10, Int_t nInit = 10, Option_t* option = "text", Bool_t useGPU = true);
#else
    void train(Int_t niter = 10, Int_t nInit = 10, Option_t* option = "text", Bool_t useGPU = false);
#endif
    //! Gets the EQM
    /*!
     \param option ["learn|test"] give the Mean Square Error (\em MSE ) on the learn database ("learn") or on the test database ("test").
     */
    Double_t getEQM(Option_t* option = "");
    //! Get the weight decay parameter
    Double_t getWeightDecay()
    {
        return _dWeightDecay;
    }

    //! Save a vector of weight if the RMS on the *test* database is better than what we currently have
    void saveIfBetter(const ColumnVector & weights, Double_t eqm_app, Double_t eqm_tst);

    //@}

    //---------------------------------------------
    /** @name Exports function
     The exports function to save an Artificial Neural Networks in an external format
     \li \b C/C++
     \li \b Fortran
     \li \b PMML

     The user fonction is the exportFunction method.
     \code
     tann->exportFunction("c++", "uranie_ann_flowrate.C","ANNflowrate");
     tann->exportFunction("fortran", "uranie_ann_flowrate.f","ANNflowrate");
     tann->exportFunction("pmml", sPMMLFile, sANNName,"new");
     \endcode

     \todo Add different format in the exportFunction method
     \code
     tann->exportFunction("C++,fortran,pmml", "uranie_ann_flowrate","ANNflowrate");
     \endcode
     **/
    //@{
    //! The only user interface method
    /*!
     \param lang (const char*) the keyword for the external format [c++|fortran|pmml]
     \param file (const char*) the filename. If empty, it is the name of the TDS with the good extension.[""]
     \param name (const char*) name of the function.
     \param option (const char*) option parameter.
     \li \b NEW : create a new PMML file
     \li \b UPDATE : Write the current ANN in the PMML file and if already exist an ANN in the PMML file, erase it. 
     \li \b BACKUP : Write the current ANN in the PMML file and if already exist ANNs in the PMML file with modelname "name_I", rename the old with "name_I+1"
     */
//     void exportFunction(const char* lang, const char* file = "",
//             const char* name = "", Option_t *option = "");
    //! Export the ANN model in a ROOT macro file 
    void exportModelCplusplus(const char* file = "", const char* name = "",
            Option_t *option = "") const;
//    /*!
//     \todo Write this method as a TModeler method
//     */
    void exportModelCplusplus(std::ofstream * sourcefile)  const;
//     {
//         cerr
//                 << "   --- Method TANNModeler::exportModelCplusplus(ofstream * sourcefile) not yet implemented"
//                 << endl;
//         cerr
//                 << "   --- Then using the method TANNModeler::exportModelCplusplus(file, name, option)"
//                 << endl;
//     }
//     ;

    //! Export the TANNModeler object in a PMML File
    void exportModelPMML(const char* file = "", const char* name = "", Option_t *option = "") const;
    //! Export the ANN model in a Fortran file 
    void exportModelFortran(const char* file, const char* name, Option_t *option = "") ;
    /*!
     \todo Write this method as a TModeler method
     */
    void exportModelFortran(std::ofstream * sourcefile) const;
//     {
//         std::cerr
//                 << "   --- Method TANNModeler::exportModelFortran(ofstream * sourcefile) not yet implemented"
//                 << endl;
//         std::cerr
//                 << "   --- Then using the method TANNModeler::exportModelFortran(file, name, option)"
//                 << endl;
//     }
//     ;
    
    
    //! Export the model in Python langage in a file (not yet implemented)
    /*!
     \warning Don't use this method. Use the main method TModeler::exportFunction.
     \param sourcefile (ofstream *) the pointer to the file to export the modeler.
     */
    void exportModelPython(std::ofstream * sourcefile) const
    {
	UNUSED(sourcefile);
        std::cerr
                << "   --- Method TANNModeler::exportModelPython(ofstream * sourcefile) not yet implemented"
                << endl;
    };
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}
    ClassDef(URANIE::Modeler::TANNModeler, ID_MODELER)
    //Classe def

private:
    void train_cpu(Int_t niter, Int_t nInit, Bool_t randomizeweights);

#if defined(WITH_GPUANN)
    void train_gpu(NNLM::NeuronType T, Int_t niter, Int_t nInit, Bool_t randomizeweights);
#endif

};
} // Fin du namespace Modeler
} // Fin du namespace URANIE

R__EXTERN URANIE::Modeler::TANNModeler *gUranieANNModeler;

#endif
