//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TGPBuilder.cxx
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 \date

 \brief Implementation of the class  URANIE::Modeler::TGPBuilder

 */

//// ROOT
#include "TMath.h"
#include "TLeaf.h"
#ifdef ROOT5
#include "TTreeFormula.h"
#else
#include "TFormula.h"
#endif
#include "TPRegexp.h"
#include "TRegexp.h"
#include "TSystem.h"
#include "TObjString.h"

//// URANIE
#include "UExceptions.h"

#include "TGPBuilder.h"
#include "TGPCostFunction.h"
#include "TGPMLCostFunction.h"
#include "TGPReMLCostFunction.h"
#include "TGPLooCostFunction.h"

#include "TMaternICorrFunction.h"
#include "TMaternIICorrFunction.h"
#include "TMaternIIICorrFunction.h"
#include "TMatern12CorrFunction.h"
#include "TMatern32CorrFunction.h"
#include "TMatern52CorrFunction.h"
#include "TMatern72CorrFunction.h"
#include "TExponentialCorrFunction.h"
#include "TGaussianCorrFunction.h"
#include "TIsotropicGaussianCorrFunction.h"

#include "TDSNtupleD.h"

#include "TAttribute.h"
#include "TUniformDistribution.h"
#include "TLogUniformDistribution.h"

#ifdef WITH_NLOPT
#include "TNlopt.h"
#include "TNloptDirect.h"
#include "TNloptCobyla.h"
#include "TSequentialRun.h"
#include "TThreadedRun.h"
#include "TGradientEstimationRun.h"
#include "TLauncher2.h"
#include "TBasicSampling.h"

#endif

#include <cmath>
#include <math.h>
#include <stdlib.h>
#include <iomanip>

#define epsilon 10 * DBL_MIN

using std::cout;
using std::endl;
using namespace URANIE::DataServer;

ClassImp (URANIE::Modeler::TGPBuilder)

namespace URANIE
{
namespace Modeler
{
// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------
TGPBuilder::TGPBuilder()
{
    initVariables();
}


TGPBuilder::TGPBuilder(TDataServer *tdsObs, const char *varexpinput, const char *varexpoutput, TCorrelationFunction* corrFunc, const char *trend, Option_t * option)
{    
    TString soption = option;
    soption.ToLower();
    initVariables();  
    
    _tds = tdsObs;
    _sInputAttributes  = TString(varexpinput); 
    _sOutputAttributes = TString(varexpoutput);
    trendString        = TString(trend);
    
    initMatrices();
    setCorrFunction(corrFunc);
    computeCovarianceMatrix();
}

TGPBuilder::TGPBuilder(TDataServer *tdsObs, const char *varexpinput, const char *varexpoutput, const char *corrFunc, const char *trend, Option_t * option)
{    
    TString soption = option;
    soption.ToLower();
    initVariables();  
    
    _tds = tdsObs;
    _sInputAttributes  = TString(varexpinput); 
    _sOutputAttributes = TString(varexpoutput);   
    trendString        = TString(trend);
    
    initMatrices();
    setCorrFunction(corrFunc);
    computeCovarianceMatrix();
}

TGPBuilder::TGPBuilder(URANIE::DataServer::TDataServer *tdsModelData)
{
    
    // Check if the tds is a proper one
    if( !TString(tdsModelData->GetName()).BeginsWith("tdsKriging") )
    {
        try
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TGPBuilder::TGPBuilder: the dataserver [%s] provided is incompatible with this constructor. Please load the file generated with TKriging::exportModelData into an EMPTY data server, i.e. with no name or title.",tdsModelData->GetName()));        
        } 
        catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            throw ue;
        }
    }

    TString varexpinput  = "";
    TString varexpoutput = "";
    Double_t varGP = 1.0;
    Double_t vMes  = 0.0;
    Double_t regul = 0.0;
    TString corrFunc("");
    Double_t *params = NULL;
    
    // initialise the GPBuilder
    initVariables();
    _tds = tdsModelData;
    
    // Retrieve the inputs and output names.
    Int_t nbAtt = _tds->getNAttributes();
    for(Int_t i = 0; i < nbAtt-1; i++)
        varexpinput += Form("%s:",_tds->getAttribute(i)->GetName());
    varexpinput.Remove(varexpinput.Length()-1); // we remove the last ":" character
    varexpoutput = _tds->getAttribute(nbAtt-1)->GetName();

    _sInputAttributes  = TString(varexpinput); 
    _sOutputAttributes = TString(varexpoutput);

    // Create the regular expressions     
    TPMERegexp reCorrFunc("URANIE::Modeler::(\\w+): \\[(.+)\\];");
    TPMERegexp reVariance("variance: (\\d\\.\\d+[eE][\\+-]\\d+);");
    TPMERegexp reTrend("trend: \"(.+)\";");
    TPMERegexp reBayes("a priori information needed;");
    TPMERegexp reVMes("measurement error variance: (.+);");
    TPMERegexp reReg("regularisation: (\\d\\.\\d+[eE][\\+-]\\d+);");

    // Parse the #TITLE field for parameters    
    TString title = _tds->GetTitle();
    try
    {
        if(reCorrFunc.Match(title) == 3)
        {
            corrFunc = reCorrFunc[1];
            TObjArray *paramList = reCorrFunc[2].Tokenize(" ");
            Int_t nbParams = paramList->GetEntries();
            params = new Double_t[nbParams];
            for(Int_t i = 0; i < nbParams; i++)
                params[i] = atof(paramList->At(i)->GetName());
    delete paramList;
        }
        else
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form("TGPBuilder::TGPBuilder: no information found about the correlation function in the following string: [%s]", title.Data()));                    
        }
        
        if(reVariance.Match(title) == 2)
            varGP = reVariance[1].Atof();
        else
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form("TGPBuilder::TGPBuilder: no information found about the variance of the Gaussian Process in the following string: [%s]", title.Data()));                    
        }
        
        if(reTrend.Match(title) == 2) 
        {
            trendString = TString(reTrend[1]);
            if(reBayes.Match(title) > 0)
            {
                cout << "!!!!!!!!!!!! IMPORTANT MESSAGE from TGPBuilder  !!!!!!!!!!!" << endl;
                cout << "!! The kriging model was created with a priori knowledge !!" << endl;
                cout << "!! on the trend parameters. Don't forget to set it using !!" << endl;
                cout << "!! setPriorData before calling buildGP.                  !!" << endl;
                cout << "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" << endl;
            }
        }
        
        if(reVMes.Match(title) == 2)
        {
            if(reVMes[1].Contains("user defined"))
            {
                cout << "!!!!!!!!!!!!!! IMPORTANT MESSAGE from TGPBuilder  !!!!!!!!!!!!!!" << endl;
                cout << "!! The kriging model was created with a user defined          !!" << endl;
                cout << "!! measurement error variance. Don't forget to set it using   !!" << endl;
                cout << "!! setMeasurementErrorVariance (or setMeasurementErrorCovMat) !!" << endl;
                cout << "!! before calling buildGP.                                    !!" << endl;
                cout << "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" << endl;
            }
            else
                vMes = reVMes[1].Atof();
        }
        if(reReg.Match(title) == 2)
            regul = reReg[1].Atof();
    }    
    catch(URANIE::Exceptions::UErrorExceptions& ue)
    {
        if(reCorrFunc.Match(title) == 3)
            delete[] params;
        ue.printMessage();
        throw ue;
    }

    // Finish the construction by setting the parameters
    initMatrices();
    setCorrFunction(corrFunc);

    // Normalisation of the correlation lengths
    // NOTE: for the normalisation of the correlation lengths, we should not 
    //       use the xMin shift: correlation lengths are distances ! 
    Int_t nbOtherParams = correlationFunction->getNParameters() - correlationFunction->getNCorrLengths();
    for(Int_t i = 0; i < correlationFunction->getNCorrLengths(); i++)
        params[i+nbOtherParams] = params[i+nbOtherParams]/(xNormParams[2*i+1] - xNormParams[2*i]);         
    correlationFunction->setParameters(params);

    // Variances    
    setVariance(varGP);
    if(vMes > 0.0)
    {
        setHasMeasurementError(kTRUE);
        setAlpha(vMes/varGP);
    }
    else if(regul > 0.0)
        setRegularisation(regul);
    
    computeCovarianceMatrix();  
    
    delete [] params;
}

TGPBuilder* TGPBuilder::buildWithDuplicateKandCandCorrFunc()
{
    TGPBuilder* ret = new TGPBuilder(*this);
    Double_t *KK = new Double_t[_nS * _nS];
    Double_t *CC = new Double_t[_nS * _nS];
    TCorrelationFunction* corrFunc = correlationFunction->clone();
    ret->setKandCandCorrFunc(KK,CC,corrFunc);
    return ret;
}

void TGPBuilder::initVariables()
{
    // Debugging log
    _blog = kFALSE;
    _retOptim=-99;
    
    // Matrix dimensions
    _nX = 0;
    _nY = 0;
    _nS = 0;
    _nP = 0;

    // Variable names
    _sInputAttributes  = TString("");
    _sOutputAttributes = TString("");
    
    // correlation function
    correlationFunction = NULL;
    
    // Initialisation of parameters    
    reg    = 0.0;
    sigma2 = 1.0;
    tolerance = 1e-6;
    alpha  = 0.0;
    _modlengthMax = -1;
    _modlengthMin = -1;
    _lengthMaxScreeningSF = -1;
    
    // Initialisation of flags    
    bHas_corrFunction      = kFALSE;
    bVariance_in_factor    = kTRUE;
    bHas_measurement_error = kFALSE;
    bHas_trend             = kFALSE;
    bUse_prior             = kFALSE;  
    

    // List of trend coefficients
    trendString   = TString("");
    trendCoefList = NULL;
    
    // Kriging matrices
    xObs        = NULL; 
    xNormParams = NULL;
    yObs        = NULL;
    Vmes        = NULL;
    K           = NULL;
    C           = NULL;

    // Trend matrices
    H     = NULL;

    // Prior matrices
    BetaPrior = NULL;
    QPrior    = NULL;
    
    // seed number
    _nSeed = 0;

    bIsbuildWithDuplicateKandCandCorrFunc = kFALSE;
}

void TGPBuilder::initMatrices()
{
    TList *in = _tds->getListOfAttributes(_sInputAttributes);
    _nX = in->GetEntries();
    delete in;

    TList *out = _tds->getListOfAttributes(_sOutputAttributes);
    _nY = out->GetEntries();
    delete out;
    
    try
    {
        if (_nY > 1)
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form("TGPBuilder::initMatrices: the number of output variables (%d) is too large. TGPBuilder can only handle 1 output.",
                         _nY));
        }
        else if (_nY == 0)
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form("TGPBuilder::initMatrices: no output variables defined"));
        }
        else if (_nX == 0)
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form("TGPBuilder::initMatrices: no input variables defined"));
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    _nS = _tds->getNPatterns();

    // Build the matrices
    if(xObs)
    {
        delete[] xObs; xObs = NULL;
        delete[] xNormParams; xNormParams = NULL;
        delete[] yObs; yObs = NULL;
        delete[] Vmes; Vmes = NULL;
        delete[] K; K = NULL;
        delete[] C; C = NULL;        
    }
    
    xObs        = new Double_t[_nX * _nS];
    xNormParams = new Double_t[_nX * 2];
    yObs        = new Double_t[_nS];
    Vmes        = new Double_t[_nS * _nS];
    K           = new Double_t[_nS * _nS];
    C           = new Double_t[_nS * _nS];
        
    // Extract arrays of data from the data server
    // WARNING: for compatibility with gpLib's correlation functions,
    //          the matrix corresponding to xObs must have the input
    //          dimensions as LINE, i.e. if we have the set of points
    //          {(a1,b1),(a2,b2),(a3,b3)}, the corresponding matrix XObs
    //          is supposed to be:
    //          [ a1  a2  a3 ]
    //          [ b1  b2  b3 ]
    //          and the expected flat array xObs = {a1,b1,a2,b2,a3,b3}.
    //          In the data server, data are stored with input dimensions
    //          as COLUMNS, i.e. data are extracted like this:
    //          [ a1  b1 ]
    //          [ a2  b2 ]
    //          [ a3  b3 ]
    //          Thus, to construct the expected flat array xObs, we simply
    //          extract the data by "line" instead of "column".
    _tds->getTuple()->extractData(xObs, _nX * _nS, _sInputAttributes.Data(), _tds->getCut(), "line");
    _tds->getTuple()->extractData(yObs, _nS, _sOutputAttributes.Data(), _tds->getCut(), "column");
    
    // Normalisation of xObs. Each dimension of xObs is projected on a [0,1] 
    // interval. Normalisation parameters are saved in xNormParams.
    // NOTE: the normalisation procedure below is not optimal in speed, but 
    //       it is clear and simple. If it becomes too slow, we can consider 
    //       improving it a bit.
    for(Int_t j = 0; j < _nX; j++)    
    {
        Double_t xMin = 1e99;
        Double_t xMax = -1e99;
        for (Int_t i = 0; i < _nS; i++)
        {
            Double_t x = xObs[i*_nX+j];
            if (xMin > x)
                xMin = x;
            if (xMax < x)
                xMax = x;
        }
        for (Int_t i = 0; i < _nS; i++)
            xObs[i*_nX+j] = (xObs[i*_nX+j]-xMin)/(xMax-xMin);
        
        xNormParams[2*j]   = xMin;  
        xNormParams[2*j+1] = xMax; 
    }

    // Create the deterministic trend
    if (trendString.CompareTo("") != 0)
        createTrend();
    
    // Initialise the "mesurement errors" matrix
    for (Int_t i = 0; i < _nS * _nS; i++)
    {
        C[i] = 0.0;
        Vmes[i] = 0.0;
    }
}

// -----------------------------------------------------------------------
//  Destructor
// -----------------------------------------------------------------------

TGPBuilder::~TGPBuilder()
{
    if (_blog)
        cout << "detruit TGPBuilder" << endl;
    if(xObs)
    {
      if(!bIsbuildWithDuplicateKandCandCorrFunc){
        delete[] xObs;
        xObs = NULL;
        delete[] xNormParams;
        xNormParams = NULL;
        delete[] yObs;
        yObs = NULL;
        delete[] Vmes;
        Vmes = NULL;
      }
      delete[] K;
      K = NULL;
      delete[] C;
      C = NULL;
    }
    if (H && !bIsbuildWithDuplicateKandCandCorrFunc)
    {
        delete[] BetaPrior;
        BetaPrior = NULL;
        delete[] QPrior;
        QPrior = NULL;
        delete[] H;
        H = NULL;
        delete trendCoefList;
        trendCoefList = NULL;
    }
    if(correlationFunction)
    {
        delete correlationFunction;
        correlationFunction = NULL;
    }

}

// -----------------------------------------------------------------------
//  Get and Set method
// -----------------------------------------------------------------------


void TGPBuilder::setLengthRange(double themin, double themax)
{
    try
    {
        if( themin < 0 ||  themax < 0 )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TGPBuilder::setLengthRange the min [%f] and max [%f] values must be positive numbers",
                                                            themin, themax));

        if( themin >= themax )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TGPBuilder::setLengthRange the min [%f] must be stricly lower than the max [%f] value",
                                                            themin, themax));

    } 
    catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    _modlengthMin = themin;
    _modlengthMax = themax;
}

void TGPBuilder::setLengthMaxScreeningSF(double sf)
{

    try
    {
        if( sf < 0 ||  sf >1 )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TGPBuilder::setLengthMaxScreeningSF the scale factor [%f] must be stricly positive and smaller or equal to 1",
                                                            sf));
    } 
    catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    _lengthMaxScreeningSF = sf;
    
}

void TGPBuilder::setCorrFunction(TCorrelationFunction* corrFunc)
{
    if(correlationFunction)
    {
        delete correlationFunction;
        correlationFunction = NULL;
    }
    
    Int_t nbDim = corrFunc->getNDimensions();
    try
    {
        if( nbDim != _nX)
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form("TGPBuilder::setCorrelationFunction: the number of input variables (%d) is different from the number of dimensions of the correlation function (%d). Please provide another function.",
                         _nX, nbDim));
        }
    } 
    catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    correlationFunction = corrFunc->clone();  
    bHas_corrFunction   = kTRUE;
}

void TGPBuilder::setCorrFunction(const char* corrFuncType)
{
    if(correlationFunction)
    {
        delete correlationFunction;
        correlationFunction = NULL;
    }
    
    if ( TString(corrFuncType).CompareTo(     "maternIII"             , TString::kIgnoreCase) == 0 ||
         TString(corrFuncType).CompareTo(     "matern3"               , TString::kIgnoreCase) == 0 ||
         TString(corrFuncType).CompareTo(     "TMaternIIICorrFunction", TString::kIgnoreCase) == 0)
        correlationFunction = new TMaternIIICorrFunction(_nX);
        
    else if ( TString(corrFuncType).CompareTo("maternII"              , TString::kIgnoreCase) == 0 ||
              TString(corrFuncType).CompareTo("matern2"               , TString::kIgnoreCase) == 0 ||
              TString(corrFuncType).CompareTo("TMaternIICorrFunction" , TString::kIgnoreCase) == 0 )
        correlationFunction = new TMaternIICorrFunction(_nX);
        
    else if ( TString(corrFuncType).CompareTo("maternI"               , TString::kIgnoreCase) == 0 ||
              TString(corrFuncType).CompareTo("matern1"               , TString::kIgnoreCase) == 0 ||
              TString(corrFuncType).CompareTo("TMaternICorrFunction"  , TString::kIgnoreCase) == 0 )
        correlationFunction = new TMaternICorrFunction(_nX);

    else if ( TString(corrFuncType).CompareTo("matern1/2"             , TString::kIgnoreCase) == 0 ||
              TString(corrFuncType).CompareTo("matern12"              , TString::kIgnoreCase) == 0 ||
              TString(corrFuncType).CompareTo("TMatern12CorrFunction" , TString::kIgnoreCase) == 0 )
        correlationFunction = new TMatern12CorrFunction(_nX);
    
    else if ( TString(corrFuncType).CompareTo("matern3/2"             , TString::kIgnoreCase) == 0 ||
              TString(corrFuncType).CompareTo("matern32"              , TString::kIgnoreCase) == 0 ||
              TString(corrFuncType).CompareTo("TMatern32CorrFunction" , TString::kIgnoreCase) == 0 )
        correlationFunction = new TMatern32CorrFunction(_nX);
        
    else if ( TString(corrFuncType).CompareTo("matern5/2"             , TString::kIgnoreCase) == 0 ||
              TString(corrFuncType).CompareTo("matern52"              , TString::kIgnoreCase) == 0 ||
              TString(corrFuncType).CompareTo("TMatern52CorrFunction" , TString::kIgnoreCase) == 0 )
        correlationFunction = new TMatern52CorrFunction(_nX);
        
    else if ( TString(corrFuncType).CompareTo("matern7/2"             , TString::kIgnoreCase) == 0 ||
              TString(corrFuncType).CompareTo("matern72"              , TString::kIgnoreCase) == 0 ||
              TString(corrFuncType).CompareTo("TMatern72CorrFunction" , TString::kIgnoreCase) == 0 )        
        correlationFunction = new TMatern72CorrFunction(_nX);
        
    else if ( TString(corrFuncType).CompareTo("isogauss"                      , TString::kIgnoreCase) == 0 ||
              TString(corrFuncType).CompareTo("isogaussian"                   , TString::kIgnoreCase) == 0 ||
              TString(corrFuncType).CompareTo("TIsotropicGaussianCorrFunction", TString::kIgnoreCase) == 0 )
        correlationFunction = new TIsotropicGaussianCorrFunction(_nX);
        
    else if ( TString(corrFuncType).CompareTo("gauss"                         , TString::kIgnoreCase) == 0 ||
              TString(corrFuncType).CompareTo("gaussian"                      , TString::kIgnoreCase) == 0 ||
              TString(corrFuncType).CompareTo("TGaussianCorrFunction"         , TString::kIgnoreCase) == 0 )
        correlationFunction = new TGaussianCorrFunction(_nX);
        
    else if ( TString(corrFuncType).CompareTo("exp"                           , TString::kIgnoreCase) == 0 ||
              TString(corrFuncType).CompareTo("exponential"                   , TString::kIgnoreCase) == 0 ||
              TString(corrFuncType).CompareTo("TExponentialCorrFunction"      , TString::kIgnoreCase) == 0 )
        correlationFunction = new TExponentialCorrFunction(_nX);
        
    else
    {
        try
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form("TGPBuilder::setCorrelationFunction: Unknown or unsupported correlation function: [%s]", corrFuncType));
        }
        catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            throw ue;
        }
        
    }
    bHas_corrFunction = kTRUE;
}


void TGPBuilder::setMeasurementErrorCovMatrix(Double_t* covMes)
{
    for (Int_t i = 0; i < (_nS * _nS); i++)
        Vmes[i] = covMes[i];

    bVariance_in_factor = kFALSE;
    bHas_measurement_error = kTRUE;
    
    computeCovarianceMatrix();
}

void TGPBuilder::setMeasurementErrorVariance(Double_t varMes)
{
    for (Int_t i = 0; i < _nS; i++)
        Vmes[i*(_nS+1)] = varMes;
    
    bVariance_in_factor = kFALSE;
    bHas_measurement_error = kTRUE;
    
    computeCovarianceMatrix();
}

void TGPBuilder::setSeed(Int_t nSeed)
{    
   _nSeed = abs(nSeed);
}

// -----------------------------------------------------------------------
//  Tools
// -----------------------------------------------------------------------

void TGPBuilder::computeCovarianceMatrix(bool print)
{
    if (bHas_corrFunction)
        correlationFunction->eval(_nS, xObs, _nS, xObs, C);
    
    if (bVariance_in_factor)
    {
        for (Int_t i = 0; i < _nS * _nS; i++)
            K[i] = C[i];
        if (bHas_measurement_error)
        {
            for (Int_t i = 0; i < _nS; i++)
                K[i*(_nS+1)] += alpha;
        }
        else
        {
            for (Int_t i = 0; i < _nS; i++)
                K[i*(_nS+1)] += reg;
        }
    }
    else
    {
        for (Int_t i = 0; i < _nS * _nS; i++)
            K[i] = sigma2 * C[i] + Vmes[i];
    }            

    if(print)
    {
        for(int is=0; is<_nS; is++)
        {
            for (Int_t i = 0; i < _nX; i++)
                cout<<xObs[is*_nX+i]<<" ";
            cout<<endl;
        }
        for(int is=0; is<_nS; is++)
        {
            for (Int_t i = 0; i < _nS; i++)
                cout << std::setw(10) << K[is*_nS+i] << " ";
            cout<<endl;        
        }
    }

}

void TGPBuilder::createTrend()
{
    if (H)
    {
        delete[] BetaPrior; BetaPrior = NULL;
        delete[] QPrior; QPrior = NULL;
        delete[] H; H = NULL;
        delete trendCoefList; trendCoefList = NULL;
    }

    Double_t* coefData;
    Int_t nbVal;

    TString localTrendString;
    
    // Decide if the trend is user defined or constructed automatically
    if(trendString.CompareTo("const", TString::kIgnoreCase) == 0)
    {
        localTrendString = "1";
        bUse_normalisation = kTRUE;
    }
    else if(trendString.CompareTo("linear", TString::kIgnoreCase) == 0)
    {
        localTrendString = Form("1:%s",_sInputAttributes.Data());
        bUse_normalisation = kTRUE;
    }
    else
    {
        localTrendString = trendString;
        bUse_normalisation = kFALSE;
    }
    
    // separate the formulas of the different coefficients.
    trendCoefList = localTrendString.Tokenize(":");

    // create the matrices    
    _nP       = trendCoefList->GetEntries();
    H         = new Double_t[_nS * _nP];
    BetaPrior = new Double_t[_nP];
    QPrior    = new Double_t[_nP * _nP];
    TDSNtupleD* tuple;

    if(bUse_normalisation)
    {
        TMatrixD xObsMat(_nS, _nX, xObs);
        tuple = new TDSNtupleD("xObs", "xObs", _sInputAttributes, xObsMat);
    }
    else
        tuple = _tds->getTuple();

    for (Int_t j = 0; j < _nP; j++)
    {
        // retrieve the values of the data server corresponding to the
        // given formula.
        // NOTE: it is unclear if the usage of TTree::Draw is thread safe.
        //       If not, we may have to find another way of computing
        //       the results of the formula. I certainly hope not !!
        TObjString* coefFormula = (TObjString*) trendCoefList->At(j);
        nbVal = tuple->Draw(coefFormula->GetName(), "", "goff");
        if (nbVal>1000000)
        {
            tuple->SetEstimate(tuple->GetEntries());
            nbVal = tuple->Draw(coefFormula->GetName(), "", "goff");
        }
        // fill the H matrix with the retrieved values.
        if (nbVal == _nS)
        {
            coefData = tuple->GetV1();
            for (Int_t i = 0; i < _nS; i++)
                H[i + j * _nS] = coefData[i];
        }
        else
        {
            // if the returned array of value is empty, we suppose it's
            // because the formula is a constant expression. Thus, we evaluate 
            // this constant by creating a TFormula object out of coefFormula 
            // and evaluate it for any input value (like 0, for example).
            // If it is not a constant, then it is an unrecognized character 
            // string and we stop here.
            TFormula f("f", coefFormula->GetName());
            if(f.GetNdim() > 0)
            {                
                Double_t cstValue = f.Eval(0);
                for (Int_t i = 0; i < _nS; i++)
                    H[i + j * _nS] = cstValue;
            }
            else
            {
                try
                {
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,Form("TGPBuilder::createTrend: unrecognized trend formula [%s]", coefFormula->GetName()));
                } catch (URANIE::Exceptions::UErrorExceptions& ue)
                {
                    ue.printMessage();
                    throw ue;
                }                
            }
        }

        // Initialise the prior data
        BetaPrior[j] = 0.0;
        for (Int_t i = 0; i < _nP; i++)
            QPrior[i + j * _nP] = 0.0;
    }
    
    if (_nP > 0)
        bHas_trend = kTRUE;
    
    if (bUse_normalisation)
        delete tuple;
}

void TGPBuilder::setPriorData(Double_t* betaPrior, Double_t* qPrior)
{
    try
    {
        if (!bHas_trend)
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TGPBuilder::setPriorData: no trend was defined");
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    for (Int_t i = 0; i < _nP; i++)
    {
        BetaPrior[i] = betaPrior[i];
        for (Int_t j = 0; j < _nP; j++)
            QPrior[i + j * _nP] = qPrior[i + j * _nP];
    }
    bUse_prior = kTRUE;
}

void TGPBuilder::setUsePrior(Bool_t use_prior)
{
    try
    {
        if (use_prior && !bHas_trend)
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TGPBuilder::setUsePrior: no trend was defined");
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    bUse_prior = use_prior;
}

// -----------------------------------------------------------------------
//  Export
// -----------------------------------------------------------------------

void TGPBuilder::updateObservations()
{
    Int_t nbCorrLengths = correlationFunction->getNCorrLengths();
    Double_t *oldCorrLengths = correlationFunction->getCorrLengths();
    Double_t *newCorrLengths = new Double_t[nbCorrLengths];
    
    // unormalise the correlation lengths
    for(Int_t i = 0; i < nbCorrLengths; i++)
        newCorrLengths[i] = oldCorrLengths[i] * (xNormParams[2*i+1] - xNormParams[2*i]); 

    // reinitialise the matrices
    initMatrices();
    
    // renormalise the correlation lengths
    for(Int_t i = 0; i < nbCorrLengths; i++){
        newCorrLengths[i] = newCorrLengths[i] / (xNormParams[2*i+1] - xNormParams[2*i]); 
    }
    
    // update the correlation function
    correlationFunction->setCorrLengths(newCorrLengths);

    delete [] newCorrLengths; newCorrLengths = NULL;
}

void TGPBuilder::exportGPData(const char* fileName)
{
    TString title("");
    // correlation function
    title += correlationFunction->ClassName();
    title += ": [";
    // correlation function parameters (not correlation lengths)
    Int_t nOtherParams = correlationFunction->getNParameters() - correlationFunction->getNCorrLengths();
    for(Int_t i = 0; i < nOtherParams; i++)
        title += Form("%.14e ", correlationFunction->getParameters()[i]);
    // Unormalised correlation lengths 
    for(Int_t i = 0; i < correlationFunction->getNCorrLengths(); i++)
        title += Form("%.14e ", correlationFunction->getCorrLengths()[i] * (xNormParams[2*i+1] - xNormParams[2*i]));    
    title.Remove(title.Length()-1);
    title += "]; ";
    // variance
    title += Form("variance: %.14e; ", sigma2);
    // trend
    if(bHas_trend)
    {
        title += Form("trend: \"%s\"; ", trendString.Data());
        // Bayesian
        if(bUse_prior)
            title += "a priori information needed; ";            
    }
    
    // measurement error variance
    if(!bVariance_in_factor)
        title += "measurement error variance: user defined; ";
    else if(alpha > 0.0)
        title += Form("measurement error variance: %.14e; ", alpha * sigma2);
    // regularisation
    else if(reg > 0.0)
        title += Form("regularisation: %.14e; ", reg);
    
    
    // Create the tds
    TString varNames(Form("%s:%s", _sInputAttributes.Data(), _sOutputAttributes.Data()));
    TObjArray *varNamesList = varNames.Tokenize(":");    
    
    TDataServer *tdsKrig = new TDataServer("tdsKriging", title.Data());
    for(Int_t i = 0; i < varNamesList->GetEntries(); i++)
    {
        TAttribute *att = new TAttribute(varNamesList->At(i)->GetName());
att->delShare();
        tdsKrig->addAttribute(att);
    }

    delete varNamesList;
    
    // Fill the tds    
    tdsKrig->createTuple();
    TDSNtupleD *t = tdsKrig->getTuple();
    Int_t nbBr = t->GetNbranches();
    Double_t *line = new Double_t[nbBr];

    for(Int_t i = 0; i < _nS; i++)
    {
        line[0] = Double_t(i+1);
        for(Int_t j = 0; j < _nX; j++)
            line[j+1] = xObs[i*_nX+j] * (xNormParams[2*j+1] - xNormParams[2*j]) + xNormParams[2*j];
        line[_nX+1] = yObs[i];
        t->Fill(line);
    }
    
    // Export the data
    tdsKrig->exportData(fileName);
    
delete [] line;   line = NULL;
    // delete the tds
    delete tdsKrig;   tdsKrig = NULL;
}

// -----------------------------------------------------------------------
//  Construction of the Gaussian Process
// -----------------------------------------------------------------------

URANIE::Modeler::TKriging* TGPBuilder::buildGP(Bool_t computeLooErrors)
{
    Int_t returnCode;
    URANIE::Modeler::TKriging *model = NULL;
    Double_t *iL    = new Double_t[_nS*_nS];
    Double_t *gamma = new Double_t[_nS];
    Double_t *hbeta=NULL, *vbeta=NULL, *iR=NULL, *M=NULL;
    Double_t *errorLoo = new Double_t[_nS];
    Double_t *varLoo = new Double_t[_nS];

    Double_t eqmLoo[1];
    Double_t vErr;
    TString varNames = _sInputAttributes;
    varNames.Append(Form(":%s",_sOutputAttributes.Data()));
    
    computeCovarianceMatrix();
    if (bVariance_in_factor)
    {
        vErr = alpha * sigma2;
        for(Int_t i = 0; i < _nS*_nS; i++)
            K[i] *= sigma2;
    }
    else
    {
        // The variance of the measurement error was set by the user, so we 
        // suppose it is known.
        vErr = -1.0;
    }

    if (bHas_trend)
    {        
        hbeta = new Double_t[_nP];
        vbeta = new Double_t[_nP*_nP];
        iR = new Double_t[_nP*_nP];
        M  = new Double_t[_nS*_nP];
        
        if (bUse_prior)
        {
            returnCode = gpModhBayes(BetaPrior, QPrior, H, K, yObs, _nS, _nP, hbeta, vbeta, gamma, iL, iR, M);
            if(computeLooErrors)
                returnCode += gpLoohBayesError(BetaPrior, QPrior, H, K, yObs, _nS, _nP, errorLoo, varLoo, eqmLoo);
        }
        else
        {
            returnCode = gpModh(H, K, yObs, _nS, _nP, hbeta, vbeta, gamma, iL, iR, M);
            if(computeLooErrors)
                returnCode += gpLoohError(H, K, yObs, _nS, _nP, errorLoo, varLoo, eqmLoo);
        }

        if (returnCode == 0)
        {
            model = new URANIE::Modeler::TKriging(_nX, _nS, xObs, xNormParams, yObs, gamma, iL, 
                                                  sigma2, vErr, correlationFunction, varNames,
                                                  _nP, hbeta, vbeta, M, iR, trendString);
        }
        delete [] hbeta;
        delete [] vbeta;
        delete [] iR;
        delete []  M;
    }
    else
    {
        returnCode = gpMod0(K, yObs, _nS, gamma, iL);
        if(computeLooErrors)
            returnCode += gpLoo0Error(K, yObs, _nS, errorLoo, varLoo, eqmLoo);
        
        if (returnCode == 0)
        {
            model = new URANIE::Modeler::TKriging(_nX, _nS, xObs, xNormParams, yObs, gamma, iL, 
                                                  sigma2, vErr, correlationFunction, varNames);
        }
    }

    delete [] iL;
    delete [] gamma;
    
    try
    {
        if (returnCode != 0)
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TGPBuilder::build: Cholesky decomposition failed");
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        delete [] errorLoo;
        delete [] varLoo;
        throw ue;
    }

    if(computeLooErrors)
        model->setLooErrors(errorLoo, varLoo, eqmLoo[0]);
    
    delete [] errorLoo;
    delete [] varLoo;
    return model;
}

// -----------------------------------------------------------------------
//  Find optimal parameters
// -----------------------------------------------------------------------

#ifndef WITH_NLOPT
void TGPBuilder::findOptimalParameters(TString criterion,
                                       Int_t screeningSize,
                                       TString optimAlgo,
                                       Int_t nbMaxOptimSteps,
                                       Bool_t reset,
                               Option_t * option)
{
    UNUSED(criterion,screeningSize,optimAlgo,nbMaxOptimSteps,reset,option);
    try
    {
        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                   Form(
                                                       "TGPBuilder::findOptimalParameters: NLOpt optimisation library not found"));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}
#else
void TGPBuilder::findOptimalParameters(TString criterion,
                                       Int_t screeningSize,
                                       TString optimAlgo,
                                       Int_t nbMaxOptimSteps,
                                       Bool_t reset,
                                       Option_t * option)
{
    if(_blog)
        cout << "TGPBuilder::findOptimalParameters:" << endl;

    //----------------------------------------------------------------------//
    // Get the number of threads from macro
    //----------------------------------------------------------------------//

    TString soption = TString(option);
    soption.ToLower();
    soption.ReplaceAll(" ", "");
    Int_t nbthreads = 1;
    if (soption.Contains("nthreads=") && screeningSize > 0)
    {

        // Get the number max of threads
        SysInfo_t* s = new SysInfo_t();
        gSystem->GetSysInfo(s);
        Int_t nbthmax = s->fCpus;
        delete s;

        TRegexp regular("nthreads=[0-9]*");
        TString out = soption(regular);
        TString snthreads = out.Remove(0, 9);
        soption.ReplaceAll(Form("nthreads=%s", out.Data()), "");

        if (snthreads.IsDigit())
            nbthreads = snthreads.Atoi();
        if(nbthreads < 1)
            nbthreads = 1;
        else if(nbthreads > nbthmax-1)
            nbthreads = nbthmax-1;
        if( _blog )
            cout << endl << " ** nbthreads read in option[" << nbthreads << "]" << endl;
    }

    vector<string> param_provided;
    if(soption.Contains("optimpar="))
    {
       
        vector<string> pouet;
        URANIE::DataServer::UEntry::StringParser(string(soption),",",pouet);
        for(unsigned int istr = 0 ; istr < pouet.size(); istr++)
        {
            TString part = TString(pouet.at(istr));
            if( part.Contains("optimpar=") )
            {
                part = part.ReplaceAll("optimpar=","");
                URANIE::DataServer::UEntry::StringParser(string(part), "-", param_provided);
            }
        }

        cout<<"Parameter provided for optimisation : ";
        for(unsigned int istr = 0 ; istr < param_provided.size(); istr++)
            cout<<" p"<<istr<<"="<<param_provided.at(istr)<<((istr!=(param_provided.size()-1))?",":"");
        cout<<endl;
    }

    
    //---------------------------------------------------------------------//
    // In the following steps, we create and fill the inputs and outputs 
    // parameters of the study.
    //---------------------------------------------------------------------//
    TList *paramInputList = new TList();
    TList *paramOutputList = new TList();
    
    // Create the output
    paramOutputList->Add( new TAttribute("cost") );

    // Construction of the parameters of the correlation function
    // The values below define the interval of variation of the correlation 
    // function parameters. They are fixed arbitrarily and might be changed 
    // in the futur.  
    Double_t lengthMin = ((_modlengthMin!=-1) ? _modlengthMin : 1e-3); // We suppose that the observations are 
    Double_t lengthMax = ((_modlengthMax!=-1) ? _modlengthMax : 10.0 ) * sqrt((double)_nX); // normalised to fit in the interval [0,1]
    Double_t lengthDef = 0.1;
    
    Double_t nuMin     = 0.5;  // Theoretical lower limit for parameter nu
    Double_t nuMax     = 7.5;
    Double_t nuDef     = 1.5;
    
    Double_t powerMin  = 0.5;
    Double_t powerMax  = 2.0; // Theoretical upper limit for power parameter.
    Double_t powerDef  = 1.0;
    
    // Below are the variance related parameters. They too might be changed.
    Double_t alphaMin  = 1e-6;
    Double_t alphaMax  = 10.0;
    Double_t alphaDef  = 0.00316; // sqrt(alphaMin * alphaMax)
    
    Double_t s2CoefMin = 1e-6; // this value is multiplied by the variance of the observed output.
    Double_t s2CoefMax = 100.0;
    Double_t s2CoefDef = 0.01;
    
    // Cases of the Matern correlation functions
    if ( TString(correlationFunction->ClassName()).Contains("TMaternICorrFunction") )
    {
        if(_blog)
            cout << "    Correlation function: TMaternICorrFunction" << endl;

        // If MaternI is used, we have a value of nu per input dimension
        TMaternICorrFunction* cf = dynamic_cast<TMaternICorrFunction*>(correlationFunction);
        Double_t* nu = cf->getParameterNu();
        for(Int_t i = 0; i < _nX; i++)
        {
            TUniformDistribution* attNu = new TUniformDistribution(Form("nu%d",i+1), nuMin, nuMax);
            if(reset)
                attNu->setDefaultValue(nuDef);
            else
            {
                try
                {
                    attNu->setDefaultValue(nu[i]);
                }
                catch (URANIE::Exceptions::UErrorExceptions& ue)
                {
                    ue.printMessage();
                    delete attNu;
                    cleanAttributeList(paramInputList);
                    paramInputList=NULL;
                    cleanAttributeList(paramOutputList);
                    paramOutputList=NULL;
                    throw ue;
                }
            }
            paramInputList->Add(attNu);
        }
        
    }
    else if ( TString(correlationFunction->ClassName()).Contains("TMaternIICorrFunction") ||
              TString(correlationFunction->ClassName()).Contains("TMaternIIICorrFunction"))
    {
        Double_t nu;
        if ( TString(correlationFunction->ClassName()).Contains("TMaternIICorrFunction") )
        {            
            if(_blog)
                cout << "    Correlation function: TMaternIICorrFunction" << endl;    
            TMaternIICorrFunction *cf = dynamic_cast<TMaternIICorrFunction*>(correlationFunction);
            nu = cf->getParameterNu();
        }
        else
        {
            if(_blog)
                cout << "    Correlation function: TMaternIIICorrFunction" << endl;
            TMaternIIICorrFunction *cf = dynamic_cast<TMaternIIICorrFunction*>(correlationFunction);
            nu = cf->getParameterNu();
        }   
        TUniformDistribution* attNu = new TUniformDistribution("nu", nuMin, nuMax);
        if(reset)
            attNu->setDefaultValue(nuDef);
        else
        {
            try
            {
                attNu->setDefaultValue(nu);
            }
            catch (URANIE::Exceptions::UErrorExceptions& ue)
            {
                ue.printMessage();
                delete attNu;
                cleanAttributeList(paramInputList);
                paramInputList=NULL;
                cleanAttributeList(paramOutputList);
                paramOutputList=NULL;
                throw ue;
            }
        }
        paramInputList->Add(attNu);
    }
    else if ( TString(correlationFunction->ClassName()).Contains("TMatern12CorrFunction") )
    {
        if(_blog)
            cout << "    Correlation function: TMatern12CorrFunction" << endl;        
    }
    else if ( TString(correlationFunction->ClassName()).Contains("TMatern32CorrFunction") )
    {
        if(_blog)
            cout << "    Correlation function: TMatern32CorrFunction" << endl;        
    }
    else if ( TString(correlationFunction->ClassName()).Contains("TMatern52CorrFunction") )
    {
        if(_blog)
            cout << "    Correlation function: TMatern52CorrFunction" << endl;        
    }
    else if ( TString(correlationFunction->ClassName()).Contains("TMatern72CorrFunction") )
    {
        if(_blog)
            cout << "    Correlation function: TMatern72CorrFunction" << endl;        
    }
    // Case of Power correlation functions
    else if ( TString(correlationFunction->ClassName()).Contains("TExponentialCorrFunction") )
    {
        if(_blog)
            cout << "    Correlation function: TExponentialCorrFunction" << endl;

        TExponentialCorrFunction *cf = dynamic_cast<TExponentialCorrFunction*>(correlationFunction);
        Double_t* p = cf->getPower();
        for(Int_t i = 0; i < _nX; i++)
        {
            TUniformDistribution* attPower = new TUniformDistribution(Form("p%d",i+1), powerMin, powerMax);
            if(reset)
                attPower->setDefaultValue(powerDef);
            else
            {
                try
                {
                    attPower->setDefaultValue(p[i]);
                }
                catch (URANIE::Exceptions::UErrorExceptions& ue)
                {
                    ue.printMessage();
                    delete attPower;
                    cleanAttributeList(paramInputList);
                    paramInputList=NULL;
                    cleanAttributeList(paramOutputList);
                    paramOutputList=NULL;
                    throw ue;
                }                
            }
            paramInputList->Add( attPower );
        }
    }
    // Case of Gaussian correlation functions
    else if ( TString(correlationFunction->ClassName()).Contains("TGaussianCorrFunction") )
    {
        if(_blog)
            cout << "    Correlation function: TGaussianCorrFunction" << endl;        
    }
    else if ( TString(correlationFunction->ClassName()).Contains("TIsotropicGaussianCorrFunction") )
    {
        if(_blog)
            cout << "    Correlation function: TIsotropicGaussianCorrFunction" << endl;
    }
    // Case of other correlation functions
    else
    {
        try
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TGPBuilder::findOptimalParameters: Unknown or unsupported correlation function: [%s]",correlationFunction->ClassName()));
        }
        catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            cleanAttributeList(paramInputList);
            paramInputList=NULL;
            cleanAttributeList(paramOutputList);
            paramOutputList=NULL;
            throw ue;
        }
    }

    // Creating the correlation lengths (this part is common to all correlation functions
    Int_t nbCorrLengths   = correlationFunction->getNCorrLengths();
    Double_t* corrLengths = correlationFunction->getCorrLengths();
    
    for(Int_t i = 0; i < nbCorrLengths; i++)
    {
        TLogUniformDistribution* l = new TLogUniformDistribution(Form("l%d",i+1), lengthMin, lengthMax);
        if(reset)
            l->setDefaultValue(lengthDef);
        else
        {
            try
            {
                l->setDefaultValue( corrLengths[i] );
            }
            catch (URANIE::Exceptions::UErrorExceptions& ue)
            {
                ue.printMessage();
                delete l;
                cleanAttributeList(paramInputList);
                paramInputList=NULL;
                cleanAttributeList(paramOutputList);
                paramOutputList=NULL;
                throw ue;
            }
        }                        
        paramInputList->Add( l );
    }

    // Dealing with the measurement error
    if (bHas_measurement_error)
    {       
        // Creating the variance parameter if it can't be factorized
        if( !bVariance_in_factor )
        {            
            // The variance of the GP is supposed to be smaller than that of the observed output.
            // NOTE: is this assumption correct ?
            _tds->computeStatistic(_sOutputAttributes);
            TAttribute* att = _tds->getAttribute(_sOutputAttributes);
            Double_t outputVar = att->getStd() * att->getStd();
    
            if(_blog)
                cout << "    creating GP Variance as a parameter" << endl;
    
            TLogUniformDistribution* var = new TLogUniformDistribution("sigma2", outputVar*s2CoefMin, outputVar*s2CoefMax);
            if(reset)
                var->setDefaultValue(outputVar * s2CoefDef);
            else
            {                
                try
                {
                    var->setDefaultValue( sigma2 );
                }
                catch (URANIE::Exceptions::UErrorExceptions& ue)
                {
                    ue.printMessage();
                    delete var;
                    cleanAttributeList(paramInputList);
                    paramInputList=NULL;
                    cleanAttributeList(paramOutputList);
                    paramOutputList=NULL;
                    throw ue;
                }
            }   
            
            // In this case, the sigma2 parameter is an input of the cost function.
            paramInputList->Add( var );
        }
        // Creating the alpha parameter 
        else
        {
            if(_blog)
                cout << "    Creating alpha parameter" << endl;
    
            TLogUniformDistribution* alphaAtt = new TLogUniformDistribution("alpha", alphaMin, alphaMax);
            if(reset)
                alphaAtt->setDefaultValue(alphaDef);
            else
            {
                try
                {
                    alphaAtt->setDefaultValue( alpha );
                }
                catch (URANIE::Exceptions::UErrorExceptions& ue)
                {
                    ue.printMessage();
                    delete alphaAtt;
                    cleanAttributeList(paramInputList);
                    paramInputList=NULL;
                    cleanAttributeList(paramOutputList);
                    paramOutputList=NULL;
                    throw ue;
                }
            }   
            
            paramInputList->Add( alphaAtt );       
            paramOutputList->Add( new TAttribute("sigma2") );
        }
    }
    else        
        paramOutputList->Add( new TAttribute("sigma2") );

    // All the input and output parameters are now set.    
    Int_t nbInputParams = paramInputList->GetEntries();
    Int_t nbOutputParams = paramOutputList->GetEntries();
    
    //---------------------------------------------------------------------//
    // In the next steps, we create the object that will allow us to find 
    // the optimal parameters.
    //---------------------------------------------------------------------//
    Double_t *optimalParamValues = new Double_t[nbInputParams*nbthreads];
    Double_t *optimalVar = new Double_t[nbthreads];
    Double_t *optimalCost = new Double_t[nbthreads];
    Double_t *optimalAlpha = new Double_t[nbthreads];
    for(int t=0;t<nbthreads;t++){
        optimalVar[t] = 1.0;
        optimalCost[t] = 1e99;
        optimalAlpha[t] = 0.0;
    }

    URANIE::Relauncher::TLauncher2 *launcher=NULL;
    URANIE::Relauncher::TRun *runner=NULL;
              
    if( screeningSize > 0)
    {
        cout << "TGPBuilder::findOptimalParameters: starting screening procedure ("<< screeningSize << " evaluations)"<< endl;
        // Construction of the cost function
        TGPCostFunction *gpCostFunc=NULL;
        if( (criterion.CompareTo("ML", TString::kIgnoreCase) == 0) && !bUse_prior )
            gpCostFunc = new TGPMLCostFunction(this);
        else if( (criterion.CompareTo("ReML", TString::kIgnoreCase) == 0) && bHas_trend )
            gpCostFunc = new TGPReMLCostFunction(this);
        else if( criterion.CompareTo("Loo", TString::kIgnoreCase) == 0 )
            gpCostFunc = new TGPLooCostFunction(this);
        else
        {            
            try
            {
                if( (criterion.CompareTo("ML", TString::kIgnoreCase) == 0) && bUse_prior )
                {
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                               "TGPBuilder::findOptimalParameters: ML criterion can't be used with a priori information. Use ReML criterion instead.");
                }
                else if( (criterion.CompareTo("ReML", TString::kIgnoreCase) == 0) && !bHas_trend )
                {
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                               "TGPBuilder::findOptimalParameters: ReML criterion can't be used without a deterministic trend. Use ML criterion instead.");
                }
                else
                {
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                               Form("TGPBuilder::findOptimalParameters: Unknown or unsupported criterion: [%s]", criterion.Data()));
                }
            }
            catch (URANIE::Exceptions::UErrorExceptions& ue)
            {
                ue.printMessage();
                delete gpCostFunc;
                cleanAttributeList(paramInputList);
                paramInputList=NULL;
                cleanAttributeList(paramOutputList);
                paramOutputList=NULL;
                delete[] optimalVar; optimalVar=NULL;
                delete[] optimalCost; optimalCost=NULL;
                delete[] optimalAlpha; optimalAlpha=NULL;
                delete[] optimalParamValues; optimalParamValues=NULL;
                throw ue;
            }
        }
        
        // Construction of the TDS storing the values of the parameters
        TDataServer tdsParam("tdsParam","parameters to optimise");    

        // Fill the TDS and GPCostFunction with the parameters
        // NOTE: only the variable input parameters are added to the TDS.
        for(Int_t i = 0; i < nbInputParams; i++)
        {
            TAttribute *att = (TAttribute*)paramInputList->At(i);
            if( _lengthMaxScreeningSF != -1 && TString(att->GetName()).BeginsWith("l") )
                att->setUpperBound(  _lengthMaxScreeningSF * lengthMax );
            tdsParam.addAttribute(att);
            gpCostFunc->addInput(att);
        }
        for(Int_t i = 0; i < nbOutputParams; i++)
            gpCostFunc->addOutput( (TAttribute*)paramOutputList->At(i) );

        // Construction of the screening
        URANIE::Sampler::TBasicSampling s(&tdsParam, "lhs", screeningSize);
        s.setSeed(_nSeed);
        s.generateSample();
        
        // Evaluation of the cost function on the screening
        if(nbthreads == 1)
            runner = new URANIE::Relauncher::TSequentialRun(gpCostFunc);
        else
            runner = new URANIE::Relauncher::TThreadedRun(gpCostFunc, nbthreads+1);

        runner->startSlave();
        if (runner->onMaster())
        {
            launcher = new URANIE::Relauncher::TLauncher2(&tdsParam, runner);
            launcher->solverLoop();
            runner->stopSlave();
            delete launcher;
        }

        delete runner;
        delete gpCostFunc;

        int nbFailure = tdsParam.getTuple()->Draw("cost","cost==1e99","goff");
        if(nbFailure!=0)
            cout<<"\033[1;35m<WARNING> TGPBuilder::Screening procedure. The cholesky decomposition has failed "<<nbFailure<<" out of "<<screeningSize<<" times </WARNING>\033[0m"<<endl;       

        Double_t *costValues = tdsParam.getTuple()->getBranchData("cost");
        // Search for the paramaters corresponding to the nbthreads minimal costs
        Int_t *minIndexes = new Int_t[nbthreads];
        LSort(costValues,minIndexes,nbthreads,screeningSize);
        
        for(Int_t i = 0; i < nbInputParams; i++)
        {
            TAttribute* att = (TAttribute*)paramInputList->At(i);
            for(int t=0;t<nbthreads;t++)
            {
                optimalParamValues[i+t*nbInputParams] = tdsParam.getValue(att->GetName(), minIndexes[t]);
                att->setDefaultValue(optimalParamValues[i+t*nbInputParams]);
                if(_blog)
                    cout << "    " << att->GetName() << " = " << optimalParamValues[i+t*nbInputParams] << endl; 
            }           
        }
        
        for(int t=0;t<nbthreads;t++)
        {
            optimalCost[t] = costValues[minIndexes[t]];
            optimalVar[t] = tdsParam.getValue("sigma2", minIndexes[t]);        
            if(bHas_measurement_error && bVariance_in_factor)
                optimalAlpha[t] = tdsParam.getValue("alpha", minIndexes[t]);  
        } 
        delete[] minIndexes;
        tdsParam.exportData("gpBuilder-screening.dat");
        delete tdsParam.getTuple(); tdsParam.setTuple( (TDSNtupleD*)NULL);     
    }
    //---------------------------------------------------------------------//
    // In the next steps, we proceed to the optimisation
    //---------------------------------------------------------------------//
    if (optimAlgo.CompareTo("") != 0)
    {
        cout << "TGPBuilder::findOptimalParameters: starting optimisation procedure ("<< optimAlgo.Data() << " algorithm)"<< endl;
        // Construction of the cost function
        TGPCostFunction *gpCostFunc=NULL;
        if( ((criterion.CompareTo("ML", TString::kIgnoreCase) == 0) && !bUse_prior) || 
            ((criterion.CompareTo("ReML", TString::kIgnoreCase) == 0) && !bHas_trend) )
            gpCostFunc = new TGPMLCostFunction(this);
        else if( ((criterion.CompareTo("ML", TString::kIgnoreCase) == 0) && bUse_prior) || 
                 ((criterion.CompareTo("ReML", TString::kIgnoreCase) == 0) && bHas_trend) )
            gpCostFunc = new TGPReMLCostFunction(this);
        else if( criterion.CompareTo("Loo", TString::kIgnoreCase) == 0 )
            gpCostFunc = new TGPLooCostFunction(this);
        else
        {
            try
            {
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("TGPBuilder::findOptimalParameters: Unknown or unsupported criterion: [%s]", criterion.Data()));
            }
            catch (URANIE::Exceptions::UErrorExceptions& ue)
            {
                ue.printMessage();
                delete gpCostFunc;
                cleanAttributeList(paramInputList);
                paramInputList=NULL;
                cleanAttributeList(paramOutputList);
                paramOutputList=NULL;
                delete[] optimalVar; optimalVar=NULL;
                delete[] optimalCost; optimalCost=NULL;
                delete[] optimalAlpha; optimalAlpha=NULL;
                delete[] optimalParamValues; optimalParamValues=NULL;
                throw ue;
            }
        }
        
        // Construction of the TDS storing the values of the parameters        
        TDataServer tdsParam("tdsParam","parameters to optimise");    
        
        // Fill the TDS and GPCostFunction with the parameters
        // NOTE: only the variable input parameters are added to the TDS.
        for(Int_t i = 0; i < nbInputParams; i++)
        {
            TAttribute *att = (TAttribute*)paramInputList->At(i);
            if( _lengthMaxScreeningSF != -1 && TString(att->GetName()).BeginsWith("l") )
                att->setUpperBound(  lengthMax );
            tdsParam.addAttribute(att);
            gpCostFunc->addInput(att);
        }
        for(Int_t i = 0; i < nbOutputParams; i++)
            gpCostFunc->addOutput( (TAttribute*)paramOutputList->At(i) );

   
        // Optimisation algorithm
        URANIE::Reoptimizer::TNloptSolver* solv=NULL;

        // Direct optimisation algorithms
        if (optimAlgo.CompareTo("NelderMead", TString::kIgnoreCase) == 0)
            solv = new URANIE::Reoptimizer::TNloptNelderMead();
        
        else if (optimAlgo.CompareTo("Subplexe", TString::kIgnoreCase) == 0)
            solv = new URANIE::Reoptimizer::TNloptSubplexe();
        
        else if (optimAlgo.CompareTo("Cobyla", TString::kIgnoreCase) == 0)
            solv = new URANIE::Reoptimizer::TNloptCobyla();
        
        else if (optimAlgo.CompareTo("Bobyqa", TString::kIgnoreCase) == 0)
            solv = new URANIE::Reoptimizer::TNloptBobyqa();
        
        else if (optimAlgo.CompareTo("Praxis", TString::kIgnoreCase) == 0)
            solv = new URANIE::Reoptimizer::TNloptPraxis();
        
        // Gradient based optimisation algorithms
        else if (optimAlgo.CompareTo("BFGS", TString::kIgnoreCase) == 0)
            solv = new URANIE::Reoptimizer::TNloptLBFGS();
        
        else if (optimAlgo.CompareTo("Newton", TString::kIgnoreCase) == 0)
            solv = new URANIE::Reoptimizer::TNloptNewton();
        
        else if (optimAlgo.CompareTo("MMA", TString::kIgnoreCase) == 0)
            solv = new URANIE::Reoptimizer::TNloptMMA();
        
        else if (optimAlgo.CompareTo("SLSQP", TString::kIgnoreCase) == 0)
            solv = new URANIE::Reoptimizer::TNloptSLSQP();
        
        else if (optimAlgo.CompareTo("VariableMetric", TString::kIgnoreCase) == 0)
            solv = new URANIE::Reoptimizer::TNloptVariableMetric();        
        else
        {
            try
            {
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("TGPBuilder::findOptimalParameters: Unknown optimisation algorithm type: [%s]",optimAlgo.Data()));
            }
            catch (URANIE::Exceptions::UErrorExceptions& ue)
            {
                ue.printMessage();
                delete gpCostFunc;
                delete solv;
                cleanAttributeList(paramInputList);
                paramInputList=NULL;
                cleanAttributeList(paramOutputList);
                paramOutputList=NULL;
                delete[] optimalVar; optimalVar=NULL;
                delete[] optimalCost; optimalCost=NULL;
                delete[] optimalAlpha; optimalAlpha=NULL;
                delete[] optimalParamValues; optimalParamValues=NULL;
                throw ue;
            }
        }
    
        // Set the starting position of the optimisation
        Double_t *startingPosition;
        if(screeningSize == 0)
        {
            startingPosition = new Double_t[nbInputParams];
            for(Int_t i = 0; i < nbInputParams; i++)
            {
                TStochasticAttribute* obj = dynamic_cast<TStochasticAttribute*>(paramInputList->At(i));
                if( ! obj->getDefaultValue( startingPosition[i] ) )
                    startingPosition[i] = obj->getTheoreticalMean();
            }
        }
        else
            startingPosition = optimalParamValues;

        // Construction of the optimiser
        // Launch the gradient computation sequence
        if(nbthreads == 1)
            runner = new URANIE::Relauncher::TSequentialRun(gpCostFunc);
        else
            runner = new URANIE::Relauncher::TThreadedRun(gpCostFunc, nbthreads+1);
        //URANIE::Relauncher::TGradientEstimationRun gradRun(runner);

        runner->startSlave();

        if( runner->onMaster() )
        {
            URANIE::Reoptimizer::TNlopt opt(&tdsParam, runner, solv);
            opt.addObjective((TAttribute*)paramOutputList->At(0));
            opt.setTolerance(tolerance);  
            opt.setMaximumEval(nbMaxOptimSteps);
            // Solving
           if(nbthreads == 1)
           {
               if(param_provided.size()==(UInt_t(nbInputParams)) )
                   for(int idim = 0; idim < nbInputParams; idim++)
                       startingPosition[idim]=atof(param_provided.at(idim).c_str());
                       
               cout<<"Starting point is ";
               for(int idim = 0; idim < nbInputParams; idim++)
                   cout<<" p"<<idim<<"=="<<startingPosition[idim]<<((idim!=nbInputParams-1) ? "," : "");
               cout<<""<<endl;
           }

            
            for(int t=0;t<nbthreads;t++)
                opt.setStartingPoint(nbInputParams,startingPosition+t*nbInputParams);

            opt.solverLoop();

            _retOptim = opt.getReturnOptim();
            runner->stopSlave();
        }

        tdsParam.getTuple()->Scan("*");
        
        if(tdsParam.getNPatterns() > 0)
        {
            // Retrieve the results    
            for(Int_t i = 0; i < nbInputParams; i++)
            {

                for(int t=0;t<nbthreads;t++)
                {
                    TAttribute* att = (TAttribute*)paramInputList->At(i);
                    optimalParamValues[i+t*nbInputParams] = tdsParam.getValue(att->GetName(), t);

                    // verifies that the value is between the bounds
                    if(att->isUpperBounded()
                       && optimalParamValues[i+t*nbInputParams] >= att->getUpperBound())
                    {
                        optimalParamValues[i+t*nbInputParams] = att->getUpperBound();
                    }   
                    else if (att->isLowerBounded()
                             && optimalParamValues[i+t*nbInputParams] <= att->getLowerBound())
                    {
                        optimalParamValues[i+t*nbInputParams] = att->getLowerBound();
                    }
                
                    if(_blog)
                        cout << "    " << att->GetName() << " = " << optimalParamValues[i+t*nbInputParams] << endl;
                }
            }
            for(int t=0;t<nbthreads;t++)
            {
                optimalCost[t] = tdsParam.getValue("cost", t);
                optimalVar[t]  = tdsParam.getValue("sigma2", t);
                if(bHas_measurement_error && bVariance_in_factor)
                    optimalAlpha[t] = tdsParam.getValue("alpha", t);        
    
                if(_blog)
                    cout << "    Optimal cost: " << optimalCost[t] << "; optimal variance: " << optimalVar[t] << endl;
            }
            
            tdsParam.exportData("gpBuilder-optim.dat");
            delete tdsParam.getTuple(); tdsParam.setTuple( (TDSNtupleD*)NULL);   
        }
        else
        {
            try
            {
                if( screeningSize > 0)
                {
                    throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                                                                 Form("TGPBuilder::findOptimalParameters: the optimisation procedure failed. The optimal parameters are chosen from the results of the screening procedure") );
                }
                else
                {
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                               Form("TGPBuilder::findOptimalParameters: the optimisation procedure failed. No optimal parameters found.") );
                }
            }
            catch (URANIE::Exceptions::UWarningExceptions& ue)
            {
                ue.printMessage();
            }
            catch (URANIE::Exceptions::UErrorExceptions& ue)
            {
                ue.printMessage();
                delete gpCostFunc;
                gpCostFunc=NULL;
                delete solv;
                solv=NULL;
                cleanAttributeList(paramInputList);
                paramInputList=NULL;
                cleanAttributeList(paramOutputList);
                paramOutputList=NULL;
                delete[] optimalVar; optimalVar=NULL;
                delete[] optimalCost; optimalCost=NULL;
                delete[] optimalAlpha; optimalAlpha=NULL;
                delete[] optimalParamValues; optimalParamValues=NULL;
                throw ue;
            }
        }
      
        delete solv; solv = NULL;
        delete gpCostFunc;   gpCostFunc= NULL;
        delete runner;    runner = NULL;
        if(screeningSize == 0) {
            delete [] startingPosition;
            startingPosition = NULL;}

    }

    //---------------------------------------------------------------------//
    // Now we save the optimal parameters found
    //---------------------------------------------------------------------//    

    // Update the GP and correlation function with the optimal parameters
    if( (screeningSize > 0) || (optimAlgo.CompareTo("") != 0) )
    {
        Double_t cost = optimalCost[0];
        sigma2 = optimalVar[0];
        correlationFunction->setParameters(optimalParamValues);
        if(bHas_measurement_error && bVariance_in_factor)
            alpha  = optimalAlpha[0];
        for(int t=1;t<nbthreads;t++){
            if(optimalCost[t] < cost){
                cost=optimalCost[t];
                sigma2 = optimalVar[t];
                correlationFunction->setParameters(optimalParamValues+t*nbInputParams);
                if(bHas_measurement_error && bVariance_in_factor)
                    alpha  = optimalAlpha[t];
            }
        }
    }
    
    //---------------------------------------------------------------------//
    // Finally, we are cleaning the memory

    //---------------------------------------------------------------------//
    delete[] optimalVar; optimalVar=NULL;
    delete[] optimalCost; optimalCost=NULL;
    delete[] optimalAlpha; optimalAlpha=NULL;
    delete[] optimalParamValues; optimalParamValues=NULL;
    // We delete the TList objects
    cleanAttributeList(paramInputList);    
    cleanAttributeList(paramOutputList);
}
#endif

  void TGPBuilder::cleanAttributeList(TList *inputlist)
  {
    TList *garbageCollector = new TList();
    TIter nextIn(inputlist);
    TAttribute* att;

    // Here, we use the "TAttribute::_nshare" variable which enumerates how many uses of
    // the TAttribute object are active. It is set to 1 at the construction, and
    // incremented when we add the attribute to a data server.
    // If the attribute isn't shared, we store it into the garbageCollector list 
    // for clean deletion. If it is shared, we simply decrement the sharing by one 
    // and leave the deletion to someone else.
    while( (att = (TAttribute*)nextIn() ) )
      {
if(att->getShare() == 1)
  garbageCollector->Add(att);
else
  att->delShare();
      }
    
    // We give ownership of the attributes to the TList, which means they will 
    // be deleted when the TList will be destroyed.
    garbageCollector->SetOwner();
    
    // We delete the TList objects
    delete inputlist;
    delete garbageCollector;

  }



// -----------------------------------------------------------------------
//  Log
// -----------------------------------------------------------------------
void TGPBuilder::printLog(Option_t *option)
{
    cout << "*******************************" << endl;
    cout << " ** TGPBuilder::printLog[" << option << "]" << endl;
    cout << "*******************************" << endl;
}

void TGPBuilder::LSort(Double_t *costValues, Int_t *minIndexes, Int_t nbthreads, Int_t n)
{
  std::vector<double> sval;
  std::vector<double>::iterator it;

// put the nbthreads first values in the vector
  for(int i=0;i<nbthreads;i++)
    sval.push_back(costValues[i]);

// sort the vector in ascendant order
  std::sort(sval.begin(),sval.end());

// put in the vector the ordered nbthreads smallest values of the array
  for(int i=nbthreads;i<n;i++){
    it = sval.end();
    it--;
    if( costValues[i] < *it ){
      sval.erase(it);
      for (it=sval.begin(); it!=sval.end(); ++it)
        if( costValues[i] < *it ) break;
      sval.insert(it,costValues[i]);
    }
  }

  int j=0;
// calculate the index of the nbthreads smallest values in the input array
  for (it=sval.begin(); it!=sval.end(); ++it)
    for(int i=0;i<n;i++)
      if( fabs(costValues[i] - *it) < epsilon ){
        minIndexes[j++] = i;
        break;
      }

}


} // Fin du namespace Modeler
} // Fin du namespace URANIE

