//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TGPCostFunction.h
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 \date   

 \brief Interface de la classe URANIE::Modeler::TGPCostFunction

 */
/**

 * @class URANIE::Modeler::TGPCostFunction
 *
 * @brief Description of the class TGPCostFunction.
 *
 * This class defines a cost function used to select the optimal hyper-parameters of 
 * a Gaussian Process (correlation lengths, variance, etc.). Its children are wrappers 
 * to the various cost functions provided by the "gaussian process Library" (gpLib), 
 * by Jean-Marc Martinez.
 *
 */
#ifndef TGPCOSTFUNCTION_H
#define TGPCOSTFUNCTION_H    

//// URANIE
#include "Modeler.h"
#include "TGPBuilder.h"
#include "TShareEval.h"

namespace URANIE
{
namespace Modeler
{
//class TGPBuilder;
class TGPCostFunction: public URANIE::Relauncher::TSimpleEval
{

    // Attributes
protected:
    Double_t* _H;    ///< pointer to the deterministic trend matrix (size: _n * _p)
    Double_t* _K;    ///< pointer to the covariance matrix (size: _n * _n)
    Double_t* _C;    ///< pointer to the correlation matrix (size: _n * _n)
    Double_t* _yObs; ///< pointer to the vector of observations outputs (size: _n)
    Int_t _n;    ///< number of observations
    Int_t _p;    ///< number of deterministic trend coefficients 

public:
    Bool_t _blog;          ///< Boolean to decide if the log information is shown or not.
    TGPBuilder* gpBuilder; ///< pointer to a gaussian process builder object.

    // Operations
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{            
    //! Standard constructor
    /*!
     A TGPCostFunction object need to be associated with a TGPBuilder 
     object which will provide the covariance matrix and the observations.
     \param gpb (TGPBuilder*): pointer to a Gaussian Process Builder object.
     */
    TGPCostFunction(TGPBuilder* gpb);

    //! Default destructor
    virtual ~TGPCostFunction();
    //@}

    //---------------------------------------------
    /** @name Get and Set methods
     **/
    //@{          
    //! Return a pointer to the GP Builder
    TGPBuilder* getBuilder()
    {
        return gpBuilder;
    }
    
    //! Return a pointer to the GP Builder covariance matrix
    Double_t* getCovarianceMatrix()
    {
        return _K;
    }
    
    //! Return a pointer to the GP Builder correlation matrix
    Double_t* getCorrelationMatrix()
    {
        return _C;
    }
    
    //! Return a pointer to the observations outputs
    Double_t* getObservationVector()
    {
        return _yObs;
    }
    
    //! Return the number of observations
    Int_t getNObs()
    {
        return _n;
    }

    void setBuilder(TGPBuilder* gpb)
    {
        gpBuilder = gpb;
    }
    //@}

    //---------------------------------------------
    /** @name Estimate
     **/
    //@{            
    //! Evaluate the cost function (implemented in the children classes)
    virtual Int_t eval(Double_t* theta, Double_t* res, int=0) = 0;
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    void setLog()
    {
        _blog = kTRUE;
    }
    void unsetLog()
    {
        _blog = kFALSE;
    }
    void changeLog()
    {
        _blog = !_blog;
    }
    Bool_t getLog()
    {
        return _blog;
    }
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Modeler::TGPCostFunction, ID_MODELER)
    //Classe de

protected:
    void duplicateGPB();
};

} // Fin du namespace Modeler
} // Fin du namespace URANIE
#endif
