//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TGPMLCostFunction.cxx
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 \date   

 \brief Implementation of the class  URANIE::Modeler::TGPMLCostFunction

 */

//// ROOT
#include "TObjArray.h"

//// URANIE
#include "TGPMLCostFunction.h"
#include "TAttribute.h"

using std::cout;
using std::endl;
using namespace URANIE::DataServer;

ClassImp (URANIE::Modeler::TGPMLCostFunction)

namespace URANIE
{
namespace Modeler
{
// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------
TGPMLCostFunction::TGPMLCostFunction(TGPBuilder* gpb): TGPCostFunction(gpb)
{
}

// -----------------------------------------------------------------------
//  Destructor
// -----------------------------------------------------------------------

TGPMLCostFunction::~TGPMLCostFunction()
{
    if (_blog)
        cout << "detruit TGPMLCostFunction" << endl;
}

URANIE::Relauncher::TEval *TGPMLCostFunction::rebuild(int rank, Bool_t chef)
{
    UNUSED(rank,chef);
    TGPMLCostFunction* ret = new TGPMLCostFunction(*this);
    ret->duplicateGPB();
    return (URANIE::Relauncher::TEval *) ret;
}
Bool_t TGPMLCostFunction::unbuild()
{
    delete gpBuilder;
    return kTRUE;
}

// -----------------------------------------------------------------------
//  Eval
// -----------------------------------------------------------------------
Int_t TGPMLCostFunction::eval(vector<URANIE::DataServer::UEntry*> *in, vector<URANIE::DataServer::UEntry*> *out, int ncalc)
{
    UNUSED(ncalc);
    double *inp = new double[in->size()];
	double *outp = new double[out->size()];
    for(unsigned int ival=0; ival<in->size(); ival++) inp[ival] = *in->at(ival)->val.dou;

    int admis=eval(inp, outp);
    
    for(unsigned int ival=0; ival<out->size(); ival++) *out->at(ival)->val.dou = outp[ival];
	delete [] inp; delete [] outp;
    return ((admis!=0) ? getOutDimension(out) : admis);
}


Int_t TGPMLCostFunction::eval(Double_t* theta, Double_t* res, int ncalc)
{
    UNUSED(ncalc);
    Int_t returnCode;
    Int_t nbCorrFuncParams;

    if (gpBuilder->hasCorrFunction())
    {
        // Update the parameters of the correlation function            
        nbCorrFuncParams = gpBuilder->getCorrFunction()->getNParameters();
        gpBuilder->getCorrFunction()->setParameters(theta);
    }
    else
        nbCorrFuncParams = 0;
    
    if (gpBuilder->hasMeasurementError())
    {
        // In this case, the array theta contains nbCorrFuncParams+1 elements        
        if(gpBuilder->hasVarianceInFactor())
            // The last element is the parameter alpha
            gpBuilder->setAlpha(theta[nbCorrFuncParams]);
        else
            // The last element is the variance of the GP
            gpBuilder->setVariance(theta[nbCorrFuncParams]);
    }

    // Update the covariance matrix
    gpBuilder->computeCovarianceMatrix();

    // Compute the cost
    if (gpBuilder->hasTrend())
    {
        if (gpBuilder->hasVarianceInFactor())
            returnCode = gpMLh(_H, _K, _yObs, _n, _p, res, 1);
        else
        {
            // WARNING: for case 0, the size of res array is 1 whereas 
            //          gpMLh fills an array of size 2. Thus, 
            //          to avoid memory corruption, we must not pass res 
            //          to gpMLh.
            Double_t cost[2];
            returnCode = gpMLh(_H, _K, _yObs, _n, _p, cost, 0);
            res[0] = cost[0];
        }
    }
    else
    {
        if (gpBuilder->hasVarianceInFactor())
            returnCode = gpML0(_K, _yObs, _n, res, 1);
        else
        {
            // WARNING: for case 0, the size of res array is 1 whereas 
            //          gpML0 fills an array of size 2. Thus, to avoid 
            //          memory corruption, we must not pass res to gpML0. 
            Double_t cost[2];
            returnCode = gpML0(_K, _yObs, _n, cost, 0);
            res[0] = cost[0];
        }
    }

    if (returnCode != 0)
    {
        res[0] = 1e99;
        res[1] = 1e99;
    }

    return 1;
}

} // Fin du namespace Modeler
} // Fin du namespace URANIE

