//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TGaussianCorrFunction.h
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 \date   

 \brief Interface de la classe URANIE::Modeler::TGaussianCorrFunction

 */
/**

 * @class URANIE::Modeler::TGaussianCorrFunction
 *
 * @brief Description of the class TGaussianCorrFunction.
 *
 * This class defines the shape of the correlation between two sets of variables 
 * as a gaussian:
 *
 * \f[ c(x1,x2) = exp\left(-\frac{||x1-x2||^2}{lc^2}\right) \f]
 *
 * where \f$lc\f$ is the correlation length.
 *
 */
#ifndef TGAUSSIANCORRFUNCTION_H
#define TGAUSSIANCORRFUNCTION_H

//// URANIE
#include "TCorrelationFunction.h"

extern "C"
{
int gpCorGauss(double *X1, double *X2, int n1, int n2, int d, double *lc,
        double *C);
}
;

namespace URANIE
{
namespace Modeler
{
class TGaussianCorrFunction: public TCorrelationFunction
{

    // Attributes
private:
public:

    // Operations
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    /*!
     \param nbDim (Int_t): number of dimension of the input variables
     \param corrLength (Double_t*): array of the correlation lengths for the 
            function (size: nbDim). If NULL, the correlation lengths are 
            initialised to 0.1. (default = NULL).             
     */
    TGaussianCorrFunction(Int_t nbDim, Double_t* corrLength = NULL);

    //! Default destructor
    virtual ~TGaussianCorrFunction();
    //@}

    //---------------------------------------------
    /** @name Get and Set methods
     **/
    //@{  

    //! Sets the parameters of the gaussian correlation function
    /*!
     \param newParams (Double_t*): array containing the parameters of the correlation function. 
     If the number of dimension of the inputs is \f$nbDim\f$, the array's 
     size is \f$nbDim\f$. It is supposed to be of the form:
     \f[ newParams[nbDim] = {lc[0],..., lc[nbDim-1]} \f]
     where \f$lc\f$ is the correlation length parameter (array of size \f$nbDim\f$).
     */
    virtual void setParameters(Double_t* newParams);

    //! Sets the correlation lengths
    /*!
     \param corrLength (Double_t*): array containing the correlation length parameter.
     If the number of dimension of the inputs is \f$nbDim\f$, the array's 
     size is \f$nbDim\f$.
     */
    void setCorrLengths(Double_t* corrLength)
    {
        setParameters(corrLength);
    }

    //@}

    //---------------------------------------------
    /** @name Tools
     **/
    //@{           
    //! Return a clone of the correlation function
    /*!
    The clone must be destroyed by the object (or user) calling the function.                
    */
    virtual TCorrelationFunction* clone();

    //! Compute the correlation between x1 and x2 
    /*!
     The order of the indices in the flat arrays are the following:
     x[i+j*d] = X[i,j]
     where X is a matrix of size (d x n), i is the line index, and j the 
     column index. The results are stored in the flat array C.  
     \warning Be aware that the matrix corresponding to flat arrays x1 and 
     x2 have input dimensions as LINE, not columns. This means 
     that for a set of point {(a1,b1),(a2,b2),(a3,b3)}, the 
     corresponding matrix X will be:
     \f$ 
     \left( 
     \begin{array}{ccc}
     a1 & a2 & a3 \\
                           b1 & b2 & b3 
     \end{array}
     \right)
     \f$                     
     and the corresponding flat array x = {a1,b1,a2,b2,a3,b3}.
     \param n1 (Int_t): number of elements of x1
     \param x1 (Double_t*): data set x1 represented as a flat array (size: d * n1)
     \param n2 (Int_t): number of elements of x2
     \param x2 (Double_t*): data set x2 represented as a flat array (size: d * n2)
     \param C  (Double_t*): correlation matrix represented as a flat array (size: n1 * n2)
     
     \exception UErrorExceptions if the gpLib function returns 1 (evaluation failed)
     */
    virtual void eval(Int_t n1, Double_t* x1, Int_t n2, Double_t* x2,
            Double_t* C);

    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Modeler::TGaussianCorrFunction, ID_MODELER)
    //Classe de
};

} // Fin du namespace Modeler
} // Fin du namespace URANIE
#endif
