//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TKernel.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   2012-11-14 12:54:45
 
 \brief Interface de la classe URANIE::Modeler::TKernel
 
 */

//ROOT
#include "TDecompSVD.h"
#include "TLeaf.h"
#include "TMath.h"

// Uranie
#include "TKernel.h"

ClassImp (URANIE::Modeler::TKernel)

namespace URANIE
{
namespace Modeler
{
///////////////////////////////////////////////////////////
///  Constructors
///////////////////////////////////////////////////////////
TKernel::TKernel(URANIE::DataServer::TDataServer *tds, TString architecture, Option_t *option):
    TModeler(tds, architecture, option), _nKernelType(kUnknown),
    _dParameter1(1.0), _dParameter2(0.0), _dParameter3(1.0), _nParameter(0),
    _matX(0,0), _vecY(0), _matK(0, 0), _dLambda(1.e-6)
{
    if (_blog) {
        cout << "construit TKernel tds[" << _tds->GetName() << "] architecture["
             << architecture << "] Option [" << option << "]" << endl;
    }

    try {
        if (_nY > 1)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form(
                                                           "TKernel constructor\n The number of output must be equal to 1.\nHere, we have %d output attributes with [%s]",
                                                           _nY, _sOutputAttributes.Data()));

    } catch (URANIE::Exceptions::UErrorExceptions& ue) {
        ue.printMessage();
        throw ue;
    }
    TString opt = option;
    opt.ToLower();

    fillData();
}

TKernel::TKernel(URANIE::DataServer::TDataServer *tds, const char *varexpinput,
                 const char *varexpoutput, Option_t * option) :
    TModeler(tds, varexpinput, varexpoutput, option), _nKernelType(kUnknown),
    _dParameter1(1.0), _dParameter2(0.0), _dParameter3(1.0), _nParameter(0),
    _matX(0,0), _vecY(0), _matK(0, 0), _dLambda(1.e-6)
{

    try {
        if (_nY > 1)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form(
                                                           "TKernel constructor\n The number of output must be equal to 1.\nHere, we have %d output attributes with [%s]",
                                                           _nY, _sOutputAttributes.Data()));

    } catch (URANIE::Exceptions::UErrorExceptions& ue) {
        ue.printMessage();
        throw ue;
    }

    _sFileName = Form("kernel_%s_%d", _sOutputAttributes.Data(), _nKernelType);
    _sFunctionName = Form("kernel_%d_%s", _nKernelType, _sOutputAttributes.Data());
    SetName(Form("Modeler[TKernel]Tds[%s]Predictor[%s]Target[%s]",
                 _tds->GetName(), varexpinput, _sOutputAttributes.Data()));
    SetTitle(Form("TKernel as modeler for TDS[%s] with Target is[%s] versus[%s]",
                  _tds->GetName(), _sOutputAttributes.Data(), varexpinput));

    fillData();
}

///////////////////////////////////////////////////////////
///  Destructor
///////////////////////////////////////////////////////////
TKernel::~TKernel()
{
    if (_blog)
        cout << " *** destroy TKernel(" << GetName() << ") Title[" << GetTitle() << "] *** " << endl;
}

///////////////////////////////////////////////////////////
///  setKernel
///////////////////////////////////////////////////////////
void TKernel::setKernel(EKernelType nKernel)
{
    _nParameter = 0;
    _nKernelType = nKernel;
}

void TKernel::setKernel(EKernelType nKernel, Double_t dparam1)
{
    _nParameter = 1;
    _nKernelType = nKernel;
    _dParameter1 = dparam1;
}

void TKernel::setKernel(EKernelType nKernel, Double_t dparam1, Double_t dparam2)
{
    _nParameter = 2;
    _nKernelType = nKernel;
    _dParameter1 = dparam1;
    _dParameter2 = dparam2;
}

void TKernel::setKernel(EKernelType nKernel, Double_t dparam1, Double_t dparam2,
                        Double_t dparam3)
{
    try {
        if (nKernel == kPoly) {
            _nParameter = 3;
            _nKernelType = nKernel;
            _dParameter1 = dparam1;
            _dParameter2 = dparam2;
            _dParameter3 = dparam3;
        } else {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form(
                                                           "TKernel::setKernel\n Just kPoly kernel have three parameter (%f, %f, %f).\n And we use the kernel [%d]",
                                                           dparam1, dparam2, dparam3, nKernel));
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue) {
        ue.printMessage();
        throw ue;
    }
}

///////////////////////////////////////////////////////////
///  fillData
///////////////////////////////////////////////////////////
void TKernel::fillData()
{
    if (_blog) {
        cout << endl << "**************************************" << endl;
        cout << " ** TKernel::fillData" << endl;
    }

    Int_t nP = _nX;
    if (_bIntercept)
        nP++;

    _matX.ResizeTo(_nS, nP);
    _vecY.ResizeTo(_nS);

    // Get The list of Leaf (from TTree) of the list of attributes ( _listOfInputAttributes)
    TList *listOfInputLeaves = new TList();
    TIter next(_listOfInputAttributes);
    if (_blog)
        cout << " ** input" << endl;
    while (URANIE::DataServer::TAttribute *att =
           (URANIE::DataServer::TAttribute *) next())
    {
        listOfInputLeaves->Add(_tds->getTuple()->GetLeaf(att->GetName()));
        if (_blog)
        {
            _tds->getTuple()->GetLeaf(att->GetName())->Print();
            att->printLog();
        }
    }

    TLeaf *leafY = NULL;
    TList *listOfOutputLeaves = new TList();


    if (_nY == 1) {
        // Case of the output (only one attribut) : Fill the vector Y
        leafY = _tds->getTuple()->GetLeaf(_sOutputAttributes.Data());
        if (_blog) {
            cout << " ** output" << endl;
            _tds->getTuple()->GetLeaf(_sOutputAttributes)->Print();
            _tds->getAttribute(_sOutputAttributes)->printLog();
        }
    } else {
        TIter nextouet(_listOfOutputAttributes);
        if (_blog)
            cout << " ** Output" << endl;
        while (URANIE::DataServer::TAttribute *att =
               (URANIE::DataServer::TAttribute *) nextouet())
        {
            listOfOutputLeaves->Add(_tds->getTuple()->GetLeaf(att->GetName()));
            if (_blog) {
                _tds->getTuple()->GetLeaf(att->GetName())->Print();
                att->printLog();
            }
        }
    }

    // Fill the matrix and vector of data
    Int_t ivar;
    for (Int_t ipat = 0; ipat < _nS; ipat++) {
        _tds->getTuple()->GetEntry(ipat);
        if ((ipat < 5) && (_blog))
            _tds->getTuple()->Show();

        if (_blog)
            cout << " ** ipat[" << ipat << "/" << _nS << "]" << endl;

        // Case of the input attributes : Fill the matrix Aw
        ivar = 0;
        if (_bIntercept)
            _matX(ipat, ivar++) = 1.0;
        TIter nextLeaf(listOfInputLeaves);
        while (TLeaf *leaf = (TLeaf *) nextLeaf())
            _matX(ipat, ivar++) = leaf->GetValue(0);

        // Case of output attribute : Fill the vector _vecY
        ivar = 0;
        if (_nY == 1) {
            _vecY(ipat) = leafY->GetValue(0);
        }
    }

    delete listOfInputLeaves;
    listOfInputLeaves = NULL;

    delete listOfOutputLeaves;
    listOfOutputLeaves = NULL;
    
    if (_blog) {
        cout << " ** Matrix _matX(" << _matX.GetNrows() << "," << _matX.GetNcols() << ")" << endl;
        _matX.Print();
        cout << " ** Vector _vecY(" << _vecY.GetNrows() << ")" << endl;
        _vecY.Print();
        cout << " ** End of TKernel::fillData" << endl;
    }
}

///////////////////////////////////////////////////////////
///  Compute the K Matrix
///////////////////////////////////////////////////////////
void TKernel::computeK(Option_t * option)
{
    TString soption = option;
    soption.ToLower();

    _matK.ResizeTo(_nS, _nS);
    _matK.Zero();

    Double_t da, db;
    for (Int_t i = 0; i < _nS; i++) {
        for (Int_t j = i; j < _nS; j++) {
            switch (_nKernelType) {
            case kLinear:
                da = 0.0;
                for (Int_t k = 0; k < _nX; k++) {
                    da += _matX(i, k) * _matX(j, k);
                }
                _matK(i, j) = _matK(j, i) = da;
                break;
            case kPoly:
                da = 0.0;
                for (Int_t k = 0; k < _nX; k++) {
                    da += _matX(i, k) * _matX(j, k);
                }
                _matK(i, j) = _matK(j, i) = TMath::Power( _dParameter1 * da + _dParameter2, _dParameter3);
                break;
            case kSigmoid:
                da = 0.0;
                for (Int_t k = 0; k < _nX; k++) {
                    da += _matX(i, k) * _matX(j, k);
                }
                _matK(i, j) = _matK(j, i) = TMath::TanH( _dParameter1 * da + _dParameter2);
                break;
            case kRBF:
                da = 0.0;
                for (Int_t k = 0; k < _nX; k++) {
                    da += (_matX(i, k) - _matX(j, k)) * (_matX(i, k) - _matX(j, k));
                }
                _matK(i, j) = _matK(j, i) = TMath::Exp( -1.0 * _dParameter1 * da);
                break;
            case kLaplace:
                da = 0.0;
                for (Int_t k = 0; k < _nX; k++) {
                    da += TMath::Abs(_matX(i, k) - _matX(j, k));
                }
                _matK(i, j) = _matK(j, i) = TMath::Exp( -1.0 * _dParameter1 * da);
                break;
            case kSpline1:
                _matK(i, j) = _matK(j, i) = 1.0 + TMath::Min(_matX(i, 1), _matX(j, 1));
                break;
            case kBernoulli1:
                da = TMath::Min(_matX(i, 1), _matX(j, 1));
                db = TMath::Min(1.0 - _matX(i, 1), 1.0 - _matX(j, 1));
                _matK(i, j) = _matK(j, i) = 5.0 / 6.0 + 0.5 * (da * da + db * db);
                break;
	    default:
		break;
            } // End Of switch
        } // End Of j
        _matK(i, i) += _dLambda;
    } // End Of i
}

///////////////////////////////////////////////////////////
///  estimate
///////////////////////////////////////////////////////////
void TKernel::estimate(Option_t * option)
{
    TString soption = option;
    soption.ToLower();
    if (_blog) {
        cout << endl << "**************************************" << endl;
        cout << " ** TKernel::estimate" << endl;
    }

    computeK();

    // Compute the linear regression
    _parameterValues.ResizeTo(_nS);
    Bool_t ok = kFALSE;
    TDecompSVD svd(_matK);
    _parameterValues = svd.Solve(_vecY, ok);

    if (ok) {
        if (_blog) {
            cout << " ** Vector of parameters solution ("
                 << _parameterValues.GetNrows() << ")" << endl;
            _parameterValues.Print();
        }
    }

    // Fill the vector of lapha
    TString salpha = TString("__alpha__");
    _tds->addAttribute(salpha);
    Double_t dalpha = 0.0;
    TBranch *tbalpha = _tds->getTuple()->Branch(salpha, &dalpha, Form("%s/D", salpha.Data()));
    for (Int_t i = 0; i < _nS; i++) {
        dalpha = _parameterValues[i];
        tbalpha->Fill();
    }

    _tds->getTuple()->ResetBranchAddress(tbalpha);
    _tds->saveTuple("UPDATE");


    if (_blog) {
        cout << " ** FINAL ::  Vector of parameters solution ("
             << _parameterValues.GetNrows() << ")" << endl;
        _parameterValues.Print();
    }
}

///////////////////////////////////////////////////////////
///  Export Functions
///////////////////////////////////////////////////////////
void TKernel::exportModelCplusplus(std::ofstream * sourcefile) const
{
    cout << " ** TKernel::exportModelCplusplus not completely implemented** " << endl;
    // Model Quality 
    *sourcefile << "  //" << endl;
    *sourcefile << "  //     QUALITY  : " << endl;
    *sourcefile << "  //" << endl;
    *sourcefile << "  //" << endl;
    *sourcefile << "  ////////////////////////////// " << endl << endl;

    // Learning database
    *sourcefile << endl;
    *sourcefile << "  // --- Learning database" << endl;

    *sourcefile << "  double X[" << _nS << "][" << _nX << "] = {" << endl
                << "  ";
    for (Int_t ivar = 0; ivar < _nS; ivar++)
    {
        *sourcefile << _matX(ivar, 1) << ", ";
        if ((ivar + 1) % 5 == 0)
            *sourcefile << endl << "  ";
    }
    if (_nS % 5 != 0)
        *sourcefile << endl;
    *sourcefile << "  };" << endl;

    // Vector of parameter alpha
    *sourcefile << endl;
    *sourcefile << "  // --- Vector alpha of parameters" << endl;

    *sourcefile << "  double Kernel_alpha[" << _nS << "] = {" << endl << "  ";
    for (Int_t ivar = 0; ivar < _nS; ivar++)
    {
        *sourcefile << _parameterValues[ivar] << ", ";
        if ((ivar + 1) % 5 == 0)
            *sourcefile << endl << "  ";
    }
    if (_nS % 5 != 0)
        *sourcefile << endl;
    *sourcefile << "  };" << endl;

    // Loop
    *sourcefile << endl;
    *sourcefile << "  // --- Initial values " << endl;
    *sourcefile << "  res[0] = 0.0;" << endl;
    if (_nKernelType == kBernoulli1)
    {
        *sourcefile << "  double da = 0.0;" << endl;
        *sourcefile << "  double db = 0.0;" << endl;
    }
    else
    {
        *sourcefile << "  double ddot = 0.0;" << endl;
    }
    *sourcefile << endl;

    *sourcefile << "  // --- Loop " << endl;
    *sourcefile << "  for(int i = 0; i < " << _nS << "; i++) {" << endl;
    switch (_nKernelType)
    {
    case kLinear:
        *sourcefile << "    ddot = 0.0;" << endl;
        *sourcefile << "    for(int k = 0; k < " << _nX << "; k++) {" << endl;
        *sourcefile << "        ddot += param[k]*X[i][k];" << endl;
        *sourcefile << "    }" << endl;
        *sourcefile << "    res[0] += Kernel_alpha[i] * ddot;" << endl;
        break;
    case kPoly:
        *sourcefile << "    ddot = 0.0;" << endl;
        *sourcefile << "    for(int k = 0; k < " << _nX << "; k++) {" << endl;
        *sourcefile << "        ddot += param[k]*X[i][k];" << endl;
        *sourcefile << "    }" << endl;
        *sourcefile << "    res[0] += Kernel_alpha[i] * TMath::Power("
                    << _dParameter1 << " * ddot + " << _dParameter2 << " , "
                    << _dParameter3 << ");" << endl;
        break;
    case kSigmoid:
        *sourcefile << "    ddot = 0.0;" << endl;
        *sourcefile << "    for(int k = 0; k < " << _nX << "; k++) {" << endl;
        *sourcefile << "        ddot += param[k]*X[i][k];" << endl;
        *sourcefile << "    }" << endl;
        *sourcefile << "    res[0] += Kernel_alpha[i] * TMath::TanH("
                    << _dParameter1 << " * ddot + " << _dParameter2 << ");" << endl;
        break;
    case kRBF:
        *sourcefile << "    ddot = 0.0;" << endl;
        *sourcefile << "    for(int k = 0; k < " << _nX << "; k++) {" << endl;
        *sourcefile
            << "        ddot += ( param[k] - X[i][k]) * ( param[k] - X[i][k]);"
            << endl;
        *sourcefile << "    }" << endl;
        *sourcefile << "    res[0] += Kernel_alpha[i] * TMath::Exp( -1.0 * "
                    << _dParameter1 << " * ddot );" << endl;
        break;
    case kLaplace:
        *sourcefile << "    ddot = 0.0;" << endl;
        *sourcefile << "    for(int k = 0; k < " << _nX << "; k++) {" << endl;
        *sourcefile << "        ddot += TMath::Abs( param[k] - X[i][k] );"
                    << endl;
        *sourcefile << "    }" << endl;
        *sourcefile << "    res[0] += Kernel_alpha[i] * TMath::Exp( -1.0 * "
                    << _dParameter1 << " * ddot );" << endl;
        break;
    case kSpline1:
        *sourcefile
            << "    res[0] += Kernel_alpha[i] * ( 1.0 + TMath::Min( param[0], X[i][0] ));"
            << endl;
        break;
    case kBernoulli1:
        *sourcefile << "    da = TMath::Min( param[0] , X[i][0]);" << endl;
        *sourcefile << "    db = TMath::Min( 1.0 - param[0], 1.0 - X[i][0]);"
                    << endl;
        *sourcefile
            << "    res[0] += Kernel_alpha[i] * ( 5.0/6.0 + 0.5 * ( da * da + db * db));"
            << endl;
        break;
    default:
	break;
    }
    *sourcefile << "  }" << endl;

}

void TKernel::exportModelFortran(std::ofstream * sourcefile) const
{
    UNUSED(sourcefile);
    cout << " ** TLinearRegression::exportModelFortran not yet implemented ** " << endl;
}

///////////////////////////////////////////////////////////
///  Log
///////////////////////////////////////////////////////////
void TKernel::printLog(Option_t *option)
{
    cout << "*******************************" << endl;
    cout << " ** TKernel::printLog[" << option << "]" << endl;
    cout << " ** _nKernelType[" << _nKernelType << "]  _dLambda[" << _dLambda
         << "]" << endl;
    cout << " ** _nParameter[" << _nParameter << "]" << endl;
    if (_nParameter > 0)
        cout << " ** _dParameter1[" << _dParameter1 << "]" << endl;
    if (_nParameter > 1)
        cout << " ** _dParameter2[" << _dParameter2 << "]" << endl;

    cout << " ** Vector of parameters (" << _parameterValues.GetNrows() << ")"
         << endl;
    _parameterValues.Print();
    cout << " ** Matrix _matK(" << _matK.GetNrows() << "," << _matK.GetNcols()
         << ")" << endl;
    _matK.Print();
    TModeler::printLog();
    cout << "*******************************" << endl;
}

} // End of namespace Modeler
} // End of namespace URANIE
