//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TMaternIICorrFunction.cxx
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 \date   

 \brief Implementation of the class  URANIE::Modeler::TMaternIICorrFunction

 */
#include <iostream>

//// ROOT
#include "TMath.h"

//// URANIE
#include "TMaternIICorrFunction.h"

using std::cout;
using std::endl;

ClassImp (URANIE::Modeler::TMaternIICorrFunction)

namespace URANIE
{
namespace Modeler
{
// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------
TMaternIICorrFunction::TMaternIICorrFunction(Int_t nbDim, Double_t nu,
        Double_t* corrLength)
{
    nbDimensions  = nbDim;
    nbParameters  = nbDimensions + 1;
    nbCorrLengths = nbDimensions;
    parameters    = new Double_t[nbParameters];
    _lCorrLengths = new Double_t[nbCorrLengths];

    parameters[0] = nu;
    _dNu = nu;
    
    if(corrLength)
    {
        for (Int_t i = 1; i < nbParameters; i++)
        {
            parameters[i] = corrLength[i - 1];
            _lCorrLengths[i - 1] = corrLength[i - 1];
        }
    }
    else
    {
        for (Int_t i = 1; i < nbParameters; i++)
        {
            parameters[i] = 0.1;
            _lCorrLengths[i - 1] = 0.1;
        }
    }
}

// -----------------------------------------------------------------------
//  Destructor
// -----------------------------------------------------------------------

TMaternIICorrFunction::~TMaternIICorrFunction()
{
    delete[] parameters;
    parameters = NULL;
    delete[] _lCorrLengths;
    _lCorrLengths = NULL;
    if (_blog)
        cout << "detruit TMaternIICorrFunction" << endl;
}

// -----------------------------------------------------------------------
//  Set methods
// -----------------------------------------------------------------------

void TMaternIICorrFunction::setParameters(Double_t* newParams)
{
    parameters[0] = newParams[0];
    _dNu = parameters[0];
    for (Int_t i = 1; i < nbParameters; i++)
    {
        parameters[i] = newParams[i];
        _lCorrLengths[i - 1] = newParams[i];
    }
}

void TMaternIICorrFunction::setParameterNu(Double_t nu)
{
    parameters[0] = nu;
    _dNu = nu;
}

void TMaternIICorrFunction::setCorrLengths(Double_t *corrLength)
{
    for (Int_t i = 1; i < nbParameters; i++)
    {
        parameters[i] = corrLength[i - 1];
        _lCorrLengths[i - 1] = corrLength[i - 1];
    }
}

// -----------------------------------------------------------------------
// Clone
// -----------------------------------------------------------------------

TCorrelationFunction* TMaternIICorrFunction::clone()
{
    TMaternIICorrFunction* cfClone = new TMaternIICorrFunction(nbDimensions, _dNu, _lCorrLengths);
    return cfClone;
}

// -----------------------------------------------------------------------
// Evaluation
// -----------------------------------------------------------------------

void TMaternIICorrFunction::eval(Int_t n1, Double_t* x1, Int_t n2, Double_t* x2, Double_t* C)
{
    if (_blog)
        std::cout << "** TMaternIICorrFunction::eval **: nu   = " << _dNu
                << "; l[0] = " << _lCorrLengths[0] << std::endl;

    Int_t errorCode = gpCorMaternII(x1, x2, n1, n2, nbDimensions, _lCorrLengths, _dNu, C);

    try
    {
        if (errorCode == 1)
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TMaternIICorrFunction::eval: bessik function failed to converge.");
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

}

// -----------------------------------------------------------------------
//  Log
// -----------------------------------------------------------------------
void TMaternIICorrFunction::printLog(Option_t *option)
{
    cout << "*******************************" << endl;
    cout << " ** TMaternIICorrFunction::printLog[" << option << "]" << endl;
    cout << " ** Number of parameters = [" << nbParameters << "]" << endl;
    cout << " ** regularity parameter = [" << parameters[0] << "]" << endl;
    for (Int_t i = 1; i < nbParameters; i++)
        cout << " ** correlation length " << i << " = [" << parameters[i] << "]"
                << endl;
    cout << "*******************************" << endl;
}

} // Fin du namespace Modeler
} // Fin du namespace URANIE

