//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TModeler.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   2008-02-07 17:14:24
 
 \brief Interface of the class URANIE::Optimize::TModeler

 \par Modifications
 \li F. Gaudier 17/06/2010
 Take into account a new option "nointercept" for exclude the intercept in certain models (as linear regression)
 By default, the intercept is TRUE.
 \li Cedric Forgues (C.S.) 2010/01/11
 TMultiLayerPerceptron.h provisoirement retir� + Methods using TMultiLayerPerceptron getANN()
 *********************************************************************    
 */
#ifndef TMODELER_H
#define TMODELER_H

#include <iostream>

//ROOT
#include "Rtypes.h"
#include "TNamed.h"
//Added to make up for TMultiLayerPerceptron comment
#include "TVector.h"
//Provisoirement retir�
#ifndef WIN32
#include "TMultiLayerPerceptron.h"
#endif
//Uranie
#include "Modeler.h"
#include "TDataServer.h"
#include "UExceptions.h"

namespace URANIE
{
namespace Modeler
{
class TModeler: public TNamed
{
public:
    Bool_t _blog;                             ///< Boolean for edit the log
    Bool_t _bdrawProgressBar; ///< Boolean to know if the progress bar has to be drawn
    Bool_t _bStoreYHat; ///< Boolean to specify if we add the \hat{} attribute in the TDS [default kTRUE]
    URANIE::DataServer::TDataServer *_tds;     //! Pointer to a TDS
    //TMultiLayerPerceptron* _mlp;               //!< Perceptron associated

    Int_t _nS;                            ///< The size of the TDataServer

    TString _sInputAttributes; ///< The string of input attributes (separated by the character ":"
    TString _sOutputAttributes; ///< The string of output attributes (separated by the character ":"
    TString _sIntermediateArchitecture; ///< The intermediate (between the first and last character \em ",") information
    TList *_listOfInputAttributes;             ///< The list of input Leaf
    TList *_listOfOutputAttributes;            ///< The list of output Leaf

    Int_t _nX;                            ///< The number of input attributes
    Int_t _nY;                            ///< The number of output attributes

    TString _sFileName; ///< The name of the file when export the model without extension
    TString _sFunctionName;  ///< The name of the function when export the model

    TVectorD _parameterValues;                 ///< The list of parameters
    Double_t _dr2;                             ///< The R2 quality
    Bool_t _bIntercept; ///< For certain models, add the intercept (Default is TRUE)
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Constructor with a dataserver
    TModeler(URANIE::DataServer::TDataServer *tds, TString architecture,
            Option_t *option = "");
    //! Default constructor with input and output attributes
    TModeler(URANIE::DataServer::TDataServer *tds, const char *varexpinput,
            const char *varexpoutput, Option_t * option = "");

    //! Default destructor
    virtual ~TModeler();
    //@}

    //---------------------------------------------
    /** @name Get and Set methods
     **/
    //@{
    //TMultiLayerPerceptron* getANN() { return _mlp;}
    //! Return the parameter vector of the model
    TVectorD getParameters()
    {
        return _parameterValues;
    }
    //! Return the R2 quality
    Double_t getR2() const
    {
        return _dr2;
    }
    //! Return the name of input attributes
    const char * getInputName(Int_t i) const;
    //! Return the name of output attribute
    const char * getOutputName(Int_t i = 0);
    //! Gets the number of input
    Int_t getNInput();

    //! Gets the boolean attribut _bIntercept
    Bool_t getIntercept()
    {
        return _bIntercept;
    }
    //! Set the "draw progress bar" flag
    /*!
     The progress bar will be drawn if and only if the "draw progress bar"
     flag is true.
     */
    void setDrawProgressBar(Bool_t bbool = kTRUE)
    {
        _bdrawProgressBar = bbool;
    }

    //! Get the "draw progress bar" flag
    /*!
     \return (Bool_t) the "draw progress bar" flag (boolean)
     */
    Bool_t getDrawProgressBar()
    {
        return _bdrawProgressBar;
    }
    //@}

    //---------------------------------------------
    /** @name Training
     **/
    //@{
    void train(Option_t* option = "text");
    //@}

    //---------------------------------------------
    /** @name Export the model in extrenal langage
     **/
    //@{
    //! Export the model in an external file with a specified langage
    /*!
     It will be the main interface function to the export.
     \param lang (const char *) the langage to export (C|Fortran)
     \param file (const char *) the file to export the modeler. If empty, use the name of the model, plus the extension of the langage.
     \param name (const char *) the name of the function in the export file. If empty, use the name of the model
     \param option (Option_t *)[""] The option to pass
     */
    virtual void exportFunction(const char* lang, const char* file = "",
            const char* name = "", Option_t *option = "");
    //! Export the model in C++ langage in a file
    /*!
     \warning Don't use this method. Use the main method TModeler::exportFunction.
     \param sourcefile (ofstream *) the pointer to the file to export the modeler.
     */
    virtual void exportModelCplusplus(std::ofstream * sourcefile) const = 0;
    //! Export the model in Fortran langage in a file
    /*!
     \warning Don't use this method. Use the main method TModeler::exportFunction.
     \param sourcefile (ofstream *) the pointer to the file to export the modeler.
     */
    virtual void exportModelFortran(std::ofstream * sourcefile) const = 0;
    //! Export the model in  a PMML file
    /*!
     \warning Don't use this method. Use the main method TModeler::exportFunction.
     \param sourcefile (ofstream *) the pointer to the file to export the modeler.
     */
    virtual void exportModelPMML(const char* file = "", 
                                 const char* name = "",Option_t *option = "") const = 0;
    //! Export the model in Python langage in a file
    /*!
     \warning Don't use this method. Use the main method TModeler::exportFunction.
     \param sourcefile (ofstream *) the pointer to the file to export the modeler.
     */
    virtual void exportModelPython(std::ofstream * sourcefile) const = 0;
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    void setLog()
    {
        _blog = kTRUE;
    }
    void unsetLog()
    {
        _blog = kFALSE;
    }
    void changeLog()
    {
        _blog = _blog ? kFALSE : kTRUE;
    }
    Bool_t getLog()
    {
        return _blog;
    }
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Modeler::TModeler, ID_MODELER)
    //Classe de
};

} // Fin du namespace Modeler
} // Fin du namespace URANIE
#endif
