//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TNisp.cxx
 \author Maud Salvoldi (OCC)
 \date   23 March 2011
 
 \brief Implementation of the class URANIE::Modeler::TNisp
 */
#include <sstream>
#include <string>
#include <vector>

// ROOT
#include "Rtypes.h"
#include "TObjString.h"

// URANIE
#include "TNisp.h"
#include "TUniformDistribution.h"
#include "TBetaDistribution.h"
#include "TLogNormalDistribution.h"
#include "TExponentialDistribution.h"

// nisp
#include "nisp_gva.h"
#include "nisp_util.h"


ClassImp (URANIE::Modeler::TNisp)

namespace URANIE
{
namespace Modeler
{

//Constructor  from a TDataServer object
TNisp::TNisp(TDataServer *tds) :
       _tds(tds), _gvx(NULL),  _gvu(NULL), _nx(0),  _ns(0), _level(0), _listAttIn(NULL), 
       _bHasWeight(kFALSE), _sWeight("uWeight"), _blog(kFALSE), _bSavePsi(kFALSE), _bIsSample(kFALSE),
       _bError(kFALSE), _bSuppress(kFALSE)
{
    Int_t numberOfVariables;
    try
    {
        if (_tds == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TNisp::constructor ERROR : the pointer of \"tds\" is empty");
        //List of Input Attribut
        _listAttIn = new TDataSpecification();
        //stochastic variables
        _gvx = new SetRandomVariable();
        //uncertain parameters
        _gvu = new SetRandomVariable();
        //number of variable
        numberOfVariables = tds->getNAttributes();
        if (numberOfVariables < 1)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TNips::TNisp -  Warning: No attribut in the tds ");
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        if(_listAttIn != NULL)
        {
            delete _listAttIn;
            _listAttIn = NULL;
        }
        if(_gvx != NULL)
        {
            delete _gvx;
            _gvx = NULL;
        }
        if(_gvu != NULL)
        {
            delete _gvu;
            _gvu = NULL;
        }
        ue.printMessage();
        throw ue;
        exit(-1);
    }
    //build the group of uncertain parameter and of the stochastic araible associated
    TAttribute * att = 0;
    TStochasticAttribute *stoatt;
    for (Int_t i = 0; i < numberOfVariables; i++)
    {
        att = tds->getAttribute(i);
        stoatt = dynamic_cast<TStochasticAttribute *>(att);
        if (stoatt != NULL)
        {
            _listAttIn->addAttribute(att);
            _nx += 1;
            try
            {
                switch (stoatt->getLawType())
                {
                case TStochasticAttribute::kNormal:
                {
                    _gvu->AddRandomVariable(
                            new RandomVariable("Normale",
                                    stoatt->getTheoreticalMean (),
                                    stoatt->getTheoreticalStdDev()));
                    if (stoatt->getStochasticBasis()
                            == TStochasticAttribute::kHermite)
                        _gvx->AddRandomVariable(new RandomVariable("Normale"));
                    else if (stoatt->getStochasticBasis()
                            == TStochasticAttribute::kLegendre)
                        _gvx->AddRandomVariable(new RandomVariable("Uniforme"));
                    else if (stoatt->getStochasticBasis()
                            == TStochasticAttribute::kLaguerre)
                        _gvx->AddRandomVariable(
                                new RandomVariable("Exponentielle"));
                    else
                    {
                        throw URANIE::Exceptions::UErrorExceptions(
                                    __FILE__,
                                    __LINE__,
                                    Form("TNisp::TNisp\n The stochastique base [%d] for the Attribut [%s] is not authorized",
                                         stoatt->getStochasticBasis(),
                                         stoatt->GetName ()));
                    }
                    break;
                }
                case TStochasticAttribute::kUniform:
                {
                    _gvu->AddRandomVariable(
                            new RandomVariable("Uniforme",
                                    stoatt->getLowerBound(),
                                    stoatt->getUpperBound()));
                    if (stoatt->getStochasticBasis()
                            == TStochasticAttribute::kHermite)
                        _gvx->AddRandomVariable(new RandomVariable("Normale"));
                    else if (stoatt->getStochasticBasis()
                            == TStochasticAttribute::kLegendre)
                        _gvx->AddRandomVariable(new RandomVariable("Uniforme"));
                    else if (stoatt->getStochasticBasis()
                            == TStochasticAttribute::kLaguerre)
                        _gvx->AddRandomVariable(
                                new RandomVariable("Exponentielle"));
                    else
                    {
                        throw URANIE::Exceptions::UErrorExceptions(
                                    __FILE__,
                                    __LINE__,
                                    Form("TNisp::TNisp\n The stochastique base [%d] for the Attribut [%s] is not authorized",
                                         stoatt->getStochasticBasis(),
                                         stoatt->GetName ()));
                    }
                    break;
                }
                case TStochasticAttribute::kExponential:
                {
                    _gvu->AddRandomVariable(
                                new RandomVariable("Exponentielle",
                                                   ((TExponentialDistribution *)stoatt)->getParameterLambda(),
                                                   ((TExponentialDistribution *)stoatt)->getParameterMode()));
                    if (stoatt->getStochasticBasis()
                            == TStochasticAttribute::kHermite)
                        _gvx->AddRandomVariable(new RandomVariable("Normale"));
                    else if (stoatt->getStochasticBasis()
                             == TStochasticAttribute::kLegendre)
                        _gvx->AddRandomVariable(new RandomVariable("Uniforme"));
                    else if (stoatt->getStochasticBasis()
                             == TStochasticAttribute::kLaguerre)
                        _gvx->AddRandomVariable(
                                    new RandomVariable("Exponentielle"));
                    else
                    {
                        throw URANIE::Exceptions::UErrorExceptions(
                                    __FILE__,
                                    __LINE__,
                                    Form("TNisp::TNisp\n The stochastique base [%d] for the Attribut [%s] is not authorized",
                                         stoatt->getStochasticBasis(),
                                         stoatt->GetName ()));
                    }
                    break;
                }
                case TStochasticAttribute::kBeta:
                {
                    _gvu->AddRandomVariable(
                            new RandomVariable("Beta",
                                               ((TBetaDistribution *)stoatt)->getParameterAlpha(),
                                               ((TBetaDistribution *)stoatt)->getParameterBeta(),
                                               stoatt->getLowerBound(),
                                               stoatt->getUpperBound()));
                    if (stoatt->getStochasticBasis()
                            == TStochasticAttribute::kJacobi)
                        _gvx->AddRandomVariable(new RandomVariable("Beta",
                                                ((TBetaDistribution *)stoatt)->getParameterAlpha(),
                                                ((TBetaDistribution *)stoatt)->getParameterBeta()));
                    else if (stoatt->getStochasticBasis()
                            == TStochasticAttribute::kHermite)
                        _gvx->AddRandomVariable(new RandomVariable("Normale"));
                    else if (stoatt->getStochasticBasis()
                            == TStochasticAttribute::kLegendre)
                        _gvx->AddRandomVariable(new RandomVariable("Uniforme"));
                    else if (stoatt->getStochasticBasis()
                            == TStochasticAttribute::kLaguerre)
                        _gvx->AddRandomVariable(
                                new RandomVariable("Exponentielle"));
                    else
                    {
                        throw URANIE::Exceptions::UErrorExceptions(
                                    __FILE__,
                                    __LINE__,
                                    Form("TNisp::TNisp\n The stochastique base [%d] for the Attribut [%s] is not authorized",
                                         stoatt->getStochasticBasis(),
                                         stoatt->GetName ()));
                    }
                    break;
                }
                case TStochasticAttribute::kLogNormal:
                {
                    _gvu->AddRandomVariable(
                            new RandomVariable("LogNormale",
                                               stoatt->getTheoreticalMean () - ((TLogNormalDistribution *)stoatt)->getParameterXMin (),
                                               // TODO: FGA 2014/05/06 Why i must take one half of the normal sigma of the normal law underlying?
                                               ((TLogNormalDistribution *)stoatt)->getUnderlyingNormalParameterSigma(),
                                               ((TLogNormalDistribution *)stoatt)->getParameterXMin ()));
                    // ,          1.0 + ((TLogNormalDistribution *)stoatt)->getParameterXMin ()));
                    if (stoatt->getStochasticBasis()
                            == TStochasticAttribute::kHermite)
                        _gvx->AddRandomVariable(new RandomVariable("Normale"));
                    else if (stoatt->getStochasticBasis()
                            == TStochasticAttribute::kLegendre)
                        _gvx->AddRandomVariable(new RandomVariable("Uniforme"));
                    else if (stoatt->getStochasticBasis()
                            == TStochasticAttribute::kLaguerre)
                        _gvx->AddRandomVariable(
                                new RandomVariable("Exponentielle"));
                    else
                    {
                        throw URANIE::Exceptions::UErrorExceptions(
                                    __FILE__,
                                    __LINE__,
                                    Form("TNisp::TNisp\n The stochastique base [%d] for the Attribut [%s] is not authorized",
                                         stoatt->getStochasticBasis(),
                                         stoatt->GetName ()));
                    }
                    break;
                }
                case TStochasticAttribute::kLogUniform:
                {
                    _gvu->AddRandomVariable(
                            new RandomVariable("LogUniforme",
                                    stoatt->getLowerBound(),
                                    stoatt->getUpperBound()));
                    if (stoatt->getStochasticBasis()
                            == TStochasticAttribute::kHermite)
                        _gvx->AddRandomVariable(new RandomVariable("Normale"));
                    else if (stoatt->getStochasticBasis()
                            == TStochasticAttribute::kLegendre)
                        _gvx->AddRandomVariable(new RandomVariable("Uniforme"));
                    else if (stoatt->getStochasticBasis()
                            == TStochasticAttribute::kLaguerre)
                        _gvx->AddRandomVariable(
                                new RandomVariable("Exponentielle"));
                    else
                    {
                        throw URANIE::Exceptions::UErrorExceptions(__FILE__,
                            __LINE__,
                            Form("TNisp::TNisp\n The stochastique base [%d] for the Attribut [%s] is not authorized",
                                 stoatt->getStochasticBasis(),
                                 stoatt->GetName ()));
                    }
                    break;
                }
                default:
                {
                    throw URANIE::Exceptions::UErrorExceptions(
                                __FILE__,
                                __LINE__,
                                Form("TNisp::TNisp constructor\n The type of law for the Attribut [%s] not authorized for the NISP library [%d]",
                                     stoatt->GetName (),
                                     stoatt->getLawType()));
                }
                } //end switch
            } catch (URANIE::Exceptions::UErrorExceptions& ue)
            {
                ue.printMessage();

                // Clean
                delete _listAttIn;
                _listAttIn = NULL;
                delete _gvx;
                _gvx = NULL;
                delete _gvu;
                _gvu = NULL;

                throw ue;
            } catch (URANIE::Exceptions::UWarningExceptions& ue)
            {
                ue.printMessage();
                throw ue;
            }
        }
    } //end for
   
    if (_blog)
        _listAttIn->printLog();
    
}
//Destructor
TNisp::~TNisp()
{

    if (_listAttIn != NULL)
    {
        delete _listAttIn;
        _listAttIn = NULL;
    }
    if (_gvx != NULL)
    {
        delete _gvx;
        _gvx = NULL;
    }

    if (_gvu != NULL)
    {
        delete _gvu;
        _gvu = NULL;
    }

    _ns = 0;
    _nx = 0;
    _level = 0;
    _blog = kFALSE;
    _bIsSample = kFALSE;
    _bSavePsi = kFALSE;
    _bError = kFALSE;
    _bSuppress = kFALSE;
}

//Realisation of a sample by the method "type", of size or degree "n"     
void TNisp::generateSample(TString type, Int_t n, Option_t *option)
{
    if (_blog)
        cout << " ** TNisp::generateSample" << endl;
    //Verification of the parameters
    try
    {
        if (n < 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TNisp::generateSample - Error: Invalide value for level or size");
        // TODO: Compare TString in lower case, then adapt also the NISP library
        TString scmpType = type;
        scmpType.ToLower ();

        if ((type != "MonteCarlo") && (type != "Lhs") && (type != "QmcSobol")
                && (type != "Quadrature") && (type != "Petras")
                && (type != "SmolyakGauss") && (type != "SmolyakTrapeze")
                && (type != "SmolyakFejer")
                && (type != "SmolyakClenshawCurtis"))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TNisp::generateSample - Error: Unkown method ");
        if (_nx < 1)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TNISP::generateSample  - ERROR:  stochastic dimension < 1");
       
        _bHasWeight = kFALSE;    
            
        if((type == "Quadrature") || (type == "Petras")
                || (type == "SmolyakGauss") || (type == "SmolyakTrapeze")
                || (type == "SmolyakFejer")
                || (type == "SmolyakClenshawCurtis"))
        {
            TAttribute * attWeight = new TAttribute(_sWeight);
	    attWeight->delShare();
            _tds->addAttribute(attWeight); // add an attribute the weights of numerical integration
            _bHasWeight = kTRUE;
        }
            
        // Option to save the gvx variable in the tds
        TString soption = TString(option);
        TString name;
        Int_t nb;
        nb = 1;
        soption.ToLower();
        if (!soption.Contains("nosavegvx"))
        {
            if (soption.Contains("savegvx"))
            {
                if (_blog)
                    cout
                            << "TNisp: Saving the stochastic variable (_gvx) in the tds"
                            << endl;
                _bSavePsi = kTRUE;
                nb = 2;
                for (Int_t j = 0; j < _nx; j++)
                {
                    name = "psi_";
                    name = name + (_listAttIn->getAttribute(j))->GetName();
                    _tds->addAttribute(new TAttribute(name));
		    _tds->getAttribute(name)->delShare();
                }
            }
            //no save
            else
            {
                nb = 1;
                _bSavePsi = kFALSE;
                _bError = kTRUE;
            }
        }
        // Create the tuple if necessary
        try
        {
            _tds->getTuple();

            if (_bIsSample == kTRUE)
            {
                (_tds->getTuple())->~TDSNtupleD();
                _tds->createTuple();
                _bSuppress = kTRUE;
            }

        } catch (URANIE::Exceptions::UWarningExceptions& ue)
        {
            ue.printMessage();
            _tds->createTuple();
        }
        if (_blog)
            _tds->getTuple()->Print();
        //construction of the sample
        _level = n;
        _gvx->BuildSample((string) type, n);
        _ns = _gvx->GetSize();
        // Transfert to the tds
        _gvu->BuildSample(_gvx);
        Int_t k, i;
        Int_t nlen = 1;
        if( _bHasWeight )
            nlen = 2; // increase the length of x for the weigths values
        Double_t * x = new Double_t[_nx * nb + nlen];
        for (k = 0; k < _ns; k++)
        {
            x[0] = k; // iterator
            if( _bHasWeight )
                x[_nx + 1] = (Double_t) _gvu->w[k+1];// add weights values
            for (i = 0; i < _nx; i++)
            {
                x[i + 1] = (Double_t) _gvu->GetSample(k + 1, i + 1);
                if (_bSavePsi == kTRUE)
                    x[i + _nx + nlen] = (Double_t) _gvx->GetSample(k + 1, i + 1);
            } // end for
            _tds->getTuple()->Fill(x); // fill the TDataServer
            
        } //end for k        
        _tds->fillOthersAttributes();
          // Clean
        delete[] x;
        x = NULL;
        //throw warning exception
        if (_bSuppress)
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                    "TNisp::generateSample - Warning: The tds is already fill \n -->suppression of this tuple!!");
        if (_bError)
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                    "TNisp::generateSample - Warning: False option name --> No saving stochastic variable (_gvx) in the tds");
        if (_blog)
            cout << " ** TNisp::generateSample - End" << endl;

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
        exit(-1);
    } catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

//Specification of a sample from another componant
void TNisp::setSample(TString methode, Option_t *option)
{
    if (_blog)
        cout << " ** TNisp::setSample" << endl;
    
    _bHasWeight = kFALSE;
    TString sMethodRef;  
    TString sSeparator = ","; // separator between method name and specified name for the weight Attribute (this separator may be several characters long)
    TList listOfMethod; // list of methods using weight Attribute 
    listOfMethod.Add(new TObjString("Petras"));       // add the Petras method generated by the Sampler module
    listOfMethod.Add(new TObjString("Quadrature"));   // add the Quadrature method (generated by NISP)
    listOfMethod.Add(new TObjString("SmolyakGauss")); // add the SmolyakGauss method (generated by NISP)
    listOfMethod.Add(new TObjString("MonteCarlo"));   // add the MonteCarlo method (generated by NISP)
    // TString cannot be used in a TList, we use a TObjString instead
    TObjString * objStr = NULL; 
    TIter nextMethod(&listOfMethod); // iterator over the list of methods
    // iterate on the list of methods to verify if the method use weight attribute
    while ((objStr = (TObjString *) nextMethod()))
    {
        sMethodRef = objStr->GetString(); // get the method name in a TString
        if(methode.BeginsWith(sMethodRef)) 
        {
            _bHasWeight = kTRUE;
            if(methode.BeginsWith(sMethodRef+sSeparator)) // verify if the method name specifies a name for the weight attribute
            {
                Int_t nSepa = methode.Index(sSeparator);
                _sWeight = methode( nSepa+sSeparator.Length(), methode.Length() - nSepa); // get the weight attribute name 
                
            }
            if(!_tds->isAttribute(_sWeight))
            {
                try
                {
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                        Form("TNisp::setSample - Error: The weight attribute [%s] is missing!! ",_sWeight.Data() ));
                }catch (URANIE::Exceptions::UErrorExceptions& ue)
                {
                    ue.printMessage();
		    listOfMethod.Delete();
                    throw ue;
                    exit(-1);
                }
            }
        }
    }
    listOfMethod.Delete();
    
    if(_blog)
    {
        if(_bHasWeight)
        {
            cout << " ** method name : " << methode <<endl;
            cout << " ** weight name : " << _sWeight <<endl;
        }
        else
            cout << " ** method name : " << methode <<endl;
    }

    try
    {

        try
        {
            _tds->getTuple();
        } catch (URANIE::Exceptions::UWarningExceptions& )
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TNisp::setSample - Error: the sample is empty!! ");
        }

        _level = -1;
        _ns = _tds->getNPatterns();
        _gvu->SetSample((string) methode, _ns);
        Int_t jcrt = 0;
        TAttribute *att = NULL;
        TLeaf *linput = NULL;
        if(_blog)
            cout<<" ** List of input attributes : "<<endl;
        for(Int_t iterInputAttribute=0; iterInputAttribute < _listAttIn->getNAttributes(); iterInputAttribute++)
        {   
            att = _listAttIn->getAttribute(iterInputAttribute);
            if (att->isInput())
            {
             
                linput = _tds->getTuple()->GetLeaf(att->GetName());
                if (_blog)
                    linput->Print();
                for (Int_t k = 0; k < _ns; k++)
                {
                    _tds->getTuple()->GetEntry(k);
                    _gvu->SetSample(k + 1, jcrt + 1, linput->GetValue());
                }      //end for
                jcrt++;
                
               
            }      //end if
        }      //end for   
        // set integration weights
        if(_bHasWeight)
        {
            linput = _tds->getTuple()->GetLeaf(_sWeight);
            if(linput == NULL)
            {
                try
                {
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                        Form("TNisp::setSample - Error:  the leaf [%s] is not found !!", _sWeight.Data()));
                }catch (URANIE::Exceptions::UErrorExceptions& ue)
                {
                    ue.printMessage();
                    throw ue;
                    exit(-1);
                } 
            }
            if (_blog)
            {
                cout<<" ** weight attribute : "<<endl;
                linput->Print();
            }
            for (Int_t k = 0; k < _ns; k++)
            {
                _tds->getTuple()->GetEntry(k);
                _gvu->w[k+1] = linput->GetValue();
            }
        }
        _gvx->BuildSample(_gvu);
        _bIsSample = kTRUE;
        // Option to save the gvx variable in the tds
        TString soption = TString(option);
        TString name;
        Double_t * x = new Double_t[_nx * 2 + 1];
        Int_t k, i;
        soption.ToLower();
        if (!soption.Contains("nosavegvx"))
        {
            if (soption.Contains("savegvx"))
            {
                if (_blog)
                    cout
                            << "TNisp::setSample Saving the stochastic variable (_gvx) in the tds"
                            << endl;
                _bSavePsi = kTRUE;
                TBranch* branchAttribut;            
                Double_t * valAtt;
                for (i = 0; i < _nx; i++)
                {
                    // verify that the attribute is an input type
                    if((_listAttIn->getAttribute(i))->isInput())
                    {
                        name = "psi_";
                        name = name + (_listAttIn->getAttribute(i))->GetName();
                        _tds->addAttribute(name);
                        valAtt = new Double_t;
                        // add branch associated with the new attribute
                        branchAttribut = _tds->getTuple()->Branch(name, 
                                                                  valAtt,
                                                                  Form("%s/D", name.Data()));
                        // fill the new branch
                        for (k = 0; k < _ns; k++)
                        {
                            valAtt[0] = (Double_t) _gvx->GetSample(k + 1, i + 1);
                            branchAttribut->Fill();
                        }
                        
                        if(_blog)
                        {
                            cout<<" ** add stochastic variable : " <<name<<endl;
                            _tds->getTuple()->GetLeaf(_tds->getAttribute(name)->GetName())->Print();
                        }
                        // resest branch address                 
                        _tds->getTuple()->ResetBranchAddress(branchAttribut);
                        // Clean
                        delete valAtt;
                    }
                }      //end for i
                
              
            } //end savegvx
            else
            {
                _bSavePsi = kFALSE;
                // Clean
                delete[] x;
                x = NULL;
                throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                        Form("TNisp::setSample\n The option [%s] is unknown : The options are (savegvx | nosavegvx)\n No saving the stochastic variable (psi_*) in the TDS (nosavegvx as default)", soption.Data()));

            }
        } //end nosavegvx
        if (_blog)
        {
            cout << " _bIsSample: " << _bIsSample << endl;
            cout << " ** TNisp::setSample - end" << endl;
        }
        // Clean
        delete[] x;
        x = NULL;
       
    } catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
        exit(-1);
    }

}
//Stochastic dimension (nx)
Int_t TNisp::getNAttributes()
{
    Int_t dimension;
    dimension = _nx;
    if (_blog)
        cout << " ** TNisp::getNAttributes: " << _nx << endl;
    return dimension;
}
//Size of the sample (ns)
Int_t TNisp::getNPatterns()
{
    Int_t size;
    size = _ns;
    if (_blog)
        cout << " ** TNisp::getNPatterns: " << _ns << endl;
    return size;
}
//Get the level
Int_t TNisp::getLevel()
{
    Int_t val;
    val = _level;
    if (_blog)
        cout << " ** TNisp::getLevel: " << _level << endl;
    return (val);
}

// Set the level
void TNisp::setLevel(Int_t level)
{
    try
    {
        if (level < 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TNisp::setLevel - Error: Invalide value for level");
        if (_level > 0)
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                    "TNisp::setLevel - Warning: the level has already been set once \n");
    }
    catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    } 
    catch(URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
        exit(-1);
    }

    _level = level;
    _gvu->degre_max = level;
    _gvx->degre_max = level;
}

void TNisp::printLog(Option_t *option)
{
    TString name;
    cout << " ** TNisp::printLog[" << option << "]" << endl;
    cout << "      Number of simulation :" << _ns << endl;
    cout << "      Number of variable   :" << _nx << endl;
    cout << "      Level                :" << _level << endl;
    cout << "      Stochastic variables:" << endl;
    for (int i = 0; i < _nx; i++)
    {
        name = "psi_";
        name = name + (_listAttIn->getAttribute(i))->GetName();
        cout << "       -->" << name << ": " << _gvx->va[i]->type << " "
                << _gvx->va[i]->a << " " << _gvx->va[i]->b << " " << _gvx->va[i]->c << " " << _gvx->va[i]->d << endl;
    }
    cout << "      Uncertain variables:" << endl;
    for (int i = 0; i < _nx; i++)
    {
        name = (_listAttIn->getAttribute(i))->GetName();
        cout << "       -->" << name << ": " << _gvu->va[i]->type << " "
                << _gvu->va[i]->a << " " << _gvu->va[i]->b << " " << _gvu->va[i]->c << " " << _gvu->va[i]->d << endl;
    }
    if (_blog)
    {
        cout << "     _bIsSample (setSample() was used ?): " << _bIsSample
                << endl;
        cout << "     _bSavePsi  (saving the gvx in tds?): " << _bSavePsi
                << endl;
    }
    cout << "fin of TNisp::printLog[" << option << "]" << endl;

}

} //end of uranie namespace
} //end of uranie namespace
