//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TNisp.h
 \author Maud Salvoldi (OCC)
 \date   23 March 2011

 \brief Interface of the class URANIE::Modeler::TNisp

 */
/**

 * @class URANIE::Modeler::TNisp
 *
 * @brief Description of the class TNisp.
 *
 *  The class TNisp is associated to the class TDataServer.
 *  It's a mirror of the class SetRandomVariable, a class of the NISP library.
 *  It's used to access to functionalities of the NIPS library in URANIE.
 *  It's a class needed to create, manage, ..  a random variables group. 
 *  
 *
 */
#ifndef TNISP_H
#define TNISP_H

//ROOT
#include "Rtypes.h"
#include "TLeaf.h"

// uranie
#include "TDataServer.h"
#include "Modeler.h"
#include "TStochAttribut.h"
#include "UExceptions.h"

// nisp
#include "nisp_gva.h"

using namespace URANIE::DataServer;
using namespace URANIE::Exceptions;

namespace URANIE
{
namespace Modeler
{
class TNisp
{
/////////////members of the class
public:
    //! Object of type TDataServer.
    TDataServer * _tds;
    //! Object of type SetRandomVariable (library Nisp).
    //! It's the stochastic variable modeled by random variable.
    SetRandomVariable * _gvx;
    //! Object of type SetRandomVariable (library Nisp).
    //! It's the uncertain parameters modeled by random variable.  
    SetRandomVariable * _gvu;
    //! Number of random variable.
    Int_t _nx;
    //! Number of simulation.
    Int_t _ns;
    //! Level of the method used to build the sample. 
    Int_t _level;
    //! Object of type TDataSpecification used to index the name of the variables.
    TDataSpecification * _listAttIn;
    //! Boolean used to verify the use (or not) of integration weight.
    Bool_t _bHasWeight;
    //! Name of the integration weight.
    TString _sWeight;

protected:
    //! Boolean for edit the log.
    Bool_t _blog;

private:
    //! Boolean used to active the saving of the stochastic variables in the tds.
    Bool_t _bSavePsi;
    //! Boolean used to verify  the use (or not) of the method setSample.x
    Bool_t _bIsSample;
    //! Boolean used to write a warning if the option name is KO
    Bool_t _bError;
    //! Boolean used to write a warning if the tuple of tds already exist
    Bool_t _bSuppress;

//////////////methode of the class
public:
    /** @name Constructor and Destructor
     **/
    //@{
    //! Constructor  from a TDataServer object
    /*!
     \brief The constructor builds a object TNisp thanks to information which come from a object TDataServer.
     The elements which constitue the object TNisp are determinated in the following way: 
     \li _gvx: The stochastic variables are determinated from the type of the attribute of the stochastic base.
     This type is find in the TDataServer tds,
     \li _gvu: The uncertain parameters come from the TDataServer named tds,
     \li _nx:  The number of random variarble is the number of attribute of th TDataServer,
     \li _ns:  The number of simulation (size of the sample) is not determinate here.
     This value is known when the sample was genarated,
     \li _tds: It's the TDataServer given in parameter.
     \brief  
     \param tds : TDataServer object. 
     */
    TNisp(TDataServer *tds);
    //! Destructor
    virtual ~TNisp();
    //@}
    //! Realisation of a sample
    /*!
     \brief Build a sample by the method "type" with the size or degree associed "level". 
     \param type : Name of the method used to build the sample;
     \param level :  Follow the build sample's method choosen "level" corresponds to the size or the degree;
     \param option : (optional) Save the stochastic variables in the tds. By defaut no save.
     */
    void generateSample(TString type, Int_t level, Option_t *option =
            "nosavegvx");
    //! Specification of a sample from another componant
    /*!
     \brief Build a sample from a another sample (LHS, ...)
     \param methode : Name of the sample;
     \param option : (optional) Save the stochastic variables in the tds. By defaut no save.
     */
    void setSample(TString methode, Option_t *option = "nosavegvx");
    //! Number of ramdom variable i.e. stochastic dimension (_nx)
    /*!
     \brief Return the number of attributes
     */
    Int_t getNAttributes();
    //! Get the number of simulation i.e. size of the sample (_size)
    /*!
     \brief Return the numbers of patterns.
     */
    Int_t getNPatterns();
    //! Level
    /*
     \brief Return the level of th method used to build the sample.
     */
    Int_t getLevel();
    
    //! set the level of the method used to build the sample
     /*!
     \brief set the level
     */
    void setLevel(Int_t level);
    
    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    void setLog()
    {
        _blog = kTRUE;
    }
    void unsetLog()
    {
        _blog = kFALSE;
    }
    void changeLog()
    {
        _blog = _blog ? kFALSE : kTRUE;
    }
    Bool_t getLog()
    {
        return _blog;
    }
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Modeler::TNisp, ID_MODELER)

};
//end of the class TNisp
}//end of the namespace Modeler
}      //end of the namespace URANIE
#endif
