//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TPolynomialChaos.h
 \author Maud Salvoldi (OCC)
 \date   23 March 2011

 \brief Interface of the class URANIE::Modeder::TPolynomialChaos

 */
/**

 * @class URANIE::Modeler::TPolynomialChaos
 *
 * @brief Description of the class TPolynomialChaos.
 *
 * The class TPolynomialChaos is associated to the classe TNisp. 
 * It's a mirror of the class PolynomialChaos, a class of the NISP library.
 * It's used to access to Chaos Polynomial functionalities of the NIPS library in URANIE.
 *
 */
#ifndef TPOLYNOMIALCHAOS_H
#define TPOLYNOMIALCHAOS_H

//ROOT
#include "Rtypes.h"
#include "TNamed.h"

// uranie
#include "TDataServer.h"
#include "Modeler.h"

//nisp
#include "TNisp.h"
#include "nisp_pc.h"
#include "nisp_util.h"

#include <ctime>

using namespace URANIE::DataServer;

namespace URANIE
{
namespace Modeler
{
class TPolynomialChaos: public TNamed
{
 protected:
   URANIE::DataServer::UMessageLogger *_fLogger;   ///  Message logger
//-----------------------members of the class
public:
    //!Object of ype TNisp 
    TNisp * _nisp;
    //!Degree of the Chaos polynomial
    Int_t _degree;
    //! Object of type polynomialChaos (library Nisp)
    PolynomialChaos * _pc;
    //!Number of stochastic variables i.e number of input
    Int_t _nx;
    //! Number of simulation 
    Int_t _np;
    // Number of output
    Int_t _ny;
    Bool_t _bStoreYHat;     ///< Boolean to specify if we add the \hat{} attribute in the TDS [default kTRUE]

    TTree *_degreeResults; ///< TDSNtupleD used to store the results of the automatisedDegree method
    double _autoDegreeFactor; ///< Factor used to scale the maximum degree knowing _nx and _np
    int _minAutoDeg; ///< Minimal value for automatic degree scan
    int _maxAutoDeg; ///< Maximal value for automatic degree scan
    int _bestAutoDeg; ///< Best value for automatic degree scan

    vector<double> *_verrloo; ///< vector to contains the err loo and store them in the tree
    int _degval; ///< memory arry for the degree in the tree
    double _eqmval; ///< memory arry for the eqm in the tree
      
private:
    //! Object of type TDataSpecification used to index the name of the variables.
    TDataSpecification * _listAttOut;
    //! Boolean for edit the log.
    Bool_t _blog;
//--------------------methodes of the class
public:
    /** @name Constructor and Destructor
     **/
    //@{
    //!default constructor from a set of stochastic variables
    /*!
     \brief The constructor builds a object TPolynomialChaos thanks to information which come from a object TDaaServer and TNisp.
     The elements which constitue the object TPolynomialChaos are determinated in the following way
     \li _nisp :  It's the TNisp object given in parameter;
     \li _pc :  Ths polynomialchaos object is built from data with come from the stochastic variables of the TNisp object;
     \li _nx :  The number of random variable. It's the number of attribute of the TDataServer;
     \li _np :  The number of simulation (size of the sample) is given by TNisp
     \li _ny :  The number of output. 	   
     \li _degree : It's the degree of the polynomial chaos.
     \brief  
     \param tds : TDataServer object.
     \param nisp : TNisp object.
     \param soutput : list of output which be use to build the PC. By default all the output of the tds are used.
     */
    TPolynomialChaos(TDataServer *tds, TNisp *nisp, TString soutput = "");

    /*! Constructor from a file in \b "nisp" \b format
     \param ct : the name of the \b "nisp" file
     */
    TPolynomialChaos(const char *ct);

    //! copy constructor
    /*!
     \brief The copy constructor
     \param TPC : TPolynomialChaos object.
     */
    TPolynomialChaos(const TPolynomialChaos & TPC);

    //!destructor
    virtual ~TPolynomialChaos();
    //@}
    //!Set the degree
    /*!
     \brief Set the degree of the polynomial chaos.
     \param degree : degree max of the polynomial.
     */
    void setDegree(Int_t degree);

    //! Change the factor that links the number of samples, coefficient and the highest possible degree
    /*!
      If no maximum is specified for automatic degree optimisation, then the maximal degree to be tested is computed knowing that 
      Ncoeff <= np /autodeg, where np is number of pattern and Ncoeff = (nx+deg)!/(nx!*deg!);
      \param autodeg: change the value of autodeg (default is 1.5). Should be >= 1.
     */
    void setAutoDegreeFactor(double autodeg);
    
    //! Change the min and max degree tested for automatise degree determination
    /*!
      If no maximum is specified, the automatic specification will be done using the _autoDegreeFactor.
      \param amin: minimum to be tested (>=1)
      \param amax: max to be tested. Should be > amin. If empty, automatic determination is done
     */
    void setAutoDegreeBoundaries(int amin, int amax=-1);

    //!Return a pointer to the TTree that contains the degree optimisation results
    TTree *getAutoDegreeResults()
    {
      return _degreeResults;
    }
    
    //! Return the bet estimated degree when optimisation is done with regression
    int getBestAutoDegree()
    {
      return _bestAutoDeg;
    }
    
    //! Computation of output in relation with input : input[0...nx-1]
    /*!
     \brief Compute the ouput Y with the values Xi of the input (i=0 to nx-1).
     \param input : table of input value.
     */
    void computeOutput(double *input);
    //! Get an output
    /*!
     \brief Return the value of a ouput.
     \param jname : name of the ouput.
     \brief
     \brief refer to URANIE:Modeler::TPolynomialChaos:getOutput(Int_t j)
     */
    Double_t getOutput(TString jname = "");
    //! Get an output
    /*!
     \brief Return the value of a ouput.
     \param j : index od the ouput.
     */
    Double_t getOutput(Int_t j);

    //! Get the Mean squared uncertainty for the output with rank
    /*!
      \param rank: the rank of the considered output
     */
    Double_t getEqmLoo(Int_t rank=0);
    
    //! Get the Leave-One-Out uncertainty for the input (input) and the output (rank)
    /*!
      \param input: the input indexes of the doe
      \param rank: the rank of the considered output
     */
    Double_t getErrLoo(Int_t input, Int_t rank=0);

    //! Dummy factorial methods
    double factorial(int n);
    
    //! Computation of coefficients
    /*!
     \brief Calculated coefficients of the polynomial chaos by the methode "type".
     \param type : Name of the method ( Integration or Regression only accepted).
     */
    void automatisedDegree(Option_t *option = "");
    //! Computation of coefficients
    /*!
     \brief Calculated coefficients of the polynomial chaos by the methode "type".
     \param type : Name of the method ( Integration or Regression only accepted).
     */
    void computeChaosExpansion(TString type, Option_t * option = "");
    //!Get the mean
    /*!
     \brief Return the mean of the attribute given by the index.
     \param j : index of the attribut.
     */
    Double_t getMean(Int_t j);
    //!Get the mean
    /*!
     \brief Return the mean of the attribute given by the name.
     \param name : name of the attribut. By defalut the forst output.
     \brief
     \brief refer to URANIE:Modeler::TPolynomialChaos:getMean(Int_t j)
     */
    Double_t getMean(TString name = "");
    //!Get the variance
    /*!
     \brief Return the variance of the attribute given by the index.
     \param j : index of the attribut.
     */
    Double_t getVariance(Int_t j);
    //!Get the variance
    /*!
     \brief Return the variance of the attribute given by the name.
     \param name : name of the attribut. By defalut the forst output.
     \brief
     \brief refer to URANIE:Modeler::TPolynomialChaos:getVariance(Int_t j)
     */
    Double_t getVariance(TString name = "");
    //!Get the Covariance
    /*!
     \brief Return the covariance.
     \param i : index of the output;
     \param j : index of the second output.
     */
    Double_t getCovariance(Int_t i, Int_t j);
    //!Get the Covariance
    /*!
     \brief Return the covariance.
     \param xname : name of the output;
     \param yname : name of the second output.
     \brief
     \brief refer to URANIE:Modeler::TPolynomialChaos:getCovariance(Int_t i, Int_t j)
     */
    Double_t getCovariance(TString xname, TString yname);
    //!Get the correlation
    /*!
     \brief Return the correlation.
     \param i : index of the output;
     \param j : index of the second output.
     */
    Double_t getCorrelation(Int_t i, Int_t j);
    //!Get the correlation
    /*!
     \brief Return the correlation
     \param xname : name of the output;
     \param yname : name of the second ouput.
     \brief
     \brief refer to URANIE:Modeler::TPolynomialChaos:getCorrelation(Int_t i, Int_t j)
     */
    Double_t getCorrelation(TString xname, TString yname);
    //!First index of sensitivity
    /*!
     \brief Return the index od sensitivity.
     \param i : index of the input (<_nx);	
     \param j : index of the output (<_ny).
     */
    Double_t getIndexFirstOrder(Int_t i, Int_t j = 0);
    //!First index of sensitivity
    /*!
     \brief Return the index od sensitivity.
     \param xname : name of the input;
     \param yname : name of the output.
     \brief
     \brief refer to URANIE:Modeler::TPolynomialChaos:getIndexFirstOrder(Int_t i, Int_t j)
     */
    Double_t getIndexFirstOrder(TString xname, TString yname = "");
    //! Total index of sensitivity
    /*!
     \brief Return the total index of sensitivity.
     \param j : index of the output (<_ny);
     \param i : index of the variable (<_nx).
     */
    Double_t getIndexTotalOrder(Int_t i, Int_t j = 0);
    //! Total index of sensitivity
    /*!
     \brief Return the total index of sensitivity.
     \param xname : name of the input;
     \param yname : name of the output.
     \brief 
     \brief refer to URANIE:Modeler::TPolynomialChaos:getIndexTotalOrder(Int_t i, Int_t j)
     */
    Double_t getIndexTotalOrder(TString xname, TString yname = "");
    //! Index of sensitivity of a set of variables
    /*!
     \brief Return the index of sensitivity of the output yname.
     \param sinput : list of input. By default all the input;
     \param yname : name of the output. By default the first ouput.
     */
    Double_t getIndex(TString sinput = "", TString yname = "");
    //! Index of interaction sensitivity of a set of variables
    /*!
     \brief Return the interaction sensitivity index of the ouput j.
     \param sinput : list of input. By default all the input;
     \param yname : name of the output. By default the first ouput.
     */
    Double_t getIndexInteraction(TString sinput = "", TString yname = "");
    //! Number of input
    /*!
     \brief Return the number of input i.e the number of input random variable. 
     */
    Int_t getDimensionInput();
    //! Number of output
    /*!
     \brief Return the number of output i.e the number of output variable 
     */
    Int_t getDimensionOutput();
    //!Number of the coefficients of the polynomial chaos
    /*!
     \brief Return the number of coefficients of the PC.
     */
    Int_t getDimensionExpansion();
    //!Get the degree
    /*!
     \brief Return the degree maximum of the PC
     */
    Int_t getDegree();
    //! Export the model in C++ langage in a file.
    /*!
     \brief  Writing in a C++ programm, the model of the polynomial chaos.
     \param file (const char *) the file to export the modeler. If empty, use the name of the model, plus the extension of the langage;
     \param name (const char *) the name of the function in the export file. If empty, use the name of the model.
     */
    void exportFunction(const char *file = "nisp",
            const char *name = "nisp_fct");
    //!Get coefficents value
    /*!
     \brief Get the coefficeint k of the PC for the output janme
     \param k: number of the coefficient (<p)
     \param jname: output name
     */
    Double_t getCoefficient(Int_t k, TString jname = "");
    /** @name Printing Log
     **/
    //@{
    void setLog()
    {
        _blog = kTRUE;
    }
    void unsetLog()
    {
        _blog = kFALSE;
    }
    void changeLog()
    {
        _blog = _blog ? kFALSE : kTRUE;
    }
    Bool_t getLog()
    {
        return _blog;
    }
    virtual void printLog(Option_t *option = "");
    //@}
private:
    //! Get the target
    /*!
     \brief Return 
     */
    void getTarget();
    //! Add a random variable
    /*!
     \brief Add  a random variable (rank i) in the set
     \param i : index of the input variable added
     */
    void setGroupAddVar(Int_t i);
    //! Write set Denormalization
    /*!
     * Write the C code to denormalize the inputs parameters
     */
    void writeCodeCToDenormalizeInput(std::ofstream * sourcefile);
    //! Freeing the set
    /*!
     \brief Freeing the set
     */
    void setGroupEmpty();
public:

    //! Set the coefficient beta of _pc
    /*!
     \brief Set the coefficient beta of _pc
     \param noutput : number of output
     \param num : number of Polynomial Chaos
     \param coef : new value of coefficient 
     */
    void setCoefficient(Int_t noutput, Int_t num, Double_t coef);

    //! Get the coefficient beta of _pc
    /*!
     \brief Get the coefficient beta of _pc
     \param noutput : number of output
     \param num : number of Polynomial Chaos
     */
    Double_t getCoefficient(Int_t noutput, Int_t num);

    //---------------A FAIRE----------juste un lien vers nisp actuellement
    /** @name No accessible - In building
     **/
    //@{
    /*!
     \todo
     \brief Build a sample for statitical analysis (quantile). Build a sample of size "np" by the method "type"
     \param type : method used;
     \param np : number of simulation;
     \param order : 
     */
    void generateSample(TString type, Int_t np, Int_t order=1);
    /*! 
     \todo
     */
    Double_t getSample(Int_t k, Int_t j);
    /*!
     \brief The value of the output yname for the example k
     \brief refer to URANIE:Modeler::TPolynomialChaos::getSample(Int_t k,Int_t j)
     */
    Double_t getSample(Int_t k, TString yname);
    /*!
     \brief Polynomial chaos is a random variable : a realisation of this random variable (GetOutput()).
     \todo
     */
    void realisation();
    /*!
     \brief Edition of the ANOVA decomposition
     \todo
     */
    void getAnova(Int_t nt);
    /*!
     \todo
     */
    void getAnovaOrdered(Double_t seuil, Int_t j);
    /*!
     \brief Return the ANOVA ordered decomposition
     \param seuil : value of the seuil;
     \param yname : name of the output.
     \brief 
     \brief refer to URANIE:Modeler::TPolynomialChaos::getAnovaOrdered(Double_t seuil,Int_t j)
     */
    void getAnovaOrdered(Double_t seuil, TString yname = "");
    /*!
     \todo
     */
    void getAnovaOrderedCoefficients(Double_t seuil, Int_t j);
    /*!
     \brief Edition of the ANOVA ordered decomposition / coefficients
     \param seuil : value of the seuil;
     \param yname : name of the output.
     \brief 
     \brief refer to URANIE:Modeler::TPolynomialChaos::getAnovaOrderedCoefficients(Double_t seuil,Int_t j)
     */
    void getAnovaOrderedCoefficients(Double_t seuil, TString yname = "");
    /*!
     \brief Return the quantile of order "alpha"
     \param alpha : order;
     \param j : index of the output.
     */
    Double_t getQuantile(Double_t alpha, Int_t j);
    /*!
     \brief Return the quantile of order "alpha"
     \param alpha : order;
     \param yname : name of the output.
     \brief 
     \brief refer to URANIE:Modeler::TPolynomialChaos::getQuantile(Double_t alpha, Int_t j)
     */
    Double_t getQuantile(Double_t alpha, TString yname);
    /*!
     \brief Return the Wilks' quantile of order "alpha" with confidence "beta"
     \param alpha : order of the quantile;
     \param beta :  confidence
     \param j : index of the output. By default j=1.
     */
    Double_t getQuantileWilks(Double_t alpha, Double_t beta, Int_t j = 1);
    /*!
     \brief Return the Wilks' quantile of order "alpha" with confidence "beta"
     \param alpha : order of the quantile;
     \param beta :  confidence
     \param yname : name of the output
     \brief 
     \brief refer to URANIE:Modeler::TPolynomialChaos::getQuantileWilks(Double_t alpha, Double_t beta, Int_t j)
     */
    Double_t getQuantileWilks(Double_t alpha, Double_t beta, TString yname);
    /*!
     \brief Return the probability of a value overrun
     \param seuil : ;
     \param j : index of the output.
     */
    Double_t getInvQuantile(Double_t seuil, Int_t j = 1);
    /*!
     \brief Return the probability of a value overrun
     \param seuil : seuil;
     \param yname : name of the output.
     \brief 
     \brief refer to URANIE:Modeler::TPolynomialChaos::getInvQuantile(Double_t seuil, Int_t j)	
     */
    Double_t getInvQuantile(Double_t seuil, TString yname);
    /*!
     \brief Read a set of target from a file
     */
    void readTarget(Char_t *file);
    /*!
     \brief Set an input
     */
    void setInput(Int_t nt, Double_t dt);
    /*!
     \brief Propagation of input which has been specified SetInput()
     */
    void propagateInput();
    /*!
     \brief Propagation of dt[1...nx]
     */
    void propagateInput(Double_t *dt);
    /*!
     \brief Saving of the PC in a file.
     \param file : name of the file;
     */
    void save(Char_t *file);
    void save(string file);
    /*!
     \brief  Preparation on the ANOVA decomposition
     */
    void setAnova();
    //@}

    ClassDef(URANIE::Modeler::TPolynomialChaos, ID_MODELER)

};
//end of the class TPolynomialChaos
}//end of the namespace Modeler
}      //end of the namespace URANIE
#endif
