//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TPolynomialMultiDimFit.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date  2012-11-28 15:14:38
 
 \brief Implementation of the class URANIE::Modeler::TPolynomialMultiDimFit
 
 */
#include <ctime>

//// ROOT
#include "TLeaf.h"
#include "TF1.h"
#include "TPad.h"
#include "TMath.h"

//// URANIE
#include "TPolynomialMultiDimFit.h"
#include "UTimer.h"
#include "TDSNtupleD.h"
#include "TAttributeFormula.h"

using std::cout;
using std::endl;

ClassImp (URANIE::Modeler::TPolynomialMultiDimFit)

namespace URANIE
{
namespace Modeler
{
// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------
TPolynomialMultiDimFit::TPolynomialMultiDimFit(
        URANIE::DataServer::TDataServer *tds, TString architecture,
        TMultiDimFit::EMDFPolyType type, Option_t *option) :
        TModeler(tds, architecture, option), _MultiDimFit(NULL)
{
    if (_blog)
        cout << " ** Constructor TPolynomialMultiDimFit option[" << option
                << "]" << endl;

    try
    {
        if (_nY > 1)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TPolynomialMultiDimFit constructor\n The number of output must be equal to 1.\nHere, we are %d output attributes with [%s]",
                            _nY, _sOutputAttributes.Data()));

        _MultiDimFit = new TMultiDimFit(_nX, type, option);
        if (_MultiDimFit == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TPolynomialMultiDimFit constructor\n The TMultiDimFit pointer is NULL");

        _sFileName = Form("_MultiDimFit_%s", _sOutputAttributes.Data());
        _sFunctionName = Form("MDF_%s", _sOutputAttributes.Data());

        TString opt = option;
        opt.ToLower();

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

TPolynomialMultiDimFit::TPolynomialMultiDimFit(
        URANIE::DataServer::TDataServer *tds, const char *varexpinput,
        const char *varexpoutput, TMultiDimFit::EMDFPolyType type,
        Option_t * option) :
        TModeler(tds, varexpinput, varexpoutput, option), _MultiDimFit(NULL)
{
    if (_blog)
        cout << " ** Constructor TPolynomialMultiDimFit option[" << option
                << "]" << endl;

    try
    {
        if (_nY > 1)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TPolynomialMultiDimFit constructor\n The number of output must be equal to 1.\nHere, we are %d output attributes with [%s]",
                            _nY, _sOutputAttributes.Data()));

        _MultiDimFit = new TMultiDimFit(_nX, type, option);
        if (_MultiDimFit == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TPolynomialMultiDimFit constructor\n The TMultiDimFit pointer is NULL");

        _sFileName = Form("_MultiDimFit_%s", _sOutputAttributes.Data());
        _sFunctionName = Form("MDF_%s", _sOutputAttributes.Data());

        TString opt = option;
        opt.ToLower();

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

// -----------------------------------------------------------------------
//  Destructor
// -----------------------------------------------------------------------

TPolynomialMultiDimFit::~TPolynomialMultiDimFit()
{
    if (_blog)
        cout << " ** Destroy TPolynomialMultiDimFit" << endl;

    if (_MultiDimFit != NULL) {
        delete _MultiDimFit; _MultiDimFit = NULL;
    }

    if (_blog)
        cout << " ** End Of Destroy TPolynomialMultiDimFit" << endl;
}

// -----------------------------------------------------------------------
//  setter methods 
// -----------------------------------------------------------------------
void TPolynomialMultiDimFit::setMinAngle(Double_t dangle)
{
    try
    {
        if (dangle > 90 || dangle <= 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TPolynomialMultiDimFit::setMinAngle\n The angle must be in [0,90), and you use [%e]",
                            dangle));
        _MultiDimFit->SetMinAngle(dangle);

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

void TPolynomialMultiDimFit::setMaxAngle(Double_t dangle)
{
    try
    {
        if (dangle > 90 || dangle <= 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TPolynomialMultiDimFit::setMaxAngle\n The angle must be in [0,90), and you use [%e]",
                            dangle));
        _MultiDimFit->SetMaxAngle(dangle);

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

void TPolynomialMultiDimFit::setMaxFunctions(Int_t n)
{
    try
    {
        if (n < 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TPolynomialMultiDimFit::setMaxFunctions\n The number of functions must be possitive, and you use [%d]",
                            n));
        _MultiDimFit->SetMaxFunctions(n);

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}
void TPolynomialMultiDimFit::setMaxStudy(Int_t n)
{
    try
    {
        if (n < 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TPolynomialMultiDimFit::setMaxStudy\n The number of max study must be possitive, and you use [%d]",
                            n));
        _MultiDimFit->SetMaxStudy(n);

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

void TPolynomialMultiDimFit::setMaxTerms(Int_t n)
{
    try
    {
        if (n < 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TPolynomialMultiDimFit::setMaxTerms\n The number of max study must be possitive, and you use [%d]",
                            n));
        _MultiDimFit->SetMaxTerms(n);

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

void TPolynomialMultiDimFit::setMaxPowers(const Int_t *powers)
{
    try
    {
        if (powers == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TPolynomialMultiDimFit::setMaxPowers\n The pointer of the vector of powers is NULL");
        _MultiDimFit->SetMaxPowers(powers);

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

void TPolynomialMultiDimFit::setMinRelativeError(Double_t derror)
{
    try
    {
        if (derror < 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TPolynomialMultiDimFit::setMinRelativeError\n The minimal relative error must be possitive, and you use [%e]",
                            derror));
        _MultiDimFit->SetMinRelativeError(derror);

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

void TPolynomialMultiDimFit::setPowerLimit(Double_t dlimit)
{
    try
    {
        if (dlimit < 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TPolynomialMultiDimFit::setPowerLimit\n The power limit must be possitive, and you use [%e]",
                            dlimit));
        _MultiDimFit->SetPowerLimit(dlimit);

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

void TPolynomialMultiDimFit::setPowers(const Int_t *powers, Int_t terms)
{
    _MultiDimFit->SetPowers(powers, terms);
}

// -----------------------------------------------------------------------
//  fillData
// -----------------------------------------------------------------------
void TPolynomialMultiDimFit::fillData(TMatrixD &Aw, TVectorD &yw)
{
    if (_blog)
    {
        cout << endl << "**************************************" << endl;
        cout << " ** TPolynomialMultiDimFit::fillData" << endl;
    }

    Int_t nP = _nX;
    if (_bIntercept)
        nP++;
    Aw.ResizeTo(_nS, nP);
    yw.ResizeTo(_nS);
    yw = 1.0;

    // Get The list of Leaf (from TTree) of the list of attributes ( _listOfInputAttributes)
    TList *listOfInputLeaves = new TList();
    TIter next(_listOfInputAttributes);
    if (_blog)
        cout << " ** input" << endl;
    while (URANIE::DataServer::TAttribute *att =
            (URANIE::DataServer::TAttribute *) next())
    {
        listOfInputLeaves->Add(_tds->getTuple()->GetLeaf(att->GetName()));
        if (_blog)
        {
            _tds->getTuple()->GetLeaf(att->GetName())->Print();
            att->printLog();
        }
    }

    TLeaf *leafY = NULL;
    TList *listOfOutputLeaves = new TList();
    if (_nY == 1)
    {
        // Case of the output (only one attribut) : Fill the vector Y
        leafY = _tds->getTuple()->GetLeaf(_sOutputAttributes.Data());
        if (_blog)
        {
            cout << " ** output" << endl;
            _tds->getTuple()->GetLeaf(_sOutputAttributes)->Print();
            _tds->getAttribute(_sOutputAttributes)->printLog();
        }
    }
    else
    {
        TIter nextout(_listOfOutputAttributes);
        if (_blog)
            cout << " ** Output" << endl;
        while (URANIE::DataServer::TAttribute *att =
                (URANIE::DataServer::TAttribute *) nextout())
        {
            listOfOutputLeaves->Add(_tds->getTuple()->GetLeaf(att->GetName()));
            if (_blog)
            {
                _tds->getTuple()->GetLeaf(att->GetName())->Print();
                att->printLog();
            }
        }
    }

    // Fill the matrix and vector of data
    Int_t ivar;
    for (Int_t ipat = 0; ipat < _nS; ipat++)
    {
        _tds->getTuple()->GetEntry(ipat);
        if ((ipat < 5) && (_blog))
            _tds->getTuple()->Show();

        if (_blog)
            cout << " ** ipat[" << ipat << "/" << _nS << "]" << endl;

        // Case of the input attributes : Fill the matrix Aw
        ivar = 0;
        if (_bIntercept)
            Aw(ipat, ivar++) = 1.0;
        TIter nextLeaf(listOfInputLeaves);
        while (TLeaf *leaf = (TLeaf *) nextLeaf())
            Aw(ipat, ivar++) = leaf->GetValue(0);

        // Case of output attribute : Fill the vector yw
        yw(ipat) = leafY->GetValue(0);
    }

    delete listOfInputLeaves;
    listOfInputLeaves = NULL;

    delete listOfOutputLeaves;
    listOfOutputLeaves = NULL;

    if (_blog)
    {
        cout << " ** Matrix Aw(" << Aw.GetNrows() << "," << Aw.GetNcols() << ")"
                << endl;
        Aw.Print();
        cout << " ** Vector Yw(" << yw.GetNrows() << ")" << endl;
        yw.Print();
    }

    if (_blog)
    {
        cout << " ** End of TPolynomialMultiDimFit::fillData" << endl;
        cout << "**************************************" << endl;
    }
}
// -----------------------------------------------------------------------
//  estimate 
// -----------------------------------------------------------------------
void TPolynomialMultiDimFit::estimate(Option_t * option)
{
    if (_blog)
    {
        cout << endl << "**************************************" << endl;
        cout << " ** TPolynomialMultiDimFit::estimate Option [" << option << "]"
                << endl;
    }

    TMatrixD X(_nS, _nX);
    TVectorD vecY(_nS);

    fillData(X, vecY);
    if (_blog)
    {
        cout << " ** _nX[" << _nX << "]" << endl;
        cout << " ** Matrix X(" << X.GetNrows() << "," << X.GetNcols() << ")"
                << endl;
        X.Print();
        cout << " ** Vector vecY(" << vecY.GetNrows() << ")" << endl;
        vecY.Print();
    }

    // Fill the data in the TMultiDimFit object
    // Do not take into account the intercept
    Double_t *dxvalues = new Double_t[_nX];
    for (Int_t is = 0; is < _nS; is++)
    {
        for (Int_t j = 0; j < _nX; j++)
            dxvalues[j] = X(is, j + 1);
        _MultiDimFit->AddRow(dxvalues, vecY[is]);
    }

    // Print parameters
    _MultiDimFit->Print("p");
    // Print statistic on learning database
    _MultiDimFit->Print("s");
    // Create histograms
    _MultiDimFit->MakeHistograms();
    // Fitting the parameters
    _MultiDimFit->FindParameterization();
    // Print results and statistics
    _MultiDimFit->Print("rcf");

    Int_t nP = _MultiDimFit->GetNCoefficients();
    _parameterValues.ResizeTo(nP);
    _parameterValues.Zero();
    const TVectorD* vecCoeff = _MultiDimFit->GetCoefficients();
    for (Int_t i = 0; i < nP; i++)
    {
        _parameterValues[i] = (*vecCoeff)[i];
    }

    if (_blog)
    {
        cout << " *** _parameterValues" << endl;
        _parameterValues.Print();
    }

    if (_bStoreYHat)
    {
        TString syhat = Form("%shat", _sOutputAttributes.Data());
        URANIE::DataServer::TAttribute *attY = _tds->getAttribute(
                _sOutputAttributes);
        _tds->addAttributeByNameTitle(syhat, Form("#hat{%s}", attY->GetTitle()));

        Double_t tabYValue = 0.0;
        TBranch *tbr = _tds->getTuple()->Branch(syhat, &tabYValue,
                Form("%s/D", syhat.Data()));

        for (Int_t is = 0; is < _nS; is++)
        {
            for (Int_t j = 0; j < _nX; j++)
                dxvalues[j] = X(is, j + 1);
            // Evaluate function. Scale to shifted values
            tabYValue = _MultiDimFit->Eval(dxvalues);
            // Fill the #hat{y} data
            tbr->Fill();
        }
        _tds->getTuple()->ResetBranchAddress(tbr);
        _tds->saveTuple("UPDATE");
    }
    // Clean
    delete[] dxvalues;
    dxvalues = NULL;

    if (_blog)
    {
        cout << endl << " ** TPolynomialMultiDimFit::estimate Option ["
                << option << "]" << endl;
        cout << "**************************************" << endl;
    }
}

// -----------------------------------------------------------------------
//  Export Function
// -----------------------------------------------------------------------
// void TPolynomialMultiDimFit::exportFunction(const char* lang, const char* file,
//         const char* name, Option_t *option)
// {
//     TString lg = lang;
//     lg.ToLower();
//     TString sComment = "**";
//     Bool_t bOK = kFALSE;
//     TString sFile = TString(file);
//     if (sFile.Length() == 0)
//         sFile = _sFileName;
// 
//     TString sFunction = TString(name);
//     if (sFunction.Length() == 0)
//         sFunction = _sFunctionName;
// 
//     try
//     {
//         if (lg.BeginsWith("c"))
//         {
//             if (!sFile.EndsWith(".C"))
//                 sFile += ".C";
//             sComment = "  //";
// 
//             // Write the function by the TMultiDimFit
//             if (_MultiDimFit)
//                 _MultiDimFit->MakeCode(sFile);
// 
//             std::ofstream * outFile = new std::ofstream(sFile, ios::app);
//             *outFile << "//" << endl;
//             *outFile
//                     << "// This was the end of File for the TMultiDimFit ROOT Classe"
//                     << endl;
//             *outFile << "//" << endl;
//             *outFile << "////////////////////////////// " << endl;
//             *outFile << endl;
//             *outFile << "  ////////////////////////////// " << endl;
//             *outFile << "  //" << endl;
//             *outFile << "  // The Uranie function" << endl;
//             *outFile << "  //" << endl;
//             *outFile << "  //" << endl;
//             *outFile << "  //     QUALITY  : " << endl;
//             *outFile << "  //" << endl;
//             *outFile << "  //              ResidualSumSq  = "
//                     << _MultiDimFit->GetResidualSumSq() << endl;
//             *outFile << "  //              Residual Min   = "
//                     << _MultiDimFit->GetResidualMin() << endl;
//             *outFile << "  //              Residual Max   = "
//                     << _MultiDimFit->GetResidualMax() << endl;
//             *outFile << "  //" << endl;
//             *outFile << "  ////////////////////////////// " << endl << endl;
//             *outFile << endl;
//             *outFile << "void " << sFunction << "(double *param, double *res)"
//                     << endl;
//             *outFile << "{" << endl;
//             *outFile << "  res[0] = MDF(param);" << endl;
//             *outFile << "}" << endl;
//             *outFile << endl;
//             *outFile << "// End Of File" << endl;
//             outFile->close();
// 
//             bOK = kTRUE;
//         }
//         if (lg.Contains("for"))
//         {
//             throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
//                     Form(
//                             "TPolynomialMultiDimFit::exportFunction\n The FORTRAN language \"%s\" is not yet implemented.",
//                             lang));
//             sFile += TString(".f");
//             sComment = "C --- ";
//             bOK = kTRUE;
//         }
//         if (lg.Contains("pmml"))
//         {
//             throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
//                     Form(
//                             "TPolynomialMultiDimFit::exportFunction\n The PMML format \"%s\" is not yet implemented.",
//                             lang));
//             bOK = kTRUE;
//         }
//         if (bOK == kFALSE)
//             throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
//                     Form(
//                             "TPolynomialMultiDimFit::exportFunction\n The chosen language \"%s\" is not accepted (C|fortran|pmml)",
//                             lang));
//     } catch (URANIE::Exceptions::UErrorExceptions& ue)
//     {
//         ue.printMessage();
//         throw ue;
//     } catch (URANIE::Exceptions::UWarningExceptions& ue)
//     {
//         ue.printMessage();
//         throw ue;
//     }
// }



void TPolynomialMultiDimFit::exportModelCplusplus(std::ofstream * sourcefile) const
{
    UNUSED(sourcefile);
    cout << " ** Nothing to do" << endl;
    throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,        
          "TPolynomialMultiDimFit::exportFunction\n The C++ format is not yet implemented.");
}

void TPolynomialMultiDimFit::exportModelFortran(std::ofstream * sourcefile) const
{
    UNUSED(sourcefile);
    cout << " ** Nothing to do" << endl;
    throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,        
          "TPolynomialMultiDimFit::exportFunction\n The Fortran format is not yet implemented.");
}

// -----------------------------------------------------------------------
//  Log
// -----------------------------------------------------------------------
void TPolynomialMultiDimFit::printLog(Option_t *option)
{
    cout << "*******************************" << endl;
    cout << " ** TPolynomialMultiDimFit::printLog[" << option << "]" << endl;
    cout << " ** TModeler::printLog[" << option << "]" << endl;
    cout << " ** Name[" << GetName() << "] Title[" << GetTitle() << "]" << endl;
    cout << " **   _sFileName[" << _sFileName << "] _sFunctionName["
            << _sFunctionName << "]" << endl;
    if (_bIntercept)
        cout << " ** _bIntercept[" << _bIntercept << " - YES]" << endl;
    else
        cout << " ** _bIntercept[" << _bIntercept << " - NO]" << endl;
    if (_MultiDimFit)
    {
        cout << "  *** _MultiDimFit properties ***" << endl;
        cout << "   ** Type[" << _MultiDimFit->GetPolyType() << "]" << endl;
        cout << "   ** MinAngle[" << _MultiDimFit->GetMinAngle() << "]" << endl;
        cout << "   ** MawAngle[" << _MultiDimFit->GetMaxAngle() << "]" << endl;
        cout << "   ** MaxFunctions[" << _MultiDimFit->GetMaxFunctions() << "]"
                << endl;
        cout << "  *** End Of _MultiDimFit properties ***" << endl;
    }
    else
    {
        cout << "  *** _MultiDimFit IS NULL ***" << endl;
    }
    cout << "*******************************" << endl;
}

} // Fin du namespace Modeler
} // Fin du namespace URANIE

