//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TPolynomialRegression.cxx
 \author Estelle Deville <estelle.deville@cea.fr>
 \date   jeu jun 5 10:35:25 CEST 2008

 \brief Implementation of the class URANIE::Modeler::TPolynomialRegression

 */
#define TFILENAME "_uranie_tpolynomialregression_complete.root"
#include <ctime>

//// ROOT
#include "TDecompSVD.h"
#include "TLeaf.h"
#include "TF1.h"
#include "TH1.h"
#include "TCanvas.h"
#include "TPad.h"
#include "TMath.h"

//// URANIE
#include "TPolynomialRegression.h"
#include "UTimer.h"
#include "TDSNtupleD.h"
#include "TAttributeFormula.h"

using std::cout;
using std::endl;

ClassImp (URANIE::Modeler::TPolynomialRegression)

namespace URANIE
{
namespace Modeler
{
// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------
TPolynomialRegression::TPolynomialRegression(
        URANIE::DataServer::TDataServer *tds, TString architecture,
        Option_t *option) :
  TModeler(tds, architecture, option), _nCriteria(kUnknown), _dCriteria(0.7),
  _bChangemaxorder(kTRUE), _OKcomplete(kFALSE), _theTLR(NULL), _completeDS(NULL), _sCompleteInputAttributes(""), 
  _sLROption(""), _drawingGarbageCollector(NULL), _dr2a(-1.), _dq2(-1.), _dPress(-1.), _ntresult(NULL), _nMinOrder(1),
  _nMaxOrder(2), _nMaxMonomes(-1), _nNumber(0), _bTrace(kFALSE), _sTraceFile("_tpr_trace_.log")
  
{
    _completeDS = new URANIE::DataServer::TDataServer(
            Form("full_%s", _tds->GetName()),
            Form("Full TDS of %s", _tds->GetTitle()));
    for (Int_t ivar = 0; ivar < _nX; ivar++)
    {
        URANIE::DataServer::TAttribute *att =
                (URANIE::DataServer::TAttribute *) _listOfInputAttributes->At(
                        ivar);
        _completeDS->addAttribute(att);
    }
    for (Int_t ivar = 0; ivar < _nY; ivar++)
    {
        URANIE::DataServer::TAttribute *att =
                (URANIE::DataServer::TAttribute *) _listOfOutputAttributes->At(
                        ivar);
        _completeDS->addAttribute(att);
    }
    _tds->getTuple()->SetBranchStatus("*", 1);
    _completeDS->setTuple(_tds->getTuple());
    // URANIE::DataServer::TDSNtupleD *outputtree =
    //         dynamic_cast<URANIE::DataServer::TDSNtupleD *>(_tds->getTuple()->CloneTree());
    // _completeDS->createTuple();
    // _completeDS->deleteTuple();
    // _completeDS->setTuple(outputtree);
    // _outputtree = new URANIE::DataServer::TDSNtupleD (_tds->getTuple() );
    // _completeDS->setTuple(_outputtree);

    _lPropMonoms = new TList();

    _ntresult = new TTree("tmpsobol", "PolynomialRegression");
    _ntresult->Branch("Inp", (void *) _sMonomials, (const char *) "Inp/C");
    _ntresult->Branch("n", &_nMonomials, (const char *) "n/I");
    _ntresult->Branch("b", &_bRetained, (const char *) "b/I");
    _ntresult->Branch("Criteria", &_dValueCriteria,
            (const char *) "Criteria/D");
    _ntresult->Branch("C", &_dAllCriteria, (const char *) "R2/D:R2A/D:Q2/D");
    _ntresult->SetMarkerColor(2);
    _ntresult->SetMarkerSize(0.8);
    _ntresult->SetMarkerStyle(8);

    _fLogger = new URANIE::DataServer::UMessageLogger("TPolynomialRegression");
    (*_fLogger) << Endl<< "TPolynomialRegression \"" << _tds->GetName() << "\"" << Endl;
    (*_fLogger) << " **  architecture : " << architecture << Endl;
    (*_fLogger) << " **      Patterns : " << _tds->getNPatterns() << Endl;

    TString opt = option;
    opt.ToLower();

    _sLROption = "nostoreyhat";
    if (_bIntercept == kFALSE)
        _sLROption += ", nointercept";

    _drawingGarbageCollector = new TObjArray();
    _drawingGarbageCollector->SetOwner(kTRUE);
}

TPolynomialRegression::TPolynomialRegression(
        URANIE::DataServer::TDataServer *tds, const char *varexpinput,
        const char *varexpoutput, Option_t * option) :
        TModeler(tds, varexpinput, varexpoutput, option), _nCriteria(kUnknown), _dCriteria(0.7),
                _bChangemaxorder(kTRUE), _OKcomplete(kFALSE), _theTLR(NULL),
                _completeDS(NULL), _sCompleteInputAttributes(""), _sLROption(""), 
                _drawingGarbageCollector(NULL), _dr2a(-1.), _dq2(-1.), _dPress(-1.), _ntresult(NULL), _nMinOrder(1), 
                _nMaxOrder(2), _nMaxMonomes(-1), _nNumber(0), _bTrace(kFALSE), _sTraceFile("_tpr_trace_.log")
{
    _completeDS = new URANIE::DataServer::TDataServer(
            Form("full_%s", _tds->GetName()),
            Form("Full TDS of %s", _tds->GetTitle()));
    for (Int_t ivar = 0; ivar < _nX; ivar++)
    {
        URANIE::DataServer::TAttribute *att =
                (URANIE::DataServer::TAttribute *) _listOfInputAttributes->At(
                        ivar);
        _completeDS->addAttribute(att);
    }
    for (Int_t ivar = 0; ivar < _nY; ivar++)
    {
        URANIE::DataServer::TAttribute *att =
                (URANIE::DataServer::TAttribute *) _listOfOutputAttributes->At(
                        ivar);
        _completeDS->addAttribute(att);
    }
    _tds->getTuple()->SetBranchStatus("*", 1);
    _completeDS->setTuple(_tds->getTuple());
    // URANIE::DataServer::TDSNtupleD *outputtree =  dynamic_cast<URANIE::DataServer::TDSNtupleD *>(_tds->getTuple()->CloneTree());
    // _completeDS->createTuple();
    // _completeDS->deleteTuple();
    // _completeDS->setTuple(outputtree);
    // //delete _completeDS->getTuple();    //_completeDS->getTuple() = NULL;
    // // _outputtree =  new URANIE::DataServer::TDSNtupleD(_tds->getTuple());
    // // _completeDS->setTuple(outputtree);
    
    _lPropMonoms = new TList();
       
    _ntresult = new TTree("tmpsobol", "PolynomialRegression");
    _ntresult->Branch("Inp", (void *) _sMonomials, (const char *) "Inp/C");
    _ntresult->Branch("n", &_nMonomials, (const char *) "n/I");
    _ntresult->Branch("b", &_bRetained, (const char *) "b/I");
    _ntresult->Branch("Criteria", &_dValueCriteria,
            (const char *) "Criteria/D");
    _ntresult->Branch("C", &_dAllCriteria, (const char *) "R2/D:R2A/D:Q2/D");
    _ntresult->SetMarkerColor(2);
    _ntresult->SetMarkerSize(0.8);
    _ntresult->SetMarkerStyle(8);

    _sFileName = Form("linearmodel_%s", _sOutputAttributes.Data());
    _sFunctionName = Form("polynomialmodel_Fct_%s", _sOutputAttributes.Data());
    SetName(
            Form(
                    "Modeler[PolynomialRegression]Tds[%s]InitialPredictor[%s]Target[%s]",
                    _tds->GetName(), varexpinput, _sOutputAttributes.Data()));
    SetTitle(
            Form(
                    "PolynomialRegression as modeler for TDS[%s] with Target is[%s] versus[%s]",
                    _tds->GetName(), _sOutputAttributes.Data(), varexpinput));
    _fLogger = new URANIE::DataServer::UMessageLogger("TPolynomialRegression");
    (*_fLogger) << Endl<< "TPolynomialRegression \"" << _tds->GetName() << "\"" << Endl;
    (*_fLogger) << " **  varexpinput  : " << varexpinput << Endl;
    (*_fLogger) << " ** varexpoutput  : " << varexpoutput << Endl;
    (*_fLogger) << " **     Patterns  : " << _tds->getNPatterns() << Endl;

    TString opt = option;
    opt.ToLower();

    _sLROption = "nostoreyhat";
    if (_bIntercept == kFALSE)
        _sLROption += ", nointercept";
    
    _drawingGarbageCollector = new TObjArray();
    _drawingGarbageCollector->SetOwner(kTRUE);
    
}

// -----------------------------------------------------------------------
//  Destructor
// -----------------------------------------------------------------------

TPolynomialRegression::~TPolynomialRegression()
{
    if (_blog)
        cout << endl << " ** Destroy TPolynomialRegression" << endl;

    if(_fLogger!=NULL)
      delete _fLogger;
    _fLogger=NULL;

    if (_lPropMonoms != NULL)
    {
        //Int_t nvar = _lPropMonoms->GetSize();
        // for (Int_t ivar = 0; ivar < nvar; ivar++)
        // {
        //     delete _lPropMonoms->At(ivar); // _lPropMonoms->At(ivar) = NULL;
        // }
	_lPropMonoms->Delete();
        delete _lPropMonoms;
        _lPropMonoms = NULL;
    }

    if( gROOT->GetList() != 0 &&  !gROOT->GetList()->IsEmpty())
    {
	if (_completeDS != NULL)
	{
	    delete _completeDS;
	    _completeDS = NULL;
	}
    }
    
    if (_theTLR != NULL)
    {
        delete _theTLR;
        _theTLR = NULL;
    }

    if( gROOT->GetList() != 0 &&  !gROOT->GetList()->IsEmpty())
    {
	if (_ntresult != NULL)
	{
	    delete _ntresult;
	    _ntresult = NULL;
	}

	if(_drawingGarbageCollector)
	    delete _drawingGarbageCollector;
	_drawingGarbageCollector=NULL;
    }
    
    if (_blog)
        cout << " ** End Of Destroy TPolynomialRegression" << endl;
}

// -----------------------------------------------------------------------
//  setMaxMonomes 
// -----------------------------------------------------------------------
void TPolynomialRegression::setMaxMonomes(Int_t n)
{
    _nMaxMonomes = n;
}

// -----------------------------------------------------------------------
//  Treat the trace information  
// -----------------------------------------------------------------------
void TPolynomialRegression::setTrace(TString straceFile)
{
    _bTrace = kTRUE;
    if (straceFile.Length() > 0)
        _sTraceFile = straceFile;
}

TString TPolynomialRegression::getTraceFile()
{
    TString sReturn = "";
    if (_bTrace)
        sReturn = _sTraceFile;
    return sReturn;
}

// -----------------------------------------------------------------------
//  estimate 
// -----------------------------------------------------------------------
void TPolynomialRegression::estimate(Option_t * option)
{

  if (_blog)
    {
      cout << endl << "**************************************" << endl;
      cout << " ** TPolynomialRegression::estimate Option [" << option << "]"
	   << endl;
    }

  URANIE::Modeler::TLinearRegression *tlr=NULL;
  
  // Treat the option
  try
    {
      TString soption = TString(option);
      soption.ToUpper();
      if (soption.CompareTo("R2") == 0)
	_nCriteria = kR2;
      else if (soption.CompareTo("R2A") == 0)
	_nCriteria = kR2A;
      else if ((soption.CompareTo("Q2") == 0) || (soption.Length() == 0))
	_nCriteria = kQ2;
      else
	throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						   "TPolynomialRegression::estimate - Error: invalid value for option (Need to be R2, Q2 or R2A)");
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
      ue.printMessage();
      throw ue;
    }

  if (_nMaxMonomes > 0)
    {
      (*_fLogger) << " ** Nb. 2nd terms : " << _nMaxMonomes
		  << "     ( Exaustive Research )" << Endl;
    }
  else
    {
      (*_fLogger) << " **     Criteria  : ";
      switch (_nCriteria)
        {
	case kR2:
	  (*_fLogger) << "R2";
	  break;
	case kR2A:
	  (*_fLogger) << "R2A";
	  break;
	case kQ2:
	  (*_fLogger) << "Q2";
	  break;
	case kUnknown:
	  (*_fLogger) << " -not defined- ";
	  break;
        }
      (*_fLogger) << " -- Target (" << _dCriteria << ")" << Endl;
    }
  (*_fLogger) << " ** _nY           : " << _nY << Endl;
  (*_fLogger) << " ** _nMaxOrder    : " << _nMaxOrder << Endl;
  (*_fLogger) << " ** _nX           : " << _nX << Endl;

  try
    {
      if (_nY != 1)
	throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						   Form(
							"TPolynomialRegression::estimate   ERROR\nThe number of output must be equal to 1.\nHere, we are %d output attributes with [%s]",
							_nY, _sOutputAttributes.Data()));

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
      ue.printMessage();
      throw ue;
    }

  try
    {
      if (_nMaxOrder > _nX)
	throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						   Form(
							"TPolynomialRegression::estimate   ERROR\nThe maximal order of polynomial is greater than the number of input.\nHere, we have %d maximal order with [%d] intput attributes",
							_nMaxOrder, _nX));

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
      ue.printMessage();
      throw ue;
    }

  // Generate all candidates 2nd order terms monomes and the complete polynome
  Bool_t bOK = _generateCompletePolynomial();

  if (_blog)
    cout << endl << "Try with this this _sCompleteInputAttributes ["
	 << _sCompleteInputAttributes.Data() << "] -- ";
  Double_t dval_complete = -1.0;
  if (bOK)
    {
      if( tlr != NULL) { delete tlr; tlr=NULL;}
      tlr = new TLinearRegression( _completeDS, _sCompleteInputAttributes, _sOutputAttributes.Data(), option);
      if (_blog)
	tlr->setLog();
      tlr->estimate();
      Double_t dval;
      dval = -1.0;
      switch (_nCriteria)
        {
        case kR2:
	  dval = tlr->getR2();
	  break;
        case kR2A:
	  dval = tlr->getR2Adjusted();
	  break;
        case kQ2:
	  dval = tlr->getQ2();
	  break;
	default:
	    break;
        }
      dval_complete = dval;
      if (_blog)
	cout << " *** Complete Criteria[" << dval_complete << "] Threshold["
	     << _dCriteria << "]" << endl;

      // Del the output hat attribute
      if (_completeDS->isAttribute(Form("%shat", _sOutputAttributes.Data())))
	_completeDS->delAttribute(Form("%shat", _sOutputAttributes.Data()));
      try
        {
	  if ((_nCriteria == kR2) && (dval_complete < _dCriteria))
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form(
							    "TPolynomialRegression::estimate\n The criteria [%f] will never be reached with the maximal order value [%d] where the R2 criteria is [%f]",
							    _dCriteria, _nMaxOrder, dval_complete));
        } catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
	  ue.printMessage();
	  delete tlr;
	  tlr=NULL;
	  throw ue;
        }
      delete tlr;
      tlr=NULL;
    }//if(bOK)

  TString sRetainedMonomials = ""; //_sInputAttributes;

  // the criteria can perhaps be reached, we look for the "elected" polynomial
  // first try : with defined input attributes without any add
  Bool_t bOKcriteria = kFALSE;
  Double_t dval;
  if (_nCriteria != kUnknown)
    (*_fLogger) << " **     Initial polynomial : " << Endl;
  if( tlr != NULL) { delete tlr; tlr=NULL;}
  tlr = new TLinearRegression( _completeDS, _sInputAttributes.Data(), _sOutputAttributes.Data(), _sLROption);
  tlr->estimate();

  // we first try with defined input attribues without any add
  dval = -1.0;
  switch (_nCriteria)
    {
    case kR2:
      dval = tlr->getR2();
      break;
    case kR2A:
      dval = tlr->getR2Adjusted();
      break;
    case kQ2:
      dval = tlr->getQ2();
      break;
    default:
	break;
    }

  if (_nCriteria != kUnknown)
    {
      (*_fLogger) << " **                            with criteria ";

      switch (_nCriteria)
        {
        case kR2:
	  (*_fLogger) << " R2 = ";
	  break;
        case kR2A:
	  (*_fLogger) << " R2A = ";
	  break;
        case kQ2:
	  (*_fLogger) << " Q2 = ";
	  break;
	default:
	    break;
        }
      // The value of the criteria
      (*_fLogger) << dval;
      // the threshold value if not exhaustive research
      if (_nMaxMonomes < 0)
	(*_fLogger) << " / " << _dCriteria;
      (*_fLogger) << Endl;
    }
  _dAllCriteria[0] = tlr->getR2();
  _dAllCriteria[1] = tlr->getR2Adjusted();
  _dAllCriteria[2] = tlr->getQ2();
  fillIndex(" ", dval, 0);
  if (_blog)
    cout << " Criteria[" << dval << "] Threshold[" << _dCriteria << "]"
	 << endl;

  if (_nMaxMonomes > 0)
    {
      estimateExaustiveResearch(option);
      if(tlr!=NULL) delete tlr;
      tlr=NULL;
      return;
    }

  Double_t _dbestcriteria = 0.0; // list containing criteria

  if (dval > _dCriteria)
    {
      _dbestcriteria = dval;
      bOKcriteria = kTRUE;
      Int_t nparam = tlr->getParameters().GetNrows();
      _parameterValues.ResizeTo(nparam);
      _parameterValues = tlr->getParameters();
    }

  TObject *sbestMonome = NULL;
  TObject *crtNomial;

  TF1 * bis = new TF1("bb", Form("%e", _dCriteria), 0, 50.0);
  _drawingGarbageCollector->Add(bis);
  bis->SetLineColor(3);
  URANIE::DataServer::UTimer timer(_lPropMonoms->GetSize(),
				   "Uranie TPolynomialRegression");

  TCanvas *can = NULL;
  Bool_t bFoundBetter = kFALSE;
  // construction of the good one TLinearRegression
  while (!bOKcriteria)
    {
      timer.Init(_lPropMonoms->GetSize());
      TString _sinputs = _sInputAttributes;

      (*_fLogger) << " **     Searching to add one more Monomial";

      if (sRetainedMonomials.Length() > 0)
        {
	  _sinputs.Append(sRetainedMonomials);
	  (*_fLogger) << " to [" << sRetainedMonomials << "]";
        }
      (*_fLogger) << " : " << Endl;

      //Int_t _numberofinputs = sRetainedMonomials.CountChar(':');

      bFoundBetter = kFALSE;

      for (Int_t ivar = 0; ivar < _lPropMonoms->GetSize(); ivar++)
        {
#ifndef WIN32  
	  if (_bdrawProgressBar)
	    timer.DrawProgressBar(ivar);
#endif    
	  // Delette the output hat attribute
	  if (_completeDS->isAttribute(
				       Form("%shat", _sOutputAttributes.Data())))
	    _completeDS->delAttribute(
				      Form("%shat", _sOutputAttributes.Data()));

	  // we add a new monom
	  crtNomial = _lPropMonoms->At(ivar);
	  if (_blog)
	    cout << "Try with this this inputs [" << _sinputs << "] + ["
		 << crtNomial->GetName() << "]" << endl;

	  if( tlr != NULL) { delete tlr; tlr=NULL;}
	  tlr = new TLinearRegression( _completeDS, Form("%s:%s", _sinputs.Data(), crtNomial->GetName()),
				       _sOutputAttributes.Data(), _sLROption);
	  tlr->estimate();
	  switch (_nCriteria)
            {
            case kR2:
	      dval = tlr->getR2();
	      break;
            case kR2A:
	      dval = tlr->getR2Adjusted();
	      break;
            case kQ2:
	      dval = tlr->getQ2();
	      break;
	    default:
		break;
            }
	  if (_blog)
	    cout << " Criteria[" << dval << "] _dbestcriteria["
		 << _dbestcriteria << "] - Threshold[" << _dCriteria
		 << "]" << endl;

	  _dAllCriteria[0] = tlr->getR2();
	  _dAllCriteria[1] = tlr->getR2Adjusted();
	  _dAllCriteria[2] = tlr->getQ2();

	  fillIndex(
                    Form("%s:%s", sRetainedMonomials.Data(),
			 crtNomial->GetName()), dval, 0);

	  // if the monom has best criteria than previous monoms : we keep this one
	  // else we go one with another monom
	  if (dval > _dbestcriteria)
            {
	      _dbestcriteria = dval;
	      sbestMonome = crtNomial;
	      _dr2 = tlr->getR2();
	      _dr2a = tlr->getR2Adjusted();
	      _dq2 = tlr->getQ2();
	      bFoundBetter = kTRUE;

	      if (dval > _dCriteria)
                {
		  bOKcriteria = kTRUE;
		  Int_t nparam = tlr->getParameters().GetNrows();
		  _parameterValues.ResizeTo(nparam);
		  _parameterValues = tlr->getParameters();
		  //               break ;
                } // if dval > _dcriteria
            }
        } // for ivar

      try
        {
	  if (bFoundBetter)
            {
	      (*_fLogger) << " **        Retained Monomial ["
			  << sbestMonome->GetName() << "] with criteria " << _dbestcriteria << "/" << _dCriteria << Endl;

	      _lPropMonoms->Remove(sbestMonome);
	      sRetainedMonomials += Form(":%s", sbestMonome->GetName());
	      _dAllCriteria[0] = _dr2;
	      _dAllCriteria[1] = _dr2a;
	      _dAllCriteria[2] = _dq2;

	      delete sbestMonome;
	      sbestMonome=NULL;
	      
	      fillIndex(sRetainedMonomials, _dbestcriteria, 1);

	      std::stringstream sstr; 
	      if(can==NULL)
	      {
		  sstr << "tplrcan" << _drawingGarbageCollector->GetEntries() ;
		  can = new TCanvas(sstr.str().c_str(),sstr.str().c_str(),1);	      
		  _drawingGarbageCollector->Add(can);
	      }
	      else
		  can->Clear();
	      
	      _ntresult->SetMarkerSize(0.8);
	      _ntresult->SetMarkerColor(2);
	      _ntresult->Draw("Criteria:n", "", "p");
	      // TGraph *gr = (TGraph*)can->GetPrimitive("Graph");
	      // sstr.str(""); sstr << "TPR_graph_"<< _drawingGarbageCollector->GetEntries();
	      // gr->SetName(sstr.str().c_str());
	      // _drawingGarbageCollector->Add(gr);
	      
	      _ntresult->SetMarkerSize(1.);
	      _ntresult->SetMarkerColor(4);
	      _ntresult->Draw("Criteria:n", "((b==1)||(n==0))", "lpsame");
	      // gr = (TGraph*)can->GetPrimitive("Graph");
	      // sstr.str(""); sstr << "TPR_graph_"<< _drawingGarbageCollector->GetEntries();
	      // gr->SetName(sstr.str().c_str());
	      // _drawingGarbageCollector->Add(gr);
	      _ntresult->SetMarkerColor(2);
	      _ntresult->SetMarkerSize(0.8);
	      bis->Draw("same");
            }
	  else
            {
	      (*_fLogger) << Endl<< URANIE::DataServer::UMessageLogger::kERROR << Endl << "***  Not found better than the last criteria " << _dbestcriteria << "/" << _dCriteria << Endl;
	      (*_fLogger) << "***  The threshold [" << _dCriteria << "] can not be reach" << Endl << URANIE::DataServer::UMessageLogger::kINFO;
	      bOKcriteria = kTRUE;
	      throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							 Form("TPolynomialRegression::estimate   ERROR\nThe threshold [%e] has not reached", _dCriteria));
            }
	  //    if ( !bOKcriteria )
	  //      _dbestcriteria = 0.0 ;

	  if(_bdrawProgressBar)
            (*_fLogger) << URANIE::DataServer::UMessageLogger::kINFO << Endl << "time elapsed " << timer.GetElapsedTime() << Endl;

	  if ((_lPropMonoms->GetSize() == 0) && (!bOKcriteria))
            {
	      throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							 Form("TPolynomialRegression::estimate\n No polynomial of %d maximal reachs the criteria %e. The last best one is %e ", _nMaxOrder, _dCriteria, _dbestcriteria));
            }
        }
      catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
	  ue.printMessage();
	  if( tlr != NULL) { delete tlr; tlr=NULL;}
	  // if( bis != NULL) delete bis;
	  // bis=NULL;
	  throw ue;
        }

    } // while

  (*_fLogger) << " **  Final Retained Monomial [" << sRetainedMonomials
	      << "] with criteria " << _dbestcriteria << "/" << _dCriteria << Endl;
  TString sfinalinputs = _sInputAttributes;
  if (sRetainedMonomials.Length() > 0)
    sfinalinputs.Append(sRetainedMonomials);

  SetName(
	  Form(
	       "Modeler[PolynomialRegression]Tds[%s]FinalPredictor[%s]Target[%s]",
	       _tds->GetName(), sfinalinputs.Data(),
	       _sOutputAttributes.Data()));
  SetTitle(
	   Form(
		"PolynomialRegression as modeler for TDS[%s] with Target is[%s] versus[%s]",
		_tds->GetName(), _sOutputAttributes.Data(),
		sfinalinputs.Data()));

  // Create the final Linear Regression on the original TDS
  TString saddedmonomials = sRetainedMonomials;
  saddedmonomials.ReplaceAll(":", " ");
  saddedmonomials += TString(" ");
  if (_blog)
    cout << " ** saddedmonomials[" << saddedmonomials << "]" << endl;
  istringstream inStream((char *) saddedmonomials.Data());
  string inName;
  //  TFile* file= _tds->getTuple()->GetCurrentFile();
  while (inStream >> inName)
    {
      TString satt(inName.c_str());
      URANIE::DataServer::TAttributeFormula *att =
	dynamic_cast<URANIE::DataServer::TAttributeFormula *>(_completeDS->getAttribute(
											satt));

      if (_blog)
        {
	  cout << " ** satt[" << satt << "] formula[" << att->getFormula()
	       << "]" << endl;
	  att->printLog();
        }

      _tds->addAttribute(satt, att->getFormula(),
			 Form("title %s", inName.c_str()));
    } // while

  // // Clean
  // delete _completeDS;
  // _completeDS = NULL;

  // Change the LROption to store the  yhat attribute in the final (or retained) Linear Regression
  TString sLROption = _sLROption;
  sLROption.ReplaceAll("nostoreyhat", "");

  _theTLR = new TLinearRegression(_tds, _sInputAttributes + sRetainedMonomials, _sOutputAttributes.Data(),
				  sLROption);
  _theTLR->estimate();

  _dq2 = _theTLR->getQ2();
  _dr2a = _theTLR->getR2Adjusted();
  _dr2 = _theTLR->getR2();
  _dPress = _theTLR->getPress();

  _parameterValues.ResizeTo(_theTLR->getParameters().GetNrows());
  _parameterValues = _theTLR->getParameters();

  //clean
  if( tlr != NULL) { delete tlr; tlr=NULL;}
  // if( bis != NULL) delete bis;
  // bis=NULL;
  if (_blog)
    {
      cout << " ** End of TPolynomialRegression::estimate" << endl;
      cout << "**************************************" << endl;
    }
}

// -----------------------------------------------------------------------
//  estimateExaustiveResearch 
// -----------------------------------------------------------------------
void TPolynomialRegression::estimateExaustiveResearch(Option_t * option)
{
    if (_blog)
    {
        cout << endl << "**************************************" << endl;
        cout << " ** TPolynomialRegression::estimateExaustiveResearch Option ["
                << option << "]" << endl;
        _lPropMonoms->Print("");
    }

    URANIE::Modeler::TLinearRegression *tlr=NULL;
    
    TString sListOfAttributes = "";
    Int_t nInteractionSecondOrder = _lPropMonoms->GetSize();
    (*_fLogger) << " ** Number of second Order Interaction (Exaustive Research) : "
            << nInteractionSecondOrder << Endl;

    Int_t nSum = 1; // Original polynome without any second order monomial

    Int_t nTotalPolynomial = nSum + nInteractionSecondOrder;
    for (Int_t i = 2; i <= _nMaxMonomes; i++)
        nTotalPolynomial +=
                (Int_t)(
                        TMath::Factorial(nInteractionSecondOrder)
                                / (TMath::Factorial(i)
                                        * TMath::Factorial(
                                                nInteractionSecondOrder - i)));
    (*_fLogger) << " ** Total Number of Polynomials (Exaustive Research) : "
            << nTotalPolynomial << Endl;

    URANIE::DataServer::UTimer timer(nTotalPolynomial,
            "Uranie TPolynomialRegression Exaustive Research");

    for (Int_t ivar = 0; ivar < nInteractionSecondOrder; ivar++)
    {
        _nNumber = 0;
        recursiveSearch(_lPropMonoms, ivar, sListOfAttributes);
        nSum += _nNumber;
        time_t seconds;
        time(&seconds);

        if (_nCriteria != kUnknown)
        {
            _ntresult->Draw("Criteria:n");
            gPad->Modified();
            gPad->Update();
        }
#ifndef WIN32
        if (_bdrawProgressBar)
        {
            timer.DrawProgressBar(nSum);
        }
#endif
        if (_bTrace)
        {
            std::ofstream fTraceFile;
            fTraceFile.open(_sTraceFile.Data(), std::ios::out);
            cout << " ************* recursiveSearch  [" << _nNumber
                    << "] :: nPCT[" << 100.0 * nSum / nTotalPolynomial << "]"
                    << endl;
            fTraceFile << 100.0 * nSum / nTotalPolynomial << endl;
            fTraceFile.flush();
            fTraceFile.close();
        }
    }
    if (_bdrawProgressBar)
        cout << endl << " ** time elapsed " << timer.GetElapsedTime() << endl;

    // Loop to search the maximum for each number of monomes
    if (_nCriteria != kUnknown)
    {
        (*_fLogger) << " ** search the maximum with [" << _nMaxMonomes
                << "] monomes maximum." << Endl;

        Double_t dval = -1.0;
        Double_t dbestCriteria = 0.0;
        TString sbestMonom;

        for(Int_t i=0; i < _nMaxMonomes +1; i++)
        {
            (*_fLogger) << " **   (" << i << "/" << _nMaxMonomes << ") monomes ";
            Long64_t ndata = _ntresult->Draw("Entry$:Criteria", Form("(n==%d)&&(b==0)", i), "goff");
            if (ndata > 1000000)
            {
                  _ntresult->SetEstimate( ndata );
                  _ntresult->Draw("Entry$:Criteria", Form("(n==%d)&&(b==0)", i), "goff");
            }
            (*_fLogger) << " size (" << _ntresult->GetSelectedRows() << ")";

            Double_t *vv = _ntresult->GetV1();
            //Double_t dmax = TMath::MaxElement(_ntresult->GetSelectedRows(), _ntresult->GetV2());

            Long_t dindmax = (Long_t)TMath::LocMax(_ntresult->GetSelectedRows(), _ntresult->GetV2());
            _ntresult->GetEntry(Long_t(vv[dindmax]));
            if ( _blog )
              _ntresult->Show();

            TString strName = _sMonomials;
            (*_fLogger) << " _sMonomials [" << _sMonomials << "]";

            strName.ReplaceAll("*", "");

	    if(tlr!=NULL) delete tlr;
            tlr = new TLinearRegression(_completeDS, Form("%s%s", _sInputAttributes.Data(), strName.Data()), _sOutputAttributes.Data(), _sLROption);
            tlr->estimate();

            _dAllCriteria[0] = tlr->getR2();
            _dAllCriteria[1] = tlr->getR2Adjusted();
            _dAllCriteria[2] = tlr->getQ2();
            switch (_nCriteria)
            {
	    case kR2:
                dval = _dAllCriteria[0];
                (*_fLogger) << " R2  [" << dval << "]";
                break;
	    case kR2A:
                dval = _dAllCriteria[1];
                (*_fLogger) << " R2A  [" << dval << "]";
                break;
	    case kQ2:
                dval = _dAllCriteria[2];
                (*_fLogger) << " Q2  [" << dval << "]";
                break;
	    default:
		break;
            }

            if ( dval > _dCriteria)
              fillIndex( TString(_sMonomials), dval, 1);

            // init the default value of best criteria
            //   with the value of criteria without 2nd order monomes
            //   - no rule for Q2
            //   - R2 and R2 > 0
            if ( i == 0 )
              dbestCriteria = 0.99 * dval;

            // if the monom has best criteria than previous monoms : we keep this one
            // else we go one with another monom
            if (dval > dbestCriteria)
            {
                dbestCriteria=dval;
                sbestMonom = TString(_sMonomials);
                _dr2 = tlr->getR2();
                _dr2a = tlr->getR2Adjusted();
                _dq2 = tlr->getQ2();
            }

            (*_fLogger) << Endl;
            // Clean
            delete tlr;
	    tlr=NULL;
        }//for i <nMonomes

        {
            // Terminate state for the Exhaustive Research
            // \todo 12/04/2013 11:57 FGA to add in the terminate state of the method and add the init state with (_generateCompletePolynomial)
            (*_fLogger) << " ** Final Retained Monomial [" << sbestMonom << "] with criteria [" << dbestCriteria << "]" << Endl;
            sbestMonom.ReplaceAll("*", "");
            TString sfinalinputs = _sInputAttributes;
            if ( sbestMonom.Length() > 0 )
            sfinalinputs.Append(sbestMonom);

            SetName(Form("Modeler[PolynomialRegression]Tds[%s]FinalPredictor[%s]Target[%s]", _tds->GetName(), sfinalinputs.Data(), _sOutputAttributes.Data()));
            SetTitle(Form("PolynomialRegression as modeler for TDS[%s] with Target is[%s] versus[%s]", _tds->GetName(), _sOutputAttributes.Data(), sfinalinputs.Data()));

            // Create the final Linear Regression on the original TDS
            TString saddedmonomials =sbestMonom;
            saddedmonomials.ReplaceAll(":", " ");
            saddedmonomials += TString(" ");
            if ( _blog )
            cout << " ** saddedmonomials[" << saddedmonomials << "]"<<endl;
            istringstream inStream((char *) saddedmonomials.Data());
            string inName;
            //  TFile* file= _tds->getTuple()->GetCurrentFile();
            while (inStream >> inName)
            {
                TString satt(inName.c_str());
                URANIE::DataServer::TAttributeFormula *att = dynamic_cast<URANIE::DataServer::TAttributeFormula *>(_completeDS->getAttribute(satt));

                if ( _blog )
                {
                    cout << " ** satt[" << satt << "] formula[" << att->getFormula() <<"]"<<endl;
                    att->printLog();
                }

		cout<<__LINE__<<" add "<<satt<<endl;
                _tds->addAttribute(satt, att->getFormula(), Form("title %s", inName.c_str()));
            } // while

            // // Clean
            // delete _completeDS; _completeDS = NULL;

            TString sLROption = _sLROption;
            sLROption.ReplaceAll("nostoreyhat", "");

            _theTLR = new TLinearRegression(_tds, _sInputAttributes + sbestMonom , _sOutputAttributes.Data(), sLROption);
            _theTLR->estimate();

            _dq2 = _theTLR->getQ2();
            _dr2a= _theTLR->getR2Adjusted();
            _dr2= _theTLR->getR2();
            _dPress = _theTLR->getPress();

            _parameterValues.ResizeTo(_theTLR->getParameters().GetNrows());
            _parameterValues = _theTLR->getParameters();
        }
    }

    if (_blog)
    {
        cout << " ** End of TPolynomialRegression::estimateExaustiveResearch"
                << endl;
        cout << "**************************************" << endl;
    }
}

void TPolynomialRegression::recursiveSearch(TList *tl, Int_t npos, TString str)
{

    if (npos < tl->GetSize())
    {
        if (str.CountChar(':') < _nMaxMonomes)
        {
            str += TString(":") + TString(tl->At(npos)->GetTitle());

            if (_blog)
                cout << " ** TPolynomialRegression::recursiveSearch npos["
                        << npos << "] str (" << str << ")" << endl;

            // Del the output hat attribute
            if (_completeDS->isAttribute(
                    Form("%shat", _sOutputAttributes.Data())))
                _completeDS->delAttribute(
                        Form("%shat", _sOutputAttributes.Data()));

            TString strName = str;
            strName.ReplaceAll("*", "");

            URANIE::Modeler::TLinearRegression * tlr = new TLinearRegression(
                    _completeDS,
                    Form("%s%s", _sInputAttributes.Data(), strName.Data()),
                    _sOutputAttributes.Data(), _sLROption);
            tlr->estimate();

            _dAllCriteria[0] = tlr->getR2();
            _dAllCriteria[1] = tlr->getR2Adjusted();
            _dAllCriteria[2] = tlr->getQ2();

            switch (_nCriteria)
            {
            case kR2:
                fillIndex(str, _dAllCriteria[0], 0);
                break;
            case kR2A:
                fillIndex(str, _dAllCriteria[1], 0);
                break;
            case kQ2:
                fillIndex(str, _dAllCriteria[2], 0);
                break;
            case kUnknown:
                fillIndex(str, -1.0, 0);
                break;
            }

            // Clean
            delete tlr;
	    tlr=NULL;
	      
            if (_blog)
                cout << "    *** str(" << str << ") Q2(" << _dAllCriteria[2]
                        << ")" << endl;
            _nNumber++;

            for (Int_t n = npos + 1; n < tl->GetSize(); n++)
                recursiveSearch(tl, n, str);
        }
    }
}

// -----------------------------------------------------------------------
//  fillIndex 
// -----------------------------------------------------------------------
void TPolynomialRegression::fillIndex(TString sinputname, Double_t dval,
        Int_t bretained)
{
    if (_blog)
        cout << " ** TPolynomialRegression::fillIndex sinputname[" << sinputname
                << "] value[" << dval << "] retained[" << bretained << "]"
                << endl;
    _nMonomials = sinputname.CountChar(':');
    sprintf(_sMonomials, "%s", sinputname.Data());
    _dValueCriteria = dval;
    _bRetained = bretained;

    _ntresult->Fill();
}

// -----------------------------------------------------------------------
//  _generateCompletePolynomial 
// -----------------------------------------------------------------------
Bool_t TPolynomialRegression::_generateCompletePolynomial()
{
    if (_blog)
        cout << endl
                << " **TPolynomialRegression::_generateCompletePolynomial _sInputAttributes["
                << _sInputAttributes << "]" << endl;
    _sCompleteInputAttributes = TString(_sInputAttributes);
    // pour le moment seulement degre 2 : il faut reflechir a l'algo pour faire degre superieur
    Int_t _ncompletenX = _nX + 1;
    _ncompletenX += _nX * (_nX + 1) / 2;
    if (_blog)
        cout << " ** _ncompletenX[" << _ncompletenX << "]" << endl;
    if (_bChangemaxorder)
    {
        TIter next(_listOfInputAttributes);
        Int_t itomax = 1;
        while (itomax < _nMaxOrder)
        {
            for (Int_t ivar = 0; ivar < _nX; ivar++)
            {
                URANIE::DataServer::TAttribute *att =
                        (URANIE::DataServer::TAttribute *) _listOfInputAttributes->At(
                                ivar);
                Int_t jvar = ivar;
                if (_bIntercept == kFALSE)
                    jvar++;
                for (Int_t ivar1 = jvar; ivar1 < _nX; ivar1++)
                {
                    URANIE::DataServer::TAttribute *att1 =
                            (URANIE::DataServer::TAttribute *) _listOfInputAttributes->At(
                                    ivar1);
                    TString sname = Form("%s%s", att->GetName(),
                            att1->GetName());
                    TString sformula = Form("%s*%s", att->GetName(),
                            att1->GetName()); // formula of the new monom
                    _lPropMonoms->Add(
                            new TNamed(sname.Data(),
                                    Form("%s*%s", att->GetName(),
                                            att1->GetName())));
                    _sCompleteInputAttributes.Append(":" + sname);

                    _completeDS->addAttribute(sname,
                            Form("%s*%s", att->GetName(), att1->GetName()));
                } // for  ivar1 : second order
                itomax++;
            } // for ivar
        } // while itomax
        _bChangemaxorder = kFALSE;

    } // if
    if (_blog)
        cout << " ** _nS [" << _nS << "]" << endl;

    try
    {
        if (_nS < _ncompletenX)
        {
            _OKcomplete = kFALSE;
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                    Form(
                            "TPolynomialRegression::generateCompletePolynomial Method:\n Impossible to generate polynomial with complete monoms\n there is not enough experiment [%d] than the number of unknowns [%d]",
                            _nS, _ncompletenX));
        }
        else
        {
            _OKcomplete = kTRUE;
        }
    } catch (URANIE::Exceptions::UExceptions& ue)
    {
        ue.printMessage();
    }
    if (_blog)
    {
        cout << "All possible monoms are " << _sCompleteInputAttributes << endl;
    }

    return _OKcomplete;
} //_generateCompletePolynomial

// -----------------------------------------------------------------------
//  Export Functions
// -----------------------------------------------------------------------
// void TPolynomialRegression::exportFunction(const char* lang, const char* file,
//         const char* name)
// {
//     _theTLR->exportFunction(lang, file, name);
// }
void TPolynomialRegression::exportModelCplusplus(std::ofstream * sourcefile) const
{
    _theTLR->exportModelCplusplus(sourcefile);
}

void TPolynomialRegression::exportModelFortran(std::ofstream * sourcefile) const
{
    _theTLR->exportModelFortran(sourcefile);
}

// -----------------------------------------------------------------------
//  Log
// -----------------------------------------------------------------------
void TPolynomialRegression::printLog(Option_t *option)
{
    cout << "*******************************" << endl;
    cout << " ** TPolynomialRegression::printLog[" << option << "]" << endl;
    cout << " ** TModeler::printLog[" << option << "]" << endl;
    cout << " ** Name[" << GetName() << "]" << endl;
    cout << " **   Title[" << GetTitle() << "]" << endl;
    if (_bTrace)
        cout << "Yes in file[" << _sTraceFile << "]";
    if (_bIntercept)
        cout << " ** _bIntercept[" << _bIntercept << " - YES]" << endl;
    else
        cout << " ** _bIntercept[" << _bIntercept << " - NO]" << endl;
    cout << " ** _sLROption[" << _sLROption << "]" << endl;
    cout << " ** Order _nMinOrder[" << _nMinOrder << "] _nMaxOrder["
            << _nMaxOrder << "]" << endl;
    cout << " ** _sOutputAttributes[" << _sOutputAttributes << "]" << endl;
    cout << " **  R2  [" << _dr2 << "]" << endl;
    cout << " **  R2a [" << _dr2a << "]" << endl;
    cout << " **  PRESS[" << _dPress << "]" << endl;
    cout << " **  Q2  [" << _dq2 << "]" << endl;
    cout << "*******************************" << endl;
}

} // Fin du namespace Modeler
} // Fin du namespace URANIE

