//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TPolynomialRegression.h
 \author Estelle Deville <estelle.deville@cea.fr>
 \date   jeu jun 5 10:35:25 CEST 2011
 
 \brief Interface de la classe URANIE::Modeler::TPolynomialRegression
 
 */
/**
 
 * @class URANIE::Modeler::TPolynomialRegression
 *
 * @brief Description of the class TPolynomialRegression.
 *
 * To write by the developper.

 \par Default value
 \li \b Criteria : \b \f$Q^2\f$
 \li \b Value : \b 0.70
 \li \b MinOder : \b 1
 \li \b MaxOrder : \b 2
 *
 */

#ifndef TPOLYNOMIALREGRESSION_H
#define TPOLYNOMIALREGRESSION_H

#define MAXLENGTHSTRING 250

//ROOT

// Uranie
#include "TModeler.h"
#include "TDataServer.h"
#include "UMessageLogger.h"
#include "TLinearRegression.h"

namespace URANIE
{
namespace Modeler
{
class TPolynomialRegression: public TModeler
{
public:
    enum ECriteria
    {
        kUnknown, kR2, kR2A, kQ2
    };
protected:
    // the mutable declaration is needed to use the logger in const methods
    URANIE::DataServer::UMessageLogger *_fLogger;   ///  Message logger

private:
    ECriteria _nCriteria;                     ///< The estimation option choosen
    Double_t _dCriteria;                       ///< The value of wanted criteria

    Bool_t _bChangemaxorder; ///  True if value of maximal order of the polynomial has changed
    Bool_t _OKcomplete;            ///  True if _complete data server can be use
    URANIE::Modeler::TLinearRegression *_theTLR; ///< The choosen non linear regression 
    URANIE::DataServer::TDataServer *_completeDS; ///< The complete data server for the maximalorder wanted
    TString _sCompleteInputAttributes; ///< The string of complete input attributes (separated by the character ":"
    TList *_lPropMonoms; ///  The list of monoms containing potential candidates
    TString _sLROption; ///< The string of option to pass to the TLinearRegression (\em "nointercept" for example)
    TObjArray *_drawingGarbageCollector; ///<Garbage collector for drawn objects
    
public:
    Double_t _dr2a;                                ///< The Adjusted R2 quality
    Double_t _dq2;                                 ///< The Q2 quality
    Double_t _dPress;                              ///< The Press quality

    TTree* _ntresult;                             ///< The TTree of results
    Char_t _sMonomials[MAXLENGTHSTRING];     ///  The list of retained monomials
    Int_t _nMonomials;                     ///  The number of retained monomials
    Int_t _bRetained;        ///  The number to known if the momonials are added
    Double_t _dValueCriteria; ///  The current value of the criteria ( fill the TTree)
    Double_t _dAllCriteria[3]; ///  The vector of 3 criteria (R2, R2A, Q2) ( fill the TTree)

    Int_t _nMinOrder;                   ///< The minimal order of the polynomial
    Int_t _nMaxOrder;                   ///< The maximal order of the polynomial
    Int_t _nMaxMonomes; ///< The maximal number of monomia to insert (exaustive research)
    Int_t _nNumber;                          ///< the current number of polynome

    Bool_t _bTrace;                    ///< The flag to known is trace is needed
    TString _sTraceFile;                         ///< The name of the trace file
    // Operations
private:
    Bool_t _generateCompletePolynomial();
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor with the TDataServer and architecture
    /*!
     If you don't want a \em intercept for the linear regression, put \b "nointercept" in the option argument.
     
     \code
     TPolynomialRegression * tpr = new TPolynomialRegression(tds, TString("i1:i2:i3,out"), "nointercept");
     tpr->setCriteria(0.977);
     tpr->estimate("Q2");
     \endcode
     
     \param tds (URANIE::DataServer::TDataServer *) the dataserver
     \param architecture (TString)[""] The list of input and output attributes to pass to the function. The names of attributes are separated by the character \em ":" and the type (\em Input or \em Output) are separated by the character \em "," 
     \param option (Option_t *)[""] The option to pass
     */
    TPolynomialRegression(URANIE::DataServer::TDataServer *tds,
            TString architecture, Option_t *option = "");
    //! Default constructor with the TDataServer and input and output attributes
    /*!
     If you don't want a \em intercept for the linear regression, put \b "nointercept" in the option argument.

     \code
     TPolynomialRegression * tpr = new TPolynomialRegression(tds, "i1:i2:i3", "out", "nointercept");
     tpr->setCriteria(0.977);
     tpr->estimate("Q2");
     \endcode

     \param tds (URANIE::DataServer::TDataServer *) the dataserver
     \param varexpinput (const char*) The list of input attributes to pass to the function separated by the caracter ":"
     \param varexpoutput (const char *)  The list of output attributes separated by the caracter ":"
     \param option (Option_t *)[""] The option to pass
     */
    TPolynomialRegression(URANIE::DataServer::TDataServer *tds,
            const char *varexpinput, const char *varexpoutput,
            Option_t * option = "");
    //! Default destructor
    virtual ~TPolynomialRegression();
    //@}

    //---------------------------------------------
    /** @name Get and Set methods
     **/
    //@{
    //! Return the Adjusted R2 statistic
    Double_t getR2Adjusted()
    {
        return _dr2a;
    }
    //! Return the Q2 statistic
    Double_t getQ2()
    {
        return _dq2;
    }
    //! Return the PRESS statistic
    Double_t getPress()
    {
        return _dPress;
    }
    //!Set the minimal order of the polynomial
    /*!
     \brief Set the minimal  order of the polynomial .
     \param iorder : order min of the polynomial.
     */
    void setMinimalOrder(Int_t iorder)
    {
        if (iorder < 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TPolynomialRegression::setMinimalOrder - Error: invalid value for minimal order (Need to be >0)");
        else if (iorder > _nMaxOrder)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TPolynomialRegression::setMinimalOrder - Error: invalid value for minimal order (Need to be lower than maximal order)");
        _nMinOrder = iorder;
    }

    //!Set the maximal order of the polynomial
    /*!
     \brief Set the maximal  order of the polynomial .
     \param iorder : order max of the polynomial.
     */
    void setMaximalOrder(Int_t iorder)
    {
        if (iorder < 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TPolynomialRegression::setMaximalOrder - Error: invalid value for maximal order (Need to be >0)");
        else if (iorder < _nMinOrder)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TPolynomialRegression::setMaximalOrder - Error: invalid value for maximal order (Need to be greater than minimal order)");
        else if (iorder > 2)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TPolynomialRegression::setMaximalOrder - Error: a maximal order greater than 2 is under development");
        _nMaxOrder = iorder;
        _bChangemaxorder = true;
    }
    //! Set the maximal number of monomia to add (exaustive research)
    /*!
     \param n (Int_t) : The limit of number of monomia to add in the polynomial
     */
    void setMaxMonomes(Int_t n);
    //! Return the maximal number of monomia
    Int_t getMaxMonomes()
    {
        return _nMaxMonomes;
    }

    //! Activate the trace of the process
    /*!
     the trace is used only in the generateCandidate method
     
     /param straceFile (TString) The name of the trace file (empty by default, then use the default name which is \b "_tpr_trace_.log" )
     */
    void setTrace(TString straceFile = "");
    //! Unset trace
    void unsetTrace()
    {
        _bTrace = kFALSE;
    }
    //! Return the trace flag
    Bool_t getTrace()
    {
        return _bTrace;
    }
    //! Return the name of the trace file
    TString getTraceFile();

    //!Set the wanted criteria
    /*!
     \brief Set the wanted criteria for estimate function .
     \param dcriteria : wanted criteria.
     */
    void setCriteria(Double_t dcriteria)
    {
        if (dcriteria < 0.)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TPolynomialRegression::setCriteria - Error: invalid value for criteria (Need to be >0)");
        else if (dcriteria > 1.)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TPolynomialRegression::setCriteria - Error: invalid value for criteria (Need to lower than 1)");
        _dCriteria = dcriteria;
    }
    //! Get the TTree of results
    /*!
     The TTree of results contains:
     \li \b Inp : the list of inputs attributes added
     \li \b n : the number of inputs attributes added
     \li \b b : a number to known if the attributes are retained (b==1) else b==0
     \li \b Criteria : the criteria used to select the monomials
     \li \b R2, \b R2A, \b Q2 : the 3 quality stored in a vector \em C.

     \code
     TTree *tt = tpr->getResultsTree();
     tt->Scan("*");
     \endcode
     to give for example
     \verbatim
     ************************************************************************************************
     *    Row   *       Inp *         n *         b *  Criteria *    C.C.R2 *   C.C.R2A *    C.C.Q2 *
     ************************************************************************************************
     *        0 *           *         0 *         1 * 0.7956222 * 0.9580832 * 0.9161665 * 0.7956222 *
     *        1 *     :z1z1 *         1 *         0 * 0.7568323 * 0.9582542 * 0.9089183 * 0.7568323 *
     *        2 *     :z1z2 *         1 *         0 * 0.7606942 * 0.9584714 * 0.9093923 * 0.7606942 *
     *        3 *     :z1z3 *         1 *         0 * 0.8093517 * 0.9679445 * 0.9300608 * 0.8093517 *
     *        4 *     :z1z4 *         1 *         0 * 0.7857819 * 0.9665406 * 0.9269978 * 0.7857819 *
     *        5 *     :z1z5 *         1 *         0 * 0.7751127 * 0.9599555 * 0.9126302 * 0.7751127 *
     *        6 *     :z1z6 *         1 *         0 * 0.7687018 * 0.9596316 * 0.9119236 * 0.7687018 *
     *        7 *     :z1z7 *         1 *         0 * 0.7510527 * 0.9595362 * 0.9117153 * 0.7510527 *
     *        8 *     :z1z8 *         1 *         0 * 0.7116922 * 0.9583941 * 0.9092235 * 0.7116922 *
     ...
     *      267 * :z3z6:z5z *         3 *         0 * 0.9572626 * 0.9947547 * 0.9860127 * 0.9572626 *
     *      268 * :z3z6:z5z *         3 *         0 * 0.9603144 * 0.9944556 * 0.9852151 * 0.9603144 *
     *      269 * :z3z6:z5z *         3 *         0 * 0.9582744 * 0.9944016 * 0.9850711 * 0.9582744 *
     *      270 * :z3z6:z5z *         3 *         0 * 0.9044754 * 0.9956092 * 0.9882912 * 0.9044754 *
     *      271 * :z3z6:z5z *         3 *         0 * 0.9566059 * 0.9944364 * 0.9851638 * 0.9566059 *
     *      272 * :z3z6:z5z *         3 *         0 * 0.9686468 * 0.9956805 * 0.9884815 * 0.9686468 *
     *      273 * :z3z6:z5z *         3 *         1 * 0.9779799 * 0.9973796 * 0.9930124 * 0.9779799 *
     ************************************************************************************************
     \endverbatim
     
     \return (TTree *) the TTree of results.
     */
    TTree * getResultsTree()
    {
        return _ntresult;
    }
    //@}
    //---------------------------------------------
    /** @name Estimate
     **/
    //@{
    void estimate(Option_t * option = "");

    //! Exhaustive research when the number of maximal monomes is no-negative
    /*!
     */
    void estimateExaustiveResearch(Option_t * option);
    void recursiveSearch(TList * tl, Int_t npos, TString str);

    //! Method to fill in the tree the value of temporal results
    /*!
     \param sinputname (const char *) : the name of the retained attributes
     \param dval (Double_t) : the value of the quality
     \param bRetained (Int_t)[0] A flag to known if the attributes are retained (1) or not retained (0)
     */
    void fillIndex(TString sinputname, Double_t dval, Int_t bRetained = 0);
    //@}

    //---------------------------------------------
    /** @name Export the model in extrenal langage
     **/
    //@{
    //! Export the model in an external file with a specified langage
    /*!
     It will be the main interface function to the export.
     \param lang (const char *) the langage to export (C|Fortran)
     \param file (const char *) the file to export the modeler. If empty, use the name of the model, plus the extension of the langage.
     \param name (const char *) the name of the function in the export file. If empty, use the name of the model
     */
//     void exportFunction(const char* lang, const char* file = "",
//             const char* name = "");
    //! Export the model in C++ langage in a file
    /*!
     \warning Don't use this method. Use the main method TModeler::exportFunction.
     \param sourcefile (ofstream *) the pointer to the file to export the modeler.
     */
    void exportModelCplusplus(std::ofstream * sourcefile) const;
    //! Export the model in Fortran langage in a file
    /*!
     \warning Don't use this method. Use the main method TModeler::exportFunction.
     \param sourcefile (ofstream *) the pointer to the file to export the modeler.
     */
    void exportModelFortran(std::ofstream * sourcefile) const;
    
    //! Export the model in a PMML file (not yet implemented)
    /*!
     \warning Don't use this method. Use the main method TModeler::exportFunction.
     \param sourcefile (ofstream *) the pointer to the file to export the modeler.
     */
    void exportModelPMML(const char* file = "", 
                         const char* name = "",Option_t *option = "") const
    {
        TString soption = option;
        soption.ToLower();
	UNUSED(file,name);
        std::cerr
                << "   --- Method TPolynomialRegression::exportModelPMML(ofstream * sourcefile) not yet implemented"
                << endl;
    };
    
    //! Export the model in Python langage in a file (not yet implemented)
    /*!
     \warning Don't use this method. Use the main method TModeler::exportFunction.
     \param sourcefile (ofstream *) the pointer to the file to export the modeler.
     */
    void exportModelPython(std::ofstream * sourcefile) const
    {
	UNUSED(sourcefile);
        std::cerr
                << "   --- Method TPolynomialRegression::exportModelPython(ofstream * sourcefile) not yet implemented"
                << endl;
    };
    
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Modeler::TPolynomialRegression, ID_MODELER)
    //Classe de
};

} // Fin du namespace Modeler
} // Fin du namespace URANIE
#endif
