//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
  \file   TkNN.cxx
  \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
  \date   Wed Jul  8 22:24:51 2015
  
  \brief Implementation of the class  URANIE::Modeler::TkNN
  
*/

//// ROOT
#include "TLeaf.h"
#include "TPad.h"
#include "TMath.h"
#include "TRegexp.h"

//// URANIE
#include "TkNN.h"
#include "TDataServer.h"

//   GPU
#if defined(WITH_GPUANN) && !defined(__CINT__)
#include "kNN_gpu.h"
#endif

#ifdef _OPENMP
#include <omp.h>
#endif
#include <algorithm>
#include <vector>


#define epsilon 10 * DBL_MIN

using std::cout;
using std::endl;

ClassImp(URANIE::Modeler::TkNN)
  
  namespace URANIE {
    namespace Modeler {
      // -----------------------------------------------------------------------
      //  Constructors
      // -----------------------------------------------------------------------
      TkNN::TkNN(URANIE::DataServer::TDataServer *tds, TString architecture, Option_t *option) :
 	TModeler(tds, architecture, option),
    _nK(-1),
    _nDistance(kEuclidean),
	_sDraw(""),
	_tdsQuery(NULL)
      {
    if( _blog )
      cout << "construit TkNN tds[" << _tds->GetName() << "] architecture[" << architecture << "] Option [" << option << "]" << endl;
	
	TString opt = option;
	opt.ToLower();
    if( _blog )
        cout << " ** opt[" << opt << "]" << endl;

	// INIT
	_tds->getTuple()->SetMarkerColor(kBlue);
	_tds->getTuple()->SetMarkerStyle(8);
	_tds->getTuple()->SetMarkerSize(1.5);
	// End Of INIT

    TString sarchi = _sIntermediateArchitecture;
    if( _blog )
        cout << " ** sarchi[" << sarchi << "]" << endl;

    sarchi.ReplaceAll("[", "");
    sarchi.ReplaceAll("]", "");

    _nK = atoi(sarchi.Data());
    if( _blog ) {
        cout << " ** sarchi[" << sarchi << "]" << endl;
        cout << " ** _nK[" << _nK << "]" << endl;
    }

      }
      
      TkNN::TkNN(URANIE::DataServer::TDataServer *tds, const char *varexpinput, const char *varexpoutput, Int_t nkernel, Option_t * option):
	TModeler(tds, varexpinput, varexpoutput, option),
	_nK(nkernel),
	_nDistance(kEuclidean),
	_sDraw(""),
	_tdsQuery(NULL)
      {

	// INIT
	_tds->getTuple()->SetMarkerColor(kBlue);
	_tds->getTuple()->SetMarkerStyle(8);
	_tds->getTuple()->SetMarkerSize(1.5);
	// End Of INIT

	
 	_sFileName = Form("linearmodel_%s", _sOutputAttributes.Data());
 	_sFunctionName = Form("linearmodel_Fct_%s", _sOutputAttributes.Data());
	
	SetName(Form("Modeler[kNN%d]Tds[%s]Predictor[%s]Target[%s]", _nK, _tds->GetName(), _sInputAttributes.Data(), _sOutputAttributes.Data()));
 	SetTitle(Form("kNN as modeler for TDS[%s] with Target is[%s] versus[%s]", _tds->GetName(), _sOutputAttributes.Data(), _sInputAttributes.Data()));
	
      }
      
      // -----------------------------------------------------------------------
      //  Destructor
      // -----------------------------------------------------------------------
      TkNN::~TkNN()
      {
	if ( _blog )
	  cout << "detruit TkNN" << endl; 
      }
      
      // -----------------------------------------------------------------------
      //  setter methods
      // -----------------------------------------------------------------------
      void TkNN::setQueryPoints(URANIE::DataServer::TDataServer *tds)
      {

	  TList *lstatts=NULL;
	try {
              lstatts = tds->getListOfAttributes(_sInputAttributes);
	} catch (URANIE::Exceptions::UErrorExceptions& ue)
	{
	    ue.printMessage();
	    delete lstatts; lstatts=NULL;
	    throw ue;
          }					       

	delete lstatts; lstatts=NULL;
	_tdsQuery = tds;
      }

      // -----------------------------------------------------------------------
      //  draw method
      // -----------------------------------------------------------------------
      void TkNN::draw(const char * vars, Option_t * option)
      {
	_sDraw = TString(vars);
	gPad->SetGrid(1,1);
	_tds->draw(_sDraw,"",option);
	if ( _tdsQuery )
	  _tdsQuery->draw(_sDraw, "","same");
	gPad->SaveAs("knn0.png");
	
      }
      // -----------------------------------------------------------------------
      //  estimate methods
      // -----------------------------------------------------------------------
      //  CPU method
      void TkNN::estimate_cpu(Option_t * option)
      {
    if( _blog ) {
      cout << endl;
      cout << "**************************************" << endl;
      cout << " ** TkNN::estimate on CPU" << endl;
      cout << "**************************************" << endl;
    }

    TString soption = TString(option);
    soption.ToLower();
    soption.ReplaceAll(" ", "");
    Int_t nbthreads = -1;
    if (soption.Contains("nthreads="))
    {
       TRegexp reg("nthreads=[0-9]*");
       TString out = soption(reg);
       TString snthreads = out.Remove(0, 9);
       soption.ReplaceAll(Form("nthreads=%s", out.Data()), "");

       if (snthreads.IsDigit())
         nbthreads = snthreads.Atoi();
       if( _blog )
	 cout << endl << " ** nbthreads read in option[" << nbthreads << "]" << endl;
    }

    // Allocate the vecteur of distance (size of nS of reference TDS)
    Int_t nsr = _tds->getNPatterns();
    Int_t nsq = _tdsQuery->getNPatterns();
    if( _blog )
      cout << " ** nsr[" << nsr << "] nsq[" << nsq << "]" << endl;

    Double_t *vecDistance = new Double_t[nsr];
    Int_t *vecIndex    = new Int_t[nsr];





    TLeaf *lr;
    TLeaf *lry = _tds->getTuple()->GetLeaf(_sOutputAttributes);


    URANIE::DataServer::TAttribute *attY = _tds->getAttribute(_sOutputAttributes);
    if ( _blog )
       attY->printLog();
    TString syhat = Form("%shat%d", _sOutputAttributes.Data(), _nK);

    _tdsQuery->addAttribute ( new URANIE::DataServer::TAttribute(syhat, Form("#hat{%s}", attY->GetTitle())));
    _tdsQuery->getAttribute(syhat)->delShare();
    Double_t *tabValue = new Double_t[1];
    TBranch *tby = _tdsQuery->getTuple()->Branch(syhat, tabValue, Form("%s/D", syhat.Data()));

    // Get The list of Leaf (from TTree) of the list of attributes ( _listOfInputAttributes)
    TList *listOfInputReferenceLeaves = new TList();
    TList *listOfInputQueryLeaves = new TList();
    TIter next(_listOfInputAttributes);
    if( _blog )
      cout << " ** input" << endl;
    while ( URANIE::DataServer::TAttribute *att = (URANIE::DataServer::TAttribute *)next()) {
      listOfInputReferenceLeaves->Add(_tds->getTuple()->GetLeaf(att->GetName()));
      listOfInputQueryLeaves->Add(_tdsQuery->getTuple()->GetLeaf(att->GetName()));
      if( _blog ) {
        _tds->getTuple()->GetLeaf(att->GetName())->Print();
        att->printLog();
      }
    }

    if( _blog ) {
      cout << endl << " **  References Leaves Attributes listing  _nX[" << _nX << "]" << endl;
      listOfInputReferenceLeaves->Print();
      cout << endl << " **  Query Leaves Attributes listing  _nX[" << _nX << "]" << endl;
      listOfInputQueryLeaves->Print();
    }

    Float_t *vecReferencePoints = new Float_t[nsr*_nX];
    Float_t *vecReferenceOutput = new Float_t[nsr];
    for(Int_t ir=0; ir<nsr; ir++) {
        _tds->getTuple()->GetEntry(ir);
        for(Int_t ii=0; ii<listOfInputReferenceLeaves->GetEntries(); ii++) {
            lr = (TLeaf *)(listOfInputReferenceLeaves->At(ii));
            vecReferencePoints[ii*nsr+ir] = (Float_t)(lr->GetValue());
        }
        vecReferenceOutput[ir] = (Float_t)(lry->GetValue());
    }
    Float_t *vecQueryPoints = new Float_t[nsq*_nX];
    for(Int_t iq=0; iq<nsq; iq++) {
        _tdsQuery->getTuple()->GetEntry(iq);
        for(Int_t ii=0; ii<listOfInputQueryLeaves->GetEntries(); ii++) {
            lr = (TLeaf *)(listOfInputQueryLeaves->At(ii));
            vecQueryPoints[ii*nsq+iq] = (Float_t)(lr->GetValue());
        }
    }

#ifdef _OPENMP
    if( _blog )
      cout << endl << " ** nbthreads read in option[" << nbthreads << "]" << endl;
    
    // Get the number of CPU to set the max of threads (nCPU-1)
    SysInfo_t* sysinfo = new SysInfo_t();
    gSystem->GetSysInfo(sysinfo);
    Int_t nbthreads_max = sysinfo->fCpus;
    delete sysinfo;

    if ( _blog )
      cout << " ** nbthreads_max[" << nbthreads_max << "]" << endl;
    if( nbthreads < 0 )
      nbthreads = nbthreads_max-1;
    else if(nbthreads > nbthreads_max)
      nbthreads = nbthreads_max-1;
    if(nbthreads < 1) nbthreads = 1;
    omp_set_num_threads(nbthreads);
    if( _blog )
      cout << " ** nbthreads used[" << nbthreads << "]" << endl;
#endif

    if ( _blog )
      cout << endl << " ** nbthreads[" << nbthreads << "]" << endl;

    switch(_nDistance) {
      case kEuclidean:
    for(Int_t iq=0; iq<nsq; iq++) {
          beginLoop(iq,nsq);
          // Compute the distance
#pragma omp parallel for
          for(Int_t ir=0; ir<nsr; ir++) {
            vecDistance[ir] = 0.0;
            for(Int_t ii=0; ii<listOfInputReferenceLeaves->GetEntries(); ii++)
              vecDistance[ir] += (vecReferencePoints[ii*nsr+ir] - vecQueryPoints[ii*nsq+iq])*(vecReferencePoints[ii*nsr+ir] - vecQueryPoints[ii*nsq+iq]);
          }
          endLoop(iq, nsr, vecDistance, vecIndex, vecReferenceOutput, tabValue, tby, nbthreads);
        }
        break;
      case kL1:
        for(Int_t iq=0; iq<nsq; iq++) {
          beginLoop(iq,nsq);
          // Compute the distance
#pragma omp parallel for
          for(Int_t ir=0; ir<nsr; ir++) {
            vecDistance[ir] = 0.0;
            for(Int_t ii=0; ii<listOfInputReferenceLeaves->GetEntries(); ii++)
              vecDistance[ir] += TMath::Abs(vecReferencePoints[ii*nsr+ir] - vecQueryPoints[ii*nsq+iq]);
          }
          endLoop(iq, nsr, vecDistance, vecIndex, vecReferenceOutput, tabValue, tby, nbthreads);
        }
        break;
      default:
        cout << " distance not yet defined" << endl;
        break;
    }

    // Clean
    delete[] vecDistance; vecDistance = NULL;
    delete[] vecIndex; vecIndex = NULL;
    _tdsQuery->getTuple()->ResetBranchAddress(tby);
    delete[] tabValue; tabValue = NULL;

    delete[] vecReferencePoints; vecReferencePoints = NULL;
    delete[] vecReferenceOutput; vecReferenceOutput = NULL;
    delete[] vecQueryPoints; vecQueryPoints = NULL;

    delete listOfInputReferenceLeaves;
    delete listOfInputQueryLeaves;

  }

void TkNN::beginLoop(Int_t iq, Int_t nsq)
{
  Int_t nStyleQuery = _tdsQuery->getTuple()->GetMarkerStyle();
  Size_t dSizeQuery = _tdsQuery->getTuple()->GetMarkerSize();

      if ( _blog )
        cout << " ** query[" << iq+1 << "/" << nsq << "]" << endl;
      _tdsQuery->getTuple()->GetEntry(iq);
      if (_blog)
        _tdsQuery->getTuple()->Show(iq);

      // Draw the current query point
      if ( _sDraw.Length() > 0 ) {
        draw(_sDraw);
        _tdsQuery->getTuple()->SetMarkerColor(kOrange);
        _tdsQuery->getTuple()->SetMarkerStyle(8);
        _tdsQuery->getTuple()->SetMarkerSize(1.5);

        _tdsQuery->draw(_sDraw, Form("%s==%d", _tdsQuery->getIteratorName(), iq+1), "same");
        gPad->Modified(); gPad->Update();

        _tdsQuery->getTuple()->SetMarkerColor(kRed);
        _tdsQuery->getTuple()->SetMarkerStyle(nStyleQuery);
        _tdsQuery->getTuple()->SetMarkerSize(dSizeQuery);
          }
      }

void TkNN::endLoop(Int_t iq, int nsr, double* vecDistance, int* vecIndex, Float_t *vecReferenceOutput, Double_t *tabValue, TBranch *tby, Int_t nbthreads)
{
      // Sort the distance
  LSort(nsr, vecDistance, vecIndex, nbthreads);

      // Make the estimation
  Double_t dyhat = 0.0;
  Double_t dsumw = 0.0;
      for(Int_t ik=0; ik<_nK; ik++) {
        Int_t nindcrt = (Int_t)(vecIndex[ik]);

        dsumw += 1.0 / TMath::Sqrt(vecDistance[nindcrt]);
        dyhat += vecReferenceOutput[nindcrt] / TMath::Sqrt(vecDistance[nindcrt]); //////// vecW[nindcrt] * lry->GetValue();
        if ( _blog )
          cout << "    **** Query [" << iq << "] reference [" << nindcrt+1 << "] y[" << vecReferenceOutput[nindcrt] << "] dist[" << vecDistance[nindcrt] << "] w[" << 1.0 / TMath::Sqrt(vecDistance[nindcrt]) << "] SumW[" << dsumw <<"] dyhat[" << dyhat <<"]" << endl;
        //	    _tds->getTuple()->Show(index[ik]);

        if ( _sDraw.Length() > 0 ) {
          _tds->getTuple()->SetMarkerColor(kGreen);
          _tds->draw(_sDraw, Form("%s==%d", _tds->getIteratorName(), vecIndex[ik]+1), "same");
          gPad->Modified(); gPad->Update();
          _tds->getTuple()->SetMarkerColor(kBlue);
        }

      }
      tabValue[0] = dyhat / dsumw;
      if ( _blog )
        cout << "    **** Query [" << iq << "] dyhat[" << tabValue[0] <<"]" << endl;
      tby->Fill();

      if ( _sDraw.Length() > 0 ) {
        // Save the graph
        gPad->Modified(); gPad->Update();
        if ( iq < 50 )
          gPad->SaveAs(Form("knn%d.png", iq+1));
      }

    }

void TkNN::LSort(int n, double* array, int* index, Int_t nbthreads)
{
  std::vector<double> sval;
  std::vector<double>::iterator it;

// put the k first values in the vector
  for(int i=0;i<_nK;i++)
    sval.push_back(array[i]);

// sort the vector in ascendant order
  std::sort(sval.begin(),sval.end());

// put in the vector the ordered k smallest values of the array
  for(int i=_nK;i<n;i++){
    it = sval.end();
    it--;
    if( array[i] < *it ){
      sval.erase(it);
      for (it=sval.begin(); it!=sval.end(); ++it)
        if( array[i] < *it ) break;
      sval.insert(it,array[i]);
    }
  }

  int j=0;
// calculate the index of the k smallest values in the input array
  if(nbthreads > 1){
    for (it=sval.begin(); it!=sval.end(); ++it)
#pragma omp parallel for
      for(int i=0;i<n;i++)
        if( fabs(array[i] - *it) < epsilon )
          index[j++] = i;
  }
  else{
  for (it=sval.begin(); it!=sval.end(); ++it)
    for(int i=0;i<n;i++)
      if( fabs(array[i] - *it) < epsilon ){
        index[j++] = i;
        break;
        }
      }

}

#if defined(WITH_GPUANN)
      //  GPU method
      void TkNN::estimate_GPU(Option_t * option)
      {

          TString soption = TString(option);
          soption.ToLower();
          soption.ReplaceAll(" ", "");

          if( _blog ) {
              cout << endl;
              cout << "**************************************" << endl;
              cout << " ** TkNN::estimate on GPU with option[" << soption << "]" << endl;
              cout << "**************************************" << endl;
          }

          URANIE::GPU::kNN knn_gpu(_nX, _nK, URANIE::GPU::kEuclidean);

          // Allocate the vecteur of distance (size of nS of reference TDS)
          Int_t nsr = _tds->getNPatterns();
          Int_t nsq = _tdsQuery->getNPatterns();

          if( _blog )
            cout << " ** nsr[" << nsr << "] nsq[" << nsq << "] nX[" << _nX << "]" << endl;

          // Get The list of Leaf (from TTree) of the list of attributes ( _listOfInputAttributes)
          TList *listOfInputReferenceLeaves = new TList();
          TList *listOfInputQueryLeaves = new TList();
          TIter next(_listOfInputAttributes);
          if( _blog )
            cout << " ** input" << endl;
          while ( URANIE::DataServer::TAttribute *att = (URANIE::DataServer::TAttribute *)next()) {
            listOfInputReferenceLeaves->Add(_tds->getTuple()->GetLeaf(att->GetName()));
            listOfInputQueryLeaves->Add(_tdsQuery->getTuple()->GetLeaf(att->GetName()));
            if( _blog ) {
              _tds->getTuple()->GetLeaf(att->GetName())->Print();
              att->printLog();
            }
          }

          // Output for the Reference NTuple
          TLeaf *lry = _tds->getTuple()->GetLeaf(_sOutputAttributes);

          if( _blog ) {
            cout << endl << " **  References Leaves Attributes listing  _nX[" << _nX << "]" << endl;
            listOfInputReferenceLeaves->Print();
            cout << endl << " **  Query Leaves Attributes listing  _nX[" << _nX << "]" << endl;
            listOfInputQueryLeaves->Print();
          }



          TLeaf *lr;
          Float_t *vecReferencePoints = new Float_t[nsr*_nX];
          Float_t *vecReferenceOutput = new Float_t[nsr];
          for(Int_t ir=0; ir<nsr; ir++) {
              _tds->getTuple()->GetEntry(ir);
              for(Int_t ii=0; ii<listOfInputReferenceLeaves->GetEntries(); ii++) {
                  lr = (TLeaf *)(listOfInputReferenceLeaves->At(ii));
                  vecReferencePoints[ii*nsr+ir] = (Float_t)(lr->GetValue());
              }
              vecReferenceOutput[ir] = (Float_t)(lry->GetValue());
          }
          knn_gpu.setReferencePoints(nsr, vecReferencePoints, vecReferenceOutput);

          Float_t *vecQueryPoints = new Float_t[nsq*_nX];
          for(Int_t iq=0; iq<nsq; iq++) {
              _tdsQuery->getTuple()->GetEntry(iq);
              for(Int_t ii=0; ii<listOfInputQueryLeaves->GetEntries(); ii++) {
                  lr = (TLeaf *)(listOfInputQueryLeaves->At(ii));
                  vecQueryPoints[ii*nsq+iq] = (Float_t)(lr->GetValue());
              }
          }
          knn_gpu.setQueryPoints(nsq, vecQueryPoints);

          if ( _blog )
              knn_gpu.printLog();
          knn_gpu.estimate();

          URANIE::DataServer::TAttribute *attY = _tds->getAttribute(_sOutputAttributes);
          if ( _blog )
             attY->printLog();
          TString syhat = Form("%shat%d", _sOutputAttributes.Data(), _nK);

          _tdsQuery->addAttribute ( new URANIE::DataServer::TAttribute(syhat, Form("#hat{%s}", attY->GetTitle())));
	  _tdsQuery->getAttribute(syhat)->delShare();
          Float_t *tabValue = new Float_t[1];
          TBranch *tby = _tdsQuery->getTuple()->Branch(syhat, tabValue, Form("%s/F", syhat.Data()));

          for(Int_t iq=0; iq<nsq; iq++) {
              if ( ( _blog ) && ( iq < 10 ) )
                cout << " ****** Query[" << iq << "] ...x[" << vecQueryPoints[iq*_nX] <<"] ";
              tabValue[0] = knn_gpu.GetQueryOutput(iq);
              if ( ( _blog ) && ( iq < 10 ) )
                  cout << "  Output[" << tabValue[0] << "]Output[" << knn_gpu.GetQueryOutput(iq) << "]" << endl;
              tby->Fill();
          }

          // Clean
          _tdsQuery->getTuple()->ResetBranchAddress(tby);
          delete[] tabValue; tabValue = NULL;

          delete[] vecReferencePoints; vecReferencePoints = NULL;
          delete[] vecReferenceOutput; vecReferenceOutput = NULL;
          delete[] vecQueryPoints; vecQueryPoints = NULL;

          delete listOfInputReferenceLeaves;
          delete listOfInputQueryLeaves;
      }
#endif


      void TkNN::estimate(Option_t * option, Bool_t useGPU)
      {
          if( _blog ) {
              cout << endl;
              cout << "**************************************" << endl;
              cout << " ** TkNN::estimate" << endl;
              cout << "**************************************" << endl;
          }

          if (useGPU)
              cout << " ** GPU computation has been requested." << endl;
          try
          {
	    if ( _blog )
              cout << " ** test if GPU is OK ..." << endl;
#ifndef WITH_GPUANN
              if (useGPU)
                  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                             Form("TkNN::estimate method\n GPU computation has been requested, but URANIE was not compiled with GPU-ANN support!"));
#endif
          } catch (URANIE::Exceptions::UErrorExceptions& ue)
          {
              ue.printMessage();
              throw ue;
          }

#ifdef WITH_GPUANN
          if (useGPU) {
              estimate_GPU(option);
          } else
#endif
              estimate_cpu(option);
      }
      
      
      // -----------------------------------------------------------------------
      //  Export Functions
      // -----------------------------------------------------------------------
      void TkNN::exportModelCplusplus(std::ofstream * sourcefile) const
      {
          *sourcefile << "  // Intercept"<< endl;
          *sourcefile << "  double y = 0.0;" << endl;
          *sourcefile << endl;
          *sourcefile << "  // Return the value"<< endl;
          *sourcefile << "  res[0] = y;" << endl;
      }
      
      void TkNN::exportModelFortran(std::ofstream * sourcefile) const
      {
          URANIE::DataServer::TAttribute *att;
          for(Int_t ivar=0; ivar<_nX; ivar++) {
              att = (URANIE::DataServer::TAttribute *)_listOfInputAttributes->At(ivar);
              *sourcefile << "      DOUBLE PRECISION " << att->GetName() << endl;
          }
          *sourcefile << "      DOUBLE PRECISION "<<_sOutputAttributes << endl;
          *sourcefile << endl;
          *sourcefile << "C --- Intercept"<< endl;
          *sourcefile << endl;
      }
      
      // -----------------------------------------------------------------------
      //  Log
      // -----------------------------------------------------------------------
      void TkNN::printLog(Option_t *option)
      {
	cout << "*******************************" << endl;
	cout << " ** TkNN::printLog[" << option << "]" << endl;
	cout << " ** References points TDS[" << _tds->GetName() << "] patterns[" << _tds->getNPatterns() << "]" << endl;
	cout << " ** _nK[" << _nK << "]" << endl;
	cout << " ** _sDraw[" << _sDraw << "]" << endl;

	switch(_nDistance) {
	case kEuclidean:{
	  cout << " ** _nDistance[kEuclidean]" << endl;
	}
	  break;
	case kManhattan:{
	  cout << " ** _nDistance[kManhattan]" << endl;
	}
	  break;
	default:{
	  cout << " ** _nDistance[not yet defined]" << endl;
	}
	}
	if ( _tdsQuery ) {
	  cout << " ** Query points TDS[" << _tdsQuery->GetName() << "] patterns[" << _tdsQuery->getNPatterns() << "]" << endl;
	} else
	  cout << " ** Query points TDS is null" << endl;
	
	
	TModeler::printLog();
	cout << "*******************************" << endl;
     }
      
    } // Fin du namespace Modeler
  } // Fin du namespace URANIE

