//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////

/** ****************************************************** **
                     GP (GPU) - Implementation
    ** ****************************************************** **/

#include <cublas.h>

#include <stdio.h>
#include <stdlib.h>

#include "GP_gpu.h"
#include "errors_gpu.h"
#include "engine_gpu.h"

#include "kernels_gpu.h"

namespace URANIE {
  namespace GPU {
  
    /* ************************************* */
    /*         GP          */
    /* ************************************* */
  
    GP::GP(int dim, Real dVariance, ECorrelationFunction t):
        _nX(dim), _dVariance(dVariance), _dTrend(0.0), _nK(-1), _nY(2),
        _nRef(0), matRef_host(0), vecRef_host(0), matRefLInverse_host(0),
      _nQuery(0), matQuery_host(0), vecQueryMean_host(0), vecQueryVariance_host(0),
      vecMinNormalisation_host(0), vecMaxNormalisation_host(0),
      vecCorrelationLengths_host(0), vecCoefficients_host(0),
      _Distance(t), _bLog(false)
    {
      initialize();
      char *line = new char[100]; //printf();

      
      Reporting::InformationLog( "Create GP on GPU.", __FILE__, __LINE__ );

      sprintf(line, "Dimension nX(%d) nY(%d)", _nX, _nY);
      Reporting::InformationLog( std::string(line) , __FILE__, __LINE__ );
      
      sprintf(line, "nMaxThreadsPerBlock of NN %d", URANIE::GPU::Engine::_nMaxThreadsPerBlock);
      Reporting::InformationLog( std::string(line) , __FILE__, __LINE__ );
    }
    
    GP::~GP()
    {

//       if( matQuery_device ) {
// 	cublasFree (matQuery_device);
//       }

      // Shutdown common (if shared equal to 1)
      GPU::Engine::Shutdown();
    }
    void GP::initialize()
    {
      // Initialize common (if not already done)
      GPU::Engine::Init();      
    }

    void GP::setReferencePoints(int nsr, Real *vecReferencePoints, Real *vecReferenceOuput, Real *matLInverse)
    {
        _nRef = nsr;
        matRef_host = vecReferencePoints;
        vecRef_host = vecReferenceOuput;
        matRefLInverse_host = matLInverse;

        if ( 0 ) {
            matRefLInverse_host     = (Real *) malloc(_nRef * _nRef * sizeof(Real));
            for (int ir=0 ; ir<_nRef; ir++) {
                for (int ii=0 ; ii<=ir; ii++) {
                    matRefLInverse_host[ir*_nRef+ii]     = 1.0f * (ir+1);
                }
            }
        }
      
      char *line = new char[100]; //printf();
      Reporting::InformationLog( "GP::setReferencePoints.", __FILE__, __LINE__ );
      sprintf(line, " - Number of references Points %d", _nRef);
      Reporting::InformationLog( std::string(line) , __FILE__, __LINE__ );

      // Just a log
      if ( _bLog ) {
          for(int index=0; index < _nRef; index++) {
              printf("   - Reference. Pat. (%04d) : ", index);
              for( int indexDim=0; indexDim<_nX; indexDim++) {
                  //                matRef_host[index*_nX+indexDim] = 1.0 * index;
                  printf(" (%02d)(%02d)[%f] ", indexDim, index*_nX+indexDim, matRef_host[index*_nX+indexDim]);
              }
              printf(" : [%f]\n", vecRef_host[index]);
          }
          printf("\nInvers Matrix L\n");
          for(int index=0; index < 10; index++) {
              printf(":(%03d):", index);
              for( int indexDim=0; indexDim<10; indexDim++) {
                  //                matRefLInverse_host[index*_nRef+indexDim] = 1.0 * index;
                  printf(" (%03d)(%04d)[%02.4f]", indexDim, index*_nRef+indexDim, matRefLInverse_host[index*_nRef+indexDim]);
              }
              printf(" :\n");
          }
          printf("\n");
      }



//       cublasStatus stat;
      
//       // Allocation on the Device and Transfer Data
//       printf("\n - Matrix matRef_device    : Device  ... ");
//       stat = cublasAlloc (_nRef*_nX, sizeof(*matRef_host), (void**)&matRef_device);
//       if (stat != CUBLAS_STATUS_SUCCESS) {
//         printf ("device memory matRef_device allocation failed");
//         cublasShutdown();
//         return ;
//       }
//       printf(" Ok - Set ...");
//       stat = cublasSetMatrix (_nX, _nRef, sizeof(*matRef_host), matRef_host, _nX, matRef_device, _nX);
//       if (stat != CUBLAS_STATUS_SUCCESS) {
//         printf ("data matRef_device download failed");
//         cublasFree (matRef_device);
//         cublasShutdown();
//         return ;
//       }
//       printf(" Done1\n");
      
//       URANIE::GPU::Engine::UpdateFreeMemory();

    }


    void GP::setCorrelationLengths(int nx, Real *vecCorrelationLengths)
    {
//        if ( nx != _nX )
//            // CuDA ERROR
        vecCorrelationLengths_host = vecCorrelationLengths;

        char *line = new char[100]; //printf();
        Reporting::InformationLog( "GP::setCorrelationLengths.", __FILE__, __LINE__ );
        sprintf(line, " - Number of Correlation Lengths %d", nx);
        Reporting::InformationLog( std::string(line) , __FILE__, __LINE__ );
        // Just a log
        for( int indexDim=0; indexDim<nx; indexDim++) {
//            vecCorrelationLengths_host[indexDim] = 1.0;
            printf(" (%02d)[%f] ", indexDim, vecCorrelationLengths_host[indexDim]);
        }
        printf("\n");
    }

    void GP::setCoefficients(int nsr, Real *vecCoefficients)
    {
//        if ( nsr != _nS )
//            // CuDA ERROR
        vecCoefficients_host = vecCoefficients;

        char *line = new char[100]; //printf();
        Reporting::InformationLog( "GP::setCoefficients.", __FILE__, __LINE__ );
        sprintf(line, " - Number of Coefficients %d", nsr);
        Reporting::InformationLog( std::string(line) , __FILE__, __LINE__ );
        // Just a log
        for( int indexDim=0; indexDim<10; indexDim++)
            printf(" (%02d)[%f] ", indexDim, vecCoefficients_host[indexDim]);
        printf("\n");
    }

    void GP::setQueryPoints(int nsq, Real *vecQueryPoints)
    {
      _nQuery = nsq;
      matQuery_host = vecQueryPoints;

      vecQueryMean_host     = (Real *) malloc(_nQuery * sizeof(Real));
      vecQueryVariance_host = (Real *) malloc(_nQuery * sizeof(Real));
      for (int i=0 ; i<_nQuery; i++) {
          vecQueryMean_host[i]     = 0.0f;
          vecQueryVariance_host[i] = 0.0f;
      }

      char *line = new char[100]; //printf();
      Reporting::InformationLog( "GP::setQueryPoints.", __FILE__, __LINE__ );
      sprintf(line, " - Number of Query Points %d", _nQuery);
      Reporting::InformationLog( std::string(line) , __FILE__, __LINE__ );

      // Just a log
      if ( _bLog ) {
          for(int index=0; index < _nQuery; index++) {
              printf("   - Query. Pat. (%04d) : ", index);
              for( int indexDim=0; indexDim<_nX; indexDim++) {
                  //                matQuery_host[index*_nX+indexDim] = 1.0 * index;
                  printf(" (%02d)(%02d)[%f] ", indexDim, index*_nX+indexDim, matQuery_host[index*_nX+indexDim]);
              }
              printf("\n");
          }
          printf("\n");
      }


    }

    void GP::setNormalisationValues(int nx, Real*vecMin, Real* vecMax)
    {
        //        if ( nx != _nX )
        //            // CuDA ERROR
        vecMinNormalisation_host = vecMin;
        vecMaxNormalisation_host = vecMax;

        char *line = new char[100]; //printf();
        Reporting::InformationLog( "GP::setNormalisationValues.", __FILE__, __LINE__ );
        sprintf(line, " - Number of Normalisation : %d", nx);
        Reporting::InformationLog( std::string(line) , __FILE__, __LINE__ );
        // Just a log
        for( int indexDim=0; indexDim<nx; indexDim++) {
//            vecMinNormalisation_host[indexDim] = 0.0;
//            vecMaxNormalisation_host[indexDim] = 2.0;
            printf(" (%02d) Min[%f] Max[%f]\n",
                   indexDim,
                   vecMinNormalisation_host[indexDim],
                   vecMaxNormalisation_host[indexDim]);
        }
        printf("\n");
    }

    void GP::estimate()
    {
      char *line = new char[100];
      
      // Determine maximum number of query that can be treated
//       max_nb_query_traited = ( memory_free * MAX_PART_OF_FREE_MEMORY_USED - size_of_float * ref_width*height ) / ( size_of_float * (height + ref_width) + size_of_int * k);
//       max_nb_query_traited = min( query_width, (max_nb_query_traited / 16) * 16 );
//       printf(" - max_nb_query_traited(%d)\n", max_nb_query_traited);
      sprintf(line, " - GP::run :: Memory Total (%.0f MBytes)", URANIE::GPU::Engine::_nMemoryTotal);
      Reporting::InformationLog( std::string(line) , __FILE__, __LINE__ );
      
      sprintf(line, " - GP::run :: Memory free (%.0f MBytes)", URANIE::GPU::Engine::_nMemoryFree);
      Reporting::InformationLog( std::string(line) , __FILE__, __LINE__ );


      // Variables for duration evaluation
      cudaEvent_t start, stop;
      cudaEventCreate(&start);
      cudaEventCreate(&stop);
      float elapsed_time;
      
      int iterations = 1;
     // Display informations
      printf("Number of reference points      : %6d\n", _nRef   );
      printf("Number of query points          : %6d\n", _nQuery );
      printf("Dimension of points             : %4d\n", _nX     );
      printf("Number of iterations            : %4d\n", iterations     );
      printf("Processing GP search           :"                );
      float* yMean     = (float *) malloc(_nQuery * sizeof(float));
      float* yVariance = (float *) malloc(_nQuery * sizeof(float));

//      // Init
//      for (int i=0 ; i<_nQuery; i++) {
//          yMean[i]     =  0.0;
//          yVariance[i] = -1.0;
//      }

      
      // Call GP search CUDA
      cudaEventRecord(start, 0);
      for (int i=0; i<iterations; i++) {
        printf("\n - Iteration(%d/%d)\n", i, iterations);
        kernel_GP(matRef_host, matRefLInverse_host, _nRef,
                  matQuery_host, _nQuery, _nX,
                  vecMinNormalisation_host, vecMaxNormalisation_host,
                  vecCorrelationLengths_host,
                  vecCoefficients_host,
                  _dVariance, _bLog,
                  yMean, yVariance);

        for(int inew=0; inew<_nQuery; inew++) {
            vecQueryMean_host[inew]     = yMean[inew] + _dTrend;
            vecQueryVariance_host[inew] = yVariance[inew];
            if ( _bLog )
                printf(" - Query (%05d / %05d) : yMean [%e] Var[]\n", inew, _nQuery, vecQueryMean_host[inew]);
        }
        printf("\n");

      }
      
      cudaEventRecord(stop, 0);
      cudaEventSynchronize(stop);
      cudaEventElapsedTime(&elapsed_time, start, stop);
      printf(" - GPU Estimation done in %f s for %d iterations (%f s by iteration)\n", elapsed_time/1000, iterations, elapsed_time/(iterations*1000));
      
      // Destroy cuda event object and free memory
      cudaEventDestroy(start);
      cudaEventDestroy(stop);

     
    }
    void GP::printLog()
    {
      char *line = new char[100]; //printf();

      Reporting::InformationLog( "No training done yet.", __FILE__, __LINE__ );

      sprintf(line, "Dimension nX(%d) nY(%d) Variance(%f) Trend(%f)", _nX, _nY, _dVariance, _dTrend);
      Reporting::InformationLog( std::string(line) , __FILE__, __LINE__ );

      for(int index=0; index < _nRef; index++) { // _nRef
          printf(" - Reference :: index(%04d) : ", index);
          for(int ii=0; ii < _nX; ii++)
              printf("(%f) :", matRef_host[index*_nX+ii]);
          printf(" :: Out (%f) :: ", vecRef_host[index]);
          printf(" Coeff (%f)\n", vecCoefficients_host[index]);
      }

      sprintf(line, "Min, Max and Correlation Lengths (%d)", _nX);
      Reporting::InformationLog( std::string(line) , __FILE__, __LINE__ );
      for( int indexDim=0; indexDim<_nX; indexDim++) {
          printf("   (%02d) Min[%f] Max[%f] Corr.[%f]\n",
                 indexDim,
                 vecMinNormalisation_host[indexDim],
                 vecMaxNormalisation_host[indexDim],
                 vecCorrelationLengths_host[indexDim]);
      }
      printf("\n");

      sprintf(line, "Number of Query points %d", _nQuery);
      Reporting::InformationLog( std::string(line) , __FILE__, __LINE__ );
      for(int index=0; index < _nQuery*_nX; index++) {
          printf(" - Query Input :: index(%04d) Value(%f)\n", index, matQuery_host[index]);
      }


    }
  } // namespace GPU

} // namespace URANIE

// End of File
