//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////

/** ****************************************************** **
                    GP (GPU)
 ** ****************************************************** **/


#ifndef __GP_GPU_INCLUDE__
#define __GP_GPU_INCLUDE__

typedef float Real;                             //!< GPU => Real (this is an implication, not a pretty arrow)


namespace URANIE
{
  namespace GPU
  {
    enum ECorrelationFunction { kUnknown, kEuclidean, kMatern32, kMatern52, kMatern72 };  ///< Correlation function
    
    class GP
    {
    public:            
      explicit GP(int nx, Real dVariance, ECorrelationFunction t);

      ~GP();
      
      /**! Size of the layer, in neurons **/
      inline int getDimension() const { return _nX; }
      /**! Size of the layer, in neurons **/
      inline int getNGP() const { return _nK; }

      inline int getNReferencePoints() const { return _nRef; }
      inline int getNQueryPoints() const { return _nQuery; }
      inline void setLog() { _bLog = true;}
      inline void unsetLog() { _bLog = false;}

      void initialize();

      void setReferencePoints(int nsr, Real *vecReferencePoints, Real *vecReferenceOutput, Real *matLInverse);
      void setCorrelationLengths(int nx, Real *vecCorrelationLengths);
      void setCoefficients(int nsr, Real *vecCoefficients);
      void setQueryPoints(int nsq, Real *vecQueryPoints);
      inline Real getQueryMean(int iq)     { return vecQueryMean_host[iq];}
      inline Real getQueryVariance(int iq) { return vecQueryVariance_host[iq];}
      void setNormalisationValues(int nx, Real*vecMin, Real* vecMax);

      void setTrends(Real dTrend) { _dTrend = dTrend;}

      void estimate();
      
      void printLog();

    public:
      Real* vecQueryMean_host;       //< Matrix of Mean Query OUTPUT values
      Real* vecQueryVariance_host;       //< Matrix of Variance Query OUTPUT values
      Real _dTrend;

    private:
      const int _nX;               //!< Dimension of Input
      const int _nY;               //!< Dimension of Ouput (2 for Mean and Sigma)
      const int _nK;               //!< the number of neighboor to take into account
      int _nRef;                   //!< Number of Reference points
      Real* matRef_host;           //< Matrix of Reference values on host
      Real* vecRef_host;           //< Matrix of Reference OUTPUT values on host
      Real *matRefLInverse_host;  //< Matrix inverse of L to compute variance
      const Real _dVariance;

      // Model Parameters
      Real* vecCorrelationLengths_host;    //< Correlation lengths in the normalised input space (size: _nX)
      Real* vecMinNormalisation_host;      //< Min values for normalisation
      Real* vecMaxNormalisation_host;      //< Max values for normalisation
      Real* vecCoefficients_host;          //< Parameters of the gaussian process (size: _nS)

//       Real* matRef_device;     //< Matrix of Reference values on DEVICE

      int _nQuery;                 //!< Number of Query points
      Real* matQuery_host;         //< Matrix of Query values
      const ECorrelationFunction _Distance;   // Activation function
      bool _bLog;
    };

    
  } // namespace GPU
  
} // namespace URANIE

#endif // #ifndef __GP_GPU_INCLUDE__
// End of File

