//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////

#include "engine_gpu.h"
#include "errors_gpu.h"
#include <cublas.h>

#include <stdio.h>
#include <stdlib.h>



bool URANIE::GPU::Engine::systemInitialized = false;
int URANIE::GPU::Engine::_nShared = 0;
int URANIE::GPU::Engine::_nMaxThreadsPerBlock = 0;
float URANIE::GPU::Engine::_nMemoryTotal = 0.0f;
float URANIE::GPU::Engine::_nMemoryFree = 0.0f;

void URANIE::GPU::Engine::Init()
{
  _nShared++;
  
  char *line = new char[100];
 
  sprintf(line, "URANIE::GPU::Engine Init Number of nShared %d",_nShared);
  Reporting::InformationLog( std::string(line) , __FILE__, __LINE__ );
 
  if( systemInitialized ) return;

    // Initialize cuBLAS (TODO: check if it's acceptable to do it here, in case of multiple instances of the class)
    cublasStatus status = cublasInit();
    GPU::checkCuBLASError( "cublasInit [<problem>]", status );

    systemInitialized = true;


    int dev = 0;
    cudaSetDevice(dev);
    cudaDeviceProp deviceProp;
    cudaGetDeviceProperties(&deviceProp, dev);

    sprintf(line, "Device %d: \"%s\"", dev, deviceProp.name);
    Reporting::InformationLog( std::string(line) , __FILE__, __LINE__ );
    
    Reporting::InformationLog( " - Compute the constantes Maxthears by Blocks" , __FILE__, __LINE__ );
    
    _nMaxThreadsPerBlock = deviceProp.maxThreadsPerBlock;
    sprintf(line, " - nMaxThreadsPerBlock[%d]", _nMaxThreadsPerBlock);
    Reporting::InformationLog( std::string(line) , __FILE__, __LINE__ );

    Reporting::InformationLog( " - Get Information on Memory" , __FILE__, __LINE__ );

    size_t memory_total;
    size_t memory_free;
    cuMemGetInfo(&memory_free, &memory_total);
    sprintf(line, " - memory_total : %zd     memory_free  : %zd", memory_total, memory_free );
    Reporting::InformationLog( std::string(line) , __FILE__, __LINE__ );
   
    _nMemoryTotal = (float)memory_total/1048576.0f; // (float)deviceProp.totalGlobalMem/1048576.0f;
    sprintf(line, " - _nMemoryTotal : %.0f MBytes", _nMemoryTotal);
    Reporting::InformationLog( std::string(line) , __FILE__, __LINE__ );

    
    _nMemoryFree = (float)memory_free/1048576.0f;
    sprintf(line, " - _nMemoryFree  : %.0f MBytes", _nMemoryFree);
    Reporting::InformationLog( std::string(line) , __FILE__, __LINE__ );

    const int nMaxTheadsBlockDevice = 512;

    // Clean
    delete[] line;
}

void URANIE::GPU::Engine::UpdateFreeMemory()
{
    if( !systemInitialized ) return;
    size_t memory_total;
    size_t memory_free;
    cuMemGetInfo(&memory_free, &memory_total);

    char *line = new char[100];
    _nMemoryFree = (float)memory_free/1048576.0f;
    sprintf(line, " - _nMemoryFree  : %.0f MBytes", _nMemoryFree);
    Reporting::InformationLog( std::string(line) , __FILE__, __LINE__ );

    // Clean
    delete[] line;
}

void URANIE::GPU::Engine::Shutdown()
{
    if( !systemInitialized ) return;
    _nShared--;
  char *line = new char[100];
 
  sprintf(line, "URANIE::GPU::Engine Shutdown Number of nShared %d",_nShared);
  Reporting::InformationLog( std::string(line) , __FILE__, __LINE__ );
 

    if ( _nShared == 0 ) {
      cublasStatus status = cublasShutdown();
      GPU::checkCuBLASError( "cublasShutdown [<problem>]", status );
      
      systemInitialized = false;
    }

    // Clean
    delete[] line;
}
