//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////

/** ****************************************************** **
               Kernels (GPU) - Implementation
 ** ****************************************************** **/


#include <cuda.h>
#include <cublas.h>
#include <map>
#include "errors_gpu.h"
#include "kernels_gpu.h"
#include "engine_gpu.h"

#include <stdio.h>


    namespace URANIE
    {
      namespace GPU
      {


#ifndef NDEBUG
void checkCuBLASError( const std::string message, int status )
{
  // If status == -1, were not using a cuBLAS helper function (returns a status) but a core one (doesn't)              
  cublasStatus cStatus = (status == -1) ? cublasGetError() : static_cast<cublasStatus>(status);
                  
  static std::map<cublasStatus, std::string> cublasErrors;
  static bool initialized = false;
  if( !initialized )
    {
      cublasErrors[ CUBLAS_STATUS_SUCCESS ] = "operation completed successfully";
      cublasErrors[ CUBLAS_STATUS_NOT_INITIALIZED ] = "CUBLAS library not initialized";
      cublasErrors[ CUBLAS_STATUS_ALLOC_FAILED ] = "resource allocation failed";
      cublasErrors[ CUBLAS_STATUS_INVALID_VALUE ] = "unsupported numerical value was passed to function";
      cublasErrors[ CUBLAS_STATUS_MAPPING_ERROR ] = "access to GPU memory space failed";
      cublasErrors[ CUBLAS_STATUS_EXECUTION_FAILED ] = "GPU program failed to execute";
      cublasErrors[ CUBLAS_STATUS_INTERNAL_ERROR ] = "an internal CUBLAS operation failed";
    }

  if( cStatus == CUBLAS_STATUS_NOT_INITIALIZED )
    {
      Reporting::Warn( message + ":\t" + cublasErrors[ cStatus ], __FILE__, __LINE__ );
      status = cublasInit();	// Retry
    }
  if( cStatus != CUBLAS_STATUS_SUCCESS )
    {
      throw Reporting::cublasException( message + ":\t" + cublasErrors[ cStatus ], __FILE__, __LINE__ );
    }
}
#endif  //  NDEBUG

#ifndef NDEBUG
      inline void cudaCheckError( const std::string kernelName )
      {
          cudaError_t error = cudaGetLastError();
          if( error != cudaSuccess )
              Reporting::Err( "Error in CUDA kernel \""+ kernelName +"\": "+cudaGetErrorString(error), __FILE__, __LINE__ );
      }
#endif



	namespace Kernels
        {

          /**!\brief Kernel: Fills a NxN identity matrix, represented by a N-dimensional unit vector (the diagonal only) **/
          

	__global__ void normalizeSamples(const Real *vecMin, const Real *vecMax, Real *dValue, const int dimN, const int dimM)
	{	
        const int nCol = threadIdx.x + blockIdx.x * blockDim.x;
        const int nRow = threadIdx.y + blockIdx.y * blockDim.y;
    	if( nCol < dimM ) {
          if ( nRow < dimN ) {
          	dValue[nRow*dimM+nCol] = ( dValue[nRow*dimM+nCol] - vecMin[nCol] ) / ( vecMax[nCol] - vecMin[nCol]);
          }	
    	}	
	}

    __global__ void normalizeXNewAndLambdaIdx(const Real *matXNew, const int idxNew, const Real *vecLambda, Real *dValue, Real* dTmpValue, const int dimN, const int dimM)
    {
        const int nCol = threadIdx.x;
        const int nRow = threadIdx.y + blockIdx.y * blockDim.y;
        if( nCol < dimM ) {
            if ( nRow < dimN ) {
                dTmpValue[nRow*dimM+nCol] = ( matXNew[idxNew*dimM+nCol] -  dValue[nRow*dimM+nCol] ) / ( vecLambda[nCol] ) ;
            }
        }
    }

    __global__ void computeNorm2Matern32(const Real *dValue, Real *dvecC, const int dimN, const int dimM)
    {
        const int i = threadIdx.x + blockIdx.x * blockDim.x;

        if( i < dimN ) {
            Real dSum = 0.0f;
            for(int nCol=0; nCol < dimM; nCol++ ) {
                dSum += dValue[i*dimM+nCol] * dValue[i*dimM+nCol];
            }
            Real dAS = sqrt( 6.0f * dSum);
            dvecC[i] = exp( -1.0f * dAS) * ( 1.0f + dAS);
        }
    }



	} // namespace Kernels
      
        namespace Wrappers
        {
          
          void gpuNormalizeSamples( dim3 numBlocks, dim3 threadsPerBlock, const Real *vecMin, const Real *vecMax, Real *dValue, const int dimN, const int dimM)
          {
            Kernels::normalizeSamples<<<numBlocks, threadsPerBlock>>>(vecMin, vecMax, dValue, dimN, dimM);
            cudaCheckError( __FUNCTION__ );
          }

          void gpuNormalizeXNewAndLambdaIdx( dim3 numBlocks, dim3 threadsPerBlock, const Real *matXNew, const int idxNew, const Real *vecLambda, Real *dValue, Real* dTmpValue, const int dimN, const int dimM)
          {
            Kernels::normalizeXNewAndLambdaIdx<<<numBlocks, threadsPerBlock>>>(matXNew, idxNew, vecLambda, dValue, dTmpValue, dimN, dimM);
            cudaCheckError( __FUNCTION__ );
          }
    } // namespace Wrappers
      
      } // namespace GPU

    } // namespace URANIE







// Constants used by the program
#define MAX_PITCH_VALUE_IN_BYTES       262144
#define MAX_TEXTURE_WIDTH_IN_BYTES     65536
#define MAX_TEXTURE_HEIGHT_IN_BYTES    32768
#define MAX_PART_OF_FREE_MEMORY_USED   0.9
#define BLOCK_DIM                      16


// Texture containing the reference points (if it is possible)
//texture<float, 2, cudaReadModeElementType> texA;



//-----------------------------------------------------------------------------------------------//
//                                            KERNELS                                            //
//-----------------------------------------------------------------------------------------------//



/**
  * Computes the distance between two matrix A (reference points) and
  * B (query points) containing respectively wA and wB points.
  * The matrix A is a texture.
  *
  * @param wA    width of the matrix A = number of points in A
  * @param B     pointer on the matrix B
  * @param wB    width of the matrix B = number of points in B
  * @param pB    pitch of matrix B given in number of columns
  * @param dim   dimension of points = height of matrices A and B
  * @param AB    pointer on the matrix containing the wA*wB distances computed
  */
__global__ void cuComputeDistanceTexture(cudaTextureObject_t texA, int wA, float * B, int wB, int pB, int dim, float* AB){
    unsigned int xIndex = blockIdx.x * blockDim.x + threadIdx.x;
    unsigned int yIndex = blockIdx.y * blockDim.y + threadIdx.y;
    if ( xIndex<wB && yIndex<wA ){
        float ssd = 0;
        for (int i=0; i<dim; i++){
            float tmp  = tex2D<float>(texA, (float)yIndex, (float)i) - B[ i * pB + xIndex ];
            ssd += tmp * tmp;
        }
        AB[yIndex * pB + xIndex] = ssd;
    }
}


/**
  * Computes the distance between two matrix A (reference points) and
  * B (query points) containing respectively wA and wB points.
  *
  * @param A     pointer on the matrix A
  * @param wA    width of the matrix A = number of points in A
  * @param pA    pitch of matrix A given in number of columns
  * @param B     pointer on the matrix B
  * @param wB    width of the matrix B = number of points in B
  * @param pB    pitch of matrix B given in number of columns
  * @param dim   dimension of points = height of matrices A and B
  * @param AB    pointer on the matrix containing the wA*wB distances computed
  */
__global__ void cuComputeDistanceGlobal( float* A, int wA, int pA, float* B, int wB, int pB, int dim,  float* AB){

	// Declaration of the shared memory arrays As and Bs used to store the sub-matrix of A and B
	__shared__ float shared_A[BLOCK_DIM][BLOCK_DIM];
	__shared__ float shared_B[BLOCK_DIM][BLOCK_DIM];
    
    // Sub-matrix of A (begin, step, end) and Sub-matrix of B (begin, step)
    __shared__ int begin_A;
    __shared__ int begin_B;
    __shared__ int step_A;
    __shared__ int step_B;
    __shared__ int end_A;
	
    // Thread index
    int tx = threadIdx.x;
    int ty = threadIdx.y;
	
	// Other variables
	float tmp;
    float ssd = 0;
	
    // Loop parameters
    begin_A = BLOCK_DIM * blockIdx.y;
    begin_B = BLOCK_DIM * blockIdx.x;
    step_A  = BLOCK_DIM * pA;
    step_B  = BLOCK_DIM * pB;
    end_A   = begin_A + (dim-1) * pA;
    
    // Conditions
	int cond0 = (begin_A + tx < wA); // used to write in shared memory
    int cond1 = (begin_B + tx < wB); // used to write in shared memory & to computations and to write in output matrix
    int cond2 = (begin_A + ty < wA); // used to computations and to write in output matrix
    
    // Loop over all the sub-matrices of A and B required to compute the block sub-matrix
    for (int a = begin_A, b = begin_B; a <= end_A; a += step_A, b += step_B) {
        
        // Load the matrices from device memory to shared memory; each thread loads one element of each matrix
        if (a/pA + ty < dim){
            shared_A[ty][tx] = (cond0)? A[a + pA * ty + tx] : 0;
            shared_B[ty][tx] = (cond1)? B[b + pB * ty + tx] : 0;
        }
        else{
            shared_A[ty][tx] = 0;
            shared_B[ty][tx] = 0;
        }
        
        // Synchronize to make sure the matrices are loaded
        __syncthreads();
        
        // Compute the difference between the two matrixes; each thread computes one element of the block sub-matrix
        if (cond2 && cond1){
            for (int k = 0; k < BLOCK_DIM; ++k){
				tmp = shared_A[k][ty] - shared_B[k][tx];
                ssd += tmp*tmp;
			}
        }
        
        // Synchronize to make sure that the preceding computation is done before loading two new sub-matrices of A and B in the next iteration
        __syncthreads();
    }
    
    // Write the block sub-matrix to device memory; each thread writes one element
    if (cond2 && cond1)
        AB[ (begin_A + ty) * pB + begin_B + tx ] = ssd;
}



/**
  * Gathers k-th smallest distances for each column of the distance matrix in the top.
  *
  * @param dist        distance matrix
  * @param dist_pitch  pitch of the distance matrix given in number of columns
  * @param ind         index matrix
  * @param ind_pitch   pitch of the index matrix given in number of columns
  * @param width       width of the distance matrix and of the index matrix
  * @param height      height of the distance matrix and of the index matrix
  * @param k           number of neighbors to consider
  */
__global__ void cuInsertionSort(float *dist, int dist_pitch, int *ind, int ind_pitch, int width, int height, int k){

	// Variables
    int l, i, j;
    float *p_dist;
	int   *p_ind;
    float curr_dist, max_dist;
    int   curr_row,  max_row;
    unsigned int xIndex = blockIdx.x * blockDim.x + threadIdx.x;
	
    if (xIndex<width){
        
        // Pointer shift, initialization, and max value
        p_dist   = dist + xIndex;
		p_ind    = ind  + xIndex;
        max_dist = p_dist[0];
        p_ind[0] = 1;
        
        // Part 1 : sort kth firt elementZ
        for (l=1; l<k; l++){
            curr_row  = l * dist_pitch;
			curr_dist = p_dist[curr_row];
			if (curr_dist<max_dist){
                i=l-1;
				for (int a=0; a<l-1; a++){
					if (p_dist[a*dist_pitch]>curr_dist){
						i=a;
						break;
					}
				}
                for (j=l; j>i; j--){
					p_dist[j*dist_pitch] = p_dist[(j-1)*dist_pitch];
					p_ind[j*ind_pitch]   = p_ind[(j-1)*ind_pitch];
                }
				p_dist[i*dist_pitch] = curr_dist;
				p_ind[i*ind_pitch]   = l+1;
			}
			else
				p_ind[l*ind_pitch] = l+1;
			max_dist = p_dist[curr_row];
		}
        
        // Part 2 : insert element in the k-th first lines
        max_row = (k-1)*dist_pitch;
        for (l=k; l<height; l++){
			curr_dist = p_dist[l*dist_pitch];
			if (curr_dist<max_dist){
                i=k-1;
				for (int a=0; a<k-1; a++){
					if (p_dist[a*dist_pitch]>curr_dist){
						i=a;
						break;
					}
				}
                for (j=k-1; j>i; j--){
					p_dist[j*dist_pitch] = p_dist[(j-1)*dist_pitch];
					p_ind[j*ind_pitch]   = p_ind[(j-1)*ind_pitch];
                }
				p_dist[i*dist_pitch] = curr_dist;
				p_ind[i*ind_pitch]   = l+1;
                max_dist             = p_dist[max_row];
            }
        }
    }
}



/**
  * Computes the square root of the first line (width-th first element)
  * of the distance matrix.
  *
  * @param dist    distance matrix
  * @param width   width of the distance matrix
  * @param pitch   pitch of the distance matrix given in number of columns
  * @param k       number of neighbors to consider
  */
__global__ void cuParallelSqrt(float *dist, int width, int pitch, int k){
    unsigned int xIndex = blockIdx.x * blockDim.x + threadIdx.x;
    unsigned int yIndex = blockIdx.y * blockDim.y + threadIdx.y;
    if (xIndex<width && yIndex<k)
        dist[yIndex*pitch + xIndex] = sqrt(dist[yIndex*pitch + xIndex]);
}







//-----------------------------------------------------------------------------------------------//
//                                   K-th NEAREST NEIGHBORS                                      //
//-----------------------------------------------------------------------------------------------//



/**
  * Prints the error message return during the memory allocation.
  *
  * @param error        error value return by the memory allocation function
  * @param memorySize   size of memory tried to be allocated
  */
void printErrorMessage(cudaError_t error, int memorySize){
    printf("==================================================\n");
    printf("MEMORY ALLOCATION ERROR  : %s\n", cudaGetErrorString(error));
    printf("Whished allocated memory : %d\n", memorySize);
    printf("==================================================\n");
}
void kernel_GP(Real* ref_host, Real* matLInv_host, int ref_width,
               Real* query_host, int query_width,
               int height,
               Real *min_host, Real *max_host,
               Real *lambda_host, Real *Coefficient_host,
               Real dVariance, bool blog,
               Real* yMean_host, Real* yVariance_host)
{
    printf(" - kernel_GP without texture\n");

    // Variables
    Real        *ref_dev;
    Real        *matLInv_device;
    Real        *query_dev;

    Real *matXtmp_device;
    Real *vecLambda_device;
    Real *vecCoeff_device;
    Real *vecDotVariance_device;
    Real *vecMin_device;
    Real *vecMax_device;
    Real *vecNormLambda_device;

    size_t memory_total;
    size_t memory_free;
    cublasStatus stat;

    // CUDA Initialisation
    printf(" - Initialisation (cublas & cuda) : \n");
    cublasInit();
    cuInit(0);


    int dev = 0;
    cudaSetDevice(dev);
    cudaDeviceProp deviceProp;
    cudaGetDeviceProperties(&deviceProp, dev);

    printf("Device %d: \"%s\"", dev, deviceProp.name);

    // Check free memory using driver API ; only (MAX_PART_OF_FREE_MEMORY_USED*100)% of memory will be used
    CUcontext cuContext;
    CUdevice  cuDevice=0;
    cuCtxCreate(&cuContext, 0, cuDevice);
    cuMemGetInfo(&memory_free, &memory_total);
    cuCtxDetach (cuContext);

    // Log on memory
    printf(" - memory_total (%zd)\n", memory_total);
    printf(" - memory_free  (%zd)\n", memory_free);


    int i, j;
    int nMaxRefQuery = ref_width;
    if ( query_width > ref_width)
        nMaxRefQuery = query_width;
    Real* matXNew_host = (Real *)malloc (height * nMaxRefQuery * sizeof (*query_host));
    if (!matXNew_host) {
        printf ("host memory matXNew_host allocation failed a");
        return;
    }
    Real* vecC_host = (Real *)malloc (ref_width * sizeof (*query_host));
    if (!vecC_host) {
        printf ("host memory vecC_host allocation failed a");
        return;
    }
    Real* vecCC_host = (Real *)malloc (ref_width * sizeof (*query_host));
    if (!vecCC_host) {
        printf ("host memory vecCC_host allocation failed a");
        return;
    }

    printf(" - Matrix ref_dev    : Device %d: \"%s\" ... ", cuDevice, deviceProp.name);
    stat = cublasAlloc (ref_width*height, sizeof(*ref_host), (void**)&ref_dev);
    if (stat != CUBLAS_STATUS_SUCCESS) {
        printf ("device memory ref_dev allocation failed");
        cublasShutdown();
        return;
    }
    printf(" Ok - Set ...");
    stat = cublasSetMatrix (ref_width, height, sizeof(*ref_host), ref_host, ref_width, ref_dev, ref_width);
    if (stat != CUBLAS_STATUS_SUCCESS) {
        printf ("data ref_dev setting failed");
        cublasFree (ref_dev);
        cublasShutdown();
        return;
    }
    printf(" Done\n   ");
    URANIE::GPU::Engine::UpdateFreeMemory();


    if ( blog ) {
        stat = cublasGetMatrix (ref_width, height, sizeof(*ref_host), ref_dev, ref_width, matXNew_host, ref_width);
        if (stat != CUBLAS_STATUS_SUCCESS) {
            printf ("data query_device upload failed");
            cublasFree (matXNew_host);
            cublasShutdown();
            return;
        }

        printf("\n Original Reference on device\n");
        for (j = 0; j < 10; j++) {
            printf(" *** Reference (%05d) : ", j);
            for (i = 0; i < height; i++) {
                printf ("(%7.5f)[%7.5f] - ", ref_host[j*height+i], matXNew_host[j*height+i]);
            }
            printf("\n");
        }
    }



    printf(" - Matrix LInverse    : Device %d: \"%s\" ... ", cuDevice, deviceProp.name);
    stat = cublasAlloc (ref_width*ref_width, sizeof(*ref_host), (void**)&matLInv_device);
    if (stat != CUBLAS_STATUS_SUCCESS) {
        printf ("device memory matLInv_device allocation failed");
        cublasShutdown();
        return;
    }
    printf(" Ok - Set ...");
    stat = cublasSetMatrix (ref_width, ref_width, sizeof(*ref_host), matLInv_host, ref_width, matLInv_device, ref_width);
    if (stat != CUBLAS_STATUS_SUCCESS) {
        printf ("data matLInv_device setting failed");
        cublasFree (matLInv_device);
        cublasShutdown();
        return;
    }
    printf(" Done\n   ");
    URANIE::GPU::Engine::UpdateFreeMemory();



    printf(" - Matrix matXtmp_device   : Device %d: \"%s\" ... ", dev, deviceProp.name);
    stat = cublasAlloc (ref_width*height, sizeof(*ref_host), (void**)&matXtmp_device);
    if (stat != CUBLAS_STATUS_SUCCESS) {
        printf ("device memory matXtmp_device allocation failed");
        cublasShutdown();
        return;
    }
    printf(" Ok - Set ...");
    stat = cublasSetMatrix (height, ref_width, sizeof(*ref_host), ref_host, height, matXtmp_device, height);
    if (stat != CUBLAS_STATUS_SUCCESS) {
        printf ("data matXtmp_device setting failed");
        cublasFree (ref_dev);
        cublasFree (matXtmp_device);
        cublasShutdown();
        return;
    }
    printf(" Done\n   ");
    URANIE::GPU::Engine::UpdateFreeMemory();

    printf(" - Matrix query_dev   : Device %d: \"%s\" ... ", dev, deviceProp.name);
    stat = cublasAlloc (query_width*height, sizeof(*query_host), (void**)&query_dev);
    if (stat != CUBLAS_STATUS_SUCCESS) {
        printf ("device memory query_dev allocation failed");
        cublasShutdown();
        return;
    }
    printf(" Ok - Set ...");
    stat = cublasSetMatrix (query_width, height, sizeof(*query_host), query_host, query_width, query_dev, query_width);
    if (stat != CUBLAS_STATUS_SUCCESS) {
        printf ("data query_dev setting failed");
        cublasFree (ref_dev);
        cublasFree (matXtmp_device);
        cublasFree (query_dev);
        cublasShutdown();
        return;
    }
    printf(" Done\n   ");
    URANIE::GPU::Engine::UpdateFreeMemory();

    if ( blog ) {
        stat = cublasGetMatrix (query_width, height, sizeof(*query_host), query_dev, query_width, matXNew_host, query_width);
        if (stat != CUBLAS_STATUS_SUCCESS) {
            printf ("data query_device upload failed");
            cublasFree (matXNew_host);
            cublasShutdown();
            return;
        }

        printf("\n Original query on device\n");
        for (j = 0; j < query_width; j++) {
            printf(" *** query (%05d) : ", j);
            for (i = 0; i < height; i++) {
                printf ("(%7.5f)[%7.5f] - ", query_host[j*height+i], matXNew_host[j*height+i]);
            }
            printf("\n");
        }
    }


    // Vector of Lambda
    printf(" - Vector vecLambda_device : Device %d: \"%s\" ... ", dev, deviceProp.name);
    stat = cublasAlloc (height, sizeof(*lambda_host), (void**)&vecLambda_device);
    if (stat != CUBLAS_STATUS_SUCCESS) {
        printf ("device memory vecLambda_device allocation failed");
        cublasShutdown();
        return;
    }
    printf(" Ok - Set ...");
    stat = cublasSetVector (height, sizeof(*lambda_host), lambda_host, 1, vecLambda_device, 1);
    if (stat != CUBLAS_STATUS_SUCCESS) {
        printf ("data vecLambda_device download failed");
        cublasFree (vecLambda_device);
        cublasShutdown();
        return;
    }
    printf(" Done\n   ");
//    URANIE::GPU::Engine::UpdateFreeMemory();

    printf(" - Vector vecCoeff_device : Device %d: \"%s\" ... ", dev, deviceProp.name);
    stat = cublasAlloc (ref_width, sizeof(*Coefficient_host), (void**)&vecCoeff_device);
    if (stat != CUBLAS_STATUS_SUCCESS) {
        printf ("device memory vecCoeff_device allocation failed");
        cublasShutdown();
        return;
    }
    printf(" Ok - Set ...");
    stat = cublasSetVector (ref_width, sizeof(*Coefficient_host), Coefficient_host, 1, vecCoeff_device, 1);
    if (stat != CUBLAS_STATUS_SUCCESS) {
        printf ("data vecCoeff_device download failed");
        cublasFree (vecCoeff_device);
        cublasShutdown();
        return;
    }
    printf(" Done\n   ");
    URANIE::GPU::Engine::UpdateFreeMemory();


    printf(" - Vector vecDotVariance_device : Device %d: \"%s\" ... ", dev, deviceProp.name);
    stat = cublasAlloc (ref_width, sizeof(*Coefficient_host), (void**)&vecDotVariance_device);
    if (stat != CUBLAS_STATUS_SUCCESS) {
        printf ("device memory vecDotVariance_device allocation failed");
        cublasShutdown();
        return;
    }
    printf(" Ok - Done\n   ");
    URANIE::GPU::Engine::UpdateFreeMemory();


    // Vector of Min
    printf(" - Vector vecMin_device : Device %d: \"%s\" ... ", dev, deviceProp.name);
    stat = cublasAlloc (height, sizeof(*min_host), (void**)&vecMin_device);
    if (stat != CUBLAS_STATUS_SUCCESS) {
        printf ("device memory vecMin_device allocation failed");
        cublasShutdown();
        return;
    }
    printf(" Ok - Set ...");
    stat = cublasSetVector (height, sizeof(*min_host), min_host, 1, vecMin_device, 1);
    if (stat != CUBLAS_STATUS_SUCCESS) {
        printf ("data vecMin_device setting failed");
        cublasFree (vecMin_device);
        cublasShutdown();
        return;
    }
    printf(" done\n   ");

    // Vector of Max
    printf(" - Vector vecMax_device : Device %d: \"%s\" ... ", dev, deviceProp.name);
    stat = cublasAlloc (height, sizeof(*max_host), (void**)&vecMax_device);
    if (stat != CUBLAS_STATUS_SUCCESS) {
        printf ("device memory vecMax_device allocation failed");
        cublasShutdown();
        return;
    }
    printf(" Ok - Set ...");
    stat = cublasSetVector (height, sizeof(*max_host), max_host, 1, vecMax_device, 1);
    if (stat != CUBLAS_STATUS_SUCCESS) {
        printf ("data vecMax_device setting failed");
        cublasFree (vecMax_device);
        cublasShutdown();
        return;
    }
    printf(" done\n   ");

    // Vector of NormLambda
    printf(" - Vector vecNormLambda_device : Device %d: \"%s\" ... ", dev, deviceProp.name);
    stat = cublasAlloc (ref_width, sizeof(*Coefficient_host), (void**)&vecNormLambda_device);
    if (stat != CUBLAS_STATUS_SUCCESS) {
        printf ("device memory vecNormLambda_device allocation failed");
        cublasShutdown();
        return;
    }
    printf(" Ok - Set ...");
    stat = cublasSetVector (ref_width, sizeof(*Coefficient_host), Coefficient_host, 1, vecNormLambda_device, 1);
    if (stat != CUBLAS_STATUS_SUCCESS) {
        printf ("data vecNormLambda_device setting failed");
        cublasFree (vecNormLambda_device);
        cublasShutdown();
        return;
    }
    printf(" done\n");


    printf("**********************************\n");
    // Compute the constantes Maxthears by Blocks
    printf(" - Compute the constantes Maxthears by Blocks \n");
    const int nMaxThreadsPerBlock = deviceProp.maxThreadsPerBlock;
    printf(" ** nMaxThreadsPerBlock[%d]   (via deviceProp.maxThreadsPerBlock)\n", nMaxThreadsPerBlock);
    const int nMaxTheadsBlockDevice = 512;
    printf(" ** nMaxTheadsBlockDevice[%d] (Imposed by the developper!)\n", nMaxTheadsBlockDevice);
    const int nThreadsPerBlocky = height;                                     // Threads per block in Y
    printf(" ** nThreadsPerBlocky[%d]\n", nThreadsPerBlocky);
    const int nMaxThreadsPerBlockx = (nMaxTheadsBlockDevice+nThreadsPerBlocky-1)/nThreadsPerBlocky;    // Number of blocks in the grid
    printf(" ** nMaxThreadsPerBlockx[%d]\n", nMaxThreadsPerBlockx);
    int nThreadsPerBlockx = nMaxThreadsPerBlockx;
//    if ( NNew < nMaxThreadsPerBlockx )
//        nThreadsPerBlockx = NNew;
    printf(" ** nThreadsPerBlockx[%d]\n", nThreadsPerBlockx);
    dim3 threadsNormalize(nThreadsPerBlocky, nThreadsPerBlockx);
    printf(" ***** threadsNormalize (nThreadsPerBlocky[%d] , nThreadsPerBlockx[%d]) \n", nThreadsPerBlocky, nThreadsPerBlockx);

    int nThreadsPerBlock = nThreadsPerBlockx * nThreadsPerBlocky;
    printf(" ** nThreadsPerBlock[%d]\n", nThreadsPerBlock);
    int numBlocksNormalize = query_width/nThreadsPerBlock+1;    // Number of blocks in the grid
//    printf(" ** numBlocksNormalize[%d]\n", numBlocksNormalize);

    // Normalize the test points on device
//    printf(" - Normalize the test points on device \n");
    dim3 sizeBlocksNormalize(1, numBlocksNormalize);
    printf(" ***** sizeBlocksNormalize (1, numBlocksNormalize[%d]) \n", numBlocksNormalize);
    printf("**********************************\n");
    URANIE::GPU::Wrappers::gpuNormalizeSamples(sizeBlocksNormalize, threadsNormalize,vecMin_device,vecMax_device, query_dev, query_width, height);

    if ( blog ) {
        stat = cublasGetMatrix (height, query_width, sizeof(*query_host), query_dev, height, matXNew_host, height);
        if (stat != CUBLAS_STATUS_SUCCESS) {
            printf ("data query_device upload failed");
            cublasFree (query_dev);
            cublasShutdown();
            return;
        }

        printf(" - Normalized query on device\n");
        for (j = 510; j < query_width; j++) {
            printf(" *** Normalized Query (%05d) : ", j);
            for (i = 0; i < height; i++) {
                printf ("(%7.5f)[%7.5f] - ", query_host[j*height+i], matXNew_host[j*height+i]);
            }
            printf("\n");
        }

        printf(" *** Correlation Lengths (%05d) : ", height);
        for (i = 0; i < height; i++) {
            printf ("(%02d)[%7.5f] - ", i, lambda_host[i]);
        }
        printf("\n");
    }

    printf("**********************************\n");
    const int maxThreadsPerBlock = 512; // deviceProp.maxThreadsPerBlock; //// 512 1024
    printf(" ** maxThreadsPerBlock[%d]\n", maxThreadsPerBlock);
    printf(" ** nThreadsPerBlocky [%d]\n", nThreadsPerBlocky);
    const int threadsPerBlockx = (maxThreadsPerBlock+nThreadsPerBlocky-1)/nThreadsPerBlocky;    // Number of blocks in the grid
    printf(" ** threadsPerBlockx[%d]\n", threadsPerBlockx);

    const int threadsPerBlock = threadsPerBlockx * nThreadsPerBlocky;            // Threads per block in X
    printf(" ** threadsPerBlock[%d]\n", threadsPerBlock);

    int numBlocks = ref_width/threadsPerBlockx+1;    // Number of blocks in the grid
    dim3 sizeBlocks(1, numBlocks);
    printf(" ***** sizeBlocks (1, numBlocks[%d]) \n", numBlocks);

    if ( 1 ) {
        printf(" ** ref_width  [%06d]\n", ref_width);
        printf(" ** query_width[%06d]\n", query_width);
    }
    printf("**********************************\n");

    Real dProdCoeff=1.0f;

//    printf(" ** CUDA kernel 'normalizeXNewAndLambda' launch with %d blocks of %d threads\n", numBlocks, threadsPerBlock);
    int nBlocksPerGridNorm = query_width/nThreadsPerBlock+1;
//    printf(" ** nThreadsPerBlock[%d]\n", nThreadsPerBlock);
//    printf(" ** nBlocksPerGridNorm[%d]\n", nBlocksPerGridNorm);
//    printf(" ** CUDA kernel 'computeNorm2Matern32' launch with %d blocks of %d threads\n", nBlocksPerGridNorm, nThreadsPerBlock);
    printf("**********************************\n");
    printf(" ** nBlocksPerGridNorm[%d]\n", nBlocksPerGridNorm);
    printf(" ** nThreadsPerBlock[%d]\n", nThreadsPerBlock);
    printf("**********************************\n");

    for(int inew=0; inew<query_width; inew++) { //// query_width
        if ( blog )
            printf(" - inew (%02d) : Variance %e : ", inew, dVariance);

        // Compute ( Xnew[j] - X[i][j]) / lambda[j] for all i=1, ref_width
        URANIE::GPU::Wrappers::gpuNormalizeXNewAndLambdaIdx(sizeBlocks, threadsNormalize, query_dev, inew, vecLambda_device, ref_dev, matXtmp_device, ref_width, height);

        if ( blog ) {
            stat = cublasGetMatrix (height, ref_width, sizeof(*query_host), matXtmp_device, height, matXNew_host, height);
            if (stat != CUBLAS_STATUS_SUCCESS) {
                printf ("data query_device upload failed");
                cublasFree (query_dev);
                cublasShutdown();
                return;
            }
        }


        //        int blocksPerGridNorm = 1;
        ////        printf(" ** CUDA kernel 'computeNorm2Matern32'' launch with %d blocks of %d threads\n", blocksPerGridNorm, N_TOT);
        //        computeNorm2Matern32<<<blocksPerGridNorm, N_TOT>>>(matXtmp_device, devPtrVecNormLambda, N, M);
        //        printf(" ** CUDA kernel 'computeNorm2Matern32'' launch with %d blocks of %d threads\n", blocksPerGridNorm, N_TOT);
        URANIE::GPU::Kernels::computeNorm2Matern32<<<nBlocksPerGridNorm, nThreadsPerBlock>>>(matXtmp_device, vecNormLambda_device, ref_width, height);

        if ( blog ) {
            stat = cublasGetVector (ref_width, sizeof(*query_host), vecCoeff_device, 1, vecC_host, 1);
            if (stat != CUBLAS_STATUS_SUCCESS) {
                printf ("data vecCoeff_device upload failed");
                cublasFree (vecCoeff_device);
                cublasShutdown();
                return;
            }

            stat = cublasGetVector (ref_width, sizeof(*query_host), vecNormLambda_device, 1, vecCC_host, 1);
            if (stat != CUBLAS_STATUS_SUCCESS) {
                printf ("data vecNormLambda_device upload failed");
                cublasFree (vecNormLambda_device);
                cublasShutdown();
                return;
            }

            printf("\n NormalizeXNewAndLambda Reference on device\n");
            for (j = 0; j < ref_width; j++) {
                printf(" *** NormalizeXNewAndLambda (%05d) : ", j);
                for (i = 0; i < height; i++) {
                    printf ("(%7.5f) ", matXNew_host[j*height+i]);
                }
                printf(": CC (%7.5f) (%7.5f) :\n", vecC_host[j], vecCC_host[j]);
            }
            Real yyp=0.0f;
            for (j = 0; j < ref_width; j++) {
                yyp += vecC_host[j] * vecCC_host[j];
                printf(" (%05d/%05d) : C (%.5f) CC(%.5f) yyp (%.5f) \n", j+1, ref_width, vecC_host[j], vecCC_host[j], yyp);
            }
            printf(" yyp %0.5f", yyp);
        }


        // Compute the dot product : vecNormLamda * vecCoeff
        dProdCoeff = cublasSdot(ref_width, vecNormLambda_device, 1, vecCoeff_device, 1);

        yMean_host[inew] = dVariance * dProdCoeff ;
        if ( blog )
            printf("dProdCoeff %e - yStar : %e\n", dProdCoeff, yMean_host[inew]);

        // Compute the variance of the GP (step 1)
        cublasSgemv ('t', ref_width, ref_width, 1.0f, matLInv_device, ref_width, vecNormLambda_device, 1, 0.0, vecDotVariance_device, 1);

        if ( blog ) {
            stat = cublasGetVector (ref_width, sizeof(*query_host), vecDotVariance_device, 1, vecCC_host, 1);
            if (stat != CUBLAS_STATUS_SUCCESS) {
                printf ("data vecCC_host upload failed");
                cublasFree (vecCoeff_device);
                cublasShutdown();
                return;
            }
        }


        // Step 2
        if ( 0 ) {
            dProdCoeff = dVariance;
            for (j = 0; j < ref_width; j++) {
                if ( blog )
                    printf(" (%04d/%04d) : yp(%.5f)\n", j, ref_width, vecCC_host[j]);
                dProdCoeff -= dVariance * dVariance * vecCC_host[j] * vecCC_host[j];
            }
            yVariance_host[inew] = dProdCoeff; // dVariance - dVariance * dVariance * dProdCoeff;
        } else {
            dProdCoeff = cublasSdot(ref_width, vecDotVariance_device, 1, vecDotVariance_device, 1);
            yVariance_host[inew] = dVariance - dVariance * dVariance * dProdCoeff;
        }

    }

    if ( blog ) {
        for(int inew=0; inew<query_width; inew++)
            printf(" - inew (%02d) : yStar [%e]\n", inew, yMean_host[inew]);
        printf("\n");
    }
    // Free memory
    cudaFree(ref_dev);
    cudaFree(matLInv_device);
    cudaFree(query_dev);
    cudaFree(vecLambda_device);
}

void kernel_GP_With_Texture(float* ref_host, int ref_width, float* query_host, int query_width, int height, int k, float* yMean_host, float* yVariance_host)
    {

        unsigned int size_of_float = sizeof(float);
        unsigned int size_of_int   = sizeof(int);

        // Variables
        float        *query_dev;
        float        *ref_dev;


        float        *dist_dev;

        cudaArray    *ref_array;
        cudaError_t  result;
        size_t       query_pitch;
        size_t	     query_pitch_in_bytes;
        size_t       ref_pitch;
        size_t       ref_pitch_in_bytes;
        unsigned int       max_nb_query_traited;
        size_t       actual_nb_query_width;
        size_t memory_total;
        size_t memory_free;

//        int          *ind_dev;
//        size_t       ind_pitch;
//        size_t       ind_pitch_in_bytes;

        // Check if we can use texture memory for reference points
        unsigned int use_texture = ( ref_width*size_of_float<=MAX_TEXTURE_WIDTH_IN_BYTES && height*size_of_float<=MAX_TEXTURE_HEIGHT_IN_BYTES );
        printf(" - use_texture(%d)\n", use_texture);

        // CUDA Initialisation
        cuInit(0);

        // Check free memory using driver API ; only (MAX_PART_OF_FREE_MEMORY_USED*100)% of memory will be used
        CUcontext cuContext;
        CUdevice  cuDevice=0;
        cuCtxCreate(&cuContext, 0, cuDevice);
        cuMemGetInfo(&memory_free, &memory_total);
        cuCtxDetach (cuContext);

        // Determine maximum number of query that can be treated
        max_nb_query_traited = ( memory_free * MAX_PART_OF_FREE_MEMORY_USED - size_of_float * ref_width*height ) / ( size_of_float * (height + ref_width) + size_of_int * k);
        max_nb_query_traited = min( query_width, (max_nb_query_traited / 16) * 16 );
        printf(" - max_nb_query_traited (%d)\n", max_nb_query_traited);
        printf(" - memory_total         (%zd)\n", memory_total);
        printf(" - memory_free          (%zd)\n", memory_free);

        // Allocation of global memory for query points
        result = cudaMallocPitch( (void **) &query_dev, &query_pitch_in_bytes, max_nb_query_traited * size_of_float, height + ref_width);
        if (result){
            printErrorMessage(result, max_nb_query_traited*size_of_float*(height+ref_width));
            return;
        }
        URANIE::GPU::Engine::UpdateFreeMemory();
        query_pitch = query_pitch_in_bytes/size_of_float;
        printf(" - query_pitch          (%zd)\n", query_pitch );
        dist_dev    = query_dev + height * query_pitch;

//        // Allocation of global memory for indexes
//        result = cudaMallocPitch( (void **) &ind_dev, &ind_pitch_in_bytes, max_nb_query_traited * size_of_int, k);
//        if (result){
//            cudaFree(query_dev);
//            printErrorMessage(result, max_nb_query_traited*size_of_int*k);
//            return;
//        }
//           URANIE::GPU::Engine::UpdateFreeMemory();
//       ind_pitch = ind_pitch_in_bytes/size_of_int;

        // Allocation of memory (global or texture) for reference points
        cudaTextureObject_t texA = 0;
        if (use_texture){

            // Allocation of texture memory
            cudaChannelFormatDesc channelDescA = cudaCreateChannelDesc<float>();
            result = cudaMallocArray( &ref_array, &channelDescA, ref_width, height );
            if (result){
                printErrorMessage(result, ref_width*height*size_of_float);
//                cudaFree(ind_dev);
                cudaFree(query_dev);
                return;
            }
            cudaMemcpyToArray( ref_array, 0, 0, ref_host, ref_width * height * size_of_float, cudaMemcpyHostToDevice );
            // Set texture parameters and bind texture to array
            cudaResourceDesc resDesc;
            memset(&resDesc, 0, sizeof(resDesc));
            resDesc.resType = cudaResourceTypeArray ;
            resDesc.res.array.array = ref_array;
            
            cudaTextureDesc texDesc;
            memset(&texDesc, 0, sizeof(texDesc));
            texDesc.addressMode[0] = cudaAddressModeClamp;
            texDesc.addressMode[1] = cudaAddressModeClamp;
            texDesc.filterMode = cudaFilterModePoint;
            texDesc.normalizedCoords = 0;
            
            //cudaTextureObject_t texA = 0;
            cudaCreateTextureObject(&texA, &resDesc, &texDesc, NULL);

        }
        else{

            // Allocation of global memory
            result = cudaMallocPitch( (void **) &ref_dev, &ref_pitch_in_bytes, ref_width * size_of_float, height);
            if (result){
                printErrorMessage(result,  ref_width*size_of_float*height);
//                cudaFree(ind_dev);
                cudaFree(query_dev);
                return;
            }
            ref_pitch = ref_pitch_in_bytes/size_of_float;
            cudaMemcpy2D(ref_dev, ref_pitch_in_bytes, ref_host, ref_width*size_of_float,  ref_width*size_of_float, height, cudaMemcpyHostToDevice);
        }
           URANIE::GPU::Engine::UpdateFreeMemory();

        // Split queries to fit in GPU memory
        for (int i=0; i<query_width; i+=max_nb_query_traited){
            printf(" - Split queries to fit in GPU memory (%d / %d / +%d)\n", i, query_width, max_nb_query_traited);

            // Number of query points considered
            actual_nb_query_width = min( max_nb_query_traited, query_width-i );

            // Copy of part of query actually being treated
            cudaMemcpy2D(query_dev, query_pitch_in_bytes, &query_host[i], query_width*size_of_float, actual_nb_query_width*size_of_float, height, cudaMemcpyHostToDevice);

            // Grids ans threads
            dim3 g_16x16(actual_nb_query_width/16, ref_width/16, 1);
            dim3 t_16x16(16, 16, 1);
            if (actual_nb_query_width%16 != 0) g_16x16.x += 1;
            if (ref_width  %16 != 0) g_16x16.y += 1;
            //
            dim3 g_256x1(actual_nb_query_width/256, 1, 1);
            dim3 t_256x1(256, 1, 1);
            if (actual_nb_query_width%256 != 0) g_256x1.x += 1;
            //
            dim3 g_k_16x16(actual_nb_query_width/16, k/16, 1);
            dim3 t_k_16x16(16, 16, 1);
            if (actual_nb_query_width%16 != 0) g_k_16x16.x += 1;
            if (k  %16 != 0) g_k_16x16.y += 1;

            // Kernel 1: Compute all the distances
            if (use_texture)
                cuComputeDistanceTexture<<<g_16x16,t_16x16>>>(texA, ref_width, query_dev, actual_nb_query_width, query_pitch, height, dist_dev);
            else
                cuComputeDistanceGlobal<<<g_16x16,t_16x16>>>(ref_dev, ref_width, ref_pitch, query_dev, actual_nb_query_width, query_pitch, height, dist_dev);

            // Kernel 2: Sort each column
//            cuInsertionSort<<<g_256x1,t_256x1>>>(dist_dev, query_pitch, ind_dev, ind_pitch, actual_nb_query_width, ref_width, k);

            // Kernel 3: Compute square root of k first elements
            cuParallelSqrt<<<g_k_16x16,t_k_16x16>>>(dist_dev, query_width, query_pitch, k);

            // Memory copy of output from device to host
//            cudaMemcpy2D(&dist_host[i], query_width*size_of_float, dist_dev, query_pitch_in_bytes, actual_nb_query_width*size_of_float, k, cudaMemcpyDeviceToHost);
//            cudaMemcpy2D(&ind_host[i],  query_width*size_of_int,   ind_dev,  ind_pitch_in_bytes,   actual_nb_query_width*size_of_int,   k, cudaMemcpyDeviceToHost);
        }

        // Free memory
        if (use_texture)
            cudaFreeArray(ref_array);
        else
            cudaFree(ref_dev);
//        cudaFree(ind_dev);
        cudaFree(query_dev);
    }


/**
  * K nearest neighbor algorithm
  * - Initialize CUDA
  * - Allocate device memory
  * - Copy point sets (reference and query points) from host to device memory
  * - Compute the distances + indexes to the k nearest neighbors for each query point
  * - Copy distances from device to host memory
  *
  * @param ref_host      reference points ; pointer to linear matrix
  * @param ref_width     number of reference points ; width of the matrix
  * @param query_host    query points ; pointer to linear matrix
  * @param query_width   number of query points ; width of the matrix
  * @param height        dimension of points ; height of the matrices
  * @param k             number of neighbor to consider
  * @param dist_host     distances to k nearest neighbors ; pointer to linear matrix
  * @param dist_host     indexes of the k nearest neighbors ; pointer to linear matrix
  *
  */
void kernel_kNN(float* ref_host, int ref_width, float* query_host, int query_width, int height, int k, float* dist_host, int* ind_host)
{
    
    unsigned int size_of_float = sizeof(float);
    unsigned int size_of_int   = sizeof(int);
    
    // Variables
    float        *query_dev;
    float        *ref_dev;
    float        *dist_dev;
    int          *ind_dev;
    cudaArray    *ref_array;
    cudaError_t  result;
    size_t       query_pitch;
    size_t	     query_pitch_in_bytes;
    size_t       ref_pitch;
    size_t       ref_pitch_in_bytes;
    size_t       ind_pitch;
    size_t       ind_pitch_in_bytes;
    unsigned int       max_nb_query_traited;
    size_t       actual_nb_query_width;
    size_t memory_total;
    size_t memory_free;
    
	
	// Check if we can use texture memory for reference points
    unsigned int use_texture = ( ref_width*size_of_float<=MAX_TEXTURE_WIDTH_IN_BYTES && height*size_of_float<=MAX_TEXTURE_HEIGHT_IN_BYTES );
    printf(" - use_texture(%d)\n", use_texture);

    // CUDA Initialisation
    cuInit(0);
    
    // Check free memory using driver API ; only (MAX_PART_OF_FREE_MEMORY_USED*100)% of memory will be used
    CUcontext cuContext;
    CUdevice  cuDevice=0;
    cuCtxCreate(&cuContext, 0, cuDevice);
    cuMemGetInfo(&memory_free, &memory_total);
    
    // Determine maximum number of query that can be treated
    max_nb_query_traited = ( memory_free * MAX_PART_OF_FREE_MEMORY_USED - size_of_float * ref_width*height ) / ( size_of_float * (height + ref_width) + size_of_int * k);
    max_nb_query_traited = min( query_width, (max_nb_query_traited / 16) * 16 );
    printf(" - max_nb_query_traited     (%d)\n", max_nb_query_traited);
    printf(" - memory_total             (%zd)\n", memory_total);
    printf(" - memory_free              (%zd)\n", memory_free);

    // Allocation of global memory for query points and for distances
    result = cudaMallocPitch( (void **) &query_dev, &query_pitch_in_bytes, max_nb_query_traited * size_of_float, height + ref_width);
    if (result){
        printErrorMessage(result, max_nb_query_traited*size_of_float*(height+ref_width));
        return;
    }
      URANIE::GPU::Engine::UpdateFreeMemory();
    query_pitch = query_pitch_in_bytes/size_of_float;
    printf(" - query_pitch              (%zd)\n", query_pitch );
    dist_dev    = query_dev + height * query_pitch;
	
    // Allocation of global memory for indexes	
    result = cudaMallocPitch( (void **) &ind_dev, &ind_pitch_in_bytes, max_nb_query_traited * size_of_int, k);
	if (result){
        cudaFree(query_dev);
        printErrorMessage(result, max_nb_query_traited*size_of_int*k);
        return;
    }
       URANIE::GPU::Engine::UpdateFreeMemory();
   ind_pitch = ind_pitch_in_bytes/size_of_int;
    
	// Allocation of memory (global or texture) for reference points
    cudaTextureObject_t texA = 0;
    if (use_texture){
	
        // Allocation of texture memory
        cudaChannelFormatDesc channelDescA = cudaCreateChannelDesc<float>();
        result = cudaMallocArray( &ref_array, &channelDescA, ref_width, height );
        if (result){
            printErrorMessage(result, ref_width*height*size_of_float);
            cudaFree(ind_dev);
            cudaFree(query_dev);
            return;
        }
        cudaMemcpyToArray( ref_array, 0, 0, ref_host, ref_width * height * size_of_float, cudaMemcpyHostToDevice );
        
        // Set texture parameters and bind texture to array
        cudaResourceDesc resDesc;
        memset(&resDesc, 0, sizeof(resDesc));
        resDesc.resType = cudaResourceTypeArray ;
        resDesc.res.array.array = ref_array;
            
        cudaTextureDesc texDesc;
        memset(&texDesc, 0, sizeof(texDesc));
        texDesc.addressMode[0] = cudaAddressModeClamp;
        texDesc.addressMode[1] = cudaAddressModeClamp;
        texDesc.filterMode = cudaFilterModePoint;
        texDesc.normalizedCoords = 0;
            
        cudaCreateTextureObject(&texA, &resDesc, &texDesc, NULL);
    }
    else{
	
		// Allocation of global memory
        result = cudaMallocPitch( (void **) &ref_dev, &ref_pitch_in_bytes, ref_width * size_of_float, height);
        if (result){
            printErrorMessage(result,  ref_width*size_of_float*height);
            cudaFree(ind_dev);
            cudaFree(query_dev);
            return;
        }
        ref_pitch = ref_pitch_in_bytes/size_of_float;
        cudaMemcpy2D(ref_dev, ref_pitch_in_bytes, ref_host, ref_width*size_of_float,  ref_width*size_of_float, height, cudaMemcpyHostToDevice);
    }
       URANIE::GPU::Engine::UpdateFreeMemory();
   
    // Split queries to fit in GPU memory
    for (int i=0; i<query_width; i+=max_nb_query_traited){
        printf(" - Split queries to fit in GPU memory (%d / %d / +%d)\n", i, query_width, max_nb_query_traited);

		// Number of query points considered
        actual_nb_query_width = min( max_nb_query_traited, query_width-i );
        
        // Copy of part of query actually being treated
        cudaMemcpy2D(query_dev, query_pitch_in_bytes, &query_host[i], query_width*size_of_float, actual_nb_query_width*size_of_float, height, cudaMemcpyHostToDevice);
        
        // Grids ans threads
        dim3 g_16x16(actual_nb_query_width/16, ref_width/16, 1);
        dim3 t_16x16(16, 16, 1);
        if (actual_nb_query_width%16 != 0) g_16x16.x += 1;
        if (ref_width  %16 != 0) g_16x16.y += 1;
        //
        dim3 g_256x1(actual_nb_query_width/256, 1, 1);
        dim3 t_256x1(256, 1, 1);
        if (actual_nb_query_width%256 != 0) g_256x1.x += 1;
		//
        dim3 g_k_16x16(actual_nb_query_width/16, k/16, 1);
        dim3 t_k_16x16(16, 16, 1);
        if (actual_nb_query_width%16 != 0) g_k_16x16.x += 1;
        if (k  %16 != 0) g_k_16x16.y += 1;
        
        // Kernel 1: Compute all the distances
        if (use_texture)
            cuComputeDistanceTexture<<<g_16x16,t_16x16>>>(texA, ref_width, query_dev, actual_nb_query_width, query_pitch, height, dist_dev);
        else
            cuComputeDistanceGlobal<<<g_16x16,t_16x16>>>(ref_dev, ref_width, ref_pitch, query_dev, actual_nb_query_width, query_pitch, height, dist_dev);
            
        // Kernel 2: Sort each column
        cuInsertionSort<<<g_256x1,t_256x1>>>(dist_dev, query_pitch, ind_dev, ind_pitch, actual_nb_query_width, ref_width, k);

        // Kernel 3: Compute square root of k first elements
        cuParallelSqrt<<<g_k_16x16,t_k_16x16>>>(dist_dev, query_width, query_pitch, k);

        // Memory copy of output from device to host
        cudaMemcpy2D(&dist_host[i], query_width*size_of_float, dist_dev, query_pitch_in_bytes, actual_nb_query_width*size_of_float, k, cudaMemcpyDeviceToHost);
        cudaMemcpy2D(&ind_host[i],  query_width*size_of_int,   ind_dev,  ind_pitch_in_bytes,   actual_nb_query_width*size_of_int,   k, cudaMemcpyDeviceToHost);
    }
    
    // Free memory
    if (use_texture)
        cudaFreeArray(ref_array);
    else
        cudaFree(ref_dev);
    cudaFree(ind_dev);
    cudaFree(query_dev);

    cuCtxDestroy(cuContext);
}

