//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////

/* To avoid errors due to CUDA-specific stuff in the activation functions header, only
   include the "pure C++ stuff" (no CUDA, which would cause the compiler to complain) */
#include "utils.h"
#include "errors.h"
#ifndef CPU_PROCESSING_ONLY
  #include <cublas.h>
#endif

#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/random/mersenne_twister.hpp>
#include <boost/random/uniform_int_distribution.hpp>
#include <boost/random/variate_generator.hpp>
#include <boost/random/normal_distribution.hpp>
#include <ctime>                        // std::time() to seed random functions
#include <boost/format.hpp>             // Error formatting printf-style
#include <sstream>
#include <map>

namespace nn = NeuralNetworkLevenbergMarquardt;

bool nn::GPU::Engine::systemInitialized = false;

void nn::GPU::Engine::Init()
{
    if( systemInitialized ) return;

#ifndef CPU_PROCESSING_ONLY     // GPU
    // Initialize cuBLAS (TODO: check if it's acceptable to do it here, in case of multiple instances of the class)
    cublasStatus status = cublasInit();
    GPU::checkCuBLASError( "cublasInit [<problem>]", status );
#endif

    systemInitialized = true;
}

void nn::GPU::Engine::Shutdown()
{
    if( !systemInitialized ) return;

#ifndef CPU_PROCESSING_ONLY         // GPU
    cublasStatus status = cublasShutdown();
    GPU::checkCuBLASError( "cublasShutdown [<problem>]", status );
#endif

    systemInitialized = false;
}

nn::Real nn::Random::StandardGaussianValue()
{
    using namespace boost;
    static const Real mean = 0.0;
    static const Real stdd = 1.0;
    static mt19937 rGen( std::time(0) );               		// Generator (Mersenne)
    static normal_distribution<Real> distrib( mean, stdd );	// Gaussian distribution ~ N(0,1) )
    static variate_generator< mt19937&, normal_distribution<Real> > randValue( rGen, distrib );

    return randValue();
}

void nn::Random::RandomPermutation(size_t len, Int * list)
{
    using namespace boost;
    static mt19937 rGen( std::time(0) );                    // Generator (Mersenne)

    // Init list
    for (size_t i = 0; i < len; i++)
        list[i] = (Int)i;

    // Create a permutation with the Devroye Algorithm - TODO: merge with TPatternsEventList.Mix()
    for (size_t i = len - 1 ; i > 0; i--)
    {
        random::uniform_int_distribution<> distrib(0, i);
        int idx = distrib(rGen);
        Int tmp = list[idx];
        list[idx] = list[i];
        list[i] = tmp;
    }
}

nn::Reporting::Exception::Exception( const std::string err, const char* filename, size_t line )
: mError(err), mFilename(filename), mLine(line)
{
    std::stringstream errorMessage;
    errorMessage << "--Error (in \"" << mFilename << "\", l. " << mLine << ")" << std::endl;
    errorMessage << "\tDescription: " << mError << std::endl;
    mFullError = errorMessage.str();
}

    
nn::Reporting::Exception::~Exception() throw()
{
}

const char* nn::Reporting::Exception::what() const throw()
{
    return mFullError.c_str();
}

void nn::Debug::printMatrixGPU( const size_t M, const size_t N, const Real* deviceMatrix, std::ostream& out )
{
  #ifndef CPU_PROCESSING_ONLY
    Real* hostMatrix = new Real[M*N];   // MxN matrix, column-major
    cublasStatus status = cublasGetMatrix( M, N, sizeof(hostMatrix[0]), deviceMatrix, M, hostMatrix, M );
    GPU::checkCuBLASError( "debug [cublasGetMatrix][1]", status );
    out << "Displaying[GPU]::" << M << "x" << N << " matrix:" << std::endl;
    for( size_t i = 0; i < M; ++i )
    {
      for( size_t j = 0; j < N; ++j )
      {
        out << hostMatrix[i+j*M] << "\t";
      }
      out << std::endl;
    }
    delete[] hostMatrix;
  #endif
}

#ifndef NDEBUG
  #ifndef CPU_PROCESSING_ONLY
      void nn::GPU::checkCuBLASError( const std::string message, int status )
      {
          // If status == -1, were not using a cuBLAS helper function (returns a status) but a core one (doesn't)              
          cublasStatus cStatus = (status == -1) ? cublasGetError() : static_cast<cublasStatus>(status);
                  
          static std::map<cublasStatus, std::string> cublasErrors;
          static bool initialized = false;
          if( !initialized )
          {
              cublasErrors[ CUBLAS_STATUS_SUCCESS ] = "operation completed successfully";
              cublasErrors[ CUBLAS_STATUS_NOT_INITIALIZED ] = "CUBLAS library not initialized";
              cublasErrors[ CUBLAS_STATUS_ALLOC_FAILED ] = "resource allocation failed";
              cublasErrors[ CUBLAS_STATUS_INVALID_VALUE ] = "unsupported numerical value was passed to function";
              cublasErrors[ CUBLAS_STATUS_MAPPING_ERROR ] = "access to GPU memory space failed";
              cublasErrors[ CUBLAS_STATUS_EXECUTION_FAILED ] = "GPU program failed to execute";
              cublasErrors[ CUBLAS_STATUS_INTERNAL_ERROR ] = "an internal CUBLAS operation failed";
          }

          if( cStatus == CUBLAS_STATUS_NOT_INITIALIZED )
          {
              Reporting::Warn( message + ":\t" + cublasErrors[ cStatus ], __FILE__, __LINE__ );
              status = cublasInit();	// Retry
          }
          if( cStatus != CUBLAS_STATUS_SUCCESS )
          {
              throw Reporting::cublasException( message + ":\t" + cublasErrors[ cStatus ], __FILE__, __LINE__ );
          }
      }
  #endif
#endif
