/**
Copyright (c) 2011, Celifan (http://www.noctua-blog.com/)
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**/

#ifndef CPU_PROCESSING_ONLY

  #include "cublas.h"
  #include "errors.h"
  #include "lu.h"

  namespace nn = NeuralNetworkLevenbergMarquardt;

  void nn::LinearAlgebra::decomposeLU( int M, int N, int padSize, float* A, int* P, float epsilon, bool lastBlock )
  {
          // Preconditions
          if(  M<=0 || N<=0 || padSize < std::max(1,M) )
          {
              std::string error;
              if( M<=0 )
                  error = "M <= 0";
              if( N<=0 )
                  error = "N <= 0";
              if( padSize < std::max(1,M) )
                  error = "padding < max(1,M)";
              NeuralNetworkLevenbergMarquardt::Reporting::Err( error, __FILE__, __LINE__ );
          }
          int minDim = std::min( M, N );
          /* If it is the last (or the only) block to compute (matrix seen as one block (regular LU) or last block of the blocked LU
          decomposition, then lastBlock == 1 and we don't try to find a pivot for the last element of the diagonal (there is no further
          element); otherwise, we also have to look (if necessary) for a pivot for the last row (since there are more rows available after) */
          for( int k = 0; k < minDim - lastBlock; ++k )
          {
              /* Find the index of the first non-zero element (or at least the one with the biggest absolute value)
              of the kth column, starting at row k - that is, below the diagonal. */ 
              int pivotRow = k-1+cublasIsamax( M-k, A+k+k*padSize, 1 ); // row relative to the current submatrix
              int kp1 = k+1;
              P[k] = pivotRow;
              if( pivotRow != k )
              {
                  // Swap the two rows to have the non-zero element on the diagonal (the LU algorithm involves dividing by it)
                  cublasSswap( N, A+pivotRow, padSize, A+k, padSize );
              }
              float diagElemValue;
              cublasGetVector( 1, sizeof(float), A+k+k*padSize, 1, &diagElemValue, 1 );
              if( fabs(diagElemValue) < epsilon ) // Oops. It wasn't non-zero after all: the whole subdiagonal column is 0
              {
                  NeuralNetworkLevenbergMarquardt::Reporting::Err( "[decomposeLU] the matrix is singular.", __FILE__, __LINE__ );
              }
              if( kp1 < M )
              {
                  // Rescale the subdiagonal column to have a 1 on the diagonal
                  cublasSscal( M-kp1, 1.0f/diagElemValue, A+kp1+k*padSize, 1 );
              }
              if( kp1 < minDim )
              {
                  // A{k+1:M,k+1:N} = A{k+1:M,k+1:N} - A{k+1:M,k}*A{k,k+1:N} (modify the lower right triangular submatrix of A)
                  cublasSger( M-kp1, N-kp1, -1.0f, A+kp1+k*padSize, 1, A+k+kp1*padSize, padSize, A+kp1*padSize+kp1, padSize );
              }
          }
          GPU::checkCuBLASError( "decomposeLU [<problem>]" );
 }

  /** \brief Performs "blocked" LU decomposition of a M-by-N matrix A, using a pivot vector.
      This is performed with cuBLAS methods and relies on DecomposeLU(); it is an in-place
      algorithm, meaning the matrix A is altered during the process: at the end, A contains
      the result of the LU decomposition (upper tringular part is U, lower part is L, assuming
      the diagonal of L consists of ones only and is therefore not stored).
      @param M			number of rows of A
      @param N			number of columns of A
      @param padSize		leading dimension of A (pad >= max(1,M)), that is the increment between two elements of the same row
      @param A			M-by-N matrix of float, to be decomposed
      @param P			M-dimensional permutation vector (P(k) is the index, in the initial matrix, of the kth row of the output one)
      @param blockSize	block size (if blockSize >= M or blockSize = 1, "regular" decomposition is called)
      @param epsilon 		"approximatively zero value" (used to test for singularities)
  **/
  void nn::LinearAlgebra::decomposeBlockedLU( int M, int N, int padSize, float *A, int* P, int blockSize, float epsilon )
  {
      // Preconditions
      if(  M<=0 || N<=0 || padSize < std::max(1,M) )
      {
          std::string error;
          if( M<=0 )
              error = "M <= 0";
          if( N<=0 )
              error = "N <= 0";
          if( padSize < std::max(1,M) )
              error = "padding < max(1,M)";
          Reporting::Err( error, __FILE__, __LINE__ );
      }
      int minSize = std::min(M,N);
      if( blockSize > minSize || blockSize == 1 )
      {
          decomposeLU( M, N, padSize, A, P, epsilon ); // Fallback on regular LU decomposition
      }
      else
      {
          /* Blocked decomposition: we start with the full A, compute 3 blocks (upper-left, lower-left and upper-right)
          parts of its decomposition, and then we iteratively consider the remaining part A', dividing at each step in
          4 new blocks. */
          for( int i = 0; i < minSize; i += blockSize )
          {
              const int realBlockSize  = std::min( minSize-i, blockSize );
              
              const bool lastIteration = (realBlockSize == minSize-i);
              
              // Decompose the current rectangular block (lower-left and top-left part of A')
              decomposeLU( M-i, realBlockSize, padSize, A+i+i*padSize, P+i, epsilon, lastIteration );
              // Adjust pivot information - TODO : write a kernel for that
              for( int p = i; p < std::min( M, i+realBlockSize )-lastIteration; ++p )
              {
                  P[p] = P[p]+i;
                  if( P[p] != p )	// DecomposeLU() swapped two parts of the rows p and P[p]: we have to swap the rest of them
                  {
                      // Apply interchanges to columns 0:i.
                      cublasSswap( i, A+p , padSize, A+P[p], padSize );
                      GPU::checkCuBLASError( "decomposeBlockedLU [cublasSswap][1]" );
                      // Apply interchanges to columns i+blockSize:N.
                      cublasSswap( N-i-realBlockSize, A+p+(i+realBlockSize)*padSize , padSize, A+P[p]+(i+realBlockSize)*padSize, padSize );
                      GPU::checkCuBLASError( "decomposeBlockedLU [cublasSswap][2]" );
                  }
              }

              /* Compute block row of U (the upper right rectangular block): solves L * X = 1.0 * U, where L is the
              (M-i)-by-R lower left part of A' we already computed at the beginning of this step, U the R-by-(N-i-R)
              upper right part (still to compute, do it's still the original data). At the end, X is stored in U. */
              cublasStrsm( 'l','l','n','u', realBlockSize, N-i-realBlockSize, 1.0f,
                          A+i+i*padSize, padSize,
                          A+i+(i+realBlockSize)*padSize, padSize
                          );
              GPU::checkCuBLASError( "decomposeBlockedLU [cublasStrsm]" );

              if( i+realBlockSize < M )
              {
                  // A'' = -1.0*LU+1.0*A'' (modify the lower right triangular submatrix of A')
                  cublasSgemm( 'n','n', M-i-realBlockSize, N-i-realBlockSize, realBlockSize,
                              -1.0f,
                              A+i+realBlockSize+i*padSize, padSize,		// L 
                              A+i+(i+realBlockSize)*padSize, padSize,	// U
                              1.0f,
                              A+i+realBlockSize+(i+realBlockSize)*padSize, padSize // A'' (lower left block of A', still to compute)
                              );
                  GPU::checkCuBLASError( "decomposeBlockedLU [cublasSgemm]" );
              }
          }
      }
  }

#endif // #ifndef CPU_PROCESSING_ONLY

