/**
Copyright (c) 2011, Celifan (http://www.noctua-blog.com/)
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**/

#ifndef __CUBLAS_LU_H__
#define __CUBLAS_LU_H__

namespace NeuralNetworkLevenbergMarquardt
{
    namespace LinearAlgebra
    {

        /** \brief Performs "regular" LU decomposition of a M-by-N matrix A, using a pivot vector.
        	This is performed with cuBLAS methods; it is an in-place algorithm, meaning the matrix A
        	is altered during the process: at the end, A contains the result of the LU decomposition
        	(upper tringular part is U, lower part is L, assuming the diagonal of L consists of ones
        	only and is therefore not stored).
            @param M		number of rows of A
            @param N		number of columns of A
            @param padSize	leading dimension of A (pad >= max(1,M)), that is the increment between two elements of the same row
            @param A		M-by-N matrix of float, to be decomposed
            @param P		M-dimensional permutation vector (P(k) is the index, in the initial matrix, of the kth row of the output one)
            @param epsilon 	"approximatively zero value" (used to test for singularities)
            @param lastBlock 	should be set to true, unless we are using blocked LU decomposition and it is /not/ the last block to compute.
        **/
        void decomposeLU( int M, int N, int padSize , float* A, int* P, float epsilon, bool lastBlock = true );

        /** \brief Performs "blocked" LU decomposition of a M-by-N matrix A, using a pivot vector.
        	This is performed with cuBLAS methods and relies on DecomposeLU(); it is an in-place
        	algorithm, meaning the matrix A is altered during the process: at the end, A contains
        	the result of the LU decomposition (upper tringular part is U, lower part is L, assuming
        	the diagonal of L consists of ones only and is therefore not stored).
            @param M			number of rows of A
            @param N			number of columns of A
            @param padSize		leading dimension of A (pad >= max(1,M)), that is the increment between two elements of the same row
            @param A			M-by-N matrix of float, to be decomposed
            @param P			M-dimensional permutation vector (P(k) is the index, in the initial matrix, of the kth row of the output one)
            @param blockSize	block size (if blockSize >= M or blockSize = 1, "regular" decomposition is called)
            @param epsilon 		"approximatively zero value" (used to test for singularities)
        **/
        void decomposeBlockedLU( int M, int N, int padSize, float *A, int* P, int blockSize,float epsilon );

    }	// --> LinearAlgebra

} // --> NeuralNetworkLevenbergMarquardt

#endif /* __CUBLAS_LU_H__ */
