//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////

/*
 * multilayerperceptron.cpp
 *
 *  Created on: 18 juil. 2014
 *      Author: adrien
 */

#include "multilayerperceptron.h"
#include "activation_functions.h"
#include <boost/timer.hpp>
#ifdef NN_DISPLAY_INFO
#include <iomanip> // For std::setw
#include <boost/progress.hpp>
#endif
#include <boost/numeric/ublas/io.hpp>   // To output the vectors with std::cout

#ifdef CPU_PROCESSING_ONLY  // For the CPU version of LU decomposition
    #include "boost/numeric/ublas/triangular.hpp"
    #include <boost/numeric/ublas/lu.hpp>
#else
    #include <cublas.h>
    #include "lu.h"
    #include "kernels_gpu.h"
#endif

namespace NeuralNetworkLevenbergMarquardt
{

const Real TwoLayersPerceptron::LAMBDA_DEFAULT_FACTOR = 0.1;
const Real TwoLayersPerceptron::LAMBDA_INCREASE_FACTOR = 10.0;
const Real TwoLayersPerceptron::LAMBDA_DECREASE_FACTOR = 0.1;
const Real TwoLayersPerceptron::LAMBDA_MAX_VALUE = 1e25;
const Real TwoLayersPerceptron::LAMBDA_MIN_VALUE = 1e-25;
const Real TwoLayersPerceptron::ERROR_STOPPING_THRESHOLD = 1e-3;
const size_t TwoLayersPerceptron::MAX_TRAINING_ITERATIONS = 1000;
const size_t TwoLayersPerceptron::MAX_NOCHANGE_ITERATIONS = 50;

/* ************************************* */
/*         NeuronLayer          */
/* ************************************* */

NeuronLayer::NeuronLayer(int m, int k, NeuronType t)
: M(m), K(k), T(t), mInitialized( false )
{
}

void NeuronLayer::setInput( const Vector& input )
{
    namespace ublas  = boost::numeric::ublas;
    if( input.size() != M && input.size() != M+1 ) // Tolerance: M+1 would be fine, just means the bias was already added
        Reporting::Err(
                boost::str( boost::format("Bad dimension for input vector (must be %1%, got %2%).") % M % input.size() ), __FILE__, __LINE__
        );
    if( !mInitialized) initialize();
    // Set the input
    static ublas::vector_range<Vector> inputMinusBias( mInput, ublas::range(0, M) );
    const ublas::vector_range<const Vector> newInputMinusBias( input, ublas::range(0, M) );
    inputMinusBias.assign( newInputMinusBias );
}

void NeuronLayer::initialize()
{
    /*
            For the w(i,j) ~ N(0,sqrt(1/M) ), see C.Bishop, 1995, "Neural Networks for Pattern
            Recognition", section 7.5: 'Initializing weights'. It allows to have values around
            unit, for the initial activation value (linear output), which, for the sigmoid
            function, is the "right" area of the curve so that the optimization process starts
            quickly enough.
            However, the final choice was to use a Gaussian distribution, standard (i.e. N(0,1)),
            and then to renormalize the weights so that the Euclidean norm of the matrix is unit.
     */
    // Real std = std::sqrt(1.0/M);

    // Resize everything
    mWeights.resize( K, M+1 );  // +1 for the bias
    mInput.resize( M+1 );       // Ditto
    mInput(M) = 1.0;            // Bias "input"
    mPreOutput.resize( K );
    mOutput.resize( K );
    // Random weights, and overall (Euclidean) norm to normalize them
    Real normWeights = 0.0;
    for( size_t i = 0; i < K; ++i )
    {
        for( size_t j = 0; j < M+1; ++j )
        {
            Real w = Random::StandardGaussianValue();
            ///+DEBUGGING: to have a deterministic matrix.
            ///w = j+i * (float) (static_cast<float>(j%3)-1.0f);
            ///-DEBUGGING
            mWeights(i,j) = w;
            normWeights += w*w;
        }
    }
    normWeights = std::sqrt(normWeights);
    for( size_t i = 0; i < K; ++i )
    {
        for( size_t j = 0; j < M+1; ++j )
        {
            mWeights(i,j) /= normWeights;
        }
    }

    // Done.
    mInitialized = true;
}

/* ************************************* */
/*         TwoLayersPerceptron          */
/* ************************************* */


TwoLayersPerceptron::TwoLayersPerceptron(int m, int k, int p, NeuronType t,
        const Real learningRate, const Real precThres, const size_t maxIterations )
: M(m),K(k),P(p),T(t), NUMBER_PARAMETERS(k*(m+p+1)+p),
  mHiddenLayer(m,k,t), mOutputLayer(k,p, kLinear),
  mLambda( learningRate ), mPrecisionThreshold( precThres ), mMaxIterations( maxIterations ),
  mMaxNoChangeIter( std::max( maxIterations/20, static_cast<size_t>(30) ) ), // Max. 5% of unfruitful epochs
  mHasBeenTrained( false ), mInitialized( false ), mHasRunOnInput( false ), mSamplesNormalized(false),
  mTrainingSetSize(0), mTestSetStartIdx(0), mActualTrainingSetSize(0),
  mTotalErrorLearn(std::numeric_limits<double>::max()), mTotalErrorTest(std::numeric_limits<double>::max()),
  mBestTotalErrorTest(std::numeric_limits<double>::max()), mBestTotalErrorLearn(std::numeric_limits<double>::max()),
  mTrainingTime(0.0)
#ifdef CPU_PROCESSING_ONLY
,mErrors(), mJacobian(), mHessian(), mGradient(), mHiddenLayerBackupWeights(), mOutputLayerBackupWeights()
#else // GPU
,mDeviceErrors(0) ,mDeviceInputValues(0),mDeviceTeacherValues(0),mDeviceOutputs(0),mDeviceIdentityRR(0),
mDeviceLMMatrix(0),mDeviceJacobian(0),mDeviceGradient(0),mDeviceHessian(0),
mDeviceWeightsHidden(0),mDeviceWeightsOutput(0),mDeviceWeightsBackup(0),mDeviceWeightsBest(0),
mDeviceHiddenOutput(0), mDeviceHiddenOutput_restr(0), mDeviceInputValues_restr(0)
#endif
{
    if( mLambda <= 0.0 )
        Reporting::Err( "Learning rate (Levenberg parameter) must be (strictly) positive.", __FILE__, __LINE__  );
    else if( precThres <= 0.0 )
        Reporting::Err( "Requested precision must be (strictly) positive.", __FILE__, __LINE__  );

    // This implementation only supports real-valued outputs (P=1)
    if (P != 1)
        Reporting::Err( "Number of output neurons (P) must be equal to 1!", __FILE__, __LINE__  );

    mMaxSampleValue = -std::numeric_limits<Real>::max();
    mMinSampleValue = std::numeric_limits<Real>::max();
}

TwoLayersPerceptron::~TwoLayersPerceptron()
{
    cleanup();
}


void TwoLayersPerceptron::cleanup()
{
    resetTeachingSamples();
    mLambda = LAMBDA_DEFAULT_FACTOR;
    mInitialized = false;
    mHasRunOnInput = false;
}


void TwoLayersPerceptron::resetTeachingSamples()
{
    // Reset the learning data
    mMaxSampleValue = -std::numeric_limits<Real>::max();
    mMinSampleValue = std::numeric_limits<Real>::max();
    mActualTrainingSetSize = 0;
    mInputValues.clear();
    mTeacherValues.clear();
    mSamplesNormalized = false;
    mHasBeenTrained = false;
}


void TwoLayersPerceptron::reset()
{
    cleanup();
    initialize();
}


void TwoLayersPerceptron::initialize()
{
    // Initialize common (if not already done)
    GPU::Engine::Init();

    // Initialize layers
    mHiddenLayer.initialize();
    mOutputLayer.initialize();

    // Initialize the damping factor
//    mLambda = LAMBDA_DEFAULT_FACTOR;

#ifdef CPU_PROCESSING_ONLY
    // For the LM algorithm, we'll need to backup the weights
    mHiddenLayerBackupWeights.resize( K, M+1 );
    mOutputLayerBackupWeights.resize( 1, K+1 );
#endif

    mInitialized = true;
}

/* Run on the input, with the current weights (must have been already trained if used in production) */

Real TwoLayersPerceptron::run( const Vector& input )
{
    if( !mInitialized ) initialize();

    mHiddenLayer.setInput( input );
    mHiddenLayer.run();
    mOutputLayer.setInput( mHiddenLayer.getOutput() );
    mOutputLayer.run();
    mHasRunOnInput = true;
    return mOutputLayer.getOutput()(0);
}

/* Train (batch mode) on the pool (Levenberg-Marquardt algorithm): backpropagate the total error, return false if something went wrong. */

bool TwoLayersPerceptron::train( const size_t nMixTrainSet, const size_t nTrials )
{
    using boost::numeric::ublas::matrix_row;

    // Measure total time taken by the training, and outputs it to console at the end if NN_DISPLAY_INFO is set
//    boost::timer t;
//    t.restart();

    // Error of the NN in a particular iteration: set to max at the beginning
    mTotalErrorLearn = std::numeric_limits<Real>::max();
    mTotalErrorTest = std::numeric_limits<Real>::max();

    // Initialize everything (weights and stuff)
    if( !mInitialized ) initialize();

    // Training set cannot be empty at this point
    const size_t N_TOT = getTrainingSampleCount();
    const size_t N_LEARN = getLearningSampleCount();
    if( N_TOT == 0 ) return false;

    // Test if training set is complete
    if( mTrainingSetSize > mActualTrainingSetSize )
    {
        Reporting::Err( "Error: number of samples is less than originally expected/specified.", __FILE__, __LINE__  );
    }
    mInputValues.resize( mTrainingSetSize, M+1, true ); // M+1 for the bias (see below)
    for( size_t n = 0; n < N_TOT; ++n ) mInputValues( n, M ) = 1.0;

    // Allocate space for the samples in the GPU and transfer them
    setupDeviceMemoryGPU();

    // Displays the training settings (as a recap)
    information();

    /** ************************************************************** ***/
    Real minTestErrorAmongstTrials = std::numeric_limits<Real>::max();
    Real minLearnErrorAmongstTrials = std::numeric_limits<Real>::max();
    Matrix bestHiddenWeights, bestOutputWeights;
    std::cout << "--------------------------------------------------------------------------" << std::endl;
    for (size_t super_i  = 0; super_i < nMixTrainSet; super_i++)
    {
        std::cout << "Training #" << super_i << " ...";
        if (super_i > 0)  // no need to shuffle the first time
        {
            initialize();
            copyWeightsToDevice();
            randomizeTrainingSetGPU();
            saveRestriction();
        }

        size_t trial = 0;

        // Indicators
        size_t nbTotalIterations = 0;
        short stopConditionMask = 0;
        bool saved = false;
        while( trial < nTrials )
        {
            size_t nbTotalIterationsTrial = 0;
            short stopConditionMaskTrial = 0;
#ifdef CPU_PROCESSING_ONLY
            //mHasBeenTrained = trainCPU( stopConditionMaskTrial, nbTotalIterationsTrial );
            !! DEPRECATED !!
#else // GPU
            mHasBeenTrained = trainGPU( stopConditionMaskTrial, nbTotalIterationsTrial );
#endif

            // Is it getting better? (Or do you feel the same?)
            if( mBestTotalErrorTest < minTestErrorAmongstTrials )
            {
                minTestErrorAmongstTrials = mBestTotalErrorTest;
                minLearnErrorAmongstTrials = mBestTotalErrorLearn;
                bestHiddenWeights = mHiddenLayer.getWeights();
                bestOutputWeights = mOutputLayer.getWeights();
                saved = true;
            }
            else
            {
//                std::cout << "---NOT saving this iteration: " << mBestTotalErrorLearn << " | " << mBestTotalErrorTest << std::endl;
            }
            stopConditionMask = stopConditionMaskTrial;
            nbTotalIterations = nbTotalIterationsTrial;

            // Reinitialize layers with random weights (Will it make it easier on you now?)
            mHiddenLayer.initialize();
            mOutputLayer.initialize();

            ++trial; // (You've got someone to blame)
        }

        // Summarizes the training for this randomization of the test base
        if (saved)
        {
            std::cout << " saved! Summary:" << std::endl;
            summarize( stopConditionMask, nbTotalIterations,  minLearnErrorAmongstTrials, minTestErrorAmongstTrials);
        }
        else
        {
            std::cout << " discarded! Summary:" << std::endl;
            summarize( stopConditionMask, nbTotalIterations,  mBestTotalErrorLearn, mBestTotalErrorTest);
        }

    }
    // Finally assign the best candidate to the members:
    mHiddenLayer.getWeights() = bestHiddenWeights;
    mOutputLayer.getWeights() = bestOutputWeights;
    mTotalErrorTest = minTestErrorAmongstTrials;
    mTotalErrorLearn = minLearnErrorAmongstTrials;

    // Deallocate the device memory dedicated to the teacher values and inputs, errors, hessian and jacobian
    cleanupDeviceMemoryGPU();

    // Trained, but no actual computation performed yet: getOutput() shouldn't be called.
    mHasRunOnInput = false;

    return mHasBeenTrained;
}


void TwoLayersPerceptron::getWeights(double *w)
{
    // In the GPU world, biais are stored at the bottom of the arrays / matrices
    // but they're stored at the begining of blocks in TANNModeler.

    const Matrix & hidden = mHiddenLayer.getWeights();
    // = K*(M+1) matrix, with K number of neurons in the hidden layer and M number of
    //   incoming links on each neurons.
    //   +1 accounts for the biais neuron/link.
    const Matrix & output = mOutputLayer.getWeights();  // 1*(K+1) -- +1 for the biais neuron in the hidden layer

    // Convention from the TANNModeler class - see doc of member _vecWeights.
    w[0] = output(0,K);  // output biais
    for (size_t i = 0; i < K; i++)
    {
        w[i*(M+2) + 1] = output(0, i);
        w[i*(M+2) + 2] = hidden(i, M);

        for (size_t j = 0; j < M; j++)
            w[i*(M+2)+j + 3] = hidden(i, j);
    }
}

#ifdef NN_DISPLAY_INFO

void TwoLayersPerceptron::information() const
{
    std::string functionType = "Sigmoid (logistic) - default";
    std::string normalizType = "[0,1]";
    switch(T)
    {
    case kLinear:
        functionType = "Linear";
        normalizType = "(-inf, inf) (none)";
        break;
    case kTanh:
        functionType = "Tanh (hyperbolic tangent)";
        normalizType = "[-1,1]";
        break;
    case kGauss:
        functionType = "Gaussian";
        normalizType = "(-inf, inf) (none)";
        break;
    case kOff:
        functionType = "None (zero function)";
        normalizType = "(-inf, inf) (none)";
        break;
    default:
        break;
    }
    std::cout << std::setfill('-') << std::setw(75) << " " << std::endl;
    std::cout << "--- Starting training on a two-layer perceptron with the Levenberg-Marquardt algorithm:" << std::endl;
    std::cout << "\tInput layer:\t" << M << " neuron"  << (M > 1 ? "s": "") << std::endl;
    std::cout << "\tHidden layer:\t" << K << " neuron" << (K > 1 ? "s": "") << std::endl;
    std::cout << "\tOutput layer:\t" << P << " neuron (mandatory)" << std::endl;
    std::cout << "\tActivation function (for the hidden neurons):\t" << functionType << std::endl;
    std::cout << "\t(output neuron has an identity function)" << std::endl;
    std::cout << "\tRequested final precision:\t" << mPrecisionThreshold << std::endl;
    std::cout << "\tInitial Levenberg factor:\t" << mLambda << std::endl;
    std::cout << "\tMaximum number of epochs:\t" << mMaxIterations << std::endl;
    std::cout << "\tMaximum number of unfruitful iterations:\t" << mMaxNoChangeIter << std::endl;
    std::cout << "\tTraining set:\t" << std::endl;
    std::cout << "\t\t" << "Total sample count:\t" << getTrainingSampleCount() << std::endl;
    std::cout << "\t\t" << "Number dedicated to pure training (the rest for validation):\t" << getLearningSampleCount() << std::endl;
    //std::cout << "\t\t" << "Maximum value:\t" << mMaxSampleValue << std::endl;
    //std::cout << "\t\t" << "Minimum value:\t" << mMinSampleValue << std::endl;
    //std::cout << "\t\t" << "Renormalization (linear) in " << normalizType << " done." << std::endl;
}


void TwoLayersPerceptron::summarize( const short stopCondMask, size_t nbIterations, Real totalErrorLearn, Real totalErrorTest ) const
{
    if( !mHasBeenTrained )
    {
        Reporting::Warn( "No training done yet.", __FILE__, __LINE__ );
    }
    else
    {
        std::string message;
        if( stopCondMask & 0x01 )
        {
            message = "requested precision reached.";
        }
        else if( stopCondMask & 0x02 )
        {
            message = "maximum number of epochs reached.";
        }
        else if( stopCondMask & 0x04 )
        {
            message = "maximum number of iterations without improvement reached.";
        }
        std::cout << std::endl;
        std::cout << "---Training ended: " <<  message << std::endl;
        std::cout << "\tMean squared error (learning): \t" << totalErrorLearn/getLearningSampleCount() << std::endl;
        std::cout << "\tMean squared error (testing): \t" << totalErrorTest/getTestingSampleCount() << std::endl;
        std::cout << "\tRequested precision:\t" << mPrecisionThreshold << std::endl;
        std::cout << "\tNumber of iterations:\t" << nbIterations << (nbIterations >= mMaxIterations ? " (maximum reached)" : "") << std::endl;
        std::cout << std::setfill('-') << std::setw(75) << " " << std::endl;
    }
}
#else
void TwoLayersPerceptron::summarize( const short, size_t ) const {}
void TwoLayersPerceptron::information() const {}
#endif

/* Return the last vector computed with run(), normalized */

const Real& TwoLayersPerceptron::getOutput() const
{
    if( !mHasRunOnInput )
        Reporting::Err( "::getOutput() mustn't be called as long as a computation has not been made with ::run()", __FILE__, __LINE__  );
    return mOutputLayer.getOutput()(0);
}

/* Return the last vector computed with run(), denormalized (actual range) */

Real TwoLayersPerceptron::getDenormalizedOutput() const
{
    return denormalizeOutput( this->getOutput() );
}

/* Denormalize (actual range) a value, based on the previous training data (samples) */

Real TwoLayersPerceptron::denormalizeOutput( const Real nOutput ) const
{
    namespace pp = Preprocessing;
    switch(T)
    {
    case kSigmoid:
        return pp::denormalize<pp::kZeroOne>( nOutput, getMinTrainingSampleValue(), getMaxTrainingSampleValue() );
    case kTanh:
        return pp::denormalize<pp::kMinusOneOne>( nOutput, getMinTrainingSampleValue(), getMaxTrainingSampleValue() );
    case kGauss:
    case kLinear:
    case kOff:
    default:
        return nOutput;
    }
}

/* Add a (x,t) pair to the training pool (allows duplicate entries) */

void TwoLayersPerceptron::addTrainingSample( const Vector& input, const Vector& teacherOutput )
{
    namespace ublas = boost::numeric::ublas;

    if( input.size() != M )
        Reporting::Err(
                boost::str( boost::format("Bad dimension for training vector (must be %1%, got %2%).") % M % input.size() ),
                __FILE__, __LINE__
        );
    else if( teacherOutput.size() != P )
        Reporting::Err(
                boost::str( boost::format("Bad dimension for teacher output vector (must be %1%, got %2%).") % P % teacherOutput.size() ),
                __FILE__, __LINE__
        );

    // If needed, increase the size of the matrix (more samples than initially specified)
    if( mTeacherValues.size() <= mActualTrainingSetSize )
    {
        //setTrainingSampleCount( mActualTrainingSetSize + 1 );
        Reporting::Err("Trying to add too many samples!", __FILE__, __LINE__);
    }
    // Access the current row of mInputValues, and set it to the input value
    MatrixRow matRow( mInputValues, mActualTrainingSetSize );
    ublas::vector_range<MatrixRow> matRowMinusBias( matRow, ublas::range(0, M) );
    matRowMinusBias.assign( input );

    // Access the current row of mTeacherValues, and set it to the teacher value
    mTeacherValues(mActualTrainingSetSize) = teacherOutput(0);

    // Update min and max values
    mMinSampleValue = std::min( mMinSampleValue, teacherOutput(0) );
    mMaxSampleValue = std::max( mMaxSampleValue, teacherOutput(0) );

    // Increment the current index
    ++mActualTrainingSetSize;

    // If we added one, then they are unlikely to be all normalized already
    if( mSamplesNormalized )
    {
        Reporting::Warn( "Adding a sample to a pool of already normalized ones will result in double normalizations for the previous ones - this might lead to very strange results. To prevent this, be sure to add all the samples at the same time, and then only to do the training.", __FILE__, __LINE__ );
        mSamplesNormalized = false;
    }
}

/* Sets the *expected* size of the training pool (as an indicator: adding more samples is still possible and will update this value) */

int TwoLayersPerceptron::setTrainingSampleCount( const size_t count, const float ratio )
{
    mTrainingSetSize = std::max( count, mTeacherValues.size() );
    if( mTeacherValues.size() < mTrainingSetSize )
    {
        mTeacherValues.resize( mTrainingSetSize, true );
        mInputValues.resize( mTrainingSetSize, M+1, true ); // +1 for the bias
    }
    mTestSetStartIdx = int(ratio * mTrainingSetSize) - 1;
    if (mTestSetStartIdx <= 1 || mTestSetStartIdx >= (mTrainingSetSize-2))
        Reporting::Err( "setTrainingSampleCount: invalid ratio specified, this results in too many or too little elements in the test database.", __FILE__, __LINE__ );
    //std::cout << "setCount " << mTestSetStartIdx << "  " << getLearningSampleCount() << "  " <<  mTrainingSetSize << std::endl;
    return mTestSetStartIdx;
}


Real TwoLayersPerceptron::getMaxTrainingSampleValue() const
{
    if( !mHasBeenTrained )
    {
        Reporting::Warn( "This function shouldn't be called before the training has been done (values might " +
                std::string("become irrelevant if new samples are added)."), __FILE__, __LINE__ );
    }
    return mMaxSampleValue;
}


Real TwoLayersPerceptron::getMinTrainingSampleValue() const
{
    if( !mHasBeenTrained )
    {
        Reporting::Warn( "This function shouldn't be called before the training has been done (values might " +
                std::string("become irrelevant if new samples are added)."), __FILE__, __LINE__ );
    }
    return mMinSampleValue;
}

/* *********** NeuronLayer::run() *********** */
 void NeuronLayer::run()
{
    if( !mInitialized) initialize();

    for( size_t j = 0; j < K; ++j )
    {
        // Compute linear combination for jth node of the layer
        Real c = 0.0;
        for( size_t i = 0; i < M+1; ++i )
        {
            c += mWeights(j,i)*mInput(i);
        }
        mPreOutput(j) = c;
        // And apply the activation function
        switch(T)
        {
            case kLinear:
                mOutput(j) = mPreOutput(j);
                break;
            case kSigmoid:
                mOutput(j) = ActivationFunction::Sigmoid( mPreOutput(j) );
                break;
            case kTanh:
                mOutput(j) = ActivationFunction::Tanh( mPreOutput(j) );
                break;
            case kGauss:
                mOutput(j) = ActivationFunction::Gauss( mPreOutput(j) );
                break;
            case kOff:
                mOutput(j) = 0.0;
                break;
        }
    }
}

/************************************************************************
          CPU-only functions
 ************************************************************************/

#ifdef CPU_PROCESSING_ONLY

/* Runs on a matrix of input. CPU version */

Vector TwoLayersPerceptron::runBatch(  const Matrix& inputMatrix  )
  {
    namespace ublas = boost::numeric::ublas;

    // Convenient shortcut for the number of training samples
    const size_t N = inputMatrix.size1();

    // Final output
    Vector output(N);

    // One-sample input vector (with bias)
    Vector input(M+1);
    input(M) = static_cast<Real>(1.0);
    ublas::vector_range<Vector> inputMinusBias( input, ublas::range(0, M) );
    // For each sample, run the NN on the input
    for( size_t n = 0; n < N; ++n )
    {
      const ublas::matrix_row<const Matrix> input_n( inputMatrix, n );
      inputMinusBias.assign( input_n );
      output(n) = run( input );
    }

    return output;
  }

/* *********** TwoLayersPerceptron::computeDeltaWeights() *********** */

const Vector TwoLayersPerceptron::computeDeltaWeights()
{
    Matrix A( mHessian ); // Now, "A = H = tJJ"

    for( size_t i = 0; i < NUMBER_PARAMETERS; ++i )
    {
        A(i,i) += mLambda;
        /*  Use, instead, A(i,i) *= (mLambda+1.0) to have "tJJ+lambda.Diag(tJJ)";
            but as a result, the matrix seems (experimentally) to become singular
            a lot more often */
    }
    // Now, "A = tJJ+lambda.I"

    Vector g( mGradient );    // g = tJ*error, where error_n = output_n - teacher_n
    // Solve Ax = g, and put the result in g: it will be our delta vector

    namespace ublas = boost::numeric::ublas;
    ublas::permutation_matrix<> piv( NUMBER_PARAMETERS );
    int isSingular = lu_factorize( A, piv );
    if( !isSingular )
        ublas::lu_substitute( A, piv, g );
    else
        Reporting::Err( "In file" + std::string(__FUNCTION__) + " : can't solve linear system (singular matrix).", __FILE__, __LINE__) ;

    return g;
}

/* *********** TwoLayersPerceptron::computeJacobian() *********** */

void TwoLayersPerceptron::computeJacobian( const size_t n )
{
    // Backpropagation method
    const Matrix& WO = mOutputLayer.getWeights();
    const Vector& IO = mOutputLayer.getInput();


    /* For the output units */
    for( size_t k = 0; k < K+1; ++k )
        mJacobian( n, K*(M+1) + k ) = IO(k);        // dF/dw(p,k)=z_k (P=1 => p=0 + identity function for activation)

    /* For the hidden units */
    const Vector& LOH = mHiddenLayer.getLinearOutput();    // Linear output (LO) of the hidden layer (H)
    const Vector& OH  = mHiddenLayer.getOutput();          // Actual output (O) of the hidden layer (H)
    const Vector& IH = mHiddenLayer.getInput();
    for( size_t m = 0; m < M+1; ++m )    // +1 for the bias
    {
        for( size_t k = 0; k < K; ++k )
        {
            Real dk = WO(0,k);     // sum_{p=0}^{P-1} w(p,k) (P=1)
            switch(T)            // [ f'(a_k)*sum_{p=0}^{P-1} w(p,k) ]
            {
                case kLinear:
                    break;
                case kSigmoid:
                    // Faster then dSigmoid( LOH(k) ), since we already have computed the sigmoid
                    dk *= ActivationFunction::dSigmoidWithFunctionValue( OH(k) );
                    break;
                case kTanh:
                    dk *= ActivationFunction::dTanhWithFunctionValue( OH(k) );
                    break;
                case kGauss:
                    dk *= ActivationFunction::dGauss( LOH(k) );
                    break;
                case kOff:
                    dk = 0.0;
                    break;
            }
            mJacobian( n, m*K+k ) = dk*IH(m);    // [ f'(a_k)*sum_{p=0}^{P-1} w(p,k) ]*x_m = f'(a_k)*w(0,k)*x_m
        }
    }
}


void TwoLayersPerceptron::computeHessian()
{
    if( mHessian.size1() != NUMBER_PARAMETERS || mHessian.size2() != NUMBER_PARAMETERS )
    {
        Reporting::Warn( "Wrongly-dimensioned Hessian: shouldn't be happening.", __FILE__, __LINE__ );
        mHessian.resize( NUMBER_PARAMETERS, NUMBER_PARAMETERS, false );
    }
    const size_t N = mJacobian.size1();
    // Compute tJJ
    for( size_t i = 0; i < NUMBER_PARAMETERS; ++i )
    {
        for( size_t j = i; j < NUMBER_PARAMETERS; ++j )
        {
            Real c = 0.0;
            for( size_t k = 0; k < N; ++k )
            {
                c += mJacobian(k,i)*mJacobian(k,j);
            }
            mHessian(i,j) = c;
            mHessian(j,i) = c; // The Hessian is symmetric.
        }
    }
}


void TwoLayersPerceptron::computeGradient()
{
    if( mGradient.size() != NUMBER_PARAMETERS )
    {
        Reporting::Warn( "Wrongly-dimensioned gradient: shouldn't be happening.", __FILE__, __LINE__ );
        mGradient.resize( NUMBER_PARAMETERS, false );
    }
    const size_t N = mJacobian.size1();
    // Compute tJ.e, where e_n = output_n - teacher_n
    for( size_t i = 0; i < NUMBER_PARAMETERS; ++i )
    {
        Real c = 0.0;
        for( size_t j = 0; j < N; ++j )
        {
            c += mJacobian(j,i)*mErrors(j);
        }
        mGradient(i) = c;
    }
}

/* *********** TwoLayersPerceptron::computeErrorForSample() *********** */

Real TwoLayersPerceptron::computeErrorForSample( const size_t n )
{
    // Access the nth component of mErrors, and put the current error inside
    const Vector& output = mOutputLayer.getOutput();
    const Real c = output(0) - mTeacherValues(n);
    mErrors(n) = c;
    return c*c;
}

/* *********** TwoLayersPerceptron::modifyWeights() *********** */

void TwoLayersPerceptron::modifyWeights( const Vector& delta )
{
    if( delta.size() != NUMBER_PARAMETERS )
        Reporting::Err(
          boost::str(
              boost::format("Bad dimension for delta weight vector (must be %1%, got %2%).") % NUMBER_PARAMETERS % delta.size()
              ), __FILE__, __LINE__
        );

    // Backup previous weights
    mHiddenLayerBackupWeights.assign( mHiddenLayer.getWeights() );
    mOutputLayerBackupWeights.assign( mOutputLayer.getWeights() );

    // And modify the weights
    Matrix& WH = mHiddenLayer.getWeights();
    for( size_t m = 0; m < M+1; ++m )    // +1 for the bias
         for( size_t k = 0; k < K; ++k )
             WH(k,m) -= delta(m*K+k);

    Matrix& WO = mOutputLayer.getWeights();
    for( size_t k = 0; k < K+1; ++k )    // +1 for the bias
         for( size_t p = 0; p < P; ++p )
             WO(p,k) -= delta( K*(M+1) + k*P+p );    // P=1 ...
}

/* *********** TwoLayersPerceptron::restoreWeights() *********** */

void TwoLayersPerceptron::restoreWeights()
{
    mHiddenLayer.getWeights().assign( mHiddenLayerBackupWeights );
    mOutputLayer.getWeights().assign( mOutputLayerBackupWeights );
}

/* Apply normalization to the samples, so that the output values lay in [0,1] */

bool TwoLayersPerceptron::normalizeSamples()
{
    using Preprocessing::normalize;
    using Preprocessing::kZeroOne;
    using Preprocessing::kMinusOneOne;

    if( !mSamplesNormalized )
    {
        // Convenient shortcut for the number of training samples
        const size_t N = getTrainingSampleCount();
        if( N <= 1 )
        {
            Reporting::Warn( "Not enough samples to apply normalization.", __FILE__, __LINE__ );
            return false;
        }
        // Finding the min and max has been done along the way already
        if( mMinSampleValue == mMaxSampleValue )
        {
            Reporting::Warn( "All values are identical: constant function to approximate?.", __FILE__, __LINE__ );
        }
        for( size_t n = 0; n < N; ++n )
        {
            switch(T)
            {
                case kGauss:
                case kLinear:
                case kOff:
                    break;
                case kSigmoid:
                    mTeacherValues(n) = normalize<kZeroOne>( mTeacherValues(n), mMinSampleValue, mMaxSampleValue );
                    break;
                case kTanh:
                    mTeacherValues(n) = normalize<kMinusOneOne>( mTeacherValues(n), mMinSampleValue, mMaxSampleValue );
                    break;
            }
        }
        mSamplesNormalized = true;
    }

    return mSamplesNormalized;
}

/* *********** Train (batch mode) on the pool (LM algorithm) - CPU version *********** */

bool TwoLayersPerceptron::trainCPU( short& stopConditionMask, size_t& nbTotalIterations )
{
  // Convenient shortcut for the number of training samples
  const size_t N = getTrainingSampleCount();

  /** Normalize the samples (or fail if we can't) **/
  if( !normalizeSamples() ) return false;

  // Allocate memory for the matrix of errors (N, so not really big: ~10k elements max?), discarding previous values.
  mErrors.resize( N, false );
  // Ditto for the Jacobian (NxR, way bigger). TODO: useful to erase the previous values, as they will be overwritten anyway?
  mJacobian.resize( N, NUMBER_PARAMETERS, false );
  // And the Hessian approximation
  mHessian.resize( NUMBER_PARAMETERS, NUMBER_PARAMETERS, false );
  // And the gradient
  mGradient.resize( NUMBER_PARAMETERS, false );

  /** Until stopping condition has been reached, iterate **/
  nbTotalIterations = 0;
  size_t nbNoChangeIterations = 0;
  // Stopping conditions
  bool stopCondThreshold = false;
  bool stopCondMaxIterat = false;
  bool stopCondNoChangeIterat = false;
  #ifdef NN_DISPLAY_PROGRESSBAR
    // Allows us to display the overall progress, in terms of epochs
    boost::progress_display progressDisplayer( mMaxIterations );
  #endif
  while( !stopCondThreshold && !stopCondMaxIterat && !stopCondNoChangeIterat )
  {
      #ifndef NDEBUG
        std::cout << "---- (" << nbTotalIterations << ")\t" << mTotalError << "||" << std::sqrt(mTotalError/N) << "\t" << mLambda;
      #endif

      /* 1. Present all input samples to the NN, compute the outputs and the error vector,
        and calculate the total error.
        2. Compute the Jacobian matrix J=(dE/dwi), NxR, with the usual BP method,
        row by row - each row correspond to a sample. */
      mTotalError = 0.0;
      for( size_t n = 0; n < N; ++n ) // For each sample
      {
          const MatrixRow input( mInputValues, n );
          run( input );        // Run the NN on the input
          mTotalError += computeErrorForSample(n);    // Computes the total error, sample by sample
          computeJacobian(n);    // Fill a row of the Jacobian
      }
      // Here, we have our Jacobian. Compute the Hessian approximation, based on it.
      computeHessian();

      // And the gradient tJ.e (e is the N-dimensional error vector)
      computeGradient();

      Real currError = mTotalError+1.0;
      /* Now, the trust region algorithm (Levenberg-Marquardt) itself
          Modify the trust region according to whether the change of weight was:
                  (i)  a good thing: then, expand the trust region
                  (ii) a bad thing: then, shrink the trust region, and revert to old weights
      */
      // Backup previous weights
      mHiddenLayerBackupWeights.assign( mHiddenLayer.getWeights() );
      mOutputLayerBackupWeights.assign( mOutputLayer.getWeights() );

      mLambda *= LAMBDA_DECREASE_FACTOR;
      while( currError > mTotalError && mLambda < LAMBDA_MAX_VALUE && nbNoChangeIterations < mMaxNoChangeIter )
      {
          mLambda *= LAMBDA_INCREASE_FACTOR;
          // If we are in this loop, it means the weight modification was no good. Restore previous weights.
          restoreWeights();

          try
          {
              /* 3. Compute \Delta(w) = ( J^T J + \lambda I )^{-1} J^T e
                where e is the N-dimensional error vector (e_i), e_i = y_i-t_i. */
              const Vector deltaWeights = computeDeltaWeights();

              /* 4. Update the weights of the NN accordingly. */
              modifyWeights( deltaWeights );
              /* Re-run on all inputs with the new weights to see if there is an improvement */
              currError = 0.0;
              mErrors.clear();
              for( size_t n = 0; n < N; ++n ) // For each sample
              {
                  const MatrixRow input( mInputValues, n );
                  run( input );                // Run the NN on the input
                  currError += computeErrorForSample(n);    // Fill a component of the error vector, and return its square
              }

              ++nbNoChangeIterations;   // One more iteration with no change (will be reset to 0 if there was)
          }
          catch( const Reporting::cublasException& e )
          {
              mLambda = ( mLambda == 1.0 ) ? 1000.0 : 1.0;
              Reporting::Warn( e.what(), __FILE__, __LINE__ );
              return false;
          }
          catch( const std::exception& e )
          {
              mLambda = ( mLambda == 1.0 ) ? 1000.0 : 1.0;
              Reporting::Warn( e.what(), __FILE__, __LINE__ );
              continue;
          }
      }

      mLambda *= LAMBDA_DECREASE_FACTOR;

      if( currError < mTotalError ) nbNoChangeIterations = 0; // Something actually improved.
      mTotalError = currError; // Save new shiny error
      ++nbTotalIterations;

      #ifdef NN_DISPLAY_PROGRESSBAR
        ++progressDisplayer;    // Update also the progress bar
      #endif

      // Stopping conditions?
      stopCondThreshold = ( mTotalError <= (mPrecisionThreshold*N) );
      stopCondMaxIterat = ( nbTotalIterations >= mMaxIterations );
      stopCondNoChangeIterat = ( nbNoChangeIterations >= mMaxNoChangeIter );
      #ifndef NDEBUG
        std::cout << std::right << "\tStop flags:\t"
                  << stopCondThreshold << " " << stopCondMaxIterat << " " << stopCondNoChangeIterat << std::endl;
      #endif
  }

  // Set the stopping condition mask, to specify why it ended
  stopConditionMask  = static_cast<short>( stopCondThreshold );
  stopConditionMask += 2*static_cast<short>( stopCondMaxIterat );
  stopConditionMask += 4*static_cast<short>( stopCondNoChangeIterat );

  return true;
}
#else // #ifdef CPU_PROCESSING_ONLY

void TwoLayersPerceptron::setupDeviceMemoryGPU()
{
    cublasStatus status;
    // Convenient shortcuts for the number of training samples
    const size_t N_TOT = getTrainingSampleCount();
    const size_t N_LEARN = getLearningSampleCount();

    // Jacobian (N_LEARNxR matrix, i.e. N_LEARNxNUMBER_PARAMETERS)
    status = cublasAlloc( N_LEARN*NUMBER_PARAMETERS, sizeof(mDeviceJacobian[0]), (void**)&mDeviceJacobian );
    GPU::checkCuBLASError( "setupDeviceMemory [cublasAlloc][mDeviceJacobian]", status );
    // Hessian (RxR matrix)
    status = cublasAlloc( NUMBER_PARAMETERS*NUMBER_PARAMETERS, sizeof(mDeviceHessian[0]), (void**)&mDeviceHessian );
    GPU::checkCuBLASError( "setupDeviceMemory [cublasAlloc][mDeviceHessian]", status );
    // Networks output errors (N_TOT 1-dimensional errors)
    status = cublasAlloc( N_TOT, sizeof(mDeviceErrors[0]), (void**)&mDeviceErrors );
    GPU::checkCuBLASError( "setupDeviceMemory [cublasAlloc][mDeviceErrors]", status );
    // Networks outputs (N_TOT 1-dimensional vectors)
    status = cublasAlloc( N_TOT, sizeof(mDeviceOutputs[0]), (void**)&mDeviceOutputs );
    GPU::checkCuBLASError( "setupDeviceMemory [cublasAlloc][mDeviceOutputs]", status );
    // Gradient (R-dimensional vector, tJ.errors)
    status = cublasAlloc( NUMBER_PARAMETERS, sizeof(mDeviceGradient[0]), (void**)&mDeviceGradient );
    GPU::checkCuBLASError( "setupDeviceMemory [cublasAlloc][mDeviceGradient]", status );
    // Teacher inputs (N_TOT (M+1)-dimensional inputs)
    status = cublasAlloc( N_TOT*(M+1), sizeof(mDeviceInputValues[0]), (void**)&mDeviceInputValues );
    GPU::checkCuBLASError( "setupDeviceMemory [cublasAlloc][mDeviceInputValues]", status );

    status = cublasAlloc( N_LEARN*(M+1), sizeof(mDeviceInputValues_restr[0]), (void**)&mDeviceInputValues_restr );
    GPU::checkCuBLASError( "setupDeviceMemory [cublasAlloc][mDeviceInputValues_restr]", status );

    // Teacher outputs (N_TOT 1-dimensional outputs)
    Real* hostTeacherValues = mTeacherValues.data().begin(); // Shortcut
    status = cublasAlloc( N_TOT, sizeof(mDeviceTeacherValues[0]), (void**)&mDeviceTeacherValues );
    GPU::checkCuBLASError( "setupDeviceMemory [cublasAlloc][mDeviceTeacherValues]", status );
    status = cublasSetVector( N_TOT, sizeof(hostTeacherValues[0]), hostTeacherValues, 1, mDeviceTeacherValues, 1 );
    GPU::checkCuBLASError( "setupDeviceMemory [cublasSetVector][mDeviceTeacherValues]", status );
    // Weights backup (for both layers)
    status = cublasAlloc( NUMBER_PARAMETERS, sizeof(mDeviceWeightsBackup[0]), (void**)&mDeviceWeightsBackup );
    GPU::checkCuBLASError( "setupDeviceMemory [cublasAlloc][mDeviceWeightsBackup]", status );
    // Weights copy (for both layers) - used to store the best candidate in the Leven-Marq algo
    status = cublasAlloc( NUMBER_PARAMETERS, sizeof(mDeviceWeightsBest[0]), (void**)&mDeviceWeightsBest );
    GPU::checkCuBLASError( "setupDeviceMemory [cublasAlloc][mDeviceWeightsBest]", status );
    // NUMBER_PARAMETERS = (M+1)*K + (K+1)*P = (M+1)*K + (K+1)
    // "Identity RxR matrix" (actually, R-dimensional unit vector) (used in ~computeDeltaWeightsGPU())
    status = cublasAlloc( NUMBER_PARAMETERS, sizeof(mDeviceIdentityRR[0]), (void**)&mDeviceIdentityRR );
    GPU::checkCuBLASError( "setupDeviceMemory [cublasAlloc][mDeviceIdentityRR]", status );
    // Matrix we use to store tJJ+lambda.Id = mDeviceHessian + mLambda * mDeviceIdentityRR
    status = cublasAlloc( NUMBER_PARAMETERS*NUMBER_PARAMETERS, sizeof(mDeviceLMMatrix[0]), (void**)&mDeviceLMMatrix );
    GPU::checkCuBLASError( "setupDeviceMemory [cublasAlloc][mDeviceLMMatrix]", status );
    // Permutation of the training set
//    status = cublasAlloc( mNumberPermutations*mTrainingSetSize, sizeof(mDevicePermutations[0]), (void**)&mDevicePermutations );
//        GPU::checkCuBLASError( "setupDeviceMemory [cublasAlloc][mDevicePermutations]", status );

    /* Sets up the parts common to training and batch run (that is, needed to run the neural network
    on an input matrix:
      - Input: M columns of input, +1 column of bias.
      - Weights (first and second layer): allocate and set
      - Intermediate result: N K-dimensional sample outputs of the hidden layer, plus the bias - one per sample
      - Networks outputs (N 1-dimensional vectors)
    */
    setupDeviceMemoryRunOnGPU();
}


void TwoLayersPerceptron::saveRestriction()
{
    const size_t N_TOT = getTrainingSampleCount();
    const size_t N_LEARN = getLearningSampleCount();

    /* Save restriction */
    for (int ii=0; ii < M+1; ii++)
    {
        cublasScopy( N_LEARN, mDeviceInputValues + N_TOT*ii, 1,  mDeviceInputValues_restr + N_LEARN*ii, 1 );
        GPU::checkCuBLASError( "setupDeviceMemoryRunOnGPU [cublasScopy][mDeviceInputValues][mDeviceInputValues_restr]" );
    }
}

void TwoLayersPerceptron::setupDeviceMemoryRunOnGPU( const Real* hostInputValues, const size_t number )
{
  cublasStatus status;
  // Convenient shortcut for the number of training samples/input vectors
  const size_t N_TOT = (hostInputValues != 0) ? number : getTrainingSampleCount();
  const size_t N_LEARN = getLearningSampleCount();

  // Input: M columns of input, +1 column of bias.
  status = cublasAlloc( N_TOT*(M+1), sizeof(mDeviceInputValues[0]), (void**)&mDeviceInputValues );
  GPU::checkCuBLASError( "setupDeviceMemoryRunOnGPU [cublasAlloc][mDeviceInputValues]", status );
  status = cublasAlloc( N_LEARN*(M+1), sizeof(mDeviceInputValues_restr[0]), (void**)&mDeviceInputValues_restr );
  GPU::checkCuBLASError( "setupDeviceMemoryRunOnGPU [cublasAlloc][mDeviceInputValues_restr]", status );

  // Transfer the samples and set the bias.
  std::vector<Real> vectorOne( N_TOT, static_cast<Real>(1.0) );
  hostInputValues = (hostInputValues != 0) ? hostInputValues : mInputValues.data().begin(); // Shortcut
  status = cublasSetVector( N_TOT*M, sizeof(hostInputValues[0]), hostInputValues, 1, mDeviceInputValues, 1 );
  GPU::checkCuBLASError( "setupDeviceMemoryRunOnGPU [cublasSetVector][mDeviceInputValues][1]", status );
  hostInputValues = &vectorOne[0]; // Shortcut (for the bias column)
  status = cublasSetVector( N_TOT, sizeof(hostInputValues[0]), hostInputValues, 1, mDeviceInputValues + N_TOT*M, 1 );
  GPU::checkCuBLASError( "setupDeviceMemoryRunOnGPU [cublasSetVector][mDeviceInputValues][2]", status );

  /* Save restriction */
  saveRestriction();
//  std::cout << "=======> [" << dum_iter << "] mDeviceInputValues_restr\n";
//  Debug::printMatrixGPU(N_LEARN, M+1, mDeviceInputValues_restr);

  // Weights (first and second layer): allocate and set
  status = cublasAlloc( K*(M+1), sizeof(mDeviceWeightsHidden[0]), (void**)&mDeviceWeightsHidden );
  GPU::checkCuBLASError( "setupDeviceMemoryRunOnGPU [cublasAlloc][mDeviceWeightsHidden]", status );
  status = cublasAlloc( (K+1), sizeof(mDeviceWeightsOutput[0]), (void**)&mDeviceWeightsOutput );
  GPU::checkCuBLASError( "setupDeviceMemoryRunOnGPU [cublasAlloc][mDeviceWeightsOutput]", status );

  copyWeightsToDevice();

  // Intermediate result: N K-dimensional sample outputs of the hidden layer, plus the bias - one per sample
  status = cublasAlloc( N_TOT*(K+1), sizeof(mDeviceHiddenOutput[0]), (void**)&mDeviceHiddenOutput );
  GPU::checkCuBLASError( "setupDeviceMemoryRunOnGPU [cublasAlloc][mDeviceHiddenOutput]", status );
  status = cublasAlloc( N_LEARN*(K+1), sizeof(mDeviceHiddenOutput_restr[0]), (void**)&mDeviceHiddenOutput_restr );
  GPU::checkCuBLASError( "setupDeviceMemoryRunOnGPU [cublasAlloc][mDeviceHiddenOutput_restr]", status );

  // Networks outputs (N 1-dimensional vectors)
  status = cublasAlloc( N_TOT, sizeof(mDeviceOutputs[0]), (void**)&mDeviceOutputs );
  GPU::checkCuBLASError( "setupDeviceMemoryRunOnGPU [cublasAlloc][mDeviceOutputs]", status );
}

void TwoLayersPerceptron::copyWeightsToDevice()
{
    Real* hostMatrix;
    cublasStatus status;
    hostMatrix = mOutputLayer.getWeights().data().begin(); // Shortcut
    status = cublasSetMatrix( 1, K+1, sizeof(hostMatrix[0]), hostMatrix, 1, mDeviceWeightsOutput, 1 );
    GPU::checkCuBLASError( "setupDeviceMemoryRunOnGPU [cublasSetMatrix][mDeviceWeightsOutput]", status );
    hostMatrix = mHiddenLayer.getWeights().data().begin(); // Shortcut
    status = cublasSetMatrix( K, M+1, sizeof(hostMatrix[0]), hostMatrix, K, mDeviceWeightsHidden, K );
    GPU::checkCuBLASError( "setupDeviceMemoryRunOnGPU [cublasSetMatrix][mDeviceWeightsHidden]", status );
}

void TwoLayersPerceptron::cleanupDeviceMemoryGPU()
{
  cublasStatus status;
  // Training-specific stuff
  status = cublasFree( mDeviceIdentityRR );
  GPU::checkCuBLASError( "cleanupDeviceMemory [cublasFree][mDeviceIdentityRR]", status );
  status = cublasFree( mDeviceLMMatrix );
  GPU::checkCuBLASError( "cleanupDeviceMemory [cublasFree][mDeviceLMMatrix]", status );
  status = cublasFree( mDeviceTeacherValues );
  GPU::checkCuBLASError( "cleanupDeviceMemory [cublasFree][mDeviceTeacherValues]", status );
  status = cublasFree( mDeviceJacobian );
  GPU::checkCuBLASError( "cleanupDeviceMemory [cublasFree][mDeviceJacobian]", status );
  status = cublasFree( mDeviceGradient );
  GPU::checkCuBLASError( "cleanupDeviceMemory [cublasFree][mDeviceGradient]", status );
  status = cublasFree( mDeviceHessian );
  GPU::checkCuBLASError( "cleanupDeviceMemory [cublasFree][mDeviceHessian]", status );
  status = cublasFree( mDeviceErrors );
  GPU::checkCuBLASError( "cleanupDeviceMemory [cublasFree][mDeviceErrors]", status );
  status = cublasFree( mDeviceWeightsBackup );
  GPU::checkCuBLASError( "cleanupDeviceMemory [cublasFree][mDeviceWeightsBackup]", status );
  status = cublasFree( mDeviceWeightsBest );
  GPU::checkCuBLASError( "cleanupDeviceMemory [cublasFree][mDeviceWeightsBest]", status );
  // And the rest
  cleanupDeviceMemoryRunOnGPU();
}

void TwoLayersPerceptron::cleanupDeviceMemoryRunOnGPU()
{
  cublasStatus status;
  status = cublasFree( mDeviceInputValues );
  GPU::checkCuBLASError( "cleanupDeviceMemoryRunOnGPU [cublasFree][mDeviceInputValues]", status );
  status = cublasFree( mDeviceInputValues_restr );
  GPU::checkCuBLASError( "cleanupDeviceMemoryRunOnGPU [cublasFree][mDeviceInputValues_restr]", status );
  status = cublasFree( mDeviceWeightsHidden );
  GPU::checkCuBLASError( "cleanupDeviceMemoryRunOnGPU [cublasFree][mDeviceWeightsHidden]", status );
  status = cublasFree( mDeviceWeightsOutput );
  GPU::checkCuBLASError( "cleanupDeviceMemoryRunOnGPU [cublasFree][mDeviceWeightsOutput]", status );
  status = cublasFree( mDeviceHiddenOutput );
  GPU::checkCuBLASError( "cleanupDeviceMemoryRunOnGPU [cublasFree][mDeviceHiddenOutput]", status );
  status = cublasFree( mDeviceHiddenOutput_restr );
  GPU::checkCuBLASError( "cleanupDeviceMemoryRunOnGPU [cublasFree][mDeviceHiddenOutput_restr]", status );
  status = cublasFree( mDeviceOutputs );
  GPU::checkCuBLASError( "cleanupDeviceMemoryRunOnGPU [cublasFree][mDeviceOutputs]", status );
}

void TwoLayersPerceptron::randomizeTrainingSetGPU()
{
    Int * perm;
    perm = new Int[mTrainingSetSize];
    Random::RandomPermutation(mTrainingSetSize, perm);

    for (int i=1; i < mTrainingSetSize; i++)
      {
        cublasSswap( M+1, mDeviceInputValues + perm[i-1], mTrainingSetSize, mDeviceInputValues + perm[i], mTrainingSetSize );
        GPU::checkCuBLASError( "randomizeTrainingSetGPU [cublasSswap][mDeviceInputValues]" );
        cublasSswap( 1, mDeviceTeacherValues + perm[i-1], 1, mDeviceTeacherValues + perm[i], 1 );
        GPU::checkCuBLASError( "randomizeTrainingSetGPU [cublasSswap][mDeviceTeacherValues]" );
      }
    delete[] perm;
}

/* Runs on a matrix of input, with the help of CUDA and CUBLAS. GPU version */
Vector TwoLayersPerceptron::runBatch(  const Matrix& input  )
{
  // Convenient shortcut for the number of training samples
  const size_t N = input.size1();

  if( input.size2() != M )
    Reporting::Err(
      boost::str( boost::format("Bad dimension for input (must be %1%, got %2%).") % M % input.size2() ), __FILE__, __LINE__
      );

  /* Setup memory and copy data to GPU */
  setupDeviceMemoryRunOnGPU( input.data().begin(), input.size1() );

  /* Compute the linear output of the first (hidden) layer with cublasSgemm.
  * C=1.0*A*B+0.0*C, A is Nx(M+1), B is (M+1)xK, C is NxK. Each matrix is followed by its pitch
  * (leading dimension), i.e. number of rows (BLAS is column-major). Here, C = mDeviceHiddenOutput,
  * A = mDeviceInputValues, B = trans(mDeviceWeightsHidden) */
  cublasSgemm( 'n', 't', N, K, M+1, 1.0, mDeviceInputValues, N, mDeviceWeightsHidden, K, 0.0, mDeviceHiddenOutput, N );
  GPU::checkCuBLASError( "runBatch [cublasSgemm][Hidden Layer]" );

  /* Transform this linear output by applying the activation function, and add the bias in the (K+1)th column */
  dim3 threadsPerBlock( 64, 1 );            // 64 threads: would a 2D organisation be clearer ?
  const int numBlocks = (N*(K+1)+threadsPerBlock.x-1)/threadsPerBlock.x;    // Number of blocks in the grid
  GPU::Wrappers::kActivationFunction( numBlocks, threadsPerBlock, N*K, N, T, mDeviceHiddenOutput );

  /* Compute the linear output of the second (output) layer with cublasSgemm. */
  cublasSgemm( 'n', 't', N, 1, K+1, 1.0, mDeviceHiddenOutput, N, mDeviceWeightsOutput, 1, 0.0, mDeviceOutputs, N );
  GPU::checkCuBLASError( "runBatch [cublasSgemm][Output Layer]" );

  /* Sets the result we will return. */
  Vector output(N);
  cublasStatus status = cublasGetVector( N, sizeof(Real), mDeviceOutputs, 1, output.data().begin(), 1 );
  GPU::checkCuBLASError( "runBatch [cublasGetVector][mDeviceOutputs][output]", status );

  /* Free memory */
  cleanupDeviceMemoryRunOnGPU();

  return output;
}

/* Run on the samples, with the help of CUDA and CUBLAS (thus, can be invoked only in the context of training with GPU) */
void TwoLayersPerceptron::runForTrainingGPU()
{
  // Convenient shortcut for the number of training samples
  const size_t N_TOT = getTrainingSampleCount();
  const size_t N_LEARN = getLearningSampleCount();

//  std::cout << "===> [" << dum_iter << "] N_TOT, N_LEARN = " << N_TOT << "," << N_LEARN << "\n";

  /* Compute the linear output of the first (hidden) layer with cublasSgemm.
  * C=1.0*A*B+0.0*C, A is Nx(M+1), B is (M+1)xK, C is NxK. Each matrix is followed by its pitch
  * (leading dimension), i.e. number of rows (BLAS is column-major). Here, C = mDeviceHiddenOutput,
  * A = mDeviceInputValues, B = trans(mDeviceWeightsHidden) */
  cublasSgemm( 'n', 't', N_TOT, K, M+1, 1.0, mDeviceInputValues, N_TOT, mDeviceWeightsHidden, K, 0.0, mDeviceHiddenOutput, N_TOT );
//  GPU::checkCuBLASError( "train [cublasSgemm][Hidden Layer]" );
//    std::cout << "=======> [" << dum_iter << "] mDeviceInputValues\n";
//    Debug::printMatrixGPU(N_TOT, M+1, mDeviceInputValues);
//    std::cout << "=======> [" << dum_iter << "] mDeviceWeightsHidden\n";
//    Debug::printMatrixGPU(K, M+1, mDeviceWeightsHidden);

  /* Transform this linear output by applying the activation function, and add the bias in the (K+1)th column */
  dim3 threadsPerBlock( 64, 1 );            // 64 threads: would a 2D organisation be clearer ?
  const int numBlocks = (N_TOT*(K+1)+threadsPerBlock.x-1)/threadsPerBlock.x;    // Number of blocks in the grid
  GPU::Wrappers::kActivationFunction( numBlocks, threadsPerBlock, N_TOT*K, N_TOT, T, mDeviceHiddenOutput );
//  std::cout << "=======> [" << dum_iter << "] mDeviceHiddenOutput\n";
//  Debug::printMatrixGPU(N_TOT, K+1, mDeviceHiddenOutput);

  /* Compute the linear output of the second (output) layer with cublasSgemm. */
  cublasSgemm( 'n', 't', N_TOT, 1, K+1, 1.0, mDeviceHiddenOutput, N_TOT, mDeviceWeightsOutput, 1, 0.0, mDeviceOutputs, N_TOT );
  GPU::checkCuBLASError( "train [cublasSgemm][Output Layer]" );

//  std::cout << "=======> [" << dum_iter << "] mDeviceWeightsOutput\n";
//    Debug::printMatrixGPU(P, K+1, mDeviceWeightsOutput);
//    std::cout << "=======> [" << dum_iter << "] mDeviceOutputs\n";
//    Debug::printMatrixGPU(N_TOT, P, mDeviceOutputs);


  /* Compute the vector of errors, and the total squared error. */
  cublasScopy( N_TOT, mDeviceOutputs, 1, mDeviceErrors, 1 );
  GPU::checkCuBLASError( "train [cublasScopy][mDeviceOutputs][mDeviceErrors]" );
  cublasSaxpy( N_TOT, -1.0, mDeviceTeacherValues, 1, mDeviceErrors, 1 ); // error=output-teacher
  GPU::checkCuBLASError( "train [cublasSaxpy]" );

//  std::cout << "=======> [" << dum_iter << "] mDeviceErrors\n";
//    Debug::printMatrixGPU(1, N_TOT, mDeviceErrors);

  /* Compute total square error on the pure learning set and on the testing set.
   * (use cublasSdot since cublasSnrm2 would return the squared root) */
  mTotalErrorLearn = cublasSdot( N_LEARN, mDeviceErrors, 1, mDeviceErrors, 1 );
  GPU::checkCuBLASError( "train [cublasSdot][errorLearn]" );
  mTotalErrorTest = cublasSdot( (N_TOT-N_LEARN), mDeviceErrors+mTestSetStartIdx, 1, mDeviceErrors+mTestSetStartIdx, 1 ); // cublasSnrm2 would return the squared root
  GPU::checkCuBLASError( "train [cublasSdot][errorTest]" );

  /* Save restriction */
  for (int ii=0; ii < K+1; ii++)
  {
      cublasScopy( N_LEARN, mDeviceHiddenOutput + N_TOT*ii, 1,  mDeviceHiddenOutput_restr + N_LEARN*ii, 1 );
      GPU::checkCuBLASError( "train [cublasScopy][mDeviceHiddenOutput][mDeviceHiddenOutput_restr]" );
  }
//  std::cout << "=======> [" << dum_iter << "] mDeviceHiddenOutput_restr\n";
//  Debug::printMatrixGPU(N_LEARN, K+1, mDeviceHiddenOutput_restr);

//  std::cout << "       runForTrainingGPU "<< mTotalErrorLearn << " | " << mTotalErrorTest << std::endl;
//  dum_iter++;
}

bool TwoLayersPerceptron::normalizeSamplesGPU()
{
    using Preprocessing::normalize;
    using Preprocessing::kZeroOne;
    using Preprocessing::kMinusOneOne;

    Reporting::Warn( "Normalization should not be done on the GPU but solely in TANNModeler!", __FILE__, __LINE__ );

    if( !mSamplesNormalized )
    {
        // Convenient shortcut for the number of training samples
        const size_t N_TOT = getTrainingSampleCount();

        if( N_TOT <= 1 )
        {
            Reporting::Warn( "Not enough samples to apply normalization.", __FILE__, __LINE__ );
            return false;
        }
        // Finding the min and max has been done along the way already
        if( mMinSampleValue == mMaxSampleValue )
        {
            Reporting::Warn( "All values are identical: constant function to approximate?", __FILE__, __LINE__ );
        }
        // Do normalization on device
        const int threadsPerBlock = 64;                                 // Threads per block
        const int numBlocks = (N_TOT+threadsPerBlock-1)/threadsPerBlock;    // Number of blocks in the grid
        GPU::Wrappers::kNormalizeSamples( numBlocks, threadsPerBlock, N_TOT, T, mDeviceTeacherValues, mMinSampleValue, mMaxSampleValue );

        mSamplesNormalized = true;
    }

    return mSamplesNormalized;
}

void TwoLayersPerceptron::computeDeltaWeightsGPU(Real lambda)
{
  const bool useBlocked = true;     // Should we use a blocked LU decomposition?
  const size_t blockLUSize = 16;    // If so, which block size?

  /* For CUBLAS, the function cublasStrsv has a limitation (which is really unlikely not to be met,
  but still)... ensure it at compilation time. */
  if (NUMBER_PARAMETERS >= 4070)
    Reporting::Err("Too many params! (cublasStrsv_Has4070RestrictionOnMatrixDimension)", __FILE__, __LINE__ );
  //COMPILE_ASSERT( (NUMBER_PARAMETERS < 4070), cublasStrsv_Has4070RestrictionOnMatrixDimension );

  cublasScopy( NUMBER_PARAMETERS*NUMBER_PARAMETERS, mDeviceHessian, 1, mDeviceLMMatrix, 1 );
  GPU::checkCuBLASError( "computeDeltaWeightsGPU [cublasScopy][mDeviceHessian][mDeviceLMMatrix]" );
  cublasSaxpy( NUMBER_PARAMETERS, lambda, mDeviceIdentityRR, 1, mDeviceLMMatrix, NUMBER_PARAMETERS+1 );
  GPU::checkCuBLASError( "computeDeltaWeightsGPU [cublasScopy][mDeviceIdentityRR][mDeviceLMMatrix]" );
  // Now, "A = tJJ+lambda.I"

  cublasStatus status;          // Error reporting for the cuBLAS helper functions (which do not use cublasGetError())
  const size_t& R = NUMBER_PARAMETERS;

  // Host variables
  int hostPivot[ R ];
  for( size_t i = 0; i < R; ++i ) hostPivot[i] = i; // initialize the pivot as identity
  // Device variable: "Delta vector", i.e. R-dimensional vector of the changes to the weights and biases
  Real* deviceDelta = 0;
  status = cublasAlloc( R, sizeof(deviceDelta[0]), (void**)&deviceDelta );
  GPU::checkCuBLASError( "computeDeltaWeightsGPU [cublasAlloc][deviceDelta]", status );

  /* First, get the LU decomposition of A */
  try
  {
      static const Real epsilon = std::numeric_limits<Real>::min();
      /* We have initialized the pivot to be identity (i.e., P[i]=i, 0<=i<M), but it doesn't matter - as
      it is overwritten by the inplace algorithm, and involves O(R) operations for nothing. NB: P[M-1] is
      never used (a permutation in [[1,R]] only requires to specify the first R-1 exchanges). (#) */
      if( useBlocked )    // Faster
          LinearAlgebra::decomposeBlockedLU( R, R, R, mDeviceLMMatrix, hostPivot, blockLUSize, epsilon );
      else
          LinearAlgebra::decomposeLU( R, R, R, mDeviceLMMatrix, hostPivot, epsilon );
  }
  catch( const std::exception& e )
  {
      std::string fullError = "In function " + std::string(__FUNCTION__) + ": can't solve linear system (singular matrix).";
      fullError += "\n(";
      fullError += e.what();
      fullError += ")";
      Reporting::Err( fullError, __FILE__, __LINE__ ) ;
      exit(0);
  }

  /* Now, we have the LU (with pivoting) decomposition of A:  PA = LU
  We want to solve Ax=g <=> LUx=Pg. 3 steps are necessary:
      (1). Compute Pg
      (2). Solve Ly=Pg with cublasStrsv()
      (3). Solve Ux=y  with cublasStrsv() */
  // Step (1):
  cublasScopy( R, mDeviceGradient, 1, deviceDelta, 1 ); // Device-space "delta" is now the gradient g
  GPU::checkCuBLASError( "computeDeltaWeightsGPU [cublasScopy][deviceDelta]" );
  for( size_t i = 0; i < R-1; ++i )   // For the "-1", see above (#)
  {
      if( i != hostPivot[i] )
      {
          cublasSswap( 1, deviceDelta + hostPivot[i], 1, deviceDelta + i, 1 );
          GPU::checkCuBLASError( "computeDeltaWeightsGPU [cublasSswap][Pg]" );
      }
  } // Device-space "delta" is now Pg

  // Step (2):
  cublasStrsv( 'l', 'n', 'u', R, mDeviceLMMatrix, R, deviceDelta, 1 );
  GPU::checkCuBLASError( "computeDeltaWeightsGPU [cublasStrsv][1][Ly=Pg]" );


  // Step (3):
  cublasStrsv( 'u', 'n', 'n', R, mDeviceLMMatrix, R, deviceDelta, 1 );
  GPU::checkCuBLASError( "computeDeltaWeightsGPU [cublasStrsv][2][Ux=y]" );

  /** 4. Update the weights of the NN accordingly. **/
  // weights = weights - delta
  cublasSaxpy( K*(M+1), -1.0, deviceDelta, 1, mDeviceWeightsHidden, 1 );
  GPU::checkCuBLASError( "computeDeltaWeightsGPU [cublasSaxpy][mDeviceWeightsHidden][deviceDelta]" );
  cublasSaxpy( P*(K+1), -1.0, deviceDelta + K*(M+1), 1, mDeviceWeightsOutput, 1 ); // P=1 ...
  GPU::checkCuBLASError( "computeDeltaWeightsGPU [cublasSaxpy][mDeviceWeightsOutput][deviceDelta]" );

  // Cleanup
  status = cublasFree( deviceDelta );
  GPU::checkCuBLASError( "computeDeltaWeightsGPU [cublasFree][deviceDelta]", status );

  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
}

/* *********** Train (batch mode) on the pool (LM algorithm) - GPU (CUDA/CUBLAS) version *********** */

bool TwoLayersPerceptron::trainGPU( short& stopConditionMask, size_t& nbTotalIterations )
{
    // Convenient shortcut for the number of training samples
    const size_t N_TOT = getTrainingSampleCount();
    const size_t N_LEARN = getLearningSampleCount();

    /** Set up identity "matrix" **/
    {
      const int threadsPerBlock = 8;                                   // Threads per block
      const int numBlocks = (NUMBER_PARAMETERS+threadsPerBlock-1)/threadsPerBlock;    // Number of blocks in the grid
      GPU::Wrappers::kIdentityMatrix( numBlocks, threadsPerBlock, NUMBER_PARAMETERS, mDeviceIdentityRR );
    }

    /** Normalize the samples (or fail if we can't) **/
    // [ABN] normalization done in TANNModeler
    //if( !normalizeSamplesGPU() ) return false;

    /** Until stopping condition has been reached, iterate **/
    stopConditionMask = 0;
    nbTotalIterations = 0;
    size_t nbNoChangeIterations = 0;
    // Stopping conditions
    bool stopCondThreshold = false;
    bool stopCondMaxIterat = false;
    bool stopCondNoChangeIterat = false;
    #ifdef NN_DISPLAY_PROGRESSBAR
      // Allows us to display the overall progress, in terms of epochs
      boost::progress_display progressDisplayer( mMaxIterations );
    #endif
    mBestTotalErrorTest = std::numeric_limits<double>::max();
    mBestTotalErrorLearn = std::numeric_limits<double>::max();
    Real lambda = mLambda;

    while( !stopCondThreshold && !stopCondMaxIterat && !stopCondNoChangeIterat )
    {
//        std::cout << " ---- Iteration (" << nbTotalIterations << ")\n";
        /** 1. Present all input samples to the NN, compute the outputs and the error vector,
          and calculate the total error on both the learning set and the testing set.
        */
        runForTrainingGPU();
//        std::cout << "   After FIRST runForTrainingGPU: "  << mTotalErrorLearn << " | " << mTotalErrorTest << std::endl;

        /* Now, we have the outputs, errors and total squared error. It's time to compute the Jacobian,
         * only on the training set. */

        /** 2. Compute the Jacobian matrix J=(dE/dwi), NxR, with the usual BP method,
          row by row - each row correspond to a sample. **/
        {
          // Same method as in Backpropagation
          /// Reminder: mDeviceJacobian is a NxR matrix J=(dei/dwj) (R = NUMBER_PARAMETERS : weights and biases)
          /// mDeviceHiddenOutput is Nx(K+1), with the last column filled with ones
          /* For the output units:  */
          // For each 0<=k<K+1, dF/dw(p,k)=z_k (P=1 => p=0 + identity function for activation)
          cublasScopy( N_LEARN*(K+1), mDeviceHiddenOutput_restr, 1,  mDeviceJacobian + N_LEARN*K*(M+1), 1 );
          GPU::checkCuBLASError( "computeJacobian [cublasScopy][mDeviceHiddenOutput_restr][mDeviceJacobian]" );

//          std::cout << "=======> [" << dum_iter << "] mDeviceJacobian INIT\n";
//          Debug::printMatrixGPU(N_LEARN, NUMBER_PARAMETERS, mDeviceJacobian);

          GPU::Wrappers::kJacobianElement( M, K, N_LEARN, T, mDeviceHiddenOutput_restr, mDeviceWeightsOutput,
                                              mDeviceInputValues_restr, mDeviceJacobian );

//          std::cout << "=======> [" << dum_iter << "] mDeviceJacobian COMPUTED\n";
//          Debug::printMatrixGPU(N_LEARN, NUMBER_PARAMETERS, mDeviceJacobian);
        }
        /** Here, we have our Jacobian. Compute the Hessian approximation, based on it. **/
        //~ computeHessian();
        {
          // mDeviceHessian = 1.0*trans(mDeviceJacobian) * mDeviceJacobian + 0.0*mDeviceHessian,
          // stored in upper storage mode (since it is symmetric, we compute half the coeffs)
          //~ cublasSsyrk( 'u', 't', NUMBER_PARAMETERS, N, 1.0, mDeviceJacobian, N, 0.0, mDeviceHessian, NUMBER_PARAMETERS );
          cublasSgemm( 't', 'n', NUMBER_PARAMETERS, NUMBER_PARAMETERS, N_LEARN,
                       1.0, mDeviceJacobian, N_LEARN, mDeviceJacobian, N_LEARN,
                       0.0, mDeviceHessian, NUMBER_PARAMETERS );
          GPU::checkCuBLASError( "computeHessian [cublasSsyrk]" );
          /*  Nota: Using cublasSsyrk() was problematic, since half of the resulting matrix was
              garbage and couldn't be employed as it was. hence the choice of cublasSgemm(). */
        }
        /** And the gradient tJ.e (e is the N-dimensional error vector, e_n = output_n - teacher_n) **/
        //~ computeGradient();
        {
          cublasSgemv( 't', N_LEARN, NUMBER_PARAMETERS, 1.0, mDeviceJacobian, N_LEARN, mDeviceErrors, 1, 0.0, mDeviceGradient, 1 );
          GPU::checkCuBLASError( "computeGradient [cublasSgemv]" );

//          std::cout << "=======> [" << dum_iter << "] mDeviceGradient\n";
//          Debug::printMatrixGPU(1, NUMBER_PARAMETERS, mDeviceGradient);
        }

        Real currErrorLearn = mTotalErrorLearn+1.0;
        Real currErrorTest;
        /** Now, the trust region algorithm (Levenberg-Marquardt) itself
            Modify the trust region according to whether the change of weight was:
                    (i)  a good thing: then, expand the trust region
                    (ii) a bad thing: then, shrink the trust region, and revert to old weights
            The LM optimization is applied using the mTotalErrorLearn as a criteria of convergence.
            We just take a look at the mTotalErrorTest on the way to save the best iteration when the ANN
            is applied on the testing part of the samples.
        **/

        lambda *= LAMBDA_DECREASE_FACTOR;
        while( currErrorLearn > mTotalErrorLearn && lambda < LAMBDA_MAX_VALUE && nbNoChangeIterations < mMaxNoChangeIter )
        {
            lambda *= LAMBDA_INCREASE_FACTOR;
            try
            {
                /** 3. Compute \f$ \Delta(w) = ( J^T J + \lambda I )^{-1} J^T e \f$
                  where e is the N-dimensional error vector \f$(e_i)\f$, \f$ e_i = y_i-t_i \f$. **/
                /* Backup the previous weights */
                //~ backupTrainingWeights();
                {
                    cublasScopy( K*(M+1), mDeviceWeightsHidden, 1,  mDeviceWeightsBackup, 1 );
                    GPU::checkCuBLASError( "backupWeights [cublasScopy][mDeviceWeightsHidden][mDeviceWeightsBackup]" );
                    cublasScopy( P*(K+1), mDeviceWeightsOutput, 1,  mDeviceWeightsBackup + K*(M+1), 1 ); // P=1
                    GPU::checkCuBLASError( "backupWeights [cublasScopy][mDeviceWeightsOutput][mDeviceWeightsBackup]" );
                }
                /* And them, create and assign the new weights */
                computeDeltaWeightsGPU(lambda);

                /* Re-run on all inputs with the new weights to see if there is an improvement */
                currErrorLearn = mTotalErrorLearn;
                currErrorTest = mTotalErrorTest;
                runForTrainingGPU();
                std::swap( currErrorLearn, mTotalErrorLearn );
                std::swap( currErrorTest, mTotalErrorTest );

                // if the test metric improved save the weights (and the current errors) as the best of all candidate
                // This is different from the next 'if' block which is really part of the Leven-Marq optim.
                if (currErrorTest < mBestTotalErrorTest)
                {
//                    std::cout << "   Saving trainGPU iter[" << nbTotalIterations <<"]- " << currErrorLearn << " | " << currErrorTest << std::endl;
                    mBestTotalErrorTest = currErrorTest;
                    mBestTotalErrorLearn = currErrorLearn;
                    cublasScopy( K*(M+1), mDeviceWeightsHidden, 1,  mDeviceWeightsBest, 1 );
                    GPU::checkCuBLASError( "saveWeights [cublasScopy][mDeviceWeightsHidden][mDeviceWeightsBest]" );
                    cublasScopy( P*(K+1), mDeviceWeightsOutput, 1,  mDeviceWeightsBest + K*(M+1), 1 ); // P=1
                    GPU::checkCuBLASError( "saveWeights [cublasScopy][mDeviceWeightsOutput][mDeviceWeightsBest]" );
//                    std::cout <<  "     mDeviceWeightsBest\n";
//                    Debug::printMatrixGPU(1,1,mDeviceWeightsBest);
                }

                if( currErrorLearn > mTotalErrorLearn ) // Well, no improvement: restore previous weights
                { //~ restoreWeights();
                  cublasScopy( K*(M+1), mDeviceWeightsBackup, 1,  mDeviceWeightsHidden, 1 );
                  GPU::checkCuBLASError( "restoreWeights [cublasScopy][mDeviceWeightsBackup][mDeviceWeightsHidden]" );
                  cublasScopy( P*(K+1), mDeviceWeightsBackup + K*(M+1), 1, mDeviceWeightsOutput, 1 ); // P=1
                  GPU::checkCuBLASError( "restoreWeights [cublasScopy][mDeviceWeightsBackup][mDeviceWeightsOutput]" );
                }


                ++nbNoChangeIterations;   // One more iteration with no change (will be reset to 0 if there was)
            }
            catch( const Reporting::cublasException& e )
            {
                lambda = ( lambda == 1.0 ) ? 1000.0 : 1.0;
                Reporting::Warn( e.what(), __FILE__, __LINE__ );
                return false;
            }
            catch( const std::exception& e )
            {
                lambda = ( lambda == 1.0 ) ? 1000.0 : 1.0;
                Reporting::Warn( e.what(), __FILE__, __LINE__ );
                continue;
            }
        }
        lambda *= LAMBDA_DECREASE_FACTOR;

        if( currErrorLearn < mTotalErrorLearn ) nbNoChangeIterations = 0;   // Something actually improved.
        mTotalErrorLearn = currErrorLearn; // Save new shiny error
        ++nbTotalIterations;

        #ifdef NN_DISPLAY_PROGRESSBAR
          ++progressDisplayer;    // Update also the progress bar
        #endif

        // Stopping conditions?
        stopCondThreshold = ( mTotalErrorLearn <= (mPrecisionThreshold*N_LEARN) );
        stopCondMaxIterat = ( nbTotalIterations >= mMaxIterations );
        stopCondNoChangeIterat = ( nbNoChangeIterations >= mMaxNoChangeIter );
//        #ifdef NN_DISPLAY_INFO
//          std::cout << std::right << "\tStop flags:\t"
//                    << stopCondThreshold << " " << stopCondMaxIterat << " " << stopCondNoChangeIterat << std::endl;
//        #endif
    }

    // Retrieve the result of the training: namely, weights and biases
    // from the best iteration in terms of error on the testing part:
    //~ retrieveTrainedWeights();
//    {
//        cublasStatus status;
//        Real* hostMatrix;
//        hostMatrix = mOutputLayer.getWeights().data().begin(); // Shortcut
//        status = cublasGetMatrix( 1, K+1, sizeof(hostMatrix[0]), mDeviceWeightsOutput, 1, hostMatrix, 1 );
//        GPU::checkCuBLASError( "retrieveTrainedWeights [cublasGetMatrix][mDeviceWeightsOutput]", status );
//        hostMatrix = mHiddenLayer.getWeights().data().begin(); // Shortcut
//        status = cublasGetMatrix( K, M+1, sizeof(hostMatrix[0]), mDeviceWeightsHidden, K, hostMatrix, K );
//        GPU::checkCuBLASError( "retrieveTrainedWeights [cublasGetMatrix][mDeviceWeightsHidden]", status );
//    }

    {
        cublasScopy( K*(M+1), mDeviceWeightsBest, 1,  mDeviceWeightsHidden, 1 );
        GPU::checkCuBLASError( "restoreWeights [cublasScopy][mDeviceWeightsBackup][mDeviceWeightsHidden]" );
        cublasScopy( P*(K+1), mDeviceWeightsBest + K*(M+1), 1, mDeviceWeightsOutput, 1 ); // P=1
        GPU::checkCuBLASError( "restoreWeights [cublasScopy][mDeviceWeightsBackup][mDeviceWeightsOutput]" );
    }

    {
      cublasStatus status;
      Real* hostMatrix;
      hostMatrix = mOutputLayer.getWeights().data().begin(); // Shortcut
      status = cublasGetMatrix( 1, K+1, sizeof(hostMatrix[0]), mDeviceWeightsOutput, 1, hostMatrix, 1 );
      GPU::checkCuBLASError( "retrieveTrainedWeights [cublasGetMatrix][mDeviceWeightsOutput]", status );
      hostMatrix = mHiddenLayer.getWeights().data().begin(); // Shortcut
      status = cublasGetMatrix( K, M+1, sizeof(hostMatrix[0]), mDeviceWeightsHidden, K, hostMatrix, K );
      GPU::checkCuBLASError( "retrieveTrainedWeights [cublasGetMatrix][mDeviceWeightsHidden]", status );
    }

    // Set the stopping condition mask, to specify why it ended
    stopConditionMask  = static_cast<short>( stopCondThreshold );
    stopConditionMask += 2*static_cast<short>( stopCondMaxIterat );
    stopConditionMask += 4*static_cast<short>( stopCondNoChangeIterat );

    return true;
}
#endif // #ifdef CPU_PROCESSING_ONLY

}  // end namespace

