//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////

#include "nn_headers.h"
#include <iomanip>
#include <algorithm>
#include <fstream>
#include <map>
#include <sys/time.h>	// gettimeofday()
#include <boost/numeric/ublas/io.hpp>

namespace nn = NeuralNetworkLevenbergMarquardt;
namespace pp = nn::Preprocessing;
using std::cout;
using std::endl;

static float rosenbrock( const float x, const float y )
{
    float tmp = y-x*x;
        tmp *= tmp;
        tmp *= 10.0f;
    float result = 1.0f - x;
        result *= result;
        result += tmp;
    return result;
}

void nn::Testing::testWithXOR( const nn::Real precision, const size_t nbRuns, const size_t nbMaxIter )
{
    const size_t nbSamples = 4;        
    TwoLayersPerceptron<2,2,1> mXORApproxNN( 0.1, precision, nbMaxIter );        
    for( size_t n = 0; n < nbRuns; ++n )
    {
    
        mXORApproxNN.setTrainingSampleCount( nbSamples );
        
        std::vector<nn::Vector> xSamples, ySamples;
        xSamples.reserve( nbSamples );
        ySamples.reserve( nbSamples );
        nn::Vector x(2);
        nn::Vector y(1);
        
        x(0)=0; x(1)=0; y(0)=0;xSamples.push_back( x );ySamples.push_back( y );
        x(0)=0; x(1)=1; y(0)=1;xSamples.push_back( x );ySamples.push_back( y );
        x(0)=1; x(1)=0; y(0)=1;xSamples.push_back( x );ySamples.push_back( y );
        x(0)=1; x(1)=1; y(0)=0;xSamples.push_back( x );ySamples.push_back( y );
        
        cout << std::setprecision(3) << std::fixed;
        cout << "--- Training on 4 samples:" << endl;
        for( size_t i = 0; i < nbSamples; ++i )
        {
            cout << "XOR(" << xSamples[i](0) << "," << xSamples[i](1) << ") = " << ySamples[i](0) << endl;
            mXORApproxNN.addTrainingSample( xSamples[i], ySamples[i] );
        }
        cout << std::resetiosflags( std::ios::showpoint | std::ios::floatfield | std::ios::adjustfield) << std::setprecision(6);
        
        /* Training */
        mXORApproxNN.train();
        
        /* Test */
        cout << "--- Testing" << endl;
        x(0) = 1.0; x(1) = 1.0;        
        mXORApproxNN.run(x);
        cout << "Input:\t\t(" << x(0) << "," << x(1) << ")\t|\t" << endl;
        cout << "Network output:\t\t" << mXORApproxNN.getOutput() << "\t|\t";
        cout << "Expected output:\t" << pp::normalize<pp::kZeroOne>( static_cast<int>(x(0))^static_cast<int>(x(1)), 0, 1 ) << endl;
        cout << "Network value:\t\t" << mXORApproxNN.getDenormalizedOutput() << "\t|\t";
        cout << "Expected value:\t\t" << ( static_cast<int>(x(0))^static_cast<int>(x(1)) ) << endl;
        cout << "---" << endl;
        x(0) = 1.0; x(1) = 0;        
        mXORApproxNN.run(x);
        cout << "Input:\t\t(" << x(0) << "," << x(1) << ")\t|\t" << endl;
        cout << "Network output:\t\t" << mXORApproxNN.getOutput() << "\t|\t";
        cout << "Expected output:\t" << pp::normalize<pp::kZeroOne>( static_cast<int>(x(0))^static_cast<int>(x(1)), 0, 1 ) << endl;
        cout << "Network value:\t\t" << mXORApproxNN.getDenormalizedOutput() << "\t|\t";
        cout << "Expected value:\t\t" << ( static_cast<int>(x(0))^static_cast<int>(x(1)) ) << endl;
        cout << "---" << endl;
        x(0) = 0.0; x(1) = 1.0;        
        mXORApproxNN.run(x);
        cout << "Input:\t\t(" << x(0) << "," << x(1) << ")\t|\t" << endl;
        cout << "Network output:\t\t" << mXORApproxNN.getOutput() << "\t|\t";
        cout << "Expected output:\t" << pp::normalize<pp::kZeroOne>( static_cast<int>(x(0))^static_cast<int>(x(1)), 0, 1 ) << endl;
        cout << "Network value:\t\t" << mXORApproxNN.getDenormalizedOutput() << "\t|\t";
        cout << "Expected value:\t\t" << ( static_cast<int>(x(0))^static_cast<int>(x(1)) ) << endl;
        
		std::cout << "--- Run #" << n << " completed. ---" << std::endl;  
		// New trial to come
    	mXORApproxNN.reset();
    }
}

void nn::Testing::testWithRosenbrock( const size_t nbSamples, const nn::Real precision, const size_t nbRuns, const size_t nbMaxIter )
{
    size_t sqrtSamples = static_cast<size_t>( std::ceil( std::sqrt( nbSamples ) ) );
    size_t nbSamplesActual = sqrtSamples*sqrtSamples;

    TwoLayersPerceptron<2,5,1, nn::kTanh> mRosenbrockApproxNN( 0.1, precision, nbMaxIter );
    mRosenbrockApproxNN.setTrainingSampleCount( nbSamplesActual );
    
    // We will, after training, run the NN on a lot of random values (handy to profile: see where the running time goes)
    const size_t nbRandomInputs = 100000;
    nn::Matrix brInputRandom( nbRandomInputs, 2 );
    for( size_t i = 0; i < brInputRandom.size1(); ++i )
    {
      brInputRandom(i,0) = nn::Random::StandardGaussianValue()*2.0-1.0;	// Not always in [-1,1], but often.
      brInputRandom(i,1) = nn::Random::StandardGaussianValue()*2.0-1.0;
    }
    
    for( size_t n = 0; n < nbRuns; ++n )
    {
    
        std::vector<nn::Vector> xSamples, ySamples;
        xSamples.reserve( nbSamplesActual );
        ySamples.reserve( nbSamplesActual );
        nn::Vector x(2);
        nn::Vector y(1);

        for( size_t i = 0; i < sqrtSamples; ++i )
        {
            for( size_t j = 0; j < sqrtSamples; ++j )
            {
                x(0) = i/(sqrtSamples-1.0)*2.0-1.0; x(1) = j/(sqrtSamples-1.0)*2.0-1.0; // in [-1,1]
                y(0) = rosenbrock(x(0),x(1));
                xSamples.push_back( x );
                ySamples.push_back( y );
            }
        }
        for( size_t i = 0; i < nbSamplesActual; ++i )
        {
            mRosenbrockApproxNN.addTrainingSample( xSamples[i], ySamples[i] );
        }
        
        /* Training */
        cout << "--- Training on " << nbSamplesActual << " samples." << endl;
        mRosenbrockApproxNN.train();
        
        /* Test */
        cout << "--- Testing" << std::fixed << endl;
        const float minSamples = mRosenbrockApproxNN.getMinTrainingSampleValue();
        const float maxSamples = mRosenbrockApproxNN.getMaxTrainingSampleValue();
        x(0) = 1.0;
        x(1) = 1.0;
        mRosenbrockApproxNN.run(x);
        cout << "Input:\t\t(" << x(0) << "," << x(1) << ")\t|\t" << endl;
        cout << "Network output:\t\t" << mRosenbrockApproxNN.getOutput() << "\t|\t";
        cout << "Expected output:\t" << pp::normalize<pp::kMinusOneOne>( rosenbrock( x(0), x(1) ), minSamples, maxSamples ) << endl;
        cout << "Network value:\t\t" << mRosenbrockApproxNN.getDenormalizedOutput() << "\t|\t";
        cout << "Expected value:\t\t" << rosenbrock( x(0), x(1) ) << endl;
        cout << "---" << endl;
        x(0) = -0.124;
        x(1) = 0.84;        
        mRosenbrockApproxNN.run(x);
        cout << "Input:\t\t(" << x(0) << "," << x(1) << ")\t|\t" << endl;
        cout << "Network output:\t\t" << mRosenbrockApproxNN.getOutput() << "\t|\t";
        cout << "Expected output:\t" << pp::normalize<pp::kMinusOneOne>( rosenbrock( x(0), x(1) ), minSamples, maxSamples ) << endl;
        cout << "Network value:\t\t" << mRosenbrockApproxNN.getDenormalizedOutput() << "\t|\t";
        cout << "Expected value:\t\t" << rosenbrock( x(0), x(1) ) << endl;
        cout << "---" << endl;
        x(0) = 0.0;
        x(1) = -1.0;        
        mRosenbrockApproxNN.run(x);
        cout << "Input:\t\t(" << x(0) << "," << x(1) << ")\t|\t" << endl;
        cout << "Network output:\t\t" << mRosenbrockApproxNN.getOutput() << "\t|\t";
        cout << "Expected output:\t" << pp::normalize<pp::kMinusOneOne>( rosenbrock( x(0), x(1) ), minSamples, maxSamples ) << endl;
        cout << "Network value:\t\t" << mRosenbrockApproxNN.getDenormalizedOutput() << "\t|\t";
        cout << "Expected value:\t\t" << rosenbrock( x(0), x(1) ) << endl;
        
        cout << "Trying the batch run:" << endl;
        // On the same values as before
        nn::Matrix brInput( 3, 2 );
        brInput(0,0) = 1.0;
        brInput(0,1) = 1.0;
        brInput(1,0) = -0.124;
        brInput(1,1) = 0.84;
        brInput(2,0) = 0.0;
        brInput(2,1) = -1.0;
        nn::Debug::printVector( mRosenbrockApproxNN.runBatch( brInput ) );
        // On a lot of random values (discard the actual result)
        mRosenbrockApproxNN.runBatch( brInputRandom );
        
        // New trial to come
        mRosenbrockApproxNN.reset();            
	}
}

void nn::Testing::testWithRefFileRosenbrock( const std::string& strFileIn, const std::string& strFileOut, const nn::Real precision, const size_t nbMaxIter )
{
    const size_t sqrtTrainingSamples = 64;
    const size_t nbTrainingSamples = sqrtTrainingSamples*sqrtTrainingSamples;   // 4096 training samples 

    TwoLayersPerceptron<2,5,1, nn::kTanh> mLVMultiPerceptron( 0.1, precision, nbMaxIter );
    mLVMultiPerceptron.setTrainingSampleCount( nbTrainingSamples );
    
    nn::Vector x(2);
    nn::Vector y(1);
    // Testing samples
    std::vector<nn::Vector> ySamples;
    nn::Matrix xSamples;    // NxM matrix for runBatch()
    
    /* Generate the traning samples */       
    for( size_t i = 0; i < sqrtTrainingSamples; ++i )
    {
        for( size_t j = 0; j < sqrtTrainingSamples; ++j )
        {
            x(0) = i/(sqrtTrainingSamples-1.0)*2.0-1.0; x(1) = j/(sqrtTrainingSamples-1.0)*2.0-1.0; // in [-1,1]
            y(0) = rosenbrock(x(0),x(1));
            mLVMultiPerceptron.addTrainingSample( x, y );
        }
    }
    
    /* Retrieve the testing samples */
    // Get number of testing samples (number of lines - 5 for the header)
    const size_t headerSize = 5;
    std::ifstream fileIn( strFileIn.c_str() );
    const size_t nbTestingSamples = std::count( std::istreambuf_iterator<char>(fileIn), std::istreambuf_iterator<char>(), '\n' ) + 1 - headerSize;
    xSamples.resize( nbTestingSamples, 2 );
    ySamples.reserve( nbTestingSamples );
    
    // Read file, restarting at the beginning
    fileIn.clear( );
    fileIn.seekg( 0, std::ios::beg );
    size_t nbTestingSamplesActual = 0;
    std::string line;
    for( size_t i = 0; i < headerSize; ++i ) std::getline( fileIn, line );  // Discard header
    while( std::getline( fileIn, line ) )
    {
        std::stringstream linestream(line);

        // Read the integers using the operator >>
        linestream >> x(0) >> x(1) >> y(0);
        xSamples( nbTestingSamplesActual, 0 ) = x(0);
        xSamples( nbTestingSamplesActual, 1 ) = x(1);
        ySamples.push_back( y );
        ++nbTestingSamplesActual;
    }
    xSamples.resize( nbTestingSamplesActual, 2, true ); // In case we had too many
    
    /* Training */
    cout << "--- Training on " << nbTrainingSamples << " samples." << endl;
    mLVMultiPerceptron.train();
        
    /* Test */
    std::ofstream fileOut( strFileOut.c_str() );
    if( fileOut.is_open() )
    {
        cout << "--- Testing on " << nbTestingSamplesActual << " samples." << endl;
        fileOut << "#COLUMN_NAMES: x| y| z| zhat" << std::endl;
        nn::Vector result = mLVMultiPerceptron.runBatch( xSamples );
        for( size_t i = 0; i < nbTestingSamplesActual; ++i )
        {
          nn::Real output = mLVMultiPerceptron.denormalizeOutput( result(i) );
          fileOut << xSamples(i,0) << "\t" << xSamples(i,1) << "\t" << ySamples[i](0) << "\t" << output << std::endl;
        }
    }
    else
    {
        Reporting::Err( "Cannot open the file \""+ strFileOut +"\".", __FILE__, __LINE__ );
    }
}

void nn::Testing::testCaseANDRA( const std::string& dir, const std::string& strFileIn, const std::string& outputPrefix,
        					const Real precision, const size_t nbMaxIter, const size_t nbTrainings, const size_t nbBatchRuns )
{
    /** Output variable "s4t" : R^17->R, 16 hidden nodes, 1400 training samples, tanh **/
    const size_t nbTrialsPerTraining = 1;
    const size_t nbTrainingSamples = 1400;
    const size_t dimInputSpace = 17;
    const size_t headerSize = 5; // 5 lines of header
    TwoLayersPerceptron< dimInputSpace,16,1, nn::kTanh > mLVMultiPerceptron( 0.1, precision, nbMaxIter );
    mLVMultiPerceptron.setTrainingSampleCount( nbTrainingSamples );
    
    // Temporary storage vectors
    nn::Vector x(dimInputSpace), y(1);
    
    // Output header
    cout << "#COLUMN_NAMES: isTraining| n| t| e" << endl;
    cout << endl;
    
	/************************************
    			Training phase
    ************************************/
   
    /* Retrieve the training samples */
    // Read file
    std::ifstream fileIn( (dir+strFileIn).c_str() );
    size_t nbTrainingSamplesActual = 0;
    std::string line;
    for( size_t i = 0; i < headerSize; ++i ) std::getline( fileIn, line );  // Discard header
    while( std::getline( fileIn, line ) && nbTrainingSamplesActual < nbTrainingSamples )
    {
        std::stringstream linestream(line);
        std::vector<nn::Real> tmp(38);
        std::copy( std::istream_iterator<nn::Real>(linestream), std::istream_iterator<nn::Real>(), tmp.begin() );
        /* We only want some of them (17 inputs, and the normalized ones; + one output).
        Amongst:
        cKd| bDe| bK| bWacc| bKd| fzDe| fzK| fzWacc| mfzDe| mfzK| mfzWacc| uaDe| uaKv| uaKh| uaWacc| uaKd| uaH| s1| s2| s3| s4| s5| cKdLn| bDeLn| bKLn| fzKLn| mfzKLn| uaKvLn| uaKhLn| s3Ln| s5Ln| uaKdLn| s1t| s2t| s3t| s4t| s5t| andra__n__iter__
        */
        x( 0 ) = tmp[22]; // cKdLn
        x( 1 ) = tmp[23]; // bDeLn
        x( 2 ) = tmp[24]; // bKdLn
        x( 3 ) = tmp[3];  // bWacc
        x( 4 ) = tmp[4];  // bKd
        x( 5 ) = tmp[5];  // fzDe
        x( 6 ) = tmp[25]; // fzKLn
        x( 7 ) = tmp[7];  // fzWacc
        x( 8 ) = tmp[8];  // mfzDe
        x( 9 ) = tmp[26]; // mfzKLn
        x( 10 ) = tmp[10]; // mfzWacc
        x( 11 ) = tmp[11]; // uaDe
        x( 12 ) = tmp[27]; // uaKvLn
        x( 13 ) = tmp[28]; // uaKhLn
        x( 14 ) = tmp[14]; // uaWacc
        x( 15 ) = tmp[31]; // uaKdLn
        x( 16 ) = tmp[16]; // uaH
        y(0) = tmp[35];     // s4t
 //~       y(0) = tmp[36];     // s5t
 
        mLVMultiPerceptron.addTrainingSample( x, y );
        ++nbTrainingSamplesActual;
    }
    fileIn.close();
    
    /* Training: several times, to have reliable statistics */
    for( size_t k = 0; k < nbTrainings; ++k )
	{
		mLVMultiPerceptron.initialize(); // reset weights  
        mLVMultiPerceptron.train( nbTrialsPerTraining );   	// re-train     
        cout	<< true << "\t"
        		<< nbTrainingSamplesActual << "\t"
        		<< mLVMultiPerceptron.getLastTrainingTime()
        		<< "\t" << mLVMultiPerceptron.getLastTrainingError() << std::endl;
	}
    
    
    /************************************
    			Testing phase
    ************************************/
    
    /* The different testing databases */
    std::map<size_t, std::string> testingFNames;
    testingFNames[1500]  = "_test_final_1500_pretraitment.dat";
    testingFNames[5000]  = "_test_final_5000_pretraitment.dat";
    testingFNames[10000] = "_test_final_10000_pretraitment.dat";
    testingFNames[15000] = "_test_final_15000_pretraitment.dat";
    testingFNames[20000] = "_test_final_20000_pretraitment.dat";
    testingFNames[50000] = "_test_final_50000_pretraitment.dat";
    // For each of them
    for( std::map<size_t, std::string>::const_iterator cit = testingFNames.begin(); cit != testingFNames.end(); ++cit )
    {
        // Testing samples
        std::vector<nn::Vector> ySamples;
        nn::Matrix xSamples( cit->first, dimInputSpace );    // NxM matrix for runBatch()
    
        // Read file
        fileIn.open( (dir+(cit->second)).c_str() );
        size_t nbTestingSamplesActual = 0;
        std::string line;
        for( size_t i = 0; i < headerSize; ++i ) std::getline( fileIn, line );  // Discard header
        while( std::getline( fileIn, line ) )
        {
            std::stringstream linestream(line);
            std::vector<nn::Real> tmp(38);
            std::copy( std::istream_iterator<nn::Real>(linestream), std::istream_iterator<nn::Real>(), tmp.begin() );
            /* We only want some of them (17 inputs, and the normalized ones; + one output).
            Amongst:
            cKd| bDe| bK| bWacc| bKd| fzDe| fzK| fzWacc| mfzDe| mfzK| mfzWacc| uaDe| uaKv| uaKh| uaWacc| uaKd| uaH| s1| s2| s3| s4| s5| cKdLn| bDeLn| bKLn| fzKLn| mfzKLn| uaKvLn| uaKhLn| s3Ln| s5Ln| uaKdLn| s1t| s2t| s3t| s4t| s5t| andra__n__iter__
            */
            x( 0 ) = tmp[22]; // cKdLn
            x( 1 ) = tmp[23]; // bDeLn
            x( 2 ) = tmp[24]; // bKdLn
            x( 3 ) = tmp[3];  // bWacc
            x( 4 ) = tmp[4];  // bKd
            x( 5 ) = tmp[5];  // fzDe
            x( 6 ) = tmp[25]; // fzKLn
            x( 7 ) = tmp[7];  // fzWacc
            x( 8 ) = tmp[8];  // mfzDe
            x( 9 ) = tmp[26]; // mfzKLn
            x( 10 ) = tmp[10]; // mfzWacc
            x( 11 ) = tmp[11]; // uaDe
            x( 12 ) = tmp[27]; // uaKvLn
            x( 13 ) = tmp[28]; // uaKhLn
            x( 14 ) = tmp[14]; // uaWacc
            x( 15 ) = tmp[31]; // uaKdLn
            x( 16 ) = tmp[16]; // uaH
            y(0) = tmp[35];     // s4t
     //~       y(0) = tmp[36];     // s5t
     		
     		// Save the y reference value
     		ySamples.push_back( y );     
            // Access the current row of xSamples, and set it to the input value
            MatrixRow matRow( xSamples, nbTestingSamplesActual );
            matRow.assign( x );

            ++nbTestingSamplesActual;
        }
        xSamples.resize( nbTestingSamplesActual, dimInputSpace, true ); // In case we had too many
        
            
        /* Test */
        std::ofstream fileOut( (dir+outputPrefix+cit->second).c_str() );
        if( fileOut.is_open() )
        {
//            cout << "# Testing on " << nbTestingSamplesActual << " samples." << endl;
            fileOut << "#COLUMN_NAMES: x| y| z| zhat" << std::endl;
            
            // Do several batch runs, to have accurate estimations of the time taken
            nn::Vector result;
            for( size_t k = 0; k < nbBatchRuns; ++k )
            {
                // Measure total time taken by the batch running, and outputs it to console         
                timeval t1, t2;
    			double elapsedTime;
    			gettimeofday( &t1, NULL ); // Start timer

    			result = mLVMultiPerceptron.runBatch( xSamples );

    			gettimeofday( &t2, NULL ); // Stop timer
                // Compute elapsed time in milliseconds
                elapsedTime = (t2.tv_sec - t1.tv_sec) * 1000.0;      // seconds to ms
                elapsedTime += (t2.tv_usec - t1.tv_usec) / 1000.0;   // microseconds to ms
                
                cout << false << "\t" << nbTestingSamplesActual << "\t" << elapsedTime << "\t" << 0.0 << std::endl;
            }
            
            for( size_t i = 0; i < nbTestingSamplesActual; ++i )
            {
              nn::Real output = mLVMultiPerceptron.denormalizeOutput( result(i) );
              fileOut << xSamples(i,0) << "\t" << xSamples(i,1) << "\t" << ySamples[i](0) << "\t" << output << std::endl;
            }
        }
        else
        {
            Reporting::Err( "Cannot open the file \""+ (dir+outputPrefix+cit->second) +"\".", __FILE__, __LINE__ );
        }
        fileIn.close();
    }
}
