//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   nisp_pc.cpp
 \author Jean-Marc Martinez <jean-marc.martinez@cea.fr>
 \date   2011/04/07 14:33:00

 \brief  Class of polynomial chaos
 */
#include <iostream>
#include <sstream>
#include <fstream>
#include <string>
#include <vector>
#include <random>
#include <cmath>

#include "nisp_pc.h"
#include "nisp_util.h"

void dindexx(unsigned long, double *, unsigned long *);
void dindex(int n1, double *a, int *index, int down);

int CalculP(int, int);
int Puissance2(int);
void ReadTarget2(double **target, int np, int ny, char *fichier);

#define HERMITE  1
#define LEGENDRE 2
#define LAGUERRE 3
#define JACOBI   4

void hermite(double *phi, double x, int no);
void legendre(double *phi, double x, int no);
void laguerre(double *phi, double x, int no);
void jacobi(double *phi, double x, double alpha, double beta, int no);

// ===========
// Methodes PolynomialChaos
// ===========

PolynomialChaos::PolynomialChaos(SetRandomVariable *gva, int nny)
{
    int i;
    if (gva->nx < 1)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg
                << "Nisp(PolynomialChaos::PolynomialChaos) : size of set random variables is < 1"
                << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    nx = gva->nx;
    for (i = 0; i < nx; i++)
    {
        if (gva->va[i]->type == "Normale")
        {
            typol.push_back("Hermite");
            funpol.push_back(HERMITE);
        }
        else if (gva->va[i]->type == "Uniforme")
        {
            typol.push_back("Legendre");
            funpol.push_back(LEGENDRE);
        }
        else if (gva->va[i]->type == "Exponentielle")
        {
            typol.push_back("Laguerre");
            funpol.push_back(LAGUERRE);
        }
        else if (gva->va[i]->type == "Beta")
        {
            typol.push_back("Jacobi");
            funpol.push_back(JACOBI);
        }
        else
        {
            std::ostringstream msg;
            msg << "NISP - ERROR" << std::endl;
            msg << "Nisp(PolynomialChaos::PolynomialChaos) : law  "
                    << gva->va[i]->type << " is not available" << std::endl;
            NispOutPutMsg(msg.str(), 1);
            return;
        }
    }
    // fonction de repartition, quantiles, ...
    gpx = new SetRandomVariable();
    for (i = 0; i < nx; i++) {
        if (gva->va[i]->type == "Beta") {
	  gpx->AddRandomVariable(new RandomVariable(gva->va[i]->type,gva->va[i]->a,gva->va[i]->b));
	}
	else {
          gpx->AddRandomVariable(new RandomVariable(gva->va[i]->type));
	}
    }
    std::random_device device;
    gpx->set_seed(device());
    no = 0; // degre
    p = 0;  // nombre de coefficients (p+1)
    ny = nny;
    x = dvector(nx + 1);
    y = dvector(ny + 1);
    eqmLoo   = dvector(ny + 1);
    moyenne  = dvector(ny + 1);
    variance = dvector(ny + 1);
    groupe = ivector(nx + 1);
    ni = Puissance2(nx) - 1; // le nombre d'indices (ni=2**nx-1)
    indices = dmatrix(ny + 1, ni + 1);
    indices_globaux = dmatrix(ny + 1, nx + 1);
    indices_setvar  = imatrix(ni + 1, nx + 1);
    np = 0;
    typecomputation = 0;
}

PolynomialChaos::~PolynomialChaos()
{
    FreeMemoryTarget();

    free_dvector(x);
    free_dvector(y);
    free_dvector(eqmLoo);
    free_dvector(moyenne);
    free_dvector(variance);
    free_ivector(groupe);
    free_dmatrix(indices, ny + 1);
    free_dmatrix(indices_globaux, ny + 1);
    free_imatrix(indices_setvar, ni + 1);

    if (no)
    {
        free_dmatrix(phi, nx + 1);
        free_dvector(psi);
        free_imatrix(indmul, p + 1);
        free_dmatrix(beta, ny + 1);
    }

    delete gpx;
    gpx = NULL;
}

void PolynomialChaos::Realisation()
{
    for (int i = 1; i <= nx; i++)
        x[i] = gpx->va[i - 1]->GetValue();
    ComputeOutput(x);
}

// JMM 08/04/2014
void PolynomialChaos::GetLog()
{
    std::ostringstream msg;
    msg << "************************************" << std::endl;
    msg << "Nisp(PolynomialChaos::GetLog) for PC" << std::endl;
    for (int i = 0; i < nx; i++) {
      if( typol[i] == "Jacobi") {
	msg << typol[i] << " " << gpx->va[i]->a << " " << gpx->va[i]->b << std::endl;
      }
      else {
        msg << typol[i] << std::endl;
      }
    }
    msg << "************************************" << std::endl;
    NispOutPutMsg(msg.str(), 0);
}
void PolynomialChaos::SetDimensionOutput(int nny)
{
    if (nny < 1)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::SetDimensionOutput) : number of output "
                << nny << " < 1" << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    free_dvector(y);
    free_dvector(eqmLoo);
    free_dvector(moyenne);
    free_dvector(variance);
    free_dmatrix(indices, ny + 1);
    free_dmatrix(indices_globaux, ny + 1);
    free_imatrix(indices_setvar, ni + 1);

    ny = nny;
    y = dvector(ny + 1);
    eqmLoo = dvector(ny + 1);
    moyenne = dvector(ny + 1);
    variance = dvector(ny + 1);
    indices = dmatrix(ny + 1, ni + 1);
    indices_globaux = dmatrix(ny + 1, nx + 1);
    indices_setvar = imatrix(ni + 1, nx + 1);
}

PolynomialChaos::PolynomialChaos(char *fichier)
{
    std::ifstream entree(fichier, std::ios::in);
    if (!entree)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg
                << "Nisp(PolynomialChaos::PolynomialChaos) : problem when open file "
                << fichier << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    std::string str, type;
    double a, b;
    int nnx, nno, pp, nny;
    entree >> str >> nnx;
    nx = nnx;
    int i, j, k;
    gpx = new SetRandomVariable(); // fonction de r�partition, quantiles, ...
    for (i = 1; i <= nnx; i++)
    {
        entree >> type;
        if (type == "Hermite")
        {
            typol.push_back("Hermite");
            funpol.push_back(HERMITE);
            gpx->AddRandomVariable(new RandomVariable("Normale"));
        }
        else if (type == "Legendre")
        {
            typol.push_back("Legendre");
            funpol.push_back(LEGENDRE);
            gpx->AddRandomVariable(new RandomVariable("Uniforme"));
        }
        else if (type == "Laguerre")
        {
            typol.push_back("Laguerre");
            funpol.push_back(LAGUERRE);
            gpx->AddRandomVariable(new RandomVariable("Exponentielle"));
        }
        else if (type == "Jacobi")
	{
	    entree >> a >> b;
	    typol.push_back("Jacobi");
            funpol.push_back(JACOBI);
            gpx->AddRandomVariable(new RandomVariable("Beta",a , b));
	}
        else
        {
            std::ostringstream msg;
            msg << "NISP - ERROR" << std::endl;
            msg << "Nisp(PolynomialChaos::PolynomialChaos) : law  " << type
                    << " is not available" << std::endl;
            NispOutPutMsg(msg.str(), 1);
            return;
        }
    }
    std::random_device device;
    gpx->set_seed(device());

    entree >> str >> nno;
    no = nno;
    entree >> str >> pp;
    p = pp;
    entree >> str >> nny;
    ny = nny;

    x = dvector(nx + 1);
    y = dvector(ny + 1);
    eqmLoo   = dvector(ny + 1);
    moyenne  = dvector(ny + 1);
    variance = dvector(ny + 1);
    groupe = ivector(nx + 1);
    ni = Puissance2(nx) - 1; // le nombre d'indices (ni=2**nx-1)
    indices = dmatrix(ny + 1, ni + 1);
    indices_globaux = dmatrix(ny + 1, nx + 1);
    indices_setvar  = imatrix(ni + 1, nx + 1);
    np = 0;

    phi = dmatrix(nx + 1, no + 1);
    psi = dvector(p + 1);
    indmul = imatrix(p + 1, nx + 1);
    beta = dmatrix(ny + 1, p + 1);
    // Indices multiples
    int IndiceMultipleEval(int nnx, int noi, int nof, int **indmul);
    IndiceMultipleEval(nx, 0, no, indmul);
    // Lecture des coefficients
    for (j = 1; j <= ny; j++)
    {
        entree >> str;
        for (k = 0; k <= p; k++)
            entree >> beta[j][k];
    }
    SetAnova();
    typecomputation = 0;
    entree.close();
}

// Utilitaire pour le calcul des esperances et variances conditionnelles
// en optimisation robuste
// *pc est le polynome construit a partir du polynome global regroupant
// variables de conception et variables aleatoires
// *ind est l'indice multiple du polynome global
// *groupe designe le groupe des variables aleatoires (complementaires
// des variables de conception
// retourne la valeur liant le multi-indice du polynome global au multi-indice
// du polynome reduit
// Remarque : groupe[0,1,....] on commence a 0
int pcond_rank_indice(int *ind, PolynomialChaos * pc, int * groupe)
{
    int i, k, flag;
    for (k = 0; k <= pc->p; k++)
    {
        flag = 1;
        for (i = 1; i <= pc->nx; i++)
        {
            if (ind[groupe[i - 1]] != pc->indmul[k][i])
                flag = 0;
        }
        if (flag)
            return (k);
    }
    if (k == pc->p + 1)
    {
        // echec
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(pcond_rank_indice) : error on the rank = " << k << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (-1);
    }
    return (-1);
}

PolynomialChaos::PolynomialChaos(PolynomialChaos *pc, int *var_opt, int nopt)
{
    if ((pc->nx <= nopt) || (nopt < 1))
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg
                << "Nisp(PolynomialChaos::PolynomialChaos) : error on the number of optimization variables = "
                << nopt << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }

    // On definit le groupe var_sto complementaire au groupe var_opt
    int i, j, k, flag, nsto, *var_sto;
    nsto = pc->nx - nopt;
    var_sto = ivector(nsto);
    for (k = 0, i = 1; i <= pc->nx; i++)
    {
        flag = 1;
        for (j = 0; j < nopt; j++)
            if (var_opt[j] == i)
                flag = 0;
        if (flag)
            var_sto[k++] = i;
    }

    for (i = 0; i < nsto; i++)
    {
        if ((var_sto[i] < 1) || (var_sto[i] > pc->nx))
        {
            std::ostringstream msg;
            msg << "NISP - ERROR" << std::endl;
            msg
                    << "Nisp(PolynomialChaos::PolynomialChaos) : error on the set of variables "
                    << std::endl;
            NispOutPutMsg(msg.str(), 1);
            return;
        }
    }
    gpx = new SetRandomVariable(); // fonction de r�partition, quantiles, ...
    for (i = 0; i < nsto; i++)
    {
        if (pc->typol[var_sto[i] - 1] == "Hermite")
        {
            typol.push_back("Hermite");
            funpol.push_back(HERMITE);
            gpx->AddRandomVariable(new RandomVariable("Normale"));
        }
        else if (pc->typol[var_sto[i] - 1] == "Legendre")
        {
            typol.push_back("Legendre");
            funpol.push_back(LEGENDRE);
            gpx->AddRandomVariable(new RandomVariable("Uniforme"));
        }
        else if (pc->typol[var_sto[i] - 1] == "Laguerre")
        {
            typol.push_back("Laguerre");
            funpol.push_back(LAGUERRE);
            gpx->AddRandomVariable(new RandomVariable("Exponentielle"));
        }
        else
        {
            std::ostringstream msg;
            msg << "NISP - ERROR" << std::endl;
            msg << "Nisp(PolynomialChaos::PolynomialChaos) : law  "
                    << pc->typol[var_sto[i] - 1] << " is not available" << std::endl;
            NispOutPutMsg(msg.str(), 1);
            return;
        }
    }
    std::random_device device;
    gpx->set_seed(device());

    nx = nsto;
    no = 0; // degre
    p = 0;  // nombre de coefficients (p+1)
    ny = pc->ny;
    x = dvector(nx + 1);
    y = dvector(ny + 1);
    eqmLoo   = dvector(ny + 1);
    moyenne  = dvector(ny + 1);
    variance = dvector(ny + 1);
    groupe = ivector(nx + 1);
    ni = Puissance2(nx) - 1; // le nombre d'indices (ni=2**nx-1)
    indices = dmatrix(ny + 1, ni + 1);
    indices_globaux = dmatrix(ny + 1, nx + 1);
    indices_setvar = imatrix(ni + 1, nx + 1);
    np = 0;

    no = pc->no;
    int CalculP(int, int);
    p = CalculP(nx, no);
    phi = dmatrix(nx + 1, no + 1);
    psi = dvector(p + 1);
    indmul = imatrix(p + 1, nx + 1);
    beta = dmatrix(ny + 1, p + 1);
    // Indices multiples
    int IndiceMultipleEval(int nnx, int noi, int nof, int **indmul);
    IndiceMultipleEval(nx, 0, no, indmul);

    rank = ivector(pc->p + 1);
    int pcond_rank_indice(int *, PolynomialChaos *, int *);
    for (int k = 0; k <= pc->p; k++)
        rank[k] = pcond_rank_indice(pc->indmul[k], this, var_sto);
    typecomputation= 0;
    free_ivector(var_sto);
}

void PolynomialChaos::SetDegree(int nno)
{
    if (nno < 0)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::SetDegree) : degree  " << nno << " is < 0"
                << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    if (nx < 1)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::SetDegree) : stochastic dimension  " << nx
                << " is < 1" << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    if (no)
    {
        free_dmatrix(phi, nx + 1);
        free_dvector(psi);
        free_imatrix(indmul, p + 1);
        free_dmatrix(beta, ny + 1);
    }
    no = nno;
    int CalculP(int, int);
    p = CalculP(nx, no);
    phi = dmatrix(nx + 1, no + 1);
    psi = dvector(p + 1);
    indmul = imatrix(p + 1, nx + 1);
    beta = dmatrix(ny + 1, p + 1);
    // Indices multiples
    int IndiceMultipleEval(int nnx, int noi, int nof, int **indmul);
    IndiceMultipleEval(nx, 0, no, indmul);
}

void PolynomialChaos::SetInput(int i, double value)
{
    if (i < 1)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::SetInput) : rank input " << i << " < 1"
                << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    if (i > nx)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::SetInput) : rank input " << i
                << " is > stochastic dimension " << nx << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    else
        x[i] = value;
}

void PolynomialChaos::PropagateInput()
{
    int i, k;
    // calcul du tableau phi[1...nx][0...no]
    for (i = 1; i <= nx; i++)
    {
        if (funpol[i - 1] == HERMITE)
            hermite(phi[i], x[i], no);
        else if (funpol[i - 1] == LEGENDRE)
            legendre(phi[i], x[i], no);
        else if (funpol[i - 1] == LAGUERRE)
            laguerre(phi[i], x[i], no);
	else
	    jacobi(phi[i], x[i], gpx->va[i-1]->a, gpx->va[i-1]->b, no);
    }
    // calcul du vecteur psi
    for (k = 0; k <= p; k++)
    {
        for (psi[k] = 1., i = 1; i <= nx; i++)
            psi[k] = psi[k] * phi[i][indmul[k][i]];
    }
}

// Juillet 08 2010
void PolynomialChaos::PropagateInput(double *input)
{
    for (int i = 1; i <= nx; i++)
        SetInput(i, input[i]);
    PropagateInput();
}

void PolynomialChaos::ComputeOutput()
{
    int j, k;
    double s;
    PropagateInput();
    for (j = 1; j <= ny; j++)
    {
        for (s = 0., k = 0; k <= p; k++)
            s += beta[j][k] * psi[k];
        y[j] = s;
    }
}

void PolynomialChaos::ComputeOutput(double *entree)
{
    int j, k;
    double s;
    PropagateInput(entree);
    for (j = 1; j <= ny; j++)
    {
        for (s = 0., k = 0; k <= p; k++)
            s += beta[j][k] * psi[k];
        y[j] = s;
    }
}

void coefficients_integration(PolynomialChaos *pc, SetRandomVariable *gva)
{
    if (pc->np != gva->np)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(coefficients_integration) : output number of PC'sample  "
                << pc->np << " != output number of sample = " << gva->np
                << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    if (gva->degre_max < pc->no)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(coefficients_integration) :  degree of PC " << pc->no
                << " > maximal value = " << gva->degre_max << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    int i, j, k;
    double * w;
    w = gva->w;
    double ** X = gva->x;
    for (k = 0; k <= pc->p; k++)
        for (j = 1; j <= pc->ny; j++)
            pc->beta[j][k] = 0.;
    for (i = 1; i <= gva->np; i++)
    {
        pc->PropagateInput(X[i]);
        for (k = 0; k <= pc->p; k++)
        {
            for (j = 1; j <= pc->ny; j++)
                pc->beta[j][k] += pc->target[i][j] * pc->psi[k] * w[i];
        }
    }
    pc->typecomputation = 0;
}

void coefficients_regression(PolynomialChaos *pc, SetRandomVariable *gva)
{
    void CoefficientsRegression(double **, double *, double **, double **, double *, int, int, int);
    if ((gva->typesampling != "Lhs") && (gva->typesampling != "LhsMaxMin")
            && (gva->typesampling != "QmcSobol")
            && (gva->typesampling != "MonteCarlo"))
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg
                << "Nisp(coefficients_regression) :  PC and sample are not compatible "
                << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    int np = gva->np;  // nombre de simulations
    int p = pc->p;
    // matrice des regresseurs (Z'Z)=matrice d'information
    double * Z = new double[np * (p + 1)];
    for (int i = 1; i <= np; i++)
    {
        pc->PropagateInput(gva->x[i]);
        for (int k = 1; k <= p + 1; k++)
            Z[(i - 1) + (k - 1) * np] = pc->psi[k - 1];
    }
    CoefficientsRegression(pc->beta, Z, pc->target, pc->errLoo, pc->eqmLoo, p, np, pc->ny);
    pc->typecomputation = 1;
    delete[] Z;
}

void PolynomialChaos::ComputeChaosExpansion(PolynomialChaos * pc, double *input,
        int *var_opt, int nopt)
{
    int i, j, k;
    for (i = 1; i <= nopt; i++)
    {
        j = var_opt[i - 1];
        if (pc->funpol[j - 1] == HERMITE)
            hermite(pc->phi[j], input[i], pc->no);
        else if (pc->funpol[j - 1] == LEGENDRE)
            legendre(pc->phi[j], input[i], pc->no);
        else if (pc->funpol[j - 1] == LAGUERRE)
            laguerre(pc->phi[j], input[i], pc->no);
	else
	    jacobi(pc->phi[j], input[i], pc->gpx->va[i-1]->a, pc->gpx->va[i-1]->b, pc->no);
    }
    // calcul du vecteur psi
    for (k = 0; k <= pc->p; k++)
    {
        for (pc->psi[k] = 1., i = 1; i <= nopt; i++)
        {
            j = var_opt[i - 1];
            pc->psi[k] = pc->psi[k] * pc->phi[j][pc->indmul[k][j]];
        }
    }
    for (j = 1; j <= ny; j++)
    {
        for (k = 0; k <= p; k++)
            beta[j][k] = 0;
        for (k = 0; k <= pc->p; k++)
            beta[j][rank[k]] += pc->psi[k] * pc->beta[j][k];
    }
    SetAnova();
}

void PolynomialChaos::ComputeChaosExpansion(SetRandomVariable *gva,
        std::string Methode)
{
    if (Methode == "Integration")
        coefficients_integration(this, gva);
    else if (Methode == "Regression")
        coefficients_regression(this, gva);
    else
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::ComputeChaosExpansion) :  method "
                << Methode << " is not known" << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    SetAnova();
}
void PolynomialChaos::Save(char *fichier)
{
    int j, k;
    std::ofstream sortie(fichier, std::ios::out);
    if (!sortie)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::Save) :  problem when open file "
                << fichier << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    sortie.setf(std::ios::scientific, std::ios::floatfield);
    sortie.precision(6);
    sortie << "nx= " << nx << " ";
    for (j = 1; j <= nx; j++) {
        sortie << typol[j - 1] << " ";
        if (typol[j - 1] == "Jacobi") {
          sortie << gpx->va[j - 1]->a << " " << gpx->va[j - 1]->b << " ";
	}
    }
    sortie << "no= " << no << " p= " << p << " ny= " << ny;
    for (j = 1; j <= ny; j++)
    {
        sortie << " Coefficients[" << j << "]= ";
        for (k = 0; k <= p; k++)
            sortie << beta[j][k] << " ";
    }
    sortie << std::endl;
    sortie.close();
}

// Calcul de la moyenne, de la variance, des indices de sobol (ANOVA) et des indices globaux
void PolynomialChaos::SetAnova()
{
    // Decomposition ANOVA (indice[0]=variance)
    int i, j, k, ip;
    for (j = 1; j <= ny; j++)
        for (i = 1; i <= ni; i++)
            indices[j][i] = 0.;
    for (j = 1; j <= ny; j++)
        for (i = 1; i <= nx; i++)
            indices_globaux[j][i] = 0.;
    for (i = 1; i <= ni; i++)
        for (j = 1; j <= nx; j++)
            indices_setvar[i][j] = 0;
    for (k = 0; k <= p; k++)
    {
        for (ip = 0, i = 1; i <= nx; i++)
        {
            if (indmul[k][i])
                ip = ip + Puissance2(i - 1);
        }
        if (ip)
        {
            for (j = 1; j <= ny; j++)
                indices[j][ip] += (beta[j][k] * beta[j][k]);
            for (i = 1; i <= nx; i++)
            { // Groupe de variables
                if (indmul[k][i])
                    indices_setvar[ip][i] = 1;
            }
        }
    }
    // Indices globaux
    for (j = 1; j <= ny; j++)
    {
        for (k = 0; k <= p; k++)
        {
            for (i = 1; i <= nx; i++)
                if (indmul[k][i])
                    indices_globaux[j][i] += (beta[j][k] * beta[j][k]);
        }
    }
    // Calcul de la moyenne
    for (j = 1; j <= ny; j++)
        moyenne[j] = beta[j][0];
    // Calcul de la variance
    for (j = 1; j <= ny; j++)
        for (variance[j] = 0., k = 1; k <= p; k++)
            variance[j] += (beta[j][k] * beta[j][k]);
    // Normalisation par la variance -> indices
    for (j = 1; j <= ny; j++)
        if (variance[j])
            for (i = 1; i <= ni; i++)
                indices[j][i] = indices[j][i] / variance[j];
    for (j = 1; j <= ny; j++)
        if (variance[j])
            for (i = 1; i <= nx; i++)
                indices_globaux[j][i] = indices_globaux[j][i] / variance[j];
}

// Moyenne
double PolynomialChaos::GetMean(int j)
{
    if (j < 1)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetMean) :  the rank of output " << j
                << " < 1" << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    if (j > ny)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetMean) :  the rank of output " << j
                << " > number of output = " << ny << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    return (moyenne[j]);
}

// Variance
double PolynomialChaos::GetVariance(int j)
{
    if (j < 1)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetVariance) :  rank of output " << j
                << " < 1" << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    if (j > ny)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetVariance) :  rank of output " << j
                << " > number of output = " << ny << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    return (variance[j]);
}

double PolynomialChaos::GetCovariance(int i, int j)
{
    if (i < 1)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg
                << "Nisp(PolynomialChaos::GetCovariance) :  rank of the first variable "
                << i << " < 1" << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    if (i > ny)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg
                << "Nisp(PolynomialChaos::GetCovariance) :  rank of the first variable "
                << i << " > number of ouput = " << ny << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    if (j < 1)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg
                << "Nisp(PolynomialChaos::GetCovariance) :  rank of the second variable "
                << j << " < 1" << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    if (j > ny)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg
                << "Nisp(PolynomialChaos::GetCovariance) :  rank of the second variable "
                << j << " > number of ouput = " << ny << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    double cov;
    int k;
    for (cov = 0., k = 1; k <= p; k++)
        cov += beta[i][k] * beta[j][k];
    return (cov);
}

double PolynomialChaos::GetCorrelation(int i, int j)
{
    if (i < 1)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg
                << "Nisp(PolynomialChaos::GetCorrelation) :  rank of the first variable "
                << i << " < 1" << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    if (i > ny)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg
                << "Nisp(PolynomialChaos::GetCorrelation) :  rank of the first variable "
                << i << " > number of ouput = " << ny << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    if (j < 1)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg
                << "Nisp(PolynomialChaos::GetCorrelation) :  rank of the second variable "
                << j << " < 1" << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    if (j > ny)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg
                << "Nisp(PolynomialChaos::GetCorrelation) :  rank of the second variable "
                << j << " > number of ouput = " << ny << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    return (GetCovariance(i, j) / sqrt(variance[i] * variance[j]));
}

double PolynomialChaos::GetIndiceFirstOrder(int i, int j)
{
    if (i < 1)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg
                << "Nisp(PolynomialChaos::GetIndiceFirstOrder) :  rank of the variable "
                << i << " < 1" << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    if (i > nx)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg
                << "Nisp(PolynomialChaos::GetIndiceFirstOrder) :  rank of the variable "
                << i << " > stochastic dimension = " << nx << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    if (j < 1)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg
                << "Nisp(PolynomialChaos::GetIndiceFirstOrder) :  rank of the output "
                << j << " < 1" << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    if (j > ny)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg
                << "Nisp(PolynomialChaos::GetIndiceFirstOrder) :  rank of the output "
                << j << " > number of ouput = " << ny << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    int ii = Puissance2(i - 1);
    return (indices[j][ii]);
}

double PolynomialChaos::GetIndiceTotalOrder(int i, int j)
{
    if (i < 1)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetCorrelation) :  rank of the variable "
                << i << " < 1" << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    if (i > nx)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetCorrelation) :  rank of the variable "
                << i << " > stochastic dimension = " << nx << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    if (j < 1)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetCorrelation) :  rank of the output "
                << j << " < 1" << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    if (j > ny)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetCorrelation) :  rank of the output "
                << j << " > number of ouput = " << ny << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    return (indices_globaux[j][i]);
}

// affiche le tableau des indices
void PolynomialChaos::GetMultipleIndices()
{
    int k, j;
    std::ostringstream msg;
    for (k = 0; k <= p; k++)
    {
        for (j = 1; j < nx; j++)
            msg << indmul[k][j] << ":";
        msg << indmul[k][nx];
        msg << std::endl;
    }
    NispOutPutMsg(msg.str(), 0);
}

// Retourne l'indice de sensibilite d'un groupe de variables - sortie 1
double PolynomialChaos::GetGroupIndice(int j)
{
    int i, k, r;
    double s;
    if (j < 1)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetGroupIndice) :  rank of the output "
                << j << " < 1" << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    if (j > ny)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetGroupIndice) :  rank of the output "
                << j << " > number of ouput = " << ny << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    for (s = 0., k = 1; k <= p; k++)
    {
        for (r = 1, i = 1; i <= nx; i++)
        {
            if (indmul[k][i] > groupe[i] * indmul[k][i])
                r = 0;
        }
        if (r)
            s += beta[j][k] * beta[j][k];
    }
    if (variance[j])
        return (s / variance[j]);
    else
        return (0.);
}

double PolynomialChaos::GetGroupIndiceInteraction(int j)
{
    int i, k, r;
    double s;
    if (j < 1)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg
                << "Nisp(PolynomialChaos::GetGroupIndiceInteraction) :  rank of the output "
                << j << " < 1" << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    if (j > ny)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg
                << "Nisp(PolynomialChaos::GetGroupIndiceInteraction) :  rank of the output "
                << j << " > number of ouput = " << ny << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    for (s = 0., k = 1; k <= p; k++)
    {
        for (r = 1, i = 1; i <= nx; i++)
        {
            if (indmul[k][i] > groupe[i] * indmul[k][i])
                r = 0;
        }
        for (i = 1; i <= nx; i++)
        {
            if ((indmul[k][i] == 0) && (groupe[i]))
                r = 0;
        }
        if (r)
            s += beta[j][k] * beta[j][k];
    }
    if (variance[j])
        return (s / variance[j]);
    else
        return (0.);
}

void PolynomialChaos::SetGroupAddVar(int i)
{
    if (i < 1)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::SetGroupAddVar) :  rank of the variable "
                << i << " < 1" << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    if (i > nx)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::SetGroupAddVar) :  rank of the variable "
                << i << " > stochastic dimension = " << nx << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    groupe[i] = 1;
}

void PolynomialChaos::SetGroupEmpty()
{
    for (int i = 1; i <= nx; i++)
        groupe[i] = 0;
}

void PolynomialChaos::GetAnova(int r)
{
    int i, j;
    std::ostringstream msg;
    for (i = 1; i <= ni; i++)
    { // ni = 2**nx - 1 = 3 dans cet exemple
        SetGroupEmpty(); // On vide le groupe
        for (j = 1; j <= nx; j++)
            if (indices_setvar[i][j])
                SetGroupAddVar(j); // on construit les 2**nx - 1 groupes
        for (j = 1; j <= nx; j++)
            msg << indices_setvar[i][j] << " "; // On affiche le groupe
        msg << " : " << GetGroupIndiceInteraction(r) << std::endl; // et on edite la valeur de l'indice du groupe
    }
    NispOutPutMsg(msg.str(), 0);
}

void PolynomialChaos::GetAnovaOrdered(double seuil, int r)
{ // seuil = 1. et r=1 par defaut
    int i, j;
    double *AnovaOrder = dvector(ni);
    for (i = 1; i <= ni; i++)
    { // ni = 2**nx - 1
        SetGroupEmpty(); // On vide le groupe
        for (j = 1; j <= nx; j++)
            if (indices_setvar[i][j])
                SetGroupAddVar(j); // on construit les 2**nx - 1 groupes
        AnovaOrder[i - 1] = GetGroupIndiceInteraction(r);
    }
    // On ordonne
    int *indx = ivector(ni);
    dindex(ni, AnovaOrder, indx, 1);

    // On affiche
    std::ostringstream msg;
    //msg << "Nisp(PolynomialChaos::GetAnovaOrdered) output = " << r << " threshold = " << seuil << "\n"<<endl;
    double dv = 0.;
    for (i = 1; i <= ni; i++)
    { // ni = 2**nx - 1 = 3 dans cet exemple
        dv += AnovaOrder[indx[i - 1]];
        for (j = 1; j <= nx; j++)
            msg << indices_setvar[indx[i - 1] + 1][j] << " "; // On affiche le groupe
        msg << " : " << AnovaOrder[indx[i - 1]] << " : cumul = " << dv << std::endl; // et on edite la valeur de l'indice du groupe
        if (dv > seuil)
            break;
    }
    NispOutPutMsg(msg.str(), 0);
    free_dvector(AnovaOrder);
    free_ivector(indx);
}

void PolynomialChaos::GetAnovaOrderedCoefficients(double seuil, int r)
{ // sortie r = 1 par defaut
    int i, k;
    std::ostringstream msg;
    if (r < 1)
    {
        msg << "NISP - ERROR" << std::endl;
        msg
                << "Nisp(PolynomialChaos::GetAnovaOrderedCoefficients) :  rank of the output "
                << r << " < 1" << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    if (r > ny)
    {
        msg << "NISP - ERROR" << std::endl;
        msg
                << "Nisp(PolynomialChaos::GetAnovaOrderedCoefficients) :  rank of the output "
                << r << " > number of ouput = " << ny << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    if (variance[r] == 0.)
    {
        msg << "NISP - ERROR" << std::endl;
        msg
                << "Nisp(PolynomialChaos::GetAnovaOrdererCoefficients) :  variance of the output "
                << r << " = 0" << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    //double *Ratio = dvector(p+1);
    double * Ratio = new double[p];
    for (k = 1; k <= p; k++)
        Ratio[k - 1] = beta[r][k] * beta[r][k] / variance[r];

    //int *indx=ivector(p);
    int * indx = new int[p];
    dindex(p, Ratio, indx, 1);
    double dv = 0.;
    //msg << "Nisp(PolynomialChaos::GetAnovaOrdererCoefficients)"<<endl;
    for (k = 0; k < p; k++)
    {
        dv += Ratio[indx[k]];
        for (i = 1; i <= nx; i++)
            msg << indmul[indx[k] + 1][i] << " ";
        msg << " : " << Ratio[indx[k]] << " : cumul = " << dv * 1.e0
                << " coef = " << beta[r][indx[k] + 1] << std::endl;
        if (dv > seuil)
            break;
    }
    NispOutPutMsg(msg.str(), 0);
    // Mise a zero des autres coefficients;
    for (; k < p; k++)
        beta[r][indx[k] + 1] = 0.;
    delete[] indx;
    delete[] Ratio;
}

void PolynomialChaos::BuildSample(std::string type, int np, int order)
{
    if (gpx->np)
    {
        free_dmatrix(sample, ny);
        free_imatrix(sampleindx, ny);
    }
    gpx->BuildSample(type, np);
    sample = dmatrix(ny, gpx->np);
    sampleindx = imatrix(ny, gpx->np);
    int j, k;
    for (k = 1; k <= gpx->np; k++)
    {
        ComputeOutput(gpx->x[k]);
        for (j = 1; j <= ny; j++)
            sample[j - 1][k - 1] = y[j];
    }
    if (order)
    {
        for (j = 0; j < ny; j++)
            dindex(gpx->np, sample[j], sampleindx[j], 0);
    }
    else
    {
        for (j = 0; j < ny; j++)
            for (k = 0; k < gpx->np; k++)
                sampleindx[j][k] = k;
    }
}
double PolynomialChaos::GetSample(int k, int j)
{
    std::ostringstream msg;
    if (j < 1)
    {
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetSample) :  rank of the output " << j
                << " < 1" << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    if (j > ny)
    {
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetSample) :  rank of the output " << j
                << " > number of ouput = " << ny << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    if (gpx->np < 1)
    {
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetSample) :  size of sample = "
                << gpx->np << " < 1 " << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    if (k < 1)
    {
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetSample) :  rank of simulation " << k
                << " < 1" << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    if (k > gpx->np)
    {
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetSample) :  rank of simulation " << k
                << " > size of sample = " << gpx->np << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    return sample[j - 1][sampleindx[j - 1][k - 1]];
}
double PolynomialChaos::GetQuantile(double alpha, int j)
{
    std::ostringstream msg;
    if (j < 1)
    {
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetQuantile) :  rank of the output " << j
                << " < 1" << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    if (j > ny)
    {
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetQuantile) :  rank of the output " << j
                << " > number of ouput = " << ny << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    if ((alpha < 0) || (alpha > 1))
    {
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetQuantile) :  value of quantile "
                << alpha << " is not in [0,1]" << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    if (gpx->np < 1)
    {
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetSample) :  size of sample = "
                << gpx->np << " < 1 " << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    int k = (int) (alpha * (unsigned long) gpx->np);
    if (k < 1)
        k = 1;
    else if (k > gpx->np)
        k = gpx->np;
    return (sample[j - 1][sampleindx[j - 1][k - 1]]);
}

double PolynomialChaos::GetQuantileWilks(double alpha, double beta, int j)
{
    std::ostringstream msg;
    if (j < 1)
    {
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetQuantileWilks) :  rank of the output "
                << j << " < 1" << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    if (j > ny)
    {
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetQuantileWilks) :  rank of the output "
                << j << " > number of ouput = " << ny << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    if ((alpha < 0) || (alpha > 1))
    {
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetQuantileWilks) :  value of quantile "
                << alpha << " is not in [0,1]" << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    if ((beta < 0.5) || (beta > 0.999999))
    {
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetQuantileWilks) :  value of confidence "
                << beta << " is not in [0.5,0.999999]" << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    int n = gpx->np;
    if (n < 1)
    {
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetQuantileWilks) :  size of sample = "
                << n << " < 1 " << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    int ns;
    ns = (int) (floor(log(1. - beta) / log(alpha))) + 1;
    if (n < ns)
    {
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetQuantileWilks) :  size of sample = "
                << n << " < " << ns << " = minimal size " << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    double InvFcn(double);
    int r = (int) ceil(InvFcn(beta) * sqrt(alpha * (1. - alpha) * n));
    /*double binomial(int, int);
     int nc=(int) floor(n*(1.-alpha));
     double s;
     for(s=0.,r=0;r<=nc;r++) {
     s += binomial(n,r) * pow(1.-alpha, (double) r) * pow(alpha, (double) (n-r));
     if(s > 1. - beta) break;
     }
     */
    int k = (int) floor(n * alpha) + r;
    //int k = n-r;
    //cout << "r = " << r << " k = " << k << " ns = " << ns << endl;
    if (k < 1)
        k = 1;
    else if (k > n)
        k = n;
    return (sample[j - 1][sampleindx[j - 1][k - 1]]);
}

double PolynomialChaos::GetInvQuantile(double seuil, int j)
{
    std::ostringstream msg;
    if (j < 1)
    {
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetInvQuantile) :  rank of the output "
                << j << " < 1" << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    if (j > ny)
    {
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetInvQuantile) :  rank of the output "
                << j << " > number of ouput = " << ny << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    if (gpx->np < 1)
    {
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetSample) :  size of sample = "
                << gpx->np << " < 1 " << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    int k;
    for (k = 1; k <= gpx->np; k++)
        if (sample[j - 1][sampleindx[j - 1][k - 1]] > seuil)
            break;
    return ((double) (k) / gpx->np);
}

void PolynomialChaos::SetSizeTarget(int nnp)
{
    if (np)
    {
        free_dmatrix(target, np + 1);
	free_dmatrix(errLoo, np + 1);
	
    };
    np = nnp;
    target = dmatrix(nnp + 1, ny + 1);
    errLoo = dmatrix(nnp + 1, ny + 1);
}
void PolynomialChaos::ReadTarget(char *fichier)
{
    ReadTarget2(target, np, ny, fichier);
}
void PolynomialChaos::FreeMemoryTarget()
{
    if (np) {
        free_dmatrix(target, np + 1);
        free_dmatrix(errLoo, np + 1);
    }
}
void PolynomialChaos::SetTarget(int k, int j, double z)
{
    target[k][j] = z;
}
int PolynomialChaos::GetDimensionInput()
{
    return (nx);
}
int PolynomialChaos::GetDimensionOutput()
{
    return (ny);
}
int PolynomialChaos::GetDimensionExpansion()
{
    return (p + 1);
}
int PolynomialChaos::GetDegree()
{
    return (no);
}
double PolynomialChaos::GetOutput(int j)
{
    std::ostringstream msg;
    if (j < 1)
    {
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetOutput) :  rank of the output " << j
                << " < 1" << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    if (j > ny)
    {
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetOutput) :  rank of the output " << j
                << " > number of ouput = " << ny << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    return (y[j]);
}

double PolynomialChaos::GetEqmLoo(int j)
{
    std::ostringstream msg;
    // JMM 3 mars 2016
    if (this->typecomputation != 1)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg
                << "Nisp(GetEqmLoo) :  Coefficients have not been computed by Regression !!! "
                << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    
    if (j < 1)
    {
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetEqmLoo) :  rank of the output " << j
                << " < 1" << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    if (j > ny)
    {
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetEqmLoo) :  rank of the output " << j
                << " > number of ouput = " << ny << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    return (eqmLoo[j]);
}

double PolynomialChaos::GetErrLoo(int i, int j)
{
    std::ostringstream msg;
    // JMM 3 mars 2016
    if (this->typecomputation != 1)
    {
        std::ostringstream msg;
        msg << "NISP - ERROR" << std::endl;
        msg
                << "Nisp(GetErrLoo) :  Coefficients have not been computed by Regression !!! "
                << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    
    if (j < 1)
    {
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetErrLoo) :  rank of the output " << j
                << " < 1" << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    if (j > ny)
    {
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetErrLoo) :  rank of the output " << j
                << " > number of ouput = " << ny << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    if (i < 1)
    {
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetErrLoo) :  rank of the input " << i
                << " < 1" << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    if (i > np)
    {
        msg << "NISP - ERROR" << std::endl;
        msg << "Nisp(PolynomialChaos::GetErrLoo) :  rank of the input " << i
                << " > number of sample = " << np << std::endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    return (errLoo[i][j]);
}
