//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   nisp_pc.h
 \author Jean-Marc Martinez <jean-marc.martinez@cea.fr>
 \date   2011/04/07 14:33:00

 \brief  Class of polynomial chaos
 */

#ifndef NISP_PC_H
#define NISP_PC_H

#include "nisp_gva.h" // Class of the set of random variables
#ifdef WIN32
#include "libnisp_Export.h"
#else
#define libnisp_EXPORT
#endif

//! Class of polynomial chaos
class libnisp_EXPORT PolynomialChaos
{
public:
    //! Stocahstic dimension (number of orthogonal polynomials)
    int nx;
    //! Set of stochastic variables
    SetRandomVariable * gpx;
    //! Number of terms of ANOVA decomposition
    int ni; // 2**nx-1
    //! Degree
    int no;
    //! Number of coefficients = p+1
    int p;
    //! Number of output
    int ny;
    //! Number of simulations
    int np;
    //! Types of orthogonal polynomials ("Hermite","Legendre","Laguerre")
    std::vector<std::string> typol;
    //! Function associated wtih orthogonal polynomials ("hermite()","legendre()","laguerre()")
    std::vector<int> funpol;
    //! Input x[1...nx]
    double *x;
    //! Output y[1...ny]
    double *y;
    //! Sample sample[1...ny][1...gpx->np]
    double **sample;
    //! Ordored sample
    int **sampleindx;
    //! Experimental design target [1...np][1...ny]
    double **target;
    //! Error Loo  [1...np][1...ny]
    double **errLoo;
    //! EQM error Loo  [1...ny]
    double *eqmLoo;
    //! Monomial of orthogonal polynomials phi[1...nx][0...no]
    double **phi;
    //! Monomial multidimensional polynomial  psi[0...p] 
    double *psi;
    //! Matrix of relation between indice k=[0...p] and multiple indice
    int **indmul;
    //! Matrix of coefficients beta[1...ny][0...p]
    double **beta;
    //! Mean moyenne[1...ny]=beta[1...ny][0]
    double *moyenne;
    //! Variance variance[1...ny]
    double *variance;
    //! Indices of sensitivity : indices[1...ny][1...ni]
    double **indices;
    //! Total sensitivity indices : indices_globaux[1...ny][1...nx]
    double **indices_globaux;
    //! Blackboard of the rank of variables in the set (groupe)
    int **indices_setvar;
    //! Set of random variable for ANOVA
    int *groupe;
    //! Expectation conditionnal
    int *rank;
    //! Type of sampling : 1 pour Regression et 0 sinon
    int typecomputation;
public:
    /** @name Constructor and Destructor
     **/
    //@{
    //! Constructor from a set of stochastic variables
    PolynomialChaos(SetRandomVariable *gva, int ny = 1);
    //! Constructor from a file
    PolynomialChaos(char *file);
    //! Constructor from another PC for conditional expectation
    PolynomialChaos(PolynomialChaos *pc, int *var_sto, int nsto);
    //! Destructor
    ~PolynomialChaos();
    //@}
    //! Definition of the degree (no)
    void SetDegree(int no);
    //! Dimension of output (ny)
    void SetDimensionOutput(int ny);
    //! Allocation of memory for the target
    void SetSizeTarget(int np);
    //! Assign a target
    void SetTarget(int k, int j, double output);
    //! Read a set of target from a file
    void ReadTarget(char *file);
    //! Freeing the target memory 
    void FreeMemoryTarget();
    //! Set an input
    void SetInput(int i, double x);
    //! Propagation of input which has been specified SetInput()
    void PropagateInput();
    //! Propagation of input[1...nx]
    void PropagateInput(double *input);
    //! Computation of output y[1...ny]
    void ComputeOutput();
    //! Computation of output in relation with input : input[1...nx]
    void ComputeOutput(double *input);
    //! Get an output (rank of output is 1 by default)
    double GetOutput(int = 1);
    //! Get EQM Loo of an output(rank of output is 1 by default)
    double GetEqmLoo(int = 1);
    //! Get Err Loo of an output(rank of output is 1 by default)
    double GetErrLoo(int i, int j = 1);
    //! Computation of coefficients by method ("Integration", "Regression")
    void ComputeChaosExpansion(SetRandomVariable *gva, std::string methode);
    //! Conditionnal Expectation
    void ComputeChaosExpansion(PolynomialChaos * pc, double *input,
            int *var_opt, int nopt);
    //! Saver the polynomial chaos in a "file"
    void Save(char *file);
    //! Preparation on the ANOVA decomposition (for Sobol'indices)
    void SetAnova();
    //! Mean
    double GetMean(int = 1);
    //! Variance
    double GetVariance(int = 1);
    //! Covariance
    double GetCovariance(int i, int j);
    //! Correlation coefficient
    double GetCorrelation(int i, int j);
    //! First indice of sensitivity
    double GetIndiceFirstOrder(int i, int j = 1);
    //! Total indice of sensitivity
    double GetIndiceTotalOrder(int i, int j = 1);
    //! Edition of the multiple indices
    void GetMultipleIndices();
    //! Indice of sensitivity of a set of variables - output 1 by default
    double GetGroupIndice(int = 1);
    //! Indice of interaction sensitivity of a set of variables - output 1 by default
    double GetGroupIndiceInteraction(int = 1);
    //! Add a random variable (rank i) in the set
    void SetGroupAddVar(int i);
    //! Freeing the set
    void SetGroupEmpty();
    //! Edition of the ANOVA decomposition
    void GetAnova(int r = 1);
    //! Edition of the ANOVA ordered decomposition
    void GetAnovaOrdered(double seuil = 1., int r = 1);
    //!  Edition of the ANOVA ordered decomposition / coefficients
    void GetAnovaOrderedCoefficients(double seuil = 1., int r = 1);
    //! Writing the program C of the ploynomial chaos in a file
    void GenerateCode(char *, char *);
    //! Number of input
    int GetDimensionInput();
    //! Number of output
    int GetDimensionOutput();
    //! Number of the coefficients of the polynomial chaos (p=number-1)
    int GetDimensionExpansion();
    //! Degree
    int GetDegree();
    //! Informations : type of the set of orthogonal polynomial
    void GetLog();
    //! Build a sample for statitical analysis (quantile)
    void BuildSample(std::string type, int np, int order = 1);
    //! The value of the output j for the example k
    double GetSample(int k, int j = 1);
    //! Polynomial chaos is a random variable : a realisation of this random variable (GetOutput()).
    void Realisation();
    //! Quantile of order alpha
    double GetQuantile(double alpha, int j = 1);
    //! Wilks's Quantile of order alpha with confidence beta
    double GetQuantileWilks(double alpha, double beta, int j = 1);
    //! Probability of a value overrun 
    double GetInvQuantile(double seuil, int j = 1);
};
#endif
