//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TestTTAnisp.cxx
 \author Thierry Crestaux (CS)
 \date   29/04/2013
 
 \brief 
 
 */

// ROOT
#include "TMath.h"
#ifdef ROOT5
#include "Api.h"
#endif

// cppunit

#include <fstream>
#include <iostream>
#include <exception>
#define DBL_PRECISION_COMPARE 1.e-12
#define DBL_PRECISION_COMPARE2 1.e-9

//uranie
#include "TestTAnisp.h"
#include "TDataServer.h"
#include "TUniformDistribution.h"
#include "TLogNormalDistribution.h"
#include "TLogTriangularDistribution.h"
#include "TLogUniformDistribution.h"
#include "TTriangularDistribution.h"
#include "TOutputFileKey.h"
#include "TPossibilityAttribute.h"
#include "TestUserFunctions.h"
#include "TPolynomialChaos.h"
#include "TAttributeFormula.h"
#include "TOutputFileDataServer.h"
#include "code_paths.h"
#include "uranie_commands.h"

#include "anisp.hpp"

TAnispTest::TAnispTest() :
        _tdsIshigami(NULL), _tdsFlowrate(NULL), _tdsNULL(NULL), _anisp(NULL), _anisp2(
                NULL), _anisp3(NULL), _foutDataServer(NULL), _codeFlowrate(NULL)
{
}

TAnispTest::~TAnispTest()
{

}

///////////////////////////////////
//  setUP                        //
///////////////////////////////////

void TAnispTest::setUp()
{

    //Fct Ishigami ///////////////////////////////////
    _tdsIshigami = new URANIE::DataServer::TDataServer("tdsIshigami",
						       "TDS for Ishigami Use Case");
    _tdsIshigami->addAttribute(
	new URANIE::DataServer::TUniformDistribution("x1",
						     -1. * TMath::Pi(), TMath::Pi()));
    _tdsIshigami->addAttribute(
	new URANIE::DataServer::TUniformDistribution("x2",
						     -1. * TMath::Pi(), TMath::Pi()));
    _tdsIshigami->addAttribute(
	new URANIE::DataServer::TUniformDistribution("x3",
						     -1. * TMath::Pi(), TMath::Pi()));
    _tdsIshigami->getAttribute("x1")->delShare();
    _tdsIshigami->getAttribute("x2")->delShare();
    _tdsIshigami->getAttribute("x3")->delShare();
    
#ifdef ROOT5
    // Get the void * from the "Ishigami" analytical function
    G__ClassInfo globalscope;
    G__MethodInfo method;
    long dummy;
    // resolve function overloading
    method = globalscope.GetMethod("FctIshigami", "double*,double*", &dummy);
    _p2fIshigami = method.PointerToFunc();
#endif

    //end Fct Ishigami //////////////////////////////////

    // FlowRate code ////////////////////////////////////
    _tdsFlowrate = new URANIE::DataServer::TDataServer("tdsFlowrate",
						       "TDs for the Flowrate Use Case");
//   _tdsFlowrate->fileDataRead("flowrate_sampler_launcher_500.dat");

    _tdsFlowrate->addAttribute(new TUniformDistribution("rw", 0.05, 0.15));
    _tdsFlowrate->addAttribute(new TUniformDistribution("r", 100.0, 50000.0));
    _tdsFlowrate->addAttribute(new TUniformDistribution("tu", 63070.0, 115600.0));
    _tdsFlowrate->addAttribute(new TUniformDistribution("tl", 63.1, 116.0));
    _tdsFlowrate->addAttribute(new TUniformDistribution("hu", 990.0, 1110.0));
    _tdsFlowrate->addAttribute(new TUniformDistribution("hl", 700.0, 820.0));
    _tdsFlowrate->addAttribute(new TUniformDistribution("l", 1120.0, 1680.0));
    _tdsFlowrate->addAttribute(new TUniformDistribution("kw", 9855.0, 12045.0));
    _tdsFlowrate->getAttribute("rw")->delShare();
    _tdsFlowrate->getAttribute("r")->delShare();
    _tdsFlowrate->getAttribute("tu")->delShare();
    _tdsFlowrate->getAttribute("tl")->delShare();
    _tdsFlowrate->getAttribute("hu")->delShare();
    _tdsFlowrate->getAttribute("hl")->delShare();
    _tdsFlowrate->getAttribute("l")->delShare();
    _tdsFlowrate->getAttribute("kw")->delShare();
    
    // The reference input file
//   TString sJDDReference = gSystem->pwd();
//   sJDDReference += TString("/flowrate_input_with_keys.in");
//   
//   _tdsFlowrate->getAttribute("rw")->setFileKey(sJDDReference, "Rw","%f");
//   _tdsFlowrate->getAttribute("r")->setFileKey(sJDDReference, "R","%f");
//   _tdsFlowrate->getAttribute("tu")->setFileKey(sJDDReference, "Tu","%f");
//   _tdsFlowrate->getAttribute("tl")->setFileKey(sJDDReference, "Tl","%f");
//   _tdsFlowrate->getAttribute("hu")->setFileKey(sJDDReference, "Hu","%f");
//   _tdsFlowrate->getAttribute("hl")->setFileKey(sJDDReference, "Hl","%f");
//   _tdsFlowrate->getAttribute("l")->setFileKey(sJDDReference, "L","%f");
//   _tdsFlowrate->getAttribute("kw")->setFileKey(sJDDReference, "Kw","%f");

// The output file of the code 
//   TOutputFileKey *fout = new TOutputFileKey("_output_flowrate_withKey_.dat");
//   // The attribute in the output file
//   TAttribute *attr = new TAttribute("y0");
//   attr->setFileKey("_output_flowrate_withKey_.dat", "yhat");
//   fout->addAttribute(attr);

    TString sJDDReferenceKey = gSystem->pwd();
    sJDDReferenceKey += TString("/flowrate_input_with_row.in");

    _tdsFlowrate->getAttribute("rw")->setFileKey(sJDDReferenceKey, "Rw", "%f",
						 TAttributeFileKey::kNewRow);
    _tdsFlowrate->getAttribute("r")->setFileKey(sJDDReferenceKey, "R", "%f",
						TAttributeFileKey::kNewRow);
    _tdsFlowrate->getAttribute("tu")->setFileKey(sJDDReferenceKey, "Tu", "%f",
						 TAttributeFileKey::kNewRow);
    _tdsFlowrate->getAttribute("tl")->setFileKey(sJDDReferenceKey, "Tl", "%f",
						 TAttributeFileKey::kNewRow);
    _tdsFlowrate->getAttribute("hu")->setFileKey(sJDDReferenceKey, "Hu", "%f",
						 TAttributeFileKey::kNewRow);
    _tdsFlowrate->getAttribute("hl")->setFileKey(sJDDReferenceKey, "Hl", "%f",
						 TAttributeFileKey::kNewRow);
    _tdsFlowrate->getAttribute("l")->setFileKey(sJDDReferenceKey, "L", "%f",
						TAttributeFileKey::kNewRow);
    _tdsFlowrate->getAttribute("kw")->setFileKey(sJDDReferenceKey, "Kw", "%f",
						 TAttributeFileKey::kNewRow);

    _foutDataServer = new TOutputFileDataServer(
	"_output_flowrate_withRow_.dat");
    _foutDataServer->addAttribute("yhat");

    _codeFlowrate = new TCode(_tdsFlowrate, UranieTests::code_path("flowrate") +
        " -s -r flowrate_input_with_row.in");

    _codeFlowrate->addOutputFile(_foutDataServer);

    // end FlowRate code ////////////////////////////////////

    //Fct Linear ///////////////////////////////////
    _tdsLinear = new URANIE::DataServer::TDataServer("tdsLinear",
						     "TDS for Linear Use Case");
    _tdsLinear->addAttribute(
	new URANIE::DataServer::TNormalDistribution("x1", 0., 2.));
    _tdsLinear->addAttribute(
	new URANIE::DataServer::TNormalDistribution("x2", 0., 2.));
    _tdsLinear->addAttribute(
	new URANIE::DataServer::TNormalDistribution("x3", 0., 3.));
    _tdsLinear->addAttribute(
	new URANIE::DataServer::TNormalDistribution("x4", 0., 0.5));
    _tdsLinear->getAttribute("x1")->delShare();
    _tdsLinear->getAttribute("x2")->delShare();
    _tdsLinear->getAttribute("x3")->delShare();
    _tdsLinear->getAttribute("x4")->delShare();
    
#ifdef ROOT5
    // Get the void * from the "Linear" analytical function
    G__ClassInfo globalscope2;
    G__MethodInfo method2;
    long dummy2;
    // resolve function overloading
    method2 = globalscope2.GetMethod("FctLinear", "double*,double*", &dummy2);
    _p2fLinear = method2.PointerToFunc();
#endif

    //end Fct Linear //////////////////////////////////

    //Fct Venise ///////////////////////////////////
    _tdsVenise = new URANIE::DataServer::TDataServer("tdsVenise",
						     "TDS for Venise Use Case");
    _tdsVenise->addAttribute(
	new URANIE::DataServer::TUniformDistribution("x1", -0.5, 0.5));
    _tdsVenise->addAttribute(
	new URANIE::DataServer::TUniformDistribution("x2", -0.5, 0.5));
    _tdsVenise->addAttribute(
	new URANIE::DataServer::TAttributeFormula("x3", "TMath::Exp(x2)"));
    _tdsVenise->addAttribute(
	new URANIE::DataServer::TAttributeFormula("x4", "TMath::Exp(x1)"));
    _tdsVenise->getAttribute("x1")->delShare();
    _tdsVenise->getAttribute("x2")->delShare();
    _tdsVenise->getAttribute("x3")->delShare();
    _tdsVenise->getAttribute("x4")->delShare();
    
#ifdef ROOT5
    // Get the void * from the "FctVenise" analytical function
    G__ClassInfo globalscope3;
    G__MethodInfo method3;
    long dummy3;
    // resolve function overloading
    method3 = globalscope3.GetMethod("FctVenise", "double*,double*", &dummy3);
    _p2fVenise = method3.PointerToFunc();
#endif
    //end Fct Venise //////////////////////////////////


#ifdef ROOT5
    _anisp = new URANIE::Modeler::TAnisp(_tdsIshigami, _p2fIshigami, "x1:x2:x3",
#else
    _anisp = new URANIE::Modeler::TAnisp(_tdsIshigami, &FctIshigami, "FctIshigami", "x1:x2:x3",
#endif
            "", "myOption");
    _anisp->setDrawProgressBar(kFALSE);
    _anisp2 = new URANIE::Modeler::TAnisp(_tdsFlowrate, _codeFlowrate,
            "myOption");
    _anisp2->setDrawProgressBar(kFALSE);
#ifdef ROOT5
    _anisp3 = new URANIE::Modeler::TAnisp(_tdsLinear, _p2fLinear, 
#else
    _anisp3 = new URANIE::Modeler::TAnisp(_tdsLinear, &FctLinear, "FctLinear", 
#endif
            "x1:x2:x3:x4", "", "myOption");
    _anisp3->setDrawProgressBar(kFALSE);
#ifdef ROOT5
    _anisp4 = new URANIE::Modeler::TAnisp(_tdsVenise, _p2fVenise, 
#else
    _anisp4 = new URANIE::Modeler::TAnisp(_tdsVenise, &FctVenise, "FctVenise",
#endif
            "x1:x3", "", "myOption");
    _anisp4->setDrawProgressBar(kFALSE);

                                          UranieTests::removeFile(gSystem->pwd(), "_*_.dat");

}

///////////////////////////////////
//  tearDown                     //
///////////////////////////////////

void TAnispTest::tearDown()
{

    if (_anisp)
    {
        delete _anisp;
	_anisp=NULL;
    }

    if (_anisp2)
    {
        delete _anisp2;
	_anisp2=NULL;
    }

    if (_anisp3)
    {
        delete _anisp3;
	_anisp3=NULL;
    }

    if (_anisp4)
    {
        delete _anisp4;
	_anisp4=NULL;
    }    

    if (_foutDataServer != NULL)
    {
	delete _foutDataServer;
	_foutDataServer=NULL;
    }

    if (_codeFlowrate)
    {
	delete _codeFlowrate;
	_codeFlowrate=NULL;
    }
      
    if (_tdsIshigami)
    {
	delete _tdsIshigami;
	_tdsIshigami=NULL;
    }

    if (_tdsFlowrate)
    {
	delete _tdsFlowrate;
	_tdsFlowrate=NULL;
    }

    if (_tdsLinear)
    {
	delete _tdsLinear;
	_tdsLinear=NULL;
    }

    if (_tdsVenise)
    {
	delete _tdsVenise;
	_tdsVenise = NULL;
    }
  
    UranieTests::removeFile(gSystem->pwd(),"test_res.dat");
    UranieTests::removeFile(gSystem->pwd(),"test_sto.dat");
    UranieTests::removeFile(gSystem->pwd(),"testVenise_data.dat");
    UranieTests::removeFile(gSystem->pwd(),"testVenise_error.dat");
    UranieTests::removeFile(gSystem->pwd(),"testVenise_indices.dat");
    UranieTests::removeFile(gSystem->pwd(),"testVenise_res.dat");
    UranieTests::removeFile(gSystem->pwd(),"testVenise_sto.dat");
    UranieTests::removeFile(gSystem->pwd(),"testIshigami_data.dat");
    UranieTests::removeFile(gSystem->pwd(),"testIshigami_error.dat");
    UranieTests::removeFile(gSystem->pwd(),"testIshigami_indices.dat");
    UranieTests::removeFile(gSystem->pwd(),"testIshigami_res.dat");
    UranieTests::removeFile(gSystem->pwd(),"testIshigami_sto.dat");
    UranieTests::removeFile(gSystem->pwd(),"testLinear_data.dat");
    UranieTests::removeFile(gSystem->pwd(),"testLinear_error.dat");
    UranieTests::removeFile(gSystem->pwd(),"testLinear_indices.dat");
    UranieTests::removeFile(gSystem->pwd(),"testLinear_res.dat");
    UranieTests::removeFile(gSystem->pwd(),"testLinear_sto.dat");
    UranieTests::removeFile(gSystem->pwd(),"testFlowRate_data.dat");
    UranieTests::removeFile(gSystem->pwd(),"testFlowRate_error.dat");
    UranieTests::removeFile(gSystem->pwd(),"testFlowRate_indices.dat");
    UranieTests::removeFile(gSystem->pwd(),"testFlowRate_res.dat");
    UranieTests::removeFile(gSystem->pwd(),"testFlowRate_sto.dat");
    UranieTests::removeFile(gSystem->pwd(),"anisp.isf");
    UranieTests::removeFile(gSystem->pwd(),"fichierAnisp_error.dat");
    UranieTests::removeFile(gSystem->pwd(),"fichierAnisp_indices.dat");
    UranieTests::removeFile(gSystem->pwd(),"testthrowExceptionrestartAnisp_2_data.dat");
    UranieTests::removeFile(gSystem->pwd(),"_dataserver_.root");
    UranieTests::removeFile(gSystem->pwd(),"tata.C");
    UranieTests::removeFile(gSystem->pwd(),"test_data.dat");
    UranieTests::removeFile(gSystem->pwd(),"test_error.dat");
    UranieTests::removeFile(gSystem->pwd(),"test_export_data.dat");
    UranieTests::removeFile(gSystem->pwd(),"test_export_error.dat");
    UranieTests::removeFile(gSystem->pwd(),"test_export_indices.dat");
    UranieTests::removeFile(gSystem->pwd(),"test_export_sto.dat");
    UranieTests::removeFile(gSystem->pwd(),"test_indices.dat");
    UranieTests::removeFile(gSystem->pwd(),"titi.C");
    UranieTests::removeFile(gSystem->pwd(),"toto.C");

}


///////////////////////////////////
//  TU                           //
///////////////////////////////////
// void TAnispTest::testgetNPatterns()
// {
//   // contructor initialization of the parameter _ns
//   CPPUNIT_ASSERT_EQUAL(0, _tdsFlowrate->getNPatterns());
//   
// 
// }

void TAnispTest::testAnispDefaultValues()
{
    TString str = "fichierAnisp_";
    // test default value of _anisp
    CPPUNIT_ASSERT_EQUAL(1, _anisp->getKMin());
    CPPUNIT_ASSERT_EQUAL(1000, _anisp->getNumberMaxOfSimulations());
    CPPUNIT_ASSERT_EQUAL(6, _anisp->getGreaterIndice());
    CPPUNIT_ASSERT_EQUAL(1000, _anisp->getMaxIndices());
    CPPUNIT_ASSERT_EQUAL(15, _anisp->getDegreeMax());
    CPPUNIT_ASSERT_EQUAL(0, _anisp->getIterationNumber());
    CPPUNIT_ASSERT(
            URANIE::Modeler::TAnisp::kAnalyticalFunction == _anisp->getCible());
    CPPUNIT_ASSERT(!_anisp->getLog());
    CPPUNIT_ASSERT_EQUAL(str, _anisp->getRootFilename());
    CPPUNIT_ASSERT_THROW(_anisp->getTPolynomialChaos(),
            URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.001, _anisp->getTolerance(),
            DBL_PRECISION_COMPARE);

    // test default value of _anips2
    CPPUNIT_ASSERT_EQUAL(1, _anisp2->getKMin());
    CPPUNIT_ASSERT_EQUAL(1000, _anisp2->getNumberMaxOfSimulations());
    CPPUNIT_ASSERT_EQUAL(6, _anisp2->getGreaterIndice());
    CPPUNIT_ASSERT_EQUAL(1000, _anisp2->getMaxIndices());
    CPPUNIT_ASSERT_EQUAL(15, _anisp2->getDegreeMax());
    CPPUNIT_ASSERT_EQUAL(0, _anisp2->getIterationNumber());
    CPPUNIT_ASSERT(URANIE::Modeler::TAnisp::kCode == _anisp2->getCible());
    CPPUNIT_ASSERT(!_anisp2->getLog());
    CPPUNIT_ASSERT_EQUAL(str, _anisp2->getRootFilename());
    CPPUNIT_ASSERT_THROW(_anisp2->getTPolynomialChaos(),
            URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.001, _anisp2->getTolerance(),
            DBL_PRECISION_COMPARE);

}

void TAnispTest::testgetKMin()
{
    // get function of the parameter _nKMin
    CPPUNIT_ASSERT_EQUAL(1, _anisp2->getKMin());

}

void TAnispTest::testsetKMin()
{
    // set function of the parameter _nKMin
    _anisp2->setKMin(2);
    CPPUNIT_ASSERT_EQUAL(2, _anisp2->getKMin());

}

void TAnispTest::testgetTolerance()
{
    // get function of the parameter _dTolerance
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.001, _anisp2->getTolerance(),
            DBL_PRECISION_COMPARE);

}

void TAnispTest::testsetTolerance()
{
    // set function of the parameter _dTolerance
    _anisp2->setTolerance(0.0001);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0001, _anisp2->getTolerance(),
            DBL_PRECISION_COMPARE);
}

void TAnispTest::testgetNumberMaxOfSimulations()
{
    //   get function of the parameter _nNumberMaxOfSimulations
    CPPUNIT_ASSERT_EQUAL(1000, _anisp2->getNumberMaxOfSimulations());
}

void TAnispTest::testsetNumberMaxOfSimulations()
{
    //   set function of the parameter _nNumberMaxOfSimulations
    _anisp2->setNumberMaxOfSimulations(2000);
    CPPUNIT_ASSERT_EQUAL(2000, _anisp2->getNumberMaxOfSimulations());
}

void TAnispTest::testgetGreaterIndice()
{
    //   get function of the parameter _nGreaterIndice
    CPPUNIT_ASSERT_EQUAL(6, _anisp2->getGreaterIndice());
}

void TAnispTest::testsetGreaterIndice()
{
    //   set function of the parameter _nGreaterIndice

    _anisp2->setGreaterIndice(3);
    CPPUNIT_ASSERT_EQUAL(3, _anisp2->getGreaterIndice());

    _anisp->setGreaterIndice(12);
    CPPUNIT_ASSERT_EQUAL(12, _anisp->getGreaterIndice());
}

void TAnispTest::testgetMaxIndices()
{
    //   get function of the parameter _nMaxIndices
    CPPUNIT_ASSERT_EQUAL(1000, _anisp2->getMaxIndices());
}

void TAnispTest::testsetMaxIndices()
{
    //   set function of the parameter _nMaxIndices
    _anisp2->setMaxIndices(5000);
    CPPUNIT_ASSERT_EQUAL(5000, _anisp2->getMaxIndices());
}

void TAnispTest::testgetDegreeMax()
{
    //   get function of the parameter _nDegreeMax
    CPPUNIT_ASSERT_EQUAL(15, _anisp2->getDegreeMax());
}

void TAnispTest::testsetDegreeMax()
{
    //   set function of the parameter _nDegreeMax
    _anisp2->setDegreeMax(5);
    CPPUNIT_ASSERT_EQUAL(5, _anisp2->getDegreeMax());
}

void TAnispTest::testgetCible()
{
    //   get function of the parameter _nCible
    CPPUNIT_ASSERT(
            URANIE::Modeler::TAnisp::kAnalyticalFunction == _anisp->getCible());
    CPPUNIT_ASSERT(URANIE::Modeler::TAnisp::kCode == _anisp2->getCible());
}

void TAnispTest::testgetIterationNumber()
{
    //   get function of the parameter _nIterationNumber
    CPPUNIT_ASSERT_EQUAL(0, _anisp2->getIterationNumber());
}

void TAnispTest::testgetRootFilename()
{
    //   get function of the parameter _sRootFilename
    TString str = "fichierAnisp_";
    CPPUNIT_ASSERT_EQUAL(str, _anisp2->getRootFilename());
}

void TAnispTest::testsetRootFilename()
{
    //   set function of the parameter _sRootFilename
    TString str = "testfilename_";
    _anisp2->setRootFilename(str);
    CPPUNIT_ASSERT_EQUAL(str, _anisp2->getRootFilename());
}

void TAnispTest::testgetDataAnispFile()
{
    //   get function of Data Anisp File name
    _anisp2->setRootFilename("toto_");
    TString str = "toto_data.dat";
    CPPUNIT_ASSERT_EQUAL(str, _anisp2->getDataAnispFile());
}

void TAnispTest::testgetResultFile()
{
    //   get function of Result File name
    _anisp2->setRootFilename("toto_");
    TString str = "toto_res.dat";
    CPPUNIT_ASSERT_EQUAL(str, _anisp2->getResultFile());
}

void TAnispTest::testgetStochasticFile()
{
    //   get function of Stochastic File name
    _anisp2->setRootFilename("toto_");
    TString str = "toto_sto.dat";
    CPPUNIT_ASSERT_EQUAL(str, _anisp2->getStochasticFile());
}

void TAnispTest::testgetErrorIndicatorFile()
{
    //   get function of Error Indicator File name
    _anisp2->setRootFilename("toto_");
    TString str = "toto_error.dat";
    CPPUNIT_ASSERT_EQUAL(str, _anisp2->getErrorIndicatorFile());
}

void TAnispTest::testgetIndicesFile()
{
    //   get function of Indices File name
    _anisp2->setRootFilename("toto_");
    TString str = "toto_indices.dat";
    CPPUNIT_ASSERT_EQUAL(str, _anisp2->getIndicesFile());
}

void TAnispTest::testsetlog()
{
    //   set function of the parameter _blog
    _anisp2->setlog(kTRUE);
    CPPUNIT_ASSERT(_anisp2->getLog());
    _anisp2->setlog(kFALSE);
    CPPUNIT_ASSERT(!_anisp2->getLog());
}

void TAnispTest::testsetAllAnispParameters()
{
    TString str = "testsetAllAnispParameters_";
    _anisp2->setAllAnispParameters(0.0001, 2, 2000, 3, 5000, kTRUE, 5, str);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0001, _anisp2->getTolerance(),
    DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_EQUAL(2, _anisp2->getKMin());
    CPPUNIT_ASSERT_EQUAL(2000, _anisp2->getNumberMaxOfSimulations());
    CPPUNIT_ASSERT_EQUAL(3, _anisp2->getGreaterIndice());
    CPPUNIT_ASSERT_EQUAL(5000, _anisp2->getMaxIndices());
    CPPUNIT_ASSERT(_anisp2->getLog());
    CPPUNIT_ASSERT_EQUAL(5, _anisp2->getDegreeMax());
    CPPUNIT_ASSERT_EQUAL(str, _anisp2->getRootFilename());
}

void TAnispTest::testsetAnispParameters()
{
    _anisp2->setAnispParameters(0.0001, 2, 2000, 3, 5000, kTRUE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0001, _anisp2->getTolerance(),
    DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_EQUAL(2, _anisp2->getKMin());
    CPPUNIT_ASSERT_EQUAL(2000, _anisp2->getNumberMaxOfSimulations());
    CPPUNIT_ASSERT_EQUAL(3, _anisp2->getGreaterIndice());
    CPPUNIT_ASSERT_EQUAL(5000, _anisp2->getMaxIndices());
    CPPUNIT_ASSERT(_anisp2->getLog());
}

void TAnispTest::testsetLightAnispParameters()
{
    _anisp2->setLightAnispParameters(0.0001, 2, 2000);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0001, _anisp2->getTolerance(),
    DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_EQUAL(2, _anisp2->getKMin());
    CPPUNIT_ASSERT_EQUAL(2000, _anisp2->getNumberMaxOfSimulations());
}

void TAnispTest::testchangeOfVariable()
{
    vector<Double_t*> vptExp;
    Double_t* x1 = new Double_t[4];
    Double_t* x2 = new Double_t[4];
    x1[0] = 1;
    x1[1] = 1;
    x1[2] = 1;
    x2[0] = -1;
    x2[1] = -1;
    x2[2] = -1;
    vptExp.push_back(x1);
    vptExp.push_back(x2);

    gUranieTAnisp = _anisp;
    URANIE::Modeler::functionLauncher(vptExp);

    Double_t* vx; //   = new Double_t[5];
    vx = _tdsIshigami->getTuple()->GetArgs();
    _tdsIshigami->getTuple()->GetEntry(0);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.141592654, vx[1], 0.000000001);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.141592654, vx[2], 0.000000001);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.141592654, vx[3], 0.000000001);
    _tdsIshigami->getTuple()->GetEntry(1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-3.141592654, vx[1], 0.000000001);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-3.141592654, vx[2], 0.000000001);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-3.141592654, vx[3], 0.000000001);

    // Clean
    delete[] x1;
    x1 = NULL;
    delete[] x2;
    x2 = NULL;
}

void TAnispTest::testgetTPolynomialChaos()
{
    TString str = "testIshigami_";
    _anisp->setAllAnispParameters(0.001, 2, 500, 6, 5000, kFALSE, 5, str);
    _anisp->runAnisp();
    URANIE::Modeler::TPolynomialChaos * pc;
    URANIE::Modeler::TPolynomialChaos P1(_anisp->getTPolynomialChaos());
    pc = &P1;

    std::ifstream result(_anisp->getResultFile().Data(), ios::in);
    Int_t num, i;
    Double_t coef, buff;
    while (!result.eof())
    {
        result >> num;
        result >> coef;
        if (result.fail())
            break;
        CPPUNIT_ASSERT_DOUBLES_EQUAL(coef, pc->getCoefficient(1, num),
                DBL_PRECISION_COMPARE);

        for (i = 0; i < 3; i++)
        {
            result >> buff;
        }

    }

    result.close();
}

void TAnispTest::testrunAnisp()
{
    TString str = "testIshigami_";
    _anisp->setAllAnispParameters(0.001, 2, 500, 5, 5000, kFALSE, 5, str);

    _anisp->runAnisp();

    URANIE::Modeler::TPolynomialChaos * pc;
    URANIE::Modeler::TPolynomialChaos P1(_anisp->getTPolynomialChaos());
    pc = &P1;

    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.5000000000000004441,
            pc->getCoefficient(1, 0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.40917673573214308202,
            pc->getCoefficient(1, 49), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_EQUAL(423, _tdsIshigami->getNPatterns());

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0, _tdsIshigami->getValue("x1", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0, _tdsIshigami->getValue("x2", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0, _tdsIshigami->getValue("x3", 0),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.993005281e+00,
            _tdsIshigami->getValue("x1", _tdsIshigami->getNPatterns() / 2),
            DBL_PRECISION_COMPARE2);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0,
            _tdsIshigami->getValue("x2", _tdsIshigami->getNPatterns() / 2),
            DBL_PRECISION_COMPARE2);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.221441469e+00,
            _tdsIshigami->getValue("x3", _tdsIshigami->getNPatterns() / 2),
            DBL_PRECISION_COMPARE2);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(-2.221441469e+00,
            _tdsIshigami->getValue("x1", _tdsIshigami->getNPatterns() - 1),
            DBL_PRECISION_COMPARE2);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0,
            _tdsIshigami->getValue("x2", _tdsIshigami->getNPatterns() - 1),
            DBL_PRECISION_COMPARE2);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-3.126465026e+00,
            _tdsIshigami->getValue("x3", _tdsIshigami->getNPatterns() - 1),
            DBL_PRECISION_COMPARE2);

    CPPUNIT_ASSERT(_tdsIshigami->isAttribute("__uranie__anisp__"));

    _tdsIshigami->computeStatistic("__uranie__anisp__");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1,
            _tdsIshigami->getAttribute("__uranie__anisp__")->getMaximum(),
            DBL_PRECISION_COMPARE);

    //////////////////////////////////////////////////////////////////////////:

    TString str2 = "testFlowRate_";
    _anisp2->setAllAnispParameters(0.1, 1, 500, 3, 5000, kFALSE, 5, str2);
	
    _anisp2->runAnisp();

    URANIE::Modeler::TPolynomialChaos P2(_anisp2->getTPolynomialChaos());
    pc = &P2;

    CPPUNIT_ASSERT_DOUBLES_EQUAL(77.647675585029730883,
            pc->getCoefficient(1, 0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.012948275333039389878,
            pc->getCoefficient(1, 171), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_EQUAL(285, _tdsFlowrate->getNPatterns());

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.000000000e-01,
            _tdsFlowrate->getValue("rw", 0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.505000000e+04,
            _tdsFlowrate->getValue("r", 0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(8.933500000e+04,
            _tdsFlowrate->getValue("tu", 0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(8.955000000e+01,
            _tdsFlowrate->getValue("tl", 0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.050000000e+03,
            _tdsFlowrate->getValue("hu", 0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(7.600000000e+02,
            _tdsFlowrate->getValue("hl", 0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.400000000e+03,
            _tdsFlowrate->getValue("l", 0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.095000000e+04,
            _tdsFlowrate->getValue("kw", 0), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.000000000e-01,
            _tdsFlowrate->getValue("rw", _tdsFlowrate->getNPatterns() / 2),
            DBL_PRECISION_COMPARE2);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(4.269231419e+04,
            _tdsFlowrate->getValue("r", _tdsFlowrate->getNPatterns() / 2),
            DBL_PRECISION_COMPARE2 * 1.0e+4);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(8.933500000e+04,
            _tdsFlowrate->getValue("tu", _tdsFlowrate->getNPatterns() / 2),
            DBL_PRECISION_COMPARE2 * 1.0e+4);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(8.955000000e+01,
            _tdsFlowrate->getValue("tl", _tdsFlowrate->getNPatterns() / 2),
            DBL_PRECISION_COMPARE2 * 1.0e+1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.050000000e+03,
            _tdsFlowrate->getValue("hu", _tdsFlowrate->getNPatterns() / 2),
            DBL_PRECISION_COMPARE2 * 1.0e+3);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(7.175735931e+02,
            _tdsFlowrate->getValue("hl", _tdsFlowrate->getNPatterns() / 2),
            DBL_PRECISION_COMPARE2 * 1.0e+2);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.400000000e+03,
            _tdsFlowrate->getValue("l", _tdsFlowrate->getNPatterns() / 2),
            DBL_PRECISION_COMPARE2 * 1.0e+3);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.095000000e+04,
            _tdsFlowrate->getValue("kw", _tdsFlowrate->getNPatterns() / 2),
            DBL_PRECISION_COMPARE2 * 1.0e+4);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.000000000e-01,
            _tdsFlowrate->getValue("rw", _tdsFlowrate->getNPatterns() - 1),
            DBL_PRECISION_COMPARE2);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.505000000e+04,
            _tdsFlowrate->getValue("r", _tdsFlowrate->getNPatterns() - 1),
            DBL_PRECISION_COMPARE2 * 1.0e+4);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(8.933500000e+04,
            _tdsFlowrate->getValue("tu", _tdsFlowrate->getNPatterns() - 1),
            DBL_PRECISION_COMPARE2 * 1.0e+4);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(8.955000000e+01,
            _tdsFlowrate->getValue("tl", _tdsFlowrate->getNPatterns() - 1),
            DBL_PRECISION_COMPARE2 * 1.0e+1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.050000000e+03,
            _tdsFlowrate->getValue("hu", _tdsFlowrate->getNPatterns() - 1),
            DBL_PRECISION_COMPARE2 * 1.0e+3);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(7.600000000e+02,
            _tdsFlowrate->getValue("hl", _tdsFlowrate->getNPatterns() - 1),
            DBL_PRECISION_COMPARE2 * 1.0e+2);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.400000000e+03,
            _tdsFlowrate->getValue("l", _tdsFlowrate->getNPatterns() - 1),
            DBL_PRECISION_COMPARE2 * 1.0e+3);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(9.938351912e+03,
            _tdsFlowrate->getValue("kw", _tdsFlowrate->getNPatterns() - 1),
            DBL_PRECISION_COMPARE2 * 1.0e+3);

    CPPUNIT_ASSERT(_tdsFlowrate->isAttribute("__uranie__anisp__"));

    _tdsFlowrate->computeStatistic("__uranie__anisp__");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1,
            _tdsFlowrate->getAttribute("__uranie__anisp__")->getMaximum(),
            DBL_PRECISION_COMPARE);

    ///////////////////////////////////////////////////////////////////////

    TString str3 = "testLinear_";
    _anisp3->setAllAnispParameters(0.001, 1, 500, 3, 5000, kFALSE, 5, str3);
    _anisp3->runAnisp();

    URANIE::Modeler::TPolynomialChaos P3(_anisp3->getTPolynomialChaos());
    pc = &P3;

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, pc->getCoefficient(1, 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(19.595917942266350309,
            pc->getCoefficient(1, 15), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_EQUAL(25, _tdsLinear->getNPatterns());

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0, _tdsLinear->getValue("x1", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0, _tdsLinear->getValue("x2", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0, _tdsLinear->getValue("x3", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0, _tdsLinear->getValue("x4", 0),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(-1.755827474e+01,
            _tdsLinear->getValue("x1", _tdsLinear->getNPatterns() / 2),
            DBL_PRECISION_COMPARE2 * 1.0e+1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.000000000e+00,
            _tdsLinear->getValue("x2", _tdsLinear->getNPatterns() / 2),
            DBL_PRECISION_COMPARE2);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.000000000e+00,
            _tdsLinear->getValue("x3", _tdsLinear->getNPatterns() / 2),
            DBL_PRECISION_COMPARE2);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.000000000e+00,
            _tdsLinear->getValue("x4", _tdsLinear->getNPatterns() / 2),
            DBL_PRECISION_COMPARE2);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(-1.041224779e+01,
            _tdsLinear->getValue("x1", _tdsLinear->getNPatterns() - 1),
            DBL_PRECISION_COMPARE2 * 1.0e+1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.000000000e+00,
            _tdsLinear->getValue("x2", _tdsLinear->getNPatterns() - 1),
            DBL_PRECISION_COMPARE2);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.000000000e+00,
            _tdsLinear->getValue("x3", _tdsLinear->getNPatterns() - 1),
            DBL_PRECISION_COMPARE2);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-2.603061948e+00,
            _tdsLinear->getValue("x4", _tdsLinear->getNPatterns() - 1),
            DBL_PRECISION_COMPARE2);

    CPPUNIT_ASSERT(_tdsLinear->isAttribute("__uranie__anisp__"));

    _tdsLinear->computeStatistic("__uranie__anisp__");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1,
            _tdsLinear->getAttribute("__uranie__anisp__")->getMaximum(),
            DBL_PRECISION_COMPARE);

    /////////////////////////////////////////////////////////////////

    TString str4 = "testVenise_";
    _anisp4->setAllAnispParameters(0.0000001, 2, 500, 5, 5000, kFALSE, 10,
            str4);
    _anisp4->runAnisp();

    URANIE::Modeler::TPolynomialChaos P4(_anisp4->getTPolynomialChaos());
    pc = &P4;

    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.2099051662958193454,
            pc->getCoefficient(1, 0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.9238792770568409196e-11,
            pc->getCoefficient(1, 54), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_EQUAL(97, _tdsVenise->getNPatterns());

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0, _tdsVenise->getValue("x1", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0, _tdsVenise->getValue("x2", 0),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(-4.903926402e-01,
            _tdsVenise->getValue("x1", _tdsVenise->getNPatterns() / 2),
            DBL_PRECISION_COMPARE2 * 1.0e-1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(5.551115123e-17,
            _tdsVenise->getValue("x2", _tdsVenise->getNPatterns() / 2),
            DBL_PRECISION_COMPARE2 * 1.0e-17);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(5.551115123e-17,
            _tdsVenise->getValue("x1", _tdsVenise->getNPatterns() - 1),
            DBL_PRECISION_COMPARE2 * 1.0e-17);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-4.975923633e-01,
            _tdsVenise->getValue("x2", _tdsVenise->getNPatterns() - 1),
            DBL_PRECISION_COMPARE2 * 1.0e-1);

    CPPUNIT_ASSERT(_tdsVenise->isAttribute("__uranie__anisp__"));

    _tdsVenise->computeStatistic("__uranie__anisp__");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1,
            _tdsVenise->getAttribute("__uranie__anisp__")->getMaximum(),
            DBL_PRECISION_COMPARE);
}

void TAnispTest::testrestartAnisp()
{

    TString str = "testIshigami_";
    _anisp->setAllAnispParameters(0.01, 2, 100, 6, 500, kTRUE, 10, str);
    _anisp->runAnisp();
    _anisp->setLightAnispParameters(0.0001, 2, 1000);
//   _anisp->setlog(kTRUE);
    _anisp->restartAnisp();

    URANIE::Modeler::TPolynomialChaos * pc;
    URANIE::Modeler::TPolynomialChaos P1(_anisp->getTPolynomialChaos());
    pc = &P1;
cout<<" number of coefficients = "<<pc->getDimensionExpansion()<<endl;
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.5000000000000004441,
            pc->getCoefficient(1, 0), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.045914725892680793851,
            pc->getCoefficient(1, 275), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_EQUAL(775, _tdsIshigami->getNPatterns());

    CPPUNIT_ASSERT(_tdsIshigami->isAttribute("__uranie__anisp__"));

    _tdsIshigami->computeStatistic("__uranie__anisp__");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2,
            _tdsIshigami->getAttribute("__uranie__anisp__")->getMaximum(),
            DBL_PRECISION_COMPARE);
    
	
    _tdsIshigami->exportData("_exportDataIshigami_.dat",
			     _anisp->getVariablesName(),"","internal");

    TString _anispVariableName = _anisp->getVariablesName();
    delete _anisp;
    _anisp=NULL;
    
    delete _tdsIshigami;
    _tdsIshigami=NULL;

    // test the restart with another TAnisp Object

    TDataServer* tdsIshigami = new URANIE::DataServer::TDataServer(
            "tdsIshigami", "TDS for Ishigami Use Case");
    tdsIshigami->addAttribute(
            new URANIE::DataServer::TUniformDistribution("x1",
                    -1. * TMath::Pi(), TMath::Pi()));
    tdsIshigami->addAttribute(
            new URANIE::DataServer::TUniformDistribution("x2",
                    -1. * TMath::Pi(), TMath::Pi()));
    tdsIshigami->addAttribute(
            new URANIE::DataServer::TUniformDistribution("x3",
                    -1. * TMath::Pi(), TMath::Pi()));
    tdsIshigami->getAttribute("x1")->delShare();
    tdsIshigami->getAttribute("x2")->delShare();
    tdsIshigami->getAttribute("x3")->delShare();

    TDataServer *tdstmp = new TDataServer("tdsIshigami",
            "TDS for Ishigami Use Case");
    tdstmp->fileDataRead("_exportDataIshigami_.dat");

    TString LaunchNumber = TString("__uranie__anisp__");
    TAttribute * att = new TAttribute(LaunchNumber);
    att->setOrigin(URANIE::DataServer::TAttribute::kInternal);
    att->delShare();
    tdsIshigami->addAttribute(att);
    tdsIshigami->setTuple(tdstmp->getTuple());

    URANIE::Modeler::TAnisp * anisp = new URANIE::Modeler::TAnisp(tdsIshigami,
#ifdef ROOT5
            _p2fIshigami, "x1:x2:x3", "");
#else
            &FctIshigami, "FctIshigami", "x1:x2:x3", "");
#endif
    anisp->setDrawProgressBar(kFALSE);
    anisp->setRootFilename(str);
    anisp->setKMin(2);
    anisp->setNumberMaxOfSimulations(3000);
    anisp->setTolerance(0.0000000000000000001);
//   anisp->setlog(kTRUE);
    anisp->restartAnisp();
    tdsIshigami->exportData("_exportDataIshigami2_.dat",
			    _anispVariableName);
    URANIE::Modeler::TPolynomialChaos P2(anisp->getTPolynomialChaos());
    pc = &P2;

    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.5000000000000008882,
            pc->getCoefficient(1, 0), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(-2.7141838864088682929e-06,
            pc->getCoefficient(1, 694), DBL_PRECISION_COMPARE);


    CPPUNIT_ASSERT_EQUAL(1511, tdsIshigami->getNPatterns());

    CPPUNIT_ASSERT(tdsIshigami->isAttribute("__uranie__anisp__"));

    tdsIshigami->computeStatistic("__uranie__anisp__");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3,
            tdsIshigami->getAttribute("__uranie__anisp__")->getMaximum(),
            DBL_PRECISION_COMPARE);
    
    delete anisp;
    delete tdsIshigami;
    delete tdstmp;
}

void TAnispTest::testgetErrorIndicator()
{
    TString str = "testIshigami_";
    _anisp->setAllAnispParameters(0.0001, 2, 500, 6, 5000, kTRUE, 10, str);
    _anisp->runAnisp();
    
    string tmpstr = string(_anisp->getErrorIndicatorFile());
    const char* file = tmpstr.c_str();
//     const char* file = string(_anisp->getErrorIndicatorFile()).c_str();
    std::ifstream result(file, ios::in);
    Int_t num, i;
    Double_t err;
    TNtupleD * ErrorIndicator = _anisp->getErrorIndicator();
    Double_t * terrind = ErrorIndicator->GetArgs();
    i = 0;
    while (!result.eof())
    {
        result >> num;
        result >> err;
        if (result.fail())
            break;
        ErrorIndicator->GetEntry(i);
        CPPUNIT_ASSERT_EQUAL(num, (Int_t)(terrind[0]));
        CPPUNIT_ASSERT_DOUBLES_EQUAL(err, terrind[1], DBL_PRECISION_COMPARE);
        i += 1;
    }

    result.close();
}

void TAnispTest::testgetIntegrationIndices()
{
    TString str = "test_";
    _anisp->setAllAnispParameters(0.01, 2, 100, 3, 5000, kTRUE, 10, str);
    _anisp->runAnisp();


    TNtupleD * IntegInd = _anisp->getIntegrationIndices();
    Double_t * terrind = IntegInd->GetArgs();

    IntegInd->GetEntry(0);
    CPPUNIT_ASSERT_EQUAL(1, (Int_t)(terrind[0]));
    CPPUNIT_ASSERT_EQUAL(1, (Int_t)(terrind[1]));
    CPPUNIT_ASSERT_EQUAL(1, (Int_t)(terrind[2]));

    IntegInd->GetEntry(1);
    CPPUNIT_ASSERT_EQUAL(2, (Int_t)(terrind[0]));
    CPPUNIT_ASSERT_EQUAL(1, (Int_t)(terrind[1]));
    CPPUNIT_ASSERT_EQUAL(1, (Int_t)(terrind[2]));

    IntegInd->GetEntry(7);
    CPPUNIT_ASSERT_EQUAL(1, (Int_t)(terrind[0]));
    CPPUNIT_ASSERT_EQUAL(3, (Int_t)(terrind[1]));
    CPPUNIT_ASSERT_EQUAL(1, (Int_t)(terrind[2]));

    _anisp->changeLog();

}

void TAnispTest::testexportModelCplusplus()
{
    _anisp->setRootFilename("testIshigami_");
//     cout<<"******testexportModelCplusplus***********"<<_anisp->getResultFile()<<endl;
    _anisp->setKMin(2);
    _anisp->runAnisp();
    _anisp->exportFunction("c++", "toto", "toto");
    gROOT->LoadMacro("toto.C");
    Double_t* param = new Double_t[7];
    Double_t* res = new Double_t[1];
    Double_t* res2 = new Double_t[1];

#ifdef ROOT5
    G__ClassInfo globalscope;
    G__MethodInfo method;
    long dummy;
    method = globalscope.GetMethod("toto", "double*,double*", &dummy);
    void * _p2ftoto = method.PointerToFunc();
#endif
    URANIE::DataServer::TDataServer* _tdsIshigami2;
    _tdsIshigami2 = new URANIE::DataServer::TDataServer("tdsIshigami",
            "TDS for Ishigami Use Case");
    _tdsIshigami2->addAttribute(
            new URANIE::DataServer::TUniformDistribution("x1",
                    -1. * TMath::Pi(), TMath::Pi()));
    _tdsIshigami2->addAttribute(
            new URANIE::DataServer::TUniformDistribution("x2",
                    -1. * TMath::Pi(), TMath::Pi()));
    _tdsIshigami2->addAttribute(
            new URANIE::DataServer::TUniformDistribution("x3",
                    -1. * TMath::Pi(), TMath::Pi()));
    _tdsIshigami2->createTuple();
    _tdsIshigami2->getAttribute("x1")->delShare();
    _tdsIshigami2->getAttribute("x2")->delShare();
    _tdsIshigami2->getAttribute("x3")->delShare();
    
    param[0] = 1;
    param[1] = 2;
    param[2] = 1;
    param[3] = -0.5;
    param[4] = -0.9;
    param[5] = 0;
    param[6] = 0;
    _tdsIshigami2->getTuple()->Fill(param);
    URANIE::Launcher::TLauncherFunction *_tlf=NULL;
#ifdef ROOT5
    _tlf = new URANIE::Launcher::TLauncherFunction(_tdsIshigami2, _p2ftoto,
#else
    _tlf = new URANIE::Launcher::TLauncherFunction(_tdsIshigami2, "toto",
#endif
            "x1:x2:x3", "y");
    _tlf->setDrawProgressBar(kFALSE);
    _tlf->run();
    res[0] = _tdsIshigami2->getValue("y", 0);

    _tdsIshigami->createTuple();
    _tdsIshigami->getTuple()->Fill(param);
    URANIE::Launcher::TLauncherFunction *_tlf2;
#ifdef ROOT5
    _tlf2 = new URANIE::Launcher::TLauncherFunction(_tdsIshigami, _p2fIshigami,
#else
    _tlf2 = new URANIE::Launcher::TLauncherFunction(_tdsIshigami, &FctIshigami,
#endif
            "x1:x2:x3", "y");
    _tlf2->setDrawProgressBar(kFALSE);
    _tlf2->run();
    res2[0] = _tdsIshigami->getValue("y", 0);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(res2[0], res[0], 0.001);

    //////////////////////////////////////////////////////////
    _anisp3->setRootFilename("testLinear_");
//     cout<<_anisp3->getResultFile()<<endl;
    _anisp3->setKMin(1);
    _anisp3->setNumberMaxOfSimulations(3000);
    _anisp3->runAnisp();
    _anisp3->exportFunction("c++", "titi", "titi");
    gROOT->LoadMacro("titi.C");

#ifdef ROOT5
    G__ClassInfo globalscope2;
    G__MethodInfo method2;
    long dummy2;
    method2 = globalscope2.GetMethod("titi", "double*,double*", &dummy2);
    void * _p2ftiti = method2.PointerToFunc();
#endif
    URANIE::DataServer::TDataServer* _tdsLinear2;
    _tdsLinear2 = new URANIE::DataServer::TDataServer("tdsLinear",
            "TDS for Linear Use Case");
    _tdsLinear2->addAttribute(
            new URANIE::DataServer::TNormalDistribution("x1", 0, 2.));
    _tdsLinear2->addAttribute(
            new URANIE::DataServer::TNormalDistribution("x2", 0., 2.));
    _tdsLinear2->addAttribute(
            new URANIE::DataServer::TNormalDistribution("x3", 0., 3.));
    _tdsLinear2->addAttribute(
            new URANIE::DataServer::TNormalDistribution("x4", 0., 0.5));
    _tdsLinear2->createTuple();
    _tdsLinear2->getAttribute("x1")->delShare();
    _tdsLinear2->getAttribute("x2")->delShare();
    _tdsLinear2->getAttribute("x3")->delShare();
    _tdsLinear2->getAttribute("x4")->delShare();
	
    param[0] = 1;
    param[1] = 1;
    param[2] = 2;
    param[3] = -0.5;
    param[4] = -2.9;
    param[5] = 0;
    param[6] = 0;
    _tdsLinear2->getTuple()->Fill(param);
    URANIE::Launcher::TLauncherFunction *_tlf3;
#ifdef ROOT5
    _tlf3 = new URANIE::Launcher::TLauncherFunction(_tdsLinear2, _p2ftiti,
#else
    _tlf3 = new URANIE::Launcher::TLauncherFunction(_tdsLinear2, "titi",
#endif
            "x1:x2:x3:x4", "y");
    _tlf3->setDrawProgressBar(kFALSE);
    _tlf3->run();
    res[0] = _tdsLinear2->getValue("y", 0);

    _tdsLinear->createTuple();
    _tdsLinear->getTuple()->Fill(param);
    URANIE::Launcher::TLauncherFunction *_tlf4;
#ifdef ROOT5
    _tlf4 = new URANIE::Launcher::TLauncherFunction(_tdsLinear, _p2fLinear,
#else
    _tlf4 = new URANIE::Launcher::TLauncherFunction(_tdsLinear, &FctLinear,
#endif
            "x1:x2:x3:x4", "y");
    _tlf4->setDrawProgressBar(kFALSE);
    _tlf4->run();
    res2[0] = _tdsLinear->getValue("y", 0);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(res2[0], res[0], 0.000001);

    delete[] param;
    param = NULL;
    delete[] res;
    res = NULL;
    delete[] res2;
    res2 = NULL;
    delete _tlf;
    delete _tlf2;
    delete _tlf3;
    delete _tlf4;
    delete _tdsLinear2;
    delete _tdsIshigami2;
}

void TAnispTest::testgetVariablesName()
{

    TString str = "rw:r:tu:tl:hu:hl:l:kw:yhat:__uranie__anisp__";
    CPPUNIT_ASSERT_EQUAL(str, _anisp2->getVariablesName());
    str = "rw:r:tu:tl:hu:yhat:__uranie__anisp__";
    CPPUNIT_ASSERT_EQUAL(str, _anisp2->getVariablesName("rw:r:tu:tl:hu"));
}

///////////////////////////////////
//  Exceptions                   //
///////////////////////////////////

void TAnispTest::throwExceptionNonSupportedLaw()
{
    // Error in the laws associated with the TDataServer attribute
    // Define the DataServer
    TDataServer *tds2 = new TDataServer("tdsflowreate", "DataBase flowreate");
    tds2->addAttribute(new TUniformDistribution("rw", 0.05, 0.15));
    tds2->addAttribute(new TUniformDistribution("r", 100.0, 50000.0));
    tds2->addAttribute(new TUniformDistribution("tu", 63070.0, 115600.0));
    tds2->addAttribute(new TUniformDistribution("tl", 63.1, 116.0));
    tds2->addAttribute(new TUniformDistribution("hu", 990.0, 1110.0));
    tds2->addAttribute(new TUniformDistribution("hl", 700.0, 820.0));
    tds2->addAttribute(new TUniformDistribution("l", 1120.0, 1680.0));
    tds2->addAttribute(new TLogUniformDistribution("kw", 9855.0, 12045.0));

    //Input File
    TString sJDDReferenceKey = gSystem->pwd();
    sJDDReferenceKey += TString("/flowrate_input_with_row.in");

    tds2->getAttribute("rw")->setFileKey(sJDDReferenceKey, "Rw", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("r")->setFileKey(sJDDReferenceKey, "R", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("tu")->setFileKey(sJDDReferenceKey, "Tu", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("tl")->setFileKey(sJDDReferenceKey, "Tl", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("hu")->setFileKey(sJDDReferenceKey, "Hu", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("hl")->setFileKey(sJDDReferenceKey, "Hl", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("l")->setFileKey(sJDDReferenceKey, "L", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("kw")->setFileKey(sJDDReferenceKey, "Kw", "%f",
            TAttributeFileKey::kNewRow);

    
    TOutputFileDataServer *foutDataServer = new TOutputFileDataServer(
            "_output_flowrate_withRow_.dat");
    foutDataServer->addAttribute("yhat");

    TCode *codeFlowrate = new TCode(tds2, Form("%s/%s", gSystem->pwd(),
                            "../../../../launCHER/teST/coDE/flowREATE/flowrate -r flowrate_input_with_row.in >>/dev/null"));
    codeFlowrate->addOutputFile(foutDataServer);

    
    CPPUNIT_ASSERT_THROW(
            URANIE::Modeler::TAnisp(tds2, codeFlowrate, "myOption"),
            URANIE::Exceptions::UErrorExceptions);
    // Clean
    tds2->getAttribute("rw")->delShare();
    tds2->getAttribute("r")->delShare();
    tds2->getAttribute("tu")->delShare();
    tds2->getAttribute("tl")->delShare();
    tds2->getAttribute("hu")->delShare();
    tds2->getAttribute("hl")->delShare();
    tds2->getAttribute("l")->delShare();
    tds2->getAttribute("kw")->delShare();
    
    delete foutDataServer;
    delete codeFlowrate;
    delete tds2;
}

void TAnispTest::throwExceptionFunctionNull()
{
    // Error in the value of analytical function = NULL
#ifdef ROOT5
	CPPUNIT_ASSERT_THROW(URANIE::Modeler::TAnisp(_tdsIshigami, (void*)NULL, "", "", ""),
		URANIE::Exceptions::UErrorExceptions);
#else
	CPPUNIT_ASSERT_THROW(URANIE::Modeler::TAnisp(_tdsIshigami, (void(*)(double*, double*)) NULL, "", "NULL", ""),
		URANIE::Exceptions::UErrorExceptions);
#endif
}

void TAnispTest::throwExceptionTCodeNull()
{
    // Error in the value of _Code = NULL
    TCode *tc = NULL;
//  CPPUNIT_ASSERT_THROW( URANIE::Modeler::TAnisp(_tdsFlowrate, tc,"rw:r:tu:tl:hu:l:kw", "myOption") ,
//                URANIE::Exceptions::UErrorExceptions );
    CPPUNIT_ASSERT_THROW(URANIE::Modeler::TAnisp(_tdsFlowrate, tc, "myOption"),
            URANIE::Exceptions::UErrorExceptions);
}

void TAnispTest::throwExceptionTCodeSeveralOutputs()
{
    // Error in the number of outputs

// Define the DataServer
    TDataServer *tds2 = new TDataServer("tdsflowreate", "DataBase flowreate");
    tds2->addAttribute(new TUniformDistribution("rw", 0.05, 0.15));
    tds2->addAttribute(new TUniformDistribution("r", 100.0, 50000.0));
    tds2->addAttribute(new TUniformDistribution("tu", 63070.0, 115600.0));
    tds2->addAttribute(new TUniformDistribution("tl", 63.1, 116.0));
    tds2->addAttribute(new TUniformDistribution("hu", 990.0, 1110.0));
    tds2->addAttribute(new TUniformDistribution("hl", 700.0, 820.0));
    tds2->addAttribute(new TUniformDistribution("l", 1120.0, 1680.0));
    tds2->addAttribute(new TNormalDistribution("kw", 9855.0, 12045.0));

    TString sJDDReferenceKey = gSystem->pwd();
    sJDDReferenceKey += TString("/flowrate_input_with_row.in");

    tds2->getAttribute("rw")->setFileKey(sJDDReferenceKey, "Rw", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("r")->setFileKey(sJDDReferenceKey, "R", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("tu")->setFileKey(sJDDReferenceKey, "Tu", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("tl")->setFileKey(sJDDReferenceKey, "Tl", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("hu")->setFileKey(sJDDReferenceKey, "Hu", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("hl")->setFileKey(sJDDReferenceKey, "Hl", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("l")->setFileKey(sJDDReferenceKey, "L", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("kw")->setFileKey(sJDDReferenceKey, "Kw", "%f",
            TAttributeFileKey::kNewRow);

    //Define two outputs                       
    TOutputFileDataServer *fout = new TOutputFileDataServer(
            "_output_flowrate_withRow_.dat");
    fout->addAttribute("yhat");
    fout->addAttribute("d");

//  _codeFlowrate =  new TCode(_tdsFlowrate, Form("%s/%s", (gSystem->pwd()).Data(), "../../../../launCHER/teST/coDE/flowREATE/flowrate -r flowrate_input_with_row.in >>/dev/null");

    TCode * mycode2 =
            new TCode(tds2,
                    Form("%s/%s", gSystem->pwd(),
                            "../../../../launCHER/teST/coDE/flowREATE/flowrate -r flowrate_input_with_row.in >>/dev/null"));
    mycode2->addOutputFile(fout);

    //Error in the TDataServer
    CPPUNIT_ASSERT_THROW(URANIE::Modeler::TAnisp(tds2, mycode2),
            URANIE::Exceptions::UErrorExceptions);

    // Clean
    tds2->getAttribute("rw")->delShare();
    tds2->getAttribute("r")->delShare();
    tds2->getAttribute("tu")->delShare();
    tds2->getAttribute("tl")->delShare();
    tds2->getAttribute("hu")->delShare();
    tds2->getAttribute("hl")->delShare();
    tds2->getAttribute("l")->delShare();
    tds2->getAttribute("kw")->delShare();

    delete fout;
    delete mycode2;
    delete tds2;
    
}

void TAnispTest::throwExceptionsetKMinMethod()
{
    //Error in the value of _nKMin (must be > 0)
    CPPUNIT_ASSERT_THROW(_anisp2->setKMin(0),
            URANIE::Exceptions::UErrorExceptions);
    //Error in the value of _nKMin (must be <= _nX)
    CPPUNIT_ASSERT_THROW(_anisp2->setKMin(12),
            URANIE::Exceptions::UErrorExceptions);
}

void TAnispTest::throwExceptionsetToleranceMethod()
{
    //Error in the value of _dTolerance (must be > 0)
    CPPUNIT_ASSERT_THROW(_anisp2->setTolerance(0),
            URANIE::Exceptions::UErrorExceptions);
}

void TAnispTest::throwExceptionsetNumberMaxOfSimulationsMethod()
{
    //Error in the value of _nNumberMaxOfSimulations (must be > 0)
    CPPUNIT_ASSERT_THROW(_anisp2->setNumberMaxOfSimulations(0),
            URANIE::Exceptions::UErrorExceptions);
}

void TAnispTest::throwExceptionsetGreaterIndiceMethod()
{
    //Error in the value of _nGreaterIndice (must be > 0)
    CPPUNIT_ASSERT_THROW(_anisp2->setGreaterIndice(0),
            URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(_anisp2->setGreaterIndice(13),
            URANIE::Exceptions::UErrorExceptions);
    // Define the DataServer
    TDataServer *tds2 = new TDataServer("tdsflowreate", "DataBase flowreate");
    tds2->addAttribute(new TUniformDistribution("rw", 0.05, 0.15));
    tds2->addAttribute(new TUniformDistribution("r", 100.0, 50000.0));
    tds2->addAttribute(new TUniformDistribution("tu", 63070.0, 115600.0));
    tds2->addAttribute(new TUniformDistribution("tl", 63.1, 116.0));
    tds2->addAttribute(new TUniformDistribution("hu", 990.0, 1110.0));
    tds2->addAttribute(new TUniformDistribution("hl", 700.0, 820.0));
    tds2->addAttribute(new TUniformDistribution("l", 1120.0, 1680.0));
    tds2->addAttribute(new TNormalDistribution("kw", 9855.0, 12045.0));

    // The reference input file
    TString sJDDReferenceKey = gSystem->pwd();
    sJDDReferenceKey += TString("/flowrate_input_with_row.in");

    tds2->getAttribute("rw")->setFileKey(sJDDReferenceKey, "Rw", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("r")->setFileKey(sJDDReferenceKey, "R", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("tu")->setFileKey(sJDDReferenceKey, "Tu", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("tl")->setFileKey(sJDDReferenceKey, "Tl", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("hu")->setFileKey(sJDDReferenceKey, "Hu", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("hl")->setFileKey(sJDDReferenceKey, "Hl", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("l")->setFileKey(sJDDReferenceKey, "L", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("kw")->setFileKey(sJDDReferenceKey, "Kw", "%f",
            TAttributeFileKey::kNewRow);

    
    TOutputFileDataServer *foutDataServer = new TOutputFileDataServer(
            "_output_flowrate_withRow_.dat");
    foutDataServer->addAttribute("yhat");

    TCode *codeFlowrate = new TCode(tds2, Form("%s/%s", gSystem->pwd(),
                            "../../../../launCHER/teST/coDE/flowREATE/flowrate -r flowrate_input_with_row.in >>/dev/null"));
    codeFlowrate->addOutputFile(foutDataServer);

    
    URANIE::Modeler::TAnisp * anisp = new URANIE::Modeler::TAnisp(tds2,
            codeFlowrate, "myOption");

    //Error in the value of _nGreaterIndice must < 7 if there is a normal variable
    CPPUNIT_ASSERT_THROW(anisp->setGreaterIndice(7),
            URANIE::Exceptions::UErrorExceptions);

    // Clean
    tds2->getAttribute("rw")->delShare();
    tds2->getAttribute("r")->delShare();
    tds2->getAttribute("tu")->delShare();
    tds2->getAttribute("tl")->delShare();
    tds2->getAttribute("hu")->delShare();
    tds2->getAttribute("hl")->delShare();
    tds2->getAttribute("l")->delShare();
    tds2->getAttribute("kw")->delShare();
    
 
    delete anisp;
    delete foutDataServer;
    delete codeFlowrate;
    delete tds2;
    
}

void TAnispTest::throwExceptionsetMaxIndicesMethod()
{
    //Error in the value of _nMaxIndices (must be > 0)
    CPPUNIT_ASSERT_THROW(_anisp2->setMaxIndices(0),
            URANIE::Exceptions::UErrorExceptions);
}

void TAnispTest::throwExceptionsetDegreeMaxMethod()
{
    //Error in the value of _nDegreeMax (must be >= 0)
    CPPUNIT_ASSERT_THROW(_anisp2->setDegreeMax(-1),
            URANIE::Exceptions::UErrorExceptions);
}

void TAnispTest::throwExceptionsetRootFilenameMethod()
{
    //Error in the value of _sRootFilename must not be empty 
    CPPUNIT_ASSERT_THROW(_anisp2->setRootFilename(""),
            URANIE::Exceptions::UErrorExceptions);
//Error in the value of _sRootFilename must not be white
    CPPUNIT_ASSERT_THROW(_anisp2->setRootFilename("    "),
            URANIE::Exceptions::UErrorExceptions);
}

void TAnispTest::throwExceptionConstructorInputAttribute()
{

    // Define the DataServer
    TDataServer *tds2 = new TDataServer("tdsflowreate", "DataBase flowreate");
    tds2->addAttribute(new TUniformDistribution("rw", 0.05, 0.15));
    tds2->addAttribute(new TUniformDistribution("r", 100.0, 50000.0));
    tds2->addAttribute(new TUniformDistribution("tu", 63070.0, 115600.0));
    tds2->addAttribute(new TUniformDistribution("tl", 63.1, 116.0));
    tds2->addAttribute(new TUniformDistribution("hu", 990.0, 1110.0));
    tds2->addAttribute(new TUniformDistribution("hl", 700.0, 820.0));
    tds2->addAttribute(new TUniformDistribution("l", 1120.0, 1680.0));
    tds2->addAttribute(new TNormalDistribution("kw", 9855.0, 12045.0));
    // add a input  TPossibilityAttribute with no reason
    tds2->addAttribute(new TPossibilityAttribute("kwww"));

    // The reference input file
    TString sJDDReferenceKey = gSystem->pwd();
    sJDDReferenceKey += TString("/flowrate_input_with_row.in");

    tds2->getAttribute("rw")->setFileKey(sJDDReferenceKey, "Rw", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("r")->setFileKey(sJDDReferenceKey, "R", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("tu")->setFileKey(sJDDReferenceKey, "Tu", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("tl")->setFileKey(sJDDReferenceKey, "Tl", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("hu")->setFileKey(sJDDReferenceKey, "Hu", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("hl")->setFileKey(sJDDReferenceKey, "Hl", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("l")->setFileKey(sJDDReferenceKey, "L", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("kw")->setFileKey(sJDDReferenceKey, "Kw", "%f",
            TAttributeFileKey::kNewRow);

    
    TOutputFileDataServer *foutDataServer = new TOutputFileDataServer(
            "_output_flowrate_withRow_.dat");
    foutDataServer->addAttribute("yhat");

    TCode *codeFlowrate = new TCode(tds2, Form("%s/%s", gSystem->pwd(),
                            "../../../../launCHER/teST/coDE/flowREATE/flowrate -r flowrate_input_with_row.in >>/dev/null"));
    codeFlowrate->addOutputFile(foutDataServer);

    
    //Error in the input attribute
    CPPUNIT_ASSERT_THROW(
            URANIE::Modeler::TAnisp(tds2, codeFlowrate, "myOption"),
            URANIE::Exceptions::UErrorExceptions);

    // Clean
    tds2->getAttribute("rw")->delShare();
    tds2->getAttribute("r")->delShare();
    tds2->getAttribute("tu")->delShare();
    tds2->getAttribute("tl")->delShare();
    tds2->getAttribute("hu")->delShare();
    tds2->getAttribute("hl")->delShare();
    tds2->getAttribute("l")->delShare();
    tds2->getAttribute("kw")->delShare();
    tds2->getAttribute("kwww")->delShare();

    delete foutDataServer;
    delete codeFlowrate;
    delete tds2;
    
}

void TAnispTest::throwExceptionrestartAnisp()
{
    //no previous run of runAnisp
    TString str = "testthrowExceptionrestartAnisp_";
    _anisp2->setRootFilename(str);
    CPPUNIT_ASSERT_THROW(_anisp2->restartAnisp(),
            URANIE::Exceptions::UErrorExceptions);

    str += "2_";
    TString str2 = str + TString("data.dat");
    const char* filename = str2.Data();
    std::ofstream file(filename, ios::out);
    file << 0 << endl;
    _anisp2->setRootFilename(str);
    CPPUNIT_ASSERT_THROW(_anisp2->restartAnisp(),
            URANIE::Exceptions::UErrorExceptions);

}

void TAnispTest::throwExceptiongetTPolynomialChaos()
{
    //no previous run of runAnisp
    CPPUNIT_ASSERT_THROW(_anisp2->getTPolynomialChaos(),
            URANIE::Exceptions::UErrorExceptions);
}

void TAnispTest::throwExceptionchangeOfVariable()
{

    _tdsIshigami->addAttribute(
            new TLogUniformDistribution("x5", 1, TMath::Pi()));
    _tdsIshigami->addAttribute(
            new TLogUniformDistribution("x6", 1, TMath::Pi()));
    _tdsIshigami->getAttribute("x5")->delShare();
    _tdsIshigami->getAttribute("x6")->delShare();
//     CPPUNIT_ASSERT_THROW(_anisp->runAnisp(),
//                          URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(_anisp->runAnisp(), Anisp::anispException); 
     
}

void TAnispTest::throwExceptionExportCplusplus()
{

    _anisp->setRootFilename("test_export_");
    _anisp->runAnisp();
    UranieTests::removeFile(gSystem->pwd(), "test_export_res.dat");
    CPPUNIT_ASSERT_THROW(_anisp->exportFunction("c++", "tata", "tata"),
            URANIE::Exceptions::UErrorExceptions);
}

void TAnispTest::throwExceptionConstructorNonAnispPatterns()
{

    // Define the DataServer
    TDataServer *tds2 = new TDataServer("tdsflowreate", "DataBase flowreate");
    tds2->addAttribute(new TUniformDistribution("rw", 0.05, 0.15));
    tds2->addAttribute(new TUniformDistribution("r", 100.0, 50000.0));
    tds2->addAttribute(new TUniformDistribution("tu", 63070.0, 115600.0));
    tds2->addAttribute(new TUniformDistribution("tl", 63.1, 116.0));
    tds2->addAttribute(new TUniformDistribution("hu", 990.0, 1110.0));
    tds2->addAttribute(new TUniformDistribution("hl", 700.0, 820.0));
    tds2->addAttribute(new TUniformDistribution("l", 1120.0, 1680.0));
    tds2->addAttribute(new TNormalDistribution("kw", 9855.0, 12045.0));

    TString sJDDReferenceKey = gSystem->pwd();
    sJDDReferenceKey += TString("/flowrate_input_with_row.in");

    tds2->getAttribute("rw")->setFileKey(sJDDReferenceKey, "Rw", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("r")->setFileKey(sJDDReferenceKey, "R", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("tu")->setFileKey(sJDDReferenceKey, "Tu", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("tl")->setFileKey(sJDDReferenceKey, "Tl", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("hu")->setFileKey(sJDDReferenceKey, "Hu", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("hl")->setFileKey(sJDDReferenceKey, "Hl", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("l")->setFileKey(sJDDReferenceKey, "L", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("kw")->setFileKey(sJDDReferenceKey, "Kw", "%f",
            TAttributeFileKey::kNewRow);

    // creation of non Anisp patterns
    TSampling *fsamp = new TSampling(tds2, "lhs", 10);
    fsamp->generateSample();

    TOutputFileDataServer *foutDataServer = new TOutputFileDataServer(
            "_output_flowrate_withRow_.dat");
    foutDataServer->addAttribute("yhat");

    TCode *codeFlowrate = new TCode(tds2, Form("%s/%s", gSystem->pwd(),
                            "../../../../launCHER/teST/coDE/flowREATE/flowrate -r flowrate_input_with_row.in >>/dev/null"));
      
    codeFlowrate->addOutputFile(foutDataServer);

     
    //Error in the TDataServer
    CPPUNIT_ASSERT_THROW(
            URANIE::Modeler::TAnisp(tds2, codeFlowrate, "myOption"),
            URANIE::Exceptions::UErrorExceptions);

    // Clean
    delete fsamp;
    tds2->getAttribute("rw")->delShare();
    tds2->getAttribute("r")->delShare();
    tds2->getAttribute("tu")->delShare();
    tds2->getAttribute("tl")->delShare();
    tds2->getAttribute("hu")->delShare();
    tds2->getAttribute("hl")->delShare();
    tds2->getAttribute("l")->delShare();
    tds2->getAttribute("kw")->delShare(); 
    
    delete foutDataServer;
    delete codeFlowrate;
    delete tds2;
    
}


void TAnispTest::throwExceptionConstructorTdsDifferentInTCode()
{

    // Define the DataServer
    TDataServer *tds2 = new TDataServer("tdsflowreate", "DataBase flowreate");
    tds2->addAttribute(new TUniformDistribution("rw", 0.05, 0.15));
    tds2->addAttribute(new TUniformDistribution("r", 100.0, 50000.0));
    tds2->addAttribute(new TUniformDistribution("tu", 63070.0, 115600.0));
    tds2->addAttribute(new TUniformDistribution("tl", 63.1, 116.0));
    tds2->addAttribute(new TUniformDistribution("hu", 990.0, 1110.0));
    tds2->addAttribute(new TUniformDistribution("hl", 700.0, 820.0));
    tds2->addAttribute(new TUniformDistribution("l", 1120.0, 1680.0));
    tds2->addAttribute(new TNormalDistribution("kw", 9855.0, 12045.0));

    TString sJDDReferenceKey = gSystem->pwd();
    sJDDReferenceKey += TString("/flowrate_input_with_row.in");

    tds2->getAttribute("rw")->setFileKey(sJDDReferenceKey, "Rw", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("r")->setFileKey(sJDDReferenceKey, "R", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("tu")->setFileKey(sJDDReferenceKey, "Tu", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("tl")->setFileKey(sJDDReferenceKey, "Tl", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("hu")->setFileKey(sJDDReferenceKey, "Hu", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("hl")->setFileKey(sJDDReferenceKey, "Hl", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("l")->setFileKey(sJDDReferenceKey, "L", "%f",
            TAttributeFileKey::kNewRow);
    tds2->getAttribute("kw")->setFileKey(sJDDReferenceKey, "Kw", "%f",
            TAttributeFileKey::kNewRow);
    
    // creation of non Anisp patterns
    TSampling *fsamp = new TSampling(tds2, "lhs", 10);
    fsamp->generateSample();

    
    //Error in the TDataServer
    CPPUNIT_ASSERT_THROW(
            URANIE::Modeler::TAnisp(tds2, _codeFlowrate, "myOption"),
            URANIE::Exceptions::UErrorExceptions);

    // Clean
    delete fsamp;
    tds2->getAttribute("rw")->delShare();
    tds2->getAttribute("r")->delShare();
    tds2->getAttribute("tu")->delShare();
    tds2->getAttribute("tl")->delShare();
    tds2->getAttribute("hu")->delShare();
    tds2->getAttribute("hl")->delShare();
    tds2->getAttribute("l")->delShare();
    tds2->getAttribute("kw")->delShare();

    delete tds2;
   
}



// End Of File

