//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTKernel.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   $Date$
 
 \brief  Fichier de test de la classe URANIE::DataServer::TKernel
 
 Created : 2012-11-23 22:27:55
 
 */

#define DBL_PRECISION_COMPARE 1.e-10
// Data Header
#include "TestTKernel.h"
#include "TString.h"
#include <iostream>
#include <ctime>
#include <iomanip>

//ROOT
#include "TMath.h"
#include "TLeaf.h"

#include "uranie_commands.h"

TKernelTest::TKernelTest() :
        _tds1(NULL), _tds2(NULL)
{
    _workingTestPath = gSystem->pwd(); 
}

TKernelTest::~TKernelTest()
{
    if (_tds1 != NULL)
    {
        delete _tds1;
        _tds1 = NULL;
    }
    if (_tds2 != NULL)
    {
        delete _tds2;
        _tds2 = NULL;
    }
}

////////////////////////////////
///                          ///
///   setUp / tearDown       ///
///                          ///
////////////////////////////////


void TKernelTest::setUp()
{
    _tds1 = new URANIE::DataServer::TDataServer("tdsPoisson",
            "Data from capitalPunishement");
    _tds1->fileDataRead("kernel.dat");

    _tds2 = new URANIE::DataServer::TDataServer("tdsBinomial",
            "Data from capitalPunishement");
}

void TKernelTest::tearDown()
{
    delete _tds1;
    _tds1 = NULL;
    delete _tds2;
    _tds2 = NULL;
    
    
    UranieTests::removeFile(_workingTestPath.Data(),"uranie_kernel_ishigami.C");
    UranieTests::removeFile(_workingTestPath.Data(),"uranie_kernel_ishigamiBernouilli.C");
}

////////////////////////////////
///                          ///
///         Tests            ///
///                          ///
////////////////////////////////

void TKernelTest::testTKernelSpline1()
{
    URANIE::Modeler::TKernel *tk = new URANIE::Modeler::TKernel(_tds1, "u",
            "y1");
    tk->unsetLog();
    Int_t npat = _tds1->getNPatterns();
    CPPUNIT_ASSERT_EQUAL(100, npat);

    tk->setKernel(URANIE::Modeler::TKernel::kSpline1);
    tk->estimate();

    CPPUNIT_ASSERT(_tds1->isAttribute("__alpha__"));

    TVectorD dparam = tk->getParameters();

    CPPUNIT_ASSERT_EQUAL(npat, _tds1->getNPatterns());
    CPPUNIT_ASSERT_EQUAL(npat, dparam.GetNrows());

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0520120141589e-10, dparam(0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, dparam(1), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.0152253030971975, dparam(2),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, dparam(3), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.0164896582540249, dparam(4),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, dparam(5), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.0167971857293323, dparam(npat - 3),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.0222344896561872, dparam(npat - 2),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.0194803214811988, dparam(npat - 1),
            DBL_PRECISION_COMPARE);

    delete tk;

}

void TKernelTest::testTKernelBernoulli1()
{
    URANIE::Modeler::TKernel *tk = new URANIE::Modeler::TKernel(_tds1, "u",
            "y1");
    tk->unsetLog();
    Int_t npat = _tds1->getNPatterns();
    CPPUNIT_ASSERT_EQUAL(100, npat);

    tk->setKernel(URANIE::Modeler::TKernel::kBernoulli1);
    tk->estimate();

    CPPUNIT_ASSERT(_tds1->isAttribute("__alpha__"));

    TVectorD dparam = tk->getParameters();

    CPPUNIT_ASSERT_EQUAL(npat, _tds1->getNPatterns());
    CPPUNIT_ASSERT_EQUAL(npat, dparam.GetNrows());

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0263869024156246, dparam(0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0318954538173212, dparam(1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0016452117543545, dparam(2),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0121001041864645, dparam(3),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0017818400080782, dparam(4),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0178223929379681, dparam(5),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0018150551302517, dparam(npat - 3),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0024026152412567, dparam(npat - 2),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0021050231265450, dparam(npat - 1),
            DBL_PRECISION_COMPARE);
    delete tk;
}

void TKernelTest::testTKernelSpline1ConstructorArchitecture()
{
    URANIE::Modeler::TKernel *tk = new URANIE::Modeler::TKernel(_tds1, "u,1,y1");
    tk->unsetLog();
    Int_t npat = _tds1->getNPatterns();
    CPPUNIT_ASSERT_EQUAL(100, npat);

    tk->setKernel(URANIE::Modeler::TKernel::kSpline1);
    tk->estimate();

    CPPUNIT_ASSERT(_tds1->isAttribute("__alpha__"));

    TVectorD dparam = tk->getParameters();

    CPPUNIT_ASSERT_EQUAL(npat, _tds1->getNPatterns());
    CPPUNIT_ASSERT_EQUAL(npat, dparam.GetNrows());


    CPPUNIT_ASSERT_EQUAL(npat, _tds1->getNPatterns());
    CPPUNIT_ASSERT_EQUAL(npat, dparam.GetNrows());

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0520120141589e-10, dparam(0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, dparam(1), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.0152253030971975, dparam(2),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, dparam(3), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.0164896582540249, dparam(4),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, dparam(5), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.0167971857293323, dparam(npat - 3),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.0222344896561872, dparam(npat - 2),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.0194803214811988, dparam(npat - 1),
            DBL_PRECISION_COMPARE);

    delete tk;
    
}



void TKernelTest::testTKernelSetKernel()
{
    URANIE::Modeler::TKernel *tk = new URANIE::Modeler::TKernel(_tds1, "u,1,y1");
    tk->unsetLog();
    URANIE::Modeler::TKernel::EKernelType kType = URANIE::Modeler::TKernel::kSpline1;
    
    tk->setKernel(kType);
    CPPUNIT_ASSERT_EQUAL(kType, tk->getKernel());
    
    kType = URANIE::Modeler::TKernel::kLinear;
    tk->setKernel(kType,1.2345678);
    CPPUNIT_ASSERT_EQUAL(kType, tk->getKernel());
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.2345678,tk->getParameter1() , DBL_PRECISION_COMPARE);
    
    
    kType = URANIE::Modeler::TKernel::kSigmoid;
    tk->setKernel(kType,1.2345678,2.3456789);
    CPPUNIT_ASSERT_EQUAL(kType, tk->getKernel());
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.2345678,tk->getParameter1() , DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.3456789,tk->getParameter2() , DBL_PRECISION_COMPARE);
    
    kType = URANIE::Modeler::TKernel::kPoly;
    tk->setKernel(kType,1.2345678,2.3456789,3.4567890);
    CPPUNIT_ASSERT_EQUAL(kType, tk->getKernel());
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.2345678,tk->getParameter1() , DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.3456789,tk->getParameter2() , DBL_PRECISION_COMPARE);

    delete tk;
    
}


void TKernelTest::testExportModelCplusplus()
{
    time_t seconds;
    string line, date;
    Bool_t bStop = kFALSE;
    Int_t test1 = 0;
    Int_t test2 = 0;
    Int_t test3 = 0;
    Int_t test4 = 0;
    URANIE::DataServer::TDataServer *tds1 = new URANIE::DataServer::TDataServer("tdsIshigami",
            "Data from ishigami");
    tds1->fileDataRead("ishigami_sample.dat");
    URANIE::Modeler::TKernel *tk = new URANIE::Modeler::TKernel(tds1, "x2,1,Ishigami");
    tk->unsetLog();
    URANIE::Modeler::TKernel::EKernelType kType = URANIE::Modeler::TKernel::kSpline1;
    
    tk->setKernel(kType);
    tk->estimate();
    tk->exportFunction("c++", "uranie_kernel_ishigami.C","KernelIshigami");
    
    std::fstream file("uranie_kernel_ishigami.C");
    time(&seconds);
    date = string(ctime(&seconds));
    while (bStop == kFALSE)
    {
        getline(file, line);
        if (line.find("  //     ** Date : ") != string::npos)
        {
            //verify that the date are the same actually and in the file
            if ((date[0] == line[24]) && (date[1] == line[25])
                    && (date[2] == line[26]) && (date[3] == line[27])
                    && (date[4] == line[28]) && (date[5] == line[29])
                    && (date[6] == line[30]) && (date[7] == line[31])
                    && (date[8] == line[32]) && (date[9] == line[33])
                    && (date[10] == line[34]) && (date[11] == line[35])
                    && (date[12] == line[36]) && (date[13] == line[37])
                    && (date[14] == line[38]) && (date[15] == line[39])
                    && (date[16] == line[40]) && (date[17] == line[41])
                    && (date[18] == line[42]) && (date[19] == line[43])
                    && (date[20] == line[44]) && (date[21] == line[45])
                    && (date[22] == line[46]) && (date[23] == line[47]))
                test1 = 1;
               
            if(test1==0)
            {
                Int_t pos = date.find_last_of(":");
                Int_t pos2 = line.find_last_of(":");
                Int_t date_seconds, line_seconds;
                date_seconds =  (date[pos-5]*10+date[pos-4])*3600
                            +(date[pos-2]*10+date[pos-1])*60
                            +(date[pos+1]*10+date[pos+2]);
                            
                line_seconds =  (line[pos2-5]*10+line[pos2-4])*3600
                            +(line[pos2-2]*10+line[pos2-1])*60
                            +(line[pos2+1]*10+line[pos2+2]);
                
               
                if(fabs(date_seconds - line_seconds + 0.0)<3)
                    test1 = 1;
            }
        }
         
        if (line.find("  -0.496091, -0.393658, -1.074, 0.150571, 0.648718,")
                != string::npos)
            test2 = 1;
        
        if (line.find("  -0.843373, 0.646949, -0.870043, 0.112959, 0.00746089, ")
                != string::npos)
            test3 = 1;
         
        if (line.find("void KernelIshigami(double *param, double *res)")
                != string::npos)
            test4 = 1;
         
        if (file.eof())
            bStop = kTRUE;
    }
    
    CPPUNIT_ASSERT_EQUAL(1, test1);
    CPPUNIT_ASSERT_EQUAL(1, test2);
    CPPUNIT_ASSERT_EQUAL(1, test3);
    CPPUNIT_ASSERT_EQUAL(1, test4);

    bStop = kFALSE;
    test1 = 0;
    test2 = 0;
    test3 = 0;
    test4 = 0;
    
    URANIE::DataServer::TDataServer *tds2 = new URANIE::DataServer::TDataServer("tdsIshigami",
            "Data from ishigami");
    tds2->fileDataRead("ishigami_sample.dat");
    
    delete tk;
    delete tds1;
    
    tk = new URANIE::Modeler::TKernel(tds2, "x2,1,Ishigami");
    tk->unsetLog();
    kType = URANIE::Modeler::TKernel::kBernoulli1;
    
    tk->setKernel(kType);
    tk->estimate();
    tk->exportFunction("c++", "uranie_kernel_ishigamiBernouilli.C","KernelIshigami");
    
    std::fstream file2("uranie_kernel_ishigamiBernouilli.C");
    time(&seconds);
    date = string(ctime(&seconds));
    while (bStop == kFALSE)
    {
        getline(file2, line);
        if (line.find("  //     ** Date : ") != string::npos)
        {
            //verify that the date are the same actually and in the file
            if ((date[0] == line[24]) && (date[1] == line[25])
                    && (date[2] == line[26]) && (date[3] == line[27])
                    && (date[4] == line[28]) && (date[5] == line[29])
                    && (date[6] == line[30]) && (date[7] == line[31])
                    && (date[8] == line[32]) && (date[9] == line[33])
                    && (date[10] == line[34]) && (date[11] == line[35])
                    && (date[12] == line[36]) && (date[13] == line[37])
                    && (date[14] == line[38]) && (date[15] == line[39])
                    && (date[16] == line[40]) && (date[17] == line[41])
                    && (date[18] == line[42]) && (date[19] == line[43])
                    && (date[20] == line[44]) && (date[21] == line[45])
                    && (date[22] == line[46]) && (date[23] == line[47]))
                test1 = 1;
               
            if(test1==0)
            {
                Int_t pos = date.find_last_of(":");
                Int_t pos2 = line.find_last_of(":");
                Int_t date_seconds, line_seconds;
                date_seconds =  (date[pos-5]*10+date[pos-4])*3600
                            +(date[pos-2]*10+date[pos-1])*60
                            +(date[pos+1]*10+date[pos+2]);
                            
                line_seconds =  (line[pos2-5]*10+line[pos2-4])*3600
                            +(line[pos2-2]*10+line[pos2-1])*60
                            +(line[pos2+1]*10+line[pos2+2]);
                
               
                if(fabs(date_seconds - line_seconds + 0.0)<3)
                    test1 = 1;
            }
        }
         
        if (line.find("  -0.590832, -0.499923, -1.22058, 0.0721238, 0.556039, ")
                != string::npos)
            test2 = 1;
        
        if (line.find("  4.78983, 0.503401, -0.995464, -0.0246984, -0.0810206, ")
                != string::npos)
            test3 = 1;
         
        if (line.find("void KernelIshigami(double *param, double *res)")
                != string::npos)
            test4 = 1;
         
        if (file2.eof())
            bStop = kTRUE;
    }
    
    CPPUNIT_ASSERT_EQUAL(1, test1);
    CPPUNIT_ASSERT_EQUAL(1, test2);
    CPPUNIT_ASSERT_EQUAL(1, test3);
    CPPUNIT_ASSERT_EQUAL(1, test4);

   
    delete tk;
    delete tds2;
    
}


////////////////////////////////
///                          ///
///       Exceptions         ///
///                          ///
////////////////////////////////

void TKernelTest::throwExceptionTKernelSetterKernel()
{
    URANIE::Modeler::TKernel *tk = new URANIE::Modeler::TKernel(_tds1, "u",
            "y1");
    tk->unsetLog();
    Int_t npat = _tds1->getNPatterns();
    CPPUNIT_ASSERT_EQUAL(100, npat);

    CPPUNIT_ASSERT_THROW(
            tk->setKernel(URANIE::Modeler::TKernel::kSpline1, 0.12, 0.23, 0.34),
            URANIE::Exceptions::UErrorExceptions);
        delete tk;
}


void TKernelTest::throwExceptionTKernelTooManyOutputs()
{

    CPPUNIT_ASSERT_THROW(URANIE::Modeler::TKernel tk(_tds1, "u","y1:y2"),
            URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(URANIE::Modeler::TKernel tk(_tds1, "u,1,y1:y2"),
            URANIE::Exceptions::UErrorExceptions);
}

