//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTPolynomialRegression.cxx
 \author Estelle Deville <estelle.deville@cea.fr>
 \date   $Date$
 
 \brief  Fichier de test de la classe URANIE::DataServer::TPolynomialRegression
 
 \bf Created : 2011/06/30 15:34:06
 
 \par Modifications
 \li M. Salvoldi - OCC/Areva group (2011/10/19)
 - add 13 units test to covert the class TPolynomialRegresion at 80%
 */

#define DBL_PRECISION_COMPARE 1.e-10

#include <iostream>
#include <iomanip>

// CPPUNIT
#include <cppunit/Test.h>
#include <cppunit/TestFailure.h>

// ROOT
#include "TString.h"
#include "TLeaf.h"
#include "TLeafC.h"
#include "TPad.h"

// Uranie
#include "TestTPolynomialRegression.h"
#include "uranie_commands.h"

TPolynomialRegressionTest::TPolynomialRegressionTest() :
        _tds(NULL), _sX("z1:z2:z3:z4:z5:z6:z7:z8:z9:z10:z11"), _sY("visco") //  Densite | tg | visco
{
    std::fstream file("modelTPR.C");
    if (!file.fail())
        remove("modelTPR.C");
    std::fstream file2("modelTPR.f");
    if (!file2.fail())
        remove("modelTPR.f");
}

TPolynomialRegressionTest::~TPolynomialRegressionTest()
{
    std::fstream file("modelTPR.C");
    if (!file.fail())
        remove("modelTPR.C");
    std::fstream file2("modelTPR.f");
    if (!file2.fail())
        remove("modelTPR.f");
}


void TPolynomialRegressionTest::setUp()
{

    UranieTests::removeFile(gSystem->pwd(), "_dataserver_.root");
    UranieTests::removeFile(gSystem->pwd(), "exercice1-pseudoconstituants.root");
	
    _tds = new URANIE::DataServer::TDataServer();
    _tds->fileDataRead("exercice1-pseudoconstituants.dat");
}

void TPolynomialRegressionTest::tearDown()
{

    delete _tds;
    _tds = NULL;
    if (gPad != NULL)
    {
        gPad->Clear();
    }

    UranieTests::removeFile(gSystem->pwd(), "_dataserver_.root");
    UranieTests::removeFile(gSystem->pwd(), "exercice1-pseudoconstituants.root");
}


void TPolynomialRegressionTest::testPolynomialRegressionArchi()
{
    URANIE::Modeler::TPolynomialRegression *tpr =
            new URANIE::Modeler::TPolynomialRegression(_tds,
                    TString("z1:z2:z3:z4:z5:z6:z7:z8:z9:z10:z11,1,visco"),
                    "nointercept");
    tpr->setCriteria(0.9);
    tpr->setDrawProgressBar(kFALSE);
    tpr->estimate("Q2");

    TVectorD dparam = tpr->getParameters();

    CPPUNIT_ASSERT_EQUAL(14, dparam.GetNrows());

    CPPUNIT_ASSERT_DOUBLES_EQUAL(321.194026848064, dparam(0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(18.510968426589, dparam(1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-148.071967162172, dparam(2),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-153.58542945954, dparam(3),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(173.172131467353, dparam(4),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-36.340218995185, dparam(5),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(288.070313073538, dparam(6),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(130.112002774998, dparam(7),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-62.246780027326, dparam(8),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-118.370849110815, dparam(9),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-587.573299798665, dparam(10),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-1238.978412065981, dparam(11),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1569.027461232204, dparam(12),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-711.101521506294, dparam(13),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.995730939571419, tpr->getR2(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.986481308642826, tpr->getR2Adjusted(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.937963144887466, tpr->getQ2(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2468.50539994009, tpr->getPress(), 1e-9);

    delete tpr;
}

void TPolynomialRegressionTest::testPolynomialRegressionPseudoConstituant()
{
    URANIE::Modeler::TPolynomialRegression *tpr =
            new URANIE::Modeler::TPolynomialRegression(_tds, _sX, _sY,
                    "nointercept");
    tpr->setCriteria(0.9);
    tpr->setDrawProgressBar(kFALSE);
    tpr->estimate("Q2");

    CPPUNIT_ASSERT(_tds->isAttribute("z5z8"));
    CPPUNIT_ASSERT(_tds->isAttribute("z3z10"));
    CPPUNIT_ASSERT(_tds->isAttribute("z1z10"));
    CPPUNIT_ASSERT(_tds->isAttribute("z3z8") == kFALSE);
    CPPUNIT_ASSERT(_tds->isAttribute("z1z11") == kFALSE);

    CPPUNIT_ASSERT(_tds->isAttribute(Form("%shat", _sY.Data())));

    TVectorD dparam = tpr->getParameters();

    CPPUNIT_ASSERT_EQUAL(14, dparam.GetNrows());

    CPPUNIT_ASSERT_DOUBLES_EQUAL(321.194026848064, dparam(0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(18.510968426589, dparam(1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-148.071967162172, dparam(2),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-153.58542945954, dparam(3),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(173.172131467353, dparam(4),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-36.340218995185, dparam(5),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(288.070313073538, dparam(6),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(130.112002774998, dparam(7),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-62.246780027326, dparam(8),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-118.370849110815, dparam(9),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-587.573299798665, dparam(10),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-1238.978412065981, dparam(11),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1569.027461232204, dparam(12),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-711.101521506294, dparam(13),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.995730939571, tpr->getR2(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.986481308642, tpr->getR2Adjusted(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.937963144887, tpr->getQ2(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2468.50539994009, tpr->getPress(), 1e-9);

    // TTree of results
    TTree *tt = tpr->getResultsTree();
    Int_t npat = 166;
    CPPUNIT_ASSERT_EQUAL(npat, (Int_t)(tt->GetEntries()));

    TLeafC *tlinp = dynamic_cast<TLeafC *>(tt->GetLeaf("Inp"));
    TLeaf *tln = tt->GetLeaf("n");
    TLeaf *tlb = tt->GetLeaf("b");
    TLeaf *tlCriteria = tt->GetLeaf("Criteria");
    TLeaf *tlCR2 = tt->GetLeaf("R2");
    TLeaf *tlCR2A = tt->GetLeaf("R2A");
    TLeaf *tlCQ2 = tt->GetLeaf("Q2");

    // First recording
    tt->GetEntry(0);

    CPPUNIT_ASSERT(!strcmp(" ", tlinp->GetValueString()));
    CPPUNIT_ASSERT_EQUAL(0, (Int_t)(tln->GetValue()));
    CPPUNIT_ASSERT_EQUAL(0, (Int_t)(tlb->GetValue()));
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.4385190371, tlCriteria->GetValue(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.9037680607, tlCR2->GetValue(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.7968436837, tlCR2A->GetValue(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.4385190371, tlCQ2->GetValue(),
            DBL_PRECISION_COMPARE);

    // Second recording
    tt->GetEntry(1);

    CPPUNIT_ASSERT(!strcmp(":z1z2", tlinp->GetValueString()));
    CPPUNIT_ASSERT_EQUAL(1, (Int_t)(tln->GetValue()));
    CPPUNIT_ASSERT_EQUAL(0, (Int_t)(tlb->GetValue()));
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.6499007261, tlCriteria->GetValue(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.9546111292, tlCR2->GetValue(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.8922014318, tlCR2A->GetValue(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.6499007261, tlCQ2->GetValue(),
            DBL_PRECISION_COMPARE);

    // Last recording
    tt->GetEntry(npat - 1);

    CPPUNIT_ASSERT(!strcmp(":z5z8:z3z10:z1z10", tlinp->GetValueString()));
    CPPUNIT_ASSERT_EQUAL(3, (Int_t)(tln->GetValue()));
    CPPUNIT_ASSERT_EQUAL(1, (Int_t)(tlb->GetValue()));
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.9379631448, tlCriteria->GetValue(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.9957309395, tlCR2->GetValue(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.9864813086, tlCR2A->GetValue(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.9379631448, tlCQ2->GetValue(),
            DBL_PRECISION_COMPARE);

    delete tpr;
}

void TPolynomialRegressionTest::testPolynomialRegressionPseudoConstituantFewAttribute1()
{
    URANIE::Modeler::TPolynomialRegression *tpr =
            new URANIE::Modeler::TPolynomialRegression(_tds, "z1:z5:z7:z11",
                    _sY, "nointercept");
    tpr->unsetLog();
    tpr->setCriteria(0.76);
    tpr->setDrawProgressBar(kFALSE);
    tpr->estimate("R2");

    CPPUNIT_ASSERT(_tds->isAttribute("z1z7"));
    CPPUNIT_ASSERT(_tds->isAttribute("z5z7"));
    CPPUNIT_ASSERT(_tds->isAttribute("z1z10") == kFALSE);
    CPPUNIT_ASSERT(_tds->isAttribute("z7z10") == kFALSE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.80416786200107, tpr->getR2(),
            DBL_PRECISION_COMPARE);

    TVectorD dparam = tpr->getParameters();

    CPPUNIT_ASSERT_EQUAL(6, dparam.GetNrows());

    CPPUNIT_ASSERT_DOUBLES_EQUAL(180.49779024659290, dparam(0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(74.23734445993517, dparam(1),
            DBL_PRECISION_COMPARE);

    delete tpr;
}

void TPolynomialRegressionTest::testPolynomialRegressionPseudoConstituantFewAttribute1Stop()
{
    URANIE::Modeler::TPolynomialRegression *tpr =
            new URANIE::Modeler::TPolynomialRegression(_tds, "z1:z5:z7:z11",
                    _sY, "nointercept");
    tpr->unsetLog();
    tpr->setCriteria(0.60);
    tpr->setDrawProgressBar(kFALSE);
    tpr->estimate("R2");

    CPPUNIT_ASSERT(_tds->isAttribute("z1z7") == kFALSE);
    CPPUNIT_ASSERT(_tds->isAttribute("z5z7") == kFALSE);
    CPPUNIT_ASSERT(_tds->isAttribute("z1z10") == kFALSE);
    CPPUNIT_ASSERT(_tds->isAttribute("z7z10") == kFALSE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.6256960399, tpr->getR2(),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT(_tds->isAttribute(Form("%shat", _sY.Data())));

    TVectorD dparam = tpr->getParameters();

    CPPUNIT_ASSERT_EQUAL(4, dparam.GetNrows());

    CPPUNIT_ASSERT_DOUBLES_EQUAL(202.39733559283, dparam(0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(79.08639205378, dparam(1),
            DBL_PRECISION_COMPARE);

    // TTree of results
    TTree *tt = tpr->getResultsTree();
    CPPUNIT_ASSERT_EQUAL(1, (Int_t)(tt->GetEntries()));

    TLeafC *tlinp = dynamic_cast<TLeafC *>(tt->GetLeaf("Inp"));
    TLeaf *tln = tt->GetLeaf("n");
    TLeaf *tlb = tt->GetLeaf("b");
    TLeaf *tlCriteria = tt->GetLeaf("Criteria");
    TLeaf *tlCR2 = tt->GetLeaf("R2");
    TLeaf *tlCR2A = tt->GetLeaf("R2A");
    TLeaf *tlCQ2 = tt->GetLeaf("Q2");

    // First recording
    tt->GetEntry(0);

    CPPUNIT_ASSERT(!strcmp(" ", tlinp->GetValueString()));
    CPPUNIT_ASSERT_EQUAL(0, (Int_t)(tln->GetValue()));
    CPPUNIT_ASSERT_EQUAL(0, (Int_t)(tlb->GetValue()));
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.6256960399, tlCriteria->GetValue(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.6256960399, tlCR2->GetValue(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.5555140473, tlCR2A->GetValue(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.5601102566, tlCQ2->GetValue(),
            DBL_PRECISION_COMPARE);

     delete tpr;
}

void TPolynomialRegressionTest::testPolynomialRegressionPseudoConstituantFewAttribute2()
{

    URANIE::Modeler::TPolynomialRegression *tpr4 =
            new URANIE::Modeler::TPolynomialRegression(_tds, "z1:z2:z3:z4", _sY,
                    "nointercept");
    tpr4->unsetLog();
    tpr4->setDrawProgressBar(kFALSE);
    tpr4->setCriteria(0.50);
    tpr4->estimate("R2A");

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.53149618253958, tpr4->getR2Adjusted(),
            DBL_PRECISION_COMPARE);

    delete tpr4;

}

void TPolynomialRegressionTest::testPolynomialRegressionPseudoConstituantFewAttribute3()
{
    //
    URANIE::Modeler::TPolynomialRegression *tpr2 =
            new URANIE::Modeler::TPolynomialRegression(_tds, "z3:z5:z8:z10",
                    _sY, "nointercept");
    tpr2->unsetLog();
    tpr2->setDrawProgressBar(kFALSE);
    tpr2->setCriteria(0.55);

    CPPUNIT_ASSERT_THROW(tpr2->estimate("R2"),
            URANIE::Exceptions::UErrorExceptions);

    delete tpr2;
}

void TPolynomialRegressionTest::testPolynomialRegressionPseudoConstituantFewAttribute3All()
{
    //

    URANIE::DataServer::TDataServer * theTDS =
            new URANIE::DataServer::TDataServer();
    theTDS->fileDataRead("tplr.dat");

    URANIE::Modeler::TPolynomialRegression *tpr2 =
            new URANIE::Modeler::TPolynomialRegression(theTDS, "x1:x2:x3", "y",
                    "nointercept");
    tpr2->unsetLog();
    tpr2->setDrawProgressBar(kFALSE);
    tpr2->setCriteria(0.9800);

    CPPUNIT_ASSERT_THROW(tpr2->estimate("Q2"),
            URANIE::Exceptions::UErrorExceptions);
    
    delete tpr2;
    delete theTDS;

}

void TPolynomialRegressionTest::testPolynomialRegressionPseudoConstituantFewAttribute4()
{
    // 
    URANIE::Modeler::TPolynomialRegression *tpr3 =
            new URANIE::Modeler::TPolynomialRegression(_tds, "z3:z5:z8:z10",
                    _sY, "nointercept");
    tpr3->unsetLog();
    tpr3->setDrawProgressBar(kFALSE);

    tpr3->setCriteria(0.55);
    tpr3->_nMaxOrder = 7;

    CPPUNIT_ASSERT_THROW(tpr3->estimate("R2"),
            URANIE::Exceptions::UErrorExceptions);

    delete tpr3;
}

void TPolynomialRegressionTest::testPolynomialRegressionPseudoConstituantExhaustive()
{
    URANIE::Modeler::TPolynomialRegression *tpr =
            new URANIE::Modeler::TPolynomialRegression(_tds, _sX, _sY,
                    "nointercept");
    tpr->unsetLog();
    tpr->setDrawProgressBar(kFALSE);

    // Default Value
    CPPUNIT_ASSERT_EQUAL(-1, tpr->getMaxMonomes());

    // Current Value
    tpr->setMaxMonomes(2);
    CPPUNIT_ASSERT_EQUAL(2, tpr->getMaxMonomes());

    tpr->estimate("Q2");

    CPPUNIT_ASSERT(_tds->isAttribute("z5z8") == kFALSE);
    CPPUNIT_ASSERT(_tds->isAttribute("z3z10") == kFALSE);
    CPPUNIT_ASSERT(_tds->isAttribute("z1z10") == kFALSE);
    CPPUNIT_ASSERT(_tds->isAttribute("z3z8") == kFALSE);
    CPPUNIT_ASSERT(_tds->isAttribute("z1z7"));
    CPPUNIT_ASSERT(_tds->isAttribute("z1z11"));

    CPPUNIT_ASSERT(_tds->isAttribute(Form("%shat", _sY.Data())));

    TVectorD dparam = tpr->getParameters();

    CPPUNIT_ASSERT_EQUAL(13, dparam.GetNrows());

    CPPUNIT_ASSERT_DOUBLES_EQUAL(186.626068434883, dparam(0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(28.670005423237, dparam(1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(24.672663834895, dparam(2),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-126.49948022591, dparam(3),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(77.640903408966, dparam(4),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(84.346585689805, dparam(5),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(54.189352322378, dparam(6),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(8.804971046443, dparam(7),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(315.889094691235, dparam(8),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-130.939005979089, dparam(9),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-168.388806355761, dparam(10),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(953.734845516051, dparam(11),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-1380.78437840366, dparam(12),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.992063601226046, tpr->getR2(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.978458346184982, tpr->getR2Adjusted(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.953805491140515, tpr->getQ2(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1838.12339230231, tpr->getPress(), 1e-9);

    // Clean
    delete tpr;
    tpr = NULL;
}

void TPolynomialRegressionTest::testPolynomialRegressionTrace()
{
    TString strDefault = "_tpr_trace_.log";
    URANIE::Modeler::TPolynomialRegression *tpr =
            new URANIE::Modeler::TPolynomialRegression(_tds, _sX, _sY,
                    "nointercept");
    tpr->unsetLog();

    // Default Value
    CPPUNIT_ASSERT(!tpr->getTrace());
    CPPUNIT_ASSERT(!strcmp("", Form("%s", tpr->getTraceFile().Data())));
    CPPUNIT_ASSERT(strcmp("titi.log", Form("%s", tpr->getTraceFile().Data())));

    // Activate Trace with default file name
    tpr->setTrace();
    CPPUNIT_ASSERT(tpr->getTrace());
    CPPUNIT_ASSERT(
            !strcmp(strDefault.Data(), Form("%s", tpr->getTraceFile().Data())));
    CPPUNIT_ASSERT(strcmp("titi.log", Form("%s", tpr->getTraceFile().Data())));

    // Desactivate the trace
    tpr->unsetTrace();
    CPPUNIT_ASSERT(!tpr->getTrace());
    CPPUNIT_ASSERT(!strcmp("", Form("%s", tpr->getTraceFile().Data())));
    CPPUNIT_ASSERT(strcmp("titi.log", Form("%s", tpr->getTraceFile().Data())));

    // Activate Trace with a specific file name
    TString strFileName = "_toto_.log";
    tpr->setTrace(strFileName);
    CPPUNIT_ASSERT(tpr->getTrace());
    CPPUNIT_ASSERT(
            !strcmp(strFileName.Data(),
                    Form("%s", tpr->getTraceFile().Data())));
    CPPUNIT_ASSERT(strcmp("titi.log", Form("%s", tpr->getTraceFile().Data())));

    // Desactivate the trace
    tpr->unsetTrace();
    CPPUNIT_ASSERT(!tpr->getTrace());
    CPPUNIT_ASSERT(!strcmp("", Form("%s", tpr->getTraceFile().Data())));
    CPPUNIT_ASSERT(strcmp("titi.log", Form("%s", tpr->getTraceFile().Data())));

    // Clean
    delete tpr;
    tpr = NULL;
}

void TPolynomialRegressionTest::testsetMinimalMaximalOrder()
{
    URANIE::Modeler::TPolynomialRegression *tpr =
            new URANIE::Modeler::TPolynomialRegression(_tds, _sX, _sY,
                    "nointercept");
    tpr->unsetLog();
    tpr->setDrawProgressBar(kFALSE);
    tpr->setMinimalOrder(0);
    tpr->setMaximalOrder(1);

    CPPUNIT_ASSERT_EQUAL(0, tpr->_nMinOrder);
    CPPUNIT_ASSERT_EQUAL(1, tpr->_nMaxOrder);

    delete tpr;
}

void TPolynomialRegressionTest::testPolynomialRegressionR2A()
{
    URANIE::Modeler::TPolynomialRegression *tpr =
            new URANIE::Modeler::TPolynomialRegression(_tds, _sX, _sY,
                    "nointercept");
    tpr->setCriteria(0.97);
    tpr->setDrawProgressBar(kFALSE);
    tpr->estimate("R2A");

    CPPUNIT_ASSERT(_tds->isAttribute("z1z7"));
    CPPUNIT_ASSERT(_tds->isAttribute("z3z8") == kFALSE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.992063601226, tpr->getR2(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.978458346184, tpr->getR2Adjusted(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.953805491140, tpr->getQ2(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1838.123392302, tpr->getPress(), 1e-9);

    delete tpr;
}

void TPolynomialRegressionTest::testPolynomialRegressionR2()
{
    URANIE::Modeler::TPolynomialRegression *tpr =
            new URANIE::Modeler::TPolynomialRegression(_tds, _sX, _sY,
                    "nointercept");
    tpr->setCriteria(0.975);
    tpr->setDrawProgressBar(kFALSE);
    tpr->estimate("R2");

    CPPUNIT_ASSERT(_tds->isAttribute("z1z7"));
    CPPUNIT_ASSERT(_tds->isAttribute("z3z8") == kFALSE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.992063601226, tpr->getR2(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.978458346184, tpr->getR2Adjusted(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.953805491140, tpr->getQ2(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1838.123392302, tpr->getPress(), 1e-9);

    delete tpr;
}

void TPolynomialRegressionTest::throwExceptionPolynomialRegressionMaxOrder()
{
    URANIE::Modeler::TPolynomialRegression *tpr =
            new URANIE::Modeler::TPolynomialRegression(_tds, _sX, _sY,
                    "nointercept");
    tpr->unsetLog();
    CPPUNIT_ASSERT_THROW(tpr->setMaximalOrder(3),
            URANIE::Exceptions::UErrorExceptions);

    // Clean
    delete tpr;
    tpr = NULL;
}

void TPolynomialRegressionTest::throwExceptionPolynomialRegressionCriteriaValue()
{
    URANIE::Modeler::TPolynomialRegression *tpr =
            new URANIE::Modeler::TPolynomialRegression(_tds, _sX, _sY,
                    "nointercept");
    tpr->unsetLog();
    CPPUNIT_ASSERT_THROW(tpr->setCriteria(1.2),
            URANIE::Exceptions::UErrorExceptions);

    // Clean
    delete tpr;
    tpr = NULL;
}

void TPolynomialRegressionTest::throwExceptionPolynomialRegressionEstimateValue()
{
    URANIE::Modeler::TPolynomialRegression *tpr =
            new URANIE::Modeler::TPolynomialRegression(_tds, _sX, _sY,
                    "nointercept");
    tpr->unsetLog();
    tpr->setDrawProgressBar(kFALSE);
    CPPUNIT_ASSERT_THROW(tpr->estimate("Q3"),
            URANIE::Exceptions::UErrorExceptions);

    // Clean
    delete tpr;
    tpr = NULL;
}

void TPolynomialRegressionTest::throwExceptionPolynomialRegressionNoPolynom()
{
    URANIE::Modeler::TPolynomialRegression *tpr =
            new URANIE::Modeler::TPolynomialRegression(_tds, _sX, _sY,
                    "nointercept");
    tpr->unsetLog();
    tpr->setCriteria(0.99999999999);
    tpr->setDrawProgressBar(kFALSE);
    CPPUNIT_ASSERT_THROW(tpr->estimate("Q2"),
            URANIE::Exceptions::UErrorExceptions);

    // Clean
    delete tpr;
    tpr = NULL;
}

void TPolynomialRegressionTest::throwExceptionPolynomialRegressionMinOrder()
{
    URANIE::Modeler::TPolynomialRegression *tpr =
            new URANIE::Modeler::TPolynomialRegression(_tds, _sX, _sY,
                    "nointercept");
    tpr->unsetLog();
    CPPUNIT_ASSERT_THROW(tpr->setMinimalOrder(-1),
            URANIE::Exceptions::UErrorExceptions);

    // Clean
    delete tpr;
    tpr = NULL;
}
void TPolynomialRegressionTest::throwExceptionPolynomialRegressionCriteriaNegative()
{
    URANIE::Modeler::TPolynomialRegression *tpr =
            new URANIE::Modeler::TPolynomialRegression(_tds, _sX, _sY,
                    "nointercept");
    tpr->unsetLog();
    CPPUNIT_ASSERT_THROW(tpr->setCriteria(-1.2),
            URANIE::Exceptions::UErrorExceptions);

    // Clean
    delete tpr;
    tpr = NULL;
}

void TPolynomialRegressionTest::throwExceptionPolynomialRegressionMaxOrderNegative()
{
    URANIE::Modeler::TPolynomialRegression *tpr =
            new URANIE::Modeler::TPolynomialRegression(_tds, _sX, _sY,
                    "nointercept");
    tpr->unsetLog();
    CPPUNIT_ASSERT_THROW(tpr->setMaximalOrder(-1),
            URANIE::Exceptions::UErrorExceptions);

    // Clean
    delete tpr;
    tpr = NULL;
}
void TPolynomialRegressionTest::throwExceptionPolynomialRegressionMaxOrderInfMinOrder()
{
    URANIE::Modeler::TPolynomialRegression *tpr =
            new URANIE::Modeler::TPolynomialRegression(_tds, _sX, _sY,
                    "nointercept");
    tpr->unsetLog();
    CPPUNIT_ASSERT_THROW(tpr->setMaximalOrder(0),
            URANIE::Exceptions::UErrorExceptions);

    // Clean
    delete tpr;
    tpr = NULL;
}
void TPolynomialRegressionTest::throwExceptionPolynomialRegressionMinOrderSupMaxOrder()
{
    URANIE::Modeler::TPolynomialRegression *tpr =
            new URANIE::Modeler::TPolynomialRegression(_tds, _sX, _sY,
                    "nointercept");
    tpr->unsetLog();
    CPPUNIT_ASSERT_THROW(tpr->setMinimalOrder(3),
            URANIE::Exceptions::UErrorExceptions);

    // Clean
    delete tpr;
    tpr = NULL;
}
void TPolynomialRegressionTest::throwExceptionPolynomialRegressionMultiOutput()
{
    TString _sInput = "z1:z2:z3:z4:z5:z6:z7:z8:z9:z10:z11";
    TString _sOutput = "visco:lnvisco";
    URANIE::Modeler::TPolynomialRegression *tpr =
            new URANIE::Modeler::TPolynomialRegression(_tds, _sInput, _sOutput,
                    "nointercept");
    tpr->unsetLog();
    tpr->setDrawProgressBar(kFALSE);
    CPPUNIT_ASSERT_THROW(tpr->estimate("Q2"),
            URANIE::Exceptions::UErrorExceptions);

    // Clean
    delete tpr;
    tpr = NULL;
}
void TPolynomialRegressionTest::testExportFunction()
{
    URANIE::Modeler::TPolynomialRegression *tpr =
            new URANIE::Modeler::TPolynomialRegression(_tds, _sX, _sY,
                    "nointercept");
    tpr->unsetLog();
    tpr->setCriteria(0.9);
    tpr->setDrawProgressBar(kFALSE);
    tpr->estimate("Q2");

    TString cFile = "modelTPR.C";
    std::ofstream * cfile = new std::ofstream(cFile.Data());
    tpr->exportModelCplusplus(cfile);
    string line;
    Bool_t bStop = kFALSE;
    Int_t test = 0;
    std::fstream file1("modelTPR.C");
    while (bStop == kFALSE)
    {
        getline(file1, line);
        if (line.find("  y += param[13]*-711.102;") != string::npos)
        {
            bStop = kTRUE;
            test = 1;
        }
        if (file1.eof())
            bStop = kTRUE;
    }
    CPPUNIT_ASSERT_EQUAL(1, test);
    delete cfile;
    cfile = NULL;

    TString fFile = "modelTPR.f";
    std::ofstream * fortranfile = new std::ofstream(fFile.Data());
    tpr->exportModelFortran(fortranfile);
    test = 0;
    bStop = kFALSE;
    std::fstream file2("modelTPR.f");
    while (bStop == kFALSE)
    {
        getline(file2, line);
        if (line.find("      visco = visco + z11 * -587.573") != string::npos)
        {
            bStop = kTRUE;
            test = 1;
        }
        if (file2.eof())
            bStop = kTRUE;
    }
    CPPUNIT_ASSERT_EQUAL(1, test);
    delete fortranfile;
    fortranfile = NULL;

    // Clean
    delete tpr;
    tpr = NULL;
}
void TPolynomialRegressionTest::throwExceptionExportFunction()
{
    URANIE::Modeler::TPolynomialRegression *tpr =
            new URANIE::Modeler::TPolynomialRegression(_tds, _sX, _sY,
                    "nointercept");
    tpr->unsetLog();
    tpr->setCriteria(0.9);
    tpr->setDrawProgressBar(kFALSE);
    tpr->estimate("Q2");
    CPPUNIT_ASSERT_THROW(
            tpr->exportFunction("toto", "model_polynomialregression",
                    "function_name"), URANIE::Exceptions::UErrorExceptions);
    // Clean
    delete tpr;
    tpr = NULL;
}
