//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestUserFunctions.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   2012-09-14 10:45:11

 \brief Implement the user function 

 The analytical function is givent by the formula
 
 \$[ f(x_1, x_2) = a * ( x_2 - x_1^2 )^2 + b * ( 1.0 - x_1 )^2 \$]

 with \f$ a=10.0, b=1.0 \f$.

 */

// ROOT
#include "TROOT.h"
#include "TMath.h"

/****************** User Function **********************/
//! The true flow rate model
/*!
 It describes the flow rate of water through a borehole from an upper aquifer to a lower aquifer separated by an impermeable rock layer.
 The response variable \f$ y\f$, the \b flow \b water through the borehole in \f$ m^{3}/yr\f$, is given by :
 \f[
 y = f(x) = \frac{2 \pi T_u\left( H_u - H_l\right)}{\ln(\frac{r}{r_{\omega}})\left[ 1 + \frac{2 L T_u}{\ln(\frac{r}{r_{\omega}}) r_{\omega}^2 K_{\omega}} + \frac{T_u}{T_l}\right]}
 \f]

 \param x (double *) the input vector in the order :
 \li \f$ r_{\omega} \in [0.05, 0.15]\f$ : radius of borehole (m)
 \li \f$ r \in [100, 50~000]\f$ : radius of influence {\ix (m)}
 \li \f$ T_u \in [63~070, 115~600]\f$ : Transmissivity of upper aquifer (\f$ m^{2}\f$/year)
 \li \f$ T_l \in [63.1, 116]\f$ : Transmissivity of lower aquifer (\f$ m^{2}\f$/year)
 \li \f$ H_u \in [990, 1~110]\f$ : Potentiometric head of upper aquifer (m)
 \li \f$ H_l \in [700, 820]\f$ : Potentiometric head of lower aquifer (m)
 \li \f$ L \in [1~120, 1~680]\f$ : length of borehole (m)
 \li \f$ K_{\omega} \in [9~855, 12~045]\f$ : hydraulic conductivity of borehole (m)
 \param y (double *) the output vector

 \see waterboreholeHoXu
 */
// void flowrateModel(double *x, double *y)
// {
//     Double_t drw = x[0], dr = x[1];
//     Double_t dtu = x[2], dtl = x[3];
//     Double_t dhu = x[4], dhl = x[5];
//     Double_t dl = x[6], dkw = x[7];
// 
//     Double_t dnum = 2.0 * TMath::Pi() * dtu * (dhu - dhl);
//     Double_t dlnronrw = TMath::Log(dr / drw);
//     Double_t dden =
//             dlnronrw
//                     * (1.0 + (2.0 * dl * dtu) / (dlnronrw * drw * drw * dkw)
//                             + dtu / dtl);
// 
//     y[0] = dnum / dden;
// }

// void flowrateModelSquared(double *x, double *y)
// {
//     Double_t drw = x[0], dr = x[1];
//     Double_t dtu = x[2], dtl = x[3];
//     Double_t dhu = x[4], dhl = x[5];
//     Double_t dl = x[6], dkw = x[7];
// 
//     Double_t dnum = 2.0 * TMath::Pi() * dtu * (dhu - dhl);
//     Double_t dlnronrw = TMath::Log(dr / drw);
//     Double_t dden =
//             dlnronrw
//                     * (1.0 + (2.0 * dl * dtu) / (dlnronrw * drw * drw * dkw)
//                             + dtu / dtl);
// 
//     y[0] = dnum / dden;
//     y[1] = x[0] * x[0];
// 
// }

//! The surrogate \em flow \em rate model (polynomial regression) proposed Ho \em and Xu (2000) is
/*!
 The polynomial regression models is given by the formula :
 \f[
 \hat{log(y)}  = 4.156 + 1.9903(\log{r_{\omega}}+2.3544) - 0.0007292(L-1400) - .003554(H_l - 760) \\
  \f]
 \f[
 \hspace*{5cm} + .0035068(H_u-1050) + .000090868(K_{\omega}-10950) - .00087286 (\log{(r)} -8.8914) \\
  \f]
 \f[
 + .000015325 (H_u-1050) (H_l - 760) + .00000026487 (L-1400)^2 \\
  \f]
 \f[
 \hspace*{5cm} - .0000071759 (H_l - 760)^2 - .0000068021 (H_u-1050)^2 \\
  \f]
 
 \param x (double *) the input vector in the order :
 \li \f$ r_{\omega} \in [0.05, 0.15]\f$ : radius of borehole (m)
 \li \f$ r \in [100, 50~000]\f$ : radius of influence {\ix (m)}
 \li \f$ T_u \in [63~070, 115~600]\f$ : Transmissivity of upper aquifer (\f$ m^{2}\f$/year)
 \li \f$ T_l \in [63.1, 116]\f$ : Transmissivity of lower aquifer (\f$ m^{2}\f$/year)
 \li \f$ H_u \in [990, 1~110]\f$ : Potentiometric head of upper aquifer (m)
 \li \f$ H_l \in [700, 820]\f$ : Potentiometric head of lower aquifer (m)
 \li \f$ L \in [1~120, 1~680]\f$ : length of borehole (m)
 \li \f$ K_{\omega} \in [9~855, 12~045]\f$ : hydraulic conductivity of borehole (m)
 \param y (double *) the output vector

 \see flowrateModel
 
 */

// void HoXuSurrogateModel(double *x, Double_t *y)
// {
//     Double_t drw = x[0], dr = x[1];
// //   Double_t dtu = x[2], dtl = x[3];
//     Double_t dhu = x[4] - 1050.0, dhl = x[5] - 760.0;
//     Double_t dl = x[6] - 1400.0, dkw = x[7] - 10950.0;
// 
//     Double_t yHoXu = 4.1560 + 1.9903 * (TMath::Log(drw) + 2.3544);
//     yHoXu -= .0007292 * dl;
//     yHoXu -= .003554 * dhl;
//     yHoXu += .0035068 * dhu;
//     yHoXu += .000090868 * dkw;
//     yHoXu += .000015325 * dhu * dhl;
//     yHoXu += .00000026487 * dl * dl;
//     yHoXu -= .0000071759 * dhl * dhl;
//     yHoXu -= .0000068021 * dhu * dhu;
//     yHoXu -= .00087286 * (TMath::Log(dr) - 8.8914);
// 
//     y[0] = TMath::Exp(yHoXu);
// 
// }

//! Sum function; adds the two first values of x in y
void fSum(double *x, double *y)
{
    y[0] = x[0] + x[1];
}

//! Product function; multiplies the two first values of x in y
// void fProd(double *x, double *y)
// {
//     y[0] = x[0] * x[1];
// }

//! Ishigami function
void FctIshigami(Double_t *x, Double_t *y)
{
    Double_t u1, u2, u3, s1, s2, a, b;
    a = 7.;
    b = 0.1;
    u1 = x[0];
    u2 = x[1];
    u3 = x[2];
    s1 = TMath::Sin(u1);
    s2 = TMath::Sin(u2);
    y[0] = s1 + a * s2 * s2 + b * u3 * u3 * u3 * u3 * s1;
}

//! Linear function
void FctLinear(Double_t *x, Double_t *y)
{
    Double_t u1, u2, u3, u4;
    u1 = x[0];
    u2 = x[1];
    u3 = x[2];
    u4 = x[3];

    y[0] = u1 + u1 * u1 * u1 + u2 + u3 + u4 + u4 * u4 * u4;
}

//! Venise function
void FctVenise(Double_t *x, Double_t *y)
{
    Double_t u1, u2, a, b, c;
    a = 1;
    b = 2;
    c = 3;
    u1 = x[0];
    u2 = x[1];
    y[0] = a * u1 * u1 + b * u1 * u2 + c * u2;
}

/****************** End of User Function **********************/
