/**
 * @file   UserFunctions.C
 * @brief  User Functions
 */
/****************** User Function **********************/
//! Ishigami model
/*!
 It describes the .
 The response variable 

 \param x (double *) the input vector
 \param y (double *) the output vector

 */
void Ishigami(Double_t *x, Double_t *y)
{
    Double_t u1, u2, u3, s1, s2, s3, a, b;
    a = 7.;
    b = 0.1;
    u1 = x[0];
    u2 = x[1];
    u3 = x[2];
    s1 = sin(u1);
    s2 = sin(u2);
    s3 = sin(u3);
    y[0] = s1 + a * s2 * s2 + b * u3 * u3 * u3 * u3 * s1;
    y[1] = u1 + 2. * u2 + 3. * u3;
}

//! Flow rate model
/*!
 It describes the flow rate of water through a borehole from an upper aquifer to a lower aquifer separated by an impermeable rock layer.
 The response variable \f$ y\f$, the \b flow \b water through the borehole in \f$ m^{3}/yr\f$, is given by :
 \f[
 y = f(x) = \frac{2 \pi T_u\left( H_u - H_l\right)}{\ln(\frac{r}{r_{\omega}})\left[ 1 + \frac{2 L T_u}{\ln(\frac{r}{r_{\omega}}) r_{\omega}^2 K_{\omega}} + \frac{T_u}{T_l}\right]}
 \f]

 \param x (double *) the input vector in the order :
 \li \f$ r_{\omega} \in [0.05, 0.15]\f$ : radius of borehole (m)
 \li \f$ r \in [100, 50~000]\f$ : radius of influence {\ix (m)}
 \li \f$ T_u \in [63~070, 115~600]\f$ : Transmissivity of upper aquifer (\f$ m^{2}\f$/year)
 \li \f$ T_l \in [63.1, 116]\f$ : Transmissivity of lower aquifer (\f$ m^{2}\f$/year)
 \li \f$ H_u \in [990, 1~110]\f$ : Potentiometric head of upper aquifer (m)
 \li \f$ H_l \in [700, 820]\f$ : Potentiometric head of lower aquifer (m)
 \li \f$ L \in [1~120, 1~680]\f$ : length of borehole (m)
 \li \f$ K_{\omega} \in [9~855, 12~045]\f$ : hydraulic conductivity of borehole (m)
 \param y (double *) the output vector

 */
void function_FlowRate(Double_t *x, Double_t *y)
{
    Double_t drw = x[0], dr = x[1], dtu = x[2], dtl = x[3];
    Double_t dhu = x[4], dhl = x[5], dl = x[6], dkw = x[7];

    Double_t dnum = 2.0 * TMath::Pi() * dtu * (dhu - dhl);
    Double_t dlnronrw = TMath::Log(dr / drw);
    Double_t dden =
            dlnronrw
                    * (1.0 + (2.0 * dl * dtu) / (dlnronrw * drw * drw * dkw)
                            + dtu / dtl);
    y[0] = dnum / dden;
}

