//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TMultiGenCode.h
 \author uranietm <uranietm@cea.fr>
 \date   mar sep 29 10:00:00 2010
 
 \brief Example of a function to be minimized which calls an external code TCode and a dataserver whith the new TMinuit2 wrapper

 This is tested for Rosenbrock function defined as follows:
 \em Rosenbrock function is d�fined by the equation
 \f[
 f(x, y) \, = \, 100( y - x^{2})^{2} + ( 1 - x )^{2}
 \f]

 The minimun is reached when \f$ ( \star{x^{0}}, \star{y^{0}} ) \, = \, ( 1.0, 1.0 ) \f$

 A possible beginning point may be : \f$ ( x^{0}, y^{0} ) \, = \, ( -1.2, 1.0 )\f$ with f equal to 24.20.


 
 \see http://root.cern.ch/drupal/content/numerical-minimization
 \see http://root.cern.ch/viewvc/trunk/math/minuit2/test/testNewMinimizer.cxx?view=markup

 */
#ifndef TMULTIGENCODE_H
#define TMULTIGENCODE_H

// ROOT
#include "Math/IFunction.h"
#include "TH1.h"
#include "TF1.h"
#include "TFile.h"
#include "TNtupleD.h"
#include "TTreeFormula.h"
#include "TPad.h"
#include "TRandom3.h"
#include "TStyle.h"
#include "TSystem.h"

#include <vector>
#include <iostream>

// URANIE
#include "Optimizer.h"
#include "TCode.h"
#include "TDataServer.h"

namespace URANIE
{
namespace Optimizer
{
class TMultiGenCode: public ROOT::Math::IBaseFunctionMultiDim
{

public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Constructor with a dataserver
    TMultiGenCode(URANIE::DataServer::TDataServer *tds,
            URANIE::Launcher::TCode *code, TString scost);
    //! Desctructor
    ~TMultiGenCode();

    //---------------------------------------------
    /** @name Methods
     **/
    //@{
    //! The init method
    void init();
    //! The clean method
    void clean();
    //@}

    //---------------------------------------------
    /** @name pure virtual methods
     **/
    //@{
    //! The clone method
    TMultiGenCode * Clone() const
    {
        return new TMultiGenCode(_tds, _code, _sCost);
    }

    //! the NDIM method
    unsigned int NDim() const
    {
        return _ninput;
    }
    //@}
    
    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    void setLog()
    {
        _blog = kTRUE;
    }
    void unsetLog()
    {
        _blog = kFALSE;
    }
    
    void changeLog()
    {
        _blog = _blog ? kFALSE : kTRUE;
    }
    Bool_t getLog()
    {
        return _blog;
    }
    
    void doNotSaveSteps()
    {
        _bintermedstep = true;
    }
    //@}
    

private:
    //---------------------------------------------
    /** @name pure virtual methods
     **/
    //@{
    //! the DoEval method
    double DoEval(const double * x) const;
    //@}

    Bool_t _blog;                            ///< Boolean for edit the log
    int _ninput;
    TString _sinput;                          ///< Name of the input attributes
    double *_dinputValue;
    double *_xin, *_codevalue;
    int *_codeindex ;
    int _noutput;
    int _ninformula;
    int _noutformula;
    TString _soutput;                         ///< Name of the output attributes
    double * _doutputValue;                       ///< The output attribute name
    TString _sCost; ///< The name of the selected cost (the number must be equal to 1)
    Int_t _nCost; ///< the current position for the selected cost in the output list of attribute
    URANIE::DataServer::TDataServer *_tds;              //! Pointeur vers un TDS
    URANIE::Launcher::TCode *_code;    //! Pointeur vers un TCode
    TNtupleD *_informTree; 
    vector<TTreeFormula*> _vinForm;
    bool _bintermedstep;
    
    ClassDef(URANIE::Optimizer::TMultiGenCode, ID_OPTIMIZER)
    //Classe de

};
} // Fin du namespace Optimizer
} // Fin du namespace URANIE

R__EXTERN URANIE::Optimizer::TMultiGenCode *gTMGC;

#endif
