//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TMultiGenSumOfSquares.h
 \author uranietm <uranietm@cea.fr>
 \date   mer sep 29 18:00:00 2010
 
 \brief Example of a function to be minimised : Sum of Square with the new TMinuit2 wrapper.

 This is tested for Rosenbrock function defined as follows:

 
 \see http://root.cern.ch/drupal/content/numerical-minimization
 \see http://root.cern.ch/viewvc/trunk/math/minuit2/test/testNewMinimizer.cxx?view=markup

 */
#ifndef TMULTIGENSUMOFSQUARES_H
#define TMULTIGENSUMOFSQUARES_H

// ROOT
#include "Math/IFunction.h"
#include "TH1.h"
#include "TF1.h"
#include "TFile.h"
#include "TPad.h"
#include "TRandom3.h"
#include "TStyle.h"
#include "TSystem.h"

#include <vector>
#include <iostream>

// URANIE
#include "Optimizer.h"
#include "TCode.h"
#include "TDataServer.h"
#include "TObjective.h"

namespace URANIE
{
namespace Optimizer
{
class TMultiGenSumOfSquares: public ROOT::Math::IBaseFunctionMultiDim
{
private:
    vector<TObjective *> _objectives;            ///< vector of objectives      
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Constructor with a dataserver
    TMultiGenSumOfSquares(URANIE::DataServer::TDataServer *tds,
            URANIE::Launcher::TCode *code);
    //! Desctructor
    ~TMultiGenSumOfSquares();

    //---------------------------------------------
    /** @name Methods
     **/
    //@{
    //! The init method
    void init();
    //! The clean method
    void clean();
    //@}
    void addObjective(TObjective *tobj);
    Int_t getNObjectives()
    {
        Int_t nres = (Int_t)(_objectives.size());
        return nres;
    }
    const char * getCriteriaName()
    {
        return _sCriteria.Data();
    }
    Double_t getSumOfWeight()
    {
        return _dSumOfWeight;
    }
    //---------------------------------------------
    /** @name pure virtual methods
     **/
    //@{
    //! The clone method
    TMultiGenSumOfSquares * Clone() const
    {
        TMultiGenSumOfSquares * theClone = new TMultiGenSumOfSquares(_tds,
                _code);
        for (Int_t indx = 0; indx < (Int_t)(_objectives.size()); indx++)
            theClone->addObjective(_objectives[indx]);
        theClone->init();
        return theClone;
    }

    //! the NDIM method
    unsigned int NDim() const
    {
        return _ninput;
    }
    //@}
    
    
     //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    void setLog()
    {
        _blog = kTRUE;
    }
    void unsetLog()
    {
        _blog = kFALSE;
    }
    
    //@}
    

private:
    //---------------------------------------------
    /** @name pure virtual methods
     **/
    //@{
    //! the DoEval method
    double DoEval(const double * x) const;
    //@}

    Bool_t _blog;                            ///< Boolean for edit the log
    int _ninput;
    double * _dinputValue;
    int _noutput;
    double * _doutputValue;
    Double_t _dSumOfWeight;                              ///< The sum of weights
    URANIE::DataServer::TDataServer *_tds;               //! Pointer vers un TDS
    URANIE::Launcher::TCode *_code;                //! Pointer vers un TCode
    TList *_listOfDataOutput;    //! List of matrix
    TString _sCriteria;      //! The name of the attribute which is the criteria

    ClassDef(URANIE::Optimizer::TMultiGenSumOfSquares, ID_OPTIMIZER)
    //Classe de

};
} // Fin du namespace Optimizer
} // Fin du namespace URANIE

R__EXTERN URANIE::Optimizer::TMultiGenSumOfSquares *gTMGSOS;

#endif
