//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTVizir.cxx
 \author Maud Salvoldi (OCC/Areva group)
 \date   $Date$
 
 \brief  Fichier de test de la classe URANIE::Optimizer::TVizir
 
 Cree le : jeu dec 22 11:27:34 CET 2011
 
 */

#define DBL_PRECISION_COMPARE 1.e-12
#define EPSILON 0.01 

// Uranie
#include "TestTBestEstimate.h"
#include "UDisableTest.h"

// ROOT
#include "TString.h"
#include "TMatrixD.h"
#include "TROOT.h"
#include "TSystem.h" 

#include <math.h>
#include <iostream>
#include <iomanip>
using namespace URANIE::Optimizer;

TBestEstimateTest::TBestEstimateTest() :
        _tds(NULL),_tbe(NULL)
{
    
}

TBestEstimateTest::~TBestEstimateTest()
{

}
void TBestEstimateTest::setUp()
{
    
}
void TBestEstimateTest::tearDown()
{
  if(_tbe != NULL)
      delete _tbe;
  
}

void readMatrixFile(TString sfile, TMatrixD &mat)
{
//function to read a sparse matrix and to produce a normal one
  std::ifstream in;
  in.open(Form("%s",sfile.Data()));

  Int_t ni, nj, nz;
  in >> ni >> nj >> nz;
  mat.ResizeTo(ni, nj);
  mat.Zero();

  Double_t dval;
  for(Int_t i=0; i< nz; i++) {
    in >> ni >> nj >> dval;
    mat(ni-1, nj-1) = dval;
  }

  in.close();
}

// void writeMatrixFile(TString sfile, TMatrixD &mat)
// {
// //function to write a normal matrix in a file as sparsed one
//   ofstream out;
//   out.open(Form("%s",sfile.Data()));
//   out.precision(18);
// 
//   Int_t ni, nj, nz;
//   ni = mat.GetNrows();
//   nj = mat.GetNcols();
//   nz = 0;
//   Int_t index=0;
//   Double_t *pData = mat.GetMatrixArray();
//   for(Int_t i=0; i < ni; i++) {
//     for(Int_t j=0; j < nj; j++) {
//       if(pData[index] != 0.0) nz++;
//       index++;      
//     }
//   }
//   out << ni << ' ' << nj << ' ' << nz << endl;
//   
//   index=0;
//   for(Int_t i=0; i < ni; i++) {
//     for(Int_t j=0; j < nj; j++) {
//       if(pData[index] != 0.0) {
//         out << i+1 << ' ' << j+1 << ' ' << pData[index] << endl;
//       }
//       index++;      
//     }
//   }
// 
// 
//   out.close();
//   delete pData;
// }

////////////////////////////////
///                          ///                            
///           test           ///
///                          ///
////////////////////////////////

void TBestEstimateTest::testTBestEstimateEmptyConstructor()
{
    _tbe = new TBestEstimate();
    _tbe->unsetLog();
    
    CPPUNIT_ASSERT(_tbe->getParameterData() == NULL);
    CPPUNIT_ASSERT(_tbe->getMeasureData() == NULL);
    CPPUNIT_ASSERT(_tbe->getResponseData() == NULL);
    CPPUNIT_ASSERT(_tbe->getOATSamplingData() == NULL);
    CPPUNIT_ASSERT_EQUAL(0,_tbe->getNInputParameter());
    CPPUNIT_ASSERT_EQUAL(0,_tbe->getNResponses());
    CPPUNIT_ASSERT_EQUAL(0,_tbe->getNTimeStepsInputParameter());
    CPPUNIT_ASSERT_EQUAL(0,_tbe->getNTimeStepsResponses());
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0,_tbe->getChi2(),DBL_PRECISION_COMPARE);
    
}

void TBestEstimateTest::testTBestEstimateConstructorWithMatrices()
{
    TMatrixD a, ca, car, cm, m, r, s;
    
    // Input data
    readMatrixFile( "vorBESTEST/a.abs",   a);   // vector of parameters \alpha (nominal values ?)
    readMatrixFile( "vorBESTEST/ca.abs",  ca);  // covariance matrix of parameters \alpha
    readMatrixFile( "vorBESTEST/r.abs",   r);   // vector of computed responses R
    readMatrixFile( "vorBESTEST/s.abs",   s);   // sensitivity matrix (\frac{\partial R}{\partial \alpha})
    readMatrixFile( "vorBESTEST/m.abs",   m);   // vector of measured responses M
    readMatrixFile( "vorBESTEST/cm.abs",  cm);  // covariance matrix of measured responses
    readMatrixFile( "vorBESTEST/car.abs", car); // "response-parameter" covariance matrix (using measured responses ?)
    
    Int_t nbParam = 3;
    Int_t nbResp = 2;

    _tbe = new TBestEstimate(nbParam, a, nbResp, m, r, ca, cm, s, car);
    _tbe->unsetLog();
    
    TMatrixD matrixA = _tbe->getParameterMatrix();    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.13862943,matrixA(0,0),DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.17328679,matrixA(14,0),DBL_PRECISION_COMPARE);
    
    TMatrixD matrixR = _tbe->getResponseMatrix();    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(129.345837,matrixR(0,0),DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.511050677,matrixR(9,0),DBL_PRECISION_COMPARE);
    
    TMatrixD matrixM = _tbe->getMeasureMatrix();    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(125.048262,matrixM(0,0),DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.520277667,matrixM(9,0),DBL_PRECISION_COMPARE);
    
    TMatrixD matrixCA = _tbe->getParameterCovMatrix();    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.000555403684,matrixCA(0,0),DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.000507478496,matrixCA(14,14),DBL_PRECISION_COMPARE);
    
    TMatrixD matrixCM = _tbe->getMeasureCovMatrix();    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(150.768617,matrixCM(0,0),DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.00156374736,matrixCM(9,9),DBL_PRECISION_COMPARE);
    
    TMatrixD matrixS = _tbe->getSensitivityMatrix();    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.9330329915E3,matrixS(0,0),DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.5456455553E0,matrixS(9,14),DBL_PRECISION_COMPARE);
    
    
    TMatrixD matrixCAR = _tbe->getParameterResponseCovMatrix();    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0,matrixCAR(0,0),DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0,matrixCAR(9,14),DBL_PRECISION_COMPARE);
    
    
    CPPUNIT_ASSERT_EQUAL(3,_tbe->getNInputParameter());
    CPPUNIT_ASSERT_EQUAL(2,_tbe->getNResponses());
    CPPUNIT_ASSERT_EQUAL(5,_tbe->getNTimeStepsInputParameter());
    CPPUNIT_ASSERT_EQUAL(5,_tbe->getNTimeStepsResponses());
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0,_tbe->getChi2(),DBL_PRECISION_COMPARE);
    
}

void TBestEstimateTest::testTBestEstimateRunBestEstimate()
{
    TMatrixD a, ca, car, cm, m, r, s;
    
    // Input data
    readMatrixFile( "vorBESTEST/a.abs",   a);   // vector of parameters \alpha (nominal values ?)
    readMatrixFile( "vorBESTEST/ca.abs",  ca);  // covariance matrix of parameters \alpha
    readMatrixFile( "vorBESTEST/r.abs",   r);   // vector of computed responses R
    readMatrixFile( "vorBESTEST/s.abs",   s);   // sensitivity matrix (\frac{\partial R}{\partial \alpha})
    readMatrixFile( "vorBESTEST/m.abs",   m);   // vector of measured responses M
    readMatrixFile( "vorBESTEST/cm.abs",  cm);  // covariance matrix of measured responses
    readMatrixFile( "vorBESTEST/car.abs", car); // "response-parameter" covariance matrix (using measured responses ?)
    
    Int_t nbParam = 3;
    Int_t nbResp = 2;

    _tbe = new TBestEstimate(nbParam, a, nbResp, m, r, ca, cm, s, car);
    _tbe->unsetLog();
    
    _tbe->runBestEstimate();
    
    TMatrixD matrixABE= _tbe->getBestEstParameterMatrix();    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.139123362101639642,matrixABE(0,0),DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.172956085292615996,matrixABE(14,0),DBL_PRECISION_COMPARE);
    
    TMatrixD matrixCABE= _tbe->getBestEstParameterCovMatrix();    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.82734575788373078e-06,matrixCABE(0,0),DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.000376712973783356814,matrixCABE(14,14),DBL_PRECISION_COMPARE);
    
    TMatrixD matrixRBE= _tbe->getBestEstResponseMatrix();    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(129.806691946390686 ,matrixRBE(0,0),DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.522484567516859677,matrixRBE(9,0),DBL_PRECISION_COMPARE);
    
    TMatrixD matrixCRBE= _tbe->getBestEstResponsesCovMatrix();    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.59079687873577313,matrixCRBE(0,0),DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.000902955308561803919,matrixCRBE(9,9),DBL_PRECISION_COMPARE);
    
    TMatrixD matrixCARBE= _tbe->getBestEstParameterResponsesCovMatrix();    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.00170497387898180751,matrixCARBE(0,0),DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.000263160082938748989,matrixCARBE(9,14),DBL_PRECISION_COMPARE);
    
    TMatrixD matrixCR= _tbe->getResponsesCovMatrix();  
    CPPUNIT_ASSERT_DOUBLES_EQUAL(483.506989924794595,matrixCR(0,0),DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0780805074839171775,matrixCR(9,9),DBL_PRECISION_COMPARE);
    
    TMatrixD matrixCD= _tbe->getDifferencesCovMatrix();  
    CPPUNIT_ASSERT_DOUBLES_EQUAL(634.275606924794602,matrixCD(0,0),DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0796442548439171838,matrixCD(9,9),DBL_PRECISION_COMPARE);
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.15893146301478334,_tbe->getChi2(),DBL_PRECISION_COMPARE);
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.15893146301478334/10,_tbe->getConsistencyMeasure(),DBL_PRECISION_COMPARE);
    
    
}

void TBestEstimateTest::testTBestEstimateSetLog()
{
    _tbe = new TBestEstimate();
    _tbe->unsetLog();    
    CPPUNIT_ASSERT(!_tbe->getLog());
    
    _tbe->setLog();    
    CPPUNIT_ASSERT(_tbe->getLog());
    
    _tbe->changeLog();    
    CPPUNIT_ASSERT(!_tbe->getLog());
    
    _tbe->changeLog();    
    CPPUNIT_ASSERT(_tbe->getLog());
    
    
}

////////////////////////////////
///                          ///                            
///     throwexception       ///
///                          ///
////////////////////////////////

void TBestEstimateTest::throwExceptionTBestEstimateMethodNotYetImplemented()
{
    TMatrixD a, ca, car, cm, m, r, s;
    
    // Input data
    readMatrixFile( "vorBESTEST/a.abs",   a);   // vector of parameters \alpha (nominal values ?)
    readMatrixFile( "vorBESTEST/ca.abs",  ca);  // covariance matrix of parameters \alpha
    readMatrixFile( "vorBESTEST/r.abs",   r);   // vector of computed responses R
    readMatrixFile( "vorBESTEST/s.abs",   s);   // sensitivity matrix (\frac{\partial R}{\partial \alpha})
    readMatrixFile( "vorBESTEST/m.abs",   m);   // vector of measured responses M
    readMatrixFile( "vorBESTEST/cm.abs",  cm);  // covariance matrix of measured responses
    readMatrixFile( "vorBESTEST/car.abs", car); // "response-parameter" covariance matrix (using measured responses ?)
    
    Int_t nbParam = 3;
    Int_t nbResp = 2;

    _tbe = new TBestEstimate(nbParam, a, nbResp, m, r, ca, cm, s, car);
    _tbe->unsetLog();
    
    CPPUNIT_ASSERT_THROW(TBestEstimate(_tds, _tds,_tds, 
                                       a, ca) ,URANIE::Exceptions::UWarningExceptions);
    
    CPPUNIT_ASSERT_THROW(TBestEstimate(_tds, _tds,
        _tds, _tds, a) ,URANIE::Exceptions::UWarningExceptions);
    
    
    CPPUNIT_ASSERT_THROW(_tbe->setParameterData(_tds),URANIE::Exceptions::UWarningExceptions);
    
    CPPUNIT_ASSERT_THROW(_tbe->setMeasureData(_tds),URANIE::Exceptions::UWarningExceptions);
    
    CPPUNIT_ASSERT_THROW(_tbe->setResponseData(_tds),URANIE::Exceptions::UWarningExceptions);
    
    CPPUNIT_ASSERT_THROW(_tbe->setOATSamplingData(_tds),URANIE::Exceptions::UWarningExceptions);
    
    CPPUNIT_ASSERT_THROW(_tbe->setSensitivityMatrix(a),URANIE::Exceptions::UWarningExceptions);
    
    CPPUNIT_ASSERT_THROW(_tbe->setParameterResponseCovMatrix(a),URANIE::Exceptions::UWarningExceptions);
    
    CPPUNIT_ASSERT_THROW(_tbe->buildSensitivityMatrix(),URANIE::Exceptions::UWarningExceptions);
    
    CPPUNIT_ASSERT_THROW(_tbe->buildMatrices(),URANIE::Exceptions::UWarningExceptions);
    
    CPPUNIT_ASSERT_THROW(_tbe->printLog(),URANIE::Exceptions::UWarningExceptions);
}











