//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
#ifndef CPPUNITEST_TOPTIMIZER_H
#define CPPUNITEST_TOPTIMIZER_H

#include <cppunit/extensions/HelperMacros.h>
//#include <cppunit/portability/CppUnitDeque.h>
#include <cppunit/Test.h>
#include <cppunit/TestFailure.h>
#include <cppunit/TestResultCollector.h>

// Uranie
#include "TOptimizer.h"
#include "TDataServer.h"
#include "TLauncher.h"
#include "TAttribute.h"
#include "Optimizer.h"
#include "UExceptions.h"
#include "TOutputFileKey.h"
#include "TOutputFileRow.h"
#include "TOutputFileDataServer.h"
#include "TOutputFileColumn.h"

class TOptimizerTest: public CPPUNIT_NS::TestFixture
{
    CPPUNIT_TEST_SUITE (TOptimizerTest);

    CPPUNIT_TEST (testRosenbrockCodeOutputKey);
    CPPUNIT_TEST (testRosenbrockCodeOutputKeyNewInput);
    CPPUNIT_TEST (testRosenbrockCodeOutputKeyNewInputAndOutput);
    CPPUNIT_TEST (testRosenbrockCodeOutputKeyNewOutputSum);
    CPPUNIT_TEST (testRosenbrockCodeOutputKeyNewOutputMult);
    CPPUNIT_TEST (testRosenbrockCodeOutputKeyNewOutputSquared);
    CPPUNIT_TEST (testRosenbrockCodeOutputRow);
    CPPUNIT_TEST (testRosenbrockCodeInputsKeyAndRowRecreate);
    CPPUNIT_TEST (testRosenbrockCodeInputKeyOutputColumn);
    CPPUNIT_TEST (testRosenbrockCodeInputFileRecreateOutputDataser);
    CPPUNIT_TEST (testRosenbrockCodeInputFileRecreateOutputKey);
    CPPUNIT_TEST (testRosenbrockCodeInputFileRecreateOutputRow);
    CPPUNIT_TEST (testRosenbrockCodeInputFileRecreateOutputColumn);

    CPPUNIT_TEST (testRosenbrockFunctionSimplex);
    CPPUNIT_TEST (testRosenbrockFunctionPointerSimplex);
    CPPUNIT_TEST (testRosenbrockFunctionSimplexNewInput);
    CPPUNIT_TEST (testRosenbrockFunctionSimplexNewInputAndOutput);
    CPPUNIT_TEST (testRosenbrockFunctionSimplexNewOutputSum);
    CPPUNIT_TEST (testRosenbrockFunctionSimplexNewOutputMult);
    CPPUNIT_TEST (testRosenbrockFunctionSimplexNewOutputSquared);
    CPPUNIT_TEST (testRosenbrockFunctionMigrad);
    CPPUNIT_TEST (testRosenbrockFunctionCombined);
    CPPUNIT_TEST (testRosenbrockFunctionScn);
    CPPUNIT_TEST (testOptimizeRosenbrockFunctionWithOneConstantVariable);
    CPPUNIT_TEST (testOptimizersetToleranceAndgetTolerance);
    CPPUNIT_TEST (testOptimizersetMaxIterationsAndgetMaxIterations);
    CPPUNIT_TEST (testOptimizersetMaxFunctionCallsAndgetMaxFunctionCalls);
    CPPUNIT_TEST (testOptimizersetMethodAndgetMethod);
    CPPUNIT_TEST (testOptimizersetPrintLevelAndgetLevel);
    CPPUNIT_TEST (testTOptimizerSetLog);

    CPPUNIT_TEST (testOptimizeObjective);
    CPPUNIT_TEST (testOptimizeSeveralObjectives);
    CPPUNIT_TEST (testOptimizeSeveralObjectivesWithUnactive);
    CPPUNIT_TEST (testOptimizerAddObjective);
    CPPUNIT_TEST (testOptimizerAddObjectiveUncertaintySumSquareError);
    CPPUNIT_TEST (testOptimizerUnactiveObjective);
    CPPUNIT_TEST (testOptimizerActiveObjective);
    CPPUNIT_TEST (testOptimizerInvertObjective);

    CPPUNIT_TEST (testOptimizerselectCost);
    CPPUNIT_TEST (testOptimizerFixParameter);
    CPPUNIT_TEST (testOptimizerUnfixParameter);
    CPPUNIT_TEST (testOptimizerinvertParameters);
    CPPUNIT_TEST (testOptimizerConstructorInput);
    CPPUNIT_TEST (testOptimizerConstructorMoreOneOutput);
    CPPUNIT_TEST (testOptimizerClean);

    CPPUNIT_TEST (throwExceptionOptimizerConstructorTdsNull);
    CPPUNIT_TEST (throwExceptionOptimizerConstructorFalseInput);
    CPPUNIT_TEST (throwExceptionOptimizerConstructorUnknowFunction);
    CPPUNIT_TEST (throwExceptionOptimizerConstructorUnknowTCode);
//  CPPUNIT_TEST( throwExceptionOptimizersetMaxIterations );
//  CPPUNIT_TEST( throwExceptionOptimizersetMaxFunctionCalls );
//  CPPUNIT_TEST( throwExceptionOptimizersetPrintLevelNoInteger );
    CPPUNIT_TEST (throwExceptionOptimizersetPrintLevelInf1);
    CPPUNIT_TEST (throwExceptionOptimizersetPrintLevelSup5);
    CPPUNIT_TEST (throwExceptionOptimizerSelectCost);
    CPPUNIT_TEST (throwExceptionOptimizerFixParameter);
    CPPUNIT_TEST (throwExceptionOptimizerUnfixParameter);

    CPPUNIT_TEST (throwExceptionOptimizeraddObjectiveTdsNull);
    CPPUNIT_TEST (throwExceptionOptimizeraddObjectiveYStarFalse);
    CPPUNIT_TEST (throwExceptionOptimizeraddObjectiveYHatFalse);
    CPPUNIT_TEST (throwExceptionOptimizeraddObjectiveOutFileNull);
    
    CPPUNIT_TEST (throwExceptionOptimizerOptimize);
    CPPUNIT_TEST (throwExceptionOptimizerUnactiveObjective);
    CPPUNIT_TEST (throwExceptionOptimizerActiveObjective);

    CPPUNIT_TEST_SUITE_END();

public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    TOptimizerTest();

    //! Default destructor
    virtual ~TOptimizerTest();

    void setUp();
    void tearDown();
    //! the clean tempory directories
    void cleanWorkingDirectories(const char *option = "without");
    static void rosenbrockP(double *x, Double_t *y);
    //@}

    //---------------------------------------------
    /** @name Unit Tests
     The list of unit test with TCode
     **/
    //@{ 
    //! Test for optimisation to the rosenbrock code and output file with format key
    void testRosenbrockCodeOutputKey();
    void testRosenbrockCodeOutputKeyNewInput();
    void testRosenbrockCodeOutputKeyNewInputAndOutput();
    void testRosenbrockCodeOutputKeyNewOutputSum();
    void testRosenbrockCodeOutputKeyNewOutputMult();
    void testRosenbrockCodeOutputKeyNewOutputSquared();
    //! Test for optimisation to the rosenbrock code and output file with format Row
    void testRosenbrockCodeOutputRow();
    //! Test for optimisation to the rosenbrock code with two inputs files
    void testRosenbrockCodeInputsKeyAndRowRecreate();
    void testRosenbrockCodeInputKeyOutputColumn();
    void testRosenbrockCodeInputFileRecreateOutputColumn();
    void testRosenbrockCodeInputFileRecreateOutputKey();
    void testRosenbrockCodeInputFileRecreateOutputRow();
    void throwExceptionOptimizeraddObjectiveTdsNull();
    void throwExceptionOptimizeraddObjectiveYStarFalse();
    void throwExceptionOptimizeraddObjectiveYHatFalse();
    void throwExceptionOptimizeraddObjectiveOutFileNull();
    void testOptimizerAddObjective();
    void testOptimizerAddObjectiveUncertaintySumSquareError();
    void testOptimizerUnactiveObjective();
    void testOptimizerActiveObjective();
    void testOptimizerInvertObjective();
    //! Test for optimisation to the rosenbrock code with TInputFileRecreate and TOuputDataServer
    void testRosenbrockCodeInputFileRecreateOutputDataser();
    void testOptimizerClean();
    void throwExceptionOptimizerOptimize();
    void throwExceptionOptimizerUnactiveObjective();
    void throwExceptionOptimizerActiveObjective();
    //@}

    //---------------------------------------------
    /** @name Unit Tests
     The list of unit test with Objective
     **/
    //@{ 
    void testOptimizeObjective();
    void testOptimizeSeveralObjectives();
    void testOptimizeSeveralObjectivesWithUnactive();
    //@}

    //---------------------------------------------
    /** @name Unit Tests with Analytical Functions
     The list of unit test with Analytical Functions
     **/
    //@{ 
    //! Test the Rosenbrock function with simplex
    void testRosenbrockFunctionSimplex();
    void testRosenbrockFunctionPointerSimplex();
    void testRosenbrockFunctionSimplexNewInput();
    void testRosenbrockFunctionSimplexNewInputAndOutput();
    void testRosenbrockFunctionSimplexNewOutputSum();
    void testRosenbrockFunctionSimplexNewOutputMult();
    void testRosenbrockFunctionSimplexNewOutputSquared();
    //! Test the Rosenbrock function with Migrad
    void testRosenbrockFunctionMigrad();
    void testRosenbrockFunctionCombined();
    void testRosenbrockFunctionScn();
    void testOptimizeRosenbrockFunctionWithOneConstantVariable();

    void testOptimizersetToleranceAndgetTolerance();
    void testOptimizersetMaxIterationsAndgetMaxIterations();
    void testOptimizersetMaxFunctionCallsAndgetMaxFunctionCalls();
    void testOptimizersetMethodAndgetMethod();
    void testOptimizersetPrintLevelAndgetLevel();
    void testTOptimizerSetLog();
    void testOptimizerselectCost();
    void testOptimizerFixParameter();
    void testOptimizerUnfixParameter();
    void testOptimizerinvertParameters();
    void testOptimizerConstructorMoreOneOutput();
    void testOptimizerConstructorInput();
    void throwExceptionOptimizerConstructorTdsNull();
    void throwExceptionOptimizerConstructorFalseInput();
    void throwExceptionOptimizerConstructorUnknowFunction();
    void throwExceptionOptimizerConstructorUnknowTCode();
    void throwExceptionOptimizersetMaxIterations();
    void throwExceptionOptimizersetMaxFunctionCalls();
    void throwExceptionOptimizersetPrintLevelNoInteger();
    void throwExceptionOptimizersetPrintLevelInf1();
    void throwExceptionOptimizersetPrintLevelSup5();
    void throwExceptionOptimizerSelectCost();
    void throwExceptionOptimizerFixParameter();
    void throwExceptionOptimizerUnfixParameter();
    //@}

private:
    TString _workingTestPath;
    TString _sReferenceDirectory;
    URANIE::DataServer::TDataServer *_tds1;
    //! Output file Key
    URANIE::Launcher::TOutputFileKey *_foutKey;
    //! Output file Row
    URANIE::Launcher::TOutputFileRow *_foutRow;
    //! Output file DataServer
    URANIE::Launcher::TOutputFileDataServer *_foutDataServer;
    //! Output file Column
    URANIE::Launcher::TOutputFileColumn *_foutCol;
    //! TCode object
    URANIE::Launcher::TCode *_tcode;
    //! TOptimizer object
    URANIE::Optimizer::TOptimizer * _topt;

    URANIE::DataServer::TAttribute *_pX;
    URANIE::DataServer::TAttribute *_pY;

    void * _p2fRosenbrock;    ///< Pointer to the Rosenbrock Analytical function
    void * _p2fRosenbrock2; ///< Pointer to the Rosenbrock2 Analytical function in R2
};

#endif  // CPPUNITEST_TOPTIMIZER_H
