//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestUserFunctions.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   2012-01-30 09:39:00

 \brief Implement the rosenbrock function

 The analytical function is givent by the formula
 
 \$[ f(x_1, x_2) = a * ( x_2 - x_1^2 )^2 + b * ( 1.0 - x_1 )^2 \$]

 with \f$ a=10.0, b=1.0 \f$.

 */

// ROOT
#include "TROOT.h"
#include "TMath.h"

void rosenbrock(Double_t *x, Double_t *y)
{
    Double_t dx1 = x[0], dx2 = x[1];
    Double_t da = 10.0, db = 1.0;

    y[0] = da * (dx2 - dx1 * dx1) * (dx2 - dx1 * dx1)
            + db * (1.0 - dx1) * (1.0 - dx1);

}

void rosenbrock2(Double_t *x, Double_t *y)
{
    Double_t dx1 = x[0], dx2 = x[1];
    Double_t da = 10.0, db = 1.0;

    y[0] = da * (dx2 - dx1 * dx1) * (dx2 - dx1 * dx1)
            + db * (1.0 - dx1) * (1.0 - dx1);
    y[1] = da * (dx2 - dx1 * dx1) * (dx2 - dx1 * dx1);

}


void rosenbrockConstraint(Double_t *x, Double_t *y)
{
    Double_t dx1 = x[0], dx2 = x[1];
    Double_t da = 10.0, db = 1.0;

    y[1] = da * (dx2 - dx1 * dx1) * (dx2 - dx1 * dx1)
            + db * (1.0 - dx1) * (1.0 - dx1);
    y[0] = 0.0;
    if (fabs(x[0] - x[1]) <= 0.5)
        y[0] = 1;

}


void funAllCost2(Double_t *param, Double_t *res)
{
    Double_t x = param[0];
    Double_t y = param[1];
    res[0] = x * x + y * y;
    res[1] = 2*(x+y) - 5;
}

void mydecodeImage(int *rien, int *size, double *intern, double *user)
{
    int i;
    (void)(rien); // Implemented to do nothing
    user[0] = 256.0;
    for (i = 1; i < *size; i++)
        user[i] = 256.0 * intern[i - 1];
}

void contHard2(double * paramt, Int_t *res)
{
    *res = 0;
    if (fabs(paramt[0] - paramt[1]) <= 0.5)
        *res = 1;
}

void mydecode2(int *rien, int *size, double *intern, double *user)
{
    int i;
    (void)(rien); // Implemented to do nothing
    for (i = 0; i < *size; i++)
        user[i] = 3.0* intern[i] -1.5;
}


// void funCostAndConstraint(Double_t *param, Double_t *res)
// {
//     Double_t x = param[0];
//     Double_t y = param[1];
//     res[0] = x * x + y * y;
//     res[1] = 2*(x+y) - 5;
//     res[2] = 0.0;
//     if( param[0] <= param[1]+1.0 )
//        res[2] = 1.0;
// }


void funCostAndConstraint2(Double_t *param, Double_t *res)
{
    Double_t x = param[0];
    Double_t y = param[1];
    res[0] = (x - 1.0) * (x - 1.0) + (y - 1.0) * (y - 1.0) + 1.0;
    res[1] = 1. / (x * x + y * y + 1.0);
    res[2] = 0.0;
    if( param[0] <= param[1]+1.0 )
       res[2] = 1.0;
}

void funAllCost(Double_t *param, Double_t *res)
{
    Double_t x = param[0];
    Double_t y = param[1];
    res[0] = (x - 1.0) * (x - 1.0) + (y - 1.0) * (y - 1.0) + 1.0;
    res[1] = 1. / (x * x + y * y + 1.0);
}



void contHard1(double * paramt, Int_t *res)
{
    *res = 0;
    if (paramt[0] <= paramt[1] + 1.0)
        *res = 1;
}



void mydecode(int *rien, int *size, double *intern, double *user)
{
    int i;
    (void)(rien); // Implemented to do nothing
    user[0] = 256.0;
    for (i = 1; i < *size; i++)
        user[i] = 256.0 * intern[i - 1];
}



double contraste(int size, double *sol, const char **connexion)
{
    int i, j;
    double min, diff;

    min = 256.0;
    for (i = 0; i < size; i++)
    {
        for (j = i + 1; j < size; j++)
        {
            if (connexion[i][j])
            {
                diff = fabs(sol[i] - sol[j]);
                if (diff < min)
                    min = diff;
            }
        }
    }
    return min;
}
void myfitness(double *in, double *out)
{
    static const char tgr[8 * 8] =
    {
    /*  0  1  2  3  4  5  6  7 */
    0, 1, 1, 1, 0, 0, 1, 1, // 0
    1, 0, 1, 1, 1, 0, 0, 0, // 1
    1, 1, 0, 0, 1, 0, 0, 0, // 2
    1, 1, 0, 0, 0, 1, 1, 0, // 3
    0, 1, 1, 0, 0, 1, 0, 1, // 4
    0, 0, 0, 1, 1, 0, 1, 0, // 5
    1, 0, 0, 1, 0, 1, 0, 1, // 6
    1, 0, 0, 0, 1, 0, 1, 0  // 7
    };
    static const char *tangram[8] =
    { &tgr[0], &tgr[8], &tgr[2 * 8], &tgr[3 * 8], &tgr[4 * 8], &tgr[5 * 8],
            &tgr[6 * 8], &tgr[7 * 8] };

    out[0] = -contraste(8, in, tangram);
}




// End Of File
