//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.2 $
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   vzr_naitre.c
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date    2008-03-19.09.17.05

 \brief naissance des indivividus
 */

#include <stdio.h>

#include "vzr_util.h"
#include "vzr_mt19937ar.h"
#include "vzr_croiser.h"

int vizir1_reevaluation(struct vizir*self, struct vzrwork *wself)
{
    int i, ret;

    pthread_mutex_lock(&(self->mutpop));
    if (self->reeval != self->N)
    {
        for (i = 0; i < self->nbgene; i++)
        {
            wself->descendant[0][i] = self->population[self->reeval][0][i];
            wself->descendant[1][i] = self->population[self->reeval][1][i];
        }
        self->reeval++;
        ret = 1;
    }
    else
        ret = 0;
    pthread_mutex_unlock(&(self->mutpop));
    return ret;
}

int vizir1_immigration(struct vizir*self, struct vzrwork *wself)
{
    int i, ret;

    pthread_mutex_lock(&(self->mutim));
    if (self->taillepile > 0)
    {
        for (i = 0; i < self->nbgene; i++)
        {
            wself->descendant[0][i] = self->pile_migrant[self->baspile][0][i];
            wself->descendant[1][i] = self->pile_migrant[self->baspile][1][i];
        }
        self->baspile = suiv(self->baspile);
        self->taillepile--;
        self->imigr++;
        ret = 1;
    }
    else
        ret = 0;
    pthread_mutex_unlock(&(self->mutim));
    return ret;
}

void accoupler(struct vizir*self, struct vzrwork *wself)
{
    int i, ichoix, parent1, parent2;

    pthread_mutex_lock(&(self->mutpop));
    /*  croisement */
    /* choix du premier parent
     * soit un element de l'elite (s'il y en a assez)
     *   plus bias est grand plus on choisit dans les elites 
     * soit un quelconque des survivants */
    ichoix = self->nbsurv;
    if (vizir1_genrand_real1() < self->bias && self->nbelite > self->consang)
        ichoix = self->nbelite;
    parent1 = vizir1_genrand_int(ichoix);
    do
    {
        /* le second parents est tjs pris parmi les survivants */
        parent2 = vizir1_genrand_int(self->nbsurv);
    } while (parent1 == parent2); /* le clonage est interdit !! */
#ifdef RANKED
    if (parent1 > parent2)
    {
        ichoix = parent1; parent1 = parent2; parent2 = ichoix;
    }
#endif
    /* recopie */
    for (i = 0; i < self->nbgene; i++)
    {
        wself->pere[0][i] = self->population[parent1][0][i];
        wself->pere[1][i] = self->population[parent1][1][i];
        wself->mere[0][i] = self->population[parent2][0][i];
        wself->mere[1][i] = self->population[parent2][1][i];
    }
    pthread_mutex_unlock(&(self->mutpop));

}

void vizir1_procreer(struct vizir*self, struct vzrwork* wself)
{
    int i;
    double seuil;

    /* le descendant sera issu d'un croisement avec ou sans mutation*/
    seuil = vizir1_genrand_real1();
    if ((!wself->oper->mutation) && seuil < self->pmut)
    {
        /* mutation sans operateur specifique
         * on ecrase un allele du pere par un allele aleatoire
         * */
        PRO_CALL(wself->oper, sample, wself->descendant);
        for (i = 0; i < self->nbgene; i++)
            wself->pere[1][i] = wself->descendant[1][i];
    }
    /* tirage des alleles de l'enfant */
    PRO_CALL(wself->oper, allele, wself->pere, wself->descendant[0]);
    PRO_CALL(wself->oper, allele, wself->mere, wself->descendant[1]);
    /* mutation (avec operateur specifique) */
    if (wself->oper->mutation && seuil < self->pmut)
    {
        PRO_CALL(wself->oper, mutation, wself->descendant);
    }
    /* homozygotie et dominance sont traite lors du calcul du phenotype   */
    if (vizir1_genrand_real1() < self->phomo)
    {
        /* on construit un homozygote */
        PRO_CALL(wself->oper, homozygote, wself->descendant);
    }
    else
    {
        /* on construit le phenotype */
        PRO_CALL(wself->oper, heterozygote, wself->descendant);
    }
}

