//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.4 $
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   vzr_query.c
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date    2003-12-08.14.20.17

 \brief initialisation et requete resultat pour vizir
 */

#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#else
#include <io.h>
#include <process.h>
#endif
#include <math.h>
#include <time.h>

#include <sched.h>

#include "vizir.h"

#include "vzr_util.h"
#include "vzr_alloc.h"

#include "vzr_mt19937ar.h"
#include "vzr_kernel.h"
#include "vzr_croiser.h"
#include "vzr_survivre.h"

/**** declaration externe ****/
typedef struct vzrclasse *Selection(void);
extern Selection vizir1_multi, vizir1_mono, multi_contraint, mono_contraint, multi_contraint2;

/*** variables locales ***/
static Selection *loi_nature[] =
{ &vizir1_multi, &vizir1_mono, &mono_contraint, &multi_contraint, &multi_contraint2 };

/**** corps des fonctions ****/
/** initialisation **/
/* creation struct */
struct vizir *vizir1_vzr_init(int mode_vzr, int nb_gene, int nb_cout, int nb_cont,
        double precis)
{
    struct vizir *ret;

    /* allocation structure */
    ret = (struct vizir*) malloc(sizeof(struct vizir));

    /* definition du probleme */
    ret->id = " ";
    /* initialisation utilisateur */
    ret->darwin = (*loi_nature[mode_vzr])();
    ret->nbgene = nb_gene;
    ret->nbfun = nb_cout;
    ret->nbsoft = nb_cont;
    ret->nbhide = 0;
    ret->prec = precis;

    /* initialisation defaut */
    ret->alpha = 0.2;
    ret->phomo = 0.5;
    ret->pmut = 0.01;
    ret->bias = 0.5;
    ret->prop = 1.01;

    pthread_mutex_init(&(ret->mutex), (pthread_mutexattr_t *) NULL);
    pthread_mutex_init(&(ret->mutim), (pthread_mutexattr_t *) NULL);
    pthread_mutex_init(&(ret->mutpop), (pthread_mutexattr_t *) NULL);
#ifdef MPI_UNSAFE
    pthread_mutex_init(&(ret->mutmpi), (pthread_mutexattr_t *) NULL);
#endif

    /* initialisation zero */
    ret->maxevalu = ret->itaille = ret->N = 0; //reinit via set_taille
    ret->ipop = ret->nbevalu = ret->nbsurv = ret->nbelite = ret->niter =
            ret->nadmis = ret->genedisc = ret->slave = 0;
    ret->imigr = ret->emigr = 0;
    ret->baspile = ret->sommetpile = ret->taillepile = 0;
    ret->fitseuil = ret->pmut1g = 0.0;
    ret->bgstate = START;
    ret->ffun = (struct vzrfitnessbuild *) NULL;
    ret->foper = (struct vzrcrossoverbuild *) NULL;
    ret->generation = (struct vzrevent *) NULL;

    return ret;
}

/* modification des defauts */
void vizir1_vzr_setLabel(struct vizir *vizir, char *name)
{
    vizir->id = name;
}

void vizir1_vzr_setSurvivalRate(struct vizir *vizir, double val)
{ /* taux de survie */
    if (val > 0. && val < 1.)
        vizir->alpha = val;
}

void vizir1_vzr_setHomozygoteRate(struct vizir *vizir, double val)
{ /* probabilite d'homozigotie */
    if (val > 0. && val < 1.)
        vizir->phomo = val;
}

void vizir1_vzr_setMutationRate(struct vizir *vizir, double val)
{ /* probabilite de mutation */
    if (val > 0. && val < 1.)
        vizir->pmut = val;
}
/* EDF */
void vizir1_vzr_setOneGeneMutationRate(struct vizir *vizir, double val)
{ /* probabilite de mutation sur un gene */
    if (val > 0. && val < 1.)
        vizir->pmut1g = val;
}

void vizir1_vzr_setCombinatorialSize(struct vizir *vizir, int val)
{ /* nombre de gene a traiter comme des variables discretes */
    if (val <= vizir->nbgene)
        vizir->genedisc = val;
    else
        printf(
                "ERREUR le nombre de genes discrets depassent le nombre total  de genes \n");
}

void vizir1_vzr_setParentsBias(struct vizir *vizir, double val)
{ /* biaisage des naissances */
    if (val > 0. && val < 1.)
        vizir->bias = val;
}

void vizir1_vzr_setOutlierFactor(struct vizir *vizir, double val)
{ /* taux d'outliers */
    if (val > 1.)
        vizir->prop = val;
}

void vizir1_vzr_setCrossover(struct vizir *vizir, struct vzrcrossoverbuild *val)
{
    vizir->foper = val;
}

/* fonction utilisateur d'�valuation */
void vizir1_vzr_setFitness(struct vizir *vizir, struct vzrfitnessbuild *val)
{
    vizir->ffun = val;
}

/* dimmentionnement du probl�me */
int vizir1_vzr_build(struct vizir *vizir, int max_eval, int taille_pop)
{
    int i, j;

    vizir->nbcrit = vizir->nbfun + vizir->nbsoft + vizir->nbhide
            + vizir->darwin->rank_field_size;
    vizir->rangdomi = vizir->nbcrit - 1;
    vizir->maxevalu = max_eval;
    vizir->itaille = taille_pop;
    vizir->reeval =  vizir->N = (int) vizir->prop * taille_pop;
    vizir->consang = (int) floor(vizir->alpha * vizir->N);

    /* allocation tableau de travail */
    // NEWvecteur(vizir->rangdomi, vizir->N);
    NEWmatrice(vizir->pop_val, vizir->N, vizir->nbcrit);
    NEWmatrice(vizir->descendant_val, vizir->N, vizir->nbcrit);
    NEWmatriceG3D(vizir->population, vizir->N, 2, vizir->nbgene);

    /* operateur croisement */
    if (vizir->foper == NULL)
    {
#ifdef COMBINATOIRE
        printf("unseted Crossing operator\n");
        return 1;
#else
        vizir->foper = vizir1_vzrcrossstd_classe(vizir->nbgene, vizir->genedisc);
#endif
    }
    /* distribution    *************/
    NEWmatriceG3D(vizir->pile_migrant, DIM_PILE, 2, vizir->nbgene);
    vizir1_init_genrand((unsigned long) (time(NULL) * getpid()));
    return 0;
}

/* initialisation specifique (pour interface C) */
struct vizir *vizir1_vzr_multi(int nb_gene, int nb_cout)
{
    return vizir1_vzr_init(MULTI, nb_gene, nb_cout, 0, 0.1);
}

struct vizir *vizir1_vzr_mono(int nb_gene, double precis)
{
    return vizir1_vzr_init(MONO, nb_gene, 1, 0, precis);
}

struct vizir *vizir1_vzr_monoConstraint(int nb_gene, double precis, int nb_cont)
{
    return vizir1_vzr_init(MONOC, nb_gene, 1, nb_cont, precis);

}

struct vizir *vizir1_vzr_multiConstraint(int nb_gene, int nb_cout, int nb_cont)
{
    return vizir1_vzr_init(MULTIC, nb_gene, nb_cout, nb_cont, 0.1);
}

/** execution **/
/* threaded execution */
void *vzr_run_bg(void *vvizir)
{
    struct vizir *vizir = (struct vizir*) vvizir;

    vizir1_vzr_run(vizir);
    //pthread_exit(0);
    return 0;
}

void vizir1_vzr_start(struct vizir *vizir)
{
    pthread_t bg;

    //pthread_mutex_lock(&(vizir->mutex));
    //if (vizir->bgstate == END) vizir->bgstate |= START;
    //pthread_mutex_unlock(&(vizir->mutex));
//fga 2015/01/12     pthread_create(&bg, NULL, vzr_run_bg, (void *) vizir);
}

int vizir1_vzr_running(struct vizir *vizir)
{
    int ret;
    pthread_mutex_lock(&(vizir->mutex));
    ret = (vizir->bgstate != END) ? 1 : 0;
    pthread_mutex_unlock(&(vizir->mutex));
    return ret;
}

void vizir1_vzr_stop(struct vizir *vizir)
{
    pthread_mutex_lock(&(vizir->mutex));
    if (vizir->bgstate != END)
        vizir->bgstate |= STOP;
    pthread_mutex_unlock(&(vizir->mutex));
    printf("%s  stopped\n", vizir->id);
}

void vizir1_vzr_wait(struct vizir *vizir)
{
    int tag;

    tag = vizir1_vzr_running(vizir);
    while (tag)
    {
        // pthread_yield();
		sched_yield();
        tag = vizir1_vzr_running(vizir);
    }
}

int vzr_immigrate(struct vizir *viz, vzrgene_t **x)
{
    /* cf vzr_putIndividu de vzr.i */
    int i, ret;

    pthread_mutex_lock(&(viz->mutim));
    if (viz->taillepile < DIM_PILE)
    {
        for (i = 0; i < viz->nbgene; i++)
        {
            viz->pile_migrant[viz->sommetpile][0][i] = x[0][i];
            viz->pile_migrant[viz->sommetpile][1][i] = x[1][i];
        }
        viz->sommetpile = suiv(viz->sommetpile);
        viz->taillepile++;
        ret = 0;
    }
    else
    {
        printf("%s  File saturee  \n", viz->id);
        ret = 1;
    }
    pthread_mutex_unlock(&(viz->mutim));
    return ret;
}

void vzr_emptyStack(struct vizir *viz)
{
    /* netoyage pour recommencer (apres synchronisation) */
    /* vide la pile des migrants */
    viz->baspile = viz->sommetpile = viz->taillepile = 0;
}

void vizir1_vzr_reset(struct vizir *viz)
{
    /* netoyage pour recommencer a zero */
    /* nombre de generation, evaluation, admis */
    viz->niter = viz->nbevalu = viz->nadmis = 0;
    viz->emigr = viz->imigr = 0;
    viz->bgstate = START;

    /* pile des migrants */
    viz->baspile = viz->sommetpile = viz->taillepile = 0;

}

void vizir1_vzr_reevaluate(struct vizir *viz)
{
    if (viz->slave == 0)
        viz->reeval = viz->nbsurv = viz->ipop = 0;
    else
        printf("reevaluate needs solver to be stopped\n");
}

/** resultat **/
/* population */
vzrgene_t **vizir1_vzr_genotype(struct vizir *vizir, int i)
{
    return vizir->population[i];
}

vzrgene_t *vizir1_vzr_phenotype(struct vizir *vizir, int i)
{
    return vizir->population[i][0];
}

/* critere */
double *vizir1_vzr_objective(struct vizir *vizir, int i)
{
    return vizir->pop_val[i];
}

double vizir1_vzr_rank(struct vizir *vizir, int i)
{
    return vizir->pop_val[i][vizir->rangdomi];
}

double vzr_constraintRank(struct vizir *vizir, int i)
{
    int rg;

    rg = vizir->darwin->rank_field_size;
    if (rg > 1)
        return vizir->pop_val[i][vizir->nbcrit - rg];
    else
        return -1000.0;
}

double vzr_objectiveRank(struct vizir *vizir, int i)
{
    if (vizir->darwin->rank_field_size == 3)
        return vizir->pop_val[i][vizir->nbcrit - 2];
    else
        return -1000.0;
}

int vzr_emigrate(struct vizir *viz, vzrgene_t **migrant)
{
    /* cf vzr_getIndividu de vzr.i */
    int i, ret;

    pthread_mutex_lock(&(viz->mutex));
    if (viz->ipop && viz->niter)
    {
        for (i = 0; i < (viz->nbgene); i++)
        {
            migrant[0][i] = viz->population[viz->N - viz->ipop][0][i];
            migrant[1][i] = viz->population[viz->N - viz->ipop][1][i];
        }
        viz->ipop--;
        viz->emigr++;
        ret = 1;
    }
    else
        ret = 0;
    pthread_mutex_unlock(&(viz->mutex));
    return ret;
}

/* runtime infos */
int vizir1_vzr_birthCount(struct vizir *viz)
{
    return viz->nbevalu;
}

int vizir1_vzr_generationCount(struct vizir *viz)
{
    return viz->niter;
}

int vizir1_vzr_survivorSize(struct vizir *viz)
{
    return (int) viz->alpha * viz->N;
}

int vizir1_vzr_populationSize(struct vizir *viz)
{
    return viz->itaille;
}

int vizir1_vzr_emigrationCount(struct vizir *viz)
{
    return viz->emigr;
}

int vizir1_vzr_immigrationCount(struct vizir *viz)
{
    return viz->imigr;
}

/* destructeur */
void vizir1_vzr_dealloc(struct vizir *viz)
{
    if (viz->generation)
        PRO_DEALLOC(viz->generation);
    if (viz->darwin->rank_field_size != 1)
      PRO_DEALLOC(viz->darwin);
    PRO_DEALLOC(viz->ffun);
    PRO_DEALLOC(viz->foper);
    DELmatrice3D(viz->population);
    DELmatrice3D(viz->pile_migrant);
    DELmatrice(viz->pop_val);
    DELmatrice(viz->descendant_val);
    pthread_mutex_destroy(&viz->mutex);
    pthread_mutex_destroy(&viz->mutim);
    pthread_mutex_destroy(&viz->mutpop);
#ifdef MPI_UNSAFE
    pthread_mutex_destroy(&viz->mutmpi);
#endif
    free(viz);
}

