//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.3 $
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TSimpleTransform.cxx
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date    2012-04-02.11.36.11

 \brief   FORM/SORM point de conception
 */

#include "TSimpleTransform.h"

#include <math.h>

#include "TNormalDistribution.h"
#include "TStochAttribut.h"

//#define TRACE

#define EPSI 1.0e-6

ClassImp (URANIE::Reliability::TSimpleTransform);

namespace URANIE
{
namespace Reliability
{

URANIE::DataServer::TNormalDistribution nd;

TSimpleTransform::TSimpleTransform()
{
}
TSimpleTransform::~TSimpleTransform()
{
    TIter next(&_input);
    TObject *att;

    while ((att = next()))
        delete att;
    att = NULL;
    delete _output[0];
    _output[0] = NULL;
    delete _output[1];
    _output[1] = NULL;

}

void TSimpleTransform::addParameter(URANIE::DataServer::TAttribute *un)
{
    if (_freezed)
        throw TEval::FreezedEval();

    _param.Add((TObject *) un);
}

void TSimpleTransform::setParameters(Int_t size,
        URANIE::DataServer::TAttribute *un, ...)
{
    va_list ap;
    int i;

    if (_freezed)
        throw TEval::FreezedEval();

    va_start(ap, un);
    _param.Add((TObject *) un);
    for (i = 1; i < size; i++)
    {
        _param.Add(va_arg(ap, TObject *));
    }
    va_end(ap);
}

URANIE::DataServer::TAttribute * TSimpleTransform::getBeta()
{
    freeze();
    return (URANIE::DataServer::TAttribute *) _output[0];
}

void TSimpleTransform::freeze()
{
    TIter next(&_param);
    URANIE::DataServer::TAttribute *att;
    URANIE::DataServer::TNormalDistribution *norm;

    if (_freezed)
        return;
    TBaseEval::freeze();

    _output.Add(new URANIE::DataServer::TAttribute("betaHL"));
    _output.Add(new URANIE::DataServer::TAttribute("form"));
    while ((att = (URANIE::DataServer::TAttribute *) next()))
    {
        _output.Add(att);
        norm = new URANIE::DataServer::TNormalDistribution(
                Form("u_%s", att->GetName()));
        //norm->setBounds(-3.1, 3.1);
        norm->setBounds(norm->getInvCDF(EPSI), norm->getInvCDF(1. - EPSI));
        _input.Add(norm);

    }
}

int TSimpleTransform::eval(double *in, double *out, int ncalc)
{
    UNUSED(ncalc);
    int i;
    double ret, u, v;
    TIter get(&_output);
    URANIE::DataServer::TStochasticAttribute *att;

    att = (URANIE::DataServer::TStochasticAttribute *) get();
    att = (URANIE::DataServer::TStochasticAttribute *) get();
    ret = 0.;
    i = 0;
    v = 0;
    while ((att = (URANIE::DataServer::TStochasticAttribute *) get()))
    {
        u = in[i];
#ifdef TRACE
        v = nd.getCDF(u);
        std::cout << u << " " << v << std::endl;
#endif
        out[i + 2] = att->getInvCDF(nd.getCDF(u));
        ret += u * u;
        i++;
    }
    out[0] = sqrt(ret);
    out[1] = nd.getCDF(-out[0]);
    return 1;
}

Int_t TSimpleTransform::eval(vector<URANIE::DataServer::UEntry*> *in, 
    vector<URANIE::DataServer::UEntry*> *out, int ncalc)
{
    double *inp = new double[in->size()];
    double *outp = new double[out->size()];
    for(unsigned int ival=0; ival<in->size(); ival++) inp[ival] = *in->at(ival)->val.dou;

    int admis=eval(inp, outp, ncalc);
    
    for(unsigned int ival=0; ival<out->size(); ival++) *out->at(ival)->val.dou = outp[ival];
    delete [] inp; delete [] outp;
    return ((admis!=0) ? getOutDimension(out) : admis);
}


}
} // namespace
