//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.4 $
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TSorm.cxx
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date    2012-07-09.13.55.00

 \brief   SORM evaluation
 */

#include "TSorm.h"

#include <math.h>

#include "TLeaf.h"

#include "TRun.h"

#include "gramschmitd.h"
#include "bracket.h"

ClassImp (URANIE::Reliability::TSorm);

//#define TRACE

#define INIT 1
#define BRACKET 2
#define SEARCH 3
#define FINI 4
#define DELTA 0.3

namespace URANIE
{
namespace Reliability
{

/*** TDemiAxeSorm ***/
TDemiAxeSorm::TDemiAxeSorm(int size, double *ref)
{
    int i;

    _size = size;
    _ref = new double[size];
    for (i = 0; i < size; i++)
        _ref[i] = ref[i];
    _xl = _ref[0];
    _state = INIT;
}

TDemiAxeSorm::~TDemiAxeSorm()
{
    delete[] _ref;
    _ref = NULL;
    ridder_clean(_solv);
}

int TDemiAxeSorm::next(double eval, double *Next)
{
    int i, ret;

    ret = 1;
    switch (_state)
    {
    case INIT:
        _fl = eval;
        _xt = (_fl < 0) ? _xl * 0.95 : _xl * 1.05;
        _state = BRACKET;
        break;
    case BRACKET:
        _ft = eval;
        if (!bracket_next(&_xl, &_fl, &_xt, &_ft))
        {
            _solv = create_ridder(_xl, _fl, _xt, _ft);
            ridder_first(_solv, &_xt);
            _state = SEARCH;
        }
        break;
    case SEARCH:
        if (!ridder_next(_solv, eval, &_xt))
        {
            _xt = ridder_result(_solv);
            _state = FINI;
        }
        break;
    case FINI:
        ret = 0;
    }
    Next[0] = _xt;
    for (i = 1; i < _size; i++)
        Next[i] = _ref[i];
    return ret;
}

double TDemiAxeSorm::courbure()
{
    return ridder_result(_solv);
}

/*** TSorm ***/
/* constructeur destructeur */
TSorm::TSorm(URANIE::DataServer::TDataServer *tds, URANIE::Relauncher::TRun *run) :
    URANIE::Relauncher::TInformMaster(tds, run)
{
    create();
}

TSorm::TSorm(URANIE::DataServer::TDataServer *tds, TFormEval *fun) :
    URANIE::Relauncher::TInformMaster(tds, fun)
{
    create();
}

TSorm::~TSorm(void)
{
    delete _lapse;
    _lapse = NULL;
}

void TSorm::create()
{
    _lapse = NULL;
    _size = _fun->getInputList()->GetEntries();
    ((TFormEval *) _fun)->addLapse(this);
}

/* definition */
void TSorm::addLapse(URANIE::DataServer::TAttribute *att,
        URANIE::Reoptimizer::TOptimFit *modif)
{
    if (!_lapse)
        _lapse = new URANIE::Reoptimizer::TFitGetter(att, _fun->getOutputList(),
                modif);
    else
        std::cout << "Sorm is single objective\n";
}

/* helpers */
void TSorm::build()
{
    URANIE::DataServer::TAttribute *tmp1, *tmp2;
    int nis = _tds->getNAttributes();
    
    // master output
    tmp1 = new URANIE::DataServer::TAttribute("factor");
    _sortie.Add( new URANIE::Relauncher::TPostVar(tmp1, _NoutCod, nis));
    tmp2 = new URANIE::DataServer::TAttribute("sorm");
    _sortie.Add( new URANIE::Relauncher::TPostVar(tmp2, _NoutCod+1, nis));
    URANIE::Relauncher::TMaster::fillDS();
    tmp1->delShare(); tmp2->delShare();
    _tds->createTuple();
    //if (!_lapse) std::cout << "TSorm need a threathhold\n" ;
}

void TSorm::fill(double id, vector<URANIE::DataServer::UEntry*> *in, 
    vector<URANIE::DataServer::UEntry*> *out, 
    vector<URANIE::DataServer::UEntry*> *all, double correct)
{
    int fin = _NinDat + _Naux + _NoutDat + 1;
    
    allFromCode(
        new vector<URANIE::DataServer::UEntry*> (in->begin(), in->end()),
        new vector<URANIE::DataServer::UEntry*> (out->begin(), out->end()),
        new vector<URANIE::DataServer::UEntry*> (all->begin()+1, all->end())
    );
    *all->at(0)->val.dou = id;
    *all->at(fin-2)->val.dou = correct;
    *all->at(fin-1)->val.dou = *out->at(1)->val.dou * correct ;
    
    // fill tds
    _tds->getTuple()->Fill(all);
    
}

void TSorm::getCurrentPattern(Int_t ipat, Double_t *res)
{
    int i;
    Double_t* in = new Double_t[_inleaves.GetEntries()];
    URANIE::Relauncher::TInputVar *var;
    URANIE::DataServer::TDSNtupleD *tup;

    // import data from sampl
    //tup = _form->getTuple();
    _sampl->GetEntry(ipat);
    i = 0;
    TIter nextLeaf(&_inleaves);
    while (TLeaf *leaf = (TLeaf*) nextLeaf()) in[i++] = leaf->GetValue(0);

    // sampling data -> code
    res[0] = ipat;
    inputForCode(in, res + 1);
	
    delete[] in;
}

/* protocole  */
void TSorm::solverLoop()
{
    int ipat, i, j;
    long id;
    int cour, nbk = 2 * (_size - 1) + 1;
    int ret;
    int dpsize = (int) _sampl->GetEntries();
    double beta, beta2, crit, plus, moins, coef;

    double* in = new double[_NinDat + _Naux + 1];
    double* rot = new double[_size * _size];
	double* pt = new double[_size];
    double* rpt = new double[_size];
    TDemiAxeSorm ** solv = new TDemiAxeSorm *[nbk];
    vector<vector<URANIE::DataServer::UEntry*>> codi, codo, all;

    /* init */
    build();
    codi.resize(nbk);
    allocInCodeEntries(nbk, codi);
    codo.resize(nbk);
    allocOutCodeEntries(nbk, codo);
    all.resize(1);
    allocDataEntries(1, all);

    /* main loop */
    for (ipat = 0; ipat < dpsize; ipat++)
    {
        /* get design point */
        getCurrentPattern(ipat, in);
        beta2 = 0.;
        for (i = 0; i < _size; i++)
            beta2 += in[i + 1] * in[i + 1];
        beta = sqrt(beta2);
        /* rotation correspondante */
        sorm_gramschmidt(_size, in + 1, rot);
#ifdef TRACE
        for (i=0; i<_size; i++) std::cout << in[i+1] << "  ";
        std::cout << std::endl;
        for (i=0; i<_size; i++)
        {
            for(j=0; j<_size; j++) std::cout << rot[i*_size+j] << "  ";
            std::cout << std::endl;
        }
        std::cout << std::endl;
#endif

        /** pseudo différence finies hessien **/
        /* init des directions */
        for (id = 0; id < nbk; id++)
        {
            pt[0] = beta;
            for (j = 1; j < _size; j++)
                pt[j] = 0.0;
            if (id)
                pt[(id - 1) / 2 + 1] = (id % 2) ? DELTA * beta : -DELTA * beta;
#ifdef TRACE
            for (i=0; i<_size; i++) std::cout << pt[i] << "  ";
            std::cout << std::endl;
#endif
            solv[id] = new TDemiAxeSorm(_size, pt);
            sorm_rotation(_size, rot, pt, rpt);
#ifdef TRACE
            for (i=0; i<_size; i++) std::cout << rpt[i] << "  ";
            std::cout << std::endl;
#endif
            inputForCode(rpt, &codi[id]);
            _run->put(id, &codi[id], &codo[id]);
        }
        /* recherche zéros */
        cour = nbk;
        while (cour)
        {
            ret = _run->get(id);
            if (ret == 0)
            {
                throw PureEvalNeed();
            }
            else
            {
                crit = _lapse->from(&codo[id]);
#ifdef TRACE
                for (i=0; i<_size; i++) std::cout << *codi[id][i]->val.dou << "  ";
                std::cout << ":>  " ;
                for (i=0; i<_size+3; i++) std::cout << *codo[id][i]->val.dou << "  ";
                std::cout << "->  " << crit;
                std::cout << std::endl;
#endif
                if (solv[id]->next(crit, pt))
                {
                    for (auto val: codo[id]) val->clearEntry();
                    sorm_rotation(_size, rot, pt, rpt);
                    inputForCode(rpt, &codi[id]);
                    _run->put(id, &codi[id], &codo[id]);
                }
                else
                {
                    cour--;
                }
            }
        }
        /* calcul coefficiant correction */
        coef = 1.;
        beta = solv[0]->courbure();
        for (id = 1; id < nbk; id += 2)
        {
            plus = 2 * (solv[id]->courbure() - beta)
                    / (DELTA * DELTA * beta * beta);
            moins = 2 * (solv[id + 1]->courbure() - beta)
                    / (DELTA * DELTA * beta * beta);
            coef *= 0.5 / sqrt(1. + beta * plus)
                    + 0.5 / sqrt(1. + beta * moins);
#ifdef TRACE
            printf("SORM %g %g %g\n", plus, moins, coef);
            printf("SORM beta %g betap %g, betam %g\n", beta
                    , solv[id]->courbure(), solv[id + 1]->courbure());
#endif
        }

        /* mise a jour données */
        fill(in[0], &codi[0], &codo[0], &all[0], coef);

        /* clean up */
        for (id = 0; id < nbk; id++)
        {
            delete solv[id];
            solv[id] = NULL;
        }

    }
	

    delete[] in;
    delete[] rot;
	delete[] pt;
	delete[] rpt;
	for(Int_t itt = 0; itt<nbk; itt++)
	{
        for (auto it : codi[itt]) delete it;
        codi[itt].clear();
        for (auto it : codo[itt]) delete it;
        codo[itt].clear();
    }
    for (auto it : all[0]) delete it;
    all[0].clear();
    delete[] solv;

}

}
} // namespace
