//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.2 $
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   ridder.c
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date    2012-07-12.11.15.46

 \brief ridder's roots method
 */

#include <math.h>
#include <stdlib.h>

#include "ridder.h"

/* user define */
#define EPSILON 1.0e-12

/* etat */
#define START 0
#define PING 1
#define PONG 2

#define UNUSED -1.11e30

#define BRACKET(a,b) (((a)>0.&&(b)<0.)||((a)<0.&&(b)>0.))

/* api */
struct ridder *create_ridder(double xl, double fl, double xt, double ft)
{
    struct ridder *ret;

    if (BRACKET(fl, ft))
    {

        ret = (struct ridder *) malloc(sizeof(struct ridder));

        ret->low = xl;
        ret->flow = fl;
        ret->top = xt;
        ret->ftop = ft;
        ret->etat = START;
        ret->iter = 0;
        ret->res = UNUSED;
        ret->eps = EPSILON;
    }
    else
        ret = NULL;

    return ret;
}

int ridder_first(struct ridder *self, double *nextx)
{
    *nextx = self->med = (self->low + self->top) / 2.;
    self->etat = PING;
    return 1;
}

int ridder_next(struct ridder *self, double fold, double *nextx)
{
    double s, xnew;

    self->iter++;
    switch (self->etat)
    {
    case PING:
        self->fmed = fold;
        s = sqrt(self->fmed * self->fmed - self->flow * self->ftop);
        if (s == 0.0)
            return 0;
        xnew = self->med
                + (self->med - self->low)
                        * ((self->flow >= self->ftop ? 1.0 : -1.0) * self->fmed
                                / s);
        if (fabs(xnew - self->res) < self->eps)
            return 0;
        *nextx = self->res = xnew;
        self->etat = PONG;
        return 1;

    case PONG:
        if (fold == 0.0)
            return 0;
        if (BRACKET(self->fmed, fold))
        {
            self->low = self->med;
            self->flow = self->fmed;
            self->top = self->res;
            self->ftop = fold;
        }
        else if (BRACKET(self->flow, fold))
        {
            self->top = self->res;
            self->ftop = fold;
        }
        else
        {
            self->low = self->res;
            self->flow = fold;
        }
        if (fabs(self->top - self->low) <= self->eps)
            return 0;
        *nextx = self->med = (self->low + self->top) / 2.;
        self->etat = PING;
        return 1;
	default:
        return 0;	
    }
}

float ridder_result(struct ridder *self)
{
    return (float) self->res;
}

void ridder_clean(struct ridder *self)
{
    free(self);
}
