// --- C++ ---
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2023 CEA/DEN
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   TCMAES.cxx
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2023-10-12
 
 \brief  CMAES algorithm
*/

#include "TCMAES.h"

#include <math.h>
#ifdef WIN32
#include <windows.h>
#endif

// ROOT
#include "TVectorD.h"
#include "TMatrixDEigen.h"
// Uranie

ClassImp(URANIE::Reoptimizer::TCMAES);

/* option */
#define POP
#define REFLEX
//#define TRACE

/*** CMAES constantes ***/
#define SQRT2 sqrt(2)

/*** Constructor ***/
URANIE::Reoptimizer::TCMAES::TCMAES(URANIE::DataServer::TDataServer *tds,
        URANIE::Relauncher::TRun *run) :
    URANIE::Reoptimizer::TOptimShare(tds, run), _mt()
{
    // init
    _crit = nullptr;

    // defaut
    _Ncycle = 100;
    _Nlambda = 32;
    _Nmu = 8;
    _repair_point = &URANIE::Reoptimizer::TCMAES::repair_point_reject;

#ifndef WIN32
    _mt.SetSeed(getpid());
#else
    _mt.SetSeed(GetCurrentProcessId());
#endif
}

URANIE::Reoptimizer::TCMAES::~TCMAES()
{
    if (_dom) delete[] _dom;
}

/*** initialisation ***/
void URANIE::Reoptimizer::TCMAES::addObjective(URANIE::DataServer::TAttribute *obj,
        TOptimFit *fit)
{
    if (!_crit) {
        _y = obj;
        _crit = new URANIE::Reoptimizer::TFitGetter(obj, _fun->getOutputList(), fit);
    }
    else std::cout << "CMAES is single objective\n";
}

void URANIE::Reoptimizer::TCMAES::setSize(int loop, int step, int pop)
{
    _Ncycle = loop;
    _Nlambda = step;
    _Nmu = pop;
}

void URANIE::Reoptimizer::TCMAES::setUnbounded()
{
    _repair_point = &URANIE::Reoptimizer::TCMAES::repair_point_unbound;
}

void URANIE::Reoptimizer::TCMAES::setRejectBound()
{
    _repair_point = &URANIE::Reoptimizer::TCMAES::repair_point_reject;
}

void URANIE::Reoptimizer::TCMAES::setReflexBound()
{
    _repair_point = &URANIE::Reoptimizer::TCMAES::repair_point_reflex;
}

// init
void URANIE::Reoptimizer::TCMAES::initCmaesConst()
{
    int size = _NinDat;
    // ref article 2002
    _alpha = 1./_Nmu;
    _ccoef = 4. / (size+4);
    _cucoef = sqrt(_ccoef * (2-_ccoef));
    _ccovorig = 2. / ((size + SQRT2)*(size+ SQRT2));
    _ccovtest = (2 * _Nmu - 1.) / ((size + SQRT2)*(size+ SQRT2) + _Nmu);
    _cdcoef = _ccoef;
    _cudcoef = sqrt(_cdcoef * (2 - _cdcoef));
    _cbeta = (1. / _cdcoef) + 1;
    _Xi = sqrt(size) * (1 - 1/(4*size) + 1/(21*size*size));
}

void URANIE::Reoptimizer::TCMAES::initDomain()
{
    int i;
    URANIE::DataServer::TAttribute *att;

    _dom = new struct domain[_NinDat];
    TList *attrlist = _tds->getListOfAttributes();
    TIter next(attrlist);

    for (int i=0; i<_NinDat; i++) {
        att = (URANIE::DataServer::TAttribute *) next();
        _dom[i].mini = att->getLowerBound();
        _dom[i].maxi = att->getUpperBound();
        _dom[i].it = _dom[i].maxi - _dom[i].mini;
    }
    delete attrlist;
}

//void URANIE::Reoptimiser::TCMAES::init
void URANIE::Reoptimizer::TCMAES::build()
{
    int i, j;

    initCmaesConst();
    initDomain();
    fillDS();
    _tds->createTuple();
    if (_tdstrc) initTrace();
    
    // struct cmaes
    _etat.stepg = 0.25 ; // 1.;
    _etat.corr = new TMatrixD(_NinDat, _NinDat);
    _etat.step = new TMatrixD(_NinDat, 1);
    _etat.accu = new TMatrixD(_NinDat, 1);
    for (i=0; i<_NinDat; i++) {
        for (j=0; j<_NinDat; j++) {
            (*_etat.corr)(i,j) = 0.0;
        }
        (*_etat.corr)(i,i) = 1.0;
        (*_etat.step)(i,0) = 0.0;
        (*_etat.accu)(i,0) = 0.0;
    }
    //struct population
    _pop.ind = new struct solution [_Nmu];
    _pop.size = _Nmu;
    _pop.etat = -_Nmu;

}

void URANIE::Reoptimizer::TCMAES::initRandomPoint(TMatrixD &pt)
{
    int i;

    for (i=0; i<_NinDat; i++) {
        // pt(i, 0) = (_dom[i].it * _mt.Rndm()) + _dom[i].mini;
        pt(i, 0) = _mt.Rndm() ;
    }
}


/*** solver loop ***/
// trace helper
void URANIE::Reoptimizer::TCMAES::echo(TMatrixD &A)
{
    int i, j;
    for (i=0; i<_NinDat; i++) {
        for (j=0; j<_NinDat; j++) std::cout << A(i,j) << " ";
        cout << std::endl;
    }
    cout << std::endl;
}

void URANIE::Reoptimizer::TCMAES::echoV(TMatrixD &v)
{
    int i;
    for (i=0; i<_NinDat; i++) std::cout << v(i, 0) << " ";
    cout << std::endl;
}

// helper
void URANIE::Reoptimizer::TCMAES::massCenter(TMatrixD &center)
{
    int nbv, nbi;
    for (nbv=0; nbv<_NinDat; nbv++) {
        center(nbv,0) = 0.;
        for (nbi=0; nbi<_pop.size; nbi++) {
            center(nbv, 0) += (*_pop.ind[nbi].point)(nbv, 0);
        }
        center(nbv, 0) /= _Nmu;
    }
}

void URANIE::Reoptimizer::TCMAES::moveCenter(TMatrixD &center)
{
    int nbv, nbi;
    for (nbv=0; nbv<_NinDat; nbv++) {
        center(nbv,0) = 0.;
        for (nbi=0; nbi<_pop.size; nbi++) {
            center(nbv, 0) += (*_pop.ind[nbi].aleas)(nbv, 0);
        }
        center(nbv, 0) /= _Nmu;
    }
}

// bound 
TMatrixD* URANIE::Reoptimizer::TCMAES::repair_point_unbound(struct solution &sol)
// no bound test
{
    int i;
    //double min = MIN;
    //double max = MIN+DOM;
    TMatrixD *point = sol.point;
    TMatrixD *rpoint = new TMatrixD(_NinDat,1);

    for (i=0; i<_NinDat; i++) (*rpoint)(i, 0) = (*point)(i, 0); 
    return rpoint ;
}

TMatrixD* URANIE::Reoptimizer::TCMAES::repair_point_reflex(struct solution &sol)
// reflex domain
{
    int i;
    //double min = MIN;
    //double max = MIN+DOM;
    TMatrixD *point = sol.point;
    TMatrixD *rpoint = new TMatrixD(_NinDat,1);
    int ret = 0;

    for (i=0; i<_NinDat; i++) {
        if ((*point)(i, 0) < -1. ) { //_dom[i].mini - _dom[i].it) {
            ret = 1;
            break;
        }
        else if ((*point)(i, 0) > 2. ) { //_dom[i].maxi + _dom[i].it) {
            ret = 1; break;
        }
        else if ((*point)(i, 0) < 0. ) { //_dom[i].mini) {
            (*rpoint)(i, 0) = - (*point)(i,0) ; //2 * _dom[i].mini - (*point)(i, 0);
        }
        else if ((*point)(i, 0) > 1. ) { //_dom[i].maxi) {
            (*rpoint)(i, 0) = 2. - (*point)(i,0) ; //2 * _dom[i].maxi - (*point)(i, 0);
        }
        else (*rpoint)(i, 0) = (*point)(i, 0);
    }
    if (ret) delete rpoint;
    return (ret) ? NULL : rpoint ;
}

TMatrixD* URANIE::Reoptimizer::TCMAES::repair_point_reject(struct solution &sol)
// reject unviable point
{
    int i;
    TMatrixD *point = sol.point;
    TMatrixD *rpoint = new TMatrixD(_NinDat, 1);
    int ret = 0;

    for (i=0; i<_NinDat; i++) {
        // if ((*point)(i, 0) < _dom[i].mini || (*point)(i, 0) > _dom[i].maxi ) {
        if ((*point)(i, 0) < 0. || (*point)(i, 0) > 1. ) {
            ret = 1;
            break;
        }
        else {
            (*rpoint)(i,0) = (*point)(i,0);
        }
    }
    if (ret) delete rpoint;
    return (ret) ? nullptr : rpoint;
}

void URANIE::Reoptimizer::TCMAES::b_from_corr(TMatrixD &BB, TMatrixD &BBB)
{
    int i,j;

    const TMatrixDEigen vpropre(*_etat.corr);
    BB = vpropre.GetEigenVectors();
    //echo(BB);
    const TMatrixD BA = vpropre.GetEigenValues();
    TVectorD length(_NinDat);
    for (i=0; i<_NinDat; i++) {
        //std::cout << BA(i) << " ";
        length(i) = sqrt(BA(i,i)); // valeurs propres > 0 correlation
    }
    //std::cout << std::endl << std::endl;
    //TMatrixD BB(5, 5);
    for (i=0; i<_NinDat; i++) {
        for (j=0; j<_NinDat; j++) BBB(i, j) = length(j) * BB(i, j);
    }
    //echo(BBB);
}

void URANIE::Reoptimizer::TCMAES::push(int ide, TMatrixD &pt, 
        vector<URANIE::DataServer::UEntry*> *codi, 
        vector<URANIE::DataServer::UEntry*> *codo)
{
    int i;
    long j;
    double ret;
    std::vector<double> veci;

    // alloc
    veci.resize(_NinDat);
    // flow
    for (i=0; i<_NinDat; i++) {
        veci[i] = (pt(i,0) * _dom[i].it) + _dom[i].mini;
        // std::cout << veci[i] << " ";
    }
    inputForCode(&veci[0], codi);
    _run->put(ide, codi, codo);
}

long URANIE::Reoptimizer::TCMAES::pull(vector<struct solution> &gen,
        vector<vector<URANIE::DataServer::UEntry*>> &codi, 
        vector<vector<URANIE::DataServer::UEntry*>> &codo,
        vector<URANIE::DataServer::UEntry*> *data)
{
    long j;

    _run->get(j);
    gen[j].fit = _crit->from(&codo[j]);
    if (_tdstrc) {
        fillSample(gen[j].nbe, _tdstrc, &codi[j], &codo[j], data);
        for (auto it : *data) it->clearEntry();
    }
    // std::cout << ">--> " << ret << std::endl;

    return j;
}

void URANIE::Reoptimizer::TCMAES::add_population(struct solution &sol)
{
    int idx, imax;

    if (_pop.etat < 0 ) { // not full
        idx = -_pop.etat - 1;
        // copie
        _pop.ind[idx] = sol; 
        _pop.etat++;
        // search worse
        if (_pop.etat == 0) {
            imax = 0;
            for (int i=1; i<_pop.size; i++) {
                if (_pop.ind[i].fit > _pop.ind[imax].fit ) imax = i;
            }
            _pop.etat = imax;
        }
    }
    else { // full replace
        if (sol.fit < _pop.ind[_pop.etat].fit) {
            // replace
            _pop.ind[_pop.etat] = sol;
            // search worse
            imax = 0;
            for (int i=1; i<_pop.size; i++) {
                if (_pop.ind[i].fit > _pop.ind[imax].fit ) imax = i;
            }
            _pop.etat = imax;
        }
        else {
            // forget
            if (sol.aleas) delete sol.aleas;
            delete sol.point;
        }
    }

}

void URANIE::Reoptimizer::TCMAES::solverLoop()
{
    //struct cmaes cma;
    //TRandom3 mt;
    double vtmp;
    //initCmaes(cma);
    std::vector<struct solution> nouveau;
    std::vector<std::vector<URANIE::DataServer::UEntry*>> codi, codo, sampl;

    build();
    nouveau.resize(_Nlambda);
    //struct population _pop;
    double ccov = (1 < _ccovtest)?
        _alpha * _ccovorig + (1-_alpha) : _alpha * _ccovorig + (1-_alpha)*_ccovtest ;
    // code uentry
    codi.resize(_Nlambda); codo.resize(_Nlambda);
    allocInCodeEntries(_Nlambda, codi);
    allocOutCodeEntries(_Nlambda, codo); 
    sampl.resize(1);
    allocDataEntries(1, sampl);
    
    // first generation
    long nbe = 0;
    /* empile */
    for (int nbph=0; nbph<_Nlambda; nbph++, nbe++) {
        nouveau[nbph].idx = nbph;
        nouveau[nbph].nbe = nbe;
        nouveau[nbph].aleas = NULL;
        nouveau[nbph].point = new TMatrixD(_NinDat,1);
        initRandomPoint(*nouveau[nbph].point);
        push(nbph, *nouveau[nbph].point, &codi[nbph], &codo[nbph]);
    }
    /* depile */
    for (int nbpl=0; nbpl<_Nlambda; nbpl++) {
        int nbee = pull(nouveau, codi, codo, &sampl[0]);
#ifdef TRACE
        echoV(*nouveau[nbee].point);
        std::cout << "  -> " << nouveau[nbee].fit << std::endl;
#endif
        add_population(nouveau[nbee]);
    }

    // generation loop
    for (int nbcy=0; nbcy<_Ncycle; nbcy++) {
        TMatrixD C(_NinDat, _NinDat);
        TMatrixD CC(_NinDat, _NinDat);
        b_from_corr(C, CC);
#ifdef TRACE
        echo(CC);
#endif
        // center mass
        TMatrixD center(_NinDat, 1);
        massCenter(center);
        // generation of lambda descendant
        _pop.etat = -_pop.size;
        // clean uentry TODO
        for (int i=0; i<_Nlambda; i++) {
            for (auto it : codi[i]) it->clearEntry();
            for (auto it : codo[i]) it->clearEntry();
        } 
        // empile
        for (int nbl=0; nbl<_Nlambda; nbl++, nbe++) {
            TMatrixD *rpoint;
            nouveau[nbl].idx = nbl;
            nouveau[nbl].nbe = nbe;
            nouveau[nbl].point = new TMatrixD(_NinDat,1);
            nouveau[nbl].aleas = new TMatrixD(_NinDat,1);
            do {
                for (int nbv=0; nbv<_NinDat; nbv++) (*nouveau[nbl].aleas)(nbv, 0) = _mt.Gaus();
                *nouveau[nbl].point = center + _etat.stepg * CC * (*nouveau[nbl].aleas);
#ifdef TRACE
                echoV(*nouveau[nbl].point);
#endif
            } while ((rpoint = (this->*_repair_point)(nouveau[nbl])) == NULL);
            push(nbl, *rpoint, &codi[nbl], &codo[nbl]);
            delete rpoint;
        }
        // depile
        for (int nbl=0; nbl<_Nlambda; nbl++) {
            int nbpl = pull(nouveau, codi, codo, &sampl[0]);
            //echoV(center);
            //echo(CC);
#ifdef TRACE
            echoV(*nouveau[nbpl].point);
            std::cout << "  -> " << nouveau[nbpl].fit << std::endl;
#endif
            add_population(nouveau[nbpl]);
        }
#ifdef TRACE
        std::cout << "  ---> " << _pop.ind[_pop.etat].fit << std::endl;
#endif
#ifdef POP
        std::cout << nbcy << std::endl;
        for (int i=0; i<_Nmu; i++) {
            echoV(*(_pop.ind[i].point));
            std::cout << "  ---> " << _pop.ind[i].fit << std::endl;
        }
        std::cout << std::endl << std::endl ;
#endif
        // new current
        //echoV(dmemo);
        //current += _etat.stepg * CC * dmemo;
        //echoV(current);
        // covariance adaptation
        TMatrixD dmemo(_NinDat, 1);
        moveCenter(dmemo);
        (*_etat.accu) = (1 - _ccoef) * (*_etat.accu) + _cucoef * sqrt(_Nmu) * CC * dmemo;
#ifdef TRACE
        std::cout << "cucoef " << _cucoef << std::endl;
        echoV(*_etat.accu);
#endif
        // *_etat.corr =  (1 - ccov) * (*_etat.corr) +  ccov * (*_etat.accu) * (*_etat.accu).T
        TMatrixD taccu(TMatrixD::kTransposed, *_etat.accu);
        TMatrixDSym tcov(TMatrixDSym::kAtA, taccu);
#ifdef TRACE
        // echoSym(tcov);
#endif
        // 2003 added correction
        TMatrixDSym taccu2(_NinDat);
        taccu2.Zero();
        //for (int i=0; i<_NinDat; i++) {
        //    for (int j=0; j<_NinDat; j++) taccu2(i,j) = 0.;
        //}
        for (int nbi=0; nbi<_Nmu; nbi++) {
            TMatrixD ttmp(TMatrixD::kTransposed, *_pop.ind[nbi].aleas);
            TMatrixDSym t(TMatrixDSym::kAtA, ttmp);
            taccu2 += t;
        }
        taccu2 *= 1./_Nmu;
        TMatrixD cct(TMatrixD::kTransposed, CC);
        TMatrixD corr2 = CC * taccu2 * cct;
#ifdef TRACE
        echo(corr2);
#endif
        // global step adaptation
        (*_etat.step) = (1 - _cdcoef) * (*_etat.step) + _cudcoef * sqrt(_Nmu) * C * dmemo;
#ifdef TRACE
        echoV(*_etat.step);
#endif
        double norm = 0;
        for (int nbv=0; nbv<_NinDat; nbv++) norm += (*_etat.step)(nbv,0) * (*_etat.step)(nbv,0);
        norm = sqrt(norm);
        _etat.stepg *= exp((norm - _Xi) / (_Xi * _cbeta));
#ifdef TRACE
        std::cout << "glob " << _etat.stepg << std::endl;
#endif
        // current
    }
    // result
    for (int nbi=0; nbi<_Nmu; nbi++) {
        int nbsel = _pop.ind[nbi].idx;
        fillSample(nbi, _tds, &codi[nbsel], &codo[nbsel], &sampl[0]);
        for (auto it : sampl[0]) it->clearEntry();
    }
    // unalloc
    for (int i=0; i<_Nlambda; i++) {
        for (auto it : codi[i]) delete it;
        codi[i].clear();
        for (auto it : codo[i]) delete it;
        codo[i].clear();
    }

}

/**** post traitement ****/
void URANIE::Reoptimizer::TCMAES::fillSample(long ev, 
    URANIE::DataServer::TDataServer *tds, vector<URANIE::DataServer::UEntry*> *in,
    vector<URANIE::DataServer::UEntry*> *out, 
    vector<URANIE::DataServer::UEntry*> *sampl)
{
    vector<URANIE::DataServer::UEntry*> *tmpi =
        new vector<URANIE::DataServer::UEntry*>(in->begin(), in->end());
    vector<URANIE::DataServer::UEntry*> *tmpo =
        new vector<URANIE::DataServer::UEntry*>(out->begin(), out->end());
    vector<URANIE::DataServer::UEntry*> *tmpa =
        new vector<URANIE::DataServer::UEntry*>(sampl->begin()+1, sampl->end());

    sampl->at(0)->pushEntry((double) ev);
    allFromCode(tmpi, tmpo, tmpa);
    tds->getTuple()->Fill(sampl);
}

