/* --- Ansi C --- */
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2023 CEA/DEN
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   TCMAES.h
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2023-10-12
 
 \brief  CMAES algorithm
*/

#ifndef __CMAES__
#define __CMAES__

// ROOT
//#include "Rtypes.h"
#include "TRandom3.h"
#include "TMatrixD.h"

//URANIE
#include "Reoptimizer.h"
#include "TOptimShare.h"

#include "TDataServer.h"
#include "TAttribute.h"
#include "TRun.h"
#include "TEval.h"
#include "TOptimFit.h"
#include "TFitGetter.h"

struct cmaes {
    TMatrixD    *corr;
    //TMatrixD    base(SIZE, SIZE);
    TMatrixD    *step;
    TMatrixD    *accu;
    double      stepg;
};

struct solution {
    TMatrixD    *aleas;
    TMatrixD    *point;
    double      fit;
    int         idx, nbe;
};

struct population {
    struct solution *ind;
    int etat, size;
};

struct domain {
    double mini, maxi, it;
};

namespace URANIE {
namespace Reoptimizer {

class TCMAES : public TOptimShare
{
private:
    void massCenter(TMatrixD &);
    void moveCenter(TMatrixD &);
    void b_from_corr(TMatrixD &bb, TMatrixD &bbb);
    void add_population(struct solution &sol);
    TMatrixD * (TCMAES::*_repair_point)(struct solution &sol);
    TMatrixD *repair_point_unbound(struct solution &sol);
    TMatrixD *repair_point_reflex(struct solution &sol);
    TMatrixD *repair_point_reject(struct solution &sol);

protected:
    int _Ncycle, _Nlambda, _Nmu;
    double _ccoef, _cucoef, _ccovorig, _ccovtest;
    double _cdcoef, _cudcoef, _cbeta, _Xi, _alpha;
    TRandom3 _mt;
    URANIE::DataServer::TAttribute *_y;
    URANIE::Reoptimizer::TFitGetter *_crit, *_valid;
    struct population _pop;
    struct cmaes _etat;
    struct domain *_dom;

    void initCmaesConst();
    void initDomain();
    void build();
    void echo(TMatrixD &m);
    void echoV(TMatrixD &v);
    void initRandomPoint(TMatrixD &pt);
    void push(int nbe, TMatrixD &pt, std::vector<URANIE::DataServer::UEntry*> *codi,
            std::vector<URANIE::DataServer::UEntry*> *codo);
    long pull(std::vector<struct solution> &gen, 
            std::vector<std::vector<URANIE::DataServer::UEntry*>> &codi,
            std::vector<std::vector<URANIE::DataServer::UEntry*>> &codo,
            std::vector<URANIE::DataServer::UEntry*> *sampl);
    void fillSample(long i, URANIE::DataServer::TDataServer *tds,
            std::vector<URANIE::DataServer::UEntry*> *codi,
            std::vector<URANIE::DataServer::UEntry*> *codo,
            std::vector<URANIE::DataServer::UEntry*> *sampl);

    
public:
    TCMAES(URANIE::DataServer::TDataServer *tds, URANIE::Relauncher::TRun *run);
    virtual ~TCMAES();

    // initialisation
    virtual void addObjective(URANIE::DataServer::TAttribute *,
            URANIE::Reoptimizer::TOptimFit *f=NULL);
    void setSize(int loop, int step, int pop);
    void setUnbounded();
    void setRejectBound();
    void setReflexBound();

    // solve
    void fitnessChanged();
    virtual void solverLoop();

    // post
    double *item(int);

    ClassDef(URANIE::Reoptimizer::TCMAES, ID_REOPTIMIZER)
};

} } //namespace

#endif
