//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.7 $
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TNlopt.cxx
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date    2012-03-20.16.04.36

 \brief   interface algorithme optimisation de la bibliotheque NLOpt
 */

#include "TNlopt.h"

#include <stdlib.h>

#include "nlopt.h"
#include "loop.h"
#include "nlopt_work.h"

#include "TMasterVar.h"
#include "TFitGetter.h"

#define SUIVI
#define IsRUNNING 16384

#define EPS 1.0e-5
#define ZERO 1.e-20

ClassImp (URANIE::Reoptimizer::TNlopt);

/* struct nlopt_work
{
    URANIE::Reoptimizer::TNlopt *self;
    nlopt_opt solv, subsolv;
    double *guess, *last, *work, *grad, minf;
    vector<URANIE::DataServer::UEntry*> *codi, *codo, *codg;
    int id, state, noerr;
    Bool_t lastg;
}; */


namespace URANIE
{
namespace Reoptimizer
{

//TNlopt
TNlopt::TNlopt(URANIE::DataServer::TDataServer *tds,
        URANIE::Relauncher::TRun *Run, TNloptSolver *solv) :
    TNloptShare(tds, Run, solv), _coef(0)
{
}

TNlopt::TNlopt(URANIE::DataServer::TDataServer *tds,
        URANIE::Relauncher::TEval *fun, TNloptSolver *solv) :
    TNloptShare(tds, fun, solv), _coef(0)
{
}

TNlopt::~TNlopt()
{
}

/* constructeur helpers */
void TNlopt::build()
{
    TNloptShare::build();
    _coef = _NinCod * 2 + 1;
}

/* objective and constraints */
void TNlopt::addObjective(URANIE::DataServer::TAttribute *obj, TOptimFit *fit)
{
    if (!_crit)
        _crit = new TFitGetter(obj, _fun->getOutputList(), fit);
    else
        std::cout << "NLOpt is single objective\n";
}

void TNlopt::addEquality(URANIE::DataServer::TAttribute *obj, TOptimFit *fit)
{
    TFitGetter *one = new TFitGetter(obj, _fun->getOutputList(), fit);
    _equal.Add(one);
}

void TNlopt::gradFromCode(double *code, double *solv)
{
    URANIE::Relauncher::TMasterVar *var;
    TIter next(&_param);
    int j = 0;

    while ((var = (URANIE::Relauncher::TMasterVar *) next()))
    {
        var->get(j, solv, code);
        j++;
    }
}

int TNlopt::critere(struct nlopt_work *work, const double *in, double *out,
        double *grad)
{
    if (grad) {
        double* tmp = new double[_NinCod];

        calcul(work, kTRUE, in);
        //out[0] = _crit->from(_NinCod, work->work + _NinCod, work->grad, tmp);
        out[0] = _crit->from(_NinCod, work->codo, work->codg, tmp);
        gradFromCode(tmp, grad);
	    delete[] tmp;
	}
    else {
        calcul(work, kFALSE, in);
        out[0] = _crit->from(work->codo);
	}
    return work->noerr;
}

int TNlopt::inequality(struct nlopt_work *work, const double *in, double *out,
        double *grad)
{
    TFitGetter *one;
    TIter next(&_const);
    int i = 0;

    if (grad) {
        double* tmp = new double[_NinCod];

        calcul(work, kTRUE, in);
        while ((one = (TFitGetter *) next()))
        {
            //out[i] = one->from(_NinCod, work->work + _NinCod, work->grad, tmp);
            out[i] = one->from(_NinCod, work->codo, work->codg, tmp);
            gradFromCode(tmp, grad + i * _NinDat);
            i++;
        }
	    delete[] tmp;
	}
    else {
        calcul(work, kFALSE, in);
        while ((one = (TFitGetter *) next()))
            out[i++] = one->from(work->codo);
    }
    return work->noerr;
}

int TNlopt::equality(struct nlopt_work *work, const double *in, double *out,
        double *grad)
{
    TFitGetter *one;
    TIter next(&_equal);
    int i = 0;

    if (grad) {
        double* tmp = new double[_NinCod];

        calcul(work, kTRUE, in);
        while ((one = (TFitGetter *) next()))
        {
            //out[i] = one->from(_NinCod, work->work + _NinCod, work->grad, tmp);
            out[i] = one->from(_NinCod, work->codo, work->codg, tmp);
            gradFromCode(tmp, grad + i * _NinDat);
            i++;
        }
	    delete[] tmp;
	}
	else {
        calcul(work, kFALSE, in);
        while ((one = (TFitGetter *) next()))
            out[i++] = one->from(work->codo);
	}
    return work->noerr;
}


/* runner indirection */
void TNlopt::put(struct nlopt_work *self, Bool_t grad, const double *in)
{
    if (grad) {
        for (auto val: *self->codo ) val->clearEntry();
        for (auto val: *self->codg ) val->clearEntry();
        inputForCode(in, self->codi);
        put(self->id, self->codi, self->codo, self->codg);
    }
    else {
        for (auto val: *self->codo ) val->clearEntry();
        inputForCode(in, self->codi);
        put(self->id, self->codi, self->codo);
    }
}

void TNlopt::put(long id, 
    vector<URANIE::DataServer::UEntry *> *in, 
    vector<URANIE::DataServer::UEntry *> *out, int ncalc)
{
    struct grad_work job =
    { NULL, NULL, NULL, 1 };

    _job[id] = job;
    _run->put(id * _coef, in, out, ncalc);
}

void TNlopt::put(long id, 
    vector<URANIE::DataServer::UEntry *> *in, 
    vector<URANIE::DataServer::UEntry *> *out, 
    vector<URANIE::DataServer::UEntry *> *grad, int ncalc)
{
    int i, j;
    struct grad_work job =
    { grad, NULL, NULL, _coef };

    job.in = new vector<vector<URANIE::DataServer::UEntry *>>(); //new double[2 * _NinCod * _NinCod];
    job.in->resize(2 * _NinCod);
    allocInCodeEntries(2 * _NinCod, *job.in);
    job.out = new vector<vector<URANIE::DataServer::UEntry *>>(); //new double[2 * _NinCod * _NoutCod];
    job.out->resize(2 * _NinCod);
    allocOutCodeEntries(2 * _NinCod, *job.out);

    _job[id] = job;

    /* generation plan OAT */
    /* copie conforme */
    for (i = 0; i < 2 * _NinCod; i++)
        for (j = 0; j < _NinCod; j++)
            (*job.in)[i][j]->copyEntry( (*in)[j]);
    /* OAT diff */
    for (i = 0; i < _NinCod; i++)
    {
        if (fabs(*(*job.in)[2 * i][i]->val.dou) < ZERO)
        {
            (*job.in)[2 * i][i]->pushEntry( ZERO);
            (*job.in)[2 * i + 1][i]->pushEntry( -ZERO);
        }
        else
        {
            (*job.in)[2 * i][i]->pushEntry( *(*in)[i]->val.dou * (1. + EPS));
            (*job.in)[2 * i + 1][i]->pushEntry( *(*in)[i]->val.dou * (1. - EPS));
        }
    }
#ifdef TRACE
    for (i=0; i<2*_NinCod; i++)
    {
        for (j=0; j<_NinCod; j++)
        std::cout << (*job.in)[i][j]->val.dou << " ";
        std::cout << std::endl;
    }
#endif

    /* submit des jobs */
    _run->put(id * _coef, in, out, ncalc);
    for (i = 0; i < 2 * _NinCod; i++)
        _run->put(id * _coef + i + 1, &(*job.in)[i], &(*job.out)[i], ncalc);
    
}

Bool_t TNlopt::readyToGet()
{   
    Bool_t ready;
    int ret;
    long id, gid;
    
    // TODO
    while ((ready =_run->readyToGet())) {
        ret = get(gid);
        if (ret == 0)
            throw new NotASingleEval();
        id = gid / _coef;
        _job[id].count--;
        if (!_job[id].count) break;
    }
    
    return ready;
}

int TNlopt::get(long &id)
{
    int ret, i, j;
    long gid;

    // if readyToGet called
    for (const auto ii : _job) {
        id = ii.first;
        if (ii.second.count == 0) break;
    }
    while (_job[id].count) {
        ret = _run->get(gid);
        if (ret == 0)
            throw new NotASingleEval();
        id = gid / _coef;
        _job[id].count--;
    }
    
    if (_job[id].grad)
    {
        /* gradient estimation */
        for (i = 0; i < _NoutCod; i++)
        {
            if ((*_job[id].grad)[i]->tag == URANIE::DataServer::UEntry::Vector) { 
                //auto tmp = new vector<double>();
                (*_job[id].grad)[i]->clearEntry();
                vector<double> *tmp = (*_job[id].grad)[i]->val.vec;
                tmp->resize(_NinCod);
                for (j = 0; j < _NinCod; j++)
                {
                    (*tmp)[j] = 
                        ( *(*_job[id].out)[2 * j][i]->val.dou - 
                          *(*_job[id].out)[2 * j + 1][i]->val.dou )
                        / (*(*_job[id].in)[2 * j][j]->val.dou -
                          *(*_job[id].in)[2 * j + 1][j]->val.dou );
                }
                //(*_job[id].grad)[i]->assignEntry (tmp);
            }
        }
        /* temp data cleanup */
        for (i=0; i<2*_NinCod; i++) {
            for (auto it : (*_job[id].in)[i]) delete it;
            (*_job[id].in)[i].clear();
            for (auto it : (*_job[id].out)[i]) delete it;
            (*_job[id].out)[i].clear();
        }
        delete _job[id].in;
        delete _job[id].out;
    }
    _job.erase(id);

    return 1;
}

void TNlopt::buildSolver(int siz, struct nlopt_work *opt)
{
    int i;
    
    // data
    _codi.resize(siz);
    allocInCodeEntries(siz, _codi);
    _codo.resize(siz);
    allocOutCodeEntries(siz, _codo);
    _codg.resize(siz);
    allocGradCodeEntries(siz, _codg);

    // solver
    for (i=0; i<siz; i++)
    {
        solver(&(opt[i]), i, _guess[i], _codi[i], _codo[i], _codg[i]);
    }
    
}

void TNlopt::cleanSolver(struct nlopt_work *opt)
{
    free(opt->last);
    free(opt->work);
    free(opt->grad);
    for (auto it : *opt->codi) delete it;
    opt->codi->clear();
    for (auto it : *opt->codo) delete it;
    opt->codo->clear();
    for (auto it : *opt->codg) delete it;
    opt->codg->clear();
    nlopt_destroy(opt->solv);
    if (opt->subsolv) nlopt_destroy(opt->subsolv);
}

}
} //namespace
