//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.7 $
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TNloptGradient.cxx
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date    2012-03-20.16.04.36

 \brief   interface algorithme optimisation de la bibliotheque NLOpt
 */

#include "TNloptGradient.h"

#include <stdlib.h>

#include "nlopt.h"
#include "loop.h"
#include "nlopt_work.h"

#include "TMasterVar.h"
#include "TFitGetter.h"

#define SUIVI
#define IsRUNNING 16384

#define EPS 1.0e-5
#define ZERO 1.e-20

ClassImp (URANIE::Reoptimizer::TNloptGradient);


namespace URANIE
{
namespace Reoptimizer
{

//TNloptGradient
TNloptGradient::TNloptGradient(URANIE::DataServer::TDataServer *tds,
        URANIE::Relauncher::TRun *Run, TNloptSolver *solv) :
    TNloptShare(tds, Run, solv)
{
}

TNloptGradient::TNloptGradient(URANIE::DataServer::TDataServer *tds,
        URANIE::Relauncher::TEval *fun, TNloptSolver *solv) :
    TNloptShare(tds, fun, solv)
{
}

TNloptGradient::~TNloptGradient()
{
}

/* objective and constraints */
void TNloptGradient::addObjective(URANIE::DataServer::TAttribute *obj, 
    TOptimFit *fit)
{
    UNUSED(obj); UNUSED(fit);
    std::cout << "NLOpt with gradient need a gradient attribute\n";
    throw new NeedAGradientAttribute();
}

void TNloptGradient::addObjective(URANIE::DataServer::TAttribute *obj, 
    URANIE::DataServer::TAttribute *grad, TOptimFit *fit)
{
    if (!_crit)
        _crit = new TFitGetter(obj, grad, _fun->getOutputList(), fit);
    else
        std::cout << "NLOpt is single objective\n";
}

void TNloptGradient::addConstraint(URANIE::DataServer::TAttribute *cnt,
        TOptimFit *fit)
{
    UNUSED(cnt); UNUSED(fit);
    std::cout << "NLOpt with gradient need a gradient attribute\n";
    throw new NeedAGradientAttribute();
}

void TNloptGradient::addConstraint(URANIE::DataServer::TAttribute *cnt,
        URANIE::DataServer::TAttribute *grd, TOptimFit *fit)
{
    TFitGetter *one = new TFitGetter(cnt, grd, _fun->getOutputList(), fit);
    _const.Add(one);
}

void TNloptGradient::addEquality(URANIE::DataServer::TAttribute *obj, 
    URANIE::DataServer::TAttribute *grd, TOptimFit *fit)
{
    TFitGetter *one = new TFitGetter(obj, grd, _fun->getOutputList(), fit);
    _equal.Add(one);
}

// evaluate
void TNloptGradient::gradFromCode(double *code, double *solv)
{
    URANIE::Relauncher::TMasterVar *var;
    TIter next(&_param);
    int j = 0;

    while ((var = (URANIE::Relauncher::TMasterVar *) next()))
    {
        var->get(j, solv, code);
        j++;
    }
}

int TNloptGradient::critere(struct nlopt_work *work, const double *in, double *out,
        double *grad)
{
    calcul(work, grad, in);
    if (grad) {
        double* tmp = new double[_NinCod];

        out[0] = _crit->from(_NinCod, work->codo, tmp);
        gradFromCode(tmp, grad);
	    delete[] tmp;
	}
    else {
        out[0] = _crit->from(work->codo);
	}
    return work->noerr;
}

int TNloptGradient::inequality(struct nlopt_work *work, const double *in, double *out,
        double *grad)
{
    TFitGetter *one;
    TIter next(&_const);
    int i = 0;

    calcul(work, grad, in);
    if (grad) {
        double* tmp = new double[_NinCod];

        while ((one = (TFitGetter *) next()))
        {
            out[i] = one->from(_NinCod, work->codo, tmp);
            gradFromCode(tmp, grad + i * _NinDat);
            i++;
        }
	    delete[] tmp;
	}
    else {
        while ((one = (TFitGetter *) next()))
            out[i++] = one->from(work->codo);
    }
    return work->noerr;
}

int TNloptGradient::equality(struct nlopt_work *work, const double *in, double *out,
        double *grad)
{
    TFitGetter *one;
    TIter next(&_equal);
    int i = 0;

    calcul(work, grad, in);
    if (grad) {
        double* tmp = new double[_NinCod];

        while ((one = (TFitGetter *) next()))
        {
            out[i] = one->from(_NinCod, work->codo, tmp);
            gradFromCode(tmp, grad + i * _NinDat);
            i++;
        }
	    delete[] tmp;
	}
	else {
        calcul(work, kFALSE, in);
        while ((one = (TFitGetter *) next()))
            out[i++] = one->from(work->codo);
	}
    return work->noerr;
}


/* runner indirection */
void TNloptGradient::put(struct nlopt_work *self, Bool_t grad, const double *in)
{
    UNUSED(grad);
    
    for (auto val: *self->codo ) val->clearEntry();
    inputForCode(in, self->codi);
    _run->put(self->id, self->codi, self->codo);
}

int TNloptGradient::get(long &id)
{
    return _run->get(id);
}

/* solver */
void TNloptGradient::buildSolver(int siz, struct nlopt_work *opt)
{
    int i;
    
    // data
    _codi.resize(siz);
    allocInCodeEntries(siz, _codi);
    _codo.resize(siz);
    allocOutCodeEntries(siz, _codo);

    // solver
    for (i=0; i<siz; i++)
    {
        solver(&(opt[i]), i, _guess[i], _codi[i], _codo[i], _codo[i]);
    }
    
}

void TNloptGradient::cleanSolver(struct nlopt_work *opt)
{
    free(opt->last);
    free(opt->work);
    free(opt->grad);
    for (auto it : *opt->codi) delete it;
    opt->codi->clear();
    for (auto it : *opt->codo) delete it;
    opt->codo->clear();
    nlopt_destroy(opt->solv);
    if (opt->subsolv) nlopt_destroy(opt->subsolv);
}


}
} //namespace
