//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.2 $
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TVizirSolver.h
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date    2012-04-26.11.24.04

 \brief classe abstraite et racine pour les solvers de Vizir
 */

#ifndef __VIZIR_SOLVER__
#define __VIZIR_SOLVER__

//ROOT
#include "Rtypes.h"
#include "TNamed.h"

// Uranie
#include "Reoptimizer.h"

struct vzragent;
struct vzrbase;
struct vzrsort;

namespace URANIE
{
namespace Reoptimizer
{

typedef enum DiversityAlgo {undef, moga, crowd, pair, ibea, knee, moead, modal} diversityAlgo; 
struct diversity {
    diversityAlgo algo;
    union {
        struct {
            int nba, nbb, nbc;
        };
        struct {
            double val; int nb;
        };
    };
};


class TVizirSolver: public TObject
{
public:
    virtual struct vzragent *build(int, int, int, double) = 0;
    virtual int individu(struct vzragent *ag, int id, double **zone, double **front,
            void **dat, int *rank, int *age, int *info=NULL, double *isol=NULL) = 0;
    virtual Bool_t isCoded() = 0;
    virtual void resume(struct vzragent *) = 0;

    ClassDef(URANIE::Reoptimizer::TVizirSolver, ID_REOPTIMIZER)
};

class TVizirSolverShare: public TVizirSolver
{
protected:
    int _pop, _eval, _step, _policy, _cut1, _cut2, _continue;
    char *_resume;
    Bool_t _doresume, _quiet;
    struct diversity _diversity;

    void defaultShare();
    void buildShare(struct vzrbase *pop);

public:
    void setSize(int pop, int neval = 0, int step = 0);
    Bool_t setResume(int pop, char *file);
    void setStrictConstraintPolicy();
    void setSoftConstraintPolicy();
    void setConstraintPolicy(int val=0);
    void setMogaDiversity(int val=0);
    void setCrowdDiversity(int vois=0);
    void setPairDiversity(int vois=0);
    void setIbeaDiversity(double k=0);
    void setKneeDiversity(int vois=0, double taux=0.0);
    void setMoeadDiversity(int cut1, int cut2=0, int vois=0);
    void setModalDiversity(int vois=0);
    void setStoppingCriteria(int stop=0);
    void setQuiet();
    
    virtual void resume(struct vzragent *ag);
    virtual int individu(struct vzragent *ag, int id, double **zone, double **front,
            void **data, int *rank, int *age, int *info, double *isol);

    virtual Bool_t isCoded() {return kTRUE;}
    
    ClassDef(URANIE::Reoptimizer::TVizirSolverShare, ID_REOPTIMIZER)
};


} } //namespace
#endif
