//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
//// $Id$
//// $Author$
//// $Date$
//// $Revision$
//// $State$
////////////////////////////////////////////////////////////////
/*!\file   TestTTangram.cxx
 *   \author Gilles Arnaud <gilles.arnaud@cea.fr>
 *   \date   mar. juin 11 07:23:28 CEST 2013
 * 
 *   \brief  test file for Vizir solver classes and TRun classes
 * 
 * */

#include "TestTTangram.h"

// CPPUNIT
#include <cppunit/extensions/HelperMacros.h>
#include <cppunit/Test.h>

#define DBL_PREC_FINE 5.e-6
#define DBL_PREC_STD  5.e-4
#define DBL_PREC_LARGE 2.0
#define DBL_PREC_HUGE 8.0
#define DBL_PREC_ROBO 0.05
#define DBL_PREC_ROBI 0.11

// ROOT
#include "Rtypes.h"
//#include "Api.h"

//URANIE
#include "UExceptions.h"
#include "UDisableTest.h"
#include "TDataServer.h"
#include "TAttribute.h"

#include "TTangramEval.h"
#include "TSequentialRun.h"
#include "TThreadedRun.h"

#include "TVizirSwarm.h"
#include "TVizirSimplex.h"
#include "TVizir2.h"
#include "TVizirIsland.h"

#include "TCMAES.h"

#ifdef WITH_PYTHON
#include "python_commands.h"
#endif

#ifdef WITH_MPI
#include "mpi_commands.h"
#endif

// USER
#include "Pheromone.h"

/* user defined */
#define TOLERANCE 0.001
#define NBmaxEVAL 80000
#define STEP 64 
#define STOP 256
#define STOP2 200
//#define TAILLEPOP 10 

#define NBGENE 7 // nombre de paramètres

using namespace URANIE::DataServer;
using namespace URANIE::Relauncher;
using namespace URANIE::Reoptimizer;

// Constructor Destructor
TTangramTest::TTangramTest()
{
}

TTangramTest::~TTangramTest()
{
}

// pre and post treatemnt
void TTangramTest::setUp()
{
}

void TTangramTest::tearDown()
{
}

// TEST SUITE
void TTangramTest::swarmSequential()
{
    /* init */
    TAttribute z1("zone1", 0., 255.), z2("zone2", 0., 255.), z3("zone3", 0.,
            255.), z4("zone4", 0., 255.), z5("zone5", 0., 255.), z6("zone6", 0.,
            255.), z7("zone7", 0., 255.), ct("contrast");

    TTangramEval code;
    code.setInputs(NBGENE, &z1, &z2, &z3, &z4, &z5, &z6, &z7);
    code.addOutput(&ct);

    TDataServer tdsvzr("tdsvzr", "tds for vizir test");
    tdsvzr.keepFinalTuple(false);
    code.addAllInputs(&tdsvzr);

    TVizirSwarm solv;
    solv.setSize(STOP, NBmaxEVAL, STEP/2);
    solv.setQuiet();
    TVizir2 vzr(&tdsvzr, &code, &solv);
    vzr.setFixedSeed();
    vzr.setTolerance(TOLERANCE);
    TMaximizeFit maxim;
    vzr.addObjective(&ct, &maxim);

    /* resolution */
    vzr.solverLoop();

    // /* test */
    validateTangram(tdsvzr, DBL_PREC_LARGE);

}

void TTangramTest::simplexThreaded()
{
    /* init */
    TAttribute z1("zone1", 0., 255.), z2("zone2", 0., 255.), z3("zone3", 0.,
            255.), z4("zone4", 0., 255.), z5("zone5", 0., 255.), z6("zone6", 0.,
            255.), z7("zone7", 0., 255.), ct("contrast");

    TTangramEval code;
    code.setInputs(NBGENE, &z1, &z2, &z3, &z4, &z5, &z6, &z7);
    code.addOutput(&ct);

    TThreadedRun run(&code, 7);
    run.startSlave();
    if (run.onMaster())
    {
        TDataServer tds("tdsvzr", "tds for vizir test");
        tds.keepFinalTuple(false);
        code.addAllInputs(&tds);
        TDataServer trc("tdstrc", "tds for vizir trace");
        trc.keepFinalTuple(false);
    
        TVizirSimplex solv;
        solv.setSize(STOP, NBmaxEVAL, STEP);
        solv.setQuiet();
        TVizir2 vzr(&tds, &run, &solv);
        //TVizirIsland vzr(&tds, &run, &solv);
        //vzr.setIsland(3);
        vzr.setTrace(&trc);
        vzr.setFixedSeed();
        vzr.setTolerance(TOLERANCE);
        TMaximizeFit maxim;
        vzr.addObjective(&ct, &maxim);

        /* resolution */
        vzr.solverLoop();

        /* test and end */
        run.stopSlave();
        validateTangram(tds, DBL_PREC_LARGE);
        validateTangram2(trc, DBL_PREC_LARGE);
    }
}

void TTangramTest::antsThreaded()
{
    /* init */
    TAttribute z1("zone1", 0., 255.), z2("zone2", 0., 255.), z3("zone3", 0.,
            255.), z4("zone4", 0., 255.), z5("zone5", 0., 255.), z6("zone6", 0.,
            255.), z7("zone7", 0., 255.), ct("contrast");

    TTangramEval code;
    code.setInputs(NBGENE, &z1, &z2, &z3, &z4, &z5, &z6, &z7);
    code.addOutput(&ct);

    TThreadedRun run(&code, 7);
    run.startSlave();
    if (run.onMaster())
    {
        TDataServer tds("tdsvzr", "tds for vizir test");
        tds.keepFinalTuple(false);
        code.addAllInputs(&tds);
    
        Pheromone ants(7, 12);
        TVizirAnts solv;
        solv.setSize(STOP, NBmaxEVAL, STEP);
        solv.setQuiet();
        solv.setPheromone(&ants);
        TVizir2 vzr(&tds, &run, &solv);
        //TVizirIsland vzr(&tds, &run, &solv);
        //vzr.setIsland(3);
        vzr.setFixedSeed();
        vzr.setTolerance(TOLERANCE);
        TMaximizeFit maxim;
        vzr.addObjective(&ct, &maxim);

        /* resolution */
        vzr.solverLoop();

        /* test and end */
        run.stopSlave();
        validateTangram(tds, DBL_PREC_LARGE);
    }
}

void TTangramTest::cmaesThreaded()
{
    /* init */
    TAttribute z1("zone1", 0., 255.), z2("zone2", 0., 255.), z3("zone3", 0.,
            255.), z4("zone4", 0., 255.), z5("zone5", 0., 255.), z6("zone6", 0.,
            255.), z7("zone7", 0., 255.), ct("contrast");

    TTangramEval code;
    code.setInputs(NBGENE, &z1, &z2, &z3, &z4, &z5, &z6, &z7);
    code.addOutput(&ct);

    TThreadedRun run(&code, 7);
    run.startSlave();
    if (run.onMaster())
    {
        TDataServer tds("tdsvzr", "tds for vizir test");
        tds.keepFinalTuple(false);
        code.addAllInputs(&tds);
        TDataServer trc("tdstrc", "tds for vizir trace");
        trc.keepFinalTuple(false);
    
        TCMAES opt(&tds, &run);
        opt.setSize(200, 28, 5);
        opt.setTrace(&trc);
        TMaximizeFit maxim;
        opt.addObjective(&ct, &maxim);

        /* resolution */
        opt.solverLoop();

        /* test and end */
        run.stopSlave();
        validateTangram3(tds, DBL_PREC_LARGE);
        validateTangram2(trc, DBL_PREC_LARGE);
    }

}

#ifdef WITH_MPI
void TTangramTest::geneticMpi()
{
    /* run MPI macro */
    Int_t nProc = 4;
    gSystem->Exec(UranieTests::mpi_command(nProc, "root -l -q -b testVizirMpi.C").data());

    /* get results */
    TDataServer tds("tdsvzr", "tds for vizir test");
    tds.keepFinalTuple(false);
    tds.fileDataRead( Form("%s/_tangram_mpi_.dat", gSystem->pwd()),false);
    validateTangram(tds, DBL_PREC_HUGE);
}

void TTangramTest::geneticBiMpi()
{
    /* run MPI macro */
    Int_t nProc = 7;
    gSystem->Exec(UranieTests::mpi_command(nProc, "root -l -q -b testVizirBiMpi.C").data());

    /* get results */
    TDataServer tds("tdsvzr", "tds for vizir test");
    tds.keepFinalTuple(false);
    tds.fileDataRead( Form("%s/_bi_mpi_.dat", gSystem->pwd()), false );
    validateBi(tds);
}

void TTangramTest::geneticUserMpi()
{
    /* run MPI macro */
    Int_t nProc = 7;
    gSystem->Exec(UranieTests::mpi_command(nProc, "root -l -q -b testVizirUserMpi.C").data());

    /* get results */
    TDataServer tds("tdsvzr", "tds for vizir test");
    tds.fileDataRead( Form("%s/_user_mpi_.dat", gSystem->pwd()) );
    validateBi(tds);
}

void TTangramTest::geneticRobust1Mpi()
{
#ifdef WIN32
    URANIE_IGNORED_TEST()
#endif
    /* run MPI macro */
    Int_t nProc = 7;
    gSystem->Exec(UranieTests::mpi_command(nProc, "root -l -q -b testVizirRobust1.C", true).data());

    /* get results */
    TDataServer tds("tdsvzr", "tds for vizir test");
    tds.fileDataRead( Form("%s/_mpi_robust1_.dat", gSystem->pwd()) );
    validateRobustBi(tds);
}

void TTangramTest::geneticRobust2Mpi()
{
#ifdef WIN32
    URANIE_IGNORED_TEST()
#endif
    /* run MPI macro */
    Int_t nProc = 7;
    gSystem->Exec(UranieTests::mpi_command(nProc, "root -l -q -b testVizirRobust2.C", true).data());

    /* get results */
    TDataServer tds("tdsvzr", "tds for vizir test");
    tds.fileDataRead( Form("%s/_mpi_robust2_.dat", gSystem->pwd()) );
    validateRobustBi(tds);
}

#ifdef WITH_PYTHON
void TTangramTest::geneticRobustPyMpi()
{
#ifdef WIN32
    URANIE_IGNORED_TEST()
#endif
    /* run MPI macro */
    Int_t nProc = 7;
    string python_command;
    UranieTests::set_python_executable(python_command);
    gSystem->Exec(UranieTests::mpi_command(nProc, python_command + " testVizirRobust.py", true).data());

    /* get results */
    TDataServer tds("tdsvzr", "tds for vizir test");
    tds.fileDataRead( Form("%s/_mpi_robust_py_.dat", gSystem->pwd()) );
    validateRobustBi(tds);
}
#endif
#endif

void TTangramTest::validateTangram(TDataServer &tds, double prec)
{
    Int_t pat = tds.getNPatterns();
    CPPUNIT_ASSERT_EQUAL(0, pat % STOP);
    // attributes
    // zone1, zone2, ..., zone7, contrast, rgpareto, generation
    CPPUNIT_ASSERT_EQUAL(10, tds.getNAttributes());
    //CPPUNIT_ASSERT_EQUAL(8, tds.getNAttributes());
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("zone1"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("zone5"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("contrast"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("rgpareto"));
    // values
    // theo 1, 1, 1, 1, 1, 1, 0
    CPPUNIT_ASSERT_DOUBLES_EQUAL(85., tds.getValue("contrast", 1), prec);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(85., tds.getValue("contrast", STEP), prec);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(85., tds.getValue("contrast", STOP-1), prec);

}

void TTangramTest::validateTangram2(TDataServer &tds, double prec)
{
    UNUSED(prec);
    Int_t pat = tds.getNPatterns();
    //CPPUNIT_ASSERT_EQUAL(0, pat % STOP);
    // attributes
    // zone1, zone2, ..., zone7, contrast, rgpareto, generation
    CPPUNIT_ASSERT_EQUAL(8, tds.getNAttributes());
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("zone2"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("zone4"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("contrast"));
    CPPUNIT_ASSERT_EQUAL(kFALSE, tds.isAttribute("rgpareto"));
    // values
    // theo 1, 1, 1, 1, 1, 1, 0
    //CPPUNIT_ASSERT_DOUBLES_EQUAL(85., tds.getValue("contrast", pat-1), prec);
    //CPPUNIT_ASSERT_DOUBLES_EQUAL(85., tds.getValue("contrast", STEP), prec);
    //CPPUNIT_ASSERT_DOUBLES_EQUAL(85., tds.getValue("contrast", STOP-1), prec);

}

void TTangramTest::validateTangram3(TDataServer &tds, double prec)
{
    Int_t pat = tds.getNPatterns();
    CPPUNIT_ASSERT_EQUAL(5, pat);
    // attributes
    // zone1, zone2, ..., zone7, contrast
    CPPUNIT_ASSERT_EQUAL(8, tds.getNAttributes());
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("zone1"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("zone5"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("contrast"));
    CPPUNIT_ASSERT_EQUAL(kFALSE, tds.isAttribute("rgpareto"));
    // values
    // theo 1, 1, 1, 1, 1, 1, 0
    CPPUNIT_ASSERT_DOUBLES_EQUAL(85., tds.getValue("contrast", 0), prec);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(85., tds.getValue("contrast", 2), prec);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(85., tds.getValue("contrast", 4), prec);

}

void TTangramTest::validateBi(TDataServer &tds)
{
    Int_t pat = tds.getNPatterns();
    CPPUNIT_ASSERT_EQUAL(STOP, pat);
    // attributes
    // plus, moins, ln2dist, rgpareto, generation
    CPPUNIT_ASSERT_EQUAL(5, tds.getNAttributes());
    //CPPUNIT_ASSERT_EQUAL(3, tds.getNAttributes());
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("plus"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("moins"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("ln2_dist"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("generation"));
    // values
    // theo 1, 1, 1, 1, 1, 1, 0
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0., tds.getValue("ln2_dist", 1), DBL_PREC_STD);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0., tds.getValue("ln2_dist", STEP),
            DBL_PREC_STD);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0., tds.getValue("ln2_dist", STOP-1),
            DBL_PREC_STD);

}

void TTangramTest::validateRobustBi(TDataServer &tds)
{
    int i;
    Int_t pat = tds.getNPatterns();
    CPPUNIT_ASSERT_EQUAL(STOP2, pat);
    // attributes
    // plus, moins, ln2dist, rgpareto, generation
    CPPUNIT_ASSERT_EQUAL(7, tds.getNAttributes());
    //CPPUNIT_ASSERT_EQUAL(3, tds.getNAttributes());
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("x"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("z"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("dist"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("generation"));
    // values
    // theo 1, 1, 1, 1, 1, 1, 0
    for (i=0; i < STOP2 && tds.getValue("cont", i) != 10.; i++);
    if (i == STOP2) return ;
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.6, tds.getValue("x", i), DBL_PREC_ROBI);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.5, tds.getValue("y", i), DBL_PREC_ROBI);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.8, tds.getValue("z", i), DBL_PREC_ROBI);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.46, tds.getValue("dist", i), DBL_PREC_ROBO);

}

