/* --- Ansi C --- */
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2009-2024 CEA/DES
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   aco_core.c
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2009-10-21
 
 \brief  recherche par colonies de fourmis
*/

#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>

#include "aco_core.h"

#include "vizir/vzrcore_api.h"
#include "vizir/vzrbase_api.h"
#include "vizir/vzrfifo_api.h"
#include "vizir/vzrpop_api.h"
#include "vizir/vzritem.h"
#include "vizir/vzr_macro.h"

#define FREQ_GLOB 8

struct acowork {
    struct aco *self;
};

/* ajout d'un chemin
 * */
static void aco_add(struct aco *self, struct vzritem *item) 
{
    int i, ret;
    int mem;
    struct vzrpop *pop = &self->base.pop;

    self->base.eval++;
    if (vzrpop_viable(pop, item)) {
        /* individual update */
        if (    pop->level &&
                item->rank <= pop->item[pop->rg[pop->level/2]].rank
            )
            ; //PRO_CALL(self->work, depose, self, item->zone, item->rank);
        if (vzrpop_add(pop, item)) {
            /* generation update */
            ret = pop->level+pop->step-1;
            mem = pop->item[pop->rg[ret]].rank;
            for FROMEACH(i, pop->level, ret)
                if (pop->item[pop->rg[i]].rank < mem)
                    mem = pop->item[pop->rg[i]].rank;
            for FROMEACH(i, pop->level, ret+1)
                if (pop->item[pop->rg[i]].rank == mem)
                    PRO_CALL(self->work, depose, self
                            , pop->item[pop->rg[i]].zone
                            , pop->item[pop->rg[i]].rank);
            /* population update */
            vzrpop_cycle(pop);
            if (!self->base.quiet) 
                printf("%s  Generation : %d, elite %d rang max %d\n"
                        , self->base.id, pop->cycle
                        , pop->elite
                        , pop->item[pop->rg[pop->level-1]].rank);
            /* global update */
            if (!(pop->cycle % FREQ_GLOB)) {
                for (i=0; i<pop->elite; i++) {
                    PRO_CALL(self->work, depose, self
                            , pop->item[pop->rg[i]].zone
                            , pop->item[pop->rg[i]].rank );
                }
            }
        }
    }

}

/* protocole
 * */
static void acowork_subscribe(void *core)
{
    struct aco *self = ((struct acowork *) core)->self;

    pthread_mutex_lock(&(self->base.lock));
    (void) vzrbase_first(&(self->base));

    if (!self->base.quiet) 
        printf("Ants    %s %d\n", self->base.id, self->base.slave);
    pthread_mutex_unlock(&(self->base.lock));
}

static void acowork_unsubscribe(void *core)
{
    struct aco *self = ((struct acowork *) core)->self;

    pthread_mutex_lock(&(self->base.lock));
    if (self->base.slave == 1) {
        printf("Ants %s converge %d\n", self->base.id, self->base.eval);
    }
    self->base.slave--;
    pthread_mutex_unlock(&(self->base.lock));
}

static int acowork_get(void *core, vzrgene_t *in)
{
    int ret;
    struct aco *self = ((struct acowork *) core)->self;

    pthread_mutex_lock(&(self->base.lock));
    if (vzrbase_end(&(self->base)) ) ret = 1;
    else {
        if (self->base.pop.again) {
            struct vzritem item;

            item.flag = APLOID;
            item.zone = in;
            item.front = NULL;
            item.data = NULL;
            vzrpop_again(&(self->base.pop), &item);
        }
        /* parcours */
        else if (self->base.migr && vzrfifo_get(self->base.migr, in)) {}
        else {
            while (! PRO_CALL(self->work, build, self, in)) ;
        }
        ret = 0;
    }
    pthread_mutex_unlock(&(self->base.lock));
    return ret;
}

static int acowork_put(void *core, int good, vzrgene_t *in, double *out, void *dat)
{
    struct aco *self = ((struct acowork *) core)->self;
    struct vzritem item;
    int ret;

    item.flag = APLOID;
    item.group = 0;
    item.front = out;
    item.zone = in;
    item.data = dat;

    /* integration individu */
    if (good) {
        pthread_mutex_lock(&(self->base.lock));
        aco_add(self, &item);
    }
    /* test convergence */
    ret = vzrpop_end(&(self->base.pop));
    if (good) pthread_mutex_unlock(&(self->base.lock));
    return ret;
}

static void acowork_dealloc(struct vzrcore *core)
{
    free(core->data);
    free(core);
}

struct vzrcore *vzrcore_aco(struct aco *self)
{
    struct vzrcore  *core;
    struct acowork *dat;
    
    /* data */
    dat = (struct acowork *) malloc(sizeof(struct acowork));
    dat->self = self;

    /* interface */
    core = (struct vzrcore *) malloc(sizeof(struct vzrcore));
    core->data = (void *) dat;
    core->subscribe = &acowork_subscribe;
    core->get = &acowork_get;
    core->put = &acowork_put;
    core->unsubscribe = &acowork_unsubscribe;
    core->dealloc = &acowork_dealloc;

    return core;
}

/* mainloop
 * */
void aco_run(struct aco *self)
{
    vzrcore_run(vzrcore_aco(self));
}

void aco_start(struct aco *self)
{
    vzrcore_start(vzrcore_aco(self));
}

void aco_wait(struct aco *self)
{
    vzrbase_wait(&(self->base));
}

int aco_running(struct aco *self)
{
    return vzrbase_running(&(self->base));
}
