/* --- Ansi C --- */
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2009-2024 CEA/DES
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   aco_query.c
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2009-10-21
 
 \brief  ants user's interface
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "aco_query.h"

#include "vizir/vzr_alloc.h"
#include "vizir/vzr_aleas.h"
#include "vizir/vzrbase_api.h"
#include "vizir/vzrfifo_api.h"
#include "vizir/vzrpop_api.h"
#include "vizir/vzritem.h"

#define FOPT 1

#define EVAPOR_DFT  0.96
#define DEPOS_DFT   0.95
#define PBEST_DFT   0.05
#define PREC_DFT    0.01

struct aco *aco_init(int nbin, int nbout, int nbcont, double prec)
{
    struct aco *ret;

    ret = (struct aco *) malloc(sizeof(struct aco));
    vzrpop_init(&(ret->base.pop), nbin, nbcont, nbout, prec);
    /* TODO in acorun ? */
    ret->evap = EVAPOR_DFT;
    ret->depo = DEPOS_DFT;
    ret->pbest = PBEST_DFT;
    ret->tmax = 1 / ((1 - ret->evap) * FOPT);

    /* working value */
    mt_init(0);
    vzrbase_init(&(ret->base));
    ret->work = NULL; ret->base.migr = NULL;

    return ret;
}

int aco_prepare(struct aco *ret, int step, int stop, int eval)
{
    double tmp, cycle;

    /* working value */
    vzrbase_build(&(ret->base), eval);
    vzrpop_aploid_build(&(ret->base.pop), stop, step);

    cycle = ((double) ret->base.maxeval / ret->base.pop.step);
    tmp = pow(ret->pbest, (1./cycle));
    ret->tmin = ret->tmax * (1 - tmp) / ((cycle/2 -1) * tmp);
    printf("%s tmax %g tmin %g\n", ret->base.id, ret->tmax, ret->tmin);
    PRO_CALL(ret->work, prepare, ret);
    
    return 1;
}

int aco_build(struct aco *ret, int ant, int cyc, int stop)
{
    printf("aco_build is deprecated, use aco_prepare instead\n");
    return aco_prepare(ret, ant, stop, cyc);
}

int aco_resume(struct aco *ret, int maxeval, char *nfile)
{
    double tmp, cycle;

    /* working value */
    vzrbase_build(&(ret->base), maxeval);
    vzrpop_resume(&(ret->base.pop), APLOID, nfile);

    cycle = ((double) ret->base.maxeval / ret->base.pop.step);
    tmp = pow(ret->pbest, (1./cycle));
    ret->tmin = ret->tmax * (1 - tmp) / ((cycle/2 -1) * tmp);
    printf("%s tmax %g tmin %g\n", ret->base.id, ret->tmax, ret->tmin);
    
    return 1;
}

void aco_dump(struct aco *self, char *nfile)
{
    vzrpop_dump(&(self->base.pop), nfile);
}

void aco_dealloc(struct aco *self)
{
    vzrbase_dealloc(&(self->base));
    vzrpop_aploid_dealloc(&(self->base.pop));
    PRO_DEALLOC(self->work);
    free(self);
}

