/* --- Ansi C --- */
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2003-2024 CEA/DES
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   vzr_core.c
 \author Michel DUMAS
 \date   2003-12-11
 
 \brief  ALGORITHME GENETIQUE DIPLOIDE A DOMINANCE VRAIE 
*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <sys/types.h>
#ifndef WIN32
#include <unistd.h>
#else
#include <io.h>
#include <process.h>
#endif

#include "vzr_core.h"

#include "vzr_util.h"

/**** Prototype ****/

#include "vizir/vzrcore_api.h"
#include "vizir/vzrbase_api.h"
#include "vizir/vzrfifo_api.h"
#include "vizir/vzrpop_api.h"
#include "vizir/vzritem.h"
#include "vzr_naitre.h"
#include "vzr_croiser.h"

/*** protocole ***/
static int vzrwork_get(void *core, vzrgene_t *in)
{
    struct vzrwork *wself = (struct vzrwork *) core;
    struct vizir *self = wself->self;

    wself->child.dip.domin = in;
    pthread_mutex_lock(&(self->base.lock));
    /* test arret non converg */
    if (vzrbase_end(&(self->base)) ) {
        pthread_mutex_unlock(&(self->base.lock));
        return 1;
    }
    else {
        pthread_mutex_unlock(&(self->base.lock));

        /* nouvel individu */
        if (reevaluation(self, wself)) {}
        else if (immigration(self, wself)) {}
        else if (self->base.pop.level == self->base.pop.stop) {
            /* croisement */
            PRO_CALL(self->eros, couple, self, wself);
            procreer(self, wself);
        }
        else {
            /* generation aleatoire */
            PRO_CALL(wself->oper, sample, &(wself->child.dip));
        }
        return 0;
    }
    
}

static int vzrwork_put(void *core, int good, vzrgene_t *in, double *out, void *dat)
{
    struct vzrwork *wself = (struct vzrwork *) core;
    struct vizir *self = wself->self;
    int ret;

    ret = 0;
    wself->child.dip.domin = in;
    wself->child.front = out;
    wself->child.data = dat;
    pthread_mutex_lock(&(self->base.lock));
    self->base.eval++;
    if (good) {
        if( vzrpop_viable(&(self->base.pop), &(wself->child))) {
            if (vzrpop_add(&(self->base.pop), &(wself->child))) {
                vzrpop_cycle(&(self->base.pop));
                if (! self->base.quiet) {
                    printf("\n%s  Generation : %d, rang max %d\n"
                        , self->base.id, self->base.pop.cycle
                        , self->base.pop.item[self->base.pop.rg[self->base.pop.level-1]].rank);
                    printf("%s\tNb d'evaluation : %d, taille de la Z.P. :  %d\n"
                        , self->base.id, self->base.eval,self->base.pop.elite);
                }
                if (self->generation) PRO_CALL(self->generation, react, self);
            }
        }
        else if (self->base.pop.dealloc_data && dat) {
            self->base.pop.dealloc_data(dat);
        }
    }
    
    ret = vzrpop_end(&(self->base.pop));
    pthread_mutex_unlock(&(self->base.lock));
    return ret;
}

static void vzrwork_in(void *core)
{
    struct vzrwork *wself = (struct vzrwork *) core;
    struct vizir *self = wself->self;

    pthread_mutex_lock(&(self->base.lock));
    if ( vzrbase_first(&(self->base)) ) {
        if (!self->base.quiet) printf("first %d\n", self->base.pop.cur);
    }

    if (!self->base.quiet) 
        printf("Genetic %s %d\n", self->base.id, self->base.slave);
    pthread_mutex_unlock(&(self->base.lock));
}

static void vzrwork_out(void *core)
{
    struct vzrwork *wself = (struct vzrwork *) core;
    struct vizir *self = wself->self;

    pthread_mutex_lock(&(self->base.lock));
    if (self->base.slave == 1) {
        if (self->base.stop) {
            printf("%s  ! ! ! ! !   La convergence n'a pas ete atteinte  dans le temps imparti  ! ! ! ! \n", self->base.id);
        }
        if (self->base.pop.stop > self->base.pop.elite) {
            printf("%s  ! ! ! ! !  La convergence n'a pas ete atteinte avec le nb max d'evaluations ! ! ! ! \n", self->base.id);
        }
        printf("Genetic %s converge %d\n", self->base.id, self->base.eval);
    }
    self->base.slave--;

    pthread_mutex_unlock(&(self->base.lock));
}

static void vzrwork_dealloc(struct vzrcore *self)
{
    struct vzrwork *wself = (struct vzrwork *) self->data;

    PRO_DEALLOC(wself->oper);
    free(wself->pere.domin);
    free(wself);
    free(self);
}

/*** creation ***/
struct vzrcore *vzrcore_vzr(struct vizir *self)
{
    struct vzrwork  *wself;
    struct vzrcore  *core;
    vzrgene_t       *tab;

    /* data */
    wself = (struct vzrwork *) malloc(sizeof(struct vzrwork));
    tab = (vzrgene_t *) calloc(5*self->base.pop.gene, sizeof(vzrgene_t));
    wself->self = self;
    //wself->save = fsav;
    wself->pere.domin = tab;
    wself->pere.reces = tab+self->base.pop.gene;
    wself->mere.domin = tab+2*self->base.pop.gene;
    wself->mere.reces = tab+3*self->base.pop.gene;
    //wself->descendant[0] = &tab[4*self->base.pop.gene];
    wself->child.flag = DIPLOID;
    wself->child.group = 0;
    wself->child.data = NULL;
    wself->child.dip.reces = tab+4*self->base.pop.gene;
    //wself->val = (double *) calloc(self->nbcrit, sizeof(double));
    wself->oper = PRO_CREATE(self->foper);

    /* interface */
    core = (struct vzrcore *) malloc(sizeof(struct vzrcore));
    core->data = (void *) wself;
    core->subscribe = &vzrwork_in;
    core->get = &vzrwork_get;
    core->put = &vzrwork_put;
    core->unsubscribe = &vzrwork_out;
    core->dealloc = &vzrwork_dealloc;

    return core;

}


