/* --- Ansi C --- */
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2008-2024 CEA/DES
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   vzr_croiser.c
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2008-03-17
 
 \brief  operateur stochastique standart de croisement
*/

#include <stdlib.h>

#include "vzr_croiser.h"

#include "genetic/vzr_struct.h"

#include "vizir/vzr_aleas.h"

//#define MICHEL

#ifndef COMBINATOIRE

#define XPOL_DEC  0.2
#define XPOL_COEF 1.4 // 1 + 2*XPOL_DEC

struct cross {
    int     nballgene, nbdiscgene;
};

static void creer(void *truc, vzrdiploid_t fils)
{
    int i;
    struct cross *self = (struct cross *) truc;
    
    for (i=0;i<self->nballgene;i++) {
        fils->domin[i] = mt_real1();
        fils->reces[i] = mt_real1();
    }
}

static void allele(void *truc, vzrdiploid_t parent, double *fils)
{
    int i;
    struct cross *self = (struct cross *) truc;

    for(i=0;i<self->nballgene;i++) {
        fils[i] = (mt_real1() > 0.5) ? parent->domin[i] : parent->reces[i];
    }
}

static void mutation(void * truc, vzrdiploid_t fils)
{
    struct cross *self = (struct cross *) truc;

    fils->domin[mt_int(self->nballgene)] = mt_real1();
    fils->reces[mt_int(self->nballgene)] = mt_real1();
}

static void heterozygote(void *truc, vzrdiploid_t fils)
{
    int i;
    double temp;
    struct cross *self = (struct cross *) truc;

    /* on construit le phenotype */
    for (i=0;i<self->nballgene;i++)
        if(mt_real1() >= 0.5){
            temp = fils->domin[i] ;
            fils->domin[i] = fils->reces[i];
            fils->reces[i] = temp;
        }

}

static void homozygote(void *truc, vzrdiploid_t fils)
{
    int i;
    double temp;
    struct cross *self = (struct cross *) truc;

    for(i=0;i<self->nballgene-self->nbdiscgene;i++){
        temp = XPOL_COEF * mt_real1() - XPOL_DEC;
        temp = temp * fils->domin[i]+ (1-temp)*fils->reces[i];
        temp = (temp > 1.)? 1. : temp;
        temp = (temp < 0.)? 0. : temp;
        fils->domin[i] = fils->reces[i]= temp;
    }
    for(i=self->nballgene-self->nbdiscgene;i<self->nballgene;i++){	
        if (mt_real1() >= 0.5)
            fils->domin[i] = fils->reces[i];
        else
            fils->reces[i] = fils->domin[i];
    }
}

static void dealloc(struct vzrcrossover *cross)
{
    free(cross);
}

static struct vzrcrossover *vzrcrossstd_objet(void*info)
{
    struct vzrcrossover *ret;
    
    ret = (struct vzrcrossover *) malloc(sizeof(struct vzrcrossover));
    ret->data = info;
    ret->sample = &creer;
    ret->allele = &allele;
    ret->mutation = &mutation;
    ret->homozygote = &homozygote;
    ret->heterozygote = &heterozygote;
    ret->dealloc = &dealloc;

    return ret;
}

static void vzrcrossstd_clean(struct vzrcrossoverbuild *self)
{
    free(self->data);
    free(self);
}

struct vzrcrossoverbuild *vzrcrossstd_classe(int nball, int nbdisc)
{
    struct cross *dat;
    struct vzrcrossoverbuild *ret;

    dat = (struct cross *) malloc(sizeof(struct cross));

    dat->nballgene = nball;
    dat->nbdiscgene = nbdisc;

    ret = (struct vzrcrossoverbuild *) malloc(sizeof(struct vzrcrossoverbuild));
    ret->data = (void *) dat;
    ret->create = &vzrcrossstd_objet;
    ret->dealloc = &vzrcrossstd_clean;

    return ret;
}

#endif

