/* --- Ansi C --- */
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2003-2024 CEA/DES
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   vzr_query.c
 \author Michel DUMAS
 \date   2003-12-08
 
 \brief  initialisation et requete resultat pour vizir
*/

#include <stdio.h>
#include <math.h>
#include <time.h>
#include <stdlib.h>

#include "vzr_query.h"

#include "vzr_util.h"

#include "vizir/vzr_alloc.h"

#include "vizir/vzr_aleas.h"
#include "vizir/vzrfifo_api.h"
#include "vizir/vzrcore_api.h"
#include "vizir/vzrbase_api.h"
#include "vizir/vzrpop_api.h"
#include "vizir/vzragent_api.h"
#include "vizir/vzritem.h"

#include "vzr_core.h"
#include "vzr_croiser.h"
#include "vzr_naitre.h"
#include "vzr_mating.h"

/**** declaration externe ****/
/*** variables locales ***/
/**** corps des fonctions ****/
/** initialisation **/
/* creation struct */
struct vizir *vzr_init(int nb_gene, int nb_cout, int nb_cont, double precis) 
{
    struct vizir  *ret;
    
    /* allocation structure */
    ret = (struct vizir*) malloc(sizeof(struct vizir));

    /* definition du probleme */
    /* initialisation utilisateur */
    vzrpop_init(&(ret->base.pop), nb_gene, nb_cont, nb_cout, precis);

    /* initialisation defaut */
    ret->alpha = 0.4;
    ret->phomo = 0.5;
    ret->pmut = 0.01;
    ret->bias = 0.5;

    //pthread_mutex_init(&(ret->mutpop),( pthread_mutexattr_t *) NULL);
    mt_init(0);
    vzrbase_init(&(ret->base)); 
    ret->base.migr = NULL;

    /* initialisation zero */
    ret->genedisc = 0;
    ret->pmut1g = 0.0;
    ret->foper = (struct vzrcrossoverbuild *) NULL;
    ret->generation = (struct vzrevent *) NULL;
    ret->eros = vzrcouple_create();
    ret->mate = NULL;
	
    return ret;
}

/* modification des defauts */
void vzr_setLabel(struct vizir *vizir, char *name)
{
    vizir->base.id = name;
}

void vzr_setSurvivalRate(struct vizir *vizir, double val)
{   /* taux de survie */
    if (val > 0. && val < 1.) 
        vizir->alpha = val;
}

/*
void vzr_setDominanceMode(struct vizir *vizir, vzrpareto_t val)
{
        vizir->pop.darwin = val;
}
*/

void vzr_setHomozygoteRate(struct vizir *vizir, double val)
{   /* probabilite d'homozigotie */
    if (val > 0. && val < 1.) 
        vizir->phomo = val;
}

void vzr_setMutationRate(struct vizir *vizir, double val)
{   /* probabilite de mutation */
    if (val > 0. && val < 1.) 
        vizir->pmut = val;
}
/* EDF */
void vzr_setOneGeneMutationRate(struct vizir *vizir, double val)
{   /* probabilite de mutation sur un gene */
    if (val > 0. && val < 1.) 
        vizir->pmut1g = val;
}

void vzr_setCombinatorialSize(struct vizir *vizir, int val)
{   /* nombre de gene a traiter comme des variables discretes */
    if (val <= vizir->base.pop.gene) 
        vizir->genedisc = val;
    else 
        printf("ERREUR le nombre de genes discrets depassent le nombre total  de genes \n");
}

void vzr_setParentsBias(struct vizir *vizir, double val)
{   /* biaisage des naissances */
    if (val > 0. && val < 1.) 
        vizir->bias = val;
}

void vzr_setCrossover(struct vizir *vizir, struct vzrcrossoverbuild *val)
{
    vizir->foper = val;
}

/* fonction utilisateur d'valuation */
void vzr_setFitness(struct vizir *vizir, struct vzrfitnessbuild *val)
{
    vizir->base.fit = val;
}

/* dimmentionnement du problme */
static int vzr_build_common(struct vizir *vizir)
{
    /* operateur croisement */
    if (vizir->foper == NULL) {
#ifdef COMBINATOIRE
        printf("unseted Crossing operator\n");
        return 0;
#else
        vizir->foper = vzrcrossstd_classe(vizir->base.pop.gene, vizir->genedisc);
#endif
    }
    /* operateur selection parent */
    if (vizir->mate) {
        if (vizir->mate->genre != None && vizir->mate->genre != vizir->base.pop.genre) {
            printf("parental selection dismatch env selection info\n");
            return 0;
        }
    }
    else vizir->mate = &vzrcoupling_std;
    
    return 1;
}

int vzr_prepare(struct vizir *vizir, int step, int stop, int max_eval)
{
    vizir->consang = (int) floor(vizir->alpha * stop);
    
    vzrbase_build(&(vizir->base), max_eval);
    vzrpop_diploid_build(&(vizir->base.pop), stop, step);

    return vzr_build_common(vizir);
}

int vzr_build(struct vizir *ret, int cyc, int stop, int step)
{
    printf("vzr_build is deprecated, use vzr_prepare instead\n");
    return vzr_prepare(ret, step, stop, cyc);
}

int vzr_resume(struct vizir *vizir, int max_eval, char *nfile)
{
    
    vzrbase_build(&(vizir->base), max_eval);
    vzrpop_resume(&(vizir->base.pop), DIPLOID, nfile);

    vizir->consang = (int) floor(vizir->alpha * vizir->base.pop.stop);

    return vzr_build_common(vizir);
}

/* initialisation specifique (pour interface C) */
struct vizir *vzr_multi(int nb_gene, int nb_cout)
{
    return vzr_init(nb_gene, nb_cout, 0, 0.1);
}

struct vizir *vzr_mono(int nb_gene, double precis)
{
    return vzr_init(nb_gene, 1, 0, precis);
}

struct vizir *vzr_monoConstraint(int nb_gene, double precis, int nb_cont)
{
    return vzr_init(nb_gene, 1, nb_cont, precis);

}

struct vizir *vzr_multiConstraint(int nb_gene, int nb_cout, int nb_cont)
{
    return vzr_init(nb_gene, nb_cout, nb_cont, 0.1);
}

/** execution **/
void vzr_run(struct vizir *self)
{
    vzrcore_run(vzrcore_vzr(self));
}

/* threaded execution */
void vzr_start(struct vizir *self)
{
    vzrcore_start(vzrcore_vzr(self));
}

int vzr_running(struct vizir *vizir)
{
    return vzrbase_running(&(vizir->base));
#ifdef OLD
#ifdef DEBUG
    static  int echo = 1;
#endif
    int     ret;
    pthread_mutex_lock(&(vizir->base.lock));
    ret = (vizir->bgstate & VZRRUN_END) ? 0 : 1 ;
#ifdef DEBUG
    if (echo && !ret) {
        //echo = 0;
        printf("%s not running\n", vizir->id);
    }
#endif
    pthread_mutex_unlock(&(vizir->base.lock));
    return ret;
#endif
}

void vzr_stop(struct vizir *vizir) 
{
    vzrbase_stop(&(vizir->base));
}

void vzr_wait(struct vizir *vizir)
{
    vzrbase_wait(&(vizir->base));
}

void vzr_reset(struct vizir *viz)
{
    /* netoyage pour recommencer a zero */
    viz->base.slave = -1;
}

void vzr_reevaluate(struct vizir *viz)
{
    if (viz->base.slave < 1) vzrpop_set_again(&(viz->base.pop));
    else printf("reevaluate needs solver to be stopped\n");
}

void vzr_dump(struct vizir *self, char *nfile)
{
    vzrpop_dump(&(self->base.pop), nfile);
}

/** resultat **/
/* population */
vzrdiploid_t vzr_genotype(struct vizir *vizir, int i)
{
    struct vzrpop *pop = &(vizir->base.pop);
    return &(pop->item[pop->rg[i]].dip);
}

vzrgene_t *vzr_phenotype(struct vizir *vizir, int i)
{
    struct vzrpop *pop = &(vizir->base.pop);
    return pop->item[pop->rg[i]].dip.domin;
}

/* critere */
double *vzr_objective(struct vizir *vizir, int i)
{
    struct vzrpop *pop = &(vizir->base.pop);
    return pop->item[pop->rg[i]].front;
}

double vzr_rank(struct vizir *vizir, int i)
{
    struct vzrpop *pop = &(vizir->base.pop);
    return pop->item[pop->rg[i]].rank;
}

/* runtime infos */
int vzr_birthCount(struct vizir *viz)
{
    return viz->base.eval;
}

int vzr_generationCount(struct vizir *viz)
{
    return viz->base.pop.cycle;
}
    
int vzr_survivorSize(struct vizir *viz)
{
    return viz->base.pop.level;
}

int vzr_populationSize(struct vizir *viz)
{
    return viz->base.pop.level;
}

/* destructeur */
void vzr_dealloc(struct vizir *viz)
{
    if (viz->generation) PRO_DEALLOC(viz->generation);
    vzrbase_dealloc(&(viz->base));
    vzrpop_aploid_dealloc(&(viz->base.pop));
    PRO_DEALLOC(viz->foper);
    PRO_DEALLOC(viz->eros);
    free(viz);
}
