/* --- Ansi C --- */
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2010-2024 CEA/DES
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   vzr_aleas.c
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2010-03-15
 
 \brief  sur couche thread-safe au generateur de nombre aleatoire mt19937
*/

#include "vzr_aleas.h"

#include <pthread.h> 
#include <time.h>
#include <sys/types.h>
#include <math.h>
#ifndef WIN32
#include <unistd.h>
#define GETPID getpid
#else
#include <io.h>
#include <process.h>
#define GETPID _getpid
#endif

/* vzr_mt19937ar.h */
extern void MT_init_by_array(unsigned long [], int);
extern unsigned long MT_genrand_int32(void);

/* local lock for statci variable */
static pthread_mutex_t mtmutex;

void mt_init(int seeded)
{
    static long done=0;
    unsigned long vec[]={
        3463657, 24476403, 22365503, 7750391,
        25420944, 26966130, 955805, 20199590
    };

    if (! done++) {/* not really thread-safe */
        pthread_mutex_init(&mtmutex,( pthread_mutexattr_t *) NULL);
        pthread_mutex_lock(&mtmutex);
        if (! seeded) {
            vec[3] = (unsigned long) GETPID();
            vec[7] = (unsigned long) time(NULL);
        }
        MT_init_by_array(vec, 8);
        pthread_mutex_unlock(&mtmutex);
    }
    else if (seeded) {
        pthread_mutex_lock(&mtmutex);
        MT_init_by_array(vec, 8);
        pthread_mutex_unlock(&mtmutex);
    }
}

/* réécriture thread-safe des fonctions standarts de mt19937
 * (en supposant que tous les appels passe par cette interface)
 * */
/* generates a random number on [0,1]-real-interval */
double mt_real1(void)
{
    double ret;
    
    pthread_mutex_lock(&mtmutex);
    ret = MT_genrand_int32()*(1.0/4294967295.0); 
    /* divided by 2^32-1 */ 
    pthread_mutex_unlock(&mtmutex);
    return ret;
}

/* generates a random number on [0,1)-real-interval */
double mt_real2(void)
{
    double ret;
    
    pthread_mutex_lock(&mtmutex);
    ret = MT_genrand_int32()*(1.0/4294967296.0); 
    /* divided by 2^32 */
    pthread_mutex_unlock(&mtmutex);
    return ret;
}

/* generates a random number on (0,1)-real-interval */
double mt_real3(void)
{
    double ret;
    
    pthread_mutex_lock(&mtmutex);
    ret = (((double)MT_genrand_int32()) + 0.5)*(1.0/4294967296.0); 
    /* divided by 2^32 */
    pthread_mutex_unlock(&mtmutex);
    return ret;
}

/* generates a random number on [0,1) with 53-bit resolution*/
double mt_res53(void) 
{ 
    unsigned long a, b;
    
    pthread_mutex_lock(&mtmutex);
    a = MT_genrand_int32()>>5;
    b = MT_genrand_int32()>>6; 
    pthread_mutex_unlock(&mtmutex);
    return (a*67108864.0+b)*(1.0/9007199254740992.0); 
} 
/* These real versions are due to Isaku Wada, 2002/01/09 added */

/* added version for vizir */
/* generates a random number on [0,max)-interval */
long mt_int(long max)
{
    long ret;
    
    pthread_mutex_lock(&mtmutex);
    ret = (long) ( MT_genrand_int32()*( ((double)max)/4294967296.0) );
    pthread_mutex_unlock(&mtmutex);
    return ret;
}

/* generate a random gaussian number (0,1) */
double mt_gauss(void)
{
    static double next;
    static int flag=0;
    double d, fac, ret;

    pthread_mutex_lock(&mtmutex);
    if (flag) {
        flag = 0;
        ret = next;
    }
    else {
        do {
            next = 2.0 *((((double)MT_genrand_int32()) + 0.5)*(1.0/4294967296.0)) - 1.0;
            ret = 2.0 *((((double)MT_genrand_int32()) + 0.5)*(1.0/4294967296.0)) - 1.0;
            d = next*next + ret*ret;
        } while (d>= 1.0 || d == 0.0);
        fac = sqrt(-2.0*log(d)/d);
        next *= fac;
        ret *= fac;
        flag = 1;
    }
    pthread_mutex_unlock(&mtmutex);
    return ret;
}

