/* --- Ansi C --- */
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   vzr_pair.c
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2014-11-06
 
 \brief  distance inter individu et appairage
*/

#include <stdio.h>
#include <stdlib.h>

#include "vizir/vzr_pair.h"
#include "vizir/vzrsort.h"

#include "vizir/vzr_macro.h"
#include "vizir/vzr_alloc.h"

#include "vizir/vzr_aleas.h"
#include "vizir/vzr_survie.h"
#include "vizir/vzr_moga.h"
#include "vizir/vzr_heap.h"

/* default value */
#define NEIG_PROF 8
#define PAIR_PROF 16
#define P_SEUIL 20
#define P_RATIO 50
#define RAND

/* flag */ 
#define FREE 0
#define ALLOC 1
#define DEAD 2

#define TRACE
#define ASSERT
//#define DEBUG


/* sorting local function */
struct heap_comp {
    struct pair *val;
};

static INLINE int hs_comp(struct heap_comp *hs, int i, int j)
{
    return (hs->val[i].cost < hs->val[j].cost);
}

#include "vizir/vzr_heapsort.h"

int vzr_rank_pair(int size, int top, int *idx, struct pair *val)
{
    struct heap_comp comp = {val};
    return hs_rank(idx, size, top, &comp);
}

/* utilitaires */
void pair_affectation(struct voisin *knn, struct local *self, int siz, struct vzrpop *pop)
{
    int i, j, k, kmem, ii, jm;
    char toggle;
    VZRTAB(char, mark, knn->size);
    VZRTAB(short, pair, knn->size);
    
    for EACH(i, knn->size) {
        mark[i] = DEAD;
        pair[i] = -1;
    }
    for EACH(i, siz) mark[pop->rg[i]] = FREE;

    k = 0;
    do {
        kmem = k;
        /* recherche du plus proche voisin */
        for EACH(ii, siz) {
            i = pop->rg[ii];
            if (!mark[i]) {
                jm = 0;
                while (jm < knn->prof && mark[knn->ref[i][jm]]) jm++;
                pair[i] = (jm != knn->prof) ? jm : -1 ;
            }
        }
        for EACH(ii, siz) {
            i = pop->rg[ii];
            if (!mark[i] && pair[i] >= 0) {
                j = knn->ref[i][pair[i]];
                if (j < i && !mark[j] && pair[j] >= 0 && knn->ref[j][pair[j]] == i) {
                    self->fit[k].one = j;
                    self->fit[k].two = i;
                    self->fit[k].cost = knn->dist[i][pair[i]];
#ifdef TRACE2
                    printf("%d = %d\n", pair[i], i);
#endif
                    mark[i] = mark[j] = ALLOC;
                    k++;
                }
            }
        }
    } while ( k != kmem );
    self->done = k;
    VZRTAB_OUT(pair);
#ifdef TRACE2
    printf("pair %d\n", k);
#endif
    
    /* solution non appaire */
    toggle = 0;
    for EACH(ii, siz) {
        i = pop->rg[ii];
        if (!mark[i]) {
            if (toggle) {
                self->fit[k].two = i;
                toggle = 0;
                k++;
            }
            else {
                self->fit[k].one = i; 
                self->fit[k].cost = 1.234e10;
                toggle = 1;
            }
        }
    }
    VZRTAB_OUT(mark);
    
#ifdef TRACE2
    for EACH(i, (siz+1)/2) 
        if (i%6 == 5) printf("%d ~ %d;\n", self->fit[i].one, self->fit[i].two);
        else printf("%d ~ %d;", self->fit[i].one, self->fit[i].two);
    printf("\n"); 
#endif
}

/* indicateur d'isolement */
static void spea_isol(void *info, int size, struct vzrpop *pop)
{
    int i;
    struct voisin *self = (struct voisin *) pop->info;

    /*** recherche des plus proches voisins ***/
    vzrpop_proche_voisin(pop, size, self); 
    
    /* indicateur d'isolement */
    for EACH(i, size) pop->item[pop->rg[i]].isol = self->dist[pop->rg[i]][0];
    
}

static void pair_isol(void *info, int size, struct vzrpop *pop)
{
    int i;
    struct voisin *self = (struct voisin *) pop->info;

    /*** recherche des plus proches voisins ***/
    vzrpop_proche_voisin2(pop, size, self); 
    
    /* indicateur d'isolement */
    for EACH(i, size) pop->item[pop->rg[i]].isol = self->dist[pop->rg[i]][self->prof/2];
    
}

/* classement par similarité en V */
void pair_vorder(struct local *loc, int siz, struct vzrpop *pop, int *simil)
{
    int i, j, p, choix, toggle;
#ifdef DEBUG
    printf("INPAIR\n");
    valid_reorder(siz, pop->rg);
#endif
    /* classement des similarités */
    for EACH(p, loc->done) simil[p] = p;
    (void) vzr_rank_pair(loc->done, 0, simil, loc->fit);
    
    /* solutions appairees, par similarité en V */
    for EACH(p, loc->done) {
        //j = simil[p];
        i = loc->fit[simil[p]].one;
        j = loc->fit[simil[p]].two;
        if (pop->item[j].rank > pop->item[i].rank)
            choix = 0;
        else if (pop->item[j].rank < pop->item[i].rank)
            choix = 1;
        else {
#ifdef RAND            
            choix = (mt_real3() > 0.5);
#else
            int ic;
            double v, *norm = ((struct voisin *) pop->info)->norm;
            
            v = 0;
            for FROMEACH(ic, pop->valid, pop->cost+pop->valid) {
                v = (pop->item[i].front[ic+pop->valid] - pop->item[j].front[ic+pop->valid] )
                        / norm[ic+pop->valid];
            }
            choix = ( v > 0.0) ? 1 : 0 ; 
#endif
        }
        if (choix) {
            pop->rg[p] = j;
            pop->rg[siz-p-1] = i;
        } else {
            pop->rg[p] = i;
            pop->rg[siz-p-1] = j;
        }
    }

    /* solutions non appairees */
    toggle = 1;
    for FROMEACH(i, p, siz-loc->done) {
        if (toggle) {
            pop->rg[i] = loc->fit[p].one;
            toggle = 0;
        }
        else {
            pop->rg[i] = loc->fit[p].two;
            toggle = 1;
            p++;
        }
    }
#ifdef DEBUG
    printf("OUTPAIR\n");
    valid_reorder(siz, pop->rg);
#endif
#ifdef TRACE2
    printf ("size %d pair %d\n", siz, loc->done);
    for EACH(i,siz)
        if (i%10 == 9) printf("%d\n", pop->rg[i]);
        else printf("%d ", pop->rg[i]);
    printf("\n"); 
#endif
}

/* classement par similarité en V */
void modal_vorder(struct local *loc, int siz, struct vzrpop *pop, int *simil)
{
    int i, j, p, choix, toggle;
#ifdef DEBUG
    printf("INPAIR\n");
    valid_reorder(siz, pop->rg);
#endif
    /* classement des similarités */
    for EACH(p, loc->done) simil[p] = p;
    (void) vzr_rank_pair(loc->done, 0, simil, loc->fit);
    
    /* solutions appairees, par similarité en V */
    for EACH(p, loc->done) {
        //j = simil[p];
        i = loc->fit[simil[p]].one;
        j = loc->fit[simil[p]].two;
        if (pop->item[j].front[pop->valid] < pop->item[i].front[pop->valid]) {
            pop->rg[p] = j;
            pop->rg[siz-p-1] = i;
        } else {
            pop->rg[p] = i;
            pop->rg[siz-p-1] = j;
        }
    }

    /* solutions non appairees */
    toggle = 1;
    for FROMEACH(i, p, siz-loc->done) {
        if (toggle) {
            pop->rg[i] = loc->fit[p].one;
            toggle = 0;
        }
        else {
            pop->rg[i] = loc->fit[p].two;
            toggle = 1;
            p++;
        }
    }
#ifdef DEBUG
    printf("OUTPAIR\n");
    valid_reorder(siz, pop->rg);
#endif
#ifdef TRACE2
    printf ("size %d pair %d\n", siz, loc->done);
    for EACH(i,siz)
        if (i%10 == 9) printf("%d\n", pop->rg[i]);
        else printf("%d ", pop->rg[i]);
    printf("\n"); 
#endif
}


/* build dealloc */
static void build_neighbor(void *info, struct vzrpop *pop)
{
    struct voisin *dat;
    int i;
    
    dat = (struct voisin *) malloc(sizeof(struct voisin));
    dat->prof = *((int *) info);
    dat->size = pop->stop + pop->step;
    NEWvecteur(double, dat->norm, pop->all);
    NEWmatrice(int, dat->ref, dat->size, dat->prof);
    NEWmatrice(double, dat->dist, dat->size, dat->prof);
    
    pop->genre = Knn;
    pop->info = dat;
}

static void build_pair(void *info, struct vzrpop *pop)
{
    struct local *dat = (struct local *) info;
    struct voisin *knn;
    int i, siz;
    
    knn = (struct voisin *) malloc(sizeof(struct voisin));
    knn->prof = dat->prof;
    knn->size = pop->stop + pop->step;
    NEWvecteur(double, knn->norm, pop->all);
    NEWmatrice(int, knn->ref, knn->size, knn->prof);
    NEWmatrice(double, knn->dist, knn->size, knn->prof);
    
    siz = (knn->size+1)/2;
    NEWvecteur(struct pair, dat->fit, siz);
    
    pop->genre = Knn;
    pop->info = knn;
}

static void spea_dealloc(struct vzrsort *self, struct vzrpop *pop)
{
    struct voisin *knn = (struct voisin *) pop->info;
    
    if (knn->norm) {
        DELvecteur(knn->norm);
        DELmatrice(knn->ref);
        DELmatrice(knn->dist);
    }
    free(knn);
    free(self->data);
    free(self);
}

static void pair_dealloc(struct vzrsort *self, struct vzrpop *pop)
{
    struct local *dat = (struct local *) self->data;
    struct voisin *knn = (struct voisin *) pop->info;
    
    if (knn->norm) {
        DELvecteur(knn->norm);
        DELmatrice(knn->ref);
        DELmatrice(knn->dist);
        DELvecteur(dat->fit);
    }
    free(knn);
    free(dat);
    free(self);
}

/* constructeur */
struct vzrsort *create_neighbor_full(int prof)
{
    struct vzrsort *ret;
    int *dat;

    dat = (int *) malloc(sizeof(int));
    *dat = prof;

    ret = (struct vzrsort *) malloc(sizeof(struct vzrsort));
    ret->data = (void *) dat;
    ret->build = &build_neighbor;
    ret->valid = &moga_viable;
    ret->rank = &pareto_egal_rank;
    ret->isol = &spea_isol;
    ret->survive = &pareto_survive;
    ret->dealloc = &spea_dealloc;
    return ret;
}

struct vzrsort *create_neighbor()
{
    return create_neighbor_full(NEIG_PROF);
}

struct vzrsort *create_pair_full(int prof, int seuil, int ratio)
{
    struct vzrsort *ret;
    struct local *dat;

    dat = (struct local *) malloc(sizeof(struct local));
    dat->prof = prof;
    dat->seuil = seuil;
    dat->ratio = ratio;
    dat->cut = 0.0;

    dat->fit = NULL;

    ret = (struct vzrsort *) malloc(sizeof(struct vzrsort));
    ret->data = (void *) dat;
    ret->build = &build_pair;
    ret->valid = &moga_viable;
    ret->rank = &pareto_egal_rank;
    ret->isol = &pair_isol;
    ret->survive = &pair_survive;
    ret->dealloc = &pair_dealloc;
    return ret;
}

struct vzrsort *create_pair()
{
    return create_pair_full(PAIR_PROF, P_SEUIL, P_RATIO);
}

struct vzrsort *create_spea_full(int prof)
{
    struct vzrsort *ret;
    int *dat;

    dat = (int *) malloc(sizeof(int));
    *dat = prof;

    ret = (struct vzrsort *) malloc(sizeof(struct vzrsort));
    ret->data = (void *) dat;
    ret->build = &build_neighbor;
    ret->valid = &spea_viable;
    ret->rank = &spea_rank;
    ret->isol = &spea_isol;
    ret->survive = &pareto_survive;
    ret->dealloc = &spea_dealloc;
    return ret;
}

struct vzrsort *create_spea()
{
    return create_spea_full(NEIG_PROF);
}

struct vzrsort *create_spea_pair_full(int prof, int seuil, int ratio)
{
    struct vzrsort *ret;
    struct local *dat;

    dat = (struct local *) malloc(sizeof(struct local));
    dat->prof = prof;
    dat->seuil = seuil;
    dat->ratio = ratio;
    dat->cut = 0.0;

    dat->fit = NULL;

    ret = (struct vzrsort *) malloc(sizeof(struct vzrsort));
    ret->data = (void *) dat;
    ret->build = &build_pair;
    ret->valid = &spea_viable;
    ret->rank = &spea_rank;
    ret->isol = &pair_isol;
    ret->survive = &pair_survive;
    ret->dealloc = &pair_dealloc;
    return ret;
}

struct vzrsort *create_spea_pair()
{
    return create_spea_pair_full(PAIR_PROF, P_SEUIL, P_RATIO); 
}

