/* --- Ansi C --- */
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   vzrpop_api.h
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2013-09-24
 
 \brief  gestion des populations
*/

#ifndef __VZRPOP_API__
#define __VZRPOP_API__

#include "vizir/vzr_config.h"
#include "vizir/vzritem.h"

/* predef */
struct vzrpop;
struct vzrsort;

/* protocole */
typedef int (*vzrpareto_t) (struct vzrpop *, double*, double*);

/* choix */
enum popinfo {
    None, Ibea, Knn, Ddknn
};

/* structure */
struct vzrpop {
    int gene, valid, cost, hiden, all;
    int stop, step;
    int policy;
    double precis;
    vzrpareto_t darwin;
    struct vzrsort *sort;
    enum popinfo genre;
    void *info;
    void (*dealloc_data) (void*);
    /* working values */
    int cur, cycle, level, again;
    int low, elite;
    struct vzritem *item;
    int *rg;
};

/* private */
/* vzrpop helpers */
static INLINE void vzrpop_set_policy(struct vzrpop *self, int mask, int val)
{
    self->policy = (self->policy & ~mask) | (val & mask);
}

static INLINE int vzrpop_end(struct vzrpop *self)
{
    return (
        !(self->policy & VZR_PARETO_CONT) && 
        self->level == self->stop &&
        self->again == 0 && 
        self->item[self->rg[self->level-1]].rank == 0 
    );
}
static INLINE void vzrpop_set_again(struct vzrpop *self)
{
    self->again = self->level;
    self->cur = (self->level == self->stop) ? self->level + self->step : self->level ;
    self->level = 0;
}

static INLINE void vzritem_cp(struct vzrpop *pop, struct vzritem *src, struct vzritem *dest)
{
    int i;

    dest->rank = src->rank;
    dest->info = src->info;
    dest->cycle = src->cycle;
    dest->isol = src->isol;
    dest->group = src->group;
    dest->data = src->data;
    for EACH(i, pop->all) dest->front[i] = src->front[i];
    if (dest->flag & DIPLOID) {
        if (src->flag & DIPLOID) {
            for EACH(i, pop->gene) {
                dest->dip.domin[i] = src->dip.domin[i];
                dest->dip.reces[i] = src->dip.reces[i];
            }
        } else {
            for EACH(i, pop->gene) {
                dest->dip.domin[i] = src->zone[i];
                dest->dip.reces[i] = src->zone[i];
            }
        }
    } 
    else {
        if (src->flag & DIPLOID) {
            for EACH(i, pop->gene) dest->zone[i] = src->dip.domin[i];
        } else {
            for EACH(i, pop->gene) dest->zone[i] = src->zone[i];
        }
    }
    dest->flag = src->flag;
}

static INLINE void vzritem_incp(struct vzrpop *pop, struct vzritem *src, struct vzritem *dest)
{
    int i;

    dest->rank = src->rank;
    dest->info = src->info;
    dest->cycle = src->cycle;
    dest->isol = src->isol;
    dest->group = src->group;
    dest->data = NULL;
    //for EACH(i, pop->all) dest->front[i] = src->front[i];
    if (dest->flag & DIPLOID) {
        if (src->flag & DIPLOID) {
            for EACH(i, pop->gene) {
                dest->dip.domin[i] = src->dip.domin[i];
                dest->dip.reces[i] = src->dip.reces[i];
            }
        } else {
            for EACH(i, pop->gene) {
                dest->dip.domin[i] = src->zone[i];
                dest->dip.reces[i] = src->zone[i];
            }
        }
    } 
    else {
        if (src->flag & DIPLOID) {
            for EACH(i, pop->gene) dest->zone[i] = src->dip.domin[i];
        } else {
            for EACH(i, pop->gene) dest->zone[i] = src->zone[i];
        }
    }
    dest->flag = src->flag;
}

void vzrpop_rerank(struct vzrpop *self, int siz);

/* api */
void vzrpop_init(struct vzrpop *self, int nbin, int nbcont, int nbout, double prec);
int vzrpop_aploid_build(struct vzrpop *self, int stop, int step);
int vzrpop_diploid_build(struct vzrpop *self, int stop, int step);
void vzrpop_aploid_dealloc(struct vzrpop *self);
void vzrpop_reset(struct vzrpop *self);

void vzrpop_dump(struct vzrpop *self, char *nfile);
void vzrpop_resume(struct vzrpop *self, int dip, char *nfile);

int vzrpop_viable(struct vzrpop *self, struct vzritem *item);
int vzrpop_add(struct vzrpop *self, struct vzritem *item);
void vzrpop_cycle(struct vzrpop *self);
int vzrpop_again(struct vzrpop *self, struct vzritem *item);

#endif
