/*  _______________________________________________________________________

 DAKOTA: Design Analysis Kit for Optimization and Terascale Applications
 Copyright (c) 2001, Sandia National Laboratories.
 This software is distributed under the GNU General Public License.
 For more information, see the README file in the top Dakota directory.
 _______________________________________________________________________ */

//      File:           LHSInput.h
//
//      Description:
#ifndef LHS_INPUT_H
#define LHS_INPUT_H

enum LHSNames
{
    NORMAL,
    LOGNORMAL,
    UNIFORM,
    LOGUNIFORM,
    TRIANGULAR,
    LOGTRIANGULAR,
    WEIBULL,
    EXPONENTIAL,
    CONSTANT,
    USERDEFINED
};

class LHSVariable
{
public:
    LHSVariable()
    {
        _Id = 0;
    }
    virtual ~LHSVariable();
    void SetName(const char *name)
    {
        _Name = new char[strlen(name) + 1];
        strcpy(_Name, name);
    }
    char *GetName()
    {
        return _Name;
    }
    void SetId(int id)
    {
        _Id = id;
    }
    int GetId()
    {
        return _Id;
    }
    virtual int GetType()
    {
        return -1;
    }
    virtual void printlog()
    {
    }
protected:
    char *_Name;
    int _Id;
private:
};

class LHSNormal: public LHSVariable
{
public:
    LHSNormal()
    {
        _atop = 1.0;
        _abot = 0.0;
    }
    ~LHSNormal()
    {
    }
    void SetRange(double a, double b, double c = 0.0)
    {
        _a = a;
        _b = b;
        _c = c;
    }
    void SetBounds(double LowerBound, double UpperBound);
    virtual void SetLowerBound(double LowerBound);
    virtual void SetUpperBound(double UpperBound);
    void Set_M_V(double m, double v)
    {
        _mean = m;
        _variation = v;
    }
    virtual int GetType()
    {
        return NORMAL;
    }
    double Get_A()
    {
        return _a;
    }
    double Get_B()
    {
        return _b;
    }
    double Get_C()
    {
        return _c;
    }
    double Get_Mean()
    {
        return _mean;
    }
    double Get_Variation()
    {
        return _variation;
    }
    double Get_PMin()
    {
        return _abot;
    }
    double Get_PMax()
    {
        return _atop;
    }
    void printlog();
protected:
    double _a;
    double _b;
    double _c;
    double _mean;
    double _variation;
    double _lower_bounds;
    double _upper_bounds;
    double _atop;
    double _abot;

private:
};

class LHSLognormal: public LHSNormal
{
public:
    LHSLognormal()
    {
    }
    int GetType()
    {
        return LOGNORMAL;
    }
protected:
private:
};

class LHSUniform: public LHSNormal
{
public:
    LHSUniform()
    {
    }
    int GetType()
    {
        return UNIFORM;
    }
protected:
private:
};

class LHSLoguniform: public LHSNormal
{
public:
    LHSLoguniform()
    {
    }
    int GetType()
    {
        return LOGUNIFORM;
    }
protected:
private:
};

class LHSTriangular: public LHSNormal
{
public:
    LHSTriangular()
    {
    }
    int GetType()
    {
        return TRIANGULAR;
    }
    void SetLowerBound(double LowerBound);
    void SetUpperBound(double UpperBound);
    double Finv(double);
protected:
private:
};

class LHSLogtriangular: public LHSTriangular
{
public:
    LHSLogtriangular()
    {
    }
    int GetType()
    {
        return LOGTRIANGULAR;
    }
protected:
private:
};

class LHSWeibull: public LHSNormal
{
public:
    LHSWeibull()
    {
    }
    int GetType()
    {
        return WEIBULL;
    }
protected:
private:
};

class LHSExponential: public LHSNormal
{
public:
    LHSExponential()
    {
    }
    int GetType()
    {
        return EXPONENTIAL;
    }
protected:
private:
};

class LHSConstant: public LHSNormal
{
public:
    LHSConstant()
    {
    }
    int GetType()
    {
        return CONSTANT;
    }
    double Get_C()
    {
        return _c;
    }
    void SetConstant(double c)
    {
        _c = c;
    }
protected:
    double _c;
private:
};

class LHSUserDefined: public LHSVariable
{
public:
    LHSUserDefined()
    {
    }
    ~LHSUserDefined()
    {
        if (_filename)
            delete[] _filename;
    }
    void SetFilename(const char *name)
    {
        _filename = new char[strlen(name) + 1];
        strcpy(_filename, name);
    }
    int GetType()
    {
        return USERDEFINED;
    }
    char *Get_Filename()
    {
        return _filename;
    }
protected:
    char *_filename;
private:
};

struct Listelem
{
    LHSVariable *Data;
    Listelem *Next;
};

class LHSInput
{
public:
    LHSInput();
    ~LHSInput()
    {
        release();
    }
    void Add(LHSVariable *ptr);
    void detruit()
    {
        Listelem *temp = h;
        h = h->Next;
        delete temp->Data;
        free(temp);
    }
    void release();

    void SetN_of_V(long int n)
    {
        _N_of_Variables = n;
    }
    void SetN_of_O(long int n)
    {
        _N_of_Observations = n;
    }
    void SetSampleFlag(int n)
    {
        _SampleFlag = n;
    }

    long int GetN_of_V()
    {
        return _N_of_Variables;
    }
    long int GetN_of_O()
    {
        return _N_of_Observations;
    }
    int GetSampleFlag()
    {
        return _SampleFlag;
    }
    Listelem *GetListPtr()
    {
        return h;
    }
    void printlog();

protected:
    long int _N_of_Variables;
    long int _N_of_Observations;
    int _SampleFlag;
    Listelem *h;
    Listelem *tail;
private:
};

#endif
