/*  _______________________________________________________________________

 DAKOTA: Design Analysis Kit for Optimization and Terascale Applications
 Copyright (c) 2001, Sandia National Laboratories.
 This software is distributed under the GNU General Public License.
 For more information, see the README file in the top Dakota directory.
 _______________________________________________________________________ */

//- Description: Implementation code for LHS class
//- Owner:       Bruce Bainbridge
//- Checked by:
//- Version:
//#define DEBUG
//      Description:  This file contains the methods of the object 
//        LatinHypercube. The LatinHypercube is used to generate a 
//        distribution of random sample values.
//
//      Limitations and abilities:
//        The following distributions can be generated:
//
//        Distribution          Inputs
//        1. Normal             mean and coefficient of variation
//        2. Lognormal          mean and coefficient of variation
//        3. Weibull            mean and coefficient of variation
//        4. Uniform            Lower Limit and Upper Limit
//        5. Loguniform         Lower Limit and Upper Limit
//        6. Constant           mean = value, coef of variation = 0
//
//      Note: most of this code was originally Fortran code where
//        all arrays were 1 dimensional. This use of 1-dim arrays to
//        represent 2-dim arrays is still used in this C++ code.
//
//**************************************************************************
#include <cstring>
#include <cstdio>
#include <iostream>
#include <fstream>
#include <ctime>
#include <cmath>
#include <cstdlib>
#include "LatinHypercube.H"
#include "LHSInput.H"
#include "LHSVecMatUtil.H"

//**************************************************************************

LatinHypercube::LatinHypercube(LHSInput *InPtr)
{
//   Description:  This is the constructor method for the object. It sets 
//      the intial values and variable values.
//
//   Input
//      ip:    the pointer to the tcl/tk Tcl_Interp structure
//      InPtr: the pointer containing the information about each variable
//             that a sample is to be generated for.
//   Output
//      created and initialize Latinhypercube object

#ifdef DEBUG
    cerr << "LatinHypercube()\n";
#endif

    double *matId;
    long int i, j;
    time_t secondes;
    //_interp = ip;
    _InPtr = InPtr;
    time(&secondes);
    _seed = (int) (secondes);          // set default seed;
    _originalseed = _seed;     // save _input seed value (ran2 changes it)

    // MSE, 11/16/00: Turning on random pairing (and turning off restricted 
    // pairing) fixes an induced correlation problem observed with lhs (but not
    // observed with pure random).  The most desirable solution would be to 
    // identify the problem with the correlation in the restricted pairing case
    // (see SetUserCorrMatrix) and map through the user correlation spec.
    // Possibles:
    //   (1) default correlation matrix is bad (e.g., zeros)??
    //   (2) rand replacement using seed improperly (most likely)
    _randomPairing = 0;       // set random pairing default False
    //_randomPairing = 1;       // set random pairing default True
    _saveRang = 0;              // pas de sauvegarde des rangs par d�faut

    _sampletype = _InPtr->GetSampleFlag();
    // _sampletype = NONRANDOM;  // set random sample default False

    _ucm = 0;
    _nobv = InPtr->GetN_of_O();
    _nvar = InPtr->GetN_of_V();
    _nvx = _nvar * (_nvar + 1) / 2;        //  Number of elements in lower tri

    bins = NULL;
    _sinfo = new stats[_nvar + 1];
    _sampx = new double[_nobv * _nvar + 1]; // Vector which contains the sampling
    _workxx = new double[_nobv * _nvar + 1]; // Double of the _sampx vector

    // allocate struct vectors
    _rank.SetSize(_nobv, _nvar);
    _workc.SetSize(_nvar);
    _cmatr.SetSize(_nvar);

    // fg::modif du 17062003 Matrice identi� par d�faut
    // Only the inferior triangular part is kept
    // 11
    // 21 22
    // 31 32 33
    // 41 42 43 44 etc
    _cmatr.ncm = _nvar;
    matId = new double[_nvx];
    for (i = 1; i <= _nvar; i++)
    {
        for (j = 1; j < i; j++)
        {
            matId[LOC1(i, j)] = 0.0;
        }
        matId[LOC1(i, i)] = 1.0;
    }
    SetUserCorrMatrix(matId);
    delete[] matId;
    matId = NULL;
    // fin::modif du 17062003
}

LatinHypercube::~LatinHypercube()
{
#ifdef DEBUG
    cerr << "LatinHypercube::~LatinHypercube()\n";
#endif

//   #ifdef DEBUG
//     LOG("~LatinHypercube : _InPtr");
//   #endif
//   delete _InPtr; _InPtr = NULL;

#ifdef DEBUG
    LOG("~LatinHypercube : _sampx");
#endif
    delete[] _sampx;
    _sampx = NULL;

#ifdef DEBUG
    LOG("~LatinHypercube : _workxx");
#endif
    delete[] _workxx;
    _workxx = NULL;

#ifdef DEBUG
    LOG("~LatinHypercube : _sinfo");
#endif
    delete[] _sinfo;
    _sinfo = NULL;

#ifdef DEBUG
    LOG("~LatinHypercube : bins fait dans la fonction printHistograms");
#endif
    if (bins != NULL)
    {
        LOG("~LatinHypercube : delete bins");
        printf("~LatinHypercube : delete nbins[%d]", nbins);

        for (int i = 0; i < nbins; i++)
        {
            printf("~LatinHypercube : delete bins i[%d]", i);
            if (bins[i] != NULL)
            {
                delete[] bins[i];
                bins[i] = NULL;
            }
        }
        delete[] bins;
        bins = NULL;
    }

#ifdef DEBUG
    LOG("~LatinHypercube : fin");
#endif
}

/*! \fn int LatinHypercube::addCorr(int i, int j, double value)
 \brief Adding of an element ins the correlation matrix.

 The element \f$ Corr(i,j) = \mathrm{value} \f$ were added in the correlation matrix.
 The test of validity of the matrix (\e i.e.  a definite positive matrix ) will be performed in the 
 \ref TestCorrMatrix method.

 Cause of the inf triangular, LOC1(i,j) != LOC1(j,i).
 This method works well for j < i

 */
int LatinHypercube::addCorr(int i, int j, double value)
{

#ifdef DEBUG
    cerr << "LatinHypercube::addCorr(" << i << ", " << j << ", " << value << ")" << endl;
#endif

    if (i < 1 || i > _nvar || j < 1 || j > _nvar || i == j || value < -1.0
            || value > 1.00)
    {
        return 0;
    }

    if (i < j)
        _cmatr.ucorr[LOC1(j, i)] = value;
    else
        _cmatr.ucorr[LOC1(i, j)] = value;

    return 1;

}
/*! \fn void LatinHypercube::TestCorrMatrix()
 \brief Checks if the correlation matrix is definite positive
 
 
 */
void LatinHypercube::TestCorrMatrix()
{

#ifdef DEBUG
    cerr << "LatinHypercube::TestCorrMatrix()\n";
#endif

    int test;
    double **a, *p;

    p = _vector(_nvar);
    a = matrix(_nvar, _nvar);

    // copy 1-dim array to 2-dim de la matrice de corr�lation de l'utilisateur
    // qui est contenue dans _cmatr.ucorr
    d1_to_d2(a, _cmatr.ucorr);

    //  test si la matrice A est d�finie positive
    // par la m�thode de Cholesky (cf Numerical Recipies)
    test = choldc(a, _nvar, p);

    // Si c'est le cas, on met � 1 la variable _ucm et on renseigne le vector _workc.uq
    if (test)
    {
        _ucm = 1;
        // d2_to_d1(a,_workc.uq);
        chlsky(_cmatr.ucorr, _workc.uq, _nvar);
    }

    delete[] p;
    for (int i = 0; i < _nvar; i++)
        delete[] a[i];
    delete[] a;

}

int LatinHypercube::SetUserCorrMatrix(double *mx)
{
//  Description:  This method first tests whether the given matrix is 
//      positive definite. If it is, it is then copied to the structure
//      _workc.uq and the user defined correlation flag is set to true
//      (_ucm = 1) and a value of 1 is returned. If it is not positive 
//      definite than a value of 0 is return and the matrix is not
//      stored in the object.
//   Input
//      mx: pointer to symetric matrix
//   Output
//      If mx is positive definite
//        copy mx to _workc.uq
//        _ucm = 1
//        return 1
//      If mx is not positive definite
//        _ucm = 0
//        return 0

#ifdef DEBUG
    cerr << "LatinHypercube::SetUserCorrMatrix()\n";
#endif

    int test;
    long int i;
    double **a, *p;

    p = _vector(_nvar);
    a = matrix(_nvar, _nvar);

    for (i = 0; i < _nvx; i++)
        _cmatr.ucorr[i] = mx[i];

    d1_to_d2(a, mx);     //  copy 1-dim array to 2-dim

//   cout << endl << "Matrice A" << endl;
//   for (int k=0; k<_nvar; k++) { // loop over rows
//     for (int l=0; l<_nvar; l++) // loop over columns
//       printf("%5.2f ", a[k][l]);
//     cout << endl;
//   }

    //  check to see if matrix is positive definite
    test = choldc(a, _nvar, p);
    free_vector(p);

    if (test)
    {
        _ucm = 1;
        d2_to_d1(a, _workc.uq);
        free_matrix(a, _nvar);
        chlsky(_cmatr.ucorr, _workc.uq, _nvar);
        return 1;
    }
    else
    {
        return 0;  // cholesky test failed
    }
}

/*! \fn void  LatinHypercube::GenSample()
 \brief Generation of the sampling
 
 This method calles the corresponding distribution 
 method sample generator for each variable defined in the Input
 object pointed to by InPtr. After the samples are generated
 then the method Mix is called which ranks the sample values
 based on the correlation matrix being used. Finally the method
 compute_Stats is called to compute values to be used in ploting
 and for checking against input statistics. These computed values
 are later printed in the results file if the method 
 print_Output() is called.
 
 */
void LatinHypercube::GenSample()
{

#ifdef DEBUG
    cerr << "LatinHypercube::GenSample()\n";
#endif

    Listelem *cur;
    LHSVariable *varptr;

    if (_bLog > 0)
        printlog();

    // generate samples for all variables
    for (cur = _InPtr->GetListPtr(); cur; cur = cur->Next)
    {
        varptr = cur->Data;
        switch (varptr->GetType())
        {
        case NORMAL:
            normal(cur->Data->GetId(), _nobv,
                    ((LHSNormal *) varptr)->Get_Mean(),
                    ((LHSNormal *) varptr)->Get_Variation(),
                    cur->Data->GetType(), ((LHSNormal *) varptr)->Get_PMin(),
                    ((LHSNormal *) varptr)->Get_PMax());
            break;
        case LOGNORMAL:
            normal(cur->Data->GetId(), _nobv,
                    ((LHSLognormal *) varptr)->Get_Mean(),
                    ((LHSLognormal *) varptr)->Get_Variation(),
                    cur->Data->GetType(), ((LHSNormal *) varptr)->Get_PMin(),
                    ((LHSNormal *) varptr)->Get_PMax());
            break;
        case CONSTANT:
            constant(cur->Data->GetId(), _nobv,
            //                ((LHSLognormal *)varptr)->Get_A()+AB_CON);
                    ((LHSConstant *) varptr)->Get_C());
            break;
        case UNIFORM:
            uniform(cur->Data->GetId(), _nobv, ((LHSUniform *) varptr)->Get_A(),
                    ((LHSUniform *) varptr)->Get_B(), cur->Data->GetType());
            break;
        case LOGUNIFORM:
            uniform(cur->Data->GetId(), _nobv,
                    ((LHSLoguniform *) varptr)->Get_A(),
                    ((LHSLoguniform *) varptr)->Get_B(), cur->Data->GetType());
            break;
        case TRIANGULAR:
            triangular(cur->Data->GetId(), _nobv,
                    ((LHSTriangular *) varptr)->Get_A(),
                    ((LHSTriangular *) varptr)->Get_B(),
                    ((LHSTriangular *) varptr)->Get_C(), cur->Data->GetType(),
                    ((LHSTriangular *) varptr)->Get_PMin(),
                    ((LHSTriangular *) varptr)->Get_PMax());
            break;
        case LOGTRIANGULAR:
            triangular(cur->Data->GetId(), _nobv,
                    ((LHSTriangular *) varptr)->Get_A(),
                    ((LHSTriangular *) varptr)->Get_B(),
                    ((LHSTriangular *) varptr)->Get_C(), cur->Data->GetType(),
                    ((LHSNormal *) varptr)->Get_PMin(),
                    ((LHSNormal *) varptr)->Get_PMax());
            break;
        case WEIBULL:
            weibull(cur->Data->GetId(), _nobv, ((LHSWeibull *) varptr)->Get_A(),
                    ((LHSWeibull *) varptr)->Get_B(),
                    ((LHSWeibull *) varptr)->Get_C());
            break;
        case EXPONENTIAL:
            exponential(cur->Data->GetId(), _nobv,
                    ((LHSExponential *) varptr)->Get_A(),
                    ((LHSExponential *) varptr)->Get_B());
            break;
        case USERDEFINED:
            user_defined(cur->Data->GetId(), _nobv,
                    ((LHSUserDefined *) varptr)->Get_Filename());
            break;
        default:
            std::cerr << "ERROR! Variable type is default!" << std::endl;
            break;

        }  //  end switch
    }  //  end for
    // V�rifie que la matrice est d�finie possitive
    TestCorrMatrix();
    //  rank samples based on correlation matrix
    Mix();
    //corcal();
    //  Compute statistics of samples to be used in ploting
    compute_Stats();

    // Affichage de la log
    if (_bLog > 0)
        print_Output("_dakota_.dat");

}

//**************************************************************************

void LatinHypercube::constant(int j, long int n, double a)
{
//   Description: This method generates a constant distribution.
//      Simple uses the value a for each sample.
//   
//   Inputs
//      j:      the variable id number that the distribution is for
//      n:      number of sample observations to create
//      a:      the constant value of the variable
//   Outputs
//      the portions of the following arrays corresponding to the jth 
//      variable sample values are set
//        _sampx[]
//        _workxx[]

#ifdef DEBUG
    cerr << "LatinHypercube::constant()\n";
#endif

    for (long int i = 0; i < n; ++i)
    {
        //_sampx[(j-1)*_nobv+i] = i;
        _sampx[(j - 1) * _nobv + i] = _workxx[(j - 1) * _nobv + i] = a;
    }  //  end for loop
}  //  end constant()

//**************************************************************************

void LatinHypercube::normal(int j, long int n, double mean, double k, int type,
        double pmin, double pmax)
//   Description:  This method generates a normal distribution or a 
//      Lognormal distribution.
//
//   Inputs
//      j:      the variable id number that the distribution is for 
//      n:      number of sample observations to create
//      mean:   mean value for sample
//      k:      std. deviation (normal) or error factor (lognormal)
//      type:   type flag  type = NORMAL or LOGNORMAL
//      pmin:   le quantile inf�rieur
//      pmax :  le quantile sup�rieur
//   Outputs
//      the portions of the following arrays corresponding to the jth  
//      variable sample values are set 
//        _sampx[] 
//        _workxx[]
{
    double probinc;
    double sig;
    double pmu;
    double strtpt;
    double r;

#ifdef DEBUG
    cerr << "LatinHypercube::normal()\n";
#endif

    probinc = (pmax - pmin) / (double) n;
    if (_sampletype == RANDOM)  // if a random sample
        probinc = (pmax - pmin);
    if (type == LOGNORMAL)
    {
        //    sig = sqrt(log(k*k+1.0));
        sig = log(k) / 1.645;
        pmu = log(mean) - 0.5 * sig * sig;
    }
    else
    {
        //    sig = mean*k;
        sig = k;
        pmu = mean;
    }

    strtpt = pmin;
    //srand48(_seed);
    for (long int i = 0; i < n; ++i)
    {
        //r = probinc * drand48() + strtpt;
        r = probinc * ran2(&_seed) + strtpt;
        if (r >= .999)
            r = .999;
        if (r <= .001)
            r = .001;
        if (type == LOGNORMAL)
        {
            _sampx[(j - 1) * _nobv + i] = _workxx[(j - 1) * _nobv + i] = exp(
                    finvnor(r) * sig + pmu);
        }
        else
        {
            _sampx[(j - 1) * _nobv + i] = _workxx[(j - 1) * _nobv + i] =
                    finvnor(r) * sig + pmu;
        }
        if (_sampletype == NONRANDOM)
            strtpt += probinc;
    }  //  end for loop
}  //  end normal()

//**************************************************************************

void LatinHypercube::uniform(int j, long int n, double a, double b, int type)
{
//   Description:  This method generates a uniform distribution or a
//      Loguniform distribution.
//
//   Inputs
//      j:      the variable id number that the distribution is for
//      n:      number of sample observations to create
//      a:      lower limit for sample
//      b:      upper limit for sample
//      type:   type flag  type = NORMAL or LOGNORMAL
//   Outputs
//      the portions of the following arrays corresponding to the jth
//      variable sample values are set
//        _sampx[]
//        _workxx[]

#ifdef DEBUG
    cerr << "LatinHypercube::uniform()\n";
#endif

    double strtpt, probinc;
    long int i, nint, k, id;

    i = 0;
    nint = 1;
    //  srand48(_seed);
    for (k = 1; k <= nint; k++)
    {
        if (type == LOGUNIFORM)
        {
            a = log10(a);
            b = log10(b);
        }
        probinc = (b - a) / (double) (n);
        if (_sampletype != NONRANDOM)
            probinc = b - a;
        for (id = 1; id <= n; id++)
        {
            i = i + 1;
            if (_sampletype == NONRANDOM)
                strtpt = (id - 1) * probinc;
            else
                strtpt = 0.0;
            if (type == LOGUNIFORM)
            {
                //_sampx[(j-1)*n+i-1] = _workxx[(j-1)*n+i-1] = pow(10,a+strtpt+probinc*ran2(&_seed));
                _sampx[(j - 1) * _nobv + i - 1] = _workxx[(j - 1) * _nobv + i
                        - 1] = pow(10, a + strtpt + probinc * ran2(&_seed));
            }
            if (type == UNIFORM)
            {
                //_sampx[(j-1)*n+i-1] = _workxx[(j-1)*n+i-1] = a+strtpt+probinc*ran2(&_seed);
                _sampx[(j - 1) * _nobv + i - 1] = _workxx[(j - 1) * _nobv + i
                        - 1] = a + strtpt + probinc * ran2(&_seed);
            }
        }
    }  //  end for loop

}  //  end uniform()

/*! \fn void  LatinHypercube::triangular(int j, long int n, double a, double b, double c, int type, double pmin, double pmax)
 \brief Triangular Law
 
 Generation of a sampling following the triangular law 
 
 \param j  (int) Index of the variable
 \param n  (int) size of the sampling
 \param a  (double) inferior bound
 \param b  (double) superior bound
 \param c  (double) medain value
 \param type  (int) Law nature (triangular/Logtriangular)
 \param pmin (double) Inferior quantile
 \param pmax (double) Superior quantile

 Outputs
 The portions of the following arrays corresponding to the jth
 variable sample values are set
 _sampx[]
 _workxx[]
 
 */
void LatinHypercube::triangular(int j, long int n, double a, double b, double c,
        int type, double pmin, double pmax)
{

#ifdef DEBUG
    cerr << "LatinHypercube::triangular()\n";
#endif

    double strtpt, probinc, dalea, r;
    double leftArea;
    long int i;

    //  srand48(_seed);
    if (type == LOGTRIANGULAR)
    {
        a = log10(a);
        b = log10(b);
        c = log10(c);
    }
    // Calcul de l'aire du triangle de gauche
    leftArea = (c - a) / (b - a);

    //
    probinc = (pmax - pmin);
    if (_sampletype == NONRANDOM)
        probinc /= (double) n;
    strtpt = pmin;

    for (i = 0; i < n; i++)
    {
        r = probinc * ran2(&_seed) + strtpt;
        if (r < leftArea)
            dalea = a + sqrt(r * (b - a) * (c - a));
        else
            dalea = b - sqrt((1.0 - r) * (b - a) * (b - c));

        if (type == LOGTRIANGULAR)
            _sampx[(j - 1) * _nobv + i] = _workxx[(j - 1) * _nobv + i] = pow(10,
                    dalea);
        else
            _sampx[(j - 1) * _nobv + i] = _workxx[(j - 1) * _nobv + i] = dalea;

        // cout << i << ": " << r << " [" <<  dalea << "] [" << _sampx[(j-1)*_nobv+i] << "]" << endl;

        if (_sampletype == NONRANDOM)
            strtpt += probinc;
    }
}
/*! \fn void  LatinHypercube::weibull(int j, long int n, double alpha, double beta, double xmin = 0.0)
 \brief Generation of a sampling by the Weibull Law
 
 \param j  (int) index of the variable
 \param n  (int) size of the sampling
 \param alpha  (double) coefficient \f$ \alpha\f$
 \param beta  (double) coefficient \f$ \beta\f$
 \param xmin  (double = 0.0) min value

 The formula of \pdf is given by the following formula:
 \f[
 f(x) = \frac{\alpha}{\beta} \left(\frac{x - x_{min}}{\beta}\right)^{\alpha - 1} \exp{- \left(\frac{x - x_{min}}{\beta}\right)^{\alpha}}
 \f]

 The repartition function is given by the following formula:
 \f[
 F(x) = 1 - exp{\left(\frac{x - x_{min}}{\beta}\right)^{\alpha}}
 \f]

 The inverse function of the repartition function is given by the following formula:

 \f[
 x_{u} = F^{-1}(u) = x_{min} + \beta \left( -1.0 * \ln\left( 1.0 - u \right)\right)^{1/\alpha}
 \f]
 Outputs
 the portions of the following arrays corresponding to the jth
 variable sample values are set
 _sampx[]
 _workxx[]
 */
void LatinHypercube::weibull(int j, long int n, double alpha, double beta,
        double xmin = 0.0)
{

#ifdef DEBUG
    cerr << "LatinHypercube::weibull()\n";
#endif
    double strtpt, probinc;
    double dtemp, invalpha, r, res;
    //double beta;
    long int i;

    invalpha = 1.0 / alpha;
    // original code doesn't seem to perform LHS function
    // this code taken from lhs.f
    probinc = 1.0 / (double) n;
    if (_sampletype == RANDOM)
        probinc = 1.0; // check for random sampling

    strtpt = 0.;
    for (i = 0; i < n; i++)
    {
        r = probinc * ran2(&_seed) + strtpt;
        dtemp = -1.0 * log(1.0 - r);
        res = xmin + pow(dtemp, invalpha) * beta;
        // cout << i << "dtemp [" << dtemp << "] Res[" << res << "]  r[" << r << "] ** " << endl;
        //_sampx[(j-1)*_nobv+i] = _workxx[(j-1)*_nobv+i] = max(res, 1.0e-10);
        _sampx[(j - 1) * _nobv + i] = _workxx[(j - 1) * _nobv + i] =
                (res > 1.0e-10) ? res : 1.0e-10;
        if (_sampletype == NONRANDOM)
        {
            //if (probinc < 1.0)
            strtpt += probinc;
        }
    }

    // original code:
    //alfha = pow(variation,-1.08);
    //beta = mean / exp(gammln( ((1/alfha) + 1) ));
    //for (i = 0;i < n; i++) {
    //  _sampx[(j-1)*_nobv+i] = beta * exp( (log(-log(ran2(&_seed))))/alfha );
    //  _workxx[(j-1)*_nobv+i] = beta * exp( (log(-log(ran2(&_seed))))/alfha );
    //}

    //qsort((double *) &_sampx[(j-1)*_nobv+1-1],_nobv,sizeof(double),
    //            doublecompare);
    //quicksort(&_sampx[(j-1)*_nobv],0,_nobv-1);
    //qsort((double *) &_workxx[(j-1)*_nobv+1-1],_nobv,sizeof(double),
    //            doublecompare);
    //quicksort(&_workxx[(j-1)*_nobv],0,_nobv-1);
}

//**************************************************************************

void LatinHypercube::exponential(int j, long int n, double lambda, double xmin)
{
#ifdef DEBUG
    cerr << "LatinHypercube::exponential(" << lambda << "]" << endl;
#endif

    double invlambda, r, res;
    double strtpt, probinc;
    long int i;

    invlambda = -1.0 / lambda;
    // original code doesn't seem to perform LHS function
    // this code taken from lhs.f
    probinc = 1.0 / (double) n;
    if (_sampletype == RANDOM)
        probinc = 1.0; // check for random sampling

    strtpt = 0.;
    for (i = 0; i < n; i++)
    {
        r = probinc * ran2(&_seed) + strtpt;
        res = log(r);
        // cout << i << "dtemp [" << dtemp << "] Res[" << res << "]  r[" << r << "] ** " << endl;
        //_sampx[(j-1)*_nobv+i] = _workxx[(j-1)*_nobv+i] = max(res, 1.0e-10);
        _sampx[(j - 1) * _nobv + i] = _workxx[(j - 1) * _nobv + i] = xmin
                + res * invlambda;
        if (_sampletype == NONRANDOM)
        {
            //if (probinc < 1.0)
            strtpt += probinc;
        }
    }

}

//**************************************************************************

void LatinHypercube::user_defined(int j, long int n, char *filename)
{
//   Description:  This method generates a User Defined distribution.
//
//   Inputs
//      j:      the variable id number that the distribution is for
//      n:      number of sample observations to create
//      filename: name of file that holds distribution
//   Outputs
//      the portions of the following arrays corresponding to the jth
//      variable sample values are set
//        _sampx[]
//        _workxx[]

// blb
// replace tcl file i/o with standard C++ calls

#ifdef DEBUG
    cerr << "LatinHypercube::user_defined()\n";
#endif

    const int histogram_div = 30;
    double probinc, strtpt, r;
    double a, b, c;
    double *value, val, cdfmax, cdfmin, delta, *h;
    double range_low, range_high;
    int k, *cdf, index=-1;
    long int i, num;
    char str1[255];

    // *********  read file data  ************************ 

    //  open channel to read file
    std::ifstream filein(filename);
    if (!filein)
    {
        std::cerr << "Error in opening " << filename << '\n';
    }
    //  get string str1
    filein >> str1;
    //  read number of samples
    sscanf(str1, "%ld", &num);
    //  Tcl_DStringFree(&str1);
    value = new double[num];

    //  read num samplesf
    for (i = 0; i < num; i++)
    {
        //Tcl_DStringInit(&str1);
        //Tcl_Gets(chan,&str1);
        filein >> str1; // get string str1
        sscanf(str1, "%lf", &val);     // read sample from str1
        value[i] = val;
        //cerr << "value[" << i << "] = " << val << '\n';
        //Tcl_DStringFree(&str1);
    }
    //Tcl_Close(NULL,chan);
    filein.close();

    // ********** generate cdf data *************************88
    //  data in file must have been previously sorted

    cdfmin = value[0];
    cdfmax = value[num - 1];
    delta = (cdfmax - cdfmin) / histogram_div;
    range_low = cdfmin;
    range_high = range_low + delta;
    cdf = new int[histogram_div + 1];
    for (i = 0; i <= histogram_div; i++)
        cdf[i] = 0;
    for (k = 1; k < histogram_div; k++)
    {
        for (i = 0; i < num; i++)
            if (value[i] >= range_low && value[i] < range_high)
                cdf[k] = cdf[k] + 1;
        cdf[k] = cdf[k] + cdf[k - 1];
        //cerr << "cdf[" << k << "] = " << cdf[k] << '\n';
        range_low = range_high;
        range_high = range_high + delta;
    }
    // last range of cdf
    for (i = 0; i < num; i++)
        if (value[i] >= range_low && value[i] <= cdfmax)
            cdf[histogram_div] = cdf[histogram_div] + 1;
    cdf[histogram_div] = cdf[histogram_div] + cdf[histogram_div - 1];
    //cerr << "cdf[" << k << "] = " << cdf[k] << '\n';
    h = new double[num + 1];
    for (i = 1; i <= histogram_div; i++)
        h[i] = double(cdf[i]) / double(num);

    // ********** caluclate lhs sample
    probinc = 1.0 / n;
    strtpt = 0.0;

    for (i = 0; i < n; i++)
    {
        r = probinc * ran2(&_seed) + strtpt;
        if (r >= .999)
            r = .999;
        if (r <= .001)
            r = .001;

        //  find interval in cdf that r falls in
        range_low = 0.0;
        range_high = h[1];
        for (k = 1; k < histogram_div; k++)
        {
            if (r >= range_low && r < range_high)
            {
                index = k;
            }
            range_low = h[k];
            range_high = h[k + 1];
        }
        // last range of cdf
        if (r >= range_low && r <= range_high)
        {
            index = histogram_div;
        }

        c = cdfmin + (index * delta);
        a = (h[index] - h[index - 1]) / delta;
        b = h[index] - (a * c);

        //cerr << "r = " << r << ", c = " << c << ", a = " << a << ", b = " << b << '\n';
        _sampx[(j - 1) * _nobv + i] = (r - b) / a;
        _workxx[(j - 1) * _nobv + i] = (r - b) / a;

        strtpt += probinc;
        //cerr << "x[" << i << "] = " << _sampx[(j-1)*_nobv+i-1] << '\n';
    }
}

//**************************************************************************
/*! \fn void  LatinHypercube::Mix()
 \brief Generation of the sampling
 
 This method takes the samples of the variables and 
 ranks them according to the correlation matrix being used.
 If a user correlation matrix is not to be used then a generic
 correlation matrix is generated which tries to approximate 
 the correlation between each variable as 0.
 
 SUBROUTINE MIX IMPLEMENTS THE SCHEME DESCRIBED IN IMAN AND 
 CONOVER (1982) -SEE USERS GUIDE- FOR INDUCING A DESIRED 
 CORRELATION STRUCTURE

 IF ICM EQUALS 0 THEN A MIX OF THE GENERATED DATA IS 
 OBTAINED.  OTHERWISE, A MIX OF RANDOM INTEGERS IS 
 GENERATED AND MULTIPLIED BY THE CORRELATION MATRIX. 
 THE RESULT IS RANKED AND THE RANKS ARE USED TO GENERATE 
 THE RANDOM SAMPLE. 
 
 FOR THE FULL RANK CASE(N.GT.NV)THE FOLLOWING PROCEDURE IS USED, 
 GENERATE A SAMPLE SUCH THAT NONE OF THE OFF-DIAGONAL ELEMENTS OF 
 THE CORRESPONDING CORRELATION MARTIX ARE EQUAL TO ONE IN ABSOLUTE
 
 VALUE.CHOLESKY FACTORIZATION IS DONE NEXT. 
 FOR THE LESS THAN FULL RANK CASE(N.LE.NV) 25 SAMPLES ARE GENERATED
 
 AND THAT SAMPLE WHOSE CORRESPONDING CORRELATION MATRIX HAS THE 
 SMALLEST MAXIMUM(IN ABSOLUTE VALUE)OFF-DIAGONAL ELEMENT IS THE 
 SAMPLE THAT IS CHOSEN. CHOLESKY FACTORIZATION IS SKIPPED IN 
 THIS CASE.


 <b> Description of the method</b>

 <b> List of the used elements</b>
 <ul>
 <li> \em _rank.xv Vector which contains scores \f$ \Phi^{-1}\left(i/n+1\right)\f$;
 <li> \em _sampx matrix of scores for each variable \f$ j\f$. It is equivalent to the matrix \f$ S\f$ .
 <li> \em _work.uq matrix \f$ P\f$ : triangular matrix associated to \f$ X\f$ tel que \f$ X = P P^{T}\f$.
 </ul>
 
 
 */

void LatinHypercube::Mix()
{

#ifdef DEBUG
    cerr << "LatinHypercube::Mix()\n";
#endif

#ifdef DEBUG_ROOT
    init_root_file("toto");
#endif

    /* System generated locals */
    long int i__1, i__2;
    double r__1=-1.0;

    /* Local variables */
    static double elem;
    static long int imin, i, j, k;
    static double dn;
    static long int jj, kk, ir, it, iw;
    static double cmxold;
    static long int ijk;
    static double cmx;
    static long int ndx, nvn;

    nvn = _nvar * _nobv;
    cmxold = (double) 9999.;
    dn = (double) (_nobv + 1);

    // Initialisation du vecteur des scores \b _rank.xv par \f$ f^{-1}_{N} (\frac{i+1}{\_nobs +1}\f$.
#ifdef DEBUG
    printf("\n Matrice des Score with Dakota\n");
#endif
    for (i = 0; i < _nobv; ++i)
    {
        r__1 = (double) (i + 1) / dn;
        _rank.xv[i] = finvnor(r__1);
#ifdef DEBUG
        printf("DAKOTA: %d\t%3d\t%f\t%f\n", i, i+1, r__1, _rank.xv[i]);
#endif
    }

    // Boucle avec un max de 25 it�rations
    for (ijk = 1; ijk <= 25; ++ijk)
    {
        L10:
#ifdef DEBUG
        cout << "**** tour["<<ijk<<"] ***"<<endl;
        LOG("tour");
#endif
        for (i = 0; i < _nobv; ++i)
        {
            _rank.iwk[i] = i + 1;
        }

        // Bouble sur chaque variable
        for (i = 1; i <= _nvar; ++i)
        {
            // permutation des deux �l�ments index�s par \em ir et \em jj
            for (j = 2; j <= _nobv; ++j)
            {
                ir = _nobv + 2 - j;
                jj = (int) (ir * ran2(&_seed)) + 1;
                kk = _rank.iwk[ir - 1];
                _rank.iwk[ir - 1] = _rank.iwk[jj - 1];
                _rank.iwk[jj - 1] = kk;
            }
#ifdef DEBUG
            printf("\n ***** var=%d\n", i);
#endif
            // remplissage de la matrice �chantillon 
            for (j = 1; j <= _nobv; ++j)
            {
                ndx = _rank.iwk[j - 1];
#ifdef DEBUG
                printf("DAKOTA permut: var%d : %2d\tndx=%3d\n", i, j, ndx - 1);
#endif
                // \em _sampx est la matrice contenant l'ensemble des indices. Une copie est assur�e avec \b _workxx
                _sampx[(i - 1) * _nobv + j - 1] = _rank.xv[ndx - 1];
            }
        }
#ifdef DEBUG
        print_matrix(_sampx,"[_sampx] Matrice _sampx avant corcal");
#endif

        if (_nvar == 1 || _randomPairing == 1)
        {
            goto L235;
        }

        /*     COMPUTE THE CORRELATION MATRIX ON THE VDW SCORES */

        corcal();
#ifdef DEBUG
        print_matrix(_sampx,"[_sampx] Matrice _sampx apr�s corcal");
#endif
        if (_nobv <= _nvar)
        {
            goto L37;
        }
#ifdef DEBUG_ROOT
        print_cmatrix(_cmatr.corr,"[_cmatr.corr] sortie E");
#endif

        for (i = 2; i <= _nvar; ++i)
        {
            for (j = 1; j < i; ++j)
            {
                r__1 = fabs(r__1);
                if (r__1 >= (double) .9999)
                {
                    ijk++;
                    std::cout << "**** tour[" << ijk << "] ***" << std::endl;
                    goto L10;
                }
            }
        }
        goto L60;
        L37: cmx = (double) 0.;
        for (i = 2; i <= _nvar; ++i)
        {
            imin = i - 1;
            i__2 = imin;
            for (j = 1; j <= i__2; ++j)
            {
                elem = (r__1 = _cmatr.corr[LOC1(i, j)], fabs(r__1));
                if (elem > cmx)
                {
                    cmx = elem;
                }
            }
        }
        if (cmx >= cmxold)
        {
            goto L50;
        }
        cmxold = cmx;
        //cerr << "This code should not be executed\n";
        L50: ;
    }
    /*     X IS AN N X NV MATRIX OF VDW SCORES. IF THIS IS A FULL RANK CASE */
    /*     (N.GT.NV) THE RANK CORRELATION MATRIX OF X MUST BE FOUND AND */
    /*     THE CHOLESKY FACTORIZATION PERFORMED. IF THIS IS NOT A FULL RANK */
    /*     CASE (N.LE.NV) THIS MATRIX OF VDW SCORES WILL BE USED TO MIX THE */
    /*     INTERVALS OF THE RAW DATA MATRIX XX */

    goto L235;
    L60:
#ifdef DEBUG
    print_cmatrix(_cmatr.corr,"[_cmatr.corr] matrice avant Choleski");
#endif
    chlsky(_cmatr.corr, _workc.q, _nvar);

#ifdef DEBUG
    print_cmatrix(_workc.q,"[_workc.q] matrice Q associ� � E");
    print_cmatrix_root(_workc.q,"q_mat","toto");
#endif
    /*     INVERT THE LOWER TRIANGULAR MATRIX Q RESULTING FROM THE */
    /*     CHOLESKY FACTORIZATION */

    matinv(_workc.q, _nvar);

#ifdef DEBUG
    print_cmatrix(_workc.q,"[_workc.q] matrice Q-1");
    print_cmatrix_root(_workc.q,"invq_mat","toto");
#endif

    if (_ucm == 0 && (_nobv != 3 || _nvar != 2))
    {
        goto L160;
    }

    /*     IF THE USER HAS SPECIFIED HIS OWN RANK CORRELATION STRUCTURE */
    /*     THIS MATRIX MULTIPLICATION LOOP IS NEEDED */

    /* Calcul de Q-1()t * Pt */

#ifdef DEBUG
    print_cmatrix(_workc.uq,"[_workc.uq] matrice uq avant multicplication ( Q-1)t Pt");
#endif
    for (i = 0; i < _nvx; i++)
        _cmatr.corr[i] = _workc.uq[i];
#ifdef DEBUG
    print_cmatrix(_cmatr.corr,"[_cmatr.corr] Matrice apr�s matinv");
#endif
#ifdef DEBUG_ROOT
    print_cmatrix_root(_cmatr.corr,"corr_mat","toto");
#endif

    for (k = 1; k <= _nvar; ++k)
    {
        for (it = 1; it <= k; ++it)
        {
            _rank.xv[it - 1] = (double) 0.;
            for (iw = it; iw <= k; ++iw)
            {
                _rank.xv[it - 1] += _cmatr.corr[LOC1(k, iw)]
                        * _workc.q[LOC1(iw, it)];
            }
        }

        for (iw = 1; iw <= _nvar; ++iw)
        {
            _workc.s[LOC1(k, iw)] = _rank.xv[iw - 1];
        }
    }
#ifdef DEBUG
    printf("End of multiplication ( Q-1)t Pt");
    print_cmatrix(_workc.s,"[_workc.s] sortie de multiplication");
#endif
#ifdef DEBUG_ROOT
    print_cmatrix_root(_workc.s,"s_mat","toto");
#endif

    goto L180;

    /*     THIS ASSUMES THE DESIRED CORRELATION STRUCTURE IS THE IDENTITY */
    /*     MATRIX, I.E. CLOSE TO ORTHOGONAL */

    L160: for (i = 1; i <= _nvx; ++i)
    {
        _workc.s[i - 1] = _workc.q[i - 1];
    }
    L180:
#ifdef DEBUG
    print_matrix(_sampx,"[_sampx] Matrice _sampx avant multicplication");
#endif
    for (k = 1; k <= _nobv; ++k)
    {
        for (it = 1; it <= _nvar; ++it)
        {
            _rank.xv[it - 1] = (double) 0.;
            for (iw = 1; iw <= it; ++iw)
            {
                _rank.xv[it - 1] += _sampx[LOC(k, iw)] * _workc.s[LOC1(it, iw)];
            }
        }
        i__1 = _nvar;
        for (iw = 1; iw <= i__1; ++iw)
        {
            _sampx[LOC(k, iw)] = _rank.xv[iw - 1];
        }
    }
#ifdef DEBUG
    print_matrix(_sampx,"[_sampx] Matrice _sampx sortie de multiplication");
#endif    
#ifdef DEBUG_ROOT
    printf_matrix_root(_sampx,"x_mat","toto");
#endif

    /*     SINCE X NO LONGER CONTAINS INTEGERS WE MUST RANK EACH COLUMN */
    /*     IN ORDER TO GET A NEW MATRIX TO USE FOR MIXING THE INTERVALS */
    /*     OF THE RAW DATA MATRIX XX */

    L235: for (j = 1; j <= _nvar; ++j)
    {
        for (i = 1; i <= _nobv; ++i)
        {
            _rank.xv[i - 1] = _sampx[LOC(i, j)];
        }
        // Utilise les valeurs de la variable j qui ont �t� mises dans le vecteur _rank.xv
        ranker();
#ifdef DEBUG
        printf("       ****************** Variable %d\n", j);
        print_rank(_nobv);
#endif
        for (i = 1; i <= _nobv; ++i)
        {
            _sampx[LOC(i, j)] = _rank.rxv[i - 1];
        }
    }
#ifdef DEBUG
    print_matrix(_sampx,"[_sampx] Matrice _sampx sortie de ranker");
#endif

    // Saves the sampling
    for (j = 1; j <= _nvar; ++j)
        for (i = 1; i <= _nobv; ++i)
            _rank.final[LOC(i, j)] = _sampx[LOC(i, j)];
    // Calculates the Spearman matrix ( correlation matrix on the ranks)
    rankcorcal();
#ifdef DEBUG
    print_cmatrix(_cmatr.rcorr,"[_cmatr.rcorr] Rang final");
#endif

    for (j = 1; j <= _nvar; ++j)
    {
        for (i = 1; i <= _nobv; ++i)
        {
            k = (int) (_sampx[LOC(i, j)] + (double) .01);
            _sampx[LOC(i, j)] = _workxx[(j - 1) * _nobv + k];
        }
    }
#ifdef DEBUG
    print_matrix(_sampx,"[_sampx] Matrice d'experience _sampx finale");
#endif

#ifdef DEBUG_ROOT
    close_root_file("toto");
#endif

}  //  end Mix()

//**************************************************************************

double LatinHypercube::finvnor(double x)
{
    double test;
    double y;
    double f;

    test = x - .5;
    if (test < 0)
    {
        f = -1.0;
        y = x;
    }
    else if (test > 0)
    {
        f = 1.0;
        y = 1.0 - x;
    }
    else
        return 0.0;

    test = y * 2.0;
    return sqrt(2.0) * f * rierfc1(test);
}  //  end finvnor()

//**************************************************************************
void LatinHypercube::rankcorcal()
{
#ifdef DEBUG
    cerr << "LatinHypercube::rankcorcal()\n";
#endif

    long int i, j, k;
    double *xm;
    double *ssq;
    double fn, fval, fvalj;

    xm = new double[_nvar];
    ssq = new double[_nvar];
    if (_nvar == 1)
    {
        _cmatr.rcorr[0] = 1.0;
        return;
    }
    //  initialize xm and ssq to 0 for each variable
    for (i = 1; i <= _nvar; i++)
    {
        xm[i - 1] = 0.0;
        ssq[i - 1] = 0.0;
    }
    //  initialize correlation matrix to 0.0
    for (i = 0; i < _nvx; i++)
        _cmatr.rcorr[i] = 0.0;
    fn = (double) _nobv;

    //  COMPUTE THE MEAN
    for (j = 1; j <= _nvar; j++)
        for (k = 1; k <= _nobv; k++)
            xm[j - 1] += _rank.final[LOC(k, j)];
    for (j = 1; j <= _nvar; j++)
    {
        xm[j - 1] /= fn;
    }
    //  SUBTRACT THE MEAN FROM ALL OF THE ELEMENTS OF THE MATRIX AND 
    //  COMPUTE THE SUM OF SQUARES
    for (j = 1; j <= _nvar; j++)
    {
        for (k = 1; k <= _nobv; k++)
        {
            fval = _rank.final[LOC(k, j)] - xm[j - 1];
            ssq[j - 1] += fval * fval;
        }
    }

    //   COMPUTE THE CORRELATION MATRIX 
    for (i = 2; i <= _nvar; i++)
    {
        for (j = 1; j <= i - 1; j++)
        {
            for (k = 1; k <= _nobv; k++)
            {
                fval = _rank.final[LOC(k, i)] - xm[i - 1];
                fvalj = _rank.final[LOC(k, j)] - xm[j - 1];
                _cmatr.rcorr[LOC1(i, j)] += fval * fvalj;
            }
        }
    }
    for (i = 2; i <= _nvar; i++)
    {
        for (j = 1; j <= i - 1; j++)
        {
            _cmatr.rcorr[LOC1(i, j)] /= sqrt(ssq[i - 1] * ssq[j - 1]);
        }
    }
#ifdef DEBUG
    print_cmatrix(_cmatr.rcorr,"[_cmatr.rcorr] Matrice de Spearman");
#endif

    //  Set Diagonal equal to 1.0
    for (i = 1; i <= _nvar; i++)
        _cmatr.rcorr[LOC1(i, i)] = 1.0;
    delete[] ssq;
    delete[] xm;
}
//**************************************************************************
void LatinHypercube::corcal()
{
#ifdef DEBUG
    cerr << "LatinHypercube::corcal()\n";
#endif

    long int i, j, k;
    double *xm;
    double *ssq;
    double fn;

    xm = new double[_nvar];
    ssq = new double[_nvar];
    if (_nvar == 1)
    {
        _cmatr.corr[0] = 1.0;
        return;
    }
    //  initialize xm and ssq to 0 for each variable
    for (i = 1; i <= _nvar; i++)
    {
        xm[i - 1] = 0.0;
        ssq[i - 1] = 0.0;
    }
    //  initialize correlation matrix to 0.0
    for (i = 0; i < _nvx; i++)
        _cmatr.corr[i] = 0.0;
    fn = (double) _nobv;

    //  COMPUTE THE MEAN
    for (j = 1; j <= _nvar; j++)
        for (k = 1; k <= _nobv; k++)
            xm[j - 1] += _sampx[(j - 1) * _nobv + k - 1];
    for (j = 1; j <= _nvar; j++)
    {
        xm[j - 1] /= fn;
    }
    //  SUBTRACT THE MEAN FROM ALL OF THE ELEMENTS OF THE MATRIX AND 
    //  COMPUTE THE SUM OF SQUARES
    for (j = 1; j <= _nvar; j++)
    {
        for (k = 1; k <= _nobv; k++)
        {
            _sampx[(j - 1) * _nobv + k - 1] -= xm[j - 1];
            ssq[j - 1] += _sampx[(j - 1) * _nobv + k - 1]
                    * _sampx[(j - 1) * _nobv + k - 1];
        }
    }

    //   COMPUTE THE CORRELATION MATRIX 
    for (i = 2; i <= _nvar; i++)
    {
        for (j = 1; j <= i - 1; j++)
        {
            for (k = 1; k <= _nobv; k++)
            {
                _cmatr.corr[j + (i * i - i) / 2 - 1] += _sampx[(i - 1) * _nobv
                        + k - 1] * _sampx[(j - 1) * _nobv + k - 1];
            }
        }
    }
    for (i = 2; i <= _nvar; i++)
    {
        for (j = 1; j <= i - 1; j++)
            _cmatr.corr[j + (i * i - i) / 2 - 1] /= sqrt(
                    ssq[i - 1] * ssq[j - 1]);
    }

    //  Set Diagonal equal to 1.0
    for (i = 1; i <= _nvar; i++)
        _cmatr.corr[i + (i * i - i) / 2 - 1] = 1.0;
    delete[] ssq;
    delete[] xm;
}  //  end corcal

//**************************************************************************

void LatinHypercube::chlsky(double *cm, double *q, long int size)
{
//  Description:  performs cholesky factorization.
//
//      Input:  cm             (array to be factored )
//      Output: q                (factored array) 
//      size i dim of array (example 4 X 4 array, size = 4)
//
//      arrays are in the form
//              example of 3 X 3 array, size = 3
//  input is a symetric matrix
//      cm[0] = the [1,1] element
//      cm[1] = the [2,1] and [1,2] elements
//      cm[2] = the [2,2] element
//      cm[3] = the [3,1] and [1,3] elements
//      cm[4] = the [3,2] and [2,3] elements
//      cm[5] = the [3,3] element
//
//  ouput is a lower trianglular matrix
//      q[0] = [1,1]
//      q[1] = [2,1]
//      q[2] = [2,2]
//      q[3] = [3,1]
//      q[4] = [3,2]
//      q[5] = [3,3]

#ifdef DEBUG
    cerr << "LatinHypercube::chlsky()\n";
#endif

    long int i1, i2, k, imin;
    double r1;

    //  initialize q to 0
    for (i1 = 1; i1 <= _nvx; i1++)
    {
        q[i1 - 1] = 0.0;
    }

    //  Set first row of q = to first row of cm
    for (i2 = 1; i2 <= size; i2++)
    {
        q[LOC1(i2, 1)] = cm[LOC1(i2, 1)];
    }

    //  Set all other values
    for (i1 = 2; i1 <= size; i1++)
    {
        imin = i1 - 1;
        for (i2 = 1; i2 <= imin; i2++)
        {
            r1 = q[LOC1(i1, i2)];
            q[LOC1(i1, i1)] += r1 * r1;
        }
        // Set diagonal values 
//    cerr << "q[" << LOC1(i1,i1) << "] = (sqrt(1.0 - (q[" << LOC1(i1,i1);
//    cerr << "]=" << q[LOC1(i1,i1)] << ") = ";

        q[LOC1(i1, i1)] = sqrt((double) 1.0 - q[LOC1(i1, i1)]);

//    cerr << q[LOC1(i1,i1)] << '\n';

        if (i1 < size)
        {
            //  Set All other off diagonal values
            for (i2 = i1 + 1; i2 <= size; i2++)
            {
                for (k = 1; k <= imin; k++)
                {
                    q[LOC1(i2, i1)] += q[LOC1(i2, k)] * q[LOC1(i1, k)];

//          cerr << "q[" << LOC1(i2,i1) << "] += " << "q["<< LOC1(i2,k) << "]";
//          cerr << "*q[" << LOC1(i1,k) << "]" << '\n';
                }
//        cerr << "q'[" << LOC1(i2,i1) << "] = (cm[" << LOC1(i2,i1);
//        cerr << "]=" << cm[LOC1(i2,i1)] << ") - (q[" << LOC1(i2,i1);
//        cerr << "]=" << q[LOC1(i2,i1)] << ") / (q[" << LOC1(i1,i1);
//        cerr << "]=" << q[LOC1(i1,i1)] << ") = ";

                q[LOC1(i2, i1)] = (cm[LOC1(i2, i1)] - q[LOC1(i2, i1)])
                        / q[LOC1(i1, i1)];

//        cerr << q[LOC1(i2,i1)] << '\n';
            }
        }  // end if
    }
}  //  end chlsky

//**************************************************************************

void LatinHypercube::matinv(double *cm, long int size)
{
    (void)(size);
//  Description:  This method is used to invert a Lower Triangular matrix
//
//      Input:  cm             (array to be factored )
//      size i dim of array (example 4 X 4 array, size = 4)
//
//      arrays are in the form
//              example of 3 X 3 array, size = 3
//  input is a symetric matrix
//      cm[0] = the [1,1] element
//      cm[1] = the [2,1] and [1,2] elements
//      cm[2] = the [2,2] element
//      cm[3] = the [3,1] and [1,3] elements
//      cm[4] = the [3,2] and [2,3] elements
//      cm[5] = the [3,3] element
//
//  ouput is a lower trianglular matrix
//      q[0] = [1,1]
//      q[1] = [2,1]
//      q[2] = [2,2]
//      q[3] = [3,1]
//      q[4] = [3,2]
//      q[5] = [3,3]

#ifdef DEBUG
    cerr << "LatinHypercube::matinv()\n";
#endif

    double temp;
    long int i, j, k, jplus;

    for (i = 1; i <= _nvar; i++)
        cm[LOC1(i, i)] = 1.0 / cm[LOC1(i, i)];

    for (k = _nvar; k > 1; k--)
    {
        for (j = k - 1; j >= 1; j--)
        {
            temp = 0.0;
            jplus = j + 1;
            for (i = jplus; i <= k; i++)
                temp += cm[LOC1(k, i)] * cm[LOC1(i, j)];
            cm[LOC1(k, j)] = -(double) temp * cm[LOC1(j, j)];
        }
    }
}

//**************************************************************************

void LatinHypercube::ranker()
{
//   Description: This method ranks the samples

#ifdef DEBUG
    cerr << "LatinHypercube::ranker()\n";
#endif

    long int i, j, k;
    double rhold, xhold, avg;
    int nties;

    for (i = 0; i < _nobv; i++)
        _rank.rxv[i] = i;

    hpsrt();
    for (i = 0; i < _nobv; i++)
    {
        _rank.iwk[i] = (int) _rank.rxv[i];
        _rank.rxv[i] = i;
    }

    //  FIND TIES
    i = 0;
    L30: i++;
    if (i >= _nobv)
        goto L80;
    if (_rank.xv[i - 1] != _rank.xv[i])
        goto L30;

    //  COUNT TIES
    nties = 2;
    L40: if (i + nties > _nobv)
        goto L50;
    if (_rank.xv[i - 1] != _rank.xv[i + nties - 1])
        goto L50;
    ++nties;
    goto L40;

    //  AVERAGE TIED RANKS
    L50: avg = 0.0;
    for (j = 1; j <= nties; ++j)
        avg += _rank.rxv[i + j - 2];
    avg /= (double) nties;
    --i;
    for (j = 1; j <= nties; ++j)
    {
        ++i;
        _rank.rxv[i - 1] = avg;
    }
    goto L30;
    L80:

    //  REORDER
    i = 0;
    L90: k = _rank.iwk[i];
    if (k != i)
        goto L100;
    ++i;
    if (i < _nobv)
        goto L90;
    return;

    L100: xhold = _rank.xv[i];
    rhold = _rank.rxv[i];
    _rank.xv[i] = _rank.xv[k];
    _rank.rxv[i] = _rank.rxv[k];
    _rank.xv[k] = xhold;
    _rank.rxv[k] = rhold;
    _rank.iwk[i] = _rank.iwk[k];
    _rank.iwk[k] = k;
    goto L90;
}  //  end ranker()

//**************************************************************************
void LatinHypercube::LOG(const char *texte)
{
    std::cout << std::endl << "__LINE__[" << __LINE__ << "] type[";
    if (_sampletype)
        std::cout << "SRS";
    else
        std::cout << "LHS";
    printf("] - Texte[%s]\n", texte);

}
//**************************************************************************
void LatinHypercube::hpsrt()
/* SUBROUTINE HPSRT IS USED IN THE RANKING OF THE SAMPLED DATA */
{
    long int i, j, l, r, temp;
    double xhold, yhold;

    l = _nobv / 2 + 1;
    r = _nobv;
    L10: if (l <= 1)
        goto L70;
    --l;
    xhold = _rank.xv[l - 1];
    yhold = _rank.rxv[l - 1];
    L20: j = l;
    L30: i = j;
    j <<= 1;
    if ((temp = j - r) < 0)
    {
        goto L40;
    }
    else if (temp == 0)
    {
        goto L50;
    }
    else
    {
        goto L60;
    }
    L40: if (_rank.xv[j - 1] < _rank.xv[j])
    {
        ++j;
    }
    L50: if (xhold >= _rank.xv[j - 1])
    {
        goto L60;
    }
    _rank.xv[i - 1] = _rank.xv[j - 1];
    _rank.rxv[i - 1] = _rank.rxv[j - 1];
    goto L30;
    L60: _rank.xv[i - 1] = xhold;
    _rank.rxv[i - 1] = yhold;
    goto L10;
    L70: xhold = _rank.xv[r - 1];
    yhold = _rank.rxv[r - 1];
    _rank.xv[r - 1] = _rank.xv[0];
    _rank.rxv[r - 1] = _rank.rxv[0];
    --r;
    if (r > 1)
    {
        goto L20;
    }
    _rank.xv[0] = xhold;
    _rank.rxv[0] = yhold;
    return;
} /* hpsrt_ */

//**************************************************************************
/* **************************************************************** */
/* FUNCTION RIERFC1 IS USED IN GENERATING THE NORMAL AND */
/* LOGNORMAL DISTRIBUTIONS */

double LatinHypercube::rierfc1(double y)
{
    /* Initialized data */

    double c1 = (double) .235777520630369;
    double c2 = (double) 1.35777520630369;
    struct
    {
        double e_1[66];
    } equiv_2 =
	  {{ .918725611735013, 0., .0168792878000327, 0., 6.603371390583e-4, 0.,
            3.2020384983938e-5, 0., 1.72060607522481e-6, 0.,
            9.81965971588191e-8, 0., 5.83049613537653e-9, 0.,
            3.56019351836136e-10, 0., 2.21968915783128e-11, 0.,
            1.40639693109741e-12, 0., 9.02597345404862e-14, 0.,
            1.54701109458613, -.331460331083896, .043300112409006,
            -.0106564004165532, .00290613542304156, -8.61872838022491e-4,
            2.67933751795053e-4, -8.60838893942933e-5, 2.83232058814598e-5,
            -9.48870819734494e-6, 3.22422655069385e-6, -1.10815778472076e-6,
            3.84464770797987e-7, -1.34439275565208e-7, 4.73255976052393e-8,
            -1.67556011100019e-8, 5.96199003969093e-9, -2.13070503291886e-9,
            7.64427040920545e-10, -2.75198005584737e-10, 9.93792246090789e-11,
            -3.59877382902119e-11, 11.0642888011036, 4.34299147561447,
            -.0233781774969295, .00423345215362947, 8.68757084192089e-6,
            -5.98261113270881e-4, 4.50490139240298e-4, -2.54858131942102e-4,
            1.2782418926134e-4, -5.97873878043957e-5, 2.66474012012582e-5,
            -1.14381836209267e-5, 4.75393030377615e-6, -1.9175958992961e-6,
            7.50806465594834e-7, -2.84791180387123e-7, 1.04187791696225e-7,
            -3.64567243689145e-8, 1.2012929613903e-8, -3.61030126779729e-9,
	     9.12356140081759e-10, -1.36851363400914e-10 }};

    /* Format strings */
//    static char fmt_60[] = "(\002 Y LESS THAN EXP(-81.) OR Y GREATER THAN \1.0\002)";
    /* System generated locals */
    double ret_val = (double) 0.0;

    /* Builtin functions */
//    double log(), sqrt();
//    int s_wsfe(), e_wsfe();
    /* Local variables */
    static double temp;
#define a ((double *)&equiv_2)
    static double d;
    static int j, k, l;
    static double w;
#define a1 ((double *)&equiv_2)
#define a2 ((double *)&equiv_2 + 22)
#define a3 ((double *)&equiv_2 + 44)
    static double td, vn, vnp1;

    /* Fortran I/O blocks */
//    static cilist io___40 = { 0, 6, 0, fmt_60, 0 };

    if (y < (double) 6.63967719958073e-36 || y > (double) 1.)
    {
        goto L50;
    }
    if (y >= (double) .5)
    {
        goto L10;
    }
    if (y >= (double) .1)
    {
        goto L20;
    }
    j = 3;
    w = sqrt(-(double) log(y));
    d = c1 * w - c2;
    goto L30;
    L10: j = 1;
    d = (double) 1. - y;
    d += d;
    goto L30;
    L20: j = 2;
    d = y * (double) 5. - (double) 1.5;
    L30: td = d + d;
    vnp1 = (double) 0.;
    vn = (double) 0.;
    for (l = 1; l <= 21; ++l)
    {
        k = 22 - l + 1;
        temp = vn;
        vn = td * vn - vnp1 + a[k + j * 22 - 23];
        /* L40: */
        vnp1 = temp;
    }
    ret_val = d * vn - vnp1 + a[j * 22 - 22] * (double) .5;
    if (j == 1)
    {
        ret_val = d * ret_val;
    }
    return ret_val;
    L50:
//    s_wsfe(&io___40);
//    e_wsfe();
    return ret_val;

} /* end rierfc1_ */

#undef a3
#undef a2
#undef a1
#undef a

//***********************************************************************

void LatinHypercube::print_matrix_to_file(double *mat, const char *file)
{
//   Description:  This method is used to print the _nvar X _nobv
//      matrix to a file
//   Input
//      mat:  pointer to _nvar X _nobv matrix
//      file: name of file to print matrix in

    //Tcl_Channel chan;
    char number[100];
    long int i, j;

    //chan = Tcl_OpenFileChannel(_interp, (char *)file, "w", 0644);
    std::ofstream fileout(file);
    if (!fileout)
    {
        std::cerr << "Cannot open <" << file << "> for output\n";
    }
    for (j = 1; j <= _nvar; j++)
    {
        sprintf(number, "     variable %2ld ", j);
        //Tcl_Write(chan,number,-1);
        fileout << number << std::endl;
    }
    //Tcl_Write(chan,"\n",-1);
    for (j = 1; j <= _nvar; j++)
    {
        sprintf(number, "-----------------");
        //Tcl_Write(chan,number,-1);
        fileout << number << std::endl;
    }
    //Tcl_Write(chan,"\n",-1);
    for (i = 1; i <= _nobv; i++)
    {
        for (j = 1; j <= _nvar; j++)
        {
            //      sprintf(number,"%16f ",mat[(j-1)*_nobv+i-1]);
            sprintf(number, "%16e ", mat[(j - 1) * _nobv + i - 1]);
            //Tcl_Write(chan,number,-1);
            fileout << number << std::endl;
        }
        //Tcl_Write(chan,"\n",-1);
    }
    //Tcl_Close(NULL,chan);
    fileout.close();
}

/*! \fn void  LatinHypercube::print_Output(const char *filename)
 \brief Saves the sampling file
 
 This method prints out results data (i.e. all
 sample values, statistical data about the samples, the
 correlation matrix used, .. etc.)
 
 \param filename  (const char *) Datasave file
 
 */

void LatinHypercube::print_Output(const char *filename)
{
#ifdef DEBUG
    cerr << "LatinHypercube::print_Output()\n";
#endif

    char line[1000];
    // Ouverture du fichier
    std::ofstream fileout(filename);
    if (!fileout)
    {
        std::cerr << "Cannot open <" << filename << "> for output\n";
    }

    fileout << "\n\t******** Latin Hypercube Data Log ********\n"
            << "\t------------------------------------------\n";
    sprintf(line, "%-40.40s %10ld\n", "Number of variables:", _nvar);
    fileout << line;
    sprintf(line, "%-40.40s %10ld\n", "Number of samples:", _nobv);
    fileout << line;
    sprintf(line, "%-40.40s %10d\n", "Random number seed:", _originalseed);
    fileout << line;
    fileout << "\n\n\t\tVariables\n" << "\t\t---------\n\n";
    // Param1 = Mean for all but:    Weibull (P1=alpha)
    // Param2 = Std Dev for all but: Weibull (P2=beta), Lognormal (P2=err factor)
    sprintf(line, "%-10s %-10s %12s %12s %12s %12s\n", "Name", "PDF",
            "Param1/Mean", "Param2/Std. Dev.", "Lower Limit", "Upper Limit");
    fileout << line << "------------------------------------------------------"
            << "-------------------\n";
    long int i = 1;
    Listelem *cur;
    LHSVariable *varptr;
    for (cur = _InPtr->GetListPtr(); cur; cur = cur->Next)
    {
        varptr = cur->Data;
        switch (varptr->GetType())
        {
        case NORMAL:
            sprintf(line, "%-10s %-10s %12.4f %12.4f %12.4s %12.4s",
                    cur->Data->GetName(), "NORMAL",
                    ((LHSNormal *) varptr)->Get_Mean(),
                    ((LHSNormal *) varptr)->Get_Variation(), "N/A", "N/A");
            fileout << line << std::endl;
            break;
        case LOGNORMAL:
            sprintf(line, "%-10s %-10s %12.4f %12.4f %12.4s %12.4s",
                    cur->Data->GetName(), "LOGNORMAL",
                    ((LHSLognormal *) varptr)->Get_Mean(),
                    ((LHSLognormal *) varptr)->Get_Variation(), "N/A", "N/A");
            fileout << line << std::endl;
            break;
        case CONSTANT:
            sprintf(line, "%-10s %-10s %12.4f %12.4f %12.4s %12.4s",
                    cur->Data->GetName(), "CONSTANT",
                    ((LHSConstant *) varptr)->Get_C(), 0.0, "N/A", "N/A");
            fileout << line << std::endl;
            break;
        case UNIFORM:
            sprintf(line, "%-10s %-10s %12.4s %12.4s %12.4f %12.4f",
                    cur->Data->GetName(), "UNIFORM", "N/A", "N/A",
                    ((LHSUniform *) varptr)->Get_A(),
                    ((LHSUniform *) varptr)->Get_B());
// MSE, 4/2/00: Bad output in UNIFORM case
//                ,((LHSUniform *)varptr)->Get_Mean()
//                ,((LHSUniform *)varptr)->Get_Variation()
//                ,"N/A"
//                ,"N/A");
            fileout << line << std::endl;
            break;
        case TRIANGULAR:
            sprintf(line, "%-10s %-10s %12.4s %12.4s %12.4f %12.4f",
                    cur->Data->GetName(), "TRIANGULAR", "N/A", "N/A",
                    ((LHSTriangular *) varptr)->Get_A(),
                    ((LHSTriangular *) varptr)->Get_B());
            fileout << line << std::endl;
            break;
        case LOGTRIANGULAR:
            sprintf(line, "%-10s %-10s %12.4s %12.4s %12.4f %12.4f",
                    cur->Data->GetName(), "LOGTRIANGULAR", "N/A", "N/A",
                    ((LHSTriangular *) varptr)->Get_A(),
                    ((LHSTriangular *) varptr)->Get_B());
            fileout << line << std::endl;
            break;
        case LOGUNIFORM:
            sprintf(line, "%-10s %-10s %12.4s %12.4s %12.4f %12.4f",
                    cur->Data->GetName(), "LOGUNIFORM", "N/A", "N/A",
                    ((LHSLoguniform *) varptr)->Get_A(),
                    ((LHSLoguniform *) varptr)->Get_B());
// MSE, 4/2/00: Bad output in LOGUNIFORM case
//                ,((LHSLoguniform *)varptr)->Get_Mean()
//                ,((LHSLoguniform *)varptr)->Get_Variation()
//                ,"N/A"
//                ,"N/A");
            fileout << line << std::endl;
            break;
        case WEIBULL:
            sprintf(line, "%-10s %-10s %12.4f %12.4f %12.4s %12.4s",
                    cur->Data->GetName(), "WEIBULL",
                    ((LHSWeibull *) varptr)->Get_A(),
                    ((LHSWeibull *) varptr)->Get_B(), "N/A", "N/A");
            fileout << line << std::endl;
            break;
        default:
            fileout << "ERROR! Variable type is default!\n" << std::endl;
            break;
        }  //  end switch
//    fileout << '\n';
        sprintf(line, "%-10s %-10s %12.4f %12.4f %12.4f %12.4f", " ",
                "Computed", _sinfo[i].mean, _sinfo[i].std, _sinfo[i].min,
                _sinfo[i].max);
        fileout << line << std::endl;
        i++;
    }  //  end for

    fileout.close();

    if (_ucm)
    {
        printf_cmatrix(_cmatr.ucorr, "User Input Correlation Matrix", filename);
    }

    printf_cmatrix(_cmatr.corr, "Correlations among sample vectors", filename);
    printf_cmatrix(_cmatr.rcorr, "Correlations among rank vectors", filename);
    printf_matrix(_sampx, "Latin Hypercube Sample vectors", filename);
    printf_matrix(_rank.final, "Latin Hypercube Sample ranking order",
            filename);
}

//! Returns the list of the variables names. 
/*!  The list of names is constitued from:
 \li names of treated variables
 \li eventually the name of the rank variables

 These informations are mentioned if the field  \ref _saveRang is set to 1.
 In this case, the variable name is constitued from string <b>Rang</b>_<em>NomVariable</em>.
 Furthermore, the first variable is the index of experience. Its variable name is iter
 
 \return List of the variables names.
 */
#ifdef root
char * LatinHypercube::GetVariablesName()
{
    char line[2000];
    Listelem *cur;
    // Nom de la variable des indices 
    sprintf(line,"#iter");

    // Nom des variables
    for (cur = _InPtr->GetListPtr(); cur; cur = cur->Next)
    sprintf(line,"%s:%s", line, cur->Data->GetName());
    // Si sauvegarde des rangs, nom des rangs de variable
    if ( _saveRang )
    for (cur = _InPtr->GetListPtr(); cur; cur = cur->Next)
    sprintf(line,"%s:rang_%s", line, cur->Data->GetName());

    // Returns the string
    return line;
}
#endif
//***********************************************************************
void LatinHypercube::printf_matrix_dlhs(const char *filename)
{
    char line[1000];
    long int i, j;
    Listelem *cur;

    std::ofstream fileout(filename);
    if (!fileout)
    {
        std::cerr << "Cannot open <" << filename << "> for output\n";
    }

    // ligne de titres des variables
    fileout << "#iter";
    // Nom des variables
    for (cur = _InPtr->GetListPtr(); cur; cur = cur->Next)
        fileout << ", " << cur->Data->GetName();
    // Si sauvegarde des rangs, nom des rangs de variable
    if (_saveRang)
        for (cur = _InPtr->GetListPtr(); cur; cur = cur->Next)
            fileout << ", rang_" << cur->Data->GetName();
    fileout << " " << std::endl;

    // Boucle sur le nombre d'observations
    for (i = 1; i <= _nobv; i++)
    {
        sprintf(line, "%ld ", i);
        fileout << line;
        for (j = 1; j <= _nvar; j++)
        {
            sprintf(line, "%14e ", _sampx[(j - 1) * _nobv + i - 1]);
            fileout << line;
        }
        if (_saveRang)
            for (j = 1; j <= _nvar; j++)
            {
                sprintf(line, "%f ", _rank.final[LOC(i, j)]);
                fileout << line;
            }
        fileout << '\n';
    }
    fileout.close();
}

void LatinHypercube::printf_matrix(double *mat, const char *title,
        const char *filename)
{
//   Description:  This method is called by print_Output(). It prints
//      a matrix of size _nvar X _nobv to the file defined in the 
//      print_Output() method.
//   Input 
//      mat:  pointer to matrix
//      title:  character string to be printed as the title of the 
//        matrix.

    //extern Tcl_Channel chan;
    char line[1000];
    long int i, j;

    std::ofstream fileout(filename, std::ios::app);
    if (!fileout)
    {
        std::cerr << "Cannot open <" << filename << "> for output\n";
    }

    fileout << "\n\t";
    fileout << title;
    fileout << "\n\n";

    sprintf(line, "%-9.9s", "obv");
    fileout << line;
    for (i = 1; i <= _nvar; i++)
    {
        sprintf(line, "variable #%-2ld", i);
        fileout << line;
    }
    fileout << '\n';
    sprintf(line,
            "---------------------------------------------------------------\n");
    fileout << line;

    for (i = 1; i <= _nobv; i++)
    {
        sprintf(line, "%4ld ", i);
        fileout << line;
        for (j = 1; j <= _nvar; j++)
        {
            sprintf(line, " %14e", mat[(j - 1) * _nobv + i - 1]);
            fileout << line;
        }
        fileout << '\n';
    }
    fileout.close();
}

//***********************************************************************

void LatinHypercube::print_matrix(double *mat, char *title)
{
//   Description:  It prints a matrix of size _nvar X _nobv to 
//      standard output.
//   Input
//      mat:  pointer to matrix
//      title:  character string to be printed as the title of the  
//        matrix.
    long int i;

    std::cout << std::endl << "LatinHypercube::print_matrix" << std::endl;

    std::cerr << "\n\t" << title << '\n' << std::endl;

    printf("%-9.9s", "obv");
    for (i = 1; i <= _nvar; i++)
        printf("variable #%-2ld", i);
    printf("\n");
    printf("---------------------------------------------------------------\n");

    for (i = 1; i <= _nobv; i++)
    {
        printf("%4ld ", i);
        for (long int j = 1; j <= _nvar; j++)
            printf(" %16e", mat[(j - 1) * _nobv + i - 1]);
        printf("\n");
    }
}

void LatinHypercube::printlog()
{

    std::cerr << std::endl << "******************************************" << std::endl;
    std::cerr << "LatinHypercube\n";
    std::cerr << "  Number of Variables = " << _nvar << '\n';
    std::cerr << "  Number of Observations = " << _nobv << '\n';
    std::cerr << "  Seed = " << _originalseed << '\n';
    std::cerr << "  method : ";
    if (_sampletype)
        std::cerr << " SRS\n";
    else
        std::cerr << "LHS\n";
    std::cerr << std::endl << "******************************************" << std::endl;

}

//***********************************************************************

void LatinHypercube::print_rank(long int num)
{
//   Description:  This method prints the rank matrixs to standard output.
//      Use for debugging.
    std::cerr << "struct rank\n\n";
    printf("%6.6s %8.8s %8.8s %8.8s\n", "#", "xv", "rxv", "iwk");
    std::cerr << "-----------------------------------\n";
    for (long int i = 0; i < num; i++)
    {
        printf("[%2ld] = %8.4f %8.4f %8d\n", i, _rank.xv[i], _rank.rxv[i],
                _rank.iwk[i]);
    }
}

//***********************************************************************

void LatinHypercube::printf_cmatrix(double *corr, const char * title,
        const char *filename)
{
//   Description:  This method is called by print_Output(). It prints
//      a correlation matrix to the file defined in the
//      print_Output() method.
//   Input
//      corr:  pointer to matrix
//      title:  character string to be printed as the title of the
//        matrix.

    //extern Tcl_Channel chan;
    char line[1000];
    long int i, k;

    std::ofstream fileout(filename, std::ios::app);
    if (!fileout)
    {
        std::cerr << "Cannot open <" << filename << "> for output\n";
    }

    fileout << "\n\t";
    fileout << title << std::endl;
    sprintf(line, "-------------------------------------------\n");
    fileout << line;
    for (k = 1; k <= _nvar; ++k)
    {
        sprintf(line, "%ld ", k);
        fileout << line;
        for (i = 1; i <= k; i++)
        {
            sprintf(line, "%8.4f ", corr[i + (k * k - k) / 2 - 1]);
            fileout << line;
        }
        fileout << '\n';
    }
    //Tcl_Write(chan,"   ",-1);
    fileout << "   " << std::endl;
    for (k = 1; k <= _nvar; ++k)
    {
        sprintf(line, "%8ld ", k);
        fileout << line;
    }
    fileout << "\n\n";
    fileout.close();
}
//***********************************************************************
void LatinHypercube::print_cmatrix_root(double *corr, char * title,
        const char *filename)
{
    long int i, k;

    char line[100];
    sprintf(line, "%s.C", filename);

    std::ofstream fileout(line, std::ios::app);
    if (!fileout)
    {
        std::cerr << "Cannot open <" << filename << "> for output\n";
    }

    fileout << std::endl;
    fileout << "TMatrix " << title << "(" << _nvar << "," << _nvar << ");"
            << std::endl;
    fileout << title << ".Zero();" << std::endl;
    for (k = 1; k <= _nvar; ++k)
    {
        for (i = 1; i <= k; i++)
        {
            fileout << title << "(" << k - 1 << "," << i - 1 << ") = "
                    << corr[i + (k * k - k) / 2 - 1] << ";" << std::endl;
        }
    }
    fileout << "std::cout << \"vecteur [" << title << "]\";" << std::endl;
    fileout << title << ".Print();" << std::endl;
    fileout.close();

}
//***********************************************************************
void LatinHypercube::init_root_file(const char *filename)
{

    char line[100];

    sprintf(line, "%s.C", filename);

    std::ofstream fileout(line, std::ios::out);
    if (!fileout)
    {
        std::cerr << "Cannot open <" << filename << "> for output\n";
    }
    fileout << "void " << filename << "() {" << std::endl;

    fileout.close();
}
//***********************************************************************
void LatinHypercube::close_root_file(const char *filename)
{

    char line[100];

    sprintf(line, "%s.C", filename);

    std::ofstream fileout(line, std::ios::app);
    if (!fileout)
    {
        std::cerr << "Cannot open <" << filename << "> for output\n";
    }
    fileout << std::endl << "}" << std::endl;

    fileout.close();
}

//***********************************************************************
void LatinHypercube::printf_matrix_root(double *mat, char * title,
        const char *filename)
{
    long int i, j;

    char line[100];

    sprintf(line, "%s.C", filename);

    std::ofstream fileout(line, std::ios::app);
    if (!fileout)
    {
        std::cerr << "Cannot open <" << filename << "> for output\n";
    }
    fileout << std::endl << "TMatrix " << title << "(" << _nobv << "," << _nvar
            << ");" << std::endl;

    for (i = 1; i <= _nobv; i++)
    {
        for (j = 1; j <= _nvar; j++)
        {
            fileout << title << "(" << i - 1 << "," << j - 1 << ") = "
                    << mat[(j - 1) * _nobv + i - 1] << ";" << std::endl;
        }
    }
    fileout << "std::cout << \"matrice [" << title << "]\";" << std::endl;
    fileout << title << ".Print();" << std::endl;

}

//***********************************************************************
void LatinHypercube::print_cmatrix(double *corr, char * title)
{
    //   Description:  This method prints a correlation matrix to
    //      stantard output. Used for debugging.
    long int i, k;

    std::cout << "\n\t" << title << '\n';
    printf("-------------------------------------------\n");
    for (k = 1; k <= _nvar; ++k)
    {
        printf("%ld ", k);
        for (i = 1; i <= k; i++)
            printf("%8.4f ", corr[i + (k * k - k) / 2 - 1]);
        printf("\n");
    }
    printf("  ");
    for (k = 1; k <= _nvar; ++k)
        printf("%8ld ", k);
    printf("\n");
    printf("\n");
}

//*************************************************************************

void LatinHypercube::compute_Stats()
{
//   Description:  This method calculate statistical values about each
//      sample. The values are later printed to the plot file printed by 
//      the method Output_Plot_Data which is used for plotting. It is also
//      printed out in the results file generated by the method
//      print_Output()
//      The following values are generated for each sample
//        mean
//        coefficient of variation
//        minimum sample value
//        maximum sample value
//        standard deviation

#ifdef DEBUG
    cerr << "LatinHypercube:: compute_Stats()\n";
#endif

    long int i, j, k;
    double minval, maxval, sum, sum2, value;
    double *xm;
    double *ssq;
    double fn, fval, fvalj;
    Listelem *cur;
//  LHSVariable *varptr;
    char line[100];

    // Statistiques univari�es
    for (cur = _InPtr->GetListPtr(); cur; cur = cur->Next)
    {
        //    varptr = cur->Data;
        j = cur->Data->GetId();
        sum = 0.0;
        sum2 = 0.0;
        minval = _sampx[(j - 1) * _nobv + 1 - 1];
        maxval = _sampx[(j - 1) * _nobv + 1 - 1];
        for (i = 1; i <= _nobv; i++)
        {
            value = _sampx[(j - 1) * _nobv + i - 1];
            sum += value;
            sum2 += value * value;
            minval = (value < minval) ? value : minval;
            maxval = (value > maxval) ? value : maxval;
        }
        _sinfo[j].min = minval;
        _sinfo[j].max = maxval;
        _sinfo[j].mean = sum / _nobv;

        // check for constant distribution type
        _sinfo[j].std =
                (sum2 / _nobv) > (_sinfo[j].mean * _sinfo[j].mean) ?
                        sqrt(
                                (sum2 / _nobv)
                                        - (_sinfo[j].mean * _sinfo[j].mean)) :
                        0.;
        //    _sinfo[j].std = sqrt( (sum2/_nobv)-(_sinfo[j].mean*_sinfo[j].mean) );
        _sinfo[j].k = _sinfo[j].std / _sinfo[j].mean;
        sprintf(line, "var %3ld %10.6f %10.6f %10.6f %10.6f %10.6f", j,
                _sinfo[j].min, _sinfo[j].max, _sinfo[j].mean, _sinfo[j].std,
                _sinfo[j].k);
        //cerr << line << '\n';
    }

    //
    // Calcul de la matrice de corr�lation de l'�chantillon
    //
    // iden que corcal mais pas de centrage des donn�es
    if (_nvar == 1)
    {
        _cmatr.corr[0] = 1.0;
        return;
    }

    // Allocation des vecteurs d'infos
    xm = new double[_nvar];
    ssq = new double[_nvar];

    //  initialize xm and ssq to 0 for each variable
    for (i = 1; i <= _nvar; i++)
    {
        xm[i - 1] = 0.0;
        ssq[i - 1] = 0.0;
    }
    //  initialize correlation matrix to 0.0
    for (i = 0; i < _nvx; i++)
        _cmatr.corr[i] = 0.0;
    fn = (double) _nobv;

    //  COMPUTE THE MEAN
    for (j = 1; j <= _nvar; j++)
        for (k = 1; k <= _nobv; k++)
            xm[j - 1] += _sampx[LOC(k, j)];
    for (j = 1; j <= _nvar; j++)
    {
        xm[j - 1] /= fn;
    }
    //  SUBTRACT THE MEAN FROM ALL OF THE ELEMENTS OF THE MATRIX AND 
    //  COMPUTE THE SUM OF SQUARES
    for (j = 1; j <= _nvar; j++)
    {
        for (k = 1; k <= _nobv; k++)
        {
            fval = _sampx[LOC(k, j)] - xm[j - 1];
            ssq[j - 1] += fval * fval;
        }
    }
    //   COMPUTE THE CORRELATION MATRIX 
    for (i = 2; i <= _nvar; i++)
    {
        for (j = 1; j <= i - 1; j++)
        {
            for (k = 1; k <= _nobv; k++)
            {
                fval = _sampx[LOC(k, i)] - xm[i - 1];
                fvalj = _sampx[LOC(k, j)] - xm[j - 1];
                _cmatr.corr[LOC1(i, j)] += fval * fvalj;
            }
        }
    }
    for (i = 2; i <= _nvar; i++)
    {
        for (j = 1; j <= i - 1; j++)
        {
            _cmatr.corr[LOC1(i, j)] /= sqrt(ssq[i - 1] * ssq[j - 1]);
        }
    }
    //  Set Diagonal equal to 1.0
    for (i = 1; i <= _nvar; i++)
        _cmatr.corr[LOC1(i, i)] = 1.0;
#ifdef DEBUG
    print_cmatrix(_cmatr.corr,"[_cmatr.corr] Matrice de corr�lation de l'�chantillon");
#endif

    delete[] ssq;
    delete[] xm;
}

/*! \fn void  LatinHypercube::Output_Plot_Data(const char *filename)
 \brief Prints in a file the list of the points in order to visualize easier.
 
 This method prints the data needed for plotting the lhs samples and histograms of the samples.

 The output formalism is as following
 \verbatim
 _nvar _nobs
 Min(var1)
 Max(var1)
 val(1,1)
 val(1,2)
 ...
 val(1,_nobs)
 Min(var2)
 Max(var2)
 val(2,1)
 val(2,2)
 ...  \endverbatim
 
 \param filename  (char *) Name of the output file
 
 */
void LatinHypercube::Output_Plot_Data(const char *filename)
{
#ifdef DEBUG
    cerr << "LatinHypercube::Output_Plot_Data()\n";
#endif

    long int i, k;
    char line[100];

    std::ofstream fileout(filename);
    if (!fileout)
    {
        std::cerr << "Cannot open <" << filename << "> for output\n";
    }

    //  sample data
    sprintf(line, "%ld %ld\n", _nvar, _nobv);
    fileout << line;

    for (i = 1; i <= _nvar; i++)
    {
        sprintf(line, "%f\n%f\n", _sinfo[i].min, _sinfo[i].max);
        fileout << line;

        for (k = 1; k <= _nobv; k++)
        {
            sprintf(line, "%f\n", _sampx[(i - 1) * _nobv + k - 1]);
            fileout << line;
        }
    }
    fileout.close();
}

//*************************************************************************

void LatinHypercube::d2_to_d1(double **a2, double *a1)
{
//   Description:  This method converts a 2 diminsional correlation
//      matrix representation to a 1 diminsional representation.

#ifdef DEBUG
    cerr << "LatinHypercube::d2_to_d1()\n";
#endif

    long int i, j;
    int ind = 0;

    for (i = 0; i < _nvar; i++)
    {
        for (j = 0; j <= i; j++)
        {
            a1[ind++] = a2[i][j];
        }
    }
}

//*************************************************************************

void LatinHypercube::d1_to_d2(double **a2, double *a1)
{
//   Description:  This method converts a 1 diminsional correlation
//      matrix representation to a 2 diminsional representation.

#ifdef DEBUG
    cerr << "LatinHypercube::d1_to_d2()\n";
#endif

    long int i, j;

    for (i = 0; i < _nvar; i++)
    {
        for (j = 0; j <= i; j++)
        {
            a2[i][j] = a1[LOC1(i + 1, j + 1)];
            a2[j][i] = a1[LOC1(i + 1, j + 1)];
        }
    }
}

//*************************************************************************

void LatinHypercube::print_2d(double **a2)
{
//  Description:  This method prints a 2 dimensional array to standard 
//      output for debugging.

#ifdef DEBUG
    cerr << "LatinHypercube::print_2d()\n";
#endif

    long int i, j;

    for (i = 0; i < _nvar; i++)
    {
        for (j = 0; j < _nvar; j++)
            printf("%10.4f", a2[i][j]);
        printf("\n");
    }
}

void LatinHypercube::PrintHistogram(int numbins)
{
    // examine distribution shape by creating a histogram of each distribution
    double binsize;
    long int i, j, k;

    nbins = numbins;
    LOG("LatinHypercube::PrintHistogram");

    if (!(bins = new long int *[_nvar]))
    {
        std::cerr
                << "\nLatinHypercube distShape error: cannot allocate bin memory.\n";
        exit(-1);
    }
    for (i = 0; i < _nvar; i++)
    {
        if (!(bins[i] = new long int[nbins]))
        {
            std::cerr
                    << "\nLatinHypercube distShape error: cannot allocate bin memory.\n";
            exit(-1);
        }
    }
    for (i = 0; i < _nvar; i++)
    {
        for (j = 0; j < nbins; j++)
            bins[i][j] = 0;
    }

    for (i = 0; i < _nvar; i++)
    {
        binsize = (_sinfo[i + 1].max - _sinfo[i + 1].min) / numbins;
        for (j = 0; j < _nobv; j++)
        {
            k = (int) ((_sampx[i * _nobv + j] - _sinfo[i + 1].min) / binsize);
//         k = min(nbins-1,k);
            k = (nbins - 1 < k) ? nbins - 1 : k;
//         k = max(k,0);
            k = (k > 0) ? k : 0;
            bins[i][k]++;
        }
    }
    std::cout << std::endl;
    for (i = 0; i < nbins; i++)
    {
        std::cout << i << "   ";
        for (j = 0; j < _nvar; j++)
            std::cout << bins[j][i] << "   ";
        std::cout << std::endl;
    }
    std::cout << std::endl;

    for (i = 0; i < _nvar; i++)
        delete[] bins[i];
    delete[] bins;
}

//*************************************************************************
// Struct utility class members
//*************************************************************************

void lhrank::SetSize(long int nobv, long int nvar)
{
    if (!(xv = new double[nobv]) || !(rxv = new double[nobv]) || !(iwk =
            new int[nobv]) || !(final = new double[nobv * nvar + 1]))
    {
        std::cerr << "\nLHS Struct rank error: cannot allocate memory.\n";
    }
}

lhrank::~lhrank()
{
    //  cerr<<"rank destructor..."<<endl;
    delete[] final;
    delete[] xv;
    delete[] rxv;
    delete[] iwk;
}

void workc::SetSize(long int nvar)
{
    long int nvector;
    nvector = (nvar * (nvar + 1) / 2);
    if (!(s = new double[nvector]) || !(q = new double[nvector]) || !(uq =
            new double[nvector]))
    {
        std::cerr << "\n LHS Struct workc error: cannot allocate memory.\n";
    }
}

workc::~workc()
{
    //  cerr<<"workc destructor..."<<endl;
    delete[] s;
    delete[] q;
    delete[] uq;
}

void cmatr::SetSize(long int nvar)
{
    long int nvector;
    nvector = (nvar * (nvar + 1) / 2);
    if (!(corr = new double[nvector]) || !(ucorr = new double[nvector])
            || !(rcorr = new double[nvector]) || !(icm = new int[nvar]))
    {
        std::cerr << "\n LHS Struct cmatr error: cannot allocate memory.\n";
    }
}

cmatr::~cmatr()
{
    //  cerr<<"cmatr destructor..."<<endl;
    delete[] corr;
    delete[] ucorr;
    delete[] rcorr;
    delete[] icm;
}

