//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TAMHCopula.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mer oct 17 18:08:40 CEST 2012
 
 \brief Interface of the class URANIE::Sampler::TAMHCopula
 
 */
/**
 @class URANIE::Sampler::TAMHCopula
 
 @brief Description of the class TAMHCopula.
 
 the \b Ali-Mikhail-Haq (\b AMH ) Copula with parameter \f$\theta \in [-1, 1]\f$
 
 \f[
 C(u_1, u_2 ; \theta) \: = \: \frac{u_1 u_2}{1 - \theta(1-u)(1-v)} \: , \: \theta\in[-1,1]
 \f]
 
 We use the Johnson algorithm (1986) to generate random variates \f$ (u,v)\f$ from a \b Ali-Mikhail-Haq distribution with parameter \f$ \theta \in [-1., 1.]\f$.
 \li Generate two independent uniform variate \f$ (u,t) \in (0,1)\f$
 \li Set \f$ a = 1-u \f$ <br>
 \f$ b = -\theta(2at+1)+2\theta^2a^2t+1\f$ <br>
 \f$ c = \theta^2(4a^2t-4at+1)-\theta(4at-4t+2)+1\f$ <br>
 \f$ v = \frac{2t(a\theta-1)^2}{b+\sqrt{c}}\f$ <br>
 \li the desired pair is \f$ (u,v)\f$
 
 */
#ifndef TAMHCOPULA_H
#define TAMHCOPULA_H

////// URANIE
#include "TArchimedianCopula.h"
#include "UExceptions.h"

namespace URANIE
{
namespace Sampler
{
class TAMHCopula: public TArchimedianCopula
{
private:
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    TAMHCopula(URANIE::DataServer::TDataServer *tds, Double_t dparam,
            Int_t nCalcul);
    //! Default destructor
    virtual ~TAMHCopula();
    //@}

    //---------------------------------------------
    /** @name Generation of the sample
     **/
    //@{
    //! Generates the matrix of U
    /*!
     \param option (Option_t *) The option to pass.
     */
    void generateUMatrix(Option_t *option);
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Sampler::TAMHCopula, ID_SAMPLER)
    // TAMHCopula Class
};

} // Fin du namespace Sampler
} // Fin du namespace URANIE
#endif
