//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TClaytonCopula.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mer oct 17 18:08:40 CEST 2012
 
 \brief Implementation de la classe URANIE::Sampler::TClaytonCopula
 
 */
//// ROOT
#include "TRandom3.h"
#include "TMath.h"

////// URANIE
#include "TClaytonCopula.h"
#include "TStochAttribut.h"
#include "TUniformDistribution.h"
#include "UExceptions.h"

ClassImp (URANIE::Sampler::TClaytonCopula)

namespace URANIE
{
namespace Sampler
{
TClaytonCopula::TClaytonCopula(URANIE::DataServer::TDataServer *tds,
        Double_t dparam, Int_t nCalcul) :
        TArchimedianCopula(tds, dparam, nCalcul)
{
    if (_blog)
        cout << " ** constructor TClaytonCopula" << endl;
    try
    {
        if (_dparam < 0.0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TClaytonCopula constructor\n The parameter %e must be possitive",
                            _dparam));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

TClaytonCopula::~TClaytonCopula()
{
    if (_blog)
        cout << " ** destroy TClaytonCopula" << endl;
}

/*******************************************/
/*           Generate the U Matrix         */
/*******************************************/
void TClaytonCopula::generateUMatrix(Option_t *option)
{
    if (_blog)
        cout << " ** TClaytonCopula::generateUMatrix _dparam[" << _dparam
                << "] [" << option << "]" << endl;

    Double_t dproba;
    Int_t is;

    // Loop on Stochastics Attributes
    URANIE::DataServer::TStochasticAttribute *curdist;
    URANIE::DataServer::TUniformDistribution *guranieCopula =
            new URANIE::DataServer::TUniformDistribution("__guranie__copula__",
                    0.0, 1.0);

    TIter next(_lstOfAttributesToSample);
    Int_t iatt = 0;
    while ((curdist = (URANIE::DataServer::TStochasticAttribute *) next()))
    {
        if (_blog)
        {
            cout << "name[" << curdist->GetName() << "] Type["
                    << curdist->getLawType() << "]" << endl;
        }

        // Generate the U-sample
        for (is = 0; is < _nS; is++)
        {
            //dproba = dpas * gRandom->Rndm() + dpmin;
            dproba = gRandom->Rndm();
            if (iatt == 0)
                _matUValue(is, iatt) = guranieCopula->getInvCDF(dproba);
            else
            {
                Double_t dv1 = _matUValue(is, iatt - 1);
                Double_t dv2 = guranieCopula->getInvCDF(dproba);

                Double_t du2 = TMath::Power(dv1, -1.0 * _dparam);
                du2 *= TMath::Power(dv2, -1.0 * _dparam / (_dparam + 1.0))
                        - 1.0;
                if (_blog)
                    cout << " ** dv1[" << dv1 << "] dv2[" << dv2
                            << "] ****** du2[" << du2 << "]" << endl;
                _matUValue(is, iatt) = TMath::Power(du2 + 1.0, -1.0 / _dparam);
            }
            if (_blog)
                cout << "** is[" << is << "] -- proba [" << dproba
                        << "] -- value[" << _matUValue(is, iatt) << "]" << endl;
        }

        //Increment the index of attribute
        iatt++;
    }

    // Clean
    delete guranieCopula;
    guranieCopula = NULL;

    if (_blog)
        cout << " ** End Of TClaytonCopula::generateUMatrix[" << option << "]"
                << endl;
}

////////////////////////////////////////////////////////////
//  Printing Log
////////////////////////////////////////////////////////////
void TClaytonCopula::printLog(Option_t *option)
{
    cout << " ** TClaytonCopula::printLog[" << option << "]" << endl;
    cout << " ** Theta [" << _dparam << "]" << endl;
}

} // Fin du namespace Sampler
} // Fin du namespace URANIE

