//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TCopula.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   sat august the 9th 10:05:13 CEST 2008

 \brief Implementation of the class URANIE::Sampler::TCopula

 */
//// ROOT
#include "TRegexp.h"
#include "TRandom3.h"
#include "TMatrixD.h"
#include "TMath.h"

//// URANIE
#include "TCopula.h"
#include "TStochAttribut.h"
#include "TUniformDistribution.h"

ClassImp (URANIE::Sampler::TCopula)

namespace URANIE
{
namespace Sampler
{
////////////////////////////////////////////////////////////
//  Constructeur : from TDataServer + Methode + size
////////////////////////////////////////////////////////////
TCopula::TCopula(URANIE::DataServer::TDataServer *tds, Option_t *option,
        Int_t ns) :
        TSamplerStochastic(tds, option, ns)
{
    if (_blog)
        cout << " ** constructor TCopula by TDataServer[" << tds->GetName()
                << "]" << endl;
}

TCopula::~TCopula()
{
    if (_blog)
        cout << " ** destroy TCopula" << endl;
}

/*******************************************/
/*           Generate the sample           */
/*******************************************/
void TCopula::generateSample(Option_t *option)
{
    if (_blog)
        cout << " ** TCopula::generateSample[" << option << "]" << endl;

    _matUValue.ResizeTo(_nS, _nX);
    _matUValue.Zero();

    parseOption(option);
   
    // Generate the matrix of U
    generateUMatrix(option);

    // Print the matrix of U Values
    if (_blog)
        _matUValue.Print();

    // Create the TTree of data
    createTuple();

    Double_t *valpattern = new Double_t[_nX + 1];
    URANIE::DataServer::TStochasticAttribute *curdist;

    if (_blog)
        cout << "Fill the TTree ..." << endl;
    for (Int_t is = 0; is < _nS; is++)
    {
        if (_blog)
            cout << "Case is[" << is << "] ...";
        valpattern[0] = (Double_t)(is + 1);
        if (_blog)
            cout << " -- val [" << valpattern[0] << "/" << _nS << "]...";
        for (Int_t jx = 0; jx < _nX; jx++)
        {
            curdist =
                    (URANIE::DataServer::TStochasticAttribute *) _lstOfAttributesToSample->At(
                            jx);
            valpattern[jx + 1] = curdist->getInvCDF(_matUValue(is, jx));
            if (_blog)
                cout << "var jx[" << jx << "/" << _nX << "]  name["
                        << curdist->GetName() << "] val[" << valpattern[jx + 1]
                        << "]...";
        }
        _ntsample->Fill(valpattern);
        if (_blog)
            cout << " OK" << endl;
    }
    if (_blog)
        cout << " Done" << endl;

    // Affect the current TTree to the TDataServer
    _tds->setTuple(_ntsample);

    // Fill the formula attributes
    fillOtherAttributes();

    // Clean
    delete[] valpattern;
    valpattern = NULL;

    if (_blog)
        cout << " ** End Of TCopula::generateSample" << endl;
}

////////////////////////////////////////////////////////////
//  Printing Log
////////////////////////////////////////////////////////////
void TCopula::printLog(Option_t *option)
{
    cout << "TCopula::printLog[" << option << "]" << endl;
    TSampler::printLog();
}

} // Fin du namespace Sampler
} // Fin du namespace URANIE

