//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TFrankCopula.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mer oct 17 18:08:40 CEST 2012
 
 \brief Implementation de la classe URANIE::Sampler::TFrankCopula
 
 */
//// ROOT
#include "TRandom3.h"
#include "TMath.h"

////// URANIE
#include "TFrankCopula.h"
#include "TStochAttribut.h"
#include "TUniformDistribution.h"
#include "UExceptions.h"

ClassImp (URANIE::Sampler::TFrankCopula)

namespace URANIE
{
namespace Sampler
{
TFrankCopula::TFrankCopula(URANIE::DataServer::TDataServer *tds,
        Double_t dparam, Int_t nCalcul) :
        TArchimedianCopula(tds, dparam, nCalcul)
{
    if (_blog)
        cout << " ** constructor TFrankCopula" << endl;
}
TFrankCopula::~TFrankCopula()
{
    if (_blog)
        cout << " ** destroy TFrankCopula" << endl;
    if (_lstOfAttributesToSample != NULL)
      delete _lstOfAttributesToSample;
    _lstOfAttributesToSample = NULL;
}

/*******************************************/
/*           Generate the U Matrix         */
/*******************************************/
void TFrankCopula::generateUMatrix(Option_t *option)
{
    if (_blog)
        cout << " ** TFrankCopula::generateUMatrix _dparam[" << _dparam << "] ["
                << option << "]" << endl;

    Double_t dproba;
    Int_t is;

    // Loop on Stochastics Attributes
    URANIE::DataServer::TStochasticAttribute *curdist;
    URANIE::DataServer::TUniformDistribution *guranieCopula =
            new URANIE::DataServer::TUniformDistribution("__guranie__copula__",
                    0.0, 1.0);

    TIter next(_lstOfAttributesToSample);
    Int_t iatt = 0;
    while ((curdist = (URANIE::DataServer::TStochasticAttribute *) next()))
    {
        if (_blog)
        {
            cout << "name[" << curdist->GetName() << "] Type["
                    << curdist->getLawType() << "]" << endl;
        }

        // Generate the U-sample
        for (is = 0; is < _nS; is++)
        {
            dproba = gRandom->Rndm();
            if (iatt == 0)
                _matUValue(is, iatt) = guranieCopula->getInvCDF(dproba);
            else
            {
                Double_t dv1 = _matUValue(is, iatt - 1);
                Double_t dv2 = guranieCopula->getInvCDF(dproba);

                Double_t du2 = dv2 * (1.0 - TMath::Exp(-1.0 * _dparam));
                du2 /= dv2 * (TMath::Exp(-1.0 * _dparam * dv1) - 1.0)
                        - TMath::Exp(-1.0 * _dparam * dv1);
                if (_blog)
                    cout << " ** dv1[" << dv1 << "] dv2[" << dv2
                            << "] ****** du2[" << du2 << "]" << endl;
                _matUValue(is, iatt) = -1.0 * TMath::Log(1.0 + du2) / _dparam;
            }
            if (_blog)
                cout << "** is[" << is << "] -- proba [" << dproba
                        << "] -- value[" << _matUValue(is, iatt) << "]" << endl;
        }

        //Increment the index of attribute
        iatt++;
    }

    // Clean
    delete guranieCopula;
    guranieCopula = NULL;

    if (_blog)
        cout << " ** End Of TFrankCopula::generateUMatrix[" << option << "]"
                << endl;
}

////////////////////////////////////////////////////////////
//  Printing Log
////////////////////////////////////////////////////////////
void TFrankCopula::printLog(Option_t *option)
{
    cout << " ** TFrankCopula::printLog[" << option << "]" << endl;
    cout << " ** Theta [" << _dparam << "]" << endl;
}

} // Fin du namespace Sampler
} // Fin du namespace URANIE

