//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TGaussianCopula.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   2012-11-28 10:37:02

 \brief Implementation de la classe URANIE::Sampler::TGaussianCopula
 
 */
//// ROOT
#include "TDecompChol.h"
#include "TRandom3.h"
#include "TMath.h"

////// URANIE
#include "TGaussianCopula.h"
#include "TStochAttribut.h"
#include "TUniformDistribution.h"
#include "TNormalDistribution.h"
#include "UExceptions.h"

ClassImp (URANIE::Sampler::TGaussianCopula)

namespace URANIE
{
namespace Sampler
{
TGaussianCopula::TGaussianCopula(URANIE::DataServer::TDataServer *tds,
        Option_t *option, Int_t nCalcul) :
        TEllipticalCopula(tds, option, nCalcul)
{
    if (_blog)
        cout << " ** constructor TGaussianCopula" << endl;
}

TGaussianCopula::~TGaussianCopula()
{
    if (_blog)
        cout << " ** destroy TGaussianCopula" << endl;
}

/*******************************************/
/*           Generate the U Matrix         */
/*******************************************/
void TGaussianCopula::generateUMatrix(Option_t *option)
{
    TSamplerStochastic::init();

    if (_blog)
        cout << " ** TGaussianCopula::generateUMatrix  [" << option << "]"
                << endl;

    Double_t dproba;
    Int_t is;

    // Loop on Stochastics Attributes
    URANIE::DataServer::TStochasticAttribute *curdist;
    URANIE::DataServer::TNormalDistribution *guranieCopula =
            new URANIE::DataServer::TNormalDistribution("_normal_law_generator",
                    0.0, 1.0);

    TIter next(_lstOfAttributesToSample);
    Int_t iatt = 0;
    while ((curdist = (URANIE::DataServer::TStochasticAttribute *) next()))
    {
        if (_blog)
        {
            cout << "name[" << curdist->GetName() << "] Type["
                    << curdist->getLawType() << "]" << endl;
        }

        // Generate the U-sample
        for (is = 0; is < _nS; is++)
        {
            dproba = gRandom->Rndm();
            _matUValue(is, iatt) = guranieCopula->getInvCDF(dproba);
            if (_blog)
                cout << "** is[" << is << "] -- proba [" << dproba
                        << "] -- value[" << _matUValue(is, iatt) << "]" << endl;
        }

        //Increment the index of attribute
        iatt++;
    }
    if (_blog)
    {
        cout << "***  Reducted Normal Sampling" << endl;
        _matUValue.Print();
    }
    
    // Cholesky Decomposition of the Correlation matrix
    TDecompChol cholc(_matCorrelation);
    try
    {
        if (cholc.Decompose() == kFALSE)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TGaussianCopula::generateUMatrix \n The Cholesky Decomposition of the Correlation matrix failed");
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
      if(guranieCopula!=NULL)
	delete guranieCopula;
      guranieCopula = NULL;
        ue.printMessage();
        throw ue;
    }

    TMatrixD matUFromC = cholc.GetU();
    if (_blog)
    {
        cout << "***  Correlation matrix" << endl;
        _matCorrelation.Print();
        cout << "***  Cholesky Decomposition" << endl;
        cholc.Print();
    }

    // Construct the dependency of the independant values 
    TMatrixD matdvalC = _matUValue * matUFromC;
    if (_blog)
    {
        cout << "***  Matrix of the dependent random values" << endl;
        matdvalC.Print();
    }

    // Pass to the CDF
    for ( is = 0; is < _nS; is++)
    {
        for (Int_t jx = 0; jx < _nX; jx++)
        {
            _matUValue(is, jx) = guranieCopula->getCDF(matdvalC(is, jx));
        }
    }
    if (_blog)
    {
        cout << "***  Gaussian Copula" << endl;
        _matUValue.Print();
    }

    // Clean
    delete guranieCopula;
    guranieCopula = NULL;

    if (_blog)
        cout << " ** End Of TGaussianCopula::generateUMatrix[" << option << "]"
                << endl;
}

////////////////////////////////////////////////////////////
//  Printing Log
////////////////////////////////////////////////////////////
void TGaussianCopula::printLog(Option_t *option)
{
    cout << " ** TGaussianCopula::printLog[" << option << "]" << endl;
}

} // Fin du namespace Sampler
} // Fin du namespace URANIE

