//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*! \file TGaussianSampling.cxx
 *  \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 *  \date   $date$
 *  \brief _descritption_.
 *
 *  Created the : Fri Oct  1 13:37:23 2004
 *

 \par Modifications
 \li Cedric Forgues (C.S.) 2010/01/22
 prototype de TMath:Sort() change pour coorespondre a celui de ROOT 5.22.00f

 */
#include <ctime>
#include <iostream>
#include <fstream>
using std::cerr;

////// ROOT
#include "TDecompChol.h"
#include "TMath.h"
#include "TDatime.h"
#include "TROOT.h"
#include "TRandom3.h"

////// URANIE
#include "TGaussianSampling.h"
#include "TAttributeFormula.h"
#include "TStochAttribut.h"
#include "TNormalDistribution.h"
#include "TLogNormalDistribution.h"
#include "TUniformDistribution.h"
#include "TLogUniformDistribution.h"
#include "TTriangularDistribution.h"
#include "TLogTriangularDistribution.h"
#include "TExponentialDistribution.h"
#include "LatinHypercube.H"

ClassImp (URANIE::Sampler::TGaussianSampling)

namespace URANIE
{
namespace Sampler
{
////////////////////////////////////////////////////////////
//  Constructeur : from TDataServer + Methode + size
////////////////////////////////////////////////////////////
TGaussianSampling::TGaussianSampling(URANIE::DataServer::TDataServer *tds,
        Option_t *option, Int_t ns) :
        TSamplerStochastic(tds, option, ns)
{
    if (_blog)
        cout << "constructor TGaussianSampling by TDataServer["
                << tds->GetName() << "]" << endl;

    TString opt = option;
    opt.ToLower();

    // Case of an SRS design
    if (opt.CompareTo("srs", TString::kExact) == 0)
    {
        setMethodName("SRS");
        _ntype = kSRS;
        SetNameTitle(Form("Sampling_%s_%d", opt.Data(), ns),
                Form("Uranie %d sample with method %s", ns, opt.Data()));
    }
    // Default case: LHS design 
    else
    {
        setMethodName("LHS");
        _ntype = kLHS;
        SetNameTitle(Form("Sampling_lhs_%d", ns),
                Form("Uranie %d sample with method lhs", ns));
    }

    // Initialisation of the matrices. The covariance matrix's diagonal 
    // elements are initialised in the init() function.
    _covmatrix.ResizeTo(_nX, _nX);

    _corrmatrix.ResizeTo(_nX, _nX);
    _corrmatrix.UnitMatrix();

    init();

}

TGaussianSampling::~TGaussianSampling()
{
    if (_blog)
        cout << "************ Destructor URANIE::Sampler::TGaussianSampling"
                << endl;
    if (_blog)
        cout
                << "************ End of Destructor URANIE::Sampler::TGaussianSampling"
                << endl;
}

/*******************************************/
/*           init                          */
/*******************************************/
void TGaussianSampling::init()
{

    try
    {

        _sIteratorNameNo = Form("GaussSampling__No__%s", _tds->GetName());
        _sIteratorNameNi = Form("GaussSampling__Ni__%s", _tds->GetName());

        // Loop on the stochastics attributes to verify they are only TNormalDistribution
        for (Int_t i = 0; i < _nX; i++)
        {
            URANIE::DataServer::TStochasticAttribute *curdist =
                    (URANIE::DataServer::TStochasticAttribute *) _lstOfAttributesToSample->At(
                            i);

            if (curdist->getLawType()
                    != URANIE::DataServer::TStochasticAttribute::kNormal)
            {
                TString msg =
                        Form(
                                "TGaussianSampling::init ERROR\n The attribute (\"%s\") is not a Normal distribution (\"TNormalDistribution\") ",
                                curdist->GetName());
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                        msg);
            }
            else
                // Initialise the diagonal of the covariance matrix 
                _covmatrix(i, i) = curdist->getTheoreticalStdDev()
                        * curdist->getTheoreticalStdDev();

            if (_blog)
                cout << "name[" << curdist->GetName() << "] Type["
                        << curdist->getLawType() << "]" << endl;
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

/*******************************************/
/*           terminate                     */
/*******************************************/
void TGaussianSampling::terminate()
{
    cout << "TGaussianSampling::terminate[]" << endl;
}

/*******************************************/
/*           Generate the sample           */
/*******************************************/
void TGaussianSampling::generateSample(Option_t *option)
{
    TString soption = option;
    soption.ToLower();
    TSamplerStochastic::init();

    parseOption(option);
   
    Int_t nS = _nS;

    TMatrixD matdvalCR(nS, _nX);
    TMatrixD matVecMean(nS, _nX);

    //------------------------- ng-20100825 --------------------------//            
    // Take into account the LHS or SRS cases            
    Double_t dpas=-1.0, dtype=-1.0, dproba;

    switch (_ntype)
    {
    case kLHS:
    {
        dpas = 1.0 / (Double_t)(nS);
        dtype = 1.0;
    }
        break;
    case kSRS:
    {
        dpas = 1.0;
        dtype = 0.0;
    }
        break;
    default:
	break;
    }
    // Create the Normal law generator
    URANIE::DataServer::TNormalDistribution *_normal_law_generator =
            new URANIE::DataServer::TNormalDistribution("_normal_law_generator",
                    0.0, 1.0);
    // Prepare the indices
    Int_t * nind = new Int_t[nS];
    for (Int_t i = 0; i < nS; i++)
        nind[i] = i;

    // Generate the sample
    for (Int_t ix = 0; ix < _nX; ix++)
    {

        // Shuffle the indices (Fisher-Yates algorithm)
        for (Int_t is = nS - 1; is > 0; is--)
        {
            Int_t indtmp = (Int_t) gRandom->Integer(is + 1);
            Int_t valtmp = nind[indtmp];
            nind[indtmp] = nind[is];
            nind[is] = valtmp;
        }

        for (Int_t is = 0; is < nS; is++)
        {
            dproba = dpas * gRandom->Rndm() + dtype * dpas * is;
            matdvalCR(nind[is], ix) = _normal_law_generator->getInvCDF(dproba);

            if (_blog)
                cout << "** [" << ix << "," << is << "] -- random value ["
                        << matdvalCR(nind[is], ix) << "]" << endl;
        }
    }
    delete[] nind;
    nind = NULL;
    delete _normal_law_generator;
    _normal_law_generator = NULL;

    //----------------------------------------------------------------//

    if (_blog)
    {
        cout << "***  Matrix of the independent random values" << endl;
        matdvalCR.Print();
        cout << "***  TGaussianSampling::generateSample" << endl;
        cout << "***  Print the covariance matrix" << endl;
        cout << "_nS[" << _nS << "]" << endl;
        cout << "_nX[" << _nX << "]" << endl;
        _covmatrix.Print();
    }

   // Choleski decomposition of the covariance matrix
    TDecompChol cholesNo(_covmatrix);
    if (cholesNo.Decompose() == kFALSE)
    {
       //_covmatrix.Print();
        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                "TGaussianSampling::CholeskyFactorization : the cholesky decomposition of the covariance matrix failed");
    }
    else
    {
        TMatrixD matUFromC = cholesNo.GetU();
        if (_blog)
        {
            cout << "***  Covariance matrix" << endl;
            _covmatrix.Print();
            cout << "***  Cholesky Decomposition" << endl;
            cholesNo.Print();
            cout << "***  Matrix U from Cholesky Decomposition " << endl;
            matUFromC.Print();
        }

        // Construct the dependency of the independant values 
        TMatrixD matdvalC = matdvalCR * matUFromC;
        if (_blog)
        {
            cout << "***  Matrix of the dependent random values" << endl;
            matdvalC.Print();
        }

        // Create the matrix of the mean values
        for (Int_t iatt = 0; iatt < _nX; iatt++)
        {
            URANIE::DataServer::TStochasticAttribute *curdist =
                    (URANIE::DataServer::TStochasticAttribute *) _lstOfAttributesToSample->At(
                            iatt);
            if (_blog)
                cout << "name[" << curdist->GetName() << "] Type["
                        << curdist->getLawType() << "] Mean["
                        << curdist->getTheoreticalMean() << "] Std["
                        << curdist->getTheoreticalStdDev() << "]" << endl;

            for (Int_t is = 0; is < nS; is++)
                matVecMean(is, iatt) = curdist->getTheoreticalMean();
        }

        // Add the mean values to the dependent values 
        TMatrixD matdval = matdvalC + matVecMean;
        if (_blog)
        {
            cout << "***  Final matrix " << endl;
            matdval.Print();
        }

        // Create and fill the TTree of the DataServer
        createTuple();
        Double_t *valpattern = new Double_t[_nX + 1];

        if (_blog)
            cout << "Fill the TTree ..." << endl;

        for (Int_t is = 0; is < nS; is++)
        {
            if (_blog)
                cout << "Case is[" << is << "] ...";
            valpattern[0] = (Double_t)(is + 1);
            if (_blog)
                cout << " -- val [" << valpattern[0] << "] ...";
            for (Int_t jx = 0; jx < _nX; jx++)
            {
                valpattern[jx + 1] = matdval(is, jx);
                if (_blog)
                    cout << "var jx[" << jx << "]  val[" << valpattern[jx + 1]
                            << "]...";
            }
            _ntsample->Fill(valpattern);
        }

        // Affect the current TTree to the TDataServer
        _tds->setTuple(_ntsample);

        // Fill the formula attributes
        fillOtherAttributes();

        // Clean
        _ntsample->ResetBranchAddresses();
        delete[] valpattern;
        valpattern = NULL;
    }
}




/*******************************************/
/*    Generate Conditional Sampling        */
/*******************************************/
void TGaussianSampling::generateConditionalSample(TVectorD  vecOf01, Int_t no, Int_t ni) {

    if ( ( _tds->hasTuple()) && ( _tds->getNPatterns() > 0 ) ) {
        if ( !_tds->isAttribute(_sIteratorNameNo) ) {
            _tds->addAttribute(_sIteratorNameNo, "0");
            _tds->addAttribute(_sIteratorNameNi, "0");
        }
    }
    /* else {
        cout<< " ************** NONONONONONONONONONO DATA BEFORE ************* " << endl;
        createTuple();
        _tds->addAttribute(_sIteratorNameNo, "0");
        _tds->addAttribute(_sIteratorNameNi, "0");
    }*/



    if ( _blog ) {
        cout << " **** function[TGaussianSampling::" << __func__ << "] no[" << no << "] ni[" << ni << "]" << endl;
        vecOf01.Print("the vector of 01");
    }

    TMatrixD matVarCov0(0, 0), matVarCov1(0, 0), matVarCov01(0, 0);

    if ( _blog ) {
        cout << " ********  _covmatrix Before *****";
        _covmatrix.Print("");
        cout << " *********************************";
    }

    splitMatrixVarCoVar(_covmatrix, vecOf01, matVarCov0, matVarCov1, matVarCov01);

    if ( _blog ) {
        cout << " ********  matVarCov0 after Splitting*****";
        matVarCov0.Print("");
        cout << " ********  matVarCov1 *****";
        matVarCov1.Print("");
        cout << " ********  matVarCov01 *****";
        matVarCov01.Print("");
    }

    Int_t nox =  matVarCov0.GetNrows();
    Int_t nix =  matVarCov1.GetNrows();
    if ( _blog ) cout << " *** nox[" << nox << "] nix[" << nix << "]" << endl;


    // Treat the 2 mean vectors
    TVectorD vecMean0(nox);
    vecMean0.Zero();
    TVectorD vecMean1(nix);
    vecMean1.Zero();
    TVectorD vecPosition(_nX);
    vecPosition.Zero();

    Double_t dmean =0.0;
    Int_t inoxcrt = 0, inixcrt=0;
    for (Int_t iatt = 0; iatt < _nX; iatt++) {
        URANIE::DataServer::TStochasticAttribute *curdist =
                (URANIE::DataServer::TStochasticAttribute *) _lstOfAttributesToSample->At(iatt);
        dmean = curdist->getTheoreticalMean();

        if ( vecOf01(iatt) < 0.5 ) {
            vecMean0(inoxcrt) = dmean;
            vecPosition(inoxcrt)= iatt+1;
            inoxcrt++;
        } else {
            vecMean1(inixcrt) = dmean;
            vecPosition(nox+inixcrt)=iatt+1;
            inixcrt++;
        }
    }

    if (_blog) {
        vecMean0.Print("Mean Vector of vecMean0");
        vecMean1.Print("Mean Vector of vecMean1");
        vecPosition.Print("Vector of Position");
    }

    // Treat the conditional
    TMatrixD matXInv(TMatrixD::kInverted, matVarCov0);
    if ( _blog ) cout << " --- matXInv := matVarCov0 inverse" << endl;
    if ( _blog ) matXInv.Print("");

    TMatrixD matCov01XInv(matVarCov01,TMatrixD::kMult,matXInv);
    if ( _blog ) cout << "   ** Matrice matCov01XInv := VarCov01 * matXInv " << endl;
    if ( _blog ) matCov01XInv.Print("");

    TMatrixD matCov01XInvCov01(matCov01XInv,TMatrixD::kMultTranspose,matVarCov01);
    if ( _blog ) cout << "   ** Matrice matCov01XInvCov01 : VarCov01 * matXInv * VarCov01.T" << endl;
    if ( _blog ) matCov01XInvCov01.Print("");

    TMatrixD matVarYX(matVarCov1,TMatrixD::kMinus,matCov01XInvCov01);
    if ( _blog ) cout << "   ** Matrice : VarCovariance of Y/X " << endl;
    if ( _blog ) matVarYX.Print("");


    TMatrixD matNormalSampleNo(no, nox);
    generatenStandardPoints(matNormalSampleNo);

    if ( _blog ) {
        cout << " ******** matrice matNormalSampleNo  ( No sample) *****";
        matNormalSampleNo.Print("");
    }

    try {
        // Choleski decomposition of the matVarCov0 matrix
         TDecompChol cholesNo(matVarCov0);

         if ( !cholesNo.Decompose() ) {
             cout << "Decomposition failed, matrix singular ?" << endl;
             cout << "condition number = " << cholesNo.GetCondition() << endl;
             throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                     "TGaussianSampling::CholeskyFactorization : the cholesky decomposition of the covariance matrix  [matVarCov0] failed");
         }

         TMatrixD matUToNo = cholesNo.GetU();
         if ( _blog ) {
             cout << " *** TDecompChol cholesNo on matrix matVarCov0";
             cout << " *** matrix:: matUToNo" << endl;
             matUToNo.Print();
             cout << " ********************************" << endl;
         }



         // Choleski decomposition of the matVarCov0 matrix
          TDecompChol cholesNi(matVarYX);
          if ( !cholesNi.Decompose() ) {
              cout << "Decomposition failed, matrix singular ?" << endl;
              cout << "condition number = " << cholesNi.GetCondition() << endl;
              throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                      "TGaussianSampling::CholeskyFactorization : the cholesky decomposition of the covariance matrix  [matVarYX] failed");
          }

          TMatrixD matUToNi = cholesNi.GetU();
          if ( _blog ) {
              cout << " *** TDecompChol cholesNi on matrix matVarYX";
              cout << " *** matrix:: matUToNi" << endl;
              matUToNi.Print();
              cout << " ********************************" << endl;
          }

         // Construct the dependency of the independant values
         TMatrixD matdvalC(matNormalSampleNo,TMatrixD::kMult,matUToNo);

         if (_blog) {
             cout << "***  Matrix of the dependent random values" << endl;
             matdvalC.Print();
             cout << "   ** Matrice matCov01XInv := VarCov01 * matXInv " << endl;
             matCov01XInv.Print("");
         }

         TMatrixD matEspCond1(matdvalC , TMatrixD::kMultTranspose, matCov01XInv);

         if ( _blog ) {
             cout << "   ** Matrice matEspCond1 := matCov01XInv * matdvalC.T " << endl;
             matEspCond1.Print("");
         }

         Int_t natt = _tds->getNAttributes();
         Long64_t npatcrt = _tds->getNPatterns();
         Double_t *valpattern = new Double_t[natt + 1];
         for(Int_t i=0; i<natt+1;i++)
             valpattern[i] = -10.;


         TVectorD vecMu0(nox);
         for(Int_t io=0; io<no; io++) {
             for(Int_t iox=0; iox<nox; iox++) {
                 vecMu0(iox) = vecMean0(iox) + matdvalC(io, iox);
                 valpattern[(Int_t)(vecPosition(iox))] = vecMu0(iox);
             }
             //if ( _blog ) vecMu0.Print(Form(" Candidate vector No [%d/%d]", io+1, no));

             // Auxillary attributes (the No attribute)
             valpattern[natt-1] = 1.0+io;

             // Ni sample
             TMatrixD matNormalSampleNi(ni, nix);
             generatenStandardPoints(matNormalSampleNi);

             if ( _blog ) {
                 cout << " ********  matNormalSampleNi after sampling *****";
                 matNormalSampleNi.Print("");
             }

             // Construct the dependency of the independant values
             TMatrixD matdvalD(matNormalSampleNi,TMatrixD::kMult,matUToNi);
             if (_blog) {
                 cout << "***  Matrix size(" << ni << ") of the dependent random values Ni(" << io+1 <<"/"<<no<<")" << endl;
                 matdvalD.Print();
             }

             TVectorD vecMu1(nix);
             for(Int_t ii=0; ii<ni; ii++) {
                 npatcrt++;
                 valpattern[0] = 1.*npatcrt;

                 for(Int_t iix=0; iix<nix; iix++) {
                     vecMu1(iix) = vecMean1(iix) + matEspCond1(io,iix) + matdvalD(ii, iix);
                     valpattern[(Int_t)(vecPosition(nox+iix))] = vecMu1 (iix);
                 }
                 //if ( _blog ) vecMu1.Print(Form(" Candidate vector Ni [%d/%d]", ii+1, ni));

                 // Auxillary attributes (the No attribute)
                 valpattern[natt] = 1.0+ii;

                 _tds->getTuple()->Fill(valpattern);
             }
         }

//         // Affect the current TTree to the TDataServer
//         _tds->setTuple(_ntsample);

//         // Fill the formula attributes
//         fillOtherAttributes();

         // Clean
         _tds->getTuple()->ResetBranchAddresses();
         delete[] valpattern;
         valpattern = NULL;

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }


    if ( _blog ) cout << " ***** End Of function[TGaussianSampling::" << __func__ << "] no[" << no << "] ni[" << ni << "]" << endl;

}


void TGaussianSampling::splitMatrixVarCoVar(TMatrixD matVarCovar, TVectorD vecOfUnit,
                                            TMatrixD &matVarCovar0, TMatrixD &matVarCovar1, TMatrixD &matVarCovar01) {

    if ( _blog ) {
        cout << " **** function[TGaussianSampling::" << __func__ << "]" << endl;
        vecOfUnit.Print("the vector of 01");
        cout << " ********  Originale matVarCovar matrix *****";
        matVarCovar.Print("");
        cout << " *************************";
    }


    Int_t nP = matVarCovar.GetNrows();
    Int_t nR0 = matVarCovar0.GetNrows();
    Int_t nR1 = matVarCovar1.GetNrows();
    Int_t nR01 = matVarCovar01.GetNrows();
    Int_t nC01 = matVarCovar01.GetNcols();
    Int_t nSum1 = vecOfUnit.Sum();

    if ( _blog ) {
        cout << " *** nP[" << nP << "] --- nR0[" << nR0 << "] --- nR1[" << nR1 << "]";
        cout << "     --- matVarCovar01[" << nR01 << "," << nC01 << "] --- nSum1[" << nSum1 << "]" << endl;
    }
    Int_t nL1 = vecOfUnit.NonZeros ();
    Int_t nL0 = vecOfUnit.GetNrows()-nL1;
    if ( _blog )
        cout << " *** nP[" << nP << "] --- nL0[" << nL0 << "] --- nL1[" << nL1 << "] --- nSum1[" << nSum1 << "]" << endl;

    matVarCovar0.ResizeTo(nL0,nL0);
    matVarCovar0.Zero();

    matVarCovar1.ResizeTo(nL1, nL1);
    matVarCovar1.Zero();

    matVarCovar01.ResizeTo(nL1, nL0);
    matVarCovar01.Zero();

    Int_t icrt11 = 0;
    Int_t jcrt11 = 0;

    Int_t icrt00 = 0;
    Int_t jcrt00 = 0;

    Int_t icrt10 = 0;
    Int_t jcrt10 = 0;

    for(Int_t irow=0; irow<nP; irow++) {
        // Case of One
        if ( vecOfUnit[irow] > 0.5 ) {
            jcrt11=0;
            jcrt10=0;
            for(Int_t jcol=0; jcol<nP; jcol++) {
                if ( vecOfUnit[jcol] > 0.5  ) {
                    if ( _blog )
                        cout << "    +++++ irow[" << irow << "] jcol[" << jcol << "] - icrt11[" << icrt11 << "] jcrt11[" << jcrt11 << "] matVarCovar[" << matVarCovar(irow,jcol) << "]" << endl;
                    matVarCovar1(icrt11,jcrt11) = matVarCovar(irow,jcol);
                    jcrt11++;
                } else {
                    if ( _blog )
                        cout << "    ///// irow[" << irow << "] jcol[" << jcol << "] - icrt10[" << icrt10 << "] jcrt10[" << jcrt10 << "] matVarCovar[" << matVarCovar[irow][jcol] << "]" << endl;
                    matVarCovar01(icrt10, jcrt10) = matVarCovar(irow,jcol);
                    jcrt10++;
                }
            }
            icrt11++;
            icrt10++;
        } else {
            // Case of Zero
            jcrt00=0;
            for(Int_t jcol=0; jcol<nP; jcol++) {
                if ( vecOfUnit[jcol] < 0.5 ) {
                    if ( _blog )
                        cout << "    ----- irow[" << irow << "] jcol[" << jcol << "] matVarCovar[" << matVarCovar(irow,jcol) << "]" << endl;
                    matVarCovar0(icrt00,jcrt00) = matVarCovar(irow,jcol);
                    jcrt00++;
                }
            }
            icrt00++;
        }
    }
    if ( _blog ) cout << " ***** End Of function[TGaussianSampling::" << __func__ << "]" << endl;

}


void TGaussianSampling::generatenStandardPoints(TMatrixD &matPoints) {

    if ( _blog ) {
        cout << " **** function[TGaussianSampling::" << __func__ << "]" << endl;
    }

    Int_t nS = matPoints.GetNrows();
    Int_t nX = matPoints.GetNcols();

    if ( _blog ) {
        cout << " *** matPoints :: nS[" << nS << "] nX[" << nX << "]" << endl;
    }

    matPoints.Zero();

    // Method
    Double_t dpas=-1.0, dtype=-1.0, dproba;

    switch (_ntype) {
    case kLHS: {
        dpas = 1.0 / (Double_t)(nS);
        dtype = 1.0;
    }
        break;
    case kSRS: {
        dpas = 1.0;
        dtype = 0.0;
    }
        break;
    default:
        break;
    }

    // Create the Normal law generator
    URANIE::DataServer::TNormalDistribution *_normal_law_generator =
            new URANIE::DataServer::TNormalDistribution("_normal_law_generator", 0.0, 1.0);
    // Prepare the indices
    Int_t * nind = new Int_t[nS];
    for (Int_t i = 0; i < nS; i++)
        nind[i] = i;

    // Generate the sample
    for (Int_t ix = 0; ix < nX; ix++)
    {
        // Shuffle the indices (Fisher-Yates algorithm)
        for (Int_t is = nS - 1; is > 0; is--)
        {
            Int_t indtmp = (Int_t) gRandom->Integer(is + 1);
            Int_t valtmp = nind[indtmp];
            nind[indtmp] = nind[is];
            nind[is] = valtmp;
        }

        for (Int_t is = 0; is < nS; is++)
        {
            dproba = dpas * gRandom->Rndm() + dtype * dpas * is;
            matPoints(nind[is], ix) = _normal_law_generator->getInvCDF(dproba);

            if (_blog)
                cout << "  *** [" << ix << "," << is << "] -- random value ["
                     << matPoints(nind[is], ix) << "]" << endl;
        }
    }

    // Clean
    delete[] nind; nind = NULL;
    delete _normal_law_generator; _normal_law_generator = NULL;

    if ( _blog )
        cout << " **** End Of function[TGaussianSampling::" << __func__ << "]" << endl;
}

/***************************************************/
/*      Correlation factor                         */
/***************************************************/

void TGaussianSampling::setUserCorrelation(URANIE::DataServer::TAttribute *x,
        URANIE::DataServer::TAttribute *y, Double_t value)
{

    if (fabs(value) > 1.0)
        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                Form(
                        "TGaussianSampling::setUserCorrelation ERROR : the correlation factor  between (\"%s\", \"%s\") is greater 1.0 : [%f]",
                        x->GetName(), y->GetName(), value));
    Int_t indx = _tds->getAttributeIndex(x);
    Int_t indy = _tds->getAttributeIndex(y);

    setUserCorrelation(indx - 1, indy - 1, value);
}

void TGaussianSampling::setUserCorrelation(TString xname, TString yname,
        double value)
{
    if (fabs(value) > 1.0)
        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                Form(
                        "TGaussianSampling::setUserCorrelation ERROR : the correlation factor  between (\"%s\", \"%s\") is greater 1.0 : [%f]",
                        xname.Data(), yname.Data(), value));
    Int_t indx = _tds->getAttributeIndex(xname);
    Int_t indy = _tds->getAttributeIndex(yname);

    setUserCorrelation(indx - 1, indy - 1, value);
}

void TGaussianSampling::setUserCorrelation(Int_t indx, Int_t indy, Double_t value)
{

    if (fabs(value) > 1.0)
        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                Form(
                        "TGaussianSampling::setUserCorrelation ERROR : the absolute value of the correlation factor between attributes (%d, %d) is greater 1.0 : [%f]",
                        indx, indy, value));
    if (indx == indy)
        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                Form(
                        "TGaussianSampling::setUserCorrelation ERROR : the two indices are equal: i = %d, j= %d]",
                        indx, indy));

    if ((indx >= _nX) || (indy >= _nX) || (indx < 0) || (indy < 0))
        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                Form(
                        "TGaussianSampling::setUserCorrelation ERROR : index i [%d] or j [%d] is out of range",
                        indx, indy));

    // Update the correlation matrix
    _corrmatrix(indx, indy) = value;
    _corrmatrix(indy, indx) = value;

    // Update the covariance matrix
    URANIE::DataServer::TNormalDistribution *attX =
            (URANIE::DataServer::TNormalDistribution*) _lstOfAttributesToSample->At(
                    indx);
    URANIE::DataServer::TNormalDistribution *attY =
            (URANIE::DataServer::TNormalDistribution*) _lstOfAttributesToSample->At(
                    indy);
    _covmatrix(indx, indy) = value * attX->getTheoreticalStdDev()
            * attY->getTheoreticalStdDev();
    _covmatrix(indy, indx) = value * attX->getTheoreticalStdDev()
            * attY->getTheoreticalStdDev();

    if (_blog)
    {
        cout << " Updated correlation matrix:" << endl;
        _corrmatrix.Print();
        cout << " Updated covariance matrix:" << endl;
        _covmatrix.Print();
    }
}

void TGaussianSampling::setUserVarianceCovarianceMatrix(Int_t indx, Int_t indy, Double_t value)
{

    try {
        if (indx == indy)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__, Form(
                                                           "TGaussianSampling::setUserVarianceCovarianceMatrix ERROR : the two indices are equal: i = %d, j= %d]",
                                                           indx, indy));

        if ((indx >= _nX) || (indy >= _nX) || (indx < 0) || (indy < 0))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__, Form(
                                                           "TGaussianSampling::setUserVarianceCovarianceMatrix ERROR : index i [%d] or j [%d] is out of range (%d)",
                                                           indx, indy, _nX));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    // Update the VarianceCovariance matrix
    _covmatrix(indx, indy) = value;
    _covmatrix(indy, indx) = value;

    // Update the covariance matrix
    URANIE::DataServer::TNormalDistribution *attX = (URANIE::DataServer::TNormalDistribution*) _lstOfAttributesToSample->At(indx);
    URANIE::DataServer::TNormalDistribution *attY = (URANIE::DataServer::TNormalDistribution*) _lstOfAttributesToSample->At(indy);
    _corrmatrix(indx, indy) = value / ( attX->getTheoreticalStdDev() * attY->getTheoreticalStdDev());
    _corrmatrix(indy, indx) = value / ( attX->getTheoreticalStdDev() * attY->getTheoreticalStdDev());

    if (_blog) {
        cout << " Updated correlation matrix:" << endl;
        _corrmatrix.Print();
        cout << " Updated covariance matrix:" << endl;
        _covmatrix.Print();
    }
}

void TGaussianSampling::setFixedSumCovariance(TString method)
{
    Int_t nbAtt = _lstOfAttributesToSample->GetEntries();
    const Int_t tabSize = nbAtt;

    Double_t *tabSigma = new Double_t[tabSize]; // Standard deviations array               
    Double_t *tab_p = new Double_t[tabSize];     // Array of p_i
    Double_t sigmaCoef=-99;          // Normalisation coefficient

    // Retrieve the standard deviations
    for (Int_t i = 0; i < nbAtt; i++)
    {
        URANIE::DataServer::TNormalDistribution *tmpAtt =
                (URANIE::DataServer::TNormalDistribution*) _lstOfAttributesToSample->At(
                        i);
        tabSigma[i] = tmpAtt->getTheoreticalStdDev();
    }

    if (method == "simple")
    {
        if (_blog)
            cout << "Computing sigma coefficient for [" << method
                    << "] method..." << endl;

        Double_t sigma2 = 0.0;
        for (Int_t i = 0; i < nbAtt; i++)
            sigma2 += tabSigma[i] * tabSigma[i];

        // Filling the array of p_i
        for (Int_t i = 0; i < nbAtt; i++)
            tab_p[i] = tabSigma[i] * tabSigma[i] / sigma2;

        sigmaCoef = TMath::Sqrt(sigma2);
    }

    else if (method == "exact")
    {
        Double_t sigmaMin, sigmaMax; // Some useful standard deviations...  

        if (_blog)
            cout << "Computing sigma coefficient for [" << method
                    << "] method..." << endl;

        // Computing the normalisation coefficient (cf. Jean-Marc's document, 3rd paragraph).    
        sigmaMax = tabSigma[TMath::LocMax(nbAtt, tabSigma)];
        sigmaMin = tabSigma[TMath::LocMin(nbAtt, tabSigma)];

        Double_t lowerS = TMath::Max(2 * sigmaMax,
                nbAtt * sigmaMin / TMath::Sqrt(nbAtt - 1.0));
        Double_t upperS = nbAtt * sigmaMax / TMath::Sqrt(nbAtt - 1.0);
        Double_t oldS=0.0, newS, step, sum = 0.0;

        // The value of sigmaCoef is somewhere between lowerS et upperS. 
        // We find it by dichotomy. Note that precision is important !
        step = 2.0 * (upperS - lowerS);
        newS = lowerS;
        while (TMath::Abs(sum - (nbAtt - 2.0)) > 1e-12 * (nbAtt - 2.0))
        {
            sum = 0.0;
            for (Int_t i = 0; i < nbAtt; i++)
                sum += TMath::Sqrt(
                        1 - 4 * tabSigma[i] * tabSigma[i] / (newS * newS));

            step = step / 2.0;
            oldS = newS;

            if (sum > nbAtt - 2.0)
                newS = newS - step;
            else
                newS = newS + step;
        }
        sigmaCoef = oldS;
        if (_blog)
            cout << "n-2 = " << nbAtt - 2 << "; Sum = " << sum << "; Sigma = "
                    << sigmaCoef << "; Sigma Max = " << sigmaMax << endl;

        // Filling the array of p_i
        for (Int_t i = 0; i < nbAtt; i++)
        {
            Double_t quot = tabSigma[i] * tabSigma[i] / (sigmaCoef * sigmaCoef);

            if (quot < 1e-12)
                tab_p[i] = tabSigma[i] * tabSigma[i] / (sigmaCoef * sigmaCoef);
            else
                tab_p[i] = 0.5 * (1 - TMath::Sqrt(1 - 4 * quot));

            //tab_p[i] = TMath::Max( 0.5 * (1 - TMath::Sqrt(1 - 4 * tabSigma[i]*tabSigma[i] / (sigmaCoef*sigmaCoef))), tabSigma[i]*tabSigma[i] / (sigmaCoef*sigmaCoef) );
        }
    }

    // Fill the covariance matrix  (cf. Jean-Marc's document, 2nd and 3rd paragraphs).
    for (Int_t i = 0; i < nbAtt; i++)
    {
        _covmatrix(i, i) = tab_p[i] * (1 - tab_p[i]);
        for (Int_t j = i + 1; j < nbAtt; j++)
        {
            _covmatrix(i, j) = -tab_p[i] * tab_p[j];
            _covmatrix(j, i) = -tab_p[i] * tab_p[j];
        }
    }
    // unnormalise the covariance matrix
    _covmatrix *= sigmaCoef * sigmaCoef;

}

//////////////////////////////////////////////////////////
//  Printing Log
//////////////////////////////////////////////////////////
void TGaussianSampling::printLog(Option_t *option)
{
    cout << "** TGaussianSampling::printLog  [" << option << "]" << endl;
    cout << "** _nS   [" << _nS << "]" << endl;
    cout << "** _nX   [" << _nX << "]" << endl;
    cout << "** _nSeed   [" << getSeed() << "]" << endl;
    cout << "** _sMethod   [" << _sMethod << "]" << endl;
    cout << " *** Correlation Matrix **" << endl;
    _corrmatrix.Print();
    cout << " *** Variance-Covariance Matrix **" << endl;
    _covmatrix.Print();

}

} // End of namespace Sampler
} // End of namespace URANIE

