//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*! \file TGaussianSampling.h
 *  \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 *  \date   $Date$
 *  \brief Sampler for gaussian random variables.
 *
 *  Created: Fri Oct  1 13:37:23 2004
 *
 */

/**
 * @class URANIE::Sampler::TGaussianSampling
 *
 * @brief Description of the class TGaussianSampling.

 This class provides a data generator for gaussian random variables.

 To generate \f$k\f$ values from \f$n\f$ gaussian random variables, we proceed as follows:

 - the user defines a list of gaussian random variables \f$ y_1, y_2, ..., y_n \f$.
 - a covariance matrix \f$C \left(n \times n\right)\f$ is constructed from the variances of the \f$y_i\f$, and possibly modified by the user (adding covariances)
 - we apply a Cholesky decomposition on the matrix \f$C\f$, such that \f$C = U^T U\f$, where \f$U \left(n \times n\right)\f$ is an upper triangular matrix.
 - we build a data matrix \f$A \left(k \times n\right)\f$ of random values, generated from gaussian random variables \f$a_1, a_2, ..., a_n\f$, with mean 0 and variance 1
 - we apply the formula \f$Y = M + A U\f$, where \f$M \left(k \times n\right)\f$ is a matrix containing the means of the variables \f$y_i\f$.

 The output matrix \f$Y \left(k \times n\right)\f$ contains the realisation of the random variables \f$y_i\f$, following the covariance matrix \f$C\f$.

 */

#ifndef TGAUSSIANSAMPLING_H
#define TGAUSSIANSAMPLING_H

#ifndef ROOT_TMatrixD
#include "TMatrixD.h"
#endif
#ifndef ROOT_TVectorD
#include "TVectorD.h"
#endif
#ifndef ROOT_TNtuple
#include "TNtuple.h"
#endif
#ifndef ROOT_TString
#include "TString.h"
#endif
#ifndef ROOT_TNtupleD
#include "TNtupleD.h"
#endif
#ifndef ROOT_TNamed
#include "TNamed.h"
#endif

// Uranie
#include "TSamplerStochastic.h"
#include "TAttribute.h"
#include "TStochAttribut.h"
#include "TDataServer.h"

namespace URANIE
{
namespace Sampler
{
class TGaussianSampling: public TSamplerStochastic
{
    // Associations

    // Attributes
public:
    enum EType
    {
        kLHS, kSRS, kUnknown
    };
    TMatrixD _corrmatrix;           //! The correlation matrix
    TMatrixD _covmatrix;            //! The covariance matrix
    EType _ntype;                //! The type of sampling (SRS, LHS)

    // Operations
private:
    void splitMatrixVarCoVar(TMatrixD matVarCovar, TVectorD vecOfUnit,
                             TMatrixD &matVarCovar0, TMatrixD &matVarCovar1, TMatrixD &matVarCovar01);

    void generatenStandardPoints(TMatrixD &matPoints);
    TString _sIteratorNameNo;        ///< The specific iterator attribute for NO integer
    TString _sIteratorNameNi;        ///< The specific iterator attribute for NI integer

public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //!  Constructor from a TDataServer, the name of the method and the size of the sample
    /*!               
     \param tds (TDataServer *) the pointer of the TDataServer. It must contains objects inheriting from TStochasticAttribute
     \param option (Option_t *) Method of sampling. It can be "lhs" (Latin Hypercube Sampling) or "srs" (Simple Random Sampling) (Default = "lhs").
     \param nCalcul (Int_t) number of values to generate for each attribute (Default = 100)
     */
    TGaussianSampling(URANIE::DataServer::TDataServer *tds, Option_t *option =
            "lhs", Int_t nCalcul = 100);

    //! Default destructor
    virtual ~TGaussianSampling();
    //@}

    //---------------------------------------------
    /** @name State machine
     For the stochastic methods, two states can be created \em init and \em terminate
     which are executed just before the generation of the sample for the first one, and after the generation for the second one. They can be used for pre and post treatment purpose.
     **/
    //@{
    //! The preprocessing step
    virtual void init();
    //! The post-processing step
    virtual void terminate();
    //@}

    //---------------------------------------------
    //@{
    //! Generates the sample 
    void generateSample(Option_t *option = "");

    void generateConditionalSample(TVectorD  vecOf01, Int_t no, Int_t ni);

    void setIteratorNameNo(TString sNameNo) {
        _sIteratorNameNo = sNameNo.Data();
    }

    void setIteratorNameNi(TString sNameNi) {
        _sIteratorNameNi = sNameNi.Data();
    }

    const char * getIteratorNameNo() {
        return _sIteratorNameNo.Data();
    }
    const char * getIteratorNameNi() {
        return _sIteratorNameNi.Data();
    }
    //@}

    //---------------------------------------------
    /** @name Correlation matrix
     Methods to specify correlation between attributes
     **/
    //@{
    //! Defines a correlation between two attributes given by their indexes
    /*!
     \warning The indexes' range is from 0 to n-1.
     
     \param indx (Int_t) the index of the first attribute
     \param indy (Int_t) the index of the second attribute
     \param value (Double_t) the correlation factor between the two attributes
     */
    void setUserCorrelation(Int_t indx, Int_t indy, double value);
    void setUserVarianceCovarianceMatrix(Int_t indx, Int_t indy, double value);

    //! Defines a correlation between two attributes given by their names
    /*!                
     \param xname (TString) the first attribute
     \param yname (TString) the second attribute
     \param value (Double_t) the correlation factor between x and y
     */
    void setUserCorrelation(TString xname, TString yname, double value);

    //! Defines a correlation between two attributes
    /*!                
     \param x (URANIE::DataServer::TAttribute *) the first attribute
     \param y (URANIE::DataServer::TAttribute *) the second attribute
     \param value (Double_t) the correlation factor between x and y
     */
    void setUserCorrelation(URANIE::DataServer::TAttribute *x,
            URANIE::DataServer::TAttribute *y, Double_t value);

    //! Defines a special covariance matrix 
    /*!
     The covariance matrix is defined such that the sum of the generated 
     values is always equal to the sum of the random variables' means.
     
     \warning THIS METHOD IS EXPERIMENTAL. Its advantage is to preserve 
     the "gaussian nature" of the generated data. This might be 
     important when working on a small number of variables. 
     
     \param method (TString) method to compute the covariance matrix. 
     Two are available: 
     the "simple" method is very simple to implement but may be loose
     on standard deviations reconstruction. 
     The "exact" method is more tricky but should preserve the 
     standard deviations.  
     (Default = "simple") 
     */
    void setFixedSumCovariance(TString method = "simple");

    //! Retrieve the correlation matrix
    TMatrixD getCorrelationMatrix()
    {
        return _corrmatrix;
    }

    //! Retrieve the covariance matrix
    TMatrixD getCovarianceMatrix()
    {
        return _covmatrix;
    }
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    //! Prints the log
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Sampler::TGaussianSampling, ID_SAMPLER)
    //Definition d'un plan d'experience
};

} // Fin du namespace Sampler
} // Fin du namespace URANIE
#endif
// fin du fichier $RCSfile$.
