//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TNeuralGas.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mar jui 21 14:54:33 2009
 
 \brief  Implements the TNeuralGas class for the Neural Gas algorithm (Martinetz and Schulten, 1991)
 
 
 */
#include <ctime>

// ROOT
#include "TVectorD.h"
#include "TRegexp.h"
#include "TLeaf.h"
#include "TMath.h"
#include "TCanvas.h"
#include "TRandom3.h"
#include "TMath.h"

// URANIE
#include "TNeuralGas.h"
#include "UExceptions.h"
#include "UTimer.h"

ClassImp (URANIE::Sampler::TNeuralGas);

namespace URANIE
{
namespace Sampler
{

TNeuralGas::TNeuralGas(URANIE::DataServer::TDataServer *tds, Option_t *option,
        Int_t ns) :
        TVQ(tds, option, ns), _ntemperature(1), _dtmax(40000.), _dlambda(0.0), _dfinalambda(0.01), _dinitalpha(0.5), _dalpha(0.0), _dfinalalpha(0.005)
{

  Int_t npat = _tds->getNPatterns();
  try
    {
      if (npat < _nS)
	throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						   Form(
							"TNeuralGas constructor\n The number of patterns %d requested is greater than the number in the TDS [%d]",
							_nS, npat));
	
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
      ue.printMessage();
      throw ue;
    }
  TString soption = option;
  soption.ToLower();

  // Compute the statistics (Min, Max) of the attributes
  _tds->computeStatistic();

  // Default Values
  _dinitlambda = (Double_t)(_nS - 1) / 3.0;
   
}

// destructor
TNeuralGas::~TNeuralGas()
{
//       if ( _subtds )
//     delete _subtds; _subtds = NULL;

  if(_fLogger != NULL)
    delete _fLogger;
  _fLogger=NULL;
}

/*******************************************/
/*           state = Init                  */
/*******************************************/
void TNeuralGas::init()
{
    if (_blog)
        cout << " ** TNeuralGas::init _nS[" << _nS << "]" << endl;

    TVQ::init();

    URANIE::DataServer::TAttribute *ngiter = new URANIE::DataServer::TAttribute(
            "__neural__gas_iter__");
//       ngiter->setOrigin(URANIE::DataServer::TAttribute::kInternal);
    ngiter->delShare();
    _subtds->addAttribute(ngiter);
    _subtds->createTuple();

    // Allocate the correlation matrix
    _datamatrix.ResizeTo(_nS, _nX + 1);
    _dvw.ResizeTo(_nS, _nX + 1);

    URANIE::DataServer::TAttribute *tatt;
    TIter next(_lstOfAttributesToSample);

    // Compute the statistic to normalize the data
    _tds->computeStatistic();

    // Generate the init sample from the original dataset
    shuffle();
    TVectorD vtemp(_nX + 2);
    Int_t jcrt = 0;

    for (Int_t i = 0; i < _nS; i++)
    {
        vtemp(0) = 1.0 * i;
        _datamatrix(i, 0) = 0.;

        _tds->getTuple()->GetEntry((Long64_t)(_vecindexorg(i)));
        next.Reset();
        jcrt = 1;
        while ((tatt = (URANIE::DataServer::TAttribute *) next()))
        {
            TLeaf *leaf = _tds->getTuple()->GetLeaf(tatt->GetName());
//       _datamatrix(i,jcrt) = leaf->GetValue();
            _datamatrix(i, jcrt) = (leaf->GetValue() - tatt->getMean())
                    / tatt->getStd();
            vtemp(jcrt) = leaf->GetValue();
            jcrt++;
        }
        vtemp(jcrt) = 0.0;

        if (_bsaveall)
            _subtds->getTuple()->Fill(vtemp.GetMatrixArray());
    }

    if (_blog)
    {
        cout << " ** Initial values" << endl;
        _datamatrix.Print();
        cout << " ** End Of Initial values" << endl;
    }

    if (_blog)
        cout << " ** End Of TNeuralGas::init _nS[" << _nS << "]" << endl;
}
/*******************************************/
/*           state = learn                 */
/*******************************************/
void TNeuralGas::learn(Int_t index)
{
    if (_blog)
        cout << " ** TNeuralGas::learn index[" << index << "]" << endl;
    _tds->getTuple()->GetEntry(index);

    TVectorD valCrt(_nX);
    // Loop on the attributes
    URANIE::DataServer::TAttribute *tatt;
    TIter next(_lstOfAttributesToSample);
    Int_t ncrt = 0;
    _svardraw = "";
    while ((tatt = (URANIE::DataServer::TAttribute *) next()))
    {
        if (_blog)
        {
            cout << "name[" << tatt->GetName() << "] " << endl;
            //tatt->printLog();
        }
        TLeaf *leaf = _tds->getTuple()->GetLeaf(tatt->GetName());
        //    valCrt(ncrt) = leaf->GetValue();
        valCrt(ncrt) = (leaf->GetValue() - tatt->getMean()) / tatt->getStd();
        ncrt++;
    }

    if (_blog)
    {
        cout << " ** Current pattern _nX(" << valCrt.GetNrows() << ")" << endl;
        valCrt.Print();
        cout << " ** The current Matrix (" << _datamatrix.GetNrows() << ","
                << _datamatrix.GetNcols() << ")" << endl;
        _datamatrix.Print();
    }

    Double_t *sval = new Double_t[_nS];

    for (Int_t i = 0; i < _nS; i++)
    {
        _dvw(i, _nX) = 0.0;
        for (Int_t j = 0; j < _nX; j++)
        {
//       cout << " ** " << i << "," << j << " :: " << valCrt(j) << " :: " << _datamatrix(i,j+1);
            _dvw(i, j) = valCrt(j) - _datamatrix(i, j + 1);
            _dvw(i, _nX) += _dvw(i, j) * _dvw(i, j);
//       cout << " ::[" << _dvw(i,j) << "] dnorm (" << _dvw(i,_nX) << ")" << endl;
        }
        sval[i] = _dvw(i, _nX);

    }
    if (_blog)
    {
        cout << " ** The matrix of difference dvw(" << _dvw.GetNrows() << ","
                << _dvw.GetNcols() << ")" << endl;
        _dvw.Print();
        cout << " ** current values alpha(" << _dalpha << ") lambda("
                << _dlambda << ")" << endl;
    }

    Int_t *sindex = new Int_t[_nS];
    TMath::Sort(_nS, sval, sindex, kFALSE);

    if (_blog)
    {
        cout << " ** Order" << endl;
        for (Int_t i = 0; i < _nS; i++)
        {
            cout << " ** i(" << i << "/" << _nS << ") index[" << sindex[i]
                    << "] - value[" << sval[sindex[i]] << "]" << endl;
        }
    }

    Double_t *valueToFill = new Double_t[_nX + 2];

    for (Int_t i = 0; i < _nS; i++)
    {
        // The index
        valueToFill[0] = 1.0 * sindex[i];

        // Update the patterns
        for (Int_t j = 0; j < _nX; j++)
        {
            _datamatrix(sindex[i], j + 1) += _dalpha
                    * exp((-1.0 * (1.0 + (Double_t)(i))) / (_dlambda))
                    * _dvw(sindex[i], j);
            valueToFill[j + 1] = _datamatrix(sindex[i], j + 1);
        }

        // The temperature
        valueToFill[_nX + 1] = 1.0 * (Double_t)(_ntemperature);

        // Renormalize the data for store
        if ((_bsaveall) || (_ntemperature == (Int_t)(_dtmax)))
        {
            next.Reset();
            ncrt = 1;
            while ((tatt = (URANIE::DataServer::TAttribute *) next()))
            {
                valueToFill[ncrt] = valueToFill[ncrt] * tatt->getStd()
                        + tatt->getMean();
                ncrt++;
            }
            // Fill the data in the TNtuple
            _subtds->getTuple()->Fill(valueToFill);
        }
    }
    // Clean
    delete[] sval;
    sval = NULL;
    delete[] sindex;
    sindex = NULL;
    delete[] valueToFill;
    valueToFill = NULL;

    if (_blog)
        cout << " ** End Of TNeuralGas::learn index[" << index << "]" << endl;
}

/*******************************************/
/*           Generate the sample           */
/*******************************************/
void TNeuralGas::generateSample(Option_t *option)
{
  if (_blog)
    cout << endl << " ** TNeuralGas::generateSample option[" << option
	 << "]" << endl;

  Int_t nbcycle = 10;
  TString soption = option;
  soption.ToLower();
    
  init();

  if (soption.Contains("loop="))
    {
      TRegexp reg("loop=[0-9]*");
      TString out = soption(reg);
      if (_blog)
	cout << " ** out[" << out << "]" << endl;
      TString snoption = out.Remove(0, 5);
      if (_blog)
	cout << "snoption[" << snoption << "]" << endl;
      if (snoption.IsDigit())
	{
	  nbcycle = snoption.Atoi();
	  _dtmax = _tds->getNPatterns() * snoption.Atoi() - 1;
	}
      else
	{
	  try
	    {
	      throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							   Form(
								"TNeuralGas::generateSample Warning : the number of loop is incorrect [%s]\nThe default value is selected(%d)",
								snoption.Data(), nbcycle));
	    }catch(URANIE::Exceptions::UWarningExceptions &ue)
	    {
	      ue.printMessage();
	      throw ue;
	    }
	}

    }
  try
    {
      if (_dtmax < 0)
	throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						   Form(
							"TNeuralGas::generateSample Error : the parameter tmax is non posiitive [%d]",
							Int_t(_dtmax)));
    }catch(URANIE::Exceptions::UErrorExceptions &ue)
    {
      ue.printMessage();
      throw ue;
    }

  if (_blog)
    {
      cout << " ** _nseed[" << _nSeed << "]" << endl;
      cout << " **  soption[" << soption << "]" << endl;
      cout << " **  _dtmax [" << _dtmax << "]" << endl;
      cout << " **  nbcycle [" << nbcycle << "]" << endl;
    }

  Int_t nS = _nS;
  if (_blog)
    _lstOfAttributesToSample->Print();
  Int_t k = 1;
  Int_t index;

  Double_t ralpha = _dfinalalpha / _dinitalpha;
  Double_t rlambda = _dfinalambda / _dinitlambda;
  Double_t t;

  _fLogger = new URANIE::DataServer::UMessageLogger("TNeuralGas");
  (*_fLogger) << Endl<< " ** Neural Gas Learning (VQ) with TDataServer Name \"" << _tds->GetName() << "\" Title \"" << _tds->GetTitle() << "\"" << Endl;
  (*_fLogger) << " ** Patterns : " << _tds->getNPatterns() << Endl;
  (*_fLogger) << " ** Attributes : " << _sattributes << Endl;
  (*_fLogger) << " ** Iteration : " << _dtmax << Endl;
  if (_bsaveall)
    (*_fLogger) << " ** Save all points [TRUE]" << Endl;
  else
    (*_fLogger) << " ** Save all points [FALSE] - Only last iteration" << Endl;

  URANIE::DataServer::UTimer timer((Int_t)(_dtmax), "Neural Gas");

  while (k <= _dtmax)
    {
      if (_blog)
	cout << " ** cycle[" << k << "/" << _dtmax << "]" << endl;
#ifndef WIN32
      if (_bdrawProgressBar)
	timer.DrawProgressBar(k);
#endif

      shuffle();

      for (Int_t i = 0; i < _vecindexorg.GetNrows(); i++)
	{
	  if (_ntemperature > (Int_t)(_dtmax))
	    break;

	  index = (Int_t)(_vecindexorg(i));
	  if (_blog)
	    cout << " ** i [" << i << "] index (" << index << "/"
		 << _vecindexorg.GetNrows() << ")" << endl;

	  t = (Double_t)(_ntemperature) / _dtmax;
	  _dalpha = _dinitalpha * pow(ralpha, t);
	  _dlambda = _dinitlambda * pow(rlambda, t);

	  learn(index);

	  k++;
	  _ntemperature++;
	}
    }
#ifndef WIN32
  if (_bdrawProgressBar)
    timer.DrawProgressBar(k);
#endif
  (*_fLogger) << " ** Time elapsed " << timer.GetElapsedTime() << Endl;
  terminate();
  
  if (_blog)
    cout << endl << " ** End Of TNeuralGas::generateSample option["
	 << option << "]" << endl;
}
////////////////////////////////////////////////////////////
//  Printing Log
////////////////////////////////////////////////////////////
void TNeuralGas::printLog(Option_t *option)
{
    cout << "TNeuralGas::PrintLog[" << option << "]" << endl;
    cout << " ** _ntemperature[" << _ntemperature << "]" << endl;
    cout << " **  TMax[" << _dtmax << "]" << endl;
    cout << " **  alpha[" << _dinitalpha << ", " << _dfinalalpha << "]" << endl;
    cout << " ** lambda[" << _dinitlambda << ", " << _dfinalambda << "]"
            << endl;
    if (_bsaveall)
        cout << " ** save all[TRUE]" << endl;
    else
        cout << " ** save all[FALSE]" << endl;
}

} // Fin du namespace Sampler
} // Fin du namespace URANIE

// fin du fichier $RCSfile$.
