//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TNeuralGas.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mar jui 21 14:54:25 2009
 
 \brief  Defines the TNeuralGas class for the Neural Gas algorithm (Martinetz and Schulten, 1991)
 */
#ifndef TNeuralGas_H
#define TNeuralGas_H

// ROOT
#include "TMatrixD.h"

// Uranie
#include "TVQ.h"

/*!
 \class URANIE::Sampler::TNeuralGas
 \brief Implement the Neural Gas algorithm (Martinetz and Schulten, 1991)
 
 The \b Neural \b Gas \b Algorithm (Martinetz and Schulten, 1991) sorts for each input signal \f$ x \f$ the units of the network according to the distance of their reference vectors to \f$ \xi\f$.
 Based on this \em rank \em order a certain number of units is adapted.
 Both the number of adapted units and the adaptation strength are decreased according to a fixed schedule.

 The reference vectors are adjusted according to
 \f[
 \Delta \xi_j \: = \: \alpha(t) \exp^{-Rank(j)/\lambda(t)} \times (\xi_j - x)
 \f]
 with the following time-dependencies:
 \f[
 \alpha(t) = \alpha_i \; (\frac{\alpha_f}{\alpha_i})^{t/t_{max}}
 \f]
 \f[
 \lambda(t) = \lambda_i \; (\frac{\lambda_f}{\lambda_i})^{t/t_{max}}
 \f]
 
 \par Default values
 \li \f$ t_{max} = 40000\f$
 \li \f$ \alpha_i = 0.5\f$
 \li \f$ \alpha_f = 0.005\f$
 \li \f$ \lambda_f = 0.01\f$
 \li \f$ \lambda_i = \frac{Card(\xi)-1}{3} \f$

 The initial cluster are selected randomly from the original dataset.
 
 In the subsample dataserver, it exists an internal attribute \em "__neural__gas_iter__" which contains the number of loop need by the learning process.

 \par Example
 \code
 TDataServer *tds = new TDataServer("tdsgeyser", "Neural Gas for Geyser dataset");
 tds->fileDataRead("geyser.dat");
 
 TNeuralGas * tvq = new TNeuralGas(tds,"", 50);
 TDataServer *tdsng = tvq->getSubSample("loop=100");

 tds->draw("x2:x1");
 tdsng->draw("x2:x1","","same");

 \endcode

 
 */

namespace URANIE
{
namespace Sampler
{
class TNeuralGas: public TVQ
{
    // Attributes
public:
    TMatrixD _datamatrix;                         ///< The matrix of data

    TMatrixD _dvw;                                ///< The matrix of difference

    TString _svardraw;
    Int_t _ntemperature;
    Double_t _dtmax;                           ///< The \f$ t_{max}\f$ parameter
    Double_t _dinitlambda;           ///< The initial value of \f$ \lambda_i \f$
    Double_t _dlambda;               ///< The current value of \f$ \lambda_i \f$
    Double_t _dfinalambda;             ///< The final value of \f$ \lambda_f \f$
    Double_t _dinitalpha;              ///< The initial value of \f$ \alpha_i\f$
    Double_t _dalpha;                  ///< The current value of \f$ \alpha_i\f$
    Double_t _dfinalalpha;              ///< The final value of \f$ \alpha_f \f$

public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Constructor with a dataserver
    /*!
     \param tds (URANIE::DataServer::TDataServer *) the original data
     \param option (Option_t) contains the options [default ""]
     \li \em "saveall" Save all the data generate by the algorithm ( _bsaveall = kTRUE).
     \param ns (Int_t) the size of the subsample
     */
    TNeuralGas(URANIE::DataServer::TDataServer *tds, Option_t *option,
            Int_t ns);
    //! Default destructor
    virtual ~TNeuralGas();
    //@}

    //---------------------------------------------
    /** @name Setting and Getting attributs
     **/
    //@{
    //! Set the lambda parameters
    /*!
     Set the initial \f$ \lambda_i \f$ and final \f$ \lambda_f \f$ lambda parameters where the \f$ \lambda \f$ time-dependency is :
     \f[
     \lambda(t) = \lambda_i \; (\frac{\lambda_f}{\lambda_i})^{t/t_{max}} \:\: \forall t \in[0, t_{max}]
     \f]
     */
    void setLambda(Double_t dinitlambda, Double_t dfinallambda)
    {
        _dinitlambda = dinitlambda;
        _dfinalambda = dfinallambda;
    }
    //! Set the alpha parameters
    /*!
     Set the initial \f$ \alpha_i \f$ and final \f$ \alpha_f \f$ alpha parameters where the \f$ \alpha \f$ time-dependency is :
     \f[
     \alpha(t) = \alpha_i \; (\frac{\alpha_f}{\alpha_i})^{t/t_{max}} \:\: \forall t \in[0, t_{max}]
     \f]
     */
    void setAlpha(Double_t dinitalpha, Double_t dfinalalpha)
    {
        _dinitalpha = dinitalpha;
        _dfinalalpha = dfinalalpha;
    }
    //! Set the \f$ t_{max}\f$ parameter
    /*!
     */
    void setTMax(Int_t dtmax)
    {
        _dtmax = 1.0 * dtmax;
    }
    //! Get the \f$ t_{max}\f$ parameter
    /*!
     */
    Int_t getTMax()
    {
        return (Int_t)(_dtmax);
    }
    //@}

    //---------------------------------------------
    /** @name State machine
     For the stochastic methods, two states can be created \em init and \em terminate
     which are executed just before the generation of the sample for the first one, and after the generation for the second one. They can be used for pre and post treatment purpose.
     **/
    //@{
    //! The preprocessing step
    void init();
    //! The learning step for the index pattern
    void learn(Int_t index);
    //@}

    //---------------------------------------------
    /** @name Generate the sample
     the generate sample procedure
     **/
    //@{
    //! Generates the sample 
    /*!
     */
    void generateSample(Option_t *option = "");
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Sampler::TNeuralGas, ID_SAMPLER)
    //Classe abstraite de g�n�ration de plan d'exp�rience num�rique
};

} // Fin du namespace Sampler
} // Fin du namespace URANIE
#endif
// fin du fichier $RCSfile$.
