//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*! \file TNormale.cxx
 * \brief Implementation of the class representing the distribution of a multidimensional or real normal law
 *
 */

////// ROOT
#include "TNormale.h"
#include "TMath.h"

using std::cout;
using std::endl;

ClassImp (TNormale)

/*! 
 * \brief Constructor by default of the classs. The vector of means is equal to 0 and the matrice of variance covariance is  * the matrix identity.
 *
 *
 */

TNormale::TNormale()
/*! \n A dimension of two is arbitrarly attributed to the vector created by default. This dimension is adjusted to the one of *the mean vector and the matrix of "correlation-standard deviation", which will be set afterward.\n  */
{
    TVectorD unit(2);
    unit.Zero();
    TMatrixD Id(2, 2);
    Id.UnitMatrix();
    initialisation(unit, Id);
}

/*! \fn  TNormale::TNormale (TVectorD M ,TMatrixD A)
 * \brief Constructor of the class. M is the mean vector we want to attribute to our gaussian vector and to its matrix of "standard deviation/correlation".*/

TNormale::TNormale(TVectorD M, TMatrixD A)
{
    initialisation(M, A);
    TString sLabelVariable(5 * _dim - 1);
    sLabelVariable += "indice:";
    for (Int_t i = 1; i < _dim; i++)
    {
        sLabelVariable += "var";
        sLabelVariable += i;
        sLabelVariable += ":";
    }
    sLabelVariable += "var";
    sLabelVariable += _dim;
    _vect = new TNtupleD("echantillon", "echantillon2", sLabelVariable.Data());
}

/*! \fn TNormale::~TNormale()
 * \brief Destructor.
 */

TNormale::~TNormale()
{
}

/*! \fn TNormale::eval(Double_t x)
 * \brief Returns the value of the density along x  of a real normal law */
Double_t TNormale::eval(Double_t x)
{
    if (_dim == 1)
    {
        return (1 / (sqrt(2 * 3.1416) * _matEcartTypeCorr(0, 0)))
                * exp(
                        (-0.5
                                / (_matEcartTypeCorr(0, 0)
                                        * _matEcartTypeCorr(0, 0)))
                                * (x - _vectMean(0)) * (x - _vectMean(0)));
    }
    else
    {
        cout << "dim[" << _dim << "]" << endl;
        cout
                << "incompatibility between the dimension of the parameter and the one of the fonction ["
                << _dim << "]" << endl;
        return 0;
    }
}

/*! \fn TNormale::eval(TVectorD X)
 * \brief Returns the value of the density along x  of a multidimensional normal law*/
Double_t TNormale::eval(TVectorD x)
{
    if (_dim > 1)
    {
        Double_t Eval = 0;
        x.ResizeTo(_dim);
        x -= getMean();
        TMatrixD A = getMatCovariance();
        Double_t det = sqrt(A.Determinant());
        A.Invert();
        for (Int_t i = 0; i < _dim; i++)
            for (Int_t j = 0; j < _dim; j++)
                Eval = Eval + A(i, j) * x(i) * x(j);
        Double_t m = TMath::Power(2 * 3.1416, _dim / 2);
        return (1 / (m * det)) * exp(-0.5 * Eval);
    }
    else
    {
        return eval(x(0));
//     cout<<"Incompatibility between the dimension of the parameter and the one of the fonction"<<endl;
//     return 0;
    }

}

/*! \fn void TNormale::acceptationRejet (  )
 * \brief Method for accepting-refusing to simulate a gaussian with a Cauchy law as instrumental distribution 

 *
 *
 */

void TNormale::acceptationRejet()
{

    if (getDim() > 1)
        cout << "The variable must be real and not vectorial " << endl;

    else
    {
        Double_t randU;
        Double_t randG;
        Double_t U;
        Double_t x;
        Int_t i = 0;

        while (i < _taille)
        {
            U = _rdm->Uniform(1.0);
            randU = _rdm->Uniform(1.0);
            x = tan(3.1416 * (randU - 0.5));

            if (U <= (1 + x * x) * exp((1 - x * x) / 2) / 2)
            {
                randG = x * _matEcartTypeCorr(0, 0) + _vectMean(0);
                _vect->Fill(i, randG);
                i++;
            }

        }
    }
}

Double_t TNormale::getObs()
{

    if (getDim() > 1)
    {
        cout << "The variable must be real and not vectorial" << endl;
        return 0;
    }

    else
    {
        Double_t unif1;
        Double_t unif2;
        Double_t obs = 0;

        do
        {
            unif1 = _rdm->Uniform(1.0);
            unif2 = _rdm->Uniform(1.0);
            obs = tan(3.1416 * (unif2 - 0.5));
        } while (unif1 > (1 + obs * obs) * exp((1 - obs * obs) / 2) / 2);
        return _vectMean(0) + obs * _matEcartTypeCorr(0, 0);
    }
}

/*! \fn void TNormale::simulationPolaire (  )
 * \brief method of simulation of a centered reduced real normal law by the algorithme of Box-Muller*/

void TNormale::simulationPolaire()
{

    if (getDim() > 1)
        cout << "The variable must be real and not vectorial" << endl;

    else
    {
        Double_t U1;
        Double_t U2;
        Double_t X1;
        Double_t X2;
        Int_t j;
        Int_t k;

        for (Int_t i = 0; i < _taille / 2; i++)
        {
            j = 2 * i;
            k = 2 * i + 1;

            U1 = _rdm->Uniform(1.0);
            U2 = _rdm->Uniform(1.0);

            X1 = _matEcartTypeCorr(0, 0)
                    * (sqrt(-2 * log(U1)) * cos(2 * 3.1416 * U2))
                    + _vectMean(0);
            _vect->Fill(j, X1);
            X2 = _matEcartTypeCorr(0, 0)
                    * (sqrt(-2 * log(U1)) * sin(2 * 3.1416 * U2))
                    + _vectMean(0);
            _vect->Fill(k, X2);
        }

        // _vect->Write();
    }
}

/*! \fn void TNormale::vectCentreReduit()
 * \brief Simulates a centered reduced gaussian vector */

TVectorD TNormale::vectCentreReduit()
{
    Double_t N = _dim;
    TVectorD x(_dim);
    if (_dim > 1)
    {
        TVectorD M = getMean();
        TMatrixD A = getMatEcartTypeCorrelation();
        TVectorD m(1);
        TMatrixD a(1, 1);
        for (Int_t i = 0; i < N; i++)
        {
            m(0) = 0;
            a(0, 0) = 1;
            setMean(m);
            setMatEcartTypeCorrelation(a);
            x(i) = getObs();
        }
        setMean(M);
        setMatEcartTypeCorrelation(A);
        return x;
    }
    else
    {
        cout
                << "Method of simulation for multidimensionalvariables and not real "
                << endl;
//       TVectorD m(1);
//       m(0)=0;
//       setMean(m);
//       TMatrixD a(1,1);
//       a(0,0)=1;
//       setMatEcartTypeCorrelation(a);

        x.Zero();
        return x;
    }
}

/*! \fn void TNormale::getObsMultiDim()
 * \brief Simulates a given gaussian vector from the decomposition of Choleski  */

TVectorD TNormale::getObsMultiDim()
{
    TVectorD x(_dim);
    x = vectCentreReduit();
    TMatrixD y1(_dim, 1);
    TMatrixD y2(_dim, 1);
    for (Int_t i = 0; i < _dim; i++)
        y1(i, 0) = x(i);
    // cout<<"matcor"<<endl;
    // (getMatCovariance()).Print(); 
    TDecompChol m(getMatCovariance());
    m.Decompose();
    //cout<<"delta"<<endl;
    //(m.GetU()).Print();
    TMatrixD racine(_dim, _dim);
    racine.Transpose(m.GetU());
    y2.Mult(racine, y1);
    for (Int_t j = 0; j < _dim; j++)
        x(j) = y2(j, 0);
    x = getMean() + x;
    return x;
}

void TNormale::simVectGaussien()
{
    TVectorD tmp(_dim);
    for (Int_t i = 0; i < _taille; i++)
    {
        tmp = getObsMultiDim();
        tmp.ResizeTo(_dim + 1);
        for (Int_t j = 0; j < _dim + 1; j++)
            tmp(_dim - j) = tmp(_dim - j - 1);
        tmp(0) = i;
        _vect->Fill(tmp.GetMatrixArray());
    }
    // _vect->Write();
}
