//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TOA.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   tue apr 15 23:16:30 CEST 2008
 
 \brief Implementation of the class URANIE::Sampler::TOA
 
 */

////// ROOT
#include "TMath.h"
#include "TRandom3.h"

////// URANIE
#include "TOA.h"
#include "TStochAttribut.h"

extern "C"
{
int bose_link(int n, int ninputs, int str, int ***AA);
int OA_strength(int q, int nrow, int ncol, int** A, int *str, int verbose);
void free_imatrix(int **m, int nrl, int nrh, int ncl, int nch);
}

ClassImp (URANIE::Sampler::TOA)

namespace URANIE
{
namespace Sampler
{
TOA::TOA(URANIE::DataServer::TDataServer *tds, Option_t *option, Int_t ns) :
        TSamplerStochastic(tds, option, ns)
{
    if (_blog)
        cout << "construit TOA  nS[" << _nS << "]" << endl;

    // compute number of symbols
    Double_t tmp = (double) _nS;
    tmp = TMath::Power(tmp, 0.5000001);
    _nSymbols = (Int_t) tmp;
    if (_blog)
        cout << " ** _nSymbols[" << _nSymbols << "]" << endl;
    int nSamples1 = _nSymbols * _nSymbols;
    if (nSamples1 < _nS)
    {
        int nSamples2 = (_nSymbols + 1) * (_nSymbols + 1);
        if ((_nS - nSamples1) < (nSamples2 - _nS))
        {
            _nS = nSamples1;
        }
        else
        {
            _nS = nSamples2;
            _nSymbols++;
        }
    }
    if ((_nS != ns) && (_blog))
        cout << "constructor OA The number of samples adjusted to _nS[" << _nS
                << "] and demanded ns[" << ns << "]" << endl;
    // set up sample pattern
    init();
}
TOA::~TOA()
{
    if (_blog)
        cout << "detruit TOA" << endl;
}
/*******************************************/
/*           Generate the sample           */
/*******************************************/
void TOA::generateSample(Option_t *option)
{
    TString soption = option;
    soption.ToLower();
    if (_blog)
        cout << "TOA::GenerateSample" << endl;
    cout << "TOA::GenerateSample nX[" << _nX << "]" << endl;

}
/*******************************************/
/*           Init                          */
/*******************************************/
void TOA::init()
{
    if (_blog)
        cout << "TOA::init[]" << endl;
    Int_t i, j, is;
    // we need to use C arrays here to connect to the C code. 
    Int_t** pTmp;
    Int_t status = bose_link(_nS, _nX, 2, &pTmp);
    if (_blog)
        cout << "status[" << status << "]" << endl;
    if (status >= 0)
    {
        if (status != _nS)
        {
            cerr << "The number of samples adjusted to " << status << endl;
        }
    }
    else
    {
        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                "TROA::init :  bose cannot generate points");
    }

    // randomize the pattern
    //Array<int> ivec(nSymbols_);

    Int_t *nind = new Int_t[_nSymbols];

    // Initialize permutation
    if (_blog)
        cout << "Initialize permutation]" << endl;
    for (is = 0; is < _nSymbols; is++)
        nind[is] = is;

    // Create a permutation with the Devroye Algorithm
    Int_t valtmp, indtmp;
    for (is = _nSymbols - 1; is > 0; is--)
    {
        indtmp = (Int_t) gRandom->Integer(is);
        valtmp = nind[indtmp];
        if (_blog)
            cout << "** is[" << is << "] val [" << nind[is] << "] indtmp["
                    << indtmp << "] val[" << nind[indtmp] << "]";
        nind[indtmp] = nind[is];
        nind[is] = valtmp;
        if (_blog)
            cout << "************* is[" << is << "] val [" << nind[is]
                    << "] indtmp[" << indtmp << "] val[" << nind[indtmp] << "]"
                    << endl;
        ;
    }

    // Loop on the attributes
    Int_t noldindex;

    for (i = 0; i < _nX; i++)
    {
        if (_blog)
            cout << "variable [" << i + 1 << "]" << endl;
        // Create a new permutation and Store the vector of the current attribute in the matrix data
        for (is = _nSymbols - 1; is > 0; is--)
        {
            indtmp = (Int_t) gRandom->Integer(is);
            valtmp = nind[indtmp];
            if (_blog)
                cout << "** is[" << is << "] val [" << nind[is] << "] indtmp["
                        << indtmp << "] val[" << nind[indtmp] << "]";
            nind[indtmp] = nind[is];
            nind[is] = valtmp;
            if (_blog)
                cout << "--- is[" << is << "] val [" << nind[is] << "] indtmp["
                        << indtmp << "] val[" << nind[indtmp] << "]";
            ;
            if (_blog)
                cout << endl;
        }

        for (j = 0; j < _nS; j++)
        {
            noldindex = pTmp[j][i];
            pTmp[j][i] = nind[pTmp[j][i]];
            if (_blog)
                cout << i << ", " << j << " - new[" << pTmp[j][i] << "] ("
                        << noldindex << ")" << endl; /// = ivec[pTmp[j][i]];
        }
    }

    delete[] nind;
    nind = NULL;

    Int_t strength;
    OA_strength(_nSymbols, _nS, _nX, pTmp, &strength, 0);
    if (_blog)
        cout << "strength[" << strength << "]" << endl;

    if (strength != 2)
    {
        cout << "Orthogonal Array Sampling : failed strength 2 test" << endl;
    }

    // Create the TTree of data
    createTuple();
    Double_t *valpattern = new Double_t[_nX + 1];
    Double_t dproba;

    if (_blog)
        cout << "Fill the TTree ..." << endl;

    TList *attList = _tds->getListOfAttributes();
    for (is = 0; is < _nS; is++)
    {
        if (_blog)
            cout << "Case is[" << is << "] ...";
        valpattern[0] = (Double_t)(is + 1);
        if (_blog)
            cout << " -- val [" << valpattern[0] << "] ...";
        URANIE::DataServer::TStochasticAttribute *stoatt;

        for (Int_t jx = 0; jx < _nX; jx++)
        {
            stoatt = (URANIE::DataServer::TStochasticAttribute *) (attList->At(
                    jx));
            //stoatt = (URANIE::DataServer::TStochasticAttribute *)(_tds->getAttribute(jx));
            dproba = (gRandom->Rndm() + (Double_t)(pTmp[is][jx]))
                    / (Double_t)(_nSymbols);
            valpattern[jx + 1] = dproba;

            valpattern[jx + 1] = stoatt->getInvCDF(dproba);

            if (_blog)
                cout << "var jx[" << jx << "]  val[" << valpattern[jx + 1]
                        << "]...";
        }
        _ntsample->Fill(valpattern);
        if (_blog)
            cout << " OK" << endl;
    }
    if (_blog)
        cout << " Done" << endl;

    // Affect the current TTree to the TDataServer
    _tds->setTuple(_ntsample);

    // Clean
    delete attList;
    attList=NULL;
    _ntsample->ResetBranchAddresses();
    delete[] valpattern;
    valpattern = NULL;
    
    free_imatrix(pTmp, 0, _nS - 1, 0, _nX - 1);
}

void TOA::printLog(Option_t *option)
{
    cout << "TOA::printLog[" << option << "]" << endl;
    cout << "** _nS   [" << _nS << "]" << endl;
    cout << "** _nX   [" << _nX << "]" << endl;
    cout << "** _nSymbols   [" << _nSymbols << "]" << endl;
}

} // Fin du namespace Sampler
} // Fin du namespace URANIE

