//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*! \file TRafuSampling.h
 *  \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 *  \date   $Date$
 *  \brief Interface of the class URANIE::Sampler::TRafuSampling.
 *
 *  Created: Fri Oct  1 13:37:23 2004
 *
 */

/**
 * @class URANIE::Sampler::TRafuSampling
 *
 * @brief Description of the class TRafuSampling.

 This class defines a TSampler object for the RaFu method.

 The RaFu method applies fuzzy logic to uncertainty propagation (citation needed). 
 To achieve this, for each fuzzy variable, one has to compute an interval inside 
 which the real value is likely to lie. This "likelyness" is the "possibility" of 
 the interval and can be interpreted as the level of risk taken when choosing a value 
 inside the corresponding interval (0 = low risk, 1 = high risk).

 The goal of this sampler object is to use the IRSN code Sunset to generate a sampling 
 containing fuzzy variables. The generated data will then be stored in a Data Server 
 and used for uncertainty propagation and data analysis.

 The current implementation creates a text file for the Sunset software, launch 
 Sunset to generate the data, and retrieve the results. 

 Sunset generates 2 output files (one for each bound of the possibility intervals). 
 The TRafuSampling object will join these two data set and store them in the Data 
 Server. However, an attribute called %s__rafu__dataset__ (where "%s" is the name 
 of the TDataServer object) is given to distinguish the two. It equals to -1.0 for 
 the lower bound, and +1.0 for the upper bound.  

 */

#ifndef TRAFUSAMPLING_H
#define TRAFUSAMPLING_H

#ifndef ROOT_TString
#include "TString.h"
#endif
#ifndef ROOT_TList
#include "TList.h"
#endif
#ifndef ROOT_TNamed
#include "TNamed.h"
#endif

// Uranie
#include "TSampler.h"
#include "TAttribute.h"
#include "TStochAttribut.h"
#include "TPossibilityAttribute.h"
#include "TDataServer.h"

namespace URANIE
{
namespace Sampler
{
class TRafuSampling: public TSampler
{
    // Associations

    // Attributes
public:
    enum EProbaSampling
    {
        kLHS, kSRS, kUnknown
    };
    enum EPossibSampling
    {
        kRandom, kSampling0, kSampling1
    };
    EProbaSampling _nProbaSampling; ///< The type of sampling for the probability (SRS, LHS)
    EPossibSampling _nPossibSampling; ///< The type of sampling for the possibility (random, 0, 1)
    Int_t _nPossib;               ///< number of TPossibilityAttribute to sample
    TList * _lstOfPossibilityAttributesToSample; ///< The list of TPossibilityAttribute to sample
    TString _sunsetInputFileName;        ///< Name of the sunset input file name
    TString _sunsetInfFileName;     ///< Name of the sunset inf output file name
    TString _sunsetSupFileName;     ///< Name of the sunset sup output file name
    // Operations
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Creates a TRafuSampling object from a TDataServer, a 
    //! sampling method and the size of the sample to generate
    /*!
     The random variables to sample are the attributes of the TDataServer.
     These attributes must derive from the TStochasticAttribute and 
     TPossibilityAttribute classes.
     The sampling method for the TStochasticAttribute objects
     can be either "proba=lhs" (Latin Hypercube Sampling) or "proba=srs" (Simple 
     Random Sampling). If none of these is given, "proba=lhs" is chosen.
     The sampling method for the TPossibilityAttribute objects can be
     either "possib=random_interval" (random possibility), "possib=interval_sampling0" 
     (possibility fixed to 0) or "possib=interval_sampling1" (possibility fixed to 1).
     If none of these is given, "possib=random_sampling" is chosen.                
     
     \param tds        (TDataServer *) pointer to the TDataServer.
     \param option     (Option_t *) option string containing the sampling method for random variables ("proba=") and fuzy variables ("possib=") (Default = "proba=lhs, possib=random_sampling").
     \param sampleSize (Int_t) number of values to generate for each random variable (Default = 1000).
     
     \exception UErrorExceptions if the DataServer doesn't contain any stochastic attribute AND possibility sampling is different from "random_interval"
     */
    TRafuSampling(URANIE::DataServer::TDataServer *tds, Option_t *option =
            "proba=lhs, possib=random_sampling", Int_t sampleSize = 1000);

    //! Default destructor
    virtual ~TRafuSampling();
    //@}

    //---------------------------------------------
    /** @name Treats the TAttributes
     **/
    //@{            
    //! Creates the List of fuzzy attributes to sample  
    /*!
     Retrieve the list of TPossibilityAttribute objects stored in the
     TDataServer.
     */
    virtual void createListOfPossibilityAttributes();

    //! Constructs the input file for the Sunset software
    /*!
     This function create and fill the file containing all details to 
     command Sunset to generate a sampling of probability and possibility 
     variables. 
     */
    virtual void fillSunsetInputFile();

    //! Generates the sunset module for the given TStochasticAttribute
    /*!
     In a Sunset input file, the details of the attribute are stored 
     in a "module". This function create a TString containing the 
     definition of this module.
     
     \param att (TStochasticAttribute *) the attribute which details are used to build the module string
     \param indentation (char *) for aesthetical reason, this parameter provides the level of indentation of the module. 
     
     \exception UErrorExceptions if the type of the attribute is not supported.
     */
    virtual TString buildProbabilityModule(
            URANIE::DataServer::TStochasticAttribute *att,
            const char *indentation);

    //! Generates the sunset module for the given TPossibilityAttribute
    /*!
     In a Sunset input file, the details of the attribute are stored 
     in a "module". This function create a TString containing the 
     definition of this module.
     
     \param att (TStochasticAttribute *) the attribute which details are used to build the module string
     \param indentation (char *) for aesthetical reason, this parameter provides the level of indentation of the module.

     \exception UErrorExceptions if the type of the attribute is not supported.                
     */
    virtual TString buildPossibilityModule(
            URANIE::DataServer::TPossibilityAttribute *att,
            const char *indentation);
    //@}

    //---------------------------------------------
    /** @name Generate the sample
     **/
    //@{
    //! Creates the TTree containing the generated sample
    virtual void createTuple();

    //! Generates the sample 
    /*!
     Creates the input file for Sunset, calls Sunset, retrieves the results 
     and stores them in the TDataServer.
     
     \param option (Option_t *) option string (Default = "")
     */
    void generateSample(Option_t *option = "");
    //@}                       

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{                
    //! Prints the log
    /*!
     \param option (Option_t *) option string (Default = "")
     */
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Sampler::TRafuSampling, ID_SAMPLER)
    //Definition d'un plan d'exp�rience
};

} // Fin du namespace Sampler
} // Fin du namespace URANIE
#endif
// fin du fichier $RCSfile$.
