//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TSamplerField.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Tue Mar  7 14:06:02 2006

 \brief  Creation of the abstract class TSamplerField


 */

#include <ctime>

////// ROOT
#include "TMath.h"
#include "TROOT.h"
#include "TRandom3.h"
#include "TSystem.h"

////// URANIE
#include "TSamplerField.h"

using namespace TMath;

ClassImp (URANIE::Sampler::TSamplerField);

namespace URANIE
{
namespace Sampler
{
//////////////////////////////////////////////////////////////
//  Constructor : from TDataServer + Methode
////////////////////////////////////////////////////////////
TSamplerField::TSamplerField(URANIE::DataServer::TDataServer *tds,
                             Option_t *option, int nS) :
        TSamplerStochastic(tds, option, 1)

{

    _nS=nS;
    TString opt = option;
    opt.ToLower();
    setMethodName(Form("SamplerField_%s", opt.Data()));

    SetTitle(Form("Champ aleatoire[%s]", opt.Data()));
    SetIsotropy (kTRUE);
    SetVariance(1.0);
    SetRandomFunction(400);
    setMethod("Spectral");
    setVariogram("gaussian");

    _graph2D = 0;
    _dSigma = 0.25;
    _nXSize = 0;
    _nYSize = 0;
    _nZSize = 0;

    createListOfAttributes();

    URANIE::DataServer::TAttribute *att = 0;
    Int_t natt = _nX;

    if (natt < 2)
        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                Form(
                        "Constructor TSamplerField ERROR : The TDataServer \"%s\" contains not enough attributes (%d). It must be contains at least 2 and at more 4",
                        _tds->GetName(), natt));
    if (natt > 4)
        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                Form(
                        "Constructor TSamplerField ERROR : The TDataServer  \"%s\" contains %d attributes and is greater 4 ( 3 +1 )",
                        _tds->GetName(), natt));
    // Case of the star attribute
    att = _tds->getAttribute(natt - 1);
    
    SetName(Form("SF_%s", att->GetName()));
    if( _nS>1 && att->getDataTypeStr()!="V" )
    {
        att->setDataType(URANIE::DataServer::TAttribute::kVector);
        _tds->addAttribute( (SIZENAME+"__"+string(att->GetName()) ).c_str() ) ;
        _tds->getAttribute((SIZENAME+"__"+string(att->GetName()) ).c_str() )->setOrigin(URANIE::DataServer::TAttribute::kInternal) ;
    }
    natt--;
    Double_t valstep = 1.23456789;
    //Case of X
    if (natt > 0)
    {
        att = _tds->getAttribute(0);
        if (att)
        {
            _nXSize = (Int_t)(att->getUpperBound())
                    - (Int_t)(att->getLowerBound()) + 1;
            if (att->getStepValue(valstep))
                _nXSize = (Int_t)(_nXSize / valstep);
            if( _nS>1 && att->getDataTypeStr()!="V")
            {
                att->setDataType(URANIE::DataServer::TAttribute::kVector);
                _tds->addAttribute( (SIZENAME+"__"+string(att->GetName()) ).c_str() ) ;
                _tds->getAttribute((SIZENAME+"__"+string(att->GetName()) ).c_str() )->setOrigin(URANIE::DataServer::TAttribute::kInternal) ;
            }
        }
    }
    if (_blog)
        cout << "_nXSize[" << _nXSize << "]" << endl;

    //Case of Y
    if (natt > 1)
    {
        att = _tds->getAttribute(1);
        if (att)
        {
            _nYSize = (Int_t)(att->getUpperBound())
                    - (Int_t)(att->getLowerBound()) + 1;
            if (att->getStepValue(valstep))
                _nYSize = (Int_t)(_nYSize / valstep);
            if( _nS>1 && att->getDataTypeStr()!="V")
            {
                att->setDataType(URANIE::DataServer::TAttribute::kVector);
                _tds->addAttribute( (SIZENAME+"__"+string(att->GetName()) ).c_str() ) ;
                _tds->getAttribute((SIZENAME+"__"+string(att->GetName()) ).c_str() )->setOrigin(URANIE::DataServer::TAttribute::kInternal) ;
            }
        }
    }
    if (_blog)
        cout << "_nYSize[" << _nYSize << "]" << endl;

   if (natt > 2)
    {
        att = _tds->getAttribute(2);
        if (att)
        {
           _nZSize = (Int_t)(att->getUpperBound())
                    - (Int_t)(att->getLowerBound()) + 1;
            if (att->getStepValue(valstep))
                _nZSize = (Int_t)(_nZSize / valstep);
            if( _nS>1 && att->getDataTypeStr()!="V")
            {
                att->setDataType(URANIE::DataServer::TAttribute::kVector);
                _tds->addAttribute( (SIZENAME+"__"+string(att->GetName()) ).c_str() ) ;
                _tds->getAttribute((SIZENAME+"__"+string(att->GetName()) ).c_str() )->setOrigin(URANIE::DataServer::TAttribute::kInternal) ;
            }
        }
    }
    if (_blog)
        cout << "_nZSize[" << _nZSize << "]" << endl;
}

// Destructor
TSamplerField::~TSamplerField()
{
    if (_graph2D)
    {
        delete _graph2D;
        _graph2D = NULL;
    }
    
    for(vector<URANIE::DataServer::UEntry*>::iterator it=in.begin(); it!=in.end(); it++)
        if( dynamic_cast<URANIE::DataServer::UEntry*>( (*it) ) != NULL )
            delete (*it);
}

///////////////////////////////////////////////////////////
//  Create the List of Attributes to simulate from the TDataServer
///////////////////////////////////////////////////////////
void TSamplerField::createListOfAttributes()
{
    if (_blog)
        cout << "TSamplerField::createListOfAttributes" << endl;

    // Loop on the attributes of the TDataServer to select only TStochasticAttributes
    URANIE::DataServer::TAttribute *att = 0;

//---- Modification to avoid calling TDataServer::getAttribute(Int_t) ----//
//---- ngi: 2010.06.24          
//            
//            Int_t natt = _tds->getNAttributes();
//            for(Int_t iatt=0; iatt<natt; iatt++) {
//                att = _tds->getAttribute(iatt);
//                if ( _blog )
//                    cout << "att [" << att->GetName() << "]";
//                if ( _blog )
//                    cout << " ********** Add";
//                _lstOfAttributesToSample->Add( att );
//                if ( _blog )
//                    cout << endl;
//            }

    TList *attList = _tds->getListOfAttributes();
    for (Int_t iatt = 0; iatt < attList->GetEntries(); iatt++)
    {
        att = (URANIE::DataServer::TAttribute*) attList->At(iatt);
        if (_blog)
            cout << "att [" << att->GetName() << "]";
        if (_blog)
            cout << " ********** Add";
        _lstOfAttributesToSample->Add(att);
        if (_blog)
            cout << endl;
    }
//-----------------------------------------------------------------------//            
    _nX = _lstOfAttributesToSample->GetEntries();

    if (_blog)
    {
        cout << "Entries : " << _nX << endl;
        _lstOfAttributesToSample->Print();
    }

    delete attList;
}

void TSamplerField::VDC()
{
    // Generation de l'axe de rotation
    Double_t px = gRandom->Gaus(0.0, 1.0);
    Double_t py = gRandom->Gaus(0.0, 1.0);
    Double_t pz = gRandom->Gaus(0.0, 1.0);
    TVector3 axe = TVector3(px, py, pz);
    //  cas test avec simul_magellan
    //  TVector3 axe =  TVector3(0.6013094782829285, -0.9636887311935425, -0.1109096109867096);
    axe *= 1.0 / axe.Mag();

    // Generation de l'angle de rotation
    Double_t theta = TwoPi() * gRandom->Rndm();
    //  cas test avec simul_magellan
    //  theta = 1.994425392949919;

    Int_t n, m, d;
    Double_t x2 = 0., x3 = 0.;
    for (Int_t ind = 0; ind < _nRandomFunction; ind++)
    {
        // Binary decomposition of ind
        n = ind + 1;
        x2 = 0.;
        d = 2;
        while (n != 0)
        {
            m = n % 2;
            x2 += (Double_t)(m) / (Double_t)(d);
            d *= 2;
            n /= 2;
        }

        // Ternary decomposition of ind
        n = ind + 1;
        x3 = 0.;
        d = 3;
        while (n != 0)
        {
            m = n % 3;
            x3 += (Double_t)(m) / (Double_t)(d);
            d *= 3;
            n /= 3;
        }

        TVector3 u = TVector3(Cos(TwoPi() * x2) * Sqrt(1.0 - x3 * x3),
                Sin(TwoPi() * x2) * Sqrt(1.0 - x3 * x3), x3);
        // Effectue la rotation le long de l'axe et d'angle theta
        u.Rotate(theta, axe);

        // sauvegarde
        _vdc(ind, 0) = u(0);
        _vdc(ind, 1) = u(1);
        _vdc(ind, 2) = u(2);

    }

    if (0)
    {
        cout << " Rotation axis px[" << px << "] py[" << py << "] pz[" << pz
                << "]" << endl;
        axe.Print();
        cout << " Angle theta[" << theta << "]" << endl;
        _vdc.Print();
    }
}
void TSamplerField::SpectralInit()
{
    // Allocation de la matrice des vecteurs
    _vdc.ResizeTo(_nRandomFunction, 3);
    _scale.ResizeTo(_nRandomFunction);
    _seed.ResizeTo(_nRandomFunction);

    _ap.ResizeTo(3, 3);
    _ap.UnitMatrix();
    if (_bIsotropy)
    {
        _ap.UnitMatrix();
        for (Int_t ii = 0; ii < 3; ii++)
            _ap(ii, ii) /= _vScaleFactor(ii);
    }

    // normation factor
    switch (_nVariogram)
    {
    case kGaussianVariogram:
    case kSinusCardinalVariogram:
        _dSigma = Sqrt(2.0 * _dVariance / (Double_t)(_nRandomFunction));
        break;
    default:
        cout << "not recognized" << endl;
    }

    // Creation trois sequences de Van Der Corput ( Quasi Random Sequences)
    VDC();
}

void TSamplerField::SpectralPrepare()
{

    Double_t ux, uy, uz;
    for (Int_t ind = 0; ind < _nRandomFunction; ind++)
    {
        // rotation of the direction
        ux = _ap(0, 0) * _vdc(ind, 0) + _ap(0, 1) * _vdc(ind, 0)
                + _ap(0, 2) * _vdc(ind, 0);
        uy = _ap(1, 0) * _vdc(ind, 1) + _ap(1, 1) * _vdc(ind, 1)
                + _ap(1, 2) * _vdc(ind, 1);
        uz = _ap(2, 0) * _vdc(ind, 2) + _ap(2, 1) * _vdc(ind, 2)
                + _ap(2, 2) * _vdc(ind, 2);
        _scale(ind) = 1. / Sqrt(ux * ux + uy * uy + uz * uz);
        _vdc(ind, 0) = ux * _scale(ind);
        _vdc(ind, 1) = uy * _scale(ind);
        _vdc(ind, 2) = uz * _scale(ind);

        //  projection of the simulation field onto a line of direction id
        // save the random seed
        _seed(ind) = (Double_t)(gRandom->Integer(gRandom->GetSeed()));

        Double_t r, phi;
        Double_t a = _scale(ind);

        switch (_nVariogram)
        {
        case kGaussianVariogram:
        {
            Double_t px = gRandom->Gaus(0.0, 1.0);
            Double_t py = gRandom->Gaus(0.0, 1.0);
            Double_t pz = gRandom->Gaus(0.0, 1.0);
            r = Sqrt(2.0 * (px * px + py * py + pz * pz)) / a;
            phi = TwoPi() * gRandom->Rndm();
        }
            break;
        case kSinusCardinalVariogram:
        {
            r = (2.0 * (gRandom->Rndm() - 1.0)) / a;
            phi = TwoPi() * gRandom->Rndm();
        }
            break;
        default:
            cout << "not recognized" << endl;
        }
    }
}

void TSamplerField::SpectralSimulate(unsigned int ievt)
{

    TVectorD val(_nXSize);
    if(ievt==0 && _nS>1) in.push_back( new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::Double, string("Iterator")) );
    
    Double_t xmin = 1., xmax = 1., deltax = 1.;
    Double_t ymin = 1., ymax = 1., deltay = 1.;
    Double_t zmin = 1., zmax = 1., deltaz = 1.;
    // Valeur par defaut du nombre de boucle a realiser pour chaque composante
    Int_t nx = 1, ny = 1, nz = 1, nw=2;

    TList *attList = _tds->getListOfAttributes();
    if (_nXSize > 0)
    {
        xmin = _tds->getAttribute(0)->getLowerBound();
        xmax = _tds->getAttribute(0)->getUpperBound();
        deltax = (xmax - xmin) / (_nXSize - 1);
        nx = _nXSize;
        if(ievt==0 && _nS>1) in.push_back( new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::Vector, string("xvec")) );
    }
    if (_nYSize > 0)
    {
        ymin = _tds->getAttribute(1)->getLowerBound();
        ymax = _tds->getAttribute(1)->getUpperBound();
        deltay = (ymax - ymin) / (_nYSize - 1);
        ny = _nYSize;
        if(ievt==0 && _nS>1) in.push_back( new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::Vector, string("yvec")) );
        nw++;
    }
    if (_nZSize > 0)
    {
        zmin = _tds->getAttribute(2)->getLowerBound();
        zmax = _tds->getAttribute(2)->getUpperBound();
        deltaz = (zmax - zmin) / (_nZSize - 1);
        nz = _nZSize;
        if(ievt==0 && _nS>1) in.push_back( new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::Vector, string("zvec")) );
        nw++;
    }
    if(ievt==0 && _nS>1) in.push_back( new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::Vector, string("weight")) );


  
    for(unsigned int iin=0; iin<in.size(); iin++)
        in.at(iin)->clearEntry();
        
    for (Int_t iz = 0; iz < nz; iz++)
    {
        for (Int_t iy = 0; iy < ny; iy++)
        {

            Double_t y = ymin + iy * deltay;
            Double_t z = zmin + iz * deltaz;
            Double_t x;
            val.Zero();

            for (Int_t ind = 0; ind < _nRandomFunction; ind++)
            {
                //  projection of the simulation field onto a line of direction id
                // save the random seed
                gRandom->SetSeed((UInt_t)(_seed(ind)));

                Double_t r, phi;
                Double_t a = _scale(ind);
                Double_t tref, dtref, t0;

                switch (_nVariogram)
                {
                case kGaussianVariogram:
                {
                    Double_t px = gRandom->Gaus(0.0, 1.0);
                    Double_t py = gRandom->Gaus(0.0, 1.0);
                    Double_t pz = gRandom->Gaus(0.0, 1.0);
                    r = Sqrt(2.0 * (px * px + py * py + pz * pz)) / a;
                    phi = TwoPi() * gRandom->Rndm();

                    tref = xmin * _vdc(ind, 0) + y * _vdc(ind, 1)
                        + z * _vdc(ind, 2);
                    dtref = deltax * _vdc(ind, 0);
                    for (Int_t ix = 0; ix < _nXSize; ix++)
                    {
                        t0 = tref + ix * dtref;
                        val(ix) += Cos(r * t0 + phi);
                    }
                }
                break;
                case kSinusCardinalVariogram:
                {
                    r = (2.0 * (Nint(gRandom->Rndm()) - 1.0)) / a;
                    phi = TwoPi() * gRandom->Rndm();

                    for (Int_t ix = 0; ix < _nXSize; ix++)
                    {
                        x = xmin + ix * deltax;
                        t0 = x * _vdc(ind, 0) + y * _vdc(ind, 1)
                            + z * _vdc(ind, 2);
                        val(ix) += Cos(r * t0 + phi);
                    }
                }
                break;
                default:
                    cout << "not recognized" << endl;
                }
            }

            // Store the data
            for (Int_t ix = 0; ix < _nXSize; ix++)
            {
                Int_t index = ievt+1; //100 * iz + 10 * iy + ix;                
                x = xmin + ix * deltax;
                val(ix) *= _dSigma;
                _valcrt = val(ix);

                if( _nS>1 )
                {
                    in.at(0)->pushEntry(index);
                    in.at(1)->pushEntry(x);
                    if (_nYSize > 0)
                        in.at(2)->pushEntry(y);
                    if (_nZSize > 0)
                        in.at(3)->pushEntry(z);

                    in.at(nw)->pushEntry(_valcrt);
                }
                else
                {
                    if (_nZSize > 0)
                    {
                        _ntsample->Fill(index, x, y, z, _valcrt);
                    }
                    else if (_nYSize > 0)
                        _ntsample->Fill(index, x, y, _valcrt);
                    else
                        _ntsample->Fill(index, x, _valcrt);  
                }
            }
        } // end of iy
    } // end of iz

    if( _nS>1 )
        _ntsample->Fill(&in);
    
    
    delete attList;
}

/*******************************************/
/*       Generate the sample               */
/*******************************************/

void TSamplerField::generateSample(Option_t *option)
{
    TString soption = option;
    soption.ToLower();
    try
    {
        // Create the tuple to store the data
        createTuple();

        for(int iis=0; iis<_nS; iis++)
        {
            switch (_nMethod)
            {
            case kSpectralMethod:
            {
                switch (_nVariogram)
                {
                case kGaussianVariogram:
                case kSinusCardinalVariogram:
                {
                    SpectralInit();
                    SpectralPrepare();
                    SpectralSimulate(iis);
                    break;
                }
                default:
                {
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                               Form("TSamplerField::generateSample - ERROR: Variogram %s not yet implemented",
                                                                    getVariogramName().Data() ));
                    break;
                }
                }
                break;
            }
            default:
            {
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("TSamplerField::generateSample - ERROR: Method %s not yet implemented",
                                                                getMethodName().Data() ));
                // cout << "TSamplerField::Init Method[" << getMethodName()
                // 	 << "] not yet implemented " << endl;
                break;
            }
            }

        }
        
	// Affect the current TTree to the TDataServer
        _tds->setTuple(_ntsample);
	// Fill the formula attributes
	fillOtherAttributes();

        Int_t npat = (Int_t) _ntsample->GetEntries();
	if (_blog)
	    cout << "** TSamplerField::generateSample npat[" << npat << "]" << endl;

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
	
}

void TSamplerField::SetScaleFactor(Double_t dval)
{
  try {
    if (dval > 0) {
        _vScaleFactor(0) = Float_t(dval);
        _vScaleFactor(1) = Float_t(dval);
        _vScaleFactor(2) = Float_t(dval);
      } else {
      throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                 Form("TSamplerField::SetScaleFactor - ERROR :\n The scale factor must be possitive [%e]", dval ) );
      }
  } catch (URANIE::Exceptions::UErrorExceptions& ue)
  {
    ue.printMessage();
    throw ue;
  }

}

void TSamplerField::SetVariance(Double_t dval)
{
  try {
    if (dval > 0) {
      _dVariance = dval;
    } else {
      throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                 Form("TSamplerField::SetVarianceNew - ERROR\n The variance must be possitive [%e]", dval ) );
      }
  } catch (URANIE::Exceptions::UErrorExceptions& ue)
  {
    ue.printMessage();
    throw ue;
  }
}

////////////////////////////////////////////////////////////
//  Method 
////////////////////////////////////////////////////////////
void TSamplerField::setMethod(TString str)
{
    TString opt = str;
    opt.ToLower();

    if (opt.Contains("spectral"))
    {
        _nMethod = TSamplerField::kSpectralMethod;
    }
    else if (opt.Contains("dilution"))
    {
        _nMethod = TSamplerField::kDilutionMethod;
    }
    else if (opt.Contains("tessellation"))
    {
        _nMethod = TSamplerField::kTessellationMethod;
    }
    else if (opt.Contains("turningbands"))
    {
        _nMethod = TSamplerField::kTurningBandsMethod;
    }
    else
    {
        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                Form(
                        "TTSamplerField::setMethod ERROR : Method \"%s\" unknown :",
                        str.Data()));
    }
}

TString TSamplerField::getMethodName()
{
    switch (_nMethod)
    {
    case kSpectralMethod:
        return TString("Spectral Method");
        break;
    case kDilutionMethod:
        return TString("Dilution Method");
        break;
    case kTessellationMethod:
        return TString("Tessellation Method");
        break;
    case kTurningBandsMethod:
        return TString("Turning Bands Method");
        break;
    default:
        return TString("unknown Bands Method");
        break;
    }
}

////////////////////////////////////////////////////////////
//  Variogramm 
////////////////////////////////////////////////////////////
void TSamplerField::setVariogram(TString str)
{
    TString opt = str;
    opt.ToLower();

    if (opt.Contains("spherical"))
    {
        _nVariogram = TSamplerField::kSphericalVariogram;
    }
    else if (opt.Contains("exponential"))
    {
        _nVariogram = TSamplerField::kExponentialVariogram;
    }
    else if (opt.Contains("stable"))
    {
        _nVariogram = TSamplerField::kStableVariogram;
    }
    else if (opt.Contains("hyperbolic"))
    {
        _nVariogram = TSamplerField::kHyperbolicVariogram;
    }
    else if (opt.Contains("gaussian"))
    {
        _nVariogram = TSamplerField::kGaussianVariogram;
    }
    else if (opt.Contains("sine"))
    {
        _nVariogram = TSamplerField::kSinusCardinalVariogram;
    }
    else
    {
        cout << "*** Error in TSamplerField::setVariogram" << endl;
        cout << "*** Variogram type [" << str << "] not found" << endl;
        _nVariogram = TSamplerField::kGaussianVariogram;
        cout << "*** Choose the default[" << getVariogramName() << "]" << endl;
    }
}

TString TSamplerField::getVariogramName()
{
    switch (_nVariogram)
    {
    case kSphericalVariogram:
        return TString("Spherical");
        break;
    case kExponentialVariogram:
        return TString("Exponential");
        break;
    case kStableVariogram:
        return TString("Stable");
        break;
    case kHyperbolicVariogram:
        return TString("Hyperbolic");
        break;
    case kGaussianVariogram:
        return TString("Gaussian");
        break;
    case kSinusCardinalVariogram:
        return TString("SineCardinal");
        break;
    default:
        return TString("unknown Variogram Method");
        break;
    }
}

////////////////////////////////////////////////////////////
//  Visualisation 
////////////////////////////////////////////////////////////

void TSamplerField::Draw2D(TString sx, TString sy, Option_t* opt)
{
    if (_ntsample)
    {
        Int_t npat = (Int_t) _ntsample->GetEntries();
        if (_blog)
            cout << "** TSamplerField::Draw2D npat[" << npat << "]" << endl;

        URANIE::DataServer::TAttribute *att = _tds->getAttribute(_nX - 1);
        TString sdraw = Form("%s:%s:%s", att->GetName(), sy.Data(), sx.Data());
        if (_blog)
            cout << "** sdraw[" << sdraw << "]" << endl;

        npat = _ntsample->Draw(sdraw, "", "");

        if (_graph2D)
        {
            delete _graph2D;
            _graph2D = NULL;
        }
        _graph2D = new TGraph2D(npat, _ntsample->GetV3(), _ntsample->GetV2(),
                _ntsample->GetV1());
        _graph2D->SetTitle(Form("graph2D [%s] option[%s]", sdraw.Data(), opt));
        _graph2D->Draw(opt);

    }
    else
    {
        cout
                << "TSamplerField::Draw2D error\n ** the ttree does not exist. Have you execute the generateSample method!"
                << endl;
    }

}

////////////////////////////////////////////////////////////
//  Printing Log
////////////////////////////////////////////////////////////
void TSamplerField::printLog(Option_t *option)
{
    cout << "TSamplerField::printLog[" << option << "] Name[" << GetName()
            << "] Title[" << GetTitle() << "] ID[" << GetID() << "]" << endl;
    TSamplerStochastic::printLog(option);
    cout << "_nXSize[" << _nXSize << "]" << endl;
    cout << "_nYSize[" << _nYSize << "]" << endl;
    cout << "_nZSize[" << _nZSize << "]" << endl;
    cout << "_RandomFunction[" << _nRandomFunction << "]" << endl;
    cout << "_nMethod[" << _nMethod << "]" << endl;
    cout << "_nVariogram[" << _nVariogram << "]" << endl;
    cout << "_bIsotropy[" << _bIsotropy << "]" << endl;
    cout << "_dVariance[" << _dVariance << "]" << endl;
    cout << "_dSigma[" << _dSigma << "]" << endl;
    cout << "** _sMethod   [" << _sMethod << "]" << endl;
    cout << "TSamplerField::fin de PrintLog[]" << endl;
}

} // Fin du namespace Sampler
} // Fin du namespace URANIE

// fin du fichier $RCSfile$.
