//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TSamplerField.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Tue Mar  7 14:06:02 2006
 
 \brief  Creation of the abstract class TSamplerField
 
 
 */
#ifndef TSamplerField_H
#define TSamplerField_H

////// ROOT
#include "TMatrixD.h"
#include "TVectorD.h"
#include "TVector3.h"
#include "TGraph2D.h"

////// URANIE
#include "TSamplerStochastic.h"

namespace URANIE
{
namespace Sampler
{
class TSamplerField: public TSamplerStochastic
{
    enum SimulationMethod
    {
        kSphericalVariogram,
        kExponentialVariogram,
        kStableVariogram,
        kHyperbolicVariogram,
        kGaussianVariogram,
        kSinusCardinalVariogram,
        kSpectralMethod,
        kDilutionMethod,
        kTessellationMethod,
        kTurningBandsMethod
    };

protected:
    TMatrixD _vdc;                 //! Matrix _nRandomFunction * 3
    TMatrixD _ap;                  //! Matrix 3 * 3 
    TVectorD _scale;               //! Vector of the norms 
    TVectorD _seed;                //! Vectors of the seeds 
    TGraph2D* _graph2D;             //! Visu 2 D
    vector<URANIE::DataServer::UEntry*> in;
    
public:
    Int_t _nXSize;
    Int_t _nYSize;
    Int_t _nZSize;
    Int_t _nRandomFunction;        ///< Number of basic random functions
    Int_t _nMethod;                ///< The method of simulation
    Int_t _nVariogram;             ///< The variogram type
    Bool_t _bIsotropy;             ///< Boolean to set Isotropy/ Anisotropy
    Double_t _dVariance;           ///< Variance
    Double_t _dSigma;              ///< Sigma
    TVector3 _vScaleFactor;        ///< Scale factor of the axis
    Double_t _valcrt;
protected:
//       virtual void createTuple();

    // The spectral method of simulation
    void SpectralInit();
    void SpectralPrepare();
    void SpectralSimulate(unsigned int ievt);

    void VDC();

public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Constructor with a dataserver
    TSamplerField(URANIE::DataServer::TDataServer *tds, Option_t *option =
                  "Gauss", int nS=1);
    //! Default destructor
    virtual ~TSamplerField();
    //@}

    //---------------------------------------------
    /** @name Generates the sample
     **/
    //@{
    //! Generates the sample 
    /*!
     */
    void generateSample(Option_t *option = "");
    //@}

    //---------------------------------------------
    /** @name Treats the TAttributes
     **/
    //@{
    //! Creates the List of attributes to simulate
    /*
     In this case, the space is represented by TAttribute attributes.
     */
    void createListOfAttributes();
    //@}

    //---------------------------------------------
    /** @name Setting and Getting attributs
     **/
    //@{
    //! Sets the method given by the name
    /*!
     \param str (TString) the method name ("spectral"|"dilution"|"tessellation"|"turningbands"| )
     */
    void setMethod(TString str);
    //! Returns the name of the method
    /*!
     \return "Spectral Method"|"Dilution Method"|"Tessellation Method"|Turning Bands Method|"unknown Bands Method"|
     */
    TString getMethodName();

    //! Sets the variogramm name
    /*!
     \param str (TString) The name of variogramm ("spherical"|"exponential"|"stable"|"hyperbolic"|"gaussian"|"sine")
     */
    void setVariogram(TString str);
    //! Returns the name of the variogramm
    /*!
     \return "Spherical"|"Exponential"|"Stable"|"Hyperbolic"|"Gaussian"|"SineCardinal"
     */
    TString getVariogramName();

    //@}

    // Setter and Getter
    void SetRandomFunction(Int_t nind)
    {
        _nRandomFunction = nind;
    }
    Int_t GetRandomFunction()
    {
        return _nRandomFunction;
    }
    void SetIsotropy(Bool_t bval)
    {
        _bIsotropy = bval;
    }
    Bool_t IsIsotropy()
    {
        return _bIsotropy;
    }
    void SetVariance(Double_t dval);
 
    Double_t GetVariance()
    {
        return _dVariance;
    }
    void SetScaleFactor(Double_t dval);

    void Draw2D(TString sx, TString sy, Option_t* opt = "TRI2");

//       void AddAttribute(Option_t *option, URANIE::DataServer::TAttribute *att, Int_t ind = 100);

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Sampler::TSamplerField, ID_SAMPLER)
    //Abstract class for generation of random design.
};

} // Fin du namespace Sampler
} // Fin du namespace URANIE
#endif
// fin du fichier $RCSfile$.
