//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*! \file TSampling.h
 *  \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 *  \date   $Date$
 *  \brief _descritption_.
 *
 *  Created: Fri Oct  1 13:37:23 2004
 *
 */
#ifndef TSAMPLING_H
#define TSAMPLING_H

#ifndef ROOT_TMatrixD
#include "TMatrixD.h"
#endif
#ifndef ROOT_TVectorD
#include "TVectorD.h"
#endif
#ifndef ROOT_TNtuple
#include "TNtuple.h"
#endif
#ifndef ROOT_TString
#include "TString.h"
#endif
#ifndef ROOT_TNtupleD
#include "TNtupleD.h"
#endif
#ifndef ROOT_TNamed
#include "TNamed.h"
#endif

// Uranie
#include "TSamplerStochastic.h"
#include "TAttribute.h"
#include "TStochAttribut.h"
#include "TDataServer.h"
#include "LHSInput.H"

namespace URANIE
{
namespace Sampler
{
class TSampling: public TSamplerStochastic
{
    // Associations

    // Attributes
public:
    enum EType
    {
        kLHS, kSRS, kUnknown
    };
    LHSInput* _fLhsInput;
    TMatrixD _corrmatrix;             ///< The correlation matrix
    TMatrixD _corrIndexPosition; ///< The permutation for giving the rank correlation matrix given by the user
    EType _ntype;                  ///< The type of sampling (SRS, LHS)
    bool _buseSVD;                  ///< useSVD instead of Cholesky to factorize the input correlation matrix
    bool _buseSVDquiet;                  ///< useSVD instead of Cholesky to factorize the input correlation matrix
    
    // Operations
private:
    //---------------------------------------------
    /** @name private methods
     **/
    //@{
    //!  Decomposes the correlation matrix
    /*!
     Applies the cholesky decomposition
     */
    void CorrelationFactorization();
    //! Computes the correlation matrix for a TMatrixD
    void computeCorrelationMatrix(TMatrixD matOrg);
    //@}

public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //!  Constructor from a TDataServer, the name of the method and the size of the sample
    /*!
     Attributes to sample are specified in the TDataServer. The name of the method is "lhs"|"srs". The size is an integer greater than zero.
     
     \par Default values
     \li Method ="lhs" with default value or when the option is uncorrect
     \li size = 100 with default value.
     
     \param tds (TDataServer *) the pointer of the TDataServer. It must contains objects from daughter classes of TStochasticAttribute
     \param option (Option_t *) Method of sampling [lhs]
     \param nCalcul (Int_t) Size of the sampler to generate [1000]
     */
    TSampling(URANIE::DataServer::TDataServer *tds, Option_t *option = "lhs",
            Int_t nCalcul = 1000);
    //! Default destructor
    virtual ~TSampling();
    //@}

    //---------------------------------------------
    /** @name Generate the sample
     We have several methods which can generate the sampler.
     \li the prefered method generateSample which used the Devroye algorithm to create permutation.
     This method does not treat the correlation (2008-04-01 10:13:47).
     \li the method from the Dakota module generateSampleByDakota
     \li a deprecated method genSample (from version v0.3 to v0.4)
     **/
    //@{
    //! Generates the sample 
    /*!
     Uses the Devroye algorithm to create permutation
     \code
     for(Int_t j=nS-1; j>1; j--)
     k = (Int_t)gRandom->Integer(j);
     swap(Ak, Aj);
     \endcode
     */
    void generateSample(Option_t *option = "");
    //! Generates the sample by the Dakota module
    /*!
     The size of sample is limited to 9990. 
     */
    void generateSampleByDakota();

    //---------------------------------------------
    /** @name Correlation matrix
     Methods to specify correlation between attributes
     **/
    //@{
    //! Defines a correlation between two attributes given by their indexes
    /*!
     \warning The indexes begin 0 to n-1.
     
     \param indx (Int_t) the index of the first attribute
     \param indy (Int_t) the index of the second attribute
     \param value (Double_t) the correlation factor between the two attributes
     */
    void setUserCorrelation(Int_t indx, Int_t indy, double value);

    //! Defines a correlation between two attributes given by their names
    /*!
     
     \param xname (TString) the first attribute
     \param yname (TString) the second attribute
     \param value (Double_t) the correlation factor between x and y
     */
    void setUserCorrelation(TString xname, TString yname, double value);

    //! Defines a correlation between two attributes
    /*!
     The correlation factor is define as the
     
     \param x (URANIE::DataServer::TAttribute *) the first attribute
     \param y (URANIE::DataServer::TAttribute *) the second attribute
     \param value (Double_t) the correlation factor between x and y
     */
    void setUserCorrelation(URANIE::DataServer::TAttribute *x,
            URANIE::DataServer::TAttribute *y, Double_t value);

    //! Set the correlation matrix
    /*!
     The correlation matrix is define from a provided matrix.

     \param the input correlation matrix requested by the user.
    */
    void setCorrelationMatrix(TMatrixD corrMat);
    
    //@}

    //! EMPTY:: TSampling::setSaveRang
    /*!
     \brief Documentation of the function setSaveRang.
     \param rang (Int_t) (default = 0)
     */
    void setSaveRang(Int_t rang = 0);
    void setTypeSampling(Option_t *option);

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    //! Prints the log
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Sampler::TSampling, ID_SAMPLER)
    //Definition d'un plan d'exp�rience
};

} // Fin du namespace Sampler
} // Fin du namespace URANIE
#endif
// fin du fichier $RCSfile$.
