//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TSamplingMixture.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Mon Jul 24 14:15:38 2006

 \brief  Interface for sampling with mixture


 */
#define DEPSILON 1.e-8

////// ROOT
#include "TRandom3.h"
#include "TLeaf.h"

////// URANIE
#include "TSamplingMixture.h"
#include "UExceptions.h"

ClassImp (URANIE::Sampler::TSamplingMixture)

namespace URANIE
{
namespace Sampler
{
//////////////////////////////////////////////////////////////
//  Constructor : from TDataServer + Methode + size
////////////////////////////////////////////////////////////
TSamplingMixture::TSamplingMixture(URANIE::DataServer::TDataServer *tds,
        Int_t ns, Double_t dThreshold, Option_t *option) :
        TSamplerDeterminist(tds, "mixture", ns), _dThreshold(dThreshold), _simplex(NULL), _bSummits(kTRUE), 
                _bCenterEdges(kTRUE),_bThirdEdges(kTRUE), _bCenterFaces(
                kTRUE), _bCentroide(kTRUE), _bCenterSegments(kTRUE), _bSaveCandidates(
                kFALSE), _sAttNameTypeOfPoints("__typePoints__uranie__samplingMixture__"),
                _sAttNameCandidate("__candidatesPoints__uranie__samplingMixture__"),
                _nGenerateSample(0), _bTrace(kFALSE), _sTraceFile("_tsmix_trace_.log")
{
    setMethodName(Form("mixture_%d", _nS));
    SetNameTitle(Form("Sampling_mixture_%d", _nS),
            Form("Uranie %d sample with mixture", _nS));

    // Treat the option
    TString opt = option;
    opt.ToLower();
    if (opt.Contains("save"))
        _bSaveCandidates = kTRUE;

    init();
}

TSamplingMixture::~TSamplingMixture()
{
    if (_blog)
        cout << " ** Destroy TSamplingMixture" << endl;

    if (_simplex)
      delete _simplex;
    _simplex = NULL;
}

/*******************************************/
/*       initPlanes        */
/*******************************************/
void TSamplingMixture::init()
//    vector<vector<double> > TSamplingMixture::initPlanes()
{
    if (_tds->getNPatterns() == 0)
    {
        // Add internal attribute for the type of points (Summit|CenterEdge|ThirdEdge|CenterFace|CenterSegment|Centroide)
        URANIE::DataServer::TAttribute* attType =
                new URANIE::DataServer::TAttribute(_sAttNameTypeOfPoints);
        attType->setOrigin(URANIE::DataServer::TAttribute::kInternal);
	attType->delShare();
        _tds->addAttribute(attType);

        // Add internal attribute for candidates points (equal to 0) or solution (greater or equal to 1 and equal to _nGenerateSample attribut)
        attType = new URANIE::DataServer::TAttribute(_sAttNameCandidate);
        attType->setOrigin(URANIE::DataServer::TAttribute::kInternal);
	attType->delShare();
        _tds->addAttribute(attType);

        // Create the TTree of data
        _tds->createTuple();
    }
    else
    {
        // Case of the TDS is not empty: Verify the internal attributes exist
        try
        {

            if (_tds->isAttribute(_sAttNameTypeOfPoints))
            {
                _tds->getAttribute(_sAttNameTypeOfPoints)->setOrigin(
                        URANIE::DataServer::TAttribute::kInternal);
            }
            else
            {
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                        Form(
                                "TSamplingMixture::init method \n The internal attribut [%s] is not found",
                                _sAttNameTypeOfPoints.Data()));
            }

            if (_tds->isAttribute(_sAttNameCandidate))
            {
                _tds->getAttribute(_sAttNameCandidate)->setOrigin(
                        URANIE::DataServer::TAttribute::kInternal);
            }
            else
            {
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                        Form(
                                "TSamplingMixture::init method \n The internal attribut [%s] is not found",
                                _sAttNameCandidate.Data()));
            }

        } catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            throw ue;
        }

    }

    vector<double> vec;

    if (_lstOfAttributesToSample != NULL)
        delete _lstOfAttributesToSample;
    _lstOfAttributesToSample = NULL;

    _lstOfAttributesToSample = _tds->getListOfAttributes();

    URANIE::DataServer::TAttribute *curatt;
    TIter next(_lstOfAttributesToSample);
    Int_t iatt = 1;
    Int_t natt = _lstOfAttributesToSample->GetEntries();
    Double_t dBound;
    while ((curatt = (URANIE::DataServer::TAttribute *) next()))
    {
        if (_blog)
        {
            cout << " ** iatt[" << iatt << "] name[" << curatt->GetName() << "]"
                    << endl;
        }

        // If Lower Bound
        if (curatt->isLowerBounded())
        {
            dBound = curatt->getLowerBound();
            try
            {
                if (dBound < DEPSILON)
                    throw URANIE::Exceptions::UInfoExceptions(__FILE__,
                            __LINE__,
                            Form(
                                    "TSamplingMixture::initPlanes\n The Lower bound of attribute[%s] %e is lower than epsilon[%e]",
                                    curatt->GetName(), dBound, DEPSILON));

                if (_blog)
                    cout << "   ** LowerBound[" << dBound << "]" << endl;
                vec.clear();
                vec.push_back(-1.0 * dBound);
                for (Int_t j = 1; j < iatt; j++)
                    vec.push_back(0);
                vec.push_back(1.0);
                for (Int_t j = iatt; j < natt; j++)
                    vec.push_back(0);

                // LOG
                if (_blog)
                {
                    for (Int_t j = 0; j < natt + 1; j++)
                        cout << " **** vec[" << j << "] = [" << vec[j] << "]"
                                << endl;
                }

                _planes.push_back(vec);
            } catch (URANIE::Exceptions::UInfoExceptions&)
            {
                //       ue.printMessage();
            }

        }

        // If Upper Bound
        if (curatt->isUpperBounded())
        {
            dBound = curatt->getUpperBound();
            if (_blog)
                cout << "   ** UpperBound[" << dBound << "]" << endl;
            vec.clear();
            vec.push_back(dBound);
            for (Int_t j = 1; j < iatt; j++)
                vec.push_back(0);
            vec.push_back(-1.0);
            for (Int_t j = iatt; j < natt; j++)
                vec.push_back(0);

            // LOG
            if (_blog)
            {
                for (Int_t j = 0; j < natt + 1; j++)
                    cout << " **** vec[" << j << "] = [" << vec[j] << "]"
                            << endl;
            }

            _planes.push_back(vec);
        }

        // increment iatt
        iatt++;
    }

}
/***************************************************/
/*  Set The List of Candidates Points */
/***************************************************/
void TSamplingMixture::setListOfCandidatesPoints()
{
    if (_blog)
        cout << " ********\n ** setListOfCandidatesPoints" << endl;
    Int_t natt = _tds->getNAttributes();
    Int_t npat = _tds->getNPatterns();
    if (_blog)
        cout << " ** natt[" << natt << "] npat[" << npat << "]" << endl;

    // Init the List of Points
    list < vector<double> > lstOfSummits;
    list < vector<double> > lstOfCenterEdges;
    list < vector<double> > lstOfThirdEdges;
    list < vector<double> > lstOfCenterFaces;
    list < vector<double> > lstOfCenterSegments;

    // the cureent vector of point
    vector<double> vecCrt;

    TLeaf *lTypeOfPoints = _tds->getTuple()->GetLeaf(_sAttNameTypeOfPoints);
    TLeaf *lCandidate = _tds->getTuple()->GetLeaf(_sAttNameCandidate);

    URANIE::DataServer::TAttribute *tatt;
    TIter next(_lstOfAttributesToSample);

    for (Int_t ipat = 0; ipat < npat; ipat++)
    {
        _tds->getTuple()->GetEntry(ipat);

        if (_blog)
            cout << " ** ipat[" << ipat << "/" << npat << "] -- TypeOfPoints["
                    << lTypeOfPoints->GetValue() << "] -- Candidate["
                    << lCandidate->GetValue() << "] - ";

        _nGenerateSample = (Int_t) lCandidate->GetValue();

        if (lCandidate->GetValue() == 0)
        {
            // Case of a candidate point
            Int_t typeOfPoint = (Int_t)(lTypeOfPoints->GetValue());
            if (_blog)
                cout << " ** candidat (" << typeOfPoint << ") - ";

            // Fill the vector of data
            vecCrt.clear();
            next.Reset();
            while ((tatt = (URANIE::DataServer::TAttribute *) next()))
            {
                TLeaf *leaf = _tds->getTuple()->GetLeaf(tatt->GetName());
                vecCrt.push_back(leaf->GetValue());
            }

            switch (typeOfPoint)
            {
            case kSummit:
            {
                if (_blog)
                    cout << " ** kSummit" << endl;
                lstOfSummits.push_back(vecCrt);
                break;
            }
            case kCenterEdge:
            {
                if (_blog)
                    cout << " ** kCenterEdge" << endl;
                lstOfCenterEdges.push_back(vecCrt);
                break;
            }
            case kThirdEdge:
            {
                if (_blog)
                    cout << " ** kThirdEdge" << endl;
                lstOfThirdEdges.push_back(vecCrt);
                break;
            }
            case kCenterFace:
            {
                if (_blog)
                    cout << " ** kCenterFace" << endl;
                lstOfCenterFaces.push_back(vecCrt);
                break;
            }
            case kCenterSegment:
            {
                if (_blog)
                    cout << " ** kCenterSegment" << endl;
                lstOfCenterSegments.push_back(vecCrt);
                break;
            }
            case kCentroide:
            {
                if (_blog)
                    cout << " ** kCentroide" << endl;
                _simplex->setCentroide(vecCrt);
                break;
            }
            }
        }
        else
        {
            if (_blog)
                cout << " not a candidate" << endl;
        }
    }

    // Set the lists of points to the Simplex object
    _simplex->setSummits(lstOfSummits);
    _simplex->setCenterEdges(lstOfCenterEdges);
    _simplex->setThirdEdges(lstOfThirdEdges);
    _simplex->setCenterFaces(lstOfCenterFaces);
    _simplex->setCenterSegments(lstOfCenterSegments);
}

/***************************************************/
/*  Fill the points in a list of candidates points */
/***************************************************/
void TSamplingMixture::fillListOfCandidatesPoints(
        list<vector<double> > lstOfPoints, EType typePoint)
{
    if (_blog)
        cout << " ********\n ** fillListOfCandidatesPoints size("
                << lstOfPoints.size() << ") type[" << typePoint << "]" << endl;
    Int_t natt = _tds->getNAttributes();
    Int_t ipat = _tds->getNPatterns();
    if (_blog)
        cout << " ** natt[" << natt << "] ipat[" << ipat << "]" << endl;
    Int_t nvar = natt + 1;

    Double_t *values = new Double_t[nvar + 2];
    list<vector<double> >::iterator it1;

    for (it1 = lstOfPoints.begin(); it1 != lstOfPoints.end(); it1++)
    {
        values[0] = (Double_t)(ipat + 1);
        for (size_t i = 0; i < (*it1).size(); i++)
            values[i + 1] = (*it1)[i];
        values[natt + 1] = typePoint;
        values[natt + 2] = 0;

        _tds->getTuple()->Fill(values);
        ipat++;
    }

    // Clean
    delete[] values;
    values = NULL;
}

/*******************************************/
/*           Treat the trace information    */
/*******************************************/
void TSamplingMixture::setTrace(TString straceFile)
{
    _bTrace = kTRUE;
    if (straceFile.Length() > 0)
        _sTraceFile = straceFile;
}

TString TSamplingMixture::getTraceFile()
{
    TString sReturn = "";
    if (_bTrace)
        sReturn = _sTraceFile;
    return sReturn;
}

/*******************************************/
/*           Generate the candidates       */
/*******************************************/
void TSamplingMixture::generateCandidates(Option_t *option)
{
    if (_blog)
        cout << " ** TSamplingMixture::generateCandidates[" << option << "]"
                << endl;

    _simplex = new Simplex(_tds->getNAttributes(), _planes, _dThreshold, _blog);
    //       _simplex->setRefFile("reference.csv");

    int ret = 0;
    std::ofstream fTraceFile;
    if (_bTrace)
        fTraceFile.open(_sTraceFile.Data(), std::ios::out);

    // Compute the summits
    if (_bSummits)
    {
        if (_blog)
            cout << " ** Compute Summits" << endl;
        if (!_simplex->Summits())
            ret = 1;
        if (_bTrace)
        {
            fTraceFile << "Summits" << endl;
            fTraceFile.flush();
        }
    }

    // Case of Edges
    if ((_bCenterEdges == kTRUE) && (_bThirdEdges == kTRUE))
    {
        if (_blog)
            cout << " ** Compute PointsOnEdges" << endl;
        if (!_simplex->PointsOnEdges())
            ret = 1;
        if (_bTrace)
        {
            fTraceFile << "CenterEdges" << endl;
            fTraceFile << "ThirdEdges" << endl;
            fTraceFile.flush();
        }
    }
    else
    {
        if (_bCenterEdges)
        {
            if (_blog)
                cout << " ** Compute CenterEdges" << endl;
            if (!_simplex->CenterEdges())
                ret = 1;
            if (_bTrace)
            {
                fTraceFile << "CenterEdges" << endl;
                fTraceFile.flush();
            }
        }

        if (_bThirdEdges)
        {
            if (_blog)
                cout << " ** Compute ThirdEdges" << endl;
            if (!_simplex->ThirdEdges())
                ret = 1;
            if (_bTrace)
            {
                fTraceFile << "ThirdEdges" << endl;
                fTraceFile.flush();
            }
        }
    }

    // Case of CenterFaces
    if (_bCenterFaces)
    {
        if (_blog)
            cout << " ** Compute CenterFaces" << endl;
        if (!_simplex->CenterFaces())
            ret = 1;
        if (_bTrace)
        {
            fTraceFile << "CenterFaces" << endl;
            fTraceFile.flush();
        }
    }

    // Case of Centroide
    if (_bCentroide || _bCenterSegments)
    {
        if (_blog)
            cout << " ** Compute Centroide" << endl;
        _simplex->Centroide();
        if (_bTrace)
        {
            fTraceFile << "Centroide" << endl;
            fTraceFile.flush();
        }
    }

    // Case of CenterSegments
    if (_bCenterSegments)
    {
        if (_blog)
            cout << " ** Compute CenterSegments" << endl;
        if (!_simplex->CenterSegments())
            ret = 1;
        if (_bTrace)
        {
            fTraceFile << "CenterSegments" << endl;
            fTraceFile.flush();
        }
    }

    if (_bTrace)
        fTraceFile.close();

    try
    {
        if (ret)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TSamplingMixture::generateCandidates\n the returm of simplex is zero.\n No points have been generated"));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    // Fill the candidates points in the TDS if _bSaveCandidates
    if (_bSaveCandidates)
    {
        // Case of summits
        if (_bSummits)
            fillListOfCandidatesPoints(_simplex->getSummits(), kSummit);

        // Case of CenterEdges
        if (_bCenterEdges)
            fillListOfCandidatesPoints(_simplex->getCenterEdges(), kCenterEdge);

        // Case of ThirdEdges
        if (_bThirdEdges)
            fillListOfCandidatesPoints(_simplex->getThirdEdges(), kThirdEdge);

        // Case of CenterFaces
        if (_bCenterFaces)
            fillListOfCandidatesPoints(_simplex->getCenterFaces(), kCenterFace);

        // Case of CenterSegments
        if (_bCenterSegments)
            fillListOfCandidatesPoints(_simplex->getCenterSegments(),
                    kCenterSegment);

        // Case of Centroide
        if (_bCentroide)
        {
            list < vector<double> > lstOfPoints;
            lstOfPoints.push_back(_simplex->getCentroide());
            fillListOfCandidatesPoints(lstOfPoints, kCentroide);
        }
    }

    if (_blog)
        cout << " ** End Of TSamplingMixture::generateCandidates[" << option
                << "]" << endl;
}

/*******************************************/
/*           Generate the sample           */
/*******************************************/
void TSamplingMixture::generateSample(Option_t *option)
{
    if (_blog)
        cout << " ** TSamplingMixture::generateSample[" << option << "] npat["
                << _tds->getNPatterns() << "] natt [" << _tds->getNAttributes()
                << "]" << endl;

    try
    {
        if (_tds->getNPatterns() == 0)
            generateCandidates("");
        else
        {
            if (_simplex == NULL)
            {
                _simplex = new Simplex(_tds->getNAttributes(), _planes,
                        _dThreshold, _blog);
                setListOfCandidatesPoints();
            }
        }

        TString soption = option;
        soption.ToLower();

        Int_t nTypeOfAlgorithm = 1;  // MD by default
        if (soption.CompareTo("mmd") == 0) // Case of MMD
            nTypeOfAlgorithm = 2;

        switch (nTypeOfAlgorithm)
        {
        case 1:
        { // Case of "OptimalPointsFromMaximalDistance"
            _simplex->OptimalPointsFromMaximalDistance(_nS);
            break;
        }
        case 2:
        { // Case of OptimalPointsFromMaximalMinDistance
            _simplex->OptimalPointsFromMaximalMinDistance(_nS);
            break;
        }
        }

        // Fill The Data
        vector < vector<double> > optimalPoints = _simplex->getOptimalPoints();
        Int_t nsopt = (Int_t)(optimalPoints.size());
        if (_nS != nsopt)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TSamplingMixture::generateSample\n The size of sampling [%d] is not equal to the size founded [%d] by the method",
                            _nS, nsopt));

        size_t natt = _tds->getNAttributes();
        if (optimalPoints[0].size() != natt)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TSamplingMixture::generateSample\n The size of vector [%lu] is not equal to the number of attributes [%zu]",
                            optimalPoints[0].size(), natt));

        vector < string > optimalTypes = _simplex->getOptimalTypes();

        // Fill the data
        Int_t ipat = _tds->getNPatterns();
        Int_t nvar = natt + 1;
        Double_t *values = new Double_t[nvar + 2];

        // Increment the number of call the generateSample method
        _nGenerateSample++;

        for (Int_t j = 0; j < nsopt; j++)
        {
            if (_blog)
                cout << " ** j[" << j << "] Type[" << optimalTypes[j]
                        << "] --- ";
            values[0] = (Double_t)(ipat + 1);
            for (size_t i = 0; i < optimalPoints[j].size(); i++)
            {
                values[i + 1] = optimalPoints[j][i];
            }
            values[nvar + 1] = -1;
            if (!strcmp(optimalTypes[j].c_str(), "Sommet"))
                values[natt + 1] = kSummit;
            else if (!strcmp(optimalTypes[j].c_str(), "CentreArete"))
                values[natt + 1] = kCenterEdge;
            else if (!strcmp(optimalTypes[j].c_str(), "TiersArete"))
                values[natt + 1] = kThirdEdge;
            else if (!strcmp(optimalTypes[j].c_str(), "CentreFace"))
                values[natt + 1] = kCenterFace;
            else if (!strcmp(optimalTypes[j].c_str(), "CentreSegment"))
                values[natt + 1] = kCenterSegment;
            else if (!strcmp(optimalTypes[j].c_str(), "Centroide"))
                values[natt + 1] = kCentroide;

            if (_blog)
                cout << "values[" << values[nvar + 1] << "]" << endl;
            values[natt + 2] = (Double_t)(_nGenerateSample);

            _tds->getTuple()->Fill(values);
            ipat++;
        }

        // Case of the centroide
        if (_bCentroide)
        {
            if (_blog)
                cout << " ** Save the Centroide in the sample [" << kCentroide
                        << "]" << endl;

            values[0] = (Double_t)(ipat++);

            vector<double> pointsOfCentroide = _simplex->getCentroide();
            for (size_t i = 0; i < pointsOfCentroide.size(); i++)
            {
                values[i + 1] = pointsOfCentroide[i];
            }
            values[natt + 1] = kCentroide;
            values[natt + 2] = (Double_t)(_nGenerateSample);
            _tds->getTuple()->Fill(values);
            ipat++;
        }
        // Clean
        delete[] values;
        values = NULL;

        // Active only the solution patterns from last optimization process
        TString sselect = Form("(%s==%d)", _sAttNameCandidate.Data(),
                _nGenerateSample);
        if (_bSaveCandidates)
            sselect += Form("|| (%s==0)", _sAttNameCandidate.Data());
        _tds->setSelect(sselect);

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}
/*******************************************/
/*      Add relationnal constraints        */
/*******************************************/
void TSamplingMixture::addConstraint(TVectorD vecConstraint)
{
    if (_blog)
    {
        cout << " ** TSamplingMixture::addConstraint Relation Constraint";
        vecConstraint.Print();
    }

    vector<double> vec;
    vec.clear();
    for (Int_t i = 0; i < vecConstraint.GetNrows(); i++)
        vec.push_back(vecConstraint[i]);
    _planes.push_back(vec);

}

////////////////////////////////////////////////////////////
//  Printing Log
////////////////////////////////////////////////////////////
void TSamplingMixture::printLog(Option_t *option)
{
    cout << "TSamplingMixture::printLog[" << option << "] Name[" << GetName()
            << "] Title[" << GetTitle() << "]" << endl;
    cout << " ** _dThreshold[" << _dThreshold << "]" << endl;
    cout << " ** _nGenerateSample[" << _nGenerateSample << "]" << endl;
    cout << " ** _sAttNameTypeOfPoints[" << _sAttNameTypeOfPoints << "]"
            << endl;
    cout << " ** _sAttNameCandidate[" << _sAttNameCandidate << "]" << endl;

    cout << " ** Save Candidates[";
    if (_bSaveCandidates)
        cout << "Yes";
    else
        cout << "No";
    cout << "]" << endl;
    cout << " ** Trace[";
    if (_bTrace)
        cout << "Yes in file[" << _sTraceFile << "]";
    else
        cout << "No";
    cout << "]" << endl;
    TSampler::printLog();
    cout << "End of TSamplingMixture::printLog" << endl;
}

} // Fin du namespace Sampler
} // Fin du namespace URANIE
// fin du fichier $RCSfile$.
