//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TSpaceFilling.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Mon Jul 24 14:15:38 2006

 \brief  Interface for the sampling by TSpaceFilling
 */
////// ROOT
#include "TMath.h"
#include "TRandom3.h"

////// URANIE
#include "TSpaceFilling.h"
#include "TStochAttribut.h"
#include "THaltonSequence.h"
#include "TSobolSequence.h"
#include "UExceptions.h"

int URANIE__FAST__OMEGA[50] =
{ 0, 0, 1, 5, 11, 1, 17, 23, 19, 25, 41, 31, 23, 87, 67, 73, 85, 143, 149, 99,
        119, 237, 267, 283, 151, 385, 157, 215, 449, 163, 337, 253, 375, 441,
        673, 773, 875, 873, 587, 849, 623, 637, 891, 943, 1171, 1225, 1335,
        1725, 1663, 2019 };
int URANIE__FAST__D[50] =
{ 4, 8, 6, 10, 20, 22, 32, 40, 38, 26, 56, 62, 46, 76, 96, 60, 86, 126, 134,
        112, 92, 128, 154, 196, 34, 416, 106, 208, 328, 198, 382, 88, 348, 186,
        140, 170, 284, 568, 302, 438, 410, 248, 448, 388, 596, 216, 100, 488,
        166, 0 };

using std::cout;
using std::endl;

ClassImp (URANIE::Sampler::TSpaceFilling)

namespace URANIE
{
namespace Sampler
{
//////////////////////////////////////////////////////////////
//  Constructor : from TDataServer + Methode + size
////////////////////////////////////////////////////////////
TSpaceFilling::TSpaceFilling(URANIE::DataServer::TDataServer *tds,
        Option_t *option, Int_t ns, EFunction nFunction) :
        TSamplerDeterminist(tds, "SpaceFilling", ns), _nValFreq(NULL), _nFunction(
                nFunction), _bPermutation(kFALSE)
{
    TString soption = option;
    soption.ToLower();
    setMethodName(Form("SpaceFilling_%d", nFunction));
    SetNameTitle(Form("Sampling_SpaceFilling_%d", ns),
            Form("Uranie %d sample with SpaceFilling method %d", ns,
                    nFunction));

    //  _lstOfAttributesToSample = _tds->getListOfAttributes("*");
    //  _nX = _lstOfAttributesToSample->GetEntries();
    //  cout << endl << " **  Attributes listing  [" << _nX << "]" << endl;
    //  _lstOfAttributesToSample->Print();

}

TSpaceFilling::~TSpaceFilling()
{
    if ( _blog )
        cout << " *** destroy TSpaceFilling(" << GetName() << ") Title[" << GetTitle() << "] *** " << endl;
    if (_nValFreq)
        delete[] _nValFreq;

    if (_lstOfAttributesToSample != NULL)
      delete _lstOfAttributesToSample;
    _lstOfAttributesToSample = NULL;

    _nValFreq = NULL;
}

///////////////////////////////////////////////////////////
//  Frequencies
///////////////////////////////////////////////////////////
void TSpaceFilling::setDefaultFrequencies()
{

    if (_blog)
    {
        cout << " ***********************************" << endl;
        cout << " ***** TSpaceFilling::setDefaultFrequencies _nX[" << _nX << "]"
                << endl;
    }
    if (_nValFreq == NULL)
        _nValFreq = new Int_t[_nX];

    if (_nX == 2)
    {
        _nValFreq[0] = 5;
        _nValFreq[1] = 9;
    }
    else
    {
        _nValFreq[0] = URANIE__FAST__OMEGA[_nX - 1];
        for (Int_t i = 1; i < _nX; i++)
            _nValFreq[i] = _nValFreq[i - 1] + URANIE__FAST__D[_nX - 2 - i + 1];
    }
    if (_blog)
    {
        for (Int_t i = 0; i < _nX; i++)
            cout << " ** setDefaultFrequencies i[" << i << "] Freq["
                    << _nValFreq[i] << "]" << endl;
    }

}
void TSpaceFilling::setFrequencies(Int_t nfreq, Int_t *valfreq)
{

    try
    {
        if (nfreq != _nX)
            throw "Error in setFrequencies. The number of frequencies is not good";
    } catch (...)
    {
        cout << "error" << endl;
        cout << "nfreq[" << nfreq << "] _nX[" << _nX << "]" << endl;
    }

    if (_nValFreq == NULL)
        _nValFreq = new Int_t[_nX];

    for (Int_t i = 0; i < _nX; i++)
    {
        _nValFreq[i] = valfreq[i];
    }
}

////////////////////////////////////////////////////////////
//  Generate the sample
////////////////////////////////////////////////////////////
void TSpaceFilling::generateSample(Option_t *option)
{
    TString soption = option;
    soption.ToLower();
    if (_blog)
        cout << "TSpaceFilling::GenerateSample" << endl;

    try
    {
        if (_tds->hasTuple())
            throw URANIE::Exceptions::UInfoExceptions(__FILE__, __LINE__,
                    Form(
                            "TSamplerStochastic::init: the TDS [%s] contains data: we need to empty it ! ",
                            _tds->GetName()));
    } catch (URANIE::Exceptions::UInfoExceptions& ue)
    {
        ue.printMessage();
  _tds->deleteTuple();
    }

    // Create the TTree of data
    createTuple();

    if (_blog)
        cout << "nvar[" << _nX << "]" << endl;

    if (_nValFreq == NULL)
    {
        setDefaultFrequencies();
    }

    Double_t * valpattern = new Double_t[_nX + 1];

    // Init the random phase-shift parameter in case of the kSaltelliB algorithm
    Double_t * valvarphi = new Double_t[_nX];
    gRandom->SetSeed(0);
    for (Int_t i = 0; i < _lstOfAttributesToSample->GetEntries(); i++)
    {
        valvarphi[i] = gRandom->Rndm() * 2 * TMath::Pi();
        if (_blog)
            cout << " ** Random Varphi[" << i << "] value[" << valvarphi[i]
                    << "]" << endl;
    }

    //// Loop on stochastic attributes to have PMin and PMax values
    URANIE::DataServer::TStochasticAttribute *stoatt;
    TIter next(_lstOfAttributesToSample);
    Int_t j = 0;
    Double_t ddelta = 1.0 / (Double_t)(_nS);
    if (_blog)
        cout << "ddelta[" << ddelta << "]" << endl;
    Double_t dproba, ds;
    // Loop on the patterns
    for (Int_t is = 0; is < _nS; is++)
    {
        valpattern[0] = (Double_t)(is + 1);
        j = 0;
        next.Reset();
        ds = -1.0 * TMath::Pi()
                + 2 * TMath::Pi() * ddelta * ((Double_t)(is) + 0.5);

        // Loop on stochastic attributes storing in the TList
        while ((stoatt = (URANIE::DataServer::TStochasticAttribute *) next()))
        {
            switch (_nFunction)
            {
            case kCukier:
            {
                dproba = TMath::Exp(TMath::Sin((Double_t)(_nValFreq[j]) * ds))
                        / TMath::Exp(1);
            }
                break;
            case kSaltelliA:
            {
                dproba = 0.5
                        + TMath::ASin(TMath::Sin((Double_t)(_nValFreq[j]) * ds))
                                / TMath::Pi();
            }
                break;
            case kSaltelliB:
            {
                dproba = 0.5
                        + TMath::ASin(
                                TMath::Sin(
                                        (Double_t)(_nValFreq[j]) * ds
                                                + valvarphi[j])) / TMath::Pi();
            }
                break;
            default:
                dproba = ds;
            }

            if (_blog)
                cout << "_nFunction[" << _nFunction << "] w[" << _nValFreq[j]
                        << "] dproba[" << dproba << "]";

            valpattern[j + 1] = stoatt->getInvCDF(dproba);
            if (_blog)
                cout << "** is[" << is << "] -- ds[" << ds << "] -- proba ["
                        << dproba << "] -- value[" << valpattern[j + 1] << "]"
                        << endl;
            j++;
        }
        _ntsample->Fill(valpattern);
    }

    // Affect the current TTree to the TDataServer
    _tds->setTuple(_ntsample);
    // Fill the formula attributes
    fillOtherAttributes();

    // Clean
    delete[] valpattern;
    valpattern = NULL;
    delete[] valvarphi;
    valvarphi = NULL;

}

////////////////////////////////////////////////////////////
//  Printing Log
////////////////////////////////////////////////////////////
void TSpaceFilling::printLog(Option_t *option)
{
    cout << "TSpaceFilling::printLog[" << option << "] Name[" << GetName()
            << "] Title[" << GetTitle() << "]" << endl;
    cout << "_nFunction[" << _nFunction << "]" << endl;
    cout << "_bPermutation[" << _bPermutation << "]" << endl;
    TSampler::printLog();
    if (_nValFreq)
    {
        cout << " ** Frequencies[";
        for (Int_t i = 0; i < _nX; i++)
        {
            cout << _nValFreq[i];
            if (i < _nX - 1)
                cout << ", ";
        }
        cout << "]" << endl;
    }
    else
        cout << "** Frequencies = NULL" << endl;

    cout << " ** Function[";
    switch (_nFunction)
    {
    case kCukier:
    {
        cout << "Cukier(73)";
    }
        break;
    case kSaltelliA:
    {
        cout << "Saltelli(99) sans Phi";
    }
        break;
    case kSaltelliB:
    {
        cout << "Saltelli(99) avec Phi";
    }
        break;
    default:
        cout << "Inconnu";
    }
    cout << "]" << endl;
    cout << "End of TSpaceFilling::printLog" << endl;
}

} // Fin du namespace Sampler
} // Fin du namespace URANIE
// fin du fichier $RCSfile$.
