//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*  Use numerical recipes in C styled memory management methods */

#include <stdlib.h>
#include <stdio.h>
#include "algebra_memory.h"

int **imatrix(int nrl, int nrh, int ncl, int nch)
{
    int i;
    int **m;

    m = (int **) malloc((unsigned) (nrh - nrl + 1) * sizeof(int*));
    if (!m)
    {
        fprintf(stderr, "Unable to allocate %d int* s.\n", nrh - nrl + 1);
        return m;
    }
    m -= nrl;

    for (i = nrl; i <= nrh; i++)
    {
        m[i] = (int *) malloc((unsigned) (nch - ncl + 1) * sizeof(int));
        if (!m[i])
        {
            fprintf(stderr,
                    "Unable to allocate %d'th row in an integer matrix.\n",
                    i - nrl + 1);
            return NULL;
        }
        m[i] -= ncl;
    }
    return m;
}

int *ivector(int nl, int nh)
{
    int *v;

    v = (int *) malloc((unsigned) (nh - nl + 1) * sizeof(int));
    if (!v)
    {
        fprintf(stderr, "Unable to allocate %d int s.\n", nh - nl + 1);
        return v;
    }
    return (v - nl);
}

/* ARGSUSED */
void free_imatrix(int **m, int nrl, int nrh, int ncl, int nch)
{
    int i;
    (void)(nch);
    
    for (i = nrh; i >= nrl; i--)
        free((char*) (m[i] + ncl));
    free((char*) (m + nrl));
}

/* ARGSUSED */
void free_ivector(int *v, int nl, int nh)
{
    (void)(nh); //  UNUSED 
    free((char*) (v + nl));
}

double **dmatrix(int nrl, int nrh, int ncl, int nch)
{
    int i;
    double **m;

    m = (double **) malloc((unsigned) (nrh - nrl + 1) * sizeof(double*));
    if (!m)
    {
        fprintf(stderr, "Unable to allocate %d double* s.\n", nrh - nrl + 1);
        return m;
    }
    m -= nrl;

    for (i = nrl; i <= nrh; i++)
    {
        m[i] = (double *) malloc((unsigned) (nch - ncl + 1) * sizeof(double));
        if (!m[i])
        {
            fprintf(stderr,
                    "Unable to allocate %d'th row in an double matrix.\n",
                    i - nrl + 1);
            return NULL;
        }
        m[i] -= ncl;
    }
    return m;
}

double *dvector(int nl, int nh)
{
    double *v;

    v = (double *) malloc((unsigned) (nh - nl + 1) * sizeof(double));
    if (!v)
    {
        fprintf(stderr, "Unable to allocate %d double s.\n", nh - nl + 1);
        return v;
    }
    return (v - nl);
}

/* ARGSUSED */
void free_dmatrix(int **m, int nrl, int nrh, int ncl, int nch)
{
    int i;
    (void)(nch);
    
    for (i = nrh; i >= nrl; i--)
        free((char*) (m[i] + ncl));
    free((char*) (m + nrl));
}

/* ARGSUSED */
void free_dvector(int **v, int nl, int nh)
{
    (void)(nh); //  UNUSED 
    free((char*) (v + nl));
}

/*  Add rows to a matrix:

 imat            pointer to that matrix
 oldrowsize      highest index+1 of old rows
 newrowsize      highest index+1 of new rows
 colsize         number of cols

 */

int grow_imatrix_byrows(int *** imat, int oldrowsize, int newrowsize,
        int colsize)
{
    int i;

    imat[0] = (int **) realloc(imat[0], (unsigned) (newrowsize) * sizeof(int*));
    if (!(imat[0]))
    {
        fprintf(stderr, "Unable to reallocate %d int* s.\n", newrowsize);
        return 0;
    }
    for (i = oldrowsize; i < newrowsize; i++)
    {
        imat[0][i] = (int *) malloc((unsigned) (colsize) * sizeof(int));
        if (!imat[0][i])
        {
            fprintf(stderr,
                    "Unable to reallocate %d'th row in an integer matrix.\n",
                    i);
            return 0;
        }
    }
    return 1;
}
