//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
#ifndef _SIMPLEX_HXX_
#define _SIMPLEX_HXX_

#include <vector>
#include <list>
#include <set>
#include <string>
#include <stdexcept>

class ErrorSimplex: public std::runtime_error
{
public:
    ErrorSimplex(const std::string& msg = "") :
            std::runtime_error(msg)
    {
    }
};

class Simplex
{
public:
    Simplex(const int ndim, const std::vector<std::vector<double> > planes,
            const double k, const bool trace = false);

    bool Summits();
    std::list<std::vector<double> > getSummits()
    {
        return _summits;
    }
    void setSummits(std::list<std::vector<double> > listOfPoints)
    {
        _summits.clear();
        _summits = listOfPoints;
    }

    bool PointsOnEdges();

    bool CenterEdges();
    std::list<std::vector<double> > getCenterEdges()
    {
        return _centerEdges;
    }
    void setCenterEdges(std::list<std::vector<double> > listOfPoints)
    {
        _centerEdges.clear();
        _centerEdges = listOfPoints;
    }

    bool ThirdEdges();
    std::list<std::vector<double> > getThirdEdges()
    {
        return _thirdEdges;
    }
    void setThirdEdges(std::list<std::vector<double> > listOfPoints)
    {
        _thirdEdges.clear();
        _thirdEdges = listOfPoints;
    }

    bool CenterFaces();
    std::list<std::vector<double> > getCenterFaces()
    {
        return _centerFaces;
    }
    void setCenterFaces(std::list<std::vector<double> > listOfPoints)
    {
        _centerFaces.clear();
        _centerFaces = listOfPoints;
    }

    void Centroide();
    std::vector<double> getCentroide()
    {
        return _centroide;
    }
    void setCentroide(std::vector<double> Point)
    {
        _centroide = Point;
    }

    bool CenterSegments();
    std::list<std::vector<double> > getCenterSegments()
    {
        return _centerSegments;
    }
    void setCenterSegments(std::list<std::vector<double> > listOfPoints)
    {
        _centerSegments.clear();
        _centerSegments = listOfPoints;
    }

    void setRefFile(const std::string refFile)
    {
        _refFile = refFile;
    }
    void OptimalPointsFromMaximalMinDistance(const int nbp, const bool iter =
            true);
    void OptimalPointsFromMaximalDistance(const int nbp);

    std::vector<std::vector<double> > getOptimalPoints()
    {
        return _optimalPoints;
    }
    std::vector<std::string> getOptimalTypes()
    {
        return _optimalTypes;
    }

private:
    size_t _ndim;                                 // dimension de l'espace
    size_t _Nplanes;                              // nombre de contraintes
    double _k;                                 // contrainte: sigma(xi)=k
    bool _trace;                               // flag de trace
    std::string _refFile;                           // fichier de reference
    std::vector<std::vector<double> > _teta;             // matrice des contraintes
    std::list<std::vector<double> > _dsummits; // valeur des contraintes pour chaque sommet

    std::list<std::vector<double> > _summits;            // sommets du simplex
    std::list<std::vector<double> > _centerEdges;        // centre des aretes
    std::list<std::vector<double> > _thirdEdges;         // tiers des aretes
    std::list<std::vector<double> > _centerFaces;        // centre des faces
    std::list<std::vector<double> > _centerSegments; // centre des segments: centroide - sommets
    std::vector<double> _centroide;                 // centroide des sommets

    std::vector<std::vector<double> > _optimalPoints;    // nbp points les plus eloignes
    std::vector<std::string> _optimalTypes;              // type des points optimaux
    std::vector<double> _optimalDists; // distance minimal entre les nbp points pour chacun des nbp points
    double _optimalDistMin;            // distance mionimal entre les nbp points
    std::vector<int> _optimalIndex;       // point le plus proche des points optimaux
    size_t _optimalIndexMin1, _optimalIndexMin2; // index des points dont la distance est minimale parmi les nbp
    size_t _nbp;                              // Nombre de points optimaux demandes
    size_t _index;           // index utilise pour la recherche des points optimaux
    std::vector<std::vector<double> > _points;           // points candidats
    double _maximalDist;        // distance maximale entre deux points candidats

    void calculTeta(const std::vector<std::vector<double> >& planes);
    void initSummits();
    std::vector<double> innerProduct(const std::vector<double>& teta);
    void checkConstraints(std::set<size_t> &cu);
    void dSummits(const size_t i0, const std::set<size_t>& cu);
    void displayPoints(const std::string title, std::list<std::vector<double> >& points);
    bool checkPoints(const std::string type, std::list<std::vector<double> >& points);
    void displayVector(const std::string name, const std::vector<double>& d);
    void isThisPointOptimal(const std::vector<double>& point, const std::string type);
    void calculateOptimalMin();
    double distance(const std::vector<double> &P1, const std::vector<double> &P2);
    void OptimalPointsFromList(std::list<std::vector<double> >& points,
            const std::string type);
    std::string getPointType(const size_t i);

};

#endif
