//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TestTBasicSampling.cxx
 \author Maud Salvoldi (OCC/Areva group)
 \date   $Date$
 
 \brief  Test File of URANIE::Sampler::TSampler class
 
 Cr�e le : mer nov 02 17:00:27 2011
 
 */

// Uranie
#include "TestTBasicSampling.h"
#include "TAttribute.h"
#include "TUniformDistribution.h"
#include "TNormalDistribution.h"
#include "TLogNormalDistribution.h"
#include "UDisableTest.h"

// ROOT
#include "TString.h"

#include <iostream>
#include <math.h>

#define DBL_PRECISION_COMPARE 1.e-10
#define REAL_PRECISION_COMPARE 1.e-6
#define EPSILON 0.04
#define CORRPRECISION 0.1 // This value is a bit high, but correlations are like this...

TBasicSamplingTest::TBasicSamplingTest() :
        _tds(NULL), _tds2(NULL), _bspl(NULL)
{
}

TBasicSamplingTest::~TBasicSamplingTest()
{
}
void TBasicSamplingTest::setUp()
{
    // Create a TDataServer
    _tds = new URANIE::DataServer::TDataServer();
    // Fill the DataServer with the three attributes of the study
    _tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("Uniform", 0., 1.));
    _tds->addAttribute(
            new URANIE::DataServer::TNormalDistribution("Normal", 0., 1.));
    _tds->getAttribute("Uniform")->delShare();
    _tds->getAttribute("Normal")->delShare();

    _tds2 = new URANIE::DataServer::TDataServer("tds2",
            "tds for correlated data");
    _tds2->addAttribute(
            new URANIE::DataServer::TUniformDistribution("x1", -1.0, 1.0));
    _tds2->addAttribute(
            new URANIE::DataServer::TNormalDistribution("x2", -5.0, 3.0));
    _tds2->addAttribute(
            new URANIE::DataServer::TLogNormalDistribution("x3", 2.0, 6.0));
    _tds2->addAttribute(
            new URANIE::DataServer::TNormalDistribution("x4", 1.0, 0.5));

    _tds2->getAttribute("x1")->delShare();
    _tds2->getAttribute("x2")->delShare();
    _tds2->getAttribute("x3")->delShare();
    _tds2->getAttribute("x4")->delShare();
}

void TBasicSamplingTest::tearDown()
{
    delete _tds;
    _tds = NULL;
    delete _tds2;
    _tds2=NULL;
}

void TBasicSamplingTest::testBasicSamplinggenerateSampleLHS()
{
  _bspl = new URANIE::Sampler::TBasicSampling(_tds, "lhs", 5000);
  _bspl->setSeed(1234);
  //   _bspl->setLog();
  _bspl->generateSample();
  _tds->computeStatistic("*");

  CPPUNIT_ASSERT_DOUBLES_EQUAL(0.5, _tds->getAttribute("Uniform")->getMean(),
			       EPSILON);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(0.288, _tds->getAttribute("Uniform")->getStd(),
			       EPSILON);

  CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, _tds->getAttribute("Normal")->getMean(),
			       EPSILON);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, _tds->getAttribute("Normal")->getStd(),
			       EPSILON);

  delete _bspl;
  _bspl = NULL;
}
void TBasicSamplingTest::testBasicSamplinggenerateSampleSRS()
{
    _bspl = new URANIE::Sampler::TBasicSampling(_tds, "srs", 5000);
     _bspl->setSeed(1234);
//   _bspl->setLog();
    _bspl->generateSample();
    _tds->computeStatistic("*");

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.5, _tds->getAttribute("Uniform")->getMean(),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.288, _tds->getAttribute("Uniform")->getStd(),
            EPSILON);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, _tds->getAttribute("Normal")->getMean(),
            0.05);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, _tds->getAttribute("Normal")->getStd(),
            EPSILON);

    delete _bspl;
    _bspl = NULL;
}

void TBasicSamplingTest::testBasicSamplinggenerateCorrSample()
{
    TMatrixD corrMat(4, 4);
    corrMat.UnitMatrix();
    corrMat[0][1] = 0.5;
    corrMat[1][0] = 0.5;
    corrMat[0][2] = 0.2;
    corrMat[2][0] = 0.2;
    corrMat[0][3] = -0.5;
    corrMat[3][0] = -0.5;
    corrMat[1][2] = 0.1;
    corrMat[2][1] = 0.1;
    corrMat[1][3] = -0.95;
    corrMat[3][1] = -0.95;
    corrMat[2][3] = -0.3;
    corrMat[3][2] = -0.3;

    _bspl = new URANIE::Sampler::TBasicSampling(_tds2, "lhs", 10000);
    _bspl->setSeed(1234);
    _bspl->setCorrelationMatrix(corrMat);
    _bspl->generateCorrSample();

    _tds2->computeStatistic();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, fabs(_tds2->getAttribute("x1")->getMean()),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, _tds2->getAttribute("x2")->getStd(),
            EPSILON);

    TMatrixD empCorrMat = _tds2->computeCorrelationMatrix();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(corrMat(0, 1), empCorrMat(0, 1),
            CORRPRECISION);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(corrMat(0, 2), empCorrMat(0, 2),
            CORRPRECISION);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(corrMat(0, 3), empCorrMat(0, 3),
            CORRPRECISION);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(corrMat(1, 2), empCorrMat(1, 2),
            CORRPRECISION);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(corrMat(1, 3), empCorrMat(1, 3),
            CORRPRECISION);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(corrMat(2, 3), empCorrMat(2, 3),
            CORRPRECISION);

    delete _bspl;
    _bspl = NULL;
}

void TBasicSamplingTest::testBasicSamplinggenerateCorrSampleSRS()
{
    TMatrixD corrMat(4, 4);
    corrMat.UnitMatrix();
    corrMat[0][1] = 0.5;
    corrMat[1][0] = 0.5;
    corrMat[0][2] = 0.2;
    corrMat[2][0] = 0.2;
    corrMat[0][3] = -0.5;
    corrMat[3][0] = -0.5;
    corrMat[1][2] = 0.1;
    corrMat[2][1] = 0.1;
    corrMat[1][3] = -0.95;
    corrMat[3][1] = -0.95;
    corrMat[2][3] = -0.3;
    corrMat[3][2] = -0.3;

    _bspl = new URANIE::Sampler::TBasicSampling(_tds2, "srs", 10000);
    _bspl->setSeed(1234);
    _bspl->setCorrelationMatrix(corrMat);
    _bspl->generateCorrSample();

    _tds2->computeStatistic();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, fabs(_tds2->getAttribute("x1")->getMean()),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, _tds2->getAttribute("x2")->getStd(),
            EPSILON);

    TMatrixD empCorrMat = _tds2->computeCorrelationMatrix();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(corrMat(0, 1), empCorrMat(0, 1),
            CORRPRECISION);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(corrMat(0, 2), empCorrMat(0, 2),
            CORRPRECISION);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(corrMat(0, 3), empCorrMat(0, 3),
            CORRPRECISION);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(corrMat(1, 2), empCorrMat(1, 2),
            CORRPRECISION);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(corrMat(1, 3), empCorrMat(1, 3),
            CORRPRECISION);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(corrMat(2, 3), empCorrMat(2, 3),
            CORRPRECISION);

    delete _bspl;
    _bspl = NULL;
}

void TBasicSamplingTest::testTBasicSamplingConstructorSmallSize()
{

    URANIE::DataServer::TDataServer *tdsB = new URANIE::DataServer::TDataServer();

    Int_t nX = 10;
    Int_t nS = 2;
     
    for (Int_t i = 0; i < 10; i++)
    {
        tdsB->addAttribute(new URANIE::DataServer::TNormalDistribution(Form("x%d",i), 0., 1));
	tdsB->getAttribute(Form("x%d",i))->delShare();
    } 
    URANIE::Sampler::TBasicSampling *tbspl = new URANIE::Sampler::TBasicSampling(tdsB, "lhs", nS);

    CPPUNIT_ASSERT_EQUAL(nX, tbspl->_nX);
    CPPUNIT_ASSERT_EQUAL(nS, tbspl->_nS);

    tbspl->setSeed(1234);
    tbspl->generateSample();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.881788, tdsB->getValue("x0",0),REAL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.67817, tdsB->getValue("x0",1),REAL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.863454, tdsB->getValue("x1",0),REAL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.776036, tdsB->getValue("x1",1),REAL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.123453, tdsB->getValue("x9",0),REAL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.896004, tdsB->getValue("x9",1),REAL_PRECISION_COMPARE);

    delete tbspl;
    tbspl = NULL;
    delete tdsB;
    tdsB=NULL;
}

void TBasicSamplingTest::testTBasicSamplingSetUserCorrelation()
{

    Int_t nS = 10000;
    Double_t dcorr1 = 0.8;
    Double_t dcorr2 = -0.5;

    // Test SRS
    _bspl = new URANIE::Sampler::TBasicSampling(_tds, "srs", 5000);
    _bspl->setUserCorrelation("Uniform","Normal",dcorr1);
    _bspl->generateCorrSample();
 
    _tds->computeStatistic("*");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.5, _tds->getAttribute("Uniform")->getMean(),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.288, _tds->getAttribute("Uniform")->getStd(),
            EPSILON);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, _tds->getAttribute("Normal")->getMean(),
            0.05);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, _tds->getAttribute("Normal")->getStd(),
            EPSILON);


    TMatrixD empCorrMatsrs = _tds->computeCorrelationMatrix();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(dcorr1, fabs(empCorrMatsrs(0, 1)),
				 CORRPRECISION);
    
    delete _bspl;
    _bspl = NULL;

     // Test LHS

    _bspl = new URANIE::Sampler::TBasicSampling(_tds2, "lhs", nS);

    _bspl->setUserCorrelation("x1","x2",dcorr1);
    _bspl->setUserCorrelation("x3","x4",dcorr2);
    _bspl->generateCorrSample();

    _tds2->computeStatistic();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, fabs(_tds2->getAttribute("x1")->getMean()),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, _tds2->getAttribute("x2")->getStd(),
            EPSILON);

    TMatrixD empCorrMat = _tds2->computeCorrelationMatrix();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(dcorr1, fabs(empCorrMat(0, 1)),
            CORRPRECISION);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, fabs(empCorrMat(0, 2)),
				 CORRPRECISION);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, fabs(empCorrMat(0, 3)),
				 CORRPRECISION);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, fabs(empCorrMat(1, 2)),
				 CORRPRECISION);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, fabs(empCorrMat(1, 3)),
				 CORRPRECISION);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dcorr2, empCorrMat(2, 3),
     				 0.25);

    delete _bspl;
    _bspl = NULL;
}
void TBasicSamplingTest::tesTBasicSamplingtSetTolerance()
{
    Double_t tol = 1.0;

    _bspl = new URANIE::Sampler::TBasicSampling(_tds2, "lhs", 10);

    _bspl->setTolerance(tol);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(tol,_bspl->_simTolerance , DBL_PRECISION_COMPARE);

    delete _bspl;
    _bspl = NULL;
}

void TBasicSamplingTest::testBasicSamplingGenerateSampleSimilar()
{
    Double_t tol = 0.2;
    Double_t dcorr = 0.9;

    _bspl = new URANIE::Sampler::TBasicSampling(_tds, "lhs", 5000);

    _bspl->setTolerance(tol);
    _bspl->setUserCorrelation("Uniform","Normal",dcorr);

    _bspl->generateCorrSample();
    _tds->computeStatistic("*");

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.5, _tds->getAttribute("Uniform")->getMean(),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.288, _tds->getAttribute("Uniform")->getStd(),
            EPSILON);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, _tds->getAttribute("Normal")->getMean(),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, _tds->getAttribute("Normal")->getStd(),
            EPSILON);

    TMatrixD empCorrMat = _tds->computeCorrelationMatrix();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, fabs(empCorrMat(0, 1)),
            CORRPRECISION);
 
    delete _bspl;
    _bspl = NULL;
}
void TBasicSamplingTest::throwExceptionBasicSamplingConstructorTdsNull()
{

    URANIE::DataServer::TDataServer * tdsnull = NULL;
    CPPUNIT_ASSERT_THROW(
            new URANIE::Sampler::TBasicSampling(tdsnull, "lhs", 5000),
            URANIE::Exceptions::UErrorExceptions);
}

void TBasicSamplingTest::throwExceptionsetCorrelationMatrix()
{

    _bspl = new URANIE::Sampler::TBasicSampling(_tds2, "lhs", 10);

    TMatrixD corrMat5(5, 5);
    corrMat5.UnitMatrix();
    CPPUNIT_ASSERT_THROW(_bspl->setCorrelationMatrix(corrMat5), URANIE::Exceptions::UErrorExceptions);

    TMatrixD corrMat3(3, 3);
    corrMat3.UnitMatrix();
    CPPUNIT_ASSERT_THROW(_bspl->setCorrelationMatrix(corrMat3), URANIE::Exceptions::UErrorExceptions);

    TMatrixD corrMat4(4, 4);
    corrMat4.UnitMatrix(); corrMat4(1,2)=0.3;
    CPPUNIT_ASSERT_THROW(_bspl->setCorrelationMatrix(corrMat4), URANIE::Exceptions::UErrorExceptions);

    corrMat4(1,2)=1.3; corrMat4(2,1)=1.3;
    CPPUNIT_ASSERT_THROW(_bspl->setCorrelationMatrix(corrMat4), URANIE::Exceptions::UErrorExceptions);

    corrMat4(1,2)=-1.3; corrMat4(2,1)=-1.3;
    CPPUNIT_ASSERT_THROW(_bspl->setCorrelationMatrix(corrMat4), URANIE::Exceptions::UErrorExceptions);

    corrMat4(1,2)=-0.3; corrMat4(2,1)=-0.30001;
    CPPUNIT_ASSERT_THROW(_bspl->setCorrelationMatrix(corrMat4), URANIE::Exceptions::UErrorExceptions);

    delete _bspl;
    _bspl = NULL;

}

void TBasicSamplingTest::throwExceptionsetUserCorrelationNameFalse()
{

    Double_t dcorr1 = 0.8;

    _bspl = new URANIE::Sampler::TBasicSampling(_tds2, "lhs", 10);

    CPPUNIT_ASSERT_THROW(_bspl->setUserCorrelation("x1","xfalse",dcorr1),
            URANIE::Exceptions::UErrorExceptions);

    CPPUNIT_ASSERT_THROW(_bspl->setUserCorrelation("xfalse","x2",dcorr1),
            URANIE::Exceptions::UErrorExceptions);

    delete _bspl;
    _bspl = NULL;

}

void TBasicSamplingTest::throwExceptionsetUserCorrelationValueFalse()
{

    _bspl = new URANIE::Sampler::TBasicSampling(_tds2, "lhs", 10);

    CPPUNIT_ASSERT_THROW(_bspl->setUserCorrelation("x1","x2",-5.0),
            URANIE::Exceptions::UErrorExceptions);

    CPPUNIT_ASSERT_THROW(_bspl->setUserCorrelation("x1","x2",4.0),
            URANIE::Exceptions::UErrorExceptions);

    delete _bspl;
    _bspl = NULL;

}
