//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTGaussianSampling.cxx
 \author Cedric Forgues <uranietm@cea.fr>
 \date   $Date$
 
 \brief  Test File of URANIE::Sampler::TGaussianSampling class
 
 Cr�e le : mer nov 21 07:59:24 2007
 
 */

// Uranie
#include "TestTGaussianSampling.h"
#include "TGaussianSampling.h"
#include "TUniformDistribution.h"
#include "TNormalDistribution.h"
#include "UDisableTest.h"
// ROOT
#include "TString.h"
#include "TSystem.h"

#include <iostream>
#include <math.h>

#define DBL_PRECISION_COMPARE 1.e-10
#define EPSILON 0.01

TGaussianSamplingTest::TGaussianSamplingTest() :
        _tdsAllNormal(NULL), _tdsANormalAndUniform(NULL)
{
}

TGaussianSamplingTest::~TGaussianSamplingTest()
{
}

void TGaussianSamplingTest::setUp()
{
    // Create a TDataServer with two normal laws
    _tdsAllNormal = new URANIE::DataServer::TDataServer();
    // Fill the DataServer with the three attributes of the study
    _tdsAllNormal->addAttribute(
            new URANIE::DataServer::TNormalDistribution("y1"));
    _tdsAllNormal->addAttribute(
            new URANIE::DataServer::TNormalDistribution("y2", -1.0, 2.));
    _tdsAllNormal->getAttribute("y1")->delShare();
    _tdsAllNormal->getAttribute("y2")->delShare();
    
    // Create a TDataServer with a normal and uniform laws
    _tdsANormalAndUniform = new URANIE::DataServer::TDataServer();
    // Fill the DataServer with the three attributes of the study
    _tdsANormalAndUniform->addAttribute(
            new URANIE::DataServer::TUniformDistribution("Uniform", 0., 1.));
    _tdsANormalAndUniform->addAttribute(
            new URANIE::DataServer::TNormalDistribution("Normal", 0., 1.));
    _tdsANormalAndUniform->getAttribute("Uniform")->delShare();
    _tdsANormalAndUniform->getAttribute("Normal")->delShare();
    
}

void TGaussianSamplingTest::tearDown()
{
    delete _tdsAllNormal;
    _tdsAllNormal = NULL;
    delete _tdsANormalAndUniform;
    _tdsANormalAndUniform = NULL;
}

////////////////////////////////
///                          ///                            
///         Tests            ///
///                          ///
////////////////////////////////

void TGaussianSamplingTest::testGaussianSampling()
{
    URANIE::Sampler::TGaussianSampling *samp =
            new URANIE::Sampler::TGaussianSampling(_tdsAllNormal, "lhs", 10000);
//  samp->setLog();
    samp->generateSample();

    _tdsAllNormal->computeStatistic();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsAllNormal->getAttribute("y1")->getMean(),
            0.0, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsAllNormal->getAttribute("y1")->getStd(),
            1.0, 0.1);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsAllNormal->getAttribute("y2")->getMean(),
            -1.0, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsAllNormal->getAttribute("y2")->getStd(),
            2.0, 0.1);

    TMatrixD matcorr = _tdsAllNormal->computeCorrelationMatrix("*");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(0, 0), 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(1, 1), 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(0, 1), 0.0, 0.1);

    delete samp;
    samp = NULL;
}

void TGaussianSamplingTest::testGaussianSamplingSRS()
{
    URANIE::Sampler::TGaussianSampling *samp =
            new URANIE::Sampler::TGaussianSampling(_tdsAllNormal, "srs", 10000);
//  samp->setLog();
    samp->generateSample();

    _tdsAllNormal->computeStatistic();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsAllNormal->getAttribute("y1")->getMean(),
            0.0, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsAllNormal->getAttribute("y1")->getStd(),
            1.0, 0.1);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsAllNormal->getAttribute("y2")->getMean(),
            -1.0, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsAllNormal->getAttribute("y2")->getStd(),
            2.0, 0.1);

    TMatrixD matcorr = _tdsAllNormal->computeCorrelationMatrix("*");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(0, 0), 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(1, 1), 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(0, 1), 0.0, 0.1);

    delete samp;
    samp = NULL;
}

void TGaussianSamplingTest::testGaussianSamplingWithCorrelation()
{
    Double_t dcorr = 0.00;

    //////////////////////////////////
    // 0.99
    //////////////////////////////////
    dcorr = 0.99;
    URANIE::Sampler::TGaussianSampling *samp =
            new URANIE::Sampler::TGaussianSampling(_tdsAllNormal, "lhs", 10000);
//  samp->setLog();
    samp->setUserCorrelation(0, 1, dcorr);
    samp->generateSample();

    TMatrixD matcorr = _tdsAllNormal->computeCorrelationMatrix("*");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(0, 0), 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(1, 1), 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(0, 1), dcorr, 0.1);

    delete samp;
    samp = NULL;
    tearDown();
    gSystem->Sleep(500);
    setUp();

    //////////////////////////////////
    // 0.45
    //////////////////////////////////
    dcorr = 0.45;
    samp = new URANIE::Sampler::TGaussianSampling(_tdsAllNormal, "lhs", 10000);
//  samp->setLog();
    samp->setUserCorrelation(0, 1, dcorr);
    samp->generateSample();

    matcorr = _tdsAllNormal->computeCorrelationMatrix("*");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(0, 0), 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(1, 1), 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(0, 1), dcorr, 0.1);

    delete samp;
    samp = NULL;
    tearDown();
    gSystem->Sleep(2000);
    setUp();
    //////////////////////////////////
    // 0.1
    //////////////////////////////////
    dcorr = 0.1;
    samp = new URANIE::Sampler::TGaussianSampling(_tdsAllNormal, "lhs", 10000);
//  samp->setLog();
    samp->setUserCorrelation(0, 1, dcorr);
    samp->generateSample();

    matcorr = _tdsAllNormal->computeCorrelationMatrix("*");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(0, 0), 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(1, 1), 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(0, 1), dcorr, 0.1);

    delete samp;
    samp = NULL;
    tearDown();
    gSystem->Sleep(2000);
    setUp();
    //////////////////////////////////
    // -0.10
    //////////////////////////////////
    dcorr = -0.10;
    samp = new URANIE::Sampler::TGaussianSampling(_tdsAllNormal, "lhs", 10000);
//  samp->setLog();
    samp->setUserCorrelation("y1", "y2", dcorr);
    samp->generateSample();

    matcorr = _tdsAllNormal->computeCorrelationMatrix("*");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(0, 0), 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(1, 1), 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(0, 1), dcorr, 0.1);

    delete samp;
    samp = NULL;
    tearDown();
    gSystem->Sleep(2000);
    setUp();
    //////////////////////////////////
    // -0.45
    //////////////////////////////////
    dcorr = -0.45;
    samp = new URANIE::Sampler::TGaussianSampling(_tdsAllNormal, "lhs", 10000);
//  samp->setLog();
    samp->setUserCorrelation(_tdsAllNormal->getAttribute("y1"),
            _tdsAllNormal->getAttribute("y2"), dcorr);
    samp->generateSample();

    matcorr = _tdsAllNormal->computeCorrelationMatrix("*");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(0, 0), 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(1, 1), 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(0, 1), dcorr, 0.1);

    delete samp;
    samp = NULL;
    tearDown();
    gSystem->Sleep(2000);
    setUp();
    //////////////////////////////////
    // -0.99
    //////////////////////////////////
    dcorr = -0.99;
    samp = new URANIE::Sampler::TGaussianSampling(_tdsAllNormal, "lhs", 10000);
//  samp->setLog();
    samp->setUserCorrelation(0, 1, dcorr);
    samp->generateSample();

    matcorr = _tdsAllNormal->computeCorrelationMatrix("*");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(0, 0), 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(1, 1), 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(0, 1), dcorr, 0.1);

    delete samp;
    samp = NULL;
}

void TGaussianSamplingTest::testGaussianSamplingWithSeed()
{
    URANIE::Sampler::TGaussianSampling *samp1 =
            new URANIE::Sampler::TGaussianSampling(_tdsAllNormal, "lhs", 1000);
//   samp1->setLog();
    samp1->setSeed(123456789);
    samp1->generateSample();

    _tdsAllNormal->computeStatistic();
    Double_t dmean1 = _tdsAllNormal->getAttribute("y1")->getMean();
    Double_t dmean2 = _tdsAllNormal->getAttribute("y2")->getMean();
    Double_t dsigma1 = _tdsAllNormal->getAttribute("y1")->getStd();
    Double_t dsigma2 = _tdsAllNormal->getAttribute("y2")->getStd();

    TMatrixD matcorr1 = _tdsAllNormal->computeCorrelationMatrix("*");
    Double_t dcovar = matcorr1(0, 1);

    delete samp1;
    samp1 = NULL;
    tearDown();
    gSystem->Sleep(2000);
    setUp();

    URANIE::Sampler::TGaussianSampling *samp2 =
            new URANIE::Sampler::TGaussianSampling(_tdsAllNormal, "lhs", 1000);
//   samp2->setLog();
    samp2->setSeed(123456789);
    samp2->generateSample();

    _tdsAllNormal->computeStatistic();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsAllNormal->getAttribute("y1")->getMean(),
            dmean1, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsAllNormal->getAttribute("y1")->getStd(),
            dsigma1, DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsAllNormal->getAttribute("y2")->getMean(),
            dmean2, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsAllNormal->getAttribute("y2")->getStd(),
            dsigma2, DBL_PRECISION_COMPARE);

    TMatrixD matcorr2 = _tdsAllNormal->computeCorrelationMatrix("*");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr2(0, 1), dcovar, DBL_PRECISION_COMPARE);

    delete samp2;
    samp2 = NULL;
}

void TGaussianSamplingTest::testGaussianSamplingWithFixedSumCovariance()
{
    Double_t sumCov = 0.50;
    Int_t NS = 1000;

    URANIE_INBUILDING_TEST()
    ;

    //////////////////////////////////
    // simple
    //////////////////////////////////

    URANIE::Sampler::TGaussianSampling *samp;
    samp = new URANIE::Sampler::TGaussianSampling(_tdsAllNormal, "lhs", NS);
    samp->setSeed(123456789);
    //samp->setFixedSumCovariance("simple");
    samp->generateSample();

    delete samp;
    samp = NULL;
    tearDown();
    setUp();

    samp = new URANIE::Sampler::TGaussianSampling(_tdsAllNormal, "lhs", NS);
    samp->setSeed(123456789);
    samp->setFixedSumCovariance("exact");
    samp->generateSample();

    delete samp;
    samp = NULL;
    // //////////////////////////////////
    // // exact
    // //////////////////////////////////
    // samp = new URANIE::Sampler::TGaussianSampling(_tdsAllNormal, "lhs", 10000);
    // samp->setFixedSumCovariance("exact");
    // samp->generateSample();

    // matcorr = _tdsAllNormal->computeCorrelationMatrix("*");
    // CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(0, 0), 1.0, DBL_PRECISION_COMPARE);
    // CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(1, 1), 1.0, DBL_PRECISION_COMPARE);
    // CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(0, 1), dcorr, 0.1);

    // delete samp;
    // samp = NULL;
    // tearDown();
    // gSystem->Sleep(2000);
    // setUp();
}

////////////////////////////////
///                          ///                            
///       Exceptions         ///
///                          ///
////////////////////////////////

void TGaussianSamplingTest::throwExceptionWithOtherLaw()
{
    CPPUNIT_ASSERT_THROW(
            new URANIE::Sampler::TGaussianSampling(_tdsANormalAndUniform, "lhs",
                    10000), URANIE::Exceptions::UErrorExceptions);
}
void TGaussianSamplingTest::throwExceptionTGaussianSamplingConstrutorTdsNull()
{
    URANIE::DataServer::TDataServer * tdsnull = NULL;
    CPPUNIT_ASSERT_THROW(
            new URANIE::Sampler::TGaussianSampling(tdsnull, "lhs", 1000),
            URANIE::Exceptions::UErrorExceptions);

}
void TGaussianSamplingTest::throwExceptionTGaussianSamplinggenerateSampleMatCorrNegative()
{
  URANIE::DataServer::TDataServer tds("tdsTest","");
    // Fill the DataServer with the three attributes of the study
    tds.addAttribute(new URANIE::DataServer::TNormalDistribution("y1"));
    tds.addAttribute(new URANIE::DataServer::TNormalDistribution("y2"));
    tds.addAttribute(new URANIE::DataServer::TNormalDistribution("y3"));
    tds.getAttribute("y1")->delShare();
    tds.getAttribute("y2")->delShare();
    tds.getAttribute("y3")->delShare();
    URANIE::Sampler::TGaussianSampling *samp =
            new URANIE::Sampler::TGaussianSampling(&tds, "lhs", 10);
    samp->setUserCorrelation(0, 1, 0.9);
    samp->setUserCorrelation(0, 2, 0.9);
    samp->setUserCorrelation(1, 2, 0.1);

    CPPUNIT_ASSERT_THROW(samp->generateSample(), URANIE::Exceptions::UErrorExceptions);

    delete samp;
    samp = NULL;
}

void TGaussianSamplingTest::throwExceptionTGaussianSamplingsetUserCorrelationSameIndex()
{
    URANIE::Sampler::TGaussianSampling *samp =
            new URANIE::Sampler::TGaussianSampling(_tdsAllNormal, "lhs", 10000);
//  samp->setLog();
    CPPUNIT_ASSERT_THROW(samp->setUserCorrelation(1, 1, 0.95),
            URANIE::Exceptions::UErrorExceptions);
        delete samp;
    samp = NULL;
}
void TGaussianSamplingTest::throwExceptionTGaussianSamplingsetUserCorrelationIndexINegative()
{
    URANIE::Sampler::TGaussianSampling *samp =
            new URANIE::Sampler::TGaussianSampling(_tdsAllNormal, "lhs", 10000);
//  samp->setLog();
    CPPUNIT_ASSERT_THROW(samp->setUserCorrelation(-1, 1, 0.95),
            URANIE::Exceptions::UErrorExceptions);
        delete samp;
    samp = NULL;
}
void TGaussianSamplingTest::throwExceptionTGaussianSamplingsetUserCorrelationIndexJNegative()
{
    URANIE::Sampler::TGaussianSampling *samp =
            new URANIE::Sampler::TGaussianSampling(_tdsAllNormal, "lhs", 10000);
//  samp->setLog();
    CPPUNIT_ASSERT_THROW(samp->setUserCorrelation(1, -1, 0.95),
            URANIE::Exceptions::UErrorExceptions);
        delete samp;
    samp = NULL;
}
void TGaussianSamplingTest::throwExceptionTGaussianSamplingsetUserCorrelationIndexValueFalse()
{
    URANIE::Sampler::TGaussianSampling *samp =
            new URANIE::Sampler::TGaussianSampling(_tdsAllNormal, "lhs", 10000);
//  samp->setLog();
    CPPUNIT_ASSERT_THROW(samp->setUserCorrelation(1, 1, 1.95),
            URANIE::Exceptions::UErrorExceptions);
        delete samp;
    samp = NULL;
}
void TGaussianSamplingTest::throwExceptionTGaussianSamplingsetUserCorrelationNameValueFalse()
{
    URANIE::Sampler::TGaussianSampling *samp =
            new URANIE::Sampler::TGaussianSampling(_tdsAllNormal, "lhs", 10000);
//  samp->setLog();
    CPPUNIT_ASSERT_THROW(samp->setUserCorrelation("y1", "y2", 1.95),
            URANIE::Exceptions::UErrorExceptions);
        delete samp;
    samp = NULL;
}
void TGaussianSamplingTest::throwExceptionTGaussianSamplingsetUserCorrelationAttributeValueFalse()
{
    URANIE::Sampler::TGaussianSampling *samp =
            new URANIE::Sampler::TGaussianSampling(_tdsAllNormal, "lhs", 10000);
//  samp->setLog();
    CPPUNIT_ASSERT_THROW(
            samp->setUserCorrelation(_tdsAllNormal->getAttribute("y1"),
                    _tdsAllNormal->getAttribute("y2"), 1.95),
            URANIE::Exceptions::UErrorExceptions);
        delete samp;
    samp = NULL;
}
void TGaussianSamplingTest::throwExceptionTGaussianSamplingsetUserCorrelationIndexISup()
{
    URANIE::Sampler::TGaussianSampling *samp =
            new URANIE::Sampler::TGaussianSampling(_tdsAllNormal, "lhs", 10000);
//  samp->setLog();
    CPPUNIT_ASSERT_THROW(samp->setUserCorrelation(3, 1, 0.95),
            URANIE::Exceptions::UErrorExceptions);
        delete samp;
    samp = NULL;
}
void TGaussianSamplingTest::throwExceptionTGaussianSamplingsetUserCorrelationIndexJSup()
{
    URANIE::Sampler::TGaussianSampling *samp =
            new URANIE::Sampler::TGaussianSampling(_tdsAllNormal, "lhs", 10000);
//  samp->setLog();
    CPPUNIT_ASSERT_THROW(samp->setUserCorrelation(0, 3, 0.95),
            URANIE::Exceptions::UErrorExceptions);
        delete samp;
    samp = NULL;
}
void TGaussianSamplingTest::throwExceptionTGaussianSamplingsetUserCorrelationFalseNameI()
{
    URANIE::Sampler::TGaussianSampling *samp =
            new URANIE::Sampler::TGaussianSampling(_tdsAllNormal, "lhs", 10000);
//  samp->setLog();
    CPPUNIT_ASSERT_THROW(samp->setUserCorrelation("toto", "y1", 0.95),
            URANIE::Exceptions::UErrorExceptions);
    delete samp;
    samp = NULL;
}
void TGaussianSamplingTest::throwExceptionTGaussianSamplingsetUserCorrelationFalseNameJ()
{
    URANIE::Sampler::TGaussianSampling *samp =
            new URANIE::Sampler::TGaussianSampling(_tdsAllNormal, "lhs", 10000);
//  samp->setLog();
    CPPUNIT_ASSERT_THROW(samp->setUserCorrelation("y1", "toto", 0.95),
            URANIE::Exceptions::UErrorExceptions);
        delete samp;
    samp = NULL;
}
void TGaussianSamplingTest::throwExceptionTGaussianSamplingsetUserCorrelationFalseAttributI()
{
    URANIE::DataServer::TAttribute * att = new URANIE::DataServer::TAttribute(
            "att");
    URANIE::Sampler::TGaussianSampling *samp =
            new URANIE::Sampler::TGaussianSampling(_tdsAllNormal, "lhs", 10000);
//  samp->setLog();
    CPPUNIT_ASSERT_THROW(
            samp->setUserCorrelation(att, _tdsAllNormal->getAttribute("y1"),
                    0.95), URANIE::Exceptions::UErrorExceptions);
        delete samp;
    samp = NULL;
        delete att;
    att = NULL;
}
void TGaussianSamplingTest::throwExceptionTGaussianSamplingsetUserCorrelationFalseAttributJ()
{
    URANIE::DataServer::TAttribute * att = new URANIE::DataServer::TAttribute(
            "att");
    URANIE::Sampler::TGaussianSampling *samp =
            new URANIE::Sampler::TGaussianSampling(_tdsAllNormal, "lhs", 10000);
//  samp->setLog();
    CPPUNIT_ASSERT_THROW(
            samp->setUserCorrelation(_tdsAllNormal->getAttribute("y1"), att,
                    0.95), URANIE::Exceptions::UErrorExceptions);
            delete samp;
    samp = NULL;
        delete att;
    att = NULL;
}
