//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTNeuralGas.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 Created : 2011-12-15 13:51:45
 
 \brief  Test File of the class URANIE::DataServer::TNeuralGas
 */

#define DBL_PRECISION_COMPARE 1.e-12

#include "TestTNeuralGas.h"
#include "TString.h"

#include <iostream>

////////////////////////////////
///                          ///
///  Constructor/Descrutor   ///
///                          ///
////////////////////////////////

TNeuralGasTest::TNeuralGasTest()
{
}

TNeuralGasTest::~TNeuralGasTest()
{
}

////////////////////////////////
///                          ///
///  setUp/tearDown          ///
///                          ///
////////////////////////////////

void TNeuralGasTest::setUp()
{
    _tds1 = new URANIE::DataServer::TDataServer("tds1", "the TDS 1");
    _tds1->fileDataRead("../../../../dataSERVER/souRCE/teST/geyser.dat");

}

void TNeuralGasTest::tearDown()
{
    delete _tds1;
    _tds1 = NULL;
}

////////////////////////////////
///                          ///
///         Tests            ///
///                          ///
////////////////////////////////

void TNeuralGasTest::testTMax()
{
    URANIE::Sampler::TNeuralGas * tvq = new URANIE::Sampler::TNeuralGas(_tds1,
            "saveall", 50);
    tvq->setDrawProgressBar(kFALSE);
    tvq->unsetLog();

    // Default value
    CPPUNIT_ASSERT_EQUAL(40000, tvq->getTMax());

    Int_t ncrt = 12546;
    tvq->setTMax(ncrt);
    CPPUNIT_ASSERT_EQUAL(ncrt, tvq->getTMax());

    delete tvq;
    tvq = NULL;
}

void TNeuralGasTest::testSaveAll()
{
    URANIE::Sampler::TNeuralGas * tvq = new URANIE::Sampler::TNeuralGas(_tds1,
            "", 50);
    tvq->setDrawProgressBar(kFALSE);
    tvq->unsetLog();

    // Default value
    CPPUNIT_ASSERT_MESSAGE("SaveAll is active", tvq->getSaveAll() == kFALSE);

    delete tvq;
    tvq = NULL;

    tvq = new URANIE::Sampler::TNeuralGas(_tds1, "saveall", 50);
    tvq->setDrawProgressBar(kFALSE);
    tvq->unsetLog();
    CPPUNIT_ASSERT_MESSAGE("SaveAll is not active", tvq->getSaveAll() == kTRUE);

    delete tvq;
    tvq = NULL;
}

void TNeuralGasTest::testGenerateSampleWithLoop3()
{
    Int_t nsize = 50;
    URANIE::Sampler::TNeuralGas * tvq = new URANIE::Sampler::TNeuralGas(_tds1,
            "", nsize);
    tvq->setDrawProgressBar(kFALSE);
    tvq->unsetLog();

    URANIE::DataServer::TDataServer *tdsng = tvq->getSubSample("loop=3");
    CPPUNIT_ASSERT_EQUAL(nsize, tdsng->getNPatterns());

    delete tvq;
    tvq = NULL;
}

void TNeuralGasTest::testGetSubSample()
{
    Int_t nsize = 50;
    URANIE::Sampler::TNeuralGas * tvq = new URANIE::Sampler::TNeuralGas(_tds1,
            "", nsize);
    tvq->setDrawProgressBar(kFALSE);
    tvq->unsetLog();

    URANIE::DataServer::TDataServer *tdsng = tvq->getSubSample();
    CPPUNIT_ASSERT_EQUAL(nsize, tdsng->getNPatterns());

    // Default value
    CPPUNIT_ASSERT_EQUAL(40000, tvq->getTMax());

    delete tvq;
    tvq = NULL;
}

void TNeuralGasTest::testGetSubSampleWithSaveAll()
{
    Int_t nsize = 50;
    URANIE::Sampler::TNeuralGas * tvq = new URANIE::Sampler::TNeuralGas(_tds1,
            "saveall", nsize);
    tvq->setDrawProgressBar(kFALSE);
    tvq->unsetLog();

    Int_t ntimes = tvq->getTMax() + 1;

    URANIE::DataServer::TDataServer *tdsng = tvq->getSubSample();
    CPPUNIT_ASSERT_EQUAL(ntimes * nsize, tdsng->getNPatterns());

    delete tvq;
    tvq = NULL;
}

////////////////////////////////
///                          ///
///       Exceptions         ///
///                          ///
////////////////////////////////

void TNeuralGasTest::throwExceptionTMaxNegative()
{
    URANIE::Sampler::TNeuralGas * tvq = new URANIE::Sampler::TNeuralGas(_tds1,
            "saveall", 50);
    tvq->setDrawProgressBar(kFALSE);
    tvq->unsetLog();

    Int_t ncrt = -12546;
    tvq->setTMax(ncrt);

    CPPUNIT_ASSERT_THROW(tvq->generateSample(),
			 URANIE::Exceptions::UErrorExceptions);
    
    delete tvq;
    tvq = NULL;
}

void TNeuralGasTest::throwExceptionLoopOptionNotDigit()
{
    URANIE::Sampler::TNeuralGas * tvq = new URANIE::Sampler::TNeuralGas(_tds1,
            "saveall", 50);
    tvq->setDrawProgressBar(kFALSE);
    tvq->unsetLog();

    CPPUNIT_ASSERT_THROW(tvq->generateSample("loop=a"),
			 URANIE::Exceptions::UWarningExceptions);
    
    delete tvq;
    tvq = NULL;
}

void TNeuralGasTest::throwExceptionNGreaterThanNPat()
{
    CPPUNIT_ASSERT_THROW(
            new URANIE::Sampler::TNeuralGas(_tds1, "saveall",
                    _tds1->getNPatterns() + 1),
            URANIE::Exceptions::UErrorExceptions);
}

// End Of File
