//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTOATDesign.cxx
 \author Nicolas Gilardi
 \date   $Date$

 \brief  Test file of the class URANIE::Sampler::TOATDesign

 Created: 2012-08-21 

 */

// CPPUNIT
#define DBL_PRECISION_COMPARE 1.e-6
#define SPL_PRECISION_COMPARE 1.e-6

// Uranie
#include "TAttribute.h"
#include "TNormalDistribution.h"
#include "TTriangularDistribution.h"
#include "TestTOATDesign.h"

//ROOT
#include "TMath.h"

#include <iomanip>

TOATDesignTest::TOATDesignTest()
{
}

TOATDesignTest::~TOATDesignTest()
{
}

void TOATDesignTest::setUp()
{
}

void TOATDesignTest::tearDown()
{
}

////////////////////////////////
///                          ///                            
///         Tests            ///
///                          ///
////////////////////////////////

void TOATDesignTest::testConstructor()
{
    cout << endl;

    URANIE::DataServer::TDataServer tdsFile("tdsFile",
            "TDS for OAT sampling loaded from a file");
    tdsFile.fileDataRead("nominalValues.dat");

    CPPUNIT_ASSERT_THROW(
            new URANIE::Sampler::TOATDesign(&tdsFile, "regular", 0),
            URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(
            new URANIE::Sampler::TOATDesign(&tdsFile, "regular", -1),
            URANIE::Exceptions::UErrorExceptions);
}

void TOATDesignTest::testSetOptions()
{
    cout << endl;

    URANIE::DataServer::TDataServer tdsFile("tdsFile",
            "TDS for OAT sampling loaded from a file");
    tdsFile.fileDataRead("nominalValues.dat");

    CPPUNIT_ASSERT_THROW(new URANIE::Sampler::TOATDesign(&tdsFile, "gromf", 2),
            URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(new URANIE::Sampler::TOATDesign(&tdsFile, "lhs", 1),
            URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(
            new URANIE::Sampler::TOATDesign(&tdsFile, "regular", 7),
            URANIE::Exceptions::UErrorExceptions);
}

void TOATDesignTest::testGetHasData()
{
    cout << endl;

    // TDS with data
    URANIE::DataServer::TDataServer tdsFile("tdsFile",
            "TDS for OAT sampling loaded from a file");
    tdsFile.fileDataRead("nominalValues.dat");

    mySampling = new URANIE::Sampler::TOATDesign(&tdsFile, "regular", 4);
    CPPUNIT_ASSERT(mySampling->getHasData());
    delete mySampling;

    // TDS without data
    URANIE::DataServer::TNormalDistribution x1("x1", 0.0, 1.0);
    URANIE::DataServer::TNormalDistribution x2("x2", -3.0, 2.0);

    URANIE::DataServer::TDataServer tdsMem("tdsMem", "");
    tdsMem.addAttribute(&x1);
    x1.setDefaultValue(0.0);
    tdsMem.addAttribute(&x2);
    x2.setDefaultValue(-3.0);

    mySampling = new URANIE::Sampler::TOATDesign(&tdsMem, "lhs", 2);
    CPPUNIT_ASSERT(mySampling->getHasData() == kFALSE);
    delete mySampling;
}

void TOATDesignTest::testGetNumberOfModifications()
{
    cout << endl;

    URANIE::DataServer::TDataServer tdsFile("tdsFile",
            "TDS for OAT sampling loaded from a file");
    tdsFile.fileDataRead("nominalValues.dat");

    // Normal situation for "regular" mode.
    mySampling = new URANIE::Sampler::TOATDesign(&tdsFile, "regular", 4);
    CPPUNIT_ASSERT(mySampling->getNumberOfModifications() == 4);
    delete mySampling;

    // Random mode accepts odd number of modifications
    mySampling = new URANIE::Sampler::TOATDesign(&tdsFile, "random", 7);
    CPPUNIT_ASSERT(mySampling->getNumberOfModifications() == 7);
    delete mySampling;
}

void TOATDesignTest::testGetSamplingMode()
{
    cout << endl;

    URANIE::DataServer::TDataServer tdsFile("tdsFile",
            "TDS for OAT sampling loaded from a file");
    tdsFile.fileDataRead("nominalValues.dat");

    // Normal situation for "regular" mode.
    mySampling = new URANIE::Sampler::TOATDesign(&tdsFile, "regular", 2);
    CPPUNIT_ASSERT(mySampling->getSamplingMode() == "regular");
    delete mySampling;

    // Normal situation for "random" mode.
    mySampling = new URANIE::Sampler::TOATDesign(&tdsFile, "random", 2);
    CPPUNIT_ASSERT(mySampling->getSamplingMode() == "random");
    delete mySampling;

    // Normal situation for "random" mode.
    mySampling = new URANIE::Sampler::TOATDesign(&tdsFile, "srs", 2);
    CPPUNIT_ASSERT(mySampling->getSamplingMode() == "random");
    delete mySampling;

    // Normal situation for "random" mode.
    mySampling = new URANIE::Sampler::TOATDesign(&tdsFile, "lhs", 2);
    CPPUNIT_ASSERT(mySampling->getSamplingMode() == "random");
    delete mySampling;

    // Unexpected string containing valid sub string: 
    // "ASLhsKCjk" contains "Lhs" -> "random"; 
    // "sRSTDEre" contains "sRS" -> "random";
    // "TDEre GAjnchdu" contains "re G" -> "regular";
    // The string "reg" is searched first and thus "regular" mode should be chosen.
    mySampling = new URANIE::Sampler::TOATDesign(&tdsFile,
            "ASLhsKCjk kshze,::sRSTDEre GAjnchdu hdzdg", 2);
    CPPUNIT_ASSERT(mySampling->getSamplingMode() == "regular");
    delete mySampling;
}

void TOATDesignTest::testGetGenerationMode()
{
    cout << endl;

    URANIE::DataServer::TDataServer tdsFile("tdsFile",
            "TDS for OAT sampling loaded from a file");
    tdsFile.fileDataRead("nominalValues.dat");

    // Normal situation for "regular" mode.
    mySampling = new URANIE::Sampler::TOATDesign(&tdsFile, "regular", 2);
    CPPUNIT_ASSERT(mySampling->getGenerationMode() == "range");
    delete mySampling;

    // Normal situation for "random" mode.
    mySampling = new URANIE::Sampler::TOATDesign(&tdsFile, "random", 2);
    CPPUNIT_ASSERT(mySampling->getGenerationMode() == "srs");
    delete mySampling;

    // Normal situation for "srs" mode.
    mySampling = new URANIE::Sampler::TOATDesign(&tdsFile, "srs", 2);
    CPPUNIT_ASSERT(mySampling->getGenerationMode() == "srs");
    delete mySampling;

    // Normal situation for "lhs" mode.
    mySampling = new URANIE::Sampler::TOATDesign(&tdsFile, "lhs", 2);
    CPPUNIT_ASSERT(mySampling->getGenerationMode() == "lhs");
    delete mySampling;
}

void TOATDesignTest::testSetRangeDouble()
{
    cout << endl;

    //-----------------------------------------//
    // TDS and attributes constructed manually //
    //-----------------------------------------//

    URANIE::DataServer::TNormalDistribution x1("x1", 0.0, 1.0);
    URANIE::DataServer::TNormalDistribution x2("x2", -3.0, 2.0);
    URANIE::DataServer::TAttribute x3("x3");

    URANIE::DataServer::TDataServer tdsMem("tdsMem", "");
    tdsMem.addAttribute(&x1);
    x1.setDefaultValue(0.0);
    tdsMem.addAttribute(&x2);
    tdsMem.addAttribute(&x3);
    x3.setDefaultValue(1.0);

    // "random" mode
    mySampling = new URANIE::Sampler::TOATDesign(&tdsMem, "random", 2);
    // Exception: unknown attribute 
    CPPUNIT_ASSERT_THROW(mySampling->setRange("gromf", 0.5),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: x2 has no default value
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x2", 0.5),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: range is negative or null    
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", 0.0),
            URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", -0.5),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: x3 is not a TStochasticAttribute object
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x3", 0.5),
            URANIE::Exceptions::UErrorExceptions);
    // This one should pass without problem:
    mySampling->setRange("x1", 0.5);
    // Exception: x1 is already in the list
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", 1.5),
            URANIE::Exceptions::UErrorExceptions);
    delete mySampling;

    // "regular" mode
    mySampling = new URANIE::Sampler::TOATDesign(&tdsMem, "regular", 2);
    // Exception: unknown attribute 
    CPPUNIT_ASSERT_THROW(mySampling->setRange("gromf", 0.5),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: x2 has no default value
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x2", 0.5),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: range is negative or null    
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", 0.0),
            URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", -0.5),
            URANIE::Exceptions::UErrorExceptions);
    // These ones should pass without problem:
    mySampling->setRange("x1", 0.5);
    mySampling->setRange("x3", 10.0, kTRUE);
    // Exception: x1 is already in the list
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", 1.5),
            URANIE::Exceptions::UErrorExceptions);
    delete mySampling;

    //-----------------------------//
    // TDS constructed from a file //
    //-----------------------------//

    URANIE::DataServer::TDataServer tdsFile("tdsFile",
            "TDS for OAT sampling loaded from a file");
    tdsFile.fileDataRead("nominalValues.dat");

    // "random" mode
    mySampling = new URANIE::Sampler::TOATDesign(&tdsFile, "random", 2);
    // Exception: unknown attribute 
    CPPUNIT_ASSERT_THROW(mySampling->setRange("gromf", 0.5),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: x1 is not a TStochasticAttribute object
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", 0.5),
            URANIE::Exceptions::UErrorExceptions);
    delete mySampling;

    // "regular" mode
    mySampling = new URANIE::Sampler::TOATDesign(&tdsFile, "regular", 2);
    // Exception: unknown attribute 
    CPPUNIT_ASSERT_THROW(mySampling->setRange("gromf", 0.5),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: range is negative or null    
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", 0.0),
            URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", -0.5),
            URANIE::Exceptions::UErrorExceptions);
    // These ones should pass without problem:
    mySampling->setRange("x1", 0.5);
    mySampling->setRange("x2", 10.0, kTRUE);
    // Exception: x1 is already in the list
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", 1.5),
            URANIE::Exceptions::UErrorExceptions);
    delete mySampling;
}

void TOATDesignTest::testSetRangeString()
{
    cout << endl;

    //-----------------------------------------//
    // TDS and attributes constructed manually //
    //-----------------------------------------//

    URANIE::DataServer::TNormalDistribution x1("x1", 0.0, 1.0);
    URANIE::DataServer::TNormalDistribution x2("x2", -3.0, 2.0);
    URANIE::DataServer::TAttribute x3("x3");
    URANIE::DataServer::TAttribute r1("r1");
    URANIE::DataServer::TAttribute r2("r2");
    URANIE::DataServer::TAttribute r3("r3");

    URANIE::DataServer::TDataServer tdsMem("tdsMem", "");
    tdsMem.addAttribute(&x1);
    x1.setDefaultValue(0.0);
    tdsMem.addAttribute(&x2);
    tdsMem.addAttribute(&x3);
    x3.setDefaultValue(1.0);
    tdsMem.addAttribute(&r1);
    r1.setDefaultValue(0.1);
    tdsMem.addAttribute(&r2);
    r2.setDefaultValue(0.0);
    tdsMem.addAttribute(&r3);

    // "random" mode
    mySampling = new URANIE::Sampler::TOATDesign(&tdsMem, "random", 2);
    // Exception: unknown changing attribute 
    CPPUNIT_ASSERT_THROW(mySampling->setRange("gromf", "r1"),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: unknown range attribute 
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", "gromf"),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: x2 has no default value
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x2", "r1"),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: r2 has an invalid default value    
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", "r2"),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: r3 has no default value    
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", "r3"),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: x3 is not a TStochasticAttribute object
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x3", "r1"),
            URANIE::Exceptions::UErrorExceptions);
    // This one should pass without problem:
    mySampling->setRange("x1", "r1");
    // Exception: x1 is already in the list
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", "r2"),
            URANIE::Exceptions::UErrorExceptions);
    delete mySampling;

    // "regular" mode
    mySampling = new URANIE::Sampler::TOATDesign(&tdsMem, "regular", 2);
    // Exception: unknown changing attribute 
    CPPUNIT_ASSERT_THROW(mySampling->setRange("gromf", "r1"),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: unknown range attribute 
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", "gromf"),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: x2 has no default value
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x2", "r1"),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: r2 has an invalid default value    
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", "r2"),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: r3 has no default value    
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", "r3"),
            URANIE::Exceptions::UErrorExceptions);
    // These ones should pass without problem:
    mySampling->setRange("x1", "r1");
    mySampling->setRange("x3", "r1", kTRUE);
    // Exception: x1 is already in the list
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", "r1"),
            URANIE::Exceptions::UErrorExceptions);
    delete mySampling;

    //-----------------------------//
    // TDS constructed from a file //
    //-----------------------------//

    URANIE::DataServer::TDataServer tdsFile("tdsFile",
            "TDS for OAT sampling loaded from a file");
    tdsFile.fileDataRead("nominalValues.dat");

    // "random" mode
    mySampling = new URANIE::Sampler::TOATDesign(&tdsFile, "random", 2);
    // Exception: unknown changing attribute 
    CPPUNIT_ASSERT_THROW(mySampling->setRange("gromf", "rx1"),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: unknown range attribute 
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", "gromf"),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: x1 is not a TStochasticAttribute object
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", "rx1"),
            URANIE::Exceptions::UErrorExceptions);
    delete mySampling;

    // "regular" mode
    mySampling = new URANIE::Sampler::TOATDesign(&tdsFile, "regular", 2);
    // Exception: unknown changing attribute 
    CPPUNIT_ASSERT_THROW(mySampling->setRange("gromf", "rx1"),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: unknown range attribute 
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", "gromf"),
            URANIE::Exceptions::UErrorExceptions);
    // These ones should pass without problem:
    mySampling->setRange("x1", "rx1");
    mySampling->setRange("x2", "x1", kTRUE);
    // Exception: x1 is already in the list
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", "rx1"),
            URANIE::Exceptions::UErrorExceptions);
    delete mySampling;
}

void TOATDesignTest::testGetChangingAttList()
{
    cout << endl;

    URANIE::DataServer::TDataServer tdsFile("tdsFile",
            "TDS for OAT sampling loaded from a file");
    tdsFile.fileDataRead("nominalValues.dat");

    // Create the sampler 
    mySampling = new URANIE::Sampler::TOATDesign(&tdsFile, "regular", 2);

    // Set the range information
    mySampling->setRange("x1", "rx1");
    mySampling->setRange("x2", 10.0, kTRUE);

    TList* changingAttList = mySampling->getChangingAttList();

    CPPUNIT_ASSERT(changingAttList->At(0) == tdsFile.getAttribute("x1"));
    CPPUNIT_ASSERT(changingAttList->At(1) == tdsFile.getAttribute("x2"));
    delete mySampling;
}

void TOATDesignTest::testGetRangeInformation()
{
    cout << endl;

    URANIE::DataServer::TDataServer tdsFile("tdsFile",
            "TDS for OAT sampling loaded from a file");
    tdsFile.fileDataRead("nominalValues.dat");

    // Create the sampler 
    mySampling = new URANIE::Sampler::TOATDesign(&tdsFile, "regular", 2);

    // Set the range information
    mySampling->setRange("x1", "rx1");
    mySampling->setRange("x2", 10.0, kTRUE);

    TMap* rangeInfo = mySampling->getRangeInformation();

    CPPUNIT_ASSERT(rangeInfo->GetValue("x1")->GetName() == TString("rx1"));
    CPPUNIT_ASSERT(rangeInfo->GetValue("x2")->GetName() == TString("1.0000000000000000e+01%"));
    delete mySampling;

}

void TOATDesignTest::testGetNominalValueTuple()
{
    URANIE::DataServer::TDataServer tdsFile("tdsFile",
            "TDS for OAT sampling loaded from a file");
    tdsFile.fileDataRead("nominalValues.dat");

    // Create the sampler 
    mySampling = new URANIE::Sampler::TOATDesign(&tdsFile, "regular", 2);

    // Set the range information
    mySampling->setRange("x1", "rx1");
    mySampling->setRange("x2", 10.0, kTRUE);

    // generate the sample (this creates the nominal value tuple)
    mySampling->generateSample();

    URANIE::DataServer::TDSNtupleD* nomVal = mySampling->getNominalValueTuple();

    Double_t * data = nomVal->getBranchData("x1");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, data[0], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(5.0, data[1], DBL_PRECISION_COMPARE);

    data = nomVal->getBranchData("rx1");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.4, data[0], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.2, data[1], DBL_PRECISION_COMPARE);

    data = nomVal->getBranchData("x2");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, data[0], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, data[1], DBL_PRECISION_COMPARE);
    delete mySampling;

}

void TOATDesignTest::testGenerateSample()
{
    cout << endl;

    //-----------------------------//
    // TDS constructed from a file //
    //-----------------------------//

    URANIE::DataServer::TDataServer tdsFile("tdsFile",
            "TDS for OAT sampling loaded from a file");
    tdsFile.fileDataRead("nominalValues.dat");

    // Create the sampler
    mySampling = new URANIE::Sampler::TOATDesign(&tdsFile, "regular", 4);

    // Set the range of attribute x1. It will be the value of attribute rx1.
    mySampling->setRange("x1", "rx1");
    // Set the range of attribute x2. It will be 10.0% of x2's nominal value.
    mySampling->setRange("x2", 10.0, kTRUE);
    // Generate an OAT sampling where only x1 and x2 will vary.
    mySampling->generateSample();

    CPPUNIT_ASSERT(tdsFile.getNAttributes() == 3 + 2); // 3 original attributes plus __nominal_set__ and __modified_att__ . 
    CPPUNIT_ASSERT(tdsFile.getNPatterns() == 2 + (2 * 2 * 4)); // 2 nominal sets, 2 changing attributes, 4 generated values for each.    

    Double_t genData1[90];
    Double_t refData1[90] =
    { 3.00, 0.4, 2.000, 1.0, -1.0, 2.80, 0.4, 2.000, 1.0, 0.0, 3.20, 0.4, 2.000,
            1.0, 0.0, 2.90, 0.4, 2.000, 1.0, 0.0, 3.10, 0.4, 2.000, 1.0, 0.0,
            3.00, 0.4, 1.900, 1.0, 2.0, 3.00, 0.4, 2.100, 1.0, 2.0, 3.00, 0.4,
            1.950, 1.0, 2.0, 3.00, 0.4, 2.050, 1.0, 2.0, 5.00, 0.2, 3.000, 2.0,
            -1.0, 4.90, 0.2, 3.000, 2.0, 0.0, 5.10, 0.2, 3.000, 2.0, 0.0, 4.95,
            0.2, 3.000, 2.0, 0.0, 5.05, 0.2, 3.000, 2.0, 0.0, 5.00, 0.2, 2.850,
            2.0, 2.0, 5.00, 0.2, 3.150, 2.0, 2.0, 5.00, 0.2, 2.925, 2.0, 2.0,
            5.00, 0.2, 3.075, 2.0, 2.0 };

    tdsFile.getTuple()->extractData(genData1, 90,
            "x1:rx1:x2:__nominal_set__:__modified_att__", "", "line");
    for (Int_t i = 0; i < 90; i++)
        CPPUNIT_ASSERT_DOUBLES_EQUAL(refData1[i], genData1[i],
                DBL_PRECISION_COMPARE);

    delete mySampling;

    //-----------------------------------------//
    // TDS and attributes constructed manually //
    //-----------------------------------------//

    URANIE::DataServer::TNormalDistribution x1("x1", 0.0, 1.0);
    URANIE::DataServer::TTriangularDistribution x2("x2", -5.0, 5.0);
    URANIE::DataServer::TAttribute r1("r1");

    URANIE::DataServer::TDataServer tdsMem("tdsMem", "");
    tdsMem.addAttribute(&x1);
    x1.setDefaultValue(5.0);
    tdsMem.addAttribute(&x2);
    x2.setDefaultValue(-1.0);
    tdsMem.addAttribute(&r1);
    r1.setDefaultValue(20.0);

    // "random" mode
    mySampling = new URANIE::Sampler::TOATDesign(&tdsMem, "lhs", 9);
    mySampling->setSeed(1234);
    mySampling->setRange("x1", "r1", kTRUE);
    mySampling->setRange("x2", 4.0);
    mySampling->generateSample();

    CPPUNIT_ASSERT(tdsMem.getNAttributes() == 3 + 2); // 3 original attributes plus __nominal_set__ and __modified_att__ . 
    CPPUNIT_ASSERT(tdsMem.getNPatterns() == 1 + (1 * 2 * 9)); // 1 nominal set, 1 changing attribute, 9 generated values.    

    Double_t genDataX1[9];
    Double_t refDataX1[9] =
    { 4.5191197, 5.2618786, 4.8344788, 4.6757962, 4.5612031, 4.7396020,
            4.6581903, 4.6150053, 4.5515529 };

    Double_t genDataX2[9];
    Double_t refDataX2[9] =
    { -1.569249, -0.199756, -1.836331, 0.0234720, -0.797519, 0.4397354,
            -2.953870, -1.040587, 0.7741520 };

    tdsMem.getTuple()->extractData(genDataX1, 9, "x1", "__modified_att__ == 1",
            "line");
    tdsMem.getTuple()->extractData(genDataX2, 9, "x2", "__modified_att__ == 2",
            "line");

    for (Int_t i = 0; i < 9; i++)
    {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(refDataX1[i], genDataX1[i],
                DBL_PRECISION_COMPARE);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(refDataX2[i], genDataX2[i],
                DBL_PRECISION_COMPARE);
    }

    delete mySampling;
    
     //-----------------------------------------//
    // TDS and attributes constructed manually //
    //-----------------------------------------//

    URANIE::DataServer::TDataServer tdsR("tdsR", "");
    tdsR.addAttribute(&x1);
    x1.setDefaultValue(5.0);
    tdsR.addAttribute(&x2);
    x2.setDefaultValue(-1.0);
    tdsR.addAttribute(&r1);
    r1.setDefaultValue(20.0);

    // "random" mode
    mySampling = new URANIE::Sampler::TOATDesign(&tdsR, "random", 9);
    mySampling->setSeed(1234);
    mySampling->setRange("x1", "r1", kTRUE);
    mySampling->setRange("x2", 4.0);
    mySampling->generateSample();

    CPPUNIT_ASSERT(tdsR.getNAttributes() == 3 + 2); // 3 original attributes plus __nominal_set__ and __modified_att__ . 
    CPPUNIT_ASSERT(tdsR.getNPatterns() == 1 + (1 * 2 * 9)); // 1 nominal set, 1 changing attribute, 9 generated values.    

    Double_t genDataY1[9];
    Double_t refDataY1[9] =
    { 4.807963, 4.895716, 4.566754, 4.534375, 4.567865, 4.546525, 4.828334, 4.841752, 5.110771};

    Double_t genDataY2[9];
    Double_t refDataY2[9] =
    { -1.267646, 0.03825239, 0.3180332, -1.11316, -2.364178,  -0.4353658,-2.616766, -0.630504,-0.8430925 };

    tdsR.getTuple()->extractData(genDataY1, 9, "x1", "__modified_att__ == 1",
            "line");
    tdsR.getTuple()->extractData(genDataY2, 9, "x2", "__modified_att__ == 2",
            "line");

    for (Int_t i = 0; i < 9; i++)
    {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(refDataY1[i], genDataY1[i],
                DBL_PRECISION_COMPARE);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(refDataY2[i], genDataY2[i],
                DBL_PRECISION_COMPARE);
    }

    delete mySampling;
    
    
}


void TOATDesignTest::testGenerateSampleWithStepValue()
{
    //-----------------------------------------//
    // TDS and attributes constructed manually //
    //-----------------------------------------//

    URANIE::DataServer::TNormalDistribution x1("x1", -2.0, 2.0);
    URANIE::DataServer::TTriangularDistribution x2("x2", -5.0, 5.0);

    URANIE::DataServer::TDataServer tdsMem("tdsMem", "");
    tdsMem.addAttribute(&x1);
    tdsMem.addAttribute(&x2);

    // In case of the value 0.0 is in the DoE
    Double_t myStepValues[2] = {0.01, 0.35};
    tdsMem.getAttribute("x1")->setStepValue(myStepValues[0]);
    tdsMem.getAttribute("x2")->setStepValue(myStepValues[1]);

    tdsMem.createTuple();
    Int_t nx = 2;
    Int_t ns = 0;

    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of patterns", ns, tdsMem.getNPatterns());
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of attributes", nx, tdsMem.getNAttributes());

    Double_t myData1[3] = {0.0, -1.0, 4.0};
    tdsMem.getTuple()->Fill(myData1);
    Double_t myData2[3] = {1.0, 0.0, 0.0};
    tdsMem.getTuple()->Fill(myData2);
    Double_t myData3[3] = {2.0, 1.0, -4.0};
    tdsMem.getTuple()->Fill(myData3);
    ns = 3;
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of patterns", ns, tdsMem.getNPatterns());
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of attributes", nx, tdsMem.getNAttributes());

    // "regular" mode
    mySampling = new URANIE::Sampler::TOATDesign(&tdsMem, "regular", 2);
    mySampling->setSeed(1234);
    mySampling->setRange("x1", 1.0, kTRUE);
    mySampling->setRange("x2", 4.0, kTRUE);
    mySampling->generateSample();

    ns = 15;
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of patterns", ns, tdsMem.getNPatterns());
    // Add the auxillary attributes
    nx+=2;
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of attributes", nx, tdsMem.getNAttributes());

    // Point 1 : myData1
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("With Step Value, Point 1 ", myData1[1], tdsMem.getValue("x1", 0), SPL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("With Step Value, Point 1", myData1[2], tdsMem.getValue("x2", 0), SPL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("With Step Value, Point 1", 0.995*myData1[1], tdsMem.getValue("x1", 1), SPL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("With Step Value, Point 1", myData1[2], tdsMem.getValue("x2", 1), SPL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("With Step Value, Point 1", 1.005*myData1[1], tdsMem.getValue("x1", 2), SPL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("With Step Value, Point 1", myData1[2], tdsMem.getValue("x2", 2), SPL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("With Step Value, Point 1", myData1[1], tdsMem.getValue("x1", 3), SPL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("With Step Value, Point 1", 0.98*myData1[2], tdsMem.getValue("x2", 3), SPL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("With Step Value, Point 1", myData1[1], tdsMem.getValue("x1", 4), SPL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("With Step Value, Point 1", 1.02*myData1[2], tdsMem.getValue("x2", 4), SPL_PRECISION_COMPARE);

    // Point 2 : myData2
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("With Step Value, Point 2", myData2[1], tdsMem.getValue("x1", 5), SPL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("With Step Value, Point 2", myData2[2], tdsMem.getValue("x2", 5), SPL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("With Step Value, Point 2", -0.5*myStepValues[0], tdsMem.getValue("x1", 6), SPL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("With Step Value, Point 2", myData2[2], tdsMem.getValue("x2", 6), SPL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("With Step Value, Point 2", 0.5*myStepValues[0], tdsMem.getValue("x1", 7), SPL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("With Step Value, Point 2", myData2[2], tdsMem.getValue("x2", 7), SPL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("With Step Value, Point 2", myData2[1], tdsMem.getValue("x1", 8), SPL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("With Step Value, Point 2", -0.5*myStepValues[1], tdsMem.getValue("x2", 8), SPL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("With Step Value, Point 2", myData2[1], tdsMem.getValue("x1", 9), SPL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("With Step Value, Point 2", 0.5*myStepValues[1], tdsMem.getValue("x2", 9), SPL_PRECISION_COMPARE);

    // Point 3 : myData3
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("With Step Value, Point 3 ", myData3[1], tdsMem.getValue("x1", 10), SPL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("With Step Value, Point 3", myData3[2], tdsMem.getValue("x2", 10), SPL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("With Step Value, Point 3", 0.995*myData3[1], tdsMem.getValue("x1", 11), SPL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("With Step Value, Point 3", myData3[2], tdsMem.getValue("x2", 11), SPL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("With Step Value, Point 3", 1.005*myData3[1], tdsMem.getValue("x1", 12), SPL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("With Step Value, Point 3", myData3[2], tdsMem.getValue("x2", 12), SPL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("With Step Value, Point 3", myData3[1], tdsMem.getValue("x1", 13), SPL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("With Step Value, Point 3", 0.98*myData3[2], tdsMem.getValue("x2", 13), SPL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("With Step Value, Point 3", myData3[1], tdsMem.getValue("x1", 14), SPL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("With Step Value, Point 3", 1.02*myData3[2], tdsMem.getValue("x2", 14), SPL_PRECISION_COMPARE);

    delete mySampling;
}
