//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TestTPetras.cxx
 \author Maud Salvoldi (OCC/Areva group)
 \date   $Date$
 
 \brief  Test File of URANIE::Sampler::TPetras class
 
 Cr�e le : mer nov 02 17:00:27 2011
 
 */

// Uranie
#include "TestTPetras.h"
#include "TAttribute.h"
#include "TUniformDistribution.h"
#include "TNormalDistribution.h"
#include "TWeibullDistribution.h"
#include "TUniformByPartsDistribution.h"
#include "TBetaDistribution.h"
#include "TCauchyDistribution.h"
#include "TLogUniformDistribution.h"
#include "TLogNormalDistribution.h"
#include "TTriangularDistribution.h"
#include "TLogTriangularDistribution.h"
#include "TExponentialDistribution.h"
#include "UDisableTest.h"

// ROOT
#include "TString.h"

#include <iostream>
#include <math.h>

#define DBL_PRECISION_COMPARE 1.e-10
#define EPSILON 0.04

TPetrasTest::TPetrasTest() :
        _tds(NULL), _sampler(NULL)
{
}

TPetrasTest::~TPetrasTest()
{
  //tearDown();
}
void TPetrasTest::setUp()
{
    // Create a TDataServer
    _tds = new URANIE::DataServer::TDataServer();
    // Fill the DataServer with the three attributes of the study
    _tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("Uniform", 0., 1.));
    _tds->addAttribute(
            new URANIE::DataServer::TNormalDistribution("Normal", 0., 1.));
    _tds->addAttribute(
            new URANIE::DataServer::TLogUniformDistribution("loguniform", .001,
                    10.));
    _tds->addAttribute(
            new URANIE::DataServer::TLogNormalDistribution("lognormal", 0.01,
                    3.0));
    _tds->addAttribute(
            new URANIE::DataServer::TTriangularDistribution("triangular", 0.0,
                    1.0, 0.5));
    _tds->addAttribute(
            new URANIE::DataServer::TLogTriangularDistribution("logtriangular",
                    0.01, 10., 5.0));
    _tds->addAttribute(
            new URANIE::DataServer::TExponentialDistribution("exp", .1, 12.0));
    _tds->addAttribute(
            new URANIE::DataServer::TWeibullDistribution("Weibull", 0.2, 0.4));
    _tds->addAttribute(
            new URANIE::DataServer::TUniformByPartsDistribution("UnifParts",
                    1.0, 4.0, 2.0));
    _tds->addAttribute(
            new URANIE::DataServer::TBetaDistribution("Beta", 0.5, 0.5));
    _tds->addAttribute(
            new URANIE::DataServer::TCauchyDistribution("Cauchy", 1.0, 2.0));
    _tds->getAttribute("Uniform")->delShare();
    _tds->getAttribute("loguniform")->delShare();
    _tds->getAttribute("Normal")->delShare();
    _tds->getAttribute("lognormal")->delShare();
    _tds->getAttribute("triangular")->delShare();
    _tds->getAttribute("logtriangular")->delShare();
    _tds->getAttribute("exp")->delShare();
    _tds->getAttribute("Weibull")->delShare();
    _tds->getAttribute("UnifParts")->delShare();
    _tds->getAttribute("Beta")->delShare();
    _tds->getAttribute("Cauchy")->delShare();
    
    // Create the sampler
    _sampler = new URANIE::Sampler::TPetras(_tds, "petras", 4);
    _sampler->unsetLog();
}

void TPetrasTest::tearDown()
{
    if (_sampler != NULL)
    {
        delete _sampler;
        _sampler = NULL;
    }
    if (_tds != NULL)
    {
        delete _tds;
        _tds = NULL;
    }

}

void TPetrasTest::testTPetrasConstructor()
{

    TString sname = "_Sampler_Method_Uranie";
    CPPUNIT_ASSERT_EQUAL(0, strcmp(sname.Data(), _sampler->getMethodName()));
    CPPUNIT_ASSERT_EQUAL(1, _sampler->_nS);
    CPPUNIT_ASSERT_EQUAL(11, _sampler->_nX);

}
void TPetrasTest::testTPetrasgenerateSample()
{
    _sampler->generateSample();
    _tds->computeStatistic("*");

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("Uniform", 0), 0.5,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("Uniform", 7766),
            0.0197543656459899, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.5, _tds->getAttribute("Uniform")->getMean(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.218853410248204,
            _tds->getAttribute("Uniform")->getStd(), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("Normal", 0), 0.0,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("Normal", 7766),
            -1.21228492938631, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, _tds->getAttribute("Normal")->getMean(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.692287427827268,
            _tds->getAttribute("Normal")->getStd(), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("loguniform", 0), 0.1,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("loguniform", 7766), 0.1,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.64063102885492,
            _tds->getAttribute("loguniform")->getMean(), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.31355720930553,
            _tds->getAttribute("loguniform")->getStd(), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("lognormal", 0), 0.008001057113,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("lognormal", 7766),
            0.008001057113, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.00890828792346087,
            _tds->getAttribute("lognormal")->getMean(), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.00439158803796448,
            _tds->getAttribute("lognormal")->getStd(), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("triangular", 0), 0.5,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("triangular", 7766), 0.5,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.5,
            _tds->getAttribute("triangular")->getMean(), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.149273128311853,
            _tds->getAttribute("triangular")->getStd(), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("logtriangular", 0),
            1.02819925869063, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("logtriangular", 7766),
            1.02819925869063, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.48155094627717,
            _tds->getAttribute("logtriangular")->getMean(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.49599663127436,
            _tds->getAttribute("logtriangular")->getStd(),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("exp", 0), 18.9314718055995,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("exp", 7766), 18.9314718055995,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(20.4231314122073,
            _tds->getAttribute("exp")->getMean(), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(6.34494898703091,
            _tds->getAttribute("exp")->getStd(), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("Weibull", 0),
            0.0800006744364424, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("Weibull", 7766),
            0.080000674436442, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.289799410712487,
            _tds->getAttribute("Weibull")->getMean(), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.572768806354293,
            _tds->getAttribute("Weibull")->getStd(), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("UnifParts", 0), 2.0,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("UnifParts", 7766), 2.0,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.12354577509182,
            _tds->getAttribute("UnifParts")->getMean(), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.680957140939055,
            _tds->getAttribute("UnifParts")->getStd(), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("Beta", 0), 0.5,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("Beta", 7766), 0.5,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.5, _tds->getAttribute("Beta")->getMean(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.264411404850702,//5052087,
            _tds->getAttribute("Beta")->getStd(), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("Cauchy", 0), 2.0,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("Cauchy", 7766), 2.0,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, _tds->getAttribute("Cauchy")->getMean(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.91516433227787,
            _tds->getAttribute("Cauchy")->getStd(), DBL_PRECISION_COMPARE);
}

void TPetrasTest::throwExceptionTPetrasConstructorTdsNull()
{

    URANIE::DataServer::TDataServer * tdsnull = NULL;
    CPPUNIT_ASSERT_THROW(new URANIE::Sampler::TPetras(tdsnull, "petras", 100),
            URANIE::Exceptions::UErrorExceptions);
}
void TPetrasTest::throwExceptionTPetrasConstructorLevelNull()
{

    CPPUNIT_ASSERT_THROW(new URANIE::Sampler::TPetras(_tds, "petras", 0),
            URANIE::Exceptions::UErrorExceptions);
}

void TPetrasTest::throwExceptionTPetrasConstructorLevelNoIntegerPositive()
{

    CPPUNIT_ASSERT_THROW(new URANIE::Sampler::TPetras(_tds, "petras", -10),
            URANIE::Exceptions::UErrorExceptions);
}
void TPetrasTest::throwExceptionTPetrasConstructorLevelHigh()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
  URANIE_IGNORED_TEST()    ;

    CPPUNIT_ASSERT_THROW(new URANIE::Sampler::TPetras(_tds, "petras", 1000),
            URANIE::Exceptions::UErrorExceptions);
}
