//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTSampler.cxx
 \author Cedric Forgues <uranietm@cea.fr>
 \date   $Date$
 
 \brief  Test File of URANIE::Sampler::TSampler class
 
 Crée le : mer nov 21 07:59:24 2007
 
 */

// Uranie
#include "TestTSampler.h"
#include "TAttribute.h"
#include "TUniformDistribution.h"
#include "TNormalDistribution.h"
#include "TWeibullDistribution.h"
#include "TUniformByPartsDistribution.h"
#include "TBetaDistribution.h"
#include "TCauchyDistribution.h"

// ROOT
#include "TString.h"

#include <iostream>
#include <math.h>

#define DBL_PRECISION_COMPARE 1.e-10
#define EPSILON 0.04

using namespace std;

TSamplerTest::TSamplerTest()
{
}

TSamplerTest::~TSamplerTest()
{
}

void TSamplerTest::setUp()
{
  // Create a TDataServer
  _tds1 = new URANIE::DataServer::TDataServer();
  // Fill the DataServer with the three attributes of the study
  _tds1->addAttribute(
		      new URANIE::DataServer::TUniformDistribution("Uniform", 0., 1.));
  _tds1->addAttribute(
		      new URANIE::DataServer::TNormalDistribution("Normal", 0., 1.));
  _tds1->addAttribute(
		      new URANIE::DataServer::TWeibullDistribution("Weibull", 2., 1.));
  _tds1->addAttribute(
		      new URANIE::DataServer::TUniformByPartsDistribution("UnifParts",
									  1.0, 4.0, 2.0));
  _tds1->addAttribute(
		      new URANIE::DataServer::TBetaDistribution("Beta", 1.0, 4.0));
  _tds1->addAttribute(
		      new URANIE::DataServer::TCauchyDistribution("Cauchy", 0.5, 2.0));

  _tds1->getAttribute("Uniform")->delShare();
  _tds1->getAttribute("Normal")->delShare();
  _tds1->getAttribute("Weibull")->delShare();
  _tds1->getAttribute("UnifParts")->delShare();
  _tds1->getAttribute("Beta")->delShare();
  _tds1->getAttribute("Cauchy")->delShare();
}

void TSamplerTest::tearDown()
{
    delete _tds1;
    _tds1 = NULL;
}

////////////////////////////////
///                          ///                            
///         Tests            ///
///                          ///
////////////////////////////////

void TSamplerTest::testSampling()
{

    // Generate the sampling from the TDataServer with lhs method
    URANIE::Sampler::TSampling *samplinglhs = new URANIE::Sampler::TSampling(
            _tds1, "lhs", 1000);
    samplinglhs->generateSample();

    _tds1->computeStatistic("Uniform:Normal:UnifParts:Weibull");

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Lhs-Uniform-mean:", 0.5,
            _tds1->getAttribute("Uniform")->getMean(), EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Lhs-Uniform-std:", 0.288,
            _tds1->getAttribute("Uniform")->getStd(), EPSILON);

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Lhs-Normal-mean", 0.0,
            _tds1->getAttribute("Normal")->getMean(), EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Lhs-Normal-std", 1.0,
            _tds1->getAttribute("Normal")->getStd(), EPSILON);

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Lhs-UniformByParts-mean", 2.25,
            _tds1->getAttribute("UnifParts")->getMean(), EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Lhs-UniformByParts-mean", 0.878,
            _tds1->getAttribute("UnifParts")->getStd(), EPSILON);

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Lhs-Weibull-mean", 2.0,
            _tds1->getAttribute("Weibull")->getMean(), 0.1);

    // Generate the sampling from the TDataServer with srs method
    URANIE::Sampler::TSampling *samplingsrs = new URANIE::Sampler::TSampling(
            _tds1, "srs", 60000);
    samplingsrs->generateSample();

    _tds1->computeStatistic("Uniform:Normal:UnifParts:Weibull");

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Srs-Uniform-mean", 0.5,
            _tds1->getAttribute("Uniform")->getMean(), EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Srs-Uniform-std", 0.288,
            _tds1->getAttribute("Uniform")->getStd(), EPSILON);

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Srs-Normal-mean", 0.0,
            _tds1->getAttribute("Normal")->getMean(), EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Srs-Normal-std", 1.0,
            _tds1->getAttribute("Normal")->getStd(), EPSILON);

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Srs-UniformByParts-mean", 2.250,
            _tds1->getAttribute("UnifParts")->getMean(), 0.01);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Srs-UniformByParts-mean", 0.878,
            _tds1->getAttribute("UnifParts")->getStd(), 0.01);

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Srs-Weibull-mean", 2.0,
            _tds1->getAttribute("Weibull")->getMean(), 0.1);

    delete samplingsrs;
    delete samplinglhs;
}

void TSamplerTest::testSamplingWithSeed()
{
    // Generate the sampling from the TDataServer with lhs method
    Int_t nS = 1000;
    URANIE::Sampler::TSampling *samp = new URANIE::Sampler::TSampling(_tds1,
            "lhs", nS);
    samp->setSeed(123456789);
    samp->generateSample();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Uniform", 0),
            0.683634692654712, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Uniform", 1),
            0.64559146620892, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Uniform", nS - 1),
            0.10917467899085, DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Normal", 0),
            0.603014413619771, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Normal", 1),
            0.822573336656192, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Normal", nS - 1),
            0.244877978768316, DBL_PRECISION_COMPARE);

    // -- ngi --
    // The precision on the Beta distribution computation has been reduced from 1e-12 to 1e-9 in order to 
    // speed up the computation. As a consequence, the generated values have to be updated.
    // -- ngi --  
//CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Beta",0), 0.179808865577009, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Beta", 0), 0.179808865301311,
            DBL_PRECISION_COMPARE);
//CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Beta",1), 0.0487459513574975, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Beta", 1), 0.0487459516152739,
            DBL_PRECISION_COMPARE);
//CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Beta",nS-1), 0.292439986601189, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Beta", nS - 1),
            0.292439986020326, DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Cauchy", 0), 2.28778884148701,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Cauchy", 1), 1.95296099789086,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Cauchy", nS - 1),
            6.69006617908894, DBL_PRECISION_COMPARE);

    // SECOND GENERATE SAMPLING
    samp->generateSample();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Uniform", 0),
            0.143721841896418, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Uniform", 1),
            0.985581324541476, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Uniform", nS - 1),
            0.116021217064699, DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Normal", 0),
            -1.10169161365532, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Normal", 1),
            -1.08403964109838, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Normal", nS - 1),
            -0.958460684910756, DBL_PRECISION_COMPARE);

    // -- ngi --
    // The precision on the Beta distribution computation has been reduced from 1e-12 to 1e-9 in order to 
    // speed up the computation. As a consequence, the generated values have to be updated.
    // -- ngi --  
//CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Beta",0), 0.106526256356614, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Beta", 0), 0.106526256538927,
            DBL_PRECISION_COMPARE);
//CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Beta",1), 0.22465820111019, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Beta", 1), 0.22465820144861,
            DBL_PRECISION_COMPARE);
//CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Beta",nS-1), 0.0485634230194307, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Beta", nS - 1),
            0.0485634231008589, DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Cauchy", 0), 1.44879230312818,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Cauchy", 1), 2.69285372755024,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Cauchy", nS - 1),
            2.47811125298466, DBL_PRECISION_COMPARE);

    // THIRD GENERATE SAMPLING WITH THE INIT SEED  ===EQUAL TO=== FIRST SAMPLING
    samp->setSeed(123456789);
    samp->generateSample();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Uniform", 0),
            0.683634692654712, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Uniform", 1),
            0.64559146620892, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Uniform", nS - 1),
            0.10917467899085, DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Normal", 0),
            0.603014413619771, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Normal", 1),
            0.822573336656192, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Normal", nS - 1),
            0.244877978768316, DBL_PRECISION_COMPARE);

    // -- ngi --
    // The precision on the Beta distribution computation has been reduced from 1e-12 to 1e-9 in order to 
    // speed up the computation. As a consequence, the generated values have to be updated.
    // -- ngi --  
//CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Beta",0), 0.179808865577009, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Beta", 0), 0.179808865301311,
            DBL_PRECISION_COMPARE);
//CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Beta",1), 0.0487459513574975, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Beta", 1), 0.0487459516152739,
            DBL_PRECISION_COMPARE);
//CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Beta",nS-1), 0.292439986601189, DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Cauchy", 0), 2.28778884148701,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Cauchy", 1), 1.95296099789086,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Cauchy", nS - 1),
            6.69006617908894, DBL_PRECISION_COMPARE);

    delete samp;
}

void TSamplerTest::testQMC()
{

    // Generate the sampling from the TDataServer with a Sobol sequence
    URANIE::Sampler::TQMC *qmcSobol = new URANIE::Sampler::TQMC(_tds1, "sobol",
            10000);
    qmcSobol->generateSample();
    _tds1->computeStatistic("Uniform:Normal:UnifParts");

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.5, _tds1->getAttribute("Uniform")->getMean(),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.288,
            _tds1->getAttribute("Uniform")->getStd(), EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, _tds1->getAttribute("Normal")->getMean(),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, _tds1->getAttribute("Normal")->getStd(),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.250,
            _tds1->getAttribute("UnifParts")->getMean(), 0.001);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.8774,
            _tds1->getAttribute("UnifParts")->getStd(), 0.001);

    // Generate the sampling from the TDataServer with a Halton sequence
    URANIE::Sampler::TQMC *qmcHalton = new URANIE::Sampler::TQMC(_tds1,
            "halton", 10000);
    qmcHalton->generateSample();
    _tds1->computeStatistic("Uniform:Normal:UnifParts");

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.5, _tds1->getAttribute("Uniform")->getMean(),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.288,
            _tds1->getAttribute("Uniform")->getStd(), EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, _tds1->getAttribute("Normal")->getMean(),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, _tds1->getAttribute("Normal")->getStd(),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.250,
            _tds1->getAttribute("UnifParts")->getMean(), 0.001);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.878,
            _tds1->getAttribute("UnifParts")->getStd(), 0.01);

    delete qmcSobol;
    delete qmcHalton;
}
