//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TestTSampling.cxx
 \author Maud Salvoldi (OCC/Areva group)
 \date   $Date$
 
 \brief  Test File of URANIE::Sampler::TSampler class
 
 Cr�e le : mer nov 02 17:00:27 2011
 
 */

// Uranie
#include "TestTSampling.h"
#include "TAttribute.h"
#include "TUniformDistribution.h"
#include "TNormalDistribution.h"
#include "TWeibullDistribution.h"
#include "TUniformByPartsDistribution.h"
#include "TBetaDistribution.h"
#include "TCauchyDistribution.h"
#include "TNormalDistribution.h"
#include "TLogUniformDistribution.h"
#include "TLogNormalDistribution.h"
#include "TTriangularDistribution.h"
#include "TLogTriangularDistribution.h"
#include "TExponentialDistribution.h"
#include "UDisableTest.h"

// ROOT
#include "TString.h"

#include <iostream>
#include <math.h>

#define DBL_PRECISION_COMPARE 1.e-10
#define EPSILON 0.04

TSamplingTest::TSamplingTest() :
        _tds1(NULL), _tds(NULL), _sampler(NULL), _sampling(NULL), _nS(1000)
{
}

TSamplingTest::~TSamplingTest()
{
    if (_sampler)
    {
        delete _sampler;
        _sampler = NULL;
    }
    if (_sampling)
    {
        delete _sampling;
        _sampling = NULL;
    }
    if (_tds)
    {
        delete _tds;
        _tds = NULL;
    }
    if (_tds1)
    {
        delete _tds1;
        _tds1 = NULL;
    }
}

void TSamplingTest::setUp()
{
  // Create a TDataServer
  _tds1 = new URANIE::DataServer::TDataServer("tds1", "my first TDS");
  // Fill the DataServer with the three attributes of the study
  _tds1->addAttribute(
		      new URANIE::DataServer::TUniformDistribution("Uniform", 0., 1.));
  _tds1->addAttribute(
		      new URANIE::DataServer::TNormalDistribution("Normal", 0., 1.));
  _tds1->addAttribute(
		      new URANIE::DataServer::TWeibullDistribution("Weibull", 2., 1.));
  _tds1->addAttribute(
		      new URANIE::DataServer::TUniformByPartsDistribution("UnifParts",
									  1.0, 4.0, 2.0));
  _tds1->addAttribute(
		      new URANIE::DataServer::TBetaDistribution("Beta", 1.0, 4.0));
  _tds1->addAttribute(
		      new URANIE::DataServer::TCauchyDistribution("Cauchy", 0.5, 2.0));

  _tds = new URANIE::DataServer::TDataServer("tdsother", "my other TDS");
  _tds->addAttribute(
		     new URANIE::DataServer::TUniformDistribution("Uniform", 0., 1.));
  _tds->addAttribute(
		     new URANIE::DataServer::TNormalDistribution("Normal", 0., 1.));
  _tds->addAttribute(
		     new URANIE::DataServer::TLogUniformDistribution("lu", .001, 10.));
  _tds->addAttribute(
		     new URANIE::DataServer::TLogNormalDistribution("ln", 0.01, 3.0));
  _tds->addAttribute(
		     new URANIE::DataServer::TTriangularDistribution("t", 5.0, 8., 6.0));
  _tds->addAttribute(
		     new URANIE::DataServer::TLogTriangularDistribution("lt", .001, 10.,
									2.5));
  _tds->addAttribute(
		     new URANIE::DataServer::TExponentialDistribution("e", .1, 12.0));

  //HelpCleaning
  _tds1->getAttribute("Uniform")->delShare();
  _tds1->getAttribute("Normal")->delShare();
  _tds1->getAttribute("Weibull")->delShare();
  _tds1->getAttribute("UnifParts")->delShare();
  _tds1->getAttribute("Beta")->delShare();
  _tds1->getAttribute("Cauchy")->delShare();
      
  _tds->getAttribute("Uniform")->delShare();
  _tds->getAttribute("Normal")->delShare();
  _tds->getAttribute("lu")->delShare();
  _tds->getAttribute("ln")->delShare();
  _tds->getAttribute("t")->delShare();
  _tds->getAttribute("lt")->delShare();
  _tds->getAttribute("e")->delShare();
    

  _sampler = new URANIE::Sampler::TSampling(_tds1, "lhs", _nS);
  //  _sampler->setLog();
  _sampling = new URANIE::Sampler::TSampling(_tds, "lhs", _nS);
  //  _sampling->setLog();
}

void TSamplingTest::tearDown()
{
    delete _sampler;
    _sampler = NULL;
    delete _sampling;
    _sampling = NULL;
    delete _tds;
    _tds = NULL;
    delete _tds1;
    _tds1 = NULL;
}

////////////////////////////////
///                          ///                            
///         Tests            ///
///                          ///
////////////////////////////////

void TSamplingTest::testSampling()
{

    // Generate the sampling from the TDataServer with lhs method
    _sampler->generateSample();

    _tds1->computeStatistic("Uniform:Normal:UnifParts:Weibull");

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Lhs-Uniform-mean:", 0.5,
            _tds1->getAttribute("Uniform")->getMean(), EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Lhs-Uniform-std:", 0.288,
            _tds1->getAttribute("Uniform")->getStd(), EPSILON);

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Lhs-Normal-mean", 0.0,
            _tds1->getAttribute("Normal")->getMean(), EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Lhs-Normal-std", 1.0,
            _tds1->getAttribute("Normal")->getStd(), EPSILON);

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Lhs-UniformByParts-mean", 2.25,
            _tds1->getAttribute("UnifParts")->getMean(), EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Lhs-UniformByParts-mean", 0.878,
            _tds1->getAttribute("UnifParts")->getStd(), EPSILON);

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Lhs-Weibull-mean", 2.0,
            _tds1->getAttribute("Weibull")->getMean(), 0.1);

    // Generate the sampling from the TDataServer with srs method
    URANIE::Sampler::TSampling *samplingsrs = new URANIE::Sampler::TSampling(
            _tds1, "srs", 60000);
//  samplingsrs->setLog();
    samplingsrs->generateSample();

    _tds1->computeStatistic("Uniform:Normal:UnifParts:Weibull");

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Srs-Uniform-mean", 0.5,
            _tds1->getAttribute("Uniform")->getMean(), EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Srs-Uniform-std", 0.288,
            _tds1->getAttribute("Uniform")->getStd(), EPSILON);

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Srs-Normal-mean", 0.0,
            _tds1->getAttribute("Normal")->getMean(), EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Srs-Normal-std", 1.0,
            _tds1->getAttribute("Normal")->getStd(), EPSILON);

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Srs-UniformByParts-mean", 2.250,
            _tds1->getAttribute("UnifParts")->getMean(), 0.015);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Srs-UniformByParts-mean", 0.878,
            _tds1->getAttribute("UnifParts")->getStd(), 0.015);

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Srs-Weibull-mean", 2.0,
            _tds1->getAttribute("Weibull")->getMean(), 0.1);

    // clean
    delete samplingsrs;
    samplingsrs = NULL;
}

void TSamplingTest::testSamplingWithSeed()
{
    // Generate the sampling from the TDataServer with lhs method
    _sampler->setSeed(123456789);
    _sampler->generateSample();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Uniform", 0),
            0.683634692654712, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Uniform", 1),
            0.64559146620892, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Uniform", _nS - 1),
            0.10917467899085, DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Normal", 0),
            0.603014413619771, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Normal", 1),
            0.822573336656192, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Normal", _nS - 1),
            0.244877978768316, DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Beta", 0), 0.179808865577009,//301311,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Beta", 1), 0.0487459513574975,//615274,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Beta", _nS - 1),
                                 0.292439986601189, DBL_PRECISION_COMPARE); //020327

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Cauchy", 0), 2.28778884148701,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Cauchy", 1), 1.95296099789086,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Cauchy", _nS - 1),
            6.69006617908894, DBL_PRECISION_COMPARE);

    // SECOND GENERATE SAMPLING
    _sampler->generateSample();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Uniform", 0),
            0.143721841896418, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Uniform", 1),
            0.985581324541476, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Uniform", _nS - 1),
            0.116021217064699, DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Normal", 0),
            -1.10169161365532, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Normal", 1),
            -1.08403964109838, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Normal", _nS - 1),
            -0.958460684910756, DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Beta", 0), 0.106526256356614,//538928,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Beta", 1), 0.22465820111019,//448619,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Beta", _nS - 1),
            .0485634230194307, DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Cauchy", 0), 1.44879230312818,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Cauchy", 1), 2.69285372755024,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Cauchy", _nS - 1),
            2.47811125298466, DBL_PRECISION_COMPARE);

    // THIRD GENERATE SAMPLING WITH THE INIT SEED  ===EQUAL TO=== FIRST SAMPLING
    _sampler->setSeed(123456789);
    _sampler->generateSample();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Uniform", 0),
            0.683634692654712, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Uniform", 1),
            0.64559146620892, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Uniform", _nS - 1),
            0.10917467899085, DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Normal", 0),
            0.603014413619771, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Normal", 1),
            0.822573336656192, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Normal", _nS - 1),
            0.244877978768316, DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Beta", 0), 0.179808865577009,//301311,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Beta", 1), 0.0487459513574975,//615274,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Beta", _nS - 1),
                                 0.292439986601189, DBL_PRECISION_COMPARE); //020327

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Cauchy", 0), 2.28778884148701,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Cauchy", 1), 1.95296099789086,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds1->getValue("Cauchy", _nS - 1),
            6.69006617908894, DBL_PRECISION_COMPARE);
}

void TSamplingTest::testTSamplingConstrutorTypeFalse()
{

    URANIE::Sampler::TSampling *slgDefault = new URANIE::Sampler::TSampling(
            _tds1, "toto", _nS);
//  slgDefault->setLog();T
    slgDefault->generateSample();

    _tds1->computeStatistic("Uniform:Normal:UnifParts:Weibull");

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Lhs-Uniform-mean:", 0.5,
            _tds1->getAttribute("Uniform")->getMean(), EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Lhs-Uniform-std:", 0.288,
            _tds1->getAttribute("Uniform")->getStd(), EPSILON);

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Lhs-Normal-mean", 0.0,
            _tds1->getAttribute("Normal")->getMean(), EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Lhs-Normal-std", 1.0,
            _tds1->getAttribute("Normal")->getStd(), EPSILON);

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Lhs-UniformByParts-mean", 2.25,
            _tds1->getAttribute("UnifParts")->getMean(), EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Lhs-UniformByParts-mean", 0.878,
            _tds1->getAttribute("UnifParts")->getStd(), EPSILON);

    // clean
    delete slgDefault;
    slgDefault = NULL;
}

void TSamplingTest::testTSamplingNsEqual10()
{

    // Generate the sampling from the TDataServer with lhs method
    URANIE::Sampler::TSampling* sampler10 = new URANIE::Sampler::TSampling(
            _tds1, "lhs", 10);
    sampler10->generateSample();

    _tds1->computeStatistic("Uniform");

    cout << "0.5 " <<  _tds1->getAttribute("Uniform")->getMean() << endl;
    cout << "0.288 " <<  _tds1->getAttribute("Uniform")->getStd() << endl;
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Lhs-Uniform-mean:", 0.5,
            _tds1->getAttribute("Uniform")->getMean(), 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Lhs-Uniform-std:", 0.288,
            _tds1->getAttribute("Uniform")->getStd(), 0.1);

    delete sampler10;

}



void TSamplingTest::testTSamplingReproduction()
{
        URANIE::Sampler::TSampling* sampler10 = new URANIE::Sampler::TSampling(
                _tds1, "lhs", 10);
        URANIE::Sampler::TSampling* sampler10_1  = new URANIE::Sampler::TSampling(
                _tds1, "lhs", 10);
        CPPUNIT_ASSERT_ASSERTION_FAIL( CPPUNIT_ASSERT_EQUAL(sampler10->getSeed(), sampler10_1->getSeed()));
        delete sampler10;
        delete sampler10_1;
}

void TSamplingTest::throwExceptionTSamplingConstrutorTdsNull()
{
    URANIE::DataServer::TDataServer * tdsnull = NULL;
    CPPUNIT_ASSERT_THROW(new URANIE::Sampler::TSampling(tdsnull, "lhs", 1000),
            URANIE::Exceptions::UErrorExceptions);
}
void TSamplingTest::testgenerateSampleByDakota()
{
    // Generate the sampling from the TDataServer with lhs method
    _sampling->setSeed(123456789);
    _sampling->generateSampleByDakota();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("Uniform", 0),
            0.278364971598621, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("Uniform", 1),
            0.490972140277564, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("Uniform", _nS - 1),
            0.916722606670773, DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("Normal", 0), 0.663501215730457,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("Normal", 1), -1.28313721972726,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("Normal", _nS - 1),
            -0.375604275581462, DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("lu", 0), 0.918571402056657,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("lu", 1), 0.0674969906396236,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("lu", _nS - 1),
            0.0011676010357636, DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("t", 0), 6.223900405436,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("t", 1), 7.53248466919203,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("t", _nS - 1), 7.37629271368483,
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("lt", 0), 0.154910172510453,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("lt", 1), 0.807725934824511,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("lt", _nS - 1),
            4.17430728251563, DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("e", 0), 12.0099201336601,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("e", 1), 12.0367024637655,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("e", _nS - 1), 12.0175708095535,
            DBL_PRECISION_COMPARE);

// Generate the sampling from the TDataServer with srs method
    URANIE::Sampler::TSampling *samplingsrs = new URANIE::Sampler::TSampling(
            _tds, "srs", 1000);
//   samplingsrs->setLog();
    samplingsrs->generateSampleByDakota();

    _tds->computeStatistic("*");

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.5, _tds->getAttribute("Uniform")->getMean(),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.288, _tds->getAttribute("Uniform")->getStd(),
            EPSILON);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(6.34, _tds->getAttribute("t")->getMean(),
            0.08);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.62, _tds->getAttribute("t")->getStd(), 0.05);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(12.04, _tds->getAttribute("e")->getMean(),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.046, _tds->getAttribute("e")->getStd(),
            EPSILON);

    delete samplingsrs;
    samplingsrs = NULL;
}
void TSamplingTest::testgenerateSampleByDakotaSeedNull()
{
    // 3 calls of generateSampleByDakota with seeds equal to 0, 1234 and 0 
    // must give different results 

    //  seed = 0 
    _sampling->setSeed(0);
    _sampling->generateSampleByDakota();
    Double_t resTir1[3] = {_tds->getValue("Uniform", 0),
			 _tds->getValue("Uniform", 1),
			 _tds->getValue("Uniform", _nS - 1)}; 

    //  seed = 1234
    _sampling->setSeed(1234);
    _sampling->generateSampleByDakota();
    Double_t resTir2[3] = {_tds->getValue("Uniform", 0),
			 _tds->getValue("Uniform", 1),
			 _tds->getValue("Uniform", _nS - 1)}; 

     //  seed = 0 (1 second later)
    gSystem->Sleep(1000);
    _sampling->setSeed(0);
    _sampling->generateSampleByDakota();
    Double_t resTir3[3] = {_tds->getValue("Uniform", 0),
			 _tds->getValue("Uniform", 1),
			 _tds->getValue("Uniform", _nS - 1)}; 

    // compare
    Bool_t testRes = kTRUE;
    for(Int_t i = 0; i < 3; i++) 
    {
	testRes = (testRes 
		   && (fabs(resTir1[i] - resTir2[i]) > DBL_PRECISION_COMPARE)
		   && (fabs(resTir2[i] - resTir3[i]) > DBL_PRECISION_COMPARE)
		   && (fabs(resTir1[i] - resTir3[i]) > DBL_PRECISION_COMPARE));
    }

    CPPUNIT_ASSERT_EQUAL(kTRUE, testRes);
}
void TSamplingTest::throwExceptionTSamplinggenerateSampleByDakotaMatCorrNegative()
{
    _sampling->setUserCorrelation(0, 1, 0.9);
    _sampling->setUserCorrelation(0, 2, 0.9);
    _sampling->setUserCorrelation(1, 2, 0.1);
    CPPUNIT_ASSERT_THROW(_sampling->generateSampleByDakota(),
            URANIE::Exceptions::UErrorExceptions);

}

void TSamplingTest::throwExceptionTSamplinggenerateSampleByDakota99991()
{

    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_IGNORED_TEST()
    ;

    URANIE::Sampler::TSampling* _samplerKO = new URANIE::Sampler::TSampling(
            _tds, "lhs", 99991);
//   _samplerKO->setLog();
    CPPUNIT_ASSERT_THROW(_samplerKO->generateSampleByDakota(),
            URANIE::Exceptions::UErrorExceptions);

    delete _samplerKO;
}
void TSamplingTest::throwExceptionTSamplinggenerateSampleByDakotaLawNoKnown()
{
    // test loi de Weibull
    URANIE::DataServer::TDataServer tdsTestWeibull("tdsTest1","");
    tdsTestWeibull.addAttribute(new URANIE::DataServer::TWeibullDistribution("Weibull", 2., 1.));
    tdsTestWeibull.getAttribute("Weibull")->delShare();
									 
    URANIE::Sampler::TSampling *samplerTestWeibull;
    samplerTestWeibull = new URANIE::Sampler::TSampling(&tdsTestWeibull, "lhs", 10);
    
    CPPUNIT_ASSERT_THROW(samplerTestWeibull->generateSampleByDakota(),
			 URANIE::Exceptions::UErrorExceptions);

    // test loi inconnue
    URANIE::DataServer::TDataServer tdsTestNoLaw("tdsTest2","");
    tdsTestNoLaw.addAttribute(new URANIE::DataServer::TBetaDistribution("betaDistrib"));
    tdsTestNoLaw.getAttribute("betaDistrib")->delShare();
								    
    URANIE::Sampler::TSampling *samplerTestNoLaw;
    samplerTestNoLaw = new URANIE::Sampler::TSampling(&tdsTestNoLaw, "lhs", 10);

    CPPUNIT_ASSERT_THROW(samplerTestNoLaw->generateSampleByDakota(),
			 URANIE::Exceptions::UErrorExceptions);

    delete samplerTestWeibull;
    samplerTestWeibull = NULL;
    delete samplerTestNoLaw;
    samplerTestNoLaw = NULL;

}


void TSamplingTest::throwExceptionTSamplinggenerateSampleMatCorrNegative()
{
    _sampling->setUserCorrelation(0, 1, 0.9);
    _sampling->setUserCorrelation(0, 2, 0.9);
    _sampling->setUserCorrelation(1, 2, 0.1);
    CPPUNIT_ASSERT_THROW(_sampling->generateSample(),
            URANIE::Exceptions::UErrorExceptions);

}
void TSamplingTest::testTSamplingsetUserCorrelation()
{
    // coraltion positive
    Double_t dcorr = 0.987650;
    _sampler->setUserCorrelation(0, 1, dcorr);
    _sampler->generateSample(); 

    TMatrixD matCorrRk = _tds1->computeCorrelationMatrix("Normal:Uniform");

    CPPUNIT_ASSERT_DOUBLES_EQUAL(matCorrRk(0, 1), dcorr, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matCorrRk(1, 1), 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matCorrRk(1, 0), dcorr, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matCorrRk(0, 0), 1.0, DBL_PRECISION_COMPARE);

    //correlation negative
    dcorr = -0.45;
    _sampler->setUserCorrelation(0, 1, dcorr);
    _sampler->generateSample();

    matCorrRk = _tds1->computeCorrelationMatrix("Normal:Uniform");

    CPPUNIT_ASSERT_DOUBLES_EQUAL(matCorrRk(0, 1), dcorr, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matCorrRk(1, 1), 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matCorrRk(1, 0), dcorr, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matCorrRk(0, 0), 1.0, DBL_PRECISION_COMPARE);

    //correlation with attribute name
    dcorr = 0.987650;
    _sampler->setUserCorrelation("Uniform", "Normal", dcorr);
    _sampler->generateSample();

    matCorrRk = _tds1->computeCorrelationMatrix("Normal:Uniform");

    CPPUNIT_ASSERT_DOUBLES_EQUAL(matCorrRk(0, 1), dcorr, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matCorrRk(1, 1), 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matCorrRk(1, 0), dcorr, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matCorrRk(0, 0), 1.0, DBL_PRECISION_COMPARE);

    //correlation with attribute object
    dcorr = 0.987650;
    _sampler->setUserCorrelation(_tds1->getAttribute("Uniform"),
            _tds1->getAttribute("Normal"), dcorr);
    _sampler->generateSample();

    matCorrRk = _tds1->computeCorrelationMatrix("Normal:Uniform");

    CPPUNIT_ASSERT_DOUBLES_EQUAL(matCorrRk(0, 1), dcorr, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matCorrRk(1, 1), 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matCorrRk(1, 0), dcorr, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matCorrRk(0, 0), 1.0, DBL_PRECISION_COMPARE);

}
void TSamplingTest::throwExceptionTSamplingsetUserCorrelationSameIndex()
{
    CPPUNIT_ASSERT_THROW(_sampler->setUserCorrelation(1, 1, 0.95),
            URANIE::Exceptions::UErrorExceptions);
}
void TSamplingTest::throwExceptionTSamplingsetUserCorrelationIndexINegative()
{
    CPPUNIT_ASSERT_THROW(_sampler->setUserCorrelation(-1, 1, 0.95),
            URANIE::Exceptions::UErrorExceptions);
}
void TSamplingTest::throwExceptionTSamplingsetUserCorrelationIndexJNegative()
{
    CPPUNIT_ASSERT_THROW(_sampler->setUserCorrelation(1, -1, 0.95),
            URANIE::Exceptions::UErrorExceptions);
}
void TSamplingTest::throwExceptionTSamplingsetUserCorrelationNameValueFalse()
{
    CPPUNIT_ASSERT_THROW(
            _sampler->setUserCorrelation("Uniform", "Normal", 1.95),
            URANIE::Exceptions::UErrorExceptions);
}
void TSamplingTest::throwExceptionTSamplingsetUserCorrelationIndexValueFalse()
{
    CPPUNIT_ASSERT_THROW(_sampler->setUserCorrelation(1, 1, 1.95),
            URANIE::Exceptions::UErrorExceptions);
}

void TSamplingTest::throwExceptionTSamplingsetUserCorrelationAttributeValueFalse()
{
    CPPUNIT_ASSERT_THROW(
            _sampler->setUserCorrelation(_tds1->getAttribute("Uniform"),
                    _tds1->getAttribute("Normal"), 1.95),
            URANIE::Exceptions::UErrorExceptions);
}
void TSamplingTest::throwExceptionTSamplingsetUserCorrelationIndexISup()
{
    CPPUNIT_ASSERT_THROW(_sampler->setUserCorrelation(8, 1, 0.95),
            URANIE::Exceptions::UErrorExceptions);
}
void TSamplingTest::throwExceptionTSamplingsetUserCorrelationIndexJSup()
{
    CPPUNIT_ASSERT_THROW(_sampler->setUserCorrelation(1, 8, 0.95),
            URANIE::Exceptions::UErrorExceptions);
}
void TSamplingTest::throwExceptionTSamplingsetUserCorrelationFalseNameI()
{
    CPPUNIT_ASSERT_THROW(_sampler->setUserCorrelation("toto", "Uniform", 0.95),
            URANIE::Exceptions::UErrorExceptions);
}
void TSamplingTest::throwExceptionTSamplingsetUserCorrelationFalseNameJ()
{
    CPPUNIT_ASSERT_THROW(_sampler->setUserCorrelation("Uniform", "toto", 0.95),
            URANIE::Exceptions::UErrorExceptions);
}
void TSamplingTest::throwExceptionTSamplingsetUserCorrelationFalseAttributI()
{
    URANIE::DataServer::TAttribute * att = new URANIE::DataServer::TAttribute(
            "att");

    CPPUNIT_ASSERT_THROW(
            _sampler->setUserCorrelation(att, _tds1->getAttribute("Uniform"),
                    0.95), URANIE::Exceptions::UErrorExceptions);

    delete att;
    att = NULL;
}
void TSamplingTest::throwExceptionTSamplingsetUserCorrelationFalseAttributJ()
{
    URANIE::DataServer::TAttribute * att = new URANIE::DataServer::TAttribute(
            "att");

    CPPUNIT_ASSERT_THROW(
            _sampler->setUserCorrelation(_tds1->getAttribute("Uniform"), att,
                    0.95), URANIE::Exceptions::UErrorExceptions);

    delete att;
    att = NULL;
}
void TSamplingTest::testTSamplingsetSaveRang()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_NODEVELOPPED_TEST()
    ;
}
void TSamplingTest::testTSamplingsetTypeSampling()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_NODEVELOPPED_TEST()
    ;

}

void TSamplingTest::throwExceptionsetCorrelationMatrix()
{

    delete _sampler;
    _sampler = new URANIE::Sampler::TSampling(_tds1, "lhs", 10);

    TMatrixD corrMat5(5, 5);
    corrMat5.UnitMatrix();
    CPPUNIT_ASSERT_THROW(_sampler->setCorrelationMatrix(corrMat5), URANIE::Exceptions::UErrorExceptions);

    TMatrixD corrMat3(7, 7);
    corrMat3.UnitMatrix();
    CPPUNIT_ASSERT_THROW(_sampler->setCorrelationMatrix(corrMat3), URANIE::Exceptions::UErrorExceptions);

    TMatrixD corrMat4(6, 6);
    corrMat4.UnitMatrix(); corrMat4(1,2)=0.3;
    CPPUNIT_ASSERT_THROW(_sampler->setCorrelationMatrix(corrMat4), URANIE::Exceptions::UErrorExceptions);

    corrMat4(1,2)=1.3; corrMat4(2,1)=1.3;
    CPPUNIT_ASSERT_THROW(_sampler->setCorrelationMatrix(corrMat4), URANIE::Exceptions::UErrorExceptions);

    corrMat4(1,2)=-1.3; corrMat4(2,1)=-1.3;
    CPPUNIT_ASSERT_THROW(_sampler->setCorrelationMatrix(corrMat4), URANIE::Exceptions::UErrorExceptions);

    corrMat4(1,2)=-0.3; corrMat4(2,1)=-0.30001;
    CPPUNIT_ASSERT_THROW(_sampler->setCorrelationMatrix(corrMat4), URANIE::Exceptions::UErrorExceptions);

    delete _sampler;
    _sampler = NULL;

}
