//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TFast.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   lun jun 2 10:02:35 CEST 2008
 
 \brief Implementation of the class URANIE::Sensitivity::TFast
 
 */
#include "TMath.h"
#include "TSystem.h"
#include "TVectorD.h"
#include "TROOT.h"
#include "TFFTRealComplex.h"
#include "TLeaf.h"
#include "TGraph.h"

#include "TH1.h"
#include "TPad.h"

// Uranie
#include "TFast.h"
#include "TLauncher.h"
#include "TLauncherFunction.h"

ClassImp (URANIE::Sensitivity::TFast)

namespace URANIE
{
namespace Sensitivity
{
TFast::TFast() :
        TSensitivity(NULL, "", 0, "", 0), _nValFreq(NULL), _nM(6), _nFunction(kSaltelliA)
{
    cout << "construit TFast" << endl;
    _nS = 11;
}

// -----------------------------------------------------------------------
//  Constructors with a function
// -----------------------------------------------------------------------
#ifdef ROOT5
TFast::TFast(URANIE::DataServer::TDataServer *tds, void *fcn, Int_t ns,
        const char *varexpinput, const char *varexpoutput) :
    TSensitivity(tds, fcn, ns, varexpinput, varexpoutput), 
#else
TFast::TFast(URANIE::DataServer::TDataServer *tds, 
        void (*fcn)(Double_t*, Double_t*),
        const char *varexpinput, const char *varexpoutput, Int_t ns) :
    TSensitivity(tds, fcn, varexpinput, varexpoutput, ns), 
#endif
    _nValFreq(NULL), _nM(6), b_plotgraph(false), _nFunction(kSaltelliA)
{
    stdInit();
}

// -----------------------------------------------------------------------
//  Constructors with a function name
// -----------------------------------------------------------------------
TFast::TFast(URANIE::DataServer::TDataServer *tds, const char* fcn, Int_t ns,
	     const char *varexpinput, const char *varexpoutput) :
  TSensitivity(tds, fcn, ns, varexpinput, varexpoutput), _nValFreq(NULL),
  _nM(6), b_plotgraph(false), _nFunction(kSaltelliA)
{
    stdInit();
}

// -----------------------------------------------------------------------
//  Constructors with a TCode
// -----------------------------------------------------------------------
TFast::TFast(URANIE::DataServer::TDataServer *tds,
	     URANIE::Launcher::TCode *code, Int_t ns) :
  TSensitivity(tds, code, ns), _nValFreq(NULL), _nM(6), b_plotgraph(false), _nFunction(kSaltelliA)
{
    stdInit();
}

// -----------------------------------------------------------------------
//  Constructors with a TRun
// -----------------------------------------------------------------------
TFast::TFast(URANIE::DataServer::TDataServer *tds,
        URANIE::Relauncher::TRun *run, Int_t ns) :
    TSensitivity(tds, run, ns), 
    _nValFreq(NULL), _nM(6), b_plotgraph(false), _nFunction(kSaltelliA)
{
    stdInit();
}

// Constructor helpers
void TFast::stdInit()
{
    _sIteratorName = Form("fast__n__iter__%s", _tds->GetName());
    setMethodName("FAST");

  try
    {
      if (_nX > 50)
	throw  URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						    Form("TFast::TFast() ERROR : the number of input (%d) must be smaller than 50",
							 _nX));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
      ue.printMessage();
      throw ue;
    }
}

// -----------------------------------------------------------------------
//  Destructor
// -----------------------------------------------------------------------
TFast::~TFast()
{
    if (_blog)
        cout << " *** destroy TFast(" << GetName() << ") Title[" << GetTitle() << "] *** " << endl;
    if (_nValFreq)
    {
        // Clean frequencies
        delete[] _nValFreq;
        _nValFreq = NULL;
    }
}
  
///////////////////////////////////////////////////////////
//  parseOption 
///////////////////////////////////////////////////////////
void TFast::parseOption(Option_t * option)
{
  
  TSensitivity::parseOption(option);
  
  if (_blog)
    cout << endl << "***********************************" << endl
	 << " ** TFast::parseOption option[" << option << "]" << endl;

  TString soption = TString(option);
  soption.ToLower();
        
  soption.ReplaceAll(",", " ");
  soption += TString(" ");
  if (_blog)
    cout << " ** soption[" << soption << "]" << endl;

  istringstream inStream((char *) soption.Data());
  string inName;
  while (inStream >> inName)
    {
      TString svar(inName.c_str());
      if (_blog)
	cout << " ** svar[" << svar << "]" << endl;

      // Case of the sampler
      if (svar == "graph")
	b_plotgraph=true;
    }
    
  if (_blog)
    cout << " ** End Of TFast::parseOption option[" << option << "]"
	 << endl << "***********************************" << endl;
}

///////////////////////////////////////////////////////////
//  Frequencies 
///////////////////////////////////////////////////////////
void TFast::setFrequencies(Int_t nfreq, Int_t *valfreq)
{
  try
    {
      if (nfreq != _nX)
	throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						   Form("TFast::setFrequencies ERROR : the TDataServer \"%s\" do not have the same number of attributes %d than required %d.",
							GetName(), _nX, nfreq));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
      ue.printMessage();
      throw ue;
    }

  if (_nValFreq == NULL)
    _nValFreq = new Int_t[_nX];

  for (Int_t i = 0; i < _nX; i++)
    {
      _nValFreq[i] = valfreq[i];
    }
}

///////////////////////////////////////////////////////////
//  generateSample 
///////////////////////////////////////////////////////////
void TFast::generateSample(Option_t * option)
{

    if (_blog)
    {
        cout << " ***********************************" << endl;
        cout << " ** TFast::generateSample[" << option << "] Begin _nS[" << _nS
                << "] _nSeed[" << _nSeed << "]" << endl;
    }

    // Create the tuple if necessary
    try
    {
        _tds->getTuple();
    } catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        _tds->createTuple();
        ue.printMessage();
    }
    if (_blog)
        _tds->getTuple()->Print();


    URANIE::Sampler::TSpaceFilling::EFunction theFunction= URANIE::Sampler::TSpaceFilling::kCukier;
    
    switch (_nFunction)
    {
    case kCukier: theFunction = URANIE::Sampler::TSpaceFilling::kCukier; break;
    case kSaltelliA: theFunction = URANIE::Sampler::TSpaceFilling::kSaltelliA; break;
    case kSaltelliB: theFunction = URANIE::Sampler::TSpaceFilling::kSaltelliB; break;
    default:
      cout << "Inconnu";
    }
        
    URANIE::Sampler::TSpaceFilling *_fsampl =
      new URANIE::Sampler::TSpaceFilling(_tds, "srs", _nS, theFunction);

    if (_nValFreq == NULL)
    {
        _fsampl->setDefaultFrequencies();
        _nValFreq = new Int_t[_nX];
        for (Int_t i = 0; i < _nX; i++)
        {
            _nValFreq[i] = _fsampl->getFrequency(i);
            if (_blog)
                cout << " ** default Frequency i[" << i << "] Freq["
                        << _nValFreq[i] << "]" << endl;
        }
    }
    else
    {
        _fsampl->setFrequencies(_nX, _nValFreq);
    }

    // Look up the max frequencies
    Int_t nMaxFreq = _nValFreq[0];
    for (Int_t i = 1; i < _nX; i++)
    {
        if (_blog)
            cout << " ** " << i << " - crt[" << _nValFreq[i] << "] nMaxFreq["
                    << nMaxFreq << "]" << endl;
        if (_nValFreq[i] > nMaxFreq)
            nMaxFreq = _nValFreq[i];
    }
    if (_blog)
        cout << " ** TFast::generateSample end nMaxFreq[" << nMaxFreq << "]"
                << endl;
    Int_t nnSDefault = 2 * _nM * nMaxFreq + 1;
    if (_blog)
        cout << " ** TFast::generateSample nMaxSize[" << nnSDefault << "] _nS[ "
                << _nS << "]" << endl;
    if (_nS < nnSDefault)
    {
        _fsampl->setSize(nnSDefault);
    }

    TString smethod = this->getMethodName();
    smethod.ToLower();
    if (smethod.Contains("rbd"))
        _fsampl->setPermutation(kTRUE);
//     cout << " ** METHOD[" << this->getMethodName()<< "] smethod[" << smethod << "]" << endl;

    _fsampl->generateSample();
    _nS = _tds->getNPatterns();
    _tds->addAttribute(_sIteratorName, Form("%d", nMaxFreq));

    delete _fsampl;
    _fsampl=NULL;
    
    if (_blog)
    {
        cout << " ** End of generateSample _nS[" << _nS << "]" << endl;
        cout << "***********************************" << endl;
    }
}

///////////////////////////////////////////////////////////
//  evaluateIndexes 
///////////////////////////////////////////////////////////
void TFast::evaluateIndexes(Option_t * option)
{

    if (_blog)
    {
        cout << "***********************************" << endl;
        cout << " ** TFast::evaluateIndexes[" << option << "]" << endl;
    }


    TFFTRealComplex *Uranie__Fast__FFTW = new TFFTRealComplex(_nS, kFALSE);
    Double_t *valY = new Double_t[_nS];
    URANIE::DataServer::TAttribute *atty = (URANIE::DataServer::TAttribute *) (_listOfOutputAttributes->At(_iy));
    TString soutputy = atty->GetName();
    TString iter_att = _tds->getIteratorName();

    _tds->computeStatistic(iter_att);
    int shift = ((_tds->getAttribute(iter_att)->getMinimum() == 0) ? 0 : 1);
    
    
    if (_blog)
      cout << " ** Case of Output atty [" << _sOutputAttribute << "]" << endl;

    // Loop on the TDS for the current leaf : leafY
    for (Int_t ipat = 0; ipat < _nS; ipat++)
      {          
          // get the current pattern         
          valY[int(_tds->getValue(iter_att, ipat))-shift] = _tds->getValue(soutputy, ipat, _iely);        
      }
    
    if(_blog)
      for(Int_t ipat = 0; ipat < _nS; ipat++) 
        cout << "i[" << ipat << "/"<<_nS<<"] val[" << valY[ipat] <<"]"<<endl;

    Double_t ddelta = 1.0 / (Double_t)(_nS + 1);
    if (_blog)
      cout << "ddelta[" << ddelta << "]" << endl;
    Double_t ds;

    TMatrixD matA(_nX, _nM);
    matA.Zero();
    TMatrixD matB(_nX, _nM);
    matB.Zero();

    // Loop on the patterns
    for (Int_t is = 0; is < _nS; is++)
      {
	ds = -1.0 * TMath::Pi()
	  + 2.0 * TMath::Pi() * ddelta * ((Double_t)(is) + 0.5);
	//           cout << "ds[" << ds << "]"<<endl;
	// Loop on the input attributes
	for (Int_t ix = 0; ix < _nX; ix++)
	  {
	    //         Double_t domega=_nValFreq[ix];

	    //         cout << "ix[" << ix+1<<"] omega[" << domega << "]"<<endl;
	    for (Int_t im = 0; im < _nM; im++)
	      {
		matA(ix, im) += valY[is]
		  * TMath::Cos(
			       (Double_t)(im + 1)
			       * (Double_t)(_nValFreq[ix]) * ds)
		  / 2.0 * TMath::Pi();
		matB(ix, im) += valY[is]
		  * TMath::Sin(
			       (Double_t)(im + 1)
			       * (Double_t)(_nValFreq[ix]) * ds)
		  / 2.0 * TMath::Pi();
	      }
	    //           if ( is > (Int_t)( 0.99 * _nS) ) { cout << endl << " ****** MatA is["<<is <<"]" << endl;  matA.Print();}
	    //           if ( is > (Int_t)( 0.99 * _nS) ) { cout << " **** MatB" << endl;  matB.Print(); cout << " ****** End of is["<<is <<"]" << endl;}
	  }
	//         if ( is > (Int_t)( 0.99 * _nS) ) { cout << endl << " ****** MatA is["<<is <<"]" << endl;  matA.Print();}
	//         if ( is > (Int_t)( 0.99 * _nS) ) { cout << " **** MatB" << endl;  matB.Print(); cout << " ****** End of is["<<is <<"]" << endl;}
      }

    if (_blog)
      {
	matA.Print();
	matB.Print();
      }

    TVectorD indFast(_nX);
    indFast.Zero();

    Double_t ds1;
    // Loop on the input attributes
    for (Int_t ix = 0; ix < _nX; ix++)
      {
	//         Double_t domega=_nValFreq[ix];
	ds1 = 0;
	if (_blog)
	  cout << "ix[" << ix + 1 << "] ds1[" << ds1 << "]" << endl;
	for (Int_t im = 0; im < _nM; im++)
	  {
	    if (_blog)
	      cout << " ** ix[" << ix + 1 << "]  im[" << im << "] ds1["
		   << ds1 << "] matA[" << matA(ix, im) << "] matB["
		   << matB(ix, im) << "]" << endl;
	    ds1 += 2.0
	      * (matA(ix, im) * matA(ix, im)
		 + matB(ix, im) * matB(ix, im))
	      / (Double_t)(_nS * _nS);
	    if (_blog)
	      cout << " ** ix[" << ix + 1 << "]  im[" << im << "] ds1["
		   << ds1 << "]" << endl;
	  }
	indFast(ix) = ds1;
	if (_blog)
	  cout << "ix[" << ix + 1 << "] ds1[" << ds1 << "]" << endl;
      }

    if (_blog)
      {
	cout << " ** Presque de D1" << endl;
	indFast.Print();
      }

    Int_t nKind = 2;
    Uranie__Fast__FFTW->Init("ES", 1, &nKind);
    
    Uranie__Fast__FFTW->SetPoints(valY);
    Uranie__Fast__FFTW->Transform();

    if (_blog)
      cout << " *** FFTW[" << Uranie__Fast__FFTW->GetDefaultFFT() << "]"
	   << endl;

    Double_t dfast = 0.0;
    Int_t nhalfn = (Int_t)(_nS / 2);
    
    TGraph *gr = new TGraph(nhalfn);
    
    Double_t dre, dim;
    for (Int_t i = 1; i < nhalfn; i++)
      {
	Uranie__Fast__FFTW->GetPointComplex(i, dre, dim);
	dfast += 2. * (dre * dre + dim * dim) / (Double_t)(_nS * _nS);
	if (_blog)
	  cout << " ** FFTW  vary  i[" << i << "/" << nhalfn << "] dre["
	       << dre << "] dim[" << dim << "] dfast[" << dfast << "]"
	       << endl;
	gr->SetPoint(i - 1, 1.0 * i,
		     (dre * dre + dim * dim) / (Double_t)(_nS * _nS));
      }
    if (_blog)
      cout << " ** dfast[" << dfast << "]" << endl;

    if(b_plotgraph)
    {
	checkCanvasCreation(true);
	gr->Draw("alp");
    }
    
    TVectorD indFastFFTW(_nX);
    indFastFFTW.Zero();

    ds1 = 0;
    sprintf(_sMethod, "%s", _sMethodName.Data());
    // Loop on the input attributes
    for (Int_t ix = 0; ix < _nX; ix++)
      {
	indFast(ix) /= dfast;
	indFastFFTW(ix) = 0.0;
	for (Int_t im = 0; im < _nM; im++)
	  {
	    Int_t nfre = (im + 1) * _nValFreq[ix];
	    Uranie__Fast__FFTW->GetPointComplex(nfre, dre, dim);
	    indFastFFTW(ix) += 2. * (dre * dre + dim * dim)
	      / (Double_t)(_nS * _nS);
	    if (_blog)
	      cout << " ** FFTW  im[" << im << "] freq[" << nfre
		   << "] dre[" << dre << "] dim[" << dim
		   << "] indFastFFTW[" << indFastFFTW(ix) << "]"
		   << endl;
	  }
	indFastFFTW(ix) /= dfast;

	URANIE::DataServer::TAttribute *att =
	  (URANIE::DataServer::TAttribute *) (_listOfInputAttributes->At(
									 ix));
	
	fillIndex(att->GetName(), "First", indFastFFTW(ix),"--first--");
	fillIndex(att->GetName(), "Total", indFastFFTW(ix),"--total--");
      }
    
    if (_blog)
      {
	_ntresult->Scan("*");
	cout << " ** First Order Sensitivity Index by FAST" << endl;
	indFast.Print();
	indFastFFTW.Print();
      }


  // Clean
  delete[] valY;
  valY = NULL;
  if(!b_plotgraph)
  {
      delete gr;
      gr=NULL;
  }
  else
      _drawingGarbageCollector->Add(gr);
  delete Uranie__Fast__FFTW;
  Uranie__Fast__FFTW = NULL;
  
  //drawIndexes("FAST Indice", "", option);
  
  if (_blog)
    {
      cout << " ** End of TFast::evaluateIndexes" << endl;
      cout << "***********************************" << endl;
    }

}
///////////////////////////////////////////////////////////
//  Log 
///////////////////////////////////////////////////////////
void TFast::printLog(Option_t *option)
{
    cout << "************** TFast::printLog[" << option << "] Method["
            << getMethodName() << "]" << endl;
    cout << "**  _sMethodName[" << _sMethodName << "]" << endl;
    cout << "**  _sFunctionName[" << _sFunctionName << "]" << endl;
    cout << "**  _nX[" << _nX << "]" << endl;
    cout << "**  _nY[" << _nY << "]" << endl;
    if (_nValFreq)
    {
        cout << " ** Frequencies[";
        for (Int_t i = 0; i < _nX; i++)
        {
            cout << _nValFreq[i];
            if (i < _nX - 1)
                cout << ", ";
        }
        cout << "]" << endl;
    }
    else
        cout << "** Frequencies = NULL" << endl;

    cout << " ** Function[";
    switch (_nFunction)
    {
    case kCukier:
    {
        cout << "Cukier(73)";
    }
        break;
    case kSaltelliA:
    {
        cout << "Saltelli(99) sans Phi";
    }
        break;
    case kSaltelliB:
    {
        cout << "Saltelli(99) avec Phi";
    }
        break;
    default:
        cout << "Inconnu";
    }
    cout << "]" << endl;
    TSensitivity::printLog();
    cout << "************** End of TFast::printLog" << endl;
}

} // Fin du namespace Sensitivity
} // Fin du namespace URANIE

