//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TFiniteDifferences.h
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 \date   wen sep 13 12:39:50 CEST 2006
 
 \brief This class provides sensitivity measures based on finite difference
 
 */
/**
 
 @class URANIE::Sensitivity::TFiniteDifferences

 @brief Description of the class TFiniteDifferences.

 The finite differences method of sensitivity analysis is among the simplest ones. 
 The resulting sensitivity index of an input variable \f$ X \f$ with respect to an output 
 \f$ Y \f$ is an estimation of \f$ \frac{\partial Y}{\partial X} \f$. 

 In this implementation of the method, the estimation is obtained by applying an 
 OAT design of experiments (cf. Sampler::TOATSampling) to the studied model. 
 Thus, for each input \f$ X \f$ varying of \f$ \pm\delta x \f$, one can compute:

 \li \f$ \frac{\partial Y}{\partial X} \approx \frac{Y(X+\delta x) - Y(X-\delta x)}{2*\delta x} \f$

 To use the class TFiniteDifferences, one needs to give it an empty TDataServer 
 containing a list of input attributes with default values, and a model. The user 
 will then call the \b computeIndexes() \b method which will:

 \li generate the OAT sampling,
 \li call the model on it, 
 \li compute the sensitivity indexes, 
 \li store them in a sensitivity matrix. 

 This matrix can be retrieved using the method \b getSensitivityMatrix() \b .

 (Nicolas Gilardi 2011.06.17)
 */
#ifndef TFiniteDifferences_H
#define TFiniteDifferences_H

// Uranie
#include "TStochAttribut.h"
#include "TSensitivity.h"

using std::cout;
using std::endl;
using namespace URANIE::DataServer;

namespace URANIE
{
namespace Sensitivity
{

//void SensitivityInteractiveMethod(Double_t *param, Double_t *out);      

class TFiniteDifferences: public TSensitivity
{

protected:

    TMatrixD _sMatrix;                               ///< The sensitivity matrix
    TString _sSensitiveAtt;                ///< The list of sensitive attributes
    TString _sSamplingOption;              ///< the options for the OAT sampling
    Int_t _nbIn;                             ///< the number of sensitive inputs
    Int_t _nbAux;                        ///< the number of non sensitive inputs
protected:

public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    TFiniteDifferences();

    //! Constructor with the TDataServer and function
    /*!      
     This constructor is used when the model is a function.
     
     \param tdsNominal (URANIE::DataServer::TDataServer *) an empty dataserver containing the input variables. Each 
     attribute representing an input variables must have a default 
     value set. It will be used as the nominal value.
     \param fcn (void *) The function to analyse
     \param outputAtt (TString)   this is a list of attribute names separated by colons. They represent all the responses of 
     the problem. They must refer to existing attributes in the "tdsNominal".          
     \param sensitiveAtt (TString) A list of attribute names separated by colons. These attributes are the one that will 
     be taken into account for the sensitivity analysis. If the string is equal to "" or "*", 
     all the attributes are taken into account (Default = "").
     \param samplingOption (TString) The options for the OAT sampling (Default = "steps=1%")
     */
#ifdef ROOT5
    TFiniteDifferences(TDataServer *tdsNominal, void *fcn, TString outputAtt,
            TString sensitiveAtt = "", TString samplingOption = "steps=1%");
#else
    TFiniteDifferences(TDataServer *tdsNominal, void (*fcn)(Double_t*,Double_t*),
            TString sensitiveAtt, TString outputAtt,
            TString samplingOption = "steps=1%");
#endif

       //! Constructor with the TDataServer and function name
    /*!      
     This constructor is used when the model is a function.
     
     \param tdsNominal (URANIE::DataServer::TDataServer *) an empty dataserver containing the input variables. Each 
     attribute representing an input variables must have a default value set. It will be used as the nominal value.
     \param const char *fcn The function to analyse
     \param outputAtt (TString)   this is a list of attribute names separated by colons. They represent all the responses of 
     the problem. They must refer to existing attributes in the "tdsNominal".          
     \param sensitiveAtt (TString) A list of attribute names separated by colons. These attributes are the one that will 
     be taken into account for the sensitivity analysis. If the string is equal to "" or "*", 
     all the attributes are taken into account (Default = "").
     \param samplingOption (TString) The options for the OAT sampling (Default = "steps=1%")
     */
    TFiniteDifferences(TDataServer *tdsNominal, const char* fcn,
                       TString sensitiveAtt, TString outputAtt,
                       TString samplingOption );

 
    //! Constructor with the TDataServer and code
    /*!       
     This constructor is used when the model is an external code.
     
     \param tdsNominal (URANIE::DataServer::TDataServer *) an empty dataserver containing the input variables. Each 
     attribute representing an input variables must have a default 
     value set. It will be used as the nominal value.
     \param fcode (URANIE::Launcher::TCode *) The code to analyse
     \param sensitiveAtt (TString) A list of attribute names separated by colons. These attributes are the one that will 
     be taken into account for the sensitivity analysis. If the string is equal to "" or "*", 
     all the attributes are taken into account (Default = "").
     \param samplingOption (TString) The options for the OAT sampling (Default = "steps=1%")
     */
    TFiniteDifferences(TDataServer *tdsNominal, URANIE::Launcher::TCode *fcode,
            TString sensitiveAtt = "", TString samplingOption = "steps=1%");

    //! Constructor with the TDataServer and TRun
    /*!       
     This constructor is used when the model is a TRun/TEval model
     
     \param tdsNominal (URANIE::DataServer::TDataServer *) an empty dataserver containing the input variables. Each 
     attribute representing an input variables must have a default 
     value set. It will be used as the nominal value.
     \param fcode (URANIE::Relauncher::TRun *) The TRun to use for analyse
     \param sensitiveAtt (TString) A list of attribute names separated by colons. These attributes are the one that will 
     be taken into account for the sensitivity analysis. If the string is equal to "" or "*", 
     all the attributes are taken into account (Default = "").
     \param samplingOption (TString) The options for the OAT sampling (Default = "steps=1%")
     */
    TFiniteDifferences(TDataServer *tdsNominal, URANIE::Relauncher::TRun *run,
            TString sensitiveAtt = "", TString samplingOption = "steps=1%");

    //! Constructor with a loaded data server
    /*!       
     This constructor is used when the user already have a dataset.
     
     \param tdsNominal (URANIE::DataServer::TDataServer *)  this TDS contains the input and output values necessary for the finite 
     differences calculation. It must also contain the attributes 
     "__nominal_set__" and "__modified_att__" which are necessary to know 
     which attribute is modified.
     \param inputAtt  (TString)   this is a list of attribute names separated by colons. They represent all the inputs of the 
     problem. They must refer to existing attributes in the "tdsNominal".
     \param outputAtt (TString)   this is a list of attribute names separated by colons. They represent all the responses of 
     the problem. They must refer to existing attributes in the "tdsNominal".
     \param sensitiveAtt (TString) A list of attribute names separated by colons. These attributes are the one that will 
     be taken into account for the sensitivity analysis. If the string is equal to "" or "*", 
     all the input attributes are taken into account (Default = "").                                       
     \warning When using this constructor with CINT, do not forget to EXPLICITELY pass TString objects for 
     "listOfInputs" and "listOfOutputs" parameters. Otherwise, you may end up calling another constructor !                                            
     */
    TFiniteDifferences(TDataServer *tdsNominal, TString inputAtt,
            TString outputAtt, TString sensitiveAtt = TString(""));

    //! Default destructor
    virtual ~TFiniteDifferences();
    //@}

    //---------------------------------------------
    /** @name Setting and Getting attributs
     **/
    //@{         
    //! Returns the sensitivity matrix
    TMatrixD getSensitivityMatrix()
    {
        return _sMatrix;
    }

    //! Returns a pointer to the internal data server
    TDataServer* getDataServer()
    {
        return _tds;
    }
    //@}

    //---------------------------------------------
    /** @name Generation of the sample
     **/
    //@{
    //! Generates the OAT sampling
    /*!
     This function fills a TDS with appropriate variables, creates and uses a
     TOATSampling to generate the design of experiment.

     \param option (Option_t *) just in case we need option in the future...
     (Default = "").
     */
    virtual void generateSample(Option_t * option = "");

    //@}

    //---------------------------------------------
    /** @name Compute the indexes
     **/
    //@{
    //! Compute the sensitivity matrix
    /*! 
     This function construct the sensitivity matrix from the OAT design 
     of experiment. The resulting matrix can be retrieved using the 
     getSensitivityMatrix() function.           
     
     \param option (Option_t *) just in case we need option in the future...
     (Default = "").
     */
//       void computeIndexes(Option_t * option = "");
    void evaluateIndexes(Option_t * option = "");

    //! Resize the matrix results at the very beginning of the algo to the number of output/input
    /*!
      One should wait this moment to know the number of output as it depends on the presence of 
      vectors as well as the newoption "output=" implemented in TSensitivity::computeIndexes()
     */
    void preTreatment();
    //@}
    
    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog();
    //@}
    ClassDef(URANIE::Sensitivity::TFiniteDifferences, ID_SENSITIVITY)
    //Definition of an variable

};

} // Fin du namespace Sensitivity
} // Fin du namespace URANIE

#endif
