//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TJohnsonRW.cxx
 \author Jean-Baptiste Blanchard <jean-baptiste.blanchard@cea.fr>
 \date   jeu jun 30 10:35:25 CEST 2020
 
 \brief Implementation of the class URANIE::Sensitivity::TJohnsonRW

  To be done :
  ==> Add b0 coefficient possible for regressor (consistent with the rest so "nointercept" option i guess)
  ==> Unitary tests
  ==> Documentation
  ==> User macros

 */

//// ROOT
#include "TFile.h"
#include "TLeaf.h"
#include "TPad.h"
#include "TCanvas.h"
#include "TDecompSVD.h"

//// URANIE
//#include "TTemporalTree.h"
#include "TJohnsonRW.h"
#include "TDSNtupleD.h"
#include "TSampling.h"

using std::cout;
using std::endl;

ClassImp (URANIE::Sensitivity::TJohnsonRW)

namespace URANIE
{
namespace Sensitivity
{
// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------
TJohnsonRW::TJohnsonRW()
{
    cout << "construit TJohnsonRW" << endl;
}

TJohnsonRW::TJohnsonRW(URANIE::DataServer::TDataServer *tds, const char *varexpinput, const char *varexpoutput, Option_t * option) :
    TSensitivity(tds, varexpinput, varexpoutput, option), _dr2(0.0), _dr2A(0.0),
    _busingCorr(true), _bmatCorrSet(false), _busingSample(true), _sSampMethod("lhs"), _localYCounter(0)
{

    if (_blog)
    {
        cout << "construit TJohnsonRW tds[" << _tds->GetName() << "] Option ["
             << option << "]" << endl;
        cout << "varexpinput[" << varexpinput << "]" << endl;
        cout << "varexpoutput[" << varexpoutput << "]" << endl;
    }

    _sIteratorName = Form("johnsonrw__n__iter__%s", _tds->GetName());
    setMethodName("JohnsonRW");
    parseOption(option);
    
    if (_blog)
    {
        cout << "_sSelectedInput[" << _sSelectedInput << "]" << endl;
        cout << "_sSelectedOutput[" << _sSelectedOutput << "]" << endl;
    }

}

TJohnsonRW::TJohnsonRW(URANIE::DataServer::TDataServer *tds, const char *fcn, Int_t ns, const char *varexpinput, const char *varexpoutput, Option_t * option):
    TSensitivity(tds, fcn, ns, varexpinput, varexpoutput, option), _dr2(0.0), _dr2A(0.0),
    _busingCorr(true), _bmatCorrSet(false), _busingSample(false), _sSampMethod("lhs"), _localYCounter(0)
{

    if (_blog)
    {
        cout << "construit TJohnsonRW tds[" << _tds->GetName() << "] Option ["
             << option << "]" << endl;
        cout << "varexpinput[" << varexpinput << "]" << endl;
        cout << "varexpoutput[" << varexpoutput << "]" << endl;
    }

    _sIteratorName = Form("johnsonrw__n__iter__%s", _tds->GetName());
    setMethodName("JohnsonRW");
    parseOption(option);
    
    if (_blog)
    {
        cout << "_sSelectedInput[" << _sSelectedInput << "]" << endl;
        cout << "_sSelectedOutput[" << _sSelectedOutput << "]" << endl;
    }

}

TJohnsonRW::TJohnsonRW(URANIE::DataServer::TDataServer *tds, void (*fcn)(Double_t*,Double_t*),
                       const char *varexpinput, const char *varexpoutput, Int_t ns, Option_t * option):
    TSensitivity(tds, fcn, varexpinput, varexpoutput, ns, option), _dr2(0.0), _dr2A(0.0),
    _busingCorr(true), _bmatCorrSet(false), _busingSample(false), _sSampMethod("lhs"), _localYCounter(0)
{

    if (_blog)
    {
        cout << "construit TJohnsonRW tds[" << _tds->GetName() << "] Option ["
             << option << "]" << endl;
        cout << "varexpinput[" << varexpinput << "]" << endl;
        cout << "varexpoutput[" << varexpoutput << "]" << endl;
    }

    _sIteratorName = Form("johnsonrw__n__iter__%s", _tds->GetName());
    setMethodName("JohnsonRW");
    parseOption(option);
    
    if (_blog)
    {
        cout << "_sSelectedInput[" << _sSelectedInput << "]" << endl;
        cout << "_sSelectedOutput[" << _sSelectedOutput << "]" << endl;
    }

}

TJohnsonRW::TJohnsonRW(URANIE::DataServer::TDataServer *tds, URANIE::Launcher::TCode *code, Int_t ns, Option_t * option):
    TSensitivity(tds, code, ns, option), _dr2(0.0), _dr2A(0.0),
    _busingCorr(true), _bmatCorrSet(false), _busingSample(false), _sSampMethod("lhs"), _localYCounter(0)
{

    if (_blog)
    {
        cout << "construit TJohnsonRW tds[" << _tds->GetName() << "] Option ["
             << option << "]" << endl;
    }

    _sIteratorName = Form("johnsonrw__n__iter__%s", _tds->GetName());
    setMethodName("JohnsonRW");
    parseOption(option);
    
    if (_blog)
    {
        cout << "_sSelectedInput[" << _sSelectedInput << "]" << endl;
        cout << "_sSelectedOutput[" << _sSelectedOutput << "]" << endl;
    }

}

TJohnsonRW::TJohnsonRW(URANIE::DataServer::TDataServer *tds, URANIE::Relauncher::TRun *run, Int_t ns, Option_t * option):
    TSensitivity(tds, run, ns, option), _dr2(0.0), _dr2A(0.0),
    _busingCorr(true), _bmatCorrSet(false), _busingSample(false), _sSampMethod("lhs"), _localYCounter(0)
{

    if (_blog)
    {
        cout << "construit TJohnsonRW tds[" << _tds->GetName() << "] Option ["
             << option << "]" << endl;
    }

    _sIteratorName = Form("johnsonrw__n__iter__%s", _tds->GetName());
    setMethodName("JohnsonRW");
    parseOption(option);

    if (_blog)
    {
        cout << "_sSelectedInput[" << _sSelectedInput << "]" << endl;
        cout << "_sSelectedOutput[" << _sSelectedOutput << "]" << endl;
    }

}

// -----------------------------------------------------------------------
//  Destructor
// -----------------------------------------------------------------------

TJohnsonRW::~TJohnsonRW()
{
    //    cout << "detruit TJohnsonRW" << endl;
}

void TJohnsonRW::parseOption(Option_t *option)
{

    //Check all pssible
    TString opt = option;
    opt.ToLower();
    if (opt.Contains("reg"))
    {
	_busingCorr=false;
        setMethodName("Reg_JohnsonRW");
    }
    if (opt.Contains("lhs"))
    {
        _sSampMethod="lhs";
    }
    if (opt.Contains("srs"))
    {
	_sSampMethod="srs";
    }

}

void TJohnsonRW::setCorrelationMatrix(TMatrixD corrMat)
{

    try
    {
	int totdim=_nX+_nY;
        if ((corrMat.GetNcols() != totdim)
            || (corrMat.GetNrows() != corrMat.GetNcols()))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form(
                                                           "TJohnsonRW::setCorrelationMatrix: invalid correlation matrix: it must be a symetrical square matrix of size %d x %d.\n\
 The one provided is %d x %d.",_nX+_nY,_nX+_nY,corrMat.GetNrows(),corrMat.GetNcols()));

        _mmatCorr.ResizeTo(corrMat);
        _mmatCorr = corrMat;
        
        for(int icol=0; icol<corrMat.GetNcols(); icol++)
	{
	    _mmatCorr(icol,icol)=1.0;
	    for(int ilig=0; ilig<corrMat.GetNrows(); ilig++)
	    {
		if( fabs(_mmatCorr(ilig,icol) -_mmatCorr(icol,ilig)) > 1e-14 )
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TJohnsonRW::setCorrelationMatrix: invalid correlation matrix: it must be a symetrical square matrix.\n\
This is not the case as M(%d,%d) != M(%d,%d). Their difference is equal to %g",ilig,icol,icol,ilig,_mmatCorr(ilig,icol) -_mmatCorr(icol,ilig)));
	    }
	}
        
        Double_t max = TMath::MaxElement(_mmatCorr.GetNoElements(),
                                         _mmatCorr.GetMatrixArray());
        Double_t min = TMath::MinElement(_mmatCorr.GetNoElements(),
                                         _mmatCorr.GetMatrixArray());
        if (max > 1.0 || min < -1.0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TJohnsonRW::setCorrelationMatrix: correlation values must belong to the interval [-1.0, 1.0] (min and max found are [%g,%g])",
                                                            min,max));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    _bmatCorrSet=true;

}

///////////////////////////////////////////////////////////
//  Generate the Sample 
///////////////////////////////////////////////////////////
void TJohnsonRW::generateSample(Option_t * option)
{
    
    TString soption = option;
    soption.ToLower();
    
    if (_blog)
    {
        cout << endl;
        cout << "*********************************" << endl;
        cout << " ** TJohnsonRW::generateSample" << endl;
        cout << " ** _nSeed[" << _nSeed << "]" << endl;
    }

    if( _bmatCorrSet==false && _busingSample==false )
    {
	Int_t npat;

	// Create the tuple if necessary
	try
	{
	    _tds->getTuple();
	} catch (URANIE::Exceptions::UWarningExceptions&)
	{
	    _tds->createTuple();
	}
    
	if (_blog)
	    _tds->getTuple()->Print();

	URANIE::Sampler::TSampling *sam = new URANIE::Sampler::TSampling(_tds, _sSampMethod.Data(), _nS);
	if( _nSeed != 0)
	    sam->setSeed( _nSeed);

        if(_bisInputCorrelated)
	    sam->setCorrelationMatrix( _minputCorr);

	sam->generateSample();
	delete sam;

	_tds->exportData("_johnsonrw_sampling_.dat", _sSelectedInput);
    }
}

void TJohnsonRW::preTreatment()
{

    //Implement new matrix method from Fabrice
    // Initialise the correlation matrix with the correct size
    if(_blog)
    {
        cout << endl << "*********************************  TJohnsonRW::preTreatment" << endl;
        cout<< "*** matCorr resize to "<<_nX+_nElY<<" x "<< _nX+_nElY<<endl;
    }

    try
    {
        try
        {
            _tds->getTuple();
        } catch (URANIE::Exceptions::UWarningExceptions&)
        {
            if( _bmatCorrSet == false)
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           "TJohnsonRW::computeIndexes Error\n the sample is empty!! ");
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
        exit(-1);
    }

    if( _bmatCorrSet == false )
    {
        // Resize the correlation matrices
        _mmatCorr.ResizeTo(_nX+_nElY, _nX+_nElY); 
        _mmatCorr = _tds->computeCorrelationMatrix(Form("%s:%s", _sSelectedInput.Data(), _vOutputNames.at(_iy).c_str()));
    }
    
    if(_blog)
    {
        _mmatCorr.Print();
        cout << endl << "*********************************" << endl;
    }


    // Prepare the SVD for all output operations

    if( _busingCorr )
    {

        //Submatrices as well
        TMatrixD Rxx(_nX, _nX);
        _mmatCorr.GetSub(0,_nX-1 ,0,_nX-1,Rxx);
        
        TDecompSVD svd(Rxx);
        bool oui=false;
        oui=svd.Decompose();

        if ( oui==kFALSE)
        {
            if (_blog)
                Rxx.Print();

            try{
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           "TJohnsonRW::evaluateIndexes: the SVD decomposition of the input correlation matrix is not good");
            }catch(URANIE::Exceptions::UErrorExceptions &ue)
            {
                ue.printMessage();
                throw ue;
            }
        }

        // Get the transformation matrix
        TMatrixD mU = svd.GetU();
        TMatrixD mV = svd.GetU();
        mV.T();

        // Get the diagonal elements
        TVectorD S(_nX);
        S = svd.GetSig();
        if(_blog)
            S.Print( );
        
        TMatrixD mS(_nX,_nX);
        for(int ii=0; ii<_nX; ii++)
            mS(ii,ii)= sqrt(S(ii));

        // Compute Lambda
        _mLambda.ResizeTo(_nX,_nX);
        _mLambda = mU * mS * mV;

        double norm= _mLambda.E2Norm();
        _mLambda.Sqr();
        _mLambda *=  1./norm;

        // Invert S for the beta estimations
        double det2=0;
        mS.Invert(&det2);

        _mBetanoY.ResizeTo(_nX,_nX);
        _mBetanoY=mU * mS * mV;

        _mRxxInv.ResizeTo(_nX,_nX);
        _mRxxInv=mU * mS * mS * mV;
        
        
        if(_blog)
        {
            cout<<"**  _mLambda Normalised **"<<endl;
            _mLambda.Print();
        }
    }
    else
    {
        //Get the data
        TMatrixD X(_nS,_nX);
        _tds->getTuple()->extractData(&X,Form("%s", _sSelectedInput.Data()));

        TDecompSVD svd(X);
        bool oui=false;
        oui=svd.Decompose();

        if ( oui==kFALSE)
        {
            if (_blog)
                X.Print();

            try{
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           "TJohnsonRW::evaluateIndexes: the SVD decomposition of the regressors matrix is not good");
            }catch(URANIE::Exceptions::UErrorExceptions &ue)
            {
                ue.printMessage();
                throw ue;
            }
        }

        // Get the transformation matrix
        TMatrixD mU = svd.GetU();
        TMatrixD mV = svd.GetV();
        TMatrixD mVt(_nX,_nX); mVt.Transpose(mV);

        // Get the diagonal elements
        TVectorD S(_nX);
        S = svd.GetSig();
        if(_blog)
            S.Print( );
        
        TMatrixD mS(_nS,_nX);
        TMatrixD mId(_nX,_nS);
        for(int ii=0; ii<_nX; ii++)
        {
            mS(ii,ii)= S(ii);
            mId(ii,ii)=1.;
        }

        // Compute Lambda
        _mLambda.ResizeTo(_nX,_nX);
        _mLambda = mV * mId * mS * mVt;
        
        double norm= _mLambda.E2Norm();
        _mLambda.Sqr();
        _mLambda *=  1./norm;
        
        // Transpose U for beta estimation
        mU.T();
        _mBetanoY.ResizeTo(_nX,_nS);
        _mBetanoY= mV * mId * mU;

        _nS = _tds->getNPatterns();

        if(_blog)
        {
            cout<<"**  _mLambda Normalised **"<<endl;
            _mLambda.Print();
        }
    }
}
  
///////////////////////////////////////////////////////////
//  evaluateIndexes 
///////////////////////////////////////////////////////////
void TJohnsonRW::evaluateIndexes(Option_t * option)
{
    TString soption = option;
    soption.ToLower();
    if (_blog)
    {
        cout << endl << "*********************************" << endl;
        cout << " ** TJohnsonRW::evaluateIndexes" << endl;
        cout << " **    _sSelectedInput[" << _sSelectedInput << "]" << endl;
        cout << " **   _sSelectedOutput[" << _sOutputAttribute << "]" << endl;
    }

    if(soption.Contains("restart"))
        _localYCounter = 0;

    string outputName=string(_sOutputAttribute);
    string key=outputName.substr(0, outputName.find("["));
    int val = (( _mAttributeElements.find( key )!=_mAttributeElements.end() ) ? _iely : _localYCounter);
    
    TMatrixD Res(_nX,1);
    if( _busingCorr)
    {

        // ===============================================================
        // =============== Correlation-based method ======================
        // ===============================================================
	
        TMatrixD Rxy = _mmatCorr.GetSub(0,_nX-1,_nX+val,_nX+val);
	// XYZ<<0<<"  "<<_nX-1<<"  "<<_localYCounter<<endl;
	// Rxy.Print();
        // Compute Beta
        TMatrixD Beta = _mBetanoY * Rxy;
        double norm=Beta.E2Norm();
        Beta.Sqr();
        Beta *=  1./norm;
        if(_blog)
        {
            cout<<"**  Beta Normalised **"<<endl;
            Beta.Print();
        }

        // Combine them to get results
        Res=_mLambda*Beta;

        // Compute the R2 (as https://en.wikipedia.org/wiki/Multiple_correlation)
        TMatrixD RxyT(1,_nX); RxyT.Transpose(Rxy);
        if(_blog)
        {
            cout<<"**  Rxy **"<<endl;
            Rxy.Print();
        }

        TMatrixD InvCorrCoeff =  RxyT * _mRxxInv * Rxy;        
        _dr2 = InvCorrCoeff(0,0);
    }
    else
    {
        // ===============================================================
        // ================ Regressor-based method =======================
        // ===============================================================
        
        // Get data for corresponding output
        TMatrixD Y; Y.ResizeTo(_nS,1);
        _tds->getTuple()->extractData(&Y,_vOutputNames.at(_iy).c_str());
        
        // Compute Beta
        TMatrixD Beta = _mBetanoY * Y;
        double norm=Beta.E2Norm();
        Beta.Sqr();
        Beta *=  1./norm;
        if(_blog)
        {
            cout<<"**  Beta Normalised **"<<endl;
            Beta.Print();
        }

        // Combine them to get results
        Res=_mLambda*Beta;

        // Get reduced correlation matrix
        TMatrixD Rxxy(_nX+1, _nX+1);
        _mmatCorr.GetSub(0,_nX,0,_nX, Rxxy);

	XYZ<<_mAttributeElements.size()<<endl;
        if( _iely!=0)
        {
            for(int ival=0; ival<_nX; ival++)
                Rxxy(ival,_nX) = Rxxy(_nX,ival) = _mmatCorr(ival,_nX+_iely);
        }

        // add nugget if determinant is low
        if(Rxxy.Determinant()<1e-15)
        {
            double pepite=1e-10;
            for(int ii=0; ii<_nX+1; ii++)
                Rxxy(ii,ii) += pepite;
        }

        // Invert it to get R2 value
        double det2=0;
        Rxxy.Invert(&det2);
        
        _dr2 =  1.0 - 1.0/ Rxxy(_nX, _nX);
    }

    // Filling informations
    Res *= 1./Res.Sum();
    if(_blog)
    {
        cout<<"**  Result **"<<endl;
        Res.Print();
    }

    // RW values
    double valSobolCrt=0;
    for(Int_t i=0; i< _nX; i++)
    {
        URANIE::DataServer::TAttribute *att = (URANIE::DataServer::TAttribute *) (_listOfInputAttributes->At(i));	
        TString satt = att->GetName();
        //Go throught the method      
        valSobolCrt = Res(i,0);
        fillIndex(satt.Data(), "First", valSobolCrt,"--first--");
        fillIndex(satt.Data(), "Total", valSobolCrt,"--total--");
    }


    // R2 values
    _dr2A = 1.0 - (1.0 - _dr2) * Double_t(_nS) / Double_t(_nS - _nX);
    fillIndex("__R2__", "First", _dr2,"--first--");
    fillIndex("__R2__", "Total", _dr2,"--total--");
    if(_nS!=0)
    {
        fillIndex("__R2A__", "First", _dr2A,"--first--");
        fillIndex("__R2A__", "Total", _dr2A,"--total--");
    }

    _localYCounter++;
}

void TJohnsonRW::printLog(Option_t *option)
{
    cout << "*******************************" << endl;
    cout << " ** TJohnsonRW::printLog[" << option << "]" << endl;
    cout << " **  _nX[" << _nX << "]" << endl;
    cout << " **  _nY[" << _nY << "]" << endl;

    cout << " ** _sMethod[" << _sMethod <<"]" << endl;
    cout << " ** _dr2[" << _dr2 << "] _dr2A[" << _dr2A << "]" << endl;

    TSensitivity::printLog();
    cout << "*******************************" << endl;
}

} // Fin du namespace Sensitivity
} // Fin du namespace URANIE

