//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
// TMorris mer sep 13 12:39:50 CEST 2006
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TMorris.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   wen sep 13 12:39:50 CEST 2006
 
 \brief Interface of class URANIE::Sensitivity::TMorris
 
 */
/**
 
 * @class URANIE::Sensitivity::TMorris
 *
 * @brief Description of the class TMorris.
 *
 The method recommended by Morris (1991) is valid for p \b even,
 and the value for \f$ \Delta \f$ is given by the relation
 \f[
 \Delta = \frac{p}{2(p-1)}
 \f]
 where \f$ p\f$ is the level.


 The number for the code evaluation is given by the relation
 \f[
 n_{evak} = r * ( n_X + 1 )
 \f]

 \section use_morris Using TMorris class inside a script
 This is the use of TMorris class. For a generic documentation, please refer
 to the section \ref use_sensitivity.
 \verbatim
 TMorris * scmo = new TMorris(tds, function, replique, level);
 TMorris * scmo = new TMorris(tds, TCode, replique, level);


 
 scmo->generateSample();
 scmo->computeIndexes();
 scmo->drawIndexes("ddpie", "", "nonewcanv,pie,first");
 ntsampl->Print();
 \endverbatim
 *
 */
#ifndef TMORRIS_H
#define TMORRIS_H

/// ROOT
#ifndef ROOT_TMatrixD
#include "TMatrixD.h"
#endif
#include "TVectorD.h"
#include "TRandom2.h"

/// URANIE
#include "TScreeming.h"

namespace URANIE
{
namespace Sensitivity
{
class TMorris: public TScreeming
{
private:
    Double_t _delta; ///< The Delta parameter
    Int_t _level; ///< The level parameter (known as p)
    TMatrixD _matB;    //! Matrix of \f$ B^{\star} \f$
    TVectorD _vecDirection;    //! Vector of direction
    TNtupleD* _ntdatamorris;           ///< The tntuple of data
    URANIE::DataServer::TDSNtupleD* _ntmorrisresult;   ///< The tntuple of result \f$ (\mu, \sigma)\f$
    TRandom2 * _rand;

    //! constructor helpers
    void stdInit();
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor with the name of a function
    TMorris(URANIE::DataServer::TDataServer *tds, const char *fcn, Int_t nReplique,
            Int_t level, Double_t delta = 0, TString sout = "");
    //! Default constructor with a function
#ifdef ROOT5
    TMorris(URANIE::DataServer::TDataServer *tds, void *fcn, Int_t nReplique,
            Int_t level, Double_t delta = 0, TString sout = "");
#else
    TMorris(URANIE::DataServer::TDataServer *tds, void (*fcn)(Double_t*,Double_t*), 
            TString sin, TString sout,
            Int_t nReplique, Int_t level, Double_t delta = 0);
#endif
    //! Default constructor with a TDataServer filling
    TMorris(URANIE::DataServer::TDataServer *tds, const char *varexpinput,
            const char *varexpoutput, Option_t * option = "");
    //! Default constructor with a tcode
    TMorris(URANIE::DataServer::TDataServer *tds,
            URANIE::Launcher::TCode *fcode, Int_t nReplique, Int_t level,
            Double_t delta = 0);
    //! Default constructor with a trun
    TMorris(URANIE::DataServer::TDataServer *tds,
            URANIE::Relauncher::TRun *run, Int_t nReplique, Int_t level,
            Double_t delta = 0);
    //! Default destructor
    ~TMorris();
    //@}
    //---------------------------------------------
    /** @name Setting and Getting attributs
     **/
    //@{
    //! Get Morris results
    URANIE::DataServer::TDSNtupleD* getMorrisResults()
    {
        return _ntmorrisresult;
    }

    TTree *getResultTuple(bool commonresulttuple=false);

    //! Sets the level
    void setLevel(Int_t val)
    {
        _level = val;
    }
    //! Gets the level
    Int_t getLevel()
    {
        return _level;
    }
    //! Sets the delta
    void setDelta(Double_t val)
    {
        _delta = val;
    }
    //! Gets the delta
    Double_t getDelta()
    {
        return _delta;
    }
    //! Gets the number of replica
    Int_t getNReplica()
    {
        return _nS;
    }
    //@}

    //---------------------------------------------
    /** @name Generation of the sample
     **/
    //@{
    virtual void generateSample(Option_t * option = "");

    //! Computes the orientation matrix \f$ B^{\star}\f$
    /*! 

     \section algo_morris_computeOrientationMatrix algorithme

     The matrix \f$ \tilde{B} = \f$ d�fined in the paper of Morris is computed
     \warning the author underlines the fact that the column  of the matrix colonne is either
     the matrix of B, or the column of B by replacing the 0 by 1 or reciprocally
     the 1 by 0. Thus, we can even compute  \f$ \Delta \tilde{B}\f$.
     */
    void computeOrientationMatrixOld();
    void computeOrientationMatrix();
    //@}

    //---------------------------------------------
    /** @name Evaluate the indexes
     **/
    //@{
    void evaluateIndexes(Option_t * option = "");

    //!PreTreatment for every output
    /*!
      Doing a pre-treatment when running over outputs.
     */
    void preTreatment();
    //@}
    
    //---------------------------------------------
    /** @name visualization and Scan the data
     **/
    //@{
    virtual void createTuple();
    //! Draws the indexes
    virtual void drawIndexes(TString sTitre, const char *select = "", Option_t * option = "", bool internalCall=kFALSE);
    void drawSample(TString svar = "", Int_t nreplique = 0, Option_t *option="");
    void drawIndexes(Option_t * option = "", bool internalCall=kFALSE);
    void drawIndexesAll(Option_t * option = "");
    void pyDrawIndexes(Option_t * option = "");
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Sensitivity::TMorris, ID_SENSITIVITY)
    //Classe de
};

} // Fin du namespace Sensitivity
} // Fin du namespace URANIE
#endif
