//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TShapley.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   2020/09/29

 \brief Header of the URANIE::Sensitivity::TShapley class

*/
/**
 *
 * @class URANIE::Sensitivity::TShapley
 *
 * @brief Description of the class TShapley
 *
 * \section sectionIntro Description of the \bf TShapley class
 *
 * \subsection subsectionIntro Introduction
 *
 * \subsection subsectionBiblio Bibliography
 * \li Shapley 1953
 * \li Owen 2014 "Sobol' indices and Shapley value", Journal of Uncertainty Quantification, 2:245-251
 * \li Owen, Prieur, 2017, "On Shapley value for measuring importance of dependent inputs",
 * \li Iooss, Prieur, 2017, "Shapley effects for Sensitivity Analysis with dependent inputs: comparaison with Sobol' indices, numerical estimation and applications"
 * \li Song, Nelson, Staum, 2016 "Shapley Effect for Global Sensitivity Analysis: Theory and Computation"

 */

#ifndef TSHAPLEY_H
#define TSHAPLEY_H

/// ROOT

/// Uranie
#include "TSensitivity.h"
#include "TGaussianSampling.h"

using namespace URANIE::DataServer;
using namespace URANIE::Sampler;

namespace URANIE
{
namespace Sensitivity
{
class TShapley: public TSensitivity
{
public:
    enum EConditionalSampling
    {
        kNorta,
        kCopulas,
        kVines,
        kUnknown
    };
    //---------------------------------------------
    /** @name Constructor and Destructor
         **/
    //@{
    /*!
     * \brief Default constructor
     */
    TShapley();

    TShapley(TDataServer *tds, const char *varexpinput,
            const char *varexpoutput, Option_t * option = "");

    /*!
     * \brief Constructor with the TDataServer and function name
     * \param tds
     * \param fcn
     * \param ns
     * \param varexpinput
     * \param varexpoutput
     * \param option [""]
     */
    TShapley(TDataServer *tds, const char *fcn, Int_t ns,
             const char *varexpinput, const char *varexpoutput,
             Option_t* option="");

    /*!
     * \brief Constructor with the TDataServer and function
     * \param tds
     * \param ns
     * \param sInputs
     * \param sOutputs
     * \param option
     */
    TShapley(TDataServer *tds, void (*fcn)(Double_t*,Double_t*),
             Int_t ns, const char *varexpinput, const char *varexpoutput,
             Option_t* option="");

    /*!
     * \brief Constructor with the TDataServer and a TCode
     * \param tds
     * \param fcode
     * \param ns
     * \param option
     */
    TShapley(TDataServer *tds, URANIE::Launcher::TCode *fcode,
             Int_t ns, Option_t * option = "");

    /*!
     * \brief Constructor with the TDataServer and a TRun
     * \param tds
     * \param run
     * \param ns
     * \param option
     */
    TShapley(TDataServer *tds, URANIE::Relauncher::TRun *run,
             Int_t ns, Option_t * option = "");


    /*!
     * \brief Default destructor
     */
    virtual ~TShapley();
    //@}
    //---------------------------------------------
    /** @name Setting and Getting attributs
     **/
    //@{
    const char * getIteratorNameNo()
    {
        return _sIteratorNameNo.Data();
    }
    const char * getIteratorNameNi()
    {
        return _sIteratorNameNi.Data();
    }
    Int_t getNv() { return _nV;}
    void setNv(Int_t nv);

    Int_t getNm() { return _nM;}
    void setNm(Int_t nm);

    Int_t getNo() { return _nO;}
    void setNo(Int_t no);

    Int_t getNi() { return _nI;}
    void setNi(Int_t ni);


    void setSamplerMethod(EConditionalSampling nMethod);

    EConditionalSampling getSamplerMethod() {
        return _nCondSampler;
    }

    TMatrixD getMatrixNormalVariance() {
        return _matNormalCovariance;
    }


    //@}

    //---------------------------------------------
    /** @name Generation of the sample
     **/
    //@{
    virtual void generateSample(Option_t * option = "");
    //@}


    //---------------------------------------------
    /** @name Compute the indexes
     **/
    //@{

    /*!
     * \brief Compute the sensitivity matrix
     * \param option
     */
    void evaluateIndexes(Option_t * option = "");
    //@}



    void printLog(Option_t *option = "");

    /*!
     * \brief Print the summary of the TShapley method
     *
     * Print just the results of the method \b NOT \b THE \b LOG
     * Just a test to generalize
     * \param option (Option_t *) option to pass [""]
     */
    void printSummary(Option_t * option = "");
    //@}


private:
    // methods
    void transfertDataTDS(TDataServer *tds);
    void splitMatrixVarCoVar(TMatrixD matVarCovar, TVectorD vecOfUnit,
                             TMatrixD &matVarCovar0, TMatrixD &matVarCovar1, TMatrixD &matVarCovar01,
                             TMatrixD matCorr, TMatrixD &matCorrtmp);
    /*!
     * \brief computeNormalCorrelationMatrix
     *
     * Compute the correlation matrix \bf "_matNormalCorr   " from the rank correlation for the Normal (Gaussian) multivariate distribution
     *
     * $f[
     * \rho \: = \: 2 \: sin ( \frac{\pi}{6} \rh_{r}
     * $f]
     *
     */
    void computeNormalCorrelationMatrix(TMatrixD matVarCovariance);

    /*!
     * \brief generateConditionnalSample
     *
     * Generate a conditionnal sample from tds
     *
     * \param tds
     * \param matVarCovar0
     * \param matVarCovar1
     * \param matVarCovar01
     */
    TDataServer *generateConditionnalSampleByNortaOriginal(TDataServer *tds0, TDataServer *tds1, TMatrixD matVarCovar0, TMatrixD matVarCovar1, TMatrixD matVarCovar01);
    void generateConditionnalSampleByNorta(TGaussianSampling *tsanormal, TDataServer *tdsres, TList *listOfLeaves, TMatrixD matVarCovar, TVectorD &vecOf01);


    // Attributes
    TString _sIteratorNameNo;        ///< The specific iterator attribute for NO integer
    TString _sIteratorNameNi;        ///< The specific iterator attribute for NI integer
    Int_t _nV;                   ///< The size of the nV sample
    Int_t _nO;                   ///< The size of the nO sample
    Int_t _nI;                   ///< The size of the nI sample
    Int_t _nM;                            ///< The size of the loop of permutations
    TMatrixD _matNormalCovariance;        ///< Normal(Gaussian) Input Covariance matrix
    EConditionalSampling _nCondSampler;   ///< The method of sampler
protected:
    void init(Option_t* option);
public:
    ClassDef(TShapley, ID_SENSITIVITY)
};
} // End Of namespace Sensitivity
} // End Of namespace URANIE

#endif
// End Of File
