////////////////////////////////////////////////////////////// 
// Copyright (C) 2013-2024 CEA/DES 
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
\file   TSobolRank.cxx
\author Rudy Chocat <rudy.chocat@cea.fr>
\date   2024-03-01 15:59:38

\brief Implement the URANIE::Sensitivity::TSobolRank class

*/

/// ROOT
#include "TPad.h"
#include "TProfile.h"
#include "TLeaf.h"
#include "TGaxis.h"
#include "TLatex.h"
#include "TLegend.h"
#include "TMap.h"
#include "TRegexp.h"
#include "TObjString.h"

#include "TDirectory.h"
#include "Math/Math.h"
#include "Math/ProbFuncMathCore.h"

/// Uranie
#include "TSensitivity.h"
#include "TSobolRank.h"
#include "TPatternsEventList.h"

using std::cout;
using std::endl;

ClassImp (URANIE::Sensitivity::TSobolRank)

namespace URANIE
{
    namespace Sensitivity
    {
    TSobolRank::TSobolRank():
        TSensitivity()
    {
        SetName("SobolRank");
        setMethodName("SobolRank");
        SetTitle("my title of the Sobol Rank indexes");
        cout << " *** Default constructor TSobolRank (empty) ***" << endl;
    }
    TSobolRank::TSobolRank(URANIE::DataServer::TDataServer *tds, const char *varexpinput, const char *varexpoutput, Option_t * option) :
        TSensitivity(tds, varexpinput, varexpoutput, option)
    {
        SetName("SobolRank");
        setMethodName("SobolRank");
        SetTitle("my title of the Sobol Rank indexes");
        if (_blog)
        {
            cout << "construit TSobolRank tds[" << _tds->GetName() << "] Option ["
                << option << "]" << endl;
            cout << "varexpinput[" << varexpinput << "]" << endl;
            cout << "varexpoutput[" << varexpoutput << "]" << endl;
        }

        _sIteratorName = Form("sobolrank__n__iter__%s", _tds->GetName());
        parseOption(option);
        
        if (_blog)
        {
            cout << "_sSelectedInput[" << _sSelectedInput << "]" << endl;
            cout << "_sSelectedOutput[" << _sSelectedOutput << "]" << endl;
        }

    }
    TSobolRank::~TSobolRank()
    {
        if ( _blog )
            cout << " *** destroy THSIC(" << GetName() << ") Title[" << GetTitle() << "] *** " << endl;
    }

    void TSobolRank::evaluateIndexes(Option_t * option)
    {

        if (_blog)
        {
            cout << endl << "*********************************" << endl;
            cout << " ** TSobolRank::evaluateIndexes" << endl;
            cout << " **    _sSelectedInput[" << _sSelectedInput << "]" << endl;
            cout << " **   _sSelectedOutput[" << _sSelectedOutput << "]" << endl;
        }
        // Compute statistic and rank
        _tds->computeStatistic();
        _tds->computeRank(_sSelectedInput);

        // Vector to store rank values
        std::vector< double > vec_yrk ;
        double S1 = 0.0;

        // Calculate S1 for each input attribute
        for(Int_t i=0; i< _nX; i++)
        {
            // Get input attribute
            URANIE::DataServer::TAttribute *att = (URANIE::DataServer::TAttribute *) (_listOfInputAttributes->At(i));	
            TString satt = att->GetName();
            TString rank_name =  Form("Rk_%s",satt.Data());

            // Retrieve rank values
            for(unsigned int it=0; it<_tds->getTuple()->GetEntries(); it++)
            {
                _tds->getTuple()->Draw( _sOutputAttribute, Form("%s==%d", rank_name.Data(),it+1), "goff" );
                vec_yrk.push_back(*_tds->getTuple()->GetV1() );
           }
            // Calculate terms for S1
            double term1 = 0.0, term2 = 0.0, term3 = 0.0;
            for(unsigned int i=0; i<_tds->getTuple()->GetEntries(); i++)
            {   
                if(i==_tds->getTuple()->GetEntries()-1)
                {   
                    term1 += vec_yrk[i] * vec_yrk[0];
                }
                else
                {
                    term1 += vec_yrk[i] * vec_yrk[i+1];
                }
                term2 += vec_yrk[i] * vec_yrk[i];
            }
            // Normalize terms
            term1 = term1 / (double)_tds->getTuple()->GetEntries() ;
            term2 = term2 / (double)_tds->getTuple()->GetEntries() ;
            term3 = _tds->getAttribute(_sOutputAttribute)->getMean();

            // Calculate S1
            S1 = ( term1 - term3*term3 ) / ( term2 - term3*term3 );

            // Clear vector for next iteration
            vec_yrk.clear();

            // Fill index using calculated S1
            fillIndex(satt.Data(), "First", S1,"--first--");
        }

    }

 } // End of namespace Sensitivity
} // End of namespace URANIE
