//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TDGSM.h
 \author GAUDIER Fabrice <fabrice.gaudier@cea.fr>
 \date   2019/06/14

 \brief Test File of the class URANIE::Sensitivity::TDGSM
 */

// Uranie
#include "TestTDGSM.h"
#include "uranie_commands.h"
#include "TDGSM.h"
#include "TDataServer.h"
#include "TLogUniformDistribution.h"
#include "TCode.h"
#include "TOutputFileKey.h"
#include "TOutputFileRow.h"
#include "TLauncher.h"
#include "TLauncherFunction.h"
#include "TestUserFunctions.h"

// ROOT
#include "TString.h"
#include "TMath.h"
#include "TSystem.h"
#include "TROOT.h"
#ifdef ROOT5
#include "Api.h"
#endif

#define EPSILON 0.001
#define SPL_PRECISION_COMPARE 1.e-6
#define DBL_PRECISION_COMPARE 1.e-12

#include <iostream>
#include <math.h>

TDGSMTest::TDGSMTest() :
        _tsen(NULL), _nsFlowrate(50), _nSIshigami(10000),
        _tds1(NULL), _tds2(NULL),
        _mycode(NULL),
        _fout(NULL), _bLog(kFALSE)
{
    _workingTestPath = gSystem->pwd();

#ifdef ROOT5
    //  gROOT->LoadMacro("UserFunctions.C");
    G__ClassInfo globalscope;
    G__MethodInfo method;
    long dummy;
    // the Ishigmai usecase
    method = globalscope.GetMethod("Ishigami", "double*,double*", &dummy);
    _p2fIshigami = method.PointerToFunc();
    // the Flowrate usecase
    method = globalscope.GetMethod("flowrateModel", "double*,double*", &dummy);
    _p2fFlowrate = method.PointerToFunc();
    // the Affine usecase
    method = globalscope.GetMethod("FonctionAffine", "double*,double*", &dummy);
    _p2fAffine = method.PointerToFunc();
#endif

    // Define the Flowrate reference results
    _refMatFlowrate.ResizeTo(1, 8);
    _refMatFlowrate.Zero();

    // Define the Ishigami reference results
    _refMatIshigami.ResizeTo(1, 3);
    _refMatIshigami(0, 0) =  7.9479277;
    _refMatIshigami(0, 1) = 24.308462;
    _refMatIshigami(0, 2) = 11.503541;
}

TDGSMTest::~TDGSMTest()
{
    cleanWorkingDirectories("Destructor");
}


void TDGSMTest::cleanWorkingDirectories(const char *option)
{
    if ( _bLog)
        cout << endl
             << " ** TDGSMTest::cleanWorkingDirectories option["
             << option << "] _workingTestPath[" << _workingTestPath << "]"
             << endl;

    UranieTests::removeFolder(_workingTestPath.Data(),"URANIE");
    UranieTests::removeFolder(_workingTestPath.Data(),"tmpLanceurUranie");

    UranieTests::removeFile(_workingTestPath.Data(),"DGSMLevelEDatabase.root");
    UranieTests::removeFile(_workingTestPath.Data(),"_dgsm_launching_.dat");
}

////////////////////////////////
///                          ///
///    setUp/tearDown        ///
///                          ///
////////////////////////////////
void TDGSMTest::setUp()
{
    gSystem->cd(_workingTestPath);
    cleanWorkingDirectories("setUp");

    _nS = -1;

    rw = new URANIE::DataServer::TUniformDistribution("rw", 0.05, 0.15);
    r = new URANIE::DataServer::TUniformDistribution("r", 100.0, 50000.0);
    tu = new URANIE::DataServer::TUniformDistribution("tu", 63070.0, 115600.0);
    tl = new URANIE::DataServer::TUniformDistribution("tl", 63.1, 116.0);
    hu = new URANIE::DataServer::TUniformDistribution("hu", 990.0, 1110.0);
    hl = new URANIE::DataServer::TUniformDistribution("hl", 700.0, 820.0);
    l = new URANIE::DataServer::TUniformDistribution("l", 1120.0, 1680.0);
    kw = new URANIE::DataServer::TUniformDistribution("kw", 9855.0, 12045.0);

    rw->setDefaultValue(0.05);
    r->setDefaultValue(33366.67);
    tu->setDefaultValue(63070.0);
    tl->setDefaultValue(116.0);
    hu->setDefaultValue(1110.0);
    hl->setDefaultValue(768.57);
    l->setDefaultValue(1200.0);
    kw->setDefaultValue(11732.14);

    x1 = new URANIE::DataServer::TUniformDistribution("x_{1}",
            -1.0 * TMath::Pi(), TMath::Pi());
    x2 = new URANIE::DataServer::TUniformDistribution("x_{2}",
            -1.0 * TMath::Pi(), TMath::Pi());
    x3 = new URANIE::DataServer::TUniformDistribution("x_{3}",
            -1.0 * TMath::Pi(), TMath::Pi());

    x1->setDefaultValue(0.);
    x2->setDefaultValue(0.);
    x3->setDefaultValue(0.);
}

void TDGSMTest::setUpFlowrate()
{
    _tds1 = new URANIE::DataServer::TDataServer("tds", "tds for Unit Tests");
    _tds1->addAttribute(rw);
    _tds1->addAttribute(r);
    _tds1->addAttribute(tu);
    _tds1->addAttribute(tl);
    _tds1->addAttribute(hu);
    _tds1->addAttribute(hl);
    _tds1->addAttribute(l);
    _tds1->addAttribute(kw);


    _nS = _nsFlowrate;
#ifdef ROOT5
    _tsen = new URANIE::Sensitivity::TDGSM(_tds1, _p2fFlowrate, _nS, "rw:r:tu:hu:hl:l:kw", "myY", "steps=2%");
#else
    _tsen = new URANIE::Sensitivity::TDGSM(_tds1, &flowrateModel, _nS, "rw:r:tu:tl:hu:hl:l:kw", "myY", "steps=2%");
#endif
    _tsen->setDrawProgressBar(kFALSE);
    _tsen->setSeed(147258369);
}

void TDGSMTest::setUpAffine()
{
    _tds2 = new URANIE::DataServer::TDataServer("tds", "tds for Unit Tests");
    _tds2->addAttribute( new URANIE::DataServer::TUniformDistribution("x1", 0, 1));
    _tds2->addAttribute( new URANIE::DataServer::TUniformDistribution("x2", 0, 100000.0));
    _tds2->getAttribute("x1")->delShare();
    _tds2->getAttribute("x2")->delShare();
        
    // In case of the halton QMC method where 0.0 is in the DoE
    _tds2->getAttribute("x1")->setStepValue(0.0004);
    _tds2->getAttribute("x2")->setStepValue(1.0);

    _nS = 20;

#ifdef ROOT5
    _tsen = new URANIE::Sensitivity::TDGSM(_tds2, _p2fAffine, _nS, "x1:x2", "AffineFct", "steps=1%");
#else
    _tsen = new URANIE::Sensitivity::TDGSM(_tds2, &FonctionAffine, _nS, "x1:x2", "AffineFct", "steps=1% qmc=halton");
#endif
    _tsen->unsetLog();
    _tsen->setDrawProgressBar(kFALSE);
    _tsen->setSeed(147258369);

}

void TDGSMTest::setUpIshigami()
{
    _tds2 = new URANIE::DataServer::TDataServer("tds", "tds for Unit Tests");
    _tds2->addAttribute(x1);
    _tds2->addAttribute(x2);
    _tds2->addAttribute(x3);

    _nS = _nSIshigami;

#ifdef ROOT5
    _tsen = new URANIE::Sensitivity::TDGSM(_tds2, _p2fIshigami, _nS, "x_{1}:x_{2}:x_{3}", "myY", "steps=0.001%");
#else
    _tsen = new URANIE::Sensitivity::TDGSM(_tds2, &Ishigami, _nS, "x_{1}:x_{2}:x_{3}", "myY", "steps=0.001%");
#endif
    _tsen->setDrawProgressBar(kFALSE);
    _tsen->setSeed(147258369);
}

void TDGSMTest::tearDown()
{
    gSystem->cd(_workingTestPath);
    cleanWorkingDirectories("tearDown");

    if (_fout) {
        delete _fout;
        _fout = NULL;
    }

    if (_mycode != NULL) {
        delete _mycode;
        _mycode = NULL;
    }
    
    if (_tsen) {
        delete _tsen;
        _tsen = NULL;
    }
    
    if (_tds1) {
        delete _tds1;
        _tds1 = NULL;
      }

    if (_tds2) {
        delete _tds2;
        _tds2 = NULL;
      }

    delete rw;
    rw = NULL;
    delete r;
    r = NULL;
    delete tu;
    tu = NULL;
    delete tl;
    tl = NULL;
    delete hu;
    hu = NULL;
    delete hl;
    hl = NULL;
    delete l;
    l = NULL;
    delete kw;
    kw = NULL;

    delete x1;
    x1 = NULL;
    delete x2;
    x2 = NULL;
    delete x3;
    x3 = NULL;
}

////////////////////////////////
///                          ///
///        validate          ///
///                          ///
////////////////////////////////

void TDGSMTest::validateValuesFlowrate()
{
    TMatrixD sMat = _tsen->getSensitivityMatrix();

    for (int i = 0; i < 8; i++) {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(_refMatFlowrate(0, i), sMat(0, i), 5e-3);
    }
}

void TDGSMTest::validateValuesAffine()
{
    TMatrixD sMat = _tsen->getSensitivityMatrix();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, sMat(0, 0), 5e-3);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(4.0, sMat(0, 1), 5e-3);
}

void TDGSMTest::validateValuesIshigami()
{
    TMatrixD sMat = _tsen->getSensitivityMatrix();

    for (int i = 0; i < 3; i++) {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(_refMatIshigami(0, i), sMat(0, i), 5e-3);
    }
}

////////////////////////////////
///                          ///
///         Tests            ///
///                          ///
////////////////////////////////
///
void TDGSMTest::testDGSMConstructor()
{
    setUpFlowrate();

    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of patterns (Constructor)", 0, _tds1->getNPatterns());
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of attributes (Constructor)", 8, _tds1->getNAttributes());

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Step size parameter (Constructor)", 0.02, _tsen->getStepSizeParameter(), SPL_PRECISION_COMPARE);

    cout << " **** name[" << _tsen->GetName() << "]" << endl;
    CPPUNIT_ASSERT_MESSAGE(Form("Name of the method (DGSM) [%s]",_tsen->GetName()) ,
                           strcmp(_tsen->GetName(), "Sensitivity_myY") == 0);
    CPPUNIT_ASSERT_MESSAGE("Name of the method (DGSM)",
                           strcmp(_tsen->getMethodName(), "DGSM") == 0);

    /// Default Values for the TSensitivity Object with the Getter Functions
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of Input getID()",
                                 ID_SENSITIVITY, _tsen->getID());
    CPPUNIT_ASSERT_MESSAGE("The pointer of the result TTree getResultTuple()",
                           _tsen->getResultTuple() != NULL);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Size of the result TTree getResultTuple()",
                                 0, (Int_t)(_tsen->getResultTuple()->GetEntries()));
    const Int_t nBranches = 8;
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of branches of the result TTree getResultTuple()",
                                 nBranches, (Int_t)(_tsen->getResultTuple()->GetNbranches()));

    TString sBranches[8] = {"Out", "Inp", "Order", "Method", "Algo", "Value", "CILower", "CIUpper"};
    for (Int_t i=0; i < nBranches; i++) {
        CPPUNIT_ASSERT_MESSAGE(Form("Status of the branch \"%s\" in the result TTree getResultTuple()", sBranches[i].Data()),
                               _tsen->getResultTuple()->GetBranchStatus(sBranches[i].Data()));
    }
    CPPUNIT_ASSERT_MESSAGE(Form("The Name of the evaluatuor function getFunctionName() [%s]", _tsen->getFunctionName().Data()),
                           strcmp(_tsen->getFunctionName(), "Eval_myY_by_rw:r:tu:tl:hu:hl:l:kw") == 0);
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("The Seed value getSeed()[%d]", _tsen->getSeed()),
                                 147258369, (Int_t)(_tsen->getSeed()));
    CPPUNIT_ASSERT_MESSAGE("The noIntermediateSaved flag",
                           _tsen->getNoIntermediateSaved() == kFALSE);
    CPPUNIT_ASSERT_MESSAGE(Form("The Name of the Iterator getIteratorName() [%s]", _tsen->getIteratorName()),
                           strcmp(_tsen->getIteratorName(), "__modified_att__") == 0);
    CPPUNIT_ASSERT_MESSAGE(Form("The Name of the time attribute getTimeName() [%s]", _tsen->getTimeName().Data()),
                           strcmp(_tsen->getTimeName(), "") == 0);
    CPPUNIT_ASSERT_MESSAGE("The draw progress bar flag",
                           _tsen->getDrawProgressBar()==kFALSE);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of Input  via getNInput()",
                                 8, _tsen->getNInput());
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of Output via getNOutput()",
                                 1, _tsen->getNOutput());
    // default constructor
    delete _tsen;

#ifdef ROOT5
    _tsen = new URANIE::Sensitivity::TDGSM(_tds1, _p2fFlowrate, _nsFlowrate, "", "myY", "");
#else
    _tsen = new URANIE::Sensitivity::TDGSM(_tds1, &flowrateModel, _nsFlowrate, "", "myY", "");
#endif
    _tsen->setDrawProgressBar(kFALSE);
    _tsen->setSeed(147258369);

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Step size parameter (Constructor)", 0.01, _tsen->getStepSizeParameter(), SPL_PRECISION_COMPARE);

    // default constructor
    delete _tsen;

    ///
    /// Default constructor (empty)
    ///

    _tsen = new URANIE::Sensitivity::TDGSM();

    CPPUNIT_ASSERT_MESSAGE("The Log flag via getLog()",
                           _tsen->getLog() == kFALSE);

    _tsen->unsetLog();
    CPPUNIT_ASSERT_MESSAGE(Form("Name of the object (DGSM)(%s)", _tsen->GetName()),
                           strcmp(_tsen->GetName(), "") == 0);
    CPPUNIT_ASSERT_MESSAGE("Name of the method (DGSM)",
                           strcmp(_tsen->getMethodName(), "DGSM") == 0);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of Input getID()",
                                 ID_SENSITIVITY, _tsen->getID());
    CPPUNIT_ASSERT_MESSAGE("The pointer of the result TTree getResultTuple()",
                           _tsen->getResultTuple() != NULL);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Size of the result TTree getResultTuple()",
                                 0, (Int_t)(_tsen->getResultTuple()->GetEntries()));
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of branches of the result TTree getResultTuple()",
                                 nBranches, (Int_t)(_tsen->getResultTuple()->GetNbranches()));
    for (Int_t i=0; i < nBranches; i++) {
        CPPUNIT_ASSERT_MESSAGE(Form("Status of the branch \"%s\" in the result TTree getResultTuple()", sBranches[i].Data()),
                               _tsen->getResultTuple()->GetBranchStatus(sBranches[i].Data()));
    }
    CPPUNIT_ASSERT_MESSAGE(Form("The Name of the evaluatuor function getFunctionName() [%s]", _tsen->getFunctionName().Data()),
                           strcmp(_tsen->getFunctionName(), "") == 0);
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("The Seed value getSeed()[%d]", _tsen->getSeed()),
                                 0, (Int_t)(_tsen->getSeed()));
    CPPUNIT_ASSERT_MESSAGE("The noIntermediateSaved flag",
                           _tsen->getNoIntermediateSaved() == kFALSE);
    CPPUNIT_ASSERT_MESSAGE(Form("The Name of the Iterator getIteratorName() [%s]", _tsen->getIteratorName()),
                           strcmp(_tsen->getIteratorName(), "__modified_att__") == 0);
    CPPUNIT_ASSERT_MESSAGE(Form("The Name of the time attribute getTimeName() [%s]", _tsen->getTimeName().Data()),
                           strcmp(_tsen->getTimeName(), "") == 0);
    CPPUNIT_ASSERT_MESSAGE("The draw progress bar flag",
                           _tsen->getDrawProgressBar());
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of Input  via getNInput()",
                                 0, _tsen->getNInput());
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of Output via getNOutput()",
                                 0, _tsen->getNOutput());

}

void TDGSMTest::testDGSMInitSample()
{
    setUpIshigami();

    Int_t ns = _tds2->getNPatterns();
    Int_t nx = 3;

    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of patterns (Ishigami)", 0, ns);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of attributes (Ishigami)", nx, _tds2->getNAttributes());

    ns = 12;
    _tsen->initSample(ns, "srs");

    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of patterns (Ishigami) after call initSample", ns, _tds2->getNPatterns());
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of attributes (Ishigami) after call initSample ", nx, _tds2->getNAttributes());

    // The auxillary attributes
    // Iterator "__modified_att__"
    TString satt = _tsen->getIteratorName();
    CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", satt.Data()), _tds2->isAttribute(satt) == kFALSE);

    satt = "__nominal_set__";
    CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", satt.Data()), _tds2->isAttribute(satt) == kFALSE);

    // Print te log
    _tsen->printLog();
}

void TDGSMTest::testDGSMGenerateSample()
{
    setUpIshigami();

    Int_t ns = _tds2->getNPatterns();
    Int_t nx = 3;

    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of patterns (Ishigami)", 0, ns);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of attributes (Ishigami)", nx, _tds2->getNAttributes());

    ns = 12;
    _tsen->initSample(ns,"lhs");

    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of patterns (Ishigami) after call initSample", ns, _tds2->getNPatterns());
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of attributes (Ishigami) after call initSample ", nx, _tds2->getNAttributes());

    _tsen->generateSample();

    // Number of patterns
    Int_t neval = 2*nx*ns;
    ns += neval;
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of patterns (Ishigami) after call generateSample", ns, _tds2->getNPatterns());

    // Add the auxillary attributes
    nx+=2;
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of attributes (Ishigami) after call generateSample ", nx, _tds2->getNAttributes());

    // The auxillary attributes
    // Iterator "__modified_att__"
    TString satt = _tsen->getIteratorName();
    CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", satt.Data()), _tds2->isAttribute(satt));

    satt = "__nominal_set__";
    CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", satt.Data()), _tds2->isAttribute(satt));
}

void TDGSMTest::testDGSMIshigamiFunction()
{
    setUpIshigami();

    Int_t nx = 3;

    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of patterns (Ishigami)", 0, _tds2->getNPatterns());
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of attributes (Ishigami)", nx, _tds2->getNAttributes());

    _tsen->unsetLog();
    _tsen->setDrawProgressBar(kFALSE);
    _tsen->generateSample();

    Int_t ns=_nSIshigami;
    Int_t neval = 2*nx+1;
    Int_t npat = TMath::FloorNint(ns/neval)*neval;
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of patterns (Ishigami) after call generateSample", npat, _tds2->getNPatterns());
    // Add the auxillary attributes
    nx+=2;
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of attributes (Ishigami)", nx, _tds2->getNAttributes());

    _tsen->unsetLog();
    _tsen->setDrawProgressBar(kFALSE);
    _tsen->computeIndexes();
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of patterns (Ishigami) after call computeIndexes", npat, _tds2->getNPatterns());
    nx+=1;
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of attributes (Ishigami) after call computeIndexes", nx, _tds2->getNAttributes());

    validateValuesIshigami();
}

void TDGSMTest::testDGSMAffineFunction()
{
    setUpAffine();
    Int_t nx = 2;

    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of patterns (Affine)", 0, _tds2->getNPatterns());
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of attributes (Affine)", nx, _tds2->getNAttributes());

    _tsen->unsetLog();
    _tsen->setDrawProgressBar(kFALSE);
    _tsen->computeIndexes();
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of patterns (Affine) after call computeIndexes", _nS, _tds2->getNPatterns());
    // Add the 2 auxillary and output attributes
    nx+=3;
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of attributes (Affine)", nx, _tds2->getNAttributes());

    validateValuesAffine();
}

void TDGSMTest::testDGSMCounterExemple3Function()
{
    Int_t nx = 4;

    TDataServer *tds = new TDataServer();
    tds->addAttribute( new TUniformDistribution("x1"));
    tds->addAttribute( new TUniformDistribution("x2"));
    tds->addAttribute( new TUniformDistribution("x3"));
    tds->addAttribute( new TUniformDistribution("x4"));
    tds->getAttribute("x1")->delShare();
    tds->getAttribute("x2")->delShare();
    tds->getAttribute("x3")->delShare();
    tds->getAttribute("x4")->delShare();

    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of patterns (CounterExemple3)", 0, tds->getNPatterns());
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of attributes (CounterExemple3)", nx, tds->getNAttributes());

    _nS = 99999;

#ifdef ROOT5
    G__ClassInfo globalscope;
    G__MethodInfo method;
    long dummy;
    // the CounterExemple3 usecase
    method = globalscope.GetMethod("CounterExemple3Function", "double*,double*", &dummy);
    void *p2fFunction = method.PointerToFunc();
    _tsen = new URANIE::Sensitivity::TDGSM(tds, p2fFunction, _nS, "x1:x2:x3:x4", "theY", "steps=0.001%");
#else
    _tsen = new URANIE::Sensitivity::TDGSM(tds, &CounterExemple3Function, _nS, "x1:x2:x3:x4", "theY", "steps=0.001%");
#endif

    _tsen->unsetLog();
    _tsen->setDrawProgressBar(kFALSE);
    _tsen->computeIndexes();
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of patterns (CounterExemple3) after call computeIndexes", _nS, tds->getNPatterns());
    // Add the 2 auxillary and output attributes
    nx+=3;
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of attributes (CounterExemple3)", nx, tds->getNAttributes());


    TMatrixD sMat = _tsen->getSensitivityMatrix();
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Indices of x1 (CounterExemple3)", 1.22, sMat(0, 0), 5e-1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Indices of x2 (CounterExemple3)", 3.26, sMat(0, 1), 2e-1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Indices of x3 (CounterExemple3)", 1.00, sMat(0, 2), 1e-2);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Indices of x4 (CounterExemple3)", 1.00, sMat(0, 3), 1e-2);

    // Clean
    delete _tsen; _tsen=NULL;
    delete tds; tds = nullptr;
}


void TDGSMTest::testDGSMCounterExemple3TDS()
{

    TDataServer *tds = new TDataServer();
    tds->fileDataRead("DGSM_NominalValues_output.dat");

    Int_t nx = 7;
    Int_t npat = 99999;
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of patterns (CounterExemple3)", npat, tds->getNPatterns());
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of attributes (CounterExemple3)", nx, tds->getNAttributes());


    _tsen = new URANIE::Sensitivity::TDGSM(tds, "x1:x2:x3:x4", "theY");

    _tsen->unsetLog();
    _tsen->setDrawProgressBar(kFALSE);
    _tsen->computeIndexes();
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of patterns (CounterExemple3) after call computeIndexes", npat, tds->getNPatterns());
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of attributes (CounterExemple3)", nx, tds->getNAttributes());

    TMatrixD sMat = _tsen->getSensitivityMatrix();

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Indices of x1 (CounterExemple3)", 1.22, sMat(0, 0), 5e-1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Indices of x2 (CounterExemple3)", 3.26, sMat(0, 1), 2e-1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Indices of x3 (CounterExemple3)", 1.00, sMat(0, 2), 2e-2);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Indices of x4 (CounterExemple3)", 1.00, sMat(0, 3), 2e-2);

    // Clean
    delete _tsen; _tsen=NULL;
    delete tds; tds = nullptr;
}

////////////////////////////////
///                          ///
///    Throw Exceptions      ///
///                          ///
////////////////////////////////
void TDGSMTest::throwExceptionDGSMConstructor()
{
    URANIE::DataServer::TDataServer *tdsnull = new URANIE::DataServer::TDataServer();

    CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TDGSM(tdsnull, "x_{1}:x_{2}:x_{3}", "Ishigami", "myOpt"),
                         URANIE::Exceptions::UErrorExceptions);
    /// Clean
    delete tdsnull;

    setUpFlowrate();
    /// False input
    CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TDGSM(_tds1, "rw:hu:TheUnKnownAtt", "Ishigami", "myOpt"),
                          URANIE::Exceptions::UErrorExceptions);

    /// steps not in percent output
    CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TDGSM(_tds1, "rw:hu:hl:tu:tl", "TheUnKnownAtt", "myOpt steps=0.02"),
                          URANIE::Exceptions::UErrorExceptions);
}

void TDGSMTest::throwExceptionDGSMInitSample()
{
    setUpFlowrate();

    CPPUNIT_ASSERT_THROW_MESSAGE("Case of the initSample method",
                                 _tsen->initSample(3);,
                                 URANIE::Exceptions::UErrorExceptions);
}

// End OF File
