//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTRegression.cxx
 \author maud Salvoldi (OCC/Areva group)
 \date   $Date$
 
 \brief  Fichier de test de la classe URANIE::Sensitivity::TRegression
 
 Cr�e le : jeu d�c 22 12:30:15 CET 2011

 
 */

// Uranie
#include "TUniformDistribution.h"
#include "TSampling.h"
#include "TLauncher2.h"
#include "TCJitEval.h"
#include "TOutputFileKey.h"
#include "TOutputFileDataServer.h"
#include "TLauncherFunction.h"
#include "TLauncher.h"
#include "TestUserFunctions.h"
#include "TestTSensitivity.h"
#include "TestTRegression.h"
#include "TRegression.h"
#include "UDisableTest.h"
#include "uranie_commands.h"

// ROOT
#include "TString.h"
#include "TCanvas.h"
#define EPSILON 0.01 
#define LVLE_EPSILON 0.001
#define DBL_PRECISION_COMPARE 1.e-12

#include <iostream>
#include <iomanip>

TRegressionTest::TRegressionTest() :
        tds1(NULL), tds(NULL), treg(NULL), tss1(NULL)
{

}

TRegressionTest::~TRegressionTest()
{
    UranieTests::removeFile(gSystem->pwd(),"Ishigami.root");
    UranieTests::removeFile(gSystem->pwd(),"1000pointsLevelEDB.root");
    UranieTests::removeFile(gSystem->pwd(), "flowrateUniformDesign.root");
    UranieTests::removeFile(gSystem->pwd(), "flowrateUniformDesign2D.root");
}


void TRegressionTest::setUp()
{
    tds1 = new URANIE::DataServer::TDataServer();

    tds = new URANIE::DataServer::TDataServer("tdsflowreate",
            "DataBase flowreate");
    tds->fileDataRead("flowrateUniformDesign.dat");
}

void TRegressionTest::tearDown()
{
    delete tss1;
    tss1 = NULL;
    delete treg;
    treg = NULL;
    delete tds1;
    tds1 = NULL;
    delete tds;
    tds = NULL;
}

////////////////////////////////
///                          ///                            
///         Tests            ///
///                          ///
////////////////////////////////
void TRegressionTest::testSRCRegression()
{
    // Load the data base in the DataServer
    tds1->fileDataRead("flowrateUniformDesign.dat");

    Double_t sobol;
    //////////////////////
    // Case SRC
    //////////////////////
    tss1 = new URANIE::Sensitivity::TRegression(tds1, "rw:r:tu:tl:hu:hl:l:kw", "ystar");
    tss1->computeIndexes();

    tss1->printLog();

    TTree * tree = tss1->getResultTuple();
    tree->SetBranchAddress("Value", &sobol);

    Double_t dsum = 0.0;

    Double_t Comparison[8] =
    { 1.021863, 0.003439, 0.003259, 0.000259, 0.08636, 0.0723253, 0.04649, 0.020611 };
    for (Int_t i = 0; i < 8; i++)
    {
        tree->GetEntry(4 * i);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(Comparison[i], sobol, 0.001);
        dsum += Comparison[i];
    }

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.96265, tss1->getR2(), 0.001);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.9502, tss1->getR2A(), 0.001);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.2546, tss1->getSumOfSquareIndexes(), 0.001);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dsum, tss1->getSumOfSquareIndexes(), 0.001);
}

void TRegressionTest::testSRRCRegression()
{
  // Load the data base in the DataServer
  tds1->fileDataRead("flowrateUniformDesign.dat");

  Double_t sobol;
  //////////////////////
  // Case SRRC
  //////////////////////
  tss1 =  new URANIE::Sensitivity::TRegression(tds1, "rw:r:tu:tl:hu:hl:l:kw", "ystar", "SRRC");
  tss1->computeIndexes();

  tss1->printLog();

  TTree * tree = tss1->getResultTuple();
  tree->SetBranchAddress("Value", &sobol);

  Double_t precision0 = 0.0001;
  Double_t precision1 = 0.001;

#ifdef WIN32
  Double_t Comparison[8] =
    {  1.1416356, 7.09e-010, 0.0121472, 0.0023139, 0.0657581, 0.0845214, 0.0422113, 0.0044922};
  Double_t dr2 = 0.97243;
  Double_t dr2A = 0.963241;
  Double_t dr2SQI = 1.35308;


  for(Int_t i=0;i<8;i++)
    {
      tree->GetEntry(4*i);
      CPPUNIT_ASSERT_DOUBLES_EQUAL(Comparison[i],sobol, precision0);
    }
  CPPUNIT_ASSERT_DOUBLES_EQUAL( dr2, tss1->getR2(), precision1);
  CPPUNIT_ASSERT_DOUBLES_EQUAL( dr2A, tss1->getR2A(), precision1);
  CPPUNIT_ASSERT_DOUBLES_EQUAL( dr2SQI, tss1->getSumOfSquareIndexes(), precision1);

#else

  Double_t Comparison_3[8] =
    { 1.14163566, 7.09182618e-10, 0.0121472846, 0.00231397146, 0.0657581852, 0.0845214431, 0.042211311, 0.00449229271 };
  Double_t dr2_3 =0.972430423;
  Double_t dr2A_3 = 0.963240564;
  Double_t dr2SQI_3 = 1.35308015;
    
  for (Int_t i = 0; i < 8; i++)
    {
      tree->GetEntry(4 * i);
      Bool_t test = ( (TMath::Abs(Comparison_3[i] - sobol) < precision0) );
      CPPUNIT_ASSERT(test);
    }
  Bool_t test;
  test = ( (TMath::Abs(dr2_3 - tss1->getR2()) < precision1) );
  CPPUNIT_ASSERT(test);

  test = ( (TMath::Abs(dr2A_3 - tss1->getR2A()) < precision1) );
  CPPUNIT_ASSERT(test);

  test = ( (TMath::Abs(dr2SQI_3 - tss1->getSumOfSquareIndexes()) < precision1) );
  CPPUNIT_ASSERT(test);
#endif
}

#ifdef WITH_LEVELE
// ===========================================
//               LevelE
// ===========================================
void TRegressionTest::checkLevelEResults(TTree *res, vector<int>* elmts)
{

  double refSRRC[26]={0.855979626621,0.677628004933,0.470198321629,0.301729745776,0.176942328525,0.105083353287,0.0705090706141,0.0757135112619,
		      0.0937122757484,0.377075747895,0.383269995188,0.304124518955,0.253993707535,0.245747760396,0.258558937006,0.292308712349,0.334656632205,
		      0.389169420777,0.777126521448,0.882822091202,0.916854634223,0.923372879994,0.922043544468,0.912648961217,0.897583975395,0.876955949051};
  double refSRC[26]={0.138347112528,0.153088624485,0.124660649945,0.0802056119754,0.0689665487195,0.0724155198188,0.0721910621717,0.0706976366331,0.0693985766098,
		     0.0543584996216,0.0358530971597,0.0452663779964,0.0721090968241,0.0929691825036,0.104279805524,0.112806150501,0.119980512357,0.125615972305,
		     0.13521733456,0.116990499157,0.136990539816,0.142464917326,0.15860832064,0.194051121855,0.191359312056,0.174029535885};

  int npts = res->Draw("Value","Inp==\"__R2__\" && Order==\"Total\" && Method==\"SRRC^2\"","goff");
  double *data=res->GetV1();
  for(int ival=0; ival<npts; ival++)
    CPPUNIT_ASSERT_DOUBLES_EQUAL(refSRRC[ ((elmts==NULL) ? ival : elmts->at(ival)) ], data[ival], LVLE_EPSILON);
  // cout<<refSRRC[ ((elmts==NULL) ? ival : elmts->at(ival)) ]<<" ?=? "<< data[ival]<<endl;

  npts = res->Draw("Value","Inp==\"__R2__\" && Order==\"Total\" && Method==\"SRC^2\"","goff");
  data=res->GetV1();
  for(int ival=0; ival<npts; ival++)
    CPPUNIT_ASSERT_DOUBLES_EQUAL(refSRC[ ((elmts==NULL) ? ival : elmts->at(ival)) ], data[ival], LVLE_EPSILON);
  // cout<<refSRC[ ((elmts==NULL) ? ival : elmts->at(ival)) ]<<" ?=? "<< data[ival]<<endl;
  
}

void TRegressionTest::testLevelERegression()
{
    // Load the data base in the DataServer
    tds1->fileDataRead("1000pointsLevelEDB.dat");

    tss1 = new URANIE::Sensitivity::TRegression(tds1, "t:kl:kc:v1:l1:r1:rc1:v2:l2:r2:rc2:w","y","SRCSRRC");
    tss1->computeIndexes();

    TTree *res=tss1->getResultTuple();
    checkLevelEResults(res);
}


void TRegressionTest::testLevelERegressionOutputRestrVarexp()
{
    // Load the data base in the DataServer
    tds1->fileDataRead("1000pointsLevelEDB.dat");

    tss1 = new URANIE::Sensitivity::TRegression(tds1, "t:kl:kc:v1:l1:r1:rc1:v2:l2:r2:rc2:w","y[0]:y[5]:y[12]","SRCSRRC");
    tss1->computeIndexes();

    TTree *res=tss1->getResultTuple();
    checkLevelEResults(res, tss1->getAttributeElements(string("y")));
}

void TRegressionTest::testLevelERegressionOutputRestrOption()
{
    // Load the data base in the DataServer
    tds1->fileDataRead("1000pointsLevelEDB.dat");

    tss1 = new URANIE::Sensitivity::TRegression(tds1, "t:kl:kc:v1:l1:r1:rc1:v2:l2:r2:rc2:w","y","SRCSRRC,output=y[8]:y[15]:y[22]");
    tss1->computeIndexes();

    TTree *res=tss1->getResultTuple();
    checkLevelEResults(res, tss1->getAttributeElements(string("y")));
}

void TRegressionTest::testLevelERegressionOutputRestrBoth()
{
    // Load the data base in the DataServer
    tds1->fileDataRead("1000pointsLevelEDB.dat");

    tss1 = new URANIE::Sensitivity::TRegression(tds1, "t:kl:kc:v1:l1:r1:rc1:v2:l2:r2:rc2:w","y[0]:y[5]:y[12]","SRCSRRC");
    tss1->computeIndexes("output=y[8]:y[15]:y[22]");

    TTree *res=tss1->getResultTuple();
    checkLevelEResults(res, tss1->getAttributeElements(string("y")));
}
#endif

void TRegressionTest::testIshigamiRegression()
{
    // Load the data base in the DataServer
    tds1->fileDataRead("Ishigami.dat");

    Double_t sobol;
    //////////////////////
    // Case SRC
    //////////////////////
    treg = new URANIE::Sensitivity::TRegression(tds1, "x_{1}:x_{2}:x_{3}",
            "Ishigami", "SRC");
    treg->computeIndexes();

    TTree * tree = treg->getResultTuple();
    tree->SetBranchAddress("Value", &sobol);

    tree->GetEntry(0);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.1833568, sobol, EPSILON);
    tree->GetEntry(4);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.00, sobol, 0.1);
    tree->GetEntry(8);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.00, sobol, 0.1);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.19, treg->getR2(), EPSILON);

    //////////////////////
    // Case SRRC
    //////////////////////
    delete treg; treg=NULL;
    treg = new URANIE::Sensitivity::TRegression(tds1, "x_{1}:x_{2}:x_{3}",
            "Ishigami", "SRRC");
    treg->computeIndexes();

    tree = treg->getResultTuple();
    tree->SetBranchAddress("Value", &sobol);

    tree->GetEntry(0);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.18435501, sobol, EPSILON);
    tree->GetEntry(4);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.00, sobol, 0.1);
    tree->GetEntry(8);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.00, sobol, 0.1);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.19, treg->getR2(), EPSILON);
}


void TRegressionTest::testIshigamiRegressionWithDrawings()
{
    // Load the data base in the DataServer
    tds1->fileDataRead("Ishigami.dat");

    treg = new URANIE::Sensitivity::TRegression(tds1, "x_{1}:x_{2}:x_{3}", "Ishigami", "SRCSRRC");
    treg->computeIndexes();

    TCanvas *Can = new TCanvas("Can","Can",1);
    treg->drawIndexes("Ishigami", "", "nonewcanv,hist,first,total");
    treg->drawIndexes("Ishigami", "", "nonewcanv,hist,first,total");
    Can->Clear();
//    treg->drawIndexes("Ishigami", "", "nonewcanv,pie,first");
//    treg->drawIndexes("Ishigami", "", "nonewcanv,pie,first");
    delete Can;

    treg->drawIndexes("Ishigami", "", "hist,first,total");
    treg->drawIndexes("Ishigami", "", "hist,first,total");
    //  treg->drawIndexes("Ishigami", "", "pie,first");
    treg->drawIndexes("Ishigami", "", "pie,first");
    
}

static void constructor_function_long_output_throw_exception(const string& outputs) {
    vector<URANIE::DataServer::TUniformDistribution> inputs;
    TestTSensitivity::add_flowrate_attributes(inputs);

    URANIE::DataServer::TDataServer dataserver("dataserver", "");
    dataserver.keepFinalTuple(false);
    for (auto& input: inputs)
        dataserver.addAttribute(&input);

    string inputs_name;
    TList *list=dataserver.getListOfAttributes();
    for (const auto& attribute: *list)
        inputs_name += attribute->GetName() + string(":");
    inputs_name.pop_back();

    URANIE::Sampler::TSampling sampling(&dataserver);
    sampling.generateSample();

    URANIE::Launcher::TLauncherFunction launcher(&dataserver, &flowrateModel, "",
        outputs.data());
    launcher.setDrawProgressBar(false);
    launcher.run();

    CPPUNIT_ASSERT_THROW(URANIE::Sensitivity::TRegression regression(&dataserver,
        inputs_name.data(), outputs.data(), "SRCSRRC"),
        URANIE::Exceptions::UErrorExceptions);
    delete list;
}

void TRegressionTest::throwExceptionConstructorTooLongOutputFunctionFirst() {
    const string outputs = "voila_un_nom_beaucoup_trop_long_pour_un_seul_attribut:encore_un_p*tain_de_nom_beaucoup_trop_long_pour_un_seul_attribut";
    constructor_function_long_output_throw_exception(outputs);
}

void TRegressionTest::throwExceptionConstructorTooLongOutputFunctionSecond() {
    const string outputs = "voila_un_nom_beaucoup_trop_long_pour_un_attribut:encore_un_p*tain_de_nom_beaucoup_trop_long_pour_un_seul_attribut";
    constructor_function_long_output_throw_exception(outputs);
}

void TRegressionTest::throwExceptionConstructorTooLongOutputCode() {
    vector<URANIE::DataServer::TUniformDistribution> inputs;
    TestTSensitivity::add_flowrate_attributes(inputs);

    URANIE::DataServer::TDataServer dataserver("dataserver", "");
    for (auto& input: inputs)
        dataserver.addAttribute(&input);

    string inputs_name;
    TList *list=dataserver.getListOfAttributes();
    for (const auto& attribute: *list)
        inputs_name += attribute->GetName() + string(":");
    inputs_name.pop_back();

    URANIE::Sampler::TSampling sampling(&dataserver, "lhs", 100);
    sampling.generateSample();

    TString input_file = "flowrate_input_with_keys.in";
    TestTSensitivity::set_file_key_to_upper_first(dataserver, input_file);

    URANIE::Launcher::TOutputFileDataServer output_file("_output_flowrate_withRow_.dat");
    output_file.addAttribute("voila_un_nom_beaucoup_trop_long_pour_un_seul_attribut");

    URANIE::Launcher::TCode code(&dataserver, "");
    code.addOutputFile(&output_file);

    URANIE::Launcher::TLauncher launcher(&dataserver, &code);
    launcher.setDrawProgressBar(false);
    launcher.run("nointermediatesaved");

    CPPUNIT_ASSERT_THROW(URANIE::Sensitivity::TRegression regression(&dataserver,
        inputs_name.data(), code.getOutputVarNames(), "SRCSRRC"),
        URANIE::Exceptions::UErrorExceptions);
    delete list;
}

void TRegressionTest::throwExceptionConstructorTooLongOutputRun() {
    vector<URANIE::DataServer::TUniformDistribution> inputs;
    TestTSensitivity::add_flowrate_attributes(inputs);
    URANIE::DataServer::TAttribute output("voila_un_nom_beaucoup_trop_long_pour_un_seul_attribut_de_runner");

    URANIE::DataServer::TDataServer dataserver("dataserver", "");
    for (auto& input: inputs)
        dataserver.addAttribute(&input);

    URANIE::Sampler::TSampling sampling(&dataserver, "lhs", 100);
    sampling.generateSample();

    URANIE::Relauncher::TCJitEval eval(&flowrateModel);
    TList *list=dataserver.getListOfAttributes();
    for (const auto& attribute: *list)
        eval.addInput(dynamic_cast<URANIE::DataServer::TAttribute*>(attribute));
    eval.addOutput(&output);

    string inputs_name;
    for (const auto& input: *eval.getInputList())
        inputs_name += input->GetName() + string(":");
    inputs_name.pop_back();

    URANIE::Relauncher::TSequentialRun run(&eval);
    run.startSlave();
    if (run.onMaster()) {
        URANIE::Relauncher::TLauncher2 launcher(&dataserver, &run);
        launcher.solverLoop();
        run.stopSlave();

        CPPUNIT_ASSERT_EQUAL(1, eval.getOutputList()->GetEntries());
        CPPUNIT_ASSERT_THROW(URANIE::Sensitivity::TRegression regression(&dataserver,
            inputs_name.data(), eval.getOutputList()->First()->GetName(), "SRCSRRC"),
            URANIE::Exceptions::UErrorExceptions);
    }
    delete list;
}

void TRegressionTest::throwExceptionConstructorTooLongOutputFromData() {
    URANIE::DataServer::TDataServer dataserver("dataserver", "");
    dataserver.fileDataRead("flowrateUniformDesign.dat");
    TList *list=dataserver.getListOfAttributes();
    const string output_name = list->Last()->GetName();

    string inputs_name;
    for (const auto& attribute: *list)
        inputs_name += attribute->GetName() + string(":");
    const string output_name_w_sep = ":" + output_name + ":";
    inputs_name.erase(inputs_name.find(output_name_w_sep), output_name_w_sep.size());

    dataserver.addAttribute("voila_un_nom_beaucoup_trop_long_pour_un_seul_attribut", output_name);
    delete list; list=dataserver.getListOfAttributes();
    CPPUNIT_ASSERT_THROW(URANIE::Sensitivity::TRegression(&dataserver, inputs_name.data(),
        list->Last()->GetName(), "SRCSRRC"),
        URANIE::Exceptions::UErrorExceptions);
    delete list;
}

void TRegressionTest::throwExceptionTRegressionConstructorTdsNull()
{

    URANIE::DataServer::TDataServer *tdsnull =
            new URANIE::DataServer::TDataServer();

    CPPUNIT_ASSERT_THROW(
            new URANIE::Sensitivity::TRegression(tdsnull,
                    "rw:r:tu:tl:hu:hl:l:kw", "ystar", "SRRC"),
            URANIE::Exceptions::UErrorExceptions);

    delete tdsnull;
}
void TRegressionTest::throwExceptionTRegressionConstructorFalseInput()
{

    CPPUNIT_ASSERT_THROW(
            new URANIE::Sensitivity::TRegression(tds, "rw:r:tu:tl:hu:hl:l:toto",
                    "ystar", "SRRC"), URANIE::Exceptions::UErrorExceptions);
}
void TRegressionTest::throwExceptionTRegressionConstructorFalseOutput()
{

    CPPUNIT_ASSERT_THROW(
            new URANIE::Sensitivity::TRegression(tds, "rw:r:tu:tl:hu:hl:l:kw",
                    "toto", "SRRC"), URANIE::Exceptions::UErrorExceptions);
}
void TRegressionTest::throwExceptionTRegressionConstructorFalseOption()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_IGNORED_TEST()
    ;

    CPPUNIT_ASSERT_THROW(
            new URANIE::Sensitivity::TRegression(tds, "rw:r:tu:tl:hu:hl:l:kw",
                    "ystar", "toto"), URANIE::Exceptions::UErrorExceptions);
}

void TRegressionTest::testTRegressionsetMethod()
{
    treg = new URANIE::Sensitivity::TRegression(tds, "rw:r:tu:tl:hu:hl:l:kw",
            "ystar", "SRC");

    CPPUNIT_ASSERT_EQUAL(1, treg->getMethod());
    treg->setMethod(URANIE::Sensitivity::TRegression::kSRRC);
    CPPUNIT_ASSERT_EQUAL(3,treg->getMethod());
    treg->setMethod(URANIE::Sensitivity::TRegression::kPCC);
    CPPUNIT_ASSERT_EQUAL(7, treg->getMethod());
    treg->clearMethod();
    treg->setMethod(URANIE::Sensitivity::TRegression::kPRCC);
    CPPUNIT_ASSERT_EQUAL(8, treg->getMethod());
}
void TRegressionTest::throwExceptionTRegressioncomputeIndexesMoreOneOutput()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_IGNORED_TEST()
    ;

    tds1->fileDataRead("flowrateUniformDesign2D.dat");
    treg = new URANIE::Sensitivity::TRegression(tds1, "rw:r:tu:tl:hu:hl:l:kw",
            "ystar:y", "SRRC");

    CPPUNIT_ASSERT_THROW(treg->computeIndexes(),
            URANIE::Exceptions::UWarningExceptions);
}

