//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTSobol.cxx
 \author maud Salvoldi (OCC/Areva group)
 \date   $Date$
 
 \brief  Fichier de test de la classe URANIE::Sensitivity::TSobol
 
 Cr�e le : jeu d�c 22 12:30:15 CET 2011
 
 */

// Uranie
#include "TestTSobol.h"
#include "TSensitivity.h"
#include "TSobol.h"
#include "UDisableTest.h"
#include "TOutputFileKey.h"
#include "TOutputFileRow.h"
#include "TOutputFileDataServer.h"
#include "TCode.h"
#include "TUniformDistribution.h"
#include "TLauncher.h"
#include "TLauncherFunction.h"
#include "TLauncher2.h"
#include "TCJitEval.h"
#include "TSequentialRun.h"
#include "TThreadedRun.h"
#include "TestUserFunctions.h"
#include "code_paths.h"
#include "uranie_commands.h"

// ROOT
#include "TString.h"
#include "TMath.h"
#include "TTreePlayer.h"
#include "TCanvas.h"
#include "TROOT.h"
#ifdef ROOT5
#include "Api.h"
#endif

#define EPSILON 0.01
#define LVLE_EPSILON 0.001 
#define DBL_PRECISION_COMPARE 1.e-12

#include <iostream>
#include <iomanip>

TSobolTest::TSobolTest() :
        _tds1(NULL), _tds(NULL), _tds2(NULL), _tsobol(NULL), _mycode(NULL),
        _p2fIshigami(NULL), _p2fIshigamiDecaled(NULL), _fout(NULL)
{
#ifdef ROOT5
//   gROOT->LoadMacro("UserFunctions.C");
    G__ClassInfo globalscope;
    G__MethodInfo method;
    long dummy;
    method = globalscope.GetMethod("Ishigami", "double*,double*", &dummy);
    _p2fIshigami = method.PointerToFunc();
    method = globalscope.GetMethod("IshigamiDecaled", "double*,double*", &dummy);
    _p2fIshigamiDecaled = method.PointerToFunc();

#endif
}


TSobolTest::~TSobolTest()
{
    UranieTests::removeFolder(gSystem->pwd(),"URANIE");
    UranieTests::removeFile(gSystem->pwd(), "testSobolIshigami.out");
#ifdef WIN32
    UranieTests::removeFile(gSystem->pwd(), "testSobolIshigamiLhs.out");
#endif
    UranieTests::removeFile(gSystem->pwd(),"sobol_launching_5000_lhs_saltelli.root");
    UranieTests::removeFile(gSystem->pwd(), "testSobolIshigamiQmcHalton.out");
    UranieTests::removeFile(gSystem->pwd(), "testSobolIshigamiSobol.out");
    UranieTests::removeFile(gSystem->pwd(), "testSobolIshigamiQmcSobol.out");
    UranieTests::removeFile(gSystem->pwd(), "testSobolIshigamiSaltelli.out");
    UranieTests::removeFile(gSystem->pwd(), "_sobol_launching_.root");
    UranieTests::removeFile(gSystem->pwd(), "_sobol_launching_.dat");
    UranieTests::removeFile(gSystem->pwd(), "_sobol_sampling_.dat");
    UranieTests::removeFile(gSystem->pwd(), "_sample_sobol_sampling_.dat");
    UranieTests::removeFile(gSystem->pwd(), "_sample_from_sampling_.dat");
    UranieTests::removeFile(gSystem->pwd(), "_sample_sobol_launching_.dat");
    UranieTests::removeFile(gSystem->pwd(), "_sample_from_launching_.dat");
    UranieTests::removeFile(gSystem->pwd(), "_sample_from_launching_extra_att_.dat");
    UranieTests::removeFile(gSystem->pwd(), "_sample_from_sampling_.root");
    UranieTests::removeFile(gSystem->pwd(), "_sample_sobol_sampling_.root");
    UranieTests::removeFile(gSystem->pwd(), "_sample_from_launching_.root");
    UranieTests::removeFile(gSystem->pwd(), "_johnsonrw_sampling_.dat");
    UranieTests::removeFile(gSystem->pwd(), "_johnsonrw_launching_.dat");
    UranieTests::removeFile(gSystem->pwd(), "DGSM_NominalValues_output.root");
    UranieTests::removeFile(gSystem->pwd(), "_sample_sobol_launching_.root");
    UranieTests::removeFile(gSystem->pwd(), "_nooutput_.dat");
    UranieTests::removeFile(gSystem->pwd(), "_nooutput_.root");
    UranieTests::removeFile(gSystem->pwd(), "_missinginput_.dat");
    UranieTests::removeFile(gSystem->pwd(), "_missinginput_.root");
    UranieTests::removeFile(gSystem->pwd(), "outpuT.dat");
    UranieTests::removeFile(gSystem->pwd(), "outpuS.dat");
    UranieTests::removeFile(gSystem->pwd(), "_nosoboliter_.dat");
    UranieTests::removeFile(gSystem->pwd(), "_nosoboliter_.root");
    UranieTests::removeFile(gSystem->pwd(), "sample_45000_sobol_launching.root");
    UranieTests::removeFile(gSystem->pwd(), "sample_5000_sobol_launching.root");
    UranieTests::removeFile(gSystem->pwd(), "_out_.dat");
    UranieTests::removeFile(gSystem->pwd(), "_sample_45000_extra_att_.dat");
    UranieTests::removeFile(gSystem->pwd(), "_sample_45000_extra_att_.root");
    UranieTests::removeFile(gSystem->pwd(), "_sample_from_launching_extra_att_.root");
    UranieTests::removeFile(gSystem->pwd(), "_sample_sobol_launching_extra_att_.dat");
    UranieTests::removeFile(gSystem->pwd(), "_sample_sobol_launching_extra_att_.root");
}

void TSobolTest::setUp()
{
    _tds1 = new URANIE::DataServer::TDataServer("tds1", "my first TDS");

    _tds = new URANIE::DataServer::TDataServer("tdsflowreate",
            "DataBase flowreate");
    _tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("rw", 0.05, 0.15));
    _tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("r", 100.0, 50000.0));
    _tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("tu", 63070.0,
                    115600.0));
    _tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("tl", 63.1, 116.0));
    _tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("hu", 990.0, 1110.0));
    _tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("hl", 700.0, 820.0));
    _tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("l", 1120.0, 1680.0));
    _tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("kw", 9855.0,
                    12045.0));
    _tds->getAttribute("rw")->delShare();
    _tds->getAttribute("r")->delShare();
    _tds->getAttribute("tu")->delShare();
    _tds->getAttribute("tl")->delShare();
    _tds->getAttribute("hu")->delShare();
    _tds->getAttribute("hl")->delShare();
    _tds->getAttribute("l")->delShare();
    _tds->getAttribute("kw")->delShare();
    
    _tds2 = new URANIE::DataServer::TDataServer("tds",
            "TDS for Ishigami function");
    _tds2->addAttribute(
            new URANIE::DataServer::TUniformDistribution("x_{1}",
                    -1.0 * TMath::Pi(), TMath::Pi()));
    _tds2->addAttribute(
            new URANIE::DataServer::TUniformDistribution("x_{2}",
                    -1.0 * TMath::Pi(), TMath::Pi()));
    _tds2->addAttribute(
            new URANIE::DataServer::TUniformDistribution("x_{3}",
                    -1.0 * TMath::Pi(), TMath::Pi()));
    _tds2->getAttribute("x_{1}")->delShare();
    _tds2->getAttribute("x_{2}")->delShare();
    _tds2->getAttribute("x_{3}")->delShare();
}

void TSobolTest::tearDown()
{
    if (_fout)
    {
        delete _fout;
        _fout = NULL;
    }

    if (_mycode != NULL)
    {
        delete _mycode;
        _mycode = NULL;
    }

    if (_tsobol)
        delete _tsobol;
    _tsobol = nullptr;

    delete _tds1;
    _tds1 = NULL;

    delete _tds2;
    _tds2 = NULL;


    delete _tds;
    _tds = NULL;
}

void IshigamiOLD(double *paramt, Double_t *res)
{
    Double_t a = 7.0;
    Double_t b = 0.1;

    Double_t x1 = paramt[0];
    Double_t x2 = paramt[1];
    Double_t x3 = paramt[2];

    Double_t dsinx1 = TMath::Sin(x1);
    Double_t dsinx2 = TMath::Sin(x2);

    res[0] = dsinx1;
    res[0] += a * dsinx2 * dsinx2;
    res[0] += b * x3 * x3 * x3 * x3 * dsinx1;
}

////////////////////////////////
///                          ///                            
///         Tests            ///
///                          ///
////////////////////////////////

void TSobolTest::setUp1()
{
#ifdef ROOT5
    _tsobol = new URANIE::Sensitivity::TSobol(_tds2, _p2fIshigami, 
            50000);
#else
    _tsobol = new URANIE::Sensitivity::TSobol(_tds2, &Ishigami, "", "Ishi", 
            50000);
#endif
    _tsobol->setDrawProgressBar(kFALSE);
    _tsobol->setSeed(123);
}

void TSobolTest::setUp2(bool bad)
{
    TString sJDDReference = gSystem->pwd();
#ifdef WIN32
    sJDDReference += "\\";
#else
    sJDDReference += "/";
#endif  
    sJDDReference += "flowrate_input_with_values_rows.in";
    _tds->getAttribute("rw")->setFileKey(sJDDReference, "Rw", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("r")->setFileKey(sJDDReference, "R", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("tu")->setFileKey(sJDDReference, "Tu", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("tl")->setFileKey(sJDDReference, "Tl", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("hu")->setFileKey(sJDDReference, "Hu", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("hl")->setFileKey(sJDDReference, "Hl", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("l")->setFileKey(sJDDReference, "L", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("kw")->setFileKey(sJDDReference, "Kw", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);

    _fout = new URANIE::Launcher::TOutputFileDataServer(
            "_output_flowrate_withRow_.dat");
    _fout->addAttribute( ((bad)?"voila_un_nom_beaucoup_trop_long_pour_un_seul_attribut":"yhat") );

    _mycode = new URANIE::Launcher::TCode(_tds, UranieTests::code_path("flowrate") + " -s -r");
    _mycode->addOutputFile(_fout);

    _tsobol = new URANIE::Sensitivity::TSobol(_tds, _mycode, 2000,"sobol");
    _tsobol->setDrawProgressBar(kFALSE);
    _tsobol->setSeed(123);
}

void TSobolTest::setUp300()
{
    TString sJDDReference = gSystem->pwd();
#ifdef WIN32
    sJDDReference += "\\";
#else
    sJDDReference += "/";
#endif  
    sJDDReference += "flowrate_input_with_values_rows.in";
    _tds->getAttribute("rw")->setFileKey(sJDDReference, "Rw", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("r")->setFileKey(sJDDReference, "R", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("tu")->setFileKey(sJDDReference, "Tu", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("tl")->setFileKey(sJDDReference, "Tl", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("hu")->setFileKey(sJDDReference, "Hu", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("hl")->setFileKey(sJDDReference, "Hl", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("l")->setFileKey(sJDDReference, "L", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("kw")->setFileKey(sJDDReference, "Kw", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);

    _fout = new URANIE::Launcher::TOutputFileKey(
            "_output_flowrate_withKey_.dat");

    std::stringstream sstr;
    for(unsigned int io=0; io<300; io++)
    {
        sstr.str("");
        sstr << "ohlabellesortienumero_" << io;
        ((URANIE::Launcher::TOutputFileKey*)_fout)->addAttribute(sstr.str().c_str(),"yhat");
    }

    _mycode = new URANIE::Launcher::TCode(_tds, UranieTests::code_path("flowrate") + " -s -r");
    _mycode->addOutputFile(_fout);

    _tsobol = new URANIE::Sensitivity::TSobol(_tds, _mycode, 2000,"sobol");
    _tsobol->setDrawProgressBar(kFALSE);
    _tsobol->setSeed(123);
}

#ifdef WITH_LEVELE
/* =============================================
//             LevelE
================================================= */
void TSobolTest::checkLevelEResults(TTree *ntresu, vector<int> *elmts)
{
  // Results reference
  double first[26]={0,0.239374007768,0.314962225821,0.29437577316,0.188023248086,0.150112529829,0.091118149347,0.0845780873583,0.14568756285,0.0454668420748,0,0,0,
		    0,0,0,0.00045778432195,0.00260509116114,0.0145181588182,0.0103323951952,0.0199936310571,0.125396495767,0.414779945306,0.516082399315,0.417284042858,0.235124559715};
  double first_min[26]={0,0.139174788751,0.218441204919,0.196726045592,0.0860478596174,0.0471944730994,0,0,0.0426796827514,0,0,0,0,0,0,0,0,0,0,0,0,0.0220304597426,
			0.325104771392,0.435739083644,0.327810271324,0.134756356431};
  double first_max[26]={0.103652380125,0.334721386083,0.405380323034,0.386242822773,0.286099812603,0.24987694439,0.192948208306,0.186594644471,0.245630707401,0.148419757522,
			0.103652380125,0.103652380125,0.103652380125,0.103652380125,0.103652380125,0.103652380125,0.104105224609,0.106228786975,0.117992978188,0.113862830807,
			0.123390298676,0.226109978604,0.497062167188,0.588266538084,0.499338780264,0.330716963876};
  double total[26]={0.426218678817,0.539905994919,0.427543328916,0.483559351715,0.404261843015,0.439759974141,0.465617226733,0.451567052171,0.480887552066,0.574785187669,
		    0.651277825573,0.311192677911,0.234006475602,0.284675786038,0.35378508487,0.398935138764,0.423104135918,0.4371509656,0.490314540523,0.501069470978,0.313202955884,
		    0.359351361115,0.536055939849,0.511657798992,0.486256904668,0.472553819479};
  double total_min[26]={0.360594191611,0.461914810684,0.361761769317,0.411414150116,0.341284907671,0.372544091708,0.395450851938,0.382989486401,0.409033451319,0.493458389611,
			0.563406503073,0.260348817065,0.194322512795,0.2375549387,0.297206985464,0.336612967186,0.357850152256,0.370239232972,0.417438926566,0.427047521289,
			0.262081618896,0.302046418443,0.458446285441,0.436527190562,0.413819066419,0.401615736537};
  double total_max[26]={0.500142642636,0.625708472466,0.501623812126,0.563865310003,0.475528262266,0.515263609231,0.54401243447,0.528411322979,0.560913921861,0.663611399204,
			0.745739744141,0.369854891192,0.28053549504,0.339343772278,0.418486707966,0.469538831412,0.49665838397,0.512353750423,0.571319692419,0.583165093503,
			0.372160612599,0.424808274944,0.621507005931,0.594799824954,0.566843394688,0.551697003153};

  double *data, *themin, *themax;
  int npts = ntresu->Draw("CILower:CIUpper:Sobol", "Inp==\"w\" && Order==\"First\" &&  Algo==\"martinez11\" " ,"goff");
  data=ntresu->GetV3();
  themin=ntresu->GetV1(); 
  themax=ntresu->GetV2();
  for(int ival=0; ival<npts; ival++)
    {
      CPPUNIT_ASSERT_DOUBLES_EQUAL(first[((elmts==NULL) ? ival : elmts->at(ival))], data[ival], LVLE_EPSILON);
      CPPUNIT_ASSERT_DOUBLES_EQUAL(first_min[((elmts==NULL) ? ival : elmts->at(ival))], themin[ival], LVLE_EPSILON);
      CPPUNIT_ASSERT_DOUBLES_EQUAL(first_max[((elmts==NULL) ? ival : elmts->at(ival))], themax[ival], LVLE_EPSILON);
    }

  npts = ntresu->Draw("CILower:CIUpper:Sobol", "Inp==\"w\" && Order==\"Total\" &&  Algo==\"martinez11\" " ,"goff");
  data=ntresu->GetV3();
  themin=ntresu->GetV1(); 
  themax=ntresu->GetV2();
  for(int ival=0; ival<npts; ival++)
    {
      CPPUNIT_ASSERT_DOUBLES_EQUAL(total[((elmts==NULL) ? ival : elmts->at(ival))], data[ival], LVLE_EPSILON);
      CPPUNIT_ASSERT_DOUBLES_EQUAL(total_min[((elmts==NULL) ? ival : elmts->at(ival))], themin[ival], LVLE_EPSILON);
      CPPUNIT_ASSERT_DOUBLES_EQUAL(total_max[((elmts==NULL) ? ival : elmts->at(ival))], themax[ival], LVLE_EPSILON);
    }  
}

void TSobolTest::testLevelESobol()
{
  delete _tds1; _tds1 = new URANIE::DataServer::TDataServer("tdsLevelE","pouet");
  _tds1->fileDataRead("sobol_launching_5000_lhs_saltelli.dat");
  _tsobol = new URANIE::Sensitivity::TSobol(_tds1, "t:kl:kc:v1:l1:r1:rc1:v2:l2:r2:rc2:w","y");
  _tsobol->computeIndexes("nointermed,saltelli");

  TTree *ntresu = (TTree*)_tsobol->getResultTuple();
  checkLevelEResults(ntresu);
  
}

void TSobolTest::testLevelESobolOutputRestrVarexp()
{
  delete _tds1; _tds1 = new URANIE::DataServer::TDataServer("tdsLevelE","pouet");
 _tds1->fileDataRead("sobol_launching_5000_lhs_saltelli.dat");
 _tsobol = new URANIE::Sensitivity::TSobol(_tds1, "t:kl:kc:v1:l1:r1:rc1:v2:l2:r2:rc2:w","y[0]:y[5]:y[22]");
 _tsobol->computeIndexes("nointermed,saltelli");

  TTree *ntresu = (TTree*)_tsobol->getResultTuple();
  checkLevelEResults(ntresu, _tsobol->getAttributeElements(string("y")));
  
}

void TSobolTest::testLevelESobolOutputRestrOption()
{
   delete _tds1; _tds1 = new URANIE::DataServer::TDataServer("tdsLevelE","pouet");
 _tds1->fileDataRead("sobol_launching_5000_lhs_saltelli.dat");
 _tsobol = new URANIE::Sensitivity::TSobol(_tds1, "t:kl:kc:v1:l1:r1:rc1:v2:l2:r2:rc2:w","y","output=y[8]:y[15]:y[22]");
 _tsobol->computeIndexes("nointermed,saltelli");

  TTree *ntresu = (TTree*)_tsobol->getResultTuple();
  checkLevelEResults(ntresu, _tsobol->getAttributeElements(string("y")));
  
}

void TSobolTest::testLevelESobolOutputRestrBoth()
{
   delete _tds1; _tds1 = new URANIE::DataServer::TDataServer("tdsLevelE","pouet");
 _tds1->fileDataRead("sobol_launching_5000_lhs_saltelli.dat");
 _tsobol = new URANIE::Sensitivity::TSobol(_tds1, "t:kl:kc:v1:l1:r1:rc1:v2:l2:r2:rc2:w","y[0]:y[5]:y[12]");
 _tsobol->computeIndexes("nointermed,saltelli,output=y[8]:y[15]:y[22]");

  TTree *ntresu = (TTree*)_tsobol->getResultTuple();
  checkLevelEResults(ntresu, _tsobol->getAttributeElements(string("y")));
  
}
#endif

void TSobolTest::testIshigamiSobol()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_IGNORED_TEST()
    ;

    cout << endl;

    gROOT->Macro("testSobolIshigami.C");

    Double_t dvalx1f, dvalx1t, dvalx2f, dvalx2t, dvalx3f, dvalx3t;
    std::ifstream outFile("testSobolIshigami.out", ios::in);

    outFile >> dvalx1f >> dvalx1t >> dvalx2f >> dvalx2t >> dvalx3f >> dvalx3t;

    //Close the file
    outFile.close();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx1f, 0.25, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx1t, 0.55, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx2f, 0.4, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx2t, 0.4, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx3f, 0.0, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx3t, 0.2, 0.1);

    // from an ASCII File
    // Load the data base in the DataServer
    _tds1->fileDataRead("_sobol_launching_.dat");

    _tsobol = new URANIE::Sensitivity::TSobol(_tds1, "x_{1}:x_{2}:x_{3}",
            "Ishigami");
    _tsobol->setDrawProgressBar(kFALSE);
    _tsobol->computeIndexes();

    Double_t sobol;
    TTree *tree = _tsobol->getResultTuple();

    tree->SetBranchAddress("Value", &sobol);

    // First Order
    tree->Draw(">>uranie_list_first", "Algorithm==\"--first--\"");
    TEventList *telfirst = (TEventList*) gDirectory->Get("uranie_list_first");

    // Total Order
    tree->Draw(">>uranie_list_total", "Algorithm==\"--total--\"");
    TEventList *teltotal = (TEventList*) gDirectory->Get("uranie_list_total");

    tree->GetEntry(telfirst->GetEntry(0));
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx1f, sobol, 0.00001);
    tree->GetEntry(teltotal->GetEntry(0));
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx1t, sobol, 0.00001);
    tree->GetEntry(telfirst->GetEntry(1));
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx2f, sobol, 0.00001);
    tree->GetEntry(teltotal->GetEntry(1));
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx2t, sobol, 0.00001);
    tree->GetEntry(telfirst->GetEntry(2));
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx3f, sobol, 0.00001);
    tree->GetEntry(teltotal->GetEntry(2));
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx3t, sobol, 0.00001);

    gROOT->Delete("uranie_list_*");

}

void TSobolTest::testIshigamiSobolWithDrawings()
{

#ifdef ROOT5
    _tsobol = new URANIE::Sensitivity::TSobol(_tds2, _p2fIshigami, 1000,"","","sobol");    
#else
    _tsobol = new URANIE::Sensitivity::TSobol(_tds2, &Ishigami,"","Ishi", 1000,"sobol");    
#endif
    _tsobol->setDrawProgressBar(kFALSE);
    _tsobol->setSeed(123);
    _tsobol->computeIndexes();

    TCanvas *Can = new TCanvas("Can","Can",1);
    _tsobol->drawIndexes("Ishigami", "", "nonewcanv,hist,first,total");
    _tsobol->drawIndexes("Ishigami", "", "nonewcanv,hist,first,total");
    Can->Clear();
//    _tsobol->drawIndexes("Ishigami", "", "nonewcanv,pie,first");
//    _tsobol->drawIndexes("Ishigami", "", "nonewcanv,pie,first");
    delete Can;

    _tsobol->drawIndexes("Ishigami", "", "hist,first,total");
    _tsobol->drawIndexes("Ishigami", "", "hist,first,total");
//    _tsobol->drawIndexes("Ishigami", "", "pie,first");
    _tsobol->drawIndexes("Ishigami", "", "pie,first");

}

void TSobolTest::testIshigamiSobolWithData()
{
    setUpIshi();
    int ns=5000;
    
    delete _tds2; _tds2 = nullptr;

    _tds2 = new URANIE::DataServer::TDataServer("tds","pouet"); _tds2->keepFinalTuple(false);
    _tds2->fileDataRead("_sample_from_launching_.dat",false);

    _tsobol = new URANIE::Sensitivity::TSobol(_tds2, Ishigami, "x_{1}:x_{2}:x_{3}", "FctIshigami", ns, "WithData");
    _tsobol->setDrawProgressBar(kFALSE);
    _tsobol->computeIndexes();
    _tds2->exportData("_out_.dat");
    validateValuesIshigami(true);

    delete _tsobol; _tsobol = nullptr;
    delete _tds2; _tds2 = nullptr;

    _tds2 = new URANIE::DataServer::TDataServer("tds","pouet"); _tds2->keepFinalTuple(false);
    _tds2->fileDataRead("_sample_from_launching_extra_att_.dat",false);

    _tsobol = new URANIE::Sensitivity::TSobol(_tds2, Ishigami, "x_{1}:x_{2}:x_{3}", "FctIshigami", ns, "WithData");
    _tsobol->setDrawProgressBar(kFALSE);
    _tsobol->computeIndexes();
    _tds2->exportData("_out_.dat");
    validateValuesIshigami(true);
    
    delete _tsobol; _tsobol = nullptr;
    delete _tds2; _tds2 = nullptr;

    _tds2 = new URANIE::DataServer::TDataServer("tds","pouet"); _tds2->keepFinalTuple(false);
    _tds2->fileDataRead("_sample_from_launching_.dat",false);
    
    URANIE::DataServer::TAttribute *out = new URANIE::DataServer::TAttribute("FctIshigami");
    out->delShare();
    
    URANIE::Relauncher::TCJitEval eval(&Ishigami);
    eval.addInput(_tds2->getAttribute("x_{1}"));
    eval.addInput(_tds2->getAttribute("x_{2}"));
    eval.addInput(_tds2->getAttribute("x_{3}"));
    eval.addOutput(out);
    
    URANIE::Relauncher::TSequentialRun run(&eval);
    run.startSlave();
    if( run.onMaster() )
    {
        _tsobol = new URANIE::Sensitivity::TSobol(_tds2, &run, ns, "WithData");
        _tsobol->setDrawProgressBar(kFALSE);
        _tsobol->computeIndexes();    
        validateValuesIshigami(true);    
        run.stopSlave();
    }

    
}

void TSobolTest::testIshigamiSobolLoadingData()
{
    int ns=5000;
    
    delete _tds2; _tds2 = nullptr;

    _tds2 = new URANIE::DataServer::TDataServer("tdsishigami","pouet");
    _tds2->fileDataRead("sample_5000_sobol_launching.dat");
    
    _tsobol = new URANIE::Sensitivity::TSobol(_tds2, "x_{1}:x_{2}:x_{3}", "FctIshigami");
    _tsobol->loadOtherSobolFile("sample_45000_sobol_launching.dat");
    _tsobol->computeIndexes();
    _tds2->exportData("_out_.dat");
    validateValuesIshigami();

}

void TSobolTest::testIshigamiSobolLoadingDataExtraAtt()
{
    int ns=5000;
    
    delete _tds2; _tds2 = nullptr;

    _tds2 = new URANIE::DataServer::TDataServer("tdsishigami","pouet");
    _tds2->fileDataRead("sample_5000_sobol_launching.dat");
    _tds2->computeRank();
    
    _tsobol = new URANIE::Sensitivity::TSobol(_tds2, "x_{1}:x_{2}:x_{3}", "FctIshigami");
    _tsobol->loadOtherSobolFile("sample_45000_sobol_launching.dat");
    _tsobol->computeIndexes();
    _tds2->exportData("_out_.dat");
    validateValuesIshigami();

    delete _tsobol; _tsobol = nullptr;
    delete _tds2; _tds2 = nullptr;

    _tds2 = new URANIE::DataServer::TDataServer("tdsishigami","pouet");
    _tds2->fileDataRead("sample_45000_sobol_launching.dat");
    _tds2->computeRank();
    _tds2->exportData("_sample_45000_extra_att_.dat","x_{1}:Rk_x_{1}:x_{2}:Rk_x_{2}:x_{3}:FctIshigami:tdsishigami__n__iter__:sobol__n__iter__tdsishigami");
    
    delete _tds2; _tds2 = nullptr;

    _tds2 = new URANIE::DataServer::TDataServer("tdsishigami","pouet");
    _tds2->fileDataRead("sample_5000_sobol_launching.dat");
    
    _tsobol = new URANIE::Sensitivity::TSobol(_tds2, "x_{1}:x_{2}:x_{3}", "FctIshigami");
    _tsobol->loadOtherSobolFile("_sample_45000_extra_att_.dat");
    _tsobol->computeIndexes();
    _tds2->exportData("_out_.dat");
    validateValuesIshigami();

}

void TSobolTest::testIshigamiSobolFromExistingFiles()
{
    setUpIshi();
    int ns=5000;
    
    delete _tds2; _tds2 = nullptr;
    
    _tds2 = new URANIE::DataServer::TDataServer("tds","pouet");
    _tds2->fileDataRead("_sample_sobol_launching_.dat");
    
    _tsobol = new URANIE::Sensitivity::TSobol(_tds2, "x_{1}:x_{2}:x_{3}", "FctIshigami");
    _tsobol->computeIndexes();
    _tds2->exportData("_out_.dat");
    validateValuesIshigami(true);

    delete _tsobol; _tsobol = nullptr;
    delete _tds2; _tds2 = nullptr;
    
    _tds2 = new URANIE::DataServer::TDataServer("tds","pouet");
    _tds2->fileDataRead("_sample_sobol_launching_extra_att_.dat");
    
    _tsobol = new URANIE::Sensitivity::TSobol(_tds2, "x_{1}:x_{2}:x_{3}", "FctIshigami");
    _tsobol->computeIndexes();
    _tds2->exportData("_out_.dat");
    validateValuesIshigami(true);
    
    delete _tsobol; _tsobol = nullptr;
    delete _tds2; _tds2 = nullptr;

    _tds2 = new URANIE::DataServer::TDataServer("tds","pouet");
    _tds2->fileDataRead("_sample_sobol_launching_.dat");
    
    _tsobol = new URANIE::Sensitivity::TSobol(_tds2, Ishigami, "x_{1}:x_{2}:x_{3}", "FctIshigami", ns);
    _tsobol->computeIndexes();
    _tds2->exportData("_out_.dat");
    validateValuesIshigami(true);
    
    delete _tsobol; _tsobol = nullptr;
    delete _tds2; _tds2 = nullptr;

    _tds2 = new URANIE::DataServer::TDataServer("tds","pouet");
    _tds2->fileDataRead("_sample_sobol_launching_.dat");
    
    URANIE::DataServer::TAttribute *out = new URANIE::DataServer::TAttribute("FctIshigami");
    out->delShare();
    
    URANIE::Relauncher::TCJitEval eval(&Ishigami);
    eval.addInput(_tds2->getAttribute("x_{1}"));
    eval.addInput(_tds2->getAttribute("x_{2}"));
    eval.addInput(_tds2->getAttribute("x_{3}"));
    eval.addOutput(out);
    
    URANIE::Relauncher::TSequentialRun run(&eval);
    run.startSlave();
    if( run.onMaster() )
    {
        _tsobol = new URANIE::Sensitivity::TSobol(_tds2, &run, ns);
        _tsobol->setDrawProgressBar(kFALSE);
        _tsobol->computeIndexes();    
        validateValuesIshigami(true);    
        run.stopSlave();
    }
        
}

void TSobolTest::testTSobolConstructorwithTCode()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_IGNORED_TEST()
    ;

    TString sJDDReference = gSystem->pwd();
#ifdef WIN32
    sJDDReference += "\\";
#else
    sJDDReference += "/";
#endif  
    sJDDReference += "flowrate_input_with_keys.in";
    _tds->getAttribute("rw")->setFileKey(sJDDReference, "Rw");
    _tds->getAttribute("r")->setFileKey(sJDDReference, "R");
    _tds->getAttribute("tu")->setFileKey(sJDDReference, "Tu");
    _tds->getAttribute("tl")->setFileKey(sJDDReference, "Tl");
    _tds->getAttribute("hu")->setFileKey(sJDDReference, "Hu");
    _tds->getAttribute("hl")->setFileKey(sJDDReference, "Hl");
    _tds->getAttribute("l")->setFileKey(sJDDReference, "L");
    _tds->getAttribute("kw")->setFileKey(sJDDReference, "Kw");

    _fout = new URANIE::Launcher::TOutputFileRow(
            "_output_flowrate_withRow_.dat");
    _fout->addAttribute("yhat");

    URANIE::Launcher::TCode *mycode = new URANIE::Launcher::TCode(_tds,
        UranieTests::code_path("flowrate") + " -s -k");
    mycode->addOutputFile(_fout);

    Int_t ns = 500;
    _tsobol = new URANIE::Sensitivity::TSobol(_tds, mycode, ns);
    _tsobol->setDrawProgressBar(kFALSE);
    _tsobol->computeIndexes("nointermed");
    Double_t sobol;
    TTree *tree = _tsobol->getResultTuple();
    tree->SetBranchAddress("Value", &sobol);
    Int_t npat = (Int_t)(tree->GetEntries());

    // from an ASCII File
    // Load the data base in the DataServer
    _tds1->fileDataRead("_sobol_launching_.dat");

    URANIE::Sensitivity::TSobol *_tsobol2 = new URANIE::Sensitivity::TSobol(
            _tds1, "rw:r:tu:tl:hu:hl:l:kw", "yhat");
    _tsobol2->setDrawProgressBar(kFALSE);
    _tsobol2->computeIndexes();
    Double_t sobol2;
    TTree *tree2 = _tsobol2->getResultTuple();
    tree2->SetBranchAddress("Value", &sobol2);

    Int_t npat2 = (Int_t)(tree2->GetEntries());

    for (Int_t i = 0; i < npat; i++)
    {
        tree2->GetEntry(i);
        tree->GetEntry(i);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(sobol2, sobol, EPSILON);
    }
    CPPUNIT_ASSERT_EQUAL(npat2, npat);

    delete mycode;
    delete _tsobol2;

}

void TSobolTest::throwExceptionTSobolConstructorTdsNull()
{

    URANIE::DataServer::TDataServer *tdsnull =
            new URANIE::DataServer::TDataServer();
    Int_t nS = 2000;
#ifdef ROOT5
	CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TSobol(tdsnull, (void*)&IshigamiOLD, nS,
		"", "", "qmcsobol, sobol"), URANIE::Exceptions::UErrorExceptions);
#else
	CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TSobol(tdsnull, &IshigamiOLD,
		"", "IshiO", nS, "qmcsobol, sobol"), URANIE::Exceptions::UErrorExceptions);
#endif


#ifdef ROOT5
	CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TSobol(NULL, (void*)&IshigamiOLD, nS,
		"", "", "qmcsobol, sobol"), URANIE::Exceptions::UErrorExceptions);
#else
	CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TSobol(NULL, &IshigamiOLD,
		"", "IshiO", nS, "qmcsobol, sobol"), URANIE::Exceptions::UErrorExceptions);
#endif

    delete tdsnull;
}

void TSobolTest::throwExceptionTSobolConstructorFalseFnct()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_IGNORED_TEST()
    ;

//  CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TSobol(_tds2, (void*) &Ishigami, nS,"","","qmcsobol, sobol"),URANIE::Exceptions::UErrorExceptions);

}

void TSobolTest::throwExceptionTSobolConstructorFalseCode()
{

    TString sJDDReference = gSystem->pwd();
#ifdef WIN32
    sJDDReference += "\\";
#else
    sJDDReference += "/";
#endif  
    sJDDReference += "flowrate_input_with_keys.in";
    _tds->getAttribute("rw")->setFileKey(sJDDReference, "Rw");
    _tds->getAttribute("r")->setFileKey(sJDDReference, "R");
    _tds->getAttribute("tu")->setFileKey(sJDDReference, "Tu");
    _tds->getAttribute("tl")->setFileKey(sJDDReference, "Tl");
    _tds->getAttribute("hu")->setFileKey(sJDDReference, "Hu");
    _tds->getAttribute("hl")->setFileKey(sJDDReference, "Hl");
    _tds->getAttribute("l")->setFileKey(sJDDReference, "L");
    _tds->getAttribute("kw")->setFileKey(sJDDReference, "Kw");

    _fout = new URANIE::Launcher::TOutputFileRow(
            "_output_flowrate_withRow_.dat");
    _fout->addAttribute("yhat");

    URANIE::Launcher::TCode *code = new URANIE::Launcher::TCode();

    Int_t ns = 500;
    CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TSobol(_tds, code, ns),
            URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TSobol(_tds, (URANIE::Launcher::TCode *)NULL, ns),
            URANIE::Exceptions::UErrorExceptions);

    delete code;
}

void TSobolTest::throwExceptionTSobolConstructorFalseInput()
{
    Int_t ns = 2000;
#ifdef ROOT5
	CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TSobol(_tds2, _p2fIshigami, ns,
		"x_{1}:x_{2}:toto", "", "qmcsobol, sobol"), URANIE::Exceptions::UErrorExceptions);
#else
	CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TSobol(_tds2, &IshigamiOLD,
		"x_{1}:x_{2}:toto", "Ishi", ns, "qmcsobol, sobol"), URANIE::Exceptions::UErrorExceptions);
#endif
}

void TSobolTest::throwExceptionTSobolConstructorFalseOutput()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_IGNORED_TEST() ;

    Int_t ns = 2000;
#ifdef ROOT5
	CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TSobol(_tds2, _p2fIshigami, ns,
		"x_{1}:x_{2}:x_{3}", "a:b:c", "sobol"), URANIE::Exceptions::UErrorExceptions);
#else
	CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TSobol(_tds2, &IshigamiOLD,
		"x_{1}:x_{2}:x_{3}", "a:b:c", ns, "sobol"), URANIE::Exceptions::UErrorExceptions);
#endif
}

void TSobolTest::throwExceptionTSobolConstructorTooLongOutput()
{

    Int_t ns = 2000;
    
    URANIE::Sensitivity::TSobol *localSob=NULL;
    CPPUNIT_ASSERT_THROW(localSob=new URANIE::Sensitivity::TSobol(_tds2, &IshigamiOLD, "x_{1}:x_{2}:x_{3}", "voila_un_nom_beaucoup_trop_long_pour_un_seul_attribut:encore_un_putain_de_nom_beaucoup_trop_long_pour_un_seul_attribut", ns), URANIE::Exceptions::UErrorExceptions);
    delete localSob; localSob=NULL;

    CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TSobol(_tds2, &IshigamiOLD, "x_{1}:x_{2}:x_{3}", "voila_un_nom_beaucoup_trop_long_pour_un_attribut:encore_un_putain_de_nom_beaucoup_trop_long_pour_un_seul_attribut", ns), URANIE::Exceptions::UErrorExceptions);

    CPPUNIT_ASSERT_THROW( setUp2(true), URANIE::Exceptions::UErrorExceptions);

    URANIE::DataServer::TAttribute out("voila_un_nom_beaucoup_trop_long_pour_un_seul_attribut_de_runner");
    out.delShare();
    
    URANIE::Relauncher::TCJitEval eval(&Ishigami);
    eval.addInput(_tds2->getAttribute("x_{1}"));
    eval.addInput(_tds2->getAttribute("x_{2}"));
    eval.addInput(_tds2->getAttribute("x_{3}"));
    eval.addOutput(&out);
    
    URANIE::Relauncher::TSequentialRun run(&eval);
    run.startSlave();
    if( run.onMaster() )
    {
        CPPUNIT_ASSERT_THROW( _tsobol = new URANIE::Sensitivity::TSobol(_tds2, &run, 50000), URANIE::Exceptions::UErrorExceptions);
    
        run.stopSlave();
    }

    
}

//void  TSobolTest::throwExceptionTSobolConstructorFalseSize()
//{
//  // THIS TEST HAS BEEN DISABLED
//  // TODO: fix me
//  URANIE_IGNORED_TEST();
//
//  Int_t ns = 0.0;
//  CPPUNIT_ASSERT_THROW( new  URANIE::Sensitivity::TSobol(_tds2, (void*) &IshigamiOLD, ns,"","","sobol"),URANIE::Exceptions::UErrorExceptions);
//}
void TSobolTest::throwExceptionTSobolConstructorFalseOption()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_IGNORED_TEST()
    ;

    Int_t ns = 2000;
#ifdef ROOT5
	CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TSobol(_tds2, (void*)&IshigamiOLD, ns, "",
		"", "toto"), URANIE::Exceptions::UErrorExceptions);
#else
	CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TSobol(_tds2, &IshigamiOLD, "", "Ishi", ns,
		"toto"), URANIE::Exceptions::UErrorExceptions);
#endif
}

void TSobolTest::setUpIshi()
{
    int ns=5000;
    _tsobol = new URANIE::Sensitivity::TSobol(_tds2, Ishigami, "x_{1}:x_{2}:x_{3}", "FctIshigami", ns);
    _tsobol->setDrawProgressBar(kFALSE);
    _tsobol->setSeed(123);
    
    _tsobol->generateSample();
    _tds2->exportData("_sample_from_sampling_.dat","x_{1}:x_{2}:x_{3}:tds__n__iter__",Form("%s<%d",_tds2->getIteratorName(),2*ns/5));
    _tsobol->computeIndexes();
    _tds2->exportData("_sample_from_launching_.dat","x_{1}:x_{2}:x_{3}:FctIshigami:tds__n__iter__",Form("%s<%d",_tds2->getIteratorName(),2*ns/5)); 
    _tds2->computeRank();
    _tds2->exportData("_sample_from_launching_extra_att_.dat","x_{1}:Rk_x_{1}:x_{2}:Rk_x_{2}:x_{3}:FctIshigami:tds__n__iter__",Form("%s<%d",_tds2->getIteratorName(),2*ns/5));
    _tds2->exportData("_sample_sobol_launching_extra_att_.dat","x_{1}:Rk_x_{1}:x_{2}:Rk_x_{2}:x_{3}:FctIshigami:sobol__n__iter__tds:tds__n__iter__");
    _tsobol->getResultTuple()->Scan("*"," (Algo==\"--first--\" || Algo==\"--total--\") && Inp!=\"__sum__\"");

    gSystem->Rename("_sobol_sampling_.dat", "_sample_sobol_sampling_.dat");
    gSystem->Rename("_sobol_launching_.dat", "_sample_sobol_launching_.dat");
    
    delete _tsobol; _tsobol = nullptr;
}

void TSobolTest::throwExceptionTSobolConstructorNoFunction()
{
    setUpIshi();

    delete _tds2; _tds2 = nullptr;

    _tds2 = new URANIE::DataServer::TDataServer("tds","pouet");
    _tds2->fileDataRead("_sample_from_sampling_.dat");
    CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TSobol(_tds2, "x_{1}:x_{2}:x_{3}", "FctIshigami"), URANIE::Exceptions::UErrorExceptions);

    delete _tds2; _tds2 = nullptr;

    _tds2 = new URANIE::DataServer::TDataServer("tds","pouet");
    _tds2->fileDataRead("_sample_sobol_sampling_.dat");
    CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TSobol(_tds2, "x_{1}:x_{2}:x_{3}", "FctIshigami"), URANIE::Exceptions::UErrorExceptions);

    delete _tds2; _tds2 = nullptr;

    _tds2 = new URANIE::DataServer::TDataServer("tds","pouet");
    _tds2->fileDataRead("_sample_from_launching_.dat");
    CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TSobol(_tds2, "x_{1}:x_{2}:x_{3}", "FctIshigami"), URANIE::Exceptions::UErrorExceptions);

}

void TSobolTest::throwExceptionTSobolConstructorWithData()
{
    
    setUpIshi();
    int ns=5000;
    
    delete _tds2; _tds2 = nullptr;

    _tds2 = new URANIE::DataServer::TDataServer("tds","pouet");
    _tds2->fileDataRead("_sample_from_sampling_.dat");
    CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TSobol(_tds2, Ishigami, "x_{1}:x_{2}:x_{3}", "FctIshigami", ns, "WithData"), URANIE::Exceptions::UErrorExceptions);

    delete _tds2; _tds2 = nullptr;

    _tds2 = new URANIE::DataServer::TDataServer("tds","pouet");
    _tds2->fileDataRead("_sample_sobol_sampling_.dat");
    CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TSobol(_tds2, Ishigami, "x_{1}:x_{2}:x_{3}", "FctIshigami", ns, "WithData"), URANIE::Exceptions::UErrorExceptions);

    delete _tds2; _tds2 = nullptr;

    _tds2 = new URANIE::DataServer::TDataServer("tds","pouet");
    _tds2->fileDataRead("_sample_sobol_launching_.dat");
    CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TSobol(_tds2, Ishigami, "x_{1}:x_{2}:x_{3}", "FctIshigami", ns, "WithData"), URANIE::Exceptions::UErrorExceptions);
        
}

void TSobolTest::throwExceptionTSobolConstructorLoadingData()
{    
    int ns=5000;
    
    delete _tds2; _tds2 = nullptr;

    _tds2 = new URANIE::DataServer::TDataServer("tdsishigami","pouet");
    _tds2->fileDataRead("sample_5000_sobol_launching.dat");
    
    _tsobol = new URANIE::Sensitivity::TSobol(_tds2, "x_{1}:x_{2}:x_{3}", "FctIshigami");

    URANIE::DataServer::TDataServer youhou("youhou","youhou");
    youhou.fileDataRead("sample_45000_sobol_launching.dat",false);
    youhou.exportData("_nooutput_.dat","x_{1}:x_{2}:x_{3}:tdsishigami__n__iter__:sobol__n__iter__tdsishigami");    
    CPPUNIT_ASSERT_THROW(_tsobol->loadOtherSobolFile("_nooutput_.dat"), URANIE::Exceptions::UErrorExceptions);

    youhou.exportData("_missinginput_.dat","x_{1}:x_{3}:FctIshigami:tdsishigami__n__iter__:sobol__n__iter__tdsishigami");    
    CPPUNIT_ASSERT_THROW(_tsobol->loadOtherSobolFile("_missinginput_.dat"), URANIE::Exceptions::UErrorExceptions);

    youhou.exportData("_nosoboliter_.dat","x_{1}:x_{2}:x_{3}:FctIshigami:tdsishigami__n__iter__");    
    CPPUNIT_ASSERT_THROW(_tsobol->loadOtherSobolFile("_nosoboliter_.dat"), URANIE::Exceptions::UErrorExceptions);

    CPPUNIT_ASSERT_THROW(_tsobol->loadOtherSobolFile("_pouetpouetpouet_.dat"), URANIE::Exceptions::UErrorExceptions);

    _tsobol->loadOtherSobolFile("sample_45000_sobol_launching.dat");
    CPPUNIT_ASSERT_THROW(_tsobol->loadOtherSobolFile("_pouetpouetpouet_.dat"), URANIE::Exceptions::UErrorExceptions);

}

void TSobolTest::throwExceptionTSobolConstructorFromExistingFile()
{
    setUpIshi();
    int ns=5000;
    
    delete _tds2; _tds2 = nullptr;

    _tds2 = new URANIE::DataServer::TDataServer("tds","pouet");
    _tds2->fileDataRead("_sample_from_launching_.dat");
    CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TSobol(_tds2, "x_{1}:x_{2}:x_{3}", "FctIshigami"), URANIE::Exceptions::UErrorExceptions);
}

void TSobolTest::throwExceptionTSobolGetValue()
{
    int ns=5000;
    
    delete _tds2; _tds2 = nullptr;

    _tds2 = new URANIE::DataServer::TDataServer("tdsishigami","pouet");
    _tds2->fileDataRead("sample_5000_sobol_launching.dat");
    
    _tsobol = new URANIE::Sensitivity::TSobol(_tds2, "x_{1}:x_{2}:x_{3}", "FctIshigami");
    _tsobol->computeIndexes();

    CPPUNIT_ASSERT_THROW(_tsobol->getValue("bwrah","x_{1}"), URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(_tsobol->getValue("First","bwrah"), URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(_tsobol->getValue("First","x_{1}","bwrah"), URANIE::Exceptions::UErrorExceptions);
}




void TSobolTest::testTSobolConstructorLhs()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_IGNORED_TEST()
    ;

    cout << endl;
    gROOT->Macro("testSobolIshigamiLhs.C");

    Double_t dvalx1f, dvalx1t, dvalx2f, dvalx2t, dvalx3f, dvalx3t;

    std::ifstream outFile("testSobolIshigamiLhs.out", ios::in);

    outFile >> dvalx1f >> dvalx1t >> dvalx2f >> dvalx2t >> dvalx3f >> dvalx3t;

    //Close the file
    outFile.close();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx1f, 0.25, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx1t, 0.55, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx2f, 0.4, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx2t, 0.4, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx3f, 0.0, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx3t, 0.2, 0.1);
}

void TSobolTest::testTSobolConstructorQMCSobol()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_IGNORED_TEST()
    ;

    cout << endl;

    gROOT->Macro("testSobolIshigamiQmcSobol.C");

    Double_t dvalx1f, dvalx1t, dvalx2f, dvalx2t, dvalx3f, dvalx3t;
    std::ifstream outFile("testSobolIshigamiQmcSobol.out", ios::in);

    outFile >> dvalx1f >> dvalx1t >> dvalx2f >> dvalx2t >> dvalx3f >> dvalx3t;

    //Close the file
    outFile.close();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx1f, 0.25, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx1t, 0.55, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx2f, 0.4, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx2t, 0.4, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx3f, 0.0, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx3t, 0.2, 0.1);
}
void TSobolTest::testTSobolConstructorQMCHalton()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_IGNORED_TEST()
    ;

    cout << endl;

    gROOT->Macro("testSobolIshigamiQmcHalton.C");

    Double_t dvalx1f, dvalx1t, dvalx2f, dvalx2t, dvalx3f, dvalx3t;
    std::ifstream outFile("testSobolIshigamiQmcHalton.out", ios::in);

    outFile >> dvalx1f >> dvalx1t >> dvalx2f >> dvalx2t >> dvalx3f >> dvalx3t;

    //Close the file
    outFile.close();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx1f, 0.25, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx1t, 0.55, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx2f, 0.4, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx2t, 0.4, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx3f, 0.0, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx3t, 0.2, 0.1);
}
void TSobolTest::testTSobolConstructorSobol()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_IGNORED_TEST()
    ;

    cout << endl;

    gROOT->Macro("testSobolIshigamiSobol.C");

    Double_t dvalx1f, dvalx1t, dvalx2f, dvalx2t, dvalx3f, dvalx3t;
    std::ifstream outFile("testSobolIshigamiSobol.out", ios::in);

    outFile >> dvalx1f >> dvalx1t >> dvalx2f >> dvalx2t >> dvalx3f >> dvalx3t;

    //Close the file
    outFile.close();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx1f, 0.25, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx1t, 0.55, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx2f, 0.4, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx2t, 0.4, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx3f, 0.0, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx3t, 0.2, 0.1);
}
void TSobolTest::testTSobolConstructorSaltelli()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_IGNORED_TEST()
    ;

    cout << endl;

    gROOT->Macro("testSobolIshigamiSaltelli.C");

    Double_t dvalx1f, dvalx1t, dvalx2f, dvalx2t, dvalx3f, dvalx3t;
    std::ifstream outFile("testSobolIshigamiSaltelli.out", ios::in);

    outFile >> dvalx1f >> dvalx1t >> dvalx2f >> dvalx2t >> dvalx3f >> dvalx3t;

    //Close the file
    outFile.close();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx1f, 0.25, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx1t, 0.55, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx2f, 0.4, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx2t, 0.4, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx3f, 0.0, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx3t, 0.2, 0.1);
}

void TSobolTest::validateValuesIshigami(bool rerun)
{
    Double_t sobol;
    TTree *tree = _tsobol->getResultTuple();
    tree->SetBranchAddress("Value", &sobol);

    // First Order
    Double_t telfirstref[3] = {0.312162, 0.443108, 0.0304265} ;//{ 0.3059204357, 0.4269101427, 0.008963322552 };
    Double_t telfirstref_rerun[3] = {0.324633954094, 0.491257913806, 0} ;//{ 0.3059204357, 0.4269101427, 0.008963322552 };    
    tree->Draw(">>uranie_list_first", "Algorithm==\"--first--\"","goff");
    TEventList *telfirst = (TEventList*) gDirectory->Get("uranie_list_first");

    // Total Order
    Double_t teltotalref[3] = {0.548125, 0.447879, 0.243674}; //{ 0.542645632, 0.429824288, 0.2374062081 };
    Double_t teltotalref_rerun[3] = {0.588254543566, 0.51799088109, 0.262125005558}; //{ 0.542645632, 0.429824288, 0.2374062081 };
    tree->Draw(">>uranie_list_total", "Algorithm==\"--total--\"","goff");
    TEventList *teltotal = (TEventList*) gDirectory->Get("uranie_list_total");

    Double_t *array_first = ((rerun) ? &telfirstref_rerun[0] : &telfirstref[0]);
    Double_t *array_total = ((rerun) ? &teltotalref_rerun[0] : &teltotalref[0]);
    for (Int_t i = 0; i < 3; i++)
    {
        tree->GetEntry(telfirst->GetEntry(i));
        CPPUNIT_ASSERT_DOUBLES_EQUAL(array_first[i], sobol, 0.00001);

        tree->GetEntry(teltotal->GetEntry(i));
        CPPUNIT_ASSERT_DOUBLES_EQUAL(array_total[i], sobol, 0.00001);
    }

    tree->GetEntry(telfirst->GetEntry(0));
    CPPUNIT_ASSERT_DOUBLES_EQUAL(array_first[0], _tsobol->getValue("First", "x_{1}"), 0.00001);
    tree->GetEntry(teltotal->GetEntry(0));
    CPPUNIT_ASSERT_DOUBLES_EQUAL(array_total[0], _tsobol->getValue("Total", "x_{1}"), 0.00001);
    tree->GetEntry(telfirst->GetEntry(1));
    CPPUNIT_ASSERT_DOUBLES_EQUAL(array_first[1], _tsobol->getValue("First", "x_{2}"), 0.00001);
    tree->GetEntry(teltotal->GetEntry(1));
    CPPUNIT_ASSERT_DOUBLES_EQUAL(array_total[1], _tsobol->getValue("Total", "x_{2}"), 0.00001);
    tree->GetEntry(telfirst->GetEntry(2));
    CPPUNIT_ASSERT_DOUBLES_EQUAL(array_first[2], _tsobol->getValue("First", "x_{3}"), 0.00001);
    tree->GetEntry(teltotal->GetEntry(2));
    CPPUNIT_ASSERT_DOUBLES_EQUAL(array_total[2], _tsobol->getValue("Total", "x_{3}"), 0.00001);


    gROOT->Delete("uranie_list_first;*");
    gROOT->Delete("uranie_list_total;*");
}

void TSobolTest::testIshigamiSobolDecaled()
{
#ifdef ROOT5
    _tsobol = new URANIE::Sensitivity::TSobol(_tds2, _p2fIshigamiDecaled, 
            50000);
#else
    _tsobol = new URANIE::Sensitivity::TSobol(_tds2, IshigamiDecaled,
            "","IshigamiDec",50000);
#endif
    _tsobol->setDrawProgressBar(kFALSE);
    _tsobol->setSeed(123);
    _tsobol->computeIndexes();
    validateValuesIshigami();
}

void TSobolTest::testIshigamiSobolComputeIndexesWithTLauncherFunction()
{
    setUp1();
    _tsobol->computeIndexes();
    validateValuesIshigami();
}

void TSobolTest::testIshigamiSobolComputeIndexesGenerateSampleWithTLauncherFunction()
{
    setUp1();
    _tsobol->generateSample();
    _tsobol->computeIndexes();
    validateValuesIshigami();
}

void TSobolTest::testIshigamiSobolComputeIndexesWithTLauncher2()
{
    URANIE::DataServer::TAttribute *out = new URANIE::DataServer::TAttribute("Ishigami");
    out->delShare();
    
    URANIE::Relauncher::TCJitEval eval(&Ishigami);
    eval.addInput(_tds2->getAttribute("x_{1}"));
    eval.addInput(_tds2->getAttribute("x_{2}"));
    eval.addInput(_tds2->getAttribute("x_{3}"));
    eval.addOutput(out);
    
    URANIE::Relauncher::TSequentialRun run(&eval);
    run.startSlave();
    if( run.onMaster() )
    {
        _tsobol = new URANIE::Sensitivity::TSobol(_tds2, &run, 50000);
        _tsobol->setDrawProgressBar(kFALSE);
        _tsobol->setSeed(123);
        _tsobol->computeIndexes();
    
        validateValuesIshigami();
    
        run.stopSlave();
    }
}

void TSobolTest::testSobolComputeIndexesWithThreadedandNot()
{
    URANIE::DataServer::TAttribute *out = new URANIE::DataServer::TAttribute("TestIshigami");
    out->delShare();
    
    URANIE::Relauncher::TCJitEval eval(&Ishigami);
    eval.addInput(_tds2->getAttribute("x_{1}"));
    eval.addInput(_tds2->getAttribute("x_{2}"));
    eval.addInput(_tds2->getAttribute("x_{3}"));
    eval.addOutput(out);
    
    URANIE::Relauncher::TSequentialRun run(&eval);
    run.startSlave();
    if( run.onMaster() )
    {
        _tsobol = new URANIE::Sensitivity::TSobol(_tds2, &run, 500);
        _tsobol->setDrawProgressBar(kFALSE);
        _tsobol->setSeed(123);
        _tsobol->computeIndexes();
    
        run.stopSlave();
    }

    TTree *T = _tsobol->getResultTuple();    
    ((TTreePlayer*)(T->GetPlayer()))->SetScanRedirect(kTRUE);
    ((TTreePlayer*)(T->GetPlayer()))->SetScanFileName("outpuS.txt");
    T->Scan("*");

    _tds2->deleteTuple();
    _tds2->delAttribute("TestIshigami");
    
    URANIE::DataServer::TAttribute *outT = new URANIE::DataServer::TAttribute("TestIshigamiT");
    outT->delShare();
    
    URANIE::Relauncher::TCJitEval eval2(&Ishigami);
    eval2.addInput(_tds2->getAttribute("x_{1}"));
    eval2.addInput(_tds2->getAttribute("x_{2}"));
    eval2.addInput(_tds2->getAttribute("x_{3}"));
    eval2.addOutput(outT);
    
    URANIE::Relauncher::TThreadedRun runT(&eval2,4);
    runT.startSlave();
    if( runT.onMaster() )
    {
        delete _tsobol;
        _tsobol = new URANIE::Sensitivity::TSobol(_tds2, &runT, 500);
        _tsobol->setDrawProgressBar(kFALSE);
        _tsobol->setSeed(123);
        _tsobol->computeIndexes();
    
        runT.stopSlave();
    }

    T = _tsobol->getResultTuple();    
    ((TTreePlayer*)(T->GetPlayer()))->SetScanRedirect(kTRUE);
    ((TTreePlayer*)(T->GetPlayer()))->SetScanFileName("outpuT.txt");
    T->Scan("*");

    Int_t nret = UranieTests::areFilesEqual(TString("outpuT.dat"),
                                            TString("outpuS.dat"));
    CPPUNIT_ASSERT_EQUAL(nret, 0);        
}


void TSobolTest::testIshigamiSobolComputeIndexesGenerateSampleCreateLauncherWithTLauncherFunction()
{
    setUp1();
    _tsobol->generateSample();
    URANIE::Launcher::TLauncherFunction *tlf =
#ifdef ROOT5
            new URANIE::Launcher::TLauncherFunction(_tds2, _p2fIshigami,
#else
            new URANIE::Launcher::TLauncherFunction(_tds2, &Ishigami, 
#endif
                    "", "p2fIshigami");
    tlf->setDrawProgressBar(kFALSE);
    tlf->run();
    _tsobol->computeIndexes();
    validateValuesIshigami();
    delete tlf;
}

void TSobolTest::validateValuesFlowrate(bool plenty)
{
    Double_t sobol;
    TTree *tree = _tsobol->getResultTuple();
    tree->SetBranchAddress("Value", &sobol);

    const Int_t nCompareValues = 8;
    // First Order
    Double_t telfirstref[nCompareValues] =
      { 0.7307960632, 0.03971801855, 0.03988093144, 0.04031315879, 0.09541457836, 0.09940252188, 0.06540122537, 0.06487680967 };
    tree->Draw(">>uranie_list_first", "Algorithm==\"--first--\" && Inp!=\"__sum__\"","goff");
    TEventList *telfirst = (TEventList*) gDirectory->Get("uranie_list_first");
    
    // Total Order
    Double_t teltotalref[nCompareValues] =
      { 0.7370938288, 1.536807516e-06, 1.052193985e-11, 1.03598907e-05, 0.06616708161, 0.05140091875, 0.05620002772, 0.01662329743 };
    tree->Draw(">>uranie_list_total", "Algorithm==\"--total--\" && Inp!=\"__sum__\"","goff");
    TEventList *teltotal = (TEventList*) gDirectory->Get("uranie_list_total");

    for (Int_t i = 0; i < nCompareValues * ((plenty)?300:1); i++)
    {        
        tree->GetEntry(telfirst->GetEntry(i));
        CPPUNIT_ASSERT_DOUBLES_EQUAL(telfirstref[i%nCompareValues], sobol, 0.00001);
        
        tree->GetEntry(teltotal->GetEntry(i));
        CPPUNIT_ASSERT_DOUBLES_EQUAL(teltotalref[i%nCompareValues], sobol, 0.00001);
    }
    gROOT->Delete("uranie_list_first;*");
    gROOT->Delete("uranie_list_total;*");
}

void TSobolTest::testFlowrateSobolCompute300Outputs()
{
    setUp300();
    _tsobol->computeIndexes("nointermed");
    validateValuesFlowrate(true);
}
    
void TSobolTest::testFlowrateSobolComputeIndexesWithTLauncherCode()
{
    setUp2();
    _tsobol->computeIndexes("nointermed");
    validateValuesFlowrate();
}

void TSobolTest::testFlowrateSobolComputeIndexesGenerateSampleWithTLauncherCode()
{
    setUp2();
    _tsobol->generateSample();
    _tsobol->computeIndexes("nointermed");
    validateValuesFlowrate();
}

void TSobolTest::testFlowrateSobolComputeIndexesGenerateSampleCreateLauncherWithTLauncherCode()
{
    setUp2();
    _tsobol->generateSample();
    URANIE::Launcher::TLauncher *tl = new URANIE::Launcher::TLauncher(_tds,
            _mycode);
    tl->setDrawProgressBar(kFALSE);
    tl->run("nointermed");
    _tsobol->computeIndexes("nointermed");
    validateValuesFlowrate();
    delete tl;
}
