////////////////////////////////////////////////////////////// 
// Copyright (C) 2013-2024 CEA/DES 
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
\file   TestTSobolRank.cxx
\author Rudy Chocat <rudy.chocat@cea.fr>
\date   2024-03-01 15:59:38

\brief Unit test for URANIE::Sensitivity::TestTSobolRank class

*/
// Uranie
#include "TUniformDistribution.h"
#include "TSampling.h"
#include "TLauncher2.h"
#include "TCJitEval.h"
#include "TOutputFileKey.h"
#include "TOutputFileDataServer.h"
#include "TLauncherFunction.h"
#include "TLauncher.h"
#include "TestUserFunctions.h"
#include "TestTSensitivity.h"
#include "TestTSobolRank.h"
#include "TSobolRank.h"
#include "UDisableTest.h"
#include "uranie_commands.h"

// ROOT
#include "TString.h"
#include "TCanvas.h"
#define EPSILON 0.01 
#define LVLE_EPSILON 0.001
#define DBL_PRECISION_COMPARE 1.e-8

#include <iostream>
#include <iomanip>

TSobolRankTest::TSobolRankTest() :
        tds1(NULL), tds(NULL), tsr(NULL), tss1(NULL)
{

}

TSobolRankTest::~TSobolRankTest()
{
    UranieTests::removeFile(gSystem->pwd(),"Ishigami.root");
    UranieTests::removeFile(gSystem->pwd(),"flowrateUniformDesign100.root");
    UranieTests::removeFile(gSystem->pwd(), "flowrateUniformDesign.root");
    UranieTests::removeFile(gSystem->pwd(), "flowrateUniformDesign2D.root");
}

void TSobolRankTest::setUp()
{
    tds1 = new URANIE::DataServer::TDataServer();

    tds = new URANIE::DataServer::TDataServer("tdsflowreate",
            "DataBase flowrate");
    tds->fileDataRead("flowrateUniformDesign100.dat");
}


void TSobolRankTest::tearDown()
{
    delete tss1;
    tss1 = NULL;
    delete tsr;
    tsr = NULL;
    delete tds1;
    tds1 = NULL;
    delete tds;
    tds = NULL;
}

////////////////////////////////
///                          ///                            
///         Tests            ///
///                          ///
////////////////////////////////
void TSobolRankTest::testSobolRank()
{
    // Load the data base in the DataServer
    tds1->fileDataRead("flowrateUniformDesign100.dat");

    //////////////////////
    // Case SRC
    //////////////////////
    tss1 = new URANIE::Sensitivity::TSobolRank(tds1, "rw:r:tu:tl:hu:hl:l:kw", "ystar");
    tss1->computeIndexes();
    tss1->printLog();

    Double_t sobol;
    TTree * tree = tss1->getResultTuple();
    tree->SetBranchAddress("Value", &sobol);

    Double_t dsum = 0.0, jsum=0.0;

    Double_t Comparison[8] =
    {0.78159130, -0.23233467,  0.04793761, -0.05606230, -0.04908156, -0.05753108, -0.07804288, -0.05852849};
    for (Int_t i = 0; i < 8; i++)
    {
        tree->GetEntry(i);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(Comparison[i], sobol, 0.00001);
        if(i<8)
        {
            dsum += Comparison[i];
            jsum += sobol;
        }
    }
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dsum, jsum, 0.001);
}

void TSobolRankTest::testIshigamiSobolRank()
{
    // Load the data base in the DataServer
    tds1->fileDataRead("Ishigami.dat");

    Double_t sobol;
    tsr = new URANIE::Sensitivity::TSobolRank(tds1, "x_{1}:x_{2}:x_{3}","Ishigami");
    tsr->computeIndexes();

    TTree * tree = tsr->getResultTuple();
    tree->SetBranchAddress("Value", &sobol);

    tree->GetEntry(0);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.29711557, sobol, EPSILON);
    tree->GetEntry(1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.44369669, sobol, EPSILON);
    tree->GetEntry(2);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.01741434, sobol, EPSILON);
}


void TSobolRankTest::testIshigamiSobolRankWithDrawings()
{
    // Load the data base in the DataServer
    tds1->fileDataRead("Ishigami.dat");

    tsr = new URANIE::Sensitivity::TSobolRank(tds1, "x_{1}:x_{2}:x_{3}","Ishigami");
    tsr->computeIndexes();

    TCanvas *Can = new TCanvas("Can","Can",1);
    tsr->drawIndexes("Ishigami", "", "nonewcanv,hist,first");
    Can->Clear();
//    treg->drawIndexes("Ishigami", "", "nonewcanv,pie,first");
//    treg->drawIndexes("Ishigami", "", "nonewcanv,pie,first");
    delete Can;

    tsr->drawIndexes("Ishigami", "", "hist,first");
    //  treg->drawIndexes("Ishigami", "", "pie,first");
    // tsr->drawIndexes("Ishigami", "", "pie,first");
    
}

static void constructor_function_long_output_throw_exception(const string& outputs) {
    vector<URANIE::DataServer::TUniformDistribution> inputs;
    TestTSensitivity::add_flowrate_attributes(inputs);

    URANIE::DataServer::TDataServer dataserver("dataserver", "");
    dataserver.keepFinalTuple(false);
    for (auto& input: inputs)
        dataserver.addAttribute(&input);

    string inputs_name;
    TList *list=dataserver.getListOfAttributes();
    for (const auto& attribute: *list)
        inputs_name += attribute->GetName() + string(":");
    inputs_name.pop_back();

    URANIE::Sampler::TSampling sampling(&dataserver);
    sampling.generateSample();

    URANIE::Launcher::TLauncherFunction launcher(&dataserver, &flowrateModel, "",
        outputs.data());
    launcher.setDrawProgressBar(false);
    launcher.run();

    CPPUNIT_ASSERT_THROW(URANIE::Sensitivity::TSobolRank sobolrank(&dataserver,
        inputs_name.data(), outputs.data() ),
        URANIE::Exceptions::UErrorExceptions);
    delete list;
}

void TSobolRankTest::throwExceptionConstructorTooLongOutputFunctionFirst() {
    const string outputs = "voila_un_nom_beaucoup_trop_long_pour_un_seul_attribut:encore_un_p*tain_de_nom_beaucoup_trop_long_pour_un_seul_attribut";
    constructor_function_long_output_throw_exception(outputs);
}

void TSobolRankTest::throwExceptionConstructorTooLongOutputFunctionSecond() {
    const string outputs = "voila_un_nom_beaucoup_trop_long_pour_un_attribut:encore_un_p*tain_de_nom_beaucoup_trop_long_pour_un_seul_attribut";
    constructor_function_long_output_throw_exception(outputs);
}

void TSobolRankTest::throwExceptionConstructorTooLongOutputCode() {
    vector<URANIE::DataServer::TUniformDistribution> inputs;
    TestTSensitivity::add_flowrate_attributes(inputs);

    URANIE::DataServer::TDataServer dataserver("dataserver", "");
    for (auto& input: inputs)
        dataserver.addAttribute(&input);

    string inputs_name;
    TList *list=dataserver.getListOfAttributes();
    for (const auto& attribute: *list)
        inputs_name += attribute->GetName() + string(":");
    inputs_name.pop_back();

    URANIE::Sampler::TSampling sampling(&dataserver, "lhs", 100);
    sampling.generateSample();

    TString input_file = "flowrate_input_with_keys.in";
    TestTSensitivity::set_file_key_to_upper_first(dataserver, input_file);

    URANIE::Launcher::TOutputFileDataServer output_file("_output_flowrate_withRow_.dat");
    output_file.addAttribute("voila_un_nom_beaucoup_trop_long_pour_un_seul_attribut");

    URANIE::Launcher::TCode code(&dataserver, "");
    code.addOutputFile(&output_file);

    URANIE::Launcher::TLauncher launcher(&dataserver, &code);
    launcher.setDrawProgressBar(false);
    launcher.run("nointermediatesaved");

    CPPUNIT_ASSERT_THROW(URANIE::Sensitivity::TSobolRank sobolrank(&dataserver,
        inputs_name.data(), code.getOutputVarNames()),
        URANIE::Exceptions::UErrorExceptions);
    delete list;
}

void TSobolRankTest::throwExceptionConstructorTooLongOutputRun() {
    vector<URANIE::DataServer::TUniformDistribution> inputs;
    TestTSensitivity::add_flowrate_attributes(inputs);
    URANIE::DataServer::TAttribute output("voila_un_nom_beaucoup_trop_long_pour_un_seul_attribut_de_runner");

    URANIE::DataServer::TDataServer dataserver("dataserver", "");
    for (auto& input: inputs)
        dataserver.addAttribute(&input);

    URANIE::Sampler::TSampling sampling(&dataserver, "lhs", 100);
    sampling.generateSample();

    URANIE::Relauncher::TCJitEval eval(&flowrateModel);
    TList *list=dataserver.getListOfAttributes();
    for (const auto& attribute: *list)
        eval.addInput(dynamic_cast<URANIE::DataServer::TAttribute*>(attribute));
    eval.addOutput(&output);

    string inputs_name;
    for (const auto& input: *eval.getInputList())
        inputs_name += input->GetName() + string(":");
    inputs_name.pop_back();

    URANIE::Relauncher::TSequentialRun run(&eval);
    run.startSlave();
    if (run.onMaster()) {
        URANIE::Relauncher::TLauncher2 launcher(&dataserver, &run);
        launcher.solverLoop();
        run.stopSlave();

        CPPUNIT_ASSERT_EQUAL(1, eval.getOutputList()->GetEntries());
        CPPUNIT_ASSERT_THROW(URANIE::Sensitivity::TSobolRank sobolrank(&dataserver,
            inputs_name.data(), eval.getOutputList()->First()->GetName() ),
            URANIE::Exceptions::UErrorExceptions);
    }
    delete list;
}

void TSobolRankTest::throwExceptionConstructorTooLongOutputFromData() {
    URANIE::DataServer::TDataServer dataserver("dataserver", "");
    dataserver.fileDataRead("flowrateUniformDesign.dat");
    TList *list=dataserver.getListOfAttributes();
    const string output_name = list->Last()->GetName();

    string inputs_name;
    for (const auto& attribute: *list)
        inputs_name += attribute->GetName() + string(":");
    const string output_name_w_sep = ":" + output_name + ":";
    inputs_name.erase(inputs_name.find(output_name_w_sep), output_name_w_sep.size());

    dataserver.addAttribute("voila_un_nom_beaucoup_trop_long_pour_un_seul_attribut", output_name);
    delete list; list=dataserver.getListOfAttributes();
    CPPUNIT_ASSERT_THROW(URANIE::Sensitivity::TSobolRank(&dataserver, inputs_name.data(),
        list->Last()->GetName() ),
        URANIE::Exceptions::UErrorExceptions);
    delete list;
}


void TSobolRankTest::throwExceptionTRegressionConstructorTdsNull()
{

    URANIE::DataServer::TDataServer *tdsnull =
            new URANIE::DataServer::TDataServer();

    CPPUNIT_ASSERT_THROW(
            new URANIE::Sensitivity::TSobolRank(tdsnull,
                    "rw:r:tu:tl:hu:hl:l:kw", "ystar"),
            URANIE::Exceptions::UErrorExceptions);

    delete tdsnull;
}


void TSobolRankTest::throwExceptionTRegressionConstructorFalseInput()
{

    CPPUNIT_ASSERT_THROW(
            new URANIE::Sensitivity::TSobolRank(tds, "rw:r:tu:tl:hu:hl:l:toto",
                    "ystar"), URANIE::Exceptions::UErrorExceptions);
}

void TSobolRankTest::throwExceptionTRegressionConstructorFalseOutput()
{

    CPPUNIT_ASSERT_THROW(
            new URANIE::Sensitivity::TSobolRank(tds, "rw:r:tu:tl:hu:hl:l:kw",
                    "toto"), URANIE::Exceptions::UErrorExceptions);
}