/**
 * @file   UserFunctions.C
 * @brief  User Functions
 */

/****************** User Function **********************/
//! The true flow rate model
/*!
 It describes the flow rate of water through a borehole from an upper aquifer to a lower aquifer separated by an impermeable rock layer.
 The response variable \f$ y\f$, the \b flow \b water through the borehole in \f$ m^{3}/yr\f$, is given by :
 \f[
 y = f(x) = \frac{2 \pi T_u\left( H_u - H_l\right)}{\ln(\frac{r}{r_{\omega}})\left[ 1 + \frac{2 L T_u}{\ln(\frac{r}{r_{\omega}}) r_{\omega}^2 K_{\omega}} + \frac{T_u}{T_l}\right]}
 \f]

 \param x (double *) the input vector in the order :
 \li \f$ r_{\omega} \in [0.05, 0.15]\f$ : radius of borehole (m)
 \li \f$ r \in [100, 50~000]\f$ : radius of influence {\ix (m)}
 \li \f$ T_u \in [63~070, 115~600]\f$ : Transmissivity of upper aquifer (\f$ m^{2}\f$/year)
 \li \f$ T_l \in [63.1, 116]\f$ : Transmissivity of lower aquifer (\f$ m^{2}\f$/year)
 \li \f$ H_u \in [990, 1~110]\f$ : Potentiometric head of upper aquifer (m)
 \li \f$ H_l \in [700, 820]\f$ : Potentiometric head of lower aquifer (m)
 \li \f$ L \in [1~120, 1~680]\f$ : length of borehole (m)
 \li \f$ K_{\omega} \in [9~855, 12~045]\f$ : hydraulic conductivity of borehole (m)
 \param y (double *) the output vector

 \see waterboreholeHoXu
 */
void flowrateModel(double *x, double *y)
{
    Double_t drw = x[0], dr = x[1];
    Double_t dtu = x[2], dtl = x[3];
    Double_t dhu = x[4], dhl = x[5];
    Double_t dl = x[6], dkw = x[7];

    Double_t dnum = 2.0 * TMath::Pi() * dtu * (dhu - dhl);
    Double_t dlnronrw = TMath::Log(dr / drw);
    Double_t dden =
            dlnronrw
                    * (1.0 + (2.0 * dl * dtu) / (dlnronrw * drw * drw * dkw)
                            + dtu / dtl);

    y[0] = dnum / dden;
}

//! The surrogate \em flow \em rate model (polynomial regression) proposed Ho \em and Xu (2000) is
/*!
 The polynomial regression models is given by the formula :
 \f[
 \hat{log(y)}  = 4.156 + 1.9903(\log{r_{\omega}}+2.3544) - 0.0007292(L-1400) - .003554(H_l - 760) \\
  \f]
 \f[
 \hspace*{5cm} + .0035068(H_u-1050) + .000090868(K_{\omega}-10950) - .00087286 (\log{(r)} -8.8914) \\
  \f]
 \f[
 + .000015325 (H_u-1050) (H_l - 760) + .00000026487 (L-1400)^2 \\
  \f]
 \f[
 \hspace*{5cm} - .0000071759 (H_l - 760)^2 - .0000068021 (H_u-1050)^2 \\
  \f]
 
 \param x (double *) the input vector in the order :
 \li \f$ r_{\omega} \in [0.05, 0.15]\f$ : radius of borehole (m)
 \li \f$ r \in [100, 50~000]\f$ : radius of influence {\ix (m)}
 \li \f$ T_u \in [63~070, 115~600]\f$ : Transmissivity of upper aquifer (\f$ m^{2}\f$/year)
 \li \f$ T_l \in [63.1, 116]\f$ : Transmissivity of lower aquifer (\f$ m^{2}\f$/year)
 \li \f$ H_u \in [990, 1~110]\f$ : Potentiometric head of upper aquifer (m)
 \li \f$ H_l \in [700, 820]\f$ : Potentiometric head of lower aquifer (m)
 \li \f$ L \in [1~120, 1~680]\f$ : length of borehole (m)
 \li \f$ K_{\omega} \in [9~855, 12~045]\f$ : hydraulic conductivity of borehole (m)
 \param y (double *) the output vector

 \see flowrateModel
 
 */

void HoXuSurrogateModel(double *x, Double_t *y)
{
    Double_t drw = x[0], dr = x[1];
//   Double_t dtu = x[2], dtl = x[3];
    Double_t dhu = x[4] - 1050.0, dhl = x[5] - 760.0;
    Double_t dl = x[6] - 1400.0, dkw = x[7] - 10950.0;

    Double_t yHoXu = 4.1560 + 1.9903 * (TMath::Log(drw) + 2.3544);
    yHoXu -= .0007292 * dl;
    yHoXu -= .003554 * dhl;
    yHoXu += .0035068 * dhu;
    yHoXu += .000090868 * dkw;
    yHoXu += .000015325 * dhu * dhl;
    yHoXu += .00000026487 * dl * dl;
    yHoXu -= .0000071759 * dhl * dhl;
    yHoXu -= .0000068021 * dhu * dhu;
    yHoXu -= .00087286 * (TMath::Log(dr) - 8.8914);

    y[0] = TMath::Exp(yHoXu);

}

//! The rosenbrock Function ( a=10.0 and b=1.0)
/*!
 The rosenbrock function is defined by the equation 
 \$[ f(x_1, x_2) = a * ( x_2 - x_1^2 )^2 + b * ( 1.0 - x_1 )^2 \$]
 with a=10.0 and b=1.0
 */
void rosenbronck(double *x, Double_t *y)
{
    Double_t dx1 = x[0], dx2 = x[1];
    Double_t da = 10.0, db = 1.0;
    double tmp1, tmp2;
    
    tmp1 = dx2-dx1*dx1;
    tmp2 = 1.0-dx1;
    y[0] = da * tmp1*tmp1 + db * tmp2*tmp2;

}

void Ishigami(double *paramt, Double_t *res)
{
Double_t a = 7.0;
Double_t b = 0.1;

Double_t x1 = paramt[0];
Double_t x2 = paramt[1];
Double_t x3 = paramt[2];

Double_t dsinx1 = TMath::Sin(x1);
Double_t dsinx2 = TMath::Sin(x2);

res[0] = dsinx1;
res[0] += a * dsinx2 * dsinx2;
res[0] += b * x3 * x3 * x3 * x3 * dsinx1;
}

/****************** End of User Function **********************/
