//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////

/*!
 @file   TCirce.h
 @author Fabrice Gaudier  <fabrice.gaudier@cea.fr>
 @date 2011-08-04 16:14:29

 @brief interface with Circe tool

 */

#ifndef TCIRCE_H
#define TCIRCE_H

#include "TUncertModeler.h"

#include <iostream>
#include <fstream>

using std::cout;
using std::endl;

// ROOT
#include "Rtypes.h"
#include "TVectorD.h"

// Uranie
#include "UncertModeler.h"
#define CIRCE_DIFF_ATTRIBUTE "__circe__diff__"
#define CIRCE_RESIDU_ATTRIBUTE "__circe__residual__"
#define CIRCE_BLOC_NAME "__circe__bloc__"

namespace URANIE
{
namespace UncertModeler
{
class TCirce: public TUncertModeler
{

    // Attributes
public:
    enum ECirceMethod
    {
        kBiais, kBiaisBloc, kBiaisBlocCInitAlea, kBiaisBlocBootstrap, kUnknown
    };
protected:
    Int_t _nSeed;                          ///< the Seed of the random generator
    Double_t _dTol;    ///< The tolerance to test the convergence (default 1e-5)
    Double_t _dLikelihood;         ///< The Likelihood for the optimal parameter
    TVectorD _vecBiais;                           ///< The final vector of biais
    TMatrixD _matC;                          ///< The final matrix of Covariance
    TMatrixD _matCorrelation;               ///< The final matrix of correlation
    Int_t _nPatterns;                             ///< The number of patterns
    Int_t _nParameters;                           ///< The number of parameters
    TString _sCirceInputFile;                     ///< The input file of Circe
    ECirceMethod _nCirceMethod; ///< the Circe method from (kBiais, kBiaisBloc, kBiaisBlocCInitAlea, kBiaisBlocBootstrap)

    Int_t _nInitCAlea;              ///< the number of Matrix C alea (default 1)

    TList * _listOfTDS;                  ///< The list of TDS which contain data
    TList * _listOfInformations;          ///< The list of information for Circe
    TList * _listOfBloc; ///< The list of bloc (default NULL : when only one is as blias.f)
    TList * _listOfSensitivity;           ///< The list of sensitivity parameter

    TVectorD _vecDiff;                          ///< The vector Y = Texp - Ycode
    TVectorD _vecYStarVariance;         ///< The vector of experimental variance
    TMatrixD _matSensibility;                     ///< The matrix of sensibility
    TMatrixD _matInitialC;                        ///< The initial matrix of C
    TVectorD _vecInitialB;                        ///< The initial vector B

    TMatrixD _matVarianceMu;                      ///< The matrix of variance of mu computed Confidence Interval by the Fisher Matrix
    TMatrixD _matVarianceSigma;                   ///< The matrix of variance of sigma computed Confidence Interval by the Fisher Matrix


    // Methods
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor 
    TCirce(URANIE::DataServer::TDataServer *tds, const char *systar,
            const char *syhat, const char *ssensi, const char *sexpuncert = "",
            Option_t * option = "");
    //! Default destructor
    virtual ~TCirce();
    //@}

    //---------------------------------------------
    /** @name Setting and Getting attributs
     Treat the case where you have a Circe Input File
     **/
    //@{
    //! Set the tolerance parameter
    void setTolerance(Double_t dtol);
    //! Get the Tolerance parameter
    Double_t getTolerance()
    {
        return _dTol;
    }
    //! Set the number of random matrix C to generate
    void setNCMatrix(Int_t n)
    {
        _nInitCAlea = n;
    }
    //! Get the number of random matrix C to generate
    Int_t getNCMatrix()
    {
        return _nInitCAlea;
    }
    //! Get the number of TDS
    Int_t getNTDS();
    //! Get the Likelihood for the optimal parameters
    Double_t getLikelihood()
    {
        return _dLikelihood;
    }
    //! Get the number of patterns
    Int_t getNPatterns()
    {
        return _nPatterns;
    }
    //! Get the number of parameters
    Int_t getNParameters()
    {
        return _nParameters;
    }
    //! Get the matrix of variance of mu
    TMatrixD getMatrixVarianceMu()
    {
        return _matVarianceMu;
    }
    //! Get the matrix of variance of sigma
    TMatrixD getMatrixVarianceSigma()
    {
        return _matVarianceSigma;
    }
    //@}

    //---------------------------------------------
    /** @name vector B and matrix C Initial and final values
     **/
    //@{
    //! Set The initial Matrix C
    void setCMatrixInitial(TMatrixD mat);
    //! Get the initial Matrix C
    TMatrixD getCMatrixInitial()
    {
        return _matInitialC;
    }
    //! Get the final matrix C of coriance (Circe)
    TMatrixD getCMatrix()
    {
        return _matC;
    }
    //! Set the initial vector B
    void setBVectorInitial(TVectorD vec);
    //! Get the initial vector B
    TVectorD getBVectorInitial()
    {
        return _vecInitialB;
    }
    //! Get the final vector of biais (Circe)
    TVectorD getBVector()
    {
        return _vecBiais;
    }
    //@}

    //---------------------------------------------
    /** @name  Methods for the algorithm
     **/
    //@{
    //! The init method
    void init(Option_t * option = "");
    //! the estimate procedure
    void estimate(Option_t * option = "");
    //! the terminate method
    void terminate(Option_t * option = "");
    //! Print the results
    void printResults(Option_t * option = "");
    //! Add data from a new TDS
    void addData(URANIE::DataServer::TDataServer *tds, const char *systar,
            const char *syhat, const char *ssensi, const char *sexpuncert = "");
    //@}

    //---------------------------------------------
    /** @name Other methods for Direct implementation
     **/
    //@{
    //! Set the Circe Input File
    void setInputFile(const char * sfile);
    //! Get the Circe Input File
    const char * getInputFile()
    {
        return _sCirceInputFile.Data();
    }
    //! Set and Read the Circe Input File
    void readInputFile(const char * sfile = "", ECirceMethod nCirceMethod =
            kBiaisBloc);
    //@}

    //---------------------------------------------
    /** @name Internal methods
     **/
    //@{
    //! The addBloc method
    void addBloc(URANIE::DataServer::TDataServer *tds, const char *sbloc,
            const char* sname = "");
    //! Generate a ramdom finite positive matrix C
    TMatrixD initCMatrix(Int_t nrows, Double_t dMulti);
    //@}

    //---------------------------------------------
    /** @name The Seed
     **/
    //@{
    //! Init the seed
    /*!
     If the value is zero, the seed is initialized with the hour in milliseconds
     
     \param ind (Int_t) the seed value
     */
    void setSeed(Int_t ind = 0);
    //! Returns the seed value
    Int_t getSeed()
    {
        return _nSeed;
    }
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    //! Print the log of the class
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::UncertModeler::TCirce, ID_UNCERTMODELER)
    //The description
};

} // End Of namespace UncertModeler
} // End Of namespace URANIE

#endif
//  End Of File TCirce.h
