//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 @file   TTester.h
 @author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 @date   Tue Jan  6 10:21:14 2009

 @brief  Interface with statistical test

 */

/**

 * @class URANIE::UncertModeler::TTester
 *
 * @brief Description of the class TTester.
 *
 * These class implements the tests of fit based on the Empirical Distribution Function (\b "EDF").
 * The EDF is a step function, calculated from the sample, wich estimates the population distribution.
 *
 * \see Goofness-of-fit techniques, R. D'Agostini and M. Stephens, Dekker
 */

#ifndef TTESTER_H
#define TTESTER_H

// ROOT
#include "TNamed.h"
#include "Math/ProbFuncMathCore.h"
#include "TList.h"
#include "TVector.h"

// Uranie
#include "UncertModeler.h"
#include "TDataServer.h"

namespace URANIE
{
namespace UncertModeler
{
class TTester: public TNamed
{
    // Attributes

protected:
    Long64_t _nS;                           ///< the size of selected values
    Double_t _pValue;                       ///< The pValue of the last test
    Double_t _dScore;                       ///< The score of the last test
    Double_t _dModifiedScore; ///< The \b Modified \b value \b score of the last test
    TString _sAttributes; ///< The string of attributes separated by the colon ":" character
    TList* _listOfLaws;                   ///< The list of law to test
    Bool_t _bDrawSame;               ///< Boolean for drawing in the current pad

public:
    enum ETestType
    {
        kKolmogorovSmirnov, kAndersonDarling, kCramerVonMises, kUnknown
    };
    ETestType _nTestType;                    ///< The type of test
    enum ECase
    {
        kCase0, kCase1, kCase2, kCase3, kCaseUnknown
    };
    ECase _nCase;
    URANIE::DataServer::TDataServer *_tds;   ///< Pointer to a TDS
    Bool_t _blog;                            ///< Boolean for edit the log
    Bool_t _bquiet;                          ///< Boolean for being quiet

public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Constructor by the type of Test
    TTester(URANIE::DataServer::TDataServer *tds, ETestType nTestType,
            const char *varexpinput, Option_t * option = "");

    //! Default constructor
    virtual ~TTester();
    //@}

    //---------------------------------------------
    /** @name Setting and Getting attributes
     **/
    //@{
    Int_t getID()
    {
        return ID_UNCERTMODELER;
    }
    //! Get the \b score of the \b last \b test
    Double_t getScore()
    {
        return _dScore;
    }
    //! Get the \b modified \b score of the \b last \b test
    Double_t getModifiedScore()
    {
        return _dModifiedScore;
    }
    //! Get the \b pValue of the \b last \b test
    Double_t getPValue()
    {
        return _pValue;
    }

    /*!
     * \brief getCaseOfGoF
     *
     * kCase0 : All the parameters of the law are specified
     * kCase3 : All the parameters of the law are not specified. Then they are estimated by MLE.
     *
     * The other cases are not treated in URANIE:
     * kCase1 : Just the position parameters are specified, but not the scaled parameters
     * kCase2 : Just the scaled parameters are specified, but not the position parameters
     *
     * \return ECase the case of Goodness-Of-Fit
     */
    ECase getCase()
    {
        return _nCase;
    }
    //@}

    //---------------------------------------------
    /** @name Computes the score
     **/
    //@{
    //! Compute the score of the statistic
    /*!
     A  test to generalize the compute the score of the statistic.
     */
    Double_t computeScore(Option_t * option = "");
    //! The true computation with a vector of information and return the current value
    /*!
     \param i (Int_t) the current index
     \param dzi (Double_t) the current value of Z = F(x)
     */
    virtual Double_t computeTheCurrentScore(Int_t i, Double_t dzi) = 0;
    //! Compute the modified value of the statistic
    virtual void computeModifiedScore(
            URANIE::DataServer::TStochasticAttribute::ELawType tlaw) = 0;

    //! Return the name of the statistic
    virtual const char * getTitleStatistic() = 0;

    //! Add in the current graph the CDF of a uniform law defined by theses parameters \f$ \min, \max \f$
    /*!
     \param min (Double_t) the minnimum value of the law
     \param max (Double_t) the maximum value of the law
     \param color (Color_t)[kBlack] the color of the line of the CDF's law
     \param stitle (TString)[] the title of the entry in the legend box
     */
    void addUniformLaw(Double_t min, Double_t max, Color_t color = kBlack, TString stitle = "");

    //! Add in the current graph the CDF of a normal law defined by theses parameters \f$ {\cal N} ( \mu, \sigma ) \f$
    /*!
     \param mu (Double_t) the mean \f$ \mu \f$ of the law
     \param sigma (Double_t) the standard-deviation \f$ \sigma \f$ of the law
     \param color (Color_t)[kBlack] the color of the line of the CDF's law
     \param stitle (TString)[] the title of the entry in the legend box
     */
    void addNormalLaw(Double_t mu, Double_t sigma, Color_t color = kBlack, TString stitle = "");

    //! Adds in the current graph the CDF of a LogNormal law defined by theses parameters \f$ \mu, \sigma \f$ of the \f$ ln(X) \sim {\cal N} ( \mu, \sigma )\f$ law.
    /*!
     \param mu (Double_t) The mean \f$ \mu \f$ of the \f$ ln(X)\f$ law
     \param sigma (Double_t) The standard-deviation \f$ \sigma \f$ of the \f$ ln(X)\f$ law
     \param color (Color_t)[kBlack] The color of the line of the CDF's law
     \param stitle (TString)[] The title of the entry in the legend box
     */
    void addLogNormalLaw(Double_t mu, Double_t sigma, Color_t color = kBlack, TString stitle = "");

    /*!
     * \brief Add in the current graph the CDF of a Weibull law defined by theses parameters \f$ {\cal W}( lambda, k) \f$.
     *
     * where the CDF formula is :
     * \f$ F(x) = 1 - \exp\left(-(x/\lambda)^k\right) \forall x \ge 0\f$
     *
     * \param lambda (Double_t) The scale parameter \f$ \beta \f$ of the Weibull law
     * \param k (Double_t) The shape parameter \f$ m \f$ of the the Weibull law
     * \param color (Color_t)[kBlack] The color of the line of the CDF's law
     * \param stitle (TString)[] The title of the entry in the legend box
     */
    void addWeibullLaw(Double_t beta, Double_t m, Color_t color = kBlack, TString stitle = "");

    /*!
     * \brief Adds in the current graph the CDF of a Gamma law defined by theses parameters \f$ {\cal G}amma ( \alpha, \beta) \f$.
     *
     * where the CDF formula is :
     * \f$ F(x) = inc_gamma( \alpha, x/scale )\f$

     * \param alpha (Double_t) The scale parameter \f$ \alpha \f$ of the Gamma law
     * \param beta (Double_t) The shape parameter \f$ \beta \f$ of the the Gamma law
     * \param color (Color_t)[kBlack] The color of the line of the CDF's law
     * \param stitle (TString)[] The title of the entry in the legend box
     */
    void addGammaLaw(Double_t alpha, Double_t beta, Color_t color = kBlack, TString stitle = "");

    //! Add in the current graph the CDF of a Beta law defined by theses parameters \f$ \alpha, \beta, \min, \max \f$
    /*!
     \param alpha (Double_t) the alpha value of the law
     \param beta (Double_t) the beta value of the law
     \param min (Double_t) the minnimum value of the law
     \param max (Double_t) the maximum value of the law
     \param color (Color_t)[kBlack] the color of the line of the CDF's law
     \param stitle (TString)[] the title of the entry in the legend box
     */
    void addBetaLaw(Double_t alpha, Double_t beta, Double_t min, Double_t max, Color_t color = kBlack, TString stitle = "");

    //! Treats the options of the computeScore method
    /*!
     The option of the computeScore method can be :
     \li "normal"
     \li"normal:normal(200,35)" to test the normal law with \f$ (\mu, \sigma) = (200, 35) \f$
     */
    void treatOption(Option_t * option);
    //@}


    //---------------------------------------------
    /** @name Computes the PValues
     * from the book D'Agostino and Stephens  "Goodness-of-Fit Techniques"
     * Depends on the case :
     * \li Case0 : The parameters of the law are defined
     * \li Case3 : The parameters of the law are not defined; then they are estimated by the MLE ("Maximum LikeLihood Estimate").
     * and the type of the EDF test (KS, CvM ou AD)
     **/
    //@{
    /*!
     * \brief Compute the PValue of the test
     * \param tlaw Type of Law to test
     */
    void computePValue(URANIE::DataServer::TStochasticAttribute::ELawType tlaw);


    /*!
     * \brief Compute the PValue of the test KS in all cases (Case0 && Case3)
     * \param dz The \f$ D^{\star} \f$ value of the KS test (modified score)
     * \return The PValue
     */
    Double_t computePValue_KS_CaseAll(Double_t dz);


    /*!
     * \brief Compute the PValue of the test AD in case Case0 (all parameters of the PDF are given)
     *
     * Internal function.
     * These function comes from the Table 4.3 Page 108 of the "Goodness-of-Fit Techniques" book
     * Distribution of \f$ A^{2}\f$, Case 0 : The table gives \f$ q = I\!\!P [ A^{2} > z ] \f$.
     * Here, we return the Pvalue = \f$ 1 -q\f$.
     *
     * For the AD test, the AD value is not modified (\see Table 4.2 page 105 : \f$ A^{2} \forall n \ge 5\f$)
     *
     * \param dz The \f$A^{2}\f$ value of the AD test (not modified)
     * \return the PValue
     */
    Double_t computePValue_AD_Case0(Double_t dz);

    /*!
     * \brief Compute the PValue of the test AD in case Case3 (all parameters of the PDF are estimated by MLE)
     * \param dz The \f$A^{2}\f$ value of the AD test (not modified)
     * \param tlaw The type of Law
     * \return the PValue
     */
    Double_t computePValue_AD_Case3(Double_t dz, URANIE::DataServer::TStochasticAttribute::ELawType tlaw);

    /*!
     * \brief Compute the PValue of the test CvM in case Case3 (all parameters of the PDF are estimated by MLE)
     * \param dz The \f$W^{2}\f$ value of the CvM test (not modified)
     * \param tlaw The type of Law
     * \return the PValue
     */
    Double_t computePValue_CvM_Case3(Double_t dz, URANIE::DataServer::TStochasticAttribute::ELawType tlaw);
    //@}

    //---------------------------------------------
    /** @name Log Printing
     **/
    //@{
    void setLog()
    {
        _blog = kTRUE;
    }
    void unsetLog()
    {
        _blog = kFALSE;
    }
    void changeLog()
    {
        _blog = _blog ? kFALSE : kTRUE;
    }
    Bool_t getLog()
    {
        return _blog;
    }
    virtual void printLog(Option_t *option = "");
    //@}

protected:
    //---------------------------------------------
    /** @name Protected Methods
     **/
    //@{
    //! Finds the parameters in the option string
    /*!
     Example of string parameters are
     \li "normal(-125.025, 125)"
     \li "normal(125.25, .235)"
     */
    void findParametersLaw(TString soption, TVectorD &vec);

    Int_t nTimesValueLTvalVector(Double_t dval, TVectorD vecValues);

    TGraph *createGraph(Color_t color = kBlack);

    //@}

    ClassDef(URANIE::UncertModeler::TTester, ID_UNCERTMODELER)
    // Tester classe
};
} // Fin du namespace UncertModeler
} // Fin du namespace URANIE
#endif
