//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTTestCramerVonMises.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   $Date$

 \brief  Test File of the class URANIE::DataServer::TTestCramerVonMises

 Created: mer nov 21 07:59:24 2007
 \par Modifications
 \li M. Salvoldi 18/01/2012
 - add 8 news unit tests
 */

#define DBL_PRECISION_COMPARE 1.e-12
#define DBL_PRECISION_COMPARE2 1.e-3

#include <iostream>

//ROOT
#include "TString.h"

// URANIE
#include "TestTTestCramerVonMises.h"
#include "TUniformDistribution.h"
#include "TSampling.h"

TTestCramerVonMisesTest::TTestCramerVonMisesTest() :
        _tdsBliss(NULL), _tcm(NULL)
{
}

TTestCramerVonMisesTest::~TTestCramerVonMisesTest()
{
}

void TTestCramerVonMisesTest::setUp()
{
    TString svar = "w";
    _tdsBliss = new URANIE::DataServer::TDataServer();
    _tdsBliss->fileDataRead("bliss.dat");
    _tdsBliss->computeStatistic(svar);
}

void TTestCramerVonMisesTest::tearDown()
{
    delete _tcm;
    _tcm = NULL;
    delete _tdsBliss;
    _tdsBliss = NULL;
}

void TTestCramerVonMisesTest::testW2blissCase3()
{
    _tcm = new URANIE::UncertModeler::TTestCramerVonMises(_tdsBliss, "w");
    _tcm->unsetLog();
    Double_t dW2 = _tcm->computeScore("normal");
    Double_t dW2m = _tcm->getModifiedScore();

    //ORIGIN: Goodness-of-fit techniques D'Agostino et Stephens (page 125)
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.037, dW2, 1.e-3); // 0.034 without MLE
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.038, dW2m, 1.e-3); // 0.035
}

void TTestCramerVonMisesTest::testW2blissCase0()
{
    _tcm = new URANIE::UncertModeler::TTestCramerVonMises(_tdsBliss, "w");
    _tcm->unsetLog();
    Double_t dW2 = _tcm->computeScore("normal(200,35)");
    Double_t dW2m = _tcm->getModifiedScore();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.187456281308, dW2, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.177404095373, dW2m, DBL_PRECISION_COMPARE);
}

void TTestCramerVonMisesTest::testUniformValues()
{
    URANIE::DataServer::TDataServer *tds = new URANIE::DataServer::TDataServer();
    tds->addAttribute(new URANIE::DataServer::TUniformDistribution("u", -5.0, 2.0));
    tds->getAttribute("u")->delShare();
    URANIE::Sampler::TSampling *fsamp = new URANIE::Sampler::TSampling(tds, "lhs", 10000);
    fsamp->generateSample();

    _tcm = new URANIE::UncertModeler::TTestCramerVonMises(tds, "u");
    _tcm->unsetLog();

    Double_t dcs = _tcm->computeScore("uniform");
    Double_t dcsm = _tcm->computeScore("uniform(-5.00000,2.00000)");
    //Double_t dcsm2 = _tcm->computeScore("uniform(-5.00000,-2.00000)");

    CPPUNIT_ASSERT_DOUBLES_EQUAL(dcs, 0.000000, DBL_PRECISION_COMPARE2);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dcsm, 0.000000, DBL_PRECISION_COMPARE2);

    delete fsamp;
    delete tds;
    tds = NULL;

}

void TTestCramerVonMisesTest::throwExceptionTestCramerVonMisesConstructorTdsNullEmpty()
{
    CPPUNIT_ASSERT_THROW(
            new URANIE::UncertModeler::TTestCramerVonMises(NULL, "w"),
            URANIE::Exceptions::UErrorExceptions);

    URANIE::DataServer::TDataServer* tds = new URANIE::DataServer::TDataServer();
    CPPUNIT_ASSERT_THROW(
            new URANIE::UncertModeler::TTestCramerVonMises(tds , "w"),
            URANIE::Exceptions::UWarningExceptions);
    delete tds;
}

void TTestCramerVonMisesTest::throwExceptionTestCramerVonMisesConstructorMoreAttribut()
{
    _tdsBliss->addAttribute("E", "5.0*w*w");

    CPPUNIT_ASSERT_THROW(
            new URANIE::UncertModeler::TTestCramerVonMises(_tdsBliss, "w:E"),
            URANIE::Exceptions::UErrorExceptions);
}

void TTestCramerVonMisesTest::throwExceptionTestCramerVonMisesConstructorFalseAttribut()
{
    CPPUNIT_ASSERT_THROW(
            new URANIE::UncertModeler::TTestCramerVonMises(_tdsBliss, "toto"),
            URANIE::Exceptions::UErrorExceptions);
}

void TTestCramerVonMisesTest::throwExceptionTestCramerVonMisescomputeScoreNoLaw()
{
    _tcm = new URANIE::UncertModeler::TTestCramerVonMises(_tdsBliss, "w");
    _tcm->unsetLog();

    CPPUNIT_ASSERT_THROW(_tcm->computeScore(),
            URANIE::Exceptions::UWarningExceptions);
}

void TTestCramerVonMisesTest::throwExceptionTestCramerVonMisescomputeScoreBadLaw()
{
    _tcm = new URANIE::UncertModeler::TTestCramerVonMises(_tdsBliss, "w");
    _tcm->unsetLog();
    CPPUNIT_ASSERT_THROW(_tcm->computeScore("invgamma"), URANIE::Exceptions::UWarningExceptions);
    CPPUNIT_ASSERT_THROW(_tcm->computeScore("alpha"), URANIE::Exceptions::UWarningExceptions);
}

void TTestCramerVonMisesTest::throwExceptionTestCramerVonMisescomputeScoreNormalBadDefined()
{
    _tcm = new URANIE::UncertModeler::TTestCramerVonMises(_tdsBliss, "w");
    _tcm->unsetLog();

    CPPUNIT_ASSERT_THROW(_tcm->computeScore("normal()"),
            URANIE::Exceptions::UErrorExceptions);
}

void TTestCramerVonMisesTest::throwExceptionTestCramerVonMisescomputeScoreUniformBadDefined()
{
    _tcm = new URANIE::UncertModeler::TTestCramerVonMises(_tdsBliss, "w");
    _tcm->unsetLog();

    CPPUNIT_ASSERT_THROW(_tcm->computeScore("uniform()"),
            URANIE::Exceptions::UErrorExceptions);
}

void TTestCramerVonMisesTest::throwExceptionTestCramerVonMisescomputeScoreLogNormalBadDefined()
{
    _tcm = new URANIE::UncertModeler::TTestCramerVonMises(_tdsBliss, "w");
    _tcm->unsetLog();

    CPPUNIT_ASSERT_THROW(_tcm->computeScore("lognormal()"),
            URANIE::Exceptions::UErrorExceptions);
}
