//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.6 $
// $State$
//////////////////////////////////////////////////////////////
/*!
  \file   TComposeEval.cxx
  \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
  \date    2011-09-26.10.09.21

  \brief   composition de fonction
*/

#include "TComposeEval.h"

ClassImp (URANIE::Relauncher::TComposeEval);

namespace URANIE
{
namespace Relauncher
{

// Class TSubEval
TSubEval::TSubEval(TEval *fun) :
    TObject(), _eval(fun)
{
}

TSubEval::~TSubEval()
{
}

TEval *TSubEval::getEval()
{
    return _eval;
}

int TSubEval::freeze(int decal, TObjArray &cst, TObjArray &in, TObjArray &out)
{
    int dec;
    URANIE::DataServer::TAttribute *att;

    dec = _dec = decal;
    //std::cout << _eval << " submulti in\n" ;
    _eval->freeze();
    TIter next_in(_eval->getInputList());
    while ((att = (URANIE::DataServer::TAttribute *) next_in()))
    {
        if (!cst.FindObject(att) && !in.FindObject(att) && !out.FindObject(att))
            in.Add(att);
    }
    TIter next_out(_eval->getOutputList());
    while ((att = (URANIE::DataServer::TAttribute *) next_out()))
    {
        out.Add(att);
        dec++;
    }
    return dec;
}

void TSubEval::match(TObjArray &allvar)
{
    TObjArray *var;
    URANIE::DataServer::TAttribute *att;
    int i;
    
    // //Input
    var = _eval->getInputList();
    _select.resize(var->GetEntries());
    TIter next(var);
    i = 0;
    while ((att = (URANIE::DataServer::TAttribute *) next()))
	_select[i++] = allvar.IndexOf(att);

    //Output
    var = _eval->getOutputList();
    _selectout.resize(var->GetEntries());
    TIter nextout(var);
    i = 0;
    while ((att = (URANIE::DataServer::TAttribute *) nextout()))
        _selectout[i++] = allvar.IndexOf(att);

}

void TSubEval::_rebuild(int rank, Bool_t chef)
{
    _eval = _eval->rebuild(rank, chef);
}

TSubEval* TSubEval::rebuild(int rank, Bool_t chef)
{
    TSubEval *ret = new TSubEval(*this);

    ret->_rebuild(rank, chef);
    return ret;
}

void TSubEval::unbuild()
{
    if (_eval->unbuild())
    {
        delete _eval;
        _eval = NULL;
    }
}

int TSubEval::eval(vector<URANIE::DataServer::UEntry*> *in, vector<URANIE::DataServer::UEntry*> *out, int ncalc)
{
    
    UNUSED(ncalc);
    vector<URANIE::DataServer::UEntry*> tmp;
    vector<URANIE::DataServer::UEntry*> tmpo;
    Int_t i;
    TEval *titi = dynamic_cast<TEval*>(_eval);

    for (i = 0; i < _eval->getInputList()->GetEntries(); i++)
	tmp.push_back( in->at(_select[i]) );

    for (i = 0; i < _eval->getOutputList()->GetEntries(); i++)
	tmpo.push_back( out->at(_selectout[i]) );

    Double_t res = titi->eval(&tmp, &tmpo, ncalc);
    tmp.clear();
    return (int) res;
}

  
// Class TComposeEval
TComposeEval::TComposeEval():
    TShareEval()
{
    _eval = new TList();
    _eval->SetOwner(kTRUE);
}

TComposeEval::~TComposeEval()
{
    delete _eval;
    _eval = NULL;
}

bool TComposeEval::isThreadSafe()
{
    TSubEval *onefun;

    TIter next(_eval);
    while ((onefun = (TSubEval *) next()))
    {
	if( !onefun->getEval()->isThreadSafe() )
	    return false;
    }
    
    return true;
}

void TComposeEval::addEval(TEval *un)
{
    _eval->Add(new TSubEval(un));
}

void TComposeEval::setEvals(Int_t size, TEval *un, ...)
{
    va_list ap;
    int i;
    TEval *AnEval;

    if (_freezed)
        throw TEval::FreezedEval();

    va_start(ap, un);
    _eval->Add((TObject *) new TSubEval(un));
    for (i = 1; i < size; i++)
    {
        AnEval = va_arg(ap, TEval*);
        _eval->Add((TObject *) new TSubEval(AnEval));
    }
    va_end(ap);
}

void TComposeEval::addConstantValue(URANIE::DataServer::TAttribute *att,
				    double val)
{
    _const.Add(att);
    _val.push_back(val);
}

void TComposeEval::freeze()
{
    // TEval *fun;
    TSubEval *onefun;
    URANIE::DataServer::TAttribute *att;
    int dec;

    if (_freezed)
        return;
    TBaseEval::freeze();

    /* extraction des listes _input _output */
    TIter next(_eval);
    dec = 0;
    while ((onefun = (TSubEval *) next()))
    {
        dec = onefun->freeze(dec, _const, _input, _output);
    }

    /* identification des inputs par codes */
    /* liste complete */
    TObjArray allput;

    TIter next_cst(&_const);
    while ((att = (URANIE::DataServer::TAttribute *) next_cst()))
        allput.Add(att);

    TIter next_in(&_input);
    while ((att = (URANIE::DataServer::TAttribute *) next_in()))
        allput.Add(att);

    TIter next_out(&_output);
    while ((att = (URANIE::DataServer::TAttribute *) next_out()))
        allput.Add(att);

    _allsize = allput.GetEntries();

    /* recherche des inputs */
    next.Reset();
    while ((onefun = (TSubEval *) next()))
        onefun->match(allput);

}

void TComposeEval::_rebuild(int rank, Bool_t chef)
{
    TSubEval *onefun;
    TList *tmp = _eval;

    _eval = new TList();
    _eval->SetOwner(kTRUE);

    TIter next(tmp);
    while ((onefun = (TSubEval*) next()))
    {
        _eval->Add(onefun->rebuild(rank, chef));
    }
}

TEval *TComposeEval::rebuild(int rank, Bool_t chef)
{
    TComposeEval *ret = new TComposeEval(*this);

    ret->_rebuild(rank, chef);

    return (TEval *) ret;
}

Bool_t TComposeEval::unbuild()
{
    TSubEval *one;

    TIter next(_eval);
    while ((one = (TSubEval*) next()))
    {
        one->unbuild();
    }
    return kTRUE;
}

int TComposeEval::eval(vector<URANIE::DataServer::UEntry*> *in, vector<URANIE::DataServer::UEntry*> *out, int ncalc)
{
    UNUSED(ncalc);
    vector<URANIE::DataServer::UEntry*> memo;
    int i, dec, ret=-1;
    TSubEval *fun;

    /* copie des constantes */
    TIter next_cst(&_const);
    i = 0;
    while (next_cst())
    {
	std::stringstream buffer; buffer << "constant_" << i;
	memo.push_back( new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::Double, buffer.str()) );
	*memo.back()->val.dou = _val[i];
	i++;
    }
    /* copie des entrées */
    for(vector<URANIE::DataServer::UEntry*>::iterator it=in->begin(); it!=in->end(); it++)
    {
	memo.push_back( (*it) );
	i++;
    }
    dec = i;

    //dec = i;
    for(vector<URANIE::DataServer::UEntry*>::iterator it=out->begin(); it!=out->end(); it++)
    {
	memo.push_back( (*it) );
    }
  
    /* composition des evals */
    TIter next(_eval);
    while ((fun = (TSubEval *) next()))
    {
	//vector<URANIE::DataServer::UEntry*> tempo(memo.begin()+dec, memo.end());
	ret = fun->eval(&memo, &memo, ncalc);
	if (!ret)
	    break;
    }

    next_cst.Reset();
    i = 0;
    while (next_cst())
    {
	delete memo.at(i++);
    }
    memo.clear();
    return ret;

}
  
}
} //namespace
