//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.3 $
// $State$
//////////////////////////////////////////////////////////////
/*!
  \file   TFlatScript.cxx
  \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
  \date    2011-09-19.15.01.02

  \brief   entree sur une ligne
*/

#include "TFlatScript.h"
#include "UEntry.h"

#include <iostream>
#include <fstream>

ClassImp (URANIE::Relauncher::TShareScript);
ClassImp (URANIE::Relauncher::TVoidScript);
ClassImp (URANIE::Relauncher::TFlatScript);

namespace URANIE
{
namespace Relauncher
{

/* TShareScript */
TShareScript::TShareScript(const char *name) :
    TCodeScript(),
    _path(name), _format("%g")
{
    _svecbeg=TString(""); _svecend=TString(""); _svecdel=TString(",");
    _sstrbeg=TString("");  _sstrend=TString("");  
}

TShareScript::~TShareScript () {}

void TShareScript::setVectorProperties( TString beg, TString delim, TString end)
{
    _svecbeg=beg; _svecend=end; _svecdel=delim;
}
  
void TShareScript::setStringProperties( TString beg, TString end)
{
    _sstrbeg=beg; _sstrend=end; 
}

/* TNoParser */
TNoParser::TNoParser(URANIE::DataServer::TAttribute *att) :
    TObject(), _att(att)
{
    _index = -1;
}

TNoParser::~TNoParser() {}

void TNoParser::build(int id)
{
    _index = id;
}

/* TVoidScript */
TVoidScript::TVoidScript(const char *name) :
    TShareScript(name)
{
    _vars.SetOwner(kTRUE);
}

TVoidScript::~TVoidScript () {}


void TVoidScript::addInput(URANIE::DataServer::TAttribute *att)
{
    _vars.Add(new TNoParser(att));
}

void TVoidScript::setInputs(Int_t size, URANIE::DataServer::TAttribute *att,
			    ...)
{
    va_list ap;
    int i;
    URANIE::DataServer::TAttribute *attr;

    _vars.Add(new TNoParser(att));
    va_start(ap, att);
    for (i = 1; i < size; i++)
    {
        attr = va_arg(ap, URANIE::DataServer::TAttribute *);
        _vars.Add(new TNoParser(attr));
        //std::cout << attr->GetName() << std::endl;
    }
    va_end(ap);
}

void TVoidScript::build(TString &ref, TObjArray *in)
{
    UNUSED(ref); //Only used for flag where a reference file is needed
    TIter next(&_vars);
    TNoParser *a;
    int id;
    //URANIE::DataServer::TAttribute *att;

    // TODO gestion multi file
    while ((a = (TNoParser *) next())) {
        if ((id = in->IndexOf(a->_att)) == -1) {
            id = in->GetEntries();
            in->Add(a->_att);
        }
        a->build(id);
    }

}

/* TFlatScript */
TFlatScript::TFlatScript(const char *name) :
    TVoidScript(name)
{
}

TFlatScript::~TFlatScript() {}

  
void TFlatScript::put(TString &ref, TString &work, Double_t *in)
{
    UNUSED(ref); //Only used for flag where a reference file is needed
    char* copy = new char[work.Length() + _path.Length() + 3];
#ifdef WIN32
    sprintf(copy, "%s\\%s", work.Data(), _path.Data());
#else
    sprintf(copy, "%s/%s", work.Data(), _path.Data());
#endif

    std::ofstream fout(copy, std::ios::out);
    TIter next(&_vars);
    TNoParser *np;

    fout.precision(SSTR_PREC);
    np = (TNoParser *) next();
    fout << in[np->_index];
    while ((np = (TNoParser *) next()))
        fout << " " << in[np->_index];
    fout << std::endl;
    fout.close();
	
    delete[] copy;
}

void TFlatScript::put(TString &ref, TString &work, vector<URANIE::DataServer::UEntry*> *in)
{
    UNUSED(ref); //Only used for flag where a reference file is needed
    char* copy = new char[work.Length() + _path.Length() + 3];
#ifdef WIN32
    sprintf(copy, "%s\\%s", work.Data(), _path.Data());
#else
    sprintf(copy, "%s/%s", work.Data(), _path.Data());
#endif


    std::ofstream fout(copy, std::ios::out);
    TIter next(&_vars);
    TNoParser *np;

    fout.precision(SSTR_PREC);
    while ((np = (TNoParser *) next()))
    {
        if(in->at(np->_index)->tag==URANIE::DataServer::UEntry::Vector)
            fout << " " << _svecbeg << in->at(np->_index)->dumpStr( _svecdel ,"","",_format) << _svecend;
        else
            fout << " " << ((in->at(np->_index)->tag==URANIE::DataServer::UEntry::String) ? _sstrbeg : "")
                 << in->at(np->_index)->dumpStr( "","","",_format)
                 << ((in->at(np->_index)->tag==URANIE::DataServer::UEntry::String) ? _sstrend : "");
    }
    fout << std::endl;
    fout.close();
	
    delete[] copy;
}

}
} //namescpace

