//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.4 $
// $State$
//////////////////////////////////////////////////////////////
/*!
  \file   TKeyResult.cxx
  \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
  \date    2012-06-07.16.29.59

  \brief   gestion format key valeur des fichiers résultats
*/

#include "TKeyResult.h"

#include <iostream>

ClassImp (URANIE::Relauncher::TKeyResult);

namespace URANIE
{
namespace Relauncher
{

/*** TKeyFinder ***/
TRegexp TKeyFinder::_value("[-+0-9.eE]+");

/* Constructeur et destructeur */
TKeyFinder::TKeyFinder(const char *name, URANIE::DataServer::TAttribute *att) :
    TNamed(name, name), _re_full(Form("[ \t]*%s[ \t]*=[ \t]*[-+0-9.eE]+[ \t]*;?[ \t]*", GetName())),
    _re_equ("=[ \t]*[-0-9.eE]"), _re_fin("[-+0-9.eE]+"), _re_equ_empty(""), _re_fin_empty(""), _att(att)
{
    _svecbeg=TString(""); _svecend=TString(""); _svecdel=TString("");
    _sstrbeg=TString("");  _sstrend=TString("");

    switch( att->getDataType() )
    {
    case URANIE::DataServer::TAttribute::kString:
	_re_full = Form("[ \t]*%s[ \t]*=[ \t]*%s[ \t]*[a-zA-Z_]+[ \t]*%s[ \t]*", GetName(), _sstrbeg.Data(), _sstrend.Data());
	_re_equ=Form("=[ \t]*%s[ \t]*[a-zA-Z_]",_sstrbeg.Data());
	_re_fin=Form("[A-Za-z_]+[ \t]*%s", _sstrend.Data());
	break;
    default:
         break;
    }   
}
  
void TKeyFinder::setVectorProperties( TString beg, TString delim, TString end)
{
    _svecbeg=beg; _svecend=end; _svecdel=delim;
    if( _svecbeg!="" || _svecend!="" || _svecdel!="" )
    {
	_re_full = Form("[ \t]*%s[ \t]*=[ \t]*%s[ \t]*[-+0-9.eE%s]+[ \t]*%s[ \t]*;?[ \t]*", GetName(), _svecbeg.Data(), _svecdel.Data(), _svecend.Data());
	_re_equ=Form("=[ \t]*%s[ \t]*[-+0-9.]", _svecbeg.Data());
	_re_fin=Form("[-+0-9.eE][ \t]*%s", _svecend.Data());
        _re_equ_empty=Form("[ \t]*%s[ \t]*=[ \t]*%s[ \t]*[-+0-9.]*", GetName(), _svecbeg.Data());
	_re_fin_empty=Form("[-+0-9.eE]*[ \t]*%s", _svecend.Data());
    }
    else
    {
	_re_equ="=[ \t]*[-0-9.eE]";
	_re_fin="[-+0-9.eE]+";
	_re_full=Form("[ \t]*%s[ \t]*=[ \t]*[-+0-9.eE]+[ \t]*;?[ \t]*", GetName());
    }
}

void TKeyFinder::setStringProperties( TString beg, TString end)
{
    _sstrbeg=beg; _sstrend=end;
    _re_full = Form("[ \t]*%s[ \t]*=[ \t]*%s[ \t]*[a-zA-Z_]+[ \t]*%s[ \t]*", GetName(), _sstrbeg.Data(), _sstrend.Data());
    _re_equ=Form("=[ \t]*%s[ \t]*[a-zA-Z_]",_sstrbeg.Data());
    _re_fin=Form("[A-Za-z_]+[ \t]*%s", _sstrend.Data());
}
  
TKeyFinder::~TKeyFinder(void)
{
}

/* public methode */
Bool_t TKeyFinder::match(TString &line)
{
    bool ismatched=line.Contains(_re_full);
    if( (_svecbeg!="" || _svecend!="" || _svecdel!="") && ismatched==false)
        ismatched=(line.Contains(_re_equ_empty)&&line.Contains(_re_fin_empty));

    //    XYZ<<"Matching "<<GetName()<<"  "<<ismatched<<endl;    
    return ismatched;

}

double TKeyFinder::get(TString &line)
{
    return getTStr(line).Atof();
}
 
  
void TKeyFinder::get(TString &line, URANIE::DataServer::UEntry *entry)
{
    bool isString = entry->tag==URANIE::DataServer::UEntry::String;
    entry->fillEntry( getTStr(line), ((isString) ? "" : _svecdel), ((isString) ? _sstrbeg : _svecbeg), ((isString) ? _sstrend :_svecend ));
}
  
TString TKeyFinder::getTStr(TString &line)
{
    Ssiz_t start=0, stop=0, len=0;

    start = line.Index(_re_equ, &len, 0);
    start+=len-1;
    if ( start < 0 )
        start=0;
    stop = line.Index(_re_fin, &len, start);
    if( _att->getDataType()==URANIE::DataServer::TAttribute::kVector && (_svecbeg!="" || _svecend!="" || _svecdel!="") )
	stop+=1;
    else
	stop+=( (_att->getDataType()==URANIE::DataServer::TAttribute::kString && _sstrend.Sizeof()!=1) ? len + 1 - _sstrend.Sizeof() : len);
  
    TString val(line);
    val.Replace(stop,line.Sizeof(),"").Replace(0,start,"");
    return val;
}
  
/*** TKeyResult ***/
/* Constructeur destructeur */
TKeyResult::TKeyResult(const char *path) :
    _path(path), _bnoVectorSpecificities(true)
{
}

TKeyResult::~TKeyResult(void)
{
    TObject *one;

    TIter next(&_vars);
    while ((one = next()))
    {
        delete one;
        one = NULL;
    }
}

void TKeyResult::setVectorProperties( TString beg, TString delim, TString end)
{
    TIter next(&_vars);
    TKeyFinder *a;
    while ((a = (TKeyFinder *) next()))
	if(a->_att->getDataType()==URANIE::DataServer::TAttribute::kVector)
	    a->setVectorProperties( beg, delim, end);

    if( beg != "" || end != "" || delim!= "") _bnoVectorSpecificities=false;
}

void TKeyResult::setStringProperties( TString beg, TString end)
{
    TIter next(&_vars);
    TKeyFinder *a;
    while ((a = (TKeyFinder *) next()))
	if(a->_att->getDataType()==URANIE::DataServer::TAttribute::kString)
	    a->setStringProperties( beg, end);
}
  
/* definition */
void TKeyResult::addOutput(URANIE::DataServer::TAttribute *att,
			   const char *hash)
{
    _vars.Add(new TKeyFinder(hash, att));
}

void TKeyResult::setOutputs(Int_t size, URANIE::DataServer::TAttribute *att,
			    const char *hash, ...)
{
    va_list ap;
    int i;
    URANIE::DataServer::TAttribute *attr;
    const char *name;

    _vars.Add(new TKeyFinder(hash, att));
    //std::cout << hash << std::endl;
    va_start(ap, hash);
    for (i = 1; i < size; i++)
    {
        attr = va_arg(ap, URANIE::DataServer::TAttribute *);
        name = va_arg(ap, const char *);
        //std::cout << name << std::endl;
        _vars.Add(new TKeyFinder(name, attr));
    }
    va_end(ap);
}

/* usage */
void TKeyResult::build(TObjArray *out)
{
    TIter next(&_vars);
    TKeyFinder *tmp;

    while ((tmp = (TKeyFinder *) next()))
        out->Add(tmp->_att);
}

int TKeyResult::get(TString &work, Double_t*res)
{
    int i, ret;
    TKeyFinder *tmp;
    TIter next(&_vars);
    TString line;
    char* path = new char[work.Length() + _path.Length() + 2];
#ifdef WIN32
    sprintf(path, "%s\\%s", work.Data(), _path.Data());
#else
    sprintf(path, "%s/%s", work.Data(), _path.Data());
#endif	
    std::ifstream fin(path, std::ios::in);
    if (!fin)
    {
        std::cout << "Cannot open ::  " << path << std::endl;
        delete[] path;
        return 0;
    }

    Bool_t* mark = new Bool_t[_vars.GetEntries()];
    i = 0;
    while ((tmp = (TKeyFinder *) next()))
    {
        mark[i++] = kFALSE;
    }
    while (line.ReadLine(fin))
    {
        if (line[0] == '#')
            continue;

        next.Reset();
        i = 0;
        while ((tmp = (TKeyFinder *) next()))
        {
            if (tmp->match(line))
            {
                res[i] = tmp->get(line);
                mark[i] = kTRUE;
                break;
            }
            else
            {
                i++;
            }
        }
    }
    fin.close();

    ret = 1;
    i = 0;
    next.Reset();
    while ((tmp = (TKeyFinder *) next()))
    {
        if (!mark[i++])
        {
            std::cout << "TKeyResult(" <<_path << "): " << tmp->_att->GetName() << " Not found\n";
            ret = 0;
        }
    }
	
    delete[] path;
    delete[] mark;
    return ((ret==0)?ret:_vars.GetEntries());
}

int TKeyResult::get(TString &work, vector<URANIE::DataServer::UEntry*> *sout)
{
    int i, ret;
    TKeyFinder *tmp;
    TIter next(&_vars);
    TString line;

    char* path = new char[work.Length() + _path.Length() + 2];
#ifdef WIN32
    sprintf(path, "%s\\%s", work.Data(), _path.Data());
#else
    sprintf(path, "%s/%s", work.Data(), _path.Data());
#endif	
    std::ifstream fin(path, std::ios::in);
    if (!fin)
    {
	std::cout << "Cannot open ::  " << path << std::endl;
	return 0;
    }

    Bool_t* mark = new Bool_t[_vars.GetEntries()];
    i = 0;
    while ((tmp = (TKeyFinder *) next()))
    {
	mark[i++] = kFALSE;
    }
    next.Reset();
    while (line.ReadLine(fin))
    {
	if (line[0] == '#')
	    continue;

	next.Reset();
	i = 0;
	while ((tmp = (TKeyFinder *) next()))
        {
            //            XYZ<<tmp->GetName()<<endl;
	    if (tmp->match(line))
	    {
		if( _bnoVectorSpecificities )
		{
		    switch(sout->at(i)->tag)
		    {
		    case URANIE::DataServer::UEntry::Vector: sout->at(i)->val.vec->push_back( tmp->get(line) ); break;
		    default:  tmp->get(line, sout->at(i)); break;
		    }
		}
		else
		    tmp->get(line, sout->at(i));
	      
		mark[i] = kTRUE;
		break;
            }
	    else
	    {
		i++;
	    }
        }
    }
    fin.close();
    ret = 1;
    i = 0;

    next.Reset();
    while ((tmp = (TKeyFinder *) next()))
    {
	if (!mark[i++])
        {
	    std::cout << "TKeyResult(" <<_path << "): " << tmp->_att->GetName() << " Not found\n";
	    ret = 0;
        }
    }

    delete[] path;
    delete[] mark;
    return ret;
}

void TKeyResult::clearFile(TString tmpdir)
{    
#ifdef WIN32
    TString fullPath = tmpdir+"\\"+_path;
    gSystem->Exec( Form("if exist %s del /f %s", fullPath.Data(), fullPath.Data()) ); 
#else
    TString fullPath = tmpdir+"/"+_path;
    gSystem->Exec( Form("rm -f %s", fullPath.Data()) );
#endif    
}

}
} //namespace

