//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.4 $
// $State$
//////////////////////////////////////////////////////////////
/*!
  \file   TKeyScript.cxx
  \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
  \date    2011-09-09.14.50.18

  \brief   jeux de données du TCodeEval
*/

#include "TKeyScript.h"

#include <iostream>
#include <fstream>

//ROOT

ClassImp (URANIE::Relauncher::TKeyScript);

namespace URANIE
{
namespace Relauncher
{

/*** TKeyParser ***/
/* Constructeur et destructeur */
TKeyParser::TKeyParser(const char *name, URANIE::DataServer::TAttribute *att) :
    TNamed(name, name), _att(att), _re_equ("=[ \t]*[-0-9.eE]"), _re_fin("[-+0-9.eE]+"),
    _re_full(Form("[ \t]*%s[ \t]*=[ \t]*[-+0-9.eE]+[ \t]*;?[ \t]*", GetName()))
{
    _line = _istart = _istop = _index = -1;
    _svecbeg=TString("["); _svecend=TString("]"); _svecdel=TString(",");
    _sstrbeg=TString("");  _sstrend=TString("");

    switch( att->getDataType() )
    {
    case URANIE::DataServer::TAttribute::kString:
	_re_full = Form("[ \t]*%s[ \t]*=[ \t]*%s[ \t]*[a-zA-Z_]+[ \t]*%s[ \t]*", GetName(), _sstrbeg.Data(), _sstrend.Data());
	_re_equ=Form("=[ \t]*%s[ \t]*[a-zA-Z_]",_sstrbeg.Data());
	_re_fin=Form("[A-Za-z_]+[ \t]*%s", _sstrend.Data());
	break;
    case URANIE::DataServer::TAttribute::kVector:
	_re_full = Form("[ \t]*%s[ \t]*=[ \t]*%s[ \t]*[-+0-9.eE%s \t]+[ \t]*%s[ \t]*;?[ \t]*", GetName(), _svecbeg.Data(), _svecdel.Data(), _svecend.Data());
	_re_equ=Form("=[ \t]*%s[ \t]*[-+0-9.]", _svecbeg.Data());
	_re_fin=Form("[-+0-9.eE][ \t]*%s", _svecend.Data());
	break;
    default:
        break;
    }   
}

void TKeyParser::setVectorProperties( TString beg, TString delim, TString end)
{
    _svecbeg=beg; _svecend=end; _svecdel=delim;
    _re_full = Form("[ \t]*%s[ \t]*=[ \t]*%s[ \t]*[-+0-9.eE%s \t]+[ \t]*%s[ \t]*;?[ \t]*", GetName(), _svecbeg.Data(), _svecdel.Data(), _svecend.Data());
    _re_equ=Form("=[ \t]*%s[ \t]*[-+0-9.]", _svecbeg.Data());
    _re_fin=Form("[-+0-9.eE][ \t]*%s", _svecend.Data());
}

void TKeyParser::setStringProperties( TString beg, TString end)
{
    _sstrbeg=beg; _sstrend=end; 
    _re_full = Form("[ \t]*%s[ \t]*=[ \t]*%s[ \t]*[a-zA-Z_]+[ \t]*%s[ \t]*", GetName(), _sstrbeg.Data(), _sstrend.Data());
    _re_equ=Form("=[ \t]*%s[ \t]*[a-zA-Z_]",_sstrbeg.Data());
    _re_fin=Form("[A-Za-z_]+[ \t]*%s", _sstrend.Data());
}

TKeyParser::~TKeyParser(void)
{
}

//Bool_t TKeyParser::IsSortable(void) {return kTRUE;}

int TKeyParser::Compare(const TObject *other) const
{
    const TKeyParser *tmp = dynamic_cast<const TKeyParser*>(other);
    return (_line < tmp->_line) ? -1 : 1;
}

void TKeyParser::build(int id, std::istream &fi)
{
  
    TString titi;
    Ssiz_t tmp;

    _index = id;
    /* parse index */
    _line = 0;
    titi.ReadLine(fi, kFALSE);

    while ( !titi.Contains(_re_full) )
    {
	_line++;
	titi.ReadLine(fi, kFALSE);

	if (fi.eof())
	    throw KeyNotFound(GetName());
    }

    _istart = titi.Index(_re_equ, &tmp, 0);
    _istart+=tmp-1;
    _istop = titi.Index(_re_fin, &tmp, _istart);
    _istop+=((_att->getDataType()==URANIE::DataServer::TAttribute::kVector) ? 1 : ( (_att->getDataType()==URANIE::DataServer::TAttribute::kString && _sstrend.Sizeof()!=1) ? tmp + 1 - _sstrend.Sizeof() : tmp));
    _istop-=_istart;

    // std::cout <<__FILE__ << "::" << __LINE__ << "  " << titi << " line " <<  _line << " istart " << _istart << " istop " << _istop << " tmp ? " << tmp <<  std::endl;
}

TString &TKeyParser::replace(TString *line, Double_t val)
{
    char strnum[32];
    sprintf(strnum, "%.16g", val);
    return line->Replace(_istart, _istop, strnum);
}

TString &TKeyParser::replace(TString *line, URANIE::DataServer::UEntry *val, TString _format)
{
    return line->Replace(_istart, _istop, val->dumpStr( _svecdel ,"","",_format) );
}

/*** TKeyScript *******/
TKeyScript::TKeyScript(const char *name) :
    _path(name), _format("%g")
{
}

TKeyScript::~TKeyScript(void)
{
    TKeyParser *one;

    TIter next(&_vars);
    while ((one = (TKeyParser*) next()))
    {
        delete one;
        one = NULL;
    }
}

void TKeyScript::setVectorProperties( TString beg, TString delim, TString end)
{
    TIter next(&_vars);
    TKeyParser *a;
    while ((a = (TKeyParser *) next()))
	if(a->_att->getDataType()==URANIE::DataServer::TAttribute::kVector)
	    a->setVectorProperties( beg, delim, end);
}
  
void TKeyScript::setStringProperties( TString beg, TString end)
{
    TIter next(&_vars);
    TKeyParser *a;
    while ((a = (TKeyParser *) next()))
	if(a->_att->getDataType()==URANIE::DataServer::TAttribute::kString)
	    a->setStringProperties( beg, end);
}
  
void TKeyScript::addInput(URANIE::DataServer::TAttribute *att, const char *hash)
{
    _vars.Add(new TKeyParser(hash, att));
}

void TKeyScript::setInputs(Int_t size, URANIE::DataServer::TAttribute *att,
			   const char *hash, ...)
{
    va_list ap;
    int i;
    URANIE::DataServer::TAttribute *attr;
    const char *name;

    _vars.Add(new TKeyParser(hash, att));
    //std::cout << hash << std::endl;
    va_start(ap, hash);
    for (i = 1; i < size; i++)
    {
        attr = va_arg(ap, URANIE::DataServer::TAttribute *);
        name = va_arg(ap, const char *);
        //std::cout << name << std::endl;
        _vars.Add(new TKeyParser(name, attr));
    }
    va_end(ap);
}

void TKeyScript::build(TString &ref, TObjArray *in)
{
    TIter next(&_vars);
    TString reff = TString::Format("%s/%s", ref.Data(), _path.Data());
    TKeyParser *a;
    int id;

    while ((a = (TKeyParser *) next()))
    {

        if ((id = in->IndexOf(a->_att)) == -1)
        {
            id = in->GetEntries();
            in->Add(a->_att);
        }
        std::ifstream f(reff, std::ios::in);
        if (!f)
	    throw FileNotFound(reff);
	
        a->build(id, f);

        f.close();
    }

    _vars.Sort();
}

void TKeyScript::put(TString &ref, TString &work, Double_t *in)
{
    char* reff = new char[ref.Length() + _path.Length() + 2];
    char* copy = new char[work.Length() + _path.Length() + 2];
#ifdef WIN32
    sprintf(reff, "%s\\%s", ref.Data(), _path.Data());
    sprintf(copy, "%s\\%s", work.Data(), _path.Data());
#else	
    sprintf(reff, "%s/%s", ref.Data(), _path.Data());
    sprintf(copy, "%s/%s", work.Data(), _path.Data());
#endif	
    std::ifstream fin(reff, std::ios::in);
    std::ofstream fout(copy, std::ios::out);
    TIter next(&_vars);
    TString a;
    int iline;
    TKeyParser *match;

    iline = 0;
    match = (TKeyParser *) next();
    while (a.ReadLine(fin, kFALSE))
    {

        if (match && match->_line == iline)
        {
            a = match->replace(&a, in[match->_index]);
            match = (TKeyParser *) next();
        }
        fout << a << std::endl;
        iline++;
    }
    fin.close();
    fout.close();
	
    delete[] reff;
    delete[] copy;
}

void TKeyScript::put(TString &ref, TString &work, vector<URANIE::DataServer::UEntry*> *in)
{
    char* reff = new char[ref.Length() + _path.Length() + 2];
    char* copy = new char[work.Length() + _path.Length() + 2];
#ifdef WIN32
    sprintf(reff, "%s\\%s", ref.Data(), _path.Data());
    sprintf(copy, "%s\\%s", work.Data(), _path.Data());
#else	
    sprintf(reff, "%s/%s", ref.Data(), _path.Data());
    sprintf(copy, "%s/%s", work.Data(), _path.Data());
#endif	
    std::ifstream fin(reff, std::ios::in);
    std::ofstream fout(copy, std::ios::out);
    TIter next(&_vars);
    TString a;
    int iline;
    TKeyParser *match;

    iline = 0;
    match = (TKeyParser *) next();
    while (a.ReadLine(fin, kFALSE))
    {
	if (match && match->_line == iline)
        {
	    a = match->replace(&a, in->at(match->_index), _format);
	    match = (TKeyParser *) next();
        }
	fout << a << std::endl;
	iline++;
    }
    fin.close();
    fout.close();
	
    delete[] reff;
    delete[] copy;
}

}
} //namespace

