//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.2 $
// $State$
//////////////////////////////////////////////////////////////
/*!
  \file   TMasterVar.cxx
  \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
  \date    2012-04-27.14.40.09

  \brief   gestion des variables
*/

#include "TMasterVar.h"
//#include "TDSNtupleD.h"

#include "UExceptions.h"

namespace URANIE
{
namespace Relauncher
{

// TMasterVar
TMasterVar::TMasterVar(URANIE::DataServer::TAttribute *ref, int id) : TObject()
{
    _ref = ref;
    _id = _id2 = id;
    //std::cout << ((TNamed *)ref)->GetName() << " with " << id << std::endl;
}

void TMasterVar::allocCodeEntry(vector<URANIE::DataServer::UEntry*> &vals)
{
    URANIE::DataServer::TAttribute* att = (URANIE::DataServer::TAttribute*) _ref;

    switch ( att->getDataType() )
    {
    case URANIE::DataServer::TAttribute::kVector :
        vals.push_back( new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::Vector, string(att->GetName())) );
        break;
    case URANIE::DataServer::TAttribute::kString :
        vals.push_back( new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::String, string(att->GetName())) );
        break;
    default:
        vals.push_back( new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::Double, string(att->GetName())) );
        break;
    }
}

void TMasterVar::allocGradEntry(vector<URANIE::DataServer::UEntry*> &vals)
{
    URANIE::DataServer::TAttribute* att = (URANIE::DataServer::TAttribute*) _ref;

    switch ( att->getDataType() )
    {
    case URANIE::DataServer::TAttribute::kVector :
    case URANIE::DataServer::TAttribute::kString :
        vals.push_back( new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::Empty, string(att->GetName())) );
        break;
    default:
        vals.push_back( new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::Vector, string(att->GetName())) );
        break;
    }
}

void TMasterVar::allocDataEntry(int dec, vector<URANIE::DataServer::UEntry*> &vals)
{
    URANIE::DataServer::TAttribute* att = (URANIE::DataServer::TAttribute*) _ref;

    if (isExported()) {
        switch ( att->getDataType() )
        {
        case URANIE::DataServer::TAttribute::kVector :
            vals[dec+_id2] = new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::Vector, string(att->GetName())) ;
            break;
        case URANIE::DataServer::TAttribute::kString :
            vals[dec+_id2] = new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::String, string(att->GetName())) ;
            break;
        default:
            vals[dec+_id2] = new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::Double, string(att->GetName())) ;
            break;
        }
    }
}

// TInputVar
TInputVar::TInputVar(URANIE::DataServer::TAttribute *ref, int id) :
    TMasterVar(ref, id)
{
}

bool TInputVar::isExported()
{
    return true;
}

bool TInputVar::isValid() 
{
    return true;
}

bool TInputVar::put(int i, const double *solv, double *code)
{
    code[i] = solv[_id];
    return true;
}

bool TInputVar::get(int i, double *solv, double *code)
{
    solv[_id] = code[i];
    return true;
}

bool TInputVar::put(int i, const double *solv, vector<URANIE::DataServer::UEntry*> *code)
{
    code->at(i)->pushEntry( solv[_id]);
    return true;
}

bool TInputVar::get(int i, double *solv, vector<URANIE::DataServer::UEntry *> *code)
{
    if (code->at(i)->tag != URANIE::DataServer::UEntry::Double) return false;
    solv[_id] = *(code->at(i)->val.dou);
    return true;
}

bool TInputVar::put(int i, const vector<URANIE::DataServer::UEntry*> *solv, vector<URANIE::DataServer::UEntry*> *code)
{
    code->at(i)->copyEntry( solv->at(_id) );
    return true;
}

bool TInputVar::get(int i, vector<URANIE::DataServer::UEntry*> *solv, vector<URANIE::DataServer::UEntry*> *code)
{
    solv->at(_id)->copyEntry( code->at(i) );
    return true;
}

// TOutputVar
TOutputVar::TOutputVar(URANIE::DataServer::TAttribute *ref, int id, int id2) :
    TMasterVar(ref, id)
{
    _temp = true;
    _id2 = id2;
}

void TOutputVar::setTemp()
{
    _temp = false;
}

void TOutputVar::reindice(int i)
{
    _id2 += i;
}

bool TOutputVar::isExported()
{
    return (_temp) ? true : false;
}

bool TOutputVar::isPre()
{
    return false ;
}

bool TOutputVar::isPost()
{
    return false ;
}

bool TOutputVar::put(int i, const double *solv, double *code)
{

    UNUSED(i,solv, code);
    return false;
}

bool TOutputVar::get(int i, double *solv, double *code)
{
    if (_temp)
        solv[i] = code[_id];
    return _temp;
}

bool TOutputVar::put(int i, const double *solv, vector<URANIE::DataServer::UEntry*> *code)
{
    UNUSED(i,solv, code);
    return false;
}

bool TOutputVar::get(int i, double *solv, vector<URANIE::DataServer::UEntry*> *code)
{
    if (code->at(_id)->tag != URANIE::DataServer::UEntry::Double) return false;
    if (_temp)
        solv[i] = *(code->at(_id)->val.dou);
    return _temp;
}

bool TOutputVar::put(int i, const vector<URANIE::DataServer::UEntry*> *solv, vector<URANIE::DataServer::UEntry*> *code)
{
    UNUSED(i,solv,code);
    return false;
}

bool TOutputVar::get(int i, vector<URANIE::DataServer::UEntry*> *solv, vector<URANIE::DataServer::UEntry*> *code)
{
    if (_temp)
	solv->at(i)->copyEntry( code->at(_id) );
    return _temp;
}

/*** Input ***/
// TSolvVar
TSolvVar::TSolvVar(URANIE::DataServer::TAttribute *ref, int id) :
    TInputVar(ref, id)
{
}

// TConstVar
TConstVar::TConstVar(URANIE::DataServer::TAttribute *ref) :
    TInputVar(ref, -1), _val(NULL)
{
    URANIE::DataServer::TAttribute* att = (URANIE::DataServer::TAttribute*) ref;
    switch ( att->getDataType() )
    {
    case URANIE::DataServer::TAttribute::kVector :
	_val = new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::Vector, string(att->GetName())+string("_ConstVar")) ;
	break;
    case URANIE::DataServer::TAttribute::kString :
	_val = new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::String, string(att->GetName())+string("_ConstVar")) ;
	break;
    default:
	_val = new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::Double, string(att->GetName())+string("_ConstVar")) ; 
	break;
    }
}

TConstVar::~TConstVar()
{
    if(_val!=NULL)
	delete _val;
    _val=NULL;
}

void TConstVar::setValue(double val)
{
    _val->assignEntry(&val);
    _id = -2;
}

void TConstVar::setValue(URANIE::DataServer::UEntry *constant)
{
    _val->copyEntry(constant);
    _id = -2;
}

bool TConstVar::isExported()
{
    return false;
}

bool TConstVar::isValid()
{
    if (_id != -2)
        throw URANIE::Exceptions::UExceptions(__FILE__, __LINE__,
					      Form("TMaster Constant Unset, %s",
						   ((TNamed *) _ref)->GetName()));
    return true;
}

bool TConstVar::put(int i, const double *solv, double *code)
{
    UNUSED(solv);
    code[i] = *_val->val.dou;
    return true;
}

bool TConstVar::get(int i, double *solv, double *code)
{
    UNUSED(i,solv,code);
    return false;
}

bool TConstVar::put(int i, const double *solv, vector<URANIE::DataServer::UEntry*> *code)
{
    UNUSED(solv);
    code->at(i)->copyEntry(_val);
    return true;
}

bool TConstVar::get(int i, double *solv, vector<URANIE::DataServer::UEntry*> *code)
{
    UNUSED(i,solv,code);
    return false;
}

bool TConstVar::put(int i, const vector<URANIE::DataServer::UEntry*> *solv, vector<URANIE::DataServer::UEntry*> *code)
{
    UNUSED(solv);
    code->at(i)->copyEntry(_val);
    return true;
}

bool TConstVar::get(int i, vector<URANIE::DataServer::UEntry*> *solv, vector<URANIE::DataServer::UEntry*> *code)
{
    UNUSED(i,solv,code);
    return false;
}

// TAuxiVar
TAuxiVar::TAuxiVar(URANIE::DataServer::TAttribute *ref, int id) :
    TInputVar(ref, id)
{
}

/*** Output ***/
// TCodeVar
TCodeVar::TCodeVar(URANIE::DataServer::TAttribute *ref, int id, int id2) :
    TOutputVar(ref, id, id2)
{
}

// TPreVar
TPreVar::TPreVar(URANIE::DataServer::TAttribute *ref, int id, int id2) :
    TOutputVar(ref, id, id2)
{
}

bool TPreVar::isPre()
{
    return true;
}

// TPostVar
TPostVar::TPostVar(URANIE::DataServer::TAttribute *ref, int id, int id2) :
    TOutputVar(ref, id, id2)
{
}

bool TPostVar::isPost()
{
    return true;
}


}
} // namespace
