//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.2 $
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TMasterVar.h
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date    2012-04-27.14.40.19

 \brief gestion des variables
 */

#ifndef __MASTER_VAR__
#define __MASTER_VAR__

#include <vector>

//ROOT
#include "Rtypes.h"
#include "TObject.h"

//URANIE
#include "TEval.h"
#include "UEntry.h"

namespace URANIE
{
namespace Relauncher
{

// Abstract level
class TMasterVar: public TObject
{
protected:
    URANIE::DataServer::TAttribute *_ref;
    int _id, _id2;

public:
    TMasterVar(URANIE::DataServer::TAttribute *, int);

    URANIE::DataServer::TAttribute *getRef(){return _ref;}
    bool match(URANIE::DataServer::TAttribute *one)
    {
        return one == _ref;
    }
    void allocCodeEntry(vector<URANIE::DataServer::UEntry*>&);
    void allocGradEntry(vector<URANIE::DataServer::UEntry*>&);
    void allocDataEntry(int, vector<URANIE::DataServer::UEntry*>&);

    virtual bool isExported() = 0;
    virtual bool put(int, const double*, double*) = 0;
    virtual bool get(int, double*, double*) = 0;

    virtual bool put(int, const double*, vector<URANIE::DataServer::UEntry*>* ) = 0;
    virtual bool get(int, double*, vector<URANIE::DataServer::UEntry *>*) = 0;

    virtual bool put(int, const vector<URANIE::DataServer::UEntry*>*, vector<URANIE::DataServer::UEntry*>*) = 0;
    virtual bool get(int, vector<URANIE::DataServer::UEntry*>*, vector<URANIE::DataServer::UEntry*>*) = 0;
};

class TInputVar: public TMasterVar
{
public:
    TInputVar(URANIE::DataServer::TAttribute *, int);

    virtual bool isExported();
    virtual bool isValid();
    virtual bool put(int, const double*, double*);
    virtual bool get(int, double*, double*);

    virtual bool put(int, const double*, vector<URANIE::DataServer::UEntry*>*);
    virtual bool get(int, double*, vector<URANIE::DataServer::UEntry *>*);

    virtual bool put(int, const vector<URANIE::DataServer::UEntry*>*, vector<URANIE::DataServer::UEntry*>*);
    virtual bool get(int, vector<URANIE::DataServer::UEntry*>*, vector<URANIE::DataServer::UEntry*>*);
};

class TOutputVar: public TMasterVar
{
protected:
    bool _temp;

public:
    TOutputVar(URANIE::DataServer::TAttribute *, int, int);

    void setTemp();
    void reindice(int);

    virtual bool isExported();
    virtual bool isPre();
    virtual bool isPost();
    virtual bool put(int, const double*, double*);
    virtual bool get(int, double*, double*);

    virtual bool put(int, const double*, vector<URANIE::DataServer::UEntry*>*);
    virtual bool get(int, double*, vector<URANIE::DataServer::UEntry *>*);

    virtual bool put(int, const vector<URANIE::DataServer::UEntry*>*, vector<URANIE::DataServer::UEntry*>*);
    virtual bool get(int, vector<URANIE::DataServer::UEntry*>*, vector<URANIE::DataServer::UEntry*>*);
};

// Concrete level
class TSolvVar: public TInputVar
{
public:
    TSolvVar(URANIE::DataServer::TAttribute *, int);
};

class TConstVar: public TInputVar
{
protected:
    URANIE::DataServer::UEntry *_val;
    bool _keepit;
    
public:
    TConstVar(URANIE::DataServer::TAttribute *);
    ~TConstVar();

    void saveValue(bool b){ _keepit=b;}
    bool getSaveValue(){ return _keepit;}
    void setValue(double);
    void setValue(URANIE::DataServer::UEntry*);
    URANIE::DataServer::UEntry *getValue(){return _val;}

    virtual bool isExported();
    virtual bool isValid();
    virtual bool put(int, const double*, double*);
    virtual bool get(int, double*, double*);

    virtual bool put(int, const double*, vector<URANIE::DataServer::UEntry*>*);
    virtual bool get(int, double*, vector<URANIE::DataServer::UEntry*>*);

    virtual bool put(int, const vector<URANIE::DataServer::UEntry*>*, vector<URANIE::DataServer::UEntry*>*);
    virtual bool get(int, vector<URANIE::DataServer::UEntry*>*, vector<URANIE::DataServer::UEntry*>*);
};

class TAuxiVar: public TInputVar
{
public:
    TAuxiVar(URANIE::DataServer::TAttribute *, int);
};

class TCodeVar: public TOutputVar
{
public:
    TCodeVar(URANIE::DataServer::TAttribute *, int, int);

};

class TPreVar: public TOutputVar
{
public:
    TPreVar(URANIE::DataServer::TAttribute *, int, int);

    virtual bool isPre();
};

class TPostVar: public TOutputVar
{
public:
    TPostVar(URANIE::DataServer::TAttribute *, int, int);

    virtual bool isPost();
};


}
} // namespace

#endif
