//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.4 $
// $State$
//////////////////////////////////////////////////////////////
/*!
  \file   TThreadedRun.cxx
  \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
  \date    2008-08-22.15.45.39

  \brief   execution multithreader de la recherche
*/

#include "TThreadedRun.h"

//#include <iostream>
#include <stdlib.h>

#ifdef WITH_PYTHON
#include "Python.h"
#endif

ClassImp (URANIE::Relauncher::TThreadedRun);

namespace URANIE
{
namespace Relauncher
{

TThreadedRun::TThreadedRun(TEval *user, long nbnod) :
    TShareRun(user)
{
    _ncpus = nbnod;
    _runcore = 0;
    pthread_mutex_init(&_mutex_in, (pthread_mutexattr_t *) NULL);
    pthread_mutex_init(&_mutex_out, (pthread_mutexattr_t *) NULL);
    pthread_cond_init(&_invide, (pthread_condattr_t *) NULL);
    pthread_cond_init(&_outvide, (pthread_condattr_t *) NULL);

    if( !user->isThreadSafe() )
	throw URANIE::Relauncher::TEval::EvalIsNotThreadable();
}

TThreadedRun::~TThreadedRun(void)
{

#ifdef IMPLICIT
    stopSlave();
#endif
    pthread_mutex_destroy(&_mutex_in);
    pthread_mutex_destroy(&_mutex_out);
    pthread_cond_destroy(&_invide);
    pthread_cond_destroy(&_outvide);
}

void TThreadedRun::calcul_loop(int id)
{

    // long ind;
    int ret=-1;
    // double *var;
    // double **res;
    struct element calc;

    //std::cout << id << " Thread\n" ;
    TEval *fun = _fun->rebuild(id);
    calc = serviceGet();
    while (calc.id >= 0)
    {
        ret = fun->eval(calc.inv, calc.outv, calc.calcnb);
        servicePut(calc.id, ret);
        calc = serviceGet();

    }

    if (fun->unbuild())
    {
        delete fun;
        fun = NULL;
    }
}

void *calcul_loop_bg(void *info)
{
    struct job *obj = (struct job *) info;
    obj->run->calcul_loop(obj->id);
    pthread_exit(0);
    return NULL;
}

void TThreadedRun::startSlave(void)
{
    int i;
    _runcore = _ncpus - 1;
    _job.resize(_runcore);
    for (i = 0; i < _runcore; i++)
    {
        _job[i].id = i + 1;
        _job[i].run = this;
        pthread_create(&(_job[i].thr), NULL, calcul_loop_bg, (void *) &_job[i]);
    }
}

Bool_t TThreadedRun::onMaster(void)
{
    return kTRUE;
}

void TThreadedRun::stopSlave(void)
{
    int i;
    //void *ret;
    if (_runcore)
    {
        for (i = 0; i < _runcore; i++)
        {
            put(-1, (vector<URANIE::DataServer::UEntry*>*) NULL, 
                (vector<URANIE::DataServer::UEntry*>*) NULL, 0);
        }
        for (i = 0; i < _runcore; i++)
        {
            pthread_join(_job[i].thr, NULL);
        }
    }
    _runcore = 0;

}

void TThreadedRun::put(long i, std::vector<URANIE::DataServer::UEntry*> *in, std::vector<URANIE::DataServer::UEntry*> *out, int icalc)
{

    bool vide;
    struct element tupl =
	{ i, icalc, in, out};

    pthread_mutex_lock(&_mutex_in);
    vide = _qin.empty();
    _qin.push(tupl);
    if (vide)
	pthread_cond_broadcast(&_invide);

    pthread_mutex_unlock(&_mutex_in);

}

struct element TThreadedRun::serviceGet()
{
    struct element tupl;

    pthread_mutex_lock(&_mutex_in);
    while (_qin.empty())
        pthread_cond_wait(&_invide, &_mutex_in);

    tupl = _qin.front();
    _qin.pop();
    //  std::cout << tupl.id << "Thread get\n";
    pthread_mutex_unlock(&_mutex_in);

    return tupl;
}

void TThreadedRun::servicePut(long id, int ret)
{
    bool vide;
    struct relement tupl;
    pthread_mutex_lock(&_mutex_out);
    vide = _qout.empty();
    tupl.id = id;
    tupl.stat = ret;

//    std::cout << tupl.id << "Thread put\n";
    _qout.push(tupl);
    if (vide)
        pthread_cond_broadcast(&_outvide);

    pthread_mutex_unlock(&_mutex_out);
}

Bool_t TThreadedRun::readyToGet()
{
    return (_qout.empty()) ? kFALSE : kTRUE ;
}

int TThreadedRun::get(long &id)
{
    struct relement tupl;

#ifdef WITH_PYTHON
    PyThreadState *pysave;
    // Py_BEGIN_ALLOW_THREADS
    pysave = (Py_IsInitialized()) ? PyEval_SaveThread() : NULL ;
#endif
//    std::cout << "Client wait\n";
    if (pthread_mutex_lock(&_mutex_out))
        perror("rehum");

    while (_qout.empty())
        pthread_cond_wait(&_outvide, &_mutex_out);

    tupl = _qout.front();
    _qout.pop();
//    std::cout << tupl.id << " " << tupl.stat << " Client get\n";

    pthread_mutex_unlock(&_mutex_out);
#ifdef WITH_PYTHON
    // Py_END_ALLOW_THREADS
    if (pysave) PyEval_RestoreThread(pysave);
#endif
    id = tupl.id;

    return tupl.stat;
}

}
} // namespace

