//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2024 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TBaseModel.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   $Date$

 \brief Implementation of the class URANIE::XMLProblem::TBaseModel

 Created: 2015/10/27 14:03

 */

// CPPUNIT
#define DBL_PRECISION_COMPARE 1.e-12

// Uranie
#include "TestTBaseModel.h"

// ROOT
#include "TString.h"

#include <iostream>

TBaseModelTest::TBaseModelTest():
  _tbm1(NULL), _tbm2(NULL), _tbm3(NULL)
{
}

TBaseModelTest::~TBaseModelTest()
{
}

void TBaseModelTest::setUp()
{
  _tbm1 = new URANIE::XMLProblem::TBaseModel();
  _tbm2 = new URANIE::XMLProblem::TBaseModel("x1:x2:x3", "out1");
  _tbm3 = new URANIE::XMLProblem::TBaseModel("myModel", "x1:x2:x3", "hidden", "out1");
}

void TBaseModelTest::tearDown()
{
  delete _tbm1;
  delete _tbm2;
  delete _tbm3;
}
////////////////////////////////
///                          ///                            
///         Tests            ///
///                          ///
////////////////////////////////
void TBaseModelTest::testConstructors()
{
  TString sDefaultStr = "_default_";
  Int_t nCrt = 10;

  // Empty object
  _tbm1->printLog();
  CPPUNIT_ASSERT(strcmp(_tbm1->GetName(), "DefaultUranieModel") == 0);
  CPPUNIT_ASSERT(strcmp(_tbm1->getInput().Data(), "") == 0);
  CPPUNIT_ASSERT(strcmp(_tbm1->getOutput().Data(), "") == 0);
  CPPUNIT_ASSERT(strcmp(_tbm1->getHidden().Data(), "") == 0);
  CPPUNIT_ASSERT_EQUAL(_tbm1->getLanguage(), URANIE::XMLProblem::TBaseModel::kEMPTY);
  CPPUNIT_ASSERT_EQUAL(_tbm1->getNormalization(), URANIE::XMLProblem::TBaseModel::kCR);
  CPPUNIT_ASSERT(strcmp(_tbm1->getFileNameExport().Data(), "") == 0);
  CPPUNIT_ASSERT(strcmp(_tbm1->getFunctionNameExport().Data(), "") == 0);

  CPPUNIT_ASSERT_EQUAL(_tbm1->getNIteration(), 1);
  CPPUNIT_ASSERT_EQUAL(_tbm1->getNInitialization(), 1);

  CPPUNIT_ASSERT_DOUBLES_EQUAL(_tbm1->getLearn(), 0.80, DBL_PRECISION_COMPARE);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(_tbm1->getTolerance(), 1.e-04, DBL_PRECISION_COMPARE);
  CPPUNIT_ASSERT_EQUAL(_tbm1->getSeed(), 0);

  // Setter and Getter on empty object
  sDefaultStr = "testModeler";
  _tbm1->setName(sDefaultStr);
  CPPUNIT_ASSERT(strcmp(_tbm1->GetName(), sDefaultStr.Data()) == 0);

  sDefaultStr = "x1:x3:x2:x4";
  _tbm1->setInput(sDefaultStr);
  CPPUNIT_ASSERT(strcmp(_tbm1->getInput(), sDefaultStr.Data()) == 0);

  sDefaultStr = "hx1:hx3";
  _tbm1->setHidden(sDefaultStr);
  CPPUNIT_ASSERT(strcmp(_tbm1->getHidden(), sDefaultStr.Data()) == 0);

  sDefaultStr = "out1:out2";
  _tbm1->setOutput(sDefaultStr);
  CPPUNIT_ASSERT(strcmp(_tbm1->getOutput(), sDefaultStr.Data()) == 0);

  _tbm1->setLanguage(URANIE::XMLProblem::TBaseModel::kC);
  CPPUNIT_ASSERT_EQUAL(_tbm1->getLanguage(), URANIE::XMLProblem::TBaseModel::kC);

  _tbm1->setLanguage(URANIE::XMLProblem::TBaseModel::kCXX);
  CPPUNIT_ASSERT_EQUAL(_tbm1->getLanguage(), URANIE::XMLProblem::TBaseModel::kCXX);

  _tbm1->setLanguage(URANIE::XMLProblem::TBaseModel::kPMML);
  CPPUNIT_ASSERT_EQUAL(_tbm1->getLanguage(), URANIE::XMLProblem::TBaseModel::kPMML);

  _tbm1->setLanguage(URANIE::XMLProblem::TBaseModel::kPYTHON);
  CPPUNIT_ASSERT_EQUAL(_tbm1->getLanguage(), URANIE::XMLProblem::TBaseModel::kPYTHON);

  _tbm1->setLanguage(URANIE::XMLProblem::TBaseModel::kFORTRAN);
  CPPUNIT_ASSERT_EQUAL(_tbm1->getLanguage(), URANIE::XMLProblem::TBaseModel::kFORTRAN);

  _tbm1->setNormalization(URANIE::XMLProblem::TBaseModel::kNO);
  CPPUNIT_ASSERT_EQUAL(_tbm1->getNormalization(), URANIE::XMLProblem::TBaseModel::kNO);

  _tbm1->setNormalization(URANIE::XMLProblem::TBaseModel::kMinusOneOne);
  CPPUNIT_ASSERT_EQUAL(_tbm1->getNormalization(), URANIE::XMLProblem::TBaseModel::kMinusOneOne);

  _tbm1->setNormalization(URANIE::XMLProblem::TBaseModel::kZeroOne);
  CPPUNIT_ASSERT_EQUAL(_tbm1->getNormalization(), URANIE::XMLProblem::TBaseModel::kZeroOne);

  sDefaultStr = "_model_.C";
  _tbm1->setFileNameExport(sDefaultStr);
  CPPUNIT_ASSERT(strcmp(_tbm1->getFileNameExport(), sDefaultStr.Data()) == 0);

  sDefaultStr = "myFunction";
  _tbm1->setFunctionNameExport(sDefaultStr);
  CPPUNIT_ASSERT(strcmp(_tbm1->getFunctionNameExport(), sDefaultStr.Data()) == 0);

  nCrt = 10;
  _tbm1->setNIteration(nCrt);
  CPPUNIT_ASSERT_EQUAL(_tbm1->getNIteration(), nCrt);
  nCrt = -10;
  CPPUNIT_ASSERT_THROW( _tbm1->setNIteration(nCrt),
                        URANIE::Exceptions::UErrorExceptions);
  nCrt = 10;
  _tbm1->setNInitialization(nCrt);
  CPPUNIT_ASSERT_EQUAL(_tbm1->getNInitialization(), nCrt);
  nCrt = -10;
  CPPUNIT_ASSERT_THROW( _tbm1->setNInitialization(nCrt),
                        URANIE::Exceptions::UErrorExceptions);

  _tbm1->setLearn(.455);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(_tbm1->getLearn(), 0.455, DBL_PRECISION_COMPARE);
  CPPUNIT_ASSERT_THROW( _tbm1->setLearn(-.015),
                        URANIE::Exceptions::UErrorExceptions);
  CPPUNIT_ASSERT_THROW( _tbm1->setLearn(1.015),
                        URANIE::Exceptions::UErrorExceptions);

  _tbm1->setTolerance(5.0e-6);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(_tbm1->getTolerance(), 5.0e-6, DBL_PRECISION_COMPARE);
  CPPUNIT_ASSERT_THROW( _tbm1->setTolerance(-.015),
                        URANIE::Exceptions::UErrorExceptions);

  _tbm1->setSeed(123654789);
  CPPUNIT_ASSERT_EQUAL(_tbm1->getSeed(), 123654789);
  CPPUNIT_ASSERT_THROW( _tbm1->setSeed(-123654789),
                        URANIE::Exceptions::UErrorExceptions);

  // Constructor with Input and Output
  _tbm2->printLog();
  CPPUNIT_ASSERT(strcmp(_tbm2->GetName(), "DefaultUranieModel") == 0);
  sDefaultStr = "x1:x2:x3";
  CPPUNIT_ASSERT(strcmp(_tbm2->getInput().Data(), sDefaultStr.Data()) == 0);
  sDefaultStr = "out1";
  CPPUNIT_ASSERT(strcmp(_tbm2->getOutput().Data(), sDefaultStr.Data()) == 0);
  CPPUNIT_ASSERT(strcmp(_tbm2->getHidden().Data(), "") == 0);
  CPPUNIT_ASSERT_EQUAL(_tbm2->getLanguage(), URANIE::XMLProblem::TBaseModel::kEMPTY);
  CPPUNIT_ASSERT_EQUAL(_tbm2->getNormalization(), URANIE::XMLProblem::TBaseModel::kCR);
  CPPUNIT_ASSERT(strcmp(_tbm2->getFileNameExport().Data(), "") == 0);
  CPPUNIT_ASSERT(strcmp(_tbm2->getFunctionNameExport().Data(), "") == 0);

  CPPUNIT_ASSERT_EQUAL(_tbm2->getNIteration(), 1);
  CPPUNIT_ASSERT_EQUAL(_tbm2->getNInitialization(), 1);

  CPPUNIT_ASSERT_DOUBLES_EQUAL(_tbm2->getLearn(), 0.80, DBL_PRECISION_COMPARE);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(_tbm2->getTolerance(), 1.e-04, DBL_PRECISION_COMPARE);
  CPPUNIT_ASSERT_EQUAL(_tbm2->getSeed(), 0);

  // Constructor with Name, Input, Hidden and Output
  sDefaultStr = "myModel";
  CPPUNIT_ASSERT(strcmp(_tbm3->GetName(), sDefaultStr.Data()) == 0);
  sDefaultStr = "x1:x2:x3";
  CPPUNIT_ASSERT(strcmp(_tbm3->getInput().Data(), sDefaultStr.Data()) == 0);
  sDefaultStr = "out1";
  CPPUNIT_ASSERT(strcmp(_tbm3->getOutput().Data(), sDefaultStr.Data()) == 0);
  sDefaultStr = "hidden";
  CPPUNIT_ASSERT(strcmp(_tbm3->getHidden().Data(), sDefaultStr.Data()) == 0);
  CPPUNIT_ASSERT_EQUAL(_tbm3->getLanguage(), URANIE::XMLProblem::TBaseModel::kEMPTY);
  CPPUNIT_ASSERT_EQUAL(_tbm3->getNormalization(), URANIE::XMLProblem::TBaseModel::kCR);
  sDefaultStr = _tbm3->GetName();
  CPPUNIT_ASSERT(strcmp(_tbm3->getFileNameExport().Data(), sDefaultStr.Data()) == 0);
  CPPUNIT_ASSERT(strcmp(_tbm3->getFunctionNameExport().Data(), sDefaultStr.Data()) == 0);

  CPPUNIT_ASSERT_EQUAL(_tbm3->getNIteration(), 1);
  CPPUNIT_ASSERT_EQUAL(_tbm3->getNInitialization(), 1);

  CPPUNIT_ASSERT_DOUBLES_EQUAL(_tbm3->getLearn(), 0.80, DBL_PRECISION_COMPARE);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(_tbm3->getTolerance(), 1.e-04, DBL_PRECISION_COMPARE);
  CPPUNIT_ASSERT_EQUAL(_tbm3->getSeed(), 0);
}

void TBaseModelTest::testInitExport()
{
  TString sDefaultStr = _tbm1->GetName();

  CPPUNIT_ASSERT(strcmp(_tbm1->getFileNameExport().Data(), "") == 0);
  CPPUNIT_ASSERT(strcmp(_tbm1->getFunctionNameExport().Data(), "") == 0);

  _tbm1->initExport();
  CPPUNIT_ASSERT(strcmp(_tbm1->getFileNameExport().Data(), sDefaultStr.Data()) == 0);
  CPPUNIT_ASSERT(strcmp(_tbm1->getFunctionNameExport().Data(), sDefaultStr.Data()) == 0);

  _tbm1->setLanguage(URANIE::XMLProblem::TBaseModel::kC);
  _tbm1->initExport();
  CPPUNIT_ASSERT(strcmp(_tbm1->getFileNameExport().Data(), Form("%s.C", sDefaultStr.Data())) == 0);
  CPPUNIT_ASSERT(strcmp(_tbm1->getFunctionNameExport().Data(), sDefaultStr.Data()) == 0);

  _tbm1->setLanguage(URANIE::XMLProblem::TBaseModel::kCXX);
  _tbm1->initExport();
  CPPUNIT_ASSERT(strcmp(_tbm1->getFileNameExport().Data(), Form("%s.C.cxx", sDefaultStr.Data())) == 0);
  CPPUNIT_ASSERT(strcmp(_tbm1->getFunctionNameExport().Data(), sDefaultStr.Data()) == 0);

  _tbm1->setLanguage(URANIE::XMLProblem::TBaseModel::kFORTRAN);
  _tbm1->initExport();
  CPPUNIT_ASSERT(strcmp(_tbm1->getFileNameExport().Data(), Form("%s.C.cxx.f", sDefaultStr.Data())) == 0);
  CPPUNIT_ASSERT(strcmp(_tbm1->getFunctionNameExport().Data(), sDefaultStr.Data()) == 0);

  _tbm1->setLanguage(URANIE::XMLProblem::TBaseModel::kPYTHON);
  _tbm1->initExport();
  CPPUNIT_ASSERT(strcmp(_tbm1->getFileNameExport().Data(), Form("%s.C.cxx.f.py", sDefaultStr.Data())) == 0);
  CPPUNIT_ASSERT(strcmp(_tbm1->getFunctionNameExport().Data(), sDefaultStr.Data()) == 0);

  _tbm1->setLanguage(URANIE::XMLProblem::TBaseModel::kPMML);
  _tbm1->initExport();
  CPPUNIT_ASSERT(strcmp(_tbm1->getFileNameExport().Data(), Form("%s.C.cxx.f.py.pmml", sDefaultStr.Data())) == 0);
  CPPUNIT_ASSERT(strcmp(_tbm1->getFunctionNameExport().Data(), sDefaultStr.Data()) == 0);
}

////////////////////////////////
///                          ///
///       Exceptions         ///
///                          ///
////////////////////////////////

// End Of File
